//  DXACLRule
//
//  The DXACLRule object contains the rules for a single ACL Entry.
//
//  Message Identifier:      DXR
//  Message Range:           7301 - 7399
//  Trace area assignment:   TRACE_ACLRULESET (22)
//  Version:                 3.11.4   Build: 102
//
//  Version History:
//
//  3.5.0		15/08/2009		Initial version
//  3.9.1		03/09/2010		XML specification changed
//  3.9.2		22/09/2010		Various fixes to ACL processing
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.3		12/11/2011		Fixes to Explorer and ACL Processor
//  3.11.4		27/11/2011		Fixed an issue with wildcard matching
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 2011
//

#include "DXACLRule.h"

//  Constructor
DXACLRule::DXACLRule(ExecEnvironment *xeParent, int iThreadID)
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Set the DX Runtime
	xeLocal = xeParent;
	ParentThread = iThreadID;

	//  Clear the Rule Chain
	arNext = NULL;

	//  Set the defaults
	szEntryName[0] = '\0';
	EntryType = 0;
	AccessLevel = ACL_LEVEL_NOACCESS;
	AccessFlagsOn = 0;
	AccessFlagsOff = 0;
	cRoleSetBuffer[0] = '\0';

	//  Trace construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7301T, this);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, ParentThread);
}

//  Destructor
DXACLRule::~DXACLRule(void)
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7302T, this);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, ParentThread);
}

//  checkSettings
//
//  This function will check the passed settings against those specified in the rule
//
//  Parameters
//
//  1  -  char *	-  Pointer to the ACL entry name
//  2  -  WORD		-  Access Level of this ACL entry
//  3  -  ACL_PRIVILEGES  -  Roles assigned to this ACL entry
//  4  -  WORD		-  ACL Flags for this ACL Entry
//  5  -  int		-  ID of the invoking thread
//
//  Returns
//
//  int	-  0 if the entry conforms to the rule
//  	-  PS_ACTION_MOD_LEVEL	- if the access level must be modified
//      -  PS_ACTION_MOD_PRIVS  - if the roles must be modified
//      -  PS_ACTION_MOD_FLAGS  - if the flag settings must be modified
//      -  PS_ACTION_MOD_TYPE   - if the type must be modified
//      The 4 PS_ACTION_MOD_XXX settings may be Or'd together to produce a composite return value
//
//  Notes:
//
int DXACLRule::checkSettings(WORD wAccessLevel, ACL_PRIVILEGES *apEntry, WORD wAccessFlags, int iThreadID)
{
	int				iModAction = 0;								//  Cumulative Modification Action
	int				iIndex;										//  Generic index
	WORD			wCalcFlags;									//  Calculted flags
	WORD			wCalcType;									//  Calculated type
	ACL_PRIVILEGES	apCalculated;								//  Calculated privileges
	char			szPrivString1[MAX_PSTRING];					//  String area for dumping privs
	char			szPrivString2[MAX_PSTRING];					//  String area for dumping privs
	char			szMsg[MAX_MSG + 1];							//  Generic Message Buffer

	//  Check that the type matches
	wCalcType = wAccessFlags & TYPE_FLAG_BITS;
	if (EntryType != wCalcType)
	{
		iModAction = iModAction | PS_ACTION_MOD_TYPE;
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7312T, wCalcType, EntryType);
		xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
	}

	//  First check that the access level is conforming
	if (wAccessLevel != AccessLevel) 
	{
		iModAction = iModAction | PS_ACTION_MOD_LEVEL;
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7307T, wAccessLevel, AccessLevel);
		xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
	}

	//  Now check the flags - there should be no flags ON that should be set OFF
	//  and no flags OFF that should be set ON.

	if ((wAccessFlags & AccessFlagsOff) != 0) iModAction = iModAction | PS_ACTION_MOD_FLAGS;
	else
	{
		wCalcFlags = wAccessFlags | AccessFlagsOn;
		if (wCalcFlags != wAccessFlags) iModAction = iModAction | PS_ACTION_MOD_FLAGS;
	}

	if (iModAction & PS_ACTION_MOD_FLAGS)
	{
		AccessFlagsUpdate = wAccessFlags | AccessFlagsOn;
		AccessFlagsUpdate = AccessFlagsUpdate & (~AccessFlagsOff);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7308T, wAccessFlags, AccessFlagsUpdate, AccessFlagsOn, AccessFlagsOff);
		xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
	}
	else AccessFlagsUpdate = wAccessFlags;

	//  Now check the Roles there should be no role ON that should be OFF and vice versa
	for (iIndex = 0; iIndex < sizeof(ACL_PRIVILEGES); iIndex++)
	{
		//  And the privilege chunks together and accumulate the total
		apCalculated.BitMask[iIndex] = apEntry->BitMask[iIndex] | apSetOn.BitMask[iIndex];
		apCalculated.BitMask[iIndex] = apCalculated.BitMask[iIndex] & (~apSetOff.BitMask[iIndex]);
	}

	//  Dump the Current and new calculated privilege array
	strcpy_s(szPrivString1, MAX_PSTRING, "0x'");
	strcpy_s(szPrivString2, MAX_PSTRING, "0x'");
	for (iIndex = 0; iIndex < 10; iIndex++)
	{
		sprintf_s(szPrivString1, MAX_PSTRING, "%s%02X", szPrivString1, apEntry->BitMask[iIndex]);
		sprintf_s(szPrivString2, MAX_PSTRING, "%s%02X", szPrivString2, apCalculated.BitMask[iIndex]);
	}
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7311T, szPrivString1, szPrivString2);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);

	if (memcmp(&apCalculated, apEntry, sizeof(ACL_PRIVILEGES)) != 0) iModAction = iModAction | PS_ACTION_MOD_PRIVS;

	//  If the privileges need to be updated then compute the new privilege array
	if (iModAction & PS_ACTION_MOD_PRIVS)
	{
		memcpy(&apUpdate, &apCalculated, sizeof(ACL_PRIVILEGES));
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7309T, cRoleSetBuffer);
		xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);

		//  Dump the Current and new calculated privilege array
		strcpy_s(szPrivString1, MAX_PSTRING, "0x'");
		strcpy_s(szPrivString2, MAX_PSTRING, "0x'");
		for (iIndex = 0; iIndex < 10; iIndex++)
		{
			sprintf_s(szPrivString1, MAX_PSTRING, "%s%02X", szPrivString1, apEntry->BitMask[iIndex]);
			sprintf_s(szPrivString2, MAX_PSTRING, "%s%02X", szPrivString2, apUpdate.BitMask[iIndex]);
		}
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7310T, szPrivString1, szPrivString2);
		xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
	}
	else memcpy(&apUpdate, apEntry, sizeof(ACL_PRIVILEGES));

	return iModAction;
}

//  matchEntry
//
//  This function will check to determine if the ACL entry matches this rule
//
//  Parameters
//
//  1  -  char *	-  Pointer to the ACL entry name
//  2  -  WORD		-  ACL Flags for this ACL Entry
//  3  -  int		-  ID of the invoking thread
//
//  Returns
//
//  BOOL	-  TRUE if the entry matches the rule
//  		-  FALSE if the entry does not match
//
//  Notes:
//
BOOL DXACLRule::matchEntry(char *szEntry, WORD wFlags, int iThreadID)
{
	STATUS			stAPIRC;											//  Return code from Notes API calls
	WORD			wTypeBits;											//  Type bits extracted from the flags
	WORD			wRetLen;											//  Returned length
	char			szAbbrevEntry[MAXUSERNAME + 1];						//  Abbreviated User Name
	char			szAbbrevRule[MAXUSERNAME + 1];						//  Abbreviated Rule
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer

	wTypeBits = wFlags & TYPE_FLAG_BITS;

	//  Abbreviate the Entry name and rule name
	stAPIRC = DNAbbreviate(0, NULL, szEntry, szAbbrevEntry, MAXUSERNAME, &wRetLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7303E, szEntry);
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}
	stAPIRC = DNAbbreviate(0, NULL, szEntryName, szAbbrevRule, MAXUSERNAME, &wRetLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7303E, szEntryName);
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	//  Attempt to match the entry to the rule
	if (NameIsPattern)
	{
		if (EntryType != 0)													//  Unspecified rule type always match
		{
			if (EntryType != wTypeBits)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7306T, szAbbrevEntry, wTypeBits, "entry", szAbbrevRule, EntryType);
				xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
				xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);
				return FALSE;						//  Not matched by type
			}
		}

		//  Match the entry name to the supplied pattern
		if (xeLocal->MatchesPatterni(szAbbrevEntry, szAbbrevRule))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7304T, szAbbrevEntry, wTypeBits, "pattern", szAbbrevRule, EntryType);
			xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
			xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);
			return TRUE;
		}
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7305T, szAbbrevEntry, wTypeBits, "pattern", szAbbrevRule, EntryType);
		xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
		xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);
		return FALSE;
	}
	else
	{
		if (strcmpi(szAbbrevRule, szAbbrevEntry) == 0)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7304T, szAbbrevEntry, wTypeBits, "entry", szAbbrevRule, EntryType);
			xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
			xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);
			return TRUE;
		}
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7305T, szAbbrevEntry, wTypeBits, "entry", szAbbrevRule, EntryType);
		xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
		xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);
	}
	return FALSE;
}

