//  DXACLRuleSet
//
//  The DXACLRuleSet object describes a set of rules that can be applied to the ACL of a database in
//  order to achieve the desired content of the ACL. The object contains the entry points that are
//  used to coerce the ACL to the desired state.
//
//  Message Identifier:      DXR
//  Message Range:           7201 - 7299
//  Trace area assignment:   TRACE_ACLRULESET (22)  - for execution
//							 TRACE_ACLRSPARSER (21) - for validation
//  Version:                 3.14.0   Build: 108
//
//  Version History:
//
//  3.5.0		15/08/2009		Initial version
//  3.9.1		03/09/2010		XML specification changed
//  3.9.2		22/09/2010		Various fixes to ACL processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.7		17/05/2011		Fixed an issue with ACL processing
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.3		12/11/2011		Fixes to Explorer and ACL Processor
//	3.14.0		21/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 2015
//

#include "DXACLRuleSet.h"

//  Constructor
DXACLRuleSet::DXACLRuleSet(ExecEnvironment *xeParent, int iThreadID)
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Set the DX Runtime
	xeLocal = xeParent;
	ParentThread = iThreadID;

	//  Clear the rule chains
	FirstAdminServerRule = NULL;
	FirstCompulsoryRule = NULL;
	FirstForbiddenRule = NULL;
	FirstPermittedRule = NULL;
	FirstActionRule = NULL;

	//  Trace construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7201T, this);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, ParentThread);
}

//  Destructor
DXACLRuleSet::~DXACLRuleSet(void)
{
	DXACLRule	*arCurrent;								//  Current position in rule chain
	DXACLRule	*arDelete;								//  Rule to delete
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7202T, this);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, ParentThread);

	//  Destruction of the Rule Set triggers destruction of all of the rules within the set
	arCurrent = FirstAdminServerRule;
	while (arCurrent != NULL)
	{
		arDelete = arCurrent;
		arCurrent = arCurrent->arNext;
		delete arDelete;
	}
	arCurrent = FirstForbiddenRule;
	while (arCurrent != NULL)
	{
		arDelete = arCurrent;
		arCurrent = arCurrent->arNext;
		delete arDelete;
	}
	arCurrent = FirstPermittedRule;
	while (arCurrent != NULL)
	{
		arDelete = arCurrent;
		arCurrent = arCurrent->arNext;
		delete arDelete;
	}
	arCurrent = FirstCompulsoryRule;
	while (arCurrent != NULL)
	{
		arDelete = arCurrent;
		arCurrent = arCurrent->arNext;
		delete arDelete;
	}
}

//  validate
//
//  This function will check the logical consitency of a Rule set and indicate the state
//
//  Parameters
//
//  Returns
//
//  BOOL	-  TRUE if the Rule Set is valid and consistent
//			-  FALSE if the Rule Set is invalid or incosnsistent
//
//  Notes:
//
BOOL DXACLRuleSet::validate(int iThreadID)
{
	//  Validate the Admin Server Entry
	xeLocal->LogTrace(TRACE_ACLRSPARSER, MSG_DXR7212T, iThreadID);
	if (!validateAdminServer(iThreadID)) return FALSE;

	//  Validate the compulsory entries
	xeLocal->LogTrace(TRACE_ACLRSPARSER, MSG_DXR7213T, iThreadID);
	if (!validateCompulsoryEntries(iThreadID)) return FALSE;

	//  Validate the forbidden entries
	xeLocal->LogTrace(TRACE_ACLRSPARSER, MSG_DXR7214T, iThreadID);
	if (!validateForbiddenEntries(iThreadID)) return FALSE;

	//  Validate the permitted entries
	xeLocal->LogTrace(TRACE_ACLRSPARSER, MSG_DXR7215T, iThreadID);
	if (!validatePermittedEntries(iThreadID)) return FALSE;
	return TRUE;
}

//  coerce
//
//  This function will coerce the passed ACL to conform to the rules defined in the Rule Set
//
//  Parameters
//
//  1  -  HANDLE	- Handle to the ACL that is to be coerced
//
//  Returns
//
//  BOOL	-  TRUE if the ACL has been updated and should be saved
//			-  FALSE if the ACL has not been updated (i.e. the ACL was already conforming)
//
//  Notes:
//
BOOL DXACLRuleSet::coerce(DHANDLE hACL, int iThreadID)
{
	STATUS				stAPIRC;					//  Return code from Notes API calls
	DXACLEnumInfo		*eiLocal;					//  Enumerator message
	char				szMsg[MAX_MSG + 1];			//  Generic message buffer

	//  Sanity checks - ACL Handle cannot be NULL
	if (hACL == NULLHANDLE)
	{
		xeLocal->LogMessage(MSG_DXR7245E, iThreadID);
		return FALSE;
	}

	//  Trace starting to coerce
	xeLocal->LogTrace(TRACE_ACLRULESET, MSG_DXR7220T, iThreadID);

	//  Clear the processing state in all rule chains
	clearRuleChainState(FirstAdminServerRule);
	clearRuleChainState(FirstCompulsoryRule);
	clearRuleChainState(FirstForbiddenRule);
	clearRuleChainState(FirstPermittedRule);

	//  Bind the ACL to the RuleSet - this action will translate the role strings used in each rule
	//  into a set of Privelege bits that encode these roles for the current ACL.
	if (!bindACL(hACL, iThreadID))
	{
		xeLocal->LogMessage(MSG_DXR7231E, iThreadID);
		return FALSE;
	}

	//  Trace starting to Enumerate
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7221T, "Forbidden");
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);

	//  Setup to enumerate the ACL - processing the forbidden entries
	FirstActionRule = NULL;
	ProcPhase = CPHASE_FORBIDDEN;
	UpdateCount = 0;
	eiLocal = new DXACLEnumInfo(this, iThreadID);
	stAPIRC = ACLEnumEntries(hACL, WrapperACLEnum, eiLocal);
	if (ERR(stAPIRC) != NOERROR)
	{
		delete eiLocal;
		eiLocal = NULL;
		xeLocal->LogMessage(MSG_DXR7222E, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}
	delete eiLocal;
	eiLocal = NULL;

	//  Now remove the forbidden entries - these will be on the action rule chain
	purgeForbiddenEntries(hACL, iThreadID);

	//  Trace starting to Enumerate
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7221T, "Permitted");
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);

	//  Setup to enumerate the ACL - processing the permitted entries
	FirstActionRule = NULL;
	ProcPhase = CPHASE_ALLOWED;
	eiLocal = new DXACLEnumInfo(this, iThreadID);
	stAPIRC = ACLEnumEntries(hACL, WrapperACLEnum, eiLocal);
	if (ERR(stAPIRC) != NOERROR)
	{
		delete eiLocal;
		eiLocal = NULL;
		xeLocal->LogMessage(MSG_DXR7222E);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}
	delete eiLocal;
	eiLocal = NULL;

	//  Update the ACL if updates were applied
	updatePermittedEntries(hACL, iThreadID);

	//  Now remove the not permitted entries - these will be on the action rule chain
	purgeForbiddenEntries(hACL, iThreadID);

	//  Now add any missing compulsory entries
	//  Trace starting to Enumerate
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7221T, "Compulsory");
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
	addMissingEntries(hACL, iThreadID);

	//  If there was an Admin Server rule specified then make a final pass over the ACL to determine
	//  what should be done.
	if (FirstAdminServerRule != NULL)
	{
		FirstActionRule = NULL;
		ProcPhase = CPHASE_ADMIN;

		//  Trace starting to Enumerate
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7221T, "Admin Server");
		xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
	
		eiLocal = new DXACLEnumInfo(this, iThreadID);
		stAPIRC = ACLEnumEntries(hACL, WrapperACLEnum, eiLocal);
		if (ERR(stAPIRC) != NOERROR)
		{
			delete eiLocal;
			eiLocal = NULL;
			xeLocal->LogMessage(MSG_DXR7222E, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}
		delete eiLocal;
		eiLocal = NULL;
		
		//  Determine what the resulting action is and process it
		if (FirstActionRule == NULL)
		{
			//  We (possibly) need to add the Admin Server - it does not already exist and the rule is not a pattern
			if (!(FirstAdminServerRule->ProcState & PS_ENTRY_EXISTS))
			{
				if (!FirstAdminServerRule->NameIsPattern)
				{
					addAdminServer(hACL, iThreadID);
				}
			}
		}
		else
		{
			//  Process the action rules that have been generated
			//  Update the ACL if updates were applied
			updatePermittedEntries(hACL, iThreadID);

			//  Now remove the not permitted entries - these will be on the action rule chain
			purgeForbiddenEntries(hACL, iThreadID);
		}
	}

	//  If the ACL was updated then signal this to the caller
	if (UpdateCount > 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7240I, UpdateCount);
		xeLocal->LogVerbose(szMsg, iThreadID);
		return TRUE;
	}
	return FALSE;
}

//  WrapperACLEnum
//
//  This function provides linkage to the actionACLEnum in the current object instance
//
//  Parameters
//
//  1  -  void *	-  Pointer to the DXACLEnumInfo object
//  2  -  char *	-  Pointer to the ACL entry name
//  3  -  WORD		-  Access Level of this ACL entry
//  4  -  ACL_PRIVILEGES  -  Roles assigned to this ACL entry
//  5  -  WORD		-  ACL Flags for this ACL Entry
//
//  Returns
//
//  Notes:
//
void LNCALLBACK DXACLRuleSet::WrapperACLEnum(void *pObj, char *szName, WORD wAccessLevel, ACL_PRIVILEGES *apEntry, WORD wAccessFlags)
{
	DXACLEnumInfo			*eiThis;					//  Enumerator Info
	DXACLRuleSet			*rsThis;					//  Rule Set instance pointer

	eiThis = (DXACLEnumInfo *) pObj;					//  Cast the pointer
	rsThis = eiThis->rsLocal;							//  Address the Rule Set
	rsThis->vectorACLEnum(szName, wAccessLevel, apEntry, wAccessFlags, eiThis->iThread);   //  vector to the real action routine
	return;
}

//  vectorACLEnum
//
//  This function is called to vector to the correct antry action routine for the processing phase
//
//  Parameters
//
//  1  -  char *	-  Pointer to the ACL entry name
//  2  -  WORD		-  Access Level of this ACL entry
//  3  -  ACL_PRIVILEGES  -  Roles assigned to this ACL entry
//  4  -  WORD		-  ACL Flags for this ACL Entry
//  5  -  int		-  ID of the invoking thread
//
//  Returns
//
//  Notes:
//
void DXACLRuleSet::vectorACLEnum(char *szName, WORD wAccessLevel, ACL_PRIVILEGES *apEntry, WORD wAccessFlags, int iThreadID)
{
	char				szMsg[MAX_MSG];							//  Generic message buffer

	//  Trace start processing this entry
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7223T, szName, ProcPhase);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);

	//  If the current entry is the Admin Server - then process it as such
	switch(ProcPhase)
	{
	case CPHASE_FORBIDDEN:
		actionACLEnumForbidden(szName, wAccessLevel, apEntry, wAccessFlags, iThreadID);
		break;
	case CPHASE_ALLOWED:
		actionACLEnumAllowed(szName, wAccessLevel, apEntry, wAccessFlags, iThreadID);
		break;
	case CPHASE_ADMIN:
		actionACLEnumAdmin(szName, wAccessLevel, apEntry, wAccessFlags, iThreadID);
		break;
	}

	return;
}

//  actionACLEnumForbidden
//
//  This function is called by the enumerator for each entry currently in the ACL to process Forbidden rules.
//
//  Parameters
//
//  1  -  char *	-  Pointer to the ACL entry name
//  2  -  WORD		-  Access Level of this ACL entry
//  3  -  ACL_PRIVILEGES  -  Roles assigned to this ACL entry
//  4  -  WORD		-  ACL Flags for this ACL Entry
//  5  -  int		-  ID of the invoking thread
//
//  Returns
//
//  Notes:
//
void DXACLRuleSet::actionACLEnumForbidden(char *szName, WORD wAccessLevel, ACL_PRIVILEGES *apEntry, WORD wAccessFlags, int iThreadID)
{
	DXACLRule			*arCurrent;									//  Current Forbidden ACL Rule
	DXACLRule			*arEOC;										//  Rule at the end of chain
	BOOL				bEntryForbidden;							//  Entry is forbidden
	int					iRuleNo;									//  Rule number in evaluation
	char				szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Check each forbidden rule in turn - if the entry matches then create an action rule for it
	if (FirstForbiddenRule == NULL) return;

	arCurrent = FirstForbiddenRule;
	iRuleNo = 1;
	while (arCurrent != NULL)
	{
		//  Check the current rule
		bEntryForbidden = FALSE;
		if (arCurrent->matchEntry(szName, wAccessFlags, iThreadID)) bEntryForbidden = TRUE;

		//  If the entry is forbidden then create a new dummy action rule
		if (bEntryForbidden)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7224T, szName, iRuleNo);
			xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
			arCurrent = new DXACLRule(xeLocal, iThreadID);
			strcpy_s(arCurrent->szEntryName, MAXUSERNAME, szName);
			arCurrent->arNext = NULL;
			arCurrent->ProcState = PS_ACTION_DELETE;
			
			//  Find the end of the chain to add the new entry
			arEOC = FirstActionRule;
			if (arEOC == NULL)
			{
				FirstActionRule = arCurrent;
				return;
			}
			while (arEOC->arNext != NULL)
			{
				arEOC = arEOC->arNext;
			}
			arEOC->arNext = arCurrent;
			return;
		}
		else
		{
			arCurrent = arCurrent->arNext;
		}
		iRuleNo++;
	}
	return;
}

//  actionACLEnumAdmin
//
//  This function is called by the enumerator for each entry currently in the ACL to process Admin Server rules.
//
//  Parameters
//
//  1  -  char *	-  Pointer to the ACL entry name
//  2  -  WORD		-  Access Level of this ACL entry
//  3  -  ACL_PRIVILEGES  -  Roles assigned to this ACL entry
//  4  -  WORD		-  ACL Flags for this ACL Entry
//  5  -  int		-  ID of the invoking thread
//
//  Returns
//
//  Notes:
//
void DXACLRuleSet::actionACLEnumAdmin(char *szName, WORD wAccessLevel, ACL_PRIVILEGES *apEntry, WORD wAccessFlags, int iThreadID)
{
	DXACLRule			*arCurrent;									//  Current Forbidden ACL Rule
	DXACLRule			*arEOC;										//  Rule at the end of chain

	//  Entry is an Admin Server
	if (wAccessFlags & ACL_FLAG_ADMIN_SERVER)
	{

		//  If the rule specifies no Admin Server then create a delete action for it
		if (FirstAdminServerRule->szEntryName[0] == '\0')
		{
			arCurrent = new DXACLRule(xeLocal, iThreadID);
			strcpy_s(arCurrent->szEntryName, MAXUSERNAME, szName);
			arCurrent->arNext = NULL;
			arCurrent->ProcState = PS_ACTION_DELETE;
			//  Find the end of the chain to add the new entry
			arEOC = FirstActionRule;
			if (arEOC == NULL)
			{
				FirstActionRule = arCurrent;
				return;
			}
			while (arEOC->arNext != NULL)
			{
				arEOC = arEOC->arNext;
			}
			arEOC->arNext = arCurrent;
			return;		
		}

		//  If the name matches then flag the entry as existing
		if (FirstAdminServerRule->matchEntry(szName, wAccessFlags, iThreadID))
		{
			FirstAdminServerRule->ProcState = FirstAdminServerRule->ProcState | PS_ENTRY_EXISTS;
			return;
		}
		else
		{
			//  Providing that the rule is not for a pattern then generate a rename action
			if (!FirstAdminServerRule->NameIsPattern)
			{
				arCurrent = new DXACLRule(xeLocal, iThreadID);
				strcpy_s(arCurrent->szEntryName, MAXUSERNAME, szName);
				strcpy_s(arCurrent->szNewName, MAXUSERNAME, FirstAdminServerRule->szEntryName);
				arCurrent->arNext = NULL;
				arCurrent->ProcState = PS_ACTION_RENAME;
				//  Find the end of the chain to add the new entry
				arEOC = FirstActionRule;
				if (arEOC == NULL)
				{
					FirstActionRule = arCurrent;
					return;
				}
				while (arEOC->arNext != NULL)
				{
					arEOC = arEOC->arNext;
				}
				arEOC->arNext = arCurrent;
				return;		
			}
		}
	}
	return;
}

//  actionACLEnumAllowed
//
//  This function is called by the enumerator for each entry currently in the ACL to process Permitted rules.
//
//  Parameters
//
//  1  -  char *	-  Pointer to the ACL entry name
//  2  -  WORD		-  Access Level of this ACL entry
//  3  -  ACL_PRIVILEGES  -  Roles assigned to this ACL entry
//  4  -  WORD		-  ACL Flags for this ACL Entry
//  5  -  int		-  ID of the invoking thread
//
//  Returns
//
//  Notes:
//
void DXACLRuleSet::actionACLEnumAllowed(char *szName, WORD wAccessLevel, ACL_PRIVILEGES *apEntry, WORD wAccessFlags, int iThreadID)
{
	BOOL				bCheckCompulsory;							//  Check compulsory rules
	BOOL				bProcCompulsoryOnly;						//  Processing only compulsory rules
	BOOL				bEntryMatched;								//  Match result
	DXACLRule			*arCurrent;									//  Current Forbidden ACL Rule
	DXACLRule			*arEOC;										//  Rule at the end of chain
	DXACLRule			*arNew;										//  New ACL (Action) Rule
	int					iRuleNo;									//  Rule number in evaluation
	int					iModAction;									//  Modification option flag
	char				szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Check the boundary condition where there are no permitted rules
	if ((FirstCompulsoryRule == NULL) && (FirstPermittedRule == NULL)) return;

	//  Setup to match the entry
	bCheckCompulsory = TRUE;										//  Check compulsory rules first
	bProcCompulsoryOnly = FALSE;									//  Not only compulsory rules
	bEntryMatched = FALSE;											//  Entry not matched
	iRuleNo = 1;
	arCurrent = FirstCompulsoryRule;								//  Start with compulsory rules
	//  Check for boundary condition where there are no compulsory rules
	if (arCurrent == NULL)
	{
		bCheckCompulsory = FALSE;										//  Check permitted rules
		arCurrent = FirstPermittedRule;									//  Switch to permitted
		if (arCurrent == NULL) bProcCompulsoryOnly = TRUE;				//  Only compulsory rules
	}

	//  Process each rule in turn
	while ((arCurrent != NULL) && (!bEntryMatched))
	{
		if (arCurrent->matchEntry(szName, wAccessFlags, iThreadID)) bEntryMatched = TRUE;

		if (!bEntryMatched)
		{
			//  Move to the next entry
			arCurrent = arCurrent->arNext;
			if ((arCurrent == NULL) && (bCheckCompulsory))
			{
				bCheckCompulsory = FALSE;										//  Check permitted rules
				arCurrent = FirstPermittedRule;									//  Switch to permitted
				if (arCurrent == NULL) bProcCompulsoryOnly = TRUE;				//  Only compulsory rules
			}
			iRuleNo++;
		}
	}

	//  Action according to the match result
	if (bEntryMatched)
	{
		//  Mark the rule as matched
		arCurrent->ProcState = arCurrent->ProcState | PS_ENTRY_EXISTS;

		//  Check that the entry conforms to the settings provided in the current rule
		iModAction = arCurrent->checkSettings(wAccessLevel, apEntry, wAccessFlags, iThreadID);
		if (iModAction == 0)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7229I, szName);
			xeLocal->LogVerbose(szMsg, iThreadID);
		}
		else
		{
			//  Create an action entry to update the settings
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7230T, szName, iModAction, iRuleNo);
			xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
			arNew = new DXACLRule(xeLocal, iThreadID);
			strcpy_s(arNew->szEntryName, MAXUSERNAME, szName);
			arNew->arNext = NULL;
			arNew->ProcState = iModAction;
			arNew->EntryType = arCurrent->EntryType;
			arNew->AccessFlagsOff = arCurrent->AccessFlagsOff;
			arNew->AccessFlagsOn = arCurrent->AccessFlagsOn;
			arNew->AccessFlagsUpdate = arCurrent->AccessFlagsUpdate;
			arNew->AccessLevel = arCurrent->AccessLevel;
			memcpy(&arNew->apSetOff, &arCurrent->apSetOff, sizeof(ACL_PRIVILEGES));
			memcpy(&arNew->apSetOn, &arCurrent->apSetOn, sizeof(ACL_PRIVILEGES));	
			memcpy(&arNew->apUpdate, &arCurrent->apUpdate, sizeof(ACL_PRIVILEGES));	

			//  Find the end of the chain to add the new entry
			arEOC = FirstActionRule;
			if (arEOC == NULL)
			{
				FirstActionRule = arNew;
			}
			else
			{
				while (arEOC->arNext != NULL)
				{
					arEOC = arEOC->arNext;
				}
				arEOC->arNext = arNew;
			}
		}
	}
	else
	{
		//  Purging is only in effect if there were some permitted rules
		if (!bProcCompulsoryOnly)
		{
			//  Entry is not permitted - create an action entry to remove it
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7228T, szName);
			xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
			arCurrent = new DXACLRule(xeLocal, iThreadID);
			strcpy_s(arCurrent->szEntryName, MAXUSERNAME, szName);
			arCurrent->arNext = NULL;
			arCurrent->ProcState = PS_ACTION_DELETE;

			//  Find the end of the chain to add the new entry
			arEOC = FirstActionRule;
			if (arEOC == NULL)
			{
				FirstActionRule = arCurrent;
			}
			else
			{
				while (arEOC->arNext != NULL)
				{
					arEOC = arEOC->arNext;
				}
				arEOC->arNext = arCurrent;
			}
		}
	}
	return;
}

//  validateAdminServer
//
//  This function will check the logical consitency of any Admin Server entry in the current Rule Set
//
//  Parameters
//
//  1  -  int				-  ID of the invoking thread
//
//  Returns
//
//  BOOL	-  TRUE if the Admin Server entry is valid and consistent
//			-  FALSE if the Rule Set is invalid or incosnsistent
//
//  Notes:
//
BOOL DXACLRuleSet::validateAdminServer(int iThreadID)
{
	int				iEntryCount;							//  Count of matching entries
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  If there is no Admin Server rule defined then return TRUE
	if (FirstAdminServerRule == NULL) return TRUE;

	// check that there are no conflicting options in the Rule
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7216T, FirstAdminServerRule->szEntryName);
	xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);
	if (FirstAdminServerRule->szEntryName[0] != '\0')
	{
		if ((FirstAdminServerRule->AccessFlagsOn & FirstAdminServerRule->AccessFlagsOff) != 0)
		{
			xeLocal->LogMessage(MSG_DXR7203E, iThreadID);
			return FALSE;
		}

		//  The entry must not confilict with either a forbidden nor a compulsory entry
		iEntryCount = countEntries(FirstForbiddenRule, FirstAdminServerRule->szEntryName, ACL_FLAG_SERVER, iThreadID);
		if (iEntryCount > 0)
		{
			xeLocal->LogMessage(MSG_DXR7243E, iThreadID);
			return FALSE;
		}
		iEntryCount = countEntries(FirstCompulsoryRule, FirstAdminServerRule->szEntryName, ACL_FLAG_SERVER, iThreadID);
		if (iEntryCount > 0)
		{
			xeLocal->LogMessage(MSG_DXR7244E, iThreadID);
			return FALSE;
		}
	}

	return TRUE;
}

//  validateCompulsoryEntries
//
//  This function will check the logical consitency of the Compulsory entries in the current Rule Set
//
//  Parameters
//
//  1  -  int				-  ID of the invoking thread
//
//  Returns
//
//  BOOL	-  TRUE if the compulsory entries are valid and consistent
//			-  FALSE if the compulsory entries are invalid or incosnsistent
//
//  Notes:
//
BOOL DXACLRuleSet::validateCompulsoryEntries(int iThreadID)
{
	DXACLRule			*arCurrent;						//  The current acl rule entry
	int					iEntryCount;					//  Count of entries
	char				szMsg[MAX_MSG + 1];				//  Generic message buffer

	arCurrent = FirstCompulsoryRule;
	//  Process each rule in turn
	while (arCurrent != NULL)
	{
		//  Trace the entry being validated
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7217T, arCurrent->szEntryName);
		xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);

		//  check that there are no conflicting options in the Rule
		if ((arCurrent->AccessFlagsOn & arCurrent->AccessFlagsOff) != 0)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7204E, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		//  check that the entries are unique
		iEntryCount = countEntries(FirstCompulsoryRule, arCurrent->szEntryName, arCurrent->EntryType, iThreadID);
		if (iEntryCount > 1)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7205E, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		//  Point to the next entry in the chain
		arCurrent = arCurrent->arNext;
	}

	return TRUE;
}

//  validateForbiddenEntries
//
//  This function will check the logical consitency of the Forbidden entries in the current Rule Set
//
//  Parameters
//
//	1  -  int				-  ID of the invoking thread
//
//  Returns
//
//  BOOL	-  TRUE if the forbidden entries are valid and consistent
//			-  FALSE if the forbidden entries are invalid or incosnsistent
//
//  Notes:
//
BOOL DXACLRuleSet::validateForbiddenEntries(int iThreadID)
{
	DXACLRule			*arCurrent;						//  The current acl rule entry
	int					iEntryCount;					//  Count of entries
	char				szMsg[MAX_MSG];					//  Generic message buffer

	arCurrent = FirstForbiddenRule;
	//  Process each rule in turn
	while (arCurrent != NULL)
	{
		//  Trace the entry being validated
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7218T, arCurrent->szEntryName);
		xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);

		// check that there are no conflicting options in the Rule
		if ((arCurrent->AccessFlagsOn & arCurrent->AccessFlagsOff) != 0)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7206E, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		//  check that the entries are unique
		iEntryCount = countEntries(FirstForbiddenRule, arCurrent->szEntryName, arCurrent->EntryType, iThreadID);
		if (iEntryCount > 1)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7207E, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		//  check that the entry is not also in the compulsory rule section
		iEntryCount = countEntries(FirstCompulsoryRule, arCurrent->szEntryName, arCurrent->EntryType, iThreadID);
		if (iEntryCount > 0)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7208E, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		//  check that the entry is not also in the permited rule section - warn if it is
		iEntryCount = countEntries(FirstPermittedRule, arCurrent->szEntryName, arCurrent->EntryType, iThreadID);
		if (iEntryCount > 0)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7209W, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
		}

		//  Point to the next entry in the chain
		arCurrent = arCurrent->arNext;
	}

	return TRUE;
}

//  validatePermittedEntries
//
//  This function will check the logical consitency of the Permitted entries in the current Rule Set
//
//  Parameters
//
//	1  -  int				-  ID of the invoking thread
//
//  Returns
//
//  BOOL	-  TRUE if the permitted entries are valid and consistent
//			-  FALSE if the permitted entries are invalid or incosnsistent
//
//  Notes:
//
BOOL DXACLRuleSet::validatePermittedEntries(int iThreadID)
{
	DXACLRule			*arCurrent;						//  The current acl rule entry
	int					iEntryCount;					//  Count of entries
	char				szMsg[MAX_MSG];					//  Generic message buffer

	arCurrent = FirstPermittedRule;
	//  Process each rule in turn
	while (arCurrent != NULL)
	{
		//  Trace the entry being validated
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7219T, arCurrent->szEntryName);
		xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);

		// check that there are no conflicting options in the Rule
		if ((arCurrent->AccessFlagsOn & arCurrent->AccessFlagsOff) != 0)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7210E, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		//  check that the entries are unique
		iEntryCount = countEntries(FirstCompulsoryRule, arCurrent->szEntryName, arCurrent->EntryType, iThreadID);
		if (iEntryCount > 1)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7211E, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		//  Point to the next entry in the chain
		arCurrent = arCurrent->arNext;
	}

	return TRUE;
}

//  PurgeForbiddenEntries
//
//  This function will purge any forbidden entries that are recorded on the action chain
//
//  Parameters
//
//  1  -  DHANDLE			-  Handle to the current ACL
//	2  -  int				-  ID of the invoking thread
//
//  Returns
//
//  Notes:
//
void DXACLRuleSet::purgeForbiddenEntries(DHANDLE hACL, int iThreadID)
{
	STATUS				stAPIRC;						//  Return code from Notes API calls
	DXACLRule			*arDelete;						//  The current action rule to be deleted
	WORD				wRetLen;							//  Returned length
	char				szCanonicalName[MAXUSERNAME +1];	//  Canonical use name
	char				szMsg[MAX_MSG + 1];				//  Generic message buffer

	while (FirstActionRule != NULL)
	{
		//  Delete the entry from the ACL
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7225I, FirstActionRule->szEntryName);
		xeLocal->LogMessage(szMsg, iThreadID);

		//  Canonicalize the entry name
		stAPIRC = DNCanonicalize(0, NULL, FirstActionRule->szEntryName, szCanonicalName, MAXUSERNAME, &wRetLen);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7237E, FirstActionRule->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7226E, FirstActionRule->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		else
		{
			stAPIRC = ACLDeleteEntry(hACL, szCanonicalName);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7226E, FirstActionRule->szEntryName);
				xeLocal->LogMessage(szMsg, iThreadID);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg, iThreadID);
			}
			else UpdateCount++;
		}

		//  Unchain and delete the action object
		arDelete = FirstActionRule;
		FirstActionRule = FirstActionRule->arNext;
		delete arDelete;
	}
	return;
}

//  countEntries
//
//  This function will count the number of matching entries in a rule chain, optionally allowing wildcard matching
//
//  Parameters
//
//  1  -  DXACLRule *	- Pointer to the start of the rule chain that is to be checked
//  2  -  char *		- Pointer to the entry name to be matched
//  3  -  WORD			- The encoded type of the entry that should be matched
//
//  Returns
//
//  int	-  count of entries in the rule chain that have matched the passed entry
//
//  Notes:
//
int DXACLRuleSet::countEntries(DXACLRule *arStart, char *szEntry, WORD wType, int iThreadID)
{
	int				iCount = 0;						//  Count of entries
	DXACLRule		*arCurrent;						//  Current Rule entry

	arCurrent = arStart;							//  Set the start point on the rule chain

	//  Process each entry in the chain in turn
	while (arCurrent != NULL)
	{
		//  Perform a match of the entry against the rule
		if (arCurrent->matchEntry(szEntry, wType, iThreadID)) iCount++;

		//  Point to the next entry in the chain
		arCurrent = arCurrent->arNext;
	}

	return iCount;
}

//  clearRuleChainState
//
//  This function will clear the processing state in all rules in a chain
//
//  Parameters
//
//  1  -  DXACLRule *	- Pointer to the start of the rule chain that is to be cleared
//
//  Returns
//
//  Notes:
//

void DXACLRuleSet::clearRuleChainState(DXACLRule *arStart)
{
	DXACLRule		*arCurrent;						//  Current Rule entry

	arCurrent = arStart;							//  Set the start point on the rule chain

	//  Process each entry in the chain in turn
	while (arCurrent != NULL)
	{
		arCurrent->ProcState = 0;
		arCurrent = arCurrent->arNext;
	}
	return;
}

//  bindACL
//
//  This function will translate all of the string encoded Role information in each rule into
//  a set of Privilege bits that encode those roles in the current ACL.
//
//  Parameters
//
//  1  -  DHANDLE			-	Handle to the ACL that is to be bound
//  2  -  int				-	ID of the invoking thread
//
//  Returns
//
//  BOOL	-  TRUE if the ACL has been bound
//			-  FALSE if the ACL could not be bound
//
//  Notes:
//
BOOL DXACLRuleSet::bindACL(DHANDLE hACL, int iThreadID)
{
	char			szAllRoles[(ACL_PRIVSTRINGMAX * ACL_PRIVCOUNT) + 1];		//  Buffer for all roles names

	//  Get the list and mask for all roles in the current ACL
	if (!xeLocal->GetACLRoles(hACL, &apAllOn, szAllRoles, 0))
	{
		xeLocal->LogMessage(MSG_DXR7232E, iThreadID);
		return FALSE;
	}

	//  Process each rule chain in turn
	if (!bindRuleChain(FirstAdminServerRule, szAllRoles, "AdminServer", iThreadID)) return FALSE;
	if (!bindRuleChain(FirstCompulsoryRule, szAllRoles, "Compulsory", iThreadID)) return FALSE;
	if (!bindRuleChain(FirstForbiddenRule, szAllRoles, "Forbidden", iThreadID)) return FALSE;
	if (!bindRuleChain(FirstPermittedRule, szAllRoles, "Permitted", iThreadID)) return FALSE;

	return TRUE;
}

//  bindRuleChain
//
//  This function will translate all of the string encoded Role information in each rule on a chain into
//  a set of Privilege bits that encode those roles in the current ACL.
//
//  Parameters
//
//  1  -  DXACLRule *		-  Pointer to the first rule on a chain
//  2  -  char *			-  Pointer to the buffer containing each of the Role strings in order
//  3  -  char *			-  Name of the chain that is being bound
//  4  -  int				-  ID of the invoking thread
//
//  Returns
//
//  BOOL	-  TRUE if the Roles has been bound
//			-  FALSE if the Roles could not be bound
//
//  Notes:
//
BOOL DXACLRuleSet::bindRuleChain(DXACLRule *arStart, char *szRoles, char *szChain, int iThreadID)
{
	DXACLRule		*arCurrent;								//  Pointer to the current rule being processed
	int				iRuleNo;								//  Rule number
	char			szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	arCurrent = arStart;									//  Point to beginning of the chain
	iRuleNo = 1;											//  Rule number
	while (arCurrent != NULL)
	{
		if (!bindRule(arCurrent, szRoles, iThreadID))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7233E, iRuleNo, szChain);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}
		arCurrent = arCurrent->arNext;						//  Point to the next rule
		iRuleNo++;
	}
	//  Trace the completion
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7234T, szChain, iRuleNo - 1);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);
	return TRUE;
}

//  bindRule
//
//  This function will translate all of the string encoded Role information in a rule into
//  a set of Privilege bits that encode those roles in the current ACL.
//
//  Parameters
//
//  1  -  DXACLRule *		-  Pointer to the first rule on a chain
//  2  -  char *			-  Pointer to the buffer containing each of the Role strings in order
//  3  -  int				-  ID of the invoking thread
//
//  Returns
//
//  BOOL	-  TRUE if the Roles has been bound
//			-  FALSE if the Roles could not be bound
//
//  Notes:
//
BOOL DXACLRuleSet::bindRule(DXACLRule *arCurrent, char *szRoles, int iThreadID)
{
	char		*pCurrentRole;						//  Pointer to start of current role
	BOOL		bAssertRole;						//  Switch indicating if Role is Asserted or Negated
	int			iPrivNum;							//  Privilege number of the Role
	int			iIndex;								//  Generic index
	int			iOnCount = 0;						//  Count of Roles Set on
	int			iOffCount = 0;						//  Count of Roles set off
	int			iNotFound = 0;						//  Count of Roles not in the current ACL.
	char		szMsg[MAX_MSG + 1];					//  Generic Message Buffer
	char		szPrivString1[MAX_PSTRING];			//  String area for dumping privs
	char		szPrivString2[MAX_PSTRING];			//  String area for dumping privs

	//  Clear the Set on & Set off privilege arrays
	memset(&arCurrent->apSetOn, 0, sizeof(ACL_PRIVILEGES));
	memset(&arCurrent->apSetOff, 0, sizeof(ACL_PRIVILEGES));

	//  If there are no rules to be bound then just return
	if (arCurrent->cRoleSetBuffer[0] == '\0') return TRUE;

	//  If there are no roles defined in the database ACL then just return
	if (szRoles[0] == '\0') return TRUE;

	//  Trace the Roles we are attemting to bind
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7241T, arCurrent->cRoleSetBuffer);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg);

	//  Loop processing each entry in the Role sring buffer
	pCurrentRole = arCurrent->cRoleSetBuffer;

	while (*pCurrentRole != '\0')
	{
		bAssertRole = TRUE;
		if (*pCurrentRole == '!')
		{
			//  Role is negated for this entry
			bAssertRole = FALSE;
			pCurrentRole++;
		}
		
		//  Get the Privilege Number associated with the current role
		iPrivNum = getPrivNum(pCurrentRole, szRoles);
		if (iPrivNum == -1)
		{
			//  Role was not found - this is an acceptable condition
			iNotFound++;
		}
		else
		{
			if (bAssertRole)
			{
				if (iPrivNum == 0)
				{
					// All roles are asserted 
					memcpy(&arCurrent->apSetOn, &apAllOn, sizeof(ACL_PRIVILEGES));
					iOnCount++;
				}
				else
				{
					ACLSetPriv(arCurrent->apSetOn, iPrivNum);
					iOnCount++;
				}
			}
			else
			{
				if (iPrivNum == 0)
				{
					// All roles are negated
					memcpy(&arCurrent->apSetOff, &apAllOn, sizeof(ACL_PRIVILEGES));
					iOffCount++;
				}
				else
				{
					ACLSetPriv(arCurrent->apSetOff, iPrivNum);
					iOffCount++;
				}
			}
		}

		//  Scan forward to the start of the next specified role or the end of the roles
		while (*pCurrentRole != ']')
		{
			*pCurrentRole++;
			if (*pCurrentRole == '\0') return FALSE;
		}
		pCurrentRole++;								//  Point to the next entry
	}

	//  Trace the activity
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7235T, arCurrent->szEntryName, iOnCount, iOffCount, iNotFound);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);

	//  Dump the set on & set off privilege arrays
	strcpy_s(szPrivString1, MAX_PSTRING, "0x'");
	strcpy_s(szPrivString2, MAX_PSTRING, "0x'");
	for (iIndex = 0; iIndex < 10; iIndex++)
	{
		sprintf_s(szPrivString1, MAX_PSTRING, "%s%02X", szPrivString1, arCurrent->apSetOn.BitMask[iIndex]);
		sprintf_s(szPrivString2, MAX_PSTRING, "%s%02X", szPrivString2, arCurrent->apSetOff.BitMask[iIndex]);
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR7242T, szPrivString1, szPrivString2);
	xeLocal->LogTrace(TRACE_ACLRULESET, szMsg, iThreadID);

	return TRUE;
}

//  getPrivNum
//
//  This function will return the Privilege Number of a Role String "[xxxx]" (no terminator) by comparing
//  it to an ordered buffer containing all role strings from the current ACL.
//
//  Parameters
//
//  1  -  char *			-  Pointer to the start of the current role string
//  2  -  char *			-  Pointer to the buffer containing each of the Role strings in order
//
//  Returns
//
//  int		-  -1    - The specified role does not exist in the all roles string buffer
//          -   0    - The Role was specified as * (i.e. All Roles)
//			-  Other - The privilege number of the specified role
//
//  Notes:
//
int DXACLRuleSet::getPrivNum(char *sCurrentRole, char *szAllRoles)
{
	char		*pRole;									//  Pointer into the required role
	char		*pAllRoles;								//  Pointer into the All Roles string buffer
	int			iPrivNum = ACL_BITPRIVCOUNT;			//  Min number of the Roles entries

	//  For safety - filter out any boudary conditions
	if (sCurrentRole[0] == '\0') return -1;
	if (szAllRoles[0] == '\0') return -1;
	
	//  Check for a request for all roles
	if (sCurrentRole[1] == '*') return 0;

	//  Setup ready to scan the role strings buffer
	pRole = sCurrentRole;
	pAllRoles = szAllRoles;

	//  Scan until we hit the end of the roles string buffer
	while (*pAllRoles != '\0')
	{
		while ((toupper(*pAllRoles) == toupper(*pRole)) && (*pAllRoles != ']') && (*pRole != ']'))
		{
			//  Advance the buffer pointers
			pAllRoles++;
			pRole++;
		}
		if ((*pAllRoles == ']') && (*pRole == ']'))
		{
			//  Perfect Match - return the Privilege Number
			return iPrivNum;
		}
		//  This is a mismatch - reset & start at the next role entry
		iPrivNum++;
		pRole = sCurrentRole;
		while (*pAllRoles != ']')
		{
			pAllRoles++;
		}
		pAllRoles++;
	}
		
	return -1;
}

//  updatePermittedEntries
//
//  This function will process any UPDATE action entries to coerce conformance with the RuleSet.
//
//  Parameters
//
//  1  -  DHANDLE			-	Handle to the ACL that is to be bound
//  2  -  int				-	ID of the invoking thread
//
//  Returns
//
//  Notes:
//
void DXACLRuleSet::updatePermittedEntries(DHANDLE hACL, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DXACLRule			*arPrevious;						//  Previous Rule Entry
	DXACLRule			*arCurrent;							//  Current Rule Entry
	WORD				wUpdFlags;							//  ACL update flags
	WORD				wRetLen;							//  Returned length
	char				szCanonicalName[MAXUSERNAME +1];	//  Canonical user name
	char				szCanonicalNewName[MAXUSERNAME +1];	//  Canonical new user name
	char				szMsg[MAX_MSG + 1];					//  Generic Message Buffer

	arPrevious = NULL;
	arCurrent = FirstActionRule;
	
	//  Process each action entry in turn
	while (arCurrent != NULL)
	{
		//  Skip any delete actions
		if (arCurrent->ProcState == PS_ACTION_DELETE)
		{
			arPrevious = arCurrent;
			arCurrent = arCurrent->arNext;
		}
		else
		{
			if (arCurrent->ProcState == PS_ACTION_RENAME)
			{
				//  Canonicalize the entry name

				stAPIRC = DNCanonicalize(0, NULL, arCurrent->szEntryName, szCanonicalName, MAXUSERNAME, &wRetLen);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR7237E, arCurrent->szEntryName);
					xeLocal->LogMessage(szMsg, iThreadID);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
					xeLocal->LogMessage(szMsg, iThreadID);
					sprintf_s(szMsg, MAX_MSG, MSG_DXR7246E, arCurrent->szEntryName);
					xeLocal->LogMessage(szMsg, iThreadID);
				}
				else
				{
					stAPIRC = DNCanonicalize(0, NULL, arCurrent->szNewName, szCanonicalNewName, MAXUSERNAME, &wRetLen);
					if (ERR(stAPIRC) != NOERROR)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR7237E, arCurrent->szNewName);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
						xeLocal->LogMessage(szMsg, iThreadID);
						sprintf_s(szMsg, MAX_MSG, MSG_DXR7246E, arCurrent->szEntryName);
						xeLocal->LogMessage(szMsg, iThreadID);
					}
					else
					{
						stAPIRC = ACLUpdateEntry(hACL,
							szCanonicalName,
							ACL_UPDATE_NAME,
							szCanonicalNewName,
							arCurrent->AccessLevel,
							&arCurrent->apUpdate,
							arCurrent->AccessFlagsUpdate);
						if (ERR(stAPIRC) != NOERROR)
						{
							sprintf_s(szMsg, MAX_MSG, MSG_DXR7246E, arCurrent->szEntryName);
							xeLocal->LogMessage(szMsg, iThreadID);
							xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
							xeLocal->LogMessage(szMsg, iThreadID);
						}
						else 
						{
							sprintf_s(szMsg, MAX_MSG, MSG_DXR7247I, arCurrent->szEntryName);
							xeLocal->LogMessage(szMsg, iThreadID);
							UpdateCount++;
						}
					}
				}

				//  Unchain and destroy the current action rule
				if (arPrevious == NULL)
				{
					//  We are at the top of the chain
					FirstActionRule = arCurrent->arNext;
					delete arCurrent;
					arCurrent = FirstActionRule;
				}
				else
				{
					arPrevious->arNext = arCurrent->arNext;
					delete arCurrent;
					arCurrent = arPrevious->arNext;
				}
			}
			else
			{
				//  Update the ACL entry
				wUpdFlags = 0;
				if (arCurrent->ProcState & PS_ACTION_MOD_LEVEL) wUpdFlags = wUpdFlags | ACL_UPDATE_LEVEL;
				if (arCurrent->ProcState & PS_ACTION_MOD_FLAGS) wUpdFlags = wUpdFlags | ACL_UPDATE_FLAGS;
				if (arCurrent->ProcState & PS_ACTION_MOD_TYPE) wUpdFlags = wUpdFlags | ACL_UPDATE_FLAGS;
				if (arCurrent->ProcState & PS_ACTION_MOD_PRIVS) wUpdFlags = wUpdFlags | ACL_UPDATE_PRIVILEGES;

				//  Canonicalize the entry name

				stAPIRC = DNCanonicalize(0, NULL, arCurrent->szEntryName, szCanonicalName, MAXUSERNAME, &wRetLen);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR7237E, arCurrent->szEntryName);
					xeLocal->LogMessage(szMsg, iThreadID);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
					xeLocal->LogMessage(szMsg, iThreadID);
					sprintf_s(szMsg, MAX_MSG, MSG_DXR7236E, arCurrent->szEntryName);
					xeLocal->LogMessage(szMsg, iThreadID);
				}
				else
				{
					if (strcmpi(szCanonicalName, DEFAULT_ACL_ENTRYNAME) == 0)
					{
						stAPIRC = ACLUpdateEntry(hACL,
							NULL,
							wUpdFlags,
							arCurrent->szEntryName,
							arCurrent->AccessLevel,
							&arCurrent->apUpdate,
							arCurrent->AccessFlagsUpdate);
					}
					else
					{
						stAPIRC = ACLUpdateEntry(hACL,
							szCanonicalName,
							wUpdFlags,
							arCurrent->szEntryName,
							arCurrent->AccessLevel,
							&arCurrent->apUpdate,
							arCurrent->AccessFlagsUpdate | arCurrent->EntryType);
					}
					if (ERR(stAPIRC) != NOERROR)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR7236E, arCurrent->szEntryName);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
						xeLocal->LogMessage(szMsg, iThreadID);
					}
					else 
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR7247I, arCurrent->szEntryName);
						xeLocal->LogMessage(szMsg, iThreadID);
						UpdateCount++;
					}
				}

				//  Unchain and destroy the current action rule
				if (arPrevious == NULL)
				{
					//  We are at the top of the chain
					FirstActionRule = arCurrent->arNext;
					delete arCurrent;
					arCurrent = FirstActionRule;
				}
				else
				{
					arPrevious->arNext = arCurrent->arNext;
					delete arCurrent;
					arCurrent = arPrevious->arNext;
				}
			}
		}
	}
	return;
}

//  addMissingEntries
//
//  This function will add any compulsory entries that are missing from the ACL
//
//  Parameters
//
//  1  -  DHANDLE			-	Handle to the ACL that is to be processed
//  2  -  int				-	ID of the invoking thread
//
//  Returns
//
//  Notes:
//
void DXACLRuleSet::addMissingEntries(DHANDLE hACL, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DXACLRule			*arCurrent;							//  Current Rule Entry
	WORD				wRetLen;							//  Returned length
	char				szCanonicalName[MAXUSERNAME +1];	//  Canonical use name
	char				szMsg[MAX_MSG + 1];					//  Generic Message Buffer

	arCurrent = FirstCompulsoryRule;
	
	while (arCurrent != NULL)
	{
		//  If the entry was not matched then it must be added to the ACL
		if (!(arCurrent->ProcState & PS_ENTRY_EXISTS))
		{
			//  Canonicalize the entry name
			stAPIRC = DNCanonicalize(0, NULL, arCurrent->szEntryName, szCanonicalName, MAXUSERNAME, &wRetLen);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7237E, arCurrent->szEntryName);
				xeLocal->LogMessage(szMsg, iThreadID);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg, iThreadID);
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7238E, arCurrent->szEntryName);
				xeLocal->LogMessage(szMsg, iThreadID);
			}
			else
			{
				stAPIRC = ACLAddEntry(hACL,
									szCanonicalName,
									arCurrent->AccessLevel,
									&arCurrent->apSetOn,
									arCurrent->AccessFlagsOn | arCurrent->EntryType);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR7238E, arCurrent->szEntryName);
					xeLocal->LogMessage(szMsg, iThreadID);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
					xeLocal->LogMessage(szMsg, iThreadID);
				}
				else 
				{
					UpdateCount++;
					sprintf_s(szMsg, MAX_MSG, MSG_DXR7239I, arCurrent->szEntryName);
					xeLocal->LogMessage(szMsg, iThreadID);
				}
			}
		}
		
		arCurrent = arCurrent->arNext;
	}
	return;
}

//  addAdminServer
//
//  This function will add the Admin Server entry to the ACL from the rule
//
//  Parameters
//
//  1  -  DHANDLE			-	Handle to the ACL that is to be processed
//  2  -  int				-	ID of the invoking thread
//
//  Returns
//
//  Notes:
//
void DXACLRuleSet::addAdminServer(DHANDLE hACL, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DXACLRule			*arCurrent;							//  Current Rule Entry
	WORD				wRetLen;							//  Returned length
	int					iIndex;								//  Generic index
	char				szCanonicalName[MAXUSERNAME +1];	//  Canonical use name
	char				szMsg[MAX_MSG + 1];					//  Generic Message Buffer

	arCurrent = FirstAdminServerRule;
	
	//  Canonicalize the entry name
	stAPIRC = DNCanonicalize(0, NULL, arCurrent->szEntryName, szCanonicalName, MAXUSERNAME, &wRetLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7237E, arCurrent->szEntryName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7238E, arCurrent->szEntryName);
		xeLocal->LogMessage(szMsg, iThreadID);
	}
	else
	{
		//  Make sure that the settings are appropriate for an Admin Server
		arCurrent->AccessLevel = ACL_LEVEL_MANAGER;
		arCurrent->EntryType = ACL_FLAG_SERVER;
		arCurrent->AccessFlagsOff = arCurrent->AccessFlagsOff | DEFAULT_ADMIN_FLAGS;
		arCurrent->AccessFlagsUpdate = arCurrent->AccessFlagsOn & (~arCurrent->AccessFlagsOff);

		//  Calculate the roles as all roles in the database less any specifically excluded
		for (iIndex = 0; iIndex < sizeof(ACL_PRIVILEGES); iIndex++)
		{
			//  And the privilege chunks together and accumulate the total
			arCurrent->apUpdate.BitMask[iIndex] = apAllOn.BitMask[iIndex] & (~arCurrent->apSetOff.BitMask[iIndex]);
		}

		stAPIRC = ACLAddEntry(hACL,
							szCanonicalName,
							ACL_LEVEL_MANAGER,
							&arCurrent->apUpdate,
							arCurrent->AccessFlagsOn | ACL_FLAG_SERVER);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7238E, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		else 
		{
			UpdateCount++;
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7239I, arCurrent->szEntryName);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
	}

	return;
}

//  clone
//
//  This function will create a clone of the complete rule set. This function is needed for multi-threaded
//  processing (coerce) of the same rule set against multiple databases (ACLs) at the same time.
//  As information from the individual ACL is bound to the Rules in the set then there needs to be a unique
//  copy of the rule set for each concurrent request.
//
//  Parameters
//
//
//  Returns
//
//  DXACLRuleSet *		-		pointer to the cloned rule set object
//                      -       NULL if the operation failed
//
//  Notes:
//
void DXACLRuleSet::clone(DXACLRuleSet *arsClone, int iThreadID)
{

	//  Copy the chains of rule objects into the cloned object
	arsClone->FirstAdminServerRule = cloneRuleChain(FirstAdminServerRule, iThreadID);
	arsClone->FirstCompulsoryRule = cloneRuleChain(FirstCompulsoryRule, iThreadID);
	arsClone->FirstForbiddenRule = cloneRuleChain(FirstForbiddenRule, iThreadID);
	arsClone->FirstPermittedRule = cloneRuleChain(FirstPermittedRule, iThreadID);

	return;
}

//  cloneRuleChain
//
//  This function will create a clone of a chain of ACL rules.
//
//  Parameters
//
//  1 -		DXACLRule *
//
//  Returns
//
//  DXACLRule *			-		pointer to the cloned first rule on the chain
//                      -       NULL if the operation failed
//
//  Notes:
//
DXACLRule * DXACLRuleSet::cloneRuleChain(DXACLRule *arStart, int iThreadID)
{
	DXACLRule		*arCurrent;								//  Current Rule on the source chain
	DXACLRule		*arFirstClone = NULL;					//  First Rule on the cloned rule chain
	DXACLRule		*arPreviousClone = NULL;				//  Previous Rule on the cloned rule chain
	DXACLRule		*arCurrentClone = NULL;					//  Current Rule on the cloned rule chain

	//  Handle the valid boundary condition where the top of the chain is NULL
	if (arStart == NULL) return NULL;

	arCurrent = arStart;									//  work from the head of the chain
	while (arCurrent != NULL)
	{
		//  Create a clone of the current rule object
		arCurrentClone = cloneRule(arCurrent, iThreadID);

		//  Perform the necessary chain & position maintenance
		if (arFirstClone == NULL) arFirstClone = arCurrentClone;
		else arPreviousClone->arNext = arCurrentClone;
		arPreviousClone = arCurrentClone;
	
		//  Iterate over the rule chain
		arCurrent = arCurrent->arNext;
	}

	return arFirstClone;
}

//  cloneRule
//
//  This function will create a clone of a single ACL rule.
//
//  Parameters
//
//  1 -		DXACLRule *
//
//  Returns
//
//  DXACLRule *			-		pointer to the cloned rule
//                      -       NULL if the operation failed
//
//  Notes:
//
DXACLRule * DXACLRuleSet::cloneRule(DXACLRule *arCurrent, int iThreadID)
{
	DXACLRule			*arClone;								//  The cloned Rule

	//  Just in case
	if (arCurrent == NULL) return NULL;

	//  Create the cloned rule
	arClone = new DXACLRule(xeLocal, iThreadID);

	//  Copy the non-transient members from the current rule to the cloned rule
	arClone->AccessFlagsOff = arCurrent->AccessFlagsOff;
	arClone->AccessFlagsOn = arCurrent->AccessFlagsOn;
	arClone->AccessLevel = arCurrent->AccessLevel;
	arClone->arNext = NULL;
	strcpy_s(arClone->cRoleSetBuffer, ROLE_BFR_SIZE, arCurrent->cRoleSetBuffer);
	arClone->EntryType = arCurrent->EntryType;
	arClone->NameIsPattern = arCurrent->NameIsPattern;
	strcpy_s(arClone->szEntryName, MAXUSERNAME, arCurrent->szEntryName);

	return arClone;
}

//  The following entries belong to the message objects that are passed into enumerators
//  Constructor
DXACLEnumInfo::DXACLEnumInfo(DXACLRuleSet *rsParent, int iThreadID)
{
	rsLocal = rsParent;
	iThread = iThreadID;
}

//  Destructor
DXACLEnumInfo::~DXACLEnumInfo(void)
{

}

