//  DXACLRuleSetParser
//
//  The DXACLRuleSetParser will parse a supplied XML document from a source that a DXResourceLoader can handle
//  it will parse the document into a complete DXACLRuleSet object with all of it's child objects.
//  A DXACLRuleSet forms a complete set of rules that describe how to change a database ACL to the desired
//  Configuration.
//
//  Message Identifier:      DXR
//  Message Range:           7101 - 7199
//  Trace area assignment:   TRACE_ACLRSPARSER (21)
//  Version:                 3.11.1   Build: 99
//
//  Version History:
//
//  3.5.0		15/08/2009		Initial version
//  3.9.1		03/09/2010		XML specification changed
//  3.9.2		22/09/2010		Various fixes to ACL processing
//  3.10.7		17/05/2011		Fixed an issue with ACL processing
//  3.10.9		04/06/2011		Some tidying up for Linux
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 2011
//

#include "DXACLRuleSetParser.h"

//  Constructor
DXACLRuleSetParser::DXACLRuleSetParser(ExecEnvironment *xeParent, int iThreadID)
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Set the DX Runtime
	xeLocal = xeParent;
	iParentThread = iThreadID;

	//  Mark the Resouce Loader as not Loaded
	rlCurrent = NULL;

	//  Trace construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7101T, this);
	xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iParentThread);
}

//  Destructor
DXACLRuleSetParser::~DXACLRuleSetParser(void)
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7102T, this);
	xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iParentThread);
}

//  parseTheseRules
//
//  Generic entry point taking the name of a resource which is then passed to the current DXResource Loader
//  to be loaded into memory. The in memory XML set is then parsed by the internal micro-parser within the
//  Object.
//
//  Parameters
//
//  1  -  char *	- The resource source specifier URL, filename or Resource ID
//  2  -  int		-  ID of the invoking thread
//
//  Returns
//
//  DXACLRuleSet *  -	Pointer to the loaded DXACLRuleSet object
//						NULL if the rule set could not be formed
//
//  API Use Notes:
//
//
DXACLRuleSet * DXACLRuleSetParser::parseTheseRules(char *szRuleSet, int iThreadID)
{
	DXACLRuleSet	*rsCurrent;					//  Pointer to the ACL Rulset to be built
	MEMHANDLE		hmRuleSet;					//  Handle of the in memory Rule Set XML
	DWORD			dwRSSize;					//  Rules Set Size
	char			*pRS;						//  Pointer to Rules SEt in-memory
	char			szMsg[MAX_MSG + 1];			//  Generic message buffer

	//  If the resource loader has not been loaded then load a default loader
	if (rlCurrent == NULL)
	{
		rlCurrent = new DXResourceLoader(xeLocal, iThreadID);
	}

	//  Load the Rule Set XML into memory
	hmRuleSet = rlCurrent->loadThisResource(szRuleSet, iThreadID);
	if (hmRuleSet == (MEMHANDLE) NULL)
	{
		//  Show the load failed and return a NULL RuleSet
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7103E, szRuleSet);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULL;
	}

	//  Resize the rule set and add terminators
	dwRSSize = OSMemoryGetSize(hmRuleSet);
	hmRuleSet = xeLocal->resizeDominoMemory(hmRuleSet, dwRSSize + 2, iThreadID);
	if (hmRuleSet == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7132E, dwRSSize);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULL;
	}

	pRS = (char *) OSMemoryLock(hmRuleSet);
	pRS[dwRSSize] = '\0';
	dwRSSize++;
	pRS[dwRSSize] = '\0';

	//  Create an empty DXACLRuleSet object to contain the parsed rules
	rsCurrent = new DXACLRuleSet(xeLocal, iThreadID);

	//  Parse the Rule Set XML to the DXACLRuleSet object
	if (!parseTheseRules(hmRuleSet, rsCurrent, iThreadID))
	{
		//  Show that the parsing failed and return a NULL RuleSet
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7105E, szRuleSet);
		xeLocal->LogMessage(szMsg, iThreadID);
		delete rsCurrent;									//  Delete the Rule Set object
		return NULL;
	}
	else
	{
		//  Show that the parsing completed and return the RuleSet
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7130I, szRuleSet);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Return the pointer to the fully populated Rule Set object
	return rsCurrent;
}

//  parseTheseRules
//
//  Generic entry point taking the Note and Item name which is then passed to the current DXResource Loader
//  to be loaded into memory. The in memory XML set is then parsed by the internal micro-parser within the
//  Object.
//
//  Parameters
//
//  1  -  NOTEHANDLE  - Handle of the note that contains the XML Rule Set to be parsed
//  2  -  char *	  - Name of the Item within the note that contains the XML Rule Set to be parsed
//  3  -  int		-  ID of the invoking thread
//
//  Returns
//
//  DXACLRuleSet *  -	Pointer to the loaded DXACLRuleSet object
//						NULL if the rule set could not be formed
//
//  API Use Notes:
//
//
DXACLRuleSet * DXACLRuleSetParser::parseTheseRules(NOTEHANDLE hnSource, char *szItemName, int iThreadID)
{
	DXACLRuleSet	*rsCurrent;					//  Pointer to the ACL Rulset to be built
	NOTEID			nidSource = 0;				//  Noteid of the source note for the Rule Set XML
	MEMHANDLE		hmRuleSet;					//  Handle of the in memory Rule Set XML
	char			szMsg[MAX_MSG + 1];			//  Generic message buffer

	//  If the resource loader has not been created then create a default loader
	if (rlCurrent == NULL)
	{
		rlCurrent = new DXResourceLoader(xeLocal, iThreadID);
	}

	//  Load the Rule Set XML into memory
	hmRuleSet = rlCurrent->loadThisResource(hnSource, szItemName, iThreadID);
	if (hmRuleSet == (MEMHANDLE) NULL)
	{
		//  Show the load failed and return a NULL RuleSet
		NSFNoteGetInfo(hnSource, _NOTE_ID, &nidSource);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7104E, szItemName, nidSource);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULL;
	}

	//  Create an empty DXACLRuleSet object to contain the parsed rules
	rsCurrent = new DXACLRuleSet(xeLocal, iThreadID);

	//  Parse the Rule Set XML to the DXACLRuleSet object
	if (!parseTheseRules(hmRuleSet, rsCurrent, iThreadID))
	{
		//  Show that the parsing failed and return a NULL RuleSet
		NSFNoteGetInfo(hnSource, _NOTE_ID, &nidSource);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7106E, szItemName, nidSource);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULL;
	}
	else
	{
		//  Show that the parsing completed and return the RuleSet
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7131I, szItemName, nidSource);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Return the pointer to the fully populated Rule Set object
	return rsCurrent;
}

//  parseTheseRules
//
//  Generic entry point taking a Memory Handle to the loaded XML for a Rule Set.
//
//  Parameters
//
//  1  -  MEMHANDLE			- Handle of the memory buffer containing the XML Rule Set to be parsed
//  2  -  DXACLRuleSet *	- Pointer to an empty ACL Rule Set object that will contain the rules
//  3  -  int				-  ID of the invoking thread
//
//  Returns
//
//  BOOL  -	TRUE if the ACL Rule Set was valid and has been loaded
//			FALSE if the Rule Set could not be loaded
//
BOOL DXACLRuleSetParser::parseTheseRules(MEMHANDLE hmRuleSet, DXACLRuleSet *arEmpty, int iThreadID)
{
	BOOL				bResult;					//  Result from the Parsing operation
	char				*pXMLBuffer;				//  Pointer to the in-memory XML

	//  Lock the memory associated with the handle
	pXMLBuffer = (char *) OSMemoryLock(hmRuleSet);

	//  Parse the in-memory Rule Set
	bResult = parseTheseRules(pXMLBuffer, arEmpty, iThreadID);

	//  Unlock and free the buffer
	OSMemoryUnlock(hmRuleSet);
	OSMemoryFree(hmRuleSet);

	//  Return the result of the parsing
	return bResult;
}

//  parseTheseRules
//
//  Specific entry point taking a pointer the loaded XML for a Rule Set.
//  This function performs the real parsing of the in-memory XML.
//  The implementation is a micro-parser that will process (specifically) ACRRuleSet XML Documents.
//  The XML implemetation supported is lightweight but conformant.
//
//  Parameters
//
//  1  -  Char *			- Pointer to the memory buffer containing the XML Rule Set to be parsed
//  2  -  DXACLRuleSet *	- Pointer to an empty ACL Rule Set object that will contain the rules
//  3  -  int		-  ID of the invoking thread
//
//  Returns
//
//  BOOL  -	TRUE if the ACL Rule Set was valid and has been loaded
//			FALSE if the Rule Set could not be loaded
//
BOOL DXACLRuleSetParser::parseTheseRules(char *pXML, DXACLRuleSet *arEmpty, int iThreadID)
{
	DXACLRule			*LatestRule;						//  The latest DXACLRule created 
	int					iFSMState;							//  The current Finite State Machine state
	int					iXMLLineType;						//  The type of line that has been parsed
	int					iRuleContext;						//  The rule context for the current rule
	int					iAdminServers = 0;					//  Count of Admin Server entries created
	int					iCompulsoryRules = 0;				//  Count of compulsory ACL rules
	int					iForbiddenRules = 0;				//  Count of forbidden ACL rules
	int					iPermittedRules = 0;				//  Count of permitted ACL rules
	char				*pCurrentXML;						//  Pointer to the current position being parsed
	char				szXMLLine[MAX_MSG + 1];				//  A line of XML
	char				szXMLNode[MAX_NODE + 1];			//  The current XML Node
	char				szMsg[MAX_MSG + 1];					//  Generic Message buffer
	
	//  The micro-parser operates as a Finite State Machine (FSM)
	//  Set up the starting state for the machine.
	pCurrentXML = pXML;
	iFSMState = XMLP_STATE_UNDEFINED;						//  Initial State is UNDEFINED

	//  Trace the start of the parsing engine

	//  Major parsing loop - keep processing the XML document until a valid completion
	//  or a terminal error state.
	while ((iFSMState != XMLP_STATE_FINISHED) && (iFSMState != XMLP_STATE_TERMERROR))
	{
		//  Extract the next line of data from the XML Document
		pCurrentXML = getNextXMLLine(pCurrentXML, szXMLLine, MAX_MSG, &iXMLLineType, iThreadID);
		//  Log the input line
		if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
		{
			//  Show the line being processed
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7108I, szXMLLine);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Trace the current state of the processing engine
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7109T, iXMLLineType, iFSMState);
			xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);
		}
		
		//  Process the current line according to the current engine state
		switch(iFSMState)
		{
		case XMLP_STATE_UNDEFINED:
			//  The only valid entry at this point is an XML Declaration
			if (iXMLLineType == XMLLINE_DECL) iFSMState = XMLP_STATE_XMLDECL;
			else
			{
				//  All of these are invalid at this point in the document, but with different errors
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
				xeLocal->LogMessage(szMsg, iThreadID);
				if (iXMLLineType == XMLLINE_EOF) xeLocal->LogMessage(MSG_DXR7110E);
				else xeLocal->LogMessage(MSG_DXR7111E, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
			}
			break;

			//  The XML declaration has been processed
		case XMLP_STATE_XMLDECL:
			//  Comments or Doctype definition are valid in this context but do not change the state
			if ((iXMLLineType == XMLLINE_DOCTYPE) || (iXMLLineType == XMLLINE_COMMENT)) break;
			//  A start of node is valid if it is an ACLRuleSet
			if (iXMLLineType == XMLLINE_START_NODE)
			{
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ACLRULESET) == 0)
				{
					iFSMState = XMLP_STATE_RSSTARTED;
					break;
				}
				//  Node is not valid at this point in the document
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}
			else
			{
				//  All of these are invalid at this point in the document, but with different errors
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
				xeLocal->LogMessage(szMsg, iThreadID);
				if (iXMLLineType == XMLLINE_EOF) xeLocal->LogMessage(MSG_DXR7118E);
				else xeLocal->LogMessage(MSG_DXR7111E, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  The ACLRuleSet node has been started
		case XMLP_STATE_RSSTARTED:
			//  Comments are valid in this context but do not change the state
			if (iXMLLineType == XMLLINE_COMMENT) break;
			//  Start of Nodes AdminServer, CompulsoryEntries, ForbiddenEntries or PermittedEntries are allowed
			if (iXMLLineType == XMLLINE_START_NODE)
			{
				//  A start of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ADMINSERVER) == 0)
				{
					//  The start of an AdminServer Node has been detected
					//  Check if the Admin Server Entry already exists - if so this is an error
					if (arEmpty->FirstAdminServerRule != NULL)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->LogMessage(MSG_DXR7114E, iThreadID);
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else
					{
						//  Create a new ACL Rule object for the Admin Server
						LatestRule = new DXACLRule(xeLocal, iThreadID);
						arEmpty->FirstAdminServerRule = LatestRule;
						iRuleContext = PCONTEXT_ADMINSERVER;
						iAdminServers++;
						//  Populate the rule from the XML
						if (!populateRuleFromXML(szXMLLine, arEmpty->FirstAdminServerRule, iRuleContext, iThreadID))
						{
							iFSMState = XMLP_STATE_TERMERROR;
						}
						else iFSMState = XMLP_STATE_RULESTARTED;
					}
					break;
				}

				if (strcmpi(szXMLNode, XMLNODE_COMPENTRIES) == 0)
				{
					//  The start of CompulsoryEntries Node has been detected
					//  Check if the set of CompulsoryEntries already exist
					if (arEmpty->FirstCompulsoryRule != NULL)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->LogMessage(MSG_DXR7115E, iThreadID);
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else
					{
						//  Change to the state for processing compulsory entries
						iFSMState = XMLP_STATE_CESTARTED;
					}
					break;
				}

				if (strcmpi(szXMLNode, XMLNODE_FORBIDENTRIES) == 0)
				{
					//  The start of ForbiddenEntries Node has been detected
					//  Check if the set of ForbiddenEntries already exist
					if (arEmpty->FirstForbiddenRule != NULL)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->LogMessage(MSG_DXR7116E, iThreadID);
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else
					{
						//  Change to the state for processing forbidden entries
						iFSMState = XMLP_STATE_FESTARTED;
					}
					break;
				}

				if (strcmpi(szXMLNode, XMLNODE_OKENTRIES) == 0)
				{
					//  The start of PermittedEntries Node has been detected
					//  Check if the set of PermittedEntries already exist
					if (arEmpty->FirstPermittedRule != NULL)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->LogMessage(MSG_DXR7117E, iThreadID);
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else
					{
						//  Change to the state for processing forbidden entries
						iFSMState = XMLP_STATE_PESTARTED;
					}
					break;
				}

				//  The start node is not recognized
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
			}

			//  A Full node definition is also valid at this point
			if (iXMLLineType == XMLLINE_FULL_NODE)
			{
				//  A start of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ADMINSERVER) == 0)
				{
					//  The start of an AdminServer Node has been detected
					//  Check if the Admin Server Entry already exists - if so this is an error
					if (arEmpty->FirstAdminServerRule != NULL)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->LogMessage(MSG_DXR7114E, iThreadID);
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else
					{
						//  Create a new ACL Rule object for the Admin Server
						LatestRule = new DXACLRule(xeLocal, iThreadID);
						arEmpty->FirstAdminServerRule = LatestRule;
						//  Populate the rule from the XML
						if (!populateRuleFromXML(szXMLLine, arEmpty->FirstAdminServerRule, PCONTEXT_ADMINSERVER, iThreadID))
						{
							iFSMState = XMLP_STATE_TERMERROR;
						}
						else iFSMState = XMLP_STATE_RSSTARTED;
					}
					break;
				}

				//  The node is not recognized
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  If we encountered an end of Node then it must be for the Rule Set
			if (iXMLLineType == XMLLINE_END_NODE)
			{
				//  An end of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ACLRULESET) == 0)
				{
					//  This marks the well-formed end of the Rule Set definition
					iFSMState = XMLP_STATE_FINISHED;
					break;
				}

				//  End of the node is nor recorgized
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7119E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  All of others are invalid at this point in the document, but with different errors
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (iXMLLineType == XMLLINE_EOF) xeLocal->LogMessage(MSG_DXR7118E);
			else xeLocal->LogMessage(MSG_DXR7111E, iThreadID);
			iFSMState = XMLP_STATE_TERMERROR;
			break;

			//  State with compulsory entries now being defined
		case XMLP_STATE_CESTARTED:
			//  Comments are valid in this context but do not change the state
			if (iXMLLineType == XMLLINE_COMMENT) break;
			//  Start of an ACLRule Node is the only one allowed
			if (iXMLLineType == XMLLINE_START_NODE)
			{
				//  A start of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ACLRULE) == 0)
				{
					//  Create a new DXACLRule object and add it to the compulsory rules chain
					LatestRule = new DXACLRule(xeLocal, iThreadID);
					LatestRule->arNext = arEmpty->FirstCompulsoryRule;
					arEmpty->FirstCompulsoryRule = LatestRule;
					iRuleContext = PCONTEXT_COMPENTRIES;
					iCompulsoryRules++;
					//  Populate the rule from the XML
					if (!populateRuleFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID))
					{
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else iFSMState = XMLP_STATE_RULESTARTED;
					break;
				}
				//  Starting any other node type is inavide at this point
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			if (iXMLLineType == XMLLINE_FULL_NODE)
			{
				//  A full of Node definition has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ACLRULE) == 0)
				{
					//  Create a new DXACLRule object and add it to the compulsory rules chain
					LatestRule = new DXACLRule(xeLocal, iThreadID);
					LatestRule->arNext = arEmpty->FirstCompulsoryRule;
					arEmpty->FirstCompulsoryRule = LatestRule;
					iRuleContext = PCONTEXT_COMPENTRIES;
					iCompulsoryRules++;
					//  Populate the rule from the XML
					if (!populateRuleFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID))
					{
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else iFSMState = XMLP_STATE_CESTARTED;
					break;
				}
				//  Starting any other node type is inavide at this point
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  If we encountered an end of Node then it must be for the CompulsoryEntries
			if (iXMLLineType == XMLLINE_END_NODE)
			{
				//  An end of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_COMPENTRIES) == 0)
				{
					//  Returns to the Rel Set Defined state
					iFSMState = XMLP_STATE_RSSTARTED;
					break;
				}

				//  End of the node is not recorgized
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7119E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  All of others are invalid at this point in the document, but with different errors
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (iXMLLineType == XMLLINE_EOF) xeLocal->LogMessage(MSG_DXR7118E, iThreadID);
			else xeLocal->LogMessage(MSG_DXR7111E, iThreadID);
			iFSMState = XMLP_STATE_TERMERROR;
			break;

			//  State with forbidden entries now being defined
		case XMLP_STATE_FESTARTED:
			//  Comments are valid in this context but do not change the state
			if (iXMLLineType == XMLLINE_COMMENT) break;
			//  Start of an ACLRule Node is the only one allowed
			if (iXMLLineType == XMLLINE_START_NODE)
			{
				//  A start of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ACLRULE) == 0)
				{
					//  Create a new DXACLRule object and add it to the compulsory rules chain
					LatestRule = new DXACLRule(xeLocal, iThreadID);
					LatestRule->arNext = arEmpty->FirstForbiddenRule;
					arEmpty->FirstForbiddenRule = LatestRule;
					iRuleContext = PCONTEXT_FORBIDENTRIES;
					iForbiddenRules++;
					//  Populate the rule from the XML
					if (!populateRuleFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID))
					{
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else iFSMState = XMLP_STATE_RULESTARTED;
					break;
				}
				//  Starting any other node type is inavide at this point
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			if (iXMLLineType == XMLLINE_FULL_NODE)
			{
				//  A full of Node definition has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ACLRULE) == 0)
				{
					//  Create a new DXACLRule object and add it to the compulsory rules chain
					LatestRule = new DXACLRule(xeLocal, iThreadID);
					LatestRule->arNext = arEmpty->FirstForbiddenRule;
					arEmpty->FirstForbiddenRule = LatestRule;
					iRuleContext = PCONTEXT_FORBIDENTRIES;
					iForbiddenRules++;
					//  Populate the rule from the XML
					if (!populateRuleFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID))
					{
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else iFSMState = XMLP_STATE_FESTARTED;
					break;
				}
				//  Starting any other node type is inavide at this point
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  If we encountered an end of Node then it must be for the ForbiddenEntries
			if (iXMLLineType == XMLLINE_END_NODE)
			{
				//  An end of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_FORBIDENTRIES) == 0)
				{
					//  Returns to the Rel Set Defined state
					iFSMState = XMLP_STATE_RSSTARTED;
					break;
				}

				//  End of the node is nor recorgized
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7119E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  All of others are invalid at this point in the document, but with different errors
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (iXMLLineType == XMLLINE_EOF) xeLocal->LogMessage(MSG_DXR7118E, iThreadID);
			else xeLocal->LogMessage(MSG_DXR7111E, iThreadID);
			iFSMState = XMLP_STATE_TERMERROR;
			break;

			//  State with permitted entries now being defined
		case XMLP_STATE_PESTARTED:
			//  Comments are valid in this context but do not change the state
			if (iXMLLineType == XMLLINE_COMMENT) break;
			//  Start of an ACLRule Node is the only one allowed
			if (iXMLLineType == XMLLINE_START_NODE)
			{
				//  A start of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ACLRULE) == 0)
				{
					//  Create a new DXACLRule object and add it to the compulsory rules chain
					LatestRule = new DXACLRule(xeLocal, iThreadID);
					LatestRule->arNext = arEmpty->FirstPermittedRule;
					arEmpty->FirstPermittedRule = LatestRule;
					iRuleContext = PCONTEXT_OKENTRIES;
					iPermittedRules++;
					//  Populate the rule from the XML
					if (!populateRuleFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID))
					{
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else iFSMState = XMLP_STATE_RULESTARTED;
					break;
				}
				//  Starting any other node type is inavide at this point
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			if (iXMLLineType == XMLLINE_FULL_NODE)
			{
				//  A full of Node definition has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ACLRULE) == 0)
				{
					//  Create a new DXACLRule object and add it to the compulsory rules chain
					LatestRule = new DXACLRule(xeLocal, iThreadID);
					LatestRule->arNext = arEmpty->FirstPermittedRule;
					arEmpty->FirstPermittedRule = LatestRule;
					iRuleContext = PCONTEXT_OKENTRIES;
					iPermittedRules++;
					//  Populate the rule from the XML
					if (!populateRuleFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID))
					{
						iFSMState = XMLP_STATE_TERMERROR;
					}
					else iFSMState = XMLP_STATE_PESTARTED;
					break;
				}
				//  Starting any other node type is inavide at this point
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  If we encountered an end of Node then it must be for the PermittedEntries
			if (iXMLLineType == XMLLINE_END_NODE)
			{
				//  An end of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_OKENTRIES) == 0)
				{
					//  Returns to the Rel Set Defined state
					iFSMState = XMLP_STATE_RSSTARTED;
					break;
				}

				//  End of the node is nor recorgized
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7119E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  All of others are invalid at this point in the document, but with different errors
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (iXMLLineType == XMLLINE_EOF) xeLocal->LogMessage(MSG_DXR7118E, iThreadID);
			else xeLocal->LogMessage(MSG_DXR7111E, iThreadID);
			iFSMState = XMLP_STATE_TERMERROR;
			break;

			//  State where a single ACLRule instance is being defined
		case XMLP_STATE_RULESTARTED:
			//  Comments are valid in this context but do not change the state
			if (iXMLLineType == XMLLINE_COMMENT) break;
			//  Start of an Option or Role Node are the only valid entries at this point
			if (iXMLLineType == XMLLINE_START_NODE)
			{
				//  A start of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_OPTION) == 0)
				{
					//  Apply the Option Setting to the current rule
					if (!applyOptionFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID)) iFSMState = XMLP_STATE_TERMERROR;
					else iFSMState = XMLP_STATE_OPTSTARTED;
					break;
				}

				if (strcmpi(szXMLNode, XMLNODE_ROLE) == 0)
				{
					//  Apply the Role Setting to the current rule
					if (!applyRoleFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID)) iFSMState = XMLP_STATE_TERMERROR;
					else iFSMState = XMLP_STATE_ROLESTARTED;
					break;
				}

				//  Starting any other node type is inavide at this point
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			if (iXMLLineType == XMLLINE_FULL_NODE)
			{
				//  A Full Node definition has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_OPTION) == 0)
				{
					//  Apply the Option Setting to the current rule
					if (!applyOptionFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID)) iFSMState = XMLP_STATE_TERMERROR;
					else iFSMState = XMLP_STATE_RULESTARTED;
					break;
				}

				if (strcmpi(szXMLNode, XMLNODE_ROLE) == 0)
				{
					//  Apply the Role Setting to the current rule
					if (!applyRoleFromXML(szXMLLine, LatestRule, iRuleContext, iThreadID)) iFSMState = XMLP_STATE_TERMERROR;
					else iFSMState = XMLP_STATE_RULESTARTED;
					break;
				}

				//  Starting any other node type is inavide at this point
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  If we encountered an end of Node then it must be for the ACLRule
			if (iXMLLineType == XMLLINE_END_NODE)
			{
				//  An end of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ACLRULE) == 0)
				{
					//  Returns state that was saved as the current Rule context
					switch(iRuleContext)
					{
					case PCONTEXT_ADMINSERVER:
						iFSMState = XMLP_STATE_RSSTARTED;
						break;
					case PCONTEXT_COMPENTRIES:
						iFSMState = XMLP_STATE_CESTARTED;
						break;
					case PCONTEXT_FORBIDENTRIES:
						iFSMState = XMLP_STATE_FESTARTED;
						break;
					case PCONTEXT_OKENTRIES:
						iFSMState = XMLP_STATE_PESTARTED;
						break;
					}
					break;
				}

				//  End of the node is nor recorgized
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7119E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  All of others are invalid at this point in the document, but with different errors
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (iXMLLineType == XMLLINE_EOF) xeLocal->LogMessage(MSG_DXR7118E, iThreadID);
			else xeLocal->LogMessage(MSG_DXR7111E, iThreadID);
			iFSMState = XMLP_STATE_TERMERROR;
			break;

			//  State where a single Option is being applied to a Rule
		case XMLP_STATE_OPTSTARTED:
			//  Comments are valid in this context but do not change the state
			if (iXMLLineType == XMLLINE_COMMENT) break;
			//  There are no valid nodes that can be started here
			if ((iXMLLineType == XMLLINE_START_NODE) || (iXMLLineType == XMLLINE_FULL_NODE))
			{
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  If we encountered an end of Node then it must be for the Option
			if (iXMLLineType == XMLLINE_END_NODE)
			{
				//  An end of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_OPTION) == 0)
				{
					//  Returns to the ACL Rule being defined
					iFSMState = XMLP_STATE_RULESTARTED;
					break;
				}

				//  End of the node is nor recorgized
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7119E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  All of others are invalid at this point in the document, but with different errors
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (iXMLLineType == XMLLINE_EOF) xeLocal->LogMessage(MSG_DXR7118E, iThreadID);
			else xeLocal->LogMessage(MSG_DXR7111E, iThreadID);
			iFSMState = XMLP_STATE_TERMERROR;
			break;

			//  State where a single Role is being applied to a Rule
		case XMLP_STATE_ROLESTARTED:
			//  Comments are valid in this context but do not change the state
			if (iXMLLineType == XMLLINE_COMMENT) break;
			//  There are no valid nodes that can be started here
			if ((iXMLLineType == XMLLINE_START_NODE) || (iXMLLineType == XMLLINE_FULL_NODE))
			{
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7113E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  If we encountered an end of Node then it must be for the Option
			if (iXMLLineType == XMLLINE_END_NODE)
			{
				//  An end of Node has been detected
				//  Extract the name of the current Node
				getXMLNode(szXMLNode, szXMLLine, iThreadID);
				if (strcmpi(szXMLNode, XMLNODE_ROLE) == 0)
				{
					//  Returns to the ACL Rule being defined
					iFSMState = XMLP_STATE_RULESTARTED;
					break;
				}

				//  End of the node is nor recorgized
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7119E, szXMLNode);
				xeLocal->LogMessage(szMsg, iThreadID);
				iFSMState = XMLP_STATE_TERMERROR;
				break;
			}

			//  All of others are invalid at this point in the document, but with different errors
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7112E, szXMLLine);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (iXMLLineType == XMLLINE_EOF) xeLocal->LogMessage(MSG_DXR7118E, iThreadID);
			else xeLocal->LogMessage(MSG_DXR7111E, iThreadID);
			iFSMState = XMLP_STATE_TERMERROR;
			break;
		}
	}

	//  Return the success or failure according to the exit state of the processing engine
	if (iFSMState == XMLP_STATE_FINISHED)
	{
		//  Perform final checks on the numbers
		if ((iCompulsoryRules + iForbiddenRules + iPermittedRules + iAdminServers) == 0)
		{
			xeLocal->LogMessage(MSG_DXR7120E, iThreadID);
			return FALSE;
		}
		//  Trace the number of rules generated
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7121T, iAdminServers, iCompulsoryRules, iForbiddenRules, iPermittedRules);
		xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);

		//  Check the logical consistency of the Rule Set
		if (checkRuleSet(arEmpty)) return TRUE;
		return FALSE;
	}
	return FALSE;
}

//  getNextXMLLine
//
//  This function extracts the next line of XML from the XML document buffer
//
//  Parameters
//
//  1  -  char *	- Pointer to the current position in the memory buffer containing the XML Rule Set to be parsed
//  2  -  int *		- Pointer to an int where the type of line will be returned
//  3  -  int		- ID of the invoking thread
//
//  Returns
//
//  cahr *  -	The updated current pointer in the memory buffer
//
char * DXACLRuleSetParser::getNextXMLLine(char *pXML, char *szXMLLine, int iMaxLine, int *iType, int iThreadID)
{
	char	*pInput;							//  Input Pointer
	char	*pLookAhead;						//  Lookahead Pointer
	char	*pOutput;							//  Output Pointer
	int		iChars;								//  Count of characters passed to output
	int		iThisType;							//  Entity type detected
	BOOL	bFullNode;							//  Flag to indicate that a Full Node has been detected

	//  Initialise the positions and counts
	pInput = pXML;
	pOutput = szXMLLine;
	*pOutput = '\0';
	iChars = 0;
	iThisType = XMLLINE_EOF;					//  Default is End-Of-File (EOF)

	//  Step 1 - eliminate all input until a start of element character is found
	while (*pInput != '<')
	{
		if (*pInput == '\0')
		{
			//  End-Of-File (EOF) detected
			*iType = XMLLINE_EOF;
			return pInput;
		}
		*pInput++;
	}

	*pOutput = *pInput;
	iChars++;
	*pOutput++;
	*pInput++;

	//  Step 2 - Determine the type of element being processed
	if (*pInput == '\0')
	{
		*pOutput = '\0';
		//  Malformed XML detected
			*iType = XMLLINE_MALFORMED;
			return pInput;
	}

	iThisType = XMLLINE_START_NODE;
	if (*pInput == '?') iThisType = XMLLINE_DECL;
	if (*pInput == '!')
	{
		pLookAhead = pInput;
		pLookAhead++;
		if (*pLookAhead == 'D') iThisType = XMLLINE_DOCTYPE;
		if (*pLookAhead == '-') iThisType = XMLLINE_COMMENT;
	}
	if (*pInput == '/') iThisType = XMLLINE_END_NODE;

	//  Step 3 - Scan for the end of the element
	bFullNode = FALSE;

	while (*pInput != '>')
	{
		if (*pInput >= ' ')
		{
			//  Copy the input to output
			*pOutput = *pInput;
			iChars++;
			//  Check for overflow
			if (iChars > iMaxLine)
			{
				*pOutput = '\0';
				*iType = XMLLINE_OVERFLOW;
				return pInput;
			}
			pOutput++;
			if (*pInput == '/') bFullNode = TRUE;
			else bFullNode = FALSE;
		}
		else
		{
			//  Detect malformed XML
			if (*pInput == '\0')
			{
				*pOutput = '\0';
				//  Malformed XML detected
				*iType = XMLLINE_MALFORMED;
				return pInput;
			}
		}
		pInput++;				
	}
	*pOutput = *pInput;
	iChars++;
	//  Check for overflow
	if (iChars > iMaxLine)
	{
		*pOutput = '\0';
		*iType = XMLLINE_OVERFLOW;
		return pInput;
	}
	pOutput++;
	*pOutput = '\0';
	pInput++;

	//  Step 4 - check if the provisional type needs to be updated
	if ((bFullNode) && (iThisType == XMLLINE_START_NODE)) iThisType = XMLLINE_FULL_NODE;

	//  Step 5 - Return the updated state
	*iType = iThisType;
	return pInput;
}

//  getXMLNode
//
//  This function will extract extract an XML Node Name from the line of XML presented
//
//  Parameters
//
//  1  -  char *	- Pointer to a buffer where the Node name will be returned
//  2  -  char *	- Pointer to the line of XML to process to find the Node name
//  3  -  int		- ID of the invoking thread
//
//  Returns
//
//  Notes:
//  1)  The output buffer (param 1) is assumed to be at least MAX_NODE + 1 bytes in length
//
void DXACLRuleSetParser::getXMLNode(char *szNode, char *szXMLLine, int iThreadID)
{
	char	*pInput;							//  Input Pointer
	char	*pOutput;							//  Output Pointer
	int		iChars;								//  Count of characters passed to output

	//  Initialise the positions and counts
	pInput = szXMLLine;
	pOutput = szNode;
	*pOutput = '\0';
	iChars = 0;

	//  Step 1 - locate the start of the Node Name
	while (*pInput != '<')
	{
		//  If End-Of-String is detected return the null node name
		if (*pInput == '\0') return;
		pInput++;
	}

	pInput++;
	//  If this is an End-Of-Element then skip past the end delimiter
	if (*pInput == '/') pInput++;

	//  Step 2 - Skip forward to the start of significance of the name
	while (*pInput <= ' ')
	{
		//  If End-Of-String is detected return the null node name
		if (*pInput == '\0') return;
		pInput++;
	}

	// Step 3 - Copy the node name to output
	while ((*pInput != ' ') && (*pInput != '>'))
	{
		if (*pInput >= ' ')
		{
			//  Copy the input to output
			*pOutput = *pInput;
			iChars++;
			//  Check for overflow
			if (iChars > MAX_NODE)
			{
				*pOutput = '\0';
				return;
			}
			pOutput++;
		}
		else
		{
			if (*pInput == '\0')
			{
				*pOutput = '\0';
				return;
			}
		}
		pInput++;
	}
	*pOutput = '\0';
	return;
}

//  populateRuleFromXML
//
//  This function will populate a DXACLRule object from the passed line of XML
//
//  Parameters
//
//  1  -  char *		- Pointer to a line of XML
//  2  -  DXACLRule *	- Pointer to the ACL Rule to be populated
//  3  -  int			- Context settings (see PCONTEXT_ values)
//  4  -  int			-  ID of the invoking thread
//
//  Returns
//
//  BOOL		-  TRUE if the Rule object could be populated
//				-  FALSE if the Rule object could not be populated
//
BOOL DXACLRuleSetParser::populateRuleFromXML(char *szXML, DXACLRule *arRule, int iContext, int iThreadID)
{
	char		szAttribute[MAX_ATTRIBUTE + 1];				//  XML Attribute Value
	char		szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Get the Name= or Pattern= attribute from the XML Node definition
	arRule->NameIsPattern = FALSE;
	getAttribute(XMLATTR_NAME, arRule->szEntryName, MAXUSERNAME, szXML, iThreadID);
	if (arRule->szEntryName[0] == '\0')
	{
		getAttribute(XMLATTR_PATTERN, arRule->szEntryName, MAXUSERNAME, szXML, iThreadID);
		arRule->NameIsPattern = TRUE;
	}

	//  Validate the Name
	if (arRule->szEntryName[0] == '\0')
	{
		xeLocal->LogMessage(MSG_DXR7122E, iThreadID);
		return FALSE;
	}

	//  If the rule is a compulsory rule then wildcards are not permitted in the name
	if ((iContext == PCONTEXT_COMPENTRIES) && (arRule->NameIsPattern))
	{
		xeLocal->LogMessage(MSG_DXR7123E, iThreadID);
		return FALSE;
	}

	//  If the Context indicates that the rule is for an Admin Server then set the Type attributes
	//  otherwise obtain them from the Node definition
	if (iContext == PCONTEXT_ADMINSERVER)
	{
		arRule->EntryType = ACL_FLAG_SERVER | ACL_FLAG_ADMIN_SERVER;
		arRule->AccessLevel = ACL_LEVEL_MANAGER;

		//  Check for the special values for the Admin Server Name
		if (strcmpi(arRule->szEntryName, ADMIN_SERVER_HOST) == 0)
		{
			//  Use the current server name
			strcpy_s(arRule->szEntryName, MAXUSERNAME, xeLocal->AbbreviatedUser);
		}
		if (strcmpi(arRule->szEntryName, ADMIN_SERVER_NONE) == 0)
		{
			arRule->szEntryName[0] = '\0';
		}
	}
	else
	{
		//  Get the entry type attribute
		getAttribute(XMLATTR_TYPE, szAttribute, MAX_ATTRIBUTE, szXML, iThreadID);
		if (!encodeEntryType(szAttribute, &arRule->EntryType))
		{
			if (iContext == PCONTEXT_FORBIDENTRIES) arRule->EntryType = 0;
			else
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7124E, szAttribute);
				xeLocal->LogMessage(szMsg, iThreadID);
				return FALSE;
			}
		}

		//  Obtain the ACL Access Level from the Node definition
		getAttribute(XMLATTR_LEVEL, szAttribute, MAX_ATTRIBUTE, szXML, iThreadID);
		if (!encodeAccessLevel(szAttribute, &arRule->AccessLevel))
		{
			if (iContext == PCONTEXT_FORBIDENTRIES) arRule->AccessLevel = 0;
			else
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7125E, szAttribute);
				xeLocal->LogMessage(szMsg, iThreadID);
				return FALSE;
			}
		}
	}

	//  Show that the ACL Rule has been validly populated
	return TRUE;
}

//  applyOptionFromXML
//
//  This function will apply an option setting to a DXACLRule object from the passed line of XML.
//  Each option consists of a type and a specification as to whether the option flag should be set on or off
//
//  Parameters
//
//  1  -  char *		- Pointer to a line of XML
//  2  -  DXACLRule *	- Pointer to the ACL Rule to be populated
//  3  -  int			- Context settings (see PCONTEXT_ values)
//  4  -  int			-  ID of the invoking thread
//
//  Returns
//
//  BOOL		-  TRUE if the Rule object could be updated
//				-  FALSE if the Rule object could not be updated
//
BOOL DXACLRuleSetParser::applyOptionFromXML(char *szXML, DXACLRule *arRule, int iContext, int iThreadID)
{
	WORD		wOption = 0;								//  The option to be set on or off
	char		szAttribute[MAX_ATTRIBUTE + 1];				//  XML Attribute Value
	char		szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Get the option type attribute
	getAttribute(XMLATTR_TYPE, szAttribute, MAX_ATTRIBUTE, szXML, iThreadID);
	if (!encodeOptionType(szAttribute, &wOption))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7126E, szAttribute);
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	//  Get the setting value of On or Off
	getAttribute(XMLATTR_SET, szAttribute, MAX_ATTRIBUTE, szXML, iThreadID);
	if (isOptionOn(szAttribute)) arRule->AccessFlagsOn = arRule->AccessFlagsOn | wOption;
	else arRule->AccessFlagsOff = arRule->AccessFlagsOff | wOption;

	return TRUE;
}

//  applyRoleFromXML
//
//  This function will apply a role setting to a DXACLRule object from the passed line of XML
//
//  Parameters
//
//  1  -  char *		- Pointer to a line of XML
//  2  -  DXACLRule *	- Pointer to the ACL Rule to be populated
//  3  -  int			- Context settings (see PCONTEXT_ values)
//  4  -  int			-  ID of the invoking thread
//
//  Returns
//
//  BOOL		-  TRUE if the Rule object could be updated
//				-  FALSE if the Rule object could not be updated
//
BOOL DXACLRuleSetParser::applyRoleFromXML(char *szXML, DXACLRule *arRule, int iContext, int iThreadID)
{
	char		szRole[MAX_ROLE + 1];						//  Role Name
	char		szAttribute[MAX_ATTRIBUTE + 1];				//  XML Attribute Value

	//  Get the role name attribute
	getAttribute(XMLATTR_NAME, szRole, MAX_ROLE, szXML, iThreadID);
	if (szRole[0] == '\0')
	{
		//  Role name is invalid
		xeLocal->LogMessage(MSG_DXR7127E, iThreadID);
		return FALSE;
	}

	//  Determine if the Role should be applied or not
	getAttribute(XMLATTR_ASSIGN, szAttribute, MAX_ATTRIBUTE, szXML, iThreadID);
	if (isOptionOn(szAttribute))
	{
		sprintf_s(arRule->cRoleSetBuffer, ROLE_BFR_SIZE, "%s[%s]", arRule->cRoleSetBuffer, szRole);
	}
	else
	{
		sprintf_s(arRule->cRoleSetBuffer, ROLE_BFR_SIZE, "%s![%s]", arRule->cRoleSetBuffer, szRole);
	}
	return TRUE;
}

//  getAttribute
//
//  This function will extract extract an XML attribute value from the line of XML presented
//
//  Parameters
//
//  1  -  char *	- The name of the attribute whoes value is to be extracted
//  2  -  char *	- Pointer to a buffer where the attribute value will be returned
//  3  -  int		- Max size of the attribute value that can be returned
//  4  -  char *	- Pointer to the line of XML to process to find the attribute value
//  5  -  int		- ID of the invoking thread
//
//  Returns
//
//  Notes:
//  1)  An empty string is returned if the value is not found
//
void DXACLRuleSetParser::getAttribute(char *szAttr, char *szValue, int iMaxValue, char *szLine, int iThreadID)
{
	int					iChars;										//  # characters copied/moved
	char				cEDelimiter;									//  End marker for the value string
	char				*pInput;									//  Pointer into iput string
	char				*pOutput;									//  Pointer into output string
	char				szUCXML[MAX_MSG + 1];						//  A line of XML foldeed to uppercase
	char				szUCAttr[MAX_ATTRIBUTE + 1];				//  An attribute name folded to uppercase
	char				szMsg[MAX_MSG + 1];							//  Generic message buffer
	
	//  Get the XML line data and fold it to uppercase
	pInput = szLine;
	pOutput = szUCXML;

	while (*pInput != '\0')
	{
		*pOutput = toupper(*pInput);
		pInput++;
		pOutput++;
	}

	*pOutput = '\0';

	//  Get the Attribute name and fold it to uppercase
	pInput = szAttr;
	pOutput = szUCAttr;

	while (*pInput != '\0')
	{
		*pOutput = toupper(*pInput);
		pInput++;
		pOutput++;
	}

	*pOutput = '\0';

	//  Trace the start of search
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7128T, szUCAttr, szUCXML);
	xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);

	//  Search for the attribute specified within the string
	pInput = strstr(szUCXML, szUCAttr);
	if (pInput == NULL)
	{
		szValue[0] = '\0';						//  Set the return value as empty
		return;
	}

	//  Compute the offset in the XML Line
	iChars = (int) (pInput - szUCXML);

	//  Point to the same position in the supplied XML Line
	pInput = szLine + iChars;

	//  Locate the first "=" in the line following the attribute
	while ((*pInput != '=') && (*pInput != '\0'))
	{
		pInput++;
	}

	//  No = was found .....  return an empty value string
	if (*pInput == '\0')
	{
		szValue[0] = '\0';						//  Set the return value as empty
		return;
	}

	//  Skip over the = sign
	pInput++;

	//  Skip over any white space to find the start of significance of the value
	while ((*pInput <= ' ') && (*pInput != 0))
	{
		pInput++;
	}

	//  No start of the value was found .....  return an empty value string
	if (*pInput == '\0')
	{
		szValue[0] = '\0';						//  Set the return value as empty
		return;
	}

	//  Determine if the value is set as a quoted string and set the appropriate delimiter
	cEDelimiter = ' ';								//  Default to white space
	if ((*pInput == '"') || (*pInput == '\''))
	{
		cEDelimiter = *pInput;
		pInput++;
	}

	//  Copy the value from the line buffer until we hit the delimiter or we run out of string
	pOutput = szValue;
	while ((*pInput != cEDelimiter) && (*pInput != '\0') && (*pInput != '>'))
	{
		*pOutput = *pInput;
		pOutput++;
		pInput++;
	}
	
	//  If we terminated on an brace and the previous character was a slash then lose the slash
	if (*pInput == '>')
	{
		pOutput--;
		if (*pOutput != '/') pOutput++;
	}

	//  Terminate the output value
	*pOutput = '\0';

	//  Trace the return from search
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7129T, szUCAttr, szValue);
	xeLocal->LogTrace(TRACE_ACLRSPARSER, szMsg, iThreadID);

	return;
}

//  hasWildcards
//
//  This function will indicate if the passed string contains any wilcard (* or ?) characters
//
//  Parameters
//
//  1  -  char *	- String to be tested for wildcard characters
//
//  Returns
//
//  BOOL	-  TRUE if the string contains any wildcard characters
//			-  FALSE if there are no wildcard characters in the string
//
//  Notes:
//
BOOL DXACLRuleSetParser::hasWildcards(char *szString)
{
	char			*pInput;							//  Pointer to wildcard

	pInput = strchr(szString, '?');
	if (pInput != NULL) return TRUE;
	pInput = strchr(szString, '*');
	if (pInput != NULL) return TRUE;
	return FALSE;
}

//  encodeEntryType
//
//  This function will compute the flag settings to encode the ACL entry type 
//
//  Parameters
//
//  1  -  char *	- String containng the ACL Entry Type desired
//  2  -  WORD *	- Pointer to the word that will contain the encoded flags
//
//  Returns
//
//  BOOL	-  TRUE if the ACL Entry type could be encoded
//			-  FALSE if the ACL Entry type could not be encoded
//
//  Notes:
//
BOOL DXACLRuleSetParser::encodeEntryType(char *szEntryType, WORD *wEntryType)
{
	//  Pass back the flag settings for this type of entry
	if (strcmpi(szEntryType, ENTRY_PERSON) == 0)
	{
		*wEntryType = ACL_FLAG_PERSON;
		return TRUE;
	}
	if (strcmpi(szEntryType, ENTRY_SERVER) == 0)
	{
		*wEntryType = ACL_FLAG_SERVER;
		return TRUE;
	}
	if (strcmpi(szEntryType, ENTRY_PERSONGROUP) == 0)
	{
		*wEntryType = ACL_FLAG_PERSON | ACL_FLAG_GROUP;
		return TRUE;
	}
	if (strcmpi(szEntryType, ENTRY_SERVERGROUP) == 0)
	{
		*wEntryType = ACL_FLAG_SERVER | ACL_FLAG_GROUP;
		return TRUE;
	}
	if (strcmpi(szEntryType, ENTRY_MIXEDGROUP) == 0)
	{
		*wEntryType = ACL_FLAG_PERSON | ACL_FLAG_SERVER | ACL_FLAG_GROUP;
		return TRUE;
	}
	if (strcmpi(szEntryType, ENTRY_UNSPECIFIED) == 0)
	{
		*wEntryType = 0;
		return TRUE;
	}

	*wEntryType = 0;
	return FALSE;
}

//  encodeAccessLevel
//
//  This function will compute the flag settings to encode the ACL access level 
//
//  Parameters
//
//  1  -  char *	- String containng the ACL Access Level desired
//  2  -  WORD *	- Pointer to the word that will contain the encoded flags
//
//  Returns
//
//  BOOL	-  TRUE if the ACL Access Level could be encoded
//			-  FALSE if the ACL Access Level could not be encoded
//
//  Notes:
//
BOOL DXACLRuleSetParser::encodeAccessLevel(char *szAccessLevel, WORD *wAccessLevel)
{
	//  Pass back the flag settings for this level of access
	if (strcmpi(szAccessLevel, LEVEL_MANAGER) == 0)
	{
		*wAccessLevel = ACL_LEVEL_MANAGER;
		return TRUE;
	}
	if (strcmpi(szAccessLevel, LEVEL_DESIGNER) == 0)
	{
		*wAccessLevel = ACL_LEVEL_DESIGNER;
		return TRUE;
	}
	if (strcmpi(szAccessLevel, LEVEL_EDITOR) == 0)
	{
		*wAccessLevel = ACL_LEVEL_EDITOR;
		return TRUE;
	}
	if (strcmpi(szAccessLevel, LEVEL_AUTHOR) == 0)
	{
		*wAccessLevel = ACL_LEVEL_AUTHOR;
		return TRUE;
	}
	if (strcmpi(szAccessLevel, LEVEL_READER) == 0)
	{
		*wAccessLevel = ACL_LEVEL_READER;
		return TRUE;
	}
	if (strcmpi(szAccessLevel, LEVEL_DEPOSITOR) == 0)
	{
		*wAccessLevel = ACL_LEVEL_DEPOSITOR;
		return TRUE;
	}
	if (strcmpi(szAccessLevel, LEVEL_NOACCESS) == 0)
	{
		*wAccessLevel = ACL_LEVEL_NOACCESS;
		return TRUE;
	}

	*wAccessLevel = 0;
	return FALSE;
}

//  encodeOptionType
//
//  This function will compute the flag settings to encode the ACL Option flag 
//
//  Parameters
//
//  1  -  char *	- String containng the ACL Option Type desired
//  2  -  WORD *	- Pointer to the word that will contain the encoded flags
//
//  Returns
//
//  BOOL	-  TRUE if the ACL Option type could be encoded
//			-  FALSE if the ACL Option type could not be encoded
//
//  Notes:
//
BOOL DXACLRuleSetParser::encodeOptionType(char *szOptionType, WORD *wOptionType)
{
	//  Pass back the flag settings for this option
	if (strcmpi(szOptionType, OPTION_NOCREATEDOCS) == 0)
	{
		*wOptionType = ACL_FLAG_AUTHOR_NOCREATE;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_NODELETEDOCS) == 0)
	{
		*wOptionType = ACL_FLAG_NODELETE;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_CREATEPRIVATEAGENTS) == 0)
	{
		*wOptionType = ACL_FLAG_CREATE_PRAGENT;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_CREATEPRIVATEFOLDERS) == 0)
	{
		*wOptionType = ACL_FLAG_CREATE_PRFOLDER;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_CREATEFOLDERS) == 0)
	{
		*wOptionType = ACL_FLAG_CREATE_FOLDER;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_CREATELS) == 0)
	{
		*wOptionType = ACL_FLAG_CREATE_LOTUSSCRIPT;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_PUBLICREADER) == 0)
	{
		*wOptionType = ACL_FLAG_PUBLICREADER;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_PUBLICWRITER) == 0)
	{
		*wOptionType = ACL_FLAG_PUBLICWRITER;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_NOMONITORS) == 0)
	{
		*wOptionType = ACL_FLAG_MONITORS_DISALLOWED;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_NOREPLICATE) == 0)
	{
		*wOptionType = ACL_FLAG_NOREPLICATE;
		return TRUE;
	}
	if (strcmpi(szOptionType, OPTION_UPDREADERAUTHOR) == 0)
	{
		*wOptionType = ACL_FLAG_ADMIN_READERAUTHOR;
		return TRUE;
	}

	*wOptionType = 0;
	return FALSE;
}

//  isOptionOn
//
//  This function will indicate if an Option should be set on or off according to the node definition
//  On/Off and Yes/No combinations are both valid
//
//  Parameters
//
//  1  -  char *	- String containng the ACL Option setting desired
//
//  Returns
//
//  BOOL	-  TRUE if the ACL Option should be set on
//			-  FALSE if the ACL Option should be set off
//
//  Notes:
//
BOOL DXACLRuleSetParser::isOptionOn(char *szSetting)
{
	if (strcmpi(szSetting, SETTING_ON) == 0) return TRUE;
	if (strcmpi(szSetting, SETTING_YES) == 0) return TRUE;
	
	// Note we do not explicitly check for Off/No - this is the assumed default (anything goes)
	return FALSE;
}

//  checkRuleSet
//
//  This function will check the logical consitency of a Rule set and indicate the state
//
//  Parameters
//
//  1  -  DXACLRuleSet *	- pointer to the ACL Rule Set that is to be checked
//
//  Returns
//
//  BOOL	-  TRUE if the Rule Set is valid and consistent
//			-  FALSE if the Rule Set is invalid or incosnsistent
//
//  Notes:
//
BOOL DXACLRuleSetParser::checkRuleSet(DXACLRuleSet *arPopulated)
{
	BOOL			bValidateResult;

	//  Perform the base logic validation
	bValidateResult = arPopulated->validate(0);

	//  Return the result
	return bValidateResult;
}

