#pragma once
//  DXACLRuleSetParser
//
//  The DXACLRuleSetParser will parse a supplied XML document from a source that a DXResourceLoader can handle
//  it will parse the document into a complete DXACLRuleSet object with all of it's child objects.
//  A DXACLRuleSet forms a complete set of rules that describe how to change a database ACL to the desired
//  Configuration.
//
//  Message Identifier:      DXR
//  Message Range:           7101 - 7199
//  Trace area assignment:   TRACE_ACLRSPARSER (21)
//  Version:                 3.11.1   Build: 99
//
//  Version History:
//
//  3.5.0		15/08/2009		Initial version
//  3.9.1		03/09/2010		XML specification changed
//  3.9.2		22/09/2010		Various fixes to ACL processing
//  3.10.7		17/05/2011		Fixed an issue with ACL processing
//  3.10.9		04/06/2011		Some tidying up for Linux
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 2011
//

//  Application includes
#include	"../ExecEnvironment.h"					//  DX Run Time
#include	"../DXResourceLoader.h"					//  Resource Loader
#include	"DXACLRuleSet.h"						//  ACL Rule Set

//  XML Parser Engine FSM States
#define		XMLP_STATE_UNDEFINED		1			//  Undefined state (i.e. parsing not started)
#define		XMLP_STATE_FINISHED			2			//  Finished state XML document was well-formed
#define		XMLP_STATE_TERMERROR		3			//  Terminal Error state the XML document was malformed
#define		XMLP_STATE_XMLDECL			4			//  XML Document Declared
#define		XMLP_STATE_RSSTARTED		5			//  ACL Rule Set started
#define		XMLP_STATE_RULESTARTED		6			//  An ACL Rule has been started
#define		XMLP_STATE_CESTARTED		7			//  Compulsory Entries started
#define		XMLP_STATE_FESTARTED		8			//  Forbidden Entries started
#define		XMLP_STATE_PESTARTED		9			//  Permitted Entries started
#define		XMLP_STATE_ROLESTARTED		10			//  A Role entry has been started
#define		XMLP_STATE_OPTSTARTED		11			//  An Option entry has been started

//  XML Line Types
#define		XMLLINE_START_NODE			1			//  A start Node <NodeName .......>
#define		XMLLINE_END_NODE			2			//  An end Node </NodeName>
#define		XMLLINE_FULL_NODE			3			//  A complete Node <NodeName ........ />
#define		XMLLINE_EOF					4			//  End of File
#define		XMLLINE_DECL				5			//  XML Declaration  <?xml ..... ?>
#define		XMLLINE_DOCTYPE				6			//  XML Doctype <!DOCTYPE    !>
#define		XMLLINE_COMMENT				7			//  XML Comment <--     -->
#define		XMLLINE_UNREC				8			//  Unrecognised element
#define		XMLLINE_OVERFLOW			9			//  Element has overflowed the line buffer
#define		XMLLINE_MALFORMED			10			//  Malformed XML element has been detected

//  XML Node Names
#define		XMLNODE_ACLRULESET			"ACLRuleSet"
#define		XMLNODE_ADMINSERVER			"AdminServer"
#define		XMLNODE_COMPENTRIES			"CompulsoryEntries"
#define		XMLNODE_FORBIDENTRIES		"ForbiddenEntries"
#define		XMLNODE_OKENTRIES			"PermittedEntries"
#define		XMLNODE_ACLRULE				"ACLRule"
#define		XMLNODE_OPTION				"Option"
#define		XMLNODE_ROLE				"Role"

//  XML Attribute Names
#define		XMLATTR_NAME				"Name"
#define		XMLATTR_PATTERN				"Pattern"
#define		XMLATTR_TYPE				"Type"
#define		XMLATTR_LEVEL				"Level"
#define		XMLATTR_SET					"Set"
#define		XMLATTR_ASSIGN				"Assign"

//  Context Settings for populating an ACLRule entry
#define		PCONTEXT_ADMINSERVER		1			//  Populating a rule for an Admin Server
#define		PCONTEXT_COMPENTRIES		2			//  Populating rules for compulsory entries
#define		PCONTEXT_FORBIDENTRIES		3			//  Populating rules for forbidden entries
#define		PCONTEXT_OKENTRIES			4			//  Populating rules for permitted entries

//  ACL Entry Types
#define		ENTRY_PERSON				"Person"
#define		ENTRY_SERVER				"Server"
#define		ENTRY_PERSONGROUP			"PersonGroup"
#define		ENTRY_SERVERGROUP			"ServerGroup"
#define		ENTRY_MIXEDGROUP			"MixedGroup"
#define		ENTRY_UNSPECIFIED			"Unspecified"

//  ACL Access Levels
#define		LEVEL_MANAGER				"Manager"
#define		LEVEL_DESIGNER				"Designer"
#define		LEVEL_EDITOR				"Editor"
#define		LEVEL_AUTHOR				"Author"
#define		LEVEL_READER				"Reader"
#define		LEVEL_DEPOSITOR				"Depositor"
#define		LEVEL_NOACCESS				"NoAccess"

//  ACL Options
#define		OPTION_NOCREATEDOCS			"NoCreateDocs"
#define		OPTION_NODELETEDOCS			"NoDeleteDocs"
#define		OPTION_CREATEPRIVATEAGENTS	"CreatePrivateAgents"
#define		OPTION_CREATEPRIVATEFOLDERS	"CreatePrivateFolders"
#define		OPTION_CREATEFOLDERS		"CreateFolders"
#define		OPTION_CREATELS				"CreateLotusScript"
#define		OPTION_PUBLICREADER			"PublicReader"
#define		OPTION_PUBLICWRITER			"PublicWriter"
#define		OPTION_NOMONITORS			"NoMonitors"
#define		OPTION_NOREPLICATE			"NoReplicate"
#define		OPTION_UPDREADERAUTHOR		"UpdateReadersAuthors"

//  Miscellaneous definitions
#define		SETTING_ON					"On"
#define		SETTING_YES					"Yes"
#define		ADMIN_SERVER_HOST			"Host"
#define		ADMIN_SERVER_NONE			"None"

//  Error Messages
#define		MSG_DXR7103E				"DXR7103E: Failed to load the requested Rule Set XML '%s'."
#define		MSG_DXR7104E				"DXR7104E: Failed to load the requested Rule Set XML from Item '%s' in Note: 0x%04X."
#define		MSG_DXR7105E				"DXR7105E: Failed to parse the requested Rule Set XML '%s'."
#define		MSG_DXR7106E				"DXR7106E: Failed to parse the requested Rule Set XML from Item '%s' in Note: 0x%04X."
#define		MSG_DXR7110E				"DXR7110E: The XML ACL Rule Set is empty."
#define		MSG_DXR7111E				"DXR7111E: The XML element was not expected at this point in the document."
#define		MSG_DXR7112E				"DXR7112E: XML parse error with: %s."
#define		MSG_DXR7113E				"DXR7113E: A '%s' Node cannot be defined at this point in the XML document."
#define		MSG_DXR7114E				"DXR7114E: You may not define more than one AdminServer rule in a Rule Set."
#define		MSG_DXR7115E				"DXR7115E: You may not define more than one set of Compulsory rules in a Rule Set."
#define		MSG_DXR7116E				"DXR7116E: You may not define more than one set of Forbidden rules in a Rule Set."
#define		MSG_DXR7117E				"DXR7117E: You may not define more than one set of Permitted rules in a Rule Set."
#define		MSG_DXR7118E				"DXR7118E: The end of the file has been reached before the end of the XML document."
#define		MSG_DXR7119E				"DXR7119E: A '%s' Node cannot be closed at this point in the XML document."
#define		MSG_DXR7120E				"DXR7120E: The rule set supplied does not contain any rules, although it is well-formed."
#define		MSG_DXR7122E				"DXR7122E: The Name= attribute in a rule was missing or empty."
#define		MSG_DXR7123E				"DXR7123E: You may not specfy wildcards in the name for a compulsory entry rule."
#define		MSG_DXR7124E				"DXR7124E: Unable to encode the ACL Entry Type '%s', rule is invalid."
#define		MSG_DXR7125E				"DXR7125E: Unable to encode the ACL Access Level '%s', rule is invalid."
#define		MSG_DXR7126E				"DXR7126E: Unable to encode the ACL Option '%s', rule is invalid."
#define		MSG_DXR7127E				"DXR7127E: The Role name has not been supplied, rule is invalid."
#define		MSG_DXR7132E				"DXR7132E: The rule set could not be resized to %i bytes."

//  Information Messages
#define		MSG_DXR7108I				"DXR7108I: XML: %s."
#define		MSG_DXR7130I				"DXR7130I: The rule set has been loaded from '%s'."
#define		MSG_DXR7131I				"DXR7131I: The rule set has been loaded from Item '%s' in Note: 0x%04X."

//  Trace Messages
#define		MSG_DXR7101T				"DXR7101T: A new DXACLRuleSetParser instance has been created @%p."
#define		MSG_DXR7102T				"DXR7102T: A DXACLRuleSetParser instance @%p has been destroyed."
#define		MSG_DXR7107T				"DXR7107T: Starting XML Parse engine."
#define		MSG_DXR7109T				"DXR7109T: Processing the current XML Line type: %i at state %i."
#define		MSG_DXR7121T				"DXR7121T: Rule Set has been parsed with %i Admin servers and %i compulsory, %i forbidden and %i permitted rules."
#define		MSG_DXR7128T				"DXR7128T: Start search for '%s=' in '%s'."
#define		MSG_DXR7129T				"DXR7129T: Determined attribute '%s=%s'."

//  Class Definition
class DXACLRuleSetParser
{
public:
	//  Constructor
	DXACLRuleSetParser(ExecEnvironment *xeParent, int iThreadID);

	//  Destructor
	~DXACLRuleSetParser(void);

	//  Public Members
	DXResourceLoader		*rlCurrent;					//  Current Resource Loader to use

	//  Public Functions
	DXACLRuleSet * parseTheseRules(char *szRuleSet, int iThreadID);	//  Parse the specified Rule Set resource
	DXACLRuleSet * parseTheseRules(NOTEHANDLE hnSource, char *szItemName, int iThreadID);

private:
	//  Private Members
	ExecEnvironment			*xeLocal;					//  DX Run Time
	int						iParentThread;				//  Parent Thread

	//  Private Functions
	BOOL parseTheseRules(MEMHANDLE hmRuleSet, DXACLRuleSet *arEmpty, int iThreadID);	//  Parse a Rule Set from Memory Handle
	BOOL parseTheseRules(char *pXML, DXACLRuleSet *arEmpty, int iThreadID);			//  Parse a Rule Set from Memory
	char * getNextXMLLine(char *pXML, char *szXMLLine, int iMaxLine, int *iType, int iThreadID);		//  Extract the next XML line from the document
	void getXMLNode(char *szNode, char *szLine, int iThreadID);						//  Extract the Node name from a line of XML
	BOOL populateRuleFromXML(char *szXML, DXACLRule *arRule, int iContext, int iThreadID);  //  Populate an XML Rule from XML
	BOOL applyOptionFromXML(char *szXML, DXACLRule *arRule, int iContext, int iThreadID);  //  Apply an Option to a Rule
	BOOL applyRoleFromXML(char *szXML, DXACLRule *arRule, int iContext, int iThreadID);  //  Apply a Role to a Rule
	void getAttribute(char *szAttr, char *szValue, int iMaxValue, char *szLine, int iThreadID);	//  Extract an attribute value from a line of XML
	BOOL hasWildcards(char *szString);									//  Checks if a string has wildcard definitions
	BOOL encodeEntryType(char *szEntryType, WORD *wEntryType);			//  Encode the presented value for the acl entry type
	BOOL encodeAccessLevel(char *szAccessLevel, WORD *wAccessLevel);	//  Encode the presented value for the acl access level
	BOOL encodeOptionType(char *szOptionType, WORD *wOptionType);		//  Encode the presented value for the option type
	BOOL isOptionOn(char *szSetting);									//  See if an option should be on or off
	BOOL checkRuleSet(DXACLRuleSet *arPopulated);						//  Check a Rule Set
};
