#pragma once
//  Attachments
//
//  This class provides functions for handling Notes Attachments
//
//  Message Identifier:      DXR
//  Message Range:           5500 - 5599
//  Trace area assignment:   TRACE_ATT (31)
//  DX Version:				 3.14.0
//  Version:                 3.14.0  Build: 109
//
//  Version History:
//
//	3.13.0		20/12/2013		Incorporated into DX Core
//	3.14.0		28/01/2015		Fix a compression overflow
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2012 - 2015
//

//  Application Includes
#include "Attachments.h"

//  Default Constructor
Attachments::Attachments()
{
}

//  Normal Constructor
Attachments::Attachments(ExecEnvironment *xeParent, int iThreadID)
{
	//  Save the runtime
	xeLocal = xeParent;
	return;
}

//  RecompressAttachment
//
//  This function will recompress the specified attachment to the target compression type specified.
//
//  Parameters
//
//  1  -  NOTEHANDLE		-  Handle of the note containing the attachment
//	2  -  BLOCKID			-  BLOCKID identifying the attachment to be recompressed
//	3  -  WORD				-  Target compression type
//  4  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  DWORD  -			Bytes reduction in size of the attachment
//						0 nothing was done (not recompressed)
//						MAXDWORD An error was encountered preventing recompression
//
//  API Use Notes:
//
//
DWORD Attachments::RecompressAttachment(NOTEHANDLE hnNote, BLOCKID Attachment, WORD CType, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLTarget;						//  Local Thread handle for the Target database
	WORD				wNameLen = 0;						//  Length of the returned item name
	WORD				wFlags;								//  Item flags
	WORD				wDataType;							//  Data Type
	BLOCKID				bidValue;							//  Blockid of the Item Value
	DWORD				dwDataLen;							//  Data Length
	UCHAR				*pValue;							//  Pointer to the value
	FILEOBJECT			*foSrc;								//  Pointer to the source file object
	FILEOBJECT			foAttachment;						//  File object
	BYTE				SeqNo;								//  Sequence number
	BYTE				DupItemID;							//  Duplicate Item ID
	NOTEID				nidContainer;						//  Note ID or the container note
	WORD				RecompressMethod;					//  Method to select for recompression
	RRV					rrvNew;								//  Replacement RRV
	RRV					rrvOld;								//  Old Object RRV
	DWORD				ObjectSize;							//  Object size (old & new)
	DWORD				OldSize;							//  Old (original) object size
	BLOCKID				bidNewValue;						//  BLOCKID for the new item value
	WORD				*pType;								//  Pointer to data type word
	BLOCKID				bidNewItem;							//  BLOCKID of the new item
	WORD				OldComp;							//  Old compression type
	BOOL				IsSigned;							//  Note is signed
	BOOL				IsSealed;							//  Note is encrypted
	UCHAR				cFileName[MAXPATH + 1];				//  Saved original file name
	char				szItemName[MAX_DESC + 1];			//  Item Name
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer


	//  Get the DB Handle to use and the ID of the container note
	NSFNoteGetInfo(hnNote, _NOTE_DB, &hdbTLTarget);
	NSFNoteGetInfo(hnNote, _NOTE_ID, &nidContainer);

	//  Get info about the target attachment
	NSFItemQueryEx(hnNote,
				 Attachment,
				 szItemName,
				 MAX_DESC,
				 &wNameLen,
				 &wFlags,
				 &wDataType,
				 &bidValue,
				 &dwDataLen,
				 &SeqNo,
				 &DupItemID);
	szItemName[wNameLen] = '\0';

	//  Check that we have got a valid type
	if (wDataType != TYPE_OBJECT)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5500E, szItemName, nidContainer, wDataType);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Check that the returned length is adequate
	if (dwDataLen < sizeof(FILEOBJECT))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5501E, szItemName, nidContainer, dwDataLen);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Read the FILEOBJECT for this attachment
	pValue = OSLockBlock(UCHAR, bidValue);
	foSrc = (FILEOBJECT *) (pValue + sizeof(WORD));
	ODSReadMemory(&foSrc, _FILEOBJECT, &foAttachment, 1);
	memcpy(cFileName, pValue + sizeof(WORD) + ODSLength(_FILEOBJECT), foAttachment.FileNameLength);
	OSUnlockBlock(bidValue);

	//  Check that the object is a FILEOBJECT
	if (foAttachment.Header.ObjectType != OBJECT_FILE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5502E, szItemName, nidContainer, foAttachment.Header.ObjectType);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Check that the item is not signed or encrypted - if so do not attempt the recompression
	if (wFlags & ITEM_SIGN || wFlags & ITEM_SEAL)
	{
		if (NSFNoteIsSignedOrSealed(hnNote, &IsSigned, &IsSealed))
		{
			if (((wFlags & ITEM_SIGN) && IsSigned) || ((wFlags & ITEM_SEAL) && IsSealed))
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5524E, szItemName, nidContainer);
				xeLocal->LogMessage(szMsg, iThreadID);
				return RETURN_NOT_RECOMPRESSED;
			}
		}
	}

	//  Check that the compression type is not the same as the target compression type
	if (foAttachment.CompressionType == CType)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5503T, DupItemID, nidContainer, CType);
		xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Test if the recompression is suppressed based on the file type derived from the name
	if (suppressCompression((char *) cFileName))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5541I, cFileName);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Select the recompression method - for the time being we only use the single buffer method.
	//  The piped method could be used for large attachments when the server is memory constrained.
	RecompressMethod = RECOMPRESS_SINGLE;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5504T, DupItemID, nidContainer, foAttachment.FileSize, foAttachment.CompressionType, CType, RecompressMethod);
	xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
	
	//  Get the size of the original object
	stAPIRC = NSFDbGetObjectSize(hdbTLTarget, foAttachment.Header.RRV, OBJECT_UNKNOWN, &ObjectSize, NULL, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5505E, nidContainer, DupItemID);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}
	OldSize = ObjectSize;
	rrvOld = foAttachment.Header.RRV;

	//  Recompress the object
	if (RecompressMethod == RECOMPRESS_SINGLE) rrvNew = RecompressObjectSingle(hdbTLTarget, &foAttachment, CType, &ObjectSize, iThreadID);
	else rrvNew = RecompressObjectPiped(hdbTLTarget, &foAttachment, CType, &ObjectSize, iThreadID);

	//  If the recompression failed report it 
	if (rrvNew == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5506E, nidContainer, DupItemID);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;		
	}

	//  Update the container note by replacing the FILEOBJECT with a copy containing the updated RRV & Compression Type
	//  Firstly delete the existing item.
	stAPIRC = NSFItemDeleteByBLOCKID(hnNote, Attachment);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5507E, nidContainer, DupItemID);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);

		//  Delete the recompressed object
		stAPIRC = NSFDbFreeObject(hdbTLTarget, rrvNew);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5508E, rrvNew);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		return RETURN_NOT_RECOMPRESSED;		
	}

	//  Allocate a new memory block to hold the updated FILEOBJECT
	stAPIRC = OSMemAlloc(0, sizeof(WORD) + sizeof(FILEOBJECT) + foAttachment.FileNameLength, &bidNewValue.pool);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5509E, sizeof(WORD) + sizeof(FILEOBJECT) + foAttachment.FileNameLength);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);

		//  Delete the recompressed object
		stAPIRC = NSFDbFreeObject(hdbTLTarget, rrvNew);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5508E, rrvNew);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		return RETURN_RECOMPRESS_ERROR;
	}

	bidNewValue.block = NULLBLOCK;

	//  Build the object
	OldComp = foAttachment.CompressionType;
	foAttachment.Header.RRV = rrvNew;
	foAttachment.CompressionType = CType;
	pValue = OSLockBlock(UCHAR, bidNewValue);
	pType = (WORD *) pValue;
	*pType = wDataType;
	foSrc = (FILEOBJECT *) (pValue + sizeof(WORD));
	ODSWriteMemory(&foSrc, _FILEOBJECT, &foAttachment, 1);
	memcpy(pValue + sizeof(WORD) + ODSLength(_FILEOBJECT), cFileName, foAttachment.FileNameLength);
	OSUnlockBlock(bidNewValue);
	
	//  Append the new block to the container note
	stAPIRC = NSFItemAppendByBLOCKID(hnNote,
									  wFlags,
									  ITEM_NAME_ATTACHMENT, 
									  (WORD) strlen(ITEM_NAME_ATTACHMENT),
									  bidNewValue,
									  dwDataLen,
									  &bidNewItem);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5510E, nidContainer);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);

		//  Delete the recompressed object
		stAPIRC = NSFDbFreeObject(hdbTLTarget, rrvNew);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5508E, rrvNew);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		return RETURN_RECOMPRESS_ERROR;
	}

	//  Trace the success state
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5511T, rrvOld, OldComp, OldSize, nidContainer, rrvNew, CType, ObjectSize);
	xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
	return OldSize - ObjectSize;
}

//  RecompressAttachmentInPlace
//
//  This function will recompress the specified attachment to the target compression type specified.
//  The existing attachment object content is replaced with recompressed stream if it is smaller than the
//  existing stream.
//
//  Parameters
//
//  1  -  NOTEHANDLE		-  Handle of the note containing the attachment
//	2  -  BLOCKID			-  BLOCKID identifying the attachment to be recompressed
//	3  -  WORD				-  Target compression type
//  4  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  DWORD  -			Bytes reduction in size of the attachment
//						0 nothing was done (not recompressed)
//						MAXDWORD An error was encountered preventing recompression
//
//  API Use Notes:
//
//
DWORD Attachments::RecompressAttachmentInPlace(NOTEHANDLE hnNote, BLOCKID Attachment, WORD CType, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLTarget;						//  Local Thread handle for the Target database
	NOTEID				nidContainer;						//  Note ID or the container note
	WORD				wNameLen = 0;						//  Length of the returned item name
	WORD				wFlags;								//  Item flags
	WORD				wDataType;							//  Data Type
	BLOCKID				bidValue;							//  Blockid of the Item Value
	DWORD				dwDataLen;							//  Data Length
	BYTE				SeqNo;								//  Sequence number
	BYTE				DupItemID;							//  Duplicate Item ID
	UCHAR				*pValue;							//  Pointer to the value
	FILEOBJECT			*foSrc;								//  Pointer to the source file object
	FILEOBJECT			foAttachment;						//  File object
	WORD				OldComp;							//  Old compression type
	BOOL				IsSigned;							//  Note is signed
	BOOL				IsSealed;							//  Note is encrypted
	WORD				RecompressMethod;					//  Method to select for recompression
	DWORD				ObjectSize;							//  Object size (old & new)
	DWORD				OldSize;							//  Old (original) object size
	WORD				NewCType;							//  New compression type
	BLOCKID				bidNewValue;						//  BLOCKID for the new item value
	WORD				*pType;								//  Pointer to data type word
	BLOCKID				bidNewItem;							//  BLOCKID of the new item
	UCHAR				cFileName[MAXPATH + 1];				//  Saved original file name
	char				szItemName[MAX_DESC + 1];			//  Item Name
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Mark off the frame
	MARK_FRAME("RecompressAttachmentInPlace()", &stAPIRC, szMsg);

	//  Get the DB Handle to use and the ID of the container note
	NSFNoteGetInfo(hnNote, _NOTE_DB, &hdbTLTarget);
	NSFNoteGetInfo(hnNote, _NOTE_ID, &nidContainer);

	//  Get info about the target attachment
	NSFItemQueryEx(hnNote,
				 Attachment,
				 szItemName,
				 MAX_DESC,
				 &wNameLen,
				 &wFlags,
				 &wDataType,
				 &bidValue,
				 &dwDataLen,
				 &SeqNo,
				 &DupItemID);
	szItemName[wNameLen] = '\0';

	//  Check that we have got a valid type
	if (wDataType != TYPE_OBJECT)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5500E, szItemName, nidContainer, wDataType);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Check that the returned length is adequate
	if (dwDataLen < sizeof(FILEOBJECT))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5501E, szItemName, nidContainer, dwDataLen);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Read the FILEOBJECT for this attachment
	pValue = OSLockBlock(UCHAR, bidValue);
	foSrc = (FILEOBJECT *) (pValue + sizeof(WORD));
	ODSReadMemory(&foSrc, _FILEOBJECT, &foAttachment, 1);
	memcpy(cFileName, pValue + sizeof(WORD) + ODSLength(_FILEOBJECT), foAttachment.FileNameLength);
	cFileName[foAttachment.FileNameLength] = '\0';
	OSUnlockBlock(bidValue);

	//  Check that the object is a FILEOBJECT
	if (foAttachment.Header.ObjectType != OBJECT_FILE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5502E, szItemName, nidContainer, foAttachment.Header.ObjectType);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Save the existing compression type
	OldComp = foAttachment.CompressionType;

	//  Check that the item encrypted - if so do not attempt the recompression
	if (wFlags & ITEM_SEAL)
	{
		if (NSFNoteIsSignedOrSealed(hnNote, &IsSigned, &IsSealed))
		{
			if (((wFlags & ITEM_SIGN) && IsSigned) || ((wFlags & ITEM_SEAL) && IsSealed))
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5524E, szItemName, nidContainer);
				xeLocal->LogMessage(szMsg, iThreadID);
				return RETURN_NOT_RECOMPRESSED;
			}
		}
	}

	//  Test if the recompression is suppressed based on the file type derived from the name
	if (suppressCompression((char *) cFileName))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5541I, cFileName);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Select the recompression method - for the time being we only use the single buffer method.
	//  The piped method could be used for large attachments when the server is memory constrained.
	RecompressMethod = RECOMPRESS_SINGLE;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5504T, DupItemID, nidContainer, foAttachment.FileSize, foAttachment.CompressionType, CType, RecompressMethod);
	xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
	
	//  Get the size of the original object
	stAPIRC = NSFDbGetObjectSize(hdbTLTarget, foAttachment.Header.RRV, OBJECT_UNKNOWN, &ObjectSize, NULL, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5505E, nidContainer, DupItemID);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}
	OldSize = ObjectSize;

	//  Recompress the object in place
	if (RecompressMethod == RECOMPRESS_SINGLE) NewCType = RecompressObjectSingleInPlace(hdbTLTarget, &foAttachment, CType, &ObjectSize, iThreadID);
	else NewCType = RecompressObjectPipedInPlace(hdbTLTarget, &foAttachment, CType, &ObjectSize, iThreadID);

	//  If the recompression was not executed
	if (NewCType == OldComp)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5536W, foAttachment.Header.RRV, nidContainer, DupItemID);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;		
	}

	//  Firstly delete the existing item.
	stAPIRC = NSFItemDeleteByBLOCKID(hnNote, Attachment);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5507E, nidContainer, DupItemID);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;		
	}

	//  Allocate a new memory block to hold the updated FILEOBJECT
	stAPIRC = OSMemAlloc(0, sizeof(WORD) + sizeof(FILEOBJECT) + foAttachment.FileNameLength, &bidNewValue.pool);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5509E, sizeof(WORD) + sizeof(FILEOBJECT) + foAttachment.FileNameLength);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_RECOMPRESS_ERROR;
	}

	bidNewValue.block = NULLBLOCK;

	//  Build the object
	foAttachment.CompressionType = NewCType;
	pValue = OSLockBlock(UCHAR, bidNewValue);
	pType = (WORD *) pValue;
	*pType = wDataType;
	foSrc = (FILEOBJECT *) (pValue + sizeof(WORD));
	ODSWriteMemory(&foSrc, _FILEOBJECT, &foAttachment, 1);
	memcpy(pValue + sizeof(WORD) + ODSLength(_FILEOBJECT), cFileName, foAttachment.FileNameLength);
	OSUnlockBlock(bidNewValue);

	//  Append the new block to the container note
	stAPIRC = NSFItemAppendByBLOCKID(hnNote,
									  wFlags,
									  ITEM_NAME_ATTACHMENT, 
									  (WORD) strlen(ITEM_NAME_ATTACHMENT),
									  bidNewValue,
									  dwDataLen,
									  &bidNewItem);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5510E, nidContainer);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_RECOMPRESS_ERROR;
	}

	//  Get the size of the original object
	stAPIRC = NSFDbGetObjectSize(hdbTLTarget, foAttachment.Header.RRV, OBJECT_UNKNOWN, &ObjectSize, NULL, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5505E, nidContainer, DupItemID);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return RETURN_NOT_RECOMPRESSED;
	}

	//  Trace the success state
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5511T,foAttachment.Header.RRV, OldComp, OldSize, nidContainer, foAttachment.Header.RRV, CType, ObjectSize);
	xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
	return OldSize - ObjectSize;
}

//  RecompressObjectSingle
//
//  This function will recompress the specified object and store the recompressed stream as a new object
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the container database
//	2  -  FILEOBJECT *		-  Pointer to the file object descriptor
//  3  -  WORD				-  Target compression type
//	4  -  DWORD				-  Size of the input and output objects
//  5  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  RRV		-			RRV of the recompressed object
//						NULL if the object could not be recompressed
//
//  API Use Notes:
//
//
RRV Attachments::RecompressObjectSingle(DBHANDLE hdbContainer, FILEOBJECT *foIn, WORD CType, DWORD *dwSize, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DHANDLE				hCBlock;							//  Handle of the compressed block
	DHANDLE				hRCBlock;							//  Handle of the recompressed block
	DWORD				dwCBSize;							//  Size of the compressed block
	DWORD				dwRCBSize;							//  Size of the recompressed block
	RRV					NewRRV;								//  RRV of the new object
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	dwCBSize = *dwSize;
	dwRCBSize = *dwSize;

	//  Read the input object into memory
	stAPIRC = NSFDbReadObject(hdbContainer, foIn->Header.RRV, 0, MAXDWORD, &hCBlock);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5512E, foIn->Header.RRV);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULL;
	}

	//  Pass the compressed memory block to the recompressor
	hRCBlock = RecompressBlock(hCBlock, foIn->CompressionType, CType, &dwRCBSize, foIn->FileSize, iThreadID);
	OSMemFree(hCBlock);
	if (hRCBlock == NULLHANDLE) return NULL;

	//  Check that the recompression did not inflate the block - if so drop it
	if (dwRCBSize > dwCBSize)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5513T, foIn->Header.RRV, dwCBSize, dwRCBSize);
		xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
		OSMemFree(hRCBlock);
		return NULL;
	}

	//  Write the recompressed object as a new object in the database
	//  Allocate the new object
	stAPIRC = NSFDbAllocObject(hdbContainer, dwRCBSize, NOTE_CLASS_DOCUMENT, 0, &NewRRV);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5514E, dwRCBSize);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemFree(hRCBlock);
		return NULL;		
	}

	//  Write the stream to the newly allocated object
	stAPIRC = NSFDbWriteObject(hdbContainer, NewRRV, hRCBlock, 0, dwRCBSize);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5515E, dwRCBSize);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemFree(hRCBlock);
		stAPIRC = NSFDbFreeObject(hdbContainer, NewRRV);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5516E, NewRRV);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);	
		}
		return NULL;		
	}

	//  Free the recompressed block
	OSMemFree(hRCBlock);

	//  Return to caller
	*dwSize = dwRCBSize;
	return NewRRV;
}

//  RecompressObjectSingleInPlace
//
//  This function will recompress the specified object and store the recompressed stream as a new object
//  The compressed stream will replace the stream in the existing object.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the container database
//	2  -  FILEOBJECT *		-  Pointer to the file object descriptor
//  3  -  WORD				-  Target compression type
//	4  -  DWORD				-  Size of the input and output objects
//  5  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  WORD		-		The new compression type of the object
//
//  API Use Notes:
//
//
WORD Attachments::RecompressObjectSingleInPlace(DBHANDLE hdbContainer, FILEOBJECT *foIn, WORD CType, DWORD *dwSize, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DHANDLE				hCBlock;							//  Handle of the compressed block
	DHANDLE				hRCBlock;							//  Handle of the recompressed block
	DWORD				dwCBSize;							//  Size of the compressed block
	DWORD				dwRCBSize;							//  Size of the recompressed block
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Mark off the frame
	MARK_FRAME("RecompressObjectSingleInPlace()", &stAPIRC, szMsg);

	dwCBSize = *dwSize;
	dwRCBSize = *dwSize;

	//  Read the input object into memory
	stAPIRC = NSFDbReadObject(hdbContainer, foIn->Header.RRV, 0, MAXDWORD, &hCBlock);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5512E, foIn->Header.RRV);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return foIn->CompressionType;
	}

	//  Pass the compressed memory block to the recompressor
	hRCBlock = RecompressBlock(hCBlock, foIn->CompressionType, CType, &dwRCBSize, foIn->FileSize, iThreadID);
	if (hRCBlock == NULLHANDLE) 
	{
		OSMemFree(hCBlock);
		return foIn->CompressionType;
	}

	//  Make sure that the new stream is smaller than the original
	if (dwRCBSize > dwCBSize)
	{
		OSMemFree(hCBlock);
		OSMemFree(hRCBlock);
		return foIn->CompressionType;	
	}

	//  Reallocate the object to hold the recompressed stream
	stAPIRC = NSFDbReallocObject(hdbContainer, foIn->Header.RRV, dwRCBSize);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5537E, foIn->Header.RRV, dwRCBSize);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemFree(hCBlock);
		OSMemFree(hRCBlock);
		return foIn->CompressionType;		
	}

	//  Write the stream to the newly allocated object
	stAPIRC = NSFDbWriteObject(hdbContainer, foIn->Header.RRV, hRCBlock, 0, dwRCBSize);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5515E, dwRCBSize);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);

		//  Restore the object to the original state
		stAPIRC = NSFDbReallocObject(hdbContainer, foIn->Header.RRV, dwCBSize);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5537E, foIn->Header.RRV, dwCBSize);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			OSMemFree(hCBlock);
			OSMemFree(hRCBlock);
			return foIn->CompressionType;		
		}

		stAPIRC = NSFDbWriteObject(hdbContainer, foIn->Header.RRV, hCBlock, 0, dwCBSize);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5515E, dwCBSize);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}

		OSMemFree(hCBlock);
		OSMemFree(hRCBlock);
		return foIn->CompressionType;		
	}

	//  Free the recompressed block
	OSMemFree(hCBlock);
	OSMemFree(hRCBlock);

	//  Return to caller
	*dwSize = dwRCBSize;
	return CType;
}

//  RecompressObjectPiped
//
//  This function will recompress the specified object and store the recompressed stream as a new object.
//  For the time being this is just a wrapper for the RecompressObjectSingle function.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the container database
//	2  -  FILEOBJECT *		-  Pointer to the file object descriptor
//  3  -  WORD				-  Target compression type
//	4  -  DWORD				-  Size of the input and output objects
//  5  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  RRV		-			RRV of the recompressed object
//						NULL if the object could not be recompressed
//
//  API Use Notes:
//
//
RRV Attachments::RecompressObjectPiped(DBHANDLE hdbContainer, FILEOBJECT *foIn, WORD CType, DWORD *dwSize, int iThreadID)
{
	return RecompressObjectSingle(hdbContainer, foIn, CType, dwSize, iThreadID);
}

//  RecompressObjectPiped
//
//  This function will recompress the specified object and store the recompressed stream as a new object.
//  For the time being this is just a wrapper for the RecompressObjectSingleInPlace function.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the container database
//	2  -  FILEOBJECT *		-  Pointer to the file object descriptor
//  3  -  WORD				-  Target compression type
//	4  -  DWORD				-  Size of the input and output objects
//  5  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  WORD		-		The new compression type of the object
//
//  API Use Notes:
//
//
WORD Attachments::RecompressObjectPipedInPlace(DBHANDLE hdbContainer, FILEOBJECT *foIn, WORD CType, DWORD *dwSize, int iThreadID)
{
	return RecompressObjectSingleInPlace(hdbContainer, foIn, CType, dwSize, iThreadID);
}

//  RecompressBlock
//
//  This function will recompress the specified memory block and return the recompressed stream as a new block
//
//  Parameters
//
//  1  -  DHANDLE			-  Handle of the input memory block
//	2  -  WORD				-  Compression type of the input memory block
//  3  -  WORD				-  Compression type of the output memory block
//	4  -  DWORD				-  Size of the input and output blocks
//  5  -  DWORD				-  Size of the uncompressed file
//  6  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  DHANDLE		-			Handle of the recompressed block
//							NULLHANDLE if the block could not be recompressed
//
//  API Use Notes:
//
//
DHANDLE Attachments::RecompressBlock(DHANDLE hBIn, WORD CTypeIn, WORD CTypeOut, DWORD *dwSize, DWORD UFSize, int iThreadID)
{
	DHANDLE			hBU;												//  Uncompressed block handle
	DHANDLE			hBOut;												//  Output block handle
	DWORD			SizeIn,SizeOut;										//  Input and Output size
	char			szMsg[MAX_MSG + 1];									//  Generic message buffer

	//  Mark off the frame
	MARK_FRAME("RecompressBlock()", &hBU, szMsg);

	SizeIn = *dwSize;
	SizeOut = *dwSize;

	//  Input phase decompress the incoming block if necessary
	switch(CTypeIn)
	{
	case COMPRESS_NONE:
		//  The source is not compressed
		hBU = hBIn;
		break;

	case COMPRESS_HUFF:
		//  The source is Huffman compressed - decompress it
		hBU = DecompressBlockHuffman(hBIn, &SizeOut, UFSize, iThreadID);
		break;

	case COMPRESS_LZ1:
		//  The source is LZ1 compressed - decompress it
		hBU = DecompressBlockLZ1(hBIn, &SizeOut, UFSize, iThreadID);
		break;

	default:
		//  Unknown compression type
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5517W, CTypeIn);
		xeLocal->LogMessage(szMsg, iThreadID);
		hBU = NULLHANDLE;
		break;
	}

	if (hBU != NULLHANDLE)
	{
		//  Check that the size of the uncompressed block matches the expected size
		if (SizeOut != UFSize)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5518E, SizeOut, UFSize);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (CTypeIn != COMPRESS_NONE) OSMemFree(hBU);
			return NULLHANDLE;
		}
		
		switch(CTypeOut)
		{
		case COMPRESS_NONE:
			//  The target is not compressed
			hBOut = hBU;

		case COMPRESS_HUFF:
			//  Compress the output using Huffman
			hBOut = CompressBlockHuffman(hBU, &SizeOut, iThreadID);
			break;

		case COMPRESS_LZ1:
			//  Compress the output using LZ1
			hBOut = CompressBlockLZ1(hBU, &SizeOut, iThreadID);
			break;

		default:
			//  Unknown compression type
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5519E, CTypeOut);
			xeLocal->LogMessage(szMsg, iThreadID);
			hBOut = NULLHANDLE;
			SizeOut = 0;
			break;
		}
	}

	//  Cleanup and return to caller
	if (CTypeIn != COMPRESS_NONE) OSMemFree(hBU);
	*dwSize = SizeOut;
	return hBOut;
}

//  CompressBlockHuffman
//
//  This function will compress the specified memory block using Huffman compression
//
//  Parameters
//
//  1  -  DHANDLE			-  Handle of the input memory block
//	2  -  DWORD				-  Size of the input and output blocks
//  3  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  DHANDLE		-			Handle of the compressed block
//							NULLHANDLE if the block could not be compressed
//
//  API Use Notes:
//
//
DHANDLE Attachments::CompressBlockHuffman(DHANDLE hBIn, DWORD *dwSize, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from notesAPI calls
	DHANDLE				hCompressed = NULLHANDLE;			//  Handle of the compressed buffer
	DWORD				CStreamLen = 0;						//  Size of the compressed object
	DWORD				DoneSoFar = 0;						//  Bytes processed
	UCHAR				*pUCB;								//  Pointer to the uncompressed stream
	UCHAR				*pCStream = NULL;					//  Compressed Stream
	DWORD				Freq[256];							//  Frequency array
	WORD				Rank[256];							//  Rank of the frequencies
	DWORD				iIndex;								//  Generic index
	DWORD				Alphabet = 0;						//  Alphabet size
	DWORD				RefCount = 0;						//  Ranking reference count
	DWORD				EntriesSwapped;						//  Entries swapped by sort
	WORD				Temp;								//  Temp - for swapping references
	HuffNode			*PrioQ[260];						//  Priority Queue of Nodes
	DWORD				QEnd = 1;							//  End of the queue
	DWORD				Branches = 0;						//  Count of branches
	DWORD				BitsNeeded;							//  Bits needed to encode the file
	DWORD				ObjectSize;							//  Object size
	DWORD				StreamLen;							//  Available Stream Length
	HuffHeader			*pHHdr;								//  Pointer to the Huffman Header
	Encode				ELT[256];							//  Encoding lookup table
	WORD				NodeID = 0;							//  Node ID
	DWORD				CheckSum = 0;						//  Computed checksum
	char				szMsg[MAX_MSG + 1];					//  Generic Message Buffer

	//  Lock the uncompressed buffer
	pUCB = OSLock(UCHAR, hBIn);

	//  Calculate the frequency of each character in the uncompressed buffer
	for (iIndex = 0; iIndex < 256; iIndex++) Freq[iIndex] = 0;
	for (iIndex = 0; iIndex < *dwSize; iIndex++) 
	{
		Freq[pUCB[iIndex]]++;
		CheckSum += pUCB[iIndex];
	}

	//  Count the alphabet size
	for (iIndex = 0; iIndex < 256; iIndex++) if (Freq[iIndex] > 0) Alphabet++;

	//  Rank the characters in frequency order
	for (iIndex = 0; iIndex < 256; iIndex++) Rank[iIndex] = 0;
	//  Populate the ranking array
	for (iIndex = 0; iIndex < 256; iIndex++)
	{
		if (Freq[iIndex] > 0)
		{
			Rank[RefCount] = (WORD) iIndex;
			RefCount++;
		}
	}

	//  Sort the ranking array into frequency order (bubble sort)
	EntriesSwapped = 1;
	while (EntriesSwapped > 0)
	{
		EntriesSwapped = 0;
		for (iIndex = 1; iIndex < Alphabet; iIndex++)
		{
			if (Freq[Rank[iIndex]] > Freq[Rank[iIndex - 1]])
			{
				//  Swap the entries
				Temp = Rank[iIndex];
				Rank[iIndex] = Rank[iIndex - 1];
				Rank[iIndex - 1] = Temp;
				EntriesSwapped++;
			}
		}
	}

	//  Build the Huffman Tree
	//  The process starts by populating the priority queue with the leaf nodes of the tree
	for (iIndex = 0; iIndex < Alphabet; iIndex++)
	{
		NodeID++;
		QEnd = enqueueNode(createHuffmanNode(Freq[Rank[iIndex]], WORD(Rank[iIndex]), NULL, NULL, NodeID), QEnd, PrioQ);
	}

	//  Add a Leaf node for the logical End-Of-Stream marker (frequency = 1)
	NodeID++;
	QEnd = enqueueNode(createHuffmanNode(1, HUFF_EOS, NULL, NULL, NodeID), QEnd, PrioQ);

	//  The tree is now constructed bottom up (i.e. from the lowest frequency leaves)
	while (QEnd > 2)
	{
		HuffNode			*hnd1;																//  Dequeued node
		HuffNode			*hnd2;																//  Dequeued node
		HuffNode			*hnNew;																//  New intermediate node

		//  Dequeue the two tail nodes (lowest frequency) from the priority queue
		hnd1 = PrioQ[QEnd - 1];
		QEnd--;
		hnd2 = PrioQ[QEnd - 1];
		QEnd--;
		NodeID++;
		hnNew = createHuffmanNode(0, 0, hnd1, hnd2, NodeID);
		QEnd = enqueueNode(hnNew, QEnd, PrioQ);
		Branches++;
	}

	//  Generate the encode values in each of the nodes
	Branches = generateEncodes(PrioQ[1]);

	//  Tree building is complete - the root node is in slot [1] of the priority queue
	//  Compute the size of the compressed object
	BitsNeeded = totalBits(PrioQ[1]);
	ObjectSize = BitsNeeded / 8;
	if (BitsNeeded & 0x00000007) ObjectSize++;
	StreamLen = ObjectSize;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5535T, BitsNeeded, ObjectSize, Branches);
	xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
	ObjectSize += sizeof(DWORD) + sizeof(WORD);										//  Fixed header
	ObjectSize += Branches * sizeof(EHN);											//  Huffman Tree

	//  Allocate the output buffer
	stAPIRC = OSMemAlloc(MEM_GROWABLE, ObjectSize + LZ1_SAFETY_BUFFER, &hCompressed);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5534E, ObjectSize + LZ1_SAFETY_BUFFER);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}
	pCStream = OSLock(UCHAR, hCompressed);
	
	//  Build the fixed part of the compressed stream
	pHHdr = (HuffHeader *) pCStream;
	pHHdr->CheckSum = CheckSum;
	pHHdr->EHTreeSize = WORD(Branches * sizeof(EHN));

	//  Populate the Huffman tree in the header
	encodeHuffmanNode(PrioQ[1], 0, pHHdr);

	//  Populate the encoding lookup table that is used to compress the stream
	for (iIndex = 0; iIndex < 256; iIndex++)
	{
		ELT[iIndex].EncLen = 0;
		ELT[iIndex].EncVal = 0;
	}
	populateEncodingTable(PrioQ[1], ELT);
	// documentEncodingTable(ELT, xeLocal);

	//  This completes all of the pre-steps we can now compress the stream
	CStreamLen = compressBufferHuffman(pUCB, *dwSize, pCStream, StreamLen + LZ1_SAFETY_BUFFER, ELT, iThreadID);

	//  Return to caller
	OSUnlock(hCompressed);
	OSUnlock(hBIn);
	*dwSize = CStreamLen;
	return hCompressed;
}

//  CompressBlockLZ1
//
//  This function will compress the specified memory block using LZ1 compression
//
//  Parameters
//
//  1  -  DHANDLE			-  Handle of the input memory block
//	2  -  DWORD				-  Size of the input and output blocks
//  3  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  DHANDLE		-			Handle of the compressed block
//							NULLHANDLE if the block could not be compressed
//
//  API Use Notes:
//
//
DHANDLE Attachments::CompressBlockLZ1(DHANDLE hBIn, DWORD *dwSize, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from notesAPI calls
	DHANDLE				hCompressed = NULLHANDLE;			//  Handle of the compressed buffer
	DWORD				BytesToCompress;					//  Number of bytes to compress
	DWORD				CStreamLen = 0;						//  Length of the compressed stream
	DWORD				ChunkSizeIn,ChunkSizeOut;			//  Size of the (de)compressed chunk
	DWORD				BytesCompressed = 0;				//  Bytes compressed
	UCHAR				*pCStream;							//  Pointer to the compressed stream
	UCHAR				*pUCB;								//  Pointer to the uncompressed stream
	LZ1Header			*pCHdr;								//  Pointer to an LZ1 header
	LZ1Stats			sCompress;							//  Compression stats
	char				szMsg[MAX_MSG + 1];					//  Generic Message Buffer

	//  Mark off the frame
	MARK_FRAME("CompressBlockLZ1()", &stAPIRC, szMsg);

	//  Envelope checks
	if (hBIn == NULLHANDLE) return NULLHANDLE;
	if (*dwSize == 0) return NULLHANDLE;

	//  Clear the stats structure
	memset(&sCompress, 0, sizeof(LZ1Stats));

	//  Lock the uncompressed stream buffer and set the length to compress
	pUCB = OSLock(UCHAR, hBIn);
	BytesToCompress = *dwSize;

	//  Process the uncompressed stream until it is exhausted
	//  The stream is compressed in 16kb (LZ1_COMPRESS_CHUNK) size chunks
	while(BytesToCompress > 0)
	{
		//  Trace the chunk processing
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5538T, sCompress.Chunks + 1, CStreamLen, BytesCompressed);
		xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);

		//  Allocate or reallocate the buffer to hold the compressed stream
		if (hCompressed == NULLHANDLE) stAPIRC = OSMemAlloc(MEM_GROWABLE, LZ1_COMPRESS_CHUNK + LZ1_SAFETY_BUFFER, &hCompressed);
		else stAPIRC = OSMemRealloc(hCompressed, CStreamLen + LZ1_COMPRESS_CHUNK + LZ1_SAFETY_BUFFER);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5532E, CStreamLen + LZ1_COMPRESS_CHUNK + LZ1_SAFETY_BUFFER);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (hCompressed != NULLHANDLE) OSMemFree(hCompressed);
			OSUnlock(hBIn);
			return NULLHANDLE;
		}
		pCStream = OSLock(UCHAR, hCompressed);

		//  Compress the next chunk of the uncompressed stream
		if (BytesToCompress > LZ1_COMPRESS_CHUNK) ChunkSizeIn = LZ1_COMPRESS_CHUNK;
		else ChunkSizeIn = BytesToCompress;
		sCompress.Chunks++;

		//  Trace the chunk processing
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5539T, sCompress.Chunks, ChunkSizeIn);
		xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);

		ChunkSizeOut = compressLZ1Chunk(pUCB + BytesCompressed, ChunkSizeIn, pCStream + CStreamLen, &sCompress, iThreadID);

		sprintf_s(szMsg, MAX_MSG, MSG_DXR5540T, sCompress.Chunks, ChunkSizeOut);
		xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);

		//  Check that the compression succeeded
		if (ChunkSizeOut == 0)
		{
			OSUnlock(hCompressed);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5533E, sCompress.Chunks);
			xeLocal->LogMessage(szMsg, iThreadID);
			OSMemFree(hCompressed);
			OSUnlock(hBIn);
			return NULLHANDLE;
		}

		//  If this was the last chunk then mark the compressed chunk as the end of the chain
		if (ChunkSizeIn == BytesToCompress)
		{
			pCHdr = (LZ1Header *) (pCStream + CStreamLen);
			pCHdr->Chunks = pCHdr->Chunks | 4;
		}

		OSUnlock(hCompressed);
		BytesCompressed += ChunkSizeIn;
		BytesToCompress -= ChunkSizeIn;
		CStreamLen += ChunkSizeOut;
	}

	//  If the compressed stream is longer than the uncompressed stream then decline to use the compressed stream
	if (CStreamLen > *dwSize)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5534I, *dwSize, CStreamLen);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemFree(hCompressed);
		OSUnlock(hBIn);
		return NULLHANDLE;
	}

	//  Unlock the uncompressed stream
	OSUnlock(hBIn);
	*dwSize = CStreamLen;
	return hCompressed;
}

//  DecompressBlockHuffman
//
//  This function will decompress the specified Huffman compressed memory block 
//
//  Parameters
//
//  1  -  DHANDLE			-  Handle of the input memory block
//	2  -  DWORD				-  Size of the input and output blocks
//  3  -  DWORD				-  Size of the uncompressed file
//  4  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  DHANDLE		-			Handle of the decompressed block
//							NULLHANDLE if the block could not be decompressed
//
//  API Use Notes:
//
//   If the decompressed object will end up larger than the expected size then the decompression is
//   treated as if it failed.
//
DHANDLE Attachments::DecompressBlockHuffman(DHANDLE hBIn, DWORD *dwSize, DWORD UFSize, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DHANDLE				hBOut = NULLHANDLE;					//  Output (decompressed) handle
	DWORD				BufferSize;							//  Buffer Size
	UCHAR				*pCBlock;							//  Pointer to the compressed memory block
	UCHAR				*pDCBfr;							//  Decompressed buffer
	HuffHeader			*pCHdr;								//  Object Header
	UCHAR				*pCStream;							//  Pointer to the compressed stream
	DWORD				BitsAvailable;						//  Compressed data length in bits
	DWORD				BitsUsed;							//  Bits consumed by decompression
	DWORD				BytesDecompressed;					//  Count of bytes decompressed
	DWORD				CChk = 0;							//  Running checksum
	WORD				NextChar;							//  Next Character
	BOOL				EOSDetected = FALSE;				//  End-Of-Stream detected
	MSBitStream			bsBase;								//  Bit streaming object
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Set the overrun size
	BufferSize = UFSize + HUFF_OR_LIMIT;

	//  Allocate a memory block to hold the decompressed output
	stAPIRC = OSMemAlloc(MEM_GROWABLE, BufferSize, &hBOut);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5520E, BufferSize);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Lock the input (compressed) buffer
	pCBlock = OSLock(UCHAR, hBIn);
	pCHdr = (HuffHeader *) pCBlock;

	//  Lock the output (decompressed) object
	pDCBfr = OSLock(UCHAR, hBOut);
	BytesDecompressed = 0;

	//  Calculate the start of the compressed data stream and initialise the bit stream
	pCStream = pCBlock + (sizeof(DWORD) + sizeof(WORD) + pCHdr->EHTreeSize);
	//  Compute the number of bits in the compressed data stream
	BitsAvailable = (*dwSize - (sizeof(DWORD) + sizeof(WORD) + pCHdr->EHTreeSize)) * 8;
	bsBase.setInputBuffer(pCStream, (BitsAvailable / 8) + 1);
	BitsUsed = 0;

	//  Decompression loop
	//
	//  The loop decompresses the input data stream one byte at a time until one of the following conditions
	//  are met (a well formed stream will trigger all three at the same time).
	//
	//	1. The decompressed byte count exceeds the output size
	//	2. All bits in the compressed data stream have been consumed
	//  3. End of stream detected
	//
	while (BytesDecompressed <= UFSize && BitsUsed < BitsAvailable && !EOSDetected)
	{
		//  Extract the next character from the compressed stream
		NextChar = getNextHDChar(&bsBase, pCHdr, &BitsUsed);
		if (NextChar == HUFF_EOS) EOSDetected = TRUE;
		else
		{
			pDCBfr[BytesDecompressed] = UCHAR(NextChar);		
			CChk += pDCBfr[BytesDecompressed];
			BytesDecompressed++;
		}
	}

	//  Unlock the input and output objects
	OSUnlock(hBIn);
	OSUnlock(hBOut);

	//  Examine the terminal state to determine if we trust the decompressed output
	if (BytesDecompressed != UFSize)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5521E, BytesDecompressed, UFSize);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemFree(hBOut);
		return NULLHANDLE;
	}

	if (!EOSDetected)
	{
		//  Residual data is available in the compressed stream and EOS was not detected
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5522E, BitsAvailable - BitsUsed);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemFree(hBOut);
		return NULLHANDLE;
	}

	if (CChk != pCHdr->CheckSum)
	{
		//  Checksum mismatch
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5523E, CChk, pCHdr->CheckSum);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemFree(hBOut);
		return NULLHANDLE;
	}

	//  Set the return size
	*dwSize = BytesDecompressed;
	return hBOut;
}

//  DecompressBlockLZ1
//
//  This function will decompress the specified LZ1 compressed memory block 
//
//  Parameters
//
//  1  -  DHANDLE			-  Handle of the input memory block
//	2  -  DWORD				-  Size of the input and output blocks
//  3  -  DWORD				-  Size of the uncompressed file
//  4  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  DHANDLE		-			Handle of the decompressed block
//							NULLHANDLE if the block could not be decompressed
//
//  API Use Notes:
//
//   If the decompressed object will end up larger than the expected size then the decompression is
//   treated as if it failed.
//
DHANDLE Attachments::DecompressBlockLZ1(DHANDLE hBIn, DWORD *dwSize, DWORD UFSize, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DHANDLE				hBOut = NULLHANDLE;					//  Output (decompressed) handle
	UCHAR				*pCBlock;							//  Pointer to the compressed memory block
	UCHAR				*pUCB;								//  Pointer to the uncompressed stream
	DWORD				BytesToProcess;						//  Number of bytes to process
	LZ1Header			*pCHdr;								//  LZ1 compressed chunk header
	DWORD				BytesDecompressed = 0;				//  Count of bytes in the decompressed stream
	DWORD				BytesDone;							//  Decompressed bytes from this chunk
	LZ1Stats			sDecompress;						//  Statistics
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer
	
	//  Lock the input (compressed) buffer
	pCBlock = OSLock(UCHAR, hBIn);
	BytesToProcess = *dwSize;

	//  Clear the stats
	memset(&sDecompress, 0, sizeof(LZ1Stats));

	//  Loop processing the compressed stream until it is exhausted
	while (BytesToProcess > sizeof(LZ1Header) && BytesDecompressed < UFSize)
	{
		pCHdr = (LZ1Header *) pCBlock;

		//  Resize or allocate the decompressed object
		if (hBOut == NULLHANDLE) stAPIRC = OSMemAlloc(MEM_GROWABLE, pCHdr->UCDLen + LZ1_SAFETY_BUFFER, &hBOut);
		else stAPIRC = OSMemRealloc(hBOut, BytesDecompressed + pCHdr->UCDLen + LZ1_SAFETY_BUFFER);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5520E, BytesDecompressed + pCHdr->UCDLen + LZ1_SAFETY_BUFFER);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			if (hBOut != NULLHANDLE) OSMemFree(hBOut);
			OSUnlock(hBIn);
			return NULLHANDLE;
		}

		//  Lock the uncompressed stream
		pUCB = OSLock(UCHAR, hBOut);

		//  Decompress the next chunk of the compressed stream
		sDecompress.Chunks++;
		BytesDone = decompressLZ1Chunk(pCBlock, pUCB + BytesDecompressed, &sDecompress, iThreadID);
		OSUnlock(hBOut);

		//  Check that the bytes added to the stream is what was expected
		if (BytesDone != pCHdr->UCDLen)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5527E, sDecompress.Chunks, BytesDone, pCHdr->UCDLen);
			xeLocal->LogMessage(szMsg, iThreadID);
			OSMemFree(hBOut);
			OSUnlock(hBIn);
			return NULLHANDLE;
		}

		//  Accumulate the decompressed output
		BytesDecompressed += BytesDone;

		//  Accumulate the compressed stream length consumed
		BytesToProcess -= pCHdr->CDLen;
		BytesToProcess -= sizeof(LZ1Header);

		//  If the previous block was the last in the stream and we have not decompressed the expected size
		//  we don't trust the output
		if ((pCHdr->Chunks == 5 || pCHdr->Chunks == 6)  && BytesDecompressed != UFSize)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5528E, sDecompress.Chunks, BytesDecompressed, UFSize);
			xeLocal->LogMessage(szMsg, iThreadID);
			OSMemFree(hBOut);
			OSUnlock(hBIn);
			return NULLHANDLE;
		}

		if (pCHdr->Chunks != 5 && pCHdr->Chunks != 6)
		{
			//  Point to the next chunk of input
			pCBlock += (pCHdr->CDLen + sizeof(LZ1Header));
		}
	}

	//  Check the terminal conditions
	if (pCHdr->Chunks != 5 && pCHdr->Chunks != 6)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5529E, sDecompress.Chunks);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemFree(hBOut);
		OSUnlock(hBIn);
		return NULLHANDLE;
	}

	if (BytesDecompressed != UFSize)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5521E, BytesDecompressed, UFSize);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemFree(hBOut);
		OSUnlock(hBIn);
		return NULLHANDLE;
	}

	//  Unlock the compressed stream and return the decompressed block
	OSUnlock(hBIn);
	*dwSize = BytesDecompressed;	
	return hBOut;
}

//  CompressLZ1Chunk
//
//  This function will compress the specified LZ1 uncompressed chunk 
//
//  Parameters
//
//	1  -  UCHAR *			-  Pointer to the buffer containing the uncompressed stream
//	2  -  DWORD				-  Length of the buffer to be compressed
//  3  -  UCHAR *			-  Pointer to the buffer to place the compressed stream
//  4  -  LZ1Stats *		-  Pointer to the LZ1 statistics structure
//  5  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  DWORD		-			Size of the compressed stream
//
//  API Use Notes:
//
DWORD Attachments::compressLZ1Chunk(UCHAR *pUCB, DWORD Size, UCHAR *pCB, LZ1Stats *pStats, int iThreadID)
{
	DWORD				CompSize = 0;						//  Compressed buffer size
	DWORD				BytesCompressed = 0;				//  Counter of bytes compressed
	DWORD				BitsWritten = 0;					//  Bits written to the compressed stream
	DWORD				Token = 0;							//  Current token number
	DWORD				MaxSrc,MaxTgt;						//  Maximum length of source and target strings
	DWORD				RepeatLen;							//  Length of repeated characters
	DWORD				iIndex;								//  Generic index
	DWORD				StringLen;							//  Length of String
	DWORD				StringStart;						//  Starting position of the string
	DWORD				TestString;							//  Scanning test string position
	DWORD				IChars,IBits;						//  Save input and output offsets
	DWORD				StringNo = 0;						//  String number
	DWORD				Offset;								//  Offset 
	LZ1Header			*pCHdr;								//  Pointer to the LZ1 header
	UCHAR				SOffset,SXOffset,SXXOffset;			//  Offset bits
	UCHAR				SLen,SXLen,SXXLen;					//  Length bits
	BOOL				Scanning;							//  Scan control
	DWORD				CCSLen = 0;							//  Counter check compressed stream length
	MSBitStream			bsBase;								//  Bit streaming object
	LZ1Stats			sSave;								//  Stats save area

	//  Save the current values of the stats in case we decide to use an uncompressed chunk
	memcpy(&sSave, pStats, sizeof(LZ1Stats));

	//  Setup the header
	pCHdr = (LZ1Header *) pCB;
	pCHdr->CDLen = 0;
	pCHdr->UCDLen = Size;
	pCHdr->Chunks = 1;

	//  Setup the bit stream over the compressed buffer
	bsBase.setOutputBuffer(pCB + sizeof(LZ1Header), LZ1_COMPRESS_CHUNK + LZ1_SAFETY_BUFFER);
	BytesCompressed = 0;

	//  Continue compressing data until the uncompressed data is exhausted or
	//  there is no point in going on because we have already inflated the chunk
	while(BytesCompressed < Size && (((BitsWritten / 8) + 1) <= Size))
	{
		//  Compute the bounds for the next token
		MaxTgt = Size - BytesCompressed;
		MaxSrc = BytesCompressed;

		//  Compute the count of repeats of the last character compressed
		RepeatLen = 0;
		if (BytesCompressed > 0)
		{
			iIndex = BytesCompressed;
			while (pUCB[iIndex] == pUCB[iIndex - 1] && iIndex < (Size - BytesCompressed))
			{
				RepeatLen++;
				iIndex++;
				if (RepeatLen == LZ1_MAX_STRING) break;
			}
		}
		
		//  Ignore a single character repeat
		if (RepeatLen < 2) RepeatLen = 0;
		if (RepeatLen > LZ1_MAX_STRING) RepeatLen = LZ1_MAX_STRING;

		//  Search for the longest matching string in the portion of the buffer that is already compressed
		//  First check that we can beat the repeat count
		StringLen = 0;
		if (MaxSrc > RepeatLen && MaxTgt > RepeatLen)
		{
			//  Setup for the scan - the distance back in the buffer that can be searched is restricted by the limit for encoding a very far string reference
			TestString = 0;
			if (BytesCompressed > LZ1_MAX_WINDOW) TestString = BytesCompressed - LZ1_MAX_WINDOW;
			Scanning = TRUE;

			//  Scan for possible matched strings
			while(Scanning)
			{
				//  Find the next match on the first character of the string
				while(pUCB[TestString] != pUCB[BytesCompressed] && TestString < (MaxSrc - StringLen)) TestString++;

				//  Determine if the possible start of a string was found
				if (pUCB[TestString] == pUCB[BytesCompressed])
				{
					iIndex = 1;
					while(pUCB[TestString + iIndex] == pUCB[BytesCompressed + iIndex] && iIndex < (MaxSrc - TestString) && iIndex < MaxTgt) iIndex++;

					//  Determine if this is the best length string found so far
					if (iIndex > 1 && iIndex >= StringLen)
					{
						StringStart = TestString;
						StringLen = iIndex;
					}
					TestString++;
					if (TestString > (MaxSrc - StringLen)) Scanning = FALSE;
					if (iIndex == LZ1_MAX_STRING) Scanning = FALSE;
				}
				else Scanning = FALSE;
			}
		}
		if (StringLen < 2) StringLen = 0;

		//  Prepare to write the next token to the compressed stream
		Token++;
		pStats->Tokens++;
		if (RepeatLen > LZ1_MAX_STRING) RepeatLen = LZ1_MAX_STRING;
		if (StringLen > LZ1_MAX_STRING) StringLen = LZ1_MAX_STRING;

		//  Compute the offset in the complete uncompressed buffer
		IChars = ((pStats->Chunks - 1) * LZ1_COMPRESS_CHUNK) + BytesCompressed;
		IBits = BitsWritten;

		//  Write the next token to the compressed data stream

		//  Determine the token to generate next
		if (RepeatLen == 0 && StringLen == 0)
		{
			//  Write character token to the compressed stream
			bsBase.next(0, 1);																	//  Character token
			bsBase.next(pUCB[BytesCompressed], 8);												//  Output the character
			BitsWritten += 9;
			BytesCompressed++;
			pStats->Chars++;
			CCSLen++;
		}
		else
		{
			StringNo++;
			if (RepeatLen > StringLen)
			{
				//  Write a repeat - which is always a near string
				bsBase.next(1, 1);																//  String token
				bsBase.next(0, 1);																//  Near string
				bsBase.next(0, 7);																//  Offset is -1
				BitsWritten += 9;

				//  Write the length
				RepeatLen -= 2;
				if (RepeatLen <= 7)
				{
					//  Single length field
					SLen = BYTE(RepeatLen);
					bsBase.next(SLen, 4);
					BitsWritten += 4;
					BytesCompressed += RepeatLen + 2;
					pStats->SStrings++;
					CCSLen += SLen + 2;
				}
				else
				{
					if (RepeatLen <= 71)
					{
						//  Double length field
						SLen = UCHAR((((RepeatLen - 8) & 0x00030) >> 4) | 0x0008);
						SXLen = UCHAR((RepeatLen - 8) & 0x000F);
						bsBase.next(SLen, 4);
						bsBase.next(SXLen, 4);
						BitsWritten += 8;
						BytesCompressed += RepeatLen + 2;
						pStats->MStrings++;
						CCSLen += (((SLen & 0x03) << 4) | SXLen) + 10;
					}
					else
					{
						//  Triple length field
						SLen = UCHAR((((RepeatLen - 72) & 0x0300) >> 8) | 0x000C);
						SXLen = UCHAR(((RepeatLen - 72) & 0x00F0) >> 4);
						SXXLen = UCHAR((RepeatLen - 72) & 0x000F);
						bsBase.next(SLen, 4);
						bsBase.next(SXLen, 4);
						bsBase.next(SXXLen, 4);
						BitsWritten += 12;
						BytesCompressed += RepeatLen + 2;
						pStats->LStrings++;
						CCSLen += (((SLen & 0x03) << 8) | (SXLen << 4) | SXXLen) + 74;
					}
				}
				pStats->Repeats++;
			}
			else
			{
				//  Write a Near/Far/Very Far string
				//  Compute the offset
				Offset = BytesCompressed - (StringStart + 1);
				if (Offset <= LZ1_MAX_NEAR)
				{
					//  Near string
					bsBase.next(1, 1);																//  String token
					bsBase.next(0, 1);																//  Near string
					SOffset = UCHAR(Offset);
					bsBase.next(SOffset, 7);														//  Write the offset
					BitsWritten += 9;
					pStats->NStrings++;
				}
				else
				{
					if (Offset <= LZ1_MAX_FAR)														//  0x077f max
					{
						//  Far string 
						bsBase.next(1, 1);															//  String Token
						bsBase.next(1, 1);															//  Far String
						Offset -= (LZ1_MAX_NEAR + 1);
						SOffset = UCHAR((Offset & 0x0FF0) >> 4);
						SXOffset = UCHAR(Offset & 0x000F);
						bsBase.next(SOffset, 7);													//  Write the offset
						bsBase.next(SXOffset, 4);													//  Write the offset
						BitsWritten += 13;
						pStats->FStrings++;
					}
					else
					{
						//  Very Far String
						bsBase.next(1, 1);															//  String token
						bsBase.next(1, 1);															//  Very Far String
						bsBase.next(1, 1);															//  Very Far String
						bsBase.next(1, 1);															//  Very Far String
						bsBase.next(1, 1);															//  Very Far String
						Offset -= (LZ1_MAX_FAR + 1);
						SOffset = UCHAR((Offset & 0x0F00) >> 8);
						SXOffset = UCHAR((Offset & 0x00F0) >> 4);
						SXXOffset = UCHAR(Offset & 0x000F);
						bsBase.next(SOffset, 4);													//  Write the offset
						bsBase.next(SXOffset, 4);													//  Write the offset
						bsBase.next(SXXOffset, 4);													//  Write the offset
						BitsWritten += 17;
						pStats->VStrings++;
					}
				}

				//  Write the length
				StringLen -= 2;
				if (StringLen <= 7)
				{
					//  Single length field
					SLen = UCHAR(StringLen);
					bsBase.next(SLen, 4);
					BitsWritten += 4;
					BytesCompressed += StringLen + 2;
					pStats->SStrings++;
					CCSLen += SLen + 2;
				}
				else
				{
					if (StringLen <= 71)
					{
						//  Double length field
						SLen = UCHAR((((StringLen - 8) & 0x0030) >> 4) | 0x0008);
						SXLen = UCHAR((StringLen - 8) & 0x000F);
						bsBase.next(SLen, 4);
						bsBase.next(SXLen, 4);
						BitsWritten += 8;
						BytesCompressed += StringLen + 2;
						pStats->MStrings++;
						CCSLen += (((SLen & 0x03) << 4) | SXLen) + 10;
					}
					else
					{
						//  Triple length field
						SLen = UCHAR((((StringLen - 72) & 0x0300) >> 8) | 0x000C);
						SXLen = UCHAR(((StringLen - 72) & 0x00F0) >> 4);
						SXXLen = UCHAR((StringLen - 72) & 0x000F);
						bsBase.next(SLen, 4);
						bsBase.next(SXLen, 4);
						bsBase.next(SXXLen, 4);
						BitsWritten += 12;
						BytesCompressed += StringLen + 2;
						pStats->LStrings++;
						CCSLen += (((SLen & 0x03) << 8) | (SXLen << 4) | SXXLen) + 74;
					}
				}
			}
		}
	}

	//  Write the trailer
	bsBase.next(0x000F, 4);
	bsBase.next(0x000F, 4);
	bsBase.next(0x000F, 4);
	bsBase.next(0x000F, 4);
	bsBase.next(1, 1);
	bsBase.next(0, 4);
	BitsWritten += 21;
	bsBase.flush();


	//  Update the LZ1 header
	CompSize = BitsWritten / 8;
	if (BitsWritten % 8) CompSize++;
	pCHdr->CDLen = CompSize;
	CompSize += sizeof(LZ1Header);

	//  Check that we have not inflated the stream - if so then change the returned buffer to an uncompressed form
	if (CompSize > Size)
	{
		CompSize = Size;
		memcpy(pCB + sizeof(LZ1Header), pUCB, CompSize);
		pCHdr->CDLen = CompSize;
		CompSize += sizeof(LZ1Header);
		pCHdr->Chunks = 2;

		//  Reset the stats
		memcpy(pStats, &sSave, sizeof(LZ1Stats));
		pStats->UCChunks++;
	}

	//  Return to caller
	return CompSize;
}

//  DecompressLZ1Chunk
//
//  This function will decompress the specified LZ1 compressed chunk 
//
//  Parameters
//
//  1  -  UCHAR *			-  Pointer to the header of the chunk to be decompressed
//	2  -  UCHAR *			-  Pointer to the buffer to place the decompressed stream
//  3  -  LZ1Stats *		-  Pointer to the LZ1 statistics structure
//  4  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  DWORD		-			Size of the decompressed stream
//
//  API Use Notes:
//
DWORD Attachments::decompressLZ1Chunk(UCHAR *pCB, UCHAR *pUCB, LZ1Stats *pStats, int iThreadID)
{
	MSBitStream			bsBase;															//  Bit streaming object
	DWORD				BitsRead = 0;													//  Number of bits read from the compressed stream
	DWORD				BytesWritten = 0;												//  Number of bytes written to the decompressed stream
	BOOL				EndOfStream = FALSE;											//  Logical end of the compressed stream
	UCHAR				Character;														//  Character read from the stream
	UCHAR				Nibble1,Nibble2,Nibble3,Nibble4,Nibble5,Nibble6,Nibble7;		//  Nibbles that make up a token
	DWORD				TokenSize;														//  Token size (used for variable length tokens)
	DWORD				Offset;															//  Source offset for a string
	DWORD				StringLength;													//  Length of a source string
	DWORD				iIndex;															//  Generic Index
	LZ1Header			*pCHdr;															//  Pointer to the chunk header
	char				szMsg[MAX_MSG + 1];												//  Generic Message Buffer

	//  Setup the bitstream by pointing to the start of the compressed data stream
	pCHdr = (LZ1Header *) pCB;
	bsBase.setInputBuffer(pCB + sizeof(LZ1Header), pCHdr->CDLen);

	//  If the chunk encoding is (2) - Chunk Not Compressed then just copy the shunk to the output
	if (pCHdr->Chunks == 2 || pCHdr->Chunks == 6)
	{
		memcpy(pUCB, pCB + sizeof(LZ1Header), pCHdr->CDLen);
		pStats->UCChunks++;
		return pCHdr->CDLen;
	}

	//  Process the compressed stream until the end - there are insufficient bits for the smallest token and the end of stream has not been reached
	while (((BitsRead + LZ1_SMALLEST_TOKEN) < (pCHdr->CDLen * 8)) && !EndOfStream)
	{
		pStats->Tokens++;

		//  Read the leading bit of the next token - If set then the token is a string otherwise it is a character
		if (bsBase.next(1))
		{
			//  Processing a string token
			//  Read the leading nibble from the token - this will determine how many nibbles make up the offset
			Nibble1 = UCHAR(bsBase.next(4));
			Nibble2 = UCHAR(bsBase.next(4));
			TokenSize = 9;

			//  Determine if the offset is for a Far or Very Far String
			if (Nibble1 & 0x08)
			{
				//  Read the next Nibble of the offset
				Nibble3 = UCHAR(bsBase.next(4));
				TokenSize += 4;

				//  Determine if this is a very far string
				if ((Nibble1 & 0x0F) == 0x0F)
				{
					//  Very Far string - compute the relative offset in the buffer
					Nibble4 = UCHAR(bsBase.next(4));
					TokenSize += 4;
					Offset = ((Nibble2 << 8) | (Nibble3 << 4) | Nibble4) + (LZ1_MAX_FAR + 2);

					//  Detect a possible END-OF-STREAM
					if ((Nibble1 & Nibble2 & Nibble3 & Nibble4) == 0x0F) EndOfStream = TRUE;

					pStats->VStrings++;
				}
				else
				{
					//  Far String - compute the relative offset in the buffer
					Offset = (((Nibble1 & 0x07) << 8) | (Nibble2 << 4) | Nibble3) + (LZ1_MAX_NEAR + 2);
					pStats->FStrings++;
				}
			}
			else
			{
				//  Near String - compute the relative offset in the buffer
				Offset = ((Nibble1 << 4) | Nibble2) + 1;
				
				//  Determine if this is a repeat or a near string
				if (Offset == 1) pStats->Repeats++;
				else pStats->NStrings++;
			}

			//  Convert the offset from relative to the current position to absolute from the start of the buffer
			Offset = BytesWritten - Offset;

			//  Read the length from the compressed stream
			Nibble5 = UCHAR(bsBase.next(4));
			TokenSize += 4;

			//  Determine if this is a medium or long string
			if (Nibble5 & 0x08)
			{
				//  Read the next nibble of the length
				Nibble6 = UCHAR(bsBase.next(4));
				TokenSize += 4;

				//  Determine if this is a long string
				if (Nibble5 & 0x04)
				{
					Nibble7 = UCHAR(bsBase.next(4));
					TokenSize += 4;

					//  Long string - compute the length
					StringLength = (((Nibble5 & 0x03) << 8) | (Nibble6 << 4) | Nibble7) + 74;
					pStats->LStrings++;
				}
				else
				{
					//  Medium string - compute the length
					StringLength = (((Nibble5 & 0x03) << 4) | Nibble6) + 10;
					pStats->MStrings++;
				}
			}
			else
			{
				//  Short string
				StringLength = Nibble5 + 2;
				pStats->SStrings++;
			}

			//  If End-Of-Stream was detected then adjust the counts
			if (EndOfStream)
			{
				pStats->Tokens--;
				pStats->VStrings--;
				BitsRead += TokenSize;
			}

			//  If we have not detected END-OF-STREAM
			if (!EndOfStream)
			{
				//  Add the string to the buffer - detect overlapping strings and do a bytewise copy
				if ((Offset + StringLength) <= BytesWritten) memcpy(pUCB + BytesWritten, pUCB + Offset, StringLength);
				else
				{
					//  Overlapping copy - this must be accomplished bytewise
					for (iIndex = 0; iIndex < StringLength; iIndex++)
						pUCB[BytesWritten + iIndex] = pUCB[Offset + iIndex];
				}

				//  Update the counts
				BytesWritten += StringLength;
				BitsRead += TokenSize;
			}
		}
		else
		{
			//  Processing a character token
			pStats->Chars++;

			//  Read the character
			Character = UCHAR(bsBase.next(8));

			//  Add the character to the decompressed stream
			pUCB[BytesWritten] = Character;
			BytesWritten++;
			BitsRead += 9;
		}

		//  Detect decompression data overruns
		if (BytesWritten > pCHdr->UCDLen)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5530E, pStats->Chunks, pStats->Tokens, BytesWritten, pCHdr->UCDLen, BitsRead);
			xeLocal->LogMessage(szMsg, iThreadID);
			EndOfStream = TRUE;
		}
	}

	//  Consider how the termination happened - warn of any inconsistency
	if (!EndOfStream)
	{
		xeLocal->LogMessage(MSG_DXR5531W, iThreadID);
	}

	//  Return to caller
	return BytesWritten;
}

//  compressBufferHuffman
//
//  This function will compress the passed buffer (Huffman endoced) using the passed encoding table
//
//  Parameters
//
//		1  -  UCHAR *				-  Pointer to the buffer to be compressed
//		2  -  DWORD					-  Size of the decompressed attachment content stream
//		3  -  UCHAR *				-  Pointer to the buffer to place the compression stream
//		4  -  DWORD					-  Size available in the stream
//		5  -  Encode[]				-  Pointer to the encoding table
//		6  -  ExecEnvironment *		-  Pointer to the current Runtime
//
//  Returns
//
//		DWORD		-	Size of the compressed object in bytes
//
//  API Use Notes:
//
//
DWORD Attachments::compressBufferHuffman(UCHAR *pUCB, DWORD InSize, UCHAR *pCB, DWORD OutSize, Encode ELT[], int iThreadID)
{
	DWORD				iIndex;											//  Generic index
	HuffHeader			*pHHdr;											//  Pointer to the Huffman Header
	DWORD				BitsWritten = 0;								//  Count of bits written
	DWORD				BytesWritten;									//  Count of bytes written
	MSBitStream			bsBase;											//  Bit Stream object

	//  Point to the header
	pHHdr = (HuffHeader *) pCB;

	//  Set the output buffer address
	bsBase.setOutputBuffer(pCB + sizeof(DWORD) + sizeof(WORD) + pHHdr->EHTreeSize, OutSize);

	//  Loop encoding each character in the input buffer
	for (iIndex = 0; iIndex < InSize; iIndex++)
	{
		bsBase.next(ELT[pUCB[iIndex]].EncVal, ELT[pUCB[iIndex]].EncLen);
		BitsWritten += ELT[pUCB[iIndex]].EncLen;
	}

	//  Write a logical end-of-stream marker 
	bsBase.next(ELT[HUFF_EOS].EncVal, ELT[HUFF_EOS].EncLen);
	BitsWritten += ELT[HUFF_EOS].EncLen;

	//  Flush the output buffer
	bsBase.flush();

	//  Compute the length of the compressed object
	BytesWritten = BitsWritten / 8;
	if (BitsWritten & 0x00000007) BytesWritten++;
	BytesWritten += sizeof(DWORD);
	BytesWritten += sizeof(WORD);
	BytesWritten += (pHHdr->EHTreeSize);
	return BytesWritten;
}

//  getNextHDChar
//
//  This function will return the next decompressed character from a Huffman compressed data stream
//
//  Parameters
//
//		1  -  MSBitStream *			-  Pointer to compressed Bit Stream
//		2  -  HuffHeader *			-  Pointer to the object header containing the Huffman Tree
//		3  -  DWORD *				-  Pointer to the dword to accumulate the number of bits used
//
//  Returns
//
//		WORD		-	The next decompressed character
//
//  API Use Notes:
//
//
WORD Attachments::getNextHDChar(MSBitStream *bsBase, HuffHeader *pHH, DWORD *pBits)
{
	WORD			Bits = 1;											//  Bits used
	WORD			Node = 0;											//  Current node

	//  Loop until the character is decompressed
	while (TRUE)
	{
		if (bsBase->next(1))
		{
			//  Left hand branch - check if this is a leaf if so return the decompressed character
			if (pHH->EHTable[Node].NLeft >= 0)
			{
				*pBits += Bits;
				return WORD(pHH->EHTable[Node].NLeft);
			}
			else
			{
				//  Advance to the next node
				Node = pHH->EHTable[Node].NLeft * -1;
				Bits++;
			}
		}
		else
		{
			//  Right hand branch - check if this is a leaf if so return the decompressed character
			if (pHH->EHTable[Node].NRight >= 0)
			{
				*pBits += Bits;
				return WORD(pHH->EHTable[Node].NRight);
			}
			else
			{
				//  Advance to the next node
				Node = pHH->EHTable[Node].NRight * -1;
				Bits++;
			}
		}
	}
}

//  createHuffmanNode
//
//  Create a new Huffman node
//
//  PARAMETERS
//
//		DWORD				- Frequency for this node
//		WORD				- Character value for this node
//		HuffNode *			- Pointer to the node below to the left
//		HuffNode *			- Pointer to the node below to the right
//		WORD				- ID to be assigned to the node
//
//  RETURNS
//
//		HuffNode *			- Pointer to new node
//
//  NOTES
//
HuffNode * Attachments::createHuffmanNode(DWORD Freq, WORD CValue, HuffNode *hnL, HuffNode *hnR, WORD Id)
{
	HuffNode			*pNode = NULL;																			//  Pointer to the node

	//  Allocate the node
	pNode = (HuffNode *) malloc(sizeof(HuffNode));
	if (pNode != NULL)
	{
		pNode->EncLen = 0;
		pNode->EncVal = 0;
		pNode->NID = Id;
		if (Freq == 0)
		{
			//  Intermediate (branch) node
			pNode->CValue = 0;
			pNode->nLeft = hnL;
			pNode->nRight = hnR;
			pNode->Freq = pNode->nLeft->Freq + pNode->nRight->Freq;
		}
		else
		{
			//  terminal (leaf) node
			pNode->CValue = CValue;
			pNode->Freq = Freq;
			pNode->nLeft = NULL;
			pNode->nRight = NULL;
		}
	}
	return pNode;
}

//  enqueueNode
//
//  Adds the passed node to the priority queue of nodes
//
//  PARAMETERS
//
//		HuffNode *			- Pointer to the node to be added to the queue
//		DWORD				- Index for the end of the priority queue
//		HuffNode *[]		- Priority queue
//
//  RETURNS
//
//		DWORD				- Index of the new end of the priority queue
//
//  NOTES
//
DWORD Attachments::enqueueNode(HuffNode *pNode, DWORD QEnd, HuffNode *PQ[])
{
	DWORD		iIndex;															//  Generic index
	HuffNode	*pTemp;															//  Temp for swapping

	//  Add the entry to the end of the queue
	PQ[QEnd] = pNode;

	//  Shuffle the entry towards the head of the priority queue
	iIndex = QEnd;
	while (iIndex > 1)
	{
		if (PQ[iIndex]->Freq > PQ[iIndex - 1]->Freq)
		{
			pTemp = PQ[iIndex - 1];
			PQ[iIndex - 1] = PQ[iIndex];
			PQ[iIndex] = pTemp;
			iIndex--;
		}
		else break;
	}

	//  Return the new end of queue index
	return QEnd + 1;
}

//  generateEncodes
//
//  Generate the encoding bit strings
//
//  PARAMETERS
//
//		HuffNode *			- Root of the tree to be generated
//
//  RETURNS
//
//		DWORD		-	Count of branch nodes
//
//  NOTES
//
DWORD Attachments::generateEncodes(HuffNode *pNode)
{
	DWORD			BranchCount = 0;									//  Branch count

	//  Boundary condition
	if (pNode->nLeft == NULL && pNode->nRight == NULL) return BranchCount;

	//  Generate the encoding for the level below
	BranchCount++;
	pNode->nLeft->EncLen = pNode->EncLen + 1;
	pNode->nLeft->EncVal = (pNode->EncVal << 1) + 1;
	pNode->nRight->EncLen = pNode->EncLen + 1;
	pNode->nRight->EncVal = pNode->EncVal << 1;

	//  Explode the subordinate nodes
	BranchCount += generateEncodes(pNode->nLeft);
	BranchCount += generateEncodes(pNode->nRight);

	return BranchCount;
}

//  encodeHuffmanNode
//
//  Encode the passed Huffman Node 
//
//  PARAMETERS
//
//		HuffNode *			- Node of the tree to be encoded
//		SWORD				- Node number of the node to be encoded
//		HuffHeadr *			- Pointer to the Huffman Compressed Object Header
//
//  RETURNS
//
//		SWORD				- The last node allocated in the table
//
//  NOTES
//
SWORD Attachments::encodeHuffmanNode(HuffNode *pNode, SWORD NNO, HuffHeader *pHHdr)
{
	SWORD			HiNode;																	//  Highest used node

	//  Default the high node to the current node
	HiNode = NNO;

	//  Boundary condition - leaf node already encoded
	if (pNode->nLeft == NULL && pNode->nRight == NULL) return HiNode;

	//  Deal with the left hand branch of the tree first
	if (pNode->nLeft->nLeft == NULL)
	{
		//  Left hand branch is a leaf node encode the character bvalue that it is encoding for
		pHHdr->EHTable[NNO].NLeft = pNode->nLeft->CValue;
	}
	else
	{
		//  Encode the left hand branch of the tree
		pHHdr->EHTable[NNO].NLeft = ((HiNode + 1) * -1);
		HiNode = encodeHuffmanNode(pNode->nLeft, HiNode + 1, pHHdr);
	}

	//  Now deal with the right hand branch of the tree
	if (pNode->nRight->nRight == NULL)
	{
		//  Right hand branch is a leaf node encode the character bvalue that it is encoding for
		pHHdr->EHTable[NNO].NRight = pNode->nRight->CValue;
	}
	else
	{
		//  Encode the right hand branch of the tree
		pHHdr->EHTable[NNO].NRight = ((HiNode + 1) * -1);
		HiNode = encodeHuffmanNode(pNode->nRight, HiNode + 1, pHHdr);
	}

	//  Return the high used node
	return HiNode;
}

//  populateEncodingTable
//
//  Store the encoding strings in the encoding table
//
//  PARAMETERS
//
//		HuffNode *			- Root of the tree to be transcribed
//		Encode[]			- Pointer to the encoding table
//
//  RETURNS
//
//  NOTES
//
void Attachments::populateEncodingTable(HuffNode *pNode, Encode pET[])
{
	//  Leaf Node
	if (pNode->nLeft == NULL && pNode->nRight == NULL)
	{
		//  Add the entry to the encoding table
		pET[pNode->CValue].EncLen = pNode->EncLen;
		pET[pNode->CValue].EncVal = pNode->EncVal;
		return;
	}

	//  This is a branch node - process the levels below
	populateEncodingTable(pNode->nLeft, pET);
	populateEncodingTable(pNode->nRight, pET);
	return;
}

//  totalBits
//
//  Compute the total number of bits needed to encode the file
//
//  PARAMETERS
//
//		HuffNode *			- Root of the tree to be transcribed
//
//  RETURNS
//
//		DWORD				- Count of bits needed
//
//  NOTES
//
DWORD Attachments::totalBits(HuffNode *pNode)
{
	DWORD			SubTotal = 0;											//  Sub Total of bits

	//  Leaf Node
	if (pNode->nLeft == NULL && pNode->nRight == NULL) return pNode->EncLen * pNode->Freq;

	//  This is a branch node - process the levels below
	SubTotal += totalBits(pNode->nLeft);
	SubTotal += totalBits(pNode->nRight);
	return SubTotal;
}

//  suppressCompression
//
//  Test s file name to see if the file type indicates that it should not be compressed
//
//  PARAMETERS
//
//		char *			- Pointer to the file name
//
//  RETURNS
//
//		BOOL			- TRUE ==> suppress, FALSE ==> don't suppress
//
//  NOTES
//
BOOL Attachments::suppressCompression(char *szFile)
{
	char			*pType;													//  Pointer to the file type
	BOOL			SawDot = FALSE;											//  Saw a dot in the file name

	//  Scan to the end of the file name
	if (szFile == NULL) return FALSE;
	pType = szFile;
	while(*pType != '\0')
	{
		if (*pType == '.') SawDot = TRUE;
		pType++;
	}
	if (!SawDot) return FALSE;

	//  Scan back to the previous dot
	while(*pType != '.') pType--;

	//  Check against each of the proscribed types
	if (strcmpi(pType, ".zip") == 0) return TRUE;
	if (strcmpi(pType, ".gzip") == 0) return TRUE;
	if (strcmpi(pType, ".tar") == 0) return TRUE;
	if (strcmpi(pType, ".gz") == 0) return TRUE;
	if (strcmpi(pType, ".tgz") == 0) return TRUE;
	if (strcmpi(pType, ".arc") == 0) return TRUE;
	if (strcmpi(pType, ".cab") == 0) return TRUE;
	if (strcmpi(pType, ".jar") == 0) return TRUE;
	if (strcmpi(pType, ".zipx") == 0) return TRUE;
	if (strcmpi(pType, ".rar") == 0) return TRUE;

	return FALSE;
}


