#pragma once
//  Attachments
//
//  This class provides functions for handling Notes Attachments
//
//  Message Identifier:      DXR
//  Message Range:           5500 - 5599
//  Trace area assignment:   TRACE_ATT (31)
//  DX Version:				 3.14.0
//  Version:                 3.14.0  Build: 109
//
//  Version History:
//
//	3.13.0		20/12/2013		Incorporated into DX Core
//	3.14.0		28/01/2015		Fix a compression overflow
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2012 - 2015
//

//  Application Includes
#include "../ExecEnvironment.h"
#include "../Misc/Bitstreams.h"

//  Local Definitons

//  Attachment process owner ID
#define			ATT_PROC_OWNER		(this + 'A' + 't' + 't')

//  Special return values
#define			RETURN_NOT_RECOMPRESSED			0
#define			RETURN_RECOMPRESS_ERROR			MAXDWORD

//  Recompression methods
#define			RECOMPRESS_SINGLE				1						//  Single buffer recompression
#define			RECOMPRESS_PIPED				2						//  Pipeline recompression

//  Limits and constants for Huffman compression/decmpression
#define			HUFF_OR_LIMIT					4096					//  Size overrun limit
#define			HUFF_EOS						0x0100					//  Huffman End-Of-Stream marker

//  Limits and constants for LZ1 compression/decompression
#define		LZ1_COMPRESS_CHUNK		(1024 * 16)						//  16k compression chunks 
#define		LZ1_MAX_STRING			1025							//  Maximum size for an LZ1 
#define		LZ1_MAX_NEAR			127								//  Maximum offset for a near string reference
#define		LZ1_MAX_FAR				1919							//  Maximum offset for a far string reference
#define		LZ1_MAX_WINDOW			(LZ1_MAX_FAR + 0x0FFF)			//  Maximum search window
#define		LZ1_SMALLEST_TOKEN		9								//  Smallest LZ1 token size
#define		LZ1_SAFETY_BUFFER		2048							//  Saftey buffer for decompression

//  Structures
//  Encoded entry for a Huffman Tree Node
typedef struct EHN {
	SWORD			NLeft;								//  Left hand subordinate node
	SWORD			NRight;								//  Right hand subordinate node
} EHN;

//  Huffman compressed object header
typedef struct HuffHeader {
	DWORD			CheckSum;							//  uncompressed file checksum
	WORD			EHTreeSize;							//  Size of the encoded Huffman Tree
	EHN				EHTable[512];						//  Table of encoded Huffman Tree nodes
} HuffHeader;

//  A Node in the Huffman Tree
typedef struct HuffNode {
	HuffNode		*nLeft;								//  Node to the left
	HuffNode		*nRight;							//  Node to the right
	DWORD			Freq;								//  Frequency
	WORD			CValue;								//  Character value
	WORD			EncLen;								//  Encoded length
	DWORD			EncVal;								//  Encoded value
	WORD			NID;								//  Node ID
} HuffNode;

//  An entry in the encoding table
typedef struct Encode {
	WORD			EncLen;								//  Encodoing length
	DWORD			EncVal;								//  Encoding string
} Encode;

//  LZ1 compressed object header
typedef struct LZ1Header {
	DWORD			CDLen;								//  Compressed data length
	DWORD			UCDLen;								//  Uncompressed data length
	WORD			Chunks;								//  Chunking flags
} LZ1Header;

//  LZ1 compression/decompression statistics structure
typedef struct LZ1Stats {
	DWORD			Chunks;								//  Number of chunks read/written
	DWORD			UCChunks;							//  Number of uncompressed chunks
	DWORD			Tokens;								//  Number of tokens read/written
	DWORD			Chars;								//  Number of Character tokens read/written
	DWORD			Repeats;							//  Number of Repeat tokens read/written
	DWORD			NStrings;							//  Number of Near Strings read/written
	DWORD			FStrings;							//  Number of Far Strings read/written
	DWORD			VStrings;							//  Number of Very Far Strings read/written
	DWORD			SStrings;							//  Number of Short Strings read/written
	DWORD			MStrings;							//  Number of Medium Strings read/written
	DWORD			LStrings;							//  Number of Long Strings read/written
} LZ1Stats;

//  Trace Messages
#define			MSG_DXR5503T		"DXR5503T: Not recompressing attachment %i in Note: 0x%.4X it is already compressed to type %i."
#define			MSG_DXR5504T		"DXR5504T: Recompressing attachment %i in Note: 0x%.4X size: %i bytes from compression type: %i to %i using method %i."
#define			MSG_DXR5511T		"DXR5511T: Replaced Obj: 0x%.4X Comp: %i Size: %i in Note: 0x%.4X with Obj: 0x%.4X Comp: %i Size: %i."
#define			MSG_DXR5513T		"DXR5513T: Recompression of RRV: 0x%.4X inflated the size from %i to %i bytes."
#define			MSG_DXR5535T		"DXR5535T: Compressed bits: %i (%i bytes), Branch Nodes: %i."
#define			MSG_DXR5538T		"DXR5538T: Compressing chunk: %i, to add to %i bytes of compressed data, %i bytes already compressed."
#define			MSG_DXR5539T		"DXR5539T: Compressing chunk: %i, an additional %i bytes of data will be compressed."
#define			MSG_DXR5540T		"DXR5540T: Compressed chunk: %i, produced %i bytes of compressed data."

//  Information Messages
#define			MSG_DXR5534I		"DXR5534I: LZ1 compression inflated the stream from %i bytes to %i bytes."
#define			MSG_DXR5541I		"DXR5541I: Recompression of file '%s' has been suppressed because of the file type."

//  Warning Messages
#define			MSG_DXR5517W		"DXR5517W: Unknown compression type: %i cannot be decompressed."
#define			MSG_DXR5531W		"DXR5531W: The compression stream was depleted without seeing a valid end of stream marker."
#define			MSG_DXR5536W		"DXR5536W: Object (RRV: 0x%.4X in Note: 0x%.4X Dup ID: %i was not recompressed."

//  Error Messages
#define			MSG_DXR5500E		"DXR5500E: Attachment processing requested on item '%s' in Note: 0x%.4X which is type: %i and not an attachment."
#define			MSG_DXR5501E		"DXR5501E: Attachment processing requested on item '%s' in Note: 0x%.4X which only has a data length of %i."
#define			MSG_DXR5502E		"DXR5502E: Attachment processing requested on item '%s' in Note: 0x%.4X which is object type: %i and not an attachment."
#define			MSG_DXR5505E		"DXR5505E: Unable to get the old attachment object size Note: 0x%.4X Dup ID: %i."
#define			MSG_DXR5506E		"DXR5506E: Failed to recompress attachment Note: 0x%.4X Dup ID: %i."
#define			MSG_DXR5507E		"DXR5507E: Failed to remove the original object from the container Note: 0x%.4X Dup ID: %i."
#define			MSG_DXR5508E		"DXR5508E: Failed to delete recompressed object 0x%.4X during error recovery."
#define			MSG_DXR5509E		"DXR5509E: Failed to allocate %i bytes for the recompressed FILEOBJECT."
#define			MSG_DXR5510E		"DXR5510E: Failed to append the recompressed FILEOBJECT to the container Note: 0x%.4X."
#define			MSG_DXR5512E		"DXR5512E: Failed to load object RRV: 0x%.4X into memory."
#define			MSG_DXR5514E		"DXR5514E: Failed to allocate a new object (Size: %i) to hold the recompressed stream."
#define			MSG_DXR5515E		"DXR5515E: Failed to write the recompressed stream (Size: %i) to the new object."
#define			MSG_DXR5516E		"DXR5516E: Failed to remove object RRV: 0x%.4X from the database during error recovery."
#define			MSG_DXR5518E		"DXR5518E: The size of the uncompressed stream %i does not match the expected size %i."
#define			MSG_DXR5519E		"DXR5519E: Unknown compression type: %i object cannot be compressed."
#define			MSG_DXR5520E		"DXR5520E: Failed to (re)allocate %i bytes for a decompressed stream."
#define			MSG_DXR5521E		"DXR5521E: Size of the decompressed stream (%i bytes) does not match the expected size (%i bytes)."
#define			MSG_DXR5522E		"DXR5522E: End of the compressed data stream was not seen there are %i bits remaining in the stream."
#define			MSG_DXR5523E		"DXR5523E: Decompressed stream checksum: 0x%.4X does not match the expected checksum: 0x%.4X."
#define			MSG_DXR5524E		"DXR5524E: Attachment processing requested on item '%s' in Note: 0x%.4X which is encrypted."
#define			MSG_DXR5525E		"DXR5525E: Failed to allocate %i bytes for a chunk of a compressed stream."
#define			MSG_DXR5526E		"DXR5526E: Failed to allocate %i bytes for a chunk of a decompressed stream."
#define			MSG_DXR5527E		"DXR5527E: Chunk %i of compressed stream decompressed to %i bytes, %i was expected."
#define			MSG_DXR5528E		"DXR5528E: Last chunk (%i) has been processed with %i bytes decompressed instead of the expected %i bytes."
#define			MSG_DXR5529E		"DXR5529E: Decompression completed on chunk: %i which was not flagged as the last chunk, object may be truncated."
#define			MSG_DXR5530E		"DXR5530E: Decompression data overrun detected on chunk: %i token: %i, bytes written: %i, expected: %i, bits consumed: %i."
#define			MSG_DXR5532E		"DXR5532E: Failed to (re)allocate %i bytes for a compressed stream."
#define			MSG_DXR5533E		"DXR5533E: The LZ1 compression engine was unable to compress chunk: %i of the uncompressed stream."
#define			MSG_DXR5534E		"DXR5534E: Failed to (re)allocate %i bytes for a compressed stream."
#define			MSG_DXR5537E		"DXR5537E: Failed to reallocate object: 0x%.4X (Size: %i) to hold the recompressed stream."

//  Class Definition
class Attachments
{
public:
	//  Default Constructor
	Attachments();

	//  Normal Constructor
	Attachments(ExecEnvironment *xeParent, int iThreadID);

	//  Public Members

	//  Public Fumctions

	//  Recompress an attachment in the passed note - identify the attachment by BLOCKID
	DWORD		RecompressAttachment(NOTEHANDLE hnNote, BLOCKID Attachment, WORD CType, int iThreadID);
	//  Recompress an attachment in place in the passed note - identify the attachment by BLOCKID
	DWORD		RecompressAttachmentInPlace(NOTEHANDLE hnNote, BLOCKID Attachment, WORD CType, int iThreadID);
	//  Recompress an object - using a single memory buffer
	RRV			RecompressObjectSingle(DBHANDLE hdbContainer, FILEOBJECT *foIn, WORD CType, DWORD *dwSize, int iThreadID);
	//  Recompress an object in place - using a single memory buffer
	WORD		RecompressObjectSingleInPlace(DBHANDLE hdbContainer, FILEOBJECT *foIn, WORD CType, DWORD *dwSize, int iThreadID);
	//  Recompress an object - using a fixed size memory pipe
	RRV			RecompressObjectPiped(DBHANDLE hdbContainer, FILEOBJECT *foIn, WORD CType, DWORD *dwSize, int iThreadID);
	//  Recompress an object - using a fixed size memory pipe
	WORD		RecompressObjectPipedInPlace(DBHANDLE hdbContainer, FILEOBJECT *foIn, WORD CType, DWORD *dwSize, int iThreadID);
	//  Recompress an in memory object
	DHANDLE		RecompressBlock(DHANDLE hBIn, WORD CTypeIn, WORD CTypeOut, DWORD *dwSize, DWORD UFSize, int iThreadID);
	//  Decompress a memory block (Huffman compression scheme)
	DHANDLE		DecompressBlockHuffman(DHANDLE hBIn, DWORD *dwSize, DWORD UFSize, int iThreadID);
	//  Decompress a memory block (LZ1 compression scheme)
	DHANDLE		DecompressBlockLZ1(DHANDLE hBIn, DWORD *dwSize, DWORD UFSize, int iThreadID);
	//  Compress a memory block (Huffman compression scheme)
	DHANDLE		CompressBlockHuffman(DHANDLE hBIn, DWORD *dwSize, int iThreadID);
	//  Compress a memory block (LZ1 compression scheme)
	DHANDLE		CompressBlockLZ1(DHANDLE hBIn, DWORD *dwSize, int iThreadID);


private:
	//  Private Members
	ExecEnvironment			*xeLocal;										//  Multi-Threading Runtime

	//  Private Functions
	WORD		getNextHDChar(MSBitStream *bsBase, HuffHeader *pHH, DWORD *pBits);	//  Get the next decompressed character
	DWORD		decompressLZ1Chunk(UCHAR *pCB, UCHAR *pUCB, LZ1Stats *pStats, int iThreadID);  //  Decompress an LZ1 chunk
	DWORD		compressLZ1Chunk(UCHAR *pUCB, DWORD Size, UCHAR *pCB, LZ1Stats *pStats, int iThreadID);  //  Compress an LZ1 chunk
	DWORD		compressBufferHuffman(UCHAR *pUCB, DWORD InSize, UCHAR *pCB, DWORD OutSize, Encode ELT[], int iThreadID);  //  Compress a buffer (HUFFMAN)
	HuffNode	*createHuffmanNode(DWORD Freq, WORD CValue, HuffNode *hnL, HuffNode *hnR, WORD Id);	//  Create a new Huffman node
	DWORD		enqueueNode(HuffNode *pNode, DWORD QEnd, HuffNode *PQ[]);					//  Add node to the priority queue
	DWORD		generateEncodes(HuffNode *pNode);											//  Generate the encoding bit strings
	SWORD		encodeHuffmanNode(HuffNode *pNode, SWORD NNO, HuffHeader *pHHdr);			//  Encode a Huffman Node
	void		populateEncodingTable(HuffNode *pNode, Encode pET[]);						//  Populate the encoding table
	DWORD		totalBits(HuffNode *pNode);													//  Count total bits
	BOOL		suppressCompression(char *szFile);											//  Suppress compression test
};

