#pragma once
//  CopyRequest
//
//  This class contains the request data for a DbCopier to copy a single database
//
//  Message Identifier:      DXR
//  Message Range:           5001 - 5499
//  Trace area assignment:   TRACE_DBC (23)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  3.6.0		20/09/2009		Prepared for QCopy 2.0
//  3.7.5		30/03/2010		Improved Error Handling in DbCopier
//  3.7.7		10/04/2010		Additional recovery options in DbCopier
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.3		05/01/2011		Pre-Process steps added to DbCopier
//  3.10.9		10/06/2011		One-Time replication option added
//	3.13.0		20/12/2013		Advanced Target Database Properties
//	3.14.1		28/01/2015		Recompress attachments optional
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 2015
//

//  Platform Includes
#include  "../Platform/PlatBase.h"										//  OS and Language includes
#include  "../Platform/NotesBase.h"										//  Notes API includes
#include  "../DXGlobals.h"												//  DX Global definitions

//  Local Definitions

//  Pre-processing actions requested
#define			PRESTART_DEL_TARGET			1							//  Delete target database

//  Restart actions requested
#define			RESTART_DEL_TARGET			1							//  Delete target database
#define			RESTART_ASSERT_FIXUP		2							//  Force fixup of source

//  Fault tolerance settings
#define			ALLOW_FOLDER_ERRORS			1							//  Allow Folder Errors
#define			ALLOW_ACCESS_ERRORS			2							//  Allow Access Errors
#define			ALLOW_COPY_ERRORS			4							//  Allow Copy Errors
#define			RECOVER_RSRC_ERRORS			8							//	Recover from Resource Errors

class CopyRequest
{
public:
	//  Constructor
	CopyRequest(void);

	//  Destructor
	~CopyRequest(void);

	//  Public Members
	//  Control variables
	BOOL		PermitRun;												//  Run Permit Flag
	BOOL		bAllowProxyFolderMessages;								//  Permit proxy messaging for folders
	DBHANDLE	hdbSource;												//  Handle of Source Database
	DBID		dbidSource;												//  DBID of the Source Database
	DBHANDLE	hdbTarget;												//  Handle of the Target Database
	DBID		dbidTarget;												//  DBID of the Target database
	DHANDLE		hitFolderTable;											//  Handle of the Folder Note ID Table
	DHANDLE		hitDesignTable;											//  Handle of the Design Note ID Table

	//  Configuration variables
	int			ReturnCode;												//  Completion code from the copy
	int			RetryCount;												//  Retry iteration count
	int			RetryLimit;												//  Retry iteration limit
	int			RestartActions;											//  Restart Recovery actions
	int			FaultTolerance;											//  Fault Tolerance
	int			PreProcessActions;										//  Pre-Processing Actions
	int			ODSLevel;												//  Requested ODS Level
	int			LocalEncrypt;											//  Local encryption strength
	BOOL		AllowRetry;												//  Show if request is can be retried
	BOOL		CopyReplica;											//  Database copy is a replica
	BOOL		CopyStubs;												//  Copy Deletion Stubs
	BOOL		CopyPrivates;											//  Copy Private Design Elements
	BOOL		RefreshDesign;											//  Refresh the target database design
	BOOL		CopyFTI;												//  Copy FTI definition
	BOOL		CopyAndBuildFTI;										//  Copy FTI definition and build the Index
	BOOL		BuildIndexes;											//  Build view indexes
	BOOL		EnableTXLogging;										//  Enable Transaction Logging
	BOOL		ConsistentACLOff;										//  Turn off enforce consistent ACL
	BOOL		CopyFolders;											//  Maintain folders in non-replica copy
	BOOL		ForceFixup;												//  Force a fixup of the source database
	BOOL		DisableAgents;											//  Disable Agents
	BOOL		OneTimeReplicate;										//  Perform a one time replication
	BOOL		LZ1Compress;											//  LZ1 Compression
	BOOL		RecompressAttachments;									//  Recompress Attachments
	BOOL		EnableDAOS;												//  Enable DAOS on the target

	//  Statistics
	//  Section 1 -- stats from the copy phase
	DWORD		dwDataDocsCopied;										//  Data documents copied
	DWORD		dwDataStubsCopied;										//  Deletion stubs copied
	DWORD		dwDataAccessErrors;										//  Data access errors
	DWORD		dwDataOtherErrors;										//  Data other errors
	DWORD		dwDataRetries;											//  Data copy retries
	DWORD		dwDesignDocsCopied;										//  Design documents copied
	DWORD		dwDesignStubsCopied;									//  Deletion stubs copied
	DWORD		dwDesignAccessErrors;									//  Design access errors
	DWORD		dwDesignOtherErrors;									//  Design other errors
	DWORD		dwDesignRetries;										//  Design copy retries
	DWORD		dwCopyPhaseElapsed;										//  Copy phase elapsed time

	//  Section 2 -- stats for folder & FTI maintenance phase
	DWORD		dwFoldersCopied;										//  Count of folders copied
	DWORD		dwFolderDocs;											//  Documents in all folders
	DWORD		dwFolderPhaseElapsed;									//  Folder phase elapsed time

	//  Section 3 -- stats from design maintenance phase
	DWORD		dwDesignPhaseElapsed;									//  Design phase elapsed time

	//  Section 4 -- stats from agent phase
	DWORD		dwAgentPhaseElapsed;									//  Agent phase elapsed time

	//  Section 5 -- stats from index maintenance phase
	DWORD		dwIndexPhaseElapsed;									//  Index phase elapsed time

	//  Section 6 -- stats from the replication phase
	DWORD		dwReplPhaseElapsed;										//  Replication phase elapsed time

	//  Section 7 -- summary statistics
	DWORD		dwTotalElapsed;											//  Total elapsed time
	DWORD		dwSizeKB;												//  Database Size in KB

	//  Section 8 -- Attachment Recompression statistics
	DWORD		dwNotesRecom;											//  Number of notes recompressed
	DWORD		dwFilesRecom;											//  Number of attachments recompressed
	DWORD		dwKBSaved;												//  Number of Kb saved

	//  Configuration Strings
	char		szTransID[MAX_DATABASE + 1];							//  Request Transaction ID
	char		szSourceServer[MAX_SERVER + 1];							//  Source Server
	char		szSourceDatabase[MAX_DATABASE + 1];						//  Source Database or directory
	char		szSourceDBReplicaID[MAX_DESC + 1];						//  Source DB Replica ID
	char		szTargetServer[MAX_SERVER + 1];							//  Target Server
	char		szTargetDatabase[MAX_DATABASE + 1];						//  Target Database or directory
	char		szGroupToAdd[MAX_GROUP + 1];							//  Group to add the the ACL
	char		szTemplateName[MAX_TEMPLATE + 1];						//  Template name to apply to the target database
	char		szAgentName[MAX_ELEMENT + 1];							//  Agent to be run in the target database

};
