#pragma once
//  DbCopier
//
//  This class provides the Database Copy functions used in Domono eXplorer (DX) utilities
//
//  Message Identifier:      DXR
//  Message Range:           5001 - 5499
//  Trace area assignment:   TRACE_DBC (23)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  1.0.0		10/02/2001		Initial version
//  3.5.3		11/09/2009		Incorporated into DX core
//  3.6.0		20/09/2009		Prepared for QCopy 2.0
//  3.6.1		10/10/2009		Additional controls added
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.7.1		06/01/2010		Prevent a race condition in Rejoin processing
//  3.7.2		12/01/2010		Improve handling of corrupt folders
//  3.7.3		06/02/2010		DbMover & DbCopier improvements
//  3.7.5		30/03/2010		Improved Error Handling in DbCopier
//  3.7.6		08/04/2010		Fixed a problem with error handling
//  3.7.7		10/04/2010		Additional recovery options in DbCopier
//  3.9.0		24/07/2010		Linkage changed for logging & request posting
//								Improved error recovery
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.3		05/01/2011		Pre-Processing actions added
//  3.10.9		07/06/2011		One time replication added
//								Attitional error recovery
//  3.10.10		14/07/2011		Performance Improvements for/to DbCopier
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.6		17/12/2011		Elapsed time calculations changed to use ms
//	3.13.0		20/12/2013		Advanced Target Database Properties
//	3.14.0		19/01/2015		x64 compatability fixes
//	3.14.1		24/02/2015		Avoid a panic in DbCopier
//								Recompress Attachments optional
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2001 - 2015
//

//  Application Includes
#include "../MTX/MTExecutive.h"
#include "../Threads/Runnable.h"
#include "CopyRequest.h"
#include "PartCopyRequest.h"
#include "Attachments.h"

//  Additional Notes API includes
#include	<repl.h>												//  Replication Services

//  Local Definitons
#define			NET_RETRY_LIMIT		7								//  Limit to retry timed out network operations
#define			DESIGN_NOTES		"design notes"
#define			DATA_NOTES			"data notes"
#define			TEXT_OK				"completed normally"
#define			TEXT_WARN			"completed with warnings"
#define			TEXT_FAILED			"failed"
#define			DESIGN_ACC_MUTEX	APP_MUTEX_9						//  Design accumulator mutex
#define			DATA_ACC_MUTEX		APP_MUTEX_10					//  Data accumulator mutes
#define			REJOIN_WAITS_MAX	1800							//  3 mins wait limit
#define			MAX_FOLDER_MSGS		25								//  Max folders for proxy messages
#define			SECRET_SQUIRREL_UPDATE_FLAG		0x00000008			//  Secret flag suppressing update tracking

//  Encryption strengths
#define			ENCRYPT_NONE		0								//  No encryption
#define			ENCRYPT_SIMPLE		1								//  Simple strength
#define			ENCRYPT_MEDIUM		2								//  Medium strength
#define			ENCRYPT_STRONG		3								//  Strong strength

//  ODS Levels
#define			ODS_DFLT			0								//  Default / None
#define			ODS_R6				60								//  R 6.0 - ODS: 43
#define			ODS_R8				80								//  R 8.0 - ODS: 48
#define			ODS_R85				85								//  R 8.5 - ODS: 51
#define			ODS_R9				90								//  R 9.0 - ODS: 52

//  Trace Messages
#define			MSG_DXR5001T		"DXR5001T: A new DbCopier object has been created @%p."
#define			MSG_DXR5111T		"DXR5111T: A DbCopier object @%p has been destroyed."
#define			MSG_DXR5112T		"DXR5112T: DbCopier has been entered for %i processing."
#define			MSG_DXR5113T		"DXR5113T: All ACL processing has been completed."
#define			MSG_DXR5114T		"DXR5114T: DbCopier is preparing to start copying."
#define			MSG_DXR5115T		"DXR5115T: A Copy Design request has been dispatched."
#define			MSG_DXR5116T		"DXR5116T: A Copy Data request has been dispatched."
#define			MSG_DXR5117T		"DXR5117T: Copy Design/Data state has changed, starting the rejoin loop."
#define			MSG_DXR5118T		"DXR5118T: Copy Design/Data rejoins have completed."
#define			MSG_DXR5119T		"DXR5119T: Starting folder maintenance."
#define			MSG_DXR5120T		"DXR5120T: A maintain folder request has been dispatched."
#define			MSG_DXR5121T		"DXR5121T: Starting Full Text Index maintenance."
#define			MSG_DXR5122T		"DXR5122T: A maintain FTI request has been dispatched."
#define			MSG_DXR5123T		"DXR5123T: Maintain Folders and FTI states have changed, starting the rejoin loop."
#define			MSG_DXR5124T		"DXR5124T: Starting design maintenance."
#define			MSG_DXR5125T		"DXR5125T: An upgrade design request has been dispatched."
#define			MSG_DXR5126T		"DXR5126T: An upgrade design request has completed."
#define			MSG_DXR5127T		"DXR5127T: Starting conditioning agent processing."
#define			MSG_DXR5128T		"DXR5128T: A run agent request has been dispatched."
#define			MSG_DXR5129T		"DXR5129T: A run agent request has completed."
#define			MSG_DXR5130T		"DXR5130T: Starting index maintenance processing."
#define			MSG_DXR5131T		"DXR5131T: Index maintenance processing has completed."
#define			MSG_DXR5172T		"DXR5172T: Restart recovery actions have been started (Restart Actions = %i)."
#define			MSG_DXR5173T		"DXR5173T: All restart recovery actions have been completed."
#define			MSG_DXR5177T		"DXR5177T: Copier dispatched with RQ: @%p, Function: %i, PRQ: @%p."
#define			MSG_DXR5187T		"DXR5187T: Starting one-time replication processing."
#define			MSG_DXR5189T		"DXR5189T: One-time replication processing has completed."
#define			MSG_DXR5214T		"DXR5214T: Performing attachment recompression on Note: 0x%.4X."
#define			MSG_DXR5218T		"DXR5218T: Attachment recompression completed saving: %i bytes (running total: %i bytes."
#define			MSG_DXR5219T		"DXR5219T: Note recompression completed saving: %i bytes (running total: %i kbytes."
#define			MSG_DXR5220T		"DXR5220T: Source database flags: '%s', DbOptions: 0x%.8X, 0x%.8X, 0x%.8X, 0x%.8X."
#define			MSG_DXR5221T		"DXR5221T: Target database flags: '%s', DbOptions: 0x%.8X, 0x%.8X, 0x%.8X, 0x%.8X."

//  Information Messages
#define			MSG_DXR5002I		"DXR5002I: Copying from Database: %s on: %s."
#define			MSG_DXR5003I		"DXR5003I: Copying to Database: %s on: %s."
#define			MSG_DXR5007I		"DXR5007I: The Database Copier has opened the source database."
#define			MSG_DXR5008I		"DXR5008I: The Database Copier has opened the target database."
#define			MSG_DXR5009I		"DXR5009I: The Database Copier has completed copying the source to the target database."
#define			MSG_DXR5011I		"DXR5011I: The Database Copier has initiated copying the Database design."
#define			MSG_DXR5012I		"DXR5012I: The Database Copier has initiated copying the Database documents."
#define			MSG_DXR5017I		"DXR5017I: %d profile documents have been added to the design collection."
#define			MSG_DXR5018I		"DXR5018I: Ready to copy %d design documents."
#define			MSG_DXR5019I		"DXR5019I: The source database contains no data notes."
#define			MSG_DXR5021I		"DXR5021I: Ready to copy %d data documents."
#define			MSG_DXR5024I		"DXR5024I: Folder note (0x%.4X) has been added to the folder maintenance table."
#define			MSG_DXR5036I		"DXR5036I: There are no folders to maintain in the copied database."
#define			MSG_DXR5037I		"DXR5037I: Starting folder maintenance on %d folders."
#define			MSG_DXR5038I		"DXR5038I: All folder maintenance actions have completed in %i seconds."
#define			MSG_DXR5040I		"DXR5040I: Deleted folder note (0x%.4X) has been skipped."
#define			MSG_DXR5041I		"DXR5041I: %d documents were added to %d folders."
#define			MSG_DXR5043I		"DXR5043I: Processing folder '%s' (0x%.4X)."
#define			MSG_DXR5056I		"DXR5056I: No response documents were removed from folder note (0x%.4X)."
#define			MSG_DXR5058I		"DXR5058I: %d response documents were removed from folder note (0x%.4X)."
#define			MSG_DXR5059I		"DXR5059I: %d entries were added to folder (0x%.4X) and %d entries were dropped."
#define			MSG_DXR5060I		"DXR5060I: Copying completed in %s%d seconds at a rate of %s%d documents per second."
#define			MSG_DXR5061I		"DXR5061I: Copying completed in %s%d seconds at a rate of %s%d Kb per second."
#define			MSG_DXR5064I		"DXR5064I: Replication has been re-enabled in the target database."
#define			MSG_DXR5068I		"DXR5068I: Group '%s' has been added to the target database ACL as Manager."
#define			MSG_DXR5071I		"DXR5071I: The 'enforce consistent ACL' setting has been turned off in the target database."
#define			MSG_DXR5075I		"DXR5075I: The design template has been set to '%s' in the target database."
#define			MSG_DXR5077I		"DXR5077I: The design of the target database has been refreshed from '%s'."
#define			MSG_DXR5083I		"DXR5083I: %d view indexes have been built in the target database."
#define			MSG_DXR5084I		"DXR5084I: The source database is not Full Text Indexed, no definitions copied."
#define			MSG_DXR5094I		"DXR5094I: Full Text Indexing has been enabled on the target database."
#define			MSG_DXR5100I		"DXR5100I: The following output was captured while running the requested agent."
#define			MSG_DXR5101I		"DXR5101I: >>%s"
#define			MSG_DXR5103I		"DXR5103I: Full Text Index was updated with %lu documents, %lu bytes were indexed."
#define			MSG_DXR5132I		"DXR5132I: Design: Docs copied %i, Stubs copied %i, Retries %i, Access errors %i, Other errors %i, elapsed %i."
#define			MSG_DXR5133I		"DXR5133I: Data: Docs copied %i, Stubs copied %i, Retries %i, Access errors %i, Other errors %i, elapsed %i."
#define			MSG_DXR5134I		"DXR5134I: Raw copying completed in %s%d seconds at a rate of %s%d documents per second."
#define			MSG_DXR5135I		"DXR5135I: Raw copying completed in %s%d seconds at a rate of %s%d Kb per second."
#define			MSG_DXR5136I		"DXR5136I: Design upgrade phase has completed in %i seconds."
#define			MSG_DXR5137I		"DXR5137I: Conditioning agent phase has completed in %i seconds."
#define			MSG_DXR5138I		"DXR5138I: Index maintenance phase has completed in %i seconds."
#define			MSG_DXR5140I		"DXR5140I: No Replica ID was supplied in the copy request, skipping verification."
#define			MSG_DXR5143I		"DXR5143I: All copy actions have completed final database size is %i Kb."
#define			MSG_DXR5144I		"DXR5144I: Background agents have been disabled in the target database."
#define			MSG_DXR5148I		"DXR5148I: Attempting to use an unvalidated Note ID table for folder (0x%.4X)."
#define			MSG_DXR5149I		"DXR5149I: Folder '%s' is damaged, attempting to re-create content."
#define			MSG_DXR5156I		"DXR5156I: Scanned %i Note IDs in the native table, %i deletion stubs were dropped."
#define			MSG_DXR5157I		"DXR5157I: %i invalid Note IDs dropped and %i could not be added to the validated ID table."
#define			MSG_DXR5158I		"DXR5158I: There were %i validated note IDs added to the validated table."
#define			MSG_DXR5163I		"DXR5163I: Purging folder has removed %i $FolderObjects."
#define			MSG_DXR5168I		"DXR5168I: This request is being retried, attempt number %i."
#define			MSG_DXR5174I		"DXR5174I: Database '%s' on '%s' has been deleted for restart recovery."
#define			MSG_DXR5175I		"DXR5175I: A Fixup has been requested on the source database for restart recovery."
#define			MSG_DXR5176I		"DXR5176I: Too many folders - proxy logging has been suppressed for further folder messages."
#define			MSG_DXR5178I		"DXR5178I: All pre-processing actions have been completed."
#define			MSG_DXR5180I		"DXR5180I: Database '%s' on '%s' has been deleted during pre-processing."
#define			MSG_DXR5184I		"DXR5184I: One-time replication of database '%s' with '%s' has completed."
#define			MSG_DXR5185I		"DXR5185I: One-time replication of D/B (pull) Added: %i Deleted: %i Updated: %i Errors: %i."
#define			MSG_DXR5186I		"DXR5186I: One-time replication of D/B (push) Added: %i Deleted: %i Updated: %i Errors: %i."
#define			MSG_DXR5188I		"DXR5188I: One-time replication processing has completed in %i seconds."
#define			MSG_DXR5192I		"DXR5192I: Replication has been disabled in the target database."
#define			MSG_DXR5193I		"DXR5193I: Replication has successfully recovered from previous errors, transaction is continuing."
#define			MSG_DXR5195I		"DXR5195I: Data copying will proceed using the NSFSearch() method."
#define			MSG_DXR5197I		"DXR5197I: Data copy multiplexor level has been set to %i (unscaled: %i), with %i available threads."
#define			MSG_DXR5203I		"DXR5203I: The source database does not contain a database icon or title."
#define			MSG_DXR5217I		"DXR5217I: Recompressed %i notes with %i attachments and saved %i Kb."

//  Warning Messages
#define			MSG_DXR5016W		"DXR5016W: Not all profile documents will be copied to the target database."
#define			MSG_DXR5026W		"DXR5026W: Copying note (0x%.4X) completed after %i retries."
#define			MSG_DXR5028W		"DXR5028W: Reading source note (0x%.4X) completed after %i retries."
#define			MSG_DXR5030W		"DXR5030W: Unable to generate a new OID for note (0x%.4X), using the old OID."
#define			MSG_DXR5032W		"DXR5032W: Writing a copy of source note (0x%.4X) completed after %i retries."
#define			MSG_DXR5039W		"DXR5039W: Errors were encountered during folder mainenance."
#define			MSG_DXR5045W		"DXR5045W: Unable to get folder contents from note (0x%.4X) in the source database."
#define			MSG_DXR5047W		"DXR5047W: Entry (0x%.4X) in folder (0x%.4X) points to an invalid note, entry dropped."
#define			MSG_DXR5048W		"DXR5048W: Entry (0x%.4X) in folder (0x%.4X) is not in the target database, entry dropped."
#define			MSG_DXR5049W		"DXR5049W: Entry (0x%.4X) in folder (0x%.4X) could not be added to the target folder."
#define			MSG_DXR5066W		"DXR5066W: Group '%s' already exists in the ACL, ACL was not updated."
#define			MSG_DXR5169W		"DXR5169W: Database '%s' on server %s' was not deleted for restart."
#define			MSG_DXR5179W		"DXR5179W: Database '%s' on server %s' was not deleted during pre-processing."
#define			MSG_DXR5198W		"DXR5198W: Unable to read the class from database '%s' on server %s'."
#define			MSG_DXR5200W		"DXR5200W: No database icon note could be read from the source database, using defaults."
#define			MSG_DXR5209W		"DXR5209W: Failed to dispatch a request to check for attachment recompression."
#define			MSG_DXR5212W		"DXR5212W: Failed to dispatch a request to recompress attachments on Note: 0x%.4X."

//  Error Messages
#define			MSG_DXR5004E		"DXR5004E: The Database Copier was unable to open the source database."
#define			MSG_DXR5005E		"DXR5005E: The Database Copier was unable to create the target replica."
#define			MSG_DXR5006E		"DXR5006E: The Database Copier was unable to create the target database."
#define			MSG_DXR5010E		"DXR5010E: Unable to create the Folder ID Table."
#define			MSG_DXR5013E		"DXR5013E: The source database contains no design notes."
#define			MSG_DXR5014E		"DXR5014E: Unable to get the modfied (design) note ID Table."
#define			MSG_DXR5015E		"DXR5015E: An error was encountered adding profiles to the design note ID Table."
#define			MSG_DXR5020E		"DXR5020E: Unable to get the modfied (data) note ID Table."
#define			MSG_DXR5022E		"DXR5022E: Unable to open the design note (0x%.4X) to determine attributes."
#define			MSG_DXR5023E		"DXR5023E: Unable to add the design note (0x%.4X) to the folder maintenance table."
#define			MSG_DXR5025E		"DXR5025E: Unable to copy note (0x%.4X) to the target replica."
#define			MSG_DXR5027E		"DXR5027E: Unable to read note (0x%.4X) from the source database."
#define			MSG_DXR5029E		"DXR5029E: Unable to copy note (0x%.4X) in memory."
#define			MSG_DXR5031E		"DXR5031E: Unable to write source note (0x%.4X) to the target database."
#define			MSG_DXR5042E		"DXR5042E: Unable to open folder note (0x%.4X) in the source database."
#define			MSG_DXR5044E		"DXR5044E: Unable to create an ID table for the target folder."
#define			MSG_DXR5046E		"DXR5046E: Unable to create an empty ID table for the source folder."
#define			MSG_DXR5050E		"DXR5050E: Unable to get the OID of folder note (0x%.4X) in the source database."
#define			MSG_DXR5051E		"DXR5051E: Unable to get the noteid of folder note (0x%.4X) in the target database."
#define			MSG_DXR5052E		"DXR5052E: Unable to add entries to the folder note (0x%.4X) in the target database."
#define			MSG_DXR5053E		"DXR5053E: Unable to create a new response document table."
#define			MSG_DXR5054E		"DXR5054E: Unable to open folder entry (0x%.4X) in the target database."
#define			MSG_DXR5055E		"DXR5055E: Unable to record a response folder entry (0x%.4X)."
#define			MSG_DXR5057E		"DXR5057E: Unable to remove response documents from folder note (0x%.4X)."
#define			MSG_DXR5062E		"DXR5062E: Unable to read the replication information from the target database."
#define			MSG_DXR5063E		"DXR5063E: Unable to set the replication information in the target database."
#define			MSG_DXR5065E		"DXR5065E: Unable to read the ACL from the target database."
#define			MSG_DXR5067E		"DXR5067E: An error was encountered adding Group '%s' to the ACL of the target database."
#define			MSG_DXR5069E		"DXR5069E: Unable to read the ACL flags in the target database."
#define			MSG_DXR5070E		"DXR5070E: Unable to write the ACL flags to the target database."
#define			MSG_DXR5072E		"DXR5072E: Unable to store the ACL in the target database."
#define			MSG_DXR5073E		"DXR5073E: Unable to read the D/B Info buffer from the target database."
#define			MSG_DXR5074E		"DXR5074E: Unable to set the design template in the target database."
#define			MSG_DXR5076E		"DXR5076E: Unable to refresh the design of the target database."
#define			MSG_DXR5078E		"DXR5078E: Unable to open the design collection in the target database."
#define			MSG_DXR5079E		"DXR5079E: Unable to read the design collection from the target database."
#define			MSG_DXR5080E		"DXR5080E: A null collection was returned while reading the design collection."
#define			MSG_DXR5081E		"DXR5081E: The design collection of the target database is empty."
#define			MSG_DXR5082E		"DXR5082E: Unable to determine the class of a design note."
#define			MSG_DXR5085E		"DXR5085E: Unable to determine the last Full Text Index time of the source database."
#define			MSG_DXR5086E		"DXR5086E: Unable to determine the size of the FTI definition object in the source database."
#define			MSG_DXR5087E		"DXR5087E: Unable to determine the size of the FTI definition object in the target database."
#define			MSG_DXR5088E		"DXR5088E: The FTI definition object already exists in the target database."
#define			MSG_DXR5089E		"DXR5089E: Unable to read the FTI definition object from the source database."
#define			MSG_DXR5090E		"DXR5090E: Unable to allocate the FTI definition object in the target database."
#define			MSG_DXR5091E		"DXR5091E: Unable to write the FTI definition object to the target database."
#define			MSG_DXR5092E		"DXR5092E: Unable to set the FTI definition as a special object in the target database."
#define			MSG_DXR5093E		"DXR5093E: Unable to enable Full Text Indexing on the target database."
#define			MSG_DXR5095E		"DXR5095E: Unable to locate agent '%s' in the target database."
#define			MSG_DXR5096E		"DXR5096E: Unable to open the requested agent in the target database."
#define			MSG_DXR5097E		"DXR5097E: Unable to create an agent run context."
#define			MSG_DXR5098E		"DXR5098E: Unable to re-direct agent output to an in-memory buffer."
#define			MSG_DXR5099E		"DXR5099E: An error was encountered trying to run the requested agent in the target database."
#define			MSG_DXR5102E		"DXR5102E: Unable to update the new Full Text Index on the target database."
#define			MSG_DXR5104E		"DXR5104E: Unable to post a request to copy notes, design copy terminated."
#define			MSG_DXR5105E		"DXR5105E: Unable to post a request to copy notes, data copy terminated."
#define			MSG_DXR5106E		"DXR5106E: Unable to post a request to maintain a folder, folder maintenance terminated."
#define			MSG_DXR5107E		"DXR5107E: Unable to post a request to maintain all folders, copy terminated."
#define			MSG_DXR5108E		"DXR5108E: Unable to post a request to maintain the FTI, copy terminated."
#define			MSG_DXR5109E		"DXR5109E: Unable to post a request to upgrade the database design, copy terminated."
#define			MSG_DXR5110E		"DXR5110E: Unable to post a request to run the conditioning agent, copy terminated."
#define			MSG_DXR5139E		"DXR5139E: Source database replica ID verification failed, copy abandoned."
#define			MSG_DXR5141E		"DXR5141E: Unable to read the replication information from the source ot target database."
#define			MSG_DXR5142E		"DXR5142E: Source database replica ID '%s' does not match '%s' supplied in the request, verfification failed."
#define			MSG_DXR5145E		"DXR5145E: Unable to set the replication information in the target database."
#define			MSG_DXR5146E		"DXR5146E: Rejoin timeout has been exceeded - see Diagnostic dump for more info."
#define			MSG_DXR5147E		"DXR5147E: Part Copy Parameter object @%p is not valid (type = %i), it will not be freed."
#define			MSG_DXR5150E		"DXR5150E: Unable to locate the $FolderIDTable item in a folder note."
#define			MSG_DXR5151E		"DXR5151E: The $FolderIDTable in a folder note is not an Object."
#define			MSG_DXR5152E		"DXR5152E: Unable to read the ID Table object (RRV: 0x%.4X)."
#define			MSG_DXR5153E		"DXR5153E: Unable to create an ID table for the validated note IDs."
#define			MSG_DXR5154E		"DXR5154E: Unable to obtain the UNID for Note ID: 0x%.4X, it will be dropped from the validated table."
#define			MSG_DXR5155E		"DXR5155E: Failed to add Note ID: 0x%.4X to the validated table."
#define			MSG_DXR5159E		"DXR5159E: Failed to purge invalid existing content from folder with Note ID: (0x%.4X)."
#define			MSG_DXR5160E		"DXR5160E: Unable to get the size of the ID Table object (RRV: 0x%.4X)."
#define			MSG_DXR5161E		"DXR5161E: Unable to reallocate the ID Table object (RRV: 0x%.4X) to %i bytes."
#define			MSG_DXR5162E		"DXR5162E: Unable to write to the ID Table object (RRV: 0x%.4X) with %i bytes."
#define			MSG_DXR5164E		"DXR5164E: Unable to update folder note (0x%.4X), after purging."
#define			MSG_DXR5165E		"DXR5165E: Unable to locate the '%s' item in the Folder note."
#define			MSG_DXR5166E		"DXR5166E: Unable to delete the '%s' item from the Folder note."
#define			MSG_DXR5167E		"DXR5167E: Unable to free the Folder object for '%s' (RRV: 0x%.4X)."
#define			MSG_DXR5170E		"DXR5170E: One or more copy actions has requested that the current transaction is aborted."
#define			MSG_DXR5171E		"DXR5171E: The current copy request has been aborted."
#define			MSG_DXR5181E		"DXR5181E: Source and Target servers are both Local, one-time replication is not possible."
#define			MSG_DXR5182E		"DXR5182E: Source and Target servers are both Remote, one-time replication is not possible."
#define			MSG_DXR5183E		"DXR5183E: One-time replication of database '%s' from '%s' has failed."
#define			MSG_DXR5190E		"DXR5190E: Errors were encountered during folder mainenance."
#define			MSG_DXR5191E		"DXR5191E: One or more failed copy actions has requested that the current transaction is aborted."
#define			MSG_DXR5194E		"DXR5194E: Attempted recovery of a copy failure has also failed, failing the transaction."
#define			MSG_DXR5196E		"DXR5196E: Search method for building copy requests has failed."
#define			MSG_DXR5199E		"DXR5199E: Unable to prepare the target database for copying."
#define			MSG_DXR5201E		"DXR5201E: Failed to delete the icon note in the target database."
#define			MSG_DXR5202E		"DXR5202E: Failed to create a new icon note in the target database."
#define			MSG_DXR5204E		"DXR5204E: Failed to copy the database icon to the target database."
#define			MSG_DXR5205E		"DXR5205E: Failed to copy the '%s' item to the icon note in the target database."
#define			MSG_DXR5206E		"DXR5206E: Failed to write the database flags to the icon note in the target database."
#define			MSG_DXR5207E		"DXR5207E: Failed to write the DAOS flag to the icon note in the target database."
#define			MSG_DXR5208E		"DXR5208E: Failed to update the icon note in the target database."
#define			MSG_DXR5210E		"DXR5210E: Failed to open note NID: 0x%.4X to check for attachment recompression."
#define			MSG_DXR5211E		"DXR5211E: Failed to get object info on note NID: 0x%.4X to check for attachment recompression."
#define			MSG_DXR5213E		"DXR5213E: Failed to open note NID: 0x%.4X to perform attachment recompression."
#define			MSG_DXR5125E		"DXR5125E: Errors were encountered recompressing attachment: %i on Note: 0x%.4X."
#define			MSG_DXR5126E		"DXR5126E: Failed to write Note: 0x%.4X with recompressed attachments."

class DbCopier :
	Runnable
{
public:
	//  Default Constructor
	DbCopier(void);

	//  Constructor
	DbCopier(MTExecutive *xeParent, int iThreadID);

	//  Destructor
	~DbCopier(void);

	//  Public Members
	int			OwningThread;											//  Creating Thread ID
	int			MultiplexScale;											//  Scaling factor for multiplexor

	//  Public Functions
	BOOL CopyThisDb(CopyRequest *crRQ, int iThreadID);					//  Entry point to initiate copying
	void ExecuteThisRequest(void *pReqObject, int iThreadID);			//  Multi thread request entry point
	static STATUS LNPUBLIC WrapperAddProfileToIDT(DBHANDLE, void *, char *, WORD, char *, WORD, NOTEID);
	static STATUS LNCALLBACK WrapperProcessSearchResult(void *ObjPtr, SEARCH_MATCH *smCurrent, ITEM_TABLE *itCurrent);
	STATUS AddProfileToIDT(PartCopyRequest *pcrqRQ, NOTEID nidProf);	//  Real processor for profile enum

private:
	//  Private Members

	//  Private Functions
	void PerformCopyAction(PartCopyRequest *pcrqRQ, int iThreadID);		//  Perform a copy action
	void PerformCopyDesign(PartCopyRequest *pcrqRQ, int iThreadID);		//  Copy Design action
	void PerformCopyData(PartCopyRequest *pcrqRQ, int iThreadID);		//  Copy Data action
	BOOL CopyDataUsingSearch(PartCopyRequest *pcrqRQ, int iThreadID);   //  Copy data using NSFSearch
	void PerformCopyNotes(PartCopyRequest *pcrqRQ, int iThreadID);		//  Copy Notes
	void PerformCRA(PartCopyRequest *pcrqRQ, int iThreadID);			//  Check for Attachment recompression
	void PerformRecompress(PartCopyRequest *pcrqRQ, int iThreadID);		//  Recompress Attachments
	void PerformRecompressInPlace(PartCopyRequest *pcrqRQ, int iThreadID);	//  Recompress Attachments in place
	void PerformMaintainFolder(PartCopyRequest *pcrqRQ, int iThreadID);		//  Maintain Folder
	void PerformUpgradeDesign(PartCopyRequest *pcrqRQ, int iThreadID);	//  Upgrade the database design
	void PerformBuildIndexes(PartCopyRequest *pcrqRQ, int iThreadID);	//  Build all view Indexes
	void PerformBuildIndex(PartCopyRequest *pcrqRQ, int iThreadID);		//  Build a single view Index
	void PerformCopyFTI(PartCopyRequest *pcrqRQ, int iThreadID);		//  Copy the Full Text Index Definition
	void PerformRunAgent(PartCopyRequest *pcrqRQ, int iThreadID);		//  Run the requested agent
	void PerformMaintainFolders(PartCopyRequest *pcrqRQ, int iThreadID);		//  Maintain Folders
	BOOL IsPrivateDesignNote(PartCopyRequest *pcrqRQ, NOTEID nidNote, STATUS *stOpenErr, int iThreadID);			//  Determine if the note is a private design
	void ShowCopyStats(CopyRequest *crRQ, int iThreadID);				//  Display copy statistics
	STATUS FilteredFolderDocAdd(PartCopyRequest *pcrqRQ, DBHANDLE hdbTarget, NOTEID nidTargetFolder, DHANDLE hitDocs, int iThreadID);  //  Add filtered entries to table
	void ReenableReplication(CopyRequest *crRQ, int iThreadID);			//  Re-enable replication in the target database
	void DisableReplication(CopyRequest *crRQ, int iThreadID);			//  Disable replication in the target database
	void ConditionACL(char *szGroup, CopyRequest *crRQ,int iThreadID);	//  Add specified group to ACL
	DWORD ComputeElapsed(clock_t *tStart, clock_t *tEnd);				//  Compute an elapsed time
	BOOL VerifyReplicaID(DBHANDLE hdbSrc, CopyRequest *crCurrent, int iThreadID); //  Verify the source database replica ID
	void DisableAgents(CopyRequest *crRQ, DBHANDLE hdbTarget, int iThreadID);				//  Disable background agents
	int SetAllRoles(DHANDLE hACL, ACL_PRIVILEGES *apCurrent);			//  Set all roles in a PRIV array
	DHANDLE BuildFolderIDTable(PartCopyRequest *pcrqRQ, DBHANDLE hdbSource, NOTEID nidFolder, int iThreadID);	//  Build folder ID table from native objects
	STATUS PurgeFolderContent(PartCopyRequest *pcrqRQ, DBHANDLE hdbTarget, NOTEID nidFolder, int iThreadID); //  Purge the folders content
	BOOL PurgeFolderObject(PartCopyRequest *pcrqRQ, NOTEHANDLE hnFolder, int iObjNo, int iThreadID);	//  Purge a $FolderObject[n] instance
	BOOL PerformRestartActions(CopyRequest *crRQ, int iThreadID);		//  Perform restart actions
	void DrainRequests(CopyRequest *crRQ, int iThreadID);				//  Drain all requests		
	BOOL PerformPreProcessActions(CopyRequest *crRQ, int iThreadID);	//  Perform pre-processing actions
	BOOL PerformReplicateDb(CopyRequest *crRQ, int iThreadID);			//  Replicate source with target database
	STATUS ProcessSearchResult(SEARCH_MATCH *smHit, PartCopyRequest *pcrqRQ);	//  Search result processor
	BOOL prepareTargetDb(CopyRequest *crRQ, DBHANDLE hdbSrc, DBHANDLE hdbTgt, int iThreadID);		//  Prepare the target database
	WORD setFlag(char Flag, char *szFlags);								//  Set a flag in the flag string
	WORD removeFlag(char Flag, char *szFlags);							//  Unset a flag in the flag string
	void copyIcon(NOTEHANDLE hnSrc, NOTEHANDLE hnTgt, CopyRequest *crRQ, int iThreadID);	//  Copy the database icon from source to target
	void copyIconTextItem(NOTEHANDLE hnSrc, NOTEHANDLE hnTgt, char *szItemName, CopyRequest *crRQ, int iThreadID);	//  Copy a text item between icon notes
	void copyIconTDItem(NOTEHANDLE hnSrc, NOTEHANDLE hnTgt, char *szItemName, CopyRequest *crRQ, int iThreadID);	//  Copy a timedate item between icon notes
	void reportTargetDb(DBHANDLE hdbTgt, int iThreadID);				//  Report on the flags and optionas in the target D/B

protected:
	//  Protected Members
	MTExecutive			*xeLocal;										//  Multi-Threading Runtime

	//  Protected Functions
	BOOL virtual ProxyLogMessage(CopyRequest *crRQ, char *szMsg, int iThreadID);			//  Proxry Log interface
};
