//  DbMover
//
//  This class provides the Database Move functions used in Domono eXplorer (DX) utilities
//
//  Message Identifier:      DXR
//  Message Range:           6001 - 6499
//  Trace area assignment:   TRACE_DBM (25)
//  DX Version:				 3.14.0
//  Version:                 3.14.0  Build: 108
//
//  Version History:
//
//  3.6.2		14/10/2009		Prepared for QMove 1.0
//  3.6.3		19/12/2009		Updates to support additional ops for QMove 1.1.0
//  3.6.4		23/12/2009		Numerous cosmetic fixes for QMove 1.1.1
//  3.7.0		25/12/2009		UNIX compatability fixes in QMove
//  3.7.3		06/02/2010		Handle Db & Dir Links
//  3.10.2		18/11/2010		MTIO Support added to DbMover
//  3.11.1		22/10/2011		VS 2011 Updates
//	3.14.0		19/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 2015
//

//  Application Includes
#include "DbMover.h"

//  Default Constructor
DbMover::DbMover(void)
{
}

//  Constructor
DbMover::DbMover(MTExecutive *xeParent, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Save the Runtime Environment
	xeLocal = xeParent;
	//  Save the Creating Thread ID
	OwningThread = iThreadID;

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR6001T, this);
	xeLocal->LogTrace(TRACE_DBM, szMsg, OwningThread);
}

//  Destructor
DbMover::~DbMover(void)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR6002T, this);
	xeLocal->LogTrace(TRACE_DBM, szMsg, OwningThread);
}

//  MoveThisDb
//
//  This is the main entry point to request move process.
//
//  Parameters
//
//  1  -  MoveRequest *		-  Pointer to the MoveRequest for this database
//  2  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  BOOL  -			TRUE if the database was processed (moved) successfully
//                  FALSE if the database could not be moved
//
//  API Use Notes:
//
//
BOOL DbMover::MoveThisDb(MoveRequest *mrRQ, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code frm Notes API calls
	WORD				wRetLen;								//  Returned Length
	DBHANDLE			hdbSrc;									//  Source database handle
	DHANDLE				hBUContext;								//  Handle of the backup context
	DHANDLE				hChgContext;							//  Handle of the change context
	DHANDLE				hChgBuffer;								//  Handle of the change buffer
	DWORD				dwFSHigh;								//  High watermark
	DWORD				dwFSLow;								//  Low watermark
	BOOL				bDbIsLogged;							//  Flag showing if D/B is logged
	UNID				uidLogID;								//  Log ID
	UNID				uidDBIID;								//  DBIID
	DWORD				dwLogExtent;							//  Log Extent
	DWORD				dwChgSize;								//  Change quantum size
	int					iQuantum;								//  Change quantum number
	BOOL				bAttemptOffline;						//  Flag to control offline attempts
	BOOL				bFixupDB;								//  Flag that signals that a fixup is needed
	int					iTakeoffAttempts;						//  # attempts to take database offline
	int					iDbConfig;								//  Database link configuration
	DWORD				dwDataDocs;								//  Count of data documents
	DWORD				dwSizeUsed;								//  Size of the database
	DWORD				dwSizeFree;								//  Free space in the database
	clock_t				ct_Start;								//  Copy Start Timestamp
	clock_t				ct_End;									//  Copy End Timestamp
	long				lElapsed;								//  Elapsed Time
	long				lTimeFactor = 1000;						//  Clock tick conversion factor
	char				*ChgBuffer;								//  Change application Buffer
	char				*pSrcChar;								//  Pointer to source character
	char				*pSrcFileName;							//  Pointer to the file name portion of the source
	char				szOSSource[MAXPATH + 1];				//  Source File Name
	char				szOSTarget[MAXPATH + 1];				//  Target File Name
	char				szOSLink[MAXPATH + 1];					//  D/B Link File Name
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Record the start time
	ct_Start = xeLocal->RunningTime->getElapsedMillis();
	iDbConfig = DB_REDIRECT_NONE;
	bFixupDB = FALSE;
	mrRQ->ReturnCode = RETURN_INCOMPLETE;

	//  Get the source file name and compute the target filename
	wRetLen = OSGetDataDirectory(szOSSource);
	szOSSource[wRetLen] = '\0';
	if ((szOSSource[wRetLen - 1] != '\\') & (szOSSource[wRetLen - 1] != '/'))
	{
		strcat_s(szOSSource, MAXPATH, "/");
	}
	strcat_s(szOSSource, MAXPATH, mrRQ->szSourceDatabase);
	normalizePath(szOSSource);

	strcpy_s(szOSTarget, MAXPATH, mrRQ->szTargetDir);
	wRetLen = (WORD) strlen(szOSTarget);
	if ((szOSTarget[wRetLen - 1] != '\\') & (szOSTarget[wRetLen - 1] != '/'))
	{
		strcat_s(szOSTarget, MAXPATH, "/");
	}

	//  Find the last directory separator in the source file name
	pSrcChar = szOSSource;
	pSrcFileName = szOSSource;
	while (*pSrcChar != '\0')
	{
		if ((*pSrcChar == '\\') || (*pSrcChar == '/')) pSrcFileName = pSrcChar + 1;
		pSrcChar++;
	}
	strcat_s(szOSTarget, MAXPATH, pSrcFileName);
	normalizePath(szOSTarget);

	if ((mrRQ->PerformCopy) && (mrRQ->PerformDelete))
	{
		if (mrRQ->PerformLink)
		{
			//  This is a move operation
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6085I, szOSSource, szOSTarget);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
		else
		{
			//  This is a retire operation
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6094I, szOSSource, szOSTarget);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
	}
	else
	{
		if (mrRQ->PerformCopy)
		{
			//  This is a backup operation
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6092I, szOSSource, szOSTarget);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
		else
		{
			//  This is a delete operation
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6093I, szOSSource);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
	}

	//  Open the source database
	hdbSrc = xeLocal->OpenDatabase("Local", mrRQ->szSourceDatabase);
	if (hdbSrc == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6054E, mrRQ->szSourceDatabase);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		mrRQ->ReturnCode = RETURN_INCOMPLETE;
		return FALSE;
	}

	//  Get the count of documents in the database and size for stats
	dwDataDocs = xeLocal->GetDataDocumentCount(hdbSrc, 0);
	NSFDbSpaceUsage(hdbSrc, &dwSizeUsed, &dwSizeFree);

	if (!mrRQ->PerformCopy) checkCopy(mrRQ, szOSSource, szOSTarget, szOSLink, hdbSrc, &iDbConfig, iThreadID);

	if (mrRQ->PerformCopy)
	{

		//  Create the target directory before starting the backup
		if (!createDirectory(mrRQ, mrRQ->szTargetDir, iThreadID))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6090E, mrRQ->szTargetDir);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		//  Start recording changes on the source database
		stAPIRC = NSFBackupStart(hdbSrc, 0, &hBUContext, &dwFSLow, &dwFSHigh);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6055E, mrRQ->szSourceDatabase);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		//  Check if the database is logged - abort if not
		stAPIRC = NSFDbGetLogInfo(hdbSrc, 0, &bDbIsLogged, &uidLogID, &uidDBIID, &dwLogExtent);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6056E, mrRQ->szSourceDatabase);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		if (!bDbIsLogged)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6057W, mrRQ->szSourceDatabase);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		//  Show the Transaction Log Information
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6098I,
					uidLogID.File.Innards[1],
					uidLogID.File.Innards[0],
					uidLogID.Note.Innards[1],
					uidLogID.Note.Innards[0],
					uidDBIID.File.Innards[1],
					uidDBIID.File.Innards[0],
					uidDBIID.Note.Innards[1],
					uidDBIID.Note.Innards[0],
					dwLogExtent);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);

		//  Check the logic of the copy operation and fill in the link file name
		if (!checkCopy(mrRQ, szOSSource, szOSTarget, szOSLink, hdbSrc, &iDbConfig, iThreadID))
		{
			ProxyLogMessage(mrRQ, MSG_DXR6099E, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		//  Check that the source and target databases are not the same
		if (strcmpi(szOSSource, szOSTarget) == 0)
		{
			ProxyLogMessage(mrRQ, MSG_DXR6109E, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		//  Check that we still hold the global run permit - fail the transaction if not
		if (!xeLocal->TestGlobalPermit())
		{
			ProxyLogMessage(mrRQ, MSG_DXR6126E, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		//  Now perform the OS Copy of the file
		if (!copyOSFile(mrRQ, szOSSource, szOSTarget, hdbSrc, hBUContext, dwFSHigh, dwFSLow, iThreadID))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6058E, szOSSource);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		//  Stop recording changes to the source database
		stAPIRC = NSFBackupStop(hdbSrc, hBUContext);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6059E, mrRQ->szSourceDatabase);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}
	
		//  Get the change info size
		stAPIRC = NSFBackupGetChangeInfoSize(hdbSrc, hBUContext, 0, &dwFSLow, &dwFSHigh);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6060E, mrRQ->szSourceDatabase);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		sprintf_s(szMsg, MAX_MSG, MSG_DXR6087I, dwFSHigh, dwFSLow);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);

		//  Create a context for applying the changes to the copied file
		stAPIRC = NSFBackupStartApplyChangeInfo(&hChgContext, szOSTarget, 0, dwFSLow, dwFSHigh);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6061E, mrRQ->szSourceDatabase);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		//  Allocate a buffer for applying the changes
		stAPIRC = OSMemAlloc(MEM_SHARE, CHG_BUFFER_SIZE, &hChgBuffer);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6062E, CHG_BUFFER_SIZE);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
			NSFDbClose(hdbSrc);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		//  Lock the buffer
		ChgBuffer = OSLock(char, hChgBuffer);

		//  Loop applying the changes
		iQuantum = 0;
		dwChgSize = 1;							//  Dummy priming for the loop.
		while (dwChgSize > 0)
		{
			//  Get the next change quantum
			stAPIRC = NSFBackupGetNextChangeInfo(hdbSrc, hBUContext, 0, ChgBuffer, CHG_BUFFER_SIZE, &dwChgSize);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR6063E, mrRQ->szSourceDatabase);
				ProxyLogMessage(mrRQ, szMsg, iThreadID);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				ProxyLogMessage(mrRQ, szMsg, iThreadID);
				NSFBackupEndApplyChangeInfo(hChgContext, APPLYEND_ABORT);
				OSUnlock(hChgBuffer);
				OSMemFree(hChgBuffer);
				NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
				NSFDbClose(hdbSrc);
				mrRQ->ReturnCode = RETURN_INCOMPLETE;
				return FALSE;
			}

			//  Apply the latest quantum of change
			iQuantum++;
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6064I, iQuantum, dwChgSize);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			stAPIRC = NSFBackupApplyNextChangeInfo(hChgContext, 0, ChgBuffer, dwChgSize);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR6065E, szOSTarget);
				ProxyLogMessage(mrRQ, szMsg, iThreadID);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				ProxyLogMessage(mrRQ, szMsg, iThreadID);
				NSFBackupEndApplyChangeInfo(hChgContext, APPLYEND_ABORT);
				OSUnlock(hChgBuffer);
				OSMemFree(hChgBuffer);
				NSFBackupEnd(hdbSrc, hBUContext, BACKUPEND_ABORT);
				NSFDbClose(hdbSrc);
				mrRQ->ReturnCode = RETURN_INCOMPLETE;
				return FALSE;
			}
		}

		NSFBackupEndApplyChangeInfo(hChgContext,0);
		NSFBackupEnd(hdbSrc, hBUContext, 0);

		//  Free the change buffer
		OSUnlockObject(hChgBuffer);
		OSMemFree(hChgBuffer);
	}
	//  Close the source database
	NSFDbClose(hdbSrc);

	if (mrRQ->PerformDelete)
	{
		//  Take the database off-line and delete it in the process
		iTakeoffAttempts = 0;
		bAttemptOffline = TRUE;

		while (bAttemptOffline)
		{
			stAPIRC = NSFTakeDatabaseOffline(mrRQ->szSourceDatabase, DB_OFFLINE_WAIT, 0);
			if ((ERR(stAPIRC) == ERR_DBINUSE) || (ERR(stAPIRC) == ERR_NSFOPEN) || (ERR(stAPIRC) == ERR_LOCK))
			{
				//  Wait a bit
				xeLocal->AppWait(DB_OFFLINE_WAIT);
				iTakeoffAttempts++;
				if (iTakeoffAttempts == MAX_TAKEOFF_ATTEMPTS) bAttemptOffline = FALSE;
			}
			else bAttemptOffline = FALSE;
		}
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6066E, szOSSource, iTakeoffAttempts);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		sprintf_s(szMsg, MAX_MSG, MSG_DXR6086I, mrRQ->szSourceDatabase, iTakeoffAttempts);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);

		//  Determine of we also have to delete the link file
		if (iDbConfig == DB_REDIRECT_DB)
		{
			if (remove(szOSLink) != 0)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR6128E, szOSLink);
				ProxyLogMessage(mrRQ, szMsg, iThreadID);
			}
			else
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR6127I, szOSLink);
				ProxyLogMessage(mrRQ, szMsg, iThreadID);
			}
		}
	}

	if (mrRQ->PerformLink)
	{
		//  Check that the redirection is not a directory level redirection if so warn and don't link
		if (iDbConfig == DB_REDIRECT_DIR)
		{
			ProxyLogMessage(mrRQ, MSG_DXR6108W, iThreadID);
		}
		else
		{
			//  Create and write the D/B link
			if (!createDbLink(mrRQ, szOSLink, szOSTarget, iThreadID))
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR6067E, szOSLink);
				ProxyLogMessage(mrRQ, szMsg, iThreadID);
				mrRQ->ReturnCode = RETURN_INCOMPLETE;
				return FALSE;
			}
		}
	}

	if (mrRQ->PerformCopy)
	{
		//  Refresh the backup before bringing it back online
		szOSTarget[strlen(szOSTarget) + 1] = '\0';					//  Need a double terminator for this
		stAPIRC = NSFRecoverDatabases(szOSTarget, NULL, DBRECOVER_WAIT | DBRECOVER_REFRESH_BACKUP, NULL, NULL);
		if (ERR(stAPIRC) != NOERROR)
		{
			//  Certain conditions may be reported back by this process which are non-fatal
			if (ERR(stAPIRC) == ERR_RM_LOG_SPACE_CRITICAL)
			{
				ProxyLogMessage(mrRQ, MSG_DXR6113W, iThreadID);
			}
			else
			{
				if (ERR(stAPIRC) == ERR_RM_MQ_LOG_ERROR)
				{
					ProxyLogMessage(mrRQ, MSG_DXR6114W, iThreadID);
					bFixupDB = TRUE;
				}
					sprintf_s(szMsg, MAX_MSG, MSG_DXR6088E, szOSTarget);
					ProxyLogMessage(mrRQ, szMsg, iThreadID);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
					ProxyLogMessage(mrRQ, szMsg, iThreadID);
					mrRQ->ReturnCode = RETURN_INCOMPLETE;
					return FALSE;
				}
		}

		if (mrRQ->PerformLink) sprintf_s(szMsg, MAX_MSG, MSG_DXR6089I, szOSTarget);
		else sprintf_s(szMsg, MAX_MSG, MSG_DXR6110I, szOSTarget);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
	}

	if ((mrRQ->PerformDelete) && (mrRQ->PerformLink) && (iDbConfig != DB_REDIRECT_DIR))
	{
		//  Bring the database back online - via the link
		stAPIRC = NSFBringDatabaseOnline(mrRQ->szSourceDatabase, 0);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6068E, mrRQ->szSourceDatabase);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}

		// If needed inform the user that a fixup is needed
		if (bFixupDB)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6115E, mrRQ->szSourceDatabase);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}
	}

	//  Record the ending time
	ct_End = xeLocal->RunningTime->getElapsedMillis();

	//  Record the stats
	mrRQ->dwSizeKB = dwSizeUsed / BYTES_PER_KB;

	//  Document the run statistics
	lElapsed = (ct_End - ct_Start) / lTimeFactor;
	mrRQ->dwTotalElapsed = (DWORD) lElapsed;
	mrRQ->dwDataDocsCopied = dwDataDocs;

	//  Show operation done
	if ((mrRQ->PerformCopy) && (mrRQ->PerformDelete))
	{
		if (mrRQ->PerformLink)
		{
			//  This is a move operation
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6069I, szOSSource, szOSTarget);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
		else
		{
			//  This is a retire operation
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6097I, szOSSource, szOSTarget);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
	}
	else
	{
		if (mrRQ->PerformCopy)
		{
			//  This is a backup operation
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6096I, szOSSource, szOSTarget);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
		else
		{
			//  This is a delete operation
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6095I, szOSSource);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
	}

	mrRQ->ReturnCode = RETURN_NOERROR;

	return TRUE;
}

//  copyOSFile
//
//  This function copy the source file to the target notifying progress to the Backup API as it proceeds.
//
//  Parameters
//
//  1  -  MoveRequest *		-  Pointer to the top level move request
//  2  -  char *			-  Pointer to the file name to be copied
//  3  -  char *			-  Pointer to the target file name
//  4  -  DBHANDLE			-  DBHANDLE of the source database
//  5  -  HANDLE			-  HANDLE of the backup context
//  6  -  DWORD				-  Size of the file (High DWORD)
//  7  -  DWORD				-  Size of the file (Low DWORD)
//  8  -  int				-  Identity of the thread making the request
//
//  Returns
//
//  BOOL					-  TRUE if the file was copied
//							-  FALSE if the file could not be copied
//
//  API Use Notes:
//
//
BOOL DbMover::copyOSFile(MoveRequest *mrUOW, char *szIn, char *szOut, DBHANDLE hdbSrc, DHANDLE hBUContext, DWORD SizeHi, DWORD SizeLo, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
#ifdef WIN32
	HANDLE				hfSrc;									//  Handle of the source file
	HANDLE				hfTarg;									//  Handle of the target file
	DWORD				dwWinError;								//  Last error code
#endif
	DHANDLE				hCpyBuffer;								//  Handle of the copy buffer
	DWORD				dwHighPos;								//  Senior part of position
	DWORD				dwFSHi;									//  Supplied File Size (high)
	DWORD				dwFSLo;									//  Supplied File Size (low)
	DWORD				dwBytesLeft;							//  Bytes left to read
	DWORD				dwBytesRead;							//  Bytes read so far
	DWORD				dwReadCount;							//  Count of buffers copied
	DWORD				dwSPHigh;								//  Source position (high)
	DWORD				dwSPLow;								//  Source position (low)
	DWORD				dwBytesReadActual;						//  Actual read length
	DWORD				dwFSDisp;								//  Display File Size
	DWORD				dwFSGbDisp;								//  Gb part of display for large D/Bs
	DWORD				dwBufferSizeDisp;						//  Display Buffer Size
	DWORD				dwIOPs;									//  Count of I/O operations (read + write)
	DWORD				dwCkpts;								//  Count of Checkpoints
	BOOL				bUseLargeFileCopy;						//  Use the large file copier
	clock_t				ct_Start;								//  Copy Start Timestamp
	clock_t				ct_End;									//  Copy End Timestamp
	long				lElapsed;								//  Elapsed time
	long				lTimeFactor = 1000;			//  Clock tick conversion factor
	char				*CpyBuffer;								//  Pointer to the copy buffer
	char				szFDDispUnits[10];						//  Display File size units
	char				szGBFraction[10];						//  Display GB Fraction size
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer
#ifdef UNIX
	int					fdSrc;									//  File Descriptor for Source
	int					fdTarg;									//  File Descriptor for Target
	off_t				offSrc;									//  Source Offset
	off_t				offResult;								//  Resulting offset from lseek
	mode_t				mCreate;								//  File Create Mode
	ssize_t				stRead;									//  Size read/written
	int					iCloseRC;								//  Return Code from Close
#endif

	//  Default to performing an inline copy
	bUseLargeFileCopy = FALSE;

	//  Set the file size to copy
	dwFSHi = SizeHi;
	dwFSLo = SizeLo;

	//  Show starting to copy 
	szGBFraction[0] = '\0';
	dwBufferSizeDisp = mrUOW->CopyBfrKb;
	if (dwFSHi == 0)
	{
		strcpy_s(szFDDispUnits, 10, DISP_SIZE_KB);
		dwFSDisp = dwFSLo / 1024;								//  Convert size to Kb
		if (dwFSDisp > 1024)
		{
			strcpy_s(szFDDispUnits, 10, DISP_SIZE_MB);
			dwFSDisp = dwFSDisp / 1024;							//  Convert size to Mb
			//  Check against the lower limit for using the large file copier
			if (dwFSDisp > INLINE_COPY_LIMMB) bUseLargeFileCopy = TRUE;
			if (dwFSDisp > 1024)
			{
				strcpy_s(szFDDispUnits, 10, DISP_SIZE_GB);
				dwFSGbDisp = dwFSDisp / 1024;							//  Convert size to Gb
				dwFSDisp = dwFSDisp - (dwFSGbDisp * 1024);				//  Remainder Mb
				dwFSDisp = dwFSDisp * 1000;
				dwFSDisp = dwFSDisp / 1024;								//  Fractional remainder
				sprintf_s(szGBFraction, 10, " %i", dwFSDisp);
				szGBFraction[0] = '.';
				dwFSDisp = dwFSGbDisp;
			}
		}
	}
	else
	{
		bUseLargeFileCopy = TRUE;
		strcpy_s(szFDDispUnits, 10, DISP_SIZE_GB);
		dwFSDisp = dwFSLo / 1024;								//  Convert size to Kb
		dwFSDisp = dwFSDisp / 1024;								//  Convert size to Mb
		dwFSGbDisp = dwFSDisp / 1024;							//  Convert size to Gb
		dwFSDisp = dwFSDisp - (dwFSGbDisp * 1024);				//  Remainder Mb
		dwFSGbDisp = dwFSGbDisp + (dwFSHi * 4);					//  Hi Order 4Gb
		dwFSDisp = dwFSDisp * 1000;
		dwFSDisp = dwFSDisp / 1024;								//  Fractional remainder
		sprintf_s(szGBFraction, 10, " %i", dwFSDisp);
		szGBFraction[0] = '.';
		dwFSDisp = dwFSGbDisp;
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR6111I, dwFSDisp, szGBFraction, szFDDispUnits, dwBufferSizeDisp);
	ProxyLogMessage(mrUOW, szMsg, iThreadID);

	//  Disable the Large File Copier - if indicated
	if (!mrUOW->UseLFC) bUseLargeFileCopy = FALSE;

	//  Determine if we should use the Large (MTIO) OS File Copier instead of an inline copy
	if (bUseLargeFileCopy) return copyLargeOSFile(mrUOW, szIn, szOut, hdbSrc, hBUContext, SizeHi, SizeLo, iThreadID);

#ifdef WIN32
	//  Open the source file
	hfSrc = CreateFile(szIn, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hfSrc == (HANDLE) HFILE_ERROR)
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6070E, szIn);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		return FALSE;
	}
#endif
#ifdef UNIX
	//  Open the source file
	fdSrc = open(szIn, O_RDONLY | O_LARGEFILE);
	if (fdSrc == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6070E, szIn);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		return FALSE;
	}
#endif
	//  Move the file pointer to the start of the file
#ifdef WIN32
	dwHighPos = 0;
	if(SetFilePointer(hfSrc, 0, (PLONG) &dwHighPos, FILE_BEGIN) == MAXWORD)
	{
		dwWinError = GetLastError();
		ProxyLogMessage(mrUOW, MSG_DXR6073E, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		CloseHandle(hfSrc);
		return FALSE;
	}
#endif
#ifdef UNIX
		offSrc = 0;
		offResult = lseek(fdSrc, offSrc, SEEK_SET);
		if (offResult == -1)
		{
			ProxyLogMessage(mrUOW, MSG_DXR6073E, iThreadID);
			xeLocal->GetUNIXAPIError(errno, szMsg);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
			close(fdSrc);
			return FALSE;
		}
#endif

	//  Create the target file
#ifdef WIN32
	hfTarg = CreateFile(szOut, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hfTarg == (HANDLE) HFILE_ERROR)
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6071E, szOut);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		CloseHandle(hfSrc);
		return FALSE;
	}
#endif

#ifdef UNIX
	mCreate = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
	fdTarg = open(szOut, O_RDWR | O_CREAT | O_EXCL | O_LARGEFILE, mCreate);
	if (fdTarg == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6071E, szOut);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		close(fdSrc);
		return FALSE;
	}
#endif

	//  Allocate a buffer for doing the copy
	stAPIRC = OSMemAlloc(MEM_SHARE, mrUOW->CopyBfrKb * BYTES_PER_KB, &hCpyBuffer);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6072E, mrUOW->CopyBfrKb);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
#ifdef WIN32
		CloseHandle(hfSrc);
		CloseHandle(hfTarg);
#endif
#ifdef UNIX
		close(fdSrc);
		close(fdTarg);
#endif
		return FALSE;
	}

	//  Lock the buffer
	CpyBuffer = OSLock(char, hCpyBuffer);

	//  Compute the initial read target
	if ((MAXDWORD - dwFSHi) >= dwFSLo)
	{
		dwBytesLeft = dwFSHi + dwFSLo;
		dwFSLo = 0;
	}
	else
	{
		dwBytesLeft = dwFSHi;
	}

	//  Loop copying a buffer at a time
	dwReadCount = 0;
	dwSPHigh = 0;
	dwSPLow = 0;
	ct_Start = xeLocal->RunningTime->getElapsedMillis();
	dwIOPs = 0;
	dwCkpts = 0;
	while (dwBytesLeft > 0)
	{
		if (dwBytesLeft < mrUOW->CopyBfrKb * BYTES_PER_KB) dwBytesRead = dwBytesLeft;
		else dwBytesRead = mrUOW->CopyBfrKb * BYTES_PER_KB;
#ifdef WIN32
		if (!ReadFile(hfSrc, CpyBuffer, dwBytesRead, &dwBytesReadActual, NULL))
		{
			dwWinError = GetLastError();
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6074E, dwBytesRead);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwWinError, szMsg);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
			CloseHandle(hfSrc);
			CloseHandle(hfTarg);
			OSUnlockObject(hCpyBuffer);
			OSMemFree(hCpyBuffer);
			return FALSE;
		}
#endif
#ifdef UNIX
		stRead = read(fdSrc, CpyBuffer, dwBytesRead);
		if (stRead == -1)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6074E, dwBytesRead);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
			xeLocal->GetUNIXAPIError(errno, szMsg);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
			close(fdSrc);
			close(fdTarg);
			OSUnlockObject(hCpyBuffer);
			OSMemFree(hCpyBuffer);
			return FALSE;
		}
		dwBytesReadActual = stRead;
#endif
		dwIOPs++;
		//  Check the length actually read
		if (dwBytesRead != dwBytesReadActual)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6075E, dwBytesRead, dwBytesReadActual);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
#ifdef WIN32
			CloseHandle(hfSrc);
			CloseHandle(hfTarg);
#endif
#ifdef UNIX
			close(fdSrc);
			close(fdTarg);
#endif
			OSUnlockObject(hCpyBuffer);
			OSMemFree(hCpyBuffer);
			return FALSE;
		}

		//  Compute the total read
        if ((MAXDWORD - dwSPLow) < dwBytesRead)
        {
            dwSPHigh++;
            dwSPLow = (dwBytesRead - (MAXDWORD - dwSPLow)) - 1;
        }
        else
        {
            dwSPLow += dwBytesRead;
        }

		//  Check that we have the Global Run Permit
		if (!xeLocal->TestGlobalPermit())
		{
			ProxyLogMessage(mrUOW, MSG_DXR6126E, iThreadID);
#ifdef WIN32
			CloseHandle(hfSrc);
			CloseHandle(hfTarg);
#endif
#ifdef UNIX
			close(fdSrc);
			close(fdTarg);
#endif
			OSUnlockObject(hCpyBuffer);
			OSMemFree(hCpyBuffer);
			return FALSE;
		}

		//  Update the read count and notify the backup if we are at a checkpoint interval
		dwReadCount++;
		if (dwReadCount % READ_CKPT_INTERVAL)
		{
			dwCkpts++;
			stAPIRC = NSFBackupSetHighWaterMark(hdbSrc, hBUContext, dwSPLow, dwSPHigh);
			if (ERR(stAPIRC) != NOERROR)
			{
				ProxyLogMessage(mrUOW, MSG_DXR6076E, iThreadID);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				ProxyLogMessage(mrUOW, szMsg, iThreadID);
#ifdef WIN32
				CloseHandle(hfSrc);
				CloseHandle(hfTarg);
#endif
#ifdef UNIX
				close(fdSrc);
				close(fdTarg);
#endif
				OSUnlockObject(hCpyBuffer);
				OSMemFree(hCpyBuffer);
				return FALSE;
			}
		}

		//  Write the buffer to the copy
#ifdef WIN32
		if (!WriteFile(hfTarg, CpyBuffer, dwBytesRead, &dwBytesReadActual, NULL))
		{
			dwWinError = GetLastError();
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6077E, dwBytesRead);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwWinError, szMsg);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
			CloseHandle(hfSrc);
			CloseHandle(hfTarg);
			OSUnlockObject(hCpyBuffer);
			OSMemFree(hCpyBuffer);
			return FALSE;
		}
#endif
#ifdef UNIX
		stRead = write(fdTarg, CpyBuffer, dwBytesRead);
		if (stRead == -1)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6077E, dwBytesRead);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
			xeLocal->GetUNIXAPIError(errno, szMsg);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
			close(fdSrc);
			close(fdTarg);
			OSUnlockObject(hCpyBuffer);
			OSMemFree(hCpyBuffer);
			return FALSE;
		}
		dwBytesReadActual = stRead;
#endif
		dwIOPs++;

		//  Check the length written
		if (dwBytesRead != dwBytesReadActual)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6078E, dwBytesRead, dwBytesReadActual);
			ProxyLogMessage(mrUOW, szMsg, iThreadID);
#ifdef WIN32
			CloseHandle(hfSrc);
			CloseHandle(hfTarg);
#endif
#ifdef UNIX
			close(fdSrc);
			close(fdTarg);
#endif
			OSUnlockObject(hCpyBuffer);
			OSMemFree(hCpyBuffer);
			return FALSE;
		}

		//  Compute the next read length
		dwBytesLeft -= dwBytesRead;
		if (dwBytesLeft == 0)
		{
			if (dwFSHi > 0)
			{
				dwBytesLeft = MAXDWORD;
				dwFSHi--;
			}
			else
			{
				dwBytesLeft = dwFSLo;
				dwFSLo = 0;
			}
		}
	}

	ct_End = xeLocal->RunningTime->getElapsedMillis();

	//  Free the copy buffer
	OSUnlockObject(hCpyBuffer);
	OSMemFree(hCpyBuffer);

	//  Close the copy
#ifdef WIN32
	if (!CloseHandle(hfTarg))
	{
		dwWinError = GetLastError();
		ProxyLogMessage(mrUOW, MSG_DXR6079E, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		CloseHandle(hfSrc);
		return FALSE;
	}
#endif
#ifdef UNIX
	iCloseRC = close(fdTarg);
	if (iCloseRC == -1)
	{
		ProxyLogMessage(mrUOW, MSG_DXR6079E, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		close(fdSrc);
		return FALSE;
	}
#endif

	//  Close the source database file
#ifdef WIN32
	if (!CloseHandle(hfSrc))
	{
		dwWinError = GetLastError();
		ProxyLogMessage(mrUOW, MSG_DXR6080E, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		return FALSE;
	}
#endif
#ifdef UNIX
	iCloseRC = close(fdSrc);
	if (iCloseRC == -1)
	{
		ProxyLogMessage(mrUOW, MSG_DXR6080E, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		return FALSE;
	}
#endif

	//  Show the completion of the copy
	lElapsed = (ct_End - ct_Start) / lTimeFactor;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR6112I, dwFSDisp, szGBFraction, szFDDispUnits, dwIOPs, lElapsed, dwCkpts);
	ProxyLogMessage(mrUOW, szMsg, iThreadID);

	return TRUE;
}

//  createDbLink
//
//  This function will create a DbLink File for the designated database pointing to it's new location
//
//  Parameters
//
//  1  -  MoveRequest *		-  Pointer to the Move Request objrct
//  2  -  char *			-  Pointer to the file name for the link file
//  3  -  char *			-  Pointer to the link target file name
//  4  -  int				-  Identity of the thread making the request
//
//  Returns
//
//  BOOL					-  TRUE if the link was created
//							-  FALSE if link could not be created
//
//  API Use Notes:
//
//
BOOL DbMover::createDbLink(MoveRequest *mrRQ, char *szDb, char *szLink, int iThreadID)
{
	HANDLE				hfTarg;									//  Handle of the target file
	DWORD				dwBytesToWrite;							//  Bytes to write
	DWORD				dwBytesWritten;							//  Bytes actually written
	DWORD				dwWinError;								//  Last windows erre
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer
#ifdef UNIX
	int					fdTarg;									//  File Descriptor for Target
	off_t				offSrc;									//  Source Offset
	off_t				offResult;								//  Resulting offset from lseek
	mode_t				mCreate;								//  File Create Mode
	ssize_t				stRead;									//  Size read/written
	int					iCloseRC;								//  Return Code from Close
#endif

	//  Create the target file
#ifdef WIN32
	hfTarg = CreateFile(szDb, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hfTarg == (HANDLE) HFILE_ERROR)
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6081E, szDb);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}
#endif
#ifdef UNIX
	mCreate = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
	fdTarg = open(szDb, O_RDWR | O_CREAT | O_EXCL, mCreate);
	if (fdTarg == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6081E, szDb);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}
#endif
	//  Format the link entry
	sprintf_s(szMsg, MAX_MSG, "%s\n", szLink);

	dwBytesToWrite = (DWORD) strlen(szMsg);

	//  Write the link
#ifdef WIN32
	if (!WriteFile(hfTarg, szMsg, dwBytesToWrite, &dwBytesWritten, NULL))
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6082E, dwBytesToWrite);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		CloseHandle(hfTarg);
		return FALSE;
	}
#endif
#ifdef UNIX
		stRead = write(fdTarg, szMsg, dwBytesToWrite);
		if (stRead == -1)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6082E, dwBytesToWrite);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetUNIXAPIError(errno, szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			close(fdTarg);
			return FALSE;
		}
		dwBytesWritten = stRead;
#endif

	//  Check the length written
	if (dwBytesWritten != dwBytesToWrite)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6083E, dwBytesToWrite, dwBytesWritten);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
#ifdef WIN32
		CloseHandle(hfTarg);
#endif
#ifdef UNIX
		close(fdTarg);
#endif
		return FALSE;
	}

	//  Close the link file
#ifdef WIN32
	if (!CloseHandle(hfTarg))
	{
		dwWinError = GetLastError();
		ProxyLogMessage(mrRQ, MSG_DXR6084E, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}
#endif
#ifdef UNIX
	iCloseRC = close(fdTarg);
	if (iCloseRC == -1)
	{
		ProxyLogMessage(mrRQ, MSG_DXR6084E, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}
#endif

	return TRUE;
}

//  createDirectory
//
//  This function will create the lowest level of directory node from the passed directory
//
//  Parameters
//
//  1  -  MoveRequest *		-  Pointer to the Move Request object
//  2  -  char *			-  Pointer to full directory path
//  3  -  int				-  Identity of the thread making the request
//
//  Returns
//
//  BOOL					-  TRUE if the directory existed or was created
//							-  FALSE if directory could not be created
//
//  API Use Notes:
//
//
BOOL DbMover::createDirectory(MoveRequest *mrRQ, char *szDir, int iThreadID)
{
	DWORD				dwWinError;								//  Last error code
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer
#ifdef UNIX
	int					iDirRC;									//  Return code from drectory functions
#endif

	//  Create the directory - return TRUE if it already exists
#ifdef WIN32
	if (!CreateDirectory(szDir, NULL))
	{
		dwWinError = GetLastError();
		if (dwWinError == ERROR_ALREADY_EXISTS) return TRUE;			//  Directory already exists
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6091E, szDir);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}
#endif
#ifdef UNIX
	iDirRC = mkdir(szDir, S_IRWXU | S_IRWXG | S_IRWXO);
	if (iDirRC == -1)
	{
		if ((errno == EEXIST) || (errno == EISDIR)) return TRUE;
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6091E, szDir);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}
#endif
	return TRUE;
}

//  ProxyLogMessage
//
//  The defulat implementation of this virtual function writes the logged messages to the normal runtime logger.
//
//  Parameters
//
//  1  -  MoveRequest *		-  Pointer to the Move REquest object
//  2  -  char *			-  Pointer to the message to be logged
//  3  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL					-  TRUE if the log action succeeded
//							-  FALSE if the log action failed
//
//  API Use Notes:
//
//
BOOL DbMover::ProxyLogMessage(MoveRequest *mrRQ, char *szMsg, int iThreadID)
{
	return xeLocal->LogMessage(szMsg, iThreadID);
}

//  checkCopy
//
//  This function will check the logic of the copy and determine if the input file points to a D/B link or
//  nothing (in which case it is within a Directory Link).
//
//  Parameters
//
//  1  -  MoveRequest *		-  Pointer to the Move Request object
//  2  -  char *			-  Pointer to the Input file name (computed)
//  3  -  char *			-  Pointer to the Output file name
//  4  -  char *			-  Pointer to the Link file name, filled by this function
//  5  -  DBHANDLE			-  Handle of the source database
//  6  -  int *				-  Configuration setting	(1 - source is database)
//														(2 - source is a database link)
//														(3 - source is within a directory link)
//  7  -  int				-  Identity of the thread making the request
//
//  Returns
//
//	BOOL		-	TRUE the move is sensible, FALSE the move is not logical.
//
//  API Use Notes:
//
//
BOOL DbMover::checkCopy(MoveRequest *mrRQ, char *szIn, char *szOut, char *szLink, DBHANDLE hdbSrc, int *Config, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	DWORD				dwHighPos;								//  High part of file offset
	DWORD				dwBytesRead;							//  Count of bytes read in
#ifdef WIN32
	HANDLE				hfSrc;									//  Handle of the source file
	DWORD				dwWinError;								//  Last error code
#endif
#ifdef UNIX
	int					fdSrc;									//  File Descriptor for Source
	off_t				offSrc;									//  Source Offset
	off_t				offResult;								//  Resulting offset from lseek
	mode_t				mCreate;								//  File Create Mode
	ssize_t				stRead;									//  Size read/written
	int					iCloseRC;								//  Return Code from Close
#endif
	char				*pChar;									//  Pointer into the D/B content
	char				cDBContent[2 * MAXPATH];				//  Buffer to read in D/B content
	char				szDbPath[MAXPATH + 1];					//  Actual D/B Path
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer
	
	//  Default set the link name to be the same as the database name
	strcpy_s(szLink, MAXPATH, szIn);
	*Config = DB_REDIRECT_NONE;

	//  Get the path to the database as determined by the Domino Server
	stAPIRC = NSFDbPathGet(hdbSrc, NULL, szDbPath);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(mrRQ, MSG_DXR6100E, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}

	//  Normalize the path 
	normalizePath(szDbPath);

	//  Check if the path reported by Notes is different
	if (strcmpi(szIn, szDbPath) != 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6101I, szIn);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6102I, szDbPath);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		strcpy_s(szIn, MAXPATH, szDbPath);
	}
	else return TRUE;

	//  Determine the type of redirection that is in effect for the source database
#ifdef WIN32
	//  Open the link file
	hfSrc = CreateFile(szLink, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hfSrc == (HANDLE) HFILE_ERROR)
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6103I, szLink);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		*Config = DB_REDIRECT_DIR;
		return TRUE;
	}
#endif
#ifdef UNIX
	//  Open the source file
	fdSrc = open(szLink, O_RDONLY | O_LARGEFILE);
	if (fdSrc == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6103I, szLink);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		*Config = DB_REDIRECT_DIR;
		return TRUE;
	}
#endif
	//  Move the file pointer to the start of the file
#ifdef WIN32
	dwHighPos = 0;
	if(SetFilePointer(hfSrc, 0, (PLONG) &dwHighPos, FILE_BEGIN) == MAXWORD)
	{
		dwWinError = GetLastError();
		ProxyLogMessage(mrRQ, MSG_DXR6073E, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		CloseHandle(hfSrc);
		return FALSE;
	}
#endif
#ifdef UNIX
	offSrc = 0;
	offResult = lseek(fdSrc, offSrc, SEEK_SET);
	if (offResult == -1)
	{
		ProxyLogMessage(mrRQ, MSG_DXR6073E, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		close(fdSrc);
		return FALSE;
	}
#endif

	//  Read the start of the D/B file to determine if it really is a D/B link
#ifdef WIN32
	if (!ReadFile(hfSrc, cDBContent, MAXPATH * 2, &dwBytesRead, NULL))
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6074E, MAXPATH * 2);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		CloseHandle(hfSrc);
		return FALSE;
	}
	CloseHandle(hfSrc);
#endif
#ifdef UNIX
	stRead = read(fdSrc, cDBContent, MAXPATH * 2);
	if (stRead == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6074E, MAXPATH * 2);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		close(fdSrc);
		return FALSE;
	}
	dwBytesRead = stRead;
	close(fdSrc);
#endif
	//  Check if the length read is less than 4 - which doesn't make much sense
	if (dwBytesRead < 4)
	{
		ProxyLogMessage(mrRQ, MSG_DXR6104E, iThreadID);
		return FALSE;
	}

	//  Check if the file has what appears to be a database (NSF) signature
	if (cDBContent[0] == '\x1a')
	{
		if (cDBContent[1] == '\0')
		{
			if (cDBContent[2] == '\0')
			{
				if (cDBContent[3] == '\x04')
				{
					ProxyLogMessage(mrRQ, MSG_DXR6105E, iThreadID);
					return FALSE;
				}
			}
		}
	}

	//  Terminate the database link value
	cDBContent[dwBytesRead] = '\0';
	pChar = cDBContent;
	while (*pChar >= ' ')
	{
		pChar++;
	}
	*pChar = '\0';
	normalizePath(cDBContent);

	//  Compare against the link value that we previously computed
	if (strcmpi(szIn, cDBContent) == 0)
	{
		*Config = DB_REDIRECT_DB;
		return TRUE;
	}

	//  Show that link database names did not match
	sprintf_s(szMsg, MAX_MSG, MSG_DXR6106E, szLink);
	ProxyLogMessage(mrRQ, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR6107E, cDBContent);
	ProxyLogMessage(mrRQ, szMsg, iThreadID);
	return FALSE;
}

//  normalizePath
//
//  This function will convert \ to / in the passed path string.
//
//  Parameters
//
//  1  -  char *			-  Pointer to the path string to be normalized
//
//  Returns
//
//
//  API Use Notes:
//
//
void DbMover::normalizePath(char *szIn)
{
	char		*pChar;								//  Pointer into the string

	pChar = szIn;
	while (*pChar != '\0')
	{
		if (*pChar == '\\') *pChar = '/';
		pChar++;
	}
	return;
}

//  ExecuteThisRequest
//
//  This is the entrypoint used by the multi-threading kernel to trigger processing in different threadpool threads.
//  This is a wrapper for the PerformMoveAction() function.
//
//  Parameters
//
//  1  -  void *			-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbMover::ExecuteThisRequest(void *pReqObject, int iThreadID)
{
	PartMoveRequest		*pmrCurrent;								//  Dispatched Request
	char				szMsg[MAX_MSG + 1];							//  Generic Message Buffer

	//  Cast the Part Copy Request
	pmrCurrent = (PartMoveRequest *) pReqObject;

	//  Trace
	sprintf_s(szMsg, MAX_MSG, MSG_DXR6116T, pmrCurrent, pmrCurrent->MoveAction, pmrCurrent->Parent);
	xeLocal->LogTrace(TRACE_DBM, szMsg, iThreadID);

	//  If this is a Move Database request - then route the contained request to the top level executor
	if (pmrCurrent->MoveAction == MOVE_DATABASE) MoveThisDb(pmrCurrent->Parent, iThreadID);
	else PerformMoveAction(pmrCurrent, iThreadID);
	return;
}

//  PerformmoveAction
//
//  This is the main routing function for all multi-threaded invocations, the function will vector the current
//  request to the appropriate executing functions determined by the Move Action defined in the request.
//
//  Parameters
//
//  1  -  PartMoveRequest *	-  Pointer to the PartMoveRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbMover::PerformMoveAction(PartMoveRequest *pmrqRQ, int iThreadID)
{
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	if (pmrqRQ == NULL) return;

	sprintf_s(szMsg, MAX_MSG, MSG_DXR6117T, pmrqRQ->MoveAction);
	xeLocal->LogTrace(TRACE_DBM, szMsg, iThreadID);

	//  Vector according to the action requested
	switch(pmrqRQ->MoveAction)
	{
	case MOVE_ARENA:
		PerformMoveArena(pmrqRQ, iThreadID);
		break;
	}
	return;
}

//  PerformmoveArena
//
//  This function is invoked to move a single arena of data from the source to the target file.
//  This action is invoked asynchronously by the CopyLargeOSFile() function.
//
//  Parameters
//
//  1  -  PartMoveRequest *	-  Pointer to the PartMoveRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbMover::PerformMoveArena(PartMoveRequest *pmrqRQ, int iThreadID)
{
	BYTE			*pCopyBuffer;							//  Pointer to the copy buffer
	int				iCopyOps;								//  Number of copy operations required
	int				iIndex;									//  Generic index
	DWORD			dwBytesToRead;							//  Number of bytes to be read
	DWORD			dwBytesRead;							//  Number of bytes read
	DWORD			dwBytesWritten;							//  Number of bytes written
#ifdef WIN32
	HANDLE			hfSrc;									//  Handle of the source file
	HANDLE			hfTarg;									//  Handle of the target file
	DWORD			dwWinError;								//  Last error code
#endif
#ifdef UNIX
	int				fdSrc;									//  File Descriptor for Source
	int				fdTarg;									//  File Descriptor for Target
	off_t			offSrc;									//  Source Offset
	off_t			offResult;								//  Resulting offset from lseek
	mode_t			mCreate;								//  File Create Mode
	ssize_t			stRead;									//  Size read/written
	ssize_t			stWrite;								//  Size read/written
	int				iCloseRC;								//  Return Code from Close
#endif
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Check that the local permit is still asserted
	if (!pmrqRQ->Parent->PermitRun)
	{
		pmrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Allocate a copy buffer 
	pCopyBuffer = (BYTE *) malloc(pmrqRQ->Parent->CopyBfrKb * BYTES_PER_KB);

	//  Compute the number of copy operations that are needed to fill the arena
	iCopyOps = pmrqRQ->ArenaSizeKb / pmrqRQ->Parent->CopyBfrKb;

	//  Obtain the local handles for the source and target files
#ifdef WIN32
	hfSrc = xeLocal->GetMappedFileHandle(pmrqRQ->hfNativeSource, pmrqRQ->szIn, pmrqRQ->dwSourceAccess, pmrqRQ->dwSourceSharing, iThreadID);
	hfTarg = xeLocal->GetMappedFileHandle(pmrqRQ->hfNativeTarget, pmrqRQ->szOut, pmrqRQ->dwTargetAccess, pmrqRQ->dwTargetSharing, iThreadID);
	if ((hfSrc == NULLHANDLE) || (hfTarg == NULLHANDLE))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6123E, pmrqRQ->ArenaNo);
		ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
		free(pCopyBuffer);
		pmrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}
#endif
#ifdef UNIX
	fdSrc = xeLocal->GetMappedFileHandle(pmrqRQ->fNativeSource, pmrqRQ->szIn, pmrqRQ->iSourceFlags, iThreadID);
	fdTarg = xeLocal->GetMappedFileHandle(pmrqRQ->fNativeTarget, pmrqRQ->szOut, pmrqRQ->iTargetFlags, iThreadID);
	if ((fdSrc == NULL) || (fdTarg == NULL))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6123E, pmrqRQ->ArenaNo);
		ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
		free(pCopyBuffer);
		pmrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}
#endif

	//  Position the source and target files to the start of the arena
#ifdef WIN32
	if(SetFilePointer(hfSrc, pmrqRQ->ArenaOffsetLo, (PLONG) &pmrqRQ->ArenaOffsetHi, FILE_BEGIN) == MAXWORD)
	{
		dwWinError = GetLastError();
		ProxyLogMessage(pmrqRQ->Parent, MSG_DXR6119E, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
		free(pCopyBuffer);
		pmrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}	
#endif
#ifdef UNIX
  //TODO:  
#endif
#ifdef WIN32
	if(SetFilePointer(hfTarg, pmrqRQ->ArenaOffsetLo, (PLONG) &pmrqRQ->ArenaOffsetHi, FILE_BEGIN) == MAXWORD)
	{
		dwWinError = GetLastError();
		ProxyLogMessage(pmrqRQ->Parent, MSG_DXR6119E, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
		free(pCopyBuffer);
		pmrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}	
#endif
#ifdef UNIX
  //TODO:  
#endif
	dwBytesToRead = pmrqRQ->Parent->CopyBfrKb * BYTES_PER_KB;

	//  Loop performing the copy ops for the completed arena
	for(iIndex = 0; iIndex < iCopyOps; iIndex++)
	{
		//  Read the next (Buffer Size) chunk of the arena
#ifdef WIN32
		if (!ReadFile(hfSrc, pCopyBuffer, dwBytesToRead, &dwBytesRead, NULL))
		{
			dwWinError = GetLastError();
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6074E, dwBytesToRead);
			ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwWinError, szMsg);
			ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
			free(pCopyBuffer);
			pmrqRQ->CompletionStatus = ACTION_FAILED;
			return;
		}
#endif
#ifdef UNIX
		stRead = read(fdSrc, pCopyBuffer, dwBytesToRead);
		if (stRead == -1)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6074E, dwBytesToRead);
			ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
			xeLocal->GetUNIXAPIError(errno, szMsg);
			ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
			free(pCopyBuffer);
			pmrqRQ->CompletionStatus = ACTION_FAILED;
			return;
		}
		dwBytesRead = stRead;
#endif

		//  Write the data buffer to the target file
#ifdef WIN32
		if (!WriteFile(hfTarg, pCopyBuffer, dwBytesRead, &dwBytesWritten, NULL))
		{
			dwWinError = GetLastError();
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6077E, dwBytesRead);
			ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwWinError, szMsg);
			ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
			free(pCopyBuffer);
			pmrqRQ->CompletionStatus = ACTION_FAILED;
			return;
		}
#endif
#ifdef UNIX
		stRead = write(fdTarg, pCopyBuffer, dwBytesRead);
		if (stRead == -1)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6077E, dwBytesRead);
			ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
			xeLocal->GetUNIXAPIError(errno, szMsg);
			ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
			free(pCopyBuffer);
			pmrqRQ->CompletionStatus = ACTION_FAILED;
			return;
		}
		dwBytesWritten = stRead;
#endif

		//  Check the bytes written - it must equal the bytes read but could be shorter than a
		//  complete copy buffer. If this was a short block then this is the end-of-file on the
		//  source.
		if (dwBytesWritten != dwBytesRead)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6078E, dwBytesRead, dwBytesWritten);
			ProxyLogMessage(pmrqRQ->Parent, szMsg, iThreadID);
			free(pCopyBuffer);
			pmrqRQ->CompletionStatus = ACTION_FAILED;
			return;
		}
		if (dwBytesWritten < dwBytesToRead)
		{
			//  Trace the short block write
			sprintf_s(szMsg, MAX_MSG, MSG_DXR6124T, dwBytesWritten, pmrqRQ->ArenaNo);
			xeLocal->LogTrace(TRACE_DBM, szMsg, iThreadID);
			free(pCopyBuffer);
			pmrqRQ->CompletionStatus = ACTION_COMPLETED;
			return;
		}
	}

	//  Mark the request completed
	pmrqRQ->CompletionStatus = ACTION_COMPLETED;

	//  Free the copy buffer
	free(pCopyBuffer);
	return;
}

//  copyLargeOSFile
//
//  This function will copy the source file to the target notifying progress to the Backup API as it proceeds.
//  The function is invoked to perform Asynch (MTIO) and is used for copying large files.
//
//  Parameters
//
//  1  -  MoveRequest *		-  Pointer to the top level move request
//  2  -  char *			-  Pointer to the file name to be copied
//  3  -  char *			-  Pointer to the target file name
//  4  -  DBHANDLE			-  DBHANDLE of the source database
//  5  -  HANDLE			-  HANDLE of the backup context
//  6  -  DWORD				-  Size of the file (High DWORD)
//  7  -  DWORD				-  Size of the file (Low DWORD)
//  8  -  int				-  Identity of the thread making the request
//
//  Returns
//
//  BOOL					-  TRUE if the file was copied
//							-  FALSE if the file could not be copied
//
//  API Use Notes:
//
//
BOOL DbMover::copyLargeOSFile(MoveRequest *mrUOW, char *szIn, char *szOut, DBHANDLE hdbSrc, DHANDLE hBUContext, DWORD SizeHi, DWORD SizeLo, int iThreadID)
{
	STATUS			stAPIRC;										//  Return Code from Notes API calls
	PartMoveRequest *pmrRQ;											//  Part Move Request
	BYTE			*pArenaTable;									//  Pointer to the Arena Table
	DWORD			dwArenaSizeMb;									//  Arena Size for the copy operations
	DWORD			dwArenaSizeBytes;								//  Arena Size for offset calculations
	DWORD			dwArenaCount;									//  Count of Arenas to be copied
	DWORD			dwArenaOffHi;									//  High part of the arena offset
	DWORD			dwArenaOffLo;									//  Low part of the arena offset
	DWORD			dwFSMb;											//  File Size in Mb
	DWORD			dwIndex;										//  Generic index
	DWORD			dwHighestCompletedArena;						//  Highest completed arena
	DWORD			dwLastCompletedArena;							//  Last completed arena
	DWORD			dwCkpts;										//  Count of checkpoints taken
	int				iRejoinRC;										//  Return code from rejoin calls
	clock_t			ct_Start;										//  Start copy time
	clock_t			ct_End;											//  End copy time
	long			lElapsed;										//  Elapsed time
	long			lTimeFactor = 1000;								//  Clock tick conversion factor
#ifdef WIN32
	HANDLE			hfSrc;											//  Handle of the source file
	HANDLE			hfTarg;											//  Handle of the target file
	DWORD			dwWinError;										//  Last error code
#endif
#ifdef UNIX
	int				fdSrc;											//  File Descriptor for Source
	int				fdTarg;											//  File Descriptor for Target
	off_t			offSrc;											//  Source Offset
	off_t			offResult;										//  Resulting offset from lseek
	mode_t			mCreate;										//  File Create Mode
	ssize_t			stWrite;											//  Size read/written
	int				iCloseRC;										//  Return Code from Close
#endif
	char			szMsg[MAX_MSG + 1];								//  Generic Message Buffer


	//  Compute the Arena Size and number of arenas to be copied
	dwArenaSizeMb = mrUOW->ArenaSize;

	if (dwArenaSizeMb < 2)
	{
		//  Use a 2Mb arena if total file size >= 8Gb otherwise 1Mb
		if (SizeHi < 2) dwArenaSizeMb = 1;							
		else dwArenaSizeMb = 2;
	}
	dwArenaSizeBytes = dwArenaSizeMb * KB_PER_MB;
	dwArenaSizeBytes = dwArenaSizeBytes * BYTES_PER_KB;
	dwFSMb = SizeLo / BYTES_PER_KB;
	dwFSMb = dwFSMb / KB_PER_MB;
	dwFSMb = dwFSMb + (SizeHi * 4 * MB_PER_GB);
	dwArenaCount = dwFSMb / dwArenaSizeMb;
	dwArenaCount++;

	//  Show that the copy will be performed using the Large OS File Copier
	sprintf_s(szMsg, MAX_MSG, MSG_DXR6118I, dwArenaCount, dwArenaSizeMb);
	ProxyLogMessage(mrUOW, szMsg, iThreadID);

#ifdef WIN32
	//  Open the source file
	hfSrc = CreateFile(szIn, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hfSrc == (HANDLE) HFILE_ERROR)
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6070E, szIn);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		return FALSE;
	}
#endif
#ifdef UNIX
	//  Open the source file
	fdSrc = open(szIn, O_RDONLY | O_LARGEFILE);
	if (fdSrc == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6070E, szIn);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		return FALSE;
	}
#endif

	//  Create the target file
#ifdef WIN32
	hfTarg = CreateFile(szOut, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hfTarg == (HANDLE) HFILE_ERROR)
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6071E, szOut);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		CloseHandle(hfSrc);
		return FALSE;
	}
#endif

#ifdef UNIX
	mCreate = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
	fdTarg = open(szOut, O_RDWR | O_CREAT | O_EXCL | O_LARGEFILE, mCreate);
	if (fdTarg == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR6071E, szOut);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		ProxyLogMessage(mrUOW, szMsg, iThreadID);
		close(fdSrc);
		return FALSE;
	}
#endif

	//  Allocate and initialise an Arena Table
	pArenaTable = (BYTE *) malloc(ARENA_TABLE_SIZE);
	memset(pArenaTable, 0, ARENA_TABLE_SIZE);

	//  Setup ready to start creating new arenas in the target file
	dwArenaOffHi = 0;
	dwArenaOffLo = 0;
	dwHighestCompletedArena = 0;
	dwCkpts = 0;

	//  Record the starting time
	ct_Start = xeLocal->RunningTime->getElapsedMillis();

	//  Create each arena in turn and dispatch a request to fill it
	for (dwIndex = 0; dwIndex < dwArenaCount; dwIndex++)
	{
		//  Check that we still have the Run Permit asserted
		if (!mrUOW->PermitRun)
		{
			xeLocal->LogTrace(TRACE_DBM, MSG_DXR6125T, iThreadID);
			DrainRequests(mrUOW, iThreadID);
#ifdef WIN32
			xeLocal->InvalidateNativeFileHandle(hfSrc, iThreadID);
			CloseHandle(hfSrc);
			xeLocal->InvalidateNativeFileHandle(hfTarg, iThreadID);
			CloseHandle(hfTarg);
#endif
#ifdef UNIX
			xeLocal->InvalidateNativeFileHandle(fdSrc, iThreadID);
			close(fdSrc);
			xeLocal->InvalidateNativeFileHandle(fdTarg, iThreadID);
			close(fdTarg);
#endif
			free(pArenaTable);
			return FALSE;
		}

		//  Create a new Part Move Request for the Arena Copy
		pmrRQ = new PartMoveRequest(mrUOW);
		pmrRQ->dbmMover = this;
		pmrRQ->MoveAction = MOVE_ARENA;
		pmrRQ->ArenaNo = dwIndex + 1;
		pmrRQ->ArenaOffsetHi = dwArenaOffHi;
		pmrRQ->ArenaOffsetLo = dwArenaOffLo;
		pmrRQ->ArenaSizeKb = dwArenaSizeMb * KB_PER_MB;
		pmrRQ->szIn = szIn;
		pmrRQ->szOut = szOut;
#ifdef WIN32
		pmrRQ->hfNativeSource = hfSrc;
		pmrRQ->hfNativeTarget = hfTarg;
		pmrRQ->dwSourceAccess = GENERIC_READ;
		pmrRQ->dwSourceSharing = FILE_SHARE_READ|FILE_SHARE_WRITE;
		pmrRQ->dwTargetAccess = GENERIC_READ|GENERIC_WRITE;
		pmrRQ->dwTargetSharing = FILE_SHARE_READ|FILE_SHARE_WRITE;
#endif
#ifdef UNIX
		pmrRQ->fNativeSource = fdSrc;
		pmrRQ->fNativeTarget = fdTarg;
		pmrRQ->iSourceFlags = O_RDONLY | O_LARGEFILE;
		pmrRQ->iTargetFlags = O_RDWR | O_EXCL | O_LARGEFILE;
#endif
		//  Calculate the offset to the end of the arena
        if ((MAXDWORD - dwArenaOffLo) < dwArenaSizeBytes)
        {
            dwArenaOffHi++;
            dwArenaOffLo = (dwArenaSizeBytes - (MAXDWORD - dwArenaOffLo)) - 1;
        }
        else
        {
            dwArenaOffLo += dwArenaSizeBytes;
        }

		//  Position to the end of the arena and set the EndOfFile
		//  Except on the last arena
		if (dwIndex < (dwArenaCount - 1))
		{
#ifdef WIN32
			if(SetFilePointer(hfTarg, dwArenaOffLo, (PLONG) &dwArenaOffHi, FILE_BEGIN) == MAXWORD)
			{
				dwWinError = GetLastError();
				ProxyLogMessage(mrUOW, MSG_DXR6119E, iThreadID);
				xeLocal->GetWINAPIError(dwWinError, szMsg);
				ProxyLogMessage(mrUOW, szMsg, iThreadID);
				xeLocal->InvalidateNativeFileHandle(hfSrc, iThreadID);
				CloseHandle(hfSrc);
				xeLocal->InvalidateNativeFileHandle(hfTarg, iThreadID);
				CloseHandle(hfTarg);
				free(pArenaTable);
				return FALSE;
			}
			if (!SetEndOfFile(hfTarg))
			{
				dwWinError = GetLastError();
				ProxyLogMessage(mrUOW, MSG_DXR6120E, iThreadID);
				xeLocal->GetWINAPIError(dwWinError, szMsg);
				ProxyLogMessage(mrUOW, szMsg, iThreadID);
				xeLocal->InvalidateNativeFileHandle(hfSrc, iThreadID);
				CloseHandle(hfSrc);
				xeLocal->InvalidateNativeFileHandle(hfTarg, iThreadID);
				CloseHandle(hfTarg);
				free(pArenaTable);
				return FALSE;
			}
#endif
#ifdef UNIX
			offResult = lseek(fdTarg, dwArenaSizeBytes, SEEK_CUR);
			if (offResult == -1)
			{
				ProxyLogMessage(mrUOW, MSG_DXR6119E, iThreadID);
				xeLocal->GetUNIXAPIError(errno, szMsg);
				ProxyLogMessage(mrUOW, szMsg, iThreadID);
				xeLocal->InvalidateNativeFileHandle(fdSrc, iThreadID);
				close(fdSrc);
				xeLocal->InvalidateNativeFileHandle(fdTarg, iThreadID);
				close(fdTarg);
				free(pArenaTable);
				return FALSE;
			}
			stWrite = write(fdTarg, pArenaTable, 0);
			if (stWrite == -1)
			{
				ProxyLogMessage(mrUOW, MSG_DXR6120E, iThreadID);
				xeLocal->GetUNIXAPIError(errno, szMsg);
				ProxyLogMessage(mrUOW, szMsg, iThreadID);
				xeLocal->InvalidateNativeFileHandle(fdSrc, iThreadID);
				close(fdSrc);
				xeLocal->InvalidateNativeFileHandle(fdTarg, iThreadID);
				close(fdTarg);
				free(pArenaTable);
				return FALSE;
			}
#endif
		}
		//  Post the current Arena Copy request for execution
		if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
									mrUOW,
									this,
									pmrRQ,
									RQATTR_REJOIN,
									20,
									iThreadID))
		{
			//  Should Not Occur (SNO) Error - as the call should block in the case of any resource shortage
			ProxyLogMessage(mrUOW, MSG_DXR6121E, iThreadID);
#ifdef WIN32
			xeLocal->InvalidateNativeFileHandle(hfSrc, iThreadID);
			CloseHandle(hfSrc);
			xeLocal->InvalidateNativeFileHandle(hfTarg, iThreadID);
			CloseHandle(hfTarg);
#endif
#ifdef UNIX
			xeLocal->InvalidateNativeFileHandle(fdSrc, iThreadID);
			close(fdSrc);
			xeLocal->InvalidateNativeFileHandle(fdTarg, iThreadID);
			close(fdTarg);
#endif
			free(pArenaTable);
			return FALSE;
		}

		//  Poll for any completed Arena Copy requests
		while(xeLocal->GetRejoinRequest(mrUOW, (void **) &pmrRQ, iThreadID) == RJR_RETURNED)
		{
			//  Mark the arena entry as complete and determine if there is a new high watermark
			dwLastCompletedArena = markArenaCompleted(pmrRQ->ArenaNo, pArenaTable, ARENA_TABLE_SIZE);
			if (dwLastCompletedArena > dwHighestCompletedArena)
			{
				//  Notify the backup context that the high watermark has moved
				stAPIRC = NSFBackupSetHighWaterMark(hdbSrc, hBUContext, pmrRQ->ArenaOffsetLo, pmrRQ->ArenaOffsetHi);
				if (ERR(stAPIRC) != NOERROR)
				{
					ProxyLogMessage(mrUOW, MSG_DXR6076E, iThreadID);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
					ProxyLogMessage(mrUOW, szMsg, iThreadID);
#ifdef WIN32
					xeLocal->InvalidateNativeFileHandle(hfSrc, iThreadID);
					CloseHandle(hfSrc);
					xeLocal->InvalidateNativeFileHandle(hfTarg, iThreadID);
					CloseHandle(hfTarg);
#endif
#ifdef UNIX
					xeLocal->InvalidateNativeFileHandle(fdSrc, iThreadID);
					close(fdSrc);
					xeLocal->InvalidateNativeFileHandle(fdTarg, iThreadID);
					close(fdTarg);
#endif
					free(pArenaTable);
					return FALSE;
				}
				dwCkpts++;											//  Increment the checkpoint count
			}
			delete pmrRQ;
		}
	}

	//  All arena copy requests have been dispatched - process them as they are completed
	iRejoinRC = xeLocal->GetRejoinRequest(mrUOW, (void **) &pmrRQ, iThreadID);
	while (iRejoinRC != RJR_NONE_EXIST)
	{
		if (iRejoinRC == RJR_RETURNED)
		{
			//  Mark the arena entry as complete and determine if there is a new high watermark
			dwLastCompletedArena = markArenaCompleted(pmrRQ->ArenaNo, pArenaTable, ARENA_TABLE_SIZE);
			if (dwLastCompletedArena > dwHighestCompletedArena)
			{
				//  Notify the backup context that the high watermark has moved
				stAPIRC = NSFBackupSetHighWaterMark(hdbSrc, hBUContext, pmrRQ->ArenaOffsetLo, pmrRQ->ArenaOffsetHi);
				if (ERR(stAPIRC) != NOERROR)
				{
					ProxyLogMessage(mrUOW, MSG_DXR6076E, iThreadID);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
					ProxyLogMessage(mrUOW, szMsg, iThreadID);
#ifdef WIN32
					xeLocal->InvalidateNativeFileHandle(hfSrc, iThreadID);
					CloseHandle(hfSrc);
					xeLocal->InvalidateNativeFileHandle(hfTarg, iThreadID);
					CloseHandle(hfTarg);
#endif
#ifdef UNIX
					xeLocal->InvalidateNativeFileHandle(fdSrc, iThreadID);
					close(fdSrc);
					xeLocal->InvalidateNativeFileHandle(fdTarg, iThreadID);
					close(fdTarg);
#endif
					free(pArenaTable);
					return FALSE;
				}
				dwCkpts++;											//  Increment the checkpoint count
			}
			delete pmrRQ;
		}
		else xeLocal->AppWait(REJOIN_WAIT);
		iRejoinRC = xeLocal->GetRejoinRequest(mrUOW, (void **) &pmrRQ, iThreadID);
	}

	//  Record the ending time
	ct_End = xeLocal->RunningTime->getElapsedMillis();

	//  Invalidate and close the source and target file handles
#ifdef WIN32
	xeLocal->InvalidateNativeFileHandle(hfSrc, iThreadID);
	CloseHandle(hfSrc);
	xeLocal->InvalidateNativeFileHandle(hfTarg, iThreadID);
	CloseHandle(hfTarg);
#endif
#ifdef UNIX
	xeLocal->InvalidateNativeFileHandle(fdSrc, iThreadID);
	close(fdSrc);
	xeLocal->InvalidateNativeFileHandle(fdTarg, iThreadID);
	close(fdTarg);
#endif

	//  Free the arena table
	free(pArenaTable);

	//  Show the completion of the copy
	lElapsed = (ct_End - ct_Start) / lTimeFactor;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR6122I, dwArenaCount, dwArenaSizeMb, lElapsed, dwCkpts);
	ProxyLogMessage(mrUOW, szMsg, iThreadID);

	return TRUE;
}

//   markArenaCompleted 
//
//   This function will mark the indicated Arena as complete and return the highest consecutive completed arena.
//
//   Parameters
//
//   1  -  DWORD		-  The Arena number to ba marked complete
//   2  -  uChar *		-  Pointer to the Arena Table
//   3  -  int			-  Size of the arena table in bytes
//
//   returns
//
//   DWORD		-	The highest arena number completed
//					-1 if the specified arena could not be marked complete
//
//   API Usage notes:
//
DWORD DbMover::markArenaCompleted(DWORD Arena, BYTE *pTable, int iTableSize)
{
	int iByteIndex = 0;
	int iIndex;
	int iHighArena = 0;
	BYTE ucMask = 0x80;
	
		
	iByteIndex = Arena >> 3;
	iIndex = Arena & 0x0007;

	if (iByteIndex > iTableSize) return -1;

	ucMask = ucMask >> iIndex;			//  Shift the mask bit by the remainder

	//  Mark entry complete
	pTable[iByteIndex] = pTable[iByteIndex] | ucMask;

	//  Find the highest consecutive arena that is now complete
	iIndex = 0;
	while ((pTable[iByteIndex] == (UCHAR) '\xff') && (iIndex < iTableSize))
	{
		iIndex++;
	}

	if (iIndex == iTableSize) return -1;

	ucMask = pTable[iByteIndex];
	iHighArena = iIndex * 8;
	while (ucMask & '\x80')
	{
		iHighArena++;
		ucMask = ucMask << 1;
	}

	return iHighArena;
}

//  DrainRequests
//
//  This function will drain and discard all requests that are on the asynchronous execution queue
//
//  Parameters
//
//  1  -  MoveRequest * -  Pointer to he top level move request
//  2  -  int			-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbMover::DrainRequests(MoveRequest *mrRQ, int iThreadID)
{
	PartMoveRequest		*pmrqRQ;							//  Part Move (Copy Arena) request
	int					iRejoinStatus;						//  Status from Rejoin 

	iRejoinStatus = xeLocal->GetRejoinRequest(mrRQ, (void **) &pmrqRQ, iThreadID);
	while (iRejoinStatus != RJR_NONE_EXIST)
	{
		if (iRejoinStatus == RJR_RETURNED)
		{
			//  Clean up the request if it is valid
			if (pmrqRQ != NULL)
			{
				delete pmrqRQ;
				pmrqRQ = NULL;	
			}
		}
		else
		{
			xeLocal->AppWait(PXR_BUSY_WAIT);
		}
		iRejoinStatus = xeLocal->GetRejoinRequest(mrRQ, (void **) &pmrqRQ, iThreadID);
	}
	return;
}
