//  DbModifier
//
//  This class provides the Database Modifier functions used in Domono eXplorer (DX) utilities
//
//  Message Identifier:      DXR
//  Message Range:           7501 - 7699
//  Trace area assignment:   TRACE_DBP (29)
//  DX Version:				 3.11.1
//  Version:                 3.11.1  Build: 99
//
//  Version History:
//
//  3.10.3		13/12/2010		Initial version
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2011
//

//  Application Includes
#include "DbModifier.h"

//  Default Constructor
DbModifier::DbModifier(void)
{
}

//  Constructor
DbModifier::DbModifier(MTExecutive *xeParent, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Save the Runtime Environment
	xeLocal = xeParent;
	//  Save the Creating Thread ID
	OwningThread = iThreadID;

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7501T, this);
	xeLocal->LogTrace(TRACE_DBPMOD, szMsg, OwningThread);
}

//  Destructor
DbModifier::~DbModifier(void)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7502T, this);
	xeLocal->LogTrace(TRACE_DBPMOD, szMsg, OwningThread);
}

//  ModifyThisDb
//
//  This is the main entry point to request modification processing.
//
//  Parameters
//
//  1  -  ModRequest *		-  Pointer to the ModRequest for this database
//  2  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  BOOL  -			TRUE if the database was processed successfully
//                  FALSE if the database could not be modified
//
//  API Use Notes:
//
//
BOOL DbModifier::ModifyThisDb(ModRequest *mrRQ, int iThreadID)
{
	DBHANDLE			hdbSrc;								//  Local Handle to the database
	BOOL				bUpdates = FALSE;					//  Indicate if updates were made
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Check that the passed request actually contains something useful to do
	if (mrRQ->dwPropsToMod == 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7503W, mrRQ->szTransID);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		mrRQ->ReturnCode = RETURN_NOERROR;
		return TRUE;
	}

	//  If we were not passed a Database Handle for the request then open the database
	if (mrRQ->hDBSrc == NULLHANDLE)
	{
		hdbSrc = xeLocal->OpenDatabase(mrRQ->szSourceServer, mrRQ->szSourceDatabase, iThreadID);
		if (hdbSrc == NULLHANDLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7504E, mrRQ->szSourceServer, mrRQ->szSourceDatabase);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			mrRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7505I, mrRQ->szSourceDatabase, mrRQ->szSourceServer);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
	}
	else hdbSrc = mrRQ->hDBSrc;

	//  Trace starting the operations
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7526T, mrRQ->dwPropsToMod, mrRQ->dwPropSetting);
	xeLocal->LogTrace(TRACE_DBPMOD, szMsg, iThreadID);

	//  Determine if any of the D/B Options need to be modified
	if ((mrRQ->dwPropsToMod & MOD_MASK_OPTS) != 0)
	{
		//  Modify the D/B Options
		bUpdates = bUpdates | ModifyDbOptions(mrRQ, hdbSrc, iThreadID);
	}

	//  Determine if the replication settings (flags) are to be modified
	if ((mrRQ->dwPropsToMod & MOD_MASK_REPF) != 0)
	{
		//  Modify the Replication Settings
		bUpdates = bUpdates | ModifyRepSettings(mrRQ, hdbSrc, iThreadID);
	}

	//  Determine if the Maximum Internet Access is to be modified
	if (mrRQ->dwPropsToMod & MOD_PROP_MIA)
	{
		bUpdates = bUpdates | ModifyMaxInternetAccess(mrRQ, hdbSrc, iThreadID);
	}

	//  Determine if the Database Quota is to be modified
	if (mrRQ->dwPropsToMod & MOD_PROP_QUOTA)
	{
		bUpdates = bUpdates | ModifyDbQuota(mrRQ, hdbSrc, iThreadID);
	}

	//  Determine if the Design Template needs to be modified
	if (mrRQ->dwPropsToMod & MOD_PROP_TPLT)
	{
		bUpdates = bUpdates | ModifyTemplate(mrRQ, hdbSrc, iThreadID);
	}

	//  Determine if Transaction Loging needs to be modified
	//  NOTE: This should be the last entry in the train of modifiers as it will force closed the DB Handle
	if (mrRQ->dwPropsToMod & MOD_PROP_TXL)
	{
		bUpdates = bUpdates | ModifyTXLogging(mrRQ, hdbSrc, iThreadID);
		hdbSrc = NULLHANDLE;
	}

	//  Show if any updates were applied by this transactions
	if (bUpdates)
	{
		ProxyLogMessage(mrRQ, MSG_DXR7506I, iThreadID);
	}

	//  Close the source database if it was opened locally
	if (mrRQ->hDBSrc == NULLHANDLE)
	{
		if (hdbSrc != NULLHANDLE) NSFDbClose(hdbSrc);
	}

	mrRQ->ReturnCode = RETURN_NOERROR;
	return TRUE;
}

//  ModifyDbOptions
//
//  This function will modify settings in the DBOPTIONS flags
//
//  Parameters
//
//  1  -  ModRequest *		-  Pointer to the ModRequest for this database
//  2  -  DBHANDLE			-  Handle of the database to be modified
//  3  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  BOOL  -			TRUE if the database was processed successfully
//                  FALSE if the database could not be modified
//
//  API Use Notes:
//
//
BOOL DbModifier::ModifyDbOptions(ModRequest *mrRQ, DBHANDLE hdbSrc, int iThreadID)
{
	STATUS		stAPIRC;									//  Return code from Notes API calls
	DWORD		dwDbOptions;								//  D/B Option flags
	DWORD		dwDbOptionsB4;								//  Original settings of the options flags
	BOOL		bUpdatesMade = FALSE;						//  Show if any flags were updated
	BOOL		bEvalFlag;									//  Evaluator for flag
	BOOL		bEvalSetting;								//  Evaluator for setting
	char		szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Get the current Db Options
	stAPIRC = NSFDbGetOptions(hdbSrc, &dwDbOptions);		//  Read the current options settings
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		ProxyLogMessage(mrRQ, MSG_DXR7507E, iThreadID);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}

	dwDbOptionsB4 = dwDbOptions;

	//  Determine if the Consistent ACL setting is to be modified (potentially)
	if (mrRQ->dwPropsToMod & MOD_PROP_CACL)
	{
		if (dwDbOptions & DBOPTION_UNIFORM_ACCESS) bEvalFlag = TRUE;
		else bEvalFlag = FALSE;
		if (mrRQ->dwPropSetting & MOD_PROP_CACL) bEvalSetting = TRUE;
		else bEvalSetting = FALSE;
		if (bEvalFlag != bEvalSetting)
		{
			dwDbOptions = dwDbOptions ^ DBOPTION_UNIFORM_ACCESS;
			bUpdatesMade = TRUE;
		}
	}

	//  Determine if the Contains Calendar Entries setting is to be modified (potentially)
	if (mrRQ->dwPropsToMod & MOD_PROP_CAL)
	{
		if (dwDbOptions & DBOPTION_HAS_CALENDAR) bEvalFlag = TRUE;
		else bEvalFlag = FALSE;
		if (mrRQ->dwPropSetting & MOD_PROP_CAL) bEvalSetting = TRUE;
		else bEvalSetting = FALSE;
		if (bEvalFlag != bEvalSetting)
		{
			dwDbOptions = dwDbOptions ^ DBOPTION_HAS_CALENDAR;
			bUpdatesMade = TRUE;
		}
	}

	//  Determine if the Full Text Indexed setting is to be modified (potentially)
	if (mrRQ->dwPropsToMod & MOD_PROP_FTI)
	{
		if (dwDbOptions & DBOPTION_FT_INDEX) bEvalFlag = TRUE;
		else bEvalFlag = FALSE;
		if (mrRQ->dwPropSetting & MOD_PROP_FTI) bEvalSetting = TRUE;
		else bEvalSetting = FALSE;
		if (bEvalFlag != bEvalSetting)
		{
			dwDbOptions = dwDbOptions ^ DBOPTION_FT_INDEX;
			bUpdatesMade = TRUE;
		}
	}

	//  If any settings were changed then set the options
	if (bUpdatesMade)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7527I, dwDbOptionsB4, dwDbOptions);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		stAPIRC = NSFDbSetOptions(hdbSrc, dwDbOptions, DBOPT_SET_MASK);
		if (ERR(stAPIRC))
		{
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, MSG_DXR7508E, iThreadID);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			return FALSE;
		}
	}
	else
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7528I, dwDbOptionsB4);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
	}

	return bUpdatesMade;
}

//  ModifyRepSettings
//
//  This function will modify the replication settings
//
//  Parameters
//
//  1  -  ModRequest *		-  Pointer to the ModRequest for this database
//  2  -  DBHANDLE			-  Handle of the database to be modified
//  3  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  BOOL  -			TRUE if the database was processed successfully
//                  FALSE if the database could not be modified
//
//  API Use Notes:
//
//
BOOL DbModifier::ModifyRepSettings(ModRequest *mrRQ, DBHANDLE hdbSrc, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	BOOL			bUpdatesMade = FALSE;						//  Show if any flags were updated
	BOOL			bEvalFlag;									//  Evaluator for flag
	BOOL			bEvalSetting;								//  Evaluator for setting
	DBREPLICAINFO	riLocal;									//  Replication Settings
	WORD			wFlagsB4;									//  Original Flag settings						
	char			szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Get the current replication settings from the database
	stAPIRC = NSFDbReplicaInfoGet(hdbSrc, &riLocal);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		ProxyLogMessage(mrRQ, MSG_DXR7509E, iThreadID);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}

	wFlagsB4 = riLocal.Flags;

	//  Determine if the Temporarily Disable Replication setting is to be modified (potentially)
	if (mrRQ->dwPropsToMod & MOD_PROP_REPT)
	{
		if (riLocal.Flags & REPLFLG_DISABLE) bEvalFlag = TRUE;
		else bEvalFlag = FALSE;
		if (mrRQ->dwPropSetting & MOD_PROP_REPT) bEvalSetting = TRUE;
		else bEvalSetting = FALSE;
		if (bEvalFlag != bEvalSetting)
		{
			riLocal.Flags = riLocal.Flags ^ REPLFLG_DISABLE;
			bUpdatesMade = TRUE;
		}
	}

	//  Determine if the Permanently Disable Replication setting is to be modified (potentially)
	if (mrRQ->dwPropsToMod & MOD_PROP_REPP)
	{
		if (riLocal.Flags & REPLFLG_NEVER_REPLICATE) bEvalFlag = TRUE;
		else bEvalFlag = FALSE;
		if (mrRQ->dwPropSetting & MOD_PROP_REPP) bEvalSetting = TRUE;
		else bEvalSetting = FALSE;
		if (bEvalFlag != bEvalSetting)
		{
			riLocal.Flags = riLocal.Flags ^ REPLFLG_NEVER_REPLICATE;
			bUpdatesMade = TRUE;
		}
	}

	//  Determine if we need to set/unset the Diable background agents setting
	if (mrRQ->dwPropsToMod & MOD_PROP_BGA)
	{
		if (riLocal.Flags & REPLFLG_NO_CHRONOS) bEvalFlag = TRUE;
		else bEvalFlag = FALSE;
		if (mrRQ->dwPropSetting & MOD_PROP_BGA) bEvalSetting = TRUE;
		else bEvalSetting = FALSE;
		if (bEvalFlag != bEvalSetting)
		{
			riLocal.Flags = riLocal.Flags ^ REPLFLG_NO_CHRONOS;
			bUpdatesMade = TRUE;
		}
	}

	//  Determine if the Do Not Catalog setting is to be modified (potentially)
	if (mrRQ->dwPropsToMod & MOD_PROP_CAT)
	{
		if (riLocal.Flags & REPLFLG_DO_NOT_CATALOG) bEvalFlag = TRUE;
		else bEvalFlag = FALSE;
		if (mrRQ->dwPropSetting & MOD_PROP_CAT) bEvalSetting = TRUE;
		else bEvalSetting = FALSE;
		if (bEvalFlag != bEvalSetting)
		{
			riLocal.Flags = riLocal.Flags ^ REPLFLG_DO_NOT_CATALOG;
			bUpdatesMade = TRUE;
		}
	}

	//  Determine if the No Browsing setting is to be modified (potentially)
	if (mrRQ->dwPropsToMod & MOD_PROP_BROWSE)
	{
		if (riLocal.Flags & REPLFLG_DO_NOT_BROWSE) bEvalFlag = TRUE;
		else bEvalFlag = FALSE;
		if (mrRQ->dwPropSetting & MOD_PROP_BROWSE) bEvalSetting = TRUE;
		else bEvalSetting = FALSE;
		if (bEvalFlag != bEvalSetting)
		{
			riLocal.Flags = riLocal.Flags ^ REPLFLG_DO_NOT_BROWSE;
			bUpdatesMade = TRUE;
		}
	}

	//  If any updates were made the set the replication settings
	if (bUpdatesMade)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7529I, wFlagsB4, riLocal.Flags);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		stAPIRC = NSFDbReplicaInfoSet(hdbSrc, &riLocal);
		if (ERR(stAPIRC) != NOERROR)
		{
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, MSG_DXR7510E, iThreadID);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			return FALSE;
		}
	}
	else
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7530I, wFlagsB4);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
	}

	return bUpdatesMade;
}

//  ModifyMaxInternetAccess
//
//  This function will modify the Max Internet Access level in the database
//
//  Parameters
//
//  1  -  ModRequest *		-  Pointer to the ModRequest for this database
//  2  -  DBHANDLE			-  Handle of the database to be modified
//  3  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  BOOL  -			TRUE if the database was processed successfully
//                  FALSE if the database could not be modified
//
//  API Use Notes:
//
//
BOOL DbModifier::ModifyMaxInternetAccess(ModRequest *mrRQ, DBHANDLE hdbSrc, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	BOOL			bUpdatesMade = FALSE;						//  Show if any flags were updated
	WORD			wAccessLevel;								//  Internet Access Level
	WORD			wAccessLevelB4;								//  Original access level
	char			szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Get the current Access Level
	stAPIRC = NSFGetMaxPasswordAccess(hdbSrc, &wAccessLevel);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		ProxyLogMessage(mrRQ, MSG_DXR7511E, iThreadID);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}

	wAccessLevelB4 = wAccessLevel;

	//  Change the access level if the request specifies a different level
	if (wAccessLevel != mrRQ->MaxInternetAccess)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7531I, wAccessLevelB4, mrRQ->MaxInternetAccess);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		stAPIRC = NSFSetMaxPasswordAccess(hdbSrc, mrRQ->MaxInternetAccess);
		if (ERR(stAPIRC) != NOERROR)
		{
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, MSG_DXR7512E, iThreadID);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			return FALSE;
		}
		bUpdatesMade = TRUE;
	}
	else
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7532I, wAccessLevelB4);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
	}

	return bUpdatesMade;
}

//  ModifyDbQuota
//
//  This function will modify the Database Quota Settings
//
//  Parameters
//
//  1  -  ModRequest *		-  Pointer to the ModRequest for this database
//  2  -  DBHANDLE			-  Handle of the database to be modified
//  3  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  BOOL  -			TRUE if the database was processed successfully
//                  FALSE if the database could not be modified
//
//  API Use Notes:
//
//
BOOL DbModifier::ModifyDbQuota(ModRequest *mrRQ, DBHANDLE hdbSrc, int iThreadID)
{
	STATUS			stAPIRC;									//  Return code from Notes API calls
	DBQUOTAINFO		qiLocal;									//  D/B Quota Info
	BOOL			bUpdatesMade = FALSE;						//  Show if any flags were updated
	WORD			wRetLen;									//  Returned length
	char			szDbPath[MAXPATH + 1];						//  Full D/B Path
	char			szCanonicalServer[MAXUSERNAME + 1];			//  Canonical Server Name
	char			szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Construct the string for the server/database
	if ((mrRQ->szSourceServer[0] == '\0') || (strcmpi(mrRQ->szSourceServer, "Local") == 0))
	{
		strcpy_s(szDbPath, MAX_DATABASE, mrRQ->szSourceDatabase);
	}
	else
	{
		//  canonicalize the server name and build the NET path for the database
		stAPIRC = DNCanonicalize(0L, NULL,mrRQ->szSourceServer, szCanonicalServer, MAXUSERNAME, &wRetLen);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7514E, mrRQ->szSourceServer);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			return FALSE;
		}
		szCanonicalServer[wRetLen] = '\0';
		stAPIRC = OSPathNetConstruct(NULL, szCanonicalServer, mrRQ->szSourceDatabase, szDbPath);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7515E, mrRQ->szSourceDatabase, szCanonicalServer);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			return FALSE;
		}	
	}

	//  Read the current settings of the D/B Quota for the database
	stAPIRC = NSFDbQuotaGet(szDbPath, &qiLocal);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		ProxyLogMessage(mrRQ, MSG_DXR7513E, iThreadID);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}

	//  Check if the warning threshold needs to be updated
	if (mrRQ->QuotaWarn != -1)
	{
		if (mrRQ->QuotaWarn != qiLocal.WarningThreshold)
		{
			qiLocal.WarningThreshold = mrRQ->QuotaWarn;
			bUpdatesMade = TRUE;
		}
	}

	//  Check if the max size needs to be updated
	if (mrRQ->QuotaLimit != -1)
	{
		if (mrRQ->QuotaLimit != qiLocal.SizeLimit)
		{
			qiLocal.SizeLimit = mrRQ->QuotaLimit;
			bUpdatesMade = TRUE;
		}
	}

	//  If updates were made then write the updated settings back to the database
	if (bUpdatesMade)
	{
		qiLocal.CurrentDbSize = DBQUOTA_NOT_SPECIFIED;
		qiLocal.MaxDbSize = DBQUOTA_NOT_SPECIFIED;

		//  Update the settings
		stAPIRC = NSFDbQuotaSetExt(szDbPath, 0, &qiLocal);
		if (ERR(stAPIRC) != NOERROR)
		{
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, MSG_DXR7516E, iThreadID);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			return FALSE;
		}
	}

	return bUpdatesMade;
}

//  ModifyTemplate
//
//  This function will modify the Template name for a database
//
//  Parameters
//
//  1  -  ModRequest *		-  Pointer to the ModRequest for this database
//  2  -  DBHANDLE			-  Handle of the database to be modified
//  3  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  BOOL  -			TRUE if the database was processed successfully
//                  FALSE if the database could not be modified
//
//  API Use Notes:
//
//
BOOL DbModifier::ModifyTemplate(ModRequest *mrRQ, DBHANDLE hdbSrc, int iThreadID)
{
	STATUS		stAPIRC;									//  Return code from Notes API calls
	BOOL		bUpdatesMade = FALSE;						//  Show if any flags were updated
	char		cDbInfo[NSF_INFO_SIZE];						//  D/B Info Buffer
	char		szOriginalTemplate[MAX_TEMPLATE + 1];		//  Original template name
	char		szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Get the D/B info Buffer
	stAPIRC = NSFDbInfoGet(hdbSrc, cDbInfo);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(mrRQ, MSG_DXR7517E, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		return FALSE;
	}

	//  Parse out the current template that the database inherits from
	NSFDbInfoParse(cDbInfo, INFOPARSE_DESIGN_CLASS, szOriginalTemplate, MAX_TEMPLATE);

	//  Determine if the template name needs to be changed
	if (strcmpi(mrRQ->szTemplateName, szOriginalTemplate) != 0)
	{
		NSFDbInfoModify(cDbInfo, INFOPARSE_DESIGN_CLASS, mrRQ->szTemplateName);

		//  Update the info in the database
		stAPIRC = NSFDbInfoSet(hdbSrc, cDbInfo);
		if (ERR(stAPIRC) != NOERROR)
		{
			ProxyLogMessage(mrRQ, MSG_DXR7518E, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			return FALSE;
		}

		bUpdatesMade = TRUE;

		//  Force a design refresh
		stAPIRC = DesignRefresh("", hdbSrc, DESIGN_FORCE, NULL, NULL);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7519W, mrRQ->szTemplateName);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
		else
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7520I, szOriginalTemplate, mrRQ->szTemplateName);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
		}
	}

	return bUpdatesMade;
}

//  ModifyTXLogging
//
//  This function will modify the Transaction Logging state for the database
//
//  Parameters
//
//  1  -  ModRequest *		-  Pointer to the ModRequest for this database
//  2  -  DBHANDLE			-  Handle of the database to be modified
//  3  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  BOOL  -			TRUE if the database was processed successfully
//                  FALSE if the database could not be modified
//
//  API Use Notes:
//
//  This modifier closes the passed file handle - so that compact can be invoked.
//
BOOL DbModifier::ModifyTXLogging(ModRequest *mrRQ, DBHANDLE hdbSrc, int iThreadID)
{
	STATUS		stAPIRC;									//  Return code from Notes API calls
	BOOL		bUpdatesMade = FALSE;						//  Show if any flags were updated
	BOOL		bDbIsLogged;								//  Current Logging State
	BOOL		bEvalSetting;								//  Evaluator for the setting
	UNID		uidLogID;									//  Log ID
	UNID		uidDBIID;									//  DBIID
	DWORD		dwLogExtent;								//  Log Extent
	OSPROG_DEF	ospProg;									//  Program invocation parm structure
	char		szOpts[MAX_OPTS + 1];						//  Compact Options
	char		szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Determine if the database is currently transaction logged
	stAPIRC = NSFDbGetLogInfo(hdbSrc, 0, &bDbIsLogged, &uidLogID, &uidDBIID, &dwLogExtent);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(mrRQ, MSG_DXR7521E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
		NSFDbClose(hdbSrc);
		return FALSE;
	}

	NSFDbClose(hdbSrc);

	//  Determine if the state should be changed
	if (mrRQ->dwPropSetting & MOD_PROP_TXL) bEvalSetting = TRUE;
	else bEvalSetting = FALSE;
	if (bEvalSetting != bDbIsLogged)
	{
		if (bDbIsLogged)
		{
			//  Database is currently transaction logged - set it off
			strcpy_s(szOpts, MAX_OPTS, OPT_TX_OFF);
		}
		else
		{
			//  Database is NOT currently transaction logged - set it on
			strcpy_s(szOpts, MAX_OPTS, OPT_TX_ON);
		}

		//  Format the command line
		ospProg.szWorkingDir[0] = '\0';
		strcpy_s(ospProg.szAppName, MAXPATH, COMPACT_APP);
		sprintf_s(ospProg.szAppCmdLine, MAXPATH, " %s %s", mrRQ->szSourceDatabase, szOpts);
		ospProg.dwFlags = 0;
		ospProg.dwReturnCode = 0;

		//  Perform the compact - invoking the task inline
#ifdef WIN32
		if (xeLocal->OSLoadSrvCommand(&ospProg, iThreadID) != RC_OSPROG_OK)
		{
			ProxyLogMessage(mrRQ, MSG_DXR7522E, iThreadID);
			return FALSE;
		}
#else
		OSGetExecutableDirectory(szMsg);
		if (szMsg[strlen(szMsg) - 1] == '/') szMsg[strlen(szMsg) - 1] = '\0';
		sprintf_s(ospProg.szAppName, MAXPATH, "%s/%s", szMsg, COMPACT_APP);
	    OSGetDataDirectory(ospProg.szWorkingDir);
		ospProg.dwReturnCode = OSLoadProgram(ospProg.szAppName, ospProg.szWorkingDir, ospProg.szAppCmdLine, OS_LOADPROG_BACKGROUND);
#endif
		//  Check the return code from the compact task
		if (ospProg.dwReturnCode != RETURN_NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7523E, ospProg.dwReturnCode, ospProg.dwReturnCode);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage((STATUS) ospProg.dwReturnCode, szMsg);
			ProxyLogMessage(mrRQ, szMsg, iThreadID);
			return FALSE;
		}

		if (mrRQ->dwPropSetting & MOD_PROP_TXL) sprintf_s(szMsg, MAX_MSG, MSG_DXR7524I, mrRQ->szSourceDatabase, TXT_TX_ON);
		else sprintf_s(szMsg, MAX_MSG, MSG_DXR7524I, mrRQ->szSourceDatabase, TXT_TX_OFF);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);

		bUpdatesMade = TRUE;
	}
	else
	{
		if (mrRQ->dwPropSetting & MOD_PROP_TXL) sprintf_s(szMsg, MAX_MSG, MSG_DXR7525I, mrRQ->szSourceDatabase, TXT_TX_ON);
		else sprintf_s(szMsg, MAX_MSG, MSG_DXR7525I, mrRQ->szSourceDatabase, TXT_TX_OFF);
		ProxyLogMessage(mrRQ, szMsg, iThreadID);
	}

	return bUpdatesMade;
}

//  ProxyLogMessage
//
//  The default implementation of this virtual function writes the logged messages to the normal runtime logger.
//
//  Parameters
//
//  1  -  ModRequest *		-  Pointer to the current modify request
//  1  -  char *			-  Pointer to the message to be logged
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL					-  TRUE if the log action succeeded
//							-  FALSE if the log action failed
//
//  API Use Notes:
//
//
BOOL DbModifier::ProxyLogMessage(ModRequest *mrRQ, char *szMsg, int iThreadID)
{
	return xeLocal->LogMessage(szMsg, iThreadID);
}

//  ExecuteThisRequest
//
//  This is the entrypoint used by the multi-threading kernel to trigger processing in different threadpool threads.
//  This is a wrapper for the ModifyThisDb() function.
//
//  Parameters
//
//  1  -  void *			-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbModifier::ExecuteThisRequest(void *pReqObject, int iThreadID)
{
	ModRequest			*mrCurrent;									//  Dispatched Request
	mrCurrent = (ModRequest *) pReqObject;
	ModifyThisDb(mrCurrent, iThreadID);
	return;
}


