//  DesignManager for DX based applications
//
//  This class provides the Design Management functions used in Domono eXplorer (DX) utilities
//
//  Message Identifier:      DXR
//  Message Range:           5500 - 5999
//  Trace area assignment:   TRACE_DM (24)
//  DX Version:				 3.14.0
//  Version:                 2.3.0  Build: 120
//
//  Version History:
//
//  28/03/2011		-  2.0.1		-  Updated for compatability with Resource Loader
//  20/06/2011		-  2.1.0		-  Tidy up for unix
//  22/10/2011		-  2.1.1		-  VS 2011 Updates
//  04/11/2011		-  2.2.0		-  Migration of D/B conditioning from kernel to Design Manager
//	21/01/2015		-  2.3.0		-  x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2001 - 2015
//

//  Application Includes
#include "DesignManager.h"

//  Constructor
DesignManager::DesignManager(ExecEnvironment *xeParent)
: dmRoot(NULL)
, xeLocal(NULL)
, rlLocal(NULL)
, dihImporter(NULLHANDLE)
, iSetsApplied(0)
, iElementsAdded(0)
, iElementsSkipped(0)
, iDMDesignImportOption(DXLIMPORTOPTION_REPLACE_ELSE_CREATE)
, bDMReplaceDbProperties(TRUE)
, bDMReplicaRequiredForReplaceOrUpdate(FALSE)
, iDMInputValidationOption(Xml_Validate_Never)
, bDMExitOnFirstFatalError(TRUE)
, iDMUnknownTokenLogOption(DXLLOGOPTION_IGNORE)
, bRepairDbErrors(TRUE)
{
	char			szMsg[MAX_MSG + 1];				//  Generic message buffer
	WORD			wBuildNo = 0;					//  Domino Build Number

	//  Constructor
	xeLocal = xeParent;								//  Store the Execution Environment

	//  Check the Domino Build
	if (xeLocal->hdbRepository != NULLHANDLE)
	{
		NSFDbGetBuildVersion(xeLocal->hdbRepository, &wBuildNo);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5556I, wBuildNo);
		xeLocal->LogMessage(szMsg);

		//  Warn if build is less than 7.0.3  (build: 266)
		if (wBuildNo < 266)
		{
			xeLocal->LogMessage(MSG_DXR5557W);
		}
	}

	//  Trace construction
	if (xeLocal->rsLocal->LogLevel > LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5501T, this);
		xeLocal->LogMessage(szMsg);
	}
}

//  Destructor
DesignManager::~DesignManager(void)
{
	char			szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Log the summay of activity
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5544I, iSetsApplied, iElementsAdded, iElementsSkipped);
	xeLocal->LogMessage(szMsg);

	//  Destroy (explicily) the Design manifest
	DestroyManifest();

	//  If a local resource manager has been created then destroy it
	if (rlLocal != NULL) delete rlLocal;

	//  If the DXL Importer exists then destroy it
	if (dihImporter != NULLHANDLE)
	{
		DXLDeleteImporter(dihImporter);
	}

	//  Trace destruction
	if (xeLocal->rsLocal->LogLevel > LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5502T, this);
		xeLocal->LogMessage(szMsg);
	}
}

//  LoadManifest
//
//  This function will load the designated Design Manifest into memory
//
//  Parameters
//
//  1  -  char *		-  Pointer to the URI of the manifest that is to be loaded
//
//  Returns
//
//  BOOL  -			TRUE if the manifest was loaded
//                  FALSE if the manifest could not be loaded
//
//  API Use Notes:
//
//
BOOL DesignManager::LoadManifest(char *szManifestID)
{
	MEMHANDLE		hManifest;						//  Handle of the manifest in-memory
	DWORD			dwManSize;						//  Size of the manifest in-memory
	char			*pManifest;						//  Pointer to the in core manifest
	char			szMsg[MAX_MSG + 1];				//  Generic message buffer

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5506I, szManifestID);
	xeLocal->LogMessage(szMsg);

	//  If the Resource Loader does not exist then create it
	if (rlLocal == NULL) rlLocal = new DXResourceLoader(xeLocal, 0);

	//  Call the Resource Loader to load the menifest into memory
	hManifest = rlLocal->loadThisResource(szManifestID, 0);
	if (hManifest == NULLHANDLE)
	{
		xeLocal->LogMessage(MSG_DXR5503E, 0);
		return FALSE;
	}

	//  Resize the allocated manifest to add terminator
	dwManSize = OSMemoryGetSize(hManifest);
	hManifest = xeLocal->resizeDominoMemory(hManifest, dwManSize + 2, 0);
	if (hManifest == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5639E, dwManSize + 2);
		xeLocal->LogMessage(szMsg, 0);
		return FALSE;
	}

	//  Lock the memory
	pManifest = (char *) OSMemoryLock(hManifest);

	//  Set the terminators
	pManifest[dwManSize] = '\0';
	dwManSize++;
	pManifest[dwManSize] = '\0';

	//  Pass the loaded resource to the manifest parser
	if (!ParseManifest(pManifest))
	{
		xeLocal->LogMessage(MSG_DXR5504E);
		OSMemoryUnlock(hManifest);
		OSMemoryFree(hManifest);
		hManifest = NULLHANDLE;
		dmRoot = NULL;
		return FALSE;
	}

	//  Free up the resource
	OSMemoryUnlock(hManifest);
	OSMemoryFree(hManifest);
	hManifest = NULLHANDLE;

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5507I, szManifestID);
	xeLocal->LogVerbose(szMsg);
	return TRUE;
}

//  DestroyManifest
//
//  This function will destroy the loaded manifest
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void DesignManager::DestroyManifest(void)
{
	//  Destroy the manifest
	if (dmRoot != NULL)
	{
		delete dmRoot;
		dmRoot = NULL;
	}
}

//  ParseManifest
//
//  This function will parse an XML Design Manifest into the Design Manager Objects
//
//  Parameters
//
//  1  -  char *		-  Pointer to the in memory XML Design Manifest
//
//  Returns
//
//  BOOL  -			TRUE if the manifest was parsed
//                  FALSE if the manifest could not be parsed
//
//  API Use Notes:
//
//
BOOL DesignManager::ParseManifest(char *pManifest)
{
	int					iState;						//  State of the parser FSM
	int					iIndex;						//  General index variable
	int					iSets;						//  Count of DesignSets
	int					iElements;					//  Count of DesignElements
	BOOL				bSetIsEmpty;				//  The current DesignSet is empty
	DesignSet			*dsCurrent;					//  Pointer to the current DesignSet
	DesignElement		*deCurrent;					//  Pointer to the current DesignElement
	DesignManifestLine	*dmlCurrent;				//  Pointer to the current parsed Design Manifest Line
	UCHAR				*pCurrentPosition;			//  Pointer into the manifest object
	UCHAR				*pTarget;					//  Pointer into the copy buffer
	char				szLine[MAX_LINE + 1];		//  Buffer containing a single line of the manifest
	char				szMsg[MAX_MSG + 1];			//  Generic message buffer

	//  Parse the Design Manifest object
	//  Although the object is native XML an ultralight in-line parser is used

	//  Setup the initial state
	dsCurrent = NULL;
	deCurrent = NULL;
	dmlCurrent = NULL;
	pCurrentPosition = (UCHAR *) pManifest;
	iState = PSTATE_NOT_STARTED;
	pTarget = (UCHAR *) szLine;
	*pTarget = '\0';
	iSets = 0;
	iElements = 0;

	while (*pCurrentPosition != '\0')
	{
		//  Test for end of line
		if ((*pCurrentPosition == '\x0d') || (*pCurrentPosition == '\x0a'))
		{
			*pTarget = '\0';						//  Terminate the string in the buffer
			//  Log the line if indicated
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5505T, szLine, iState);
			xeLocal->LogVerbose(szMsg);
		
			//  Construct a new parsed line and determine the action based on the current state
			dmlCurrent = new DesignManifestLine(szLine);
			//  Log the parsed entry
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5518T, dmlCurrent->szName, dmlCurrent->LineType);
			xeLocal->LogTrace(TRACE_DM, szMsg);

			if (dmlCurrent->LineType == DM_LINE_INVALID)
			{
				xeLocal->LogMessage(MSG_DXR5508E);
				delete dmlCurrent;
				return FALSE;
			}
			if (dmlCurrent->LineType == DM_LINE_UNDEFINED)
			{
				xeLocal->LogMessage(MSG_DXR5509E);
				delete dmlCurrent;
				return FALSE;
			}
			//  Process any non-ignorable lines
			if ((dmlCurrent->LineType != DM_LINE_STRUCT) && (dmlCurrent->LineType != DM_LINE_COMMENT))
			{
				switch(iState)
				{
				case PSTATE_NOT_STARTED:
					//  The only valid element in this state is the manifest definition
					if (dmlCurrent->LineType != DM_LINE_MANIFEST)
					{
						xeLocal->LogMessage(MSG_DXR5510E);
						delete dmlCurrent;
						return FALSE;
					}
					//  Create and populate the base manifest object
					dmRoot = new DesignManifest();
					strcpy_s(dmRoot->ManifestName, MAX_DM_MANIFEST, dmlCurrent->szName);
					strcpy_s(dmRoot->ManifestVersion, MAX_DM_VERSION, dmlCurrent->szVersion);
					iState = PSTATE_MANIFEST_DEFINED;
					break;

				case PSTATE_MANIFEST_DEFINED:
					//  The only valid element in this state is a DesignSet definition or the end of the manifest
					if ((dmlCurrent->LineType != DM_LINE_SET) && (dmlCurrent->LineType != DM_LINE_MANIFEST_END))
					{
						xeLocal->LogMessage(MSG_DXR5511E);
						delete dmlCurrent;
						return FALSE;
					}
					//  Manifset can be closed if some sets have been defined
					if (dmlCurrent->LineType == DM_LINE_MANIFEST_END)
					{
						if (iSets == 0)
						{
							xeLocal->LogMessage(MSG_DXR5516E);
							delete dmlCurrent;
							return FALSE;
						}
						//  show that the manifest has been completed
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5517I, iSets, iElements);
						xeLocal->LogMessage(szMsg);
						delete dmlCurrent;
						return TRUE;
					}

					//  Create and populate a new DesignSet and add it to the manifest
					dsCurrent = new DesignSet;
					strcpy_s(dsCurrent->SetName, MAX_DM_SET, dmlCurrent->szName);
					for (iIndex = 0; iIndex < MAX_DM_SETS; iIndex++)
					{
						if (dmRoot->dsList[iIndex] == NULL)
						{
							dmRoot->dsList[iIndex] = dsCurrent;
							break;
						}
					}
					iSets++;
					iState = PSTATE_SET_OPEN;
					bSetIsEmpty = TRUE;
					break;

				case PSTATE_SET_OPEN:
					//  Either a set can be closed or an element can be defined
					//  Note: an empty set cannot be defined
					if ((dmlCurrent->LineType != DM_LINE_ELEMENT) && (dmlCurrent->LineType != DM_LINE_SET_END))
					{
						xeLocal->LogMessage(MSG_DXR5512E);
						delete dmlCurrent;
						return FALSE;
					}
					//  Check for an empty design set
					if (dmlCurrent->LineType == DM_LINE_SET_END)
					{
						if (bSetIsEmpty)
						{
							xeLocal->LogMessage(MSG_DXR5513E);
							delete dmlCurrent;
							return FALSE;
						}
						//  Close the current design set
						dsCurrent = NULL;
						iState = PSTATE_MANIFEST_DEFINED;
						break;
					}
					//  Create a new DesignElement and populate it
					bSetIsEmpty = FALSE;
					deCurrent = new DesignElement();
					strcpy_s(deCurrent->ElementName, MAX_DM_ELEMENT, dmlCurrent->szName);
					deCurrent->ElementType = dmlCurrent->ElementType;
					if ( dmlCurrent->bSign ) deCurrent->DesignOptions |= DE_OPTION_SIGN_ELEMENT;
					if ( dmlCurrent->szUNID[0] != '\0' ) strcpy_s(deCurrent->ElementUNID, MAX_DM_UNID, dmlCurrent->szUNID);
					if ( dmlCurrent->bReqUNIDMatch ) deCurrent->DesignOptions |= DE_OPTION_REQ_MATCH;

					sprintf_s(szMsg, MAX_MSG, MSG_DXR5621T, deCurrent->ElementName, deCurrent->ElementType, dmlCurrent->szUNID, deCurrent->DesignOptions);
					xeLocal->LogTrace(TRACE_DM, szMsg);

					iState = PSTATE_ELEMENT_INDEF;
					break;

				case PSTATE_ELEMENT_INDEF:
					//  Element is being defined, the URL should be provided as the XML node value
					if (dmlCurrent->LineType != DM_LINE_VALUE)
					{
						xeLocal->LogMessage(MSG_DXR5514E);
						delete dmlCurrent;
						return FALSE;
					}
					//  Add the URL to the Element
					strcpy_s(deCurrent->DesignURL, MAX_DM_URL, dmlCurrent->szValue);
					iState = PSTATE_ELEMENT_URLDEF;
					break;

				case PSTATE_ELEMENT_URLDEF:
					//  The URL has been defined the Element definition should now be closed
					if (dmlCurrent->LineType != DM_LINE_ELEMENT_END)
					{
						xeLocal->LogMessage(MSG_DXR5515E);
						delete dmlCurrent;
						return FALSE;
					}
					//  Add the design element to the current DesignSet
					for (iIndex = 0; iIndex < MAX_DM_ELEMENTS; iIndex++)
					{
						if (dsCurrent->deList[iIndex] == NULL)
						{
							dsCurrent->deList[iIndex] = deCurrent;
							break;
						}
					}
					deCurrent = NULL;
					iState = PSTATE_SET_OPEN;
					iElements++;
					break;				
				}
			}

			//  Reset ready to process the next line
			delete dmlCurrent;
			pTarget = (UCHAR *) szLine;
			*pTarget = '\0';
			pCurrentPosition++;
		}

		if (*pCurrentPosition >= ' ')
		{
			*pTarget = *pCurrentPosition;
			pTarget++;
		}
		pCurrentPosition++;
	}

	return TRUE;
}

//  ApplyDesignSet
//
//  This function will apply a design set to the current repository database
//
//  Parameters
//
//  1  -  char *		-  Pointer to the Design Set name
//
//  Returns
//
//  BOOL  -			TRUE if the design set was applied
//                  FALSE if the design set could not be applied
//
//  API Use Notes:
//
//
BOOL DesignManager::ApplyDesignSet(char *szSetName)
{
	return CoreApplyDesignSet(szSetName, xeLocal->hdbRepository);
}

//  ApplyDesignSet
//
//  This function will apply a design set to the specified database
//
//  Parameters
//
//  1  -  char *		-  Pointer to the Design Set name
//  2  -  DBHANDLE		-  Handle of the database to which the design set is to be applied
//
//  Returns
//
//  BOOL  -			TRUE if the design set was applied
//                  FALSE if the design set could not be applied
//
//  API Use Notes:
//
//
BOOL DesignManager::ApplyDesignSet(char *szSetName, DBHANDLE hDB)
{
	return CoreApplyDesignSet(szSetName, hDB);
}

//  CoreApplyDesignSet
//
//  This function will apply a design set to the specified database
//
//  Parameters
//
//  1  -  char *		-  Pointer to the Design Set name
//  2  -  DBHANDLE		-  Handle of the database to which the design set is to be applied
//
//  Returns
//
//  BOOL  -			TRUE if the design set was applied
//                  FALSE if the design set could not be applied
//
//  API Use Notes:
//
//
BOOL DesignManager::CoreApplyDesignSet(char *szSetName, DBHANDLE hDB)
{
	STATUS			stAPIRC;								//  Return code from Notes API calls
	HCOLLECTION		hDesignColl;							//  Design collection handle
	int				iIndex;
	DesignSet		*dsCurrent;								//  Pointer to the requested Design Set
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Apply a Design Set to the repository

	//  If the design manager is inactive then return silently
	if (dmRoot == NULL) return TRUE;

	//  Log the fact that the design set has been requested
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5519I, szSetName);
	xeLocal->LogVerbose(szMsg);

	//  Open the design collection of the target database
	stAPIRC = NIFOpenCollection(hDB, 
								hDB, 
								(NOTE_ID_SPECIAL | NOTE_CLASS_DESIGN), 
								(WORD) OPEN_DO_NOT_CREATE, 
								NULLHANDLE, 
								&hDesignColl,
								NULL, 
								NULL, 
								NULL, 
								NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5613E);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);

		if (bRepairDbErrors && (ERR(stAPIRC) == ERR_COLLECTION_NOT_CREATED))
		{
			xeLocal->LogMessage(MSG_DXR5638W);
			stAPIRC = NIFOpenCollection(hDB, 
										hDB, 
										(NOTE_ID_SPECIAL | NOTE_CLASS_DESIGN), 
										0, 
										NULLHANDLE, 
										&hDesignColl,
										NULL, 
										NULL, 
										NULL, 
										NULL);
			if (ERR(stAPIRC) != NOERROR)
			{
				xeLocal->LogMessage(MSG_DXR5613E);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				return FALSE;
			}

		}
		else if (bRepairDbErrors && (ERR(stAPIRC) == ERR_SPECIAL_ID))
		{
			xeLocal->LogMessage(MSG_DXR5637W);

			if (!CopyDesignCollection(hDB, 0))
			{
				xeLocal->LogMessage(MSG_DXR5636E);
				return FALSE;
			}

			stAPIRC = NIFOpenCollection(hDB, 
										hDB, 
										(NOTE_ID_SPECIAL | NOTE_CLASS_DESIGN), 
										(WORD) OPEN_DO_NOT_CREATE, 
										NULLHANDLE, 
										&hDesignColl,
										NULL, 
										NULL, 
										NULL, 
										NULL);
			if (ERR(stAPIRC) != NOERROR)
			{
				xeLocal->LogMessage(MSG_DXR5613E);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				return FALSE;
			}
		}
		else
		{
			return FALSE;
		}
	}
	
	//  Locate the requested design set
	dsCurrent = NULL;
	dsCurrent = FindDesignSet(szSetName);
	if (dsCurrent == NULL)									//  Design set could not be found
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5520E, szSetName);
		xeLocal->LogMessage(szMsg);
		NIFCloseCollection(hDesignColl);
		return TRUE;
	}

	iSetsApplied++;

	//  Process each design element in the set in turn
	for (iIndex = 0; iIndex < MAX_DM_ELEMENTS; iIndex++)
	{
		if (dsCurrent->deList[iIndex] == NULL) break;
		if (!ApplyDesignElement(dsCurrent->deList[iIndex], hDB, hDesignColl))
		{
			NIFCloseCollection(hDesignColl);
			return FALSE;
		}
		if (dmRoot == NULL)
		{
			NIFCloseCollection(hDesignColl);	
			return FALSE;
		}
	}

	//  Close the design collection of the target database
	stAPIRC = NIFCloseCollection(hDesignColl);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5620E);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	//  Log the fact that the design set has been applied
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5521I, szSetName);
	xeLocal->LogVerbose(szMsg);
	return TRUE;
}

//  FindDesignSet
//
//  This function will locate the requested design set
//
//  Parameters
//
//  1  -  char *		-  Pointer to the Design Set name
//
//  Returns
//
//  DesignSet *			- Pointer to the requested DesignSet Object
//						  NULL if the Design Set could not be located
//
//  API Use Notes:
//
//
DesignSet * DesignManager::FindDesignSet(char *szSetName)
{
	int			iIndex;

	//  Locate the requested Design Set
	for (iIndex = 0; iIndex < MAX_DM_SETS; iIndex++)
	{
		if (dmRoot->dsList[iIndex] == NULL) return NULL;
		if (strcmp(dmRoot->dsList[iIndex]->SetName, szSetName) == 0) return dmRoot->dsList[iIndex];
	}
	return NULL;
}

//  ApplyDesignElement
//
//  This function will apply the designated design element to the specified database
//
//  Parameters
//
//  1  -  DesignElement *		-  Pointer to the Design Element to be applied
//  2  -  DBHANDLE				-  Handle of the database to which the design set is to be applied
//
//
//  Returns
//
//  BOOL  -			TRUE if the design set was applied
//                  FALSE if the design set could not be applied
//
//  API Use Notes:
//
//
BOOL DesignManager::ApplyDesignElement(DesignElement *deCurrent, DBHANDLE hDB, HCOLLECTION hDesignColl)
{
	STATUS				stAPIRC;				//  Return code from Notes API calls
	DXRESOURCEHANDLE	hdrDXLImp;				//  Resource handle for the imported DXL
	NOTEID				nidDesign;				//  NoteID of the design Note
	WORD				wNoteClass;				//  Class of Design Note to look for
	BOOL				bNoCheck;				//  Do not check for the existance of the element
	char				szMsg[MAX_MSG +1];		//  Generic message buffer

	//  Apply an individual design element to the repository

	//  Trace the request
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5522T, deCurrent->ElementName);
	xeLocal->LogTrace(TRACE_DM, szMsg);

	//  See if the Design Element is already in the database
	wNoteClass = NOTE_CLASS_ALL;				//  Set default to all design types

	if (deCurrent->ElementType == DM_ELEMENT_VIEW) wNoteClass = NOTE_CLASS_VIEW;
	if (deCurrent->ElementType == DM_ELEMENT_FORM) wNoteClass = NOTE_CLASS_FORM;
	if (deCurrent->ElementType == DM_ELEMENT_SFORM) wNoteClass = NOTE_CLASS_FORM;
	if (deCurrent->ElementType == DM_ELEMENT_AGENT) wNoteClass = NOTE_CLASS_FILTER;
	if (deCurrent->ElementType == DM_ELEMENT_SFIELD) wNoteClass = NOTE_CLASS_FIELD;
	if (deCurrent->ElementType == DM_ELEMENT_LIB) wNoteClass = NOTE_CLASS_FILTER;
	if (deCurrent->ElementType == DM_ELEMENT_FOLDER) wNoteClass = NOTE_CLASS_VIEW;
	if (deCurrent->ElementType == DM_ELEMENT_OUTLINE) wNoteClass = NOTE_CLASS_FILTER;
	if (deCurrent->ElementType == DM_ELEMENT_SACTION) wNoteClass = NOTE_CLASS_FORM;

	bNoCheck = FALSE;
	if (deCurrent->ElementType == DM_ELEMENT_DBICON) bNoCheck = TRUE;
	if (deCurrent->ElementType == DM_ELEMENT_HABOUT) bNoCheck = TRUE;
	if (deCurrent->ElementType == DM_ELEMENT_HUSING) bNoCheck = TRUE;

	//  Perform any required pre-processing for this element
	if ((deCurrent->ElementType == DM_ELEMENT_HABOUT) || (deCurrent->ElementType == DM_ELEMENT_HUSING))
	{
		if (!PreProcessElement(deCurrent, hDB))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5555W, deCurrent->ElementName);
			xeLocal->LogMessage(szMsg);
			return TRUE;
		}
	}

	if (!bNoCheck)
	{
		if (!FindDesignElementNote(deCurrent, wNoteClass, hDB, hDesignColl, &nidDesign))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5566E, deCurrent->ElementName, deCurrent->ElementUNID);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}

		if ((nidDesign == (NOTEID) NULL ) && (deCurrent->DesignOptions & DE_OPTION_REQ_MATCH))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5567E, deCurrent->ElementName, deCurrent->ElementUNID);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}

		if (nidDesign != (NOTEID) NULL)
		{
			//  If Design update is forced then delete the design element
			if (xeLocal->rsLocal->UpdateDesign)
			{
				stAPIRC = NSFNoteDelete(hDB, nidDesign, UPDATE_FORCE);
				if (ERR(stAPIRC) != NOERROR)
				{	
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5554E, deCurrent->ElementName);
					xeLocal->LogMessage(szMsg);
					xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
					xeLocal->LogMessage(szMsg);
					return FALSE;
				}
			}
			else
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5545I, deCurrent->ElementName);
				xeLocal->LogVerbose(szMsg);
				iElementsSkipped++;
				return TRUE;
			}
		}
	}

	//  Create the DXLImporter if it doesn't exist
	if (dihImporter == NULLHANDLE)
	{
		if (!CreateImporter())
		{
			xeLocal->LogMessage(MSG_DXR5522E);
			//  Destroy the Design Manifest to prevent further use
			delete dmRoot;
			dmRoot = NULL;
			return FALSE;
		}
	}

	//  Update the design of this element - call the DXL Importer to import the required element
	strcpy_s(szCurrentReadURL, MAX_DM_URL, deCurrent->DesignURL);

	//  Get a DXRESOURCEHANDLE to the DXL that will be imported
	hdrDXLImp = rlLocal->loadThisResourceOnDemand(deCurrent->DesignURL, 0);

	//  Perform DXLimport
	stAPIRC = DXLImport(dihImporter, rlLocal->WrapperXMLReader, hDB, hdrDXLImp);
	if (ERR(stAPIRC) != NOERROR)
	{
		//  Report the error and log the import log
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5525E, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);

		//  Dump the import errors
		DumpXMLImportErrors(deCurrent);

		//  Dispose of the DXRESOURCEHANDLE
		rlLocal->disposeResourceHandle(hdrDXLImp, 0);

		//  Destroy the DXL Importer
		DXLDeleteImporter(dihImporter);
		dihImporter = NULLHANDLE;

		//  Destroy the Design Manifest to prevent further use
		delete dmRoot;
		dmRoot = NULL;

		return FALSE;
	}

	//  Check to see if the import suffered errors
	if (DXLImportWasErrorLogged(dihImporter))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5573E, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);

		//  Dump the import errors
		DumpXMLImportErrors(deCurrent);

		//  Dispose of the DXRESOURCEHANDLE
		rlLocal->disposeResourceHandle(hdrDXLImp, 0);

		//  Destroy the DXL Importer
		DXLDeleteImporter(dihImporter);
		dihImporter = NULLHANDLE;
		//  Destroy the Design Manifest to prevent further use
		delete dmRoot;
		dmRoot = NULL;

		return FALSE;
	}

	//  Dispose of the DXRESOURCEHANDLE
	rlLocal->disposeResourceHandle(hdrDXLImp, 0);

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5568I, deCurrent->ElementName);
	xeLocal->LogMessage(szMsg);

	//  Check that what we imported now exists!
	nidDesign = 0;
	if (!bNoCheck)
	{
		if (!FindDesignElementNote(deCurrent, wNoteClass, hDB, hDesignColl, &nidDesign))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5561E, deCurrent->ElementName);
			xeLocal->LogMessage(szMsg);

			//  Destroy the DXL Importer
			DXLDeleteImporter(dihImporter);
			dihImporter = NULLHANDLE;

			//  Destroy the Design Manifest to prevent further use
			delete dmRoot;
			dmRoot = NULL;

			return FALSE;
		}

		if (nidDesign == (NOTEID) NULL)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5561E, deCurrent->ElementName);
			xeLocal->LogMessage(szMsg);

			//  Destroy the DXL Importer
			DXLDeleteImporter(dihImporter);
			dihImporter = NULLHANDLE;

			//  Destroy the Design Manifest to prevent further use
			delete dmRoot;
			dmRoot = NULL;

			return FALSE;
		}

		sprintf_s(szMsg, MAX_MSG, MSG_DXR5564T, deCurrent->ElementName, nidDesign);
		xeLocal->LogTrace(TRACE_DM, szMsg);
	}
	else nidDesign = (NOTEID) NULL;

	//  See if the type just imported requires post processing
	if (!PostProcessElement(deCurrent, wNoteClass, hDB, hDesignColl, nidDesign))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5546W, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);

		//  Destroy the DXL Importer
		DXLDeleteImporter(dihImporter);
		dihImporter = NULLHANDLE;

		//  Destroy the Design Manifest to prevent further use
		delete dmRoot;
		dmRoot = NULL;

		return FALSE;
	}

	iElementsAdded++;
	return TRUE;
}

//  FindDesignElementNote
//
//  This function will attempt to find the design note using the database design collection
//
//  Parameters
//
//  1		-	DesignElement *			- Pointer to the DesignElement object that is to be located
//  2		-	WORD					- NOTE_CLASS of the design elements that is to be locatred
//  3		-	DBHANDLE				- Handle of the database to search
//	4		-	HCOLLECTION				- Handle of the design collection in the database to be searched
//  5		-	NOTEID *				- Pointer to a NOTEID in which the Note ID of the design note will be returned
//
//  Returns
//
//  BOOL  -			TRUE if the requested design element was located
//                  FALSE if the design element could not be found
//
//  API Use Notes:
//
//
BOOL DesignManager::FindDesignElementNote(DesignElement *deCurrent, WORD wNoteClass, DBHANDLE hDB, HCOLLECTION hDesignColl, NOTEID *nidDesign)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	BOOL				bEntryMatch;
	BYTE				*pBuffer;
	COLLECTIONPOSITION	cpIndex;
	COLLECTIONDATA		*cdIndex;								//  Ponter to collection info
	DHANDLE				hBuffer;
	DHANDLE				hcdIndex;								//  Handle to collection info
	DWORD				dwEntriesFound;
	DWORD				dwIndex;
	DWORD				dwEntriesReturned;
	DWORD				dwDocCount;
	NOTEHANDLE			hnEntry;
	NOTEID				nidEntry;
	WORD				wEntryNoteClass;
	WORD				wRetLen;
	int					iIndex;
	int					iIndex2;
	char				szTitle[MAX_MSG + 1];
	char				szTitle2[MAX_MSG + 1];
	char				szEntryUNID[MAXUNIDSTRING + 1];
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	WORD				wTragetLen;
	BYTE				*pbyteSource;
	BYTE				*pbyteTarget;
	NLS_PINFO			pnlsInfo;
	NLS_STATUS			nlstAPI;
	char				szLMBCSElementName[MAX_DM_ELEMENT * 2];

	*nidDesign = (NOTEID) NULL;

	//  Convert element name to LMBCS to handle extended ASCII characters in the
	//  element name
	szLMBCSElementName[0] = '\0';

	pbyteSource = (BYTE *) deCurrent->ElementName;
	pbyteTarget = (BYTE *) szLMBCSElementName;

	pnlsInfo = NULL;
	pnlsInfo = OSGetLMBCSCLS();

	if (pnlsInfo == NULL)
	{
		xeLocal->LogMessage(MSG_DXR5628E);
		return FALSE;
	}

	wTragetLen = (sizeof(char) * ((MAX_DM_ELEMENT * 2) - 1));

	nlstAPI = NLS_translate(pbyteSource, 
							NLS_NULLTERM, 
							pbyteTarget, 
							&wTragetLen, 
							(WORD) (NLS_NULLTERMINATE | NLS_SOURCEISPLATFORM),
							pnlsInfo);
	if (nlstAPI != NLS_SUCCESS)
	{
		xeLocal->LogMessage(MSG_DXR5629E);
		return FALSE;
	}
	
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5630T, deCurrent->ElementName, szLMBCSElementName);
	xeLocal->LogTrace(TRACE_DM, szMsg);

	//  Get the latest state of the collection
	NIFUpdateCollection(hDesignColl);
	NIFGetCollectionData(hDesignColl, &hcdIndex);
	cdIndex = (COLLECTIONDATA *) OSLockObject(hcdIndex);
	dwDocCount = cdIndex->DocCount;
	OSUnlockObject(hcdIndex);
	OSMemFree(hcdIndex);

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5633T, dwDocCount);
	xeLocal->LogTrace(TRACE_DM, szMsg);

	//  Now try and find the element
	stAPIRC = NIFFindByName(hDesignColl, szLMBCSElementName, (FIND_PARTIAL | FIND_UPDATE_IF_NOT_FOUND | FIND_CASE_INSENSITIVE | FIND_ACCENT_INSENSITIVE), &cpIndex, &dwEntriesFound);
	if (ERR(stAPIRC) != NOERROR)
	{
		if (ERR(stAPIRC) != ERR_NOT_FOUND)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5614E, deCurrent->ElementName);
			xeLocal->LogMessage(szMsg);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}

		sprintf_s(szMsg, MAX_MSG, MSG_DXR5611W, deCurrent->ElementName, dwDocCount);
		xeLocal->LogMessage(szMsg);
		
		return TRUE;
	}

	if (dwEntriesFound == 0)
	{
		//  Get the latest state of the collection
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5611W, deCurrent->ElementName, dwDocCount);
		xeLocal->LogMessage(szMsg);
		return TRUE;
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5622T, dwEntriesFound);
	xeLocal->LogTrace(TRACE_DM, szMsg);

	hBuffer = NULLHANDLE;
	stAPIRC = NIFReadEntries(hDesignColl,
							&cpIndex,
							NAVIGATE_CURRENT,
							0,
							NAVIGATE_NEXT,
							dwEntriesFound,
							(WORD) (READ_MASK_NOTEID | READ_MASK_NOTECLASS),
							&hBuffer,
							NULL,
							NULL,
							&dwEntriesReturned,
							NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5615E, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	if (hBuffer == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5616E, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5623T, dwEntriesFound);
	xeLocal->LogTrace(TRACE_DM, szMsg);

	pBuffer = (BYTE *) OSLockObject(hBuffer);

	for (dwIndex = 0; dwIndex < dwEntriesReturned; dwIndex++)
	{
		nidEntry = *(NOTEID *) pBuffer;
		pBuffer += sizeof (NOTEID);
		wEntryNoteClass = *(WORD *) pBuffer;
		pBuffer += sizeof (WORD);
		bEntryMatch = FALSE;

		//  If the design element type matches, consider it as a candidate
		if (!((wNoteClass == NOTE_CLASS_ALL) || (wEntryNoteClass == wNoteClass)) )
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5634T, nidEntry);
			xeLocal->LogTrace(TRACE_DM, szMsg);
		}
		else
		{
			stAPIRC = NSFNoteOpen(hDB, nidEntry, (WORD) (OPEN_SUMMARY | OPEN_EXPAND), &hnEntry);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5617E, nidEntry, deCurrent->ElementName);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				OSUnlockObject(hBuffer);
				OSMemFree(hBuffer);
				return FALSE;
			}

			szTitle[0] = '\0';
			wRetLen = NSFItemGetTextListEntry(hnEntry, FIELD_TITLE, 0, szTitle, (WORD) MAX_MSG);
			szTitle[wRetLen] = '\0';

			sprintf_s(szMsg, MAX_MSG, MSG_DXR5624T, szTitle, nidEntry);
			xeLocal->LogTrace(TRACE_DM, szMsg);

			//  Check to see if there is a "|" in the title, in which case get everything
			//  to the left - Some design elements don't use a text list for the TITLE item!
			szTitle2[0] = '\0';
			iIndex2 = 0;
			for (iIndex = 0; iIndex < (int) strlen(szTitle); iIndex++)
			{
				if (szTitle[iIndex] == '|') break;
				else
				{
					szTitle2[iIndex2] = szTitle[iIndex];
					iIndex2++;
				}
			}
			szTitle2[iIndex2] = '\0';

			sprintf_s(szMsg, MAX_MSG, MSG_DXR5631T, szTitle2);
			xeLocal->LogTrace(TRACE_DM, szMsg);

			//  If the title(0) value matches the design element name, its still a candidate
			if (strcmpi(szTitle2, szLMBCSElementName) == 0)
			{
				if (deCurrent->ElementUNID[0] != '\0')
				{
					//  As we have a UNID, check to see if UNID matches
					szEntryUNID[0] = '\0';
					xeLocal->GetNoteUNID(hnEntry, NULL, szEntryUNID, MAXUNIDSTRING+1, FALSE, 0);
					
					if (strcmpi(deCurrent->ElementUNID, szEntryUNID) == 0) bEntryMatch = TRUE;
					else
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5635T, nidEntry);
						xeLocal->LogTrace(TRACE_DM, szMsg);
					}
				}
				else
				{
					//  We don't have a UNID to double check therefore we have a match
					bEntryMatch = TRUE;
				}

				if (bEntryMatch)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5619T, nidEntry, deCurrent->ElementName);
					xeLocal->LogTrace(TRACE_DM, szMsg);

					//  Check to see if we have already found an entry!
					if (*nidDesign != (NOTEID) NULL)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5618E, deCurrent->ElementName);
						xeLocal->LogMessage(szMsg);
						NSFNoteClose(hnEntry);
						OSUnlockObject(hBuffer);
						OSMemFree(hBuffer);
						*nidDesign = (NOTEID) NULL;
						return FALSE;
					}

					*nidDesign = nidEntry;
				}
			}

			stAPIRC = NSFNoteClose(hnEntry);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5617E, nidEntry, deCurrent->ElementName);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				OSUnlockObject(hBuffer);
				OSMemFree(hBuffer);
				return FALSE;
			}
		} 
	} 

	OSUnlockObject(hBuffer);
	OSMemFree(hBuffer);

	if (*nidDesign == (NOTEID) NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5611W, deCurrent->ElementName, dwDocCount);
		xeLocal->LogMessage(szMsg);
	}

	return TRUE;
}

//  CreateImporter
//
//  This function will ceate and configure a DXL Importer
//
//  Parameters
//
//
//  Returns
//
//  BOOL  -			TRUE if the imorter was created and configured
//                  FALSE if the importer couldnot be created
//
//  API Use Notes:
//
//
BOOL DesignManager::CreateImporter(void)
{
	STATUS					stAPIRC;						//  Return code from Notes API calls
	BOOL					bValue;							//  Set the option
	DXL_IMPORT_PROPERTY		dipDesign;						//  Design Import Property
	DXLIMPORTOPTION			dioOpts;						//  DXL Importer Options
	DXLLOGOPTION			dloOpts;						//  DXL log options
	Xml_Validation_Option   xvoOpts;						//  XML Validation Opts
	char					szMsg[MAX_MSG + 1];				//  Generic message buffer

	//  Create and condition a DXL Importer
	stAPIRC = DXLCreateImporter(&dihImporter);
	if (ERR(stAPIRC) != NOERROR)
	{
		//  Unable to create the importer
		xeLocal->LogMessage(MSG_DXR5523E);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	//  Set properties required for Design Manager use of the importer
	dipDesign = iDesignImportOption;
	dioOpts = (DXLIMPORTOPTION) iDMDesignImportOption;
	stAPIRC = DXLSetImporterProperty(dihImporter, dipDesign, &dioOpts);
	if (ERR(stAPIRC) != NOERROR)
	{
		//  Unable to create the importer
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5524E, dipDesign, dioOpts);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	dipDesign = iReplaceDbProperties;
	bValue = bDMReplaceDbProperties;
	stAPIRC = DXLSetImporterProperty(dihImporter, dipDesign, &bValue);
	if (ERR(stAPIRC) != NOERROR)
	{
		//  Unable to create the importer
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5524E, dipDesign, bValue);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	dipDesign = iReplicaRequiredForReplaceOrUpdate;
	bValue = bDMReplicaRequiredForReplaceOrUpdate;
	stAPIRC = DXLSetImporterProperty(dihImporter, dipDesign, &bValue);
	if (ERR(stAPIRC) != NOERROR)
	{
		//  Unable to create the importer
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5524E, dipDesign, bValue);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	dipDesign = iInputValidationOption;
	xvoOpts = (Xml_Validation_Option) iDMInputValidationOption;
	stAPIRC = DXLSetImporterProperty(dihImporter, dipDesign, &xvoOpts);
	if (ERR(stAPIRC) != NOERROR)
	{
		//  Unable to create the importer
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5524E, dipDesign, xvoOpts);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	dipDesign = iExitOnFirstFatalError;
	bValue = bDMExitOnFirstFatalError;
	stAPIRC = DXLSetImporterProperty(dihImporter, dipDesign, &bValue);
	if (ERR(stAPIRC) != NOERROR)
	{
		//  Unable to create the importer
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5524E, dipDesign, bValue);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	dipDesign = iUnknownTokenLogOption;
	dloOpts = (DXLLOGOPTION) iDMUnknownTokenLogOption;
	stAPIRC = DXLSetImporterProperty(dihImporter, dipDesign, &dloOpts);
	if (ERR(stAPIRC) != NOERROR)
	{
		//  Unable to create the importer
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5524E, dipDesign, dloOpts);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	return TRUE;
}

//  PostProcessElement
//
//  This function executes any necessary post processing after importing an individual Design Element
//
//  Parameters
//
//  1  -  DesignElement *		-  Pointer to the Design Element that has been applied
//  2  -  WORD					-  Note class of the design element being supplied to function
//  3  -  DBHANDLE				-  Handle of the database to which the design set is to be applied
//  4  -  HCOLLECTION			-  Handle of the design collection in the target database
//  5  -  NOTEID				-  NoteID of the design element that has been loaded
//
//
//  Returns
//
//  BOOL  -			TRUE if the post processing was completed
//                  FALSE if the post processing could not be completed
//
//  API Use Notes:
//
//
BOOL DesignManager::PostProcessElement(DesignElement *deCurrent, WORD wNoteClass, DBHANDLE hDB, HCOLLECTION hDesignColl, NOTEID nidDE)
{
	STATUS				stAPIRC;						//  Return code from Notes API calls
	BLOCKID				bidItem;						//  BlockID of the Item Header
	BLOCKID				bidValue;						//  BlockID of the Item Value
	DWORD				dwValueLen;						//  Value length
	NOTEHANDLE			hnCarrier;						//  Handle to the carrier note
	NOTEHANDLE			hnDesign;						//  Handle to the Design note
	NOTEID				nidIR;							//  NoteID of the image resource
	NOTEID				nidDesign;						//  NoteID of the design element
	WORD				wValueType;						//  Datatype of the value
	char				szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Perform Post Processing according to the Design Element type
	if (deCurrent->ElementType == DM_ELEMENT_DBICON)
	{
		//  If this was a DbICON element then the actual ICON data is transferred in a dummy image resource
		//  copy it into the icon note and delete the dummy image resource
		stAPIRC = NIFFindDesignNote(hDB, DUMMY_CARRIER_ELEMENT, NOTE_CLASS_ALL, &nidIR);
		if (ERR(stAPIRC) != NOERROR)
		{
			xeLocal->LogMessage(MSG_DXR5547E);
			return FALSE;
		}
		
		//  open the carrier note and the dbIcon note
		stAPIRC = NSFNoteOpen(hDB, nidIR, 0, &hnCarrier);
		if (ERR(stAPIRC) != NOERROR)
		{	
			xeLocal->LogMessage(MSG_DXR5548E);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}

		stAPIRC = NSFNoteOpenExt(hDB, NOTE_ID_SPECIAL | NOTE_CLASS_ICON, 0, &hnDesign);
		if (ERR(stAPIRC) != NOERROR)
		{	
			NSFNoteClose(hnCarrier);
			xeLocal->LogMessage(MSG_DXR5549E);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}

		//  Delete the Icon Bitmap item from the note
		stAPIRC = NSFItemDelete(hnDesign, DESK_ICON_ITEM, (WORD) strlen(DESK_ICON_ITEM));
		if (ERR(stAPIRC) != NOERROR)
		{	
			NSFNoteClose(hnDesign);
			NSFNoteClose(hnCarrier);
			xeLocal->LogMessage(MSG_DXR5550E);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}

		//  Get the BLOCKID of the Icon Bitmap item in the carrier document
		stAPIRC = NSFItemInfo(hnCarrier, DESK_ICON_ITEM, (WORD) strlen(DESK_ICON_ITEM), &bidItem, &wValueType, &bidValue, &dwValueLen);
		if (ERR(stAPIRC) != NOERROR)
		{	
			NSFNoteClose(hnDesign);
			NSFNoteClose(hnCarrier);
			xeLocal->LogMessage(MSG_DXR5551E);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}

		//  Copy the Icon Bitmap item from the carrier to the DbIcon note
		stAPIRC = NSFItemCopy(hnDesign, bidItem);
		if (ERR(stAPIRC) != NOERROR)
		{	
			NSFNoteClose(hnDesign);
			NSFNoteClose(hnCarrier);
			xeLocal->LogMessage(MSG_DXR5552E);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}

		//  Sign the design note
		if ((xeLocal->rsLocal->SignDesign) || (deCurrent->DesignOptions & DE_OPTION_SIGN_ELEMENT))
		{
			stAPIRC = NSFNoteSign(hnDesign);
			if (ERR(stAPIRC) != NOERROR)
			{	
				NSFNoteClose(hnDesign);
				NSFNoteClose(hnCarrier);
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5563E, "sign", deCurrent->ElementName, (NOTE_ID_SPECIAL | NOTE_CLASS_ICON));
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				return FALSE;
			}
			else
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5640I, deCurrent->ElementName, (NOTE_ID_SPECIAL | NOTE_CLASS_ICON), xeLocal->AbbreviatedUser);
				xeLocal->LogVerbose(szMsg);
			}
		}

		//  Save the design note
		stAPIRC = NSFNoteUpdate(hnDesign, UPDATE_FORCE);
		if (ERR(stAPIRC) != NOERROR)
		{	
			NSFNoteClose(hnDesign);
			NSFNoteClose(hnCarrier);
			xeLocal->LogMessage(MSG_DXR5552E);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}
		NSFNoteClose(hnDesign);
		NSFNoteClose(hnCarrier);

		//  Delete the carrier note
		stAPIRC = NSFNoteDelete(hDB, nidIR, UPDATE_FORCE);
		if (ERR(stAPIRC) != NOERROR)
		{	
			xeLocal->LogMessage(MSG_DXR5553E);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}
		return TRUE;
	}
	else
	{
		//  Determine the note id of the element
		if (deCurrent->ElementType == DM_ELEMENT_HABOUT)
		{
			nidDesign = (NOTE_ID_SPECIAL | NOTE_CLASS_INFO);
		}
		else if (deCurrent->ElementType == DM_ELEMENT_HUSING)
		{
			nidDesign = (NOTE_ID_SPECIAL | NOTE_CLASS_HELP);
		}
		else
		{
			if (nidDE == (NOTEID) NULL)
			{
				if (!FindDesignElementNote(deCurrent, wNoteClass, hDB, hDesignColl, &nidDesign))
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5561E, deCurrent->ElementName);
					xeLocal->LogMessage(szMsg);
					return FALSE;
				}

				if (nidDesign == (NOTEID) NULL)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5561E, deCurrent->ElementName);
					xeLocal->LogMessage(szMsg);
					return FALSE;
				}

				sprintf_s(szMsg, MAX_MSG, MSG_DXR5565T, deCurrent->ElementName, nidDesign);
				xeLocal->LogMessage(szMsg);
			}
			else nidDesign = nidDE;
		}

		//  Based on the element type apply any post import fixes
		if (deCurrent->ElementType == DM_ELEMENT_AGENT)
		{
			FettleAgents(hDB, nidDesign, 0);
		}

		if ((xeLocal->rsLocal->SignDesign) || (deCurrent->DesignOptions & DE_OPTION_SIGN_ELEMENT))
		{
			//  If we are opening a view, agent or library we should not use any flags!
			if ((deCurrent->ElementType == DM_ELEMENT_VIEW) || (deCurrent->ElementType == DM_ELEMENT_AGENT) || (deCurrent->ElementType == DM_ELEMENT_LIB))
			{
				stAPIRC = NSFNoteOpenExt(hDB, nidDesign, 0, &hnDesign);
			}
			else
			{
				stAPIRC = NSFNoteOpenExt(hDB, nidDesign, OPEN_EXPAND, &hnDesign);
			}
			if (ERR(stAPIRC) != NOERROR)
			{	
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5563E, "open", deCurrent->ElementName, nidDesign);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				return FALSE;
			}

			stAPIRC = NSFNoteSign(hnDesign);
			if (ERR(stAPIRC) != NOERROR)
			{	
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5563E, "sign", deCurrent->ElementName, nidDesign);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				NSFNoteClose(hnDesign);
				return FALSE;
			}
			else
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5640I, deCurrent->ElementName, nidDesign, xeLocal->AbbreviatedUser);
				xeLocal->LogVerbose(szMsg);
			}

			stAPIRC = NSFNoteUpdate(hnDesign, UPDATE_FORCE);
			if (ERR(stAPIRC) != NOERROR)
			{	
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5563E, "save", deCurrent->ElementName, nidDesign);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				NSFNoteClose(hnDesign);
				return FALSE;
			}

			stAPIRC = NSFNoteClose(hnDesign);
			if (ERR(stAPIRC) != NOERROR)
			{	
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5563E, "close", deCurrent->ElementName, nidDesign);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				return FALSE;
			}

		}

		return TRUE;
	}

	return TRUE;
}

//  PreProcessElement
//
//  This function executes any necessary preprocessing before importing an individual Design Element
//
//  Parameters
//
//  1  -  DesignElement *		-  Pointer to the Design Element to be applied
//  2  -  DBHANDLE				-  Handle of the database to which the design set is to be applied
//
//
//  Returns
//
//  BOOL  -			TRUE if the preprocessing was completed
//                  FALSE if the preprocessing could not be completed
//
//  API Use Notes:
//
//
BOOL DesignManager::PreProcessElement(DesignElement *deCurrent, DBHANDLE hDB)
{
	STATUS			stAPIRC;							//  Status code returend from Notes API calls
	NOTEID			nidTarget;							//  NoteID of the note to be processed
	char			szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Processing for help documents - delete the existing ones if they already exist

	if ((deCurrent->ElementType == DM_ELEMENT_HABOUT) || (deCurrent->ElementType == DM_ELEMENT_HUSING))
	{
		if (deCurrent->ElementType == DM_ELEMENT_HABOUT)
		{
			nidTarget = NOTE_ID_SPECIAL | NOTE_CLASS_INFO;
		}
		else
		{
			nidTarget = NOTE_ID_SPECIAL | NOTE_CLASS_HELP;
		}

		//  Delete the requested note
		stAPIRC = NSFNoteDelete(hDB, nidTarget, UPDATE_FORCE);
		if (ERR(stAPIRC) != NOERROR)
		{	
			if (ERR(stAPIRC) == ERR_SPECIAL_ID) return TRUE;
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5554E, deCurrent->ElementName);
			xeLocal->LogMessage(szMsg);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			return FALSE;
		}
		return TRUE;
	}
	return FALSE;
}

//  ApplyAllDesignSets
//
//  This function will apply each Design Set in the Design Manifest to the Repository database
//
//  Parameters
//
//
//  Returns
//
//  BOOL  -			TRUE if all Design Sets were applied
//                  FALSE if all Design Sets could not be applied
//
//  API Use Notes:
//
//
BOOL DesignManager::ApplyAllDesignSets(void)
{
	return CoreApplyAllDesignSets(xeLocal->hdbRepository);
}

//  ApplyAllDesignSets
//
//  This function will apply each Design Set in the Design Manifest to the specified database
//
//  Parameters
//
//  1  -  DBHANDLE				-  Handle of the database to which the design sets are to be applied
//
//
//  Returns
//
//  BOOL  -			TRUE if all Design Sets were applied
//                  FALSE if all Design Sets could not be applied
//
//  API Use Notes:
//
//
BOOL DesignManager::ApplyAllDesignSets(DBHANDLE hDB)
{
	return CoreApplyAllDesignSets(hDB);
}

//  CoreApplyAllDesignSets
//
//  This function will apply each Design Set in the Design Manifest to the specified database
//
//  Parameters
//
//  1  -  DBHANDLE				-  Handle of the database to which the design sets are to be applied
//
//
//  Returns
//
//  BOOL  -			TRUE if all Design Sets were applied
//                  FALSE if all Design Sets could not be applied
//
//  API Use Notes:
//
//
BOOL DesignManager::CoreApplyAllDesignSets(DBHANDLE hDB)
{
	int			iFailCount = 0;
	int			iIndex;

	//  Apply all design sets in the manifest
	for (iIndex = 0; iIndex < MAX_DM_SETS; iIndex++)
	{
		if (dmRoot->dsList[iIndex] == NULL) return(iFailCount == 0);
		if (!ApplyDesignSet(dmRoot->dsList[iIndex]->SetName, hDB))
		{
			iFailCount++; 
			return FALSE;
		}
		if (dmRoot == NULL) return FALSE;
	}
	return(iFailCount == 0);
}

//  GetBuildInfo_s
//
//  This function will return the Build information for the Design Manager
//
//  Parameters
//
//  1  -  char *		-  Pointer to the buffer where the build info will be stored
//  2  -  int			-  Length of the buffer where the build info will be stored
//
//  Returns
//
//  API Use Notes:
//
//
void DesignManager::GetBuildInfo_s(char *szBuild, int iSize)
{
	//  Return the build identity
	sprintf_s(szBuild, iSize, DM_IDENTITY, DM_BUILD);
	return;
}

//  GetDefaultDbTitle_s
//
//  This function will return the default database titke for the current manifest
//
//  Parameters
//
//  1  -  char *		-  Pointer to the buffer where the title will be stored
//  2  -  int			-  Length of the buffer where the title will be stored
//
//  Returns
//
//  API Use Notes:
//
//
void DesignManager::GetDefaultDbTitle_s(char *szTitle, int iSize)
{
	//  Construct a default Database Title
	sprintf_s(szTitle, iSize, "New %s v%s Database", dmRoot->ManifestName, dmRoot->ManifestVersion);
	return;
}

//  DumpXMLImportErrors
//
//  This function dumps any XML errors to the log
//
//  Parameters
//
//  1  -  DesignElement *	-  Pointer to the current design element being processed
//
//  Returns
//
//  DWORD		-	Length of XML data returned
//
//  API Use Notes:
//
//
BOOL DesignManager::DumpXMLImportErrors(DesignElement *deCurrent)
{
	STATUS					stAPIRC;					//  return code from Notes API calls
	char					szMsg[MAX_MSG];				//  Generic message buffer

	//  Report on the import log, if there is anything in it
	if (DXLImportWasErrorLogged(dihImporter))
	{
		DXL_IMPORT_PROPERTY		dipLogHandle;						//  Log Handle Property
		MEMHANDLE				hMemLog;							//  Handle to the in memory log
		char					*pLog;								//  Pointer to Log data
		char					*pTarg;								//  Pointer to target data

		dipLogHandle = iResultLog;
		hMemLog = NULLHANDLE;
		stAPIRC = DXLGetImporterProperty(dihImporter, dipLogHandle, &hMemLog);
		if (ERR(stAPIRC) == NOERROR)
		{
			pLog = (char *) OSMemoryLock(hMemLog);
			// Break the log into separate lines
			pTarg = szMsg;
			while (*pLog != '\0')
			{
				if ((*pLog == '\x0d') || (*pLog == '\x0a') || (pTarg == (szMsg + MAX_MSG - 1)))
				{
					//  End of line or output buffer full, log the output
					*pTarg = '\0';
					xeLocal->LogMessage(szMsg);
					if (*pLog == '\x0d') pLog++;
					else
					{
						if (*pLog != '\x0a') pLog--;
					}
					pTarg = szMsg;
				}
				else
				{
					*pTarg = *pLog;
					pTarg++;
				}
				pLog++;
			}

			OSMemoryUnlock(hMemLog);
			OSMemoryFree(hMemLog);
			hMemLog = NULLHANDLE;
		}
		else
		{
			//  Unable to get the Log Handle
			xeLocal->LogMessage(MSG_DXR5526E);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
		}
	}
	return TRUE;
}

//  WrapperLSCompileError
//
//  This function is passed into the NSFNoteLSCompile function as a callback to
//  call for each lotusscript compilation error found on a note.
//
//  Parameters
//
//  1  -  DesignElement *	-  Pointer to the current design element being processed
//
//  Returns
//
//  DWORD		-	Length of XML data returned
//
//  API Use Notes:
//
//
STATUS LNCALLBACK DesignManager::WrapperLSCompileError(LSCOMPILE_ERR_INFO *pInfo, void *pContext)
{
	DesignManager	*dmLocal;

	dmLocal = (DesignManager *) pContext;
	return dmLocal->DumpLSCompileError(pInfo, dmLocal);
}

//  DumpLSCompileError
//
//  This function dumps a lotusscript compilation error into the log
//
//  Parameters
//
//  1  -  DesignElement *	-  Pointer to the current design element being processed
//
//  Returns
//
//  DWORD		-	Length of XML data returned
//
//  API Use Notes:
//
//
STATUS DesignManager::DumpLSCompileError(LSCOMPILE_ERR_INFO *pInfo, DesignManager *dmCurrent)
{
	DesignManager			*dmLocal;
	LSCOMPILE_ERR_INFO		*pErrInfo;
	char					szMsg[MAX_MSG];				//  Generic message buffer

	pErrInfo = pInfo;
	dmLocal = dmCurrent;

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5576E, pErrInfo->Line, pErrInfo->pErrFile, pErrInfo->pErrText);
	dmLocal->xeLocal->LogMessage(szMsg);


	return NOERROR;
}

//  FixViewSelectionFormula
//
//  This function performs the post processing of a view design element
//
//  Parameters
//
//  1  -  DesignElement *		-  Pointer to the Design Element to be applied
//  2  -  DBHANDLE				-  Handle of the database to which the design set is to be applied
//
//
//  Returns
//
//  BOOL  -			TRUE if the post processing was completed
//                  FALSE if the post processing could not be completed
//
//  API Use Notes:
//
//
BOOL DesignManager::FixViewSelectionFormula(DesignElement *deCurrent, DBHANDLE hDB, NOTEHANDLE hnDesign)
{
	STATUS					stAPIRC;					//  return code from Notes API calls
	BLOCKID					bidValue;
	BOOL					bIncludeRepCons = FALSE;
	BOOL					bShowHierarchy = FALSE;
	BOOL					bFormulaIsEmpty;
	DWORD					dwLength;
	FORMULAHANDLE			hSelFormula = NULLHANDLE;
	FORMULAHANDLE			hColFormula = NULLHANDLE;
    DHANDLE					hFormulaText;
	NOTEID					nidDesign;
	VIEW_FORMAT_HEADER		formatHeader;
	VIEW_FORMAT_HEADER		*pHeaderFormat;
    VIEW_TABLE_FORMAT		tableFormat;
    VIEW_TABLE_FORMAT		*pTableFormat;
    VIEW_COLUMN_FORMAT		viewColumn;
    VIEW_COLUMN_FORMAT		*pColumnFormat;
	WORD					wDataType;
	WORD					wColumn;
	WORD					wRetFormulaTextLen;
	WORD					wRetColFomrulaLen;
    WORD					wSelFormulaLen;
	WORD					wdc;						//  "We Don't Care"
	char 					*pData;
    char 					*pPackedData;
    char 					*pFormula;
	char					*pFormulaText;
	char					szColItemName[MAX_COLUMN_NAME_SIZE];
	char					szFormula[MAX_COLUMN_FORMULA_SIZE];
	char					szMsg[MAX_MSG];				//  Generic message buffer
	void					*tempPtr;

	if (deCurrent == NULL) return FALSE;
	if (hDB == NULLHANDLE ) return FALSE;
	if (hnDesign == NULLHANDLE ) return FALSE;

	NSFNoteGetInfo(hnDesign, _NOTE_ID, &nidDesign);

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5582T, deCurrent->ElementName);
    xeLocal->LogVerbose(szMsg);

	//  Get the selection formula
	stAPIRC = NSFItemInfo(hnDesign,
                          VIEW_FORMULA_ITEM,
                          (WORD) strlen(VIEW_FORMULA_ITEM),
                          NULL,
                          &wDataType,
                          &bidValue,
                          &dwLength);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5577E, VIEW_FORMULA_ITEM, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	if (wDataType != TYPE_FORMULA)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5585E, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	if (dwLength >= MAX_COLUMN_FORMULA_SIZE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5584E, MAX_COLUMN_FORMULA_SIZE, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	//  Lock the value, decompile the formula and then unlock the value again
	pData = OSLockBlock(char, bidValue);
    pData += sizeof(WORD);

	stAPIRC = NSFFormulaDecompile(pData, TRUE, &hFormulaText, &wRetFormulaTextLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5593E, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		OSUnlockBlock(bidValue);
		return FALSE;
	}
	OSUnlockBlock(bidValue);

	//  Lock the formula text so we can copy into our buffer and then free the formula text
	pFormulaText = OSLock(char, hFormulaText);

	szFormula[0] = '\0';
	memcpy_s(szFormula, MAX_COLUMN_FORMULA_SIZE-1, pFormulaText, (rsize_t) wRetFormulaTextLen); 
	szFormula[wRetFormulaTextLen] = '\0';

	OSUnlock(hFormulaText);
	OSMemFree(hFormulaText);

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5600T, szFormula);
    xeLocal->LogVerbose(szMsg);

	//  Now build the selection formula ready to append the column formulas
	stAPIRC = NSFFormulaCompile(NULL, 0, szFormula, (WORD) strlen(szFormula), &hSelFormula, &wSelFormulaLen, &wdc, &wdc, &wdc, &wdc, &wdc);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5586E, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	//  Get the view format item
	stAPIRC = NSFItemInfo(hnDesign,
                          VIEW_VIEW_FORMAT_ITEM,
                          sizeof(VIEW_VIEW_FORMAT_ITEM) - 1,
                          NULL,
                          &wDataType,
                          &bidValue,
                          &dwLength);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5577E, VIEW_VIEW_FORMAT_ITEM, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		OSMemFree(hSelFormula);
		return FALSE;
	}

	//  Lock for the view format item value
	pData = OSLockBlock(char, bidValue);
    pData += sizeof(WORD);

	//  Get the view format item header
	tempPtr = pData;
    ODSReadMemory( &tempPtr, _VIEW_FORMAT_HEADER, &formatHeader, 1 );
    pHeaderFormat = &formatHeader;

	//  Check the header format version
	if (pHeaderFormat->Version != VIEW_FORMAT_VERSION)
    {
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5578E, VIEW_FORMAT_VERSION, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
        OSUnlockBlock(bidValue);
		OSMemFree(hSelFormula);
        return FALSE;
    }

	if ((pHeaderFormat->ViewStyle & VIEW_CLASS_MASK) != VIEW_CLASS_TABLE)
	{

        OSUnlockBlock(bidValue);
		OSMemFree(hSelFormula);
        return FALSE;
	}

	//  Read the table format
	tempPtr = pData;
	ODSReadMemory( &tempPtr, _VIEW_TABLE_FORMAT, &tableFormat, 1 );
	pTableFormat = &tableFormat;

	//  Determine if we need to append the $Conflict item to the view summary items
	if (pTableFormat->Flags & VIEW_TABLE_FLAG_CONFLICT)
	{
		bIncludeRepCons = TRUE;
		xeLocal->LogVerbose(MSG_DXR5603T);
	}

	//  Determine if we need to append the $REF item to the view summary items
	if (!(pTableFormat->Flags & VIEW_TABLE_FLAG_FLATINDEX))
	{
		bShowHierarchy = TRUE;
		xeLocal->LogVerbose(MSG_DXR5602T);
	}

	//  Point past the table format
	pData += ODSLength(_VIEW_TABLE_FORMAT);

	//  Get a pointer to the packed data, which is located after the column format structures.
	//  This data does not need to be converted into host-specific byte ordering
	pPackedData = pData + (ODSLength(_VIEW_COLUMN_FORMAT) * pTableFormat->Columns);

	//  Now unpack and read the column data
	for (wColumn = 1; wColumn <= pTableFormat->Columns; wColumn++)
	{
		//  Get the fixed portion of the column descriptor and validate the signature.
		tempPtr = pData;
		ODSReadMemory( &tempPtr, _VIEW_COLUMN_FORMAT, &viewColumn, 1 );
		pColumnFormat = &viewColumn;
         
		pData += ODSLength(_VIEW_COLUMN_FORMAT);

        //  Column number
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5579T, wColumn, pTableFormat->Columns);
        xeLocal->LogVerbose(szMsg);

		//  Get the column item name then move past it
		szColItemName[0] = '\0';
		memcpy_s(szColItemName, MAX_COLUMN_NAME_SIZE-1, pPackedData,pColumnFormat->ItemNameSize); 
		szColItemName[pColumnFormat->ItemNameSize] = '\0';    
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5580T, wColumn, szColItemName);
        xeLocal->LogVerbose(szMsg);

		pPackedData += pColumnFormat->ItemNameSize;

		//  Move post the item title
		pPackedData += pColumnFormat->TitleSize;

		//  Get the formula, advance data ptr to next item.
		pFormula = pPackedData;
		pPackedData += pColumnFormat->FormulaSize;

		if (pColumnFormat->Flags1 & VCF1_M_Hidden)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5595T, wColumn);
			xeLocal->LogVerbose(szMsg);
		}

		bFormulaIsEmpty = FALSE;
		if (pColumnFormat->FormulaSize == 0)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5596T, wColumn);
			xeLocal->LogVerbose(szMsg);
			bFormulaIsEmpty = TRUE;
		}
		else
		{
			if (pColumnFormat->FormulaSize >= MAX_COLUMN_FORMULA_SIZE)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5583E, wColumn, MAX_COLUMN_FORMULA_SIZE, deCurrent->ElementName);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				OSUnlockBlock(bidValue);
				OSMemFree(hSelFormula);
				return FALSE;
			}

			stAPIRC = NSFFormulaDecompile(pFormula, FALSE, &hFormulaText, &wRetFormulaTextLen);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5581E, wColumn, deCurrent->ElementName);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				OSUnlockBlock(bidValue);
				OSMemFree(hSelFormula);
				return FALSE;
			}
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5597T, wColumn, wRetFormulaTextLen);
			xeLocal->LogVerbose(szMsg);

			//  Get pointer to formula text and copy it into our buffer
			pFormulaText = OSLock(char, hFormulaText);

			szFormula[0] = '\0';
			memcpy_s(szFormula, MAX_COLUMN_FORMULA_SIZE-1, pFormulaText, (rsize_t) wRetFormulaTextLen); 
			szFormula[wRetFormulaTextLen] = '\0';

			sprintf_s(szMsg, MAX_MSG, MSG_DXR5599T, strlen(szFormula));
			xeLocal->LogVerbose(szMsg);
			
			//  Dump column formula if tracing
			if ( xeLocal->rsLocal->TraceArea == 24 )
			{
				xeLocal->DumpMemory(szFormula, wRetFormulaTextLen, MSG_DXR5601T);
			}

			//  Release the formula text handle
			OSUnlock(hFormulaText);
			OSMemFree(hFormulaText);
		}

		//  Append the item name and formula to the selection formula
		stAPIRC = NSFFormulaSummaryItem(hSelFormula, szColItemName, (WORD) strlen(szColItemName));
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5587E, wColumn, deCurrent->ElementName);
			xeLocal->LogMessage(szMsg);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			OSUnlockBlock(bidValue);
			OSMemFree(hSelFormula);
			return FALSE;
		}

		if (!bFormulaIsEmpty)
		{
			//  Now re-compile the formula ready for appending to the selection formula
			stAPIRC = NSFFormulaCompile(szColItemName, (WORD) strlen(szColItemName), szFormula, (WORD) strlen(szFormula), &hColFormula, &wRetColFomrulaLen, &wdc, &wdc, &wdc, &wdc, &wdc);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5586E, deCurrent->ElementName);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				OSUnlockBlock(bidValue);
				OSMemFree(hSelFormula);
				return FALSE;
			}
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5598T, wColumn, wRetColFomrulaLen);
			xeLocal->LogVerbose(szMsg);
	        
			stAPIRC = NSFFormulaMerge(hColFormula, hSelFormula);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5588E, wColumn, deCurrent->ElementName);
				xeLocal->LogMessage(szMsg);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg);
				OSMemFree(hColFormula);
				hColFormula = NULLHANDLE;
				OSUnlockBlock(bidValue);
				OSMemFree(hSelFormula);
				return FALSE;
			}

			OSMemFree(hColFormula);
			hColFormula = NULLHANDLE;
		}

		//  Move to end of column format data
		pPackedData += pColumnFormat->ConstantValueSize;

	}

	//  Unlock the VIEW_VIEW_FORMAT_ITEM item value
	OSUnlockBlock(bidValue);

	//  Now we need to append the $Conflict item if required
	if (bIncludeRepCons)
	{
		szColItemName[0] = '\0';
		strcpy_s(szColItemName, MAX_COLUMN_NAME_SIZE, VIEW_CONFLICT_ITEM);
		stAPIRC = NSFFormulaSummaryItem(hSelFormula, szColItemName, (WORD) strlen(szColItemName));
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5589E, szColItemName, deCurrent->ElementName);
			xeLocal->LogMessage(szMsg);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			OSMemFree(hSelFormula);
			return FALSE;
		}
		xeLocal->LogVerbose(MSG_DXR5605T);
	}

	//  Now we need to append the $REF item if required
	if (bShowHierarchy)
	{
		szColItemName[0] = '\0';
		strcpy_s(szColItemName, MAX_COLUMN_NAME_SIZE, FIELD_LINK);
		stAPIRC = NSFFormulaSummaryItem(hSelFormula, szColItemName, (WORD) strlen(szColItemName));
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5589E, szColItemName, deCurrent->ElementName);
			xeLocal->LogMessage(szMsg);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg);
			OSMemFree(hSelFormula);
			return FALSE;
		}
		xeLocal->LogVerbose(MSG_DXR5604T);
	}

	stAPIRC = NSFFormulaGetSize(hSelFormula, &wSelFormulaLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5590E, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		OSMemFree(hSelFormula);
		return FALSE;
	}

	//  Dump $Formula item if tracing
	if ( xeLocal->rsLocal->TraceArea == 24 )
	{
		tempPtr = OSLockObject(hSelFormula);
		xeLocal->DumpMemory(tempPtr, wSelFormulaLen, "$Formula item");
		OSUnlockObject(hSelFormula);
	}

	stAPIRC = NSFItemDelete(hnDesign, VIEW_FORMULA_ITEM, (WORD) strlen(VIEW_FORMULA_ITEM));
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5594E, VIEW_FORMULA_ITEM, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		OSMemFree(hSelFormula);
		return FALSE;
	}

	stAPIRC = NSFItemAppend(hnDesign, (WORD) ITEM_SIGN, VIEW_FORMULA_ITEM, (WORD) strlen(VIEW_FORMULA_ITEM), TYPE_FORMULA, OSLockObject(hSelFormula), (DWORD) wSelFormulaLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5591E, VIEW_FORMULA_ITEM, deCurrent->ElementName);
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg);
		OSMemFree(hSelFormula);
		return FALSE;
	}

	OSUnlockObject(hSelFormula);
	OSMemFree(hSelFormula);

	return TRUE;
}

//  FettleAgents
//
//  This function wull perform any specific post import processing on agents.
//
//   At this time the only fix needed is for scheduled agents to correct the run between time
//   which is not exported correctly in DXL.
//
//  Parameters
//
//  1  -  DBHANDLE				-  Handle of the database being processed
//  2  -  NOTEID 				-  NoteID of the Agent design note
//  3  -  int					-  ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
//
void DesignManager::FettleAgents(DBHANDLE hdbDB, NOTEID nidAgent, int iThreadID)
{
	STATUS				stAPIRC;					//  Return code from Notes API Calls
	NOTEHANDLE			hnAgent;					//  Handle of the agent note
	BLOCKID				bidItem;					//  BlockID of the Item
	BLOCKID				bidValue;					//  BlockID of the Item Value
	WORD				wItemType;					//  The Item Type
	DWORD				dwValueLen;					//  Value length
	BOOL				bWindowAdjusted;			//  Flag - the time window was adjusted
	ODS_ASSISTSTRUCT	AssistInfo;					//  Assistant Information
	UCHAR				*ivPtr;						//  Pointer to the item value
	char				szMsg[MAX_MSG + 1];			//  Generic message buffer

	//  Safety chacks - if we do not have a NoteID then return
	if (nidAgent == 0) return;

	//  Open the agent note
	stAPIRC = NSFNoteOpen(hdbDB, nidAgent, 0, &hnAgent);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5530E, nidAgent);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Get the contents of the $AssistInfo Item
	stAPIRC = NSFItemInfo(hnAgent, ASSIST_INFO_ITEM, (WORD) strlen(ASSIST_INFO_ITEM),
						&bidItem,
						&wItemType,
						&bidValue,
						&dwValueLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5531E, nidAgent);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		NSFNoteClose(hnAgent);
		return;
	}

	//  Get the Assistant Info Structure data
	ivPtr = OSLockBlock(UCHAR, bidValue);
	memcpy(&AssistInfo, ivPtr + sizeof(WORD), sizeof(ODS_ASSISTSTRUCT));
	OSUnlockBlock(bidValue);

	//  Check that the agent trigger is scheduled - if not then just leave
	if (AssistInfo.wTriggerType == ASSISTTRIGGER_TYPE_SCHEDULED)
	{
		bWindowAdjusted = FALSE;
		//  Process according to the interval type
		switch(AssistInfo.wIntervalType)
		{
		case ASSISTINTERVAL_TYPE_MINUTES:
			//  Check for start = 0 and end = 0  (which results in never being scheduled)
			if ((AssistInfo.dwTime1 == 0) && (AssistInfo.dwTime2 == 0))
			{	
				AssistInfo.dwTime2 = TICKS_IN_DAY;						//  Run All day
				bWindowAdjusted = TRUE;
			}
			break;
		case ASSISTINTERVAL_TYPE_DAYS:
			if (AssistInfo.dwTime1 == 0)
			{
				AssistInfo.dwTime1 = 6 * 60 * (DWORD) TICKS_IN_DAY;			//  Run at 06:00
				bWindowAdjusted = TRUE;
			}
			break;
		case ASSISTINTERVAL_TYPE_WEEK:
			if (AssistInfo.dwTime1 == 0)
			{
				AssistInfo.dwTime1 = 6 * 60 * (DWORD) TICKS_IN_DAY;			//  Run at 06:00
				bWindowAdjusted = TRUE;
			}
			if (AssistInfo.dwTime2 == 0)
			{
				AssistInfo.dwTime2 = 3;									//  Run on Tuesday
				bWindowAdjusted = TRUE;
			}
			break;
		case ASSISTINTERVAL_TYPE_MONTH:
			if (AssistInfo.dwTime1 == 0)
			{
				AssistInfo.dwTime1 = 6 * 60 * (DWORD) TICKS_IN_DAY;			//  Run at 06:00
				bWindowAdjusted = TRUE;
			}
			if (AssistInfo.dwTime2 == 0)
			{
				AssistInfo.dwTime2 = 3;									//  Run on 3rd of the month
				bWindowAdjusted = TRUE;
			}
			break;
		}

		//  If the window has been adjusted the update the structure in the note
		if (bWindowAdjusted)
		{
			ivPtr = OSLockBlock(UCHAR, bidValue);
			memcpy(ivPtr + sizeof(WORD), &AssistInfo, sizeof(ODS_ASSISTSTRUCT));
			OSUnlockBlock(bidValue);

			stAPIRC = NSFNoteUpdate(hnAgent, UPDATE_FORCE);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5532E, nidAgent);
				xeLocal->LogMessage(szMsg, iThreadID);
				xeLocal->GetAPIMessage(stAPIRC, szMsg);
				xeLocal->LogMessage(szMsg, iThreadID);
			}
		}
	}

	//  Close the Note
	NSFNoteClose(hnAgent);
	return;
}

//  CopyDesignCollection
//
//  This function will copy the design collection from the local names.nsf to the passed database
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle to the database
//  2  -  int				-  ID of the calling thread
//
//  Returns
//
//  BOOL	-			TRUE if the design collection was copied
//						FALSE if the design collection could not be copied
//
//  API Use Notes:
//
//
BOOL DesignManager::CopyDesignCollection(DBHANDLE hdbX, int iThreadID)
{
	STATUS					stAPIRC;						//  Status code returned from Notes API calls
	DBHANDLE				hdbNAB;							//  D/B handle of the local NAB
	NOTEHANDLE				hnDesignColl;					//  Handle of the reference design collection note
	NOTEHANDLE				hnNewDC;						//  Handle of the new design collection note
	NOTEID					nidDesignColl;					//  Note ID of the design collection
	NOTEID					nidNewDC;						//  Note ID of the new design collection
	HCOLLECTION				hDesignColl;					//  Handle of the design collection
	DWORD					dwItemLen;						//  Item data length
	BLOCKID					bidItem;						//  Blockid of the item
	BLOCKID					bidValue;						//  Blockid of the item value
	WORD					wType;							//  Item Type
	WORD					wNoteClass;						//  Class of the new design collection note
	char					szAPIMsg[MAX_MSG + 1];			//  API message buffer

	//  Open the local NAB
	hdbNAB = xeLocal->OpenDatabase("", "names.nsf", iThreadID);
	if (hdbNAB == NULLHANDLE)
	{
		xeLocal->LogMessage(MSG_DXR5641E, iThreadID);
		return FALSE;
	}

	//  Get the Note ID of the design collection
	stAPIRC = NSFDbGetSpecialNoteID(hdbNAB, (WORD) (SPECIAL_ID_NOTE | NOTE_CLASS_DESIGN), &nidDesignColl);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5642E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFDbClose(hdbNAB);
		return FALSE;
	}

	//  Create the new note that will become the design collection
	stAPIRC = NSFNoteCreate(hdbX, &hnNewDC);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5645E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFDbClose(hdbNAB);
		return FALSE;
	}

	//  Set the Note Class
	wNoteClass = NOTE_CLASS_DESIGN | NOTE_CLASS_DEFAULT;
    NSFNoteSetInfo(hnNewDC, _NOTE_CLASS, &wNoteClass);
	
	//  Append the text fields to the new design collection note
	dwItemLen = (DWORD) strlen(DC_VIEW_VER);
	stAPIRC = NSFItemAppend(hnNewDC, 
							ITEM_SUMMARY, 
							VIEW_DESIGN_VER_ITEM, 
							(WORD) strlen(VIEW_DESIGN_VER_ITEM),
							TYPE_TEXT,
							DC_VIEW_VER,
							dwItemLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5646E, VIEW_DESIGN_VER_ITEM);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnNewDC);
		NSFDbClose(hdbNAB);
		return FALSE;
	}
	dwItemLen = (DWORD) strlen(DC_VIEW_VERV2);
	stAPIRC = NSFItemAppend(hnNewDC, 
							ITEM_SUMMARY, 
							VIEW_DESIGN_VER_ITEM_V2, 
							(WORD) strlen(VIEW_DESIGN_VER_ITEM_V2),
							TYPE_TEXT,
							DC_VIEW_VERV2,
							dwItemLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5646E, VIEW_DESIGN_VER_ITEM_V2);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnNewDC);
		NSFDbClose(hdbNAB);
		return FALSE;
	}
	dwItemLen = (DWORD) strlen(DC_VIEW_CLASSES);
	stAPIRC = NSFItemAppend(hnNewDC, 
							ITEM_SUMMARY, 
							VIEW_CLASSES_ITEM, 
							(WORD) strlen(VIEW_CLASSES_ITEM),
							TYPE_TEXT,
							DC_VIEW_CLASSES,
							dwItemLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5646E, VIEW_CLASSES_ITEM);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnNewDC);
		NSFDbClose(hdbNAB);
		return FALSE;
	}

	//  Open the existing design collection note and copy the $FORMULA itemto the new collection note
	stAPIRC = NSFNoteOpen(hdbNAB, nidDesignColl, 0, &hnDesignColl);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5647E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnNewDC);
		NSFDbClose(hdbNAB);
		return FALSE;
	}

	//  Get the Blockid of the $FORMULA item
	stAPIRC = NSFItemInfo(hnDesignColl,
							VIEW_FORMULA_ITEM,
							(WORD) strlen(VIEW_FORMULA_ITEM),
							&bidItem,
							&wType,
							&bidValue,
							&dwItemLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5648E, VIEW_FORMULA_ITEM);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnDesignColl);
		NSFNoteClose(hnNewDC);
		NSFDbClose(hdbNAB);
		return FALSE;
	}

	//  Copy the item to the new design collection note
	stAPIRC = NSFItemCopy(hnNewDC, bidItem);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5649E, VIEW_FORMULA_ITEM);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnDesignColl);
		NSFNoteClose(hnNewDC);
		NSFDbClose(hdbNAB);
		return FALSE;
	}

	//  Copy the $Collation object from source to target
	//  Get the Blockid of the $COLLATION item
	stAPIRC = NSFItemInfo(hnDesignColl,
							VIEW_COLLATION_ITEM,
							(WORD) strlen(VIEW_COLLATION_ITEM),
							&bidItem,
							&wType,
							&bidValue,
							&dwItemLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5648E, VIEW_COLLATION_ITEM);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnDesignColl);
		NSFNoteClose(hnNewDC);
		NSFDbClose(hdbNAB);
		return FALSE;
	}

	//  Copy the item to the new design collection note
	stAPIRC = NSFItemCopy(hnNewDC, bidItem);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5649E, VIEW_COLLATION_ITEM);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnDesignColl);
		NSFNoteClose(hnNewDC);
		NSFDbClose(hdbNAB);
		return FALSE;
	}

	//  close the existing design collection note
	NSFNoteClose(hnDesignColl);
	NSFDbClose(hdbNAB);

	//  Save, get the noteid and close the new design collection note
	stAPIRC = NSFNoteUpdate(hnNewDC, UPDATE_FORCE);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5643E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnNewDC);
		return FALSE;
	}
	NSFNoteGetInfo(hnNewDC, _NOTE_ID, &nidNewDC);
	NSFNoteClose(hnNewDC);

	//  Open and close the Design Collection to force a rebuild
	stAPIRC = NIFOpenCollection(hdbX, 
								hdbX, 
								(NOTE_ID_SPECIAL | NOTE_CLASS_DESIGN), 
								(WORD) OPEN_REBUILD_INDEX, 
								NULLHANDLE, 
								&hDesignColl,
								NULL, 
								NULL, 
								NULL, 
								NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5644W, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		return TRUE;
	}

	NIFCloseCollection(hDesignColl);
	return TRUE;
}

//  ValidateDesign
//
//  This function will validate and fix any non-conformance in the database design
//
//  Parameters
//
//  1  -  DBHANDLE		-  Handle of the database to be validated
//
//  Returns
//
//  BOOL  -			TRUE if the design was valid or corrected
//                  FALSE if the design could not be corrected
//
//  API Use Notes:
//
//
BOOL DesignManager::ValidateDesign(DBHANDLE hDB)
{
	STATUS					stAPIRC;						//  Status code returned from Notes API calls
	HCOLLECTION				hDefColl;						//  Handle of the default view
	NOTEID					nidView;						//  NoteID of (a view)
	char					szMsg[MAX_MSG + 1];				//  Generic message buffer

	//  Check that there is a default view in the database - if not fix it
	stAPIRC = NIFOpenCollection(hDB, 
								hDB, 
								(NOTE_ID_SPECIAL | NOTE_CLASS_VIEW), 
								(WORD) 0, 
								NULLHANDLE, 
								&hDefColl,
								NULL, 
								NULL, 
								NULL, 
								NULL);
	if (ERR(stAPIRC) == NOERROR)
	{
		NIFCloseCollection(hDefColl);
		return TRUE;
	}

	//  See if we can try to fix the problem
	if ((ERR(stAPIRC) == ERR_SPECIAL_ID) || (ERR(stAPIRC) == ERR_NOT_FOUND))
	{
		//  Find any view in the database
		nidView = FindAnyView(hDB, 0);
		if (nidView != 0)
		{
			stAPIRC = xeLocal->SetSpecialNoteID(hDB, NOTE_CLASS_VIEW, nidView, 0);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5665E, nidView);
				xeLocal->LogMessage(szMsg, 0);
				xeLocal->GetAPIMessage(stAPIRC, szMsg);
				xeLocal->LogMessage(szMsg, 0);
				return FALSE;
			}
			return TRUE;
		}
		else WriteDefaultView(hDB, 0);
		return TRUE;
	}
	else
	{
		xeLocal->LogMessage(MSG_DXR5650E, 0);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, 0);
		return FALSE;
	}
}

//  FindAnyView
//
//  This function will return the NoteID of any view in the database
//
//  Parameters
//
//  1  -  DBHANDLE		-  Handle of the database to be validated
//  2  -  int			-  ID of the thread making the call
//
//  Returns
//
//  NOTEID		-			NoteID of an arbitrary view in the database
//							NULL if no views exist in the D/B
//
//  API Use Notes:
//
//
NOTEID DesignManager::FindAnyView(DBHANDLE hDB, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
    TIMEDATE			tdStart;							//  Modified since time
	TIMEDATE			tdFinish;							//  Ending modification time
	WORD				wNoteClass;							//  Note types to be retrieved
	NOTEID				nidRetNID;							//  NoteID returned from IDScan
	DHANDLE				hitViewTable;						//  Handle of the ID table
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Set the Note Class to extract
	wNoteClass = NOTE_CLASS_VIEW;

	TimeConstant(TIMEDATE_WILDCARD, &tdStart);

	//  Get the Modified Note table for all Design Notes - modified up till now
	stAPIRC = NSFDbGetModifiedNoteTable(hDB, wNoteClass, tdStart, &tdFinish, &hitViewTable);
	if (ERR(stAPIRC) != NOERROR)
	{
		//  If no view notes were available then return with no note
		if (ERR(stAPIRC) == ERR_NO_MODIFIED_NOTES) return 0;

		//  Nasty little error
		xeLocal->LogMessage(MSG_DXR5651E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return 0;
	}

	//  Get the first NoteID in the table and return it
	if (IDScan(hitViewTable, TRUE, &nidRetNID))
	{
		IDDestroyTable(hitViewTable);
		return nidRetNID;
	}

	xeLocal->LogMessage(MSG_DXR5652E, iThreadID);
	IDDestroyTable(hitViewTable);
	return 0;
}

//  WriteDefaultView
//
//  This function will create a default View in the passed database
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle to the database
//  2  -  int				-  ID of the calling thread
//
//  Returns
//
//  BOOL	-			TRUE if the default view was written
//						FALSE if the default view could not be written
//
//  API Use Notes:
//
//
BOOL DesignManager::WriteDefaultView(DBHANDLE hdbX, int iThreadID)
{
	STATUS					stAPIRC;						//  Status code returned from Notes API calls
	STATUS					stCompileErr;					//  Formula Compile Error Code
	NOTEHANDLE				hnView;							//  Handle of the default view note
	NOTEID					nidFind;						//  Note ID of the default view note
	DHANDLE					hSelFormula;					//  Handle of the selection formula
	DHANDLE					hC0Formula;						//  Handle of Column 0 formula
	WORD					wSelFormulaLen;					//  Length of selection formula
	WORD					wC0FormulaLen;					//  Length of Column 0 formula
	WORD					wNoteClass;						//  Note class to set
	WORD					wLineNo;						//  Compile Error Line number
	WORD					wColumnNo;						//  Compile Error Column number
	WORD					wColumnOffset;					//  Compile Error Offset
	WORD					wErrLen;						//  Compile Error Length
	WORD					wVFBufferLen;					//  Calculated length of the VIEWFORMAT buffer
	BYTE					*pSelFormula;					//  Pointer to the selection formula
	DHANDLE					hVFBuffer;						//  Handle to the VIEWFORMAT buffer
	BYTE					*pVFBuffer;						//  Pointer to the VIEWFORMAT buffer
	BYTE					*pVFBufPart;					//  Pointer into the VIEWFORMAT buffer
	BYTE					*pColFormula;					//  Pointer to the column formula

	//  VIEWFORMAT component structures
    VIEW_TABLE_FORMAT		vtfDefault;						//  VIEW_TABLE_FORMAT structure
    VIEW_COLUMN_FORMAT		vcfDefault;						//  VIEW_COLUMN_FORMAT structure
    VIEW_TABLE_FORMAT2		vtf2Default;					//  VIEW_TABLE_FORMAT2 structure
	VIEW_COLUMN_FORMAT2		vcf2Default;					//  VIEW_COLUMN_FORMAT2 structure
	VIEW_TABLE_FORMAT3		vtf3Default;					//  VIEW_TABLE_FORMAT3 structure

	char					szAPIMsg[MAX_MSG + 1];			//  API message buffer

	//  Create the new view note
	stAPIRC = NSFNoteCreate(hdbX, &hnView);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5653E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		return FALSE;
	}

	//  Set the Note Class
	wNoteClass = NOTE_CLASS_VIEW | NOTE_CLASS_DEFAULT;
    NSFNoteSetInfo(hnView, _NOTE_CLASS, &wNoteClass);

	//  Set the view name to default ""
	stAPIRC = NSFItemSetText(hnView, VIEW_TITLE_ITEM, DEF_VIEW_TITLE, MAXWORD);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5654E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnView);
		return FALSE;
	}

	//  Compile the Selection Formula
	stAPIRC = NSFFormulaCompile(NULL, 
								0,
								DEF_VIEW_SELECT,
								(WORD) strlen(DEF_VIEW_SELECT),
								&hSelFormula,
								&wSelFormulaLen,
								&stCompileErr,
								&wLineNo,
								&wColumnNo,
								&wColumnOffset,
								&wErrLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5655E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5656E, wLineNo, wColumnNo, wColumnOffset, wErrLen);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(ERR(stCompileErr), szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnView);
		return FALSE;
	}

	//  Now compile the formula for column1
	stAPIRC = NSFFormulaCompile(DEF_VIEW_C0NAME, 
								(WORD) strlen(DEF_VIEW_C0NAME),
								DEF_VIEW_C0FORMULA,
								(WORD) strlen(DEF_VIEW_C0FORMULA),
								&hC0Formula,
								&wC0FormulaLen,
								&stCompileErr,
								&wLineNo,
								&wColumnNo,
								&wColumnOffset,
								&wErrLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5657E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5656E, wLineNo, wColumnNo, wColumnOffset, wErrLen);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(ERR(stCompileErr), szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		OSMemFree(hSelFormula);
		NSFNoteClose(hnView);
		return FALSE;
	}

	//  Set the Column 0 Summary Item
	stAPIRC = NSFFormulaSummaryItem(hSelFormula, DEF_VIEW_SELECT, (WORD) strlen(DEF_VIEW_SELECT));
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5658E, DEF_VIEW_C0NAME);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		OSMemFree(hC0Formula);
		OSMemFree(hSelFormula);
		NSFNoteClose(hnView);
		return FALSE;
	}

	//  Now merge the column 0 formula into the view formula
	stAPIRC = NSFFormulaMerge(hC0Formula, hSelFormula);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5659E, DEF_VIEW_C0NAME);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		OSMemFree(hC0Formula);
		OSMemFree(hSelFormula);
		NSFNoteClose(hnView);
		return FALSE;
	}

	//  Update the length of the selection formula
	stAPIRC = NSFFormulaGetSize(hSelFormula, &wSelFormulaLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5660E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		OSMemFree(hC0Formula);
		OSMemFree(hSelFormula);
		NSFNoteClose(hnView);
		return FALSE;
	}

	//  Append the composite formula to the view note ($FORMULA)
	pSelFormula = (BYTE *) OSLockObject(hSelFormula);
	stAPIRC = NSFItemAppend(hnView, ITEM_SUMMARY, VIEW_FORMULA_ITEM, (WORD) strlen(VIEW_FORMULA_ITEM), TYPE_FORMULA, pSelFormula, wSelFormulaLen);
	OSUnlockObject(hSelFormula);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5661E, wSelFormulaLen);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		OSMemFree(hC0Formula);
		OSMemFree(hSelFormula);
		NSFNoteClose(hnView);
		return FALSE;
	}

	//  Free the merged selection formula
	OSMemFree(hSelFormula);

	//  Construct the $VIEWFORMAT item for the default view
	//  Start by calculating the legth of the buffer that will be needed
	wVFBufferLen = ODSLength(_VIEW_TABLE_FORMAT) +
					ODSLength(_VIEW_COLUMN_FORMAT) +
					(WORD) strlen(DEF_VIEW_C0NAME) +
					(WORD) strlen(DEF_VIEW_C0TITLE) +
					wC0FormulaLen +
					ODSLength(_VIEW_TABLE_FORMAT2) +
					ODSLength(_VIEW_COLUMN_FORMAT2) +
					ODSLength(_VIEW_TABLE_FORMAT3);

	//  Allocate the buffer to hold the VIEWFORMAT
	stAPIRC = OSMemAlloc(0, (DWORD) wVFBufferLen, &hVFBuffer);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5662E, wVFBufferLen);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		OSMemFree(hC0Formula);
		NSFNoteClose(hnView);
		return FALSE;
	}

	//  Lock and clear the buffer
	pVFBuffer = (BYTE *) OSLockObject(hVFBuffer);
	memset(pVFBuffer, 0, wVFBufferLen);
	pVFBufPart = pVFBuffer;

	//  Populate the VIEW_TABLE_FORMAT and add it to the buffer
	memset(&vtfDefault, 0, sizeof(VIEW_TABLE_FORMAT));
	vtfDefault.Header.Version = VIEW_FORMAT_VERSION;
	vtfDefault.Header.ViewStyle = VIEW_STYLE_TABLE;
	vtfDefault.Columns = 1;
	vtfDefault.ItemSequenceNumber = 0;
	vtfDefault.Flags = VIEW_TABLE_FLAG_FLATINDEX;
	vtfDefault.Flags2 = 0;
	ODSWriteMemory(&pVFBufPart, _VIEW_TABLE_FORMAT, &vtfDefault, 1);

	//  Populate the VIEW_COLUMN_FORMAT and add it to the buffer
	memset(&vcfDefault, 0, sizeof(VIEW_COLUMN_FORMAT));
	vcfDefault.Signature = VIEW_COLUMN_FORMAT_SIGNATURE;
	vcfDefault.Flags1 = 0;
	vcfDefault.DisplayWidth = 80;
	vcfDefault.FontID = DEFAULT_BOLD_FONT_ID;
	vcfDefault.Flags2 = 0;
	vcfDefault.FormatDataType = 0;
	vcfDefault.ListSep = 0;
	vcfDefault.NumberFormat.Format = NFMT_GENERAL;
	vcfDefault.FormulaSize = wC0FormulaLen;
	vcfDefault.ItemNameSize = (WORD) strlen(DEF_VIEW_C0NAME);
	vcfDefault.TitleSize = (WORD) strlen(DEF_VIEW_C0TITLE);
	vcfDefault.ConstantValueSize = 0;
	ODSWriteMemory(&pVFBufPart, _VIEW_COLUMN_FORMAT, &vcfDefault, 1);

	//  Append the column name, column title and formula to the buffer
	memcpy(pVFBufPart, DEF_VIEW_C0NAME, strlen(DEF_VIEW_C0NAME));
	pVFBufPart += strlen(DEF_VIEW_C0NAME);
	memcpy(pVFBufPart, DEF_VIEW_C0TITLE, strlen(DEF_VIEW_C0TITLE));
	pVFBufPart += strlen(DEF_VIEW_C0TITLE);
	pColFormula = (BYTE *) OSLockObject(hC0Formula);
	memcpy(pVFBufPart, pColFormula, wC0FormulaLen);
	pVFBufPart += wC0FormulaLen;
	OSUnlock(hC0Formula);
	OSMemFree(hC0Formula);

	//  Populate the VIEW_TABLE_FORMAT2 and add it to the buffer
	memset(&vtf2Default, 0, sizeof(VIEW_TABLE_FORMAT2));
	vtf2Default.Length = ODSLength(_VIEW_TABLE_FORMAT2);
	vtf2Default.BackgroundColor = NOTES_COLOR_WHITE;
	vtf2Default.V2BorderColor = 0;
	vtf2Default.TitleFont = DEFAULT_BOLD_FONT_ID;
	vtf2Default.UnreadFont = DEFAULT_FONT_ID;
	vtf2Default.TotalsFont = DEFAULT_FONT_ID;
	vtf2Default.AutoUpdateSeconds = 0;
	vtf2Default.wSig = VALID_VIEW_FORMAT_SIG;
	ODSWriteMemory(&pVFBufPart, _VIEW_TABLE_FORMAT2, &vtf2Default, 1);

	//  Populate the VIEW_COLUMN_FORMAT2 and add it to the buffer
	memset(&vcf2Default, 0, sizeof(VIEW_COLUMN_FORMAT2));
	vcf2Default.Signature = VIEW_COLUMN_FORMAT_SIGNATURE2;
	vcf2Default.HeaderFontID = DEFAULT_FONT_ID;
	vcf2Default.ColumnColor.Flags = COLOR_VALUE_FLAGS_SYSTEMCOLOR;
	vcf2Default.HeaderFontColor.Flags = COLOR_VALUE_FLAGS_SYSTEMCOLOR;
	ODSWriteMemory(&pVFBufPart, _VIEW_COLUMN_FORMAT2, &vcf2Default, 1);

	//  Populate the VIEW_TABLE_FORMAT3 and add it to the buffer
	memset(&vtf3Default, 0, sizeof(VIEW_TABLE_FORMAT3));
	vtf3Default.Length = ODSLength(_VIEW_TABLE_FORMAT3);
	ODSWriteMemory(&pVFBufPart, _VIEW_TABLE_FORMAT3, &vtf3Default, 1);

	//  Append the VIEWFORMAT item to the default view note
	stAPIRC = NSFItemAppend(hnView,
							ITEM_SUMMARY,
							VIEW_VIEW_FORMAT_ITEM,
							(WORD) strlen(VIEW_VIEW_FORMAT_ITEM),
							TYPE_VIEW_FORMAT,
							pVFBuffer,
							wVFBufferLen);
	OSUnlockObject(hVFBuffer);
	OSMemFree(hVFBuffer);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR5663E, wVFBufferLen);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnView);
		return FALSE;
	}

	//  Update the View Note
	stAPIRC = NSFNoteUpdate(hnView, UPDATE_FORCE);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR5664E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szAPIMsg);
		xeLocal->LogMessage(szAPIMsg, iThreadID);
		NSFNoteClose(hnView);
		return FALSE;
	}
	
	//  Close the note
	NSFNoteGetInfo(hnView, _NOTE_ID, &nidFind);
	NSFNoteClose(hnView);
	return TRUE;
}

