//  DXDirScanner
//
//  This class is the eXplorer class that handles the scanning of the directory level in the hierarchy according to
//  the scope and constraints from the search request.
//
//  Message Identifier:      DXR
//  Message Range:           1301 - 1399
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.11.5
//  Version:                 3.11.5  Build: 103
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.9.2		22/09/2010		Avoided a recursion in DX processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.5		18/02/2011		Report on D/Bs which cannot be opened
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.5		30/11/2011		Recursion avoidance improvements
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2011
//

//  Application includes
#include "DXDirScanner.h"

//  Single-Threaded Constructor
DXDirScanner::DXDirScanner(ExecEnvironment *xeParent, int iThreadID)
: xeLocal(NULL)
, dxdsNextLevel(NULL)
, dxbsInstance(NULL)
, dxeDeferred(NULL)
, dxrCurrent(NULL)
{
	char			szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Alternate Constructor
	xeLocal = xeParent;									//  Store the Execution Environment
	iParentThread = iThreadID;							//  Save the ID of the owning thread

	//  Trace construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1319T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);
}

//  Destructor
DXDirScanner::~DXDirScanner(void)
{
	char			szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Destroy (explicily) any lower level scanners
	if (dxdsNextLevel != NULL) delete dxdsNextLevel;
	if (dxbsInstance != NULL) delete dxbsInstance;

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1321T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);
}

//  ProcessRequest
//
//  This function is the main entry point for dispatching a Domino eXplorer request.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the invoking thread
//
//  Returns
//
//  BOOL  -			TRUE if the scan was successful
//                  FALSE if the scan was not successful
//
//  API Use Notes:
//
//
BOOL DXDirScanner::ProcessRequest(DXRequest *dxrRQ, int iThreadID)
{
	//  Entry point used for single threaded request execution
	return CoreProcessRequest(dxrRQ, iThreadID);
}

//  ScanDirectory
//
//  This function will scan a directory and generate any applicable scan requests for the same or next level
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the invoking thread
//
//  Returns
//
//  BOOL  -			TRUE if the scan was successful
//                  FALSE if the scan was not successful
//
//  API Use Notes:
//
//
BOOL DXDirScanner::ScanDirectory(DXRequest *dxrRQ, int iThreadID)
{
	STATUS		stAPIRC;								//  Return code from Notes API calls
	DBHANDLE	hdbEntity;								//  Handle of the entity
	USHORT		usDbType;								//  Database type returned ny NSFDbGetMode
	WORD		wSearchFor;								//  Search type restrictions
	char		szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Scan the requested directory
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1303I, dxrRQ->szDirectory);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Open the passed entity and verify it is a directory
	if (dxrRQ->hdbCache == NULLHANDLE)
	{
		hdbEntity = xeLocal->OpenDatabase(dxrRQ->szServer, dxrRQ->szDirectory, &usDbType, iThreadID);
		if (hdbEntity == NULLHANDLE)
		{
			//  Unable to access the requested directory
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1305E, dxrRQ->szDirectory);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		if (usDbType != DB_DIRECTORY)
		{
			//  Requested item is not a directory
			NSFDbClose(hdbEntity);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1308E, dxrRQ->szDirectory);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}
	}
	else
	{
		//  Already verified and the handle has been cached
		hdbEntity = dxrRQ->hdbCache;
	}

	//  Determine the search scope
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1307T, dxrRQ->szDirectory, dxrRQ->szMatchDirectory);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	wSearchFor = FILE_ANYNOTEFILE + FILE_DIRS + FILE_NOUPDIRS;
	bProcessOnlyDirs = FALSE;
	if (dxrRQ->Wildcards & WILDCARD_DIRECTORY)
	{
		//  If the current directory does not (yet) match then restrict search results to directories
		if (!xeLocal->MatchesPatterni(dxrRQ->szDirectory, dxrRQ->szMatchDirectory))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1327T, dxrRQ->szDirectory);
			xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
			bProcessOnlyDirs = TRUE;
		}
	}

	iSearchThread = iThreadID;

	stAPIRC = NSFSearch(hdbEntity,
						NULLHANDLE,
						NULL,
						SEARCH_FILETYPE + SEARCH_SUMMARY,
						wSearchFor,
						NULL,
						WrapperProcessSearchResult,
						this,
						NULL);

	//  If the search was terminated early due to a recursion then treat as a normal termination
	if ((ERR(stAPIRC) == ERR_NOEXIST) && (dxrRQ->ReturnCode == RETURN_RECURSION))
	{
		//  Clear the return codes
		stAPIRC = NOERROR;
		dxrRQ->ReturnCode = RETURN_NOERROR;
	}
	if (ERR(stAPIRC) != NOERROR)
	{
		DXException		*dxeLocal;

		//  First check if there is a deferred exception pending - if so throw it
		if (dxeDeferred != NULL)
		{
			throw(*dxeDeferred);
		}

		//  This should never fail - throw an exception
		dxeLocal = new DXException;
		sprintf_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1309E, dxrRQ->szDatabase);
		dxeLocal->ExceptionType = ERRORTYPE_NOTESAPI;
		dxeLocal->ExceptionCode = stAPIRC;
		dxeLocal->FaultingThread = iThreadID;
		dxrRQ->ReturnCode = RETURN_INCOMPLETE;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
		dxeLocal->LineNumber = __LINE__;

		//  Close the current database/directory 
		if (hdbEntity != NULLHANDLE)
		{
			NSFDbClose(hdbEntity);
		}

		//  Throw the new exception
		throw(*dxeLocal);
	}

	//  Close the current database handle
	if (hdbEntity != NULLHANDLE)
	{
		NSFDbClose(hdbEntity);
	}

	//  Delete the lower level directory or database scanners
	if (dxdsNextLevel != NULL)
	{
		delete dxdsNextLevel;
		dxdsNextLevel = NULL;
	}
	if (dxbsInstance != NULL)
	{
		delete dxbsInstance;
		dxbsInstance = NULL;
	}

	//  Trace the exit
	xeLocal->LogTrace(TRACE_EXPLORER, MSG_DXR1318T, iThreadID);
	return TRUE;
}

//  WrapperProcessSearchResult
//
//  This is a static function that provides a wrapper for an NSFSearch callback routine.
//
//  Parameters
//
//  1  -  void *			- Pointer to this DXDirScanner object
//  2  -  SEARCH_MATCH *	- Pointer to the SEARCH_MATCH structure
//  3  -  ITEM_TABLE *		- Pointer to the search summary buffer for the current hit
//
//  Returns
//
//  STATUS		-		See NSFSearch for uses of the return value from this callback
//
//  API Use Notes:
//
//
STATUS LNCALLBACK DXDirScanner::WrapperProcessSearchResult(void *objPtr, SEARCH_MATCH *lpSM, ITEM_TABLE *lpITResult)
{
	//  The static wrapper is needed to provide linkage for the callback entry point

	DXDirScanner		*dxdsCurrent;

	//  Cast the void pointer back to this object type and call the real callback routine
	dxdsCurrent = (DXDirScanner *) objPtr;
	return dxdsCurrent->ProcessSearchResult(lpSM, lpITResult, dxdsCurrent->iSearchThread);
}

//  ProcessSearchResult
//
//  This is the NSFSearch callback routine.
//
//  Parameters
//
//  1  -  SEARCH_MATCH *	- Pointer to the SEARCH_MATCH structure
//  2  -  ITEM_TABLE *		- Pointer to the search summary buffer for the current hit
//  3  -  int				- ID of the thread performing the search
//
//  Returns
//
//  STATUS		-		See NSFSearch for uses of the return value from this callback
//
//  API Use Notes:
//
//
STATUS DXDirScanner::ProcessSearchResult(SEARCH_MATCH *lpSM, ITEM_TABLE *lpITResult, int iThreadID)
{
	BOOL				bProcessThisRequest;						//  Control Switch
	DXRequest			*dxrNextLevel;								//  Next Level Request
	DXFilter			*dxfSelector;								//  Filter Object
	SEARCH_MATCH		smCurrent;									//  Local SEARCH_MATCH structure
	char				*pSource;									//  Pointer into strings
	char				*pTarget;									//  Pointer into strings
	char				szNewFileName[MAXPATH + 1];					//  File name from this search hit
	char				szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  All exceptions are caught here and deferred until exit from the callback
	//  Process a search result

	try
	{
		//  Copy the passed structures to local storage in case they are not word aligned
		memcpy((char *) &smCurrent, (char *) lpSM, sizeof(SEARCH_MATCH));

		//  Check that the current item was matched
		if (!smCurrent.SERetFlags & SE_FMATCH) return(NOERROR);

		//  Get the file (directory) name returned by NSFSearch
		if (!NSFGetSummaryValue(lpITResult, FILE_PATH_ITEM, szNewFileName, MAXPATH))
		{
			DXException		*dxeLocal;

			//  Post a deferred exception
			dxeLocal = new DXException();
			strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1310E);
			dxeLocal->ExceptionType = ERRORTYPE_FAIL;
			dxeLocal->ExceptionCode = ERRORCODE_ITM_PARSE;
			dxeLocal->FaultingThread = iThreadID;
			strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
			dxeLocal->LineNumber = __LINE__;
			dxeDeferred = dxeLocal;
			return ERR_CANNOT_OPEN_NSF;
		}

		//  Trace and Debug Message
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1311T, szNewFileName);
		xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);

		//  Build a new explorer request from the returned item
		dxrNextLevel = new DXRequest();
		dxrNextLevel->InheritFrom(dxrCurrent);
		strcpy_s(dxrNextLevel->szDomain, MAX_DOMAIN, dxrCurrent->szDomain);
		strcpy_s(dxrNextLevel->szServer, MAX_SERVER, dxrCurrent->szServer);
		strcpy_s(dxrNextLevel->szDirectory, MAX_DATABASE, dxrCurrent->szDirectory);
		strcpy_s(dxrNextLevel->szDatabase, MAX_DATABASE, szNewFileName);
	
		//  Determine if the request is for a directory or a database
		if (IsRequestForDirectory(dxrNextLevel, iThreadID))
		{
			//  Detect a recursion situation where the search has returned an entry that cannot be opened because it does not exist
			if (dxrNextLevel->stNoAccess == ERR_NOEXIST)
			{
				dxrCurrent->ReturnCode = RETURN_RECURSION;
				return ERR_NOEXIST;
			}
			if (dxrNextLevel->AnchorScope == RSCOPE_DIRECTORY)
			{
				pSource = szNewFileName;
				pTarget = dxrNextLevel->szDirectory;
				while (*pSource != '\0')
				{
					if ((*pSource == '\\') || (*pSource == '/')) *pTarget = '/';
					else *pTarget = *pSource;
					pSource++;
					pTarget++;
				}
				*pTarget = '\0';
				dxrNextLevel->szDatabase[0] = '\0';

				//  Trace Directory processing
				sprintf_s(szMsg, MAX_MSG, MSG_DXR1326T, dxrNextLevel->szDirectory, dxrNextLevel->RQLevel, dxrNextLevel->szMatchDirectory);
				xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);

				//  Determine if we should skip this directory entry
				bProcessThisRequest = TRUE;

				//  Determine if the entry matches
				if (dxrCurrent->Wildcards & WILDCARD_DIRECTORY)
				{
					if (!xeLocal->MatchesPatterni(dxrNextLevel->szDirectory, dxrNextLevel->szMatchDirectory)) bProcessThisRequest = FALSE;
				}
				if (bProcessThisRequest)
				{
					//  Check to see if recursion is inhibited - if so skip the request
					if (!dxrCurrent->RecurseDirectories) bProcessThisRequest = FALSE;
				}
				if (bProcessThisRequest)
				{
					if ((dxrNextLevel->Filters & FILTER_DIRECTORY) && (dxrNextLevel->dxfDirectory != NULL))
					{
						dxfSelector = (DXFilter *) dxrNextLevel->dxfDirectory;
						if (!dxfSelector->FilterThisDirectory(dxrNextLevel, iThreadID)) bProcessThisRequest = FALSE;
					}
				}
				
				//  If recursion into sub-directories is not allowed then skip this entry
				if (!dxrNextLevel->RecurseDirectories)
				{
					dxrCurrent->dwDirectoriesDropped++;
					//  Delete the spawned request and continue
					delete dxrNextLevel;
					return NOERROR;
				}
				//  Check that we do not have a recursion issue
				//  It has been observed that directory paths can exceed MAXPATH and this can cause
				//  the explorer to keep processing the same directory
				if (strcmpi(dxrCurrent->szDirectory, dxrNextLevel->szDirectory) == 0)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR1322E, dxrCurrent->szDirectory);
					xeLocal->LogMessage(szMsg);
					dxrCurrent->dwDirectoriesDropped++;

					//  Delete the spawned request and continue
					delete dxrNextLevel;
					dxrCurrent->ReturnCode = RETURN_RECURSION;
					return ERR_NOEXIST;
				}

				if (bProcessThisRequest)
				{
					//  Request is for a directory - pass to the next level Directory Scanner
					if (dxdsNextLevel == NULL)
					{
						dxdsNextLevel = new DXDirScanner(xeLocal, iThreadID);
					}
					if (!dxdsNextLevel->ProcessRequest(dxrNextLevel, iThreadID))
					{
						DXException		*dxeLocal;
					
						//  Accumulate scores to parent
						dxrNextLevel->PromoteScores(dxrCurrent);

						//  Post a deferred exception
						dxeLocal = new DXException();
						strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1313E);
						dxeLocal->ExceptionType = ERRORTYPE_FAIL;
						dxeLocal->ExceptionCode = ERRORCODE_NON_SPEC;
						dxeLocal->FaultingThread = iThreadID;
						strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
						dxeLocal->LineNumber = __LINE__;
						dxeDeferred = dxeLocal;
						return ERR_CANNOT_OPEN_NSF;					//  Force an exit to the callback loop
					}
					else
					{
						//  Explosion completed
						if (dxrNextLevel->AnchorScope == RSCOPE_DIRECTORY)
						{
							sprintf_s(szMsg, MAX_MSG, MSG_DXR1314I, dxrNextLevel->szDirectory);
							xeLocal->LogVerbose(szMsg);
						}

						//  Accumulate scores to parent
						dxrNextLevel->PromoteScores(dxrCurrent);
						delete dxrNextLevel;
						return NOERROR;								//  Signal continue with the search
					}
				}
				else
				{
					//  Directory has been dropped from this search scope
					dxrCurrent->dwDirectoriesDropped++;
					sprintf_s(szMsg, MAX_MSG, MSG_DXR1323I, dxrNextLevel->szDirectory);
					xeLocal->LogVerbose(szMsg, iThreadID);
					delete dxrNextLevel;
					return NOERROR;
				}
			}
			else
			{
				//  If we are only processing directories then discard any database
				if (bProcessOnlyDirs)
				{
					delete dxrNextLevel;
					return NOERROR;									//  Signal continue with the search
				}
				//  Adjust the database name to eliminate the directory
				pSource = dxrNextLevel->szDatabase;
				while (*pSource != '\0')
				{
					if ((*pSource == '\\') || (*pSource == '/'))
					{
						strcpy_s(dxrNextLevel->szDatabase, MAX_DATABASE, pSource + 1);
						pSource = dxrNextLevel->szDatabase;
					}
					else pSource++;
				}
				if (dxrNextLevel->hdbCache != NULLHANDLE)
				{
					//  Request is for a database - pass to the Database Scanner
					if (dxbsInstance == NULL)
					{
						dxbsInstance = new DXDbScanner(xeLocal, iThreadID);
					}

					//  Determine if the request should be processed
					bProcessThisRequest = TRUE;
					if (dxrCurrent->Wildcards & WILDCARD_DATABASE)
					{
						if (!xeLocal->MatchesPatterni(dxrNextLevel->szDatabase, dxrNextLevel->szMatchDatabase)) bProcessThisRequest = FALSE;
					}
					if (bProcessThisRequest)
					{
						if ((dxrCurrent->Filters & FILTER_DATABASE) && (dxrCurrent->dxfDatabase != NULL))
						{	
							dxfSelector = (DXFilter *) dxrCurrent->dxfDatabase;
							if (!dxfSelector->FilterThisDatabase(dxrNextLevel, iThreadID)) bProcessThisRequest = FALSE;
						}
					}

					if (bProcessThisRequest)
					{
						if (!dxbsInstance->ProcessRequest(dxrNextLevel, iThreadID))
						{
							DXException		*dxeLocal;

							//  Accumulate scores to parent
							dxrNextLevel->PromoteScores(dxrCurrent);

							//  Post a deferred exception
							dxeLocal = new DXException();
							strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1315E);
							dxeLocal->ExceptionType = ERRORTYPE_FAIL;
							dxeLocal->ExceptionCode = ERRORCODE_NON_SPEC;
							dxeLocal->FaultingThread = iThreadID;
							strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
							dxeLocal->LineNumber = __LINE__;
							dxeDeferred = dxeLocal;
							return ERR_CANNOT_OPEN_NSF;						//  Force an exit from the callback loop
						}
						else
						{
							//  Explosion completed
							sprintf_s(szMsg, MAX_MSG, MSG_DXR1316I, dxrNextLevel->szDatabase);
							xeLocal->LogVerbose(szMsg);

							//  Accumulate scores to parent
							dxrNextLevel->PromoteScores(dxrCurrent);
							delete dxrNextLevel;
							return NOERROR;									//  Signal continue with the search
						}
					}
					else
					{
						//  Database has been dropped from this search scope
						dxrCurrent->dwDatabasesDropped++;
						sprintf_s(szMsg, MAX_MSG, MSG_DXR1324I, dxrNextLevel->szDirectory, dxrNextLevel->szDatabase);
						xeLocal->LogVerbose(szMsg, iThreadID);
						delete dxrNextLevel;
						return NOERROR;
					}
				}
				else
				{
					//  There was no access to the current database - signal the scan to continue
					//  Pass the request to the NoAccess reporter - if one is present
		
					//  Check if a D/B Reporter is needed
					if (dxrNextLevel->Reporters & DXREP_DATABASE)
					{
						DXReporter   *dxrepDb;

						//  D/B level reporter is required, invoke it
						dxrepDb = (DXReporter *) dxrNextLevel->dxrDatabase;
						dxrepDb->ReportNoAccess(dxrNextLevel, iThreadID);
					}
					delete dxrNextLevel;
					return NOERROR;
				}
			}
		}
		else
		{	
			//  Unable to access the requested resource - report and continue
			//  NOTE:  This processing path is deprecated.
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1308E, dxrNextLevel->szDatabase);
			xeLocal->LogMessage(szMsg);
			delete dxrNextLevel;
			return NOERROR;
		}
	}
	catch (DXException& dxeCallback)
	{
		//  A lower level explorer exception has been trapped - defer it and exit the callback loop
		dxeCallback.IsPercolated = true;					//  Mark the exception as percolated
		dxeDeferred = &dxeCallback;
	}
	catch (std::exception eCallback)
	{
		DXException		*dxeLocal;

		//  A lower level exception has been trapped - wrapper the exception and defer it
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_UHX;
		dxeLocal->ExceptionCode = ERRORCODE_NON_SPEC;
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->LocalExceptionMessage, MAX_MSG, eCallback.what());
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1317E);
		dxeLocal->DestroyContext = true;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
		dxeLocal->LineNumber = __LINE__;	
		dxeDeferred = dxeLocal;
	}

	//   Pass a bad return code to terminate the callback
	return ERR_CANNOT_OPEN_NSF;
}

//  IsRequestForDirectory
//
//  This function will determine if the request is for a directory or a database
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the invoking Thread
//
//  Returns
//
//  BOOL  -			TRUE if the request scope was determined
//                  FALSE if the request scope could not be determined
//
//  API Use Notes:
//
//
BOOL DXDirScanner::IsRequestForDirectory(DXRequest *dxrRQ, int iThreadID)
{
	DBHANDLE	hdbEntity;								//  Handle of the entity
	USHORT		usDbType;								//  Database type returned ny NSFDbGetMode

	char				szMsg[MAX_MSG + 1];				//  Generic message buffer

	//  Mark the request incomplete
	dxrRQ->ReturnCode = RETURN_INCOMPLETE;
	dxrRQ->stNoAccess = NOERROR;

	//  Open the passed entity and verify it is a directory

	hdbEntity = xeLocal->OpenDatabase(dxrRQ->szServer, dxrRQ->szDatabase, &usDbType, iThreadID);
	if (hdbEntity == NULLHANDLE)
	{
		//  Save the status from the open in the request
		dxrRQ->stNoAccess = xeLocal->GetLastOpenError();
		//  Unable to access the requested database/directory
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1305E, dxrRQ->szDatabase);
		xeLocal->LogMessage(szMsg, iThreadID);
		dxrRQ->hdbCache = NULLHANDLE;
		dxrRQ->AnchorScope = RSCOPE_DATABASE;
		return TRUE;
	}

	//  Cache the D/B handle in the request
	dxrRQ->hdbCache = hdbEntity;

	//  Set the scope
	dxrRQ->ReturnCode = RETURN_NOERROR;

	//  Trace the mode
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1325T, dxrRQ->szDatabase, usDbType);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);

	if (usDbType == DB_DIRECTORY)
	{
		dxrRQ->AnchorScope = RSCOPE_DIRECTORY;
	}
	else
	{
		dxrRQ->AnchorScope = RSCOPE_DATABASE;
	}
	return TRUE;
}

//  DestroySubContexts
//
//  This function will destroy any scan contexts bolow the level of this scanner
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void DXDirScanner::DestroySubContexts(void)
{
	//  Destroy a Database Scanner
	if (dxbsInstance != NULL)
	{
		delete dxbsInstance;
		dxbsInstance = NULL;
	}
	
	//  Destroy a lower level Directory Scanner
	if (dxdsNextLevel != NULL)
	{
		delete dxdsNextLevel;
		dxdsNextLevel = NULL;
	}
}

//  CoreProcessRequest
//
//  This function is the main execution routine for requests at the Directory scan level
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the invoking request
//
//  Returns
//
//  BOOL  -			TRUE if the scan was successful
//                  FALSE if the scan was not successful
//
//  API Use Notes:
//
//
BOOL DXDirScanner::CoreProcessRequest(DXRequest *dxrRQ, int iThreadID)
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Process a scan request for a directory, or pass through a lower level request.
	dxrCurrent = dxrRQ;

	try
	{
		//  Determine if the scope of the search is a directory
		if (dxrRQ->AnchorScope == RSCOPE_DIRECTORY)
		{
			dxrRQ->SetStart(xeLocal);
			//  Process the request locally
			if (ScanDirectory(dxrRQ, iThreadID))
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR1304I, dxrRQ->szDirectory);
				xeLocal->LogVerbose(szMsg, iThreadID);

				dxrRQ->dwDirectoriesProcessed++;
				dxrRQ->ReturnCode = RETURN_NOERROR;
				//  Document the request
				dxrRQ->SetEnd(xeLocal);
				if (xeLocal->rsLocal->LogLevel > LOGLEVEL_VERBOSE) dxrRQ->DocumentRequest(xeLocal, iThreadID);
				return TRUE;
			}
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1301E, dxrRQ->szDatabase);
			xeLocal->LogMessage(szMsg, iThreadID);
			dxrRQ->dwDirectoriesProcessed++;
			dxrRQ->ReturnCode = RETURN_CERR;
			//  Document the request
			dxrRQ->SetEnd(xeLocal);
			if (xeLocal->rsLocal->LogLevel > LOGLEVEL_VERBOSE) dxrRQ->DocumentRequest(xeLocal, iThreadID);
			return FALSE;
		}
		else
		{		
			//  Pass the request through to a lower level
			xeLocal->LogTrace(TRACE_EXPLORER, MSG_DXR1320T);

			//  If the lower level Database Scanner does not exist then create it
			if (dxbsInstance == NULL)
			{
				dxbsInstance = new DXDbScanner(xeLocal, iThreadID);
			}
			
			//  Process the request at a lower level in the same thread
			return dxbsInstance->ProcessRequest(dxrRQ, iThreadID);   //  Process the request		
		}
	}
	catch(DXException& dxeCurrent)
	{
		//  Destroy the lower level scanners if they are unstable
		if (dxeCurrent.DestroyContext)
		{
			DestroySubContexts();
		}

		//  Mark the Exception as percolated
		dxeCurrent.IsPercolated = true;

		//  Rethrow the exception to the next level - percolate
		throw;
	}
	catch(std::exception& eCurrent)
	{
		DXException		*dxeLocal;

		//  This level and lower levels may be unstable - destroy them
		DestroySubContexts();

		//  Create a new exception to wrapper the information from the original
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_UHX;
		dxeLocal->ExceptionCode = ERRORCODE_NON_SPEC;
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->LocalExceptionMessage, MAX_MSG, eCurrent.what());
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1302E);
		dxeLocal->DestroyContext = true;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
		dxeLocal->LineNumber = __LINE__;
		throw(*dxeLocal);
	}
	return FALSE;
}
