//  DXNoteScanner
//
//  This class is the eXplorer class that handles the scanning of the notes within a database according
//  to the search criteria
//
//  Message Identifier:      DXR
//  Message Range:           1801 - 1899
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.14.0
//  Version:                 3.14.0  Build: 108
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.9.2		22/09/2010		Avoided a recursion in DX processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.11.1		22/10/2011		VS 2011 Updates
//	3.14.0		21/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2015
//
//

#include "DXNoteScanner.h"

DXNoteScanner::DXNoteScanner(ExecEnvironment *xeParent, int iThreadID)
{
	//  Constructor
	int					iIndex;								//  Generic Index
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Set local members
	OwningThread = iThreadID;
	xeLocal = xeParent;
	hitNotesToScan = NULLHANDLE;
	dwNotesInserted = 0;
	bDbLocallyOpened = FALSE;
	bDbSearchPerformed = FALSE;
	dwGlobalScanFlags = 0;
	wGlobalNoteClasses = 0;

	//  Initialize internal formula data structure
	for (iIndex = 0; iIndex < MAX_SEARCH_ENTRIES; iIndex ++)
	{
		nsieTable[iIndex].szSearchString[0] = '\0';
		nsieTable[iIndex].hfSearch = NULLHANDLE;
		nsieTable[iIndex].phfSearch = NULL;
		nsieTable[iIndex].hcSearch = NULL;
	}

	// Trace construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1801T, this);
	xeLocal->LogTrace(TRACE_DBSCAN, szMsg);
}

DXNoteScanner::~DXNoteScanner(void)
{
	//  Destructor
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Destroy internal base IDTable if required
	if ( hitNotesToScan != NULLHANDLE )
	{
		IDDestroyTable(hitNotesToScan);
		hitNotesToScan = NULLHANDLE;
	}

	//  Trace the destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1802T, this);
	xeLocal->LogTrace(TRACE_DBSCAN, szMsg);
}

//  ScanDb
//
//  This is the main entry point into the DXNoteScanner
//
//  Parameters
//
//  1  -  DXNoteScanReq *	-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL	-  TRUE if actions were successful
//			-  FALSE if actions failed
//
//  API Use Notes:
//
//
BOOL DXNoteScanner::ScanDb(DXNoteScanReq *nsrCurrent, int iThreadID)
{
	DXNoteScanReq		*nsrRQ;								//  Pointer to DXNoteScanReq object
	clock_t				ct_Start;							//  Scan Start Timestamp
	clock_t				ct_End;								//  Scan End Timestamp
	long				lElapsed;							//  Elapsed Time
	long				lThroughput;						//  Throughput total
	long				lTimeFactor = CLOCKS_PER_SEC;		//  Clock tick conversion factor
	char				szLT[2];							//  Qualifier for zero second timing
	char				szGT[2];							//  Qualifier for zero second timing
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Sanity check the invocation

	if ( nsrCurrent == NULL )
	{
		xeLocal->LogMessage(MSG_DXR1803E, iThreadID);
		return FALSE;
	}
	nsrRQ = nsrCurrent;

	//  Show that we are starting
	xeLocal->LogVerbose(MSG_DXR1810I, iThreadID);
	if (!CheckSourceReq(nsrRQ, iThreadID) )
	{
		nsrRQ->iReturnCode = RETURN_CERR;
		return FALSE;
	}

	//  Open the database if required (i.e. we don't have a D/B Handle yet).
	if (nsrRQ->hdbToScan == NULLHANDLE)
	{
		nsrRQ->hdbToScan = xeLocal->OpenDatabase(nsrRQ->szServer, nsrRQ->szDbPath, iThreadID);
		if (nsrRQ->hdbToScan == NULLHANDLE)
		{
			xeLocal->LogMessage(MSG_DXR1805E, iThreadID);
			nsrRQ->iReturnCode = RETURN_CERR;
			return FALSE;
		}
		xeLocal->LogVerbose(MSG_DXR1806I, iThreadID);
		bDbLocallyOpened = TRUE;
	}
	else
	{
		xeLocal->LogVerbose(MSG_DXR1807I, iThreadID);
	}

	//  Record the start time
	ct_Start = clock();

	//  Perform applicable database search to get base IDTable
	if (!SearchDb(nsrRQ, iThreadID))
	{
		xeLocal->LogMessage(MSG_DXR1842E, iThreadID);

		//  We need to cleanup internal formula data structure table
		CleanupIntTable(iThreadID);

		//  Close db if required
		if ((nsrRQ->hdbToScan != NULLHANDLE) && (bDbLocallyOpened))
		{
			NSFDbClose(nsrRQ->hdbToScan);
			xeLocal->LogMessage(MSG_DXR1814I, iThreadID);
		}

		//  Destroy internal base IDTable if required
		if (hitNotesToScan != NULLHANDLE)
		{
			IDDestroyTable(hitNotesToScan);
			hitNotesToScan = NULLHANDLE;
		}

		nsrRQ->iReturnCode = RETURN_CERR;
		return FALSE;
	}
	nsrRQ->dwNotesToScan = IDEntries(hitNotesToScan);

	//  Check to see if we got any notes that matched the base search criteria
	if (nsrRQ->dwNotesToScan == 0)
	{
		xeLocal->LogMessage(MSG_DXR1843W, iThreadID);
		
		//  We need to cleanup internal formula data structure table
		CleanupIntTable(iThreadID);

		//  Close db if required
		if ((nsrRQ->hdbToScan != NULLHANDLE) && (bDbLocallyOpened))
		{
			NSFDbClose(nsrRQ->hdbToScan);
			xeLocal->LogMessage(MSG_DXR1814I, iThreadID);
		}

		//  Destroy internal base IDTable if required
		if (hitNotesToScan != NULLHANDLE)
		{
			IDDestroyTable(hitNotesToScan);
			hitNotesToScan = NULLHANDLE;
		}

		nsrRQ->iReturnCode = RETURN_NOERROR;
		return TRUE;
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR1824I, nsrRQ->dwNotesToScan);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Now process the base IDTable and populate our search request
	if (!ProcessNotes(nsrRQ, iThreadID))
	{
		xeLocal->LogMessage(MSG_DXR1844E, iThreadID);

		//  We need to cleanup internal formula data structure table
		CleanupIntTable(iThreadID);

		//  Close db if required
		if ( (nsrRQ->hdbToScan != NULLHANDLE) && (bDbLocallyOpened))
		{
			NSFDbClose(nsrRQ->hdbToScan);
			xeLocal->LogMessage(MSG_DXR1814I, iThreadID);
		}

		//  Destroy internal base IDTable if required
		if ( hitNotesToScan != NULLHANDLE )
		{
			IDDestroyTable(hitNotesToScan);
			hitNotesToScan = NULLHANDLE;
		}

		nsrRQ->iReturnCode = RETURN_CERR;
		return FALSE;
	}

	//  Record the ending time
	ct_End = clock();

	//  We need to cleanup internal formula data structure table
	CleanupIntTable(iThreadID);

	//  Close the database if required
	if ((nsrRQ->hdbToScan != NULLHANDLE) && (bDbLocallyOpened))
	{
		NSFDbClose(nsrRQ->hdbToScan);
		xeLocal->LogMessage(MSG_DXR1814I, iThreadID);
	}

	//  Destroy internal base IDTable if required
	if ( hitNotesToScan != NULLHANDLE )
	{
		IDDestroyTable(hitNotesToScan);
		hitNotesToScan = NULLHANDLE;
	}

	xeLocal->LogVerbose(MSG_DXR1813I, iThreadID);

	//  Document the run statistics
	lElapsed = (ct_End - ct_Start) / lTimeFactor;

	//  Defend against zero divides when total elapsed time is less than a second
	strcpy_s(szLT, 3, " ");
	strcpy_s(szGT, 3, " ");
	if (lElapsed == 0)
	{
		strcpy_s(szLT, 3, "<");
		strcpy_s(szGT, 3, ">");
		lElapsed = 1;
	}

	nsrRQ->dwElapsed = (DWORD) lElapsed;

	//  Compute and display the throughput in documents per second
	lThroughput = (nsrRQ->dwNotesScanned) / nsrRQ->dwElapsed;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1815I, szLT, lElapsed, szGT, lThroughput);
	xeLocal->LogVerbose(szMsg, iThreadID);

	nsrRQ->iReturnCode = RETURN_NOERROR;
	return TRUE;
}

//  CheckSourceReq
//
//  This checks the supplied search entry table to make sure its valid
//
//  Parameters
//
//  1  -  DXNoteScanReq *	-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL	-  TRUE if actions were successful
//			-  FALSE if actions failed
//
//  API Use Notes:
//
//
BOOL DXNoteScanner::CheckSourceReq(DXNoteScanReq *nsrCurrent, int iThreadID)
{
	BOOL				bRC;									//  Return code of function
	DXNoteScanReq		*nsrRQ;									//  Pointer to DXNoteScanReq object
	int					iIndex;									//  Index counter
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	nsrRQ = nsrCurrent;
	bRC = TRUE;

	//  Check source db handle / server & path
	if (nsrRQ->hdbToScan == NULLHANDLE)
	{
		if (nsrRQ->szServer[0] == '\0')
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1811E, "Server");
			xeLocal->LogMessage(szMsg, iThreadID);
			bRC = FALSE;
		}
		if (nsrRQ->szDbPath[0] == '\0')
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1811E, "DbPath");
			xeLocal->LogMessage(szMsg, iThreadID);
			bRC = FALSE;
		}
	}

	//  Check at least the first search entry IDTable is not NULL
	if (nsrRQ->nsObject->dxnsEntry[0].hitMatches == NULLHANDLE)
	{
		xeLocal->LogMessage(MSG_DXR1812E, iThreadID);
		bRC = FALSE;
	}
	else
	{
		for (iIndex = 0; iIndex < MAX_SEARCH_ENTRIES; iIndex++)
		{
			//  Check we don't have any unsupported search entries
			//  Check to make sure flags does not contain DELSTUBS or NOACCESS since
			//  you cannot run a formula against a note without a valid open note handle!
			if (nsrRQ->nsObject->dxnsEntry[iIndex].szSearchString[0] != '\0')
			{
				if ((nsrRQ->nsObject->dxnsEntry[iIndex].dwSearchFlags & FLAG_NOTE_SCAN_DELSTUBS) |
					(nsrRQ->nsObject->dxnsEntry[iIndex].dwSearchFlags & FLAG_NOTE_SCAN_NOACCESS))
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR1847E, iIndex);
					xeLocal->LogMessage(szMsg, iThreadID);
					bRC = FALSE;
				}
			}

			//  Reset added count incase caller has not!
			nsrRQ->nsObject->dxnsEntry[iIndex].dwCount = 0;
		}
	}
	return bRC;
}

//  SearchDb
//
//  This functions determines whether an NSFSearch is required or not, and compiles the formulas
//  if required
//
//  Parameters
//
//  1  -  DXNoteScanReq *	-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL	-  TRUE if actions were successful
//			-  FALSE if actions failed
//
//  API Use Notes:
//
//
BOOL DXNoteScanner::SearchDb(DXNoteScanReq *nsrCurrent, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API Calls
	STATUS				stCompile;								//  Status of formula compile
	BOOL				bPerformNSFSearch;						//  Indicate whether we should perform a NSFSearch
	BOOL				bContinue;								//  Continue check
	BOOL				bAllFormulasMatch;						//  Indicator that all search formulas specified are the same
	BOOL				bSearchFound;							//  Indicates if one or more search strings was found
	BOOL				bInvalidSearchCriteria;					//  Indicates if one or more search entries are invalid
	FORMULAHANDLE		hfNSFSearch;							//  Handle to compiled formula used with NSFSearch
	TIMEDATE			tdStart;								//  TimeDate used by NSFDbGetModifiedNoteTable
	TIMEDATE			tdFinish;								//  TimeDate used by NSFDbGetModifiedNoteTable
	WORD				wFormulaLen;							//  Length of the returned compiled formula
	WORD				wErrorLine;								//  Error line returned by NSFFormulaCompile
	WORD				wErrorColumn;							//  Error column returned by NSFFormulaCompile
	WORD				wErrorOffset;							//  Error offset returned by NSFFormulaCompile
	WORD				wErrorLength;							//  Error length returned by NSFFormulaCompile
	WORD				wSearchStringsUsed;						//  # of search strings added to NSFSearch
	DXNoteScanReq		*nsrRQ;									//  Pointer to DXNoteScanReq object
	char				szNSFSearch[MAX_SEARCH_FORMULA_LENGTH];	//  NSFSearch formula buffer
	char				szPrevFormula[MAX_SEARCH_STRING];		//  Buffer to contain previous search string
	int					iIndex;									//  Index counter
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	nsrRQ = nsrCurrent;

	//  Do an initial check of the supplied search entries to get our global flags/note class values
	//  and determine whether or not we need to handle any search strings
	bContinue = TRUE;
	bPerformNSFSearch = TRUE;
	bSearchFound = FALSE;
	iIndex = 0;
	bAllFormulasMatch = TRUE;
	szPrevFormula[0] = '\0';
	bInvalidSearchCriteria = FALSE;
	xeLocal->LogTrace(TRACE_DBSCAN, MSG_DXR1850T, iThreadID);
	while (bContinue)
	{

		if (iIndex >= MAX_SEARCH_ENTRIES) bContinue = FALSE;
		else if (nsrRQ->nsObject->dxnsEntry[iIndex].hitMatches == NULLHANDLE) bContinue = FALSE;
		else
		{

			sprintf_s(szMsg, MAX_MSG, 
					  MSG_DXR1851T, 
					  nsrRQ->nsObject->dxnsEntry[iIndex].hitMatches,
					  nsrRQ->nsObject->dxnsEntry[iIndex].wNoteClass,
					  nsrRQ->nsObject->dxnsEntry[iIndex].dwSearchFlags,
					  nsrRQ->nsObject->dxnsEntry[iIndex].szSearchString);
			xeLocal->LogTrace(TRACE_DBSCAN, szMsg, iThreadID);

			//  Determine which note classes we should be searching for
			wGlobalNoteClasses = (wGlobalNoteClasses | nsrRQ->nsObject->dxnsEntry[iIndex].wNoteClass);

			//  Setup global flags
			dwGlobalScanFlags = (dwGlobalScanFlags | nsrRQ->nsObject->dxnsEntry[iIndex].dwSearchFlags);

			//  See if we have a search string for this search entry
			if (nsrRQ->nsObject->dxnsEntry[iIndex].szSearchString[0] == '\0') bPerformNSFSearch = FALSE;
			else
			{
				//  Indicate we found at least one search string
				bSearchFound = TRUE;

				//  Check to see if formula matches previous formula
				if (szPrevFormula[0] == '\0') strcpy_s(szPrevFormula, MAX_SEARCH_STRING, nsrRQ->nsObject->dxnsEntry[iIndex].szSearchString);
				else
				{
					if ( strcmpi(szPrevFormula, nsrRQ->nsObject->dxnsEntry[iIndex].szSearchString) != 0 )
					{
						bAllFormulasMatch = FALSE;
					}
				}
			} 
		} 
		iIndex++;
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR1849T, iIndex-1, wGlobalNoteClasses, dwGlobalScanFlags, bSearchFound, bAllFormulasMatch);
	xeLocal->LogTrace(TRACE_DBSCAN, szMsg, iThreadID);


	//  Now setup search string handling (if we need to).  If all search strings were the same
	//  we don't need to setup individual search entries and we can keep the NSFSearch simple.  If
	//  search strings were different, then we need to setup individual search strings, and do the
	//  necessary formula actions and store results in internal table for use later by note processing
	szNSFSearch[0] = '\0';
	wSearchStringsUsed = 0;
	xeLocal->LogTrace(TRACE_DBSCAN, MSG_DXR1852T, iThreadID);
	if (bSearchFound)
	{
		if ((bAllFormulasMatch) && (bPerformNSFSearch))
		{
			//  As we need to do a NSFSearch and all formulas are the same...
			//  we only need to construct the search string and build the internal
			//  formula data structure
			if ((strlen(szPrevFormula) + strlen(FORMULA_START)) > MAX_SEARCH_FORMULA_LENGTH-1)
			{
				//  NSFSearch Formula too long!
				sprintf_s(szMsg, MAX_MSG, MSG_DXR1809E, MAX_SEARCH_FORMULA_LENGTH-1);
				xeLocal->LogMessage(szMsg, iThreadID);
				bInvalidSearchCriteria = TRUE;
			}
			else
			{
				//  Create beginning of select search string
				sprintf_s(szNSFSearch, MAX_SEARCH_FORMULA_LENGTH, FORMULA_START, szPrevFormula);
				wSearchStringsUsed++;
			}
		}
		else
		{
			//  One or more search strings do not match, therefore we need to iterate over them
			//  and build the internal formula data structure (and maybe NSFSearch search string)
			iIndex = 0;
			bContinue = TRUE;
			while (bContinue)
			{
				if (iIndex >= MAX_SEARCH_ENTRIES) bContinue = FALSE;
				else if (nsrRQ->nsObject->dxnsEntry[iIndex].hitMatches == NULLHANDLE) bContinue = FALSE;
				else if (nsrRQ->nsObject->dxnsEntry[iIndex].szSearchString[0] != '\0')
				{
					//  We have a search string to handle

					//  Copy search string into internal formula data structure
					strcpy_s(nsieTable[iIndex].szSearchString, MAX_SEARCH_STRING, nsrRQ->nsObject->dxnsEntry[iIndex].szSearchString);

					//  Lets compile the formula now and place its handle into the internal table
					//  This lets us validate this formula is OK before we add it to szNSFSearch
					stAPIRC = NSFFormulaCompile(NULL,
												0, 
												nsieTable[iIndex].szSearchString,
												(WORD) strlen(nsieTable[iIndex].szSearchString),
												&nsieTable[iIndex].hfSearch,
												&wFormulaLen,
												&stCompile,
												&wErrorLine,
												&wErrorColumn,
												&wErrorOffset,
												&wErrorLength);
					if (ERR(stAPIRC) != NOERROR)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR1808E, iIndex, nsieTable[iIndex].szSearchString);
						xeLocal->LogMessage(szMsg, iThreadID);
						sprintf_s(szMsg, MAX_MSG, MSG_DXR1819E, wErrorLine, wErrorColumn, wErrorOffset, wErrorLength);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
						xeLocal->LogMessage(szMsg, iThreadID);
						bInvalidSearchCriteria = TRUE;
					}
					else if ((bPerformNSFSearch) && (!bInvalidSearchCriteria))
					{
						if (szNSFSearch[0] == '\0')
						{
							//  Check length
							if ((strlen(nsieTable[iIndex].szSearchString) + strlen(FORMULA_START)) > MAX_SEARCH_FORMULA_LENGTH-1)
							{
								//  NSFSearch Formula too long!
								sprintf_s(szMsg, MAX_MSG, MSG_DXR1809E, MAX_SEARCH_FORMULA_LENGTH-1);
								xeLocal->LogMessage(szMsg, iThreadID);
								bInvalidSearchCriteria = TRUE;
							}
							else
							{
								//  Create beginning of select search string
								sprintf_s(szNSFSearch, MAX_SEARCH_FORMULA_LENGTH, FORMULA_START, nsieTable[iIndex].szSearchString);
								wSearchStringsUsed++;
							}
						}
						else
						{
							//  Check length
							if ((strlen(nsieTable[iIndex].szSearchString) + strlen(szNSFSearch) + strlen(FORMULA_CONCAT) + strlen(FORMULA_END)) > MAX_SEARCH_FORMULA_LENGTH-1)
							{
								//  NSFSearch Formula too long!
								sprintf_s(szMsg, MAX_MSG, MSG_DXR1809E, MAX_SEARCH_FORMULA_LENGTH-1);
								xeLocal->LogMessage(szMsg, iThreadID);
								bInvalidSearchCriteria = TRUE;
							}
							else
							{
								//  Concat search string onto NSFSearch search string
								strcat_s(szNSFSearch, MAX_SEARCH_FORMULA_LENGTH, FORMULA_CONCAT);
								strcat_s(szNSFSearch, MAX_SEARCH_FORMULA_LENGTH, nsieTable[iIndex].szSearchString);
								strcat_s(szNSFSearch, MAX_SEARCH_FORMULA_LENGTH, FORMULA_END);
								wSearchStringsUsed++;
							}
						}
					}
				}
				iIndex++;
			}
		} 

		//  Check to see if any errors occured during search string processing
		if (bInvalidSearchCriteria)
		{
			return FALSE;
		}

		//  If no errors occured so far, we need to finish building the internal formula
		//  data structure
		for (iIndex = 0; iIndex < MAX_SEARCH_ENTRIES; iIndex++)
		{
			if (nsieTable[iIndex].hfSearch != NULLHANDLE)
			{
				nsieTable[iIndex].phfSearch = (char *) OSLockObject(nsieTable[iIndex].hfSearch);
				if (nsieTable[iIndex].phfSearch == NULL)
				{
					//  Oh I don't like this... lets get out of here quickly!
					return FALSE;
				}

				stAPIRC = NSFComputeStart(0, nsieTable[iIndex].phfSearch, &nsieTable[iIndex].hcSearch);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR1848E, iIndex);
					xeLocal->LogMessage(szMsg, iThreadID);
					return FALSE;
				}
			}
		}

		sprintf_s(szMsg, MAX_MSG, MSG_DXR1853T, wSearchStringsUsed, szNSFSearch);
		xeLocal->LogTrace(TRACE_DBSCAN, szMsg, iThreadID);

	}

	if (!bPerformNSFSearch)
	{
		//  At least one of the search entries wants to scan all notes, therefore do a simple
		//  NSFDbGetModifiedNoteTable
		xeLocal->LogTrace(TRACE_DBSCAN, MSG_DXR1821T, iThreadID);

		//  Get the Modified Note table for all requested notes
		TimeConstant(TIMEDATE_WILDCARD, &tdStart);
		stAPIRC = NSFDbGetModifiedNoteTable(nsrRQ->hdbToScan, wGlobalNoteClasses, tdStart, &tdFinish, &hitNotesToScan);
		if (ERR(stAPIRC) != NOERROR)
		{
			if (ERR(stAPIRC) == ERR_NO_MODIFIED_NOTES)
			{
				xeLocal->LogMessage(MSG_DXR1817I, iThreadID);
				return TRUE;
			}
			else
			{
				xeLocal->LogMessage(MSG_DXR1818E, iThreadID);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg, iThreadID);
				return FALSE;
			}
		}
	}
	else
	{
		//  All search entries contained a SearchString, therefore perform NSFDbSearch
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1822T, szNSFSearch);
		xeLocal->LogTrace(TRACE_DBSCAN, szMsg, iThreadID);

		//  We need to create the base IDTable
		stAPIRC = IDCreateTable(sizeof(NOTEID), &hitNotesToScan);
		if (ERR(stAPIRC) != NOERROR)
		{
			xeLocal->LogMessage(MSG_DXR1854E, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		//  We need to compile the formula and place it into the internal table
		stAPIRC = NSFFormulaCompile(NULL,
									0, 
									szNSFSearch,
									(WORD) strlen(szNSFSearch),
									&hfNSFSearch,
									&wFormulaLen,
									&stCompile,
									&wErrorLine,
									&wErrorColumn,
									&wErrorOffset,
									&wErrorLength);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1820E, szNSFSearch);
			xeLocal->LogMessage(szMsg, iThreadID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1819E, wErrorLine, wErrorColumn, wErrorOffset, wErrorLength);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		stAPIRC = NSFSearch(nsrRQ->hdbToScan,
							hfNSFSearch,
							NULL,
							0,
							wGlobalNoteClasses,
							NULL,
							WrapperProcessSearchResult,
							this,
							NULL);
		if (ERR(stAPIRC) != NOERROR)
		{
			xeLocal->LogMessage(MSG_DXR1823E, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			OSMemFree(hfNSFSearch);
			return FALSE;
		}

		OSMemFree(hfNSFSearch);
	}	
	return TRUE;
}

//  WrapperProcessSearchResult
//
//  This functions provides the LNCALLBACK routine used by NSFSearch
//
//  Parameters
//
//  1  -  DXNoteScanReq *	-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  STATUS	-  NOERROR if actions were successful
//			-  ERR_ if actions failed
//
//  API Use Notes:
//
//
STATUS LNCALLBACK DXNoteScanner::WrapperProcessSearchResult(void *ObjPtr, SEARCH_MATCH *smCurrent, ITEM_TABLE *itCurrent)
{
	DXNoteScanner			*dxnsLocal;					//  Pointer to NoteScanner object
	dxnsLocal = (DXNoteScanner *) ObjPtr;
	return dxnsLocal->ProcessSearchResult(smCurrent, itCurrent, dxnsLocal);
}

//  ProcessSearchResult
//
//  This function is used by the NSFSearch CALLBACK routine
//
//  Parameters
//
//  1  -  DXNoteScanReq *	-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  STATUS	-  NOERROR if actions were successful
//			-  ERR_ if actions failed
//
//  API Use Notes:
//
//
STATUS DXNoteScanner::ProcessSearchResult(SEARCH_MATCH *smHit, ITEM_TABLE *itCurrent, DXNoteScanner *dxnsCurrent)
{
	STATUS					stAPIRC;					//  Return code from Notes API calls
	DXNoteScanner			*dxnsLocal;					//  Pointer to NoteScanner object
	SEARCH_MATCH			smCurrent;					//  Local copy of search match structure
	BOOL					bInserted;					//  Inserted flag

	dxnsLocal = dxnsCurrent;

	//  Copy the passed structures to local storage in case they are not word aligned
	memcpy((char *) &smCurrent, (char *) smHit, sizeof(SEARCH_MATCH));

	//  Check that the current item was matched
	if (!smCurrent.SERetFlags & SE_FMATCH) return(NOERROR);

	//  Add the found NOTEID to the results table and increment the count
	stAPIRC = IDInsert(dxnsLocal->hitNotesToScan, smCurrent.ID.NoteID, &bInserted);
	if ((ERR(stAPIRC) == NOERROR) && (bInserted)) dxnsLocal->dwNotesInserted++;
	return stAPIRC;
}

//  CleanupIntTable
//
//  This functions cleanups up the internal DXNoteScanIntEntry array
//
//  Parameters
//  1  -  int				-  Identity of the thread that is making the call
//
//  Returns
//  BOOL	-  TRUE if actions were successful
//			-  FALSE if actions failed
//
//  API Use Notes:
//
//
BOOL DXNoteScanner::CleanupIntTable(int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API Calls
	BOOL				bCleanupOK;								//  Indicate if cleanup was ok or not
	int					iIndex;									//  Index counter
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	bCleanupOK = TRUE;

	//  Initialize internal formula array
	iIndex = 0;
	for (iIndex = 0; iIndex < MAX_SEARCH_ENTRIES; iIndex ++)
	{
		if (nsieTable[iIndex].hcSearch != NULL)
		{
			stAPIRC = NSFComputeStop(nsieTable[iIndex].hcSearch);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR1816E, iIndex);
				xeLocal->LogMessage(szMsg, iThreadID);
				bCleanupOK = FALSE;
			}
		}

		if (nsieTable[iIndex].phfSearch != NULL)
		{
			OSUnlockObject(nsieTable[iIndex].hfSearch);
		}

		if (nsieTable[iIndex].hfSearch != NULLHANDLE)
		{
			OSMemFree(nsieTable[iIndex].hfSearch);
		}
	}
	return bCleanupOK;
}

//  ProcessNotes
//
//  This function takes the global IDTable and scans each note to see which
//  search entries it matches
//
//  Parameters
//
//  1  -  DXNoteScanReq *	-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL	-  TRUE if actions were successful
//			-  FALSE if actions failed
//
//  API Use Notes:
//
//
BOOL DXNoteScanner::ProcessNotes(DXNoteScanReq *nsrCurrent, int iThreadID)
{
	BOOL				bFirst;									//  Indicate first entry in IDScan
	BOOL				bContinue;								//  Whether IDScan should continue
	DWORD				dwRequestsPosted;						//  Number of part scan reqs posted
	DWORD				dwRequestsProcessed;					//  Number of part scan reqs posted
	NOTEID				nidRetNID;								//  Note ID being processed
	DXNoteScanReq		*nsrRQ;									//  Pointer to DXNoteScanReq object
	DXPartNoteScanReq	*pnsrRQ;								//  Pointer to DXPartNoteScanReq object
	long				lModNum;								//  Progress indicator
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	nsrRQ = nsrCurrent;

	//  Calculate number to be used by output mod computation
	if (nsrRQ->dwNotesToScan <= 100) lModNum = 1;
	else
	{
		if (nsrRQ->dwNotesToScan <= 1000) lModNum = 10;
		else
		{
			if (nsrRQ->dwNotesToScan <= 5000) lModNum = 50;
			else 
			{
				if (nsrRQ->dwNotesToScan <= 10000) lModNum = 100;
				else lModNum = 200;
			}
		}
	}

	dwRequestsPosted = 0;
	dwRequestsProcessed = 0;
	bFirst = TRUE;
	bContinue = TRUE;
	pnsrRQ = NULL;

	while ((IDScan(hitNotesToScan, bFirst, &nidRetNID)) && (bContinue))
	{
		bFirst = FALSE;

		//  If the scan request does not exist then create it
		if (pnsrRQ == NULL) 
		{
			pnsrRQ = new DXPartNoteScanReq(nsrRQ);
			pnsrRQ->iScanAction = REQ_SCAN_DATA_NOTES;
		}

		// Add the current noteid to the copy request
		pnsrRQ->nidToScan[pnsrRQ->iNotesToScan] = nidRetNID;
		pnsrRQ->iNotesToScan++;

		//  Check if the request is full - if so dispatch it
		if (pnsrRQ->iNotesToScan == SCAN_NOTE_MULTIPLEX)
		{
			PerformScanAction(pnsrRQ, iThreadID);
			dwRequestsPosted++;

			AccumulateTotals(nsrRQ, pnsrRQ, iThreadID);
			dwRequestsProcessed++;

			//  Clean up the request if it is valid
			if (pnsrRQ != NULL)
			{
				//  Update the action in memory in case a race causes the freed object to be returned to another thread
				pnsrRQ->iScanAction = REQ_SCAN_FREED;
				delete pnsrRQ;
				pnsrRQ = NULL;	
			}

			if ( dwRequestsProcessed % lModNum == 0 )
			{
				sprintf_s(szMsg, MAX_MSG,
						  MSG_DXR1826I, 
						  dwRequestsProcessed,
						  dwRequestsPosted,
						  nsrRQ->dwNotesScanned,
						  nsrRQ->dwNotesToScan);
				xeLocal->LogVerbose(szMsg, iThreadID);
			}
		}

		if ( (nsrRQ->dwMaxErrors > 0) && (nsrRQ->dwErrors >= nsrRQ->dwMaxErrors) )
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1845E, nsrRQ->dwErrors, nsrRQ->dwMaxErrors);
			xeLocal->LogMessage(szMsg, iThreadID);
			bContinue = FALSE;
		}

	}

	//  Check to see if there is an outstanding DXPartNoteScanReq object
	if (pnsrRQ != NULL)
	{
		if (bContinue)
		{
			PerformScanAction(pnsrRQ, iThreadID);
			dwRequestsPosted++;

			AccumulateTotals(nsrRQ, pnsrRQ, iThreadID);
			dwRequestsProcessed++;

			//  Clean up the request if it is valid
			if (pnsrRQ != NULL)
			{
				//  Update the action in memory in case a race causes the freed object to be returned to another thread
				pnsrRQ->iScanAction = REQ_SCAN_FREED;
				delete pnsrRQ;
				pnsrRQ = NULL;	
			}

			sprintf_s(szMsg, MAX_MSG,
					  MSG_DXR1826I,
					  dwRequestsProcessed,
					  dwRequestsPosted,
					  nsrRQ->dwNotesScanned,
					  nsrRQ->dwNotesToScan);
			xeLocal->LogVerbose(szMsg, iThreadID);
			dwRequestsPosted++;
		}
		else
		{
			//  Cleanup request object
			delete pnsrRQ;
			pnsrRQ = NULL;
		}
	}

	if (!bContinue) return FALSE;
	else return TRUE;

}

//  PerformScanAction
//
//  This is the main routing function for all multi-threaded invocations, the function will vector the current
//  request to the appropriate executing functions determined by the Scan Action defined in the request.
//
//  Parameters
//
//  1  -  DbScanReq::PartDbScanReq *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int							-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DXNoteScanner::PerformScanAction(DXPartNoteScanReq *ppnsCurrent, int iThreadID)
{
	DXPartNoteScanReq	*ppnsRQ;							//  Pointer to DXNoteScanReq object
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	if (ppnsCurrent == NULL) return;
	ppnsRQ = ppnsCurrent;

	sprintf_s(szMsg, MAX_MSG, MSG_DXR1829T, ppnsRQ->iScanAction);
	xeLocal->LogTrace(TRACE_DBSCAN, szMsg, iThreadID);

	//  Vector according to the action requested
	switch(ppnsRQ->iScanAction)
	{
	case REQ_SCAN_DATA_NOTES:
		PerformScanNotes(ppnsRQ, iThreadID);
		break;
	}
	return;
}

//  PerformScanNotes
//
//  This is the generic function that scan notes in the source databese.
//  The scanning is performed in chunks - the maximum size of the chunks is determined at compile time.
//
//  Parameters
//
//  1  -  DXPartNoteScanReq *	-  Pointer to the DXPartNoteScanReq object for this scheduling
//  2  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DXNoteScanner::PerformScanNotes(DXPartNoteScanReq *ppnsCurrent, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	BLOCKID				bidObjectItem;						//  BlockID used by NSFNoteHasObjects
	BOOL				bNoteSigned;						//  Whether note is signed
	BOOL				bNoteEncrypted;						//  Whether note is encrypted
	BOOL				bHasFileAttach;						//  Whether note has a file attachment or not
	BOOL				bHasOtherObject;					//  Whether note has any other type of object
	BOOL				*pbHasFileAttach;					//  Pointer to whether note has a file attachment or not
	BOOL				*pbHasOtherObject;					//  Pointer to whether note has any other type of object
	DBHANDLE			hdbTLSource;						//  Local Thread handle for the Source database
	DWORD				dwOpenFlags;						//  NSFNoteOpenExt flags
	OID					oidToScan;							//  OID of note being processed
	TIMEDATE			tdMod;								//  Modification date of note being processed
	DXPartNoteScanReq	*ppnsRQ;							//  Pointer to DXPartNoteScanReq object
	DXNoteScanNoteProps nsnpToScan;							//  DXNoteScanNoteProps data structure
	int					iIndex;								//  Index counter for NoteID array
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	if (ppnsCurrent == NULL) return;
	ppnsRQ = ppnsCurrent;

	//  Get the mapped thread local database handles
	hdbTLSource = ppnsRQ->nsrParent->hdbToScan;
	if (hdbTLSource == NULLHANDLE)
	{
		ppnsRQ->iErrors += ppnsRQ->iNotesToScan;
		xeLocal->LogMessage(MSG_DXR1830E, iThreadID);
		return;
	}

	dwOpenFlags = (OPEN_EXPAND | OPEN_RAW_MIME);
	ppnsRQ->iNotesScanned = 0;
	//  Scan each of the requested notes
	for (iIndex = 0; iIndex < ppnsRQ->iNotesToScan; iIndex++)
	{
		//  Reset data structure values
		nsnpToScan.iNIDArrayIndex = iIndex;
		nsnpToScan.nidRetNID = 0;
		nsnpToScan.hnToScan = NULLHANDLE;
		nsnpToScan.wNoteclass = 0;
		nsnpToScan.dwNoteFlags = 0;

		//  Get NoteID to be scanned
		nsnpToScan.nidRetNID = ppnsRQ->nidToScan[iIndex];

		//  If note is a deletion stub... remove deletion stub bit
		if (nsnpToScan.nidRetNID & RRV_DELETED)
		{
			nsnpToScan.nidRetNID = nsnpToScan.nidRetNID ^ RRV_DELETED;
			nsnpToScan.dwNoteFlags = (nsnpToScan.dwNoteFlags | FLAG_NOTE_SCAN_DELSTUBS);
		}
		else
		{
			nsnpToScan.dwNoteFlags = (nsnpToScan.dwNoteFlags | FLAG_NOTE_SCAN_NOT_DELSTUB);
		}

		ppnsRQ->iNotesScanned++;

		stAPIRC = NSFDbGetNoteInfo(hdbTLSource, nsnpToScan.nidRetNID, &oidToScan, &tdMod, &nsnpToScan.wNoteclass);

		//  First check to see if we are searching for this type of note class
		//  Now continue with checking return code
		if (ERR(stAPIRC) != NOERROR)
		{
			if (ERR(stAPIRC) == ERR_NOTE_DELETED)
			{
				nsnpToScan.dwNoteFlags = (nsnpToScan.dwNoteFlags | FLAG_NOTE_SCAN_DELSTUBS);

				//  Now check to see if note findings match any of the search requests
				CheckSearchEntriesForMatch2(ppnsRQ, &nsnpToScan, iThreadID);
			}
			else
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR1840E, nsnpToScan.nidRetNID);
				xeLocal->LogMessage(szMsg, iThreadID);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg, iThreadID);
				ppnsRQ->iErrors++;
			}
		}
		else
		{
			if (nsnpToScan.dwNoteFlags & FLAG_NOTE_SCAN_DELSTUBS)
			{
				//  Check to see if note findings match any of the search requests
				CheckSearchEntriesForMatch2(ppnsRQ, &nsnpToScan, iThreadID);
			}
			else
			{
				//  As we are not dealing with a deletion stub, lets open it and see which
				//  search entries want to know about this note

				// Open the data note
				stAPIRC = NSFNoteOpenExt(hdbTLSource, nsnpToScan.nidRetNID, dwOpenFlags, &nsnpToScan.hnToScan);
				if (ERR(stAPIRC) != NOERROR)
				{
					if (ERR(stAPIRC) == ERR_NOACCESS)
					{
						if (dwGlobalScanFlags & FLAG_NOTE_SCAN_NOACCESS)
						{
							nsnpToScan.dwNoteFlags = (nsnpToScan.dwNoteFlags | FLAG_NOTE_SCAN_NOACCESS);

							//  Now check to see if note findings match any of the search requests
							CheckSearchEntriesForMatch2(ppnsRQ, &nsnpToScan, iThreadID);
						}

					}
					else
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR1832E, nsnpToScan.nidRetNID);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
						xeLocal->LogMessage(szMsg, iThreadID);
						ppnsRQ->iErrors++;
					}
				}
				else
				{

					//  See if any search entry wants to know about signed and/or encrypted notes
					if ( (dwGlobalScanFlags & FLAG_NOTE_SCAN_SIGNED) | (dwGlobalScanFlags & FLAG_NOTE_SCAN_ENCRYPTED) )
					{
						if ( NSFNoteIsSignedOrSealed(nsnpToScan.hnToScan, &bNoteSigned, &bNoteEncrypted) )
						{
							
							if ( (dwGlobalScanFlags & FLAG_NOTE_SCAN_SIGNED) && (bNoteSigned) )
							{
								nsnpToScan.dwNoteFlags = (nsnpToScan.dwNoteFlags | FLAG_NOTE_SCAN_SIGNED);
							}

							if ( (dwGlobalScanFlags & FLAG_NOTE_SCAN_ENCRYPTED) && (bNoteEncrypted) )
							{
								//  Additional check if requested by caller
								if ( NSFItemIsPresent(nsnpToScan.hnToScan, ITEM_NAME_NOTE_SEALDATA, (WORD) strlen(ITEM_NAME_NOTE_SEALDATA)) )
								{
									nsnpToScan.dwNoteFlags = (nsnpToScan.dwNoteFlags | FLAG_NOTE_SCAN_ENCRYPTED);
								}
							}
						}
					}

					//  See if any search entry wants to know if note has attachments
					if ((dwGlobalScanFlags & FLAG_NOTE_SCAN_OBJECT_FILE) | (dwGlobalScanFlags & FLAG_NOTE_SCAN_OBJECT_OTHER))
					{
						//  Check to see if note has an object
						if (NSFNoteHasObjects(nsnpToScan.hnToScan, &bidObjectItem))
						{
							//  Check what we want to scan for
							bHasFileAttach = FALSE;
							bHasOtherObject = FALSE;
							pbHasFileAttach = NULL;
							pbHasOtherObject = NULL;
							if (dwGlobalScanFlags & FLAG_NOTE_SCAN_OBJECT_FILE) pbHasFileAttach = &bHasFileAttach;
							if (dwGlobalScanFlags & FLAG_NOTE_SCAN_OBJECT_OTHER) pbHasOtherObject = &bHasOtherObject;

							// Now check if this note has a OBJECT_FILE
							if (ScanForObjectItems(ppnsRQ, hdbTLSource, nsnpToScan.hnToScan, pbHasFileAttach, pbHasOtherObject, iThreadID) )
							{

								if ( (dwGlobalScanFlags & FLAG_NOTE_SCAN_OBJECT_FILE) && (bHasFileAttach) )
								{
									nsnpToScan.dwNoteFlags = (nsnpToScan.dwNoteFlags | FLAG_NOTE_SCAN_OBJECT_FILE);
								}

								if ( (dwGlobalScanFlags & FLAG_NOTE_SCAN_OBJECT_OTHER) && (bHasOtherObject) )
								{
									nsnpToScan.dwNoteFlags = (nsnpToScan.dwNoteFlags | FLAG_NOTE_SCAN_OBJECT_OTHER);
								}
							}
						}
					}

					//  Check to see if anyone is searching for RepCons
					if (dwGlobalScanFlags & FLAG_NOTE_SCAN_REPCONS)
					{
						// See if the ITEM_NAME_CONFLICT item is present on the note
						if (NSFItemIsPresent(nsnpToScan.hnToScan, ITEM_NAME_CONFLICT, (WORD)strlen(ITEM_NAME_CONFLICT)))
						{
							nsnpToScan.dwNoteFlags = (nsnpToScan.dwNoteFlags | FLAG_NOTE_SCAN_REPCONS);
						}
					}

					//  Now check to see if note findings match any of the search requests
					CheckSearchEntriesForMatch2(ppnsRQ, &nsnpToScan, iThreadID);

					//  Close the note now we are finished
					stAPIRC = NSFNoteClose(nsnpToScan.hnToScan);
					if (ERR(stAPIRC) != NOERROR)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR1833E, nsnpToScan.nidRetNID);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
						xeLocal->LogMessage(szMsg, iThreadID);
						ppnsRQ->iErrors++;
					}
				}
			}
		}
	}
}

//  CheckSearchEntriesForMatch2
//
//  This scans each search entry's flag settings to see if any contain the supplied
//  flag value.  If so, insert
//
//  Parameters
//
//  1  -  DXPartNoteScanReq *	-  Pointer to the DXPartNoteScanReq object for this scheduling
//  2  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
BOOL DXNoteScanner::CheckSearchEntriesForMatch2(DXPartNoteScanReq *ppnsCurrent, DXNoteScanNoteProps *nsnpCurrent, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes c api calls
	BOOL				bMatch;								//  Match result flag
	DWORD				*dwSearchEntryFlags;				//  Pointer to search entry flags value
	HCOMPUTE			hcFormula;							//  Handle to compiled formula to use
	DXPartNoteScanReq	*ppnsRQ;							//  Pointer to DXPartNoteScanReq object
	DXNoteScanNoteProps *pnsnpToScan;						//  Pointer to DXNoteScanNoteProps data structure
	int					iIndex;								//  Index counter for search entries iteration
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	if (ppnsCurrent == NULL) return FALSE;
	if (nsnpCurrent == NULL) return FALSE;
	ppnsRQ = ppnsCurrent;
	pnsnpToScan = nsnpCurrent;

	hcFormula = NULLHANDLE;

	for (iIndex = 0; iIndex < MAX_SEARCH_ENTRIES; iIndex++)
	{
		if (ppnsRQ->nsrParent->nsObject->dxnsEntry[iIndex].hitMatches == NULLHANDLE)
		{
			//  We have found the first entry in the search table with a NULL IDTable... so exit
			return TRUE;
		}

		//  First we need to make sure we are searching for this type of note class
		if (pnsnpToScan->wNoteclass & ppnsRQ->nsrParent->nsObject->dxnsEntry[iIndex].wNoteClass)
		{
			dwSearchEntryFlags = &ppnsRQ->nsrParent->nsObject->dxnsEntry[iIndex].dwSearchFlags;

			//  Now we need to make sure any other scan flags specified match as well
			if (((pnsnpToScan->dwNoteFlags & *dwSearchEntryFlags ) == *dwSearchEntryFlags) || (*dwSearchEntryFlags == 0))
			{
				//  Check to see if we have a search formula
				if (nsieTable[iIndex].phfSearch != NULL)
				{
					bMatch = FALSE;
					//  Double check we are dealing with a deletion stub because the scan entry has
					//  simply specified all data notes and a search string
					if (!(pnsnpToScan->dwNoteFlags & FLAG_NOTE_SCAN_DELSTUBS))
					{
						//  Get thread level HCOMPUTE handle
						stAPIRC = NSFComputeStart(0, nsieTable[iIndex].phfSearch, &hcFormula);
						if (ERR(stAPIRC) != NOERROR)
						{
							sprintf_s(szMsg, MAX_MSG, MSG_DXR1855E, iIndex);
							xeLocal->LogMessage(szMsg, iThreadID);
							ppnsRQ->iErrors++;
							return FALSE;
						}
						else
						{
							//  Evaluate the formula to see if we have a match
							stAPIRC = NSFComputeEvaluate(hcFormula,
														pnsnpToScan->hnToScan,
														0,
														0,
														&bMatch,
														0, 
														0);
							if (ERR(stAPIRC) != NOERROR)
							{
								sprintf_s(szMsg, MAX_MSG, MSG_DXR1846E, iIndex, pnsnpToScan->nidRetNID);
								xeLocal->LogMessage(szMsg, iThreadID);
								ppnsRQ->iErrors++;
								bMatch = FALSE;
							}

							//  Remove thread level HCOMPUTE handle
							stAPIRC = NSFComputeStop(hcFormula);
							if (ERR(stAPIRC) != NOERROR)
							{
								sprintf_s(szMsg, MAX_MSG, MSG_DXR1856E, iIndex);
								xeLocal->LogMessage(szMsg, iThreadID);
								ppnsRQ->iErrors++;
								return FALSE;
							}
							hcFormula = NULLHANDLE;
						}
					}
				}
				else bMatch = TRUE;

				if (bMatch)
				{
					//  Add noteid to applicable entry in results array
					ppnsRQ->nidResults[iIndex][pnsnpToScan->iNIDArrayIndex] = pnsnpToScan->nidRetNID;
					ppnsRQ->iAdded[iIndex]++;

				}
			}
		}
	}
	return TRUE;
}

//  ScanForObjectItems
//
//  This scans a note for all items of TYPE_OBJECT and checks to see whether they are
//  a File Attachment or Other type of object.  
//
//  Parameters
//
//  1  -  DXPartNoteScanReq *	-  Pointer to the DXPartNoteScanReq object for this scheduling
//  2  -  DBHANDLE				-  Handle to database containing object
//  3  -  NOTEHANDLE			-  Handle of open note to be scanned for objects
//  4  -  BOOL *				-  Pointer to BOOL indicating if the scan found a TYPE_FILE object
//  5  -  BOOL *				-  Pointer to BOOL indicating if the scan found an Other object type
//  6  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL	-  TRUE if actions were successful
//			-  FALSE if actions failed
//
//  API Use Notes:
//
//
BOOL DXNoteScanner::ScanForObjectItems(DXPartNoteScanReq *ppnsCurrent, DBHANDLE hdbTLSource, NOTEHANDLE hnToScan, BOOL *bHasFileAttach, BOOL *bHasOtherObject, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code of Notes C API calls
	BLOCK				bFieldToProcess;					//  Indicates there is another item on the note to process
	BLOCKID				bidItemToProcess;					//  BLOCKID of current item's header
	BLOCKID				bidItemValueToProcess;				//  BLOCKID of current item's value
	BLOCKID				bidPrevItem;						//  BLOCKID of prev item used by NSFItemInfoNext
	DWORD				dwValueLen;							//  Length of item being processed
	DWORD				dwItemCount;						//  Item counter
	NOTEID				nidToScan;							//  NoteID of note being processed
	WORD				wDataType;							//  Data type of item being processed
	WORD				wObjType;							//  ObjectType of item object
	DXPartNoteScanReq	*ppnsRQ;							//  Pointer to DXPartNoteScanReq object
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	if (ppnsCurrent == NULL ) return FALSE;
	ppnsRQ = ppnsCurrent;

	if (hdbTLSource == NULLHANDLE) 
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1834E, "DBHANDLE");
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}
	if (hnToScan == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1834E, "NOTEHANDLE");
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}
	if ( (bHasFileAttach == NULL) && (bHasOtherObject == NULL) )
	{
		xeLocal->LogMessage(MSG_DXR1839E, iThreadID);
		return FALSE;
	}

	//  Get NoteID of note being processed
	NSFNoteGetInfo(hnToScan, _NOTE_ID, &nidToScan);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1835T, nidToScan);
	xeLocal->LogTrace(TRACE_DBSCAN, szMsg, iThreadID);

	//  Reset return values to FALSE (if required)
	if (bHasFileAttach != NULL) *bHasFileAttach = FALSE;
	if (bHasOtherObject != NULL) *bHasOtherObject = FALSE;

	//  Get first item on note
	dwItemCount = 0;
	stAPIRC = NSFItemInfo(hnToScan, NULL, 0, &bidItemToProcess, &wDataType, &bidItemValueToProcess, &dwValueLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1836E, dwItemCount, nidToScan);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	bFieldToProcess = TRUE;
	while (bFieldToProcess)
	{
		bidPrevItem = bidItemToProcess;

		//  Check to see if we are interested in this item
		if (wDataType == TYPE_OBJECT)
		{
			//  Now we need to see if this is a File Attachment or not
			if (GetObjectInfo(bidItemValueToProcess, NULL, &wObjType, NULL, 0, NULL, 0, NULL, NULL, NULL, iThreadID))
			{
				if (wObjType == OBJECT_FILE)
				{
					if (bHasFileAttach != NULL) *bHasFileAttach = TRUE;
				}
				else
				{
					if (bHasOtherObject != NULL) *bHasFileAttach = TRUE;
				}
			}
			else
			{
				return FALSE;
			}
		}

		//  Quick check to see if we can exit out early from this process!
		if ((bHasFileAttach != NULL) && (bHasOtherObject == NULL))
		{
			if (*bHasFileAttach) return TRUE;
		}
		else if ((bHasFileAttach == NULL) && (bHasOtherObject != NULL))
		{
			if (*bHasOtherObject) return TRUE;
		}
		else
		{
			if ((*bHasFileAttach) && (*bHasOtherObject)) return TRUE;
		}
		
		// See if there is another item on the note
		dwItemCount++;
		stAPIRC = NSFItemInfoNext(hnToScan, bidPrevItem, NULL, 0, &bidItemToProcess, &wDataType, &bidItemValueToProcess, &dwValueLen);
		if (ERR(stAPIRC) != NOERROR)
		{
			if (ERR(stAPIRC) == ERR_ITEM_NOT_FOUND) bFieldToProcess = FALSE;
			else
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR1836E, dwItemCount, nidToScan);
				xeLocal->LogMessage(szMsg, iThreadID);
				xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
				xeLocal->LogMessage(szMsg, iThreadID);
				return FALSE;
			}
		}
	}
	return TRUE;
}

//  AccumulateTotals
//
//  This takes the totals from the DXPartNoteScanReq and updates the totals in the
//  DXNoteScanReq
//
//  Parameters
//
//  1  -  DXNoteScanReq *		-  Pointer to the DXNoteScanReq object for this scheduling
//  2  -  DXPartNoteScanReq *	-  Pointer to the DXPartNoteScanReq object for this scheduling
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DXNoteScanner::AccumulateTotals(DXNoteScanReq *nsrCurrent, DXPartNoteScanReq *ppnsCurrent, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes c api calls
	DXNoteScanReq		*nsrRQ;								//  Pointer to DXNoteScanReq object
	DXPartNoteScanReq	*ppnsRQ;							//  Pointer to DXPartNoteScanReq object
	int					iIndex;								//  Index counter
	int					iIndexY;
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	if (nsrCurrent == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1841E, "DXNoteScanReq");
		xeLocal->LogMessage(szMsg, iThreadID);
	}
	if (ppnsCurrent == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1841E, "DXPartNoteScanReq");
		xeLocal->LogMessage(szMsg, iThreadID);
	}
	nsrRQ = nsrCurrent;
	ppnsRQ = ppnsCurrent;

	for (iIndex = 0; iIndex < MAX_SEARCH_ENTRIES; iIndex++)
	{
		for (iIndexY = 0; iIndexY < SCAN_NOTE_MULTIPLEX; iIndexY++)
		{
			if (ppnsRQ->nidResults[iIndex][iIndexY] != 0)
			{
				stAPIRC = IDInsert(nsrRQ->nsObject->dxnsEntry[iIndex].hitMatches, ppnsRQ->nidResults[iIndex][iIndexY], NULL);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR1831E, ppnsRQ->nidResults[iIndex][iIndexY], iIndex);
					xeLocal->LogMessage(szMsg, iThreadID);
					nsrRQ->dwErrors++;
				}
				else
				{
					nsrRQ->nsObject->dxnsEntry[iIndex].dwCount++;
				}
			}
		}
	}

	nsrRQ->dwNotesScanned += ppnsRQ->iNotesScanned;
	nsrRQ->dwErrors += (DWORD) ppnsRQ->iErrors;
	return;
}

//  GetObjectInfo
//
//  Using a specified BLOCKID, get the FILEOBJECT information
//
//  Parameters
//
//  1  -  BLOCKID			-  BLOCKID of object item
//  2  -  char *			-  Pointer to FileName buffer
//  3  -  int				-  Length of FileName buffer
//  4  -  char *			-  Pointer to File extension buffer
//  5  -  int				-  Length of File extension buffer
//  6  -  DWORD *			-  Pointer to Filesize DWORD value
//  7  -  WORD *			-  Pointer to HostType WORD value
//  8  -  WORD *			-  Pointer to CompType WORD value
//  9  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL					-  TRUE if the log action succeeded
//							-  FALSE if the log action failed
//
//  API Use Notes:
//
//
BOOL DXNoteScanner::GetObjectInfo(BLOCKID bidItemValue, DWORD *dwObjID, WORD *wObjType, char *szFileName, int iFileNameLen, char *szExt, int iExtLen, DWORD *dwFileSize, WORD *wHostType, WORD *wCompType, int iThreadID)
{
	FILEOBJECT				objFile;							//  Structure used during processing
	BOOL					bContinue;							//  Continue flag while reading filename right to left
	BOOL					bExtFound;							//  Indicates that extension was found
	int						iExtIndex;							//  Used during file extension processing
	int						iSourceFileNameLen;					//  Used during file extension processing
	int						iSourceFileNameIndex;				//  Index used while looking for file extension
	char					szSourceFileName[LARGE_MAXPATH];	//  Temp buffer for filename/ext handling
	char					*pObject;							//  Pointer to locked object
	char					*pszFileObject;						//  Pointer for processing FILEOBJECT
	char					szChar;								//  Used when testing char
	char					szMsg[MAX_MSG + 1];					//  Generic Message Buffer

	//  Check the supplied BlockID
	if (ISNULLBLOCKID(bidItemValue))
	{
		xeLocal->LogMessage(MSG_DXR1837E, iThreadID);
		return FALSE;
	}

	//  Get pointer to start of block
	pObject = NULL;
	pObject = OSLockBlock(char, bidItemValue);
	if (pObject == NULL)
	{
		xeLocal->LogMessage(MSG_DXR1838E, iThreadID);
		return FALSE;
	}
	pszFileObject = pObject;

	//  Step over type word
	pszFileObject += sizeof(WORD);

	//  Read FILEOBJECT data
	ODSReadMemory( &pszFileObject, _FILEOBJECT, &objFile, 1 );

	//  Get the object ID (if required)
	if (dwObjID != NULL) *dwObjID = objFile.Header.RRV;

	//  Get the object Type (if required)
	if (wObjType != NULL) *wObjType = objFile.Header.ObjectType;

	//  Get the file size (if required)
	if (dwFileSize != NULL) *dwFileSize = objFile.FileSize;

	//  Get the HostType (if required)
	if (wHostType != NULL) *wHostType = objFile.HostType;

	//  Get the Compression Type ( if required )
	if (wCompType != NULL) *wCompType = objFile.CompressionType;

	//  Get filename and/or file extension
	if ((szFileName != NULL) || (szExt != NULL))
	{
		szSourceFileName[0] = '\0';

		//  Copy filename into buffer
		if (szSourceFileName[0] == '\0')
		{
			if (objFile.FileNameLength < iFileNameLen) 
			{
				memcpy( szSourceFileName, pszFileObject, objFile.FileNameLength );
				szSourceFileName[objFile.FileNameLength] = '\0';
			}
			else
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR1839W, objFile.FileNameLength);
				xeLocal->LogMessage(szMsg, iThreadID);
			}
		}

		if (szSourceFileName[0] != '\0')
		{
			iSourceFileNameLen = (int) strlen(szSourceFileName);

			//  Copy FileName into supplied buffer (if required)
			if (szFileName != NULL)
			{
				szFileName[0] = '\0';
				strcpy_s(szFileName, iFileNameLen, szSourceFileName);
			}

			//  Copy File Extension into supplied buffer (if required)
			if (szExt != NULL)
			{
				szExt[0] = '\0';
				iSourceFileNameIndex = iSourceFileNameLen;
				bContinue = TRUE;
				bExtFound = FALSE;

				// Find position of right most period
				while (bContinue)
				{
					if (iSourceFileNameIndex < 0) bContinue = FALSE;
					else
					{
						szChar = szSourceFileName[iSourceFileNameIndex];
						if ( szChar == '.' )
						{
							bContinue = FALSE;
							bExtFound = TRUE;
						}
						else if (szChar == '\\') bContinue = FALSE;
						else if (szChar == '/') bContinue = FALSE;
						else iSourceFileNameIndex--;
					}
				}

				//  If extension found, now copy ext string into buffer
				if (bExtFound)
				{
					iSourceFileNameIndex++;
					iExtIndex = 0;

					if ((iSourceFileNameLen - iSourceFileNameIndex) < (iExtLen-1))
					{
						while (iSourceFileNameIndex < iSourceFileNameLen)
						{
							szExt[iExtIndex] = szSourceFileName[iSourceFileNameIndex];
							iSourceFileNameIndex++;
							iExtIndex++;
						}
						szExt[iExtIndex] = '\0';
					}
				}
			}
		}
	}

	OSUnlockBlock(bidItemValue);
	return TRUE;
}

