//  DXRequest
//
//  This class encapsulates the request information that drives the Domino Explorer (DX)
//
//  Message Identifier:      DXR
//  Message Range:           1901 - 1999
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.11.1
//  Version:                 3.11.1  Build: 99
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2209		Re-based ready for 4.0.0
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.5		18/02/2011		Report on D/Bs which cannot be opened
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2011
//

#include "DXRequest.h"

//  Default Constructor
DXRequest::DXRequest(void)
: AnchorScope(RSCOPE_NONE)
, dxrParent(NULL)
, Wildcards(0)
, RecurseDirectories(TRUE)
, ReturnCode(0)
, RQLevel(0)
, hdbCache(NULLHANDLE)
, ScannersNeeded(0)
, Reporters(0)
, Filters(0)
, dwServersProcessed(0)
, dwDirectoriesProcessed(0)
, dwDatabasesProcessed(0)
, dwServersDropped(0)
, dwDirectoriesDropped(0)
, dwDatabasesDropped(0)
{
	//  Initialise the scope fields
	szDomain[0] = '\0';
	szServer[0] = '\0';
	szDirectory[0] = '\0';
	szDatabase[0] = '\0';
	szMatchDomain[0] = '\0';
	szMatchServer[0] = '\0';
	szMatchDirectory[0] = '\0';
	szMatchDatabase[0] = '\0';

	// Set default start and finish times
	ctStart = 0;
	ctEnd = 0;
}

//  Destructor
DXRequest::~DXRequest(void)
{
}

//  ParseRequest
//
//  This function will parse a two string raw Explorer request into the current request object
//
//  Parameters
//  1  -  ExecEnvironment *		-  Pointer to the current runtime environment
//  2  -  char *				-  First part of the request
//  3  -  char *				-  Second part of the request
//
//  Returns
//
//  API Use Notes:
//
//

BOOL DXRequest::ParseRequest(ExecEnvironment *xeBase, char *szRQ1, char *szRQ2)
{
	return ParseRequest(xeBase, szRQ1, szRQ2, 0);
}

//  ParseRequest
//
//  This function will parse a two string raw Explorer request into the current request object
//
//  Parameters
//
//  1  -  ExecEnvironment *		-  Pointer to the current runtime environment
//  2  -  char *				-  First part of the request
//  3  -  char *				-  Second part of the request
//  4  -  int					-  ID of the thread making the request
//
//  Returns
//
//  API Use Notes:
//
//
BOOL DXRequest::ParseRequest(ExecEnvironment *xeBase, char *szRQ1, char *szRQ2, int iThreadID)
{
	char		*pScope;											//  Pointer into the scope
	char		*pTarget;											//  Target pointer
	char		*pLastDelimiter;									//  Pointer to the last directory delimiter
	char		*pTerminator;										//  Pointer to terminator atom
	char		szMsg[MAX_MSG + 1];									//  Generic Message Buffer

	//  Trace the start of parsing
	xeBase->LogTrace(TRACE_EXPLORER, MSG_DXR1904T, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1902I, szRQ1);
	xeBase->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1903I, szRQ2);
	xeBase->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);

	//  Elimiate the boundary condition where either scope part is null
	if ((szRQ1[0] == '\0') || (szRQ2[0] == '\0'))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1902I, szRQ1);
		xeBase->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1903I, szRQ2);
		xeBase->LogMessage(szMsg, iThreadID);
		xeBase->LogMessage(MSG_DXR1905E, iThreadID);
		return FALSE;
	}

	//  Determine the form of the first part of the scope specification
	pScope = szRQ1;
	while (*pScope == ' ') pScope++;

	if (*pScope == '\0')
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1902I, szRQ1);
		xeBase->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1903I, szRQ2);
		xeBase->LogMessage(szMsg, iThreadID);
		xeBase->LogMessage(MSG_DXR1905E, iThreadID);
		return FALSE;
	}

	//  Assume that atom represents a server name - it could be qualified by a domain
	pTarget = szServer;
	while ((*pScope != '@') && (*pScope != '\0'))
	{
		*pTarget = *pScope;
		pTarget++;
		pScope++;
	}
	*pTarget = '\0';
	if (*pScope == '@')
	{
		*pScope++;
		pTarget = szDomain;
		while (*pScope != '\0')
		{
			*pTarget = *pScope;
			pTarget++;
			pScope++;
		}
		*pTarget = '\0';
	}
	// Set the match keys and keys according to the current settings
	if (HasWildcards(szDomain))
	{
		Wildcards = Wildcards | WILDCARD_DOMAIN;
		strcpy_s(szMatchDomain, MAX_DOMAIN, szDomain);
		szDomain[0] = '\0';
	}
	if (HasWildcards(szServer))
	{
		Wildcards = Wildcards | WILDCARD_SERVER;
		strcpy_s(szMatchServer, MAX_SERVER, szServer);
		szServer[0] = '\0';
	}

	//  Trace the determination so far
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1906T, AnchorScope, szDomain, szMatchDomain);
	xeBase->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1907T, AnchorScope, szServer, szMatchServer);
	xeBase->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);

	//  Determine the form of the second part of the scope specification
	pScope = szRQ2;
	while (*pScope == ' ') pScope++;

	//  If the scope at this level is universal then there is nothing more to do
	if ((*pScope != '*') && (*pScope != '\0'))
	{
		pTarget = szDirectory;
		pLastDelimiter = NULL;
		pTerminator = NULL;
		while (*pScope != '\0')
		{
			*pTarget = *pScope;
			if ((*pScope == '\\') || (*pScope == '/')) 
			{
				*pTarget = '/';
				pLastDelimiter = pTarget;
			}
			if (*pScope == '.') pTerminator = pTarget;
			pTarget++;
			pScope++;
		}
		*pTarget = '\0';
		
		//  Determine if we have a possible database in the string
		if (pTerminator != NULL)
		{
			if (xeBase->IsThisADatabase(pTerminator))
			{
				if (pLastDelimiter == NULL)
				{
					strcpy_s(szDatabase, MAX_DATABASE, szDirectory);
					szDirectory[0] = '\0';
				}
				else
				{
					strcpy_s(szDatabase, MAX_DATABASE, pLastDelimiter + 1);
					*pLastDelimiter = '\0';
				}
			}
			
		}
	}

	//  Determine the new Anchor Scope & wildcard settings
	if (HasWildcards(szDirectory))
	{
		Wildcards = Wildcards | WILDCARD_DIRECTORY;
		strcpy_s(szMatchDirectory, MAX_DIRECTORY, szDirectory);

		//  Note that it is possible to find an initial radix in the match key that can be used as a
		//  valid starting position for the scan.
		pScope = szDirectory;
		pLastDelimiter = szDirectory;
		while ((*pScope != '*') && (*pScope != '?') && (*pScope != '\0'))
		{
			if (*pScope == '/') pLastDelimiter = pScope;
			pScope++;
		}
		*pLastDelimiter = '\0';
	}

	if (HasWildcards(szDatabase))
	{
		Wildcards = Wildcards | WILDCARD_DATABASE;
		strcpy_s(szMatchDatabase, MAX_DATABASE, szDatabase);
		szDatabase[0] = '\0';
	}

	//  Determine the Anchor Scope
	if ((Wildcards == WILDCARD_NONE) && (szDatabase[0] != '\0')) AnchorScope = RSCOPE_DATABASE;
	else
	{
		if ((Wildcards & (~WILDCARD_DATABASE)) == WILDCARD_NONE) AnchorScope = RSCOPE_DIRECTORY;
		else
		{
			if ((Wildcards & (~(WILDCARD_DATABASE | WILDCARD_DIRECTORY))) == WILDCARD_NONE) 
			{
				//  If we have at least a partial directory key to work from then the scope is DIR
				if (szDirectory[0] != '\0') AnchorScope = RSCOPE_DIRECTORY;
				else AnchorScope = RSCOPE_SERVER;
			}
			else
			{
				if (szDomain[0] != '\0') AnchorScope = RSCOPE_DOMAIN;
				else
				{
					if (AnchorScope == RSCOPE_NONE) AnchorScope = RSCOPE_UNIVERSE;
				}
			}
		}
	}
	
	//  Trace the determination of the scope
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1908T, AnchorScope, szDirectory, szMatchDirectory);
	xeBase->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1909T, AnchorScope, szDatabase, szMatchDatabase);
	xeBase->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);

	return TRUE;
}

//  BaseInheritFrom
//
//  This function will populate the basic members of the current request from a passed request
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//
//  Returns
//
//  API Use Notes:
//
//
void DXRequest::BaseInheritFrom(DXRequest *dxrParent)
{
	//  Inherit the context variables from another request
	hdbCache = dxrParent->hdbCache;
	ScannersNeeded = dxrParent->ScannersNeeded;
	Reporters = dxrParent->Reporters;
	Filters = dxrParent->Filters;
	Wildcards = dxrParent->Wildcards;
	strcpy_s(szMatchDomain, MAX_DOMAIN, dxrParent->szMatchDomain);
	strcpy_s(szMatchServer, MAX_SERVER, dxrParent->szMatchServer);
	strcpy_s(szMatchDirectory, MAX_DIRECTORY, dxrParent->szMatchDirectory);
	strcpy_s(szMatchDatabase, MAX_DATABASE, dxrParent->szMatchDatabase);
	RQLevel = dxrParent->RQLevel + 1;

	//  Inherit the reporters
	dxrUniverse = dxrParent->dxrUniverse;
	dxrDomain = dxrParent->dxrDomain;
	dxrServer = dxrParent->dxrServer;
	dxrDirectory = dxrParent->dxrDirectory;
	dxrDatabase = dxrParent->dxrDatabase;

	//  Inherit the filters
	dxfUniverse = dxrParent->dxfDomain;
	dxfDomain = dxrParent->dxfDomain;
	dxfServer = dxrParent->dxfServer;
	dxfDirectory = dxrParent->dxfDirectory;
	dxfDatabase = dxrParent->dxfDatabase;

	return;
}

//  BasePromoteScores
//
//  This function will accumulate the scanning scores from this request into a higher level request
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//
//  Returns
//
//  API Use Notes:
//
//
void DXRequest::BasePromoteScores(DXRequest *dxrParent)
{
	//  Promote scores to the parent request
	dxrParent->dwServersProcessed += dwServersProcessed;
	dxrParent->dwDirectoriesProcessed += dwDirectoriesProcessed;
	dxrParent->dwDatabasesProcessed += dwDatabasesProcessed;
	dxrParent->dwServersDropped += dwServersDropped;
	dxrParent->dwDirectoriesDropped += dwDirectoriesDropped;
	dxrParent->dwDatabasesDropped += dwDatabasesDropped;
}

//  DocumentRequest
//
//  This function will serialize the current request into a DXRequest Notes Document in the current reqpository
//
//  Parameters
//
//  1  -  ExecEnvironment *		-  Pointer to the current runtime environment
//
//  Returns
//
//  API Use Notes:
//
//
void DXRequest::DocumentRequest(ExecEnvironment *xeBase)
{
	return DocumentRequest(xeBase, 0);
}

//  DocumentRequest
//
//  This function will serialize the current request into a DXRequest Notes Document in the current reqpository
//
//  Parameters
//
//  1  -  ExecEnvironment *		-  Pointer to the current runtime environment
//  2  -  int					-  ID of the thread making the request
//
//  Returns
//
//  API Use Notes:
//
//
void DXRequest::DocumentRequest(ExecEnvironment *xeBase, int iThreadID)
{
	STATUS			stAPIRC;						//  Status code returned by Notes API calls
	NOTEHANDLE		hnDXR;							//  Handle to the documentation note
	char			szMsg[MAX_MSG];					//  Generic message buffer
	char			szFieldValue[MAX_MSG];			//  Formatted data value
	NUMBER			nFieldValue;					//  General number
	long			lElapsed;						//  Elapsed time
	long			lTimeFactor = CLOCKS_PER_SEC;	//  Tme computation

	//  This function will create a DXRequest document in the repository and populate it with
	//  the current request settings and results.

	//  Exit if there is no repository
	if (xeBase->hdbRepository == NULLHANDLE) return;

	//  If the end time has not been set then set it
	if (ctStart == ctEnd)
	{
		ctEnd = xeBase->RunningTime->getElapsed();
	}

	//  Create a new DXRequest document in the repository
	stAPIRC = NSFNoteCreate(xeBase->hdbRepository, &hnDXR);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Set the basic fields
	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, "Form", (WORD) strlen("Form"),
							TYPE_TEXT, DXR_FORM_NAME, (DWORD) strlen(DXR_FORM_NAME));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, REQUESTER_ITEM, (WORD) strlen(REQUESTER_ITEM),
							TYPE_TEXT, xeBase->CommonUser, (DWORD) strlen(xeBase->CommonUser));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Decode the scope and document it
	DecodeScope(szFieldValue, MAX_MSG);
	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, SCOPE_ITEM, (WORD) strlen(SCOPE_ITEM),
							TYPE_TEXT, szFieldValue, (DWORD) strlen(szFieldValue));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Write the scope specifiers
	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, DOMAIN_ITEM, (WORD) strlen(DOMAIN_ITEM),
							TYPE_TEXT, szDomain, (DWORD) strlen(szDomain));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, SERVER_ITEM, (WORD) strlen(SERVER_ITEM),
							TYPE_TEXT, szServer, (DWORD) strlen(szServer));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, DIRECTORY_ITEM, (WORD) strlen(DIRECTORY_ITEM),
							TYPE_TEXT, szDatabase, (DWORD) strlen(szDatabase));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, DATABASE_ITEM, (WORD) strlen(DATABASE_ITEM),
							TYPE_TEXT, szDatabase, (DWORD) strlen(szDatabase));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Write the match specifiers
	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, MDOMAIN_ITEM, (WORD) strlen(MDOMAIN_ITEM),
							TYPE_TEXT, szMatchDomain, (DWORD) strlen(szDomain));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, MSERVER_ITEM, (WORD) strlen(MSERVER_ITEM),
							TYPE_TEXT, szMatchServer, (DWORD) strlen(szServer));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, MDIRECTORY_ITEM, (WORD) strlen(MDIRECTORY_ITEM),
							TYPE_TEXT, szMatchDatabase, (DWORD) strlen(szDatabase));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, MDATABASE_ITEM, (WORD) strlen(MDATABASE_ITEM),
							TYPE_TEXT, szMatchDatabase, (DWORD) strlen(szDatabase));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}


	if (RecurseDirectories)
	{
		strcpy_s(szFieldValue, MAX_MSG, "True");
	}
	else
	{
		strcpy_s(szFieldValue, MAX_MSG, "False");
	}

	stAPIRC = NSFItemAppend(hnDXR, ITEM_SUMMARY, RECURSE_ITEM, (WORD) strlen(RECURSE_ITEM),
							TYPE_TEXT, szFieldValue, (DWORD) strlen(szFieldValue));
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Write the statistics values
	nFieldValue = dwServersProcessed;
	stAPIRC = NSFItemSetNumber(hnDXR, SERVERP_ITEM, &nFieldValue);
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}
	nFieldValue = dwDirectoriesProcessed;
	stAPIRC = NSFItemSetNumber(hnDXR, DIRP_ITEM, &nFieldValue);
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}
	nFieldValue = dwDatabasesProcessed;
	stAPIRC = NSFItemSetNumber(hnDXR, DBP_ITEM, &nFieldValue);
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Compute and write the elapsed time
	if (ctStart == ctEnd)
	{
		lElapsed = 0;
	}
	else
	{
		lElapsed = (ctEnd - ctStart) / lTimeFactor;
	}
	nFieldValue = lElapsed;
	stAPIRC = NSFItemSetNumber(hnDXR, ELAPSED_ITEM, &nFieldValue);
	if (ERR(stAPIRC) != NOERROR)
	{
		NSFNoteClose(hnDXR);
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
		return;
	}
	
	//  Update the note and close it
	stAPIRC = NSFNoteUpdate(hnDXR, UPDATE_FORCE);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeBase->LogMessage(MSG_DXR1901E, iThreadID);
		xeBase->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeBase->LogMessage(szMsg, iThreadID);
	}
	NSFNoteClose(hnDXR);
	return;
}

//  DecodeScope
//
//  This function will decode the request scope into a meaningful string
//
//  Parameters
//
//  1  -  char *		-  Pointer to the string buffer for the decoded value
//  2  -  int			-  Size of the string buffer
//
//  Returns
//
//  API Use Notes:
//
//
void DXRequest::DecodeScope(char *szScope, int iSize)
{
	if (AnchorScope == RSCOPE_UNIVERSE)
	{
		strcpy_s(szScope, iSize, "Entire Infrastructure");
		return;
	}

	if (AnchorScope == RSCOPE_DOMAIN)
	{
		strcpy_s(szScope, iSize, "Domain");
		return;
	}

	if (AnchorScope == RSCOPE_SERVER)
	{
		strcpy_s(szScope, iSize, "Server");
		return;
	}

	if (AnchorScope == RSCOPE_DIRECTORY)
	{
		strcpy_s(szScope, iSize, "Directory");
		return;
	}

	if (AnchorScope == RSCOPE_DATABASE)
	{
		strcpy_s(szScope, iSize, "Database");
		return;
	}
}

//  SetStart
//
//  This function will record the starting time for executing this request
//
//  Parameters
//
//  1   -  ExecEnvironment *	-	Pointer to the Runtime
//
//  Returns
//
//  API Use Notes:
//
//
void DXRequest::SetStart(ExecEnvironment *xeBase)
{
	//  Set the start and finish time
	ctStart = xeBase->RunningTime->getElapsed();
	ctEnd = xeBase->RunningTime->getElapsed();
}

//  SetEnd
//
//  This function will record the completion time for executing this request
//
//  Parameters
//
//  1   -  ExecEnvironment *	-	Pointer to the Runtime
//
//  Returns
//
//  API Use Notes:
//
//
void DXRequest::SetEnd(ExecEnvironment *xeBase)
{
	//  Set the finish time
	ctEnd = xeBase->RunningTime->getElapsed();
}

//  HasWildcards
//
//  This function will check the passed string to see if it contains any wildcard characters
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string to be chacked
//
//  Returns
//
//  BOOL					-  TRUE if there are wildcards in the string
//							-  FALSE if there are no wildcards present
//
//  API Use Notes:
//
//
BOOL DXRequest::HasWildcards(char *szAtom)
{
	char			*pAtom;

	pAtom = szAtom;
	while (*pAtom != '\0')
	{
		if (*pAtom == '*') return TRUE;
		if (*pAtom == '?') return TRUE;
		pAtom++;
	}
	return FALSE;
}

//  CreateNewInstance
//
//  This function will create a new DXRequest instance and return it to the caller. 
//  This function can be overridden in extending classes, to allow extending classes to be handled by
//  the DX core functions.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
DXRequest * DXRequest::CreateNewInstance(void)
{
	return new DXRequest();	
}

//  InheritFrom
//
//  This function will populate the basic members of the current request from a passed request
//  This function can be overridden in extending classes.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//
//  Returns
//
//  API Use Notes:
//
//  In any replacement invoke the BaseInheritFrom() protected method to perform existing inheritance
//
void DXRequest::InheritFrom(DXRequest *dxrParent)
{
	return BaseInheritFrom(dxrParent);
}

//  PromoteScores
//
//  This function will accumulate the scanning scores from this request into a higher level request
//  This function can be overridden in extending classes.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//
//  Returns
//
//  API Use Notes:
//
//  In any replacement invoke the BasePromoteScores() protected method to perform existing promotion
//
void DXRequest::PromoteScores(DXRequest *dxrParent)
{
	return BasePromoteScores(dxrParent);
}

