//  DXSStash
//
//  This class stores and delivers Server scan requests for the DX Spider
//
//  Message Identifier:      DXR
//  Message Range:           1101 - 1199
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.11.1
//  Version:                 3.11.1  Build: 99
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2209		Re-based ready for 4.0.0
//  3.9.2		22/09/2010		Avoided a recursion in DX processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2010
//

#include "DXSStash.h"

//  Default Constructor
DXSStash::DXSStash(void)
: iStashed(0)
, xeBase(NULL)
{
}

//  Normal Constructor
DXSStash::DXSStash(ExecEnvironment *xeParent, int iThreadID)
: iStashed(0)
, xeBase(NULL)
{
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Constructor
	xeBase = xeParent;										//  Store the Execution Environment
	iParentThread = iThreadID;								//  Save the ID of the owning thread

	//  Trace the construction
	if (xeBase->rsLocal->LogLevel > LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1190T, this);
		xeBase->LogMessage(szMsg, iParentThread);
	}	
}

//  Destructor
DXSStash::~DXSStash(void)
{
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Trace the destruction
	if (xeBase->rsLocal->LogLevel > LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1191T, this);
		xeBase->LogMessage(szMsg, iParentThread);
	}
}

//  HasNext
//
//  This function will check if there is a next element to be itereated over in the stash
//
//  Parameters
//
//  Returns
//  BOOL  -			TRUE if there is another element to iterate
//                  FALSE if the end of the stash has been reached in the iteration
//
//  API Use Notes:
//
//

BOOL DXSStash::HasNext(void)
{
	int			iIndex;

	//  NOTE:  Has Next is speculative, the state of stash elements can change
	if (iStashed == 0) return FALSE;					//  Stash is empty

	//  Check each stash member
	for (iIndex = 0; iIndex < iStashed; iIndex++)
	{
		if (dxrStash[iIndex]->ReturnCode == RETURN_STASHED) return TRUE;			//  Entry available
	}
	return FALSE;
}

//  Next
//
//  This function will return the next element to be itereated over in the stash
//
//  Parameters
//
//  Returns
//  DXRequest *		-  Address of the enext entry in the Stash
//					-  NULL if there are no more entries
//                 
//
//  API Use Notes:
//
//

DXRequest * DXSStash::Next(void)
{
	int			iIndex;

	//  NOTE:  HasNext is speculative, callers of next() should check for a NULL return
	if (iStashed == 0) return NULL;					//  Stash is empty

	//  Check each stash member
	for (iIndex = 0; iIndex < iStashed; iIndex++)
	{
		if (dxrStash[iIndex]->ReturnCode == RETURN_STASHED) return dxrStash[iIndex];			//  Return Entry
	}
	return NULL;
}

//  Add
//
//  This function will add a new DXRRequest object to the stash
//
//  Parameters
//
//  1  -  DXRRequest *		-	Address of the entry to be added
//
//  Returns
//                 
//
//  API Use Notes:
//
//

void DXSStash::Add(DXRequest * dxrRQ)
{
	//  Add a request element to the stash if it is not already there
	if (!IsInStash(dxrRQ))
	{
		dxrStash[iStashed] = dxrRQ;
		iStashed++;
	}
}

//  Add
//
//  This function will check if a new DXRRequest object is already in the stash
//
//  Parameters
//
//  1  -  DXRRequest *		-	Address of the entry to be added
//
//  Returns
//         
//  BOOL			-	TRUE if the entry is already in the stash
//					-	FALSE if the entry is not in the stash
//
//  API Use Notes:
//
//

BOOL DXSStash::IsInStash(DXRequest *dxrRQ)
{
	int			iIndex;

	//  Test if a request is already in the stash
	if (iStashed == 0) return FALSE;					//  Stash is empty

	//  Check each stash member
	for (iIndex = 0; iIndex < iStashed; iIndex++)
	{
		if (strcmp(dxrRQ->szServer, dxrStash[iIndex]->szServer) == 0) return TRUE;			//  Entry found
	}
	return FALSE;
}

//  Empty
//
//  This function will drain (delete) all existing requests from the stash
//
//  Parameters
//
//
//  Returns
//         
//  API Use Notes:
//
//

void DXSStash::Empty(void)
{
	int			iIndex;

	//  Empty the content of the existing stash
	if (iStashed == 0) return;					//  Stash is already empty

	//  Eliminate each stash member
	for (iIndex = 0; iIndex < iStashed; iIndex++)
	{
		delete dxrStash[iIndex];
		dxrStash[iIndex] = NULL;
	}

	iStashed = 0;
	return;
}

