//  DXSpider
//
//  This class is the top level eXplorer class, it will orchestrate the search strategy that best matches
//  the scope and constraints from the search request.
//
//  Message Identifier:      DXR
//  Message Range:           1101 - 1199
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.14.0
//  Version:                 3.14.0  Build: 108
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2209		Re-based ready for 4.0.0
//  3.9.2		22/09/2010		Avoided a recursion in DX processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.11.1		22/10/2011		VS 2011 Updates
//	3.14.0		21/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2015
//

//  Application Includes
#include "DXSpider.h"

//  Single-Thread constructor
DXSpider::DXSpider(ExecEnvironment *xeParent, int iThreadID)
:xeLocal(NULL)
, dxssInstance(NULL)
, dxsStash(NULL)
{
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer
	//  Constructor
	xeLocal = xeParent;
	iParentThread = iThreadID;

	//  Trace message
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1103T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);
}

//  Destructor
DXSpider::~DXSpider(void)
{
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Destroy the Server Scanner and the stash if they exist
	if (dxsStash != NULL) delete dxsStash;
	if (dxssInstance != NULL) delete dxssInstance;

	//  Trace message
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1104T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);
}

//  ProcessRequest
//
//  This function is the main entry point for dispatching a Domino eXplorer request.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  Thread ID of the requester
//
//  Returns
//
//  BOOL  -			TRUE if the search was successful
//                  FALSE if the search was not successful
//
//  API Use Notes:
//
//
BOOL DXSpider::ProcessRequest(DXRequest *dxrRQ, int iThreadID) 
{
	//  Single Thread - Execute a request
	return CoreProcessRequest(dxrRQ, iThreadID);
}

//  InitiallyPopulateStash
//
//  This function will perform the initial population of the Server stach object.
//
//  Parameters
//
//  1  -  int		- ID of the calling thread
//
//  Returns
//
//  BOOL  -			TRUE if the initial population was successful
//                  FALSE if the initial population was not successful
//
//  API Use Notes:
//
//
BOOL DXSpider::InitiallyPopulateStash(int iThreadID)
{
	STATUS			stAPIRC;								//  Return code from Notes API calls
	DBHANDLE		hdbNAB;									//  Handle to the name and address book
	NOTEID			nidDesign;								//  Noteid of design element

	//  This function will find a NAB to start the spidering process from

	//  Open the local "names.nsf" and determine if it is a Server (i.e. public) or a personal NAB
	stAPIRC = NSFDbOpen(DEFAULT_NAB_DB, &hdbNAB);
	if (ERR(stAPIRC) != NOERROR)
	{
		DXException		*dxeLocal;

		//  Create aand throw an exception
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_NOTESAPI;
		dxeLocal->ExceptionCode = ERR(stAPIRC);
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1105S);
		dxeLocal->DestroyContext = true;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
		dxeLocal->LineNumber = __LINE__;
		
		//  Throw the exception
		throw(*dxeLocal);
	}

	//  Determine if the NAB is a public or a personal NAB
	stAPIRC = NIFFindDesignNote(hdbNAB, NETAUTH_NAMESPACE_SERVERS, NOTE_CLASS_VIEW, &nidDesign);
	if ((ERR(stAPIRC) != NOERROR) && (ERR(stAPIRC) != ERR_NOT_FOUND))
	{
		DXException		*dxeLocal;

		NSFDbClose(hdbNAB);									//  Close the NAB

		//  Create aand throw an exception
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_NOTESAPI;
		dxeLocal->ExceptionCode = ERR(stAPIRC);
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1106E);
		dxeLocal->DestroyContext = true;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
		dxeLocal->LineNumber = __LINE__;
		
		//  Throw the exception
		throw(*dxeLocal);
	}

	if (ERR(stAPIRC) == NOERROR)
	{
		//  Name and Address book is a public NAB, we can pass this to UpdateStash to populate the requests
		if (dxssInstance->UpdateStash(hdbNAB, iThreadID))
		{
			DXException		*dxeLocal;

			NSFDbClose(hdbNAB);									//  Close the NAB

			//  Create aand throw an exception
			dxeLocal = new DXException();
			dxeLocal->ExceptionType = ERRORTYPE_NOTESAPI;
			dxeLocal->ExceptionCode = ERR(stAPIRC);
			dxeLocal->FaultingThread = iThreadID;
			strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1107E);
			dxeLocal->DestroyContext = true;
			strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
			dxeLocal->LineNumber = __LINE__;
		
			//  Throw the exception
			throw(*dxeLocal);
		}

		xeLocal->LogTrace(TRACE_EXPLORER, MSG_DXR1108T, iThreadID);
		NSFDbClose(hdbNAB);
		return TRUE;
	}

	if (ERR(stAPIRC) == ERR_NOT_FOUND)
	{
		//  Name and Address book is a personal NAB, we need to find a server to perform the initial population of the stash
		hdbNAB = SwapNAB(hdbNAB, iThreadID);
		if (hdbNAB == NULLHANDLE)
		{
			DXException		*dxeLocal;

			//  Create aand throw an exception
			dxeLocal = new DXException();
			dxeLocal->ExceptionType = ERRORTYPE_NOTESAPI;
			dxeLocal->ExceptionCode = ERR(stAPIRC);
			dxeLocal->FaultingThread = iThreadID;
			strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1109E);
			dxeLocal->DestroyContext = true;
			strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
			dxeLocal->LineNumber = __LINE__;
		
			//  Throw the exception
			throw(*dxeLocal);
		}

		if (!dxssInstance->UpdateStash(hdbNAB, iThreadID))
		{
			DXException		*dxeLocal;

			NSFDbClose(hdbNAB);									//  Close the NAB

			//  Create aand throw an exception
			dxeLocal = new DXException();
			dxeLocal->ExceptionType = ERRORTYPE_NOTESAPI;
			dxeLocal->ExceptionCode = ERR(stAPIRC);
			dxeLocal->FaultingThread = iThreadID;
			strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1107E);
			dxeLocal->DestroyContext = true;
			strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
			dxeLocal->LineNumber = __LINE__;
		
			//  Throw the exception
			throw(*dxeLocal);
		}

		xeLocal->LogTrace(TRACE_EXPLORER, MSG_DXR1108T);
		NSFDbClose(hdbNAB);
		return TRUE;
	}
	return FALSE;
}

//  SwapNAB
//
//  This function will attempt to find another NAB to be used for the spidering process.
//
//  Parameters
//
//  1  -  DBHANDLE		-  DBHANDLE of the NAB that is currently being used by the spidering process
//  2  -  int			-  ID of the calling Thread
//
//  Returns
//
//  DBHANDLE  -		DBHANDLE of the new NAB
//                  NULLHANDLE if no alternate NAB can be found
//
//  API Use Notes:
//
//
DBHANDLE DXSpider::SwapNAB(DBHANDLE hdbPNAB, int iThreadID)
{
	STATUS			stAPIRC;											//  Return codes from Notes API calls
	DBHANDLE		hdbNAB;												//  Handle to NAB database
	DHANDLE			hRespBuffer;										//  Response buffer from NSGetServerList
	WORD			*pwServerCount;										//  Pointer to the count of servers in the list
	WORD			*pwServerNameLen;									//  Server Name Lengths
	int				iIndex;												//  Index of server names
	char			szMailServer[MAXENVVALUE + 1];						//  Mail server
	char			szPortList[MAXENVVALUE + 1];						//  List of ports that are in use
	char			szMsg[MAX_MSG + 1];									//  Generic message buffer
	char			*pPort;												//  Pointer to the current port
	char			*pServerList;										//  Pointer to the server list response
	char			*pServerName;										//  Pointer to the server name

	//  Try and find a server NAB to start the spider process from
	xeLocal->LogTrace(TRACE_EXPLORER, MSG_DXR1110T);

	//  First try using the mailserver= setting from the notes.ini
	if (OSGetEnvironmentString("Mailserver", szMailServer, MAXENVVALUE))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1111I, szMailServer);
		xeLocal->LogVerbose(szMsg, iThreadID);

		//  Attempt to open the NAB on the target server
		hdbNAB = xeLocal->OpenDatabase(szMailServer, DEFAULT_NAB_DB, iThreadID);
		if (hdbNAB != NULLHANDLE)
		{
			NSFDbClose(hdbPNAB);
			return hdbNAB;
		}
	}

	//  Get the list of available ports from notes.ini
	if (OSGetEnvironmentString(ENV_PORTS_ENTRY, szPortList, MAXENVVALUE))
	{
		if (strlen(szPortList) > 0)
		{
			//  Work through each port in turn until we get a useable server
			pPort = szPortList + strlen(szPortList);					//  Point to the end of the string
			while (pPort > szPortList)
			{
				while ((*pPort != ',') && (pPort > szPortList))
				{
					pPort--;
				}
				if (pPort > szPortList)
				{
					*pPort = '\0';
					pPort++;
				}

				//  Get the list of servers known on the current port
				sprintf_s(szMsg, MAX_MSG, MSG_DXR1112T, pPort);
				xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);

				stAPIRC = NSGetServerList(pPort, &hRespBuffer);
				if (ERR(stAPIRC) == NOERROR)
				{
					//  Process the returned list of server names
					pServerList = (char *) OSLockObject(hRespBuffer);
					pwServerCount = (WORD *) pServerList;
					pwServerNameLen = pwServerCount + sizeof(WORD);
					pServerName = pServerList + (sizeof(WORD) * (*pwServerCount + 1));
					for (iIndex = 0; iIndex < *pwServerCount; iIndex++)
					{
						memcpy(szMailServer, pServerName, pwServerCount[iIndex]);
						szMailServer[pwServerCount[iIndex]] = '\0';
						sprintf_s(szMsg, MAX_MSG, MSG_DXR1111I, szMailServer);
						xeLocal->LogVerbose(szMsg, iThreadID);

						//  Attempt to open the NAB on the target server
						hdbNAB = xeLocal->OpenDatabase(szMailServer, DEFAULT_NAB_DB, iThreadID);
						if (hdbNAB != NULLHANDLE)
						{
							OSUnlockObject(hRespBuffer);
							OSMemFree(hRespBuffer);
							NSFDbClose(hdbPNAB);
							return hdbNAB;
						}
						pServerName += pwServerCount[iIndex];
					}
				}
			}
		}
	}

	//  No starting server was found
	NSFDbClose(hdbPNAB);
	return NULLHANDLE;
}

//  ValidateDomain
//
//  This function will check that if a parent request specifies a domain then the child request
//  belongs to the specified domain.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a Parent DXRequest object
//  2  -  DXRequest *		-  Pointer to a Child DXRequest object
//  3  -  int				-  ID of the invoking Thread
//
//  Returns
//
//  BOOL  -			TRUE if the child is from the specified domain
//                  FALSE if the child does not belong to the specified domain 
//
//  API Use Notes:
//
//
BOOL DXSpider::ValidateDomain(DXRequest *dxrRQ, DXRequest *dxrSRQ, int iThreadID)
{
	//  Check that the domain is as requested
	if (dxrRQ->AnchorScope != RSCOPE_DOMAIN)
	{
		//  If there is a wildcard match string specified then attempt to match it
		if (dxrRQ->Wildcards & WILDCARD_DOMAIN)
		{
			if (xeLocal->MatchesPatterni(dxrSRQ->szDomain, dxrRQ->szMatchDomain)) return TRUE;
			return FALSE;
		}
		else return TRUE;
	}
	if (strcmp(dxrSRQ->szDomain, dxrRQ->szDomain) == 0) return TRUE;
	return FALSE;
}

//  CoreProcessRequest
//
//  This function is the main execution path for an eXplorer request that requires spidering.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the invoking Thread
//
//  Returns
//
//  BOOL  -			TRUE if the spider was successful
//                  FALSE if the spider was not successful
//
//  API Use Notes:
//
//
BOOL DXSpider::CoreProcessRequest(DXRequest *dxrRQ, int iThreadID) 
{
	DXRequest			*dxrSRQ;						//  Server Requests
	DXFilter			*dxfSelector;					//  Selection Filter
	BOOL				bProcessThisRequest;			//  Control switch
	char				szMsg[MAX_MSG + 1];				//  Generic message buffer

	//  Process a scan request
	dxrRQ->ReturnCode = RETURN_INCOMPLETE;				//  Mark the request as incomplete

	//  Determine if the request is at the universal scope level. If not it should be treated as a pass through
	//  and passed to the level below.

	if ((dxrRQ->AnchorScope == RSCOPE_UNIVERSE) || (dxrRQ->AnchorScope == RSCOPE_DOMAIN))
	{
		dxrRQ->SetStart(xeLocal);
		try
		{
			//  Build a stash if one does not exist or drain the existing stash
			if (dxsStash == NULL)
			{
				dxsStash = new DXSStash(xeLocal, iThreadID);
			}
			else
			{
				dxsStash->Empty();
			}

			//  Create the lower level scanner (Server Scanner)
			if (dxssInstance == NULL)
			{
				//  Create a new Server Scanner
				dxssInstance = new DXServerScanner(xeLocal, iThreadID);

				//  Set the stash address in the Server Scanner
				dxssInstance->dxsStash = dxsStash;
			}

			//  Initially populate the stash
			if (!InitiallyPopulateStash(iThreadID))
			{
				return FALSE;
			}

			//  Now process the content of the stash until it is exhausted
			while (dxsStash->HasNext())
			{
				dxrSRQ = dxsStash->Next();
				if (dxrSRQ != NULL)
				{
					dxrSRQ->ReturnCode = RETURN_NOERROR;					//  Mark it no longer stashed

					//  Trace the details
					sprintf_s(szMsg, MAX_MSG, MSG_DXR1114T, dxrSRQ->szServer, dxrSRQ->szDomain);
					xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
					
					//  If the top level request is for the domain level then screen the stashed request
					if (ValidateDomain(dxrRQ, dxrSRQ, iThreadID))
					{
						//  If there is a Domain Level filter in effect then see if the request is to be
						//  filtered
						bProcessThisRequest = TRUE;
						if ((dxrRQ->Filters & FILTER_DOMAIN) && (dxrRQ->dxfDomain != NULL))
						{
							dxfSelector = (DXFilter *) dxrRQ->dxfDomain;
							bProcessThisRequest = dxfSelector->FilterThisDomain(dxrSRQ, iThreadID);
						}

						if (bProcessThisRequest)
						{

							//  Inherit additional information from the parent request
							dxrSRQ->InheritFrom(dxrRQ);

							//  Pass the request to the Server Scanner
							if (!dxssInstance->ProcessRequest(dxrSRQ, iThreadID))
							{
								//  Processed with errors
								sprintf_s(szMsg, MAX_MSG, MSG_DXR1115E, dxrSRQ->szServer);
								xeLocal->LogMessage(szMsg, iThreadID);
							}
							//  Accumulate the scores
							dxrSRQ->PromoteScores(dxrRQ);
							delete dxrSRQ;
						}
						else
						{
							//  Server has been dropped from this search scope
							dxrRQ->dwServersDropped++;
							sprintf_s(szMsg, MAX_MSG, MSG_DXR1116I, dxrSRQ->szServer, dxrSRQ->szDomain);
							xeLocal->LogVerbose(szMsg, iThreadID);
						}
					}	
					else
					{
						dxrRQ->dwServersDropped++;
						sprintf_s(szMsg, MAX_MSG, MSG_DXR1113I, dxrSRQ->szServer, dxrSRQ->szDomain, dxrRQ->szDomain);
						xeLocal->LogVerbose(szMsg, iThreadID);
					}
				}
			}
		}
		catch(DXException& dxeCurrent)
		{
			//  Destroy the server scanner if it is now unstable
			if (dxeCurrent.DestroyContext)
			{
				delete dxssInstance;
				dxssInstance = NULL;
			}
			dxeCurrent.IsPercolated = true;
			//  Rethrow the exception to the next level - percolate
			throw;
		}
		catch(std::exception eCurrent)
		{
			DXException		*dxeLocal;

			//  Destroy the server scanner as it may now be unstable
			delete dxssInstance;
			dxssInstance = NULL;

			//  Create a new exception to wrapper the information from the original
			dxeLocal = new DXException();
			dxeLocal->ExceptionType = ERRORTYPE_UHX;
			dxeLocal->ExceptionCode = ERRORCODE_NON_SPEC;
			dxeLocal->FaultingThread = iThreadID;
			strcpy_s(dxeLocal->LocalExceptionMessage, MAX_MSG, eCurrent.what());
			strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1102E);
			dxeLocal->DestroyContext = true;
			strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
			dxeLocal->LineNumber = __LINE__;
		
			//  Throw the exception
			throw(*dxeLocal);
		}
		//  Document the request
		dxrRQ->SetEnd(xeLocal);
		if (xeLocal->LogGetLevel() > LOGLEVEL_VERBOSE) dxrRQ->DocumentRequest(xeLocal);
		return TRUE;
	}

	//  Trace the passthrough action
	xeLocal->LogTrace(TRACE_EXPLORER, MSG_DXR1101T, iThreadID);

	//  Pass the request to the next level scanner
	try
	{
		if (dxssInstance == NULL)
		{
			//  Create a new Server Scanner
			dxssInstance = new DXServerScanner(xeLocal, iThreadID);
		}
	
		//  Pass the request to the Server Scanner
		if (dxssInstance->ProcessRequest(dxrRQ, iThreadID))
		{
			//  Processed ok
			return TRUE;
		}
		else
		{
			//  Processed with errors
			return FALSE;
		}
	}
	catch(DXException& dxeCurrent)
	{
		//  Destroy the server scanner if it is now unstable
		if (dxeCurrent.DestroyContext)
		{
			delete dxsStash;
			dxsStash = NULL;
			delete dxssInstance;
			dxssInstance = NULL;
		}
		dxeCurrent.IsPercolated = true;
		//  Rethrow the exception to the next level - percolate
		throw;
	}
	catch(std::exception eCurrent)
	{
		DXException		*dxeLocal;

		//  Destroy the server scanner as it may now be unstable
		delete dxsStash;
		dxsStash = NULL;
		delete dxssInstance;
		dxssInstance = NULL;

		//  Create a new exception to wrapper the information from the original
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_UHX;
		dxeLocal->ExceptionCode = ERRORCODE_NON_SPEC;
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->LocalExceptionMessage, MAX_MSG, eCurrent.what());
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1102E);
		dxeLocal->DestroyContext = true;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
		dxeLocal->LineNumber = __LINE__;
		
		//  Throw the exception
		throw(*dxeLocal);
	}
	return FALSE;
}

