#include "DominoExplorer.h"
//  DominoExplorer
//
//  This class provides the root scanning capabilities for Domino eXplorer (DX) applications.
//
//  Message Identifier:      DXR
//  Message Range:           1001 - 1099
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.11.5
//  Version:                 3.11.5  Build: 103
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2209		Re-based ready for 4.0.0
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.5		30/11/2011		Recursion avoidance improvements
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2011
//

//  Constructor taking ExecEnvironment
DominoExplorer::DominoExplorer(ExecEnvironment *xeParent)
: IsStarted(FALSE)
, IsAvailable(FALSE)
, dxsRoot(NULL)
{
	char				szMsg[MAX_MSG + 1];							//  Generic message buffer
	xeLocal = xeParent;												//  Save the execution environment
	iParentThread = 0;												//  Set the parent thread ID
	
	//  Trace the creation
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1017T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);

	//  Mark the Explorer as available and ready	
	IsStarted = TRUE;
	IsAvailable = TRUE;
}

//  Constructor taking ExecEnvironment and owner thread ID
DominoExplorer::DominoExplorer(ExecEnvironment *xeParent, int iThreadID)
: IsStarted(FALSE)
, IsAvailable(FALSE)
, dxsRoot(NULL)
{
	char				szMsg[MAX_MSG + 1];							//  Generic message buffer
	xeLocal = xeParent;												//  Save the execution environment
	iParentThread = iThreadID;										//  Set the parent thread ID
	
	//  Trace the creation
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1017T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);

	//  Mark the Explorer as available and ready	
	IsStarted = TRUE;
	IsAvailable = TRUE;
}

//  Destructor
DominoExplorer::~DominoExplorer(void)
{
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Destroy the spider if it exists
	if (dxsRoot != NULL)
	{
		delete dxsRoot;
	}

	//  Trace the deletion
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1022T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);
}

//  ProcessRequest
//
//  This function is the main entry point for dispatching a Domino eXplorer request.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//
//  Returns
//
//  BOOL  -			TRUE if the search was successful
//                  FALSE if the search was not successful
//
//  API Use Notes:
//
//
BOOL DominoExplorer::ProcessRequest(DXRequest *dxrRQ)
{
	//  This  function is a wrapper for the CoreProcessRequest() function
	return CoreProcessRequest(dxrRQ, 0);
}

//  ProcessRequest
//
//  This function is the main entry point for dispatching a Domino eXplorer request.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the search was successful
//                  FALSE if the search was not successful
//
//  API Use Notes:
//
//
BOOL DominoExplorer::ProcessRequest(DXRequest *dxrRQ, int iThreadID)
{
	//  This  function is a wrapper for the CoreProcessRequest() function
	return CoreProcessRequest(dxrRQ, iThreadID);
}

//  GetBuildInfo_s
//
//  This function will return the build information about the Domino eXplorer
//
//  Parameters
//
//  1  -  char *		-  Pointer to a string buffer where the build info will be returned
//  2  -  int			-  Size of the string buffer
//
//  Returns
//
//  API Use Notes:
//
//
void DominoExplorer::GetBuildInfo_s(char *szBuild, int iSize)
{
	//  Return the build identity
	sprintf_s(szBuild, iSize, DX_IDENTITY, DX_BUILD);
	return;
}

//  CoreProcessRequest
//
//  This function is the main execution function for dispatching a Domino eXplorer request.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the thread making the request
//
//  Returns
//
//  BOOL  -			TRUE if the search was successful
//                  FALSE if the search was not successful
//
//  API Use Notes:
//
//
BOOL DominoExplorer::CoreProcessRequest(DXRequest *dxrRQ, int iThreadID)
{
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Process a Domino Explorer request

	//  Check that the request scope is valid - if it is RSCOPE_NONE then it implies that the 
	//  request has not been parsed

	if (dxrRQ->AnchorScope == RSCOPE_NONE)
	{
		xeLocal->LogMessage(MSG_DXR1031E, iThreadID);
		dxrRQ->ReturnCode = RETURN_RQ_REJECTED;
		return FALSE;
	}

	if (dxrRQ->AnchorScope > RSCOPE_UNIVERSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1032E, dxrRQ->AnchorScope);
		xeLocal->LogMessage(szMsg, iThreadID);
		dxrRQ->ReturnCode = RETURN_RQ_REJECTED;
		return FALSE;
	}

	//  Set the initial request level
	dxrRQ->RQLevel = 1;

	sprintf_s(szMsg, MAX_MSG, MSG_DXR1001T, dxrRQ, dxrRQ->AnchorScope, dxrRQ->Wildcards);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	
	//  Show the Request Values
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1030T, "Domain: Key:", dxrRQ->szDomain, "Match:", dxrRQ->szMatchDomain);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1030T, "Server: Key:", dxrRQ->szServer, "Match:", dxrRQ->szMatchServer);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1030T, "Directory: Key:", dxrRQ->szDirectory, "Match:", dxrRQ->szMatchDirectory);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1030T, "Database: Key:", dxrRQ->szDatabase, "Match:", dxrRQ->szMatchDatabase);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1020T, dxrRQ->ScannersNeeded, dxrRQ->Reporters, dxrRQ->Filters);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);

	//  Attempt to execute the eXplorer request that was passed
	try
	{
		//  Create a Spider (top level explorer scanner) - if it does not already exist
		if (dxsRoot == NULL)
		{
			dxsRoot = new DXSpider(xeLocal, iThreadID);
		}

		//  Pass the request to the Spider
		if (dxsRoot->ProcessRequest(dxrRQ, iThreadID))
		{
			//  Request was successfully processed - check the processing return code
			if (dxrRQ->ReturnCode == RETURN_NOERROR)
			{
				xeLocal->LogMessage(MSG_DXR1013I);
			}
			else
			{
				if (dxrRQ->ReturnCode == RETURN_CWARN)
				{
					xeLocal->LogMessage(MSG_DXR1014W);
				}
				else
				{
					if (dxrRQ->ReturnCode == RETURN_CERR)
					{
						xeLocal->LogMessage(MSG_DXR1015E);
					}
					else
					{
						xeLocal->LogMessage(MSG_DXR1016E);
					}
				}
			}

			//  Document then log the request
			dxrRQ->DocumentRequest(xeLocal, iThreadID);

			sprintf_s(szMsg, MAX_MSG, MSG_DXR1023I, dxrRQ->dwServersProcessed, dxrRQ->dwServersDropped);
			xeLocal->LogMessage(szMsg);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1024I, dxrRQ->dwDirectoriesProcessed, dxrRQ->dwDirectoriesDropped);
			xeLocal->LogMessage(szMsg);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1025I, dxrRQ->dwDatabasesProcessed, dxrRQ->dwDatabasesDropped);
			xeLocal->LogMessage(szMsg);
			return TRUE;
		}
		else
		{
			//  Request failed
			dxrRQ->DocumentRequest(xeLocal);
			xeLocal->LogMessage(MSG_DXR1012E);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1023I, dxrRQ->dwServersProcessed, dxrRQ->dwServersDropped);
			xeLocal->LogMessage(szMsg);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1024I, dxrRQ->dwDirectoriesProcessed, dxrRQ->dwDirectoriesDropped);
			xeLocal->LogMessage(szMsg);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1025I, dxrRQ->dwDatabasesProcessed, dxrRQ->dwDatabasesDropped);
			xeLocal->LogMessage(szMsg);
		}
	}
	catch (DXException& dxeCurrent)
	{
		//  Show the message(s)
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1011S, dxeCurrent.ExceptionType, dxeCurrent.ExceptionCode, dxeCurrent.FaultingThread);
		xeLocal->LogMessage(szMsg);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1021S, dxeCurrent.LineNumber, dxeCurrent.OriginatingFunction);
		xeLocal->LogMessage(szMsg);
		if (dxeCurrent.IsPercolated) xeLocal->LogMessage(MSG_DXR1018S);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1019S, dxeCurrent.OriginatingFunction, dxeCurrent.LineNumber);
		xeLocal->LogMessage(dxeCurrent.DXExceptionMessage);
		if (dxeCurrent.ExceptionType == ERRORTYPE_NOTESAPI)
		{
			xeLocal->GetAPIMessage(dxeCurrent.ExceptionCode, szMsg);
			xeLocal->LogMessage(szMsg);
		}
		if (dxeCurrent.LocalExceptionMessage[0] != '\0')
		{
			xeLocal->LogMessage(dxeCurrent.LocalExceptionMessage);
		}

		//  Destroy the spider as it may now be unstable
		if (dxeCurrent.DestroyContext)
		{
			delete dxsRoot;
			dxsRoot = NULL;
		}
	}
	catch (std::exception& eGeneric)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1010S, eGeneric.what());
		xeLocal->LogMessage(szMsg);

		//  Destroy the spider as it may now be unstable
		delete dxsRoot;
		dxsRoot = NULL;
	}
	return FALSE;
}

