//  DXResourceLoader
//
//  Resource Loader (usually XML documents) for DX applications. The resource loader is capable of loading an XML
//  document or other resource from a remote HTTP source, a native file, a resource embedded in the current application,
//  a Text List field in a Note or a Rich Text Item in a Note. The resource is loaded into a Domino memory allocation
//  and a handle (unlocked) is returned to the caller.
//  Files can be uploaded to a new Domino document.
//
//  Message Identifier:      DXR
//  Message Range:           7001 - 7099
//  Trace area assignment:   20
//  Version:                 3.14.0			Build:  108
//
//  Version History:
//
//  1.0.0		12/08/2009		Initial version
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.3		17/12/2010		Spring cleaning
//  3.10.6		19/03/2011		Upload capability added to DXResourceLoader
//  3.10.9		04/06/2011		CURL implementation for HTTP on UNIX
//  3.11.1		22/10/2011		VS 2011 Updates
//	3.14.0		21/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 20115
//

#include "DXResourceLoader.h"

//  Constructor
DXResourceLoader::DXResourceLoader(ExecEnvironment *xeParent, int iThreadID)
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Set the DX Runtime
	xeLocal = xeParent;
	ParentThread = iThreadID;

	//  Initialise the stateful variables
#ifdef WIN32
	hWHFacility = NULLHANDLE;
#endif
#ifdef UNIX
	hcFacility = NULLHANDLE;
#endif

	//  Trace construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7001T, this);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, ParentThread);
}

//  Destructor
DXResourceLoader::~DXResourceLoader(void)
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

#ifdef WIN32
	if (hWHFacility != NULLHANDLE)
	{
		//  If the WinHTTP Component has been created then destroy it
		WinHttpCloseHandle(hWHFacility);
	}
#endif
#ifdef UNIX
	if (hcFacility != NULLHANDLE)
	{
		//  If the CURL library and session is in use then terminate it
		curl_easy_cleanup(hcFacility);
		curl_global_cleanup();
	}
#endif

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7002T, this);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, ParentThread);
}

//  loadThisResource
//
//  Generic entry point taking the name of a resource, the name is parsed and then a more specific entry point is
//  used depending on the type of resource source that was specified.
//
//  Parameters
//
//  1  -  char *  - The resource source specifier URL, filename or Resource ID
//  2  -  int     - ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE  -  Handle to the Domino Memory allocation containing the loaded resource
//                NULLHANDLE if the resource could not be loaded.
//
//  API Use Notes:
//
//  1. It is the responsibility of the caller to free the returned memory handle using OSMemoryFree() once used.
//  2. Application resources must be of type "DXRESOURCE" to be located and used by DX components.
//
MEMHANDLE DXResourceLoader::loadThisResource(char *szResourceSource, int iThreadID)
{
	int			iSourceType = RSOURCE_INVALID;				//  Resource Source Type
	char		szMsg[MAX_MSG + 1];							//  Generic Message Buffer

	//  Determine the resource source type
	iSourceType = determineSourceType(szResourceSource);

	//  Log the start of the request
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7004I, szResourceSource, iSourceType);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Process According to the type
	switch(iSourceType)
	{
	case RSOURCE_ARSRC:
		//  Application Resource
		return loadResourceFromApplication(szResourceSource, iThreadID);
	case RSOURCE_OSFILE:
		//  Operating System File
		return loadResourceFromFile(szResourceSource, iThreadID);
	case RSOURCE_URL:
		//  URL
		return loadResourceFromURL(szResourceSource, iThreadID);
	case RSOURCE_INVALID:
		//  An invalid or unrecognized source type
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7003E, szResourceSource);
		xeLocal->LogMessage(szMsg, iThreadID);
		break;
	}
	return NULLHANDLE;
}

//  loadThisResourceOnDemand
//
//  Generic entry point taking the name of a resource, the name is parsed and then a more specific entry point is
//  used depending on the type of resource source that was specified.
//
//  Parameters
//
//  1  -  char *  - The resource source specifier URL, filename or Resource ID
//  2  -  int     - ID of the invoking thread
//
//  Returns
//
//  DXRESOURCEHANDLE  -  Handle to the DXResource that will be used to retrieve the data
//						 NULLHANDLE if the resource could not be created
//
//  API Use Notes:
//
//  1. It is the responsibility of the caller to free the returned DXRESOURCEHANDLE by using the dispose() function
//  2. Application resources must be of type "DXRESOURCE" to be located and used by DX components.
//
DXRESOURCEHANDLE DXResourceLoader::loadThisResourceOnDemand(char *szResourceSource, int iThreadID)
{
	int			iSourceType = RSOURCE_INVALID;				//  Resource Source Type
	MEMHANDLE	hBaseResource;								//  Memory handle of base resource
	char		szMsg[MAX_MSG + 1];							//  Generic Message Buffer

	//  Determine the resource source type
	iSourceType = determineSourceType(szResourceSource);

	//  Log the start of the request
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7004I, szResourceSource, iSourceType);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Process According to the type
	switch(iSourceType)
	{
	case RSOURCE_ARSRC:
		//  Application Resource
		hBaseResource = loadResourceFromApplication(szResourceSource, iThreadID);
		//  Change the request into a Load On Demand from a memory resource
		return loadResourceOnDemandFromMemory(hBaseResource, iThreadID);
	case RSOURCE_OSFILE:
		//  Operating System File
		return loadResourceOnDemandFromFile(szResourceSource, iThreadID);
	case RSOURCE_URL:
		//  URL
		return loadResourceOnDemandFromURL(szResourceSource, iThreadID);
	case RSOURCE_INVALID:
		//  An invalid or unrecognized source type
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7003E, szResourceSource);
		xeLocal->LogMessage(szMsg, iThreadID);
		break;
	}
	return NULLHANDLE;
}

//  loadThisResource
//
//  Generic entry point taking a handle to a source Note and the name of an Item containing the resource.
//  the function will vector to a more specific function depending on the type of the Item.
//
//  Parameters
//
//  1  -  NOTEHANDLE  - Handle to the Note containing the resource source
//  2  -  char *      - String buffer (NULL terminated) containg the name of the Item containing the resource source
//  3  -  int		  - ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE  -  Handle to the Domino Memory allocation containing the loaded resource
//                NULLHANDLE if the resource could not be loaded.
//
MEMHANDLE DXResourceLoader::loadThisResource(NOTEHANDLE hnSource, char *szItemName, int iThreadID)
{
	STATUS			stAPIRC;					//  Return code from Notes API calls
	DWORD			dwValueLen;					//  Returned value length
	WORD			wType;						//  Returned Item type
	BLOCKID			bidItem;					//  The Blockid for the specified item
	BLOCKID			bidValue;					//  The Blockid for the specified items value
	char			szMsg[MAX_MSG + 1];			//  Generic Message Buffer

	//  Make sure that the Item Name is not a null string
	if (szItemName[0] == '\0')
	{
		xeLocal->LogMessage(MSG_DXR7040E, iThreadID);
		return NULLHANDLE;
	}

	// Detrmine if the Item is a Text List or Rich Text Item the vector to the appropriate loader
	// if the item is any other type then the load request is invalid
	stAPIRC = NSFItemInfo(hnSource, szItemName, (WORD) strlen(szItemName), &bidItem, &wType, &bidValue, &dwValueLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7041E, szItemName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Process the request according to the Item Type
	switch(wType)
	{
	case TYPE_TEXT_LIST:
		//  Load the resource from a Text List Item
		return loadResourceFromText(bidValue, dwValueLen, szItemName, iThreadID);

	case TYPE_COMPOSITE:
		//  Load the resource from a Rich Text Item
		return loadResourceFromRTI(bidValue, dwValueLen, szItemName, iThreadID);

	default:
		//  This is an unsupported item type reject it
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7042E, szItemName, iThreadID);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}
	return NULLHANDLE;
}

//  loadThisResourceOnDemand
//
//  Generic entry point taking a handle to a source Note and the name of an Item containing the resource.
//  the function will vector to a more specific function depending on the type of the Item.
//
//  Parameters
//
//  1  -  NOTEHANDLE  - Handle to the Note containing the resource source
//  2  -  char *      - String buffer (NULL terminated) containg the name of the Item containing the resource source
//  3  -  int		  - ID of the invoking thread
//
//  Returns
//
//  DXRESOURCEHANDLE  -  Handle to the DXResource that will be used to retrieve the data
//						 NULLHANDLE if the resource could not be created
//
//  API Use Notes:
//
//  1. It is the responsibility of the caller to free the returned DXRESOURCEHANDLE by using the dispose() function
//
DXRESOURCEHANDLE DXResourceLoader::loadThisResourceOnDemand(NOTEHANDLE hnSource, char *szItemName, int iThreadID)
{
	MEMHANDLE		hResource;					//  Handle of the loaded Resource

	//  Load the resource into memory
	hResource = loadThisResource(hnSource, szItemName, iThreadID);
	if (hResource == NULLHANDLE) return NULLHANDLE;

	//  Load on-demand as a memory source
	return loadResourceOnDemandFromMemory(hResource, iThreadID);
}

//  loadThisResourceOnDemand
//
//  Specific entry point taking a handle to a memory buffer containing the resource.
//
//  Parameters
//
//  1  -  MEMHANDLE		- Handle to the memory buffer containing the resource
//  2  -  int			- ID of the invoking thread
//
//  Returns
//
//  DXRESOURCEHANDLE  -  Handle to the DXResource that will be used to retrieve the data
//						 NULLHANDLE if the resource could not be created
//
//  API Use Notes:
//
//  1. It is the responsibility of the caller to free the returned DXRESOURCEHANDLE by using the dispose() function
//
DXRESOURCEHANDLE DXResourceLoader::loadThisResourceOnDemand(MEMHANDLE hResource, int iThreadID)
{
	return loadResourceOnDemandFromMemory(hResource, iThreadID);	
}

//  loadResourceFromApplication
//
//  Specific entry point taking the name of a resource, the resource will be loaded from the application executable
//
//  Parameters
//
//  1  -  char *  - The resource source ID
//  3  -  int		  - ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE  -  Handle to the Domino Memory allocation containing the loaded resource
//                NULLHANDLE if the resource could not be loaded.
//
MEMHANDLE DXResourceLoader::loadResourceFromApplication(char *szResourceID, int iThreadID)
{
#ifdef WIN32
	HRSRC			hDXRes;							//  Handle to the application resource
	HGLOBAL			hmDXRes;						//  Handle to the resource contents
	MEMHANDLE		hDXResMem;						//  Handle to Domino Memory for the resource contents
	DWORD			dwMemSize;						//  Size of memory for Domino memory request
	DWORD			dwDXResLen;						//  Size of the resource
	char			*pDXRes;						//  Pointer to the loaded resource contents
	char			*pDXResMem;						//  Pointer to the Domino memory buffer for the resource
#endif
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

#ifdef WIN32
	//  Attempt to locate the specified resource in the current application
	hDXRes = FindResource(NULL, szResourceID, APPRSRC_TYPE);
	if (hDXRes == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7005E, szResourceID);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->LogWINAPIError(iThreadID);					//  Show the underlying cause in the log
		return NULLHANDLE;							//  Show that the load request failed
	}

	//  Load the resource into memory
	hmDXRes = LoadResource(NULL, hDXRes);
	if (hmDXRes == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7006E, szResourceID);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->LogWINAPIError(iThreadID);					//  Show the underlying cause in the log
		return NULLHANDLE;							//  Show that the load request failed
	}

	//  Determine the length of the loaded resource source 
	dwDXResLen = SizeofResource(NULL, hDXRes);

	//  Check for a zero length loaded
	if (dwDXResLen == 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7007E, szResourceID);
		xeLocal->LogMessage(szMsg, iThreadID);
		FreeResource(hmDXRes);						//  Free the memory allocation
		return NULLHANDLE;							//  Show that the load request failed
	}

	//  Obtain Domino Memory for the resource source
	dwMemSize = (DWORD) dwDXResLen + 1;						//  Allow space for NULL terminator
	hDXResMem = xeLocal->getDominoMemory(dwMemSize, iThreadID);			//  Allocate the memory buffer
	if (hDXResMem == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7008E, dwMemSize, szResourceID);
		xeLocal->LogMessage(szMsg, iThreadID);
		FreeResource(hmDXRes);						//  Free the memory allocation
		return NULLHANDLE;							//  Show that the load request failed
	}

	//  Lock the native memory 
	pDXRes = (char *) LockResource(hmDXRes);
	//  Lock the Domino memory
	pDXResMem = (char *) OSMemoryLock(hDXResMem);
	//  Copy the resource contents from windows memory to Domino the Domino memory buffer
	memcpy(pDXResMem, pDXRes, dwDXResLen);
	//  Ensure there is a NULL terminator
	pDXResMem[dwDXResLen] = '\0';

	//  Unlock and free the windows memory
	UnlockResource(hmDXRes);
	FreeResource(hmDXRes);

	//  Unlock the Domino memory
	OSMemoryUnlock(hDXResMem);

	//  Show that the resource has been loaded
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7009I, szResourceID, dwDXResLen);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Return the Domino memory buffer handle
	return hDXResMem;
#endif

#ifdef UNIX
	//  Show that the resource has been loaded
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7075E, szResourceID);
	xeLocal->LogMessage(szMsg, iThreadID);

	//  Return with a failed load
	return NULLHANDLE;
#endif

}

//  loadResourceFromFile
//
//  Specific entry point taking the name of a file, the resource will be loaded from the specified file
//
//  Parameters
//
//  1  -  char *  - The resource source File Name
//  2  -  int		  - ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE  -  Handle to the Domino Memory allocation containing the loaded resource
//                NULLHANDLE if the resource could not be loaded.
//
MEMHANDLE DXResourceLoader::loadResourceFromFile(char *szFileName, int iThreadID)
{
	DXResource		*drTemp;						//  Resouce created for the duration of this call
	MEMHANDLE		hReturn;						//  Handle to be returned
	DWORD			dwReadLen;						//  Length Read on a single read
	DWORD			dwTotalRead;					//  Total length read
	char			*pDXResMem;						//  Pointer to the Domino memory buffer for the resource
	char			*pReadPointer;					//  Current address to read data into
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Create a resource to hold the transient information
	drTemp = new DXResource(xeLocal, iThreadID);
	drTemp->pRLoader = this;

	//  Allocate a buffer to retrieve the resource contents into - the buffer is set to the default size
	drTemp->dwMemSize = DEF_FILE_RCV_BFRSIZE;							//  Set default buffer size
	drTemp->hResBuffer = xeLocal->getDominoMemory(drTemp->dwMemSize, iThreadID);			//  Allocate the memory buffer
	if (drTemp->hResBuffer == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7008E, drTemp->dwMemSize, szFileName);
		xeLocal->LogMessage(szMsg, iThreadID);
		delete drTemp;
		return NULLHANDLE;							//  Show that the load request failed
	}

	//  Lock the Domino memory
	pDXResMem = (char *) OSMemoryLock(drTemp->hResBuffer);

	//  Setup to read the resource contents from the passed File
	//  If this is a file type URI the remove the URI prefix
	if (strlen(szFileName) > strlen(FILETYPE_URI))
	{
		if (memcmpi(szFileName, FILETYPE_URI, strlen(FILETYPE_URI)) == 0)
			strcpy_s(drTemp->szCurrentFile, MAXPATH, szFileName + strlen(FILETYPE_URI));
		else strcpy_s(drTemp->szCurrentFile, MAXPATH, szFileName);
	}
	else strcpy_s(drTemp->szCurrentFile, MAXPATH, szFileName);

	dwReadLen = 1;											//  Dummy setting to enter the while loop
	dwTotalRead = 0;
	pReadPointer = pDXResMem;
	while (dwReadLen > 0)
	{
		//  Check if there is sufficient (minimum) space in the buffer, if not then grow the buffer
		if ((drTemp->dwMemSize - dwTotalRead) <= MIN_FREE_BUFFERSPACE)
		{
			//  Grow the Buffer by one increment
			OSMemoryUnlock(drTemp->hResBuffer);				//  Unlock the current buffer
			drTemp->dwMemSize = drTemp->dwMemSize + DEF_FILE_RCV_BFRSIZE;
			drTemp->hResBuffer = xeLocal->resizeDominoMemory(drTemp->hResBuffer, drTemp->dwMemSize, iThreadID);
			if (drTemp->hResBuffer == NULLHANDLE)
			{
				//  Memory reallocation has failed
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7013E, drTemp->dwMemSize, szFileName);
				xeLocal->LogMessage(szMsg, iThreadID);
				OSMemoryFree(drTemp->hResBuffer);
				drTemp->hResBuffer = NULLHANDLE;
				delete drTemp;
				return NULLHANDLE;							//  Show that the load request failed
			}

			//  Lock the Domino memory
			pDXResMem = (char *) OSMemoryLock(drTemp->hResBuffer);
			//  Recalculate the Address to read data into
			pReadPointer = pDXResMem + dwTotalRead;
		}

		//  Read Data from the HTTP stream
		dwReadLen = XMLFILEReader(pReadPointer, drTemp->dwMemSize - dwTotalRead, drTemp, iThreadID);
		dwTotalRead = dwTotalRead + dwReadLen;
		pReadPointer = pReadPointer + dwReadLen;
	}
	*pReadPointer = '\0';
	OSMemoryUnlock(drTemp->hResBuffer);

	//  Check that data was actually read - if not free the buffer
	if (dwTotalRead == 0)
	{
		OSMemoryFree(drTemp->hResBuffer);
		drTemp->hResBuffer = NULLHANDLE;
		delete drTemp;
		return NULLHANDLE;
	}

	//  Resize (shrink) the Domino memory buffer to the size of the resource content
	drTemp->hResBuffer = xeLocal->resizeDominoMemory(drTemp->hResBuffer, dwTotalRead, iThreadID);
	drTemp->dwMemSize = dwTotalRead;

	//  Show that the resource has been loaded
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7009I, szFileName, dwTotalRead);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Return the handle to the memory buffer
	hReturn = drTemp->hResBuffer;
	delete drTemp;
	return hReturn;
}

//  loadResourceFromURL
//
//  Specific entry point taking the URL of a resource, the resource will be loaded from the specified URL
//
//  Parameters
//
//  1  -  char *  - The resource source URL
//  2  -  int		  - ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE -  Handle to the Domino Memory allocation containing the loaded resource
//               NULLHANDLE if the resource could not be loaded.
//
MEMHANDLE DXResourceLoader::loadResourceFromURL(char *szURL, int iThreadID)
{
	DXResource		*drTemp;						//  Resouce created for the duration of this call
	MEMHANDLE		hReturn;						//  Handle to be returned
	DWORD			dwReadLen;						//  Length Read on a single read
	DWORD			dwTotalRead;					//  Total length read
	char			*pDXResMem;						//  Pointer to the Domino memory buffer for the resource
	char			*pReadPointer;					//  Current address to read data into
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Create a resource to hold the transient information
	drTemp = new DXResource(xeLocal, iThreadID);
	drTemp->pRLoader = this;

	//  Allocate a buffer to retrieve the resource contents into - the buffer is set to the default size
	drTemp->dwMemSize = DEF_HTTP_RCV_BFRSIZE;							//  Set default buffer size
	drTemp->hResBuffer = xeLocal->getDominoMemory(drTemp->dwMemSize, iThreadID);			//  Allocate the memory buffer
	if (drTemp->hResBuffer == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7008E, drTemp->dwMemSize, szURL);
		xeLocal->LogMessage(szMsg, iThreadID);
		delete drTemp;
		return NULLHANDLE;							//  Show that the load request failed
	}

	//  Lock the Domino memory
	pDXResMem = (char *) OSMemoryLock(drTemp->hResBuffer);

	//  Setup to read the resource contents from the passed URL
	strcpy_s(drTemp->szCurrentReadURL, MAX_URL, szURL);				//  Set the URL to read the content from
	dwReadLen = 1;											//  Dummy setting to enter the while loop
	dwTotalRead = 0;
	pReadPointer = pDXResMem;
	while (dwReadLen > 0)
	{
		//  Check if there is sufficient (minimum) space in the buffer, if not then grow the buffer
		if ((drTemp->dwMemSize - dwTotalRead) <= MIN_FREE_BUFFERSPACE)
		{
			//  Grow the Buffer by one increment
			OSMemoryUnlock(drTemp->hResBuffer);				//  Unlock the current buffer
			drTemp->dwMemSize = drTemp->dwMemSize + DEF_HTTP_RCV_BFRSIZE;
			drTemp->hResBuffer = xeLocal->resizeDominoMemory(drTemp->hResBuffer, drTemp->dwMemSize, iThreadID);
			if (drTemp->hResBuffer == NULLHANDLE)
			{
				//  Memory reallocation has failed
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7013E, drTemp->dwMemSize, szURL);
				xeLocal->LogMessage(szMsg, iThreadID);
				OSMemoryFree(drTemp->hResBuffer);
				drTemp->hResBuffer = NULLHANDLE;
				delete drTemp;
				return NULLHANDLE;							//  Show that the load request failed
			}

			//  Lock the Domino memory
			pDXResMem = (char *) OSMemoryLock(drTemp->hResBuffer);
			//  Recalculate the Address to read data into
			pReadPointer = pDXResMem + dwTotalRead;
		}

		//  Read Data from the HTTP stream
		dwReadLen = XMLHTTPReader(pReadPointer, drTemp->dwMemSize - dwTotalRead, drTemp, iThreadID);
		dwTotalRead = dwTotalRead + dwReadLen;
		pReadPointer = pReadPointer + dwReadLen;
	}
	*pReadPointer = '\0';
	OSMemoryUnlock(drTemp->hResBuffer);

	//  Check that data was actually read - if not free the buffer
	if (dwTotalRead == 0)
	{
		OSMemoryFree(drTemp->hResBuffer);
		drTemp->hResBuffer = NULLHANDLE;
		delete drTemp;
		return NULLHANDLE;
	}

	//  Resize (shrink) the Domino memory buffer to the size of the resource content
	dwTotalRead++;							//  Allow for the NULL terminator
	drTemp->hResBuffer = xeLocal->resizeDominoMemory(drTemp->hResBuffer, dwTotalRead, iThreadID);
	drTemp->dwMemSize = dwTotalRead;

	//  Show that the resource has been loaded
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7009I, szURL, dwTotalRead);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Return the handle to the memory buffer
	hReturn = drTemp->hResBuffer;
	delete drTemp;
	return hReturn;
}

//  loadResourceOnDemandFromURL
//
//  Specific entry point taking the URL of a resource, the resource will be loaded from the specified URL
//
//  Parameters
//
//  1  -  char *  - The resource source URL
//  2  -  int		  - ID of the invoking thread
//
//  Returns
//
//  DXRESOURCEHANDLE  -  Handle to the DXResource that will be used to retrieve the data
//						 NULLHANDLE if the resource could not be created
//
DXRESOURCEHANDLE DXResourceLoader::loadResourceOnDemandFromURL(char *szURL, int iThreadID)
{
	DXResource		*drNew;									//  Resouce created for the duration of this call
	char			szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Create a resource to hold the transient information
	drNew = new DXResource(xeLocal, iThreadID);
	drNew->pRLoader = this;

	//  Allocate a buffer to retrieve the resource contents into - the buffer is set to the default size
	drNew->dwMemSize = DEF_HTTP_RCV_BFRSIZE;											//  Set default buffer size
	drNew->hResBuffer = xeLocal->getDominoMemory(drNew->dwMemSize, iThreadID);			//  Allocate the memory buffer
	if (drNew->hResBuffer == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7008E, drNew->dwMemSize, szURL);
		xeLocal->LogMessage(szMsg, iThreadID);
		delete drNew;
		return NULLHANDLE;									//  Show that the load request failed
	}

	//  Setup to read the resource contents from the passed URL
	strcpy_s(drNew->szCurrentReadURL, MAX_URL, szURL);					//  Set the URL to read the content from
	drNew->iRMCLass = RMCLASS_PRIMARY | RMCLASS_HTTPLOD;
	drNew->iRMState = RMSTATE_INIT;

	//  Show that the resource has been prepared
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7054I, szURL);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Return the handle to the DXResource
	return drNew;
}

//  loadResourceOnDemandFromFile
//
//  Specific entry point taking a file as a resource, the resource will be loaded from the specified file.
//
//  Parameters
//
//  1  -  char *		- The resource filename
//  2  -  int			- ID of the invoking thread
//
//  Returns
//
//  DXRESOURCEHANDLE  -  Handle to the DXResource that will be used to retrieve the data
//						 NULLHANDLE if the resource could not be created
//
DXRESOURCEHANDLE DXResourceLoader::loadResourceOnDemandFromFile(char *szFileName, int iThreadID)
{
	DXResource		*drNew;							//  Resouce created for the duration of this call
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Create a resource to hold the transient information
	drNew = new DXResource(xeLocal, iThreadID);
	drNew->pRLoader = this;
	drNew->dwMemSize = 0;

	//  Setup to read the resource contents from the passed URL
	//  If this is a file type URI the remove the URI prefix
	if (strlen(szFileName) > strlen(FILETYPE_URI))
	{
		if (memcmpi(szFileName, FILETYPE_URI, strlen(FILETYPE_URI)) == 0)
			strcpy_s(drNew->szCurrentFile, MAXPATH, szFileName + strlen(FILETYPE_URI));
		else strcpy_s(drNew->szCurrentFile, MAXPATH, szFileName);
	}
	else strcpy_s(drNew->szCurrentFile, MAXPATH, szFileName);
	drNew->iRMCLass = RMCLASS_PRIMARY | RMCLASS_FILELOD;
	drNew->iRMState = RMSTATE_INIT;

	//  Show that the resource has been prepared
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7054I, szFileName);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Return the handle to the DXResource
	return drNew;
}

//  loadResourceOnDemandFromMemory
//
//  Specific entry point taking a memory handle to an already loaded resource and wrapping it in a
//  DXResource object to allow it to be used On Demand.
//
//  Parameters
//
//  1  -  MEMHANDLE		- A memory handle to the loaded resource
//  2  -  int			- ID of the invoking thread
//
//  Returns
//
//  DXRESOURCEHANDLE  -  Handle to the DXResource that will be used to retrieve the data
//						 NULLHANDLE if the resource could not be created
//
DXRESOURCEHANDLE DXResourceLoader::loadResourceOnDemandFromMemory(MEMHANDLE hResMem, int iThreadID)
{
	DXResource		*drNew;						//  Resouce created for the duration of this call
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Safety check - make sure that we are not creating a resource for a NULLHANDLE
	if (hResMem == NULLHANDLE)
	{
		xeLocal->LogMessage(MSG_DXR7055E, iThreadID);
		return NULLHANDLE;
	}

	//  Create a resource to hold the transient information
	drNew = new DXResource(xeLocal, iThreadID);
	drNew->pRLoader = this;
	drNew->dwMemSize = OSMemoryGetSize(hResMem);
	drNew->hResBuffer = hResMem;
	drNew->iRMCLass = RMCLASS_PRIMARY | RMCLASS_MEMLOD;
	drNew->iRMState = RMSTATE_INIT;

	//  Show that the resource has been prepared
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7056I, hResMem);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Return the handle to the DXResource
	return drNew;
}

//  loadResourceFromText
//
//  Specific entry point taking the Blockid of a text list item, the resource will be loaded from the item
//
//  Parameters
//
//  1  -  BLOCKID	- The Blockid of the item value
//  2  -  DWORD		- The length of the text list
//  3  -  char *	- Pointer to the null terminated Item Name 
//  4  -  int		  - ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE -  Handle to the Domino Memory allocation containing the loaded resource
//               NULLHANDLE if the resource could not be loaded.
//
MEMHANDLE DXResourceLoader::loadResourceFromText(BLOCKID bidVal, DWORD dwLen, char *szItem, int iThreadID)
{
	MEMHANDLE		hDXResMem;						//  Handle of the resource memory buffer
	DWORD			dwMemSize;						//  Size of the memory buffer allocated for the resource
	DWORD			dwLoadSize;						//  Size of string data loaded to the memory buffer
	int				iEntries;						//  Count of entries
	int				iIndex;							//  Index into the entry arrays
	char			*pDXResMem;						//  Pointer to the allocated memory buffer
	unsigned char	*pItemValue;					//  Pointer to the item value buffer
	LIST			*pListHeader;					//  Pointer to the list header
	USHORT			*pEntryLen;						//  Pointer to the array of string lengths
	char			*pEntryValue;					//  Pointer into the packed array of string data
	char			*pMemBuffer;					//  Pointer into the memory buffer
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Allocate a Domino memory buffer for the resource contents
	dwMemSize = (DWORD) dwLen + 1;					//  Allow space for NULL terminator
	hDXResMem = xeLocal->getDominoMemory(dwMemSize, iThreadID);			//  Allocate the memory buffer
	if (hDXResMem == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7043E, dwMemSize, szItem);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;							//  Show that the load request failed
	}

	//  Lock the Domino memory
	pDXResMem = (char *) OSMemoryLock(hDXResMem);

	//  Lock the Item Value Buffer
	pItemValue = OSLockBlock(unsigned char, bidVal); 
	
	//  Get the count of entries 
	pListHeader = (LIST *) (pItemValue + sizeof(WORD));
	iEntries = pListHeader->ListEntries;

	//  Trace the start of building the buffer
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7044T, szItem, iEntries, dwLen);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	//  Initialise pointers ready for doing the data transfer
	//  The value buffer layout for a TYPE_TEXT_LIST Entry is
	//
	//   WORD  - Data Type
	//   LIST  - Count of entries
	//   USHORT[] - array of USHORTS with the length of each string entry
	//   char  - packed strings buffer
	//
	pEntryLen = (USHORT *) (pItemValue + sizeof(WORD) + sizeof(LIST));
	pEntryValue = (char *) (pItemValue + sizeof(WORD) + sizeof(LIST) + (sizeof(USHORT) * iEntries));
	pMemBuffer = pDXResMem;
	dwLoadSize = 0;

	//  Transfer the text lines from the list to the buffer - append a CR/LF sequence to each line
	for (iIndex = 0; iIndex < iEntries; iIndex++)
	{
		memcpy(pMemBuffer, pEntryValue, pEntryLen[iIndex]);
		pEntryValue = pEntryValue + pEntryLen[iIndex];			//  Point to the next packed string
		pMemBuffer = pMemBuffer + pEntryLen[iIndex];			//  Point to the end of used space in the memory buffer
		pMemBuffer[0] = 0x0d;									//  Insert CR
		pMemBuffer[1] = 0x0a;									//  Insert LF
		pMemBuffer++;
		pMemBuffer++;
		dwLoadSize = dwLoadSize + pEntryLen[iIndex] + 2;		//  Increment the count of bytes loaded 
	}
	
	//  Ensure that there is a NULL Terminator
	pMemBuffer[0] = '\0';

	//  Trace the completion of building the buffer
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7045T, szItem, iEntries, dwLoadSize);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	//  Unlock the Item Value Buffer
	OSUnlockBlock(bidVal);

	//  Unlock the memory buffer
	OSMemoryUnlock(hDXResMem);

	//  Show that the resource has been loaded
	if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7009I, szItem, dwLoadSize);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Return the handle to the memory buffer
	return hDXResMem;
}

//  loadResourceFromRTI
//
//  Specific entry point taking the Blockid of a rich text item, the resource will be loaded from the item
//
//  Parameters
//
//  1  -  BLOCKID	- The Blockid of the item value
//  2  -  DWORD		- The length of the composite buffer
//  3  -  char *	- Pointer to the null terminated Item Name 
//  4  -  int		  - ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE -  Handle to the Domino Memory allocation containing the loaded resource
//               NULLHANDLE if the resource could not be loaded.
//
MEMHANDLE DXResourceLoader::loadResourceFromRTI(BLOCKID bidVal, DWORD dwLen, char *szItem, int iThreadID)
{
	STATUS			stAPIRC;						//  Return code from Notes API calls
	DXResource		*drTemp;						//  Transient DXresource
	MEMHANDLE		hReturn;						//  Handle to be returned to the caller
	DWORD			dwLoadSize;						//  Size of string data loaded to the memory buffer
	char			*pDXResMem;						//  Pointer to the allocated memory buffer
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Create a resource to hold the transient information
	drTemp = new DXResource(xeLocal, iThreadID);
	drTemp->pRLoader = this;

	//  Allocate a Domino memory buffer for the resource contents
	drTemp->dwMemSize = (DWORD) dwLen + 1;					//  Allow space for NULL terminator
	drTemp->hResBuffer = xeLocal->getDominoMemory(drTemp->dwMemSize, iThreadID);			//  Allocate the memory buffer
	if (drTemp->hResBuffer == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7043E, drTemp->dwMemSize, szItem);
		xeLocal->LogMessage(szMsg, iThreadID);
		delete drTemp;
		return NULLHANDLE;							//  Show that the load request failed
	}

	//  Lock the Domino memory
	pDXResMem = (char *) OSMemoryLock(drTemp->hResBuffer);

	//  Setup the context variables that will be used by the enumerators action routine
	drTemp->pEnumMemBuffer = pDXResMem;		
	drTemp->dwEnumBufferUsed = 0;
	
	//  Trace starting of the enumeration of the rich text (composite) buffer
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7046T, szItem, dwLen);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	//  Enumerate the Rich Text Item composite buffer, all of the real work is performed in the enumerator action routine
	stAPIRC = EnumCompositeBuffer(bidVal, dwLen, WrapperprocessThisCDRec, drTemp);
	if (ERR(stAPIRC) != NOERROR)
	{
		if (ERR(stAPIRC) != ERR_ODS_ENUM_COMPLETE)
		{
			//  Error during the enumeration issue the message as signal the resource not loaded
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7048E, szItem);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Free the allocated memory buffer
			OSMemoryUnlock(drTemp->hResBuffer);
			OSMemoryFree(drTemp->hResBuffer);
			drTemp->hResBuffer = NULLHANDLE;
			delete drTemp;
			return NULLHANDLE;
		}
	}

	//  Trace completion of the enumeration of the rich text (composite) buffer
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7047T, szItem, drTemp->dwEnumBufferUsed);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	//  Ensure a valid NULL termination
	pDXResMem[drTemp->dwEnumBufferUsed] = '\0';

	//  Unlock the memory buffer
	OSMemoryUnlock(drTemp->hResBuffer);

	//  Resize the allocated memory buffer before returning it
	dwLoadSize = drTemp->dwEnumBufferUsed + 1;
	drTemp->hResBuffer = xeLocal->resizeDominoMemory(drTemp->hResBuffer, dwLoadSize, iThreadID);

	//  Show that the resource has been loaded
	if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7009I, szItem, dwLoadSize);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Return the handle to the memory buffer
	hReturn = drTemp->hResBuffer;
	delete drTemp;
	return hReturn;
}

//  processThisCDRec
//
//  Function to process the current CD Record during enumeration.
//  If the record type contains text then it will extract the text content and append it to the memory buffer.
//
//  Parameters
//
//  1  -  CHAR *	- Pointer to the CD Record returned by the Enumerator
//  2  -  WORD		- The CD Record Type
//  3  -  DWORD		- The CD Record Length
//
//  Returns
//
//  STATUS	-	   Used to signal to the enumerator to continue (NOERROR)
//
STATUS DXResourceLoader::processThisCDRec(char *pCDRec, WORD wCDType, DWORD dwCDLen, DXResource *drRead)
{
	CDTEXT		*pCDText;						//  Pointer to a TEXT CD Record

	//  Process according to the type of CD record passed
	switch(wCDType)
	{
	case SIG_CD_TEXT:
		//  Move the text into the allocated memory buffer and adjust the pointers
		pCDText = (CDTEXT *) pCDRec;
		memcpy(drRead->pEnumMemBuffer, pCDRec + sizeof(CDTEXT), pCDText->Header.Length - sizeof(CDTEXT));
		drRead->pEnumMemBuffer = drRead->pEnumMemBuffer + (pCDText->Header.Length - sizeof(CDTEXT));
		drRead->dwEnumBufferUsed = drRead->dwEnumBufferUsed + (pCDText->Header.Length - sizeof(CDTEXT));
		break;
	case SIG_CD_PARAGRAPH:
		//  Insert a CR/LF sequence into the allocated memory buffer and adjust pointers
		drRead->pEnumMemBuffer[0] = 0x0d;									//  Insert CR
		drRead->pEnumMemBuffer[1] = 0x0a;									//  Insert LF
		drRead->pEnumMemBuffer = drRead->pEnumMemBuffer + 2;
		drRead->dwEnumBufferUsed = drRead->dwEnumBufferUsed + 2;
		break;
	}
	//  All other types of record are ignored
	return NOERROR;
}

#ifdef WIN32
//  XMLHTTPReader   (WinHTTP implementation)
//
//  Function to provide an in memory XML stream from an HTTP source, which can be consumed in-line or not
//
//  Parameters
//
//  1  -  char *		- Pointer to the buffer to use
//  2  -  DWORD			- Size in bytes of the freespace in the buffer
//  3  -  DXResource *	- Pointer to the DXResource object that contains the resource binfo
//  4  -  int			- ID of the invoking thread
//
//  Returns
//
//  DWORD	-  Number of bytes of data now available in the buffer
//
DWORD DXResourceLoader::XMLHTTPReader(char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID)
{
	URL_COMPONENTS			urlcCurrent;				//  URL Components derived from the current request
	char					szMsg[MAX_MSG + 1];			//  Generic message buffer
	wchar_t					wcFullURL[MAX_URL + 1];		//  Full URL
	wchar_t					wcHostName[MAX_URL + 1];	//  Host Name parsed from the request URL
	wchar_t					wcURLPath[MAX_URL + 1];		//  Path component of the request URL
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
	size_t					iConverted;					//  Count of characters converted
#endif
	char					szTempStr[MAX_URL + 1];		//  Temp string for conversions
	DWORD					dwRespCode;					//  Response code from HTTP Request
	DWORD					dwRetLen;					//  Length of responses
	DWORD					dwErrorCode;				//  Error Code from WinHTTP Function calls
	char					*cCache;					//  Cache buffer for returned data

	//  Check if the WinHTTP component is loaded - if not load it
	if (hWHFacility == NULLHANDLE)
	{
		//  Setup the connection configuration
		setConnectConfig(drRead);
		if (drRead->iCurrentConnectConfig == HTTPP_DIRECT)
		{
			//  Create the facility using direct connection setup
			hWHFacility = WinHttpOpen(L"DX Resource Loader 2.0", 
									WINHTTP_ACCESS_TYPE_NO_PROXY, 
									WINHTTP_NO_PROXY_NAME,
									WINHTTP_NO_PROXY_BYPASS,
									0);
		}
		else
		{
			//  Create the facilitty using the default WinHTTP Proxy settings
			//  NOTE: These are not the IE browser proxy settings you must use the WinHTTP Proxycfg.exe configurator
			if (drRead->iCurrentConnectConfig == HTTPP_WPROXY)
			{
				hWHFacility = WinHttpOpen(L"DX Resource Loader 2.0", 
										WINHTTP_ACCESS_TYPE_DEFAULT_PROXY, 
										WINHTTP_NO_PROXY_NAME,
										WINHTTP_NO_PROXY_BYPASS,
										0);
			}
		}
		if (hWHFacility == NULLHANDLE)
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			//  Log the error and return 0 bytes read
			xeLocal->LogMessage(MSG_DXR7014E, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
		}

		//  Trace the creation of WinHTTP
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7016T, drRead->iCurrentConnectConfig);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);
	}

	//  Check if we have already made the connection to the source - if not open the connection
	if (drRead->hWHConnection == NULLHANDLE)
	{
		ZeroMemory(&urlcCurrent, sizeof(URL_COMPONENTS));
		urlcCurrent.dwStructSize = sizeof(URL_COMPONENTS);
		urlcCurrent.dwHostNameLength = sizeof(wcHostName);					//  Flag that Host Name needed
		urlcCurrent.lpszHostName = wcHostName;								//  Set the buffer

		//  Use the URL cracker to extract the server name from the requested URL
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		mbstowcs_s(&iConverted, wcFullURL, strlen(drRead->szCurrentReadURL) + 1, drRead->szCurrentReadURL, _TRUNCATE);
#else
		mbstowcs(wcFullURL, drRead->szCurrentReadURL, MAX_URL);
#endif
		if (!WinHttpCrackUrl(wcFullURL, 0, 0, &urlcCurrent))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			//  Crack failed the URL is probably invalid
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7015E, drRead->szCurrentReadURL);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
		}

		//  Trace the connection attempt
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		wcstombs_s(&iConverted, szTempStr, MAX_URL, wcHostName, _TRUNCATE);
#else
		wcstombs(szTempStr, wcHostName, MAX_URL);
#endif
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7017T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Build the connection
		drRead->hWHConnection = WinHttpConnect(hWHFacility, wcHostName, INTERNET_DEFAULT_HTTP_PORT, 0);
		if (drRead->hWHConnection == NULLHANDLE)
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7018E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
		}

		//  Trace the connection
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7019T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);
	}

	//  Check if we have already issued the request - if not issue it
	if (drRead->hWHRequest == NULLHANDLE)
	{
		ZeroMemory(&urlcCurrent, sizeof(URL_COMPONENTS));
		urlcCurrent.dwStructSize = sizeof(URL_COMPONENTS);
		urlcCurrent.dwUrlPathLength = sizeof(wcURLPath);		
		urlcCurrent.lpszUrlPath = wcURLPath;

		// Use the URL cracker to extract the URL path from the requested URL
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		mbstowcs_s(&iConverted, wcFullURL, strlen(drRead->szCurrentReadURL) + 1, drRead->szCurrentReadURL, _TRUNCATE);
#else
		mbstowcs(wcFullURL, drRead->szCurrentReadURL, MAX_URL);
#endif
		if (!WinHttpCrackUrl(wcFullURL, 0, 0, &urlcCurrent))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			//  Crack failed the URL is probably invalid
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7020E, drRead->szCurrentReadURL);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		wcstombs_s(&iConverted, szTempStr, MAX_URL, wcURLPath, _TRUNCATE);
#else
		wcstombs(szTempStr, wcURLPath, MAX_URL);
#endif
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7021T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Open the requesst
		drRead->hWHRequest = WinHttpOpenRequest(drRead->hWHConnection,
										L"GET",
										wcURLPath,
										NULL,
										WINHTTP_NO_REFERER,
										WINHTTP_DEFAULT_ACCEPT_TYPES,
										WINHTTP_FLAG_ESCAPE_PERCENT);
		if (drRead->hWHRequest == NULLHANDLE)
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7022E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Trace the request open
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7024T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Add header
		if (!WinHttpAddRequestHeaders(	drRead->hWHRequest, 
										L"Keep-Alive: close",
                                         (ULONG)-1L,
                                         WINHTTP_ADDREQ_FLAG_ADD))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7059E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			// It may be that this request may be satisfied by re-issuing it using a different connection configuration
			return retryXMLReader(cBuffer, dwMaxToRead, dwErrorCode, drRead, iThreadID);
		}

		//  Send the request
		if (!WinHttpSendRequest(drRead->hWHRequest,
								WINHTTP_NO_ADDITIONAL_HEADERS,
								0,
								WINHTTP_NO_REQUEST_DATA,
								0,
								0,
								0))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7023E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			// It may be that this request may be satisfied by re-issuing it using a different connection configuration
			return retryXMLReader(cBuffer, dwMaxToRead, dwErrorCode, drRead, iThreadID);
		}

		//  Trace the request send
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7025T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Retrieve the request results
		if (!WinHttpReceiveResponse(drRead->hWHRequest, NULL))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7026E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Trace Response detected
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7027T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Check the response code
		dwRetLen = sizeof(DWORD);
		if (!WinHttpQueryHeaders(drRead->hWHRequest, 
								WINHTTP_QUERY_STATUS_CODE +	WINHTTP_QUERY_FLAG_NUMBER, 
								WINHTTP_HEADER_NAME_BY_INDEX,
								&dwRespCode,
								&dwRetLen,
								WINHTTP_NO_HEADER_INDEX))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7028E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Trace the response code received for the request
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7029T, dwRespCode, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Determine if the response code is ok
		if (dwRespCode != HTTP_STATUS_OK)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7030E, dwRespCode, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}
	}

	//  Response data is available
	
	//  Determine if there is residual data left from the previous call to XMLReader
	//  If there is then return that data
	if (drRead->hRespBuffer != NULLHANDLE)
	{
		//  Lock the buffer
		cCache = (char *) OSMemoryLock(drRead->hRespBuffer);
		if (dwMaxToRead >= drRead->dwRespBfrSize)
		{
			//  There is room to store all of the cached data in the XML Reader Buffer
			//  Trace the size of data being returned
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7031T, drRead->dwRespBfrSize);
			xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

			//  Return the data that is available and free the overflow cache
			memcpy(cBuffer, cCache, drRead->dwRespBfrSize);
			dwRetLen = drRead->dwRespBfrSize;
			OSMemoryUnlock(drRead->hRespBuffer);
			OSMemoryFree(drRead->hRespBuffer);
			drRead->hRespBuffer = NULLHANDLE;
			drRead->dwRespBfrSize = 0;
			return dwRetLen;
		}
		else
		{	
			//  Only a part of the cached data can be returned to the XML processor
			//  Trace the partial return of cached data
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7032T, dwMaxToRead, drRead->dwRespBfrSize);
			xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

			//  Return the data 
			memcpy(cBuffer, cCache, dwMaxToRead);
			dwRetLen = drRead->dwRespBfrSize;

			//  Adjust the buffer and resize it
			memcpy(cCache, cCache + dwMaxToRead, drRead->dwRespBfrSize - dwMaxToRead);
			OSMemoryUnlock(drRead->hRespBuffer);
			drRead->hRespBuffer = xeLocal->resizeDominoMemory(drRead->hRespBuffer, drRead->dwRespBfrSize - dwMaxToRead, iThreadID);
			drRead->dwRespBfrSize = drRead->dwRespBfrSize - dwMaxToRead;
			return dwRetLen;
		}
	}

	//  Move returned data to the buffer - if there is more data than buffer space then cache it

	//  Determine the size of data that is available 
	if (!WinHttpQueryDataAvailable(drRead->hWHRequest, &dwRetLen))
	{
		dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
		xeLocal->LogMessage(MSG_DXR7033E, iThreadID);
		xeLocal->GetWINAPIError(dwErrorCode, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		//  Close the request
		WinHttpCloseHandle(drRead->hWHRequest);
		drRead->hWHRequest = NULLHANDLE;
		//  Close the connection
		WinHttpCloseHandle(drRead->hWHConnection);
		drRead->hWHConnection = NULLHANDLE;
		return 0;
	}

	//  If the data available is greater than the buffer passed by the requester - allocate an overflow cache
	if (dwRetLen > dwMaxToRead)
	{
		drRead->hRespBuffer = xeLocal->getDominoMemory(dwRetLen, iThreadID);
		if (drRead->hRespBuffer == NULLHANDLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7034E, dwRetLen);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Lock the buffer
		cCache = (char *) OSMemoryLock(drRead->hRespBuffer);

		//  Read the data into the cache buffer
		drRead->dwRespBfrSize = dwRetLen;
		if (!WinHttpReadData(drRead->hWHRequest, cCache, drRead->dwRespBfrSize, &dwRetLen))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7035E, dwRetLen);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Free the overflow cache buffer
			OSMemoryUnlock(drRead->hRespBuffer);
			OSMemoryFree(drRead->hRespBuffer);
			drRead->hRespBuffer = NULLHANDLE;
			drRead->dwRespBfrSize = 0;
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Trace the fact that response data has been transferred to the overflow cache buffer
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7036T, dwRetLen);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Transfer the max data possible into the requesters buffer and adjust the cache

		//  Trace the partial return of cached data
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7032T, dwMaxToRead, drRead->dwRespBfrSize);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Return the data to the callers buffer
		memcpy(cBuffer, cCache, dwMaxToRead);

		//  Adjust and resize the overflow cache
		dwRetLen = drRead->dwRespBfrSize;
		OSMemoryUnlock(drRead->hRespBuffer);
		drRead->hRespBuffer = xeLocal->resizeDominoMemory(drRead->hRespBuffer, drRead->dwRespBfrSize - dwMaxToRead, iThreadID);
		drRead->dwRespBfrSize = drRead->dwRespBfrSize - dwMaxToRead;
		return dwRetLen;
	}
	else
	{
		//  Check if there is now zero bytes available - this signals the end of the stream
		if (dwRetLen == 0)
		{
			//  Trace the normal end of stream encountered
			xeLocal->LogTrace(TRACE_RESOURCELOADER, MSG_DXR7037T, iThreadID);

			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Return all of the available data - read it directly into the callers buffer
		if (!WinHttpReadData(drRead->hWHRequest, cBuffer, dwMaxToRead, &dwRetLen))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7038E, dwRetLen);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Trace the return of data
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7039T, dwRetLen);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);
		//  Return the number of bytes available
		return dwRetLen;
	}
}
#endif
#ifdef UNIX
//  XMLHTTPReader   (CURL implementation)
//
//  Function to provide an in memory XML stream from an HTTP source, which can be consumed in-line or not
//
//  Parameters
//
//  1  -  char *		- Pointer to the buffer to use
//  2  -  DWORD			- Size in bytes of the freespace in the buffer
//  3  -  DXResource *	- Pointer to the DXResource object that contains the resource binfo
//  4  -  int			- ID of the invoking thread
//
//  Returns
//
//  DWORD	-  Number of bytes of data now available in the buffer
//
DWORD DXResourceLoader::XMLHTTPReader(char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID)
{
	DWORD			dwRetLen;									//  Length of data to return
	long			lHTTPResponse;								//  HTTP Response Code
	char			*cCache;									//  Cache buffer for returned data
	CURLcode		ccResponse;									//  CURL Error Code
	char			szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Check if the CURL component is loaded - if not load it
	if (hcFacility == NULLHANDLE)
	{
		//  Perform the global initialisation
		curl_global_init(CURL_GLOBAL_ALL);
		//  Initiate the CURL session
		hcFacility = curl_easy_init();

		//  Make sure that the DXResource is correctly set up for the initial loading of the content
		drRead->hRespBuffer = NULLHANDLE;
		drRead->dwRespBfrSize = 0;
		drRead->iOwningThread = iThreadID;
	}

	//  The current state is detrmined by the state of the resource
	//  If we have not yet loaded the content then do so now
	if (drRead->hRespBuffer == NULLHANDLE)
	{
		//  Set the URL to retrieve
		curl_easy_setopt(hcFacility, CURLOPT_URL, drRead->szCurrentReadURL);

		//  Set the callback to use for loading the content
		curl_easy_setopt(hcFacility, CURLOPT_WRITEFUNCTION, WrapperCURLHTTPReader);

		//  Set the data object to be passed to the callback
		curl_easy_setopt(hcFacility, CURLOPT_WRITEDATA, drRead);

		//  Set the User Agent identity
		curl_easy_setopt(hcFacility, CURLOPT_USERAGENT, "DX Resource Loader 2.0");

		//  Load the content
		ccResponse = curl_easy_perform(hcFacility);
		if (ccResponse != CURLE_OK)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7023E, drRead->szCurrentReadURL);
			xeLocal->LogMessage(szMsg, iThreadID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7076E, ccResponse, curl_easy_strerror(ccResponse));
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;							//  Signal end of stream to the caller
		}

		//  Check the HTTP response code
		curl_easy_getinfo(hcFacility, CURLINFO_RESPONSE_CODE, &lHTTPResponse);
		if (lHTTPResponse != 200)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7030E, lHTTPResponse, drRead->szCurrentReadURL); 
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;							//  Signal end of stream to the caller
		}

		//  Check the resource to see if the load succeeded
		if ((drRead->hRespBuffer == NULLHANDLE) || (drRead->dwRespBfrSize == 0))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7030E, 0, drRead->szCurrentReadURL); 
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;							//  Signal end of stream to the caller
		}

		//  Update the buffer size to the actual size read
		drRead->dwRespBfrSize = drRead->dwMemOffset;

		//  Show that resource is loaded
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7077T, drRead->szCurrentReadURL, drRead->dwRespBfrSize);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Set the current read offset in the buffer
		drRead->dwMemOffset = 0;
	}

	//  Return data to the caller buffer from the resource response buffer
	if (drRead->dwMemOffset < drRead->dwRespBfrSize)
	{
		//  Compute how many bytes will be returned
		dwRetLen = drRead->dwRespBfrSize - drRead->dwMemOffset;
		if (dwRetLen > dwMaxToRead) dwRetLen = dwMaxToRead;

		//  Lock the buffer
		cCache = (char *) OSMemoryLock(drRead->hRespBuffer);

		//  Trace the partial return of cached data
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7032T, dwRetLen, drRead->dwRespBfrSize);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Return the data to the callers buffer
		memcpy(cBuffer, cCache, dwRetLen);

		//  Adjust the stream read offset
		drRead->dwMemOffset += dwRetLen;

		//  Signal the return of data
		return dwRetLen;
	}

	//  Show nothing is left to read -- i.e. END-OF-STREAM
	return 0;
}

//  WrapperCURLHTTPReader
//
//  Callback from CURL to load returned data into local memory
//
//  Parameters
//
//  1  -  void *			- Pointer to the buffer to use
//  2  -  size_t			- Size in bytes of the Buffer data
//  3  -  size_t			- Number of Buffers
//  4  -  void *			- Pointer to the DXResource object that contains the resource binfo
//
//  Returns
//
//  size_t	-  Number of bytes of data added to the local memory
//
size_t DXResourceLoader::WrapperCURLHTTPReader(void *pBuffer, size_t iBfrSize, size_t iBuffers, void *pContext)
{
	size_t				iTotalBytes;					//  Total size of the data buffer made available
	DWORD				dwBytesAvailable;				//  Bytes Available
	char				*cSource;						//  Pointer to the source buffer
	DXResource			*drCurrent;						//  Passed DXResource
	DXResourceLoader	*rlCurrent;						//  Resource Loader

	//  Compute te total bytes available
	iTotalBytes = iBfrSize * iBuffers;

	//  Cast the context to the passed DXResource
	drCurrent = (DXResource *) pContext;
	rlCurrent = (DXResourceLoader *) drCurrent->pRLoader;

	//  Cast/Assign the parameters
	cSource = (char *) pBuffer;
	dwBytesAvailable = iTotalBytes;

	//  Load the returned data to the response buffer
	iTotalBytes = rlCurrent->CURLHTTPReader(cSource, dwBytesAvailable, drCurrent, drCurrent->iOwningThread);

	return iTotalBytes;
}

//  CURLHTTPReader
//
//  Callback from CURL to load returned data into local memory
//
//  Parameters
//
//  1  -  char *			- Pointer to the buffer to use
//  2  -  DWORD				- Size in bytes of the Buffer data
//  3  -  DXResource *		- Pointer to the DXResource object that contains the resource binfo
//	4  -  int				- ID of the calling thread
//
//  Returns
//
//  size_t	-  Number of bytes of data added to the local memory
//
size_t DXResourceLoader::CURLHTTPReader(char *cBuffer, DWORD dwBytesAvailable, DXResource *drRead, int iThreadID)
{
	char			*cCache;									//  Pointer to the cache buffer
	DWORD			dwBytesToAllocate;							//  Size of memory to allocate
	char			szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  If the buffer has not yet been allocated in the resource then allocated to
	//  the greater of the passed data size or the default response buffer size
	if (drRead->hRespBuffer == NULLHANDLE)
	{
		dwBytesToAllocate = DEF_HTTP_RCV_BFRSIZE;
		if (dwBytesAvailable > dwBytesToAllocate) dwBytesToAllocate = dwBytesAvailable;

		//  Allocate the buffer to hold the returned data
		drRead->hRespBuffer = xeLocal->getDominoMemory(dwBytesToAllocate, iThreadID);
		if (drRead->hRespBuffer == NULLHANDLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7034E, dwBytesToAllocate);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
		}

		//  Lock the buffer
		cCache = (char *) OSMemoryLock(drRead->hRespBuffer);

		//  Read the data into the cache buffer
		drRead->dwRespBfrSize = dwBytesToAllocate;

		//  Trace the fact that response data has been transferred to the overflow cache buffer
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7036T, dwBytesAvailable);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Move the data to the cache buffer
		memcpy(cCache, cBuffer, dwBytesAvailable);

		//  Unlock the buffer
		OSMemoryUnlock(drRead->hRespBuffer);

		//  Set the offset
		drRead->dwMemOffset = dwBytesAvailable;

		//  return to the caller
		return dwBytesAvailable;
	}

	//  The buffer is already allocated - determine if there is enough free space to accomodate
	//  the returned data.
	dwBytesToAllocate = drRead->dwRespBfrSize - drRead->dwMemOffset;
	if (dwBytesAvailable > dwBytesToAllocate)
	{
		//  Reallocate the buffer to accomodate the additonal data
		dwBytesToAllocate = drRead->dwMemOffset + dwBytesAvailable;
		drRead->hRespBuffer = xeLocal->resizeDominoMemory(drRead->hRespBuffer, dwBytesToAllocate, iThreadID);
		drRead->dwRespBfrSize = dwBytesToAllocate;
	}

	//  Trace the fact that response data has been transferred to the overflow cache buffer
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7036T, dwBytesAvailable);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	//  Lock the buffer and copy in the datra
	cCache = (char *) OSMemoryLock(drRead->hRespBuffer);
	cCache += drRead->dwMemOffset;

	//  Move the data to the cache buffer
	memcpy(cCache, cBuffer, dwBytesAvailable);

	//  Unlock the buffer
	OSMemoryUnlock(drRead->hRespBuffer);

	//  Update the bytes read
	drRead->dwMemOffset += dwBytesAvailable;

	//  Return to caller
	return dwBytesAvailable;
}

#endif
//  XMLMEMReader
//
//  Function to provide an in memory XML stream from a in-memory source, which can be consumed in-line or not
//
//  Parameters
//
//  1  -  char *		- Pointer to the buffer to use
//  2  -  DWORD			- Size in bytes of the freespace in the buffer
//  3  -  DXResource *	- Pointer to the DXResource object that contains the resource binfo
//  4  -  int			- ID of the invoking thread
//
//  Returns
//
//  DWORD	-  Number of bytes of data now available in the buffer
//
DWORD DXResourceLoader::XMLMEMReader(char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID)
{
	char			*pSource;									//  Pointer to the Source point in the buffer
	DWORD			dwBytesToMove;								//  Number of bytes to move
	
	//  Test for the boundary condition where the source buffer is exhausted
	if (drRead->dwMemSize - drRead->dwMemOffset == 0) return 0;

	//  Setup to move data from the source buffer to the interface
	pSource = (char *) OSMemoryLock(drRead->hResBuffer);
	pSource = pSource + drRead->dwMemOffset;
	dwBytesToMove = drRead->dwMemSize - drRead->dwMemOffset;
	if (dwBytesToMove > dwMaxToRead) dwBytesToMove = dwMaxToRead;

	//  Move the next quantum of data to the callers buffer
	memcpy(cBuffer, pSource, dwBytesToMove);

	//  Unlock the source buffer
	OSMemoryUnlock(drRead->hResBuffer);

	//  Update the current position in the buffer
	drRead->dwMemOffset = drRead->dwMemOffset + dwBytesToMove;

	//  Return the size of data passe to the interface
	return dwBytesToMove;
}

//  XMLFILEReader
//
//  Function to provide an in memory XML stream from a file source, which can be consumed in-line or not
//
//  Parameters
//
//  1  -  char *		- Pointer to the buffer to use
//  2  -  DWORD			- Size in bytes of the freespace in the buffer
//  3  -  DXResource *	- Pointer to the DXResource object that contains the resource binfo
//  4  -  int			- ID of the invoking thread
//
//  Returns
//
//  DWORD	-  Number of bytes of data now available in the buffer
//
DWORD DXResourceLoader::XMLFILEReader(char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID)
{
#ifdef WIN32
	errno_t			errorCode;								//  Error code from library functions
	SIZE_T			iDXResLen;								//  Length of the available buffer
	SIZE_T			iDXRead;								//  Length read
#endif
#ifdef UNIX
	int				errorCode;								//  Error code from library functions
	int				iDXResLen;								//  Length of the available buffer
	int				iDXRead;								//  Length read
#endif
	char			szFError[MAX_MSG + 1];					//  File I/O error reason
	char			szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  if the file is not yet open then open it
	if (drRead->pFileDesc == NULL)
	{
		//  Attempt to open the file
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		errorCode = fopen_s(&drRead->pFileDesc, drRead->szCurrentFile, "rb");				//  Open the file for read only
		if (errorCode != EOK)
		{
			strerror_s(szFError, MAX_MSG, errorCode);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7010E, drRead->szCurrentFile, szFError);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
		}
#else
		drRead->pFileDesc = fopen(drRead->szCurrentFile, "rb");
		if (drRead->pFileDesc == NULL)
		{
#ifdef WIN32
			errorCode = GetLastError();
			xeLocal->GetWINAPIError(errorCode, szMsg);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7010E, drRead->szCurrentFile, szFError);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
#endif
#ifdef UNIX
			errorCode = errno;
			xeLocal->GetUNIXAPIError(errorCode, szMsg);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7010E, drRead->szCurrentFile, szFError);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
#endif
		}
#endif

		//  Position to the start of the file
		errorCode = fseek(drRead->pFileDesc, 0, SEEK_SET);
		if (errorCode != EOK)
		{
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
			strerror_s(szFError, MAX_MSG, errno);
#else
#ifdef WIN32
			xeLocal->GetWINAPIError(errorCode, szMsg);
#endif
#ifdef UNIX
			xeLocal->GetUNIXAPIError(errorCode, szMsg);
#endif
#endif
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7051E, drRead->szCurrentFile, szFError);
			xeLocal->LogMessage(szMsg, iThreadID);
			fclose(drRead->pFileDesc);							//  Close the file handle
			drRead->pFileDesc = NULL;
			return 0;
		}
		drRead->bEOFReached = FALSE;
	}

	//  If the EOF has already been reached then close the file and return 0
	if (drRead->bEOFReached)
	{
		//  Trace the normal end of stream encountered
		xeLocal->LogTrace(TRACE_RESOURCELOADER, MSG_DXR7037T, iThreadID);
		fclose(drRead->pFileDesc);							//  Close the file handle
		drRead->pFileDesc = NULL;
		return 0;
	}

	//  Read the max bytes possible directly into the callers buffer
	iDXResLen = dwMaxToRead;
	iDXRead = fread(cBuffer, 1, iDXResLen, drRead->pFileDesc);
	if (iDXRead == 0)
	{
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
			strerror_s(szFError, MAX_MSG, errno);
#else
#ifdef WIN32
			xeLocal->GetWINAPIError(errorCode, szMsg);
#endif
#ifdef UNIX
			xeLocal->GetUNIXAPIError(errorCode, szMsg);
#endif
#endif
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7050E, drRead->szCurrentFile, szFError, iDXResLen, iDXRead);
		xeLocal->LogMessage(szMsg, iThreadID);
		fclose(drRead->pFileDesc);							//  Close the file handle
		drRead->pFileDesc = NULL;
		return 0;
	}

	//  Trace the return of data
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7039T, iDXRead);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	//  Determine if the EOF has been reached
	if (iDXRead < dwMaxToRead) drRead->bEOFReached = TRUE;

	//  Return the bytes read
	return (DWORD) iDXRead;
}

#ifdef WIN32
//  retryXMLReader
//
//  Wrapper function that determines if the XMLReader function can be re-driven in another configuration.
//  If the function can be re-driven then it will reconfigure and re-drive the function.
//
//  Parameters
//
//  1  -  char *			- Pointer to the buffer to use
//  2  -  DWORD				- Size in bytes of the freespace in the buffer
//  3  -  DWORD				- Error code the was returned by the send request
//  4  -  DXResource *		- Pointer to the DXResource object
//  5  -  int				- ID of the invoking thread
//
//  Returns
//
//  DWORD	-  Number of bytes of data now available in the buffer
//
DWORD DXResourceLoader::retryXMLReader(char *cBuffer, DWORD dwMaxToRead, DWORD dwError, DXResource *drRead, int iThreadID)
{
	//  Currently the HTTP Proxy Strategy must be HTTPP_FALLBACK to attempt a re-drive of the request
	if (drRead->HTTPPStrategy != HTTPP_FALLBACK) return 0;

	//  Check that we are not already in the alternate connection configuration
	if (drRead->iCurrentConnectConfig == HTTPP_WPROXY) return 0;

	//  Determine if the error code inidicates that re-driving the request is possible
	switch(dwError)
	{
	case ERROR_WINHTTP_CANNOT_CONNECT:
		//  Re-drivable
		break;
	case ERROR_WINHTTP_NAME_NOT_RESOLVED:
		//  Re-drivable
		break;
	case ERROR_WINHTTP_TIMEOUT:
		//  Re-drivable
		break;
	default:
		//  All others are not considered re-drivable
		return 0;
	}

	//  Close down the HTTP facility so that it can be brought back up in the alternate configuration
	WinHttpCloseHandle(hWHFacility);
	hWHFacility = NULLHANDLE;

	//  Trace re-driving the request
	xeLocal->LogTrace(TRACE_RESOURCELOADER, MSG_DXR7052T, iThreadID);

	//  Redrive the request
	return XMLHTTPReader(cBuffer, dwMaxToRead, drRead, iThreadID);
}
#endif

//  determineResourceType
//
//  Function to determine the type of the resource source that has been requested
//
//  Parameters
//
//  1  -  char *  - The resource source specifier
//
//  Returns
//
//  int		-  Encoded Resource Source Type (See: RSOURCE_ definitions)
//
int DXResourceLoader::determineSourceType(char *szResourceSource)
{
	char		*pSubStr;						//  Pointer to a substring location

	//  Reject an empty string
	if (szResourceSource[0] == '\0') return RSOURCE_INVALID;

	//  If the first character is a hash '#" then this is an application resource identifier
	if (szResourceSource[0] == '#') return RSOURCE_ARSRC;

	//  If the first character is a letter and the second character is a colon ':' then this is a file specifier
	if (strlen(szResourceSource) > 1)
	{
		if (((szResourceSource[0] >= 'a') &&  (szResourceSource[0] >= 'z')) ||
			((szResourceSource[0] >= 'A') &&  (szResourceSource[0] >= 'Z')))
		{
			if (szResourceSource[1] == ':') return RSOURCE_OSFILE;
		}
	}

	//  Check if this is a file:// style of URI then this is a file specifier
	if (strlen(szResourceSource) > strlen(FILETYPE_URI))
	{
		if (memcmpi(szResourceSource, FILETYPE_URI, strlen(FILETYPE_URI)) == 0) return RSOURCE_OSFILE;
	}

	//  Check if the string contains a URI delimiter :// if so then assume the source is an URL
	pSubStr = strstr(szResourceSource, GENERICTYPE_URI);
	if (pSubStr != NULL) return RSOURCE_URL;

	//  The source type is unrecognized or invalid
	return RSOURCE_INVALID;
}

//  setConnectConfig
//
//  Set the connection configuration according to the HTTP Proxy strategy and the current config
//
//  Parameters
//
//  1  -  DXResource *			-  Pointer to the DXResource object
//
//  Returns
//
void DXResourceLoader::setConnectConfig(DXResource *drRead)
{
	//  Set the connection configuration
	switch(drRead->HTTPPStrategy)
	{
	case HTTPP_NONE:
		// SNO condition
		drRead->iCurrentConnectConfig = HTTPP_NONE;
		break;
	case HTTPP_DIRECT:
		// Always use Direct
		drRead->iCurrentConnectConfig = HTTPP_DIRECT;
		break;
	case HTTPP_WPROXY:
		// Always use the WinHTTP Proxy
		drRead->iCurrentConnectConfig = HTTPP_WPROXY;
		break;
	case HTTPP_FALLBACK:
		//  This method will try Direct fist and will then try WinHTTP proxy
		if (drRead->iCurrentConnectConfig == HTTPP_NONE) drRead->iCurrentConnectConfig = HTTPP_DIRECT;
		else drRead->iCurrentConnectConfig = HTTPP_WPROXY;
		break;
	}
	return;
}

//  WrapperprocessThisCDRec
//
//  Function to vector to the processThisCDRec function, this is a static wrapper for the real code to provide linkage
//
//  Parameters
//
//  1  -  CHAR *	- Pointer to the CD Record returned by the Enumerator
//  2  -  WORD		- The CD Record Type
//  3  -  DWORD		- The CD Record Length
//  4  -  void *	- Context - this contains a pointer to the current instance of the DXResourceLoader in use
//
//  Returns
//
//  STATUS	-	   Used to signal to the enumerator to continue (NOERROR)
//
STATUS LNCALLBACK DXResourceLoader::WrapperprocessThisCDRec(char *pCDRec, WORD wCDType, DWORD dwCDLen, void *pContext)
{
	DXResource				*drCurrent;						//  Current Resource being processed
	DXResourceLoader		*rlCurrent;						//  Resource Loader

	//  Cast the Resource and Resource loader objects
	drCurrent = (DXResource *) pContext;
	rlCurrent = (DXResourceLoader *) drCurrent->pRLoader;
	return rlCurrent->processThisCDRec(pCDRec, wCDType, dwCDLen, drCurrent);
}

//  WrapperXMLReader
//
//  This function provides the interface for a DXL Importer or other XML reader to the read XML data
//
//  Parameters
//
//  1  -  unsigned char *		-  Pointer to the buffer where the XML data is to be stored
//  2  -  DWORD					-  Max bytes of data that can be stored in the buffer
//  3  -  void *				-  Pointer to this Design Manager object
//
//
//  Returns
//
//  DWORD		-	Length of XML data returned
//
//  API Use Notes:
//
//
DWORD LNCALLBACK DXResourceLoader::WrapperXMLReader(unsigned char *ucBuffer, DWORD dwMaxToRead, void *pContext)
{
	DXResource				*drCurrent;						//  Current Resource being processed
	DXResourceLoader		*rlCurrent;						//  Resource Loader

	//  Cast the Resource and Resource loader objects
	drCurrent = (DXResource *) pContext;
	rlCurrent = (DXResourceLoader *) drCurrent->pRLoader;

	//  Return using the appropriate On Demand reader routine
	if (drCurrent->iRMCLass & RMCLASS_HTTPLOD)
	{
		return rlCurrent->XMLHTTPReader((char *) ucBuffer, dwMaxToRead, drCurrent, drCurrent->iOwningThread);
	}

	if (drCurrent->iRMCLass & RMCLASS_FILELOD)
	{
		return rlCurrent->XMLFILEReader((char *) ucBuffer, dwMaxToRead, drCurrent, drCurrent->iOwningThread);
	}

	if (drCurrent->iRMCLass & RMCLASS_MEMLOD)
	{
		return rlCurrent->XMLMEMReader((char *) ucBuffer, dwMaxToRead, drCurrent, drCurrent->iOwningThread);
	}

	drCurrent->xeLocal->LogMessage(MSG_DXR7053E, drCurrent->iOwningThread);
	return 0;
}

//  cloneResourceHandle
//
//  This function will clone an exiting resource to produce a copy that can be used in a different context.
//
//  Parameters
//
//  1  -  DXRESOURCEHANDLE		- Handle of the resource that is to be cloned
//  2  -  int					- ID of the invoking thread
//
//  Returns
//
//  DXRESOURCEHANDLE  -  Handle to the DXResource that will be used to retrieve the data
//						 NULLHANDLE if the resource could not be created
//
//  API Use Notes:
//
//  1. It is the responsibility of the caller to free the returned DXRESOURCEHANDLE by using the dispose() function
//
DXRESOURCEHANDLE DXResourceLoader::cloneResourceHandle(DXRESOURCEHANDLE hdrPrime, int iThreadID)
{
	DXResource			*drPrime;							//  The primary resource
	DXResource			*drClone;							//  The cloned resource
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Cast the handle to the resource
	drPrime = (DXResource *) hdrPrime;

	//  Create the new DXResource
	drClone = new DXResource(xeLocal, iThreadID);

	//  Common settings
	drClone->iRMCLass = (drPrime->iRMCLass & RMCLASS_TYPE) | RMCLASS_CLONE;
	drClone->iRMState = RMSTATE_INIT;

	//  Complete the cloning process according to the type
	switch(drPrime->iRMCLass & RMCLASS_TYPE)
	{
	case RMCLASS_HTTPLOD:
		//  This is an URL type resource
		strcpy_s(drClone->szCurrentReadURL, MAX_URL, drPrime->szCurrentReadURL);
	
		//  Allocate a buffer to retrieve the resource contents into - the buffer is set to the default size
		drClone->dwMemSize = DEF_HTTP_RCV_BFRSIZE;												//  Set default buffer size
		drClone->hResBuffer = xeLocal->getDominoMemory(drClone->dwMemSize, iThreadID);			//  Allocate the memory buffer
		if (drClone->hResBuffer == NULLHANDLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7008E, drClone->dwMemSize, drClone->szCurrentReadURL);
			xeLocal->LogMessage(szMsg, iThreadID);
			delete drClone;
			return NULLHANDLE;							//  Show that the load request failed
		}
		break;

	case RMCLASS_FILELOD:
		//  This is a File type resource
		strcpy_s(drClone->szCurrentFile, MAXPATH, drPrime->szCurrentFile);
		break;

	case RMCLASS_MEMLOD:
		//  This is a memory type resource
		drClone->dwMemOffset = 0;
		drClone->dwMemSize = drPrime->dwMemSize;
		drClone->hResBuffer = drPrime->hResBuffer;
		break;

	default:
		//  Whatever type of resource this is it is not cloneable
		delete drClone;
		return NULLHANDLE;
	}

	//  Trace the cloning event
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7057T, drClone, drPrime);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	return drClone;
}

//  disposeResourceHandle
//
//  This function will dispose of an exiting resource - underlying resource will also be conditionally disposed of.
//
//  Parameters
//
//  1  -  DXRESOURCEHANDLE		- Handle of the resource that is to be cloned
//  2  -  int					- ID of the invoking thread
//
//  Returns
//
//  API Use Notes:
//
//
void DXResourceLoader::disposeResourceHandle(DXRESOURCEHANDLE hdrUsed, int iThreadID)
{
	DXResource			*drUsed;							//  Pointer to the resource
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Safety check - don't try disposing of a NULL resource
	if (hdrUsed == NULLHANDLE) return;

	//  Cast the handle to the resource address
	drUsed = (DXResource *) hdrUsed;

	//  Complete the cloning process according to the type
	switch(drUsed->iRMCLass & RMCLASS_TYPE)
	{
	case RMCLASS_HTTPLOD:
		//  This is an URL type resource - cleanup the HTTP request & connection and free the buffers
#ifdef WIN32
		if (drUsed->hWHRequest != NULLHANDLE)
		{
			WinHttpCloseHandle(drUsed->hWHRequest);
			drUsed->hWHRequest = NULLHANDLE;
		}
		if (drUsed->hWHConnection != NULLHANDLE)
		{
			WinHttpCloseHandle(drUsed->hWHConnection);
			drUsed->hWHConnection = NULLHANDLE;
		}
#endif
		if (drUsed->hRespBuffer != NULLHANDLE)
		{
			OSMemoryFree(drUsed->hRespBuffer);
			drUsed->hRespBuffer = NULLHANDLE;
		}
		if (drUsed->hResBuffer != NULLHANDLE)
		{
			OSMemoryFree(drUsed->hResBuffer);
			drUsed->hResBuffer = NULLHANDLE;
		}
		break;

	case RMCLASS_FILELOD:
		//  This is a File type resource
		if (drUsed->pFileDesc != NULL) fclose(drUsed->pFileDesc);				//  Close the file handle
		break;

	case RMCLASS_MEMLOD:
		//  This is a memory type resource - if it is the primary resource then free the memory buffer
		if (drUsed->iRMCLass & RMCLASS_PRIMARY)
		{
			if (drUsed->hResBuffer != NULLHANDLE)
			{
				OSMemoryFree(drUsed->hResBuffer);
				drUsed->hResBuffer = NULLHANDLE;
			}
		}
		break;
	}

	//  Delete the resource
	delete drUsed;
	drUsed = NULL;

	//  Trace the action
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7058T, drUsed);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	return;
}

//  uploadThisResource
//
//  This function will upload the passed file in the provided context
//
//  Parameters
//
//  1  -  char *				-	The name of the file to be uploaded
//  2  -  int					-	The type of file to be uploaded see UPLOAD_FILE_
//  3  -  DXUploadContext *		-	Pointer to the upload context to be used
//  4  -  int					-	ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE		-  Handle of the in-memory response buffer
//
//  API Use Notes:
//
//
MEMHANDLE DXResourceLoader::uploadThisResource(char *szFileName, int iFileType, DXUploadContext *ducThis, int iThreadID)
{
	MEMHANDLE		hmResource;							//  Handle of the in-memory resource
	MEMHANDLE		hmResponse;							//  Handle of the returned response
	DWORD			dwResourceSize;						//  Size of the in-memory resource
	char			*pLoadedResource;					//  Pointer to the loaded resource
	char			szResourceName[MAXPATH + 1];		//  Complete resource name
	char			szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Sanity checks
	if (szFileName == NULL)
	{
		xeLocal->LogMessage(MSG_DXR7060E, iThreadID);
		return NULLHANDLE;
	}
	if (szFileName[0] == '\0')
	{
		xeLocal->LogMessage(MSG_DXR7060E, iThreadID);
		return NULLHANDLE;
	}

	//  Load the file into memory
	sprintf_s(szResourceName, MAXPATH, "%s%s", FILETYPE_URI, szFileName);
	hmResource = loadThisResource(szResourceName, iThreadID);
	if (hmResource == NULLHANDLE) return NULLHANDLE;

	//  Trace the loaded file
	dwResourceSize = OSMemoryGetSize(hmResource);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7074T, szFileName, dwResourceSize);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);
	pLoadedResource = (char *) OSMemoryLock(hmResource);
	xeLocal->DumpMemory(pLoadedResource, dwResourceSize, "Loaded File");
	OSMemoryUnlock(hmResource);

	//  Invoke the in-memory upload
	hmResponse = uploadThisResource(hmResource, iFileType, ducThis, iThreadID);

	//  Free the in-memory resource
	OSMemoryFree(hmResource);

	//  Return the upload result
	return hmResponse;
}

//  uploadThisResource
//
//  This function will upload the passed in-memory resource in the provided context
//
//  Parameters
//
//  1  -  MEMHANDLE				-	Handle of the in-memory resource
//  2  -  int					-	The type of file to be uploaded see UPLOAD_FILE_
//  3  -  DXUploadContext *		-	Pointer to the upload context to be used
//  4  -  int					-	ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE			-  Handle to the in-memory response buffer
//
//  API Use Notes:
//
//
MEMHANDLE DXResourceLoader::uploadThisResource(MEMHANDLE hmResource, int iFileType, DXUploadContext *ducThis, int iThreadID)
{
	MEMHANDLE			hmForm;								//  Memory handle of the form
	MEMHANDLE			hmPayload;							//  Memory handle of the payload
	DWORD				dwFrmSize;							//  Size of the loaded form
	int					iSepCount;							//  Separator count
	char				*pForm;								//  Pointer to the in-memory form
	char				*pSource;							//  Pointer into source string
	char				*pTarget;							//  Pointer into target string
	char				szOpenFormURL[MAX_URL + 1];			//  Open Form URL
	char				szPostFormURL[MAX_URL + 1];			//  Post Form URL
	char				szClickAction[MAX_URL + 1];			//  Click Action Value
	char				szFileUploadItem[MAX_URL + 1];		//  File Upload Item name
	char				szTargetURL[MAX_URL + 1];			//  Complete target URL
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Sanity checks
	if (hmResource == NULLHANDLE) 
	{
		xeLocal->LogMessage(MSG_DXR7061E, iThreadID);
		return NULLHANDLE;
	}
	if (ducThis == NULL) 
	{
		xeLocal->LogMessage(MSG_DXR7062E, iThreadID);
		return NULLHANDLE;
	}
	if (ducThis->szFormURL[0] == '\0')
	{
		xeLocal->LogMessage(MSG_DXR7063E, iThreadID);
		return NULLHANDLE;
	}

	//  Step 1 - Open the form, this will validate the URL and we use the form content to build
	//           the upload stream.
	sprintf_s(szOpenFormURL, MAXPATH, "%s?OpenForm", ducThis->szFormURL);
	hmForm = loadThisResource(szOpenFormURL, iThreadID);
	if (hmForm == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7064E, ducThis->szFormURL);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Get the click action and the upload name from the form
	pForm = (char *) OSMemoryLock(hmForm);
	dwFrmSize = OSMemoryGetSize(hmForm);
	strcpy_s(szClickAction, MAX_URL, ducThis->szSubmitAction);
	getFormAction(pForm, dwFrmSize, szPostFormURL, szFileUploadItem, szClickAction);
	OSMemoryUnlock(hmForm);
	OSMemoryFree(hmForm);

	//  Check that we have a valid Post Form URL
	if (szPostFormURL[0] == '\0')
	{
		xeLocal->LogMessage(MSG_DXR7067E, iThreadID);
		return NULLHANDLE;
	}
	//  Check that we have a valid upload & click action
	if (szFileUploadItem[0] == '\0')
	{
		//  No file upload control
		xeLocal->LogMessage(MSG_DXR7065E, iThreadID);
		return NULLHANDLE;
	}
	if (szClickAction[0] == '\0')
	{
		if (ducThis->szSubmitAction[0] == '\0') sprintf_s(szMsg, MAX_MSG, MSG_DXR7066E, FORM_SUB_ACTION);
		else sprintf_s(szMsg, MAX_MSG, MSG_DXR7066E, ducThis->szSubmitAction);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Assemble the payload content
	hmPayload = buildPayload(szClickAction, szFileUploadItem, ducThis, iFileType, hmResource, iThreadID);
	if (hmPayload == NULLHANDLE)
	{
		xeLocal->LogMessage(MSG_DXR7068E, iThreadID);
		return NULLHANDLE;
	}

	//  Construct the full target URL
	iSepCount = 0;
	pSource = ducThis->szFormURL;
	pTarget = szTargetURL;
	//  Copy the protocol & host name from the Form URL
	while ((iSepCount < 3) && (*pSource != '\0'))
	{
		if (*pSource == '/')
		{
			iSepCount++;
		}
		*pTarget = *pSource;
		pSource++;
		pTarget++;
	}
	pTarget--;
	pSource = szPostFormURL;
	while (*pSource != '\0')
	{
		*pTarget = *pSource;
		pSource++;
		pTarget++;
	}
	*pTarget = '\0';

	sprintf_s(szMsg, MAX_MSG, MSG_DXR7073I, szTargetURL);
	xeLocal->LogMessage(szMsg, iThreadID);

	//  Pass the payload to the post routine, we get the handle to the response in return
	hmForm = NULLHANDLE;
	hmForm = postFormToURL(hmPayload, szTargetURL, iThreadID);
	return hmForm;
}

//  getFormAction
//
//  This function will extract the click action and the file upload item name from a loaded form HTML.
//
//  Parameters
//
//  1  -  char *				-	Pointer to the in-memory form HTML
//  2  -  DWORD					-	Size of the form HTML
//  3  -  char *				-	Pointer to the buffer where the Post URL will be stored
//  4  -  char *				-	Pointer to the buffer where the File Upload item will be stored
//  5  -  char *				-	Pointer to the buffer where the click value will be stored
//
//  Returns
//
//  API Use Notes:
//
//
void DXResourceLoader::getFormAction(char *pForm, DWORD dwLength, char *szPostURL, char *szFileUpload, char *szClickAction)
{
	char				*pSOHTML;							//  Pointer to start of HMTL
	char				*pEOHTML;							//  Pointer to end of HTML
	char				*pSOFORM;							//  Pointer to start of FORM
	char				*pEOFORM;							//  Pointer to end of FORM
	char				*pInput;							//  Pointer to input element
	char				*pLabel;							//  Pointer to a label
	char				*pTarget;							//  Pointer into a target string
	char				cSep;								//  Separator character
	char				szReqAction[MAX_URL + 1];			//  Requitred Action
	char				szValue[MAX_URL + 1];				//  Arbitrary value

	//  Save the requested action if there is one
	strcpy_s(szReqAction, MAX_URL, szClickAction);

	//  Clear the outputs
	szPostURL[0] = '\0';
	szFileUpload[0] = '\0';
	szClickAction[0] = '\0';

	//  Locate the start and the end of the HTML
	pSOHTML = findElement(pForm, dwLength, "<html>");
	if (pSOHTML == NULL) return;
	pEOHTML = findElement(pForm, dwLength, "/html>");
	if (pEOHTML == NULL) return;

	//  Locate the start and the end of the form within the HTML
	pSOFORM = findElement(pSOHTML, (DWORD) (pEOHTML - pSOHTML), "<form");
	if (pSOFORM == NULL) return;
	pEOFORM = findElement(pSOFORM, (DWORD) (pEOHTML - pSOFORM), "</form>");
	if (pEOFORM == NULL) return;

	//  Find the Post URL in the form element
	pLabel = findElement(pSOFORM, 100, "action=");
	if (pLabel == NULL) return;

	//  Copy the action content into the Post URL, watch out for &amp;
	pLabel += 7;
	cSep = *pLabel;
	pLabel++;
	pTarget = szPostURL;
	while (*pLabel != cSep)
	{
		if (*pLabel == '&')
		{
			*pTarget = '&';
			pTarget++;
			while (*pLabel != ';')
			{
				pLabel++;
			}
			pLabel++;
		}
		else
		{
			*pTarget = *pLabel;
			pTarget++;
			pLabel++;
		}
	}
	*pTarget = '\0';

	//  Loop processing each input element in turn look for the file upload control and
	//  the submitting action
	pInput = findElement(pSOFORM, (DWORD) (pEOFORM - pSOFORM), "<input");
	while (pInput != NULL)
	{
		//  Determine the type of input element
		pLabel = findElement(pInput, 20, "type=");
		if (pLabel != NULL)
		{
			pLabel += 5;
			cSep = *pLabel;
			pLabel++;
			pTarget = szValue;
			while (*pLabel != cSep)
			{
				*pTarget = *pLabel;
				pTarget++;
				pLabel++;
			}
			*pTarget = '\0';

			//  See if the type is button
			if (strcmpi(szValue, "button") == 0)
			{
				//  Do we sill need a submit action
				if (szClickAction[0] == '\0')
				{
					//  Get the action name
					szValue[0] = '\0';
					pLabel = findElement(pInput, 100, "value=");
					if (pLabel != NULL)
					{
						pTarget = szValue;
						pLabel += 6;
						cSep = *pLabel;
						pLabel++;
						while (*pLabel != cSep)
						{
							*pTarget = *pLabel;
							pTarget++;
							pLabel++;
						}
						*pTarget = '\0';
					}

					//  If we just want the default or this is the selected action then use it
					if ((szReqAction[0] == '\0') || (strcmpi(szReqAction, szValue) == 0))
					{
						// find the onclick= element
						pLabel = findElement(pInput, 100, "onclick=");
						if (pLabel != NULL)
						{
							while (*pLabel != '(') pLabel++;
							pLabel++;
							cSep = *pLabel;
							pLabel++;
							pTarget = szClickAction;
							while (*pLabel != cSep)
							{
								*pTarget = *pLabel;
								pTarget++;
								pLabel++;
							}
							*pTarget = '\0';
						}
					}
				}
			}

			//  If this is a type="file" then we use this to get the Item Name
			if (strcmpi(szValue, "file") == 0)
			{
				pLabel = findElement(pInput, 100, "name=");
				if (pLabel != NULL)
				{
					pLabel += 5;
					cSep = *pLabel;
					pLabel++;
					pTarget = szFileUpload;
					while (*pLabel != cSep)
					{
						*pTarget = *pLabel;
						pTarget++;
						pLabel++;
					}
					*pTarget = '\0';
				}
			}
		}

		if ((szFileUpload[0] != '\0') && (szClickAction[0] != '\0')) return;
		pInput++;
		pInput = findElement(pInput, (DWORD) (pEOFORM - pInput), "<input");
	}
	return;
}

//  findElement
//
//  This function will locate the next ocurrence of required element within a search space
//
//  Parameters
//
//  1  -  char *				-	Pointer to the search space
//  2  -  DWORD					-	Size of the search space
//
//  Returns
//
//  char *			-	pointer to the occurrence or NULL if not found within the space
//
//  API Use Notes:
//
//
char * DXResourceLoader::findElement(char *pSpace, DWORD dwSize, char *SearchString)
{
	DWORD		dwLeftToSearch;						//  Bytes left to search
	char		*pSearch;							//  Primary search pointer

	//  Initial setup
	dwLeftToSearch = dwSize;
	pSearch = pSpace;

	while (dwLeftToSearch > 0)
	{
		if (toupper(*pSearch) == toupper(SearchString[0]))
		{
			if (memcmpi(pSearch, SearchString, strlen(SearchString)) == 0) return pSearch;
		}

		pSearch++;
		dwLeftToSearch--;
	}
	
	//  Not found
	return NULL;
}

//  buildPayload
//
//  This function will construct the payload for a file upload operation
//
//  Parameters
//
//  1  -  char *				-	Pointer to the click value 
//  2  -  char *				-	Pointer to the File Upload item 
//  3  -  DXUploadContext *		-	Pointer to the DXUploadContext object
//  4  -  int					-   File Type
//  5  -  MEMHANDLE				-	Handle of the in-memory file contents
//  6  -  int					-	ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE			-  Handle of the constructed payload, NULLHANDLE if build failed
//
//  API Use Notes:
//
//
MEMHANDLE DXResourceLoader::buildPayload(char *szCA, char *szFUN, DXUploadContext *ducThis, int iFileType, MEMHANDLE hmFile, int iThreadID)
{
	STATUS			stAPIRC;							//  Return code from Notes API calls
	MEMHANDLE		hmPayload;							//  Handle to memory buffer for payload
	DWORD			dwBfrSize;							//  Size of the payload buffer
	DWORD			dwFileSize;							//  Size of the payload file
	DWORD			dwAllocated;						//  Size allocated in the payload buffer
	DWORD			dwFreeSpace;						//  Free space in the payload buffer
	DWORD			dwFileOH;							//  Overhead size in the file
	DXUCItem		*uciCurrent;						//  Current DXUCItem object
	char			*pBuffer;							//  Pointer to the start of buffer
	char			*pInsertPoint;						//  Current insert point in the payload buffer
	char			szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Compute the initial size of the payload content buffer
	dwFileSize = OSMemoryGetSize(hmFile);
	dwBfrSize = dwFileSize + PAYLOAD_BFR_OHEAD;

	//  Allocate a memory buffer to hold the payload content
	stAPIRC = OSMemoryAllocate(MEM_GROWABLE, dwBfrSize, &hmPayload);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7069E, dwBfrSize);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Lock the buffer for the payload
	pBuffer = (char *) OSMemoryLock(hmPayload);
	pInsertPoint = pBuffer;

	//  Construct a DXUCItem for the __Click item and format it in the buffer
	uciCurrent = new DXUCItem();
	strcpy_s(uciCurrent->szItemName, MAX_DESC, "__Click");
	strcpy_s(uciCurrent->szEncodedValue, MAX_MSG, szCA);
	uciCurrent->pNextItem = NULL;
	pInsertPoint += appendUCItem(pInsertPoint, uciCurrent, iThreadID);
	delete uciCurrent;

	//  Loop through the chain of DXUCItems supplied and append each into the buffer
	uciCurrent = ducThis->pFirstItem;
	while (uciCurrent != NULL)
	{
		pInsertPoint += appendUCItem(pInsertPoint, uciCurrent, iThreadID);
		uciCurrent = uciCurrent->pNextItem;
	}

	//  Check that there is enough free space in the payload buffer for the file
	dwFreeSpace = DWORD(dwBfrSize - (pInsertPoint - pBuffer));
	//  Increase the buffer size until it will accomodate the file
	while (dwFreeSpace < dwFileSize)
	{
		dwAllocated = (DWORD) (pInsertPoint - pBuffer);
		OSMemoryUnlock(hmPayload);
		dwBfrSize += PAYLOAD_BFR_OHEAD;
		stAPIRC = OSMemoryReallocate(hmPayload, dwBfrSize);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7069E, dwBfrSize);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			OSMemoryFree(hmPayload);
			return NULLHANDLE;
		}
		pBuffer = (char *) OSMemoryLock(hmPayload);
		pInsertPoint = pBuffer + dwAllocated;
		dwFreeSpace = DWORD(dwBfrSize - (pInsertPoint - pBuffer));
	}

	//  Append the file to the payload buffer
	pInsertPoint += appendFile(pInsertPoint, szFUN, ducThis->szFileName, iFileType, hmFile, &dwFileOH, iThreadID);
	pInsertPoint += dwFileOH;

	//  Shrink the memory allocation back to the used size
	dwAllocated = (DWORD) (pInsertPoint - pBuffer);
	OSMemoryUnlock(hmPayload);
	stAPIRC = OSMemoryReallocate(hmPayload, dwAllocated);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7069E, dwAllocated);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(ERR(stAPIRC), szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		OSMemoryFree(hmPayload);
		return NULLHANDLE;
	}

	//  Notify that the Payload has been constructed & Dump it
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7072I, dwAllocated);
	xeLocal->LogMessage(szMsg, iThreadID);
	pInsertPoint = (char *) OSMemoryLock(hmPayload);
	xeLocal->DumpMemory(pInsertPoint, dwAllocated, "Payload");
	OSMemoryUnlock(hmPayload);

	return hmPayload;
}

//  appendUCItem
//
//  This function will append a new DXUCItem content to the payload buffer
//
//  Parameters
//
//  1  -  char *				-	Pointer to the current insert position in the payload buffer
//  2  -  DXUCItem *			-	Pointer to the DXUCItem object to be appended
//  3  -  int					-	ID of the invoking thread
//
//  Returns
//
//  DWORD			-  Size of the item appended in the buffer
//
//  API Use Notes:
//
//
DWORD DXResourceLoader::appendUCItem(char *pBfr, DXUCItem *uciThis, int iThreadID)
{
	DWORD			dwBytesAdded;				//  Count of bytes added to the payload buffer
	char			*pInsertPoint;				//  Insertion Point
	char			szChunk[MAX_MSG + 1];		//  Chunk to be added to the buffer
	char			szMsg[MAX_MSG + 1];			//  Generic message buffer

	dwBytesAdded = 0;
	pInsertPoint = pBfr;

	//  First add the multipart boundary
	sprintf_s(szChunk, MAX_MSG, "--%s\r\n", FORM_BOUNDARY);
	memcpy(pInsertPoint, szChunk, strlen(szChunk));
	dwBytesAdded += (DWORD) strlen(szChunk);
	pInsertPoint += strlen(szChunk);

	//  Now add the content descriptor
	sprintf_s(szChunk, MAX_MSG, FORM_CONTENT_DESC, uciThis->szItemName);
	memcpy(pInsertPoint, szChunk, strlen(szChunk));
	dwBytesAdded += (DWORD) strlen(szChunk);
	pInsertPoint += strlen(szChunk);

	//  Now add the item value
	sprintf_s(szChunk, MAX_MSG, "%s\r\n", uciThis->szEncodedValue);
	memcpy(pInsertPoint, szChunk, strlen(szChunk));
	dwBytesAdded += (DWORD) strlen(szChunk);
	pInsertPoint += strlen(szChunk);

	//  Trace the addition
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7070T, uciThis->szItemName, uciThis->szEncodedValue, dwBytesAdded);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	return dwBytesAdded;
}

//  appendFile
//
//  This function will append a File to the payload buffer
//
//  Parameters
//
//  1  -  char *				-	Pointer to the current insert position in the payload buffer
//  2  -  char *				-	Pointer to the File Upload Item name
//  3  -  char *				-   Pointer to the file name
//  4  -  int					-   File Type
//  5  -  MEMHANDLE				-   Memory Handle of the file contents
//  6  -  DWORD *				-   Pointer to where the prefix length will be set
//  7  -  int					-	ID of the invoking thread
//
//  Returns
//
//  DWORD			-  Size of the item appended in the buffer
//
//  API Use Notes:
//
//
DWORD DXResourceLoader::appendFile(char *pBfr, char *szFUN, char *szFN, int iFT, MEMHANDLE hmFile, DWORD *dwPfxLen, int iThreadID)
{
	DWORD			dwBytesAdded;				//  Count of bytes added to the payload buffer
	DWORD			dwPrefixLength;				//  Prefix Length
	DWORD			dwFileSize;					//  File content size
	char			*pInsertPoint;				//  Insertion Point
	char			*pFileContents;				//  Pointer to the File Contents
	char			szChunk[MAX_MSG + 1];		//  Chunk to be added to the buffer
	char			szMsg[MAX_MSG + 1];			//  Generic message buffer

	dwBytesAdded = 0;
	dwPrefixLength = 0;
	*dwPfxLen = 0;
	pInsertPoint = pBfr;

	//  First add the multipart boundary
	sprintf_s(szChunk, MAX_MSG, "--%s\r\n", FORM_BOUNDARY);
	memcpy(pInsertPoint, szChunk, strlen(szChunk));
	dwPrefixLength += (DWORD) strlen(szChunk);
	pInsertPoint += strlen(szChunk);

	//  Add the file descriptor
	sprintf_s(szChunk, MAX_MSG, FORM_FILE_DESC, szFUN, szFN);
	memcpy(pInsertPoint, szChunk, strlen(szChunk));
	dwPrefixLength += (DWORD) strlen(szChunk);
	pInsertPoint += strlen(szChunk);

	//  Add the file content type
	switch(iFT)
	{
	case UPLOAD_FILE_TEXT:
		sprintf_s(szChunk, MAX_MSG, FORM_CONTENT_TYPE, FILE_TYPE_TEXT);
		break;
	case UPLOAD_FILE_XML:
		sprintf_s(szChunk, MAX_MSG, FORM_CONTENT_TYPE, FILE_TYPE_XML);
		break;
	case UPLOAD_FILE_BINARY:
		sprintf_s(szChunk, MAX_MSG, FORM_CONTENT_TYPE, FILE_TYPE_BINARY);
		break;
	default:
		sprintf_s(szChunk, MAX_MSG, FORM_CONTENT_TYPE, FILE_TYPE_BINARY);
	}
	memcpy(pInsertPoint, szChunk, strlen(szChunk));
	dwPrefixLength += (DWORD) strlen(szChunk);
	pInsertPoint += strlen(szChunk);

	//  Append the file contents to the payload buffer
	dwFileSize =  OSMemoryGetSize(hmFile);
	pFileContents = (char *) OSMemoryLock(hmFile);
	memcpy(pInsertPoint, pFileContents, dwFileSize);
	dwBytesAdded += dwFileSize;
	pInsertPoint += dwFileSize;

	//  Finally add the multipart boundary
	sprintf_s(szChunk, MAX_MSG, "\r\n--%s--\r\n", FORM_BOUNDARY);
	memcpy(pInsertPoint, szChunk, strlen(szChunk));
	dwPrefixLength += (DWORD) strlen(szChunk);
	pInsertPoint += strlen(szChunk);

	OSMemoryUnlock(hmFile);

	//  Trace the addition
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7071T, szFN, dwBytesAdded, dwPrefixLength);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	*dwPfxLen = dwPrefixLength;
	return dwBytesAdded;
}

//  postFormToURL
//
//  Specific entry point taking the URL of a target form and a handle to a pre-constructed payload. The
//  payload will be posted to the specific URL and a memory buffer will be returned containing any
//  response data.
//
//  Parameters
//
//  1  -  MEMHANDLE			-  Handle of the payload to be posted
//  2  -  char *			- The resource source URL
//  3  -  int				- ID of the invoking thread
//
//  Returns
//
//  MEMHANDLE -  Handle to the Domino Memory allocation containing the response
//               NULLHANDLE if the post action failed.
//
//  API Usage:
//
//  This version implements only un-authenticated posting, this will be extended in future versions to
//  handle basc authentication.
//

MEMHANDLE DXResourceLoader::postFormToURL(MEMHANDLE hmPayload, char *szURL, int iThreadID)
{
	DXResource		*drTemp;						//  Resouce created for the duration of this call
	MEMHANDLE		hReturn;						//  Handle to be returned
	DWORD			dwReadLen;						//  Length Read on a single read
	DWORD			dwTotalRead;					//  Total length read
	char			*pDXResMem;						//  Pointer to the Domino memory buffer for the resource
	char			*pReadPointer;					//  Current address to read data into
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Create a resource to hold the transient information
	drTemp = new DXResource(xeLocal, iThreadID);
	drTemp->pRLoader = this;

	//  Allocate a buffer to retrieve the resource contents into - the buffer is set to the default size
	drTemp->dwMemSize = DEF_HTTP_RCV_BFRSIZE;							//  Set default buffer size
	drTemp->hResBuffer = xeLocal->getDominoMemory(drTemp->dwMemSize, iThreadID);			//  Allocate the memory buffer
	if (drTemp->hResBuffer == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7008E, drTemp->dwMemSize, szURL);
		xeLocal->LogMessage(szMsg, iThreadID);
		delete drTemp;
		return NULLHANDLE;							//  Show that the load request failed
	}

	//  Lock the Domino memory
	pDXResMem = (char *) OSMemoryLock(drTemp->hResBuffer);

	//  Setup to read the resource contents from the passed URL
	strcpy_s(drTemp->szCurrentReadURL, MAX_URL, szURL);		//  Set the URL to read the content from
	dwReadLen = 1;											//  Dummy setting to enter the while loop
	dwTotalRead = 0;
	pReadPointer = pDXResMem;
	while (dwReadLen > 0)
	{
		//  Check if there is sufficient (minimum) space in the buffer, if not then grow the buffer
		if ((drTemp->dwMemSize - dwTotalRead) <= MIN_FREE_BUFFERSPACE)
		{
			//  Grow the Buffer by one increment
			OSMemoryUnlock(drTemp->hResBuffer);				//  Unlock the current buffer
			drTemp->dwMemSize = drTemp->dwMemSize + DEF_HTTP_RCV_BFRSIZE;
			drTemp->hResBuffer = xeLocal->resizeDominoMemory(drTemp->hResBuffer, drTemp->dwMemSize, iThreadID);
			if (drTemp->hResBuffer == NULLHANDLE)
			{
				//  Memory reallocation has failed
				sprintf_s(szMsg, MAX_MSG, MSG_DXR7013E, drTemp->dwMemSize, szURL);
				xeLocal->LogMessage(szMsg, iThreadID);
				OSMemoryFree(drTemp->hResBuffer);
				drTemp->hResBuffer = NULLHANDLE;
				delete drTemp;
				return NULLHANDLE;							//  Show that the load request failed
			}

			//  Lock the Domino memory
			pDXResMem = (char *) OSMemoryLock(drTemp->hResBuffer);
			//  Recalculate the Address to read data into
			pReadPointer = pDXResMem + dwTotalRead;
		}

		//  Read Data from the HTTP stream
		dwReadLen = FormHTTPWriter(hmPayload, pReadPointer, drTemp->dwMemSize - dwTotalRead, drTemp, iThreadID);
		dwTotalRead = dwTotalRead + dwReadLen;
		pReadPointer = pReadPointer + dwReadLen;
	}
	*pReadPointer = '\0';
	OSMemoryUnlock(drTemp->hResBuffer);

	//  Check that data was actually read - if not free the buffer
	if (dwTotalRead == 0)
	{
		OSMemoryFree(drTemp->hResBuffer);
		drTemp->hResBuffer = NULLHANDLE;
		delete drTemp;
		return NULLHANDLE;
	}

	//  Resize (shrink) the Domino memory buffer to the size of the resource content
	dwTotalRead++;							//  Allow for the NULL terminator
	drTemp->hResBuffer = xeLocal->resizeDominoMemory(drTemp->hResBuffer, dwTotalRead, iThreadID);
	drTemp->dwMemSize = dwTotalRead;

	//  Show that the resource has been loaded
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7009I, szURL, dwTotalRead);
	xeLocal->LogVerbose(szMsg, iThreadID);

	//  Return the handle to the memory buffer
	hReturn = drTemp->hResBuffer;
	delete drTemp;
	return hReturn;
}

#ifdef WIN32
//  FormHTTPWriter				(WinHTTP implementation)
//
//  Function to provide an in memory response stream from an HTTP post
//
//  Parameters
//
//  1  -  MEMHANDLE				-  Handle to the in-memory payload to be posted
//  2  -  char *				-  Pointer to the buffer to use
//  3  -  DWORD					-  Size in bytes of the freespace in the buffer
//  4  -  DXResource *			-  Pointer to the DXResource object that contains the resource content
//  5  -  int					-  ID of the invoking thread
//
//  Returns
//
//  DWORD	-  Number of bytes of data now available in the buffer
//
//  API Usage;
//
//
DWORD DXResourceLoader::FormHTTPWriter(MEMHANDLE hmPayload, char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID)
{
	URL_COMPONENTS			urlcCurrent;				//  URL Components derived from the current request
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
	size_t					iConverted;					//  Count of characters converted
#endif
	DWORD					dwRespCode;					//  Response code from HTTP Request
	DWORD					dwRetLen;					//  Length of responses
	DWORD					dwErrorCode;				//  Error Code from WinHTTP Function calls
	DWORD					dwPayloadLen;				//  Length of the payload
	char					*cCache;					//  Cache buffer for returned data
	char					*pPayload;					//  Pointer to the payload data
	char					szTempStr[MAX_URL + 1];		//  Temp string for conversions
	char					szHeader[MAX_URL + 1];		//  Header value buffer
	wchar_t					wcFullURL[MAX_URL + 1];		//  Full URL
	wchar_t					wcHostName[MAX_URL + 1];	//  Host Name parsed from the request URL
	wchar_t					wcURLPath[MAX_URL + 1];		//  Path component of the request URL
	wchar_t					wcHeader[MAX_URL + 1];		//  HTTP Header to be added to the request
	char					szMsg[MAX_MSG + 1];			//  Generic message buffer

	//  Check if the WinHTTP component is loaded - if not load it
	if (hWHFacility == NULLHANDLE)
	{
		//  Setup the connection configuration
		setConnectConfig(drRead);
		if (drRead->iCurrentConnectConfig == HTTPP_DIRECT)
		{
			//  Create the facility using direct connection setup
			hWHFacility = WinHttpOpen(L"DX Resource Loader 2.0", 
									WINHTTP_ACCESS_TYPE_NO_PROXY, 
									WINHTTP_NO_PROXY_NAME,
									WINHTTP_NO_PROXY_BYPASS,
									0);
		}
		else
		{
			//  Create the facility using the default WinHTTP Proxy settings
			//  NOTE: These are not the IE browser proxy settings you must use the WinHTTP Proxycfg.exe configurator
			if (drRead->iCurrentConnectConfig == HTTPP_WPROXY)
			{
				hWHFacility = WinHttpOpen(L"DX Resource Loader 2.0", 
										WINHTTP_ACCESS_TYPE_DEFAULT_PROXY, 
										WINHTTP_NO_PROXY_NAME,
										WINHTTP_NO_PROXY_BYPASS,
										0);
			}
		}
		if (hWHFacility == NULLHANDLE)
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			//  Log the error and return 0 bytes read
			xeLocal->LogMessage(MSG_DXR7014E, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
		}

		//  Trace the creation of WinHTTP
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7016T, drRead->iCurrentConnectConfig);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);
	}

	//  Check if we have already made the connection to the source - if not open the connection
	if (drRead->hWHConnection == NULLHANDLE)
	{
		ZeroMemory(&urlcCurrent, sizeof(URL_COMPONENTS));
		urlcCurrent.dwStructSize = sizeof(URL_COMPONENTS);
		urlcCurrent.dwHostNameLength = sizeof(wcHostName);					//  Flag that Host Name needed
		urlcCurrent.lpszHostName = wcHostName;								//  Set the buffer

		//  Use the URL cracker to extract the server name from the requested URL
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		mbstowcs_s(&iConverted, wcFullURL, strlen(drRead->szCurrentReadURL) + 1, drRead->szCurrentReadURL, _TRUNCATE);
#else
		mbstowcs(wcFullURL, drRead->szCurrentReadURL, MAX_URL);
#endif
		if (!WinHttpCrackUrl(wcFullURL, 0, 0, &urlcCurrent))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			//  Crack failed the URL is probably invalid
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7015E, drRead->szCurrentReadURL);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
		}

		//  Trace the connection attempt
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		wcstombs_s(&iConverted, szTempStr, MAX_URL, wcHostName, _TRUNCATE);
#else
		wcstombs(szTempStr, wcHostName, MAX_URL);
#endif
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7017T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Build the connection
		drRead->hWHConnection = WinHttpConnect(hWHFacility, wcHostName, INTERNET_DEFAULT_HTTP_PORT, 0);
		if (drRead->hWHConnection == NULLHANDLE)
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7018E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;
		}

		//  Trace the connection
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7019T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);
	}

	//  Check if we have already issued the request - if not issue it
	if (drRead->hWHRequest == NULLHANDLE)
	{
		ZeroMemory(&urlcCurrent, sizeof(URL_COMPONENTS));
		urlcCurrent.dwStructSize = sizeof(URL_COMPONENTS);
		urlcCurrent.dwUrlPathLength = sizeof(wcURLPath);		
		urlcCurrent.lpszUrlPath = wcURLPath;

		// Use the URL cracker to extract the URL path from the requested URL
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		mbstowcs_s(&iConverted, wcFullURL, strlen(drRead->szCurrentReadURL) + 1, drRead->szCurrentReadURL, _TRUNCATE);
#else
		mbstowcs(wcFullURL, drRead->szCurrentReadURL, MAX_URL);
#endif
		if (!WinHttpCrackUrl(wcFullURL, 0, 0, &urlcCurrent))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			//  Crack failed the URL is probably invalid
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7020E, drRead->szCurrentReadURL);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		wcstombs_s(&iConverted, szTempStr, MAX_URL, wcURLPath, _TRUNCATE);
#else
		wcstombs(szTempStr, wcURLPath, MAX_URL);
#endif
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7021T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Open the requesst
		drRead->hWHRequest = WinHttpOpenRequest(drRead->hWHConnection,
										L"POST",
										wcURLPath,
										NULL,
										WINHTTP_NO_REFERER,
										WINHTTP_DEFAULT_ACCEPT_TYPES,
										WINHTTP_FLAG_ESCAPE_PERCENT);
		if (drRead->hWHRequest == NULLHANDLE)
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7022E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Trace the request open
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7024T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Format and add the Content TYpe HTTP Header
		sprintf_s(szHeader, MAX_URL, HDR_CONTENT_TYPE, FORM_BOUNDARY);
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
		mbstowcs_s(&iConverted, wcHeader, strlen(szHeader) + 1, szHeader, _TRUNCATE);
#else
		mbstowcs(wcHeader, szHeader, MAX_URL);
#endif

		if (!WinHttpAddRequestHeaders(drRead->hWHRequest, 
									  wcHeader,
                                      (ULONG)-1L,
                                      WINHTTP_ADDREQ_FLAG_ADD | WINHTTP_ADDREQ_FLAG_REPLACE))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7059E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Get the size and lock the payload buffer
		dwPayloadLen = OSMemoryGetSize(hmPayload);
		pPayload = (char *) OSMemoryLock(hmPayload);

		//  Send the request
		if (!WinHttpSendRequest(drRead->hWHRequest,
								WINHTTP_NO_ADDITIONAL_HEADERS,
								0,
								pPayload,
								dwPayloadLen,
								dwPayloadLen,
								0))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7023E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			OSMemoryFree(hmPayload);
			return 0;
		}

		//  Trace the request send
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7025T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Retrieve the request results
		if (!WinHttpReceiveResponse(drRead->hWHRequest, NULL))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7026E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			OSMemoryFree(hmPayload);
			return 0;
		}

		//  Trace Response detected
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7027T, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		OSMemoryFree(hmPayload);

		//  Check the response code
		dwRetLen = sizeof(DWORD);
		if (!WinHttpQueryHeaders(drRead->hWHRequest, 
								WINHTTP_QUERY_STATUS_CODE +	WINHTTP_QUERY_FLAG_NUMBER, 
								WINHTTP_HEADER_NAME_BY_INDEX,
								&dwRespCode,
								&dwRetLen,
								WINHTTP_NO_HEADER_INDEX))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7028E, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Trace the response code received for the request
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7029T, dwRespCode, szTempStr);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Determine if the response code is ok
		if (dwRespCode != HTTP_STATUS_OK)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7030E, dwRespCode, szTempStr);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}
	}

	//  Response data is available
	
	//  Determine if there is residual data left from the previous call to XMLReader
	//  If there is then return that data
	if (drRead->hRespBuffer != NULLHANDLE)
	{
		//  Lock the buffer
		cCache = (char *) OSMemoryLock(drRead->hRespBuffer);
		if (dwMaxToRead >= drRead->dwRespBfrSize)
		{
			//  There is room to store all of the cached data in the XML Reader Buffer
			//  Trace the size of data being returned
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7031T, drRead->dwRespBfrSize);
			xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

			//  Return the data that is available and free the overflow cache
			memcpy(cBuffer, cCache, drRead->dwRespBfrSize);
			dwRetLen = drRead->dwRespBfrSize;
			OSMemoryUnlock(drRead->hRespBuffer);
			OSMemoryFree(drRead->hRespBuffer);
			drRead->hRespBuffer = NULLHANDLE;
			drRead->dwRespBfrSize = 0;
			return dwRetLen;
		}
		else
		{	
			//  Only a part of the cached data can be returned to the XML processor
			//  Trace the partial return of cached data
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7032T, dwMaxToRead, drRead->dwRespBfrSize);
			xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

			//  Return the data 
			memcpy(cBuffer, cCache, dwMaxToRead);
			dwRetLen = drRead->dwRespBfrSize;

			//  Adjust the buffer and resize it
			memcpy(cCache, cCache + dwMaxToRead, drRead->dwRespBfrSize - dwMaxToRead);
			OSMemoryUnlock(drRead->hRespBuffer);
			drRead->hRespBuffer = xeLocal->resizeDominoMemory(drRead->hRespBuffer, drRead->dwRespBfrSize - dwMaxToRead, iThreadID);
			drRead->dwRespBfrSize = drRead->dwRespBfrSize - dwMaxToRead;
			return dwRetLen;
		}
	}

	//  Move returned data to the buffer - if there is more data than buffer space then cache it

	//  Determine the size of data that is available 
	if (!WinHttpQueryDataAvailable(drRead->hWHRequest, &dwRetLen))
	{
		dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
		xeLocal->LogMessage(MSG_DXR7033E, iThreadID);
		xeLocal->GetWINAPIError(dwErrorCode, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		//  Close the request
		WinHttpCloseHandle(drRead->hWHRequest);
		drRead->hWHRequest = NULLHANDLE;
		//  Close the connection
		WinHttpCloseHandle(drRead->hWHConnection);
		drRead->hWHConnection = NULLHANDLE;
		return 0;
	}

	//  If the data available is greater than the buffer passed by the requester - allocate an overflow cache
	if (dwRetLen > dwMaxToRead)
	{
		drRead->hRespBuffer = xeLocal->getDominoMemory(dwRetLen, iThreadID);
		if (drRead->hRespBuffer == NULLHANDLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7034E, dwRetLen);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Lock the buffer
		cCache = (char *) OSMemoryLock(drRead->hRespBuffer);

		//  Read the data into the cache buffer
		drRead->dwRespBfrSize = dwRetLen;
		if (!WinHttpReadData(drRead->hWHRequest, cCache, drRead->dwRespBfrSize, &dwRetLen))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7035E, dwRetLen);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Free the overflow cache buffer
			OSMemoryUnlock(drRead->hRespBuffer);
			OSMemoryFree(drRead->hRespBuffer);
			drRead->hRespBuffer = NULLHANDLE;
			drRead->dwRespBfrSize = 0;
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Trace the fact that response data has been transferred to the overflow cache buffer
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7036T, dwRetLen);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Transfer the max data possible into the requesters buffer and adjust the cache

		//  Trace the partial return of cached data
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7032T, dwMaxToRead, drRead->dwRespBfrSize);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Return the data to the callers buffer
		memcpy(cBuffer, cCache, dwMaxToRead);

		//  Adjust and resize the overflow cache
		dwRetLen = drRead->dwRespBfrSize;
		OSMemoryUnlock(drRead->hRespBuffer);
		drRead->hRespBuffer = xeLocal->resizeDominoMemory(drRead->hRespBuffer, drRead->dwRespBfrSize - dwMaxToRead, iThreadID);
		drRead->dwRespBfrSize = drRead->dwRespBfrSize - dwMaxToRead;
		return dwRetLen;
	}
	else
	{
		//  Check if there is now zero bytes available - this signals the end of the stream
		if (dwRetLen == 0)
		{
			//  Trace the normal end of stream encountered
			xeLocal->LogTrace(TRACE_RESOURCELOADER, MSG_DXR7037T, iThreadID);

			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Return all of the available data - read it directly into the callers buffer
		if (!WinHttpReadData(drRead->hWHRequest, cBuffer, dwMaxToRead, &dwRetLen))
		{
			dwErrorCode = GetLastError();				//  Capture the error code from WinHTTP
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7038E, dwRetLen);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetWINAPIError(dwErrorCode, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
			//  Close the request
			WinHttpCloseHandle(drRead->hWHRequest);
			drRead->hWHRequest = NULLHANDLE;
			//  Close the connection
			WinHttpCloseHandle(drRead->hWHConnection);
			drRead->hWHConnection = NULLHANDLE;
			return 0;
		}

		//  Trace the return of data
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7039T, dwRetLen);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);
		//  Return the number of bytes available
		return dwRetLen;
	}
}
#endif 

#ifdef UNIX
//  FormHTTPWriter				(CURL implementation)
//
//  Function to provide an in memory response stream from an HTTP post
//
//  Parameters
//
//  1  -  MEMHANDLE				-  Handle to the in-memory payload to be posted
//  2  -  char *				-  Pointer to the buffer to use
//  3  -  DWORD					-  Size in bytes of the freespace in the buffer
//  4  -  DXResource *			-  Pointer to the DXResource object that contains the resource content
//  5  -  int					-  ID of the invoking thread
//
//  Returns
//
//  DWORD	-  Number of bytes of data now available in the buffer
//
//  API Usage;
//
//
DWORD DXResourceLoader::FormHTTPWriter(MEMHANDLE hmPayload, char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID)
{
	DWORD				dwRetLen;									//  Length of data to return
	long				lHTTPResponse;								//  HTTP Response Code
	char				*cCache;									//  Cache buffer for returned data
	CURLcode			ccResponse;									//  CURL Error Code
	struct curl_slist	*slHdrs = NULL;								//  List of headers
	char				szHeader[MAX_URL + 1];						//  Header value buffer
	char				szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Check if the CURL component is loaded - if not load it
	if (hcFacility == NULLHANDLE)
	{
		//  Perform the global initialisation
		curl_global_init(CURL_GLOBAL_ALL);
		//  Initiate the CURL session
		hcFacility = curl_easy_init();

		//  Make sure that the DXResource is correctly set up for the initial loading of the response
		drRead->hRespBuffer = NULLHANDLE;
		drRead->dwEnumBufferUsed = 0;
		drRead->iOwningThread = iThreadID;
	}

	//  The current state is detrmined by the state of the resource
	//  If we have not yet loaded the content then do so now
	if (drRead->hRespBuffer == NULLHANDLE)
	{
		//  Use the response buffer members to hold the data to be posted
		drRead->hRespBuffer = hmPayload;
		drRead->dwRespBfrSize = OSMemoryGetSize(hmPayload);
		drRead->dwEnumBufferUsed = 0;
		drRead->iOwningThread = iThreadID;

		//  Set the functionality to POST
		curl_easy_setopt(hcFacility, CURLOPT_POST, 1L);

		//  Set the URL to post
		curl_easy_setopt(hcFacility, CURLOPT_URL, drRead->szCurrentReadURL);

		//  Set the callback to use for loading the post data
		curl_easy_setopt(hcFacility, CURLOPT_READFUNCTION, WrapperCURLHTTPWriter);

		//  Set the data object to be passed to the callback
		curl_easy_setopt(hcFacility, CURLOPT_READDATA, drRead);

		//  Set the callback to use for loading the response
		curl_easy_setopt(hcFacility, CURLOPT_WRITEFUNCTION, WrapperCURLHTTPReader);

		//  Set the data object to be passed to the callback
		curl_easy_setopt(hcFacility, CURLOPT_WRITEDATA, drRead);

		//  Set the User Agent identity
		curl_easy_setopt(hcFacility, CURLOPT_USERAGENT, "DX Resource Loader 2.0");

		//  Set the size of content that will be posted
		curl_easy_setopt(hcFacility, CURLOPT_POSTFIELDSIZE, drRead->dwRespBfrSize);

		//  Set the content type header
		sprintf_s(szHeader, MAX_URL, HDR_CONTENT_TYPE, FORM_BOUNDARY);
		slHdrs = curl_slist_append(slHdrs, szHeader);
		curl_easy_setopt(hcFacility, CURLOPT_HTTPHEADER, slHdrs);

		//  Load the content
		ccResponse = curl_easy_perform(hcFacility);
		if (ccResponse != CURLE_OK)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7023E, drRead->szCurrentReadURL);
			xeLocal->LogMessage(szMsg, iThreadID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7076E, ccResponse, curl_easy_strerror(ccResponse));
			xeLocal->LogMessage(szMsg, iThreadID);
			curl_slist_free_all(slHdrs);
			return 0;							//  Signal end of stream to the caller
		}

		//  Free the header list
		curl_slist_free_all(slHdrs);

		//  Check the HTTP response code
		curl_easy_getinfo(hcFacility, CURLINFO_RESPONSE_CODE, &lHTTPResponse);
		if (lHTTPResponse != 200)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7030E, lHTTPResponse, drRead->szCurrentReadURL); 
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;							//  Signal end of stream to the caller
		}

		//  Check the resource to see if the load succeeded
		if ((drRead->hRespBuffer == NULLHANDLE) || (drRead->dwRespBfrSize == 0))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR7030E, 0, drRead->szCurrentReadURL); 
			xeLocal->LogMessage(szMsg, iThreadID);
			return 0;							//  Signal end of stream to the caller
		}

		//  Update the buffer size to the actual size read
		drRead->dwRespBfrSize = drRead->dwMemOffset;

		//  Show that resource is loaded
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7077T, drRead->szCurrentReadURL, drRead->dwRespBfrSize);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Set the current read offset in the buffer
		drRead->dwMemOffset = 0;
	}

	//  Return data to the caller buffer from the resource response buffer
	if (drRead->dwMemOffset < drRead->dwRespBfrSize)
	{
		//  Compute how many bytes will be returned
		dwRetLen = drRead->dwRespBfrSize - drRead->dwMemOffset;
		if (dwRetLen > dwMaxToRead) dwRetLen = dwMaxToRead;

		//  Lock the buffer
		cCache = (char *) OSMemoryLock(drRead->hRespBuffer);

		//  Trace the partial return of cached data
		sprintf_s(szMsg, MAX_MSG, MSG_DXR7032T, dwRetLen, drRead->dwRespBfrSize);
		xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

		//  Return the data to the callers buffer
		memcpy(cBuffer, cCache, dwRetLen);

		//  Adjust the stream read offset
		drRead->dwMemOffset += dwRetLen;

		//  Signal the return of data
		return dwRetLen;
	}

	//  Show nothing is left to read -- i.e. END-OF-STREAM
	return 0;
}

//  WrapperCURLHTTPWriter
//
//  Callback from CURL to pass memory content 
//
//  Parameters
//
//  1  -  void *			- Pointer to the buffer to fill
//  2  -  size_t			- Size in bytes of the Buffer space
//  3  -  size_t			- Number of Buffers
//  4  -  void *			- Pointer to the DXResource object that contains the resource info
//
//  Returns
//
//  size_t	-  Number of bytes of data added to the CURL Memory Buffer
//
size_t DXResourceLoader::WrapperCURLHTTPWriter(void *pBuffer, size_t iBfrSize, size_t iBuffers, void *pContext)
{
	size_t				iTotalBytes;					//  Total size of the data buffer made available
	DWORD				dwBytesAvailable;				//  Bytes Available
	char				*cTarget;						//  Pointer to the source buffer
	DXResource			*drCurrent;						//  Passed DXResource
	DXResourceLoader	*rlCurrent;						//  Resource Loader

	//  Compute the total bytes available
	iTotalBytes = iBfrSize * iBuffers;

	//  Cast the context to the passed DXResource
	drCurrent = (DXResource *) pContext;
	rlCurrent = (DXResourceLoader *) drCurrent->pRLoader;

	//  Cast/Assign the parameters
	cTarget = (char *) pBuffer;
	dwBytesAvailable = iTotalBytes;

	//  Load the returned data to the CURL Stream
	iTotalBytes = rlCurrent->CURLHTTPWriter(cTarget, dwBytesAvailable, drCurrent, drCurrent->iOwningThread);
	return iTotalBytes;
}

//  CURLHTTPWriter
//
//  Callback from CURL to load returned data from local memory
//
//  Parameters
//
//  1  -  char *			- Pointer to the buffer to loads
//  2  -  DWORD				- Size in bytes of the available space in the Buffer
//  3  -  DXResource *		- Pointer to the DXResource object that contains the resource info
//	4  -  int				- ID of the calling thread
//
//  Returns
//
//  size_t	-  Number of bytes of data added to the local memory
//
size_t DXResourceLoader::CURLHTTPWriter(char *cBuffer, DWORD dwBytesAvailable, DXResource *drRead, int iThreadID)
{
	char			*cCache;									//  Pointer to the cache buffer
	DWORD			dwBytesToCopy;								//  Size of memory to copy
	char			szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Calculate the number of bytes to copy
	dwBytesToCopy = drRead->dwRespBfrSize - drRead->dwEnumBufferUsed;
	if (dwBytesToCopy > dwBytesAvailable) dwBytesToCopy = dwBytesAvailable;

	//  Trace the fact that stream data has been transferred
	sprintf_s(szMsg, MAX_MSG, MSG_DXR7078T, dwBytesToCopy);
	xeLocal->LogTrace(TRACE_RESOURCELOADER, szMsg, iThreadID);

	//  Check for the boundary condition where end of stream has been reached then reset the resource ready for response processing
	if (dwBytesToCopy == 0)
	{
		OSMemoryFree(drRead->hRespBuffer);
		drRead->hRespBuffer = NULLHANDLE;
		drRead->dwRespBfrSize = 0;
		drRead->dwEnumBufferUsed = 0;
		return 0;									//  Signal end-of-stream
	}

	//  Lock the buffer
	cCache = (char *) OSMemoryLock(drRead->hRespBuffer);

	//  Write the data into the CURL buffer
	memcpy(cBuffer, cCache + drRead->dwEnumBufferUsed, dwBytesToCopy);

	//  Unlock the buffer
	OSMemoryUnlock(drRead->hRespBuffer);

	//  Adjust the offset
	drRead->dwEnumBufferUsed += dwBytesToCopy;

	//  return to the caller
	return dwBytesToCopy;
}
#endif

