#pragma once
//  DXResourceLoader
//
//  Resource Loader (usually XML documents) for DX applications. The resource loader is capable of loading an XML
//  document or other resource from a remote HTTP source, a native file, a resource embedded in the current application,
//  a Text List field in a Note or a Rich Text Item in a Note. The resource is loaded into a Domino memory allocation
//  and a handle (unlocked) is returned to the caller.
//  Files can be uploaded to a new Domino document.
//
//  Message Identifier:      DXR
//  Message Range:           7001 - 7099
//  Trace area assignment:   20
//  Version:                 3.14.0			Build:  108
//
//  Version History:
//
//  1.0.0		12/08/2009		Initial version
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.3		17/12/2010		Spring cleaning
//  3.10.6		19/03/2011		Upload capability added to DXResourceLoader
//  3.10.9		04/06/2011		CURL implementation for HTTP on UNIX
//  3.11.1		22/10/2011		VS 2011 Updates
//	3.14.0		21/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2009 - 20115
//

//  Runtime Includes
#include "ExecEnvironment.h"
#include "DXGlobals.h"
#include "DXResource.h"
#include "DXUploadContext.h"

//  Additional Platform includes
#ifdef WIN32
#include	<winhttp.h>											//  WinHTTP API
#endif
#ifdef UNIX
#include	<curl/curl.h>										//  CURL API
#endif

//  Additional Notes API includes
#include	<editods.h>											//  CD Record definitions

//  Error Messages
#define		MSG_DXR7003E				"DXR7003E: Resource source '%s' is of an unrecognized or invalid type."
#define		MSG_DXR7005E				"DXR7005E: Application resource source '%s' could not be located."
#define		MSG_DXR7006E				"DXR7006E: Application resource source '%s' could not be loaded."
#define		MSG_DXR7007E				"DXR7007E: Application resource source '%s' was loaded with zero size."
#define		MSG_DXR7008E				"DXR7008E: Unable to allocate %i bytes of memory for resource source '%s'."
#define		MSG_DXR7010E				"DXR7010E: Unable to open file '%s' because: %s."
#define		MSG_DXR7013E				"DXR7013E: Unable to grow a memory buffer to %i bytes for resource source '%s'."
#define		MSG_DXR7014E				"DXR7014E: Unable to initialise the WinHTTP component."
#define		MSG_DXR7015E				"DXR7015E: Unable to crack the requested URL: %s."
#define		MSG_DXR7018E				"DXR7018E: Unable to make the http connection to host: %s."
#define		MSG_DXR7020E				"DXR7020E: Unable to crack the requested URL: %s."
#define		MSG_DXR7022E				"DXR7022E: Unable to open the http request for: %s."
#define		MSG_DXR7023E				"DXR7023E: Unable to send the http request for: %s."
#define		MSG_DXR7026E				"DXR7026E: Unable to get a response to the http request for: %s."
#define		MSG_DXR7028E				"DXR7028E: Unable to query the response code for the http request for: %s."
#define		MSG_DXR7030E				"DXR7030E: A response code of %i was received from the http request for: %s."
#define		MSG_DXR7033E				"DXR7033E: Unable to determine if data is available in response to a http request."
#define		MSG_DXR7034E				"DXR7034E: Failed to allocate an overflow cache memory buffer of %i bytes."
#define		MSG_DXR7035E				"DXR7035E: Failed to read %i bytes of response data into the overflow cache buffer."
#define		MSG_DXR7038E				"DXR7038E: Failed to read %i bytes of response data into the callers buffer."
#define		MSG_DXR7040E				"DXR7040E: A request was made to load a resource from an unnamed Notes Item."
#define		MSG_DXR7041E				"DXR7041E: Unable to obtain information about Notes Item: '%s'."
#define		MSG_DXR7042E				"DXR7042E: Item '%s' is of a type that is not supported by the Resource Loader."
#define		MSG_DXR7043E				"DXR7043E: Unable to allocate %i bytes of memory for resource source from item '%s'."
#define		MSG_DXR7048E				"DXR7048E: Failed enumeration of Rich Text Item '%s'."
#define		MSG_DXR7050E				"DXR7050E: Error encountered reading '%s' - %s (RQ: %i - Ret: %i)."
#define		MSG_DXR7051E				"DXR7051E: Error encountered while seeking Start/EOF in file '%s' - %s."
#define		MSG_DXR7053E				"DXR7053E: Unable to service an On Demand request because there is no On Demand handler for this resource type."
#define		MSG_DXR7055E				"DXR7055E: Unable to create an On Demand request for an empty memory handle."
#define		MSG_DXR7059E				"DXR7059E: Unable to add a http header to the http request for: %s."
#define		MSG_DXR7060E				"DXR7060E: A file upload has been requested with an empty file name."
#define		MSG_DXR7061E				"DXR7061E: A file upload has been requested and no memory handle has been passed."
#define		MSG_DXR7062E				"DXR7062E: A file upload has been requested and no upload context has been passed."
#define		MSG_DXR7063E				"DXR7063E: A file upload has been requested but no form url was specified in the upload context."
#define		MSG_DXR7064E				"DXR7064E: Unable to open the requested form '%s' to upload a file."
#define		MSG_DXR7065E				"DXR7065E: The specified form does not contain a file upload control."
#define		MSG_DXR7066E				"DXR7066E: The specified form does not contain a '%s' action to allow submission."
#define		MSG_DXR7067E				"DXR7067E: The specified form does not contain a Post URL."
#define		MSG_DXR7068E				"DXR7068E: Failed to construct a file upload payload."
#define		MSG_DXR7069E				"DXR7069E: Failed to allocate a %i byte buffer for the payload."
#define		MSG_DXR7075E				"DXR7075E: Resource: '%s' could not be loaded on this platform."
#define		MSG_DXR7076E				"DXR7076E: CURL Error: (%i) - %s."

//  Warning Messages

//  Information Messages
#define		MSG_DXR7004I				"DXR7004I: Loading resource source '%s' of type %i."
#define		MSG_DXR7009I				"DXR7009I: Resource source '%s' has been loaded, Length = %i bytes."
#define		MSG_DXR7054I				"DXR7054I: Resource source '%s' has been prepared for loading On Demand."
#define		MSG_DXR7056I				"DXR7056I: Memory Resource @0x%.4lX has been prepared for loading On Demand."
#define		MSG_DXR7072I				"DXR7072I: The payload for a file upload has been constructed (%i bytes)."
#define		MSG_DXR7073I				"DXR7073I: Posting the current payload to: '%s'."

//  Trace Messages
#define		MSG_DXR7001T				"DXR7001T: A new DXResourceLoader instance has been created @%p."
#define		MSG_DXR7002T				"DXR7002T: A DXResourceLoader instance @%p has been destroyed."
#define		MSG_DXR7016T				"DXR7016T: The WinHTTP component has been initialised in config: %i."
#define		MSG_DXR7017T				"DXR0017T: Attempting connection to host: %s."
#define		MSG_DXR7019T				"DXR0019T: Successfully configured connection to host: %s."
#define		MSG_DXR7021T				"DXR7021T: Attempting request for resource: %s from host."
#define		MSG_DXR7024T				"DXR7024T: Successfully opened the request for: %s."
#define		MSG_DXR7025T				"DXR7025T: Successfully sent the request for: %s."
#define		MSG_DXR7027T				"DXR7027T: A response has been detected for the request for: %s."
#define		MSG_DXR7029T				"DXR7029E: A response code of %i was received from the http request for: %s."
#define		MSG_DXR7031T				"DXR7031T: Returning %i bytes to the buffer from overflow cache."
#define		MSG_DXR7032T				"DXR7032T: Returning %i bytes (of %i bytes) to the buffer from overflow cache."
#define		MSG_DXR7036T				"DXR7036T: %i bytes of response data have been transferred to the overflow cache buffer."
#define		MSG_DXR7037T				"DXR7037T: End of content stream has been processed normally."
#define		MSG_DXR7039T				"DXR7039T: Returning %i bytes to the buffer from response data."
#define		MSG_DXR7044T				"DXR7044T: Building resource from item '%s' with %i entries in %i bytes."
#define		MSG_DXR7045T				"DXR7045T: Memory Buffer has been loaded from item '%s' with %i lines in %i bytes."
#define		MSG_DXR7046T				"DXR7046T: Starting enumeration of Rich Text Item '%s', Buffersize = %i."
#define		MSG_DXR7047T				"DXR7047T: Completed enumeration of Rich Text Item '%s', Text Buffersize = %i."
#define		MSG_DXR7049T				"DXR7049T: About to load file '%s' containing %i bytes of content."
#define		MSG_DXR7052T				"DXR7052T: Attempting to re-drive the load request using the WinHTTP Proxy settings."
#define		MSG_DXR7057T				"DXR7057T: A new cloned resource @%p has been created from the resource @%p."
#define		MSG_DXR7058T				"DXR7058T: A resource @%p has been disposed of."
#define		MSG_DXR7070T				"DXR7070T: Appended Item '%s' : '%s' to payload (%i bytes)."
#define		MSG_DXR7071T				"DXR7071T: Appended File '%s' (%i bytes) plus %i bytes of prefix and suffix data."
#define		MSG_DXR7074T				"DXR7074T: File '%s' has been loaded into memory size = %i bytes."
#define		MSG_DXR7077T				"DXR7077T: Resource source '%s' has loaded a buffer, Length = %i bytes."
#define		MSG_DXR7078T				"DXR7078T: %i bytes have been transferred into the POST buffer."

//  Miscellaneous Definitions
#define		FILETYPE_URI				"file://"				//  A File type URI
#define		GENERICTYPE_URI				"://"					//  A Generic URI
#define		APPRSRC_TYPE				"DXRESOURCE"			//  DX Resource type
#define		FORM_SUB_ACTION				"Submit"				//  Submit action on a form
#define		FORM_BOUNDARY				"---------------------------7db2deb310b5"  //  Mulipart Boundary
#define		FORM_CONTENT_DESC			"Content-Disposition: form-data; name=\"%s\"\r\n\r\n"  //  Multipart content descriptor
#define		FORM_FILE_DESC				"Content-Disposition: form-data; name=\"%s\"; filename=\"%s\"\r\n"  //  File descriptor
#define		FORM_CONTENT_TYPE			"Content-Type: %s\r\n\r\n"	//  File content type
#define		FILE_TYPE_TEXT				"text/plain"			//  Text
#define		FILE_TYPE_XML				"text/xml"				//  XML
#define		FILE_TYPE_BINARY			"application/octet-stream"		//  Binary stream
#define		HDR_CONTENT_TYPE			"Content-Type: multipart/form-data; boundary=%s"	//  Content type HTTP Header
#ifndef     EOK
#define		EOK							0						//  Copacetic error code
#endif
#define		DEF_HTTP_RCV_BFRSIZE		1024 * 32				//  32k default buffer for HTTP receives
#define		INC_HTTP_RCV_BFRSIZE		1024 * 8				//  8k increments for growing the HTTP receive buffer
#define		MIN_FREE_BUFFERSPACE		1024 * 4				//  Minimum free memory in the buffer for HTTP receives
#define		DEF_FILE_RCV_BFRSIZE		1024 * 48				//  48k default buffer for FILE receives
#define		INC_FILE_RCV_BFRSIZE		1024 * 12				//  12k increments for growing the FILE receive buffer
#define		PAYLOAD_BFR_OHEAD			(1024 * 16)				//  16k Payload overhead for fields

//  Class Definition
class DXResourceLoader
{
public:
	//  Constructor
	DXResourceLoader(ExecEnvironment *xeParent, int iThreadID);

	//  Destructor
	~DXResourceLoader(void);

	//  Public Members

	//  Public Functions
	MEMHANDLE loadThisResource(char *szResourceSource, int iThreadID);						//  Load the resource from specified source
	MEMHANDLE loadThisResource(NOTEHANDLE hnSource, char *szItemName, int iThreadID);		//  Load the resource from the specified Item
	DXRESOURCEHANDLE loadThisResourceOnDemand(char *szResourceSource, int iThreadID);		//  Load the resource from specified source On Demand
	DXRESOURCEHANDLE loadThisResourceOnDemand(NOTEHANDLE hnSource, char *szItemName, int iThreadID);		//  Load the resource from the specified Item On Demand
	DXRESOURCEHANDLE loadThisResourceOnDemand(MEMHANDLE hResource, int iThreadID);			//  Load the resource from memory on-demand
	DXRESOURCEHANDLE cloneResourceHandle(DXRESOURCEHANDLE hdrPrime, int iThreadID);			//  Clone a resource handle
	MEMHANDLE uploadThisResource(char *szFileName, int iFileType, DXUploadContext *ducThis, int iThreadID);  //  Upload a resource by name
	MEMHANDLE uploadThisResource(MEMHANDLE hmResource, int iFileType, DXUploadContext *ducThis, int iThreadID);  //  Upload a resource by handle
	void disposeResourceHandle(DXRESOURCEHANDLE hdrUsed, int iThreadID);					//  Dispose of a resource handle
	static STATUS LNCALLBACK WrapperprocessThisCDRec(char *pCDRec, WORD wCDType, DWORD dwCDLen, void *pContext);
	static DWORD LNCALLBACK WrapperXMLReader(unsigned char *ucBuffer, DWORD dwMaxToRead, void *pContext);
#ifdef UNIX
	static size_t WrapperCURLHTTPReader(void *pBuffer, size_t iBfrSize, size_t iBuffers, void *pContext);  //  CURL Reader Exit
	static size_t WrapperCURLHTTPWriter(void *pBuffer, size_t iBfrSize, size_t iBuffers, void *pContext);  //  CURL Writer Exit
#endif

private:
	//  Private Members
	ExecEnvironment	*xeLocal;									//  DX Runtime
	int			ParentThread;									//  Owning Thread
#ifdef WIN32
	HINTERNET hWHFacility;										//  Handle to the WinHTTP Component
#endif
#ifdef UNIX
	CURL *hcFacility;											//  Handle to the CURL Session
#endif

	//  Private Functions
	int determineSourceType(char *szResourceSource);			//  Determine the resource source type
	MEMHANDLE loadResourceFromApplication(char *szResourceID, int iThreadID);	//  Load the specified Application Resource by ID
	MEMHANDLE loadResourceFromFile(char *szFileName, int iThreadID);			//  Load the specified File Resource
	MEMHANDLE loadResourceFromURL(char *szURL, int iThreadID);					//  Load the specified URL Resource
	MEMHANDLE loadResourceFromText(BLOCKID bidVal, DWORD dwLen, char *szItem, int iThreadID);   //  Load the Resource from a text list field
	MEMHANDLE loadResourceFromRTI(BLOCKID bidVal, DWORD dwLen, char *szItem, int iThreadID);	//  Load the Resource from a Rich Text field
	MEMHANDLE postFormToURL(MEMHANDLE hmPayload, char *szURL, int iThreadID);	//  Post a form to a URL
	DXRESOURCEHANDLE loadResourceOnDemandFromURL(char *szURL, int iThreadID);					//  Load the specified URL Resource On Demand
	DXRESOURCEHANDLE loadResourceOnDemandFromFile(char *szFileName, int iThreadID);					//  Load the specified URL Resource On Demand
	DXRESOURCEHANDLE loadResourceOnDemandFromMemory(MEMHANDLE hResMem, int iThreadID);			//  Load the specified URL Resource On Demand

	DWORD XMLHTTPReader(char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID);		//  XML Reader routine
	DWORD XMLMEMReader(char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID);		//  XML Reader routine
	DWORD XMLFILEReader(char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID);		//  XML Reader routine
#ifdef UNIX
	size_t CURLHTTPReader(char *cBuffer, DWORD dwBytesAvailable, DXResource *drRead, int iThreadID);		//  CURL HTTP Reader routine
	size_t CURLHTTPWriter(char *cBuffer, DWORD dwBytesAvailable, DXResource *drPost, int iThreadID);		//  CURL HTTP Writer routine
#endif
	DWORD FormHTTPWriter(MEMHANDLE hmPayload, char *cBuffer, DWORD dwMaxToRead, DXResource *drRead, int iThreadID);		//  Form Posting Routine
	STATUS processThisCDRec(char *pCDRec, WORD wCDType, DWORD dwCDLen, DXResource *drRead);			//  CD Record enumerate action routine
	void setConnectConfig(DXResource *drRead);									//  Setup the connection configuration
#ifdef WIN32
	DWORD retryXMLReader(char *cBuffer, DWORD dwMaxToRead, DWORD dwError, DXResource *drRead, int iThreadID);  //  Retry the XML Read (if possible)
#endif
	void getFormAction(char *pForm, DWORD dwLength, char *szPostURL, char *szFileUpload, char *szClickAction);  //  Get the form actions
	char *findElement(char *pSpace, DWORD dwSize, char *SearchString);			//  Find the specified element in a search space
	MEMHANDLE buildPayload(char *szCA, char *szFUN, DXUploadContext *ducThis, int iFileType, MEMHANDLE hmFile, int iThreadID);  //  Assemble upload payload
	DWORD appendUCItem(char *pBfr, DXUCItem *uciThis, int iThreadID);			//  Append a new item to the buffer
	DWORD appendFile(char *pBfr, char *szFUN, char *szFN, int iFT, MEMHANDLE hmFile, DWORD *dxPfxLen, int iThreadID);   //  Append a file to the payload buffer
};
