//  ElapsedTimer
//
//  This class provides a means of deriving elapsed real (Wall Clock) times.
//  Calls to the getElapsed() function will return the number of "Ticks" since the object was created.
//  Calls to the getElapsedMillis() function will return the number of milliseconds since the object was created.
//  ExecEnvironment implements a default ElapsedTimer (->RunningTime).
//
//  Message Identifier:      N/A
//  Message Range:           0000 - 0000
//  Trace area assignment:   N/A
//  DX Version:				 3.11.6
//  Version:                 3.11.6  Build: 104
//
//  Version History:
//
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2011
//

//  Runtime Includes
#include "ElapsedTimer.h"

//  Constructor
ElapsedTimer::ElapsedTimer()
{
	//  Record the starting time point
#ifdef WIN32
	StartTime = clock();												//  Initial Time Value
#endif
#ifdef UNIX
	gettimeofday(&StartTime, NULL);
#endif
}

//  Destructor
ElapsedTimer::~ElapsedTimer()
{
}

//  getElapsed
//
//  The getElapsed() function will return the number of "Ticks" since the object was created.
//
//  Parameters
//
//  Returns
//  clock_t			-			Number of elapsed "Ticks" since the object was created
//
//  API Use Notes:
//
//
clock_t ElapsedTimer::getElapsed()
{
	long			lTimeFactor = CLOCKS_PER_SEC;			//  Clock tick conversion factor
	clock_t			tResult;								//  Result
#ifdef WIN32
	clock_t			tCurrentTime;							//  Current Time
#endif
#ifdef UNIX
	timeval			tvCurrentTime;							//  Current Time
#endif

	//  Get the latest time
#ifdef WIN32
	tCurrentTime = clock();										//  Get the current running time
#endif
#ifdef UNIX
	gettimeofday(&tvCurrentTime, NULL);					//  Get the current Epoch Time
#endif

	//  Compute the result i.e. time difference between the current & start times
#ifdef WIN32
	tResult = tCurrentTime - StartTime;							//  Compute the difference
#endif
#ifdef UNIX
	if (StartTime.tv_usec > tvCurrentTime.tv_usec)
	{
		tvCurrentTime.tv_usec += 1000000;						//  Borrow a second
		tvCurrentTime.tv_sec--;
	}
	tvCurrentTime.tv_usec = tvCurrentTime.tv_usec - StartTime.tv_usec;
	tvCurrentTime.tv_sec = tvCurrentTime.tv_sec - StartTime.tv_sec;

	//  Convert the microseconds difference to ticks
	tResult = tvCurrentTime.tv_usec * (lTimeFactor / 1000000);
	//  Add the seconds difference in ticks
	tResult = tResult + (tvCurrentTime.tv_sec * lTimeFactor);
#endif

	return tResult;
}

//  getElapsedMillis
//
//  The getElapsedMillis() function will return the number of milliseconds since the object was created.
//
//  Parameters
//
//  Returns
//
//  clock_t			-			Number of elapsed milliseconds since the object was created
//
//  API Use Notes:
//
//
clock_t ElapsedTimer::getElapsedMillis()
{
	long			lTimeFactor = CLOCKS_PER_SEC;			//  Clock tick conversion factor
	clock_t			tResult;								//  Result
#ifdef WIN32
	clock_t			tCurrentTime;							//  Current Time
#endif
#ifdef UNIX
	timeval			tvCurrentTime;							//  Current Time
#endif

	//  Get the latest time
#ifdef WIN32
	tCurrentTime = clock();									//  Get the current running time
#endif
#ifdef UNIX
	gettimeofday(&tvCurrentTime, NULL);						//  Get the current Epoch Time
#endif

	lTimeFactor = lTimeFactor / 1000;						//  Conversion for ticks per millisecond

	//  Compute the result i.e. time difference between the current & start times
#ifdef WIN32
	tResult = tCurrentTime - StartTime;							//  Compute the difference
	tResult = tResult / lTimeFactor;							//  Diference in milliseconds 
#endif
#ifdef UNIX
	if (StartTime.tv_usec > tvCurrentTime.tv_usec)
	{
		tvCurrentTime.tv_usec += 1000000;						//  Borrow a second
		tvCurrentTime.tv_sec--;
	}
	tvCurrentTime.tv_usec = tvCurrentTime.tv_usec - StartTime.tv_usec;
	tvCurrentTime.tv_sec = tvCurrentTime.tv_sec - StartTime.tv_sec;

	//  Convert the microseconds difference to milliseconds
	tResult = tvCurrentTime.tv_usec / 1000;
	//  Add the seconds difference in ticks
	tResult = tResult + (tvCurrentTime.tv_sec * 1000);
#endif

	return tResult;
}
