//  DXDbScanner
//
//  This class is the eXplorer class that handles the scanning of the database level in the hierarchy according to
//  the scope and constraints from the search request.
//
//  Message Identifier:      DXR
//  Message Range:           1401 - 1499
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.11.1
//  Version:                 3.11.1  Build: 99
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.5		18/02/2011		Report on D/Bs which cannot be opened
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2011
//

#include "DXDbScanner.h"

//  Multi Threaded Constructor
DXDbScanner::DXDbScanner(MTExecutive *xeParent, int iThreadID)
:xeLocal(NULL)
{
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Alternate Constructor
	xeLocal = xeParent;										//  Store the Execution Environment
	iParentThread = iThreadID;								//  Store the owning thread ID

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1400T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);
}

//  Destructor
DXDbScanner::~DXDbScanner(void)
{
	char				szMsg[MAX_MSG];					//  Generic message buffer

	//  Trace the destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1412T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);
}

//  ProcessRequest
//
//  This function is the main entry point for dispatching a Domino eXplorer request.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the invoking Thread
//
//  Returns
//
//  BOOL  -			TRUE if the scan was successful
//                  FALSE if the scan was not successful
//
//  API Use Notes:
//
//
BOOL DXDbScanner::ProcessRequest(DXRequest *dxrRQ, int iThreadID)
{
	BOOL			bResult;								//  Return from Core Call

	//  Entry point used for single threaded request execution
	
	bResult = CoreProcessRequest(dxrRQ, iThreadID);
	return bResult; 
}

//  DestroySubContexts
//
//  This function will destroy any scan contexts below the level of this scanner
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void DXDbScanner::DestroySubContexts(void)
{
	return;
}

//  GetDbInfo
//
//  This function will obtain information from the target database to update the eXplorer request
//
//  Parameters
//
//  1  -  DBHANDLE		-  DBHANDLE of the database from which to get information
//  2  -  DXRequest *	-  Pointer to the DX Request to fill the data in
//  3  -  int			-  ID of the invoking Thread
//
//  Returns
//
//  BOOL		-  TRUE if the information could be completed
//				   FALSE is the information could not be completed
//
//  API Use Notes:
//
//
BOOL DXDbScanner::GetDbInfo(DBHANDLE hdbTarget, DXRequest *dxrRQ, int iThreadID)
{
	//  Get Database Information

	//  If the Server Name is blank i.e. Local then put the current user name in the request 
	if (strlen(dxrRQ->szServer) == 0)
	{
		strcpy_s(dxrRQ->szServer, MAX_SERVER, xeLocal->AbbreviatedUser);
	}

	return TRUE;
}

//  CoreProcessRequest
//
//  This function is the main function for processing an eXplorer request at the Database level
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the invoking Thread
//
//  Returns
//
//  BOOL  -			TRUE if the scan was successful
//                  FALSE if the scan was not successful
//
//  API Use Notes:
//
//
BOOL DXDbScanner::CoreProcessRequest(DXRequest *dxrRQ, int iThreadID)
{
	DBHANDLE	hdbTarget;								//  Handle to the current target database
	USHORT		usDbType;								//  Database type returned by NSFDbGetMode
	DXReporter  *dxrepDb;								//  Address of the D/B Reporter Exit Object
	char		szDbPath[MAX_DATABASE + 1];				//  Constructed D/B Name
	char		szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Process a scan request for a database.
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1417T, dxrRQ, this, iThreadID);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);

	try
	{
		//  Check that the Global Run Permit is still asserted
		if (!xeLocal->TestGlobalPermit())
		{
			xeLocal->LogMessage(MSG_DXR1420E, iThreadID);
			return FALSE;
		}

		//  All requests to this level require the database to be open.
		//  Open the database 
		if (dxrRQ->szDirectory[0] != '\0')
		{
			sprintf_s(szDbPath, MAX_DATABASE, "%s/%s", dxrRQ->szDirectory, dxrRQ->szDatabase);
		}
		else
		{
			strcpy_s(szDbPath, MAX_DATABASE, dxrRQ->szDatabase);
		}
		hdbTarget = xeLocal->OpenDatabase(dxrRQ->szServer, szDbPath, &usDbType, iThreadID);
		if (hdbTarget == NULLHANDLE)
		{
			//  Unable to access the requested directory
			//  Pass the request to the NoAccess reporter - if needed
		
			//  Check if a D/B Reporter is needed
			if (dxrRQ->Reporters & DXREP_DATABASE)
			{
				DXReporter   *dxrepDb;

				//  D/B level reporter is required, invoke it
				dxrepDb = (DXReporter *) dxrRQ->dxrDatabase;
				dxrepDb->ReportNoAccess(dxrRQ, iThreadID);
			}
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1402E, dxrRQ->szDatabase, dxrRQ->szDirectory);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		if (usDbType == DB_DIRECTORY)
		{
			//  Requested item is a directory
			NSFDbClose(hdbTarget);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1403E, szDbPath);
			xeLocal->LogMessage(szMsg, iThreadID);
			return FALSE;
		}

		//  Capture supplemental information for the request
		dxrRQ->SetStart(xeLocal);
		//  Show that we are starting the scan
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1404I, dxrRQ->szDatabase, dxrRQ->szDirectory);
		xeLocal->LogVerbose(szMsg, iThreadID);

		//  Accumulate the scores
		dxrRQ->dwDatabasesProcessed++;

		//  Process the database through the appopriate reporters
		if (!GetDbInfo(hdbTarget, dxrRQ, iThreadID))
		{
			//  Close the database if it was opened at this level
			if (hdbTarget != NULLHANDLE)
			{
				NSFDbClose(hdbTarget);
				hdbTarget = NULLHANDLE;
			}
			return FALSE;
		}

		//  Check if a D/B Reporter is needed
		if (dxrRQ->Reporters & DXREP_DATABASE)
		{
			//  D/B level reporter is required, invoke it
			dxrepDb = (DXReporter *) dxrRQ->dxrDatabase;
			dxrepDb->ReportOnThisDb(dxrRQ, hdbTarget, iThreadID);
		}

		//  Close the database if it was opened at this level
		if (hdbTarget != NULLHANDLE)
		{
			//  In case the reporter has invoked multi-threaded processing then we need to invalidate the 
			//  file handle to cause the thread to purge the thread local entries
			xeLocal->InvalidateNativeDBH(hdbTarget, iThreadID);
			NSFDbClose(hdbTarget);
			hdbTarget = NULLHANDLE;
		}

		//  Show that the scan has been completed
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1405I, dxrRQ->szDatabase, dxrRQ->szDirectory);
		xeLocal->LogVerbose(szMsg, iThreadID);
		//  Document the request
		dxrRQ->SetEnd(xeLocal);
		if (xeLocal->rsLocal->LogLevel > LOGLEVEL_VERBOSE) dxrRQ->DocumentRequest(xeLocal);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1418T, dxrRQ->ReturnCode, dxrRQ, this, iThreadID);
		xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
		return TRUE;
	}
	catch(DXException& dxeCurrent)
	{
		//  Close the database if it was opened at this level
		if (hdbTarget != NULLHANDLE)
		{
			xeLocal->InvalidateNativeDBH(hdbTarget, iThreadID);
			NSFDbClose(hdbTarget);
			hdbTarget = NULLHANDLE;
		}

		//  Destroy the lower level scanners if they are unstable
		if (dxeCurrent.DestroyContext)
		{
			DestroySubContexts();
		}

		//  Mark the Exception as percolated
		dxeCurrent.IsPercolated = true;

		//  Rethrow the exception to the next level - percolate
		throw;
	}
	catch(std::exception& eCurrent)
	{
		DXException		*dxeLocal;

		//  Close the database if it was opened at this level
		if (hdbTarget != NULLHANDLE)
		{
			xeLocal->InvalidateNativeDBH(hdbTarget, iThreadID);
			NSFDbClose(hdbTarget);
			hdbTarget = NULLHANDLE;
		}

		//  This level and lower levels may be unstable - destroy them
		DestroySubContexts();

		//  Create a new exception to wrapper the information from the original
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_UHX;
		dxeLocal->ExceptionCode = ERRORCODE_NON_SPEC;
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->LocalExceptionMessage, MAX_MSG, eCurrent.what());
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1401E);
		dxeLocal->DestroyContext = true;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
		dxeLocal->LineNumber = __LINE__;
		throw(*dxeLocal);
	}
	return FALSE;
}

//  ExecuteThisRequest
//
//  This is the entry point used by the multi-threading kernel to trigger processing in different threadpool threads.
//
//  Parameters
//
//  1  -  void *			-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DXDbScanner::ExecuteThisRequest(void *pReqObject, int iThreadID)
{
	BOOL				bResult;							//  Result from invocation
	DXRequest			*dxrRQ;								//  Request object

	dxrRQ = (DXRequest *) pReqObject;
	bResult = ProcessRequest(dxrRQ, iThreadID);
	return;
}

