#pragma once
//  DXDirScanner
//
//  This class is the eXplorer class that handles the scanning of the directory level in the hierarchy according to
//  the scope and constraints from the search request.
//
//  Message Identifier:      DXR
//  Message Range:           1301 - 1399
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.11.5
//  Version:                 3.11.5  Build: 103
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.9.2		22/09/2010		Avoided a recursion in DX processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.5		18/02/2011		Report on D/Bs which cannot be opened
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.5		30/11/2011		Recursion avoidance improvements
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2011
//

//  Application Includes
#include "../MTX/MTExecutive.h"					//  Runtime environment
#include "../DXException.h"
#include "DXFilter.h"
#include "DXRequest.h"
#include "DXDbScanner.h"

//  Additional API Includes
#include <nsfsearc.h>

//  Trace Messages
#define		MSG_DXR1307T		"DXR1307T: Determining search scope for directory: '%s' for match: '%s'."
#define		MSG_DXR1311T		"DXR1311T: NSFSearch returned matched entry '%s'."
#define		MSG_DXR1316I		"DXR1316I: Explosion completed for database '%s'."
#define		MSG_DXR1318T		"DXR1318T: Scan Directory function is exiting."
#define		MSG_DXR1319T		"DXR1319T: A new DXDirScanner instance has been created @%p."
#define		MSG_DXR1320T		"DXR1320T: Directory scanner is passing request to a lower level."
#define		MSG_DXR1321T		"DXR1321T: A DXDirScanner instance @%p has been destroyed."
#define		MSG_DXR1325T		"DXR1325T: Entity: '%s' has been identified as mode: %i."
#define		MSG_DXR1326T		"DXR1326T: Processing directory '%s' at level: %i for match: '%s'."
#define		MSG_DXR1327T		"DXR1327T: Restricting search in '%s' to directories only."
#define		MSG_DXR1329T		"DXR1329T: DX Request @%p with scope: %i has been dispatched at priority: %i."
#define		MSG_DXR1330T		"DXR1330T: DX Request @%p has rejoined for parent %p."

//  Information Messages
#define		MSG_DXR1303I		"DXR1303I: Starting scan of directory '%s'."
#define		MSG_DXR1304I		"DXR1304I: Scan completed for directory '%s'."
#define		MSG_DXR1314I		"DXR1314I: Explosion completed for directory '%s'."
#define		MSG_DXR1323I		"DXR1323I: Directory '%s' has been dropped as it is unmatched or filtered."
#define		MSG_DXR1324I		"DXR1324I: Database '%s/%s' has been dropped as it is unmatched or filtered."

//  Information Messages
#define		MSG_DXR1301E		"DXR1301E: An error was encountered while scanning directory '%s'."
#define		MSG_DXR1302E		"DXR1302E: An unspecified exception has been trapped."
#define		MSG_DXR1305E		"DXR1305E: Unable to open entity '%s' to explode it."
#define		MSG_DXR1308E		"DXR1308E: Directory scan was requested for '%s', which is not a directory."
#define		MSG_DXR1309E		"DXR1309E: NSFSearch has failed for directory '%s'."
#define		MSG_DXR1310E		"DXR1310E: Unable to extract the filename from an NSFSearch result."
#define		MSG_DXR1312E		"DXR1312E: Unable to determine if an entity is a Database or Directory."
#define		MSG_DXR1313E		"DXR1313E: A lower level directory scanner has failed."
#define		MSG_DXR1315E		"DXR1315E: A Database scanner has failed."
#define		MSG_DXR1317E		"DXR1317E: A Database scanner has thrown an unhandled exception."
#define		MSG_DXR1322E		"DXR1322E: A Recursion has been avoided on directory '%s'."
#define		MSG_DXR1328E		"DXR1328E: Failed to dispatch a DX Request (@%p) for execution."
#define		MSG_DXR1331E		"DXR1331E: The run permit has been lost, all searches will be aborted."

//  Miscellaneous Definitions
#define		FILE_PATH_ITEM		"$Path"

class DXDirScanner :
	public Runnable
{
public:
	//  Constructor
	DXDirScanner(MTExecutive *xeParent, int iThreadID);

	//  Destructor
	~DXDirScanner(void);

	//  Public Members
	int iParentThread;												//  ID of the owning thread

	//  Public Functions
	BOOL ProcessRequest(DXRequest *dxrRQ, int iThreadID);			//  Process an explorer request
	static STATUS LNCALLBACK WrapperProcessSearchResult(void *, SEARCH_MATCH *, ITEM_TABLE *);	//  Callback
	STATUS ProcessSearchResult(SEARCH_MATCH *, ITEM_TABLE *, DXRequest *dxrRQ, int iThreadID);		//  Processing routine for search results
	void virtual ExecuteThisRequest(void *pReqObject, int iThreadID);	//  Multi thread request entry point

private:
	//  Private Members
	MTExecutive *xeLocal;											//  Execution Environment
	DXDirScanner *dxdsNextLevel;									//  Next Level directory scanner
	DXDbScanner *dxbsInstance;										//  Database scanner
	DXException *dxeDeferred;										//  Deferred exception

	//  Private Functions
	BOOL ScanDirectory(DXRequest *dxrRQ, int iThreadID);			//  Scan a directory
	BOOL IsRequestForDirectory(DXRequest *dxrRQ, int iThreadID);	//  Check if request is for a Directory
	void DestroySubContexts(void);									//  Destroy any lower level contexts
	BOOL CoreProcessRequest(DXRequest *dxrRQ, int iThreadID);		//  Internal Request processor
	void DispatchRequest(DXRequest *dxrRQ, int iThreadID);			//  Dispatch a request
};
