#pragma once
//  DXRequest
//
//  This class encapsulates the request information that drives the Domino Explorer (DX)
//
//  Message Identifier:      DXR
//  Message Range:           1901 - 1999
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.11.1
//  Version:                 3.11.1  Build: 99
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2209		Re-based ready for 4.0.0
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.5		18/02/2011		Report on D/Bs which cannot be opened
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2011
//

//  Framework Includes
#include "../ExecEnvironment.h"

//  Trace Messages
#define		MSG_DXR1904T		"DXR1904T: Commencing parsing of the following Explorer Request scope."
#define		MSG_DXR1906T		"DXR1906T: Domain Scope: %i Key: '%s' Match: '%s'." 
#define		MSG_DXR1907T		"DXR1907T: Server Scope: %i Key: '%s' Match: '%s'." 
#define		MSG_DXR1908T		"DXR1908T: Directory Scope: %i Key: '%s' Match: '%s'." 
#define		MSG_DXR1909T		"DXR1909T: Database Scope: %i Key: '%s' Match: '%s'." 

//  Information Messages
#define		MSG_DXR1902I		"DXR1902I: Scope1: '%s'."
#define		MSG_DXR1903I		"DXR1903I: Scope2: '%s'."

//  Error Messages
#define		MSG_DXR1901E		"DXR1901E: Unable to document the current DX request."
#define		MSG_DXR1905E		"DXR1905E: An invalid null part of the request scope was detected."

//  Local Definitions
#define		DXR_FORM_NAME		"DXRequest"
#define		SCOPE_ITEM			"Scope"
#define		REQUESTER_ITEM		"Requester"
#define		DOMAIN_ITEM			"Domain"
#define		SERVER_ITEM			"Server"
#define		DIRECTORY_ITEM		"Directory"
#define		DATABASE_ITEM		"Database"
#define		MDOMAIN_ITEM		"DomainPattern"
#define		MSERVER_ITEM		"ServerPattern"
#define		MDIRECTORY_ITEM		"DirectoryPattern"
#define		MDATABASE_ITEM		"DatabasePattern"
#define		RECURSE_ITEM		"RecurseDirectories"
#define		SERVERP_ITEM		"ServersProcessed"
#define		DIRP_ITEM			"DirectoriesProcessed"
#define		DBP_ITEM			"DatabasesProcessed"
#define		ELAPSED_ITEM		"ElapsedTime"

#define		BASE_DISP_PRIORITY	20									//  Base dispatch priority for DX Requests

class DXRequest
{
public:
	//  Default Constructor

	DXRequest(void);

	//  Destructor

	~DXRequest(void);

	//  Public Members
	DXRequest	*dxrParent;									//  Address of the parent request
	void		*rqExec;									//  Address of the executing object
	int			AnchorScope;								//  The Anchor scope for the request
	UINT		Wildcards;									//  Wildcard levels in the scope
	BOOL		RecurseDirectories;							//  Recurse into subdirectories
	BOOL		bProcessOnlyDirs;							//  Mode - only process directories
	UINT		ScannersNeeded;								//  Scanners that are needed for this request
	UINT		Reporters;									//  Standard reporters that are needed for this request
	UINT		Filters;									//  Filters that apply to this request
	int			iSearchThread;								//  ID of the thread perforimg searches
	int			RQLevel;									//  Request level
	int			ReturnCode;									//  Return code
	DBHANDLE	hdbCache;									//  Cached D/B Handle
	STATUS		stNoAccess;									//  Status code from failed to open D/B
	//  The current request position
	char		szDomain[MAX_DOMAIN];						//  Specific Domain scope
	char		szServer[MAX_SERVER];						//  Specific Server scope
	char		szDirectory[MAX_DIRECTORY];					//  Specific Directory scope
	char		szDatabase[MAX_DATABASE];					//  Specific Database scope
	//  Wildcard match strings for the different levels
	char		szMatchDomain[MAX_DOMAIN];					//  Wildcard Domain scope
	char		szMatchServer[MAX_SERVER];					//  Wildcard Server scope
	char		szMatchDirectory[MAX_DIRECTORY];			//  Wildcard Directory scope
	char		szMatchDatabase[MAX_DATABASE];				//  Wildcard Database scope

	//  Reporters
	void		*dxrUniverse;								//  Universal Reporter
	void		*dxrDomain;									//  Domain Reporter
	void		*dxrServer;									//  Server Reporter
	void		*dxrDirectory;								//  Directory Reporter
	void		*dxrDatabase;								//  Database Reporter

	//  Filters
	void		*dxfUniverse;								//  Universal Filter
	void		*dxfDomain;									//  Domain Filter
	void		*dxfServer;									//  Server Filter
	void		*dxfDirectory;								//  Directory Filter
	void		*dxfDatabase;								//  Database Filter

	//  Processing Statistics
	DWORD		dwServersProcessed;							//  Number of servers processed
	DWORD		dwDirectoriesProcessed;						//  Number of Directories processed
	DWORD		dwDatabasesProcessed;						//  Number of Databases processed
	DWORD		dwServersDropped;							//  Number of servers dropped
	DWORD		dwDirectoriesDropped;						//  Number of Directories dropped
	DWORD		dwDatabasesDropped;							//  Number of Databases dropped
	clock_t		ctStart;									//  Processing Start Time
	clock_t		ctEnd;										//  Processing End Time

	//  Public Functions
	BOOL ParseRequest(ExecEnvironment *xeBase, char *szRQ1, char *szRQ2);			//  Parse a raw request
	BOOL ParseRequest(ExecEnvironment *xeBase, char *szRQ1, char *szRQ2, int iThreadID);	//  Parse a raw request
	virtual DXRequest * CreateNewInstance(void);			//  Create a new instance of a DXRequest
	virtual void InheritFrom(DXRequest *dxrParent);			//  Inherit context from parent request
	virtual void PromoteScores(DXRequest *dxrParent);		//  Promote scores to parent request
	void DocumentRequest(ExecEnvironment *xeBase);			//  Document the current request
	void DocumentRequest(ExecEnvironment *xeBase, int iThreadID);			//  Document the current request
	void SetStart(ExecEnvironment *xeBase);					//  Set the request service start time
	void SetEnd(ExecEnvironment *xeBase);					//  Set the request service end time

	//  Private Members
private:

	//  Private Functions
	void DecodeScope(char *szScope, int iSize);				//  Decode the current scope
	BOOL HasWildcards(char *szAtom);						//  Check for wild cards

	//  Protected Members & Functions
protected:
	void BaseInheritFrom(DXRequest *dxrParent);				//  Inherit context from parent request
	void BasePromoteScores(DXRequest *dxrParent);			//  Promote scores to parent request
};
