//  DXServerScanner
//
//  This class is the eXplorer class that handles the scanning of the server level in the hierarchy according to
//  the scope and constraints from the search request.
//  This is the muti-threaded version of the class.
//
//  Message Identifier:      DXR
//  Message Range:           1201 - 1299
//  Trace area assignment:   TRACE_EXPLORER (05)
//  DX Version:				 3.11.1
//  Version:                 3.11.1  Build: 81
//
//  Version History:
//
//  1.0.0		14/09/2004		Initial version
//  3.5.3		21/08/2209		Re-based ready for 4.0.0
//  3.9.2		22/09/2010		Avoided a recursion in DX processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2004 - 2011
//

#include "DXServerScanner.h"

//  Single-Threaded Constructor
DXServerScanner::DXServerScanner(MTExecutive *xeParent, int iThreadID)
: xeLocal(NULL)
, dxdsLocal(NULL)
, dxsStash(NULL)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Save the execution environment
	xeLocal = xeParent;
	iParentThread = iThreadID;

	//  Trace message
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1227T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);
}

//  Destructor
DXServerScanner::~DXServerScanner(void)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Destroy any lower level scanners
	if (dxdsLocal != NULL) delete dxdsLocal;

	//  Trace message
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1228T, this);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iParentThread);
}

//  ProcessRequest
//
//  This function is the main entry point for dispatching a Domino eXplorer request.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the thread invoking the request
//
//  Returns
//
//  BOOL  -			TRUE if the search was successful
//                  FALSE if the search was not successful
//
//  API Use Notes:
//
//
BOOL DXServerScanner::ProcessRequest(DXRequest *dxrRQ, int iThreadID)
{	
	//  Entry point used for single threaded request execution
	return CoreProcessRequest(dxrRQ, iThreadID);
}

//  IsServerAvailable
//
//  This function will check if the currently targetted server is available (running & accessible)
//
//  Parameters
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the thread invoking the request
//
//  Returns
//  BOOL  -			TRUE if the server is available
//                  FALSE if the server is not available
//
//  API Use Notes:
//
//

BOOL DXServerScanner::IsServerAvailable(DXRequest *dxrRQ, int iThreadID)
{
	//  Test for availability of the server by opening the NAB. This function will also update the server name
	//  in the Request

	STATUS			stAPIRC;							//  Return code from Notes API calls
	DBHANDLE		hdbServerNAB;						//  Handle to the NAB
	char			szMsg[MAX_MSG + 1];					//  Generic message buffer
	char			szNABPathX[MAXPATH +1];				//  Expanded path to the NAB
	char			szServerName[MAXUSERNAME + 1];		//  Server Name
	char			szServerAbbrev[MAXUSERNAME + 1];	//  Abbreviated Server Name

	//  Exit if Server is local
	if (strlen(dxrRQ->szServer) == 0) return TRUE;
	if (strcmpi(dxrRQ->szServer, "LOCAL") == 0) return TRUE;

	sprintf_s(szMsg, MAX_MSG, MSG_DXR1223T, dxrRQ->szServer);
	xeLocal->LogTrace(TRACE_EXPLORER, szMsg);

	//  Attempt to open the NAB on the target server
	hdbServerNAB = xeLocal->OpenDatabase(dxrRQ->szServer, DEFAULT_NAB_DB, iThreadID);
	if (hdbServerNAB == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1222E, dxrRQ->szServer);
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	//  Read back the details from the opened NAB to normalize the server name in the Requset
	stAPIRC = NSFDbPathGet(hdbServerNAB, NULL, szNABPathX);
	if (ERR(stAPIRC) != NOERROR)
	{
		DXException		*dxeLocal;

		//  Close the NAB
		NSFDbClose(hdbServerNAB);

		//  There are no "expected" resons for this to fail - throw an exception
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_NOTESAPI;
		dxeLocal->ExceptionCode = ERR(stAPIRC);
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1224E);
		dxeLocal->DestroyContext = true;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
		dxeLocal->LineNumber = __LINE__;
		throw(*dxeLocal);
	}

	stAPIRC = OSPathNetParse(szNABPathX, NULL, szServerName, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		DXException		*dxeLocal;

		//  Close the NAB
		NSFDbClose(hdbServerNAB);

		//  There are no "expected" resons for this to fail - throw an exception
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_NOTESAPI;
		dxeLocal->ExceptionCode = ERR(stAPIRC);
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1225E);
		dxeLocal->DestroyContext = true;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG, __FUNCTION__);
		dxeLocal->LineNumber = __LINE__;
		throw(*dxeLocal);
	}

	//  Abbreviate the Server name
	stAPIRC = DNAbbreviate(0L, NULL, szServerName, szServerAbbrev, MAXUSERNAME, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		DXException		*dxeLocal;

		//  Close the NAB
		NSFDbClose(hdbServerNAB);

		//  There are no "expected" resons for this to fail - throw an exception
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_NOTESAPI;
		dxeLocal->ExceptionCode = ERR(stAPIRC);
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1226E);
		dxeLocal->DestroyContext = true;
		strcpy_s(dxeLocal->OriginatingFunction, MAX_MSG,__FUNCTION__);
		dxeLocal->LineNumber = __LINE__;
		throw(*dxeLocal);
	}

	//  Update the server name in the request
	strcpy_s(dxrRQ->szServer, MAX_SERVER, szServerAbbrev);

	//  If there is a stash the update the stash with this NAB
	if (dxsStash != NULL)
	{
		UpdateStash(dxrRQ, hdbServerNAB, iThreadID);
	}
	//  Close the NAB
	NSFDbClose(hdbServerNAB);
	return TRUE;
}

//  UpdateStash
//
//  This function will update the server stash using the NAB on the current server as a source.
//
//  Parameters
//
//  1  -  DXRequest *   -  Pointer to the top level request
//  2  -  DBHANDLE		-  DBHANDLE of the NAB on the server that has just been opened.
//  3  -  int			-  ID of the calling thread
//
//  Returns
//
//  BOOL  -			TRUE if the stash as updated
//                  FALSE if the stash was not updated
//
//  API Use Notes:
//
//
BOOL DXServerScanner::UpdateStash(DXRequest *dxrRQ, DBHANDLE hdbNAB, int iThreadID)
{
	STATUS				stAPIRC;					//  Return code from Notes API calls
	NOTEID				nidDesign;					//  Design note for the view
	NOTEID				*nidList;					//  Returned list of notes
	NOTEHANDLE			hnServer;					//  Handle to a Server document
	DHANDLE				hNoteList;					//  Handle of buffer of noteids
	HCOLLECTION			hcView;						//  Handle to the view collection
	COLLECTIONPOSITION	cpView;						//  Collection position
	DWORD				dwEntries;					//  Number of entries returned
	DWORD				dwIndex;					//  Index into the returned NoteIDs
	WORD				wSignalFlags;				//  Flags returned from NIF calls
	int					iUpdates;					//  Count of updated entries
	char				szMsg[MAX_MSG + 1];			//  Generic message buffer

	//  Populate the stash with server search requests

	//  Open the $Servers view and process all of the documents in the view
	stAPIRC = NIFFindDesignNote(hdbNAB, NETAUTH_NAMESPACE_SERVERS, NOTE_CLASS_VIEW, &nidDesign);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR1229E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	stAPIRC = NIFOpenCollection(hdbNAB, hdbNAB, nidDesign, 0L, NULLHANDLE, &hcView, NULLHANDLE, NULL, NULLHANDLE, NULLHANDLE);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR1230E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	cpView.Level = 0;
	cpView.Tumbler[0] = 0;
	
	stAPIRC = NIFReadEntries(hcView,				// handle to this collection
							&cpView,				// where to start in collection
							NAVIGATE_NEXT,			// order to use when skipping
							1L,						// number to skip
							NAVIGATE_NEXT,			// order to use when reading
							0xFFFFFFFF,				// max number to read
							READ_MASK_NOTEID,		// info we want
							&hNoteList,				// handle to info buffer (return)
							NULL,					// length of info buffer (return)
							NULL,					// entries skipped (return)
							&dwEntries,				// entries read (return)
							&wSignalFlags);			// share warning and more signal flag (return)

	if (ERR(stAPIRC) != NOERROR)
	{
		NIFCloseCollection(hcView);
		xeLocal->LogMessage(MSG_DXR1231E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	//  Check that there was something returned
	if (hNoteList == NULLHANDLE)
	{
		NIFCloseCollection(hcView);
		xeLocal->LogMessage(MSG_DXR1232E, iThreadID);
		return FALSE;
	}

	if (dwEntries == 0)
	{
		OSMemFree(hNoteList);
		NIFCloseCollection(hcView);
		xeLocal->LogMessage(MSG_DXR1232E, iThreadID);
		return FALSE;
	}

	//  Process all of the returned entries  
	iUpdates = 0;
	nidList = (NOTEID *) OSLockObject (hNoteList);
	for (dwIndex = 0; dwIndex < dwEntries; dwIndex++)
	{
		//  Check that the Global Run Permit is still asserted
		if (!xeLocal->TestGlobalPermit())
		{
			xeLocal->LogMessage(MSG_DXR1244E, iThreadID);
			OSUnlock(hNoteList);	
			OSMemFree(hNoteList);
			NIFCloseCollection(hcView);
			return FALSE;
		}

		//  Open the note and pass it to the stash updater
		stAPIRC = NSFNoteOpen(hdbNAB, nidList[dwIndex], 0L, &hnServer);
		if (ERR(stAPIRC) != NOERROR)
		{
			xeLocal->LogMessage(MSG_DXR1233E, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		else
		{
			if (UpdateStashFromNote(dxrRQ, hnServer, iThreadID)) iUpdates++;
			NSFNoteClose(hnServer);
		}
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR1234I, iUpdates);
	xeLocal->LogVerbose(szMsg, iThreadID);
	OSUnlock(hNoteList);
	OSMemFree(hNoteList);
	NIFCloseCollection(hcView);
	return TRUE;
}

//  UpdateStashFromNote
//
//  This function will update the server stash using server document passed as a source.
//
//  Parameters
//
//  1  -  DXRequest *   -  Pointer to the top level request
//  2  -  NOTEHANDLE	-  NOTEHANDLE of the server document note to be used as a source
//  3  -  int			-  ID of the thread that is invoking this function
//
//  Returns
//
//  BOOL  -			TRUE if the stash as updated
//                  FALSE if the stash was not updated
//
//  API Use Notes:
//
//
BOOL DXServerScanner::UpdateStashFromNote(DXRequest *dxrRQ, NOTEHANDLE hnServer, int iThreadID)
{
	STATUS			stAPIRC;							//  Return cde from Notes API calls
	DXRequest		*dxrNew;								//  Explorer Request
	WORD			wRetLen;							//  Returned length
	char			szServerName[MAXUSERNAME + 1];		//  Server Name
	char			szServerAbbrev[MAXUSERNAME + 1];	//	Abbreviated Server Name
	char			szDomainName[MAX_DOMAIN + 1];		//  Domain Name
	char			szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Create a stashed server request from a server document

	//  Get the server name and abbreviate it
	wRetLen = NSFItemGetText(hnServer, MAIL_SERVERNAME_ITEM, szServerName, MAXUSERNAME);
	if (wRetLen == 0) 
	{
		xeLocal->LogMessage(MSG_DXR1235E, iThreadID);
		return FALSE;
	}
	szServerName[wRetLen] = '\0';
	stAPIRC = DNAbbreviate(0L, NULL, szServerName, szServerAbbrev, MAXUSERNAME, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR1236E, iThreadID);
		return FALSE;
	}

	//  Get the Domain name
	wRetLen = NSFItemGetText(hnServer, MAIL_DOMAINNAME_ITEM, szDomainName, MAX_DOMAIN);
	if (wRetLen == 0) 
	{
		xeLocal->LogMessage(MSG_DXR1237E, iThreadID);
		return FALSE;
	}
	szDomainName[wRetLen] = '\0';

	//  Build and populate the request
	dxrNew = dxrRQ->CreateNewInstance();
	dxrNew->InheritFrom(dxrRQ);
	strcpy_s(dxrNew->szServer, MAX_SERVER, szServerAbbrev);
	strcpy_s(dxrNew->szDomain, MAX_DOMAIN, szDomainName);
	dxrNew->AnchorScope = RSCOPE_SERVER;
	dxrNew->ReturnCode = RETURN_STASHED;
	dxsStash->Add(dxrNew);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR1238I, szServerAbbrev, szDomainName);
	xeLocal->LogMessage(szMsg, iThreadID);
	return TRUE;
}

//  CoreProcessRequest
//
//  This function is the main entry point for executing a Server Scan Domino eXplorer request.
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the thread that is invoking the function
//
//  Returns
//
//  BOOL  -			TRUE if the scan was successful
//                  FALSE if the scan was not successful
//
//  API Use Notes:
//
//
BOOL DXServerScanner::CoreProcessRequest(DXRequest *dxrRQ, int iThreadID)
{
	DXRequest		*dxrRoot;							//  Root Directory flag
	DXRequest		*dxrReturned;						//  Request returned from execution
	int				iRejoinRC;							//  Return code from Rejoin calls
	BOOL			bProcessThisRequest;				//  Cntrol flag
	DXFilter		*dxfSelector;						//  Selection Filter
	char			szMsg[MAX_MSG + 1];					//  Generic message buffer

	try
	{
		//  Skip the server availability check if the search scope is a database
		if (dxrRQ->AnchorScope != RSCOPE_DATABASE)
		{
			//  Determine if the Server is available
			if (!IsServerAvailable(dxrRQ, iThreadID))
			{
				xeLocal->LogMessage(MSG_DXR1220E, iThreadID);
				return FALSE;
			}
		}

		//  Construct a directory scanner, if one does not already exist
		if (dxdsLocal == NULL)
		{
			dxdsLocal = new DXDirScanner(xeLocal, iThreadID);				//  Create a new directory scanner
		}

		//  Check if this server should be processed
		bProcessThisRequest = TRUE;
		if (dxrRQ->Wildcards & WILDCARD_SERVER)
		{
			if (!xeLocal->MatchesPatterni(dxrRQ->szServer, dxrRQ->szMatchServer)) bProcessThisRequest = FALSE;
		}

		if (bProcessThisRequest)
		{
			if ((dxrRQ->Filters & FILTER_SERVER) && (dxrRQ->dxfServer != NULL))
			{
				dxfSelector = (DXFilter *) dxrRQ->dxfServer;
				if (!dxfSelector->FilterThisServer(dxrRQ, iThreadID)) bProcessThisRequest = FALSE;
			}
		}

		if (bProcessThisRequest)
		{
			//  Determine if the request is to be processed at this level or passed through
			if ((dxrRQ->AnchorScope == RSCOPE_UNIVERSE) || (dxrRQ->AnchorScope == RSCOPE_DOMAIN) || (dxrRQ->AnchorScope == RSCOPE_SERVER))
			{

				dxrRQ->SetStart(xeLocal);
				//  Trace the request
				sprintf_s(szMsg, MAX_MSG, MSG_DXR1239T, dxrRQ->szServer);
				xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
			
				//  Create a new request for the root directory of the current server
				dxrRoot = dxrRQ->CreateNewInstance();
				dxrRoot->InheritFrom(dxrRQ);
				dxrRoot->AnchorScope = RSCOPE_DIRECTORY;
				dxrRoot->RecurseDirectories = dxrRQ->RecurseDirectories;
				strcpy_s(dxrRoot->szServer, MAX_SERVER, dxrRQ->szServer);
				dxrRoot->szDirectory[0] = '\0';
				dxrRoot->szDatabase[0] = '\0';
				dxrRoot->ReturnCode = RETURN_INCOMPLETE;

				//  Set the processing information in the child request
				dxrRoot->dxrParent = dxrRQ;
				dxrRoot->rqExec = dxdsLocal;

				//  Dispatch the request for execution - decide on using asynch or synch
				//  execution based on the CMLS recommendation
				if (xeLocal->GetCMLSRecommendation(RQATTR_CMLSL1, iThreadID) == CMLS_LANE_FULL)
				{
					// Synchronous execution
					dxdsLocal->ProcessRequest(dxrRoot, iThreadID);
					dxrRQ->ReturnCode = dxrRoot->ReturnCode;
					dxrRQ->dwServersProcessed++;							//  Increment the count of servers processed
					dxrRoot->PromoteScores(dxrRQ);							//  Accumulate Scores
					if (xeLocal->LogGetLevel() > LOGLEVEL_VERBOSE) dxrRoot->DocumentRequest(xeLocal);
					delete dxrRoot;
				}
				else
				{
					//  Asynch dispatch
					DispatchRequest(dxrRoot, iThreadID);
				}
			
				//  Wait for the request to rejoin

				//  All Directory (Root) Requests have been dispatched - now get all of the completed requests
				//  and accumulate the totals into the parent request

				iRejoinRC = xeLocal->GetRejoinRequest(dxrRQ, (void **) &dxrReturned, iThreadID);
				while (iRejoinRC != RJR_NONE_EXIST)
				{	
					if (iRejoinRC == RJR_RETURNED)
					{
						//  Accumulate the totals into the parent request and dispose of the child
						sprintf_s(szMsg, MAX_MSG, MSG_DXR1243T, dxrReturned, dxrReturned->dxrParent);
						xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
						dxrReturned->PromoteScores(dxrReturned->dxrParent);
						dxrReturned->dxrParent->ReturnCode = dxrReturned->ReturnCode;
						dxrReturned->dxrParent->dwServersProcessed++;
						if (xeLocal->LogGetLevel() > LOGLEVEL_VERBOSE) dxrReturned->DocumentRequest(xeLocal);
						delete dxrReturned;
						return TRUE;
					}
					else xeLocal->AppWait(PXR_BUSY_WAIT);
					iRejoinRC = xeLocal->GetRejoinRequest(dxrRQ, (void **) &dxrReturned, iThreadID);
				}
				dxrRQ->SetEnd(xeLocal);
			}
			else
			{
				//  Set the processing information in the current request
				dxrRQ->rqExec = dxdsLocal;

				//  Request is to be passed through to a lower level
				return dxdsLocal->ProcessRequest(dxrRQ, iThreadID);
			}
		}
		else
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR1240I, dxrRQ->szServer);
			xeLocal->LogVerbose(szMsg);
		}
	}
	catch(DXException&)
	{
		//  Destroy the directory scanner as it may now be unstable
		delete dxdsLocal;
		dxdsLocal = NULL;

		//  Rethrow the exception to the next level - percolate
		throw;
	}
	catch(std::exception& eCurrent)
	{
		DXException		*dxeLocal;

		//  Destroy the directory scanner as it may now be unstable
		delete dxdsLocal;
		dxdsLocal = NULL;

		//  Create a new exception to wrapper the information from the original
		dxeLocal = new DXException();
		dxeLocal->ExceptionType = ERRORTYPE_UHX;
		dxeLocal->ExceptionCode = ERRORCODE_NON_SPEC;
		dxeLocal->FaultingThread = iThreadID;
		strcpy_s(dxeLocal->LocalExceptionMessage, MAX_MSG, eCurrent.what());
		strcpy_s(dxeLocal->DXExceptionMessage, MAX_MSG, MSG_DXR1221E);
		throw(*dxeLocal);
	}
	return FALSE;
}

//  ExecuteThisRequest
//
//  This is the entry point used by the multi-threading kernel to trigger processing in different threadpool threads.
//
//  Parameters
//
//  1  -  void *			-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DXServerScanner::ExecuteThisRequest(void *pReqObject, int iThreadID)
{
	ProcessRequest((DXRequest *) pReqObject, iThreadID);
	return;
}

//  DispatchRequest
//
//  This function will dispatch a DX request for asynchronous processing
//
//  Parameters
//
//  1  -  DXRequest *		-  Pointer to a DXRequest object that defines the eXplorer search scope and constraints
//  2  -  int				-  ID of the invoking Thread
//
//  Returns
//
//  API Use Notes:
//
//
void DXServerScanner::DispatchRequest(DXRequest *dxrDisp, int iThreadID)
{
	DXRequest	*dxrReturned;									//  Request returned from execution
	int			iRejoinRC;										//  Return code from Rejoin calls
	char		szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  As the CMLS dispatch may block for an extended period of time the perform a
	//  rejoin polling loop before attempting the dispatch.
	//  Loop polling the Rejoin pool for any requests that have completed for this owner
	iRejoinRC = xeLocal->GetRejoinRequest(dxrDisp->dxrParent, (void **) &dxrReturned, iThreadID);
	while (iRejoinRC == RJR_RETURNED)
	{
		//  Accumulate the totals into the parent request and dispose of the child
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1243T, dxrReturned, dxrReturned->dxrParent);
		xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
		dxrReturned->PromoteScores(dxrReturned->dxrParent);
		delete dxrReturned;
		iRejoinRC = xeLocal->GetRejoinRequest(dxrDisp->dxrParent, (void **) &dxrReturned, iThreadID);
	}

	//  Dispatch the request - Pripority is increased as we increase the 'depth' of the request
	if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
								dxrDisp->dxrParent,
								dxrDisp->rqExec,
								dxrDisp,
								RQATTR_REJOIN | RQATTR_CMLSL1,
								dxrDisp->RQLevel * BASE_DISP_PRIORITY,
								iThreadID))
	{
		//  This is an SNO (Should Not Occur) condition - this interface will block on resource shortage
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1241E, dxrDisp);
		xeLocal->LogMessage(szMsg, iThreadID);
	}
	else
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1242T, dxrDisp, dxrDisp->AnchorScope, dxrDisp->RQLevel * BASE_DISP_PRIORITY);
		xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
	}

	//  Loop polling the Rejoin pool for any requests that have completed for this owner
	iRejoinRC = xeLocal->GetRejoinRequest(dxrDisp->dxrParent, (void **) &dxrReturned, iThreadID);
	while (iRejoinRC == RJR_RETURNED)
	{
		//  Accumulate the totals into the parent request and dispose of the child
		sprintf_s(szMsg, MAX_MSG, MSG_DXR1243T, dxrReturned, dxrReturned->dxrParent);
		xeLocal->LogTrace(TRACE_EXPLORER, szMsg, iThreadID);
		dxrReturned->PromoteScores(dxrReturned->dxrParent);
		delete dxrReturned;
		iRejoinRC = xeLocal->GetRejoinRequest(dxrDisp->dxrParent, (void **) &dxrReturned, iThreadID);
	}
	return;
}

