//  CommandHandler
//
//  This class provides the standard handling of the MQ command stream for DX Add-In tasks.
//  The class can be extended to provide support for additional commands or additional processing for
//  the standard command set.
//
//  Message Identifier:      DXR
//  Message Range:           0900 - 0999
//  Trace area assignment:   TRACE_CORERT (01)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  3.9.0		24/07/2010		Initial Version
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.3		13/12/2010		Debug Helper added
//  3.10.4		14/01/2011		Command handler extended to support automated commands
//  3.10.6		28/03/2011		Enable command injection
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.14.1		20/01/2015		x64 compatability
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2015
//

//  Application Includes
#include "CommandHandler.h"

//  Default Constructor
CommandHandler::CommandHandler(void)
{
}

//  Constructor
CommandHandler::CommandHandler(ExecEnvironment *xeParent, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Save the Runtime Environment
	xeLocal = xeParent;
	ParentThread = iThreadID;
	wInjectedCommand = MQ_COMMAND_NULL;
	szStatus[0] = '\0';
	State = 0;
#if (defined(_DEBUG) && defined(WIN32))
	DebugHelper = NULL;
#endif

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0901T, this);
	xeLocal->LogTrace(TRACE_CORERT, szMsg, ParentThread);
}

//  Destructor
CommandHandler::~CommandHandler(void)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0902T, this);
	xeLocal->LogTrace(TRACE_DBC, szMsg, ParentThread);
}

//  HandleCommands
//
//  This function is invoked to allow the handle to check the MQ for any commands.
//  If a command is detected in the MQ it will be executed and then control is returned to the caller.
//
//  Parameters
//
//  1   -		int		- Thread ID of the invoking thread
//
//  Returns
//
//
//  API Use Notes:
//
//
void CommandHandler::HandleCommands(int iThreadID)
{
	STATUS				stAPIRC;						//  Return code from Notes API calls
	WORD				wRetLen;						//  Size of the returned command string
	WORD				wCommand;						//  Encoded command word
	BOOL				bCmdIsCustom;					//  The command is a custom command
	char				szCommandBuffer[MAX_CMD + 1];	//  Command buffer
	char				szCommandOptions[MAX_CMD + 1];	//  Command options
	char				szMsg[MAX_MSG + 1];				//  Generic message buffer

	//  If the state is already terminal the just return
	if (State & CH_STATE_TERMINAL) return;

	//  Check that the MQ handle is valid
	if (xeLocal->hmqProcessor == NULLHANDLE)
	{
		xeLocal->LogMessage(MSG_DXR0903E, iThreadID); 
		AddInLogMessageText(MSG_DXR0903E, NOERROR);
		//  Set the state to show that the Command Handle cannot function
		State = State | (CH_STATE_INVALID | CH_STATE_TERMINAL);
		return;
	}

	//  Check if there is a quit message on the processor queue
	if (MQIsQuitPending(xeLocal->hmqProcessor))
	{
		xeLocal->LogMessage(MSG_DXR0904I, iThreadID); 
		AddInLogMessageText(MSG_DXR0904I, NOERROR);
		//  Set the state to show that the application is quitting
		State = State | CH_STATE_QUIT;
		State = State | CH_STATE_TERMINAL;
		return;
	}

	//  Read the next command from the Message Queue (if any)
	stAPIRC = MQGet(xeLocal->hmqProcessor, szCommandBuffer, MAX_CMD, 0, 0, &wRetLen);
	if (ERR(stAPIRC) == ERR_MQ_EMPTY)
	{
		//  No commands are in the queue - determine if there is an automatically generated command available
		szCommandOptions[0] = '\0';
		//  If there is an injected command then process that
		if (wInjectedCommand != MQ_COMMAND_NULL)
		{
			wCommand = wInjectedCommand;
			wInjectedCommand = MQ_COMMAND_NULL;
		}
		else wCommand = GetAutoCommand(szCommandOptions, iThreadID);

		//  If there were no automated commands then just return
		if (wCommand == MQ_COMMAND_NULL) return;

		//  Pass the command to the custom command handler to process
		wCommand = CustomCommandHandler(wCommand, szCommandOptions, iThreadID);

		//  Process any system commands generated by the custom handler
		while (wCommand != MQ_COMMAND_NULL)
		{
			wCommand = SystemCommandHandler(wCommand, szCommandOptions, iThreadID);		
		}
		return;
	}

	if (ERR(stAPIRC) == ERR_MQ_QUITTING)
	{
		//  Quit was detected while reading the message
		xeLocal->LogMessage(MSG_DXR0904I, iThreadID); 
		AddInLogMessageText(MSG_DXR0904I, NOERROR);
		//  REvoke the global DCF permit
		xeLocal->RevokeGlobalPermit();
		//  Set the state to show that the application is quitting
		State = State | CH_STATE_QUIT;
		return;
	}

	if ((ERR(stAPIRC) != NOERROR) && (ERR(stAPIRC) != ERR_MQ_BFR_TOO_SMALL))
	{
		//  An error has been detected on the MQ - mark the Command Handler as inoperative
		xeLocal->LogMessage(MSG_DXR0905E, iThreadID); 
		AddInLogMessageText(MSG_DXR0905E, ERR(stAPIRC));
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		State = State | CH_STATE_INVALID;
		return;
	}

	//  Warn if the command received has been truncated
	if (ERR(stAPIRC) == ERR_MQ_BFR_TOO_SMALL)
	{
		xeLocal->LogMessage(MSG_DXR0906W, iThreadID); 
		AddInLogMessageText(MSG_DXR0906W, NOERROR);
	}

	//  Echo the command to the log and the console
	szCommandBuffer[wRetLen] = '\0';
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0907I, szCommandBuffer);
	xeLocal->LogMessage(szMsg, iThreadID); 
	AddInLogMessageText(szMsg, NOERROR);

	//  Parse the command 
	bCmdIsCustom = FALSE;
	wCommand = ParseSystemCommand(szCommandBuffer, szCommandOptions, iThreadID);
	if (wCommand == MQ_COMMAND_NULL) return;
	if (wCommand == MQ_COMMAND_INVALID)
	{
		//  Determine if the command is a valid custom command
		wCommand = ParseCustomCommand(szCommandBuffer, szCommandOptions, iThreadID);
		if (wCommand == MQ_COMMAND_INVALID)
		{
			//  If the command is still invalid then show it and return;
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0908W, szCommandBuffer);
			xeLocal->LogMessage(szMsg, iThreadID); 
			AddInLogMessageText(szMsg, NOERROR);
			return;
		}
		bCmdIsCustom = TRUE;
	}

	//  Pass the command first to the Custom command handler - which may modify the command
	wCommand = CustomCommandHandler(wCommand, szCommandOptions, iThreadID);

	//  If the command was a custom command or has been cancelled by the handler then just return
	if ((bCmdIsCustom) || (wCommand == MQ_COMMAND_NULL)) return;

	//  Process the command in the system handler
	while (wCommand != MQ_COMMAND_NULL)
	{
		wCommand = SystemCommandHandler(wCommand, szCommandOptions, iThreadID);		
	}

	return;
}

//  ParseSystemCommand
//
//  This function will encode a Tell command string that is received, it will encode the command and 
//  return the encoded values and any additional keywords that were supplied.
//
//  Parameters
//
//  1  -  char *		-  The command buffer that was received
//  2  -  char *		-  The address of a buffer where any command options will be set
//  3  -  int			-  The Thread ID of the invoking thread
//
//  Returns
//  WORD  -				The encoded command value
//
//  API Use Notes:
//
//
WORD CommandHandler::ParseSystemCommand(char *szCommand, char *szOptions, int iThreadID)
{
	//  Parse the supplied command into an encoded command and optional parameters
	WORD		wEncCmd;

	char		szTempBuffer[MAX_CMD + 1];
	char		*lpCmdPtr;
	char		*pNextToken;

	int			iIndex;

	wEncCmd = MQ_COMMAND_INVALID;
	szOptions[0] = '\0';								//  Clear the parameter buffer
	if (strlen(szCommand) == 0)
	{
		return(MQ_COMMAND_NULL);
	}

	strcpy_s(szTempBuffer, MAX_CMD, szCommand);
	for (iIndex = 0; szTempBuffer[iIndex]; iIndex++)
	{
		szTempBuffer[iIndex] = toupper(szTempBuffer[iIndex]);
	}
	lpCmdPtr = strtok_s(szTempBuffer, " ,", &pNextToken);

	if (strcmp(lpCmdPtr, MQ_CMD_STRING_STOP) == 0)
		wEncCmd = MQ_COMMAND_STOP;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_QUIT) == 0)
		wEncCmd = MQ_COMMAND_STOP;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_ABORT) == 0)
		wEncCmd = MQ_COMMAND_ABORT;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_SUSPEND) == 0)
		wEncCmd = MQ_COMMAND_SUSPEND;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_RESUME) == 0)
		wEncCmd = MQ_COMMAND_RESUME;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_VERBOSE) == 0)
		wEncCmd = MQ_COMMAND_VERBOSE;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_LOUD) == 0)
		wEncCmd = MQ_COMMAND_LOUD;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_TERSE) == 0)
		wEncCmd = MQ_COMMAND_TERSE;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_QUIET) == 0)
		wEncCmd = MQ_COMMAND_QUIET;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_ECHO) == 0)
		wEncCmd = MQ_COMMAND_ECHO;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_NOECHO) == 0)
		wEncCmd = MQ_COMMAND_NOECHO;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_TRACE) == 0)
		wEncCmd = MQ_COMMAND_TRACE;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_DEBUG) == 0)
		wEncCmd = MQ_COMMAND_DEBUG;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_STATUS) == 0)
		wEncCmd = MQ_COMMAND_STATUS;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_STATS) == 0)
		wEncCmd = MQ_COMMAND_STATS;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_DISABLE) == 0)
		wEncCmd = MQ_COMMAND_DISABLE;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_ENABLE) == 0)
		wEncCmd = MQ_COMMAND_ENABLE;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_PANIC) == 0)
		wEncCmd = MQ_COMMAND_PANIC;
#ifdef _DEBUG
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_MEM) == 0)
		wEncCmd = MQ_COMMAND_MEMORY;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_MEMORY) == 0)
		wEncCmd = MQ_COMMAND_MEMORY;
	if (strcmp(lpCmdPtr, MQ_CMD_STRING_DUMP) == 0)
		wEncCmd = MQ_COMMAND_DUMP;
#endif

	if (strlen(szCommand) > strlen(lpCmdPtr))
	{
		strcpy_s(szOptions, MAX_CMD, szCommand + strlen(lpCmdPtr) + 1);
	}

	return(wEncCmd);
}

//  ParseCustomCommand
//
//  This function is a stub as will be overridden in any extending class that wishes to implement
//  custom commands.
//
//  Parameters
//
//  1  -  char *		-  The command buffer that was received
//  2  -  char *		-  The address of a buffer where any command options will be set
//  3  -  int			-  The Thread ID of the invoking thread
//
//  Returns
//
//  WORD  -				The encoded command value
//
//  API Use Notes:
//
//
WORD CommandHandler::ParseCustomCommand(char *szCommand, char *szOptions, int iThreadID)
{
	return MQ_COMMAND_INVALID;
}

//  CustomCommandHandler
//
//  This function is a stub as will be overridden in any extending class that wishes to implement
//  custom commands.
//
//  Parameters
//
//  1  -  WORD			-  The encoded command 
//  2  -  char *		-  The address of a buffer containing any command options will be set
//  3  -  int			-  The Thread ID of the invoking thread
//
//  Returns
//
//  WORD  -				The encoded command value
//
//  API Use Notes:
//
//  The handler should process it's own custom commands but it will also be passed all of the system
//  commands before they are processed by the system command handler. The handler can change the command
//  tell the system to ignore it by returning(MQ_COMMAND_NULL) or can optionally perform additional actions
//  and then return the same value and have the system handler process it.
//
WORD CommandHandler::CustomCommandHandler(WORD wCommand, char *szOptions, int iThreadID)
{
	return wCommand;
}

//  SystemCommandHandler
//
//  This function is the handle for standard system commands
//
//  Parameters
//
//  1  -  WORD			-  The encoded command 
//  2  -  char *		-  The address of a buffer containing any command options will be set
//  3  -  int			-  The Thread ID of the invoking thread
//
//  Returns
//
//  WORD  -				The encoded command value
//
//  API Use Notes:
//
//  The handler is invoked in a loop until it returns MQ_COMMAND_NULL.
//  This allows allows the command processors to invoke subsequent commands (e.g. STATUS after a state change)
//  or to mofify the command that is being executed (e.g. ECHO OFF is re-driven as a NOECHO command).
//
WORD CommandHandler::SystemCommandHandler(WORD wCommand, char *szOptions, int iThreadID)
{
	WORD			wNextCommand;							//  The next command to be executed
	int				iTraceArea;								//  Area for trace or debug
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	wNextCommand = MQ_COMMAND_NULL;							//  End of the command chain

	switch(wCommand)
	{
	case MQ_COMMAND_QUIT:
		State = State | CH_STATE_QUIT;						//  Alter the state to show a quit signal was detected
		State = State | CH_STATE_TERMINAL;
		break;
	case MQ_COMMAND_STOP:
		//  Check if the Now keyword was supplied in which case recycle the command as an Abort
		if (strcmpi(szOptions, "Now") == 0)
		{
			wNextCommand = MQ_COMMAND_ABORT;
		}
		else
		{
			State = State | CH_STATE_STOP;					//  Alter the state to show stop requested
			State = State | CH_STATE_TERMINAL;
		}
		break;
	case MQ_COMMAND_ABORT:
		State = State | CH_STATE_ABORT;						//  Alter the state to show that abort requested
		State = State | CH_STATE_TERMINAL;
		xeLocal->SetFastDrainMode();						//  Set Fast Drain Mode
		xeLocal->RevokeGlobalPermit();						//  Revoke the Global Run Permit
		break;
	case MQ_COMMAND_SUSPEND:
		xeLocal->LogMessage(MSG_DXR0909I, iThreadID); 
		AddInLogMessageText(MSG_DXR0909I, NOERROR);
		State = State | CH_STATE_SUSPENDED;					//  Alter the state to show entered the suspended state
		break;
	case MQ_COMMAND_RESUME:
		xeLocal->LogMessage(MSG_DXR0910I, iThreadID); 
		AddInLogMessageText(MSG_DXR0910I, NOERROR);
		State = State & (!CH_STATE_SUSPENDED);				//  Alter the state to show left the suspended state
		break;
	case MQ_COMMAND_VERBOSE:
		xeLocal->LogSetVerbose();							//  Set logging level to verbose
		break;
	case MQ_COMMAND_TERSE:
		xeLocal->LogSetNormal();							//  Set logging level to terse
		break;
	case MQ_COMMAND_LOUD:
		wNextCommand = MQ_COMMAND_VERBOSE;					//  Recycle as a Verbose command
		break;
	case MQ_COMMAND_QUIET:
		wNextCommand = MQ_COMMAND_TERSE;					//  Recycle as a Terse command
		break;
	case MQ_COMMAND_ECHO:
		//  Check if OFF has been specified as a command option in which case recycle as NOECHO
		if (strcmpi(szOptions, "Off") == 0)
		{
			wNextCommand = MQ_COMMAND_NOECHO;
		}
		else
		{
			xeLocal->LogSetEchoOn();						//  Set the logging state to echo to console
		}
		break;
	case MQ_COMMAND_NOECHO:
		xeLocal->LogSetEchoOff();							//  Set the logging state to NOT echo to console
		break;
	case MQ_COMMAND_TRACE:
		iTraceArea = atoi(szOptions);						//  Get the requested trace area
		xeLocal->LogSetTrace(iTraceArea);					//  Put tracing into effect
		break;
	case MQ_COMMAND_DEBUG:
		iTraceArea = atoi(szOptions);						//  Get the requested trace area
		xeLocal->LogSetDebug(iTraceArea);					//  Put debugging into effect
		break;
	case MQ_COMMAND_REFRESH:
		State = State | CH_STATE_REFRESH;					//  Alter the state to show refresh required
		break;
	case MQ_COMMAND_STATUS:
		//  If there is a status string set then display it - otherwise do nothing
		if (szStatus[0] != '\0')
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0911I, xeLocal->rsLocal->APPName, szStatus);
			xeLocal->LogMessage(szMsg, iThreadID); 
			AddInLogMessageText(szMsg, NOERROR);
		}
		break;
	case MQ_COMMAND_STATS:
		//  Check for either the debug or thread keyword that will alter the behaviour of the stats display
		iTraceArea = 0;
		if (strcmpi(szOptions, "Debug") == 0)
		{
			iTraceArea = STATS_DEBUG;
		}
		else
		{
			if (strcmpi(szOptions, "Thread") == 0)
			{
				//  Thread also sets Debug
				iTraceArea = STATS_DEBUG | STATS_THREAD;
			}
		}
		//  Invoke the stats display
		xeLocal->ShowStats(iTraceArea, iThreadID);
		break;
	case MQ_COMMAND_DISABLE:
		//  Revoke the application DCF permit
		xeLocal->RevokeGlobalPermit();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0912I, xeLocal->rsLocal->APPName);
		xeLocal->LogMessage(szMsg, iThreadID);
		break;
	case MQ_COMMAND_ENABLE:
		//  Reinstate the application DCF permit
		xeLocal->AssertGlobalPermit();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0913I, xeLocal->rsLocal->APPName);
		xeLocal->LogMessage(szMsg, iThreadID);
		break;
	case MQ_COMMAND_PANIC:
		//  Perform an operator induced PANIC
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0914I, szOptions);
		xeLocal->LogMessage(szMsg, iThreadID);
		AddInLogMessageText(szMsg, NOERROR);
		xeLocal->Panic("Operator", 0, szOptions, xeLocal, this, 0, iThreadID);
		break;
#if (defined(_DEBUG) && defined(WIN32))
	case MQ_COMMAND_MEMORY:
		//  Display the current memory usage statistics
		if (DebugHelper != NULL)
		{
			DebugHelper->ReportMemoryUsage(FALSE, TRUE, iThreadID);
		}
		break;
	case MQ_COMMAND_DUMP:
		//  Create a core dump of the application
		if (DebugHelper != NULL)
		{
			DebugHelper->CreateMemoryDump(iThreadID);
		}
		break;
#endif
	}

	return wNextCommand;
}

//  GetAutoCommand
//
//  This function is a stub as will be overridden in any extending class that wishes to implement
//  automated commands. This interface will be called regularly by kernel, if the function wishes
//  to trigger an automated command then it should return the encoded command.
//
//  Parameters
//
//	1  -  char *		-  Pointer to a buffer where command options can be set
//  2  -  int			-  The Thread ID of the invoking thread
//
//  Returns
//
//  WORD  -				The encoded command value
//
//  API Use Notes:
//
//
WORD CommandHandler::GetAutoCommand(char *szOptions, int iThreadID)
{
	//  Nothing to execute
	return MQ_COMMAND_NULL;
}

