#pragma once
//  CommandHandler
//
//  This class provides the standard handling of the MQ command stream for DX Add-In tasks.
//  The class can be extended to provide support for additional commands or additional processing for
//  the standard command set.
//
//  Message Identifier:      DXR
//  Message Range:           0900 - 0999
//  Trace area assignment:   TRACE_CORERT (01)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  3.9.0		24/07/2010		Initial Version
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.3		13/12/2010		Debug Helper added
//  3.10.4		14/01/2011		Command handler extended to support automated commands
//  3.10.6		28/03/2011		Enable command injection
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.14.1		20/01/2015		x64 compatability
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2015
//

//  Application Includes
#include "../ExecEnvironment.h"
#if (defined(_DEBUG) && defined(WIN32))
#include "../Debug/Helper.h"
#endif

//  Tell (MQ) Command strings
#define		MQ_CMD_STRING_QUIT			"QUIT"
#define		MQ_CMD_STRING_STOP			"STOP"
#define		MQ_CMD_STRING_ABORT			"ABORT"
#define		MQ_CMD_STRING_SUSPEND		"SUSPEND"
#define		MQ_CMD_STRING_RESUME		"RESUME"
#define		MQ_CMD_STRING_VERBOSE		"VERBOSE"
#define		MQ_CMD_STRING_LOUD			"LOUD"
#define		MQ_CMD_STRING_TERSE			"TERSE"
#define		MQ_CMD_STRING_QUIET			"QUIET"
#define		MQ_CMD_STRING_ECHO			"ECHO"
#define		MQ_CMD_STRING_NOECHO		"NOECHO"
#define		MQ_CMD_STRING_TRACE			"TRACE"
#define		MQ_CMD_STRING_DEBUG			"DEBUG"
#define		MQ_CMD_STRING_REFRESH		"REFRESH"
#define		MQ_CMD_STRING_STATUS		"STATUS"
#define		MQ_CMD_STRING_STATS			"STATS"
#define		MQ_CMD_STRING_DISABLE		"DISABLE"
#define		MQ_CMD_STRING_ENABLE		"ENABLE"
#define		MQ_CMD_STRING_PANIC			"PANIC"
#ifdef _DEBUG
#define		MQ_CMD_STRING_MEM			"MEM"
#define		MQ_CMD_STRING_MEMORY		"MEMORY"
#define		MQ_CMD_STRING_DUMP			"DUMP"
#endif


//  Encoded Tell (MQ) commands
#define		MQ_COMMAND_NULL				0
#define		MQ_COMMAND_INVALID			1
#define		MQ_COMMAND_QUIT				2
#define		MQ_COMMAND_STOP				3
#define		MQ_COMMAND_ABORT			4
#define		MQ_COMMAND_SUSPEND			5
#define		MQ_COMMAND_RESUME			6
#define		MQ_COMMAND_VERBOSE			7
#define		MQ_COMMAND_TERSE			8
#define		MQ_COMMAND_LOUD				9
#define		MQ_COMMAND_QUIET			10
#define		MQ_COMMAND_ECHO				11
#define		MQ_COMMAND_NOECHO			12
#define		MQ_COMMAND_TRACE			13
#define		MQ_COMMAND_DEBUG			14
#define		MQ_COMMAND_REFRESH			15
#define		MQ_COMMAND_STATUS			16
#define		MQ_COMMAND_STATS			17
#define		MQ_COMMAND_DISABLE			18
#define		MQ_COMMAND_ENABLE			19
#define		MQ_COMMAND_PANIC			20
#ifdef _DEBUG
#define		MQ_COMMAND_MEMORY			21
#define		MQ_COMMAND_DUMP				22
#endif
#define		MQ_COMMAND_USER				100				//  Custom commands start at 100

//  State values
#define		CH_STATE_QUIT				0x0001			//  A quit message has been detected
#define		CH_STATE_STOP				0x0002			//  A Stop command has beeen received
#define		CH_STATE_ABORT				0x0004			//  An Abort (or Stop Now) command has been received
#define		CH_STATE_SUSPENDED			0x0008			//  A suspend command has been recieved (cleared by resume)
#define		CH_STATE_REFRESH			0x0010			//  A refresh is requested
#define		CH_STATE_TERMINAL			0x0020			//  The state is regarded as terminal
#define		CH_STATE_INVALID			0x0040			//  The command handler is having MQ problems

//  Trace Messages
#define			MSG_DXR0901T		"DXR0901T: A new CommandHandler object has been created @%p."
#define			MSG_DXR0902T		"DXR0902T: A CommandHandler object @%p has been destroyed."

//  Information Messages
#define			MSG_DXR0904I		"DXR0904I: The Command Handler has recieved a Quit signal from the message queue."
#define			MSG_DXR0907I		"DXR0907I: Command received: %s."
#define			MSG_DXR0909I		"DXR0909I: The application is entering the suspended state."
#define			MSG_DXR0910I		"DXR0910I: The application is resuming operation from the suspended state."
#define			MSG_DXR0911I		"DXR0911I: %s - %s."
#define			MSG_DXR0912I		"DXR0912I: The application global permit for %s has been revoked."
#define			MSG_DXR0913I		"DXR0913I: The application global permit for %s has been reinstated."
#define			MSG_DXR0914I		"DXR0914I: An operator panic is being run because: %s."

//  Warning Messages
#define			MSG_DXR0906W		"DXR0906W: The following command from the Message Queue has been truncated."
#define			MSG_DXR0908W		"DXR0908W: Unrecognised command '%s' is being ignored."

//  Error Messages
#define			MSG_DXR0903E		"DXR0903E: The Command Handler has been invoked without a valid MQ handle."
#define			MSG_DXR0905E		"DXR0905E: The Command Handler has detected an error on the message queue."

class CommandHandler
{
public:
	//  Default Constructor
	CommandHandler(void);

	//  Constructor
	CommandHandler(ExecEnvironment *xeParent, int iThreadID);

	//  Destructor
	~CommandHandler(void);

	//  Public Members
	int						State;											//  State that the Addin should observe
#if (defined(_DEBUG) && defined(WIN32))
	Helper					*DebugHelper;									//  Debuging helper object
#endif
	char					szStatus[MAX_MSG + 1];							//  Status line to display
	WORD					wInjectedCommand;							//  Injected Command

	//  Public Functions
	void HandleCommands(int iThreadID);										//  Handle any commands in the MQ

private:
	//  Private Members

	//  Private Functions
	WORD	ParseSystemCommand(char *szCommand, char *szOptions, int iThreadID);			//  Parse a System Command
	WORD	SystemCommandHandler(WORD wCommand, char *szOptions, int iThreadID);			//  System command handler

protected:
	//  Protected Members
	ExecEnvironment			*xeLocal;										//  Multi-Threading Runtime
	int						ParentThread;									//  Parent Thread ID

	//  Protected Functions
	WORD	virtual ParseCustomCommand(char *szCommand, char *szOptions, int iThreadID);	//  Parse a Custom Command
	WORD	virtual CustomCommandHandler(WORD wCommand, char *szOptions, int iThreadID);	//  Process Custom Commands
	WORD	virtual GetAutoCommand(char *szOptions, int iThreadID);							//  Get any automated commands that are ready
};
