//  MTExecutive
//
//  This class extends the common runtime suport for Domino eXplorer (DX) applications to provide additional
//  capabilities for multi-threaded applications.
//
//  Message Identifier:      DXR
//  Message Range:           0501 - 0899
//  Trace area assignment:   TRACE_CORERT (01)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  1.0.0		10/02/2001		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.0		18/09/2009		New log handler for multithreaded applications
//  3.7.0		23/12/2009		Added race auto-correction to Scheduler
//  3.7.1		06/01/2010		Prevent a race condition in Rejoin processing
//  3.7.3		06/02/2010		DbMover & DbCopier improvements
//  3.7.4		19/03/2010		Fixed various issues with log/repository configuration
//  3.8.0		15/05/2010		DCF stub calls added to core
//								Load & Execute functions added
//  3.8.3       25/08/2010		Prevent escape % characters being passed to console logger
//  3.9.0		24/07/2010		Linkage changed for logging & request posting and rejoining
//								A CommandHandler can be hooked into the monitor processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.2		18/11/2010		MTIO support added to DbMover
//  3.10.3		13/12/2010		Db Property Modifier added
//								Debug Helper added
//	3.10.4		14/01/2011		Command handler extended to support automated commands
//  3.10.5		18/02/2011		DX changes to support reporting on D/Bs that cannot be opened
//  3.10.7		17/05/2011		Fixed an issue with ACL processing
//  3.10.9		04/06/2011		Sone tidying up for Linux
//  3.10.10		14/07/2011		Performance Improvements for/to DbCopier
//  3.10.11		21/07/2011		ObjectCache added to Kernel
//  3.10.12		25/08/2011		Fix timing reporting in OpenDatabase
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.2		04/11/2011		Migration of D/B conditioning from kernel to Design Manager
//  3.11.3		12/11/2011		Fixes to Explorer and ACL Processor
//  3.11.4		27/11/2011		Adopt TimeBasedArray, fixes for ACL processor
//  3.11.5		30/11/2011		Explorer recursion improvements
//								Document count changes for R5 servers
//								Transaction scheduler last run time caclulation changed
//								ObjectCache coherence improvements
//								Additional duplicate match messages in NEResolver
//  3.11.6		10/12/2011		OpenDatabase handles Replica IDs
//								Log writer improvements
//  3.12.0		21/12/2011		DXIP introduced
//	3.13.0		20/12/2013		Additional D/B Create options
//	3.14.0		19/01/2015		x64 compatability fixes
//	3.14.1		24/02/2015		Avoid a panic in DbCopier
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2001 - 2015
//

//  Application includes
#include "MTExecutive.h"

//  Constructor
MTExecutive::MTExecutive(RunSettings *rsIn, int argc, char *argv[])
: MTFunctionsAvailable(FALSE)
{
	char		szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Initialise the members
	IsInitialised = FALSE;
	rsLocal = NULL;
	hdbRepository = NULLHANDLE;
	RepositoryCreated = FALSE;
	LoggingState = LOGSTATE_NOTOPEN;
	hnLogDoc = NULLHANDLE;
	bNRTInitialised = FALSE;
	hDefDescriptor = NULLHANDLE;
	hmqProcessor = NULLHANDLE;

	//  Set the address of the Run Settings object - this MUST be done before other initialisation processing
	rsLocal = rsIn;

	//  Initialise variables
	CanonicalUser[0] = '\0';
	AbbreviatedUser[0] = '\0';
	CommonUser[0] = '\0';

	//  Trace starting creation
	LogTrace(TRACE_CORERT, MSG_DXR0510T);

	//  Invoke the Base Class initialisation
	IsInitialised = InitEnvironment(argc, argv);	//  Perform the Runtime initialisation

	//  If the base runtime has not initialised then skip the rest of the processing
	if (!IsInitialised) 
	{
		LogMessage(MSG_DXR0506E);
		return;
	}

	//  Trace creation of the ThreadManager
	LogTrace(TRACE_CORERT, MSG_DXR0507T);

	//  Create the Thread Manager
	tmLocal = new ThreadManager(this);

	//  Trace creation of the ThreadManager
	LogTrace(TRACE_CORERT, MSG_DXR0508T);

	//  Start the Thread Manager
	if (!tmLocal->Start())
	{
		LogMessage(MSG_DXR0501E);
		return;
	}

	//  Trace creation of the ThreadManager
	LogTrace(TRACE_CORERT, MSG_DXR0509T);

	//  Mark multi-threaded functions as available
	MTFunctionsAvailable = TRUE;

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0503T, this);
	LogTrace(TRACE_CORERT, szMsg);
}

//  Constructor
MTExecutive::MTExecutive(RunSettings *rsIn, ThreadManagerPolicy *tmpIn, int argc, char *argv[])
: MTFunctionsAvailable(FALSE)
{
	char		szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Initialise the members
	IsInitialised = FALSE;
	rsLocal = NULL;
	hdbRepository = NULLHANDLE;
	RepositoryCreated = FALSE;
	LoggingState = LOGSTATE_NOTOPEN;
	hnLogDoc = NULLHANDLE;
	bNRTInitialised = FALSE;
	hDefDescriptor = NULLHANDLE;
	hmqProcessor = NULLHANDLE;

	//  Set the address of the Run Settings object - this MUST be done before other initialisation processing
	rsLocal = rsIn;

	//  Initialise variables
	CanonicalUser[0] = '\0';
	AbbreviatedUser[0] = '\0';
	CommonUser[0] = '\0';

	//  Trace starting creation
	LogTrace(TRACE_CORERT, MSG_DXR0510T);

	//  Invoke the Base Class initialisation
	IsInitialised = InitEnvironment(argc, argv);	//  Perform the Runtime initialisation

	//  If the base runtime has not initialised then skip the rest of the processing
	if (!IsInitialised) 
	{
		LogMessage(MSG_DXR0506E);
		return;
	}

	//  Trace creation of the ThreadManager
	LogTrace(TRACE_CORERT, MSG_DXR0507T);

	//  Create the Thread Manager
	tmLocal = new ThreadManager(this);

	//  Set the policy
	tmLocal->Policy = tmpIn;

	//  Trace creation of the ThreadManager
	LogTrace(TRACE_CORERT, MSG_DXR0508T);

	//  Start the Thread Manager
	if (!tmLocal->Start())
	{
		LogMessage(MSG_DXR0501E);
		return;
	}

	//  Trace creation of the ThreadManager
	LogTrace(TRACE_CORERT, MSG_DXR0509T);

	//  Mark multi-threaded functions as available
	MTFunctionsAvailable = TRUE;

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0503T, this);
	LogTrace(TRACE_CORERT, szMsg);
}


//  Destructor
MTExecutive::~MTExecutive(void)
{
	char		szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  If multithreading is still active then close it down
	if (MTFunctionsAvailable)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0504T, this);
		LogTrace(TRACE_CORERT, szMsg);
		Close();
	}
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0505T, this);
	LogTrace(TRACE_CORERT, szMsg);
}

//  Close
//
//  This function will close down all runtime support capabilities that were initialised for this application,
//  capabilities include the thread manager, the logging interface and the Notes runtime. 
//  This function overrides the Close() function in the base ExecEnvironment class.
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if all functions were closed successfully
//                  FALSE if not all functions could be closed
//
//  API Use Notes:
//
//
BOOL MTExecutive::Close()
{
	LogVerbose(MSG_DXR0506I);

	//  Invalidate the Repository DB hnadle (if in use) in case it has been mapped into worker threads
	if (hdbRepository != NULLHANDLE) InvalidateNativeDBH(hdbRepository, 0);

	//  If the Thread Manager is active the drain it
	if (MTFunctionsAvailable)
	{
		if (!tmLocal->Drain())
		{
			LogMessage(MSG_DXR0502E);
		}
		MTFunctionsAvailable = FALSE;
	}

	//  Destroy the Thread Manager
	if (tmLocal != NULL)
	{
		delete tmLocal;
		tmLocal = NULL;
	}

	//  Invoke the base close for all core functions
	return CoreClose();
}

//  CoreLogMessage
//
//  This function will write a message to the appropriate logger interface, either writing it directly
//  to the main application thread logger or routing it via the asynchronous logger.
//  This function overrides the CoreLogMessage() function in the base ExecEnvironment class.
//
//  Parameters
//
//  1  -  char *		-  Pointer to the message string to be written
//  2  -  int			-  The ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the message was logged (or queued)
//                  FALSE if the message could not be logged
//
//  API Use Notes:
//
//
BOOL MTExecutive::CoreLogMessage(char *szMsg, TIMEDATE *tdEvt, int iThreadID)
{
	//  If the logging is being done by the Monitor thread asynchronlously then the message is
	//  routed to the log pool. Otherwise it is written to the logging interface directly.

	if (LoggingState <= LOGSTATE_ACTIVE) 
	{
		//  If log status is ok then write the message & Application Logging is allowed
		if ((LoggingState == LOGSTATE_ACTIVE) && (!rsLocal->NoAppLog))
		{
			if (!LogEvent(szMsg, tdEvt, iThreadID))
			{
				//  Event logging failed - terminate the event log
				LoggingState = LOGSTATE_CLOSED;
				LogMessage(MSG_DXR0007E);
				CloseLogDoc();
				return FALSE;
			}
		}

		//  If requested or logging not available then echo the message to the console
		if ((rsLocal->EchoLog) || (LoggingState < LOGSTATE_ACTIVE))
		{
			//  If the application is running on a Server then echo to the Domino Console
			if ((RunningOnServer) && (bNRTInitialised))
			{
				//  Eliminate Escapae characters
				RemoveEscapes(szMsg, MAX_MSG+1);
				AddInLogMessageText(szMsg, NOERROR);
			}
			else
			{
				std::cout << szMsg << std::endl;
			}
		}
		return TRUE;
	}
	else
	{
		//  If requested then echo the message to the console
		if (rsLocal->EchoLog)
		{
			//  If the application is running on a Server then echo to the Domino Console
			if ((RunningOnServer) && (bNRTInitialised))
			{
				//  Eliminate Escapae characters
				RemoveEscapes(szMsg, MAX_MSG+1);
				AddInLogMessageText(szMsg, NOERROR);
			}
			else
			{
				std::cout << szMsg << std::endl;
			}
		}
		return tmLocal->CorePostALogMessage(szMsg, iThreadID);
	}
}

//  GetBuildInfo
//
//  This function will return the current build information for this Runtime.
//
//  Parameters
//
//  1  -  char *			-  Pointer to a string buffer that will receive the 
//  2  -  int				-  The size of the buffer where the build info will be returned
//
//  Returns
//
//  API Use Notes:
//
//
void MTExecutive::GetBuildInfo_s(char * szBuild, int iSize)
{
	//  Return the build identity
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1400))
	sprintf_s(szBuild, iSize, MRT_IDENTITY, MRT_BUILD);
#else
	sprintf(szBuild, MRT_IDENTITY, MRT_BUILD);
#endif
	return;
}

//  PostARequest
//
//  This function will post a request into the execution queue for the Thread Pool Manager.
//  The function is a wrapper for the CorePostARequest() function in the ThreadManager
//
//  Parameters
//
//	1  -  int		Flags that determins how the request will be posted. See PXR_ 
//  2  -  void *	Pointer to the owning object for this request, can be NULL if the request will not rejoin
//  3  -  void *	Pointer to the object that will execute the request
//  4  -  void *	Pointer to the parameter object for this request
//  5  -  UNINT		Attributes of the request. See RQATTR_
//  6  -  int		The initial priority of the request
//  7  -  int		The Thread ID of the requester
//
//  Returns
//
//  BOOL  -			TRUE if the request was posted to the queue successfully
//                  FALSE if the request could not be posted
//
//  API Use Notes:
//
//
BOOL MTExecutive::PostARequest(int iFlags, void *pOwner, void *pExecutor, void *pParms, UINT Attrs, int iPriority, int iThreadID)
{
	if (!MTFunctionsAvailable) return FALSE;
	return tmLocal->CorePostARequest(iFlags, pOwner, pExecutor, pParms, Attrs, iPriority, iThreadID);
}

//  IsRejoinRecommended
//
//  This function will determine if the state of the internal work pools indicates that Rejoin
//  processing should be done now.
//
//  Parameters
//
//  1  -  void *	Pointer to the owning object for this request
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE is rejoin processing should be done, FALSE if not
//
//  API Use Notes:
//
//
BOOL MTExecutive::IsRejoinRecommended(void *pOwner, int iThreadID)
{
	return tmLocal->CoreIsRejoinRecommended(pOwner, iThreadID);
}

//  GetRejoinRequest
//
//  This function will retrieve an Exec Request that has been executed by the Thread Pool Manager and
//  is ready to rejoin the main application flow.
//
//  Parameters
//
//  1  -  void *	Pointer to the owning object for this request
//  2  -  void **	Pointer to a pointer that will be populated with a pointer to the parameter object that is ready
//  3  -  int		ID of the thread that is making the request
//
//  Returns
//
//  int  -			See RJR_ for the values returned
//
//  API Use Notes:
//
//
int MTExecutive::GetRejoinRequest(void *pOwner, void **pParms, int iThreadID)
{
	if (!MTFunctionsAvailable) 
	{
		*pParms = NULL;
		return RJR_NONE_EXIST;
	}
	return tmLocal->CoreGetRejoinRequest(pOwner, pParms, iThreadID);
}

//  PeekQueues
//
//  This function will peek all of the transaction queues and return the count of requests owned by the
//  designated owner.
//
//  Parameters
//
//  1  -  void *	Pointer to the owning object for any requests
//  3  -  int		ID of the thread that is making the request
//
//  Returns
//
//  DWORD  -		Count of transactions for the designated owner
//
//  API Use Notes:
//
//
DWORD MTExecutive::PeekQueues(void *pOwner, int iThreadID)
{
	if (!MTFunctionsAvailable) return 0;
	return tmLocal->CorePeekQueues(pOwner, iThreadID);	
}

//  GetMappedDBH
//
//  This function will return the DB Handle that can be used on the current thread that maps to
//  the database identified by the native DB Handle that was passed.
//
//  Parameters
//
//	1  -  DBHANDLE	Native D/B handle for the requested database
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  DBHANDLE  -		Local (to this thread) DB Handle for the database
//                  NULLHANDLE - A local DB Handle could not be made available
//
//  API Use Notes:
//
//
DBHANDLE MTExecutive::GetMappedDBH(DBHANDLE hdbNative, int iThreadID)
{
	if (!MTFunctionsAvailable) return NULLHANDLE;
	return tmLocal->CoreGetMappedDBH(hdbNative, iThreadID);	
}

//  InvalidateNativeDBH
//
//  This function will mark a Native Database Handle as no longer needed in the DB Handle mapping table.
//  This will allow any threads to release the per thread handle opened for the database.
//
//  Parameters
//
//	1  -  DBHANDLE	Native D/B handle that is no longer needed
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the DB Handle was marked as no longer needed
//                  FALSE if the DB Handle could not be marked as no longer needed
//
//  API Use Notes:
//
//
BOOL MTExecutive::InvalidateNativeDBH(DBHANDLE hdbNative, int iThreadID)
{
	if (!MTFunctionsAvailable) return FALSE;
	return tmLocal->CoreInvalidateNativeDBH(hdbNative, iThreadID);	
}

//  SetFastDrainMode
//
//  This function block the dispatcher from starting any new requests and modify the behaviour of
//  the Rejoin functions so that they ignore queued requests..
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if mode was set successfully
//                  FALSE if the mode could not be set
//
//  API Use Notes:
//
//
BOOL MTExecutive::SetFastDrainMode(void)
{
	if (!MTFunctionsAvailable) return FALSE;
	return tmLocal->CoreSetFastDrainMode();	
}

//  ShowThreadStats
//
//  This function will display the Thread manager statistics
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//

void MTExecutive::ShowThreadStats()
{
	tmLocal->CoreShowThreadStats(0, 0);
	return;
}

//  LogFlush
//
//  This function will force an update (write) of the current log document if the master (0) thread still
//  owns the log or it will post a signal for the Logger (Monitor) thread.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//

void MTExecutive::LogFlush()
{
	//  If the master thread is in charge then just flush the log document
	if (LoggingState == LOGSTATE_ACTIVE)
	{
		if (hnLogDoc != NULLHANDLE) NSFNoteUpdate(hnLogDoc, UPDATE_FORCE);				//  Update the note
		return;
	}

	//  If the asynchronous logger is active then post a flusf request to the logger thread
	if (LoggingState == LOGSTATE_ASYNCH)
	{
		tmLocal->CoreLogFlush();
	}
}

//  AcquireAppMutex
//
//  This function will gain control of one of the application defined mutexes
//
//  Parameters
//  1  -  int			-  The ID of the application mutex to be acquired
//  2  -  int			-  The ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the Mutex was acquired
//                  FALSE if the Mutex could not be acquired
//
//  API Use Notes:
//
//
BOOL MTExecutive::AcquireAppMutex(int iMutexID, int iThreadID)
{
	//  Pass the call to the core lock routine
	return tmLocal->CoreAcquireAppMutex(tmLocal->ThreadPool, iMutexID, iThreadID);
}

//  FreeAppMutex
//
//  This function will relinquish control of one of the application defined mutexes
//
//  Parameters
//
//  1  -  int			-  The ID of the application mutex to be freed
//  2  -  int			-  The ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the Mutex was freed
//                  FALSE if the Mutex could not be freed
//
//  API Use Notes:
//
//
BOOL MTExecutive::FreeAppMutex(int iMutexID, int iThreadID)
{
	//  Pass the call to the core unlock routine
	return tmLocal->CoreFreeAppMutex(tmLocal->ThreadPool, iMutexID, iThreadID);
}

//  DiagnosticDump
//
//  This function will trigger a disgnostic dump of key control blocks to the log
//
//  Parameters
//
//	1		-	int			-  Scope of the dump
//  2       -   int			-  ID of the thread making the request
//
//  Returns
//
//  API Use Notes:
//
//
void MTExecutive::DiagnosticDump(int iDumpScope, int iThreadID)
{
	return tmLocal->CoreDump(iDumpScope, iThreadID);
}

//  ShowStats
//
//  This function will show statistics from the Runtime 
//
//  Parameters
//
//  1  -  int		-  options describing what statistics to show
//  2  -  int		-  Thread ID of the requesting thread
//
//  Returns
//
//
//  API Use Notes:
//
//
void MTExecutive::ShowStats(int iOpts, int iThreadID)
{
	tmLocal->CoreShowThreadStats(iOpts, iThreadID);
	return;
}

//  SetLastOpenError
//
//  This function will store the last open error code so that it can be interrogated by the caller
//
//  Parameters
//
//  1  -  STATUS	-  Last error code
//  2  -  int		-  Thread ID of the requesting thread
//
//  Returns
//
//
//  API Use Notes:
//
//
void MTExecutive::SetLastOpenError(STATUS stOpen, int iThreadID)
{
	//  If the runtime is not fully initialised yet then store the RC in stLastOpenError
	if (!MTFunctionsAvailable)
	{
		stLastOpenError = stOpen;
		return;
	}
	tmLocal->CoreSetLastOpenError(stOpen, iThreadID);
	return;
}

//  GetLastOpenError
//
//  This function will store the last open error code so that it can be interrogated by the caller
//
//  Parameters
//
//  1  -  int		-  Thread ID of the requesting thread
//
//  Returns
//
//	STATUS		-  The last API return code from D/B Open calls
//
//  API Use Notes:
//
//
STATUS MTExecutive::GetLastOpenError(int iThreadID)
{
	//  If the runtime is not fully initialised yet then get the RC from stLastOpenError
	if (!MTFunctionsAvailable)
	{
		return stLastOpenError;
	}
	return tmLocal->CoreGetLastOpenError(iThreadID);
}

//  AttachCommandHandler
//
//  This function will attach the passed Command Handler to the Thread Monitor
//
//  Parameters
//
//  1  -  CommandHandler *		-	Pointer to Command Handler to be attached
//
//  Returns
//
//
//  API Use Notes:
//
//
void MTExecutive::AttachCommandHandler(CommandHandler *chApp)
{
	tmLocal->CoreAttachCommandHandler(chApp);
	return;
}

//  GetThreadCount
//
//  This function will return the count of currently active threads
//
//  Parameters
//
//
//  Returns
//
//  int			-	The count of threads that are currently active
//
//  API Use Notes:
//
//
int MTExecutive::GetThreadCount()
{
	return tmLocal->CoreGetThreadCount();
}

//  GetCMLSRecommendation
//
//  This function will determine if lane restrictions are in place for the indicated CMLS execution level
//
//  Parameters
//
//  1  -  int		-	The CMLS level 
//
//  Returns
//
//	int		-  The CMLS dispatch recommendation
//
//  API Use Notes:
//
//
int MTExecutive::GetCMLSRecommendation(int iLevel, int iThreadID)
{
	return tmLocal->CoreGetCMLSRecommendation(iLevel, iThreadID);
}

//  GetMappedFileHandle
//
//  This function will return the File Handle that can be used on the current thread that maps to
//  the file identified by the native file Handle that was passed.
//
//  Parameters
//
//	1  -  HANDLE	- Native file handle for the requested file   (int for UNIX)
//  2  -  char *	- Pointer to the file name
//  3  -  DWORD		- Access Flags   (int for UNIX)
//  4  -  DWORD		- Sharing Mode   (omitted for UNIX)
//  5  -  int		ID of the thread that is making the request
//
//  Returns
//
//  HANDLE  -		Local (to this thread) file Handle for the file   (int for UNIX)
//                  NULLHANDLE - A local File Handle could not be made available (NULL for UNIX)
//
//  API Use Notes:
//
//
#ifdef WIN32
HANDLE MTExecutive::GetMappedFileHandle(HANDLE hfNative, char *szFile, DWORD dwAccess, DWORD dwSharing, int iThreadID)
{
	if (!MTFunctionsAvailable) return NULLHANDLE;
	return tmLocal->CoreGetMappedFileHandle(hfNative, szFile, dwAccess, dwSharing, iThreadID);
}
#endif
#ifdef UNIX
int MTExecutive::GetMappedFileHandle(int fNative, char *szFile, int iFlags, int iThreadID)
{
	if (!MTFunctionsAvailable) return -1;
	return tmLocal->CoreGetMappedFileHandle(fNative, szFile, iFlags, iThreadID);
}
#endif

//  InvalidateNativeFileHandle
//
//  This function will mark a Native File Handle as no longer needed in the File Handle mapping table.
//  This will allow any threads to release the per thread handle opened for the file.
//
//  Parameters
//
//	1  -  HANDLE	Native file handle that is no longer needed   (int for UNIX)
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the File Handle was marked as no longer needed
//                  FALSE if the File Handle could not be marked as no longer needed
//
//  API Use Notes:
//
//
#ifdef WIN32
BOOL MTExecutive::InvalidateNativeFileHandle(HANDLE hfNative, int iThreadID)
{
	if (!MTFunctionsAvailable) return FALSE;
	return tmLocal->CoreInvalidateNativeFileHandle(hfNative, iThreadID);	
	return FALSE;
}
#endif
#ifdef UNIX
BOOL MTExecutive::InvalidateNativeFileHandle(int fNative, int iThreadID)
{
	if (!MTFunctionsAvailable) return FALSE;
//	return tmLocal->CoreInvalidateNativeFileHandle(fNative, iThreadID);	
	return FALSE;
}
#endif

