#pragma once
//  MTExecutive
//
//  This class extends the common runtime suport for Domino eXplorer (DX) applications to provide additional
//  capabilities for multi-threaded applications.
//
//  Message Identifier:      DXR
//  Message Range:           0501 - 0899
//  Trace area assignment:   TRACE_CORERT (01)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  1.0.0		10/02/2001		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.0		18/09/2009		New log handler for multithreaded applications
//  3.7.0		23/12/2009		Added race auto-correction to Scheduler
//  3.7.1		06/01/2010		Prevent a race condition in Rejoin processing
//  3.7.3		06/02/2010		DbMover & DbCopier improvements
//  3.7.4		19/03/2010		Fixed various issues with log/repository configuration
//  3.8.0		15/05/2010		DCF stub calls added to core
//								Load & Execute functions added
//  3.8.3       25/08/2010		Prevent escape % characters being passed to console logger
//  3.9.0		24/07/2010		Linkage changed for logging & request posting and rejoining
//								A CommandHandler can be hooked into the monitor processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.2		18/11/2010		MTIO support added to DbMover
//  3.10.3		13/12/2010		Db Property Modifier added
//								Debug Helper added
//	3.10.4		14/01/2011		Command handler extended to support automated commands
//  3.10.5		18/02/2011		DX changes to support reporting on D/Bs that cannot be opened
//  3.10.7		17/05/2011		Fixed an issue with ACL processing
//  3.10.9		04/06/2011		Sone tidying up for Linux
//  3.10.10		14/07/2011		Performance Improvements for/to DbCopier
//  3.10.11		21/07/2011		ObjectCache added to Kernel
//  3.10.12		25/08/2011		Fix timing reporting in OpenDatabase
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.2		04/11/2011		Migration of D/B conditioning from kernel to Design Manager
//  3.11.3		12/11/2011		Fixes to Explorer and ACL Processor
//  3.11.4		27/11/2011		Adopt TimeBasedArray, fixes for ACL processor
//  3.11.5		30/11/2011		Explorer recursion improvements
//								Document count changes for R5 servers
//								Transaction scheduler last run time caclulation changed
//								ObjectCache coherence improvements
//								Additional duplicate match messages in NEResolver
//  3.11.6		10/12/2011		OpenDatabase handles Replica IDs
//								Log writer improvements
//  3.12.0		21/12/2011		DXIP introduced
//	3.13.0		20/12/2013		Additional D/B Create options
//	3.14.0		19/01/2015		x64 compatability fixes
//	3.14.1		24/02/2015		Avoid a panic in DbCopier
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2001 - 2015
//

//  Application Includes
#include	"../ExecEnvironment.h"					//  Base DX common runtime
#include	"../Threads/ThreadManager.h"			//  Thread Manager

//  Identification Definitions
#define		MRT_BUILD			109
#define		MRT_IDENTITY		"DXCommon MT Runtime 3.14.1 (build: %i)"

//  Trace Messages
#define		MSG_DXR0503T		"DXR0503T: A new MTExecutive has been has been created @%p."
#define		MSG_DXR0504T		"DXR0504T: The MTExecutive @%p is closing down in the destructor."
#define		MSG_DXR0505T		"DXR0505T: An MTExecutive @%p has been destroyed."
#define		MSG_DXR0507T		"DXR0507T: MTExecutive is creating the Thread Manager."
#define		MSG_DXR0508T		"DXR0508T: MTExecutive is starting the Thread Manager."
#define		MSG_DXR0509T		"DXR0509T: MTExecutive has started the Thread Manager."
#define		MSG_DXR0510T		"DXR0510T: Creation of a new MTExecutive has started."

//  Information Messages
#define		MSG_DXR0506I		"DXR0506I: The MTExecutive is closing down."

//  Error Messages
#define		MSG_DXR0501E		"DXR0501E: Unable to start the Thread Manager, multithreading functions are not available."
#define		MSG_DXR0502E		"DXR0502E: Unable to drain the thread manager, it will be destroyed."
#define		MSG_DXR0506E		"DXR0506E: Failed to create an MTExecutive because the base runtime failed to initialise."

class MTExecutive :
	public ExecEnvironment
{
public:	
	//  Constructor
	MTExecutive(RunSettings *rsIn, int argc, char *argv[]);
	MTExecutive(RunSettings *rsIn, ThreadManagerPolicy *tmpIn, int argc, char *argv[]);

	//  Destructor
	~MTExecutive(void);

	//  Public Members
	BOOL MTFunctionsAvailable;											//  Flag indicating if Multi-Threading is available

	//  Public Functions

	//  Request Handling Functions
	BOOL PostARequest(int iFlags, void *pOwner, void *pExecutor, void *pParms, UINT Attrs, int iPriority, int iThreadID);		//  Post an Exec Request
	int GetRejoinRequest(void *pOwner, void **pParms, int iThreadID);   //  Get request that is ready to rejoin
	BOOL IsRejoinRecommended(void *pOwner, int iThreadID);				//  Get Rejoin process recommendation
	DWORD PeekQueues(void *pOwner, int iThreadID);						//  Peek queues for any request owned by

	//  Lightweight Mutex Functions
	BOOL AcquireAppMutex(int iMutexID, int iThreadID);					//  Acquire one of the application mutexes
	BOOL FreeAppMutex(int iMutexID, int iThreadID);						//  Free one of the application mutexes
	
	//  Constrained Multi-Lane Scheduling (CMLS) Functions
	int GetThreadCount(void);											//  Get the count of active threads
	int GetCMLSRecommendation(int iLevel, int iThreadID);				//  Check CMLS lane restrictions

	//  Thread Local File Handle Support Functions
#ifdef WIN32
	HANDLE GetMappedFileHandle(HANDLE hfNative, char *szFile, DWORD dwAccess, DWORD dwSharing, int iThreadID);			//  Get a locally mapped File Handle
	BOOL InvalidateNativeFileHandle(HANDLE hfNative, int iThreadID);	//  Invalidate a Native File Handle
#endif
#ifdef UNIX
	int GetMappedFileHandle(int fNative, char *szFile, int iFlags, int iThreadID);	//  Get a locally mapped File Handle
	BOOL InvalidateNativeFileHandle(int fNative, int iThreadID);		//  Invalidate a Native File Handle
#endif

	//  Miscellaneous Functions
	void ShowThreadStats();												//  Show the thread manager stats
	void DiagnosticDump(int iDumpScope, int iThreadID);					//  Perform a diagnostic dump
	void AttachCommandHandler(CommandHandler *chApp);					//  Attach a CommandHandler

	//  Overidden Public Functions
	using ExecEnvironment::Close;
	using ExecEnvironment::CoreLogMessage;
	using ExecEnvironment::GetBuildInfo_s;
	using ExecEnvironment::LogFlush;
	using ExecEnvironment::ShowStats;
	using ExecEnvironment::GetLastOpenError;
	using ExecEnvironment::SetFastDrainMode;
	using ExecEnvironment::GetMappedDBH;
	using ExecEnvironment::InvalidateNativeDBH;

	BOOL Close(void);													//  Shutdown the executive
	void GetBuildInfo_s(char *szBuild, int iSize);						//  Get Build Info
	BOOL virtual CoreLogMessage(char *szMsg, TIMEDATE *tdEvt, int iThreadID);		//  Write a log message
	void virtual LogFlush(void);										//  Force a flush of the log
	void virtual ShowStats(int iOpts, int iThreadID);					//  Show statistics
	STATUS virtual GetLastOpenError(int iThreadID);						//  Get last D/B Open Error
	BOOL virtual SetFastDrainMode();									//  Set Fast draining mode

	//  Thread Local DBHANDLE Support Functions
	DBHANDLE virtual GetMappedDBH(DBHANDLE hdbNative, int iThreadID);			//  Get a locally mapped Database Handle
	BOOL virtual InvalidateNativeDBH(DBHANDLE hdbNative, int iThreadID);		//  Invalidate a Native Database Handle

private:
	//  Private Members
	ThreadManager		*tmLocal;										//  Thread Manager

	//  Private Functions

protected:
	//  Protected

	//  Overidden Protected Functions
	using ExecEnvironment::SetLastOpenError;

	void virtual SetLastOpenError(STATUS stOpen, int iThreadID);		//  Set last D/B open Error code
};
