//  TransactionHandler
//
//  This class provides the standard handling of the transaction queues.
//  The class must be extended to provide the necessary interfaces for handling the serialisation and
//  marshalling of transaction documents to/from Request Objects.
//
//  Message Identifier:      DXR
//  Message Range:           2810 - 2899
//  Trace area assignment:   TRACE_TXH (07)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  3.10.1		13/10/2010		Initial Version
//  3.10.3		13/12/2010		Debug Helper added
//  3.10.8		30/05/2011		Transaction scheduling added to transaction handler
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.5		30/11/2011		Calculation of last run time changed
//  3.14.1		20/01/2015		x64 compatability
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2015
//

//  Application Includes
#include "TransactionHandler.h"

//  Default Constructor
TransactionHandler::TransactionHandler(void)
{
}

//  Constructor
TransactionHandler::TransactionHandler(MTExecutive *xeParent, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Save the Runtime Environment
	xeLocal = xeParent;
	ParentThread = iThreadID;

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2801T, this);
	xeLocal->LogTrace(TRACE_TXH, szMsg, ParentThread);
}

//  Destructor
TransactionHandler::~TransactionHandler(void)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2802T, this);
	xeLocal->LogTrace(TRACE_TXH, szMsg, ParentThread);
}

//  ProcessQueue
//
//  This function is invoked to process a transaction queue until it is told to stop
//
//  Parameters
//
//  1  -  TransactionQueue *		-  Pointer to the Transaction Queue descriptor object
//  2  -  int						-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::ProcessQueue(TransactionQueue *tqNew, int iThreadID)
{
	STATUS			stAPIRC;								//  Return code from Notes API calls
	DBHANDLE		hdbQueue;								//  Local DB Handle for the queue database
	NOTEID			nidReadyView;							//  NoteID of the Ready Queue
	NOTEID			nidDelayedView;							//  NoteID of the Delayed Queue
	NOTEID			nidIPView;								//  NoteID of the In Progress Queue
	NOTEID			nidSchedView;							//  NoteID of the Scheduled Queue
	DWORD			dwCycleStartCount;						//  Count of dispatched transactions
	int				iIdleLoops;								//  Count of continuous idle loops
	int				iIndex;									//  Generic Index
	clock_t			tStartInterval;							//  Begin Interval Time
	clock_t			tEndInterval;							//  Ending Interval Time
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	tqNew->ReturnCode = RETURN_NOTSTARTED;					// Mark the queue as not started

	//  Test for unset controls and set the defaults
	if (tqNew->MinDelay == 0) tqNew->MinDelay = DEFAULT_DELAY_QTIME;
	if (tqNew->ReQTXCycle == 0) tqNew->ReQTXCycle = DEFAULT_DELAY_QCYCLE;
	if (tqNew->DelayCycleSecs == 0) tqNew->DelayCycleSecs = DEFAULT_SLEEP_CYCLE;
	if (tqNew->MaxReqCount == 0) tqNew->MaxReqCount = DEFAULT_MAX_TRANS;

	//  Trace the handling of a new queue
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2803T, tqNew->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg, iThreadID);

	//  If we are not passed a DB Handle then we need to open the specified database
	if (tqNew->hdbQueue == NULLHANDLE)
	{
		//  Check that we have at least a D/B Name
		if (tqNew->szQDbPath[0] == '\0')
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2804E, tqNew->szQName);
			xeLocal->LogMessage(szMsg, iThreadID);
			tqNew->ReturnCode = RETURN_RQ_REJECTED;
			return;
		}

		//  Open the database
		hdbQueue = xeLocal->OpenDatabase(tqNew->szQServer, tqNew->szQDbPath, iThreadID);
		if (hdbQueue == NULLHANDLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2805E, tqNew->szQName, tqNew->szQServer, tqNew->szQDbPath);
			xeLocal->LogMessage(szMsg, iThreadID);
			tqNew->ReturnCode = RETURN_RQ_REJECTED;
			return;
		}
	}
	else
	{
		//  We were passed a database handle - we need to reopen it for use in this thread
		hdbQueue = xeLocal->GetMappedDBH(tqNew->hdbQueue, iThreadID);
		if (hdbQueue == NULLHANDLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2806E, tqNew->szQName);
			xeLocal->LogMessage(szMsg, iThreadID);
			tqNew->ReturnCode = RETURN_RQ_REJECTED;
			return;
		}
	}

	//  Make sure that we can locate the necessary views for this queue
	stAPIRC = NIFFindView(hdbQueue, tqNew->szReadyQName, &nidReadyView);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2807E, "Ready", tqNew->szReadyQName, tqNew->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		NSFDbClose(hdbQueue);
		tqNew->ReturnCode = RETURN_RQ_REJECTED;
		return;
	}

	//  If we need the In Progress view then locate it
	if (tqNew->wQueueProtocols & QPFLAG_REQUEIP_STARTUP)
	{
		stAPIRC = NIFFindView(hdbQueue, tqNew->szInProgressQName, &nidIPView);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2807E, "In Progress", tqNew->szInProgressQName, tqNew->szQName);
			xeLocal->LogMessage(szMsg, iThreadID);
			NSFDbClose(hdbQueue);
			tqNew->ReturnCode = RETURN_RQ_REJECTED;
			return;
		}
	}

	//  If we need the Delayed view then locate it
	if (tqNew->wQueueProtocols & QPFLAG_REQUE_DELAY)
	{
		stAPIRC = NIFFindView(hdbQueue, tqNew->szDelayedQName, &nidDelayedView);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2807E, "Delayed", tqNew->szDelayedQName, tqNew->szQName);
			xeLocal->LogMessage(szMsg, iThreadID);
			NSFDbClose(hdbQueue);
			tqNew->ReturnCode = RETURN_RQ_REJECTED;
			return;
		}
	}

	//  If we need the Scheduled view then locate it
	if (tqNew->wQueueProtocols & QPFLAG_MONITOR_SCHED)
	{
		stAPIRC = NIFFindView(hdbQueue, tqNew->szSchedQName, &nidSchedView);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2807E, "Scheduled", tqNew->szSchedQName, tqNew->szQName);
			xeLocal->LogMessage(szMsg, iThreadID);
			NSFDbClose(hdbQueue);
			tqNew->ReturnCode = RETURN_RQ_REJECTED;
			return;
		}
	}

	//  If requested then put any In Progress transactions onto the Delayed or Ready Queues
	if (tqNew->wQueueProtocols & QPFLAG_REQUEIP_STARTUP)
	{
		RequeueInProgress(tqNew, hdbQueue, nidIPView, iThreadID);
	}

	//  Mark the queue as active
	tqNew->ReturnCode = RETURN_INCOMPLETE;					// Mark the queue as active

	//  Loop processing the queue until the Local Permit is revoked
	if (!xeLocal->TestGlobalPermit()) tqNew->LocalPermit = FALSE;
	dwCycleStartCount = 0;
	iIdleLoops = 0;
	while (tqNew->LocalPermit)
	{
		if (!xeLocal->TestGlobalPermit()) tqNew->LocalPermit = FALSE;

		//  If there are any executing transactions then Poll the Rejoin Queue to see of they are completed
		if (tqNew->dwStarted > tqNew->dwCompleted && tqNew->LocalPermit) PollCompleted(tqNew, hdbQueue, iThreadID);

		//  If the queue is temporarily suspended - then skip over any active processing
		if (!xeLocal->TestGlobalPermit()) tqNew->LocalPermit = FALSE;
		if (!tqNew->QueueIsSuspended && tqNew->LocalPermit)
		{
			//  Attempt to dispatch ready transactions
			if (!DispatchReady(tqNew, hdbQueue, nidReadyView, iThreadID))
			{
				//  If needed - check for any delayed transactions and requeue them if appropriate
				iIdleLoops++;
				if (tqNew->wQueueProtocols & QPFLAG_REQUE_DELAY)
				{
					if ((tqNew->wQueueProtocols & QPFLAG_DELAY_IDLE) && (iIdleLoops >= IDLE_DELAY_TRIGGER))
					{
						iIdleLoops = 0;
						RequeueDelayed(tqNew, hdbQueue, nidDelayedView, iThreadID);
					}
				}	
			}
			else
			{
				//  A transaction has been dispatched
				iIdleLoops = 0;
				if (tqNew->MaxRunLimit > 0)
				{
					//  Queue run limit is in effect
					if (tqNew->dwStarted == tqNew->MaxRunLimit)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR2824I, tqNew->szQName, tqNew->dwStarted);
						xeLocal->LogMessage(szMsg, iThreadID);
						tqNew->LocalPermit = FALSE;					//  Remove the local permit
					}
				}
			}
		
			//  If there is cycle processng of the delayed queue then perform that
			if (tqNew->wQueueProtocols & QPFLAG_REQUE_DELAY)
			{
				if ((tqNew->wQueueProtocols & QPFLAG_DELAY_CYCLE) && ((int) (tqNew->dwStarted - dwCycleStartCount) >= tqNew->ReQTXCycle))
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR2822T, tqNew->dwStarted, dwCycleStartCount, tqNew->ReQTXCycle);
					xeLocal->LogTrace(TRACE_TXH, szMsg);
					dwCycleStartCount = tqNew->dwStarted;
					RequeueDelayed(tqNew, hdbQueue, nidDelayedView, iThreadID);
				}
			}
		}

		if (!xeLocal->TestGlobalPermit()) tqNew->LocalPermit = FALSE;

		//  Check for and schedule any transactions that are now due for execution
		if (tqNew->LocalPermit)
		{
			if (tqNew->wQueueProtocols & QPFLAG_MONITOR_SCHED) ScheduleTransactions(tqNew, hdbQueue, nidSchedView, iThreadID);
		}

		//  Now pause for the cycle delay time for this queue - poll for any completions
		if (!xeLocal->TestGlobalPermit()) tqNew->LocalPermit = FALSE;
		tStartInterval = xeLocal->RunningTime->getElapsed();
		for (iIndex = 0; iIndex < tqNew->DelayCycleSecs; iIndex++)
		{
			if (!xeLocal->TestGlobalPermit()) tqNew->LocalPermit = FALSE;
			//  If the permit has been revoked on the queue then don't wait or requeue
			if (tqNew->LocalPermit)
			{
				xeLocal->AppWait(1000);										//  Delay for 1 second
				if (tqNew->dwStarted > tqNew->dwCompleted) PollCompleted(tqNew, hdbQueue, iThreadID);
			}
		}
		tEndInterval = xeLocal->RunningTime->getElapsed();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2823T, tEndInterval - tStartInterval);
		xeLocal->LogTrace(TRACE_TXH, szMsg);
	}

	//  Drain any transactions already posted 
	if (!xeLocal->TestGlobalPermit()) tqNew->LocalPermit = FALSE;
	if (tqNew->LocalPermit)
	{
		while (PollCompleted(tqNew, hdbQueue, iThreadID) != RJR_NONE_EXIST) SLEEPSECONDS(tqNew->DelayCycleSecs);
	}

	//  If we were passed a DBHandle then invalidate it - if we opened a handle then close it.
	if (tqNew->hdbQueue == NULLHANDLE) NSFDbClose(hdbQueue);
	else xeLocal->InvalidateNativeDBH(tqNew->hdbQueue, iThreadID);

	//  Mark the queue as closed
	tqNew->ReturnCode = RETURN_NOERROR;					// Mark the queue as closed

	return;
}

//  PollCompleted
//
//  This function will Poll the rejoin pool for any completed transaction
//
//  Parameters
//
//  1  -  TransactionQueue *	- Pointer to the Queue descriptor object
//  2  -  DBHANDLE				-  Handle of the Transaction Queue Database
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  int			-  The last rejoin return code
//
//  API Use Notes:
//
//
int TransactionHandler::PollCompleted(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, int iThreadID)
{
	int			iProcessed;										//  Count processed on this invocation
	int			iRejoinRC;										//  Return code from Rejoin calls
	void		*txReturned;									//  Transaction object returned

	iProcessed = 0;
	iRejoinRC = xeLocal->GetRejoinRequest(tqCurrent, (void **) &txReturned, iThreadID);
	while (iRejoinRC == RJR_RETURNED)
	{
		tqCurrent->dwCompleted++;								//  Increment the completed count
		iProcessed++;

		//  Invoke the Serialize Interface to dispose of the transaction object
		SerializeTransaction(txReturned, hdbQueue, tqCurrent, iThreadID);

#if (defined(_DEBUG) && defined(WIN32))
		//  If a Debug Helper has been inserted then show the memory usage statistics
		if (DebugHelper != NULL)
		{
			DebugHelper->ReportMemoryUsage(FALSE, TRUE, iThreadID);
		}
#endif
		iRejoinRC = xeLocal->GetRejoinRequest(tqCurrent, (void **) &txReturned, iThreadID);
	}

	return iRejoinRC;
}

//  DispatchReady
//
//  This function will dispatch any ready transactions 
//
//  Parameters
//
//  1  -  TransactionQueue *	-  Pointer to the Queue descriptor object
//  2  -  DBHANDLE				-  Handle of the Transaction Queue Database
//  3  -  NOTEID				-  Note ID of the Ready View
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL		-		TRUE if anything was dispatched or max concurrent transactions were executing
//						FALSE if there was no work to do.
//
//  API Use Notes:
//
//
BOOL TransactionHandler::DispatchReady(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, NOTEID nidView, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	HCOLLECTION			hcQueue;								//  Handle of the In Progress View
	COLLECTIONPOSITION	cpQueue;								//  Position in the In Progress view
	DHANDLE				hBufferQueueNoteIDs;					//	Handle to buffer of NoteIDs 
	NOTEID				*nidList;								//  Pointer to the returned Note IDs
	NOTEHANDLE			hnTransaction;							//  Handle to the transaction document
	DWORD				dwQueueEntriesFound;					//  Count of entries located
	DWORD				dwMaxTrans;								//  Max Transactions
	DWORD				dwIndex;								//  Generic index
	DWORD				dwDeleted = 0;							//  Deleted entries
	WORD				wQueueSignalFlag;						//  Signal flags
	int					Disposition;							//  Transaction Disposition
	void				*txObject;								//  Transaction Object
	void				*txExecutor;							//  Transaction Executor
	UINT				Attrs;									//  Transaction Attributes
	int					Priority;								//  Transaction Priority
	int					iRunning;								//  Running transactions
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Check if we still have a Permit to process
	if (!tqCurrent->LocalPermit) return TRUE;

	//  Check if we have the max concurrent transactions already executing
	iRunning = (int) (tqCurrent->dwStarted - tqCurrent->dwCompleted);
	if (iRunning >= tqCurrent->MaxConcurrent) return TRUE;

	//  Open the view of Ready transactions
	stAPIRC = NIFOpenCollection(hdbQueue, hdbQueue, nidView, 0, NULLHANDLE, &hcQueue, NULLHANDLE, NULL, NULLHANDLE, NULLHANDLE);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2809E, "Ready", tqCurrent->szReadyQName, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		tqCurrent->LocalPermit = FALSE;
		return TRUE;
	}

	//  Set position to the start of the view and determine how many trans can be read
	cpQueue.Level = 0;
	cpQueue.Tumbler[0] = 0;
	if (tqCurrent->MaxConcurrent == 0) dwMaxTrans = DEFAULT_MAX_TRANS;
	else dwMaxTrans = tqCurrent->MaxConcurrent - iRunning;

	if (dwMaxTrans == 0)
	{
		NIFCloseCollection(hcQueue);
		return TRUE;
	}

	//  Read the transactions from the view - we only read that max allowed on a single scheduling
	stAPIRC = NIFReadEntries(hcQueue,					// handle to this collection
							&cpQueue,					// where to start in collection
							NAVIGATE_NEXT,				// order to use when skipping
							1L,							// number to skip
							NAVIGATE_NEXT,				// order to use when reading
							dwMaxTrans,					// max number to read
							READ_MASK_NOTEID,			// info we want
							&hBufferQueueNoteIDs,		// handle to info buffer (return)
							NULL,						// length of info buffer (return)
							NULL,						// entries skipped (return)
							&dwQueueEntriesFound,		// entries read (return)
							&wQueueSignalFlag);			// share warning and more signal flag (return)

	//  Check the return code
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2810E, "Ready", tqCurrent->szReadyQName, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		NIFCloseCollection(hcQueue);
		tqCurrent->LocalPermit = FALSE;
		return TRUE;
	}

	//  Check for nothing returned
	if (hBufferQueueNoteIDs == NULLHANDLE)
	{
		NIFCloseCollection(hcQueue);
		return FALSE;
	}

	//  Check again
	if (dwQueueEntriesFound == 0)
	{
		OSMemFree(hBufferQueueNoteIDs);
		NIFCloseCollection (hcQueue);
		return FALSE;
	}

	//  We have entries to process - trace the start of processing
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2816T, dwQueueEntriesFound, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg);

	//  Process each of the transactions in turn
	nidList = (NOTEID *) OSLockObject (hBufferQueueNoteIDs);
	for (dwIndex = 0; dwIndex < dwQueueEntriesFound; dwIndex++)
	{
		if (nidList[dwIndex] & RRV_DELETED) dwDeleted++;
		else
		{
			//  Open the transaction
			stAPIRC = NSFNoteOpen(hdbQueue, nidList[dwIndex], 0, &hnTransaction);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2812E, nidList[dwIndex], "Ready", tqCurrent->szQName);
				xeLocal->LogMessage(szMsg, iThreadID);
			}
			else
			{
				//  Invoke the marshalling interface to create the transaction object and determine what action to take
				Disposition = MarshallTransaction(nidList[dwIndex], hnTransaction, tqCurrent, &txObject, &txExecutor, &Attrs, &Priority, iThreadID);

				//  Take action based on the returned disposition
				switch(Disposition)
				{
				case TX_DISPOSITION_NORMAL:
					//  Dispatch the transaction request and mark the transaction document In Progress
					if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
												tqCurrent,
												txExecutor,
												txObject,
												Attrs,
												Priority,
												iThreadID))
					{
						//  This is an SNO (Should Not Occur) condition - this interface will block on resource shortage
						sprintf_s(szMsg, MAX_MSG, MSG_DXR2818E, txObject, tqCurrent->szQName);
						xeLocal->LogMessage(szMsg, iThreadID);
						tqCurrent->LocalPermit = FALSE;
						MarkTransactionError(nidList[dwIndex], hnTransaction, tqCurrent, iThreadID);
					}
					else 
					{
						MarkTransactionInProgress(nidList[dwIndex], hnTransaction, tqCurrent, iThreadID);
						tqCurrent->dwStarted++;
					}
					break;

				case TX_DISPOSITION_ERROR:
					//  Mark the transaction in error
					MarkTransactionError(nidList[dwIndex], hnTransaction, tqCurrent, iThreadID);
					break;

				case TX_DISPOSITION_DELAY:
					//  Mark the transaction delayed
					MarkTransactionDelayed(nidList[dwIndex], hnTransaction, FALSE, tqCurrent, iThreadID);
					break;

				case TX_DISPOSITION_PANIC:
					//  Mark the transaction in error
					MarkTransactionError(nidList[dwIndex], hnTransaction, tqCurrent, iThreadID);
					tqCurrent->LocalPermit = FALSE;
					break;
					
				default:
					//  Mark the transaction in error
					MarkTransactionError(nidList[dwIndex], hnTransaction, tqCurrent, iThreadID);
					break;
				}

				//  Save and close the transaction
				NSFNoteUpdate(hnTransaction, UPDATE_FORCE);
				NSFNoteClose(hnTransaction);
			}
		}
	}

	//  Trace completion
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2813T, dwQueueEntriesFound, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg, iThreadID);

	//  Tidy up and leave
	OSUnlockObject (hBufferQueueNoteIDs);
	OSMemFree (hBufferQueueNoteIDs);
	NIFCloseCollection (hcQueue);

	return TRUE;
}

//  ReQueueInProgress
//
//  This function will requeue any transaction that were In Progress at startup
//
//  Parameters
//
//  1  -  TransactionQueue *	- Pointer to the Queue descriptor object
//  2  -  DBHANDLE				-  Handle of the Transaction Queue Database
//  3  -  NOTEID				-  Note ID of the In Progress View
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::RequeueInProgress(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, NOTEID nidView, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	HCOLLECTION			hcQueue;								//  Handle of the In Progress View
	COLLECTIONPOSITION	cpQueue;								//  Position in the In Progress view
	DHANDLE				hBufferQueueNoteIDs;					//	Handle to buffer of NoteIDs 
	NOTEID				*nidList;								//  Pointer to the returned Note IDs
	NOTEHANDLE			hnTransaction;							//  Handle to the transaction document
	DWORD				dwQueueEntriesFound;					//  Count of entries located
	DWORD				dwIndex;								//  Generic index
	DWORD				dwDeleted = 0;							//  Deleted entries
	WORD				wQueueSignalFlag;						//  Signal flags
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace the requeue action
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2808T, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg, iThreadID);

	//  Open the view of In Progress transactions
	stAPIRC = NIFOpenCollection(hdbQueue, hdbQueue, nidView, 0, NULLHANDLE, &hcQueue, NULLHANDLE, NULL, NULLHANDLE, NULLHANDLE);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2809E, "In Progress", tqCurrent->szInProgressQName, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Set position to the start of the view
	cpQueue.Level = 0;
	cpQueue.Tumbler[0] = 0;

	//  Read the transactions from the view - this can be accomplished with a single read
	stAPIRC = NIFReadEntries(hcQueue,					// handle to this collection
							&cpQueue,					// where to start in collection
							NAVIGATE_NEXT,				// order to use when skipping
							1L,							// number to skip
							NAVIGATE_NEXT,				// order to use when reading
							0xFFFFFFFF,					// max number to read
							READ_MASK_NOTEID,			// info we want
							&hBufferQueueNoteIDs,		// handle to info buffer (return)
							NULL,						// length of info buffer (return)
							NULL,						// entries skipped (return)
							&dwQueueEntriesFound,		// entries read (return)
							&wQueueSignalFlag);			// share warning and more signal flag (return)

	//  Check the return code
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2810E, "In Progress", tqCurrent->szInProgressQName, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		NIFCloseCollection(hcQueue);
		return;
	}

	//  Check for nothing returned
	if (hBufferQueueNoteIDs == NULLHANDLE)
	{
		NIFCloseCollection(hcQueue);
		return;
	}

	//  Check again
	if (dwQueueEntriesFound == 0)
	{
		OSMemFree(hBufferQueueNoteIDs);
		NIFCloseCollection (hcQueue);
		return;
	}

	//  We have entries to process - trace the start of processing
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2811T, dwQueueEntriesFound, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg);

	//  Process each of the transactions in turn
	nidList = (NOTEID *) OSLockObject (hBufferQueueNoteIDs);
	for (dwIndex = 0; dwIndex < dwQueueEntriesFound; dwIndex++)
	{
		if (nidList[dwIndex] & RRV_DELETED) dwDeleted++;
		else
		{
			//  Open the transaction
			stAPIRC = NSFNoteOpen(hdbQueue, nidList[dwIndex], 0, &hnTransaction);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2812E, nidList[dwIndex], "In Progress", tqCurrent->szQName);
				xeLocal->LogMessage(szMsg, iThreadID);
			}
			else
			{
				//  Invoke the appropriate marking process for this transaction
				if (tqCurrent->wQueueProtocols & QPFLAG_REQUEIP_READY) MarkTransactionReady(nidList[dwIndex], hnTransaction, TRUE, tqCurrent, iThreadID);
				else MarkTransactionDelayed(nidList[dwIndex], hnTransaction, TRUE, tqCurrent, iThreadID);

				//  Save and close the transaction
				NSFNoteUpdate(hnTransaction, UPDATE_FORCE);
				NSFNoteClose(hnTransaction);
			}
		}
	}

	//  Trace completion
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2813T, dwQueueEntriesFound, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg, iThreadID);

	//  Tidy up and leave
	OSUnlockObject (hBufferQueueNoteIDs);
	OSMemFree (hBufferQueueNoteIDs);
	NIFCloseCollection (hcQueue);
	return;
}

//  ReQueueDelayed
//
//  This function will attempt to requeue any transaction that have been delayed
//
//  Parameters
//
//  1  -  TransactionQueue *	- Pointer to the Queue descriptor object
//  2  -  DBHANDLE				-  Handle of the Transaction Queue Database
//  3  -  NOTEID				-  Note ID of the In Progress View
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::RequeueDelayed(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, NOTEID nidView, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	HCOLLECTION			hcQueue;								//  Handle of the In Progress View
	COLLECTIONPOSITION	cpQueue;								//  Position in the In Progress view
	DHANDLE				hBufferQueueNoteIDs;					//	Handle to buffer of NoteIDs 
	NOTEID				*nidList;								//  Pointer to the returned Note IDs
	NOTEHANDLE			hnTransaction;							//  Handle to the transaction document
	DWORD				dwQueueEntriesFound;					//  Count of entries located
	DWORD				dwIndex;								//  Generic index
	DWORD				dwDeleted = 0;							//  Deleted Entries
	WORD				wQueueSignalFlag;						//  Signal flags
	TIMEDATE			tdMod;									//  Last Modified Timestamp
	TIMEDATE			tdCurrent;								//  Current time
	LONG				lTimeDiff;								//  Time difference
	int					iRequeues = 0;							//  Count of requeues
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace the requeue action
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2820T, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg, iThreadID);

	//  Open the view of In Progress transactions
	stAPIRC = NIFOpenCollection(hdbQueue, hdbQueue, nidView, 0, NULLHANDLE, &hcQueue, NULLHANDLE, NULL, NULLHANDLE, NULLHANDLE);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2809E, "Delayed", tqCurrent->szInProgressQName, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Set position to the start of the view
	cpQueue.Level = 0;
	cpQueue.Tumbler[0] = 0;

	//  Read the transactions from the view - this can be accomplished with a single read
	stAPIRC = NIFReadEntries(hcQueue,					// handle to this collection
							&cpQueue,					// where to start in collection
							NAVIGATE_NEXT,				// order to use when skipping
							1L,							// number to skip
							NAVIGATE_NEXT,				// order to use when reading
							0xFFFFFFFF,					// max number to read
							READ_MASK_NOTEID,			// info we want
							&hBufferQueueNoteIDs,		// handle to info buffer (return)
							NULL,						// length of info buffer (return)
							NULL,						// entries skipped (return)
							&dwQueueEntriesFound,		// entries read (return)
							&wQueueSignalFlag);			// share warning and more signal flag (return)

	//  Check the return code
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2810E, "Delayed", tqCurrent->szInProgressQName, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		NIFCloseCollection(hcQueue);
		return;
	}

	//  Check for nothing returned
	if (hBufferQueueNoteIDs == NULLHANDLE)
	{
		NIFCloseCollection(hcQueue);
		return;
	}

	//  Check again
	if (dwQueueEntriesFound == 0)
	{
		OSMemFree(hBufferQueueNoteIDs);
		NIFCloseCollection (hcQueue);
		return;
	}

	//  We have entries to process - trace the start of processing
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2811T, dwQueueEntriesFound, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg);

	//  Process each of the transactions in turn
	nidList = (NOTEID *) OSLockObject (hBufferQueueNoteIDs);
	for (dwIndex = 0; dwIndex < dwQueueEntriesFound; dwIndex++)
	{
		if (nidList[dwIndex] & RRV_DELETED) dwDeleted++;
		else
		{
			//  Open the transaction
			stAPIRC = NSFNoteOpen(hdbQueue, nidList[dwIndex], 0, &hnTransaction);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2812E, nidList[dwIndex], "Delayed", tqCurrent->szQName);
				xeLocal->LogMessage(szMsg, iThreadID);
			}
			else
			{
				//  Determine how long the transaction has been on the queue
				NSFNoteGetInfo(hnTransaction, _NOTE_MODIFIED, &tdMod);
				OSCurrentTIMEDATE(&tdCurrent);
				lTimeDiff = TimeDateDifference(&tdCurrent, &tdMod);
				if (lTimeDiff > tqCurrent->MinDelay)
				{
					MarkTransactionReady(nidList[dwIndex], hnTransaction, FALSE, tqCurrent, iThreadID);
					NSFNoteUpdate(hnTransaction, UPDATE_FORCE);
					iRequeues++;
				}

				//  Close the transaction
				NSFNoteClose(hnTransaction);
			}
		}
	}

	//  Trace completion
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2813T, dwQueueEntriesFound, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg, iThreadID);

	//  Show requeue count (if there were any)
	if (iRequeues > 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2821I, iRequeues, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Tidy up and leave
	OSUnlockObject (hBufferQueueNoteIDs);
	OSMemFree (hBufferQueueNoteIDs);
	NIFCloseCollection (hcQueue);
	return;
}

//  ScheduleTransactions
//
//  This function will attempt schedule any transactions that are now due for execution
//
//  Parameters
//
//  1  -  TransactionQueue *	- Pointer to the Queue descriptor object
//  2  -  DBHANDLE				-  Handle of the Transaction Queue Database
//  3  -  NOTEID				-  Note ID of the Scheduled View
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::ScheduleTransactions(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, NOTEID nidView, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	HCOLLECTION			hcQueue;								//  Handle of the In Progress View
	COLLECTIONPOSITION	cpQueue;								//  Position in the In Progress view
	DHANDLE				hBufferQueueNoteIDs;					//	Handle to buffer of NoteIDs 
	NOTEID				*nidList;								//  Pointer to the returned Note IDs
	NOTEHANDLE			hnTransaction;							//  Handle to the transaction document
	DWORD				dwQueueEntriesFound;					//  Count of entries located
	DWORD				dwIndex;								//  Generic index
	DWORD				dwDeleted = 0;							//  Deleted Entries
	WORD				wQueueSignalFlag;						//  Signal flags
	int					iReadied = 0;							//  Count of Tx added to the ready queue
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace the schedule action
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2825T, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg, iThreadID);

	//  Open the view of scheduled transactions
	stAPIRC = NIFOpenCollection(hdbQueue, hdbQueue, nidView, 0, NULLHANDLE, &hcQueue, NULLHANDLE, NULL, NULLHANDLE, NULLHANDLE);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2809E, "Scheduled", tqCurrent->szSchedQName, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return;
	}

	//  Set position to the start of the view
	cpQueue.Level = 0;
	cpQueue.Tumbler[0] = 0;

	//  Read the transactions from the view - this can be accomplished with a single read
	stAPIRC = NIFReadEntries(hcQueue,					// handle to this collection
							&cpQueue,					// where to start in collection
							NAVIGATE_NEXT,				// order to use when skipping
							1L,							// number to skip
							NAVIGATE_NEXT,				// order to use when reading
							0xFFFFFFFF,					// max number to read
							READ_MASK_NOTEID,			// info we want
							&hBufferQueueNoteIDs,		// handle to info buffer (return)
							NULL,						// length of info buffer (return)
							NULL,						// entries skipped (return)
							&dwQueueEntriesFound,		// entries read (return)
							&wQueueSignalFlag);			// share warning and more signal flag (return)

	//  Check the return code
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2810E, "Scheduled", tqCurrent->szSchedQName, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		NIFCloseCollection(hcQueue);
		return;
	}

	//  Check for nothing returned
	if (hBufferQueueNoteIDs == NULLHANDLE)
	{
		NIFCloseCollection(hcQueue);
		return;
	}

	//  Check again
	if (dwQueueEntriesFound == 0)
	{
		OSMemFree(hBufferQueueNoteIDs);
		NIFCloseCollection (hcQueue);
		return;
	}

	//  We have entries to process - trace the start of processing
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2826T, dwQueueEntriesFound, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg);

	//  Process each of the transactions in turn
	nidList = (NOTEID *) OSLockObject (hBufferQueueNoteIDs);
	for (dwIndex = 0; dwIndex < dwQueueEntriesFound; dwIndex++)
	{
		if (nidList[dwIndex] & RRV_DELETED) dwDeleted++;
		else
		{
			//  Open the transaction
			stAPIRC = NSFNoteOpen(hdbQueue, nidList[dwIndex], 0, &hnTransaction);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2812E, nidList[dwIndex], "Scheduled", tqCurrent->szQName);
				xeLocal->LogMessage(szMsg, iThreadID);
			}
			else
			{
				//  Check if this transaction should now be copied to the ready queue
				iReadied = iReadied + CheckSchedTransaction(tqCurrent, hnTransaction, iThreadID);

				//  Close the transaction
				NSFNoteClose(hnTransaction);
			}
		}
	}

	//  Trace completion
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2827T, dwQueueEntriesFound, tqCurrent->szQName);
	xeLocal->LogTrace(TRACE_TXH, szMsg, iThreadID);

	//  Show the scheduled count (if there were any)
	if (iReadied > 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2828I, iReadied, tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Tidy up and leave
	OSUnlockObject (hBufferQueueNoteIDs);
	OSMemFree (hBufferQueueNoteIDs);
	NIFCloseCollection (hcQueue);
	return;
}

//  CheckSchedTransaction
//
//  This function will check an individual scheduled transaction to see if it is ready to be copied to
//  the ready queue
//
//  Parameters
//
//  1  -  NOTEHANDLE			-  Handle of the transaction note to be checked
//  2  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  int			-  Count added to the ready queue (0 or 1)
//
//  API Use Notes:
//
//
int TransactionHandler::CheckSchedTransaction(TransactionQueue *tqCurrent, NOTEHANDLE hnTransaction, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	NOTEHANDLE			hnNewTrans;								//  New copy of the scheduled transaction
	TIMEDATE			tdCurrent;								//  Current time
	TIMEDATE			tdFirstRun;								//  First Run Date/Time
	TIMEDATE			tdLastRun;								//  Last Run Date/Time
	TIMEDATE			tdNextRun;								//  Next Run Date/Time
	WORD				wDataType;								//  Used for field checks
	WORD				wRetLen;								//  Returned length
	BLOCKID				bidValue;								//  Used for field checks
	DWORD				dwDataLen;								//  Used for field checks
	DBHANDLE			hdbThis;								//  Local D/B Handle
	OID					oidNewTrans;							//  OID for the new transaction
	int					iReturnCount = 0;						//  Returned Count
	char				szReschedInterval[MAX_ELEMENT + 1];		//  Reschecule interval
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	OSCurrentTIMEDATE(&tdCurrent);
	if (hnTransaction == NULLHANDLE) return iReturnCount;

	//  Sanity check for the schedule information ... set defaults if not available
	stAPIRC = NSFItemInfo(hnTransaction, DEFAULT_TRANSDOC_SCHEDFIRST, 
										(WORD) strlen(DEFAULT_TRANSDOC_SCHEDFIRST),
										NULL,
										&wDataType,
										&bidValue,
										&dwDataLen);
	if (ERR(stAPIRC) == ERR_ITEM_NOT_FOUND)
	{
		//  Set the field in the document to the current time
		stAPIRC = NSFItemSetTime(hnTransaction, DEFAULT_TRANSDOC_SCHEDFIRST, &tdCurrent);
	}
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR2829E, iThreadID);
		return iReturnCount;
	}

	stAPIRC = NSFItemInfo(hnTransaction, DEFAULT_TRANSDOC_SCHEDINTERVAL, 
										(WORD) strlen(DEFAULT_TRANSDOC_SCHEDINTERVAL),
										NULL,
										&wDataType,
										&bidValue,
										&dwDataLen);
	if (ERR(stAPIRC) == ERR_ITEM_NOT_FOUND)
	{
		stAPIRC = NSFItemSetText(hnTransaction, DEFAULT_TRANSDOC_SCHEDINTERVAL,
												DEFAULT_TRANSDOC_INTERVAL_NEVER,
												(WORD) strlen(DEFAULT_TRANSDOC_INTERVAL_NEVER));
	}
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR2829E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return iReturnCount;
	}

	//  Now perform the scheduling checks
	if (!NSFItemGetTime(hnTransaction, DEFAULT_TRANSDOC_SCHEDFIRST, &tdFirstRun))
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2830E, DEFAULT_TRANSDOC_SCHEDFIRST);
		xeLocal->LogMessage(szMsg, iThreadID);
		return iReturnCount;
	}

	//  If the first run date/time is later than the wall clock time then skip out
	if (TimeDateCompare(&tdFirstRun, &tdCurrent) == 1) return iReturnCount;

	//  Get the Last Run time
	if(!NSFItemGetTime(hnTransaction, DEFAULT_TRANSDOC_SCHEDLAST, &tdLastRun))
	{
		//  There is no last run time - so this should be run now
		OSCurrentTIMEDATE(&tdNextRun);
	}
	else
	{
		//  Get the reschedule interval to compute when this transaction should next be run
		wRetLen = NSFItemGetText(hnTransaction, DEFAULT_TRANSDOC_SCHEDINTERVAL, szReschedInterval, MAX_ELEMENT);
		szReschedInterval[wRetLen] = '\0';

		//  If this transaction is never to be resheduled then leave it
		if (strcmp(szReschedInterval, DEFAULT_TRANSDOC_INTERVAL_NEVER) == 0) return iReturnCount;

		//  Compute the next run time according to the interval
		memcpy(&tdNextRun, &tdLastRun, sizeof(TIMEDATE));
		if (strcmp(szReschedInterval, DEFAULT_TRANSDOC_INTERVAL_DAILY) == 0)
		{
			if (!CycleTime(&tdNextRun, 1, 0))
			{
				xeLocal->LogMessage(MSG_DXR2831E, iThreadID);
				return iReturnCount;
			}
		}
		else
		{
			if (strcmp(szReschedInterval, DEFAULT_TRANSDOC_INTERVAL_WEEKLY) == 0)
			{
				if (!CycleTime(&tdNextRun, 7, 0))
				{
					xeLocal->LogMessage(MSG_DXR2831E, iThreadID);
					return iReturnCount;
				}
			}
			else
			{
				if (strcmp(szReschedInterval, DEFAULT_TRANSDOC_INTERVAL_MONTHLY) == 0)
				{
					if (!CycleTime(&tdNextRun, 0, 1))
					{
						xeLocal->LogMessage(MSG_DXR2831E, iThreadID);
						return iReturnCount;
					}
				}
				else
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR2832E, szReschedInterval);
					xeLocal->LogMessage(szMsg, iThreadID);
					return iReturnCount;
				}
			}
		}
		
		//  Check if the transaction is due to be run again
		if (TimeDateCompare(&tdLastRun, &tdNextRun) > -1) return iReturnCount;
	}

	//  The transaction is eligible for scheduling 
	//  Firstly update the scheduled document (for safety), then create the new Ready transaction

	stAPIRC = NSFItemSetTime(hnTransaction, DEFAULT_TRANSDOC_SCHEDLAST, &tdNextRun);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2832E, szReschedInterval);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return iReturnCount;
	}

	stAPIRC = NSFNoteUpdate(hnTransaction, UPDATE_FORCE);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2832E, szReschedInterval);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return iReturnCount;
	}
	
	//  Create a copy of the scheduled transaction and set the status to "NEW"
	stAPIRC = NSFNoteCopy(hnTransaction, &hnNewTrans);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR2834E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return iReturnCount;
	}

	//  Create a new OID for the note & clear the NoteID
	NSFNoteGetInfo(hnTransaction, _NOTE_DB, &hdbThis);
	stAPIRC = NSFDbGenerateOID(hdbThis, &oidNewTrans);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR2834E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return iReturnCount;
	}
	NSFNoteSetInfo(hnNewTrans, _NOTE_OID, &oidNewTrans);
	NSFNoteSetInfo(hnNewTrans, _NOTE_ID, NULL);
	NSFNoteSetInfo(hnNewTrans, _NOTE_DB, &hdbThis);

	MarkTransactionReady(0, hnNewTrans, FALSE, tqCurrent, iThreadID);

	//  Save the transaction
	NSFNoteUpdate(hnNewTrans, UPDATE_FORCE);
	iReturnCount = 1;
	return iReturnCount;
}

//  MarkTransactionReady
//
//  This is the default implementation for marking a passed transaction "Ready" - this implementation follows
//  the convention commonly used in transaction driven DX tooling.
//
//  Parameters
//
//  1  -  NOTEID				-  Note ID of the transaction note
//  2  -  NOTEHANDLE			-  Handle of the open transaction note
//  3  -  BOOL					-  Flag indicating if (TRUE) the transaction came from the In Progress view
//  4  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  5  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::MarkTransactionReady(NOTEID nidTX, NOTEHANDLE hnTX, BOOL bInProgress, TransactionQueue *tqCurrent, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	char				szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Mark the transaction as Ready
	stAPIRC = NSFItemSetText(hnTX, DEFAULT_TRANSDOC_STATUS, DEFAULT_TRANSDOC_STATUS_NEW, MAXWORD);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2814E, nidTX, "Ready", tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);		
	}
	return;
}

//  MarkTransactionDelayed
//
//  This is the default implementation for marking a passed transaction "Delayed" - this implementation follows
//  the convention commonly used in transaction driven DX tooling.
//
//  Parameters
//
//  1  -  NOTEID				-  Note ID of the transaction note
//  2  -  NOTEHANDLE			-  Handle of the open transaction note
//  3  -  BOOL					-  Flag indicating if (TRUE) the transaction came from the In Progress view
//  4  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  5  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::MarkTransactionDelayed(NOTEID nidTX, NOTEHANDLE hnTX, BOOL bInProgress, TransactionQueue *tqCurrent, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	char				szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Mark the transaction as Delayed
	stAPIRC = NSFItemSetText(hnTX, DEFAULT_TRANSDOC_STATUS, DEFAULT_TRANSDOC_STATUS_DELAYED, MAXWORD);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2814E, nidTX, "Delayed", tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);		
	}
	return;
}

//  MarkTransactionError
//
//  This is the default implementation for marking a passed transaction "Error" - this implementation follows
//  the convention commonly used in transaction driven DX tooling.
//
//  Parameters
//
//  1  -  NOTEID				-  Note ID of the transaction note
//  2  -  NOTEHANDLE			-  Handle of the open transaction note
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::MarkTransactionError(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	char				szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Mark the transaction as Error
	stAPIRC = NSFItemSetText(hnTX, DEFAULT_TRANSDOC_STATUS, DEFAULT_TRANSDOC_STATUS_ERROR, MAXWORD);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2814E, nidTX, "Error", tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);		
	}
	return;
}

//  MarkTransactionInProgress
//
//  This is the default implementation for marking a passed transaction "In Progress" - this implementation follows
//  the convention commonly used in transaction driven DX tooling.
//
//  Parameters
//
//  1  -  NOTEID				-  Note ID of the transaction note
//  2  -  NOTEHANDLE			-  Handle of the open transaction note
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::MarkTransactionInProgress(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	char				szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Mark the transaction as In Progress
	stAPIRC = NSFItemSetText(hnTX, DEFAULT_TRANSDOC_STATUS, DEFAULT_TRANSDOC_STATUS_INPROGRESS, MAXWORD);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2814E, nidTX, "In Progress", tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);		
	}

	return;
}

//  MarkTransactionCompleted
//
//  This is the default implementation for marking a passed transaction "Completed" - this implementation follows
//  the convention commonly used in transaction driven DX tooling.
//
//  Parameters
//
//  1  -  NOTEID				-  Note ID of the transaction note
//  2  -  NOTEHANDLE			-  Handle of the open transaction note
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::MarkTransactionCompleted(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	char				szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Mark the transaction as In Progress
	stAPIRC = NSFItemSetText(hnTX, DEFAULT_TRANSDOC_STATUS, DEFAULT_TRANSDOC_STATUS_COMPLETED, MAXWORD);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2814E, nidTX, "Completed", tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);		
	}
	return;
}

//  MarkTransactionRetried
//
//  This is the default implementation for marking a passed transaction "Retried" - this implementation follows
//  the convention commonly used in transaction driven DX tooling.
//
//  Parameters
//
//  1  -  NOTEID				-  Note ID of the transaction note
//  2  -  NOTEHANDLE			-  Handle of the open transaction note
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::MarkTransactionRetried(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, int iThreadID)
{
	STATUS				stAPIRC;								//  Return code from Notes API calls
	char				szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Mark the transaction as Retried
	stAPIRC = NSFItemSetText(hnTX, DEFAULT_TRANSDOC_STATUS, DEFAULT_TRANSDOC_STATUS_RETRIED, MAXWORD);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2814E, nidTX, "Retried", tqCurrent->szQName);
		xeLocal->LogMessage(szMsg, iThreadID);		
	}
	return;
}

//  SerializeTransaction
//
//  There is no default behaviour for this part of the interface. The implementation here will
//  issue an error message and revoke the execution permit for this queue.
//
//  Parameters
//
//  1  -  void *				-  Pointer to the transaction object
//  2  -  DBHANDLE				-  Handle of the transaction queue database
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//  NOTE:  ******  This function MUST be overridden in the inheriting class
//
//  At a minimum the implementation for this function must do the following :-
//
//  1)  Mark the transaction document that this object refers to with a completion status
//  2)  Delete the transaction object
//
void TransactionHandler::SerializeTransaction(void *txObject, DBHANDLE hdbQueue, TransactionQueue *tqCurrent, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];								//  Generic Message Buffer

	tqCurrent->LocalPermit = FALSE;									//  Force the queue to shutdown
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2815E, tqCurrent->szQName);
	xeLocal->LogMessage(szMsg, iThreadID);
	return;
}

//  MarshallTransaction
//
//  There is no default behaviour for this part of the interface. The implementation here will
//  issue an error message and revoke the execution permit for this queue.
//
//  Parameters
//
//  1  -  NOTEID				-  Note ID of the transaction note
//  2  -  NOTEHANDLE			-  Handle of the open transaction note
//  3  -  TransactionQueue *	-  Pointer to the Transaction Queue object for this transaction
//  4  -  void **				-  Pointer for the returmed address to the transaction object
//  5  -  void **				-  Pointer for the returned address for the execution object
//  6  -  UINT *				-  Pointer for the returned Attributes for the transaction
//  7  -  int *					-  Pointer for the returned Priority for the transaction
//  8  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  int			-   The transaction disposition
//
//	TX_DISPOSITION_NORMAL - Dispatch the transaction object and mark the transaction In Progress
//	TX_DISPOSITION_ERROR  - Mark the transaction Error
//  TX_DISPOSITION_DELAY  - Mark the transaction delayed
//  TX_DISPOSITION_PANIC  - Critical Error, stop the queue
//
//  API Use Notes:
//
//  NOTE:  ******  This function MUST be overridden in the inheriting class
//
//
int TransactionHandler::MarshallTransaction(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, void **txObject, void **xxObject, UINT *Attrs, int *Priority, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];								//  Generic Message Buffer

	tqCurrent->LocalPermit = FALSE;									//  Force the queue to shutdown
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2817E, tqCurrent->szQName);
	xeLocal->LogMessage(szMsg, iThreadID);
	return TX_DISPOSITION_PANIC;
}

//  CycleTime
//
//  This function will cycle a timedate by the quantity specified. The time/date is increased continuously until it is in the future and then
//  reduced by one increment.
//
//  Parameters
//
//  1  -  TIMEDATE *			-  The TIMEDATE to be cycled
//  2  -  int 					-  Pointer for the returned Priority for the transaction
//  3  -  int					-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL		-   TRUE if the operation succeeded and FALSE if it failed
//
//  API Use Notes:
//
//
BOOL TransactionHandler::CycleTime(TIMEDATE *tdCycle, int Days, int Months)
{
	TIMEDATE			tdCurrent;									//  Current time/date
	TIME				t_Adjust;									//  TIME structure for making fine adjustments

	//  Get the current time
	OSCurrentTIMEDATE(&tdCurrent);
	
	//  Cycle the time until it is in the future
	while (TimeDateCompare(&tdCurrent, tdCycle) == 1)
	{
		if (TimeDateAdjust(tdCycle, 0, 0, 0, Days, Months, 0)) return FALSE;

	}

	//  Reduce the cycle time by one interval
	if (TimeDateAdjust(tdCycle, 0, 0, 0, Days * -1, Months * -1, 0)) return FALSE;

	//  Final fine tuning adjust the seconds and hundredths to zero to prevent time creep
	memcpy(&t_Adjust.GM, tdCycle, sizeof(TIMEDATE));
	if (TimeGMToLocal(&t_Adjust)) return FALSE;
	t_Adjust.second = 0;
	t_Adjust.hundredth = 0;
	if (TimeLocalToGM(&t_Adjust)) return FALSE;
	memcpy(tdCycle, &t_Adjust.GM, sizeof(TIMEDATE));
	return TRUE;
}

//  ExecuteThisRequest
//
//  This is the entry point used by the multi-threading kernel to trigger processing in different threadpool threads.
//
//  Parameters
//
//  1  -  void *			-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void TransactionHandler::ExecuteThisRequest(void *pReqObject, int iThreadID)
{
	ProcessQueue((TransactionQueue *) pReqObject, iThreadID);
	return;
}

