#pragma once
//  TransactionHandler
//
//  This class provides the standard handling of the transaction queues.
//  The class must be extended to provide the necessary interfaces for handling the serialisation and
//  marshalling of transaction documents to/from Request Objects.
//
//  Message Identifier:      DXR
//  Message Range:           2810 - 2899
//  Trace area assignment:   TRACE_TXH (07)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  3.10.1		13/10/2010		Initial Version
//  3.10.3		13/12/2010		Debug Helper added
//  3.10.8		30/05/2011		Transaction scheduling added to transaction handler
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.5		30/11/2011		Calculation of last run time changed
//  3.14.1		20/01/2015		x64 compatability
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2010 - 2015
//

//  Application Includes
#include "MTExecutive.h"
#include "TransactionQueue.h"
#if (defined(_DEBUG) && defined(WIN32))
#include "../Debug/Helper.h"
#endif

//  Local Definitions
//  Default Transaction Item Names
#define 		DEFAULT_TRANSDOC_STATUS				"Status"
#define			DEFAULT_TRANSDOC_SCHEDFIRST			"FirstRunTime"
#define			DEFAULT_TRANSDOC_SCHEDLAST			"LastRunTime"
#define			DEFAULT_TRANSDOC_SCHEDINTERVAL		"RepeatInterval"

//  Default Transaction Status field values		
#define			DEFAULT_TRANSDOC_STATUS_NEW			"NEW"
#define			DEFAULT_TRANSDOC_STATUS_INPROGRESS	"INPROGRESS"
#define			DEFAULT_TRANSDOC_STATUS_COMPLETED	"COMPLETED"
#define			DEFAULT_TRANSDOC_STATUS_ERROR		"ERROR"
#define			DEFAULT_TRANSDOC_STATUS_DELAYED		"DELAYED"
#define			DEFAULT_TRANSDOC_STATUS_RETRIED		"RETRIED"

//  Default Transaction ReScheduling Interval field values
#define			DEFAULT_TRANSDOC_INTERVAL_NEVER		"Never"
#define			DEFAULT_TRANSDOC_INTERVAL_DAILY		"Daily"
#define			DEFAULT_TRANSDOC_INTERVAL_WEEKLY	"Weekly"
#define			DEFAULT_TRANSDOC_INTERVAL_MONTHLY	"Monthly"

//  Transaction Marshalling Interface Dispositions
#define			TX_DISPOSITION_NORMAL				1
#define			TX_DISPOSITION_ERROR				2
#define			TX_DISPOSITION_DELAY				4
#define			TX_DISPOSITION_PANIC				8

//  Miscellaneous Definition
#define			DEFAULT_MAX_TRANS					25			//  max of 25 transactions per dispatch
#define			IDLE_DELAY_TRIGGER					10			//  10 continuous Idle cycles to process delayed
#define			DEFAULT_DELAY_QTIME					10*60		//  10 minutes
#define			DEFAULT_DELAY_QCYCLE				200			//  200 transactions
#define			DEFAULT_SLEEP_CYCLE					30			//  30 seconds

//  Trace Messages
#define			MSG_DXR2801T		"DXR2801T: A new TransactionHandler object has been created @%p."
#define			MSG_DXR2802T		"DXR2802T: A TransactionHandler object @%p has been destroyed."
#define			MSG_DXR2803T		"DXR2803T: Initialisation is in progress for Transaction Queue: '%s'."
#define			MSG_DXR2808T		"DXR2808T: Requeue started for in progress transactions on Queue: '%s'."
#define			MSG_DXR2811T		"DXR2811T: Processing %i In Progress transactions on Queue '%s'."
#define			MSG_DXR2813T		"DXR2813T: Finished processing %i In Progress transactions on Queue '%s'."
#define			MSG_DXR2816T		"DXR2816T: Attempting to dispatch %i Ready transactions on Queue '%s'."
#define			MSG_DXR2819T		"DXR2819T: Finished dispatching %i Ready transactions on Queue '%s'."
#define			MSG_DXR2820T		"DXR2820T: Requeue started for Delayed transactions on Queue: '%s'."
#define			MSG_DXR2822T		"DXR2822T: Cycle triggered requeue for Started: %i Last Cycle: %i Limit: %i."
#define			MSG_DXR2823T		"DXR2823T: Dispatch cycle has resumed after resting for %i ticks."
#define			MSG_DXR2825T		"DXR2825T: Checking for any scheduled transactions that are due to execute on Queue: '%s'."
#define			MSG_DXR2826T		"DXR2826T: Checking %i scheduled transactions on Queue '%s'."
#define			MSG_DXR2827T		"DXR2827T: Finished checking %i scheduled transactions on Queue '%s'."

//  Information Messages
#define			MSG_DXR2821I		"DXR2821I: Finished returning %i transactions to the Ready Queue on Queue '%s'."
#define			MSG_DXR2824I		"DXR2824I: Queue '%s' is draining due the the run limit of %i transactions having been reached."
#define			MSG_DXR2828I		"DXR2828I: Added %i scheduled transactions to the Ready Queue on Queue '%s'."

//  Warning Messages

//  Error Messages
#define			MSG_DXR2804E		"DXR2804E: Unable to initialise Trasaction Queue '%s' no Handle or Database Name was specified."
#define			MSG_DXR2805E		"DXR2805E: Unable to open Queue '%s' - Database '%s' on '%s'."
#define			MSG_DXR2806E		"DXR2806E: Unable to get a local DBHandle for Transaction Queue '%s'."
#define			MSG_DXR2807E		"DXR2807E: Unable to locate the %s ('%s') view for Transaction Queue '%s'."
#define			MSG_DXR2809E		"DXR2809E: Unable to open the %s ('%s') view for Transaction Queue '%s'."
#define			MSG_DXR2810E		"DXR2810E: Error reading from the %s ('%s') view for Transaction Queue '%s'."
#define			MSG_DXR2812E		"DXR2812E: Unable to open transaction (Note ID: 0x%.4X) from the %s view in Queue '%s', skipping transaction."
#define			MSG_DXR2814E		"DXR2814E: Unable to update transaction (Note ID: 0x%.4X) to Status: %s in Queue '%s'."
#define			MSG_DXR2815E		"DXR2815E: The Queue handler for '%s' has not implemented the SerializeTransaction interface, the Queue will be closed."
#define			MSG_DXR2817E		"DXR2817E: The Queue handler for '%s' has not implemented the MarshallTransaction interface, the Queue will be closed."
#define			MSG_DXR2818E		"DXR2818E: Unable to dispatch a request %p to the Pending Pool, closing Queue '%s'."
#define			MSG_DXR2829E		"DXR2829E: Unable to set a default value on a transaction, it will not be scheduled."
#define			MSG_DXR2830E		"DXR2830E: Unable to read the '%s' field on a transaction, it will not be scheduled."
#define			MSG_DXR2831E		"DXR2831E: Unable to calculate the next run time for a scheduled transaction, it will not be scheduled."
#define			MSG_DXR2832E		"DXR2832E: A reschedule interval of '%s' did not yield a next run time, transaction will not be scheduled."
#define			MSG_DXR2833E		"DXR2833E: Unable to set a current value on a transaction, it has not be scheduled."
#define			MSG_DXR2834E		"DXR2834E: Unable to set create a copy of a transaction, it has not be scheduled."

class TransactionHandler :
	public Runnable
{
public:
	//  Default Constructor
	TransactionHandler(void);

	//  Constructor
	TransactionHandler(MTExecutive *xeParent, int iThreadID);

	//  Destructor
	~TransactionHandler(void);

	//  Public Members
#if (defined(_DEBUG) && defined(WIN32))
	Helper					*DebugHelper;									//  Debuging helper object
#endif

	//  Public Functions
	void virtual ExecuteThisRequest(void *pReqObject, int iThreadID);	//  Multi thread request entry point
	void ProcessQueue(TransactionQueue *tqNew, int iThreadID);			//  Process a transaction queue

private:
	//  Private Members

	//  Private Functions
	void RequeueInProgress(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, NOTEID nidView, int iThreadID);  //  Requeue In Progress Transactions
	int PollCompleted(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, int iThreadID);						//  Poll for completed transactions
	BOOL DispatchReady(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, NOTEID nidView, int iThreadID);		//  Dispatch any ready transactions
	void RequeueDelayed(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, NOTEID nidView, int iThreadID);		//  Requeue Delayed Transactions
	void ScheduleTransactions(TransactionQueue *tqCurrent, DBHANDLE hdbQueue, NOTEID nidView, int iThreadID);  //  Schedule any due transactions
	int CheckSchedTransaction(TransactionQueue *tqCurrent, NOTEHANDLE hnTransaction, int iThreadID);			//  Check if transaction should be readied
	BOOL CycleTime(TIMEDATE *tdCycle, int Days, int Months);												//  Cycle a Timestamp

protected:
	//  Protected Members
	MTExecutive		*xeLocal;											//  Runtime
	int				ParentThread;										//  Owning Thread

	//  Protected Functions - these define the interface needed to manage transactions
	//  only the needed entry points have to be defined in the inheriting class.
	//  Transaction marking functions
	void virtual MarkTransactionDelayed(NOTEID nidTX, NOTEHANDLE hnTX, BOOL bInProgress, TransactionQueue *tqCurrent, int iThreadID);
	void virtual MarkTransactionReady(NOTEID nidTX, NOTEHANDLE hnTX, BOOL bInProgress, TransactionQueue *tqCurrent, int iThreadID);
	void virtual MarkTransactionError(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, int iThreadID);
	void virtual MarkTransactionInProgress(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, int iThreadID);
	void virtual MarkTransactionCompleted(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, int iThreadID);
	void virtual MarkTransactionRetried(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, int iThreadID);
	//  Serialize a Transaction object and dispose of it
	void virtual SerializeTransaction(void *txObject, DBHANDLE hdbQueue, TransactionQueue *tqCurrent, int iThreadID);
	//  Marshall a Transaction object and determine the disposition of it and the transaction document
	int virtual MarshallTransaction(NOTEID nidTX, NOTEHANDLE hnTX, TransactionQueue *tqCurrent, void **txObject, void **xxObject, UINT *Attrs, int *Priority, int iThreadID);

};
