//  AutoScaleFrequencyAnalyser Class
//
//  This class provides a frequency analyser that automatically scales to maintain the maximum granularity of the data
//
//  Message Identifier:      N/A
//  Message Range:           N/A
//  Trace area assignment:   N/A
//  Version:                 3.11.5  Build: 103
//
//  Version History:
//
//  3.11.1		22/10/2011		Incorporated into kernel
//  3.11.5		30/11/2011		Cache coherence improvements
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2011
//

//  Application Includes
#include	"AutoScaleFrequencyAnalyser.h"

//  Default Constructor
AutoScaleFrequencyAnalyser::AutoScaleFrequencyAnalyser()
{
	pPool = NULL;												//  Initialise the storage pool pointer
	ASFANumBuckets = ASFA_NUM_BUCKETS;							//  Set the default number of buckets
	ASFANumOutliers = ASFA_NUM_OUTLIERS;						//  Set the default number of outliers
	ASFARenormBuckets = ASFA_RENORM_BUCKETS;					//  Set the bucket limit for re-normalisation
	InitPool();													//  Initialise the storage pool
}

//  Normal Constructor
//
//  Parameters
//
//  1		-	int				-	The count of buckets to allocate in the object
//  2		-	int				-	The count of outliers for this object
//  3		-	int				-	The count of buckets that limit re-normalisation
//
//  Returns
//
//  API Use Notes:
//
//
AutoScaleFrequencyAnalyser::AutoScaleFrequencyAnalyser(int Bkts, int Outls, int RenormBkts)
{
	pPool = NULL;												//  Initialise the storage pool pointer
	ASFANumBuckets = Bkts;										//  Set the number of buckets
	ASFANumOutliers = Outls;									//  Set the number of outliers
	ASFARenormBuckets = RenormBkts;								//  Set the bucket limit for re-normalisation
	InitPool();													//  Initialise the storage pool
}

//  Destructor
AutoScaleFrequencyAnalyser::~AutoScaleFrequencyAnalyser()
{
	if (pPool != NULL) delete pPool;							//  Delete the storage pool if allocated
}

//  DisallowOutliers
//
//  This function will turn off the possibility of recording outliers
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void AutoScaleFrequencyAnalyser::DisallowOutliers()
{
	OutliersPermitted = FALSE;
	return;
}

//  RecordDataPoint
//
//  This function will record a new data point
//
//  Parameters
//
//  1		-	DWORD				-	New data value to be added to the data set
//
//  Returns
//
//  API Use Notes:
//
//
void AutoScaleFrequencyAnalyser::RecordDataPoint(DWORD dwDP)
{
#ifdef _DEBUG
	printf("+DEBUG: Recording: %i in L: %i, H: %i, S: %i, DIB=%i.\r\n", dwDP, B0LowValue, BnHighValue, BucketSize, DataInBuckets); 
#endif
	//  If the data point is an outlier (currntly) then it was filtered
	if (FilterOutliers(dwDP)) return;
	Increment(dwDP);											//  Increment the bucket counter
}

//  FilterOutliers
//
//  This function will filter out any outlier data points
//
//  Parameters
//
//  1		-	DWORD				-	New data value to be added to the data set
//
//  Returns
//
//  BOOL		-  TRUE if the value was filtered as an outlier, FALSE if not
//
//  API Use Notes:
//
//
BOOL AutoScaleFrequencyAnalyser::FilterOutliers(DWORD dwDP)
{
	DWORD			dwNewDP;							//  Replacement Data Point

	//  If outliers are not permitted then check the bounds of the buckets and adjust if necessary
	if (!OutliersPermitted)
	{
		if ((dwDP < B0LowValue) || (dwDP >= BnHighValue)) AdjustBounds(dwDP);
		return FALSE;
	}

	//  Detect low end outliers
	if ((dwDP < B0LowValue) || (DataInBuckets == 0))
	{
#ifdef _DEBUG
		printf("+DEBUG: %i is low outlier, %i low outliers already recorded.\r\n", dwDP, LXCount);
#endif
		//  See if the low end outlier array is full 
		if (LXCount == ASFANumOutliers)
		{
			dwNewDP = SwapForHighestLowOutlier(dwDP);
			AdjustBounds(dwNewDP);
			Increment(dwNewDP);
		}
		else
		{
			//  Record the data point as a low end outlier
			LowOutliers[LXCount] = dwDP;
			LXCount++;
		}
		return TRUE;
	}

	//  Detect high end outliers
	if (dwDP >= BnHighValue)
	{
#ifdef _DEBUG
		printf("+DEBUG: %i is high outlier, %i high outliers already recorded.\r\n", dwDP, HXCount);
#endif
		//  See if the high end outliers array is full
		if (HXCount == ASFANumOutliers)
		{
			dwNewDP = SwapForLowestHighOutlier(dwDP);
			AdjustBounds(dwNewDP);
			Increment(dwNewDP);
		}
		else
		{
			//  Record the data point as a high end outlier
			HighOutliers[HXCount] = dwDP;
			HXCount++;
		}
		return TRUE;
	}

	return FALSE;
}

//  AdjustBounds
//
//  This function will adjust the bounds of the recording array so that the passed data point will now fit the data space
//
//  Parameters
//
//  1		-	DWORD				-	New data value to be added to the data set
//
//  Returns
//
//  API Use Notes:
//
//
void AutoScaleFrequencyAnalyser::AdjustBounds(DWORD dwDP)
{
	DWORD				dwNewB0Low;									//  New lower bound for bucket zero
	DWORD				dwNewBnHigh;								//  New higher bound for bucket n
	DWORD				dwNewSize;									//  New Bucket size
	DWORD				dwLowSample;								//  Low Sample value
	DWORD				dwHighSample;								//  High Sample
	int					iIndex;										//  Generic Index
	int					jIndex;										//  Generic Index

#ifdef _DEBUG
	printf("+DEBUG: Adjusting bounds to accomodate %i.\r\n", dwDP); 
#endif
	//  Check for the boundary condition where there are no data points in the bucket yet
	if (DataInBuckets == 0)
	{
		GetSuggestedBounds(dwDP, dwDP, &dwNewB0Low, &dwNewBnHigh, &dwNewSize);
		//  Set the new bounds and return
		B0LowValue = dwNewB0Low;
		BnHighValue = dwNewBnHigh;
		BucketSize = dwNewSize;
#ifdef _DEBUG
		printf("+DEBUG: New bounds to accomodate %i are L: %i, H: %i, S: %i.\r\n", dwDP, B0LowValue, BnHighValue, BucketSize); 
#endif
	}
	else
	{
		//  Get new suggested bounds for the bucket array
		//  Find the lowest and highest current values in the data space
		GetSampleBounds(&dwLowSample, &dwHighSample);

#ifdef _DEBUG
		printf("+DEBUG: Bounds initially determined as L: %i - H: %i.\r\n", dwLowSample, dwHighSample);
#endif
		if (dwDP < dwLowSample) dwLowSample = dwDP;
		if (dwDP > dwHighSample) dwHighSample = dwDP;
#ifdef _DEBUG
		printf("+DEBUG: Bounds adjusted to L: %i - H: %i.\r\n", dwLowSample, dwHighSample);
#endif
		GetSuggestedBounds(dwLowSample, dwHighSample, &dwNewB0Low, &dwNewBnHigh, &dwNewSize);

		//  Remap the data from the old data space into the new data space
		dwLowSample = dwNewB0Low;
		for (iIndex = 0; iIndex < ASFANumBuckets; iIndex++)
		{
			TempBucket[iIndex] = GetBucketValue(dwLowSample, dwNewSize);
			dwLowSample += dwNewSize;
		}
		memcpy(&Bucket[0], &TempBucket[0], sizeof(DWORD) * ASFANumBuckets);

		//  Set the new bounds and return
		B0LowValue = dwNewB0Low;
		BnHighValue = dwNewBnHigh;
		BucketSize = dwNewSize;

#ifdef _DEBUG
		printf("+DEBUG: New bounds to accomodate %i are L: %i, H: %i, S: %i.\r\n", dwDP, B0LowValue, BnHighValue, BucketSize);
#endif
	}
#ifdef _DEBUG
		printf("+DEBUG: Eliminating %i low outliers.\r\n", LXCount);
#endif

	//  Eliminate any outliers that now fit into the data space
	for (iIndex = 0; iIndex < LXCount; iIndex++)
	{
		if (LowOutliers[iIndex] >= B0LowValue) 
		{
			Increment(LowOutliers[iIndex]);
			//  Shuffle up the remaining entries
			for (jIndex = iIndex; jIndex < (LXCount - 1); jIndex++)
			{
				LowOutliers[jIndex] = LowOutliers[jIndex + 1];
			}
			LowOutliers[LXCount - 1] = 0;
			LXCount--;
			iIndex--;
		}
	}

#ifdef _DEBUG
		printf("+DEBUG: After eliminating %i low outliers remain.\r\n", LXCount);
		if (LXCount > 0)
		{
			printf("+DEBUG: Low Outliers: %i", HighOutliers[0]);
			for (iIndex = 1; iIndex < LXCount; iIndex++)
			{
				printf(", %i", LowOutliers[iIndex]);
			}
			printf(".\r\n");
		}
#endif

#ifdef _DEBUG
		printf("+DEBUG: Eliminating %i high outliers.\r\n", HXCount);
#endif

		for (iIndex = 0; iIndex < HXCount; iIndex++)
	{
		if (HighOutliers[iIndex] < BnHighValue) 
		{
			Increment(HighOutliers[iIndex]);
			//  Shuffle up the remaining entries
			for (jIndex = iIndex; jIndex < (HXCount - 1); jIndex++)
			{
				HighOutliers[jIndex] = HighOutliers[jIndex + 1];
			}
			HighOutliers[HXCount - 1] = 0;
			HXCount--;
			iIndex--;
		}
	}

#ifdef _DEBUG
		printf("+DEBUG: After eliminating %i high outliers remain.\r\n", HXCount);
		if (HXCount > 0)
		{
			printf("+DEBUG: High Outliers: %i", HighOutliers[0]);
			for (iIndex = 1; iIndex < HXCount; iIndex++)
			{
				printf(", %i", HighOutliers[iIndex]);
			}
			printf(".\r\n");
		}
#endif
	
	return;
}

//  SwapForHighestLowOutlier
//
//  This function will swap the supplied value for the highest value in the low outliers array
//
//  Parameters
//
//  1		-	DWORD				-	New candidate low outlier
//
//  Returns
//
//  API Use Notes:
//
//
DWORD AutoScaleFrequencyAnalyser::SwapForHighestLowOutlier(DWORD dwDP)
{
	DWORD				dwCandidate;							//  Candidate highest value
	DWORD				dwSwap;									//  Value to be swapped
	int					iIndex;									//  Generic index

	dwCandidate = dwDP;

	for (iIndex = 0; iIndex < LXCount; iIndex++)
	{
		if (LowOutliers[iIndex] > dwCandidate)
		{
			dwSwap = dwCandidate;
			dwCandidate = LowOutliers[iIndex];
			LowOutliers[iIndex] = dwSwap;
		}
	}

#ifdef _DEBUG
	printf("+DEBUG: Exchanged %i for %i in low outliers.\r\n", dwCandidate, dwDP);
#endif
	return dwCandidate;
}

//  SwapForLowestHighOutlier
//
//  This function will swap the supplied value for the lowest value in the high outliers array
//
//  Parameters
//
//  1		-	DWORD				-	New candidate high outlier
//
//  Returns
//
//  API Use Notes:
//
//
DWORD AutoScaleFrequencyAnalyser::SwapForLowestHighOutlier(DWORD dwDP)
{
	DWORD				dwCandidate;							//  Candidate highest value
	DWORD				dwSwap;									//  Value to be swapped
	int					iIndex;									//  Generic index

	dwCandidate = dwDP;

	for (iIndex = 0; iIndex < HXCount; iIndex++)
	{
		if (HighOutliers[iIndex] < dwCandidate)
		{
			dwSwap = dwCandidate;
			dwCandidate = HighOutliers[iIndex];
			HighOutliers[iIndex] = dwSwap;
		}
	}

#ifdef _DEBUG
	printf("+DEBUG: Exchanged %i for %i in high outliers.\r\n", dwCandidate, dwDP);
#endif

	return dwCandidate;
}

//  Increment
//
//  This function will increment the bucket value in which the specified data point lies
//
//  Parameters
//
//  1		-	DWORD				-	Data Point
//
//  Returns
//
//  API Use Notes:
//
//
void AutoScaleFrequencyAnalyser::Increment(DWORD dwDP)
{
	DWORD			dwBucketLimit;									//  Lower limit of the next bucket
	int				iIndex;											//  Generic Index

	dwBucketLimit = B0LowValue;

	for (iIndex = 0; iIndex < ASFANumBuckets; iIndex++)
	{
		dwBucketLimit += BucketSize;
		if (dwDP < dwBucketLimit)
		{
			Bucket[iIndex]++;
			DataInBuckets++;
#ifdef _DEBUG
			printf("+DEBUG: Setting Bucket: %i to %i for value: %i.\r\n", iIndex, Bucket[iIndex], dwDP);
#endif
			return;
		}
	}
	return;
}

//  GetSuggestedBounds
//
//  This function will suggest new bounds for the bucket array given a low and high sample
//
//  Parameters
//
//  1		-	DWORD				-	Low sample value to be accomodated
//  2		-	DWORD				-   High sample value to be accomodated
//  3		-   DWORD *				-   Returned low boundary value for bucket 0 (lowest bucket)
//  4		-	DWORD *				-	Returned high boundary value for bucket n (highest bucket)
//  5		-	DWORD *				-   Returned bucket size
//
//  Returns
//
//  API Use Notes:
//
//
void AutoScaleFrequencyAnalyser::GetSuggestedBounds(DWORD lo, DWORD hi, DWORD *B0, DWORD *Bn, DWORD *BSize)
{
	int					iIndex;							//  Generic Index
	DWORD				dwNewSize;						//  New bucket size
	BOOL				bSatisfactory;					//  Solution is satisfactory
	DWORD				dwNewB0Low;						//  New Bucket 0 low value
	int					iLowIndex;						//  Low bucket index
	int					iHighIndex;						//  High bucket index
	
#ifdef _DEBUG
	printf("+DEBUG: Suggesting bounds for L: %i, H: %i.\r\n", lo, hi);
#endif
	//  Check for the boundary condition where the bucket array is unbounded
	if (DataInBuckets == 0)
	{
		if (lo < (DWORD) ASFANumBuckets)
		{
			*B0 = 0;
			*Bn = ASFANumBuckets;
			*BSize = 1;
		}
		else
		{
			iIndex = ASFANumBuckets / 2;								//  Midpoint
			*B0 = (lo - iIndex) - 1;
			*Bn = (lo + iIndex) + 1;
			*BSize = 1;
		}
#ifdef _DEBUG
		printf("+DEBUG: Suggested bounds for L: %i, H: %i are L: %i, H: %i, S: %i.\r\n", lo, hi, *B0, *Bn, *BSize);
#endif
		return;
	}

	//  Calculate the next prospective bucket size
	dwNewSize = BucketSize;
	bSatisfactory = FALSE;

	//  Loop until we have found a satisfactory solution
	while (!bSatisfactory)
	{
		//  Compute the new bucket 0 low value
		if (lo < B0LowValue) dwNewB0Low = lo / dwNewSize;
		else dwNewB0Low = B0LowValue / dwNewSize;
		dwNewB0Low = dwNewB0Low * dwNewSize;

		//  Compute the bucket indexes in the new data space
		iLowIndex = (lo - dwNewB0Low) / dwNewSize;
		iHighIndex = (hi - dwNewB0Low) / dwNewSize;

#ifdef _DEBUG
		printf("+DEBUG: Trying solution L: %i, S: %i, LX=%i, HX=%i.\r\n", dwNewB0Low, dwNewSize, iLowIndex, iHighIndex);
#endif
		//  Check that the interval will fit the new data space
		if ((iHighIndex - iLowIndex) < ASFANumBuckets)
		{
			//  Shift the origin so that the data points are spaces around the center of the array
			iIndex = (ASFANumBuckets - (iHighIndex - iLowIndex)) / 2;
			iIndex = iLowIndex - iIndex;
			dwNewB0Low = dwNewB0Low + (iIndex * dwNewSize);
			while (iIndex < 0)
			{
				iIndex++;
				dwNewB0Low = dwNewB0Low + dwNewSize;
			}
			bSatisfactory = TRUE;
		}
		else
		{
			//  Try the next prospective bucket size
			dwNewSize = GetNextBucketSize(dwNewSize);
		}
	}

	// Return the solution
	*BSize = dwNewSize;
	*B0 = dwNewB0Low;
	*Bn = dwNewB0Low + (ASFANumBuckets * dwNewSize);
#ifdef _DEBUG
		printf("+DEBUG: Suggested bounds for L: %i, H: %i are L: %i, H: %i, S: %i.\r\n", lo, hi, *B0, *Bn, *BSize);
#endif
	return;
}

//  GetBucketValue
//
//  This function will compute the value of a notional bucket from one or more buckets in the current array
//
//  Parameters
//
//  1		-	DWORD				-	Low value of the notional bucket
//  2		-	DWORD				-   Size of the notional bucket
//
//  Returns
//
//  DWORD			-		The value of the notion bucket
//
//  API Use Notes:
//
//
DWORD AutoScaleFrequencyAnalyser::GetBucketValue(DWORD lo, DWORD bSize)
{
	DWORD			dwAccumulator = 0;								//  Accumulator for the bucket value
	DWORD			dwBucketLimit;									//  Lower limit of the next bucket
	int				iIndex;											//  Generic Index

	dwBucketLimit = B0LowValue;

	//  If the notional bucket is outside the current data space return 0
	if (lo < dwBucketLimit) return 0;

	//  Find the first bucket that fits the notional data value
	iIndex = 0;
	while (dwBucketLimit < lo)
	{
		iIndex++;
		dwBucketLimit += BucketSize;
	}

	//  If the start point is outside the existing data space then return 0
	if (iIndex >= ASFANumBuckets) return 0;

	//  Loop through the buckets in the data space accumulating the new value
	while (dwBucketLimit < (lo + bSize))
	{
		if (iIndex >= ASFANumBuckets) break;
		dwAccumulator += Bucket[iIndex];
		iIndex++;
		dwBucketLimit += BucketSize;
	}

	return dwAccumulator;
}

//  GetNextBucketSize
//
//  This function will return the next increment of bucket size
//
//  Parameters
//
//  1		-		DWORD		-	The previous bucket size
//
//  Returns
//
//  DWORD			-		The next bucket size to use
//
//  API Use Notes:
//
//
DWORD AutoScaleFrequencyAnalyser::GetNextBucketSize(DWORD dwOldSize)
{
	if (dwOldSize == 1) return 5;
	if (dwOldSize == 5) return 10;
	if (dwOldSize == 10) return 50;
	if (dwOldSize == 50) return 100;
	if (dwOldSize == 100) return 500;
	if (dwOldSize == 500) return 1000;
	if (dwOldSize == 1000) return 5000;
	if (dwOldSize == 5000) return 10000;
	if (dwOldSize == 10000) return 50000;
	if (dwOldSize == 50000) return 100000;
	if (dwOldSize == 100000) return 500000;
	if (dwOldSize == 500000) return 1000000;
	if (dwOldSize == 1000000) return 5000000;
	if (dwOldSize == 5000000) return 10000000;
	return 100000000;
}

//  GetFrequency
//
//  This function will return the frequency of a particular bucket also the low and high bounds of the bucket
//
//  Parameters
//
//  1		-	int					-   The index of the requested bucket
//  2		-	DWORD *				-	Returned Low value of the bucket
//  3		-	DWORD *				-   Returned High value of the  bucket
//
//  Returns
//
//  DWORD			-		The frequency recorded for the bucket
// 
//  API Use Notes:
//
//
DWORD AutoScaleFrequencyAnalyser::GetFrequency(int Ordinate, DWORD *lo, DWORD *hi)
{
	DWORD				dwLowLimit;											//  Low outlier limit
	DWORD				dwHighLimit;										//  High outlier limit
	DWORD				dwNewDataPoint;										//  New data point 
	int					iIndex;												//  Generic index
	int					jIndex;												//  Generic index

	//  Sanity check
	if ((Ordinate < 0) || (Ordinate >= ASFANumBuckets)) return -1;
	
	//  If the array has not be Re-Normalised then do so now
	if (!ReNormalised)
	{
#ifdef _DEBUG
		printf("+DEBUG: Starting normalisation.\r\n");
#endif
		OutliersPermitted = FALSE;
		//  Compute the new high and low limits
		dwLowLimit = B0LowValue - (ASFARenormBuckets * BucketSize);
		dwHighLimit = BnHighValue + (ASFARenormBuckets * BucketSize);

		//  Incorporate and low outliers that are within the limit
		for (iIndex = 0; iIndex < LXCount; iIndex++)
		{
			if (LowOutliers[iIndex] >= dwLowLimit)
			{
#ifdef _DEBUG
				printf("+DEBUG: Adopting low outlier %i value=%i against limit=%i.\r\n", iIndex, LowOutliers[iIndex], dwLowLimit);
#endif
				dwNewDataPoint = LowOutliers[iIndex];
				//  Shuffle the array to lose the outlier
				for (jIndex = iIndex; jIndex < (LXCount - 1); jIndex++)
				{
					LowOutliers[jIndex] = LowOutliers[jIndex + 1];
				}
				LowOutliers[LXCount - 1] = 0;
				LXCount--;
				iIndex--;
				//  Adopt the outlier
				RecordDataPoint(dwNewDataPoint);
			}
		}

		//  Incorporate and high outliers that are within the limit
		for (iIndex = 0; iIndex < HXCount; iIndex++)
		{
			if (HighOutliers[iIndex] < dwHighLimit)
			{
#ifdef _DEBUG
				printf("+DEBUG: Adopting high outlier %i value=%i against limit=%i.\r\n", iIndex, HighOutliers[iIndex], dwHighLimit);
#endif
				dwNewDataPoint = HighOutliers[iIndex];
				//  Shuffle the array to lose the outlier
				for (jIndex = iIndex; jIndex < (HXCount - 1); jIndex++)
				{
					HighOutliers[jIndex] = HighOutliers[jIndex + 1];
				}
				HighOutliers[LXCount - 1] = 0;
				HXCount--;
				iIndex--;
				//  Adopt the outlier
				RecordDataPoint(dwNewDataPoint);
			}
		}

		ReNormalised = TRUE;
	}

	//  Compute the high and low bounds of the bucket
	*lo = (Ordinate * BucketSize) + B0LowValue;
	*hi = *lo + BucketSize - 1;
	return Bucket[Ordinate];
}

//  GetLowOutlierCount
//
//  This function will return the count of low outliers
//
//  Parameters
//
//  Returns
//
//  int			-		The count of low outliers
// 
//  API Use Notes:
//
//
int AutoScaleFrequencyAnalyser::GetLowOutlierCount()
{
	return LXCount;
}

//  GetHighOutlierCount
//
//  This function will return the count of high outliers
//
//  Parameters
//
//  Returns
//
//  int			-		The count of high outliers
// 
//  API Use Notes:
//
//
int AutoScaleFrequencyAnalyser::GetHighOutlierCount()
{
	return HXCount;
}

//  GetLowOutlier
//
//  This function will return the value of a particular low outlier
//
//  Parameters
//
//  1		-	int					-   The index of the requested outlier
//
//  Returns
//
//  DWORD			-		The value of the outlier
// 
//  API Use Notes:
//
//
DWORD AutoScaleFrequencyAnalyser::GetLowOutlier(int Ordinate)
{
	return LowOutliers[Ordinate];
}

//  GetHighOutlier
//
//  This function will return the value of a particular high outlier
//
//  Parameters
//
//  1		-	int					-   The index of the requested outlier
//
//  Returns
//
//  DWORD			-		The value of the outlier
// 
//  API Use Notes:
//
//
DWORD AutoScaleFrequencyAnalyser::GetHighOutlier(int Ordinate)
{
	return HighOutliers[Ordinate];
}

//  GetSampleBounds
//
//  This function will return the bounds of the data sample
//
//  Parameters
//
//  1		-	DWORD *					-   The returned low sample boundary
//  2		-	DWORD *					-	The returned high sample value
//
//  Returns
//
//  API Use Notes:
//
//
void AutoScaleFrequencyAnalyser::GetSampleBounds(DWORD *lo, DWORD *hi)
{
	int					iIndex;												//  Generic index

	*lo = 0;
	*hi = 0;
	iIndex = 0;
	while (iIndex < ASFANumBuckets)
	{
		if (Bucket[iIndex] != 0)
		{
			if (*lo == 0) 
			{
				*lo = (iIndex * BucketSize) + B0LowValue;
			}
			*hi = ((iIndex + 1) * BucketSize) + B0LowValue - 1;
		}
		iIndex++;
	}
	return;
}

//  InitPool
//
//  This function will initialise the underlying storage for the object
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void AutoScaleFrequencyAnalyser::InitPool()
{
	//  Allocate storage for the pool
	pPool = (UCHAR *)malloc(sizeof(DWORD) * ((2 * ASFANumBuckets) + (2 * ASFANumOutliers)));

	//  Clear the pool
	memset(pPool, 0, sizeof(DWORD) * ((2 * ASFANumBuckets) + (2 * ASFANumOutliers)));

	//  Compute the pointers to the different arrays in the pool
	Bucket = (DWORD *) pPool;
	TempBucket = (DWORD *) (pPool + (sizeof(DWORD) * ASFANumBuckets));
	LowOutliers = (DWORD *) (pPool + (sizeof(DWORD) * (ASFANumBuckets * 2)));
	HighOutliers = (DWORD *) (pPool + (sizeof(DWORD) * ((ASFANumBuckets * 2) + ASFANumOutliers)));

	//  Initialise the sate of the array
	BucketSize = 0;							//  Clear the current bucket size
	B0LowValue = 0;							//  Clear the lower limit of bucket zero
	BnHighValue = 0;						//  Clear the high limit of the last bucket
	DataInBuckets = 0;						//  Mark the buckets empty
	OutliersPermitted = TRUE;				//  Outliers are allowed
	LXCount = 0;							//  No lower bound outliers recorded
	HXCount = 0;							//  No higher bound outliers recorded
	ReNormalised = FALSE;					//  Array has not been renormalised
}
