#pragma once
//  Bitstreams
//
//  These classes allow the handling of bit strings as streams
//
//  Message Identifier:      N/A
//  Message Range:           None
//  Trace area assignment:   None
//  DX Version:				 3.13.0
//  Version:                 3.13.0  build: 107
// 
//  Version History:
//
//  3.13.0		27/12/2013		-  Incorporated into kernel
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2011-2015
//

//  Windows WIN32
#ifdef WIN32
//  Platform/Language includes
#include	<stdlib.h>
#if ((defined(_MSC_VER)) && (_MSC_VER >= 1700))
#include	<stdint.h>
#else
typedef		unsigned __int8			uint8_t;
typedef		unsigned __int16		uint16_t;
typedef		unsigned __int32		uint32_t;
#endif

//  Pseudo intrinsic types
#ifndef		BYTE
typedef		unsigned char			BYTE;	
#endif
#endif

//  Unix generic
#ifdef UNIX
//  Platform/Language includes
#include	<stdlib.h>
#include	<stdint.h>
//  Pseudo intrinsic types
#ifndef		BYTE
typedef		unsigned char			BYTE;	
#endif
#endif

//  ByteStream
//
//  The ByteStream class virtualises a memory buffer as a serial stream, the class provides the base memory 
//  buffer support for the BitStream classes.
//
class ByteStream
{
public:
	//  Constructors
	ByteStream()
	{
		Buffer = NULL;																				//  Clear the buffer address		
		BytesRead = 0;																				//  Clear the count of bytes consumed
		BytesWritten = 0;																			//  Clear the count of bytes written
		EndOfStream = true;																			//  Set the end-of-stream indicator
		return;
	};

	//  Destructor
	~ByteStream() {};

	//  Public Members

	//  Public Functions

	//  setInputBuffer
	//
	//  Sets the address and size of the underlying buffer to provide the stream input.
	//
	//  PARAMETERS
	//
	//		BYTE *				-	Address of the input buffer to use
	//		uint32_t			-	Size of the buffer	
	//
	//	RETURNS
	//
	//	NOTES
	//
	void setInputBuffer(BYTE *pNewBuffer, uint32_t BfrSize)
	{
		Buffer = pNewBuffer;																		//  Set the underlying buffer address
		BufferSize = BfrSize;																		//  Set the buffer size
		BytesRead = 0;																				//  Clear the count of bytes read
		if (Buffer != NULL && BufferSize > 0) EndOfStream = false;									//  Clear the end-of-stream indicator
		return;
	};

	//  setOutputBuffer
	//
	//  Sets the address and size of the underlying buffer to provide the stream storage.
	//
	//  PARAMETERS
	//
	//		BYTE *				-	Address of the input buffer to use
	//		uint32_t			-	Size of the buffer	
	//
	//	RETURNS
	//
	//	NOTES
	//
	void setOutputBuffer(BYTE *pNewBuffer, uint32_t BfrSize)
	{
		Buffer = pNewBuffer;																		//  Set the underlying buffer address
		BufferSize = BfrSize;																		//  Set the buffer size
		BytesWritten = 0;																			//  Clear the count of bytes written
		if (Buffer != NULL && BufferSize > 0) EndOfStream = false;									//  Clear the end-of-stream indicator
		return;
	};

	//  next
	//
	//  Reads the next byte from the input buffer
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		BYTE		-	The next byte available from the buffer, returns zero if no data is available
	//
	//	NOTES
	//
	BYTE next()
	{
		//  If end-of-stream has already been detected return 0
		if (EndOfStream) return 0;

		//  Increment the count of bytes read and test for end-of-stream
		BytesRead++;
		if (BytesRead == BufferSize) EndOfStream = true;

		//  Return the next available byte
		if (Buffer == NULL) return 0;
		return Buffer[BytesRead - 1];
	};

	//  next
	//
	//  Writes the next byte to the output buffer
	//
	//  PARAMETERS
	//
	//		BYTE		-	The next byte to be written to the output
	//
	//	RETURNS
	//
	//	NOTES
	//
	void next(BYTE NextByte)
	{
		//  Check to see if the output buffer is full - if so dump the byte
		if (EndOfStream) return;

		//  Increment the count of bytes written and check for buffer full
		BytesWritten++;
		if (BytesWritten == BufferSize) EndOfStream = true;

		//  Check if this is a NULL buffer
		if (Buffer == NULL) return;

		//  Post the byte to the next position in the buffer
		Buffer[BytesWritten - 1] = NextByte;
		return;
	};

	//  eos
	//
	//  Tests if the input buffer is at the end-of-stream
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		bool			-	true if the end-of-stream is asserted otherwise false
	//
	//	NOTES
	//
	bool eos()
	{
		return EndOfStream;
	};

	//  full
	//
	//  Tests if the output buffer is full
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		bool			-	true if the buffer is full otherwise false
	//
	//	NOTES
	//
	bool full()
	{
		return EndOfStream;
	};

	//  getBytesRead
	//
	//  Returns the count of bytes that have been read from the stream
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t			-	count of bytes that have been consumed (read)
	//
	//	NOTES
	//
	uint32_t getBytesRead()
	{
		return BytesRead;
	};

	//  getBytesWritten
	//
	//  Returns the count of bytes that have been written to the stream
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t			-	count of bytes that have been written
	//
	//	NOTES
	//
	uint32_t getBytesWritten()
	{
		return BytesWritten;
	};

private:
	//  Private Members
	BYTE				*Buffer;																	//  Base buffer address
	uint32_t			BufferSize;																	//  Size of the underlying buffer
	bool				EndOfStream;																//  End of stream has been read
	uint32_t			BytesRead;																	//  Number of bytes consumed
	uint32_t			BytesWritten;																//  Number of bytes written

	//  Private Functions

};

//  MSBitStream
//
//  The MSBitStream class virtualises access to a stream of bytes treating it as a continuous stream of bits. The byte order is MSB first to support 
//  various compression implementations of Huffman & Lempel-Ziv algorithms.
//  This implementation supports a maximum of 32 bits in an individual bit string.
//
class MSBitStream
{
public:
	//  Constructors
	MSBitStream()
	{
		BitOffset = 0;
		
		return;
	};

	//  Destructor
	~MSBitStream() {};

	//  Public Members

	//  Public Functions

	//  setInputBuffer
	//
	//  Sets the address and size of the underlying buffer to provide the stream input.
	//
	//  PARAMETERS
	//
	//		BYTE *				-	Address of the input buffer to use
	//		uint32_t			-	Size of the buffer	
	//
	//	RETURNS
	//
	//	NOTES
	//
	void setInputBuffer(BYTE *pNewBuffer, uint32_t BfrSize)
	{
		//  Set the underlying buffer in the Byte Stream
		bsBase.setInputBuffer(pNewBuffer, BfrSize);

		//  Clear the bit offset
		BitOffset = 0;
		BitsRead = 0;

		//  Fill the byte array from the underlying byte stream
		ByteArray[0] = bsBase.next();
		ByteArray[1] = bsBase.next();
		ByteArray[2] = bsBase.next();
		return;
	};

	//  setOutputBuffer
	//
	//  Sets the address and size of the underlying buffer to provide the output stream memory.
	//
	//  PARAMETERS
	//
	//		BYTE *				-	Address of the input buffer to use
	//		uint32_t			-	Size of the buffer	
	//
	//	RETURNS
	//
	//	NOTES
	//
	void setOutputBuffer(BYTE *pNewBuffer, uint32_t BfrSize)
	{
		//  Set the underlying buffer in the Byte Stream
		bsBase.setOutputBuffer(pNewBuffer, BfrSize);

		//  Clear the bit offset
		BitOffset = 0;
		BitsWritten = 0;

		//  Clear the byte array
		ByteArray[0] = 0;
		ByteArray[1] = 0;
		ByteArray[2] = 0;
		return;
	};

	//  next
	//
	//  Reads the next bit string from the stream
	//
	//  PARAMETERS
	//
	//		uint32_t			-	Size of the bit string to read	
	// 
	//	RETURNS
	//
	//		uint32_t			-	String
	//
	//	NOTES
	//
	uint32_t next(uint32_t Bits)
	{
		uint16_t				Junior,Senior;												//  Junior and senior part of the string

		//  Envelope check
		if (Bits > 32) return 0;

		//  Read the parts of the string
		if (Bits > 16)
		{
			Senior = next(Bits - 16);
			Junior = next16(16);
		}
		else
		{
			Junior = next16(Bits);
			Senior = 0;
		}
		return (Senior << 16) | Junior;
	};

	//  next
	//
	//  Writes the next bit string to the stream
	//
	//  PARAMETERS
	//
	//		uint32_t			-	Bit string to be written
	//		uint32_t			-	Size of the bit string to write	
	// 
	//	RETURNS
	//
	//	NOTES
	//
	void next(uint32_t Out, uint32_t Bits)
	{
		if (Bits > 32) return;
		if (Bits > 16)
		{
			next16(unsigned short(Out >> 16), Bits - 16);
			next16(unsigned short(Out & 0x0000FFFF), 16);
		}
		else next16(unsigned short(Out), Bits);
		return;
	};

	//  flush
	//
	//  Frces any buffered bits to be written to the the current output stream
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//	NOTES
	//
	void flush()
	{
		//  Output any complete BYTES
		while (BitOffset > 0)
		{
			bsBase.next(ByteArray[0]);
			ByteArray[0] = ByteArray[1];
			ByteArray[1] = ByteArray[2];
			ByteArray[2] = 0;
			if (BitOffset < 8) BitOffset = 0;
			else BitOffset -= 8;
		}
		return;
	};

	//  getBytesRead
	//
	//  Returns the total number of BYTES that have been read from the intput stream. The value is obtained from
	//  the underlying BYTEStream
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t		-		total number of BYTES read
	//
	//	NOTES
	//
	uint32_t getBytesRead() 
	{
		return bsBase.getBytesRead();
	};

	//  getBytesWritten
	//
	//  Returns the total number of BYTES that have been written to the output stream. The value is obtained from
	//  the underlying BYTEStream
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t		-		total number of BYTES written
	//
	//	NOTES
	//
	uint32_t getBytesWritten() 
	{
		return bsBase.getBytesWritten();
	};

	//  getBitsRead
	//
	//  Returns the total number of bits that have been read from the intput stream.
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t		-		total number of bits read
	//
	//	NOTES
	//
	uint32_t getBitsRead() 
	{
		return (bsBase.getBytesRead() * 8) + BitOffset;
	};

	//  getBitsWritten
	//
	//  Returns the total number of bits that have been written to the outtput stream.
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t		-		total number of bits written
	//
	//	NOTES
	//
	uint32_t getBitsWritten() 
	{
		return (bsBase.getBytesWritten() * 8) + BitOffset;
	};

	//  isStreamByteAligned
	//
	//  Indicates if the stream is currently aligned on a BYTE boundary or not
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		bool		-		true if the stream is aligned otherwise false
	//
	//	NOTES
	//
	bool isStreamByteAligned()
	{
		if (BitOffset == 0) return true;
		return false;
	};

private:
	//  Private Members
	ByteStream			bsBase;																//  Underlying Byte Stream
	uint32_t			BitOffset;															//  Offset of the next Bit to be read/written
	uint32_t			BitsRead;															//  Bits read (consumed) from the stream
	uint32_t			BitsWritten;														//  Bits written to the stream
	BYTE				ByteArray[3];														//  Array of bytes read from the underlying stream

	//  Private Functions

	//  next16
	//
	//  Reads the next bit string (up to 16 bits) from the stream
	//
	//  PARAMETERS
	//
	//		uint16_t			-	Size of the bit string to read	
	// 
	//	RETURNS
	//
	//		uint16_t			-	String
	//
	//	NOTES
	//
	uint16_t next16(uint32_t Bits)
	{
		uint32_t			Acc = 0;														//  Accumulator for the bit string
		uint32_t			Mask = 0xffffffff;												//  Mak for the Accumulator values

		//  Build the bitstring for the next n Bits
		Acc = ByteArray[0];
		Acc <<= 8;
		Acc |= ByteArray[1];
		Acc <<= 8;
		Acc |= ByteArray[2];

		Acc <<= BitOffset;
		Acc >>= (24 - Bits);

		Mask <<= (32 - Bits);
		Mask >>= (32 - Bits);
		Acc &= Mask;
		BitOffset += Bits;
		BitsRead += Bits;

		//  Replace any whole bytes that have been consumed
		while (BitOffset >= 8)
		{
			ByteArray[0] = ByteArray[1];
			ByteArray[1] = ByteArray[2];
			ByteArray[2] = bsBase.next();
			BitOffset -= 8;
		}

		//  Return the string
		return uint16_t(Acc);
	};

	//  next16
	//
	//  Writes the next bit string (up to 16 bits) to the stream
	//
	//  PARAMETERS
	//
	//		uint16_t			-   Value to be written
	//		uint16_t			-	Size of the bit string to read	
	// 
	//	RETURNS
	//
	//	NOTES
	//
	void next16(uint16_t Out, uint16_t Bits)
	{
		unsigned int			Acc = 0;									//  Bit string accumulator

		//  Load the existing BYTE array into the accumulator
		Acc = ByteArray[0];
		Acc <<= 8;
		Acc |= ByteArray[1];
		Acc <<= 8;
		Acc |= ByteArray[2];

		//  Accumulate the output bits
		Acc |= (Out << ((24 - BitOffset) - Bits));

		//  Store back the accumulated value
		ByteArray[2] = BYTE(Acc & 0x000000FF);
		ByteArray[1] = BYTE((Acc >> 8) & 0x000000FF);
		ByteArray[0] = BYTE((Acc >> 16) & 0x000000FF);
		
		BitOffset += Bits;
		BitsWritten += Bits;

		//  Output any complete BYTES
		while (BitOffset >= 8)
		{
			bsBase.next(ByteArray[0]);
			ByteArray[0] = ByteArray[1];
			ByteArray[1] = ByteArray[2];
			ByteArray[2] = 0;
			BitOffset -= 8;
		}
		return;
	};

};

//  LSBitStream
//
//  The LSBitStream class virtualises access to a stream of bytes treating it as a continuous stream of bits. The byte order is LSB first to support 
//  compressed image formats such as GIF.
//  This implementation supports a maximum of 32 bits in an individual bit string.
//
class LSBitStream
{
	//  Constructors
	LSBitStream()
	{
		BitOffset = 0;
		return;
	};

	//  Destructor
	~LSBitStream() {};

	//  Public Members

	//  Public Functions

	//  setInputBuffer
	//
	//  Sets the address and size of the underlying buffer to provide the stream input.
	//
	//  PARAMETERS
	//
	//		BYTE *				-	Address of the input buffer to use
	//		uint32_t			-	Size of the buffer	
	//
	//	RETURNS
	//
	//	NOTES
	//
	void setInputBuffer(BYTE *pNewBuffer, uint32_t BfrSize)
	{
		//  Set the underlying buffer in the Byte Stream
		bsBase.setInputBuffer(pNewBuffer, BfrSize);

		//  Clear the bit offset
		BitOffset = 0;
		BitsRead = 0;

		//  Fill the byte array from the underlying byte stream
		ByteArray[0] = bsBase.next();
		ByteArray[1] = bsBase.next();
		ByteArray[2] = bsBase.next();
		return;
	};

	//  setOutputBuffer
	//
	//  Sets the address and size of the underlying buffer to provide the output stream memory.
	//
	//  PARAMETERS
	//
	//		BYTE *				-	Address of the input buffer to use
	//		uint32_t			-	Size of the buffer	
	//
	//	RETURNS
	//
	//	NOTES
	//
	void setOutputBuffer(BYTE *pNewBuffer, uint32_t BfrSize)
	{
		//  Set the underlying buffer in the Byte Stream
		bsBase.setOutputBuffer(pNewBuffer, BfrSize);

		//  Clear the bit offset
		BitOffset = 0;
		BitsWritten = 0;

		//  Clear the byte array
		ByteArray[0] = 0;
		ByteArray[1] = 0;
		ByteArray[2] = 0;
		return;
	};

	//  next
	//
	//  Reads the next bit string from the stream
	//
	//  PARAMETERS
	//
	//		uint32_t			-	Size of the bit string to read	
	// 
	//	RETURNS
	//
	//		uint32_t			-	String
	//
	//	NOTES
	//
	uint32_t next(uint32_t Bits)
	{
		uint16_t				Junior,Senior;												//  Junior and senior part of the string

		//  Envelope check
		if (Bits > 32) return 0;

		//  Read the parts of the string
		if (Bits > 16)
		{
			Junior = next16(16);
			Senior = next(Bits - 16);
		}
		else
		{
			Junior = next16(Bits);
			Senior = 0;
		}
		return (Senior << 16) | Junior;
	};

	//  next
	//
	//  Writes the next bit string to the stream
	//
	//  PARAMETERS
	//
	//		uint32_t			-	Bit string to be written
	//		uint32_t			-	Size of the bit string to write	
	// 
	//	RETURNS
	//
	//	NOTES
	//
	void next(uint32_t Out, uint32_t Bits)
	{
		if (Bits > 32) return;
		if (Bits > 16)
		{
			next16(unsigned short(Out & 0x0000FFFF), 16);
			next16(unsigned short(Out >> 16), Bits - 16);
		}
		else next16(unsigned short(Out), Bits);
		return;
	};

	//  flush
	//
	//  Frces any buffered bits to be written to the the current output stream
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//	NOTES
	//
	void flush()
	{
		//  Output any complete BYTES
		while (BitOffset > 0)
		{
			bsBase.next(ByteArray[0]);
			ByteArray[0] = ByteArray[1];
			ByteArray[1] = ByteArray[2];
			ByteArray[2] = 0;
			if (BitOffset < 8) BitOffset = 0;
			else BitOffset -= 8;
		}
		return;
	};

	//  getBytesRead
	//
	//  Returns the total number of BYTES that have been read from the intput stream. The value is obtained from
	//  the underlying BYTEStream
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t		-		total number of BYTES read
	//
	//	NOTES
	//
	uint32_t getBytesRead() 
	{
		return bsBase.getBytesRead();
	};

	//  getBytesWritten
	//
	//  Returns the total number of BYTES that have been written to the output stream. The value is obtained from
	//  the underlying BYTEStream
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t		-		total number of BYTES written
	//
	//	NOTES
	//
	uint32_t getBytesWritten() 
	{
		return bsBase.getBytesWritten();
	};

	//  getBitsRead
	//
	//  Returns the total number of bits that have been read from the intput stream.
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t		-		total number of bits read
	//
	//	NOTES
	//
	uint32_t getBitsRead() 
	{
		return (bsBase.getBytesRead() * 8) + BitOffset;
	};

	//  getBitsWritten
	//
	//  Returns the total number of bits that have been written to the outtput stream.
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		uint32_t		-		total number of bits written
	//
	//	NOTES
	//
	uint32_t getBitsWritten() 
	{
		return (bsBase.getBytesWritten() * 8) + BitOffset;
	};

	//  isStreamByteAligned
	//
	//  Indicates if the stream is currently aligned on a BYTE boundary or not
	//
	//  PARAMETERS
	//
	//	RETURNS
	//
	//		bool		-		true if the stream is aligned otherwise false
	//
	//	NOTES
	//
	bool isStreamByteAligned()
	{
		if (BitOffset == 0) return true;
		return false;
	};

private:
	//  Private Members
	ByteStream			bsBase;																//  Underlying Byte Stream
	uint32_t			BitOffset;															//  Offset of the next Bit to be read/written
	uint32_t			BitsRead;															//  Bits read (consumed) from the stream
	uint32_t			BitsWritten;														//  Bits written to the stream
	BYTE				ByteArray[3];														//  Array of bytes read from the underlying stream

	//  Private Functions

	//  next16
	//
	//  Reads the next bit string (up to 16 bits) from the stream
	//
	//  PARAMETERS
	//
	//		uint16_t			-	Size of the bit string to read	
	// 
	//	RETURNS
	//
	//		uint16_t			-	String
	//
	//	NOTES
	//
	uint16_t next16(uint32_t Bits)
	{
		uint32_t			Acc = 0;														//  Accumulator for the bit string

		//  Build the bitstring for the next n Bits
		Acc = ByteArray[2];
		Acc <<= 8;
		Acc |= ByteArray[1];
		Acc <<= 8;
		Acc |= ByteArray[0];
		Acc >>= BitOffset;
		Acc <<= (24 - Bits);
		Acc &= 0x00FFFFFF;
		Acc >>= (24 - Bits);

		BitOffset += Bits;
		BitsRead += Bits;

		//  Replace any complete bytes that were consumed
		while (BitOffset >= 8)
		{
			ByteArray[0] = ByteArray[1];
			ByteArray[1] = ByteArray[2];
			ByteArray[2] = bsBase.next();
			BitOffset -= 8;
		}

		//  Return the string
		return uint16_t(Acc);
	};

	//  next16
	//
	//  Writes the next bit string (up to 16 bits) to the stream
	//
	//  PARAMETERS
	//
	//		uint16_t			-   Value to be written
	//		uint16_t			-	Size of the bit string to read	
	// 
	//	RETURNS
	//
	//	NOTES
	//
	void next16(uint16_t Out, uint16_t Bits)
	{
		unsigned int			Acc = 0;									//  Bit string accumulator

		//  Load the existing BYTE array into the accumulator
		Acc = ByteArray[0];
		Acc <<= 8;
		Acc |= ByteArray[1];
		Acc <<= 8;
		Acc |= ByteArray[2];

		//  Accumulate the output bits
		Acc |= (Out << BitOffset);

		//  Store back the accumulated value
		ByteArray[2] = BYTE(Acc & 0x000000FF);
		ByteArray[1] = BYTE((Acc >> 8) & 0x000000FF);
		ByteArray[0] = BYTE((Acc >> 16) & 0x000000FF);
		
		BitOffset += Bits;

		//  Output any complete BYTES
		while (BitOffset >= 8)
		{
			bsBase.next(ByteArray[2]);
			ByteArray[2] = ByteArray[1];
			ByteArray[1] = ByteArray[0];
			ByteArray[0] = 0;
			BitOffset -= 8;
		}
		return;
	};
};
