//  ObjectCache
//
//  This is the base class for an Object Cache supporting both Most Recently Used (MRU) cacheing and
//  Most Frequently Used (MFU) cacheing of aritrary objects.
//
//  Message Identifier:      DXR
//  Message Range:           2901 - 2999
//  Trace area assignment:   N/A
//  Version:                 3.11.5  Build: 103
//
//  Version History:
//
//  1.0.0		13/02/2011		Initial version
//  3.10.11		19/07/2011		Incorporated into DXCommon Kernel
//  3.11.1		22/10/2011		VS 2011 readiness
//  3.11.5		30/11/2011		Cache coherence improvements
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2011
//

//  Application Includes
#include	"ObjectCache.h"

//  Default Constructor
ObjectCache::ObjectCache()
: Cache(NULL)
, CacheEntries(0)
, dwSearches(0)
, dwSearched(0)
, dwNoHits(0)
, dwReorgs(0)
, dwWrites(0)
, CacheMutex(CACHE_MUTEX_FREE)
, dwWaits(0)
, dwWaitQuanta(0)
, dwRSkips(0)
, dwMFUIProms(0)
, dwMFUIPQs(0)
{

}

//  Destructor
ObjectCache::~ObjectCache()
{
	//  Free the cache allocation if it has been made
	if (Cache != NULL)
	{
		free(Cache);
		Cache = NULL;
	}
}

//  initCache
//
//  This function is invoked by constructors in extending classes in order to initialise the 
//  underlying cache infrastructure.
//
//  Parameters
//
//  1  -  int				-  Maximum size of the cache in number of entries
//  2  -  int				-  Cache strategy MFU/MRU
//  3  -  int				-  ID of the mutext to use synchronize cache activities
//  4  -  int				-  ID of the thread initialising the cache
//
//  Returns
//
//
//  API Use Notes:
//
//
void ObjectCache::initCache(int iMaxEntries, int iStrategy, int iThreadID)
{
	//  Safety checks
	CacheSize = iMaxEntries;
	if (CacheSize < CACHE_MIN_ENTRIES) CacheSize = CACHE_MIN_ENTRIES;
	if (CacheSize > CACHE_MAX_ENTRIES) CacheSize = CACHE_MAX_ENTRIES;
	CacheStrategy = iStrategy;
	if ((CacheStrategy != CACHE_STRATEGY_MFU) && (CacheStrategy != CACHE_STRATEGY_MRU))
		CacheStrategy = CACHE_STRATEGY_MRU;

	//  Allocate the cache storage
	Cache = (CacheEntry *) malloc(CacheSize * sizeof(CacheEntry));
	memset(Cache, 0, CacheSize * sizeof(CacheEntry));
	CacheMutex = CACHE_MUTEX_FREE;
	return;
}

//  getCachedObject
//
//  This function will retrieve an object from the cache.
//  If the object does not exist in the cache then it will be added.
//
//  Parameters
//
//  1  -  char *			-  Pointer to a key string to be mateched against objects 
//  2  -  int				-  ID of the thread requesting the object
//
//  Returns
//
//  void *					-  Pointer to an object that matches the supplied key
//
//  API Use Notes:
//
//
void * ObjectCache::getCachedObject(char *szKey, int iThreadID)
{
	return getCachedObjectExt(szKey, TRUE, iThreadID);
}

//  getCachedObjectExt
//
//  This function will retrieve an object from the cache.
//  If the object does not exist in the cache then it will be added only if the Create flag is on.
//
//  Parameters
//
//  1  -  char *			-  Pointer to a key string to be mateched against objects 
//  2  -  BOOL				-  Flag indicating the bahaviour if an entry is not in the cache
//  3  -  int				-  ID of the thread requesting the object
//
//  Returns
//
//  void *					-  Pointer to an object that matches the supplied key
//
//  API Use Notes:
//
//  If the key is not matched (i.e. the object is not in the cache) then the functon either returns
//  a new object that is built to match the supplied key (bCreate = TRUE) or returns NULL (bCreate = FALSE).
//
void * ObjectCache::getCachedObjectExt(char *szKey, BOOL bCreate, int iThreadID)
{
	DWORD			dwCacheEntry;								//  Cache entry index
	DWORD			dwSafeRefCount;								//  Safe Reference Count
	CacheEntry		*pCEntry;									//  Pointer to the current cache entry
	CacheEntry		ceNew;										//  New cache entry
	void			*pRetObj;									//  Object to be treturned

	//  If the cache is not initialised then return NULL
	if (Cache == NULL) return NULL;

	//  Acquire the cache mutex
	lockTheCache(iThreadID);

	//  Increment the count of cache searches
	dwSearches++;
	dwSafeRefCount = dwSearches;

	//  Loop processing each object that is currently in the cache to determine if it is the one selected.
	//  If it is the selected one then return it to the caller.
	pCEntry = Cache;
	dwCacheEntry = 0;

	while (dwCacheEntry < CacheEntries)
	{
		pRetObj = pCEntry->ObjPtr;
		if (pRetObj == NULL) dwCacheEntry = CacheEntries;
		else
		{
			//  Increment the count of entries searched
			dwSearched++;
			if (selectObject(pRetObj, szKey, iThreadID))
			{
				//  The current entry is selected reorganise the cache and return the object
				pCEntry->RefCount++;
				reorganiseCache(pCEntry, dwCacheEntry, dwSafeRefCount, iThreadID);

				//  Free the cache mutex
				unlockTheCache(iThreadID);

				//  Return the object pointer
				return pRetObj;
			}
			dwCacheEntry++;
			pCEntry++;
		}
	}

	//  Increment the count of no hits
	dwNoHits++;

	//  Return a NULL pointer if we are not asked to create a new cache entry
	if (!bCreate) 
	{
		//  Free the cache mutex
		unlockTheCache(iThreadID);
		return NULL;
	}

	//  Make sure that we hold (or still hold) the cache mutex
	lockTheCache(iThreadID);

	//  The entry does not exist in the cache it needs to be created and inserted into the cache
	//  before returning it to the user.
	ceNew.ObjPtr = createObject(szKey, iThreadID);
	ceNew.RefCount = 1;

	//  Make sure that we hold (or still hold) the cache mutex
	lockTheCache(iThreadID);

	//  Insert the object into a new cache entry
	insertCacheEntry(&ceNew, iThreadID);

	//  Free the cache mutex
	unlockTheCache(iThreadID);
	//  Return the new cache object
	return ceNew.ObjPtr;
}

//  selectObject
//
//  This function will be implemented in any extending class, this default implementation 
//  will always return FALSE (not selected).
//
//  Parameters
//
//  1  -  void *			-  Pointer to the object to be selected or not
//  2  -  char *			-  Pointer to a key string to be mateched against objects 
//  3  -  int				-  ID of the thread requesting the object
//
//  Returns
//
//  BOOL					-  TRUE is the object matches the key, FALSE if not
//
//  API Use Notes:
//
//
BOOL ObjectCache::selectObject(void *pObject, char *szKey, int iThreadID)
{
	return FALSE;
}

//  createObject
//
//  This function will be implemented in any extending class, this default implementation 
//  will always return a pointer to the Cache object itself which is a PUI.
//  Patently Useless Implementation !
//
//  Parameters
//
//  1  -  char *			-  Pointer to a key string to be mateched against objects 
//  2  -  int				-  ID of the thread requesting the object
//
//  Returns
//
//  void *		-		A pointer to the new object
//
//  API Use Notes:
//
//  The contract for this interface asserts that an object is ALWAYS returned. In the case where an
//  object cannot be created from the specified key then an empty or dummy object should be returned.
//
void *ObjectCache::createObject(char *szKey, int iThreadID)
{
	return this;
}

//  cacheOut
//
//  This function will be implemented in any extending class, this default implementation 
//  does nothing.
//
//  Parameters
//
//  1  -  void *			-  Pointer to the object to be selected or not
//  2  -  BOOL				-  If TRUE then the object should be written to persistent storage
//  2  -  int				-  ID of the thread requesting the object
//
//  Returns
//
//
//  API Use Notes:
//
//
void ObjectCache::cacheOut(void *pObject, BOOL bWrite, int iThreadID)
{
	return;
}

//  reorganiseCache
//
//  This function will move the current cache entry within the cache based on the change in
//  reference count for Most Frequently Used (MFU) strategy or position for Most Recently Used (MRU).
//
//  Parameters
//
//  1  -  CacheEntry *		-  Pointer to the ChacheEntry that has been updated
//  2  -  DWORD				-  Index of the entry within the cache
//  3  -  int				-  ID of the calling thread 
//
//  Returns
//
//
//  API Use Notes:
//
//
void ObjectCache::reorganiseCache(CacheEntry *pUPEntry, DWORD dwIndex, DWORD dwSafeCount, int iThreadID)
{
	CacheEntry			*pSrcEntry;					//  Source Cache Entry
	CacheEntry			*pTargEntry;				//  Target Cache Entry
	CacheEntry			ceTemp;						//  Temp storage for cache entry
	DWORD				dwCurrentSrcIndex;			//  Current source entry index in the cache

	//  Skip the boundary condition where there is only a single Cache entry
	if (CacheEntries == 1)  return;

	//  Skip the boundary conbdition where it is MRU strategy and the entry is already at the top
	if ((dwIndex == 0) && (CacheStrategy == CACHE_STRATEGY_MRU)) return;

	//  Reacquire the mutex for safety
	lockTheCache(iThreadID);

	//  Check that the ref count is the same as safety count, if not then skip the reorg
	if (dwSafeCount != dwSearches)
	{
		dwRSkips++;
		return;
	}

	dwCurrentSrcIndex = dwIndex;
	pSrcEntry = pUPEntry;
	pTargEntry = pUPEntry - 1;
	dwReorgs++;

	//  Determine the strategy
	if (CacheStrategy == CACHE_STRATEGY_MFU)
	{
		//  Most Frequently Used - promote the updated entry up the cache until the reference count
		//  is less than or equal to the entry above.
		while ((dwCurrentSrcIndex > 0) && (pSrcEntry->RefCount > pTargEntry->RefCount))
		{
			//  Swap the source and target entries
			ceTemp.ObjPtr = pTargEntry->ObjPtr;
			ceTemp.RefCount = pTargEntry->RefCount;
			pTargEntry->ObjPtr = pSrcEntry->ObjPtr;
			pTargEntry->RefCount = pSrcEntry->RefCount;
			pSrcEntry->ObjPtr = ceTemp.ObjPtr;
			pSrcEntry->RefCount = ceTemp.RefCount;
			pSrcEntry = pTargEntry;
			pTargEntry--;
			dwCurrentSrcIndex--;
		}
	}
	else
	{
		//  Most Recently Used - promote the updated entry to the head of the cache
		ceTemp.ObjPtr = pSrcEntry->ObjPtr;
		ceTemp.RefCount = pSrcEntry->RefCount;
		pTargEntry = pSrcEntry;
		pSrcEntry--;
		while (dwCurrentSrcIndex > 0)
		{
			pTargEntry->ObjPtr = pSrcEntry->ObjPtr;
			pTargEntry->RefCount = pSrcEntry->RefCount;
			pSrcEntry--;
			pTargEntry--;
			dwCurrentSrcIndex--;
		}
		pSrcEntry = Cache;
		pSrcEntry->ObjPtr = ceTemp.ObjPtr;
		pSrcEntry->RefCount = ceTemp.RefCount;
	}
	return;
}

//  insertCacheEntry
//
//  This function will insert a new entry into the cache
//
//  Parameters
//
//  1  -  CacheEntry *		-  Pointer to the ChachEntry that has been updated
//  2  -  int				-  ID of the calling thread 
//
//  Returns
//
//
//  API Use Notes:
//
//
void ObjectCache::insertCacheEntry(CacheEntry *pNewEntry, int iThreadID)
{
	CacheEntry			ceTemp;									//  Temp Cache entry
	CacheEntry			*pCEntry;								//  Pointer to a cache entry
	CacheEntry			*pSrcEntry;								//  Source Entry
	CacheEntry			*pTargEntry;							//  Target Entry
	void				*pCachedObject;							//  Pointer to the cached object

	//  Point to the first entry
	pCEntry = Cache;

	//  Test for the boundary condition where the cache is empty
	if (CacheEntries == 0)
	{
		pCEntry->ObjPtr = pNewEntry->ObjPtr;
		pCEntry->RefCount = pNewEntry->RefCount;
		CacheEntries++;
		return;
	}

	//  Determine the cache strategy 
	if (CacheStrategy == CACHE_STRATEGY_MRU)
	{
		//  Strategy is Most Recently Used - insert the entry at the start of the cache
		//  If the cache is full then firstly remove the last cache entry
		if (CacheEntries == CacheSize)
		{
			pSrcEntry = Cache + (CacheEntries - 1);
			pCachedObject = pSrcEntry->ObjPtr;
			pSrcEntry->ObjPtr = NULL;
			pSrcEntry->RefCount = 0;
			CacheEntries--;
			cacheOut(pCachedObject, TRUE, iThreadID);
		}

		//  Move all entries down in the cache
		pTargEntry = Cache + CacheEntries;
		pSrcEntry = pTargEntry;
		pSrcEntry--;

		while (pTargEntry != Cache)
		{
			pTargEntry->ObjPtr = pSrcEntry->ObjPtr;
			pTargEntry->RefCount = pSrcEntry->RefCount;
			pTargEntry--;
			pSrcEntry--;
		}

		//  Insert the new entry at the head of the cache and return
		pCEntry->ObjPtr = pNewEntry->ObjPtr;
		pCEntry->RefCount = pNewEntry->RefCount;
	}
	else
	{
		//  Strategy is Most Frequently Used - the new entry is inserted at the last position in
		//  and then promoted to the first position with refcount = 1
		//  If the cache is full then firstly remove the last cache entry
		if (CacheEntries == CacheSize)
		{
			pSrcEntry = Cache + (CacheEntries - 1);
			pCachedObject = pSrcEntry->ObjPtr;
			pSrcEntry->ObjPtr = NULL;
			pSrcEntry->RefCount = 0;
			CacheEntries--;
			cacheOut(pCachedObject, TRUE, iThreadID);
		}
		pTargEntry = Cache + CacheEntries;
		pTargEntry->ObjPtr = pNewEntry->ObjPtr;
		pTargEntry->RefCount = pNewEntry->RefCount;

		//  Now promote the entry up the cache
		pSrcEntry = pTargEntry;
		pTargEntry--;
		if (pSrcEntry->RefCount == pTargEntry->RefCount) dwMFUIProms++;
		while ((pSrcEntry->RefCount == pTargEntry->RefCount) && (pSrcEntry != Cache))
		{
			ceTemp.ObjPtr = pTargEntry->ObjPtr;
			ceTemp.RefCount = pTargEntry->RefCount;
			pTargEntry->ObjPtr = pSrcEntry->ObjPtr;
			pTargEntry->RefCount = pSrcEntry->RefCount;
			pSrcEntry->ObjPtr = ceTemp.ObjPtr;
			pSrcEntry->RefCount = ceTemp.RefCount;
			pSrcEntry--;
			pTargEntry--;
			dwMFUIPQs++;
		}
	}
	CacheEntries++;
	return;
}

//  drain
//
//  This function will drain the cache by passing each cached object to the cacheOut() interface.
//
//  Parameters
//
//  1  -  int				-  ID of the calling thread 
//
//  Returns
//
//
//  API Use Notes:
//
//
void ObjectCache::drain(int iThreadID)
{
	CacheEntry			*pCEntry;							//  Pointer to cache entry

	pCEntry = Cache;
	while (CacheEntries > 0)
	{
		cacheOut(pCEntry->ObjPtr, TRUE, iThreadID);
		pCEntry++;
		CacheEntries--;
	}
	return;
}

//  dump
//
//  This function will dump the current content of the cache
//
//  Parameters
//
//  1  -  int				-  ID of the calling thread 
//
//  Returns
//
//
//  API Use Notes:
//
//
void ObjectCache::dump(int iThreadID)
{
	CacheEntry			*pCEntry;							//  Pointer to cache entry

	pCEntry = Cache;
	while (CacheEntries > 0)
	{
		cacheOut(pCEntry->ObjPtr, FALSE, iThreadID);
		pCEntry++;
		CacheEntries--;
	}
	return;
}

//  lockTheCache
//
//  This function assert the Cache Mutex on behalf of the calling thread
//
//  Parameters
//
//  1  -  int				-  ID of the calling thread 
//
//  Returns
//
//
//  API Use Notes:
//
//
void ObjectCache::lockTheCache(int iThreadID)
{
	DWORD				dwCWaits = 0;									//  Wait count
	DWORD				dwCWaitQuanta = 0;								//  Wait Qanta
	while (CacheMutex != iThreadID)
	{
#ifdef WIN32
		while (InterlockedCompareExchange(&CacheMutex, iThreadID, CACHE_MUTEX_FREE) != CACHE_MUTEX_FREE)
		{
			dwCWaits = 1;
			dwCWaitQuanta++;
			SLEEP(MUTEX_WAIT_MILLIS);
		}
#endif
#ifdef UNIX
		while (DXAtomic_cmpxchg(&CacheMutex, CACHE_MUTEX_FREE, iThreadID) != CACHE_MUTEX_FREE)
		{
			dwCWaits = 1;
			dwCWaitQuanta++;
			SLEEP(MUTEX_WAIT_MILLIS);
		}
#endif
	}
	dwWaits += dwCWaits;
	dwWaitQuanta += dwCWaitQuanta;
	return;
}

//  unlockTheCache
//
//  This function release the Cache Mutex on behalf of the calling thread
//
//  Parameters
//
//  1  -  int				-  ID of the calling thread 
//
//  Returns
//
//
//  API Use Notes:
//
//
void ObjectCache::unlockTheCache(int iThreadID)
{
	if (CacheMutex == iThreadID) CacheMutex = CACHE_MUTEX_FREE;
	return;
}

//  getEntries
//
//  This function will return the count of entries currently in the cache
//
//  Parameters
//
//  Returns
//
//  DWORD				-		Count of entries in the cache
//
//  API Use Notes:
//
//
DWORD ObjectCache::getEntries()
{
	return CacheEntries;
}

//  getNextCachedObject
//
//  This function provides an iterator over the contents of the cache. Given the address of an object that
//  is currently in the cache it will return the next entry. If the current address provided is NULL
//  then it returns the first object in the cache.
//
//  Parameters
//
//  1  -  void *			-  Pointer to an object that is currently in the cache or NULL
//  2  -  int				-  ID of the thread requesting the object
//
//  Returns
//
//  void *					-  Pointer to the next object in the cache (NULL if there are no more)
//
//  API Use Notes:
//
//
void * ObjectCache::getNextCachedObject(void *pCurrentObject, int iThreadID)
{
	DWORD			dwCacheEntry;								//  Cache entry index
	void			*pObject;									//  Object pointer
	CacheEntry		*pCEntry;									//  Pointer to the current cache entry

	//  If the cache is not established then return NULL
	if (Cache == NULL) return NULL;

	//  Return NULL if cache is empty
	if (CacheEntries == 0) return NULL;

	//  Lock the cache
	lockTheCache(iThreadID);

	//  Establish position at the start of the cache
	pCEntry = Cache;
	dwCacheEntry = 1;

	//  If the first entry was requested then return it
	if (pCurrentObject == NULL)
	{
		pObject = pCEntry->ObjPtr;
		unlockTheCache(iThreadID);
		return pObject;
	}

	//  Iterate over the cache until we match the passed object
	while ((pCEntry->ObjPtr != pCurrentObject) && (dwCacheEntry < CacheEntries))
	{
		//  Point to the next entry
		pCEntry++;
		dwCacheEntry++;
	}

	//  If we ended up pointing at the last entry then it means either that we did not match the
	//  passed pointer or that we matched the last entry. In either case return NULL.
	if (dwCacheEntry == CacheEntries)
	{
		unlockTheCache(iThreadID);
		return NULL;		
	}

	//  Point to the next entry and return the object
	pCEntry++;
	pObject = pCEntry->ObjPtr;
	unlockTheCache(iThreadID);
	return pObject;
}
