#pragma once
//  ObjectCache
//
//  This is the base class for an Object Cache supporting both Most Recently Used (MRU) cacheing and
//  Most Frequently Used (MFU) cacheing of aritrary objects.
//
//  Message Identifier:      DXR
//  Message Range:           2901 - 2999
//  Trace area assignment:   N/A
//  Version:                 3.11.5  Build: 103
//
//  Version History:
//
//  1.0.0		13/02/2011		Initial version
//  3.10.11		19/07/2011		Incorporated into DXCommon Kernel
//  3.11.1		22/10/2011		VS 2011 readiness
//  3.11.5		30/11/2011		Cache coherence improvements
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2011
//

//  Platform Includes
#include	"../Platform/PlatBase.h"

//  Runtime Includes
#include	"../MTX/MTExecutive.h"

//  Local Definitions
#define			CACHE_MIN_ENTRIES		64			//  Minimum number of cache entries
#define			CACHE_MAX_ENTRIES		4096		//  Maximum number of cache entries

//  Cache Strategy settings
#define			CACHE_STRATEGY_MFU		1			//  Most Frequently Used
#define			CACHE_STRATEGY_MRU		2			//  Most Recently Used

//  Cache Mutex Values
#define			CACHE_MUTEX_FREE		-1			//  Mutex is free
#define			MUTEX_WAIT_MILLIS		25			//  25 ms wait for the mutex

//  Trace Messages

//  Local Structures
typedef struct {
	DWORD		RefCount;							//  Refrence Count
	void		*ObjPtr;							//  Object Entry pointer
} CacheEntry;

//  Class Definition
class ObjectCache
{
public:
	//  Default Constructor
	ObjectCache(void);

	//  Destructor
	~ObjectCache(void);

	//  Public Members
	DWORD					dwSearches;						//  Count of cache searches
	DWORD					dwSearched;						//  Count of cache entries searched
	DWORD					dwNoHits;						//  Count of No Hits
	DWORD					dwReorgs;						//  Count of reorganisations
	DWORD					dwWrites;						//  Count of Writes
	DWORD					dwWaits;						//  Count of waits on the cache mutex
	DWORD					dwWaitQuanta;					//  Count of wait intervals on the cache mutex
	DWORD					dwRSkips;						//  Count of reorgs skipped
	DWORD					dwMFUIProms;					//  Count of MFU initial promotes
	DWORD					dwMFUIPQs;						//  Count of MFU initial promote slots

	//  Public Functions
	void *getCachedObject(char *szKey, int iThreadID);		//  Retrieve a cached entry
	void *getCachedObjectExt(char *szKey, BOOL bCreate, int iThreadID); //  Flexible Retrieval
	void *getNextCachedObject(void *pCurrentObject, int iThreadID);		//  Iterator
	void drain(int iThreadID);								//  Drain the cache
	void dump(int iThreadID);								//  Dump the cache
	DWORD getEntries(void);									//  Get the count of entries in the cache

private:
	//  Private Members
	FORCE_ALIGN volatile UINT			CacheMutex;						//  Cache Mutex

	//  Private Functions
	void reorganiseCache(CacheEntry *pUPEntry, DWORD dwIndex, DWORD dwSafeCount, int iThreadID);	//  Reorganise the cache
	void insertCacheEntry(CacheEntry *pNewEntry, int iThreadID);				//  Insert a new entry

protected:
	//  Protected Members
	CacheEntry				*Cache;							//  The cache storage
	DWORD					CacheSize;						//  Size of the cache in entries
	DWORD					CacheEntries;					//  Number of entries in the cache
	int						CacheStrategy;					//  Cache Strategy

	//  Protected Functions
	void initCache(int iMaxEntries, int iStrategy, int iThreadID);				//  Initialise the cache
	virtual BOOL selectObject(void *pObject, char *szKey, int iThreadID);		//  Determine if this is the selected object
	virtual void *createObject(char *szKey, int iThreadID);						//  Create a new object
	virtual void cacheOut(void *pObject, BOOL bWrite, int iThreadID);			//  Remove a cache entry
	void lockTheCache(int iThreadID);											//  Lock the cache
	void unlockTheCache(int iThreadID);											//  Unlock the cache
};

