//  TimeBasedArray
//
//  This class provides the Time Based Array
//
//  Message Identifier:      N/A
//  Message Range:           None
//  Trace area assignment:   None
//  DX Version:				 3.14.0
//  Version:                 3.14.0  build: 108
// 
//  Version History:
//
//	1.0.0 -		18/11/2011		-  Initial Build
//  3.11.4		27/11/2011		-  Incorporated into kernel
//	3.14.0		21/01/2015		-  x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2011 - 2015
//

//  Application Includes
#include "TimeBasedArray.h"

//  Default Constructor
TimeBasedArray::TimeBasedArray()
{
	VarCount = 1;									//  Default to a single counter
	Hourly = TRUE;									//  Hourly recording defaults to on
	pAStorage = NULL;								//  Show array not initialised
	TimeConstant(TIMEDATE_MINIMUM, &tdAnchor);		//  Set the ancor to minimum value
}

//  Normal Constructor
//
//  Parameters
//
//  1	-	int			- Count of the variables to be handled in the array
//  2	-	BOOL		- TRUE is hourly spectrum is to be used, FALSE if not
//
TimeBasedArray::TimeBasedArray(int iVars, BOOL bRecordHourly)
{
	if (iVars < 1) VarCount = 1;
	else VarCount = iVars;
	Hourly = bRecordHourly;
	TimeConstant(TIMEDATE_MINIMUM, &tdAnchor);		//  Set the anchor to minimum value
	pAStorage = NULL;								//  Show array not initialised
	initArray();									//  Initialise the Array Storage	
}

//  Destructor
TimeBasedArray::~TimeBasedArray()
{
	//  If there is a storage array then free it
	if (pAStorage != NULL) free(pAStorage);
}

//  initArray
//
//  This function will initialise the storage array 
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void TimeBasedArray::initArray()
{
	DWORD dwArraySize = 0;											//  Array size
	
	dwOffHourly = dwArraySize;								//  Offset within a variable to the Hourly array

	//  Compute the per variable storage requirement and relative offsets

	//  If hourly recording is active then allow space for it
	if (Hourly) dwArraySize += TBA_HOURS_RECORDED * sizeof(DWORD);	//  Space for hourly recording
	dwOffDaily = dwArraySize;										//  Offset within a variable to the Daily array

	dwArraySize += TBA_DAYS_RECORDED * sizeof(DWORD);				//  Space for daily recording
	dwOffWeekly = dwArraySize;										//  Offset within a variable to the Weekly array
	dwArraySize += TBA_WEEKS_RECORDED * sizeof(DWORD);				//  Space for weekly recording
	dwOffMonthly = dwArraySize;										//  Offset within a variable to the Monthly array
	dwArraySize += TBA_MONTHS_RECORDED * sizeof(DWORD);				//  Space for monthly recording
	dwOffYearly = dwArraySize;										//  Offset within a variable to the Yearly array
	dwArraySize += TBA_YEARS_RECORDED * sizeof(DWORD);				//  Space for yearly recording
	dwOffTotal = dwArraySize;										//  Offset within a variable to the Total array
	dwArraySize += sizeof(DWORD);									//  Space for total recording
	dwVarSize = dwArraySize;										//  Size of a variable array collection

	//  Compute the total storage requirement
	dwArraySize = dwArraySize * VarCount;

	//  Allocate and initialise storage for the array
	pAStorage = (UCHAR *) malloc(dwArraySize);
	if (pAStorage != NULL) memset(pAStorage, 0, dwArraySize); 
	return;
}

//  increment
//
//  This function will increment a counter in the array
//
//  Parameters
//
//  1  -  int				-	Ordinal index of the variable
//  2  -  TIMEDATE *		-	Timestamp of the event to be counted
//
//  Returns
//
//  API Use Notes:
//
//
void TimeBasedArray::increment(int iVarIndex, TIMEDATE *tdEvent)
{
	//  Increment is converted to an add 1 function call
	add(iVarIndex, tdEvent, 1);
	return;
}

//  decrement
//
//  This function will decrement a counter in the array
//
//  Parameters
//
//  1  -  int				-	Ordinal index of the variable
//  2  -  TIMEDATE *		-	Timestamp of the event to be counted
//
//  Returns
//
//  API Use Notes:
//
//
void TimeBasedArray::decrement(int iVarIndex, TIMEDATE *tdEvent)
{
	//  Increment is converted to a subtrace 1 function call
	subtract(iVarIndex, tdEvent, 1);
	return;
}

//  add
//
//  This function will add a value to a counter in the array
//
//  Parameters
//
//  1  -  int				-	Ordinal index of the variable
//  2  -  TIMEDATE *		-	Timestamp of the event to be counted
//  3  -  DWORD				-   Value to be added
//
//  Returns
//
//  API Use Notes:
//
//
void TimeBasedArray::add(int iVarIndex, TIMEDATE *tdEvent, DWORD dwValue)
{
	int			iIndexPoints[6];							//  Index Point Array
	UCHAR		*pOurVar;									//  Pointer to our variable in storage
	DWORD		*dwArray;									//  Pointer to the required array

	//  Determine if this event is later than the anchor event time - if so we need to update the anchor
	if (TimeDateCompare(tdEvent, &tdAnchor) > 0)
	{
		adjustAnchor(tdEvent);
	}

	//  Get the array index points for all of the component arrays
	getIndexPoints(tdEvent, iIndexPoints);

	//  Calculate the base address for our variable
	if (iVarIndex < 1) pOurVar = pAStorage;
	else pOurVar = pAStorage + (dwVarSize * (iVarIndex - 1));

	//  Update each of the spectrum elements - if within the time scope
	//  Start with the hourly index
	if (iIndexPoints[TBA_HOURLY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffHourly);
		dwArray[iIndexPoints[TBA_HOURLY_INDEX]] += dwValue;
	}

	//  Next the daily index
	if (iIndexPoints[TBA_DAILY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffDaily);
		dwArray[iIndexPoints[TBA_DAILY_INDEX]] += dwValue;
	}

	//  Next the weekly index
	if (iIndexPoints[TBA_WEEKLY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffWeekly);
		dwArray[iIndexPoints[TBA_WEEKLY_INDEX]] += dwValue;
	}

	//  Next the monthly index
	if (iIndexPoints[TBA_MONTHLY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffMonthly);
		dwArray[iIndexPoints[TBA_MONTHLY_INDEX]] += dwValue;
	}

	//  Next the yearly index
	if (iIndexPoints[TBA_YEARLY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffYearly);
		dwArray[iIndexPoints[TBA_YEARLY_INDEX]] += dwValue;
	}

	//  Finally the totals index
	if (iIndexPoints[TBA_TOTAL_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffTotal);
		dwArray[iIndexPoints[TBA_TOTAL_INDEX]] += dwValue;
	}
	return;
}

//  subtract
//
//  This function will subtract a value from a counter in the array
//
//  Parameters
//
//  1  -  int				-	Ordinal index of the variable
//  2  -  TIMEDATE *		-	Timestamp of the event to be counted
//  3  -  DWORD				-   Value to be subtracted
//
//  Returns
//
//  API Use Notes:
//
//
void TimeBasedArray::subtract(int iVarIndex, TIMEDATE *tdEvent, DWORD dwValue)
{
	int			iIndexPoints[6];							//  Index Point Array
	UCHAR		*pOurVar;									//  Pointer to our variable in storage
	DWORD		*dwArray;									//  Pointer to the required array

	//  Determine if this event is later than the anchor event time - if so we need to update the anchor
	if (TimeDateCompare(tdEvent, &tdAnchor) > 0)
	{
		adjustAnchor(tdEvent);
	}

	//  Get the array index points for all of the component arrays
	getIndexPoints(tdEvent, iIndexPoints);

	//  Calculate the base address for our variable
	if (iVarIndex < 1) pOurVar = pAStorage;
	else pOurVar = pAStorage + (dwVarSize * (iVarIndex - 1));

	//  Update each of the spectrum elements - if within the time scope
	//  Start with the hourly index
	if (iIndexPoints[TBA_HOURLY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffHourly);
		dwArray[iIndexPoints[TBA_HOURLY_INDEX]] -= dwValue;
	}

	//  Next the daily index
	if (iIndexPoints[TBA_DAILY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffDaily);
		dwArray[iIndexPoints[TBA_DAILY_INDEX]] -= dwValue;
	}

	//  Next the weekly index
	if (iIndexPoints[TBA_WEEKLY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffWeekly);
		dwArray[iIndexPoints[TBA_WEEKLY_INDEX]] -= dwValue;
	}

	//  Next the monthly index
	if (iIndexPoints[TBA_MONTHLY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffMonthly);
		dwArray[iIndexPoints[TBA_MONTHLY_INDEX]] -= dwValue;
	}

	//  Next the yearly index
	if (iIndexPoints[TBA_YEARLY_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffYearly);
		dwArray[iIndexPoints[TBA_YEARLY_INDEX]] -= dwValue;
	}

	//  Finally the totals index
	if (iIndexPoints[TBA_TOTAL_INDEX] > -1)
	{
		dwArray = (DWORD *) (pOurVar + dwOffTotal);
		dwArray[iIndexPoints[TBA_TOTAL_INDEX]] -= dwValue;
	}
	return;
}

//  isArrayInitialised
//
//  This function will test to see if the array was initialised or not
//
//  Parameters
//
//  Returns
//
//  BOOL			-	TRUE if the array was initialised, FALSE if not.
//
//  API Use Notes:
//
//
BOOL TimeBasedArray::isArrayInitialised(void)
{
	if (pAStorage == NULL) return FALSE;
	return TRUE;
}

//  marshall
//
//  This function will load the array values from fields on a Notes document
//
//  Parameters
//
//  1  -  char *		- Pointer to a string containing the name of the latest timestamp
//  2  -  NOTEHANDLE	- Handle of the note to load the array from
//
//  Returns
//
//  BOOL			-	TRUE if the array was loaded, FALSE if not.
//
//  API Use Notes:
//
//
BOOL TimeBasedArray::marshall(char *szLatestTDFieldName, NOTEHANDLE hnSNote)
{
	UCHAR			*pOurVar;										//  Pointer to our variable in storage
	DWORD			*dwArray;										//  Pointer to the required array
	int				iVarNum;										//  Variable number
	char			szBaseName[TBA_MAX_BASENAME];					//  Base field name
	char			szArrayName[TBA_MAX_ITEMNAME];					//  Array field name

	//  A few sanity checks before we start the real work
	if (szLatestTDFieldName[0] == '\0') return FALSE;
	if (hnSNote == NULLHANDLE) return FALSE;

	//  Retrieve the anchor timestamp
	if (!NSFItemGetTime(hnSNote, szLatestTDFieldName, &tdAnchor)) return FALSE;

	//  Process each vaiable in turn
	iVarNum = 1;
	while (iVarNum <= VarCount)
	{
		//  Compute the address of the arrays for this variable
		pOurVar = pAStorage + (dwVarSize * (iVarNum - 1));

		//  Get the base name for this variable
		getFieldName(iVarNum, szBaseName);
		
		//  If hourly stats are being colected then retrieve them and load the array
		if (Hourly)
		{
			//  Construct the field name for the Hourly stats
			sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_HOURLY);

			//  Compute the address of the hourly array for this variable
			dwArray = (DWORD *) (pOurVar + dwOffHourly);

			//  Load the hourly array for this variable
			loadArray(dwArray, TBA_HOURS_RECORDED, szArrayName, hnSNote);
		}

		//  Load the Daily stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_DAILY);
		dwArray = (DWORD *) (pOurVar + dwOffDaily);
		loadArray(dwArray, TBA_DAYS_RECORDED, szArrayName, hnSNote);

		//  Load the Weekly stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_WEEKLY);
		dwArray = (DWORD *) (pOurVar + dwOffWeekly);
		loadArray(dwArray, TBA_WEEKS_RECORDED, szArrayName, hnSNote);

		//  Load the Monthly stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_MONTHLY);
		dwArray = (DWORD *) (pOurVar + dwOffMonthly);
		loadArray(dwArray, TBA_MONTHS_RECORDED, szArrayName, hnSNote);

		//  Load the Yearly stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_YEARLY);
		dwArray = (DWORD *) (pOurVar + dwOffYearly);
		loadArray(dwArray, TBA_YEARS_RECORDED, szArrayName, hnSNote);

		//  Load the Total stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_TOTAL);
		dwArray = (DWORD *) (pOurVar + dwOffTotal);
		loadArray(dwArray, 1, szArrayName, hnSNote);

		iVarNum++;
	}
	return TRUE;
}

//  serialize
//
//  This function will store the array values into fields on a Notes document
//
//  Parameters
//
//  1  -  char *		- Pointer to a string containing the name of the latest timestamp
//  2  -  NOTEHANDLE	- Handle of the note to load the array from
//
//  Returns
//
//  BOOL			-	TRUE if the array was loaded, FALSE if not.
//
//  API Use Notes:
//
//
BOOL TimeBasedArray::serialize(char *szLatestTDFieldName, NOTEHANDLE hnSNote)
{
	UCHAR			*pOurVar;										//  Pointer to our variable in storage
	DWORD			*dwArray;										//  Pointer to the required array
	int				iVarNum;										//  Variable number
	char			szBaseName[TBA_MAX_BASENAME];					//  Base field name
	char			szArrayName[TBA_MAX_ITEMNAME];					//  Array field name

	//  A few sanity checks before we start the real work
	if (szLatestTDFieldName[0] == '\0') return FALSE;
	if (hnSNote == NULLHANDLE) return FALSE;

	//  Store the anchor timestamp
	NSFItemSetTime(hnSNote, szLatestTDFieldName, &tdAnchor);

	//  Process each vaiable in turn
	iVarNum = 1;
	while (iVarNum <= VarCount)
	{
		//  Compute the address of the arrays for this variable
		pOurVar = pAStorage + (dwVarSize * (iVarNum - 1));

		//  Get the base name for this variable
		getFieldName(iVarNum, szBaseName);
		
		//  If hourly stats are being colected then store the array
		if (Hourly)
		{
			//  Construct the field name for the Hourly stats
			sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_HOURLY);

			//  Compute the address of the hourly array for this variable
			dwArray = (DWORD *) (pOurVar + dwOffHourly);

			//  Store the hourly array for this variable
			storeArray(dwArray, TBA_HOURS_RECORDED, szArrayName, hnSNote);
		}

		//  Store the Daily stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_DAILY);
		dwArray = (DWORD *) (pOurVar + dwOffDaily);
		storeArray(dwArray, TBA_DAYS_RECORDED, szArrayName, hnSNote);

		//  Store the Weekly stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_WEEKLY);
		dwArray = (DWORD *) (pOurVar + dwOffWeekly);
		storeArray(dwArray, TBA_WEEKS_RECORDED, szArrayName, hnSNote);

		//  Store the Monthly stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_MONTHLY);
		dwArray = (DWORD *) (pOurVar + dwOffMonthly);
		storeArray(dwArray, TBA_MONTHS_RECORDED, szArrayName, hnSNote);

		//  Store the Yearly stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_YEARLY);
		dwArray = (DWORD *) (pOurVar + dwOffYearly);
		storeArray(dwArray, TBA_YEARS_RECORDED, szArrayName, hnSNote);

		//  Store the Total stats
		sprintf_s(szArrayName, TBA_MAX_ITEMNAME, "%s%s", szBaseName, TBA_SUFF_TOTAL);
		dwArray = (DWORD *) (pOurVar + dwOffTotal);
		storeArray(dwArray, 1, szArrayName, hnSNote);

		iVarNum++;
	}
	return TRUE;
}

//  adjustAnchor
//
//  This function will adjust the current setting of the anchor timestamp
//
//  Parameters
//
//  1  -  TIMEDATE *		-	The new anchor timestamp to be set
//
//  Returns
//
//  API Use Notes:
//
//
void TimeBasedArray::adjustAnchor(TIMEDATE *tdNewAnchor)
{
	int			iIndexPoints[6];							//  Index Point Array
	TIMEDATE	tdOldAnchor;								//  Old anchor timestamp
	int			iVarNum;									//  Number of the variable being shuffled
	UCHAR		*pOurVar;									//  Pointer to our variable in storage
	DWORD		*dwArray;									//  Pointer to the required array

	//  Save the existing timedate
	memcpy(&tdOldAnchor, &tdAnchor, sizeof(TIMEDATE));

	//  Set the new anchor
	memcpy(&tdAnchor, tdNewAnchor, sizeof(TIMEDATE));

#ifdef _DEBUG
	printf("DEBUG: Adjusting Anchor :- ");
#endif

	//  Compute the index of the old anchor within the new anchor space
	getIndexPoints(&tdOldAnchor, iIndexPoints);

	//  Shuffle each of the recording arrays for each variable by the needed amount
	iVarNum = 0;
	while (iVarNum < VarCount)
	{
		//  Calculate the address of the storage for this variable
		pOurVar = pAStorage + (dwVarSize * iVarNum);

		//  Shuffle the hourly array
		dwArray = (DWORD *) (pOurVar + dwOffHourly);
		shuffleArray(dwArray, TBA_HOURS_RECORDED, iIndexPoints[TBA_DAILY_INDEX] * 24);

		//  Shuffle the daily array
		dwArray = (DWORD *) (pOurVar + dwOffDaily);
		shuffleArray(dwArray, TBA_DAYS_RECORDED, iIndexPoints[TBA_DAILY_INDEX]);

		//  Shuffle the weekly array
		dwArray = (DWORD *) (pOurVar + dwOffWeekly);
		shuffleArray(dwArray, TBA_WEEKS_RECORDED, iIndexPoints[TBA_WEEKLY_INDEX]);

		//  Shuffle the monthly array
		dwArray = (DWORD *) (pOurVar + dwOffMonthly);
		shuffleArray(dwArray, TBA_MONTHS_RECORDED, iIndexPoints[TBA_MONTHLY_INDEX]);

		//  Shuffle the yearly array
		dwArray = (DWORD *) (pOurVar + dwOffYearly);
		shuffleArray(dwArray, TBA_YEARS_RECORDED, iIndexPoints[TBA_YEARLY_INDEX]);
		
		iVarNum++;
	}
	return;
}

//  getIndexPoints
//
//  This function will build the Index Point Array for the supplied timestamp
//
//  Parameters
//
//  1  -  TIMEDATE *		-	The timestamp to be indexed
//  2  -  int[] *			-   Pointer to the index point array
//
//  Returns
//
//  API Use Notes:
//
//
void TimeBasedArray::getIndexPoints(TIMEDATE *tdEvent, int *pIPA)
{
	TIME			t_Anchor;											//  Anchor timestamp structure
	TIME			t_Event;											//  Event timestamp structure
	TIME			t_WeekAnchor;										//  Weekly Anchor timestamp structure
	TIME			t_DayAnchor;										//  Daily Anchor timestamp structure
	TIME			t_HourAnchor;										//  Hourly Anchor timestamp structure
	int				iRawIndex;											//  Raw computed index value
	LONG			lTimeDelta;											//  Time Delta

	//  Set all of the index points to "Out-Of-Scope" except for Total
	pIPA[TBA_HOURLY_INDEX] = -1;
	pIPA[TBA_DAILY_INDEX] = -1;
	pIPA[TBA_WEEKLY_INDEX] = -1;
	pIPA[TBA_MONTHLY_INDEX] = -1;
	pIPA[TBA_YEARLY_INDEX] = -1;
	pIPA[TBA_TOTAL_INDEX] = 0;

	//  Build the anchor and event TIME structures
	memcpy(&t_Anchor.GM, &tdAnchor, sizeof(TIMEDATE));
	if (TimeGMToLocal(&t_Anchor)) return;
	memcpy(&t_Event.GM, tdEvent, sizeof(TIMEDATE));
	if (TimeGMToLocal(&t_Event)) return;

	//  If the new event was sourced in a different time zone from the anchor the convert it to the anchor zone
	if ((t_Event.zone != t_Anchor.zone) || (t_Event.dst != t_Anchor.dst))
	{
		t_Event.zone = t_Anchor.zone;
		t_Event.dst = t_Anchor.dst;
		TimeGMToLocalZone(&t_Event);
	}

	//  Compute the yearly index point
	iRawIndex = t_Anchor.year - t_Event.year;
	if (iRawIndex < TBA_YEARS_RECORDED) pIPA[TBA_YEARLY_INDEX] = iRawIndex;

	//  Compute the monthly index point
	iRawIndex = ((12 * t_Anchor.year) + t_Anchor.month) - ((12 * t_Event.year) + t_Event.month);
	if (iRawIndex < TBA_MONTHS_RECORDED) pIPA[TBA_MONTHLY_INDEX] = iRawIndex;

	//  Compute the weekly anchor timestamp this is the sunday following the anchor timestamp
	memcpy(&t_WeekAnchor, &t_Anchor, sizeof(TIME));
	while (t_WeekAnchor.weekday != 1)
	{
		TimeDateAdjust(&t_WeekAnchor.GM, 0, 0, 0, 1, 0, 0);
		TimeGMToLocal(&t_WeekAnchor);
	}

	//  Now set the TOD of the anchor to the TOD of the event
	t_WeekAnchor.hour = t_Event.hour;
	t_WeekAnchor.minute = t_Event.minute;
	t_WeekAnchor.second = t_Event.second;
	t_WeekAnchor.hundredth = t_Event.hundredth;
	TimeLocalToGM(&t_WeekAnchor);

	//  Compute the weekly index point
	lTimeDelta = TimeDateDifference(&t_WeekAnchor.GM, tdEvent);
	lTimeDelta = lTimeDelta / 60;												//  Convert to minutes
	lTimeDelta = lTimeDelta / 60;												//  Convert to hours
	lTimeDelta = lTimeDelta / 24;												//  Convert to days
	iRawIndex = (int) lTimeDelta / 7;											//  Convert to weeks
	if (iRawIndex < TBA_WEEKS_RECORDED) pIPA[TBA_WEEKLY_INDEX] = iRawIndex;

	//  Compute the daily index point - set the TOD of the Anchor to TOD of the event
	memcpy(&t_DayAnchor, &t_Anchor, sizeof(TIME));
	t_DayAnchor.hour = t_Event.hour;
	t_DayAnchor.minute = t_Event.minute;
	t_DayAnchor.second = t_Event.second;
	t_DayAnchor.hundredth = t_Event.hundredth;
	TimeLocalToGM(&t_DayAnchor);
	lTimeDelta = TimeDateDifference(&t_DayAnchor.GM, tdEvent);
	lTimeDelta = lTimeDelta / 60;												//  Convert to minutes
	lTimeDelta = lTimeDelta / 60;												//  Convert to hours
	iRawIndex = (int) lTimeDelta / 24;											//  Convert to days
	if (iRawIndex < TBA_DAYS_RECORDED) pIPA[TBA_DAILY_INDEX] = iRawIndex;

	//  Compute the hourly index point
	memcpy(&t_HourAnchor, &t_Anchor, sizeof(TIME));
	t_HourAnchor.hour = 23;
	t_HourAnchor.minute = t_Event.minute;
	t_HourAnchor.second = t_Event.second;
	t_HourAnchor.hundredth = t_Event.hundredth;
	TimeLocalToGM(&t_HourAnchor);
	lTimeDelta = TimeDateDifference(&t_HourAnchor.GM, tdEvent);
	lTimeDelta = lTimeDelta / 60;												//  Convert to minutes
	iRawIndex = (int) lTimeDelta / 60;											//  Convert to hours
	if (iRawIndex < TBA_HOURS_RECORDED) pIPA[TBA_HOURLY_INDEX] = iRawIndex;
#ifdef _DEBUG
	//  Dump the result of the index analysis
	printf("DEBUG: A=%i/%i/%i E=%i/%i/%i (%i:%i) I=%i,%i,%i,%i,%i,%i.\r\n",
			t_Anchor.year,
			t_Anchor.month,
			t_Anchor.day,
			t_Event.year,
			t_Event.month,
			t_Event.day,
			t_Event.hour,
			t_Event.minute,
			pIPA[TBA_HOURLY_INDEX],
			pIPA[TBA_DAILY_INDEX],
			pIPA[TBA_WEEKLY_INDEX],
			pIPA[TBA_MONTHLY_INDEX],
			pIPA[TBA_YEARLY_INDEX],
			pIPA[TBA_TOTAL_INDEX]);
#endif

	return;
}

//  shuffleArray
//
//  This function will shuffle the passed array by the given amount
//
//  Parameters
//
//  1  -  DWORD[] *			-	The array to be shuffled
//  2  -  int				-   The size of the array
//  3  -  int				-	The number of slots by which to shuffle the array
//
//  Returns
//
//  API Use Notes:
//
//
void TimeBasedArray::shuffleArray(DWORD *dwArray, int iSize, int iShift)
{
	int				iIndex;											//  Generic index

	//  If no shuffle is required then return
	if (iShift == 0) return;

	//  If the old anchor is now out of scope of the array then just clear the array
	if (iShift == -1)
	{
		memset(dwArray, 0, iSize * sizeof(DWORD));
		return;
	}

	// Cycle the array
	for (iIndex = iSize - 1; iIndex >= 0; iIndex--)
	{
		if ((iIndex - iShift) < 0) dwArray[iIndex] = 0;
		else dwArray[iIndex] = dwArray[iIndex - iShift];
	}
	return;
}

//  getFieldName
//
//  This function will return a pointer to the base field name for a variable.
//
//  Parameters
//
//  1  -  int				-	Ordinal index of the variable
//  2  -  char *			-   Pointer to the buffer to populate
//
//  Returns
//
//  API Use Notes:
//
//  This is a default implementation and is intended to be replaced in any extending class.
//
void TimeBasedArray::getFieldName(int iVarNum, char *Bfr)
{
	sprintf_s(Bfr, TBA_MAX_BASENAME, "%s%.3i", TBA_DEF_BASENAME, iVarNum);
	return;
}

//  loadArray
//
//  This function will load an individual array from a Notes Item
//
//  Parameters
//
//  1  -  DWORD[] *			-	The array to be shuffled
//  2  -  int				-   The size of the array
//  3  -  char *			-	The item name from which to load the array.
//  4  -  NOTEHANDLE		-	Handle of the note to load the array from
//
//  Returns
//
//  API Use Notes:
//
//  This implementation will extend or truncate the in memory array to the size passed, this allows the code to dynamically adjust the stored
//  arrays according to what the latest code needs. If the passed item is not found in a note then the array is initialised to zeroes, this
//  allows additional variables to be added to a note on demand.
//
void TimeBasedArray::loadArray(DWORD *dwArray, int iSize, char *szItem, NOTEHANDLE hnSNote)
{
	STATUS			stAPIRC;							//  Return codes from Notes API calls
	BLOCKID			bidItem;							//  BlockID for the item
	BLOCKID			bidValue;							//  BlockID for the item value
	WORD			wItemType;							//  Data Type of the item
	DWORD			dwValueSize;						//  Size of the item value
	int				iIndex;								//  Generic Index
	UCHAR			*pValue;							//  Pointer to the item value in memory
	NUMBER			*pnValue;							//  Pointer to an individual numeric item value
	RANGE			*prValue;							//  Pointer to the range 


	//  Any error condition forces the presumption that the item does not (yet) exist. The array is initialised with zeroes.
	for (iIndex = 0; iIndex < iSize; iIndex++)
	{
		dwArray[iIndex] = 0;
	}

	//  Get the information about the requested array item
	stAPIRC = NSFItemInfo(hnSNote, szItem, (WORD) strlen(szItem), &bidItem, &wItemType, &bidValue, &dwValueSize);
	if (ERR(stAPIRC) != NOERROR) return;

	//  Check the type NUMBER_RANGE (or the degenerate case of NUMBER) are permissable - any other type is treated as if the item was not found
	if ((wItemType != TYPE_NUMBER_RANGE) && (wItemType != TYPE_NUMBER)) return;

	//  Lock the Value Block to access the values
	pValue = NULL;
	pValue = OSLockBlock(UCHAR, bidValue);
	if (pValue == NULL) return;
	pValue += sizeof(WORD);								//  Move past the item type definition word

	//  Deal with the degenerate case where the value is not a list
	if (wItemType == TYPE_NUMBER)
	{
		pnValue = (NUMBER *) pValue;
		dwArray[0] = (DWORD) *pnValue;
		//  Free the value block
		OSUnlockBlock(bidValue);
		return;
	}
	
	//  Point to the RANGE structure
	prValue = (RANGE *) pValue;

	//  If there are no entrries in the list then return with the array initialised
	if (prValue->ListEntries == 0)
	{
		//  Free the value block
		OSUnlockBlock(bidValue);
		return;
	}

	//  Point to the start of the list items
	pValue += sizeof(RANGE);
	pnValue = (NUMBER *) pValue;

	//  Iterate over the list of NUMBER entries assigning them into the array
	for (iIndex = 0; iIndex < prValue->ListEntries; iIndex++)
	{
		//  Do not assign the value if the array should now be truncated and we are past the new bounds
		if (iIndex < iSize) dwArray[iIndex] = (DWORD) pnValue[iIndex];
	}

	//  Free the value block
	OSUnlockBlock(bidValue);
	return;
}

//  storeArray
//
//  This function will store an individual array in a Notes Item
//
//  Parameters
//
//  1  -  DWORD[] *			-	The array to be shuffled
//  2  -  int				-   The size of the array
//  3  -  char *			-	The item name from which to load the array.
//  4  -  NOTEHANDLE		-	Handle of the note to store the array in
//
//  Returns
//
//  API Use Notes:
//
//  If an array is passed with a size of 1 then the item is stored as a NUMBER field with the summary flag on. Otherwise the array is stored as
//  a NUMBER_RANGE with the summary flag off.
//
void TimeBasedArray::storeArray(DWORD *dwArray, int iSize, char *szItem, NOTEHANDLE hnSNote)
{
	int				iBufferSize;							//  Size of the buffer needed 
	NUMBER			nSingleValue;							//  Number to be added as a single value
	UCHAR			*pDataBuffer;							//  Data Buffer
	UCHAR			*pValue;								//  Pointer to the value buffer
	NUMBER			*pnValue;								//  Pointer to an individual numeric item value
	RANGE			*prValue;								//  Pointer to the range 
	int				iIndex;									//  Generic index

	//  Determine if the item should be stored as a NUMBER rather than a RANGE (list)
	if (iSize == 1)
	{
		nSingleValue = dwArray[0];							//  Capture the value
		
		//  Remove any existing item with the specified name
		NSFItemDelete(hnSNote, szItem, (WORD) strlen(szItem));

		//  Append the new value
		NSFItemAppend(hnSNote, ITEM_SUMMARY, szItem, (WORD) strlen(szItem), TYPE_NUMBER, &nSingleValue, sizeof(NUMBER));
		return;
	}

	//  Compute the storage needed for the item value data
	iBufferSize = sizeof(RANGE) + (iSize * sizeof(NUMBER));

	//  Allocate storage fot the value data
	pDataBuffer = (UCHAR *) malloc(iBufferSize);

	//  Set the pointers for the RANGE and the start of the NUMBER array
	pValue = pDataBuffer;
	prValue = (RANGE *) pValue;
	pValue += sizeof(RANGE);
	pnValue = (NUMBER *) pValue;

	//  Initialise the range
	prValue->ListEntries = iSize;
	prValue->RangeEntries = 0;

	//  Assign the entries for the number array
	for (iIndex = 0; iIndex < iSize; iIndex++)
	{
		pnValue[iIndex] = dwArray[iIndex];
	}

	//  Remove any existing item with the specified name
	NSFItemDelete(hnSNote, szItem, (WORD) strlen(szItem));

	//  Append the new value
	NSFItemAppend(hnSNote, 0, szItem, (WORD) strlen(szItem), TYPE_NUMBER_RANGE, pDataBuffer, iBufferSize);

	//  Free the value data buffer
	free(pDataBuffer);
	return;
}

//  getArraySize
//
//  This function will return the size of one of the time based arrays
//
//  Parameters
//
//  1  -  int				-   The index of the time based element
//
//  Returns
//
//  int			-	Size of the Array
//
//  API Use Notes:
//
//
int TimeBasedArray::getArraySize(int iTimeIndex)
{
	if (iTimeIndex == TBA_HOURLY_INDEX) return TBA_HOURS_RECORDED;
	if (iTimeIndex == TBA_DAILY_INDEX) return TBA_DAYS_RECORDED;
	if (iTimeIndex == TBA_WEEKLY_INDEX) return TBA_WEEKS_RECORDED;
	if (iTimeIndex == TBA_MONTHLY_INDEX) return TBA_MONTHS_RECORDED;
	if (iTimeIndex == TBA_YEARLY_INDEX) return TBA_YEARS_RECORDED;
	if (iTimeIndex == TBA_TOTAL_INDEX) return 1;
	return -1;										//  Not implemented
}

//  gertCounterValue
//
//  This function will return a counter value from one of the arrays
//
//  Parameters
//
//  1  -  int				-	Ordinal index of the variable
//  2  -  int				-   The index of the time based element
//  3  -  int				-   The index of th element in the array to be returned (zero based)
//
//  Returns
//
//  API Use Notes:
//
//
DWORD TimeBasedArray::getCounterValue(int iVarIndex, int iTimeIndex, int iElement)
{
	UCHAR		*pOurVar;									//  Pointer to our variable in storage
	DWORD		*dwArray;									//  Pointer to the required array
	int			iMaxElements;								//  Max elements in the particular array

	//  Calculate the pointer to storage for the requested variable
	if ((iVarIndex > 0) && (iVarIndex <= VarCount)) pOurVar = pAStorage + (dwVarSize * (iVarIndex - 1));
	else pOurVar = pAStorage;

	//  Calculate the pointer to the appropriate array
	switch(iTimeIndex)
	{
	case TBA_HOURLY_INDEX:
		dwArray = (DWORD *) (pOurVar + dwOffHourly);
		break;
	case TBA_DAILY_INDEX:
		dwArray = (DWORD *) (pOurVar + dwOffDaily);
		break;
	case TBA_WEEKLY_INDEX:
		dwArray = (DWORD *) (pOurVar + dwOffWeekly);
		break;
	case TBA_MONTHLY_INDEX:
		dwArray = (DWORD *) (pOurVar + dwOffMonthly);
		break;
	case TBA_YEARLY_INDEX:
		dwArray = (DWORD *) (pOurVar + dwOffYearly);
		break;
	case TBA_TOTAL_INDEX:
		dwArray = (DWORD *) (pOurVar + dwOffTotal);
		break;
	default:
		dwArray = (DWORD *) (pOurVar + dwOffTotal);
	}

	//  Get the max elements in the particular array
	iMaxElements = getArraySize(iTimeIndex);
	if ((iElement < 0) || (iElement >= iMaxElements)) return dwArray[0];
	return dwArray[iElement];
}


