//  NEResolver
//
//  This class provides Named Entity resolution services providing the ability to retrieve a particular identity given an
//  arbitrary identity.
//
//  NOTE: This class uses the Windows LDAP Client API rather than the Domino LDAP Client API, this is because the LDAP sessions have to
//        be shared with authenticater classes that use special AD features that are not accessible through the Domino LDAP API.
//
//  Message Identifier:      DXR
//  Message Range:           8001 - 8499
//  Trace area assignment:   TRACE_NER (30)
//  DX Version:				 3.11.1
//  Version:                 3.11.1  Build: 99
//
//  Version History:
//
//  3.11.1		22/10/2011		VS 2011 readiness
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2011
//

//  Application Includes
#include "NEResolver.h"

//  Default Constructor
NEResolver::NEResolver()
{
}

//  Normal Constructor
NEResolver::NEResolver(ExecEnvironment *xeParent, int iThreadID)
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Set the DX Runtime
	xeLocal = xeParent;
	iParentThread = iThreadID;

	//  Clear the environment definition settings
	hlsSession = NULL;
	szDomRefServer[0] = '\0';
	szBaseDN[0] = '\0';
	szMatchDN[0] = '\0';
	ShowDuplicates = FALSE;

	//  Build a new Named Entity Cache
	necCurrent = new NECache(xeParent, CACHE_SIZE, CACHE_STRATEGY_MRU, iThreadID);

	//  Trace construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8001T, this);
	xeLocal->LogTrace(TRACE_NER, szMsg, iParentThread);
}

//  Destructor
NEResolver::~NEResolver()
{
	char		szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Eliminate the Named Entity Cache
	if (necCurrent != NULL)
	{
		necCurrent->drain(iParentThread);
		delete necCurrent;
	}

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8002T, this);
	xeLocal->LogTrace(TRACE_NER, szMsg, iParentThread);
}

//  getAffinity
//
//  This function will return the affinity for an arbitrarily named entity
//
//  Parameters
//
//  1  -  char *			-  Pointer to the arbitrary ID
//  2  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  int			-	See NE_ID_XXXXX for possible values and meanings
//
//  API Use Notes:
//
//
int NEResolver::getAffinity(char *szAID, int iThreadID)
{
	NamedEntity		*neCurrent;											//  Named Entity Object
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer

	//  Sanity check the input
	if (szAID == NULL) return NE_ID_UNRESOLVEABLE;
	if (szAID[0] == '\0') return NE_ID_UNRESOLVEABLE;
	if (strlen(szAID) > MAXUSERNAME) return NE_ID_UNRESOLVEABLE;

	//  Trace the fact that we are starting the resolver process
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8005T, "Affinity", szAID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  Get the current or new Named Entity object from the cache
	neCurrent = (NamedEntity *) necCurrent->getCachedObject(szAID, iThreadID);

	//  If this is a new entry then we must resolve the affinity of the arbitrary ID
	if (neCurrent->AIDAff == NE_ID_UNRESOLVED)
	{
		resolveAffinity(neCurrent, iThreadID);
	}

	//  Dump the Named Entity
	xeLocal->LogTrace(TRACE_NER, MSG_DXR8011T, iThreadID);
	dumpNamedEntity(neCurrent, iThreadID);

	return neCurrent->AIDAff;
}

//  getDomDN
//
//  This function will return the Domino Distinguished Name for an arbitrarily named entity
//
//  Parameters
//
//  1  -  char *			-  Pointer to the arbitrary ID
//	2  -  char *			-  Pointer to the buffer to hold the returned name
//  3  -  int				-  Size of the buffer used to hold the returned name
//  4  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  int			-	See NERET_XXXX  definitions for possible values and meanings
//
//  API Use Notes:
//
//
int NEResolver::getDomDN(char *szAID, char *szDomDN, int iDomDNSize, int iThreadID)
{
	STATUS			stAPIRC;											//  Return code from Notes API calls
	DHANDLE			hLookupResults;										//  Handle to the lookup results
	NamedEntity		*neCurrent;											//  Named Entity Object
	WORD			wMatches;											//  Number of matches
	UCHAR			*pLookupResults;									//  Pointer to lookup results
	UCHAR			*pResultName;										//  Result name entry
	UCHAR			*pResultMatch;										//  Result match entry
	int				iIndex;												//  Generic index
	char			szLookupKey[MAXUSERNAME + 1];						//  Lookup key to use
	char			szItemList[2 * MAX_DESC];							//  Item Name List
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer
	
	//  Sanity check the input
	if ((szAID == NULL) || (szDomDN == NULL)) return NERET_INVPARM;
	if (szAID[0] == '\0') return NERET_INVPARM;
	if (strlen(szAID) > MAXUSERNAME) return NERET_INVPARM;
	if (iDomDNSize < MAXUSERNAME) return NERET_INVPARM;

	szDomDN[0] = '\0';

	//  Get the current or new Named Entity object from the cache
	neCurrent = (NamedEntity *) necCurrent->getCachedObject(szAID, iThreadID);

	//  Trace the fact that we are starting the resolver process
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8005T, "DomDN", szAID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);


	//  If this is a new entry then we must start by resolving the affinity of the arbitrary ID
	if (neCurrent->AIDAff == NE_ID_UNRESOLVED)
	{
		resolveAffinity(neCurrent, iThreadID);
		//  Dump the Named Entity
		xeLocal->LogTrace(TRACE_NER, MSG_DXR8011T, iThreadID);
		dumpNamedEntity(neCurrent, iThreadID);

		if (neCurrent->AIDAff == NE_ID_UNRESOLVEABLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR8009T, neCurrent->AID);
			xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
			neCurrent->AIDAff = NE_ID_UNRESOLVEABLE;
			return NERET_UNRESOLVED;
		}
	}

	//  If the Domino DN is already resolved then return it
	if (neCurrent->DDNQ == NE_DN_FULLYQUALIFIED)
	{
		strcpy_s(szDomDN, MAXUSERNAME, neCurrent->DDN);
		return NERET_RESOLVED;
	}

	//  If we already have a fully qualifeid e-Mail address then use that for the search
	if (neCurrent->EMDNQ == NE_DN_FULLYQUALIFIED)
	{
		strcpy_s(szLookupKey, MAXUSERNAME, neCurrent->EMAIL);
	}
	else
	{
		//  If the Domino DN is already marked as not found then return that signal again
		if (neCurrent->DDNQ == NE_DN_NOTFOUND) return NERET_NOTFOUND;

		//  Check that we have a key to attempt the resolution
		if ((neCurrent->DDNQ == NE_DN_EMPTY) && (neCurrent->EMDNQ == NE_DN_EMPTY)) return NERET_UNRESOLVED;

		//  Setup the search key
		if (neCurrent->DDNQ == NE_DN_EMPTY) strcpy_s(szLookupKey, MAXUSERNAME, neCurrent->EMAIL);
		else strcpy_s(szLookupKey, MAXUSERNAME, neCurrent->DDN);
	}

	//  Build the item list
	strcpy_s(szItemList, (2 * MAX_DESC), MAIL_FULLNAME_ITEM);
	strcpy_s(szItemList + (strlen(szItemList) + 1), (2 * MAX_DESC) -  (strlen(szItemList) + 1), MAIL_INTERNETADDRESS_ITEM);

	//  Lookup the name in the Domino Reference Directory
	stAPIRC = NAMELookup(szDomRefServer,
							0,
							1,
							LOCAL_MAIL_NAMEFIELD_LOOKUP,
							1,
							szLookupKey,
							2,
							szItemList,
							&hLookupResults);

	if (ERR(stAPIRC) != NOERROR)
	{
		//  Fault encountered doing the lookup
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8013E, szLookupKey);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		neCurrent->DDNQ = NE_DN_NOTFOUND;
		return NERET_UNRESOLVED;
	}

	//  Process the returned result
	pLookupResults = (UCHAR *) OSLockObject(hLookupResults);

	//  Position to the results for the first (only) name in the query
	pResultName = NULL;
	pResultName = (UCHAR *) NAMELocateNextName(pLookupResults, pResultName, &wMatches);

	//  If nothing was found then mark the entry as not found, cleanup and exit
	if (wMatches == 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8014I, szLookupKey);
		xeLocal->LogVerbose(szMsg, iThreadID);
		neCurrent->DDNQ = NE_DN_NOTFOUND;
		//  Unlock and free the search results
		OSUnlockObject(hLookupResults);	
		OSMemFree(hLookupResults);
		return NERET_UNRESOLVED;
	}

	//  If multiple matches were returned then that is just as bad, treat it as not found
	if (wMatches > 1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8015I, szLookupKey, wMatches);
		xeLocal->LogVerbose(szMsg, iThreadID);
		neCurrent->DDNQ = NE_DN_NOTFOUND;
		//  Unlock and free the search results
		OSUnlockObject(hLookupResults);	
		OSMemFree(hLookupResults);
		return NERET_UNRESOLVED;
	}

	//  Position to the first (only) match 
	pResultMatch = NULL;
	pResultMatch = (UCHAR *) NAMELocateNextMatch(pLookupResults, pResultName, pResultMatch);

	//  Process the FullName Item - we want the first entry in the list
	iIndex = 0;
	if (NAMEGetTextItem(pResultMatch, 0, iIndex, szLookupKey, MAXUSERNAME) == NOERROR)
	{
		strcpy_s(neCurrent->DDN, MAXUSERNAME, szLookupKey);
		neCurrent->DDNQ = NE_DN_FULLYQUALIFIED;
	}

	//  If we have an e-Mail address returned then set it and mark as qualified
	iIndex = 0;
	if (NAMEGetTextItem(pResultMatch, 1, iIndex, szLookupKey, MAXUSERNAME) == NOERROR)
	{
		strcpy_s(neCurrent->EMAIL, MAXUSERNAME, szLookupKey);
		neCurrent->EMDNQ = NE_DN_FULLYQUALIFIED;
	}

	//  Unlock and free the search results
	OSUnlockObject(hLookupResults);	
	OSMemFree(hLookupResults);

	//  Trace the result
	xeLocal->LogTrace(TRACE_NER, MSG_DXR8016T, iThreadID);
	dumpNamedEntity(neCurrent, iThreadID);
	strcpy_s(szDomDN, MAXUSERNAME, neCurrent->DDN);
	return NERET_RESOLVED;
}

//  getDomAN
//
//  This function will return the Domino Abbreviated Name for an arbitrarily named entity
//
//  Parameters
//
//  1  -  char *			-  Pointer to the arbitrary ID
//	2  -  char *			-  Pointer to the buffer to hold the returned name
//  3  -  int				-  Size of the buffer used to hold the returned name
//  4  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  int			-	See NERET_XXXX  definitions for possible values and meanings
//
//  API Use Notes:
//
//
int NEResolver::getDomAN(char *szAID, char *szDomAN, int iDomANSize, int iThreadID)
{
	STATUS			stAPIRC;											//  Return code from Notes API Calls
	int				iResolveRC;											//  Return code from internal resolver calls
	WORD			wRetLen;											//  Returned Length
	char			szDomDN[MAXUSERNAME + 1];							//  Domino Distinguished Name
	char			szAbbreviated[MAXUSERNAME + 1];						//  Domino Abbreviated Name
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer
	
	//  Sanity check the input
	if ((szAID == NULL) || (szDomAN == NULL)) return NERET_INVPARM;
	if (szAID[0] == '\0') return NERET_INVPARM;
	if (strlen(szAID) > MAXUSERNAME) return NERET_INVPARM;
	if (iDomANSize < MAXUSERNAME) return NERET_INVPARM;

	szDomAN[0] = '\0';
	
	//  Trace the fact that we are starting a resolve Domino AN
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8005T, "DomAN", szAID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  Start off by obtaining the Domino Distinguished Name
	iResolveRC = getDomDN(szAID, szDomDN, MAXUSERNAME, iThreadID);
	if (iResolveRC != NERET_RESOLVED) return iResolveRC;

	//  Abbreviate the name
	stAPIRC = DNAbbreviate(0L, NULL, szDomDN, szAbbreviated, MAXUSERNAME, &wRetLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8029E, szDomDN);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC,szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NERET_UNRESOLVED;
	}

	szAbbreviated[wRetLen] = '\0';

	//  Check that we actually got something
	if (strlen(szAbbreviated) == 0) return NERET_UNRESOLVED;

	//  Copy the abbreviated name to the caller
	strcpy_s(szDomAN, MAXUSERNAME, szAbbreviated);
	return NERET_RESOLVED;
}

//  getDomCN
//
//  This function will return the Domino Common Name for an arbitrarily named entity
//
//  Parameters
//
//  1  -  char *			-  Pointer to the arbitrary ID
//	2  -  char *			-  Pointer to the buffer to hold the returned name
//  3  -  int				-  Size of the buffer used to hold the returned name
//  4  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  int			-	See NERET_XXXX  definitions for possible values and meanings
//
//  API Use Notes:
//
//
int NEResolver::getDomCN(char *szAID, char *szDomCN, int iDomCNSize, int iThreadID)
{
	STATUS			stAPIRC;											//  Return code from Notes API Calls
	int				iResolveRC;											//  Return code from internal resolver calls
	char			szDomDN[MAXUSERNAME + 1];							//  Domino Distinguished Name
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer
	DN_COMPONENTS	dncParsed;											//  Components structure
	
	//  Sanity check the input
	if ((szAID == NULL) || (szDomCN == NULL)) return NERET_INVPARM;
	if (szAID[0] == '\0') return NERET_INVPARM;
	if (strlen(szAID) > MAXUSERNAME) return NERET_INVPARM;
	if (iDomCNSize < MAXUSERNAME) return NERET_INVPARM;

	szDomCN[0] = '\0';
	
	//  Trace the fact that we are starting a resolve Domino AN
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8005T, "DomCN", szAID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  Start off by obtaining the Domino Distinguished Name
	iResolveRC = getDomDN(szAID, szDomDN, MAXUSERNAME, iThreadID);
	if (iResolveRC != NERET_RESOLVED) return iResolveRC;

	//  Extract the Common Name from the Distinguished Name
	stAPIRC = DNParse(0L, NULL, szDomDN, &dncParsed, sizeof(DN_COMPONENTS));
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8028E, szDomDN);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC,szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NERET_UNRESOLVED;
	}

	//  Make sure that we actually get a CN
	if (dncParsed.CNLength == 0) return NERET_UNRESOLVED;

	//  Copy the common name to the caller
	strncpy_s(szDomCN, MAXUSERNAME, dncParsed.CN, dncParsed.CNLength);
	szDomCN[dncParsed.CNLength] = '\0';
	return NERET_RESOLVED;
}

//  getADDN
//
//  This function will return the Active Directory Distinguished Name for an arbitrarily named entity
//
//  Parameters
//
//  1  -  char *			-  Pointer to the arbitrary ID
//	2  -  char *			-  Pointer to the buffer to hold the returned name
//  3  -  int				-  Size of the buffer used to hold the returned name
//  4  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  int			-	See NERET_XXXX  definitions for possible values and meanings
//
//  API Use Notes:
//
//
int NEResolver::getADDN(char *szAID, char *szADDN, int iADDNSize, int iThreadID)
{
	NamedEntity		*neCurrent;											//  Named Entity Object
	ULONG			ulLDAPErr;											//	LDAP Error Code
	char			*pLES;												//  LDAP Error String
	LDAPMessage		*pSearchResults;									//  Handle of search results
	LDAPMessage		*pMatch;											//  Matched Entry
	BerElement		*pbeAttr;											//  BER Element
	ULONG			ulMatches;											//  Number of Matches
	char			*pRetAttr;											//  Returned Attribute
	char			**pRetValue;										//  Returned Attribute Value
	char			*pDN;												//  Returned Distinguished Name
	char			*pReqAttrs[5];										//  Array of attributes to be returned
	char			szAttrName[MAX_DESC + 1];							//  Attribute Name
	char			szQuery[MAX_MSG + 1];								//  Query String
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer
	
	//  Sanity check the input
	if ((szAID == NULL) || (szADDN == NULL)) return NERET_INVPARM;
	if (szAID[0] == '\0') return NERET_INVPARM;
	if (strlen(szAID) > MAXUSERNAME) return NERET_INVPARM;
	if (iADDNSize < MAXUSERNAME) return NERET_INVPARM;

	szADDN[0] = '\0';

	//  Get the current or new Named Entity object from the cache
	neCurrent = (NamedEntity *) necCurrent->getCachedObject(szAID, iThreadID);

	//  Trace the fact that we are starting the resolver process
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8005T, "ADDN", szAID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  If this is a new entry then we must start by resolving the affinity of the arbitrary ID
	if (neCurrent->AIDAff == NE_ID_UNRESOLVED)
	{
		resolveAffinity(neCurrent, iThreadID);
		//  Dump the Named Entity
		xeLocal->LogTrace(TRACE_NER, MSG_DXR8011T, iThreadID);
		dumpNamedEntity(neCurrent, iThreadID);

		if (neCurrent->AIDAff == NE_ID_UNRESOLVEABLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR8009T, neCurrent->AID);
			xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
			neCurrent->AIDAff = NE_ID_UNRESOLVEABLE;
			return NERET_UNRESOLVED;
		}
	}

	//  If the AD DN is already resolved then return it
	if (neCurrent->ADDNQ == NE_DN_FULLYQUALIFIED)
	{
		strcpy_s(szADDN, MAXUSERNAME, neCurrent->ADDN);
		return NERET_RESOLVED;
	}

	//  If the AD DN is already marked as not found then return that signal again
	if (neCurrent->ADDNQ == NE_DN_NOTFOUND) return NERET_NOTFOUND;

	//  Check that we have a key to attempt the resolution
	if ((neCurrent->ADDNQ == NE_DN_EMPTY) && (neCurrent->EMDNQ == NE_DN_EMPTY)) return NERET_UNRESOLVED;

	//  If we do not have an LDAP session then bale out
	if (hlsSession == NULL) return NERET_UNRESOLVED;

	//  Build the query string according to the information that is availble
	//  If we have a CN then this could also possible be a UID
	if (neCurrent->ADDNQ == NE_DN_EMPTY) sprintf_s(szQuery, MAX_MSG, LDAP_QUERY_EMAIL, neCurrent->EMAIL);
	else 
	{
		if (neCurrent->EMDNQ == NE_DN_EMPTY) sprintf_s(szQuery, MAX_MSG, LDAP_QUERY_CN, neCurrent->ADDN, neCurrent->ADDN);
		else sprintf_s(szQuery, MAX_MSG, LDAP_QUERY_EITHER, neCurrent->ADDN, neCurrent->ADDN, neCurrent->EMAIL);
	}

	//  Setup the attributes to be returned
	pReqAttrs[0] = LDAP_ATTR_1;
	pReqAttrs[1] = LDAP_ATTR_2;
	pReqAttrs[2] = LDAP_ATTR_3;
	pReqAttrs[3] = LDAP_ATTR_4;
	pReqAttrs[4] = NULL;

	sprintf_s(szMsg, MAX_MSG, MSG_DXR8024T, szQuery);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8025T, szBaseDN);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  Issue the query
	pSearchResults = NULL;
	ulLDAPErr = ldap_search_s(hlsSession, szBaseDN, LDAP_SCOPE_SUBTREE, szQuery, pReqAttrs, 0, &pSearchResults); 
	if (ulLDAPErr != LDAP_SUCCESS)
	{
		pLES = ldap_err2string(ulLDAPErr);
		if (pLES == NULL) pLES = NO_LDAP_MSG;
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8017E, "search", ulLDAPErr, pLES);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8024T, szQuery);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8025T, szBaseDN);
		xeLocal->LogMessage(szMsg, iThreadID);
		if (pSearchResults != NULL) ldap_msgfree(pSearchResults);
		neCurrent->ADDNQ = NE_DN_NOTFOUND;
		return NERET_UNRESOLVED;
	}

	//  Get the match count filtered against the specified DN
	ulMatches = getFilteredMatchCount(pSearchResults, iThreadID);
	if (ulMatches == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8018E, szQuery);
		xeLocal->LogMessage(szMsg, iThreadID);
		ulLDAPErr = LdapGetLastError();
		pLES = ldap_err2string(ulLDAPErr);
		if (pLES == NULL) pLES = NO_LDAP_MSG;
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8017E, "countmatches", ulLDAPErr, pLES);
		xeLocal->LogMessage(szMsg, iThreadID);
		if (pSearchResults != NULL) ldap_msgfree(pSearchResults);
		neCurrent->ADDNQ = NE_DN_NOTFOUND;
		return NERET_UNRESOLVED;
	}

	//  If we found not matches then set the entry as not found and return
	if (ulMatches == 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8019T, szQuery);
		xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
		if (pSearchResults != NULL) ldap_msgfree(pSearchResults);
		neCurrent->ADDNQ = NE_DN_NOTFOUND;
		return NERET_UNRESOLVED;
	}

	//  If we found multiple matches then treat it as not found
	if (ulMatches > 1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8020T, szQuery);
		xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
		if (pSearchResults != NULL) ldap_msgfree(pSearchResults);
		neCurrent->ADDNQ = NE_DN_NOTFOUND;
		return NERET_UNRESOLVED;
	}

	//  Extract the returned attributes and set as appropriate
	pMatch = NULL;
	pMatch = getFirstFilteredMatch(pSearchResults, iThreadID);
	if (pMatch == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8021E, szQuery);
		xeLocal->LogMessage(szMsg, iThreadID);
		if (pSearchResults != NULL) ldap_msgfree(pSearchResults);
		neCurrent->ADDNQ = NE_DN_NOTFOUND;
		return NERET_UNRESOLVED;
	}

	//  Get the matched DN and set it in the Named Entity
	pDN = NULL;
	pDN = ldap_get_dn(hlsSession, pMatch);
	if (pDN == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8022T, "DN");
		xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
	}
	else
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8023T, "DN", pDN);
		xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
		strcpy_s(neCurrent->ADDN, MAXUSERNAME, pDN);
		neCurrent->ADDNQ = NE_DN_FULLYQUALIFIED;
		ldap_memfree(pDN);
	}

	//  Loop through extracting each of the returned values and processing as appropriate
	pRetAttr = NULL;
	pRetAttr = ldap_first_attribute(hlsSession, pMatch, &pbeAttr);
	while (pRetAttr != NULL)
	{
		strcpy_s(szAttrName, MAX_DESC, pRetAttr);
		//  Get the pointer to the first value returned for this attribute
		pRetValue = NULL;
		pRetValue = ldap_get_values(hlsSession, pMatch, pRetAttr);
		if (pRetValue == NULL)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR8022T, pRetAttr);
			xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
		}
		else
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR8023T, pRetAttr, *pRetValue);
			xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
			
			//  If this was the EMail then capture the value as appropriate
			if (strcmpi(pRetAttr, LDAP_ATTR_3) == 0)
			{
				if (neCurrent->EMDNQ == NE_DN_EMPTY)
				{
					strcpy_s(neCurrent->EMAIL, MAXUSERNAME, *pRetValue);
					neCurrent->EMDNQ = NE_DN_FULLYQUALIFIED;
				}
			}

			//  If this was the UID then capture the value
			if (strcmpi(pRetAttr, LDAP_ATTR_2) == 0)
			{
				strcpy_s(neCurrent->ADUID, MAXUSERNAME, *pRetValue);
			}

			//  If this was the Login then capture the value
			if (strcmpi(pRetAttr, LDAP_ATTR_4) == 0)
			{
				strcpy_s(neCurrent->ADLogin, MAXUSERNAME, *pRetValue);
			}

			ldap_value_free(pRetValue);
		}
		//  Free up the retrieved attribute and get the next one
		ldap_memfree(pRetAttr);
		pRetAttr = NULL;		
		pRetAttr = ldap_next_attribute(hlsSession, pMatch, pbeAttr);
	}

	//  Free the complete message
	if (pSearchResults != NULL) ldap_msgfree(pSearchResults);

	//  Trace the result
	xeLocal->LogTrace(TRACE_NER, MSG_DXR8027T, iThreadID);
	dumpNamedEntity(neCurrent, iThreadID);

	strcpy_s(szADDN, MAXUSERNAME, neCurrent->ADDN);
	return NERET_RESOLVED;
}

//  getADCN
//
//  This function will return the Active Directory Common Name for an arbitrarily named entity
//
//  Parameters
//
//  1  -  char *			-  Pointer to the arbitrary ID
//	2  -  char *			-  Pointer to the buffer to hold the returned name
//  3  -  int				-  Size of the buffer used to hold the returned name
//  4  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  int			-	See NERET_XXXX  definitions for possible values and meanings
//
//  API Use Notes:
//
//
int NEResolver::getADCN(char *szAID, char *szADCN, int iADCNSize, int iThreadID)
{
	int				iResolveRC;											//  Return code from internal resolver calls;
	char			*pSource;											//  Pointer into source string
	char			*pTarget;											//  Pointer into target string
	char			szADDN[MAXUSERNAME + 1];							//  AD Distinguished Name
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer

	//  Sanity check the input
	if ((szAID == NULL) || (szADCN == NULL)) return NERET_INVPARM;
	if (szAID[0] == '\0') return NERET_INVPARM;
	if (strlen(szAID) > MAXUSERNAME) return NERET_INVPARM;
	if (iADCNSize < MAXUSERNAME) return NERET_INVPARM;

	szADCN[0] = '\0';

	//  Trace the fact that we are starting the resolver process
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8005T, "ADCN", szAID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  Start off by obtaining the AD Distinguished Name
	iResolveRC = getDomDN(szAID, szADDN, MAXUSERNAME, iThreadID);
	if (iResolveRC != NERET_RESOLVED) return iResolveRC;

	//  Check that we actually got something back
	if (strlen(szADDN) == 0) return NERET_UNRESOLVED;

	//  Pass back the Common Name
	pSource = szADDN + 3;
	pTarget = szADCN;
	while ((*pSource != ',') && (*pSource != '\0'))
	{
		*pTarget = *pSource;
		pTarget++;
		pSource++;
	}
	*pTarget = '\0';
	return NERET_RESOLVED;
}

//  getADUID
//
//  This function will return the Active Directory UID for an arbitrarily named entity
//
//  Parameters
//
//  1  -  char *			-  Pointer to the arbitrary ID
//	2  -  char *			-  Pointer to the buffer to hold the returned name
//  3  -  int				-  Size of the buffer used to hold the returned name
//  4  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  int			-	See NERET_XXXX  definitions for possible values and meanings
//
//  API Use Notes:
//
//
int NEResolver::getADUID(char *szAID, char *szADUID, int iADUIDSize, int iThreadID)
{
	NamedEntity		*neCurrent;											//  Resolved Named Entity
	int				iResolveRC;											//  Return code from internal resolver calls;
	char			szADDN[MAXUSERNAME + 1];							//  AD Distinguished Name
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer

	//  Sanity check the input
	if ((szAID == NULL) || (szADUID == NULL)) return NERET_INVPARM;
	if (szAID[0] == '\0') return NERET_INVPARM;
	if (strlen(szAID) > MAXUSERNAME) return NERET_INVPARM;
	if (iADUIDSize < MAXUSERNAME) return NERET_INVPARM;

	szADUID[0] = '\0';

	//  Trace the fact that we are starting the resolver process
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8005T, "ADUID", szAID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  Start off by obtaining the AD Distinguished Name
	iResolveRC = getDomDN(szAID, szADDN, MAXUSERNAME, iThreadID);
	if (iResolveRC != NERET_RESOLVED) return iResolveRC;

	//  Locate the cached Named Entity
	neCurrent = (NamedEntity *) necCurrent->getCachedObject(szAID, iThreadID);

	//  Check that the AD credentials were resolved
	if (neCurrent->ADDNQ != NE_DN_FULLYQUALIFIED) return NERET_UNRESOLVED;

	//  Check that a UID was located
	if (strlen(neCurrent->ADUID) == 0) return NERET_UNRESOLVED;

	//  Pass back the UID
	strcpy_s(szADUID, MAXUSERNAME, neCurrent->ADUID);
	return NERET_RESOLVED;
}

//  getADLoginUID
//
//  This function will return the Active Directory Login for an arbitrarily named entity
//
//  Parameters
//
//  1  -  char *			-  Pointer to the arbitrary ID
//	2  -  char *			-  Pointer to the buffer to hold the returned name
//  3  -  int				-  Size of the buffer used to hold the returned name
//  4  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  int			-	See NERET_XXXX  definitions for possible values and meanings
//
//  API Use Notes:
//
//
int NEResolver::getADLogin(char *szAID, char *szADLogin, int iADLoginSize, int iThreadID)
{
	NamedEntity		*neCurrent;											//  Resolved Named Entity
	int				iResolveRC;											//  Return code from internal resolver calls;
	char			szADDN[MAXUSERNAME + 1];							//  AD Distinguished Name
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer

	//  Sanity check the input
	if ((szAID == NULL) || (szADLogin == NULL)) return NERET_INVPARM;
	if (szAID[0] == '\0') return NERET_INVPARM;
	if (strlen(szAID) > MAXUSERNAME) return NERET_INVPARM;
	if (iADLoginSize < MAXUSERNAME) return NERET_INVPARM;

	szADLogin[0] = '\0';

	//  Trace the fact that we are starting the resolver process
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8005T, "ADLogin", szAID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  Start off by obtaining the AD Distinguished Name
	iResolveRC = getDomDN(szAID, szADDN, MAXUSERNAME, iThreadID);
	if (iResolveRC != NERET_RESOLVED) return iResolveRC;

	//  Locate the cached Named Entity
	neCurrent = (NamedEntity *) necCurrent->getCachedObject(szAID, iThreadID);

	//  Check that the AD credentials were resolved
	if (neCurrent->ADDNQ != NE_DN_FULLYQUALIFIED) return NERET_UNRESOLVED;

	//  Check that a UID was located
	if (strlen(neCurrent->ADLogin) == 0) return NERET_UNRESOLVED;

	//  Pass back the UID
	strcpy_s(szADLogin, MAXUSERNAME, neCurrent->ADLogin);
	return NERET_RESOLVED;
}

//  getEmailAddress
//
//  This function will return the SMTP e-Mail address for an arbitrarily named entity
//
//  Parameters
//
//  1  -  char *			-  Pointer to the arbitrary ID
//	2  -  char *			-  Pointer to the buffer to hold the returned address
//  3  -  int				-  Size of the buffer used to hold the returned name
//  4  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  int			-	See NERET_XXXX  definitions for possible values and meanings
//
//  API Use Notes:
//
//
int NEResolver::getEmailAddress(char *szAID, char *szEMAIL, int iEMSize, int iThreadID)
{
	NamedEntity		*neCurrent;											//  Named Entity Object
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer
	
	//  Sanity check the input
	if ((szAID == NULL) || (szEMAIL == NULL)) return NERET_INVPARM;
	if (szAID[0] == '\0') return NERET_INVPARM;
	if (strlen(szAID) > MAXUSERNAME) return NERET_INVPARM;
	if (iEMSize < MAXUSERNAME) return NERET_INVPARM;

	szEMAIL[0] = '\0';

	//  Get the current or new Named Entity object from the cache
	neCurrent = (NamedEntity *) necCurrent->getCachedObject(szAID, iThreadID);

	//  Trace the fact that we are starting the resolver process
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8005T, "EMAIL", szAID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  If this is a new entry then we must start by resolving the affinity of the arbitrary ID
	if (neCurrent->AIDAff == NE_ID_UNRESOLVED)
	{
		resolveAffinity(neCurrent, iThreadID);
		//  Dump the Named Entity
		xeLocal->LogTrace(TRACE_NER, MSG_DXR8011T, iThreadID);
		dumpNamedEntity(neCurrent, iThreadID);

		if (neCurrent->AIDAff == NE_ID_UNRESOLVEABLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR8009T, neCurrent->AID);
			xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
			neCurrent->AIDAff = NE_ID_UNRESOLVEABLE;
			return NERET_UNRESOLVED;
		}

		//  Now attempt to resolve the attributes from the appropriate domain
		if (neCurrent->AIDAff & (NE_ID_DOMINO | NE_ID_EMAIL))
		{
			if (getDomDN(szAID, szMsg, MAX_MSG, iThreadID) == NERET_RESOLVED)
			{
				getADDN(szAID, szMsg, MAX_MSG, iThreadID);
			}
		}
		else
		{
			if (getADDN(szAID, szMsg, MAX_MSG, iThreadID) == NERET_RESOLVED)
			{
				getDomDN(szAID, szMsg, MAX_MSG, iThreadID);
			}
		}
	}

	//  Check if we now have a fully qualified e-Mail address, if so return it
	if (neCurrent->EMDNQ & NE_DN_FULLYQUALIFIED)
	{
		strcpy_s(szEMAIL, MAXUSERNAME, neCurrent->EMAIL);
		return NERET_RESOLVED;
	}

	return NERET_UNRESOLVED;
}

//  resolveAffinity
//
//  This function will attempt to resolve the affinity of the arbitrary ID in the Named Entity
//
//  Parameters
//
//  1  -  NamedEntity *		-  Pointer to NamedEntity
//  2  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//
//  API Use Notes:
//
//
void NEResolver::resolveAffinity(NamedEntity *neResolve, int iThreadID)
{
	char			*pSrc;												//  Pointer to source character
	char			*pTarg;												//  Pointer to target character
	char			cQuoteChar;											//  Character used to close a quoted string
	char			szSPattern[MAXUSERNAME + 1];						//  Buffer for the string pattern
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer
	
	//  Trace starting to resolve the affinity
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8008T, neResolve->AID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	// Default the result to unresolvable
	neResolve->AIDAff = NE_ID_UNRESOLVEABLE;

	// Get the string pattern of the arbitrary entity
	getASP(neResolve->AID, szSPattern, MAXUSERNAME);
	//  Trace entity pattern established
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8010T, szSPattern);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);

	//  If the pattern matched was just a quoted string then redrive the pattern eveluation with the strings stripped
	if (strcmp(szSPattern, "q") == 0)
	{
		pSrc = neResolve->AID;
		pTarg = neResolve->AID;
		while (*pSrc == ' ')
		{
			pSrc++;
		}
		cQuoteChar = *pSrc;
		if (cQuoteChar == MS_OPEN_QUOTE) cQuoteChar = MS_CLOSE_QUOTE;
		if (cQuoteChar == MS_OPEN_DQUOTE) cQuoteChar = MS_CLOSE_DQUOTE;
		pSrc++;
		while ((*pSrc != cQuoteChar) && (*pSrc != '\0'))
		{
			*pTarg = *pSrc;
			pTarg++;
			pSrc++;
		}
		*pTarg = '\0';

		// Get the string pattern of the arbitrary entity
		getASP(neResolve->AID, szSPattern, MAXUSERNAME);
		//  Trace entity pattern established
		sprintf_s(szMsg, MAX_MSG, MSG_DXR8010T, szSPattern);
		xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
	}

	//  See if the pattern is recognized as a standard pattern
	if (strcmp(szSPattern, PATT_EMAIL_1) == 0)
	{
		strcpy_s(neResolve->EMAIL, MAXUSERNAME, neResolve->AID);
		neResolve->EMDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_EMAIL;
		return;
	}

	if (strcmp(szSPattern, PATT_EMAIL_2) == 0)
	{
		strcpy_s(neResolve->EMAIL, MAXUSERNAME, neResolve->AID);
		if (!extractAtom(neResolve->ADDN, MAXUSERNAME, neResolve->AID, szSPattern, 1)) return;
		neResolve->ADDNQ = NE_DN_UNRESOLVED;
		neResolve->EMDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_EMAIL;
		return;
	}

	if (strcmp(szSPattern, PATT_EMAIL_3) == 0)
	{
		if (!extractAtom(neResolve->EMAIL, MAXUSERNAME, neResolve->AID, szSPattern, 2)) return;
		if (!extractAtom(neResolve->EMAIL + (int) strlen(neResolve->EMAIL), MAXUSERNAME - (int) strlen(neResolve->EMAIL), neResolve->AID, szSPattern, 3)) return;
		if (!extractAtom(neResolve->EMAIL + (int) strlen(neResolve->EMAIL), MAXUSERNAME - (int) strlen(neResolve->EMAIL), neResolve->AID, szSPattern, 4)) return;
		neResolve->EMDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_EMAIL;
		return;
	}

	if (strcmp(szSPattern, PATT_EMAIL_4) == 0)
	{
		if (!extractAtom(neResolve->ADDN, MAXUSERNAME, neResolve->AID, szSPattern, 2)) return;
		neResolve->ADDNQ = NE_DN_UNRESOLVED;
		if (!extractAtom(neResolve->EMAIL, MAXUSERNAME, neResolve->AID, szSPattern, 2)) return;
		if (!extractAtom(neResolve->EMAIL + (int) strlen(neResolve->EMAIL), MAXUSERNAME - (int) strlen(neResolve->EMAIL), neResolve->AID, szSPattern, 3)) return;
		if (!extractAtom(neResolve->EMAIL + (int) strlen(neResolve->EMAIL), MAXUSERNAME - (int) strlen(neResolve->EMAIL), neResolve->AID, szSPattern, 4)) return;
		neResolve->EMDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_EMAIL;
		return;
	}

	if (strcmp(szSPattern, PATT_EMAIL_RFC822_1) == 0)
	{
		if (!extractAtom(neResolve->EMAIL, MAXUSERNAME, neResolve->AID, szSPattern, 3)) return;
		if (!extractAtom(neResolve->EMAIL + (int) strlen(neResolve->EMAIL), MAXUSERNAME - (int) strlen(neResolve->EMAIL), neResolve->AID, szSPattern, 4)) return;
		if (!extractAtom(neResolve->EMAIL + (int) strlen(neResolve->EMAIL), MAXUSERNAME - (int) strlen(neResolve->EMAIL), neResolve->AID, szSPattern, 5)) return;
		neResolve->EMDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_EMAIL;
		return;
	}

	if (strcmp(szSPattern, PATT_EMAIL_RFC822_2) == 0)
	{
		if (!extractAtom(neResolve->ADDN, MAXUSERNAME, neResolve->AID, szSPattern, 3)) return;
		neResolve->ADDNQ = NE_DN_UNRESOLVED;
		if (!extractAtom(neResolve->EMAIL, MAXUSERNAME, neResolve->AID, szSPattern, 3)) return;
		if (!extractAtom(neResolve->EMAIL + (int) strlen(neResolve->EMAIL), MAXUSERNAME - (int) strlen(neResolve->EMAIL), neResolve->AID, szSPattern, 4)) return;
		if (!extractAtom(neResolve->EMAIL + (int) strlen(neResolve->EMAIL), MAXUSERNAME - (int) strlen(neResolve->EMAIL), neResolve->AID, szSPattern, 5)) return;
		neResolve->EMDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_EMAIL;
	}

	if ((strcmp(szSPattern, PATT_AD_ABBREVIATED) == 0) || (strcmp(szSPattern, PATT_AD_OVERQUALIFIED) == 0))
	{
		if (!extractAtom(neResolve->ADDN, MAXUSERNAME, neResolve->AID, szSPattern, 3)) return;
		neResolve->ADDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_AD;
		return;
	}

	if (strcmp(szSPattern, PATT_DOMINO_SN) == 0)
	{
		strcpy_s(neResolve->DDN, MAXUSERNAME, neResolve->AID);
		neResolve->DDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_DOMINO;
		return;
	}

	if ((strcmp(szSPattern, PATT_DOMINO_ALT_SN1) == 0) || (strcmp(szSPattern, PATT_DOMINO_ALT_SN2) == 0))
	{
		if (!extractAtom(neResolve->DDN, MAXUSERNAME, neResolve->AID, szSPattern, 1)) return;
		if (!extractAtom(neResolve->DDN + (int) strlen(neResolve->DDN), MAXUSERNAME - (int) strlen(neResolve->DDN), neResolve->AID, szSPattern, 2)) return;
		neResolve->DDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_DOMINO;
		return;
	}

	if (strcmp(szSPattern, PATT_ANY_CN) == 0)
	{
		strcpy_s(neResolve->ADDN, MAXUSERNAME, neResolve->AID);
		neResolve->ADDNQ = NE_DN_UNRESOLVED;
		strcpy_s(neResolve->DDN, MAXUSERNAME, neResolve->AID);
		neResolve->DDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_DOMINO | NE_ID_AD;
		return;
	}

	//  Check for Domino abbreviated name "s/s......"
	if (((szSPattern[0] == 's') || (szSPattern[0] == 'd')) && (szSPattern[1] == '/') && (szSPattern[2] == 's'))
	{
		strcpy_s(neResolve->DDN, MAXUSERNAME, neResolve->AID);
		neResolve->DDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_DOMINO;
		return;
	}

	//  Check for Domino abbreviated name "s/s......"
	if ((szSPattern[0] == 's') && (szSPattern[1] == 'u') && (szSPattern[2] == '/') && (szSPattern[2] == '/'))
	{
		strcpy_s(neResolve->DDN, MAXUSERNAME, neResolve->AID);
		neResolve->DDNQ = NE_DN_UNRESOLVED;
		neResolve->AIDAff = NE_ID_DOMINO;
		return;
	}

	//  Check for and discriminate between Domino & AD DN
	if ((szSPattern[0] == 's') && (szSPattern[1] == '=') && (szSPattern[2] == 's'))
	{
		if (szSPattern[3] == '/')
		{
			strcpy_s(neResolve->DDN, MAXUSERNAME, neResolve->AID);
			neResolve->DDNQ = NE_DN_UNRESOLVED;
			neResolve->AIDAff = NE_ID_DOMINO;
			return;
		}
		if (szSPattern[3] == ',')
		{
			strcpy_s(neResolve->ADDN, MAXUSERNAME, neResolve->AID);
			neResolve->ADDNQ = NE_DN_UNRESOLVED;
			neResolve->AIDAff = NE_ID_AD;
			return;
		}
	}
	return;
}

//  getASP
//
//  This function will return a normalised pattern for an Arbitrary String
//
//  Parameters
//
//  1  -  char *			-  Arbitrary String 
//  2  -  char *			-  Pointer to the buffer to hold the returned pattern
//  3  -  int				-  Size of the buffer to hold the returned pattern
//
//  Returns
//
//
//  API Use Notes:
//
//
void NEResolver::getASP(char *szAS, char *szASP, int iPatternSize)
{
//  Define the states for the scanning engine
#define			SS_IN_LIMBO						1							//  Pattern element not Started
#define			SS_DIGEST_WHITESPACE			2							//  Whitespace digestion
#define			SS_DIGEST_QSTRING				3							//  Quoted string digestion
#define			SS_DIGEST_STRING				4							//  String digestion
#define			SS_DIGEST_DSTRING				5							//  Digest dotted string

	char			*pChar;													//  Pointer to the current character in the string
	char			cQuoteChar;												//  Quote character in use
	char			cEndPattern;											//  Ending pattern
	int				iPatternOffset;											//  Offset into the pattern buffer
	int				iState;													//  Current state of the scanning engine
#ifdef _DEBUG
	char			szMsg[MAX_MSG + 1];										//  Generic Message Buffer
#endif

	//  Perform the initial setup
	szASP[0] = '\0';
	iPatternOffset = 0;
	pChar = szAS;
	iState = SS_IN_LIMBO;

	//  Scanning loop
	while (*pChar != '\0')
	{
#ifdef _DEBUG
		sprintf_s(szMsg, MAX_MSG, "DEBUG: Digesting '%c' (%i) at state %i.", *pChar, *pChar, iState);
		xeLocal->LogMessage(szMsg, 0);
#endif
		switch(iState)
		{
		case SS_IN_LIMBO:
			//  Check for whitespace
			if (*pChar == ' ')
			{
				//  Digest Whitespace
				iState = SS_DIGEST_WHITESPACE;
			}
			else
			{
				//  Check for the start of a quoted string
				if ((*pChar == '"') || (*pChar == '\'') || (*pChar == MS_OPEN_QUOTE) || (*pChar == MS_OPEN_DQUOTE))
				{
					//  Quoted String is in progress
					cQuoteChar = *pChar;
					if (cQuoteChar == MS_OPEN_QUOTE) cQuoteChar = MS_CLOSE_QUOTE;
					if (cQuoteChar == MS_OPEN_DQUOTE) cQuoteChar = MS_CLOSE_DQUOTE;
					iState = SS_DIGEST_QSTRING;
				}
				else
				{
					//  Check for special characters
					if (isSpecialCharacter(*pChar))
					{
						//  Add to pattern
						if (iPatternOffset < iPatternSize)
						{
							szASP[iPatternOffset] = *pChar;
							iPatternOffset++;
						}
						else 
						{
							szASP[iPatternOffset] = '!';
						}
					}
					else
					{
						//  This character is part of a string
						if (*pChar == '.') iState = SS_DIGEST_DSTRING;
						else iState = SS_DIGEST_STRING;
					}
				}
			}
			break;

		case SS_DIGEST_WHITESPACE:
			//  Digesting white space
			if (*pChar != ' ') 
			{
				pChar--;
				iState = SS_IN_LIMBO;
			}
			break;

		case SS_DIGEST_QSTRING:
			//  Digesting a quoted string
			if (*pChar == cQuoteChar) 
			{
				//  Add to pattern
				if (iPatternOffset < iPatternSize)
				{
					szASP[iPatternOffset] = 'q';
					iPatternOffset++;
				}
				else 
				{
					szASP[iPatternOffset] = '!';
				}
				iState = SS_IN_LIMBO;
			}
			break;

		case SS_DIGEST_DSTRING:
			//  Digesting a diotted string

			//  Check for the start of a quoted string
			if ((*pChar == '"') || (*pChar == '\'') || (*pChar == MS_OPEN_QUOTE) || (*pChar == MS_OPEN_DQUOTE))
			{
				//  Quoted String is in progress
				cQuoteChar = *pChar;
				if (cQuoteChar == MS_OPEN_QUOTE) cQuoteChar = MS_CLOSE_QUOTE;
				if (cQuoteChar == MS_OPEN_DQUOTE) cQuoteChar = MS_CLOSE_DQUOTE;
				//  Add to pattern
				if (iPatternOffset < iPatternSize)
				{
					szASP[iPatternOffset] = 'd';
					iPatternOffset++;
				}
				else 
				{
					szASP[iPatternOffset] = '!';
				}
				iState = SS_DIGEST_QSTRING;
			}
			else
			{
				//  Check for special characters
				if (isSpecialCharacter(*pChar))
				{
					//  Add to pattern
					if (iPatternOffset < iPatternSize)
					{
						szASP[iPatternOffset] = 'd';
						iPatternOffset++;
					}
					else 
					{
						szASP[iPatternOffset] = '!';
					}
					//  Add to pattern
					if (iPatternOffset < iPatternSize)
					{
						szASP[iPatternOffset] = *pChar;
						iPatternOffset++;
					}
					else 
					{
						szASP[iPatternOffset] = '!';
					}
					iState = SS_IN_LIMBO;
				}
			}
			break;

		case SS_DIGEST_STRING:
			//  Digesting a string

			//  Check for transition to a dotted string
			if (*pChar == '.') iState = SS_DIGEST_DSTRING ;
			else
			{
				//  Check for the start of a quoted string
				if ((*pChar == '"') || (*pChar == '\'') || (*pChar == MS_OPEN_QUOTE) || (*pChar == MS_OPEN_DQUOTE))
				{
					//  Quoted String is in progress
					cQuoteChar = *pChar;
					if (cQuoteChar == MS_OPEN_QUOTE) cQuoteChar = MS_CLOSE_QUOTE;
					if (cQuoteChar == MS_OPEN_DQUOTE) cQuoteChar = MS_CLOSE_DQUOTE;
					//  Add to pattern
					if (iPatternOffset < iPatternSize)
					{
						szASP[iPatternOffset] = 's';
						iPatternOffset++;
					}
					else 
					{
						szASP[iPatternOffset] = '!';
					}
					iState = SS_DIGEST_QSTRING;
				}
				else
				{
					//  Check for special characters
					if (isSpecialCharacter(*pChar))
					{
						//  Add to pattern
						if (iPatternOffset < iPatternSize)
						{
							szASP[iPatternOffset] = 's';
							iPatternOffset++;
						}
						else 
						{
							szASP[iPatternOffset] = '!';
						}
						//  Add to pattern
						if (iPatternOffset < iPatternSize)
						{
							szASP[iPatternOffset] = *pChar;
							iPatternOffset++;
						}
						else 
						{
							szASP[iPatternOffset] = '!';
						}
					}
				}
			}
			break;
		}
		pChar++;
	}

	//  Determine how to terminate the pattern according to the current state of the scanning engine
	szASP[iPatternOffset] = '\0';
	szASP[iPatternOffset + 1] = '\0';
	if (iState == SS_IN_LIMBO) return;
	if (iState == SS_DIGEST_WHITESPACE) return;
	if (iState == SS_DIGEST_STRING) cEndPattern = 's';
	if (iState == SS_DIGEST_QSTRING) cEndPattern = 'u';
	if (iState == SS_DIGEST_DSTRING) cEndPattern = 'd';
	//  Add to pattern
	if (iPatternOffset < iPatternSize)
	{
		szASP[iPatternOffset] = cEndPattern;
		iPatternOffset++;
	}
	else 
	{
		szASP[iPatternOffset] = '!';
	}
	szASP[iPatternOffset + 1] = '\0';
	return;
}

//  isSpecialCharacter
//
//  This function will determine if the supplied character is a special character in a string pattern
//
//  Parameters
//
//  1  -  char 				-  Supplied Character
//
//  Returns
//
//
//  API Use Notes:
//
//
BOOL NEResolver::isSpecialCharacter(char cIn)
{
	if (cIn == '@') return TRUE;
	if (cIn == '/') return TRUE;
	if (cIn == '\\') return TRUE;
	if (cIn == '*') return TRUE;
	if (cIn == '?') return TRUE;
	if (cIn == ';') return TRUE;
	if (cIn == '#') return TRUE;
	if (cIn == ',') return TRUE;
	if (cIn == '=') return TRUE;
	if (cIn == ':') return TRUE;
	if (cIn == '<') return TRUE;
	if (cIn == '>') return TRUE;

	return FALSE;
}

//  extractAtom
//
//  This function will return the requested atom from the passed string
//
//  Parameters
//
//  1  -  char *			-  Pointer to the buffer to hold the returned atom
//  2  -  int				-  Size of the buffer to hold the returned atom
//  3  -  char *			-  Source String 
//  4  -  char *			-  Source String pattern
//  5  -  int				-  The ordinal number of the atom to be extracted
//
//  Returns
//
//  BOOL - TRUE if the atom was extracted and FALSE if not
//
//  API Use Notes:
//
//
BOOL NEResolver::extractAtom(char *szTarget, int iTargetSize, char *szSource, char *szPattern, int iToken)
{
	int				iCurrentToken;								//  Current Token
	int				iOutputLen;									//  Output string length
	char			cQuoteChar;									//  Close quote character
	char			*pSrc;										//  Pointer into the source string
	char			*pTarget;									//  Pointer into the target string
	char			*pPattern;									//  Pointer into the pattern
#ifdef _DEBUG
	char			szMsg[MAX_MSG + 1];							//  Generic Message Buffer
#endif

	//  Setup toperform the extraction
	iCurrentToken = 1;
	iOutputLen = 0;
	pSrc = szSource;
	pTarget = szTarget;
	pPattern = szPattern;

	//  Eliminate any leading whitespace
	while (*pSrc == ' ')
	{
		pSrc++;
	}
#ifdef _DEBUG
	sprintf_s(szMsg, MAX_MSG, "DEBUG: Src='%s', Patt='%s', Token=%i.", szSource, szPattern, iToken);
	xeLocal->LogMessage(szMsg, 0);
#endif
	//  Locate the correct token
	while (iCurrentToken < iToken)
	{
#ifdef _DEBUG
		sprintf_s(szMsg, MAX_MSG, "DEBUG: Eliminating token %i - %c (%i) starting at '%s'.", iCurrentToken, *pPattern, *pPattern, pSrc);
		xeLocal->LogMessage(szMsg, 0);
#endif
		if (isSpecialCharacter(*pPattern))
		{
			//  Skip through the source until we find the character
			while (*pSrc != *pPattern)
			{
				if (*pSrc == '\0') return FALSE;
				pSrc++;
			}
			pSrc++;								//  Eliminate the character
		}
		else
		{
			if ((*pPattern == 'q') || (*pPattern == 'u'))
			{
				// Locate the start of the quoted string
				while ((*pSrc != '"') && (*pSrc != '\'') && (*pSrc != MS_OPEN_QUOTE) && (*pSrc != MS_OPEN_DQUOTE))
				{
					if (*pSrc == '\0') return FALSE;
					pSrc++;
				}

				if (*pPattern == 'q')
				{
					//  Eliminate a quoted string
					cQuoteChar = *pSrc;
					if (cQuoteChar == MS_OPEN_QUOTE) cQuoteChar = MS_CLOSE_QUOTE;
					if (cQuoteChar == MS_OPEN_DQUOTE) cQuoteChar = MS_CLOSE_DQUOTE;
					pSrc++;
					while (*pSrc != cQuoteChar)
					{
						if (*pSrc == '\0') return FALSE;
						pSrc++;
					}
					pSrc++;
				}
				else
				{
					//  Eliminate an unbalanced quoted string
					pSrc++;
					while (!isSpecialCharacter(*pSrc))
					{
						if (*pSrc == '\0') return FALSE;
						pSrc++;
					}
				}

			}
			else
			{
				if ((*pPattern == 's') || (*pPattern == 'd'))
				{
					// Skip leading whitespace
					while (*pSrc == ' ')
					{
						pSrc++;
					}

					//  Eliminate a string or dotted string
					while (!isSpecialCharacter(*pSrc))
					{
						if (*pSrc == '\0') return FALSE;
						if ((*pSrc == '"') || (*pSrc == '\'') || (*pSrc == MS_OPEN_QUOTE) || (*pSrc == MS_OPEN_DQUOTE)) break;
						pSrc++;
					}
				}				
			}
		}
		pPattern++;
		if (*pPattern == '\0') return FALSE;						//  Overrun the end of the pattern
		iCurrentToken++;
	}

	//  Check that we are on the correct token and not at the end of the string
	if (iCurrentToken != iToken) return FALSE;
	if (*pSrc == '\0') return FALSE;
#ifdef _DEBUG
	sprintf_s(szMsg, MAX_MSG, "DEBUG: Extracting Patt=%c (%i), Token=%i from '%s' starting at '%s'.", *pPattern, *pPattern, iToken, szSource, pSrc);
	xeLocal->LogMessage(szMsg, 0);
#endif
	//  Copy the atom to the target buffer
	if (isSpecialCharacter(*pPattern))
	{
		*pTarget = *pSrc;
		pTarget++;
		*pTarget = '\0';
		return TRUE;
	}

	if (*pPattern == 'q')
	{
		cQuoteChar = *pSrc;
		if (cQuoteChar == MS_OPEN_QUOTE) cQuoteChar = MS_CLOSE_QUOTE;
		if (cQuoteChar == MS_OPEN_DQUOTE) cQuoteChar = MS_CLOSE_DQUOTE;
		pSrc++;
		while (*pSrc != cQuoteChar)
		{
			if (*pSrc == '\0') return FALSE;
			*pTarget = *pSrc;
			iOutputLen++;
			if (iOutputLen == iTargetSize) return FALSE;
			pSrc++;
			pTarget++;
		}
		*pTarget = '\0';
		return TRUE;
	}

	if ((*pPattern == 's') || (*pPattern == 'd'))
	{
		while (!isSpecialCharacter(*pSrc))
		{
			if (*pSrc == '\0') return TRUE;
			if ((*pSrc == '"') || (*pSrc == '\'') || (*pSrc == MS_OPEN_QUOTE) || (*pSrc == MS_OPEN_DQUOTE)) break;
			*pTarget = *pSrc;
			iOutputLen++;
			if (iOutputLen == iTargetSize) return FALSE;
			pSrc++;
			pTarget++;
		}
		*pTarget = '\0';
		return TRUE;
	}

	if (*pPattern == 'u')
	{
		while (*pSrc != '\0')
		{
			*pTarget = *pSrc;
			iOutputLen++;
			if (iOutputLen == iTargetSize) return FALSE;
			pSrc++;
			pTarget++;
		}
		*pTarget = '\0';
		return TRUE;
	}

	return FALSE;
}

//  setDomRefServer
//
//  This function will set the value for the Domino Reference Server Name
//
//  Parameters
//
//  1  -  char *		-  Pointer to the server name to set
//
//  Returns
//
//  BOOL   -   TRUE is the value was set, FALSE if not
//
//  API Use Notes:
//
//
BOOL NEResolver::setDomRefServer(char *szServer)
{
	if (szServer == NULL) return FALSE;
	if (strlen(szServer) > MAX_SERVER) return FALSE;
	strcpy_s(szDomRefServer, MAX_SERVER, szServer);
	return TRUE;
}

//  setLDAPBaseDN
//
//  This function will set the Base DN to be used in LDAP searches
//
//  Parameters
//
//  1  -  char *		-  Pointer to the distinguished name to set as a search base.
//
//  Returns
//
//  BOOL   -   TRUE is the value was set, FALSE if not
//
//  API Use Notes:
//
//
BOOL NEResolver::setLDAPBaseDN(char *szDN)
{
	if (szDN == NULL) return FALSE;
	if (strlen(szDN) > MAX_MSG) return FALSE;
	setVirtualBaseDN(szDN);
	return TRUE;
}

//  setVirtualBaseDN
//
//  This function will set the Base DN and if needed the Match DN to be used in LDAP searches
//
//  Parameters
//
//  1  -  char *		-  Pointer to the distinguished name to set as a search base
//
//  Returns
//
//
//  API Use Notes:
//
//
void NEResolver::setVirtualBaseDN(char *szDN)
{
	char			*pScan;										//  Character string scanning position
	char			*pStartQual;								//  Start of qualified (i.e. Search Base DN)

	//  Clear the DNs
	szBaseDN[0] = '\0';
	szMatchDN[0] = '\0';

	//  Find the split point between the fully qualified part of the passed DN and the unqualified (containing wildcard) parts.
	pStartQual = szDN;											//  Default to the complete string is qualified
	pScan = szDN;
	while (*pScan != '\0')
	{
		if ((*pScan == '*') || (*pScan == '?'))
		{
			//  Find the next delimiter
			while ((*pScan != ',') && (*pScan != '\0'))
			{
				pScan++;
			}
			pStartQual = pScan;
		}
		else pScan++;
	}

	//  Determine the qualification state of the virtual DN and set the base and match accordingly
	if (pStartQual == szDN)
	{
		//  The complete DN is qualified set it as the Base DN and leave the match DN empty
		strcpy_s(szBaseDN, MAX_MSG, szDN);
		return;
	}

	if (*pStartQual == '\0')
	{
		//  The complete DN is unqualified set it as the match DN and leave the base DN empty
		strcpy_s(szMatchDN, MAX_MSG, szDN);
		return;
	}

	//  The passed DN is partially qualified set the Base DN to the qualified part and the match DN to the complete string
	strcpy_s(szMatchDN, MAX_MSG, szDN);
	strcpy_s(szBaseDN, MAX_MSG, pStartQual + 1);
	return;
}

//  setLDAPSessionHandle
//
//  This function will set the LDAP session handle to use for searches
//
//  Parameters
//
//  1  -  LDAP *		-  Pointer to the LDAP Session Handle
//
//  Returns
//
//
//  API Use Notes:
//
//
void NEResolver::setLDAPSessionHandle(LDAP *hlsSet)
{
	hlsSession = hlsSet;
}

//  getLDAPSessionHandle
//
//  This function will return the LDAP session handle used for searches
//
//  Parameters
//
//
//  Returns
//
//  LDAP *		-  Pointer to the LDAP Session Handle
//
//  API Use Notes:
//
//
LDAP * NEResolver::getLDAPSessionHandle()
{
	return hlsSession;
}

//  dumpNamedEntity
//
//  This function will dump the contents of the passed named entity
//
//  Parameters
//
//  1  -  NamedEntity *		-  Pointer to the Named Entity to be dumped
//  2  -  int				-  ID of the calling thread
//
//  Returns
//
//
//  API Use Notes:
//
//
void NEResolver::dumpNamedEntity(NamedEntity *neDump, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer

	if (xeLocal->LogGetLevel() < LOGLEVEL_TRACE) return;	
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8012T, "Arbitrary ID", neDump->AIDAff, neDump->AID);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8012T, "Domino ID", neDump->DDNQ, neDump->DDN);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8012T, "Active Directory ID", neDump->ADDNQ, neDump->ADDN);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR8012T, "Email ID", neDump->EMDNQ, neDump->EMAIL);
	xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
	return;
}

//  getFilteredMatchCount
//
//  This function will return the count of entries in a search rusult that match the DN Match filter (if any applies)
//
//  Parameters
//
//  1  -  LDAPMessage *		-  Handle of the raw search results message
//  2  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  ULONG			-	Count of matched entries
//
//  API Use Notes:
//
//
ULONG NEResolver::getFilteredMatchCount(LDAPMessage *lmRaw, int iThreadID)
{
	ULONG			ulMatchingCount;									//  Count of entries that match
	LDAPMessage		*pMatch;											//  Matched Entry
	int				iPatternCommas;										//  Pattern Commas
	int				iDNCommas;											//  DN Commas
	char			*pDN;												//  Returned Distinguished Name
	char			*pDNHier;											//  Pointer to the hierarchic part of the DN
	char			szSavedDN[MAX_MSG + 1];								//  Save the first matching DN
	char			szMsg[MAX_MSG + 1];									//  Generic Message Buffer

	//  Get the native matching count
	ulMatchingCount = ldap_count_entries(hlsSession, lmRaw);

	//  If there is no Match DN filter in effect then all of the raw search results will match
	if (szMatchDN[0] == '\0')
	{
		return ulMatchingCount;
	}

	//  Check that there is actually something in the message to filter
	if (ulMatchingCount > 0)
	{
		//  We need to loop through each of the returned results
		ulMatchingCount = 0;
		pMatch = NULL;
		pMatch = ldap_first_entry(hlsSession, lmRaw);
		while (pMatch != NULL)
		{
			//  Get the DN for the supplied result
			pDN = NULL;
			pDN = ldap_get_dn(hlsSession, pMatch);
			if (pDN != NULL)
			{
				//  Determine the alignment point in the hierarchy 
				//  Firstly count the number of seperators in the pattern Base DN
				iPatternCommas = 0;
				iDNCommas = 0;
				pDNHier = szMatchDN;
				while (*pDNHier != '\0')
				{
					if (*pDNHier == ',') iPatternCommas++;
					pDNHier++;
				}
				//  Work back through the returned DN to determine the start of significance
				pDNHier = pDN;
				pDNHier = pDNHier + strlen(pDNHier);
				pDNHier--;
				while ((iDNCommas <= iPatternCommas) && (pDNHier > pDN))
				{
					if (*pDNHier == ',') iDNCommas++;
					pDNHier--;
				}
				if (iDNCommas > iPatternCommas) pDNHier+= 2;						//  point past the comma
#ifdef _DEBUG		
				printf("DEBUG: Matching '%s'.\r\n", pDNHier);
				printf("DEBUG: Against '%s'.\r\n", szMatchDN);
#endif
				if (xeLocal->MatchesPatterni(pDNHier, szMatchDN)) 
				{
					if (ulMatchingCount == 0)
					{
						//  Just save the first matching entry for future reference
						strcpy_s(szSavedDN, MAX_MSG, pDN);
					}
					else
					{
						//  If show duplicates is on then show the appropriate entries
						if (ulMatchingCount == 1)
						{
							//  Show the saved entry
							sprintf_s(szMsg, MAX_MSG, MSG_DXR8030T, szSavedDN);
							if (ShowDuplicates) xeLocal->LogMessage(szMsg, iThreadID);
							else xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
						}

						//  Show the current entry as a duplicate
						sprintf_s(szMsg, MAX_MSG, MSG_DXR8030T, pDN);
						if (ShowDuplicates) xeLocal->LogMessage(szMsg, iThreadID);
						else xeLocal->LogTrace(TRACE_NER, szMsg, iThreadID);
					}
					ulMatchingCount++;
				}
				ldap_memfree(pDN);
			}
			pMatch = ldap_next_entry(hlsSession, pMatch);
		}	
	}

	return ulMatchingCount;
}

//  getFirstFilteredMatch
//
//  This function will return the a pointer to the LDAP Message for the first (matched) entry in the search results
//
//  Parameters
//
//  1  -  LDAPMessage *		-  Handle of the raw search results message
//  2  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  LDAPMessage *			-  Pointer (handle) to the matched entry
//
//  API Use Notes:
//
//
LDAPMessage * NEResolver::getFirstFilteredMatch(LDAPMessage *lmRaw, int iThreadID)
{
	LDAPMessage		*pMatch;											//  Matched Entry
	char			*pDN;												//  Returned Distinguished Name
	char			*pDNHier;											//  Pointer to the hierarchic part of the DN

	//  Get the first matching entry
	pMatch = NULL;
	pMatch = ldap_first_entry(hlsSession, lmRaw);

	//  If there is no Match DN filter in effect then all of the raw search results will match
	if (szMatchDN[0] == '\0')
	{
		return pMatch;
	}

	//  Loop processing each candidate in turn
	while (pMatch != NULL)
	{
		//  Get the DN for the supplied result
		pDN = NULL;
		pDN = ldap_get_dn(hlsSession, pMatch);
		if (pDN != NULL)
		{
			pDNHier = pDN;
			while ((*pDNHier != ',') && (*pDNHier != '\0'))
			{
				pDNHier++;
			}
			if (*pDNHier == ',') pDNHier++;
			if (xeLocal->MatchesPatterni(pDNHier, szMatchDN))
			{
				ldap_memfree(pDN);
				return pMatch;
			}
			ldap_memfree(pDN);
			pMatch = ldap_next_entry(hlsSession, pMatch);
		}	
	}
	return pMatch;
}

//  getNextFilteredMatch
//
//  This function will return the a pointer to the LDAP Message for the next (matched) entry in the search results
//
//  Parameters
//
//  1  -  LDAPMessage *		-  Handle of the previous matched entry message
//  2  -  int				-  ID of the thread requesting to set the list
//
//  Returns
//
//  LDAPMessage *			-  Pointer (handle) to the matched entry
//
//  API Use Notes:
//
//
LDAPMessage * NEResolver::getNextFilteredMatch(LDAPMessage *lmRaw, int iThreadID)
{
	LDAPMessage		*pMatch;											//  Matched Entry
	char			*pDN;												//  Returned Distinguished Name
	char			*pDNHier;											//  Pointer to the hierarchic part of the DN

	//  Get the first matching entry
	pMatch = NULL;
	pMatch = ldap_next_entry(hlsSession, lmRaw);

	//  If there is no Match DN filter in effect then all of the raw search results will match
	if (szMatchDN[0] == '\0')
	{
		return pMatch;
	}

	//  Loop processing each candidate in turn
	while (pMatch != NULL)
	{
		//  Get the DN for the supplied result
		pDN = NULL;
		pDN = ldap_get_dn(hlsSession, pMatch);
		if (pDN != NULL)
		{
			pDNHier = pDN;
			while ((*pDNHier != ',') && (*pDNHier != '\0'))
			{
				pDNHier++;
			}
			if (*pDNHier == ',') pDNHier++;
			if (xeLocal->MatchesPatterni(pDNHier, szMatchDN))
			{
				ldap_memfree(pDN);
				return pMatch;
			}
			ldap_memfree(pDN);
			pMatch = ldap_next_entry(hlsSession, pMatch);
		}	
	}
	return pMatch;
}
