#pragma once
//  NEResolver
//
//  This class provides Named Entity resolution services providing the ability to retrieve a particular identity given an
//  arbitrary identity.
//
//  NOTE: This class uses the Windows LDAP Client API rather than the Domino LDAP Client API, this is because the LDAP sessions have to
//        be shared with authenticater classes that use special AD features that are not accessible through the Domino LDAP API.
//
//  Message Identifier:      DXR
//  Message Range:           8001 - 8499
//  Trace area assignment:   TRACE_NER (30)
//  DX Version:				 3.11.1
//  Version:                 3.11.1  Build: 99
//
//  Version History:
//
//  3.11.1		22/10/2011		VS 2011 readiness
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2011
//

//  Platform Includes
#include "../Platform/PlatBase.h"
#include "../Platform/NotesBase.h"

//  Additional Notes Includes
#include <lookup.h>

//  Windows LDAP Includes
#include	<Winldap.h>

//  Runtime Includes
#include "../ExecEnvironment.h"
#include "NECache.h"
#include "NamedEntity.h"

//  Local Definitions
#define		CACHE_SIZE					200											//  Size of the Named Entity cache
#define		NO_LDAP_MSG					"Unknown LDAP Error"

//  Special Characters
#define		MS_OPEN_QUOTE				'\x91'										//  Open Quote
#define		MS_CLOSE_QUOTE				'\x92'										//  Close Quote
#define		MS_OPEN_DQUOTE				'\x93'										//  Open Double Quote
#define		MS_CLOSE_DQUOTE				'\x94'										//  Close Double Quote

// Standard String Patterns
#define		PATT_EMAIL_1				"d@d"										//  E-Mail Address Format # 1
#define		PATT_EMAIL_2				"s@d"										//  E-Mail Address Format # 2
#define		PATT_EMAIL_3				"<d@d>"										//  E-Mail Address Format # 3
#define		PATT_EMAIL_4				"<s@d>"										//  E-Mail Address Format # 4
#define		PATT_EMAIL_RFC822_1			"q<d@d>"									//  RFC822 E-Mail Address Format # 1
#define		PATT_EMAIL_RFC822_2			"q<s@d>"									//  RFC822 E-Mail Address Format # 2
#define		PATT_AD_ABBREVIATED			"s\\s"										//  AD Domain\CN format
#define		PATT_AD_OVERQUALIFIED		"s\\s@d"									//  Overqualified AD Domain\CN and SMTP domain
#define		PATT_DOMINO_SN				"d"											//  Domino Short Name
#define		PATT_DOMINO_ALT_SN1			"du"										//  Alternate Domino Short Name
#define		PATT_DOMINO_ALT_SN2			"su"										//  Alternate Domino Short Name/Common Name
#define		PATT_ANY_CN					"s"											//  Domnio or AD CN

//  LDAP Search Strings
#define		LDAP_QUERY_CN				"(&(|(cn=%s)(uid=%s))(ObjectCategory=person))"
#define		LDAP_QUERY_EMAIL			"(&(mail=%s)(ObjectCategory=person))"
#define		LDAP_QUERY_EITHER			"(&(|(cn=%s)(uid=%s)(mail=%s))(ObjectCategory=person))"
#define		LDAP_ATTR_1					"CN"										//  Common Name
#define		LDAP_ATTR_2					"UID"										//  Unique ID
#define		LDAP_ATTR_3					"Mail"										//  Email Address
#define		LDAP_ATTR_4					"sAMAccountName"							//  Administrative Account Name

//  Return codes
#define		NERET_INVPARM				1											//  Invalid parameters were passed to a resolver call
#define		NERET_RESOLVED				2											//  The ID was fully resolved
#define		NERET_NOTFOUND				3											//  The ID was constructed but could not be found
#define		NERET_UNRESOLVED			4											//  The ID is unresolved

//  Trace Messages
#define		MSG_DXR8001T				"DXR8001T: A new NEResolver instance has been created @%p."
#define		MSG_DXR8002T				"DXR8002T: A NEResolver instance @%p has been destroyed."
#define		MSG_DXR8005T				"DXR8005T: Starting '%s' resolve for entity '%s'."
#define		MSG_DXR8006T				"DXR8006T: '%s' was resolved to '%s' from the existing Named Entity."
#define		MSG_DXR8007T				"DXR8007T: '%s' was resolved to '%s' from the existing Named Entity, but was not found in AD."
#define		MSG_DXR8008T				"DXR8008T: Attempting to resolve the affinity for '%s'."
#define		MSG_DXR8009T				"DXR8009T: Unable to resolve the affinity for '%s'."
#define		MSG_DXR8010T				"DXR8010T: Entity pattern has been established as '%s'."
#define		MSG_DXR8011T				"DXR8011T: Affinity has been determined for the Named Entity that follows."
#define		MSG_DXR8012T				"DXR8012T: Named Entity '%s' qualification: 0x%.4x value: '%s'."
#define		MSG_DXR8016T				"DXR8016T: Domino resolution has determined the Named Entity that follows."
#define		MSG_DXR8019T				"DXR8019T: LDAP search for '%s' found no matches."
#define		MSG_DXR8020T				"DXR8020T: LDAP search for '%s' found multiple matches."
#define		MSG_DXR8022T				"DXR8022T: No value was returned for attribute '%s'."
#define		MSG_DXR8023T				"DXR8023T: Attribute: %s = '%s'."
#define		MSG_DXR8024T				"DXR8024T: Issuing LDAP Search for '%s'."
#define		MSG_DXR8025T				"DXR8025T: Issuing LDAP Search in Base DN: '%s'."
#define		MSG_DXR8027T				"DXR8027T: Active Directory resolution has determined the Named Entity that follows."
#define		MSG_DXR8030T				"DXR8030T: Duplicate entry '%s' found in AD search."

//  Information Messages
#define		MSG_DXR8014I				"DXR8014I: Domino lookup of '%s' returned nothing."
#define		MSG_DXR8015I				"DXR8015I: Domino lookup of '%s' returned %i matches, authoratative result not possible."

//  Warning Messages
#define		MSG_DXR8003W				"DXR8003W: The supplied list of Domnio Organisations is too long, it will be trunncated."
#define		MSG_DXR8004W				"DXR8004W: The supplied list of AD Domains is too long, it will be truncated."

//  Error Messages
#define		MSG_DXR8013E				"DXR8013E: An error was encountered while attempting to lookup '%s'."
#define		MSG_DXR8017E				"DXR8017E: LDAP operation: %s failed with RC = 0x%.4x - %s."
#define		MSG_DXR8018E				"DXR8018E: Unable to determine matches for query '%s'."
#define		MSG_DXR8021E				"DXR8021E: Failed to get a handle to an LDAP search results entry for query '%s'."
#define		MSG_DXR8028E				"DXR8028E: Domino failed to parse '%s'."
#define		MSG_DXR8029E				"DXR8029E: Domino failed to abbreviate '%s'."

//  Class Definition for the NEResolver Class
class NEResolver
{
public:
	//  Default Constructor
	NEResolver(void);

	//  Normal Constructor
	NEResolver(ExecEnvironment *xeParent, int iThreadID);

	//  Destructor
	~NEResolver(void);

	//  Public Members
	BOOL	ShowDuplicates;															//  Display duplicate entries

	//  Public Functions

	//  Environment Conditioning Functions
	BOOL setDomRefServer(char *szServer);											//  Set the Domino Reference Server Name
	void setLDAPSessionHandle(LDAP *hlsSet);										//  Set the LDAP Session Handle
	BOOL setLDAPBaseDN(char *szDN);													//  Set the Base DN for LDAP searches
	LDAP *getLDAPSessionHandle(void);												//  Get the current LDAP Session Handle

	//  Identity Mapping Functions
	int getAffinity(char *szAID, int iThreadID);									//  Get the affinity of an arbitrary Identity
	int	getDomDN(char *szAID, char *szDomDN, int iDomDNSize, int iThreadID);		//  Get the Domino Distinguished Name from an arbitrary Identity
	int getDomAN(char *szAID, char *szDomAN, int iDomANSize, int iThreadID);		//  Get the Domino Abbreviated Name from an arbitrary Identity
	int getDomCN(char *szAID, char *szDomCN, int iDomCNSize, int iThreadID);		//  Get the Domino Common Name from an arbitrary Identity
	int	getADDN(char *szAID, char *szADDN, int iADDNSize, int iThreadID);			//  Get the AD Distinguished Name from an arbitrary Identity
	int getADCN(char *szAID, char *szADCN, int iADCNSize, int iThreadID);			//  Get the AD Common Name from an arbitrary Identity
	int getADUID(char *szAID, char *szADUID, int iADUIDSize, int iThreadID);		//  Get the AD UID from an arbitrary Identity
	int getADLogin(char *szAID, char *szADLogin, int iADLoginSize, int iThreadID);  //  Get the AD Login from an arbitrary Identity
	int getEmailAddress(char *szAID, char *szEMAIL, int iEMSize, int iThreadID);	//  Get the E-Mail address from an arbitrary Identity

private:
	//  Private Members
	int							iParentThread;									//  ID of the owning thread
	NECache						*necCurrent;									//  Current cache of Named Entities
	LDAP						*hlsSession;									//  LDAP Session handle
	char						szDomRefServer[MAX_SERVER + 1];					//  Domino Reference Server Name
	char						szBaseDN[MAX_MSG + 1];							//  Base DN for LDAP searches
	char						szMatchDN[MAX_MSG + 1];							//  DN pattern to be matched for returns

	//  Private Functions
	void resolveAffinity(NamedEntity *neResolve, int iThreadID);				//  Resolve the affinity of the arbitrary ID
	void getASP(char *szAS, char *szASP, int iPatternSize);						//  Get the Arbitrary String Pattern
	BOOL isSpecialCharacter(char cIn);											//  Determine if the supplied character is special
	BOOL extractAtom(char *szTarget, int iTargetSize, char *szSource, char *szPattern, int iToken);				//  Extract an atom from a string
	void dumpNamedEntity(NamedEntity *neDump, int iThreadID);					//  Dump the named entity
	void setVirtualBaseDN(char *szDN);											//  Set the Base DN and the Match DN.
	ULONG getFilteredMatchCount(LDAPMessage *lmRaw, int iThreadID);				//  Get the count of filtered search matches
	LDAPMessage *getFirstFilteredMatch(LDAPMessage *lmRaw, int iThreadID);		//  Get the first filtered match from a search
	LDAPMessage *getNextFilteredMatch(LDAPMessage *lmRaw, int iThreadID);		//  Get the next filtered match from a search

protected:
	//  Protrcted Members
	ExecEnvironment				*xeLocal;										//  Current Runtime

};



