//  ThreadDispatcher
//
//  This class provides the dispatching function for requests for execution that are submitted to the
//  the Thread Pool manager queues.
//
//  Message Identifier:      DXR
//  Message Range:           2501 - 2599
//  Trace area assignment:   TRACE_MT (02)
//  DX Version:				 3.12.0
//  Version:                 3.12.0  Build: 106
//
//  Version History:
//
//  1.0.0		07/06/2003		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.1		30/09/2009		Fixed stats & counter integrity for dispatcher
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.9.0		24/07/2010		Linkage changed for logging & request posting and rejoining
//  3.9.1		02/09/2010		Changes to the ACL Processors
//								Fix for false RJR_NONE_EXIST signals
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2003-2010
//

//  Application Includes
#include "ThreadDispatcher.h"

//  Constructor
ThreadDispatcher::ThreadDispatcher(ExecEnvironment *xeParent, UCHAR *pTP, ThreadManagerPolicy *tmpParent)
: NextThread(0)
{
	char			szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Save the Runtime & policy addresses
	xeLocal = xeParent;
	tmpLocal = tmpParent;

	//  Save the Thread Pool address
	pThreadPool = pTP;

	//  Trace the construction 
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2505T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg);
}

//  Destructor
ThreadDispatcher::~ThreadDispatcher(void)
{
	char			szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Trace the destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2506T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg);
}

//  Start
//
//  This function will start the dispatching process.
//
//  Parameters
//
//  Returns
//
//  UINT  -			Status of exiting from the dispatching loop
//
//  API Use Notes:
//
//
UINT ThreadDispatcher::Start(void)
{
	STATUS						stAPIRC;								//  Return code from Notes API calls
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char						szMsg[MAX_MSG];							//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header

	//  Perform any pre-work and mark the thread as not started
	NextThread = 0;
	stAPIRC = NotesInitThread();
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR2507E, DISPATCHER_THREAD_ID);		//  Show the error
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, DISPATCHER_THREAD_ID);				//  Show the API error message
		pThreadPoolHeader->TDispState = TDSTATE_DRAINED;
#ifdef WIN32
		_endthreadex(THREAD_RC_ERROR);									//  Terminate the thread
#endif

#ifdef UNIX
		pthread_exit(NULL);
#endif
		return THREAD_RC_OK;
	}

	pThreadPoolHeader->TDispState = TDSTATE_NOTSTARTED;
	//  Trace the state change  T:5
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2500T, DISPATCHER_THREAD_ID);				//  Show the state change

	while (pThreadPoolHeader->TDispState == TDSTATE_NOTSTARTED)
	{
		SLEEP(IDLE_WAITINTERVAL);
	}

	//  Check that the transition requested is to the started state
	if (pThreadPoolHeader->TDispState == TDSTATE_START)
	{
		pThreadPoolHeader->TDispState = TDSTATE_RUNNING;
		//  Trace the state change
		xeLocal->LogTrace(TRACE_MT, MSG_DXR2501T, DISPATCHER_THREAD_ID);			//  Show the state change
	}

	//  Running state loop - monitor the Request ready queue and dispatch available entries
	while ((pThreadPoolHeader->TDispState == TDSTATE_RUNNING) || (pThreadPoolHeader->TDispState == TDSTATE_BLOCKED))
	{
		MonitorDispSemaphores();								//  Check the request dispatch semaphores for posted requests
		UpdateAvailableThreadCount(pThreadPool);				//  Update the available thread counts
		Dispatch(pThreadPool);									//  Dispatch any ready requests
		MonitorRejoinSemaphores();								//  Check the rejoin semaphores for posted requests
		MonitorRejoinPollSemaphores();							//  Check the rejoin poll semaphores
		SLEEP(tmpLocal->DispatcherWaitCycle);
	}

	//  Check that the transition requested is to the quit state
	if (pThreadPoolHeader->TDispState == TDSTATE_QUIT)
	{
		pThreadPoolHeader->TDispState = TDSTATE_QUITTING;
		//  Trace the state change
		xeLocal->LogTrace(TRACE_MT, MSG_DXR2502T, DISPATCHER_THREAD_ID);				//  Show the state change
	}
	else
	{
		pThreadPoolHeader->TDispState = TDSTATE_QUITTING;
		xeLocal->LogMessage(MSG_DXR2503E, DISPATCHER_THREAD_ID);				//  Show inconsistent state change
	}

	NotesTermThread();
	
	//  Mark the thread as drained and terminate the thread
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2508T, DISPATCHER_THREAD_ID);				//  Show the state change
	pThreadPoolHeader->TDispState = TDSTATE_DRAINED;
#ifdef WIN32
		_endthreadex(THREAD_RC_OK);									//  Terminate the thread
#endif

#ifdef UNIX
		pthread_exit(NULL);
#endif
	return THREAD_RC_OK;
}

//  StartWrapper
//
//  This function is the static wrapper for the Start() function allowing it to be started as an independent thread.
//
//  Parameters
//
//  Returns
//
//  UINT  -			Status indication  
//
//  API Use Notes:
//
//
#ifdef WIN32
UINT LNPUBLIC ThreadDispatcher::StartWrapper(void * pObject)
#endif
#ifdef UNIX
void * LNPUBLIC ThreadDispatcher::StartWrapper(void * pObject)
#endif
{
	ThreadDispatcher		*tdThis;

	//  Wrapper function calls the Start function in the Object
	tdThis = (ThreadDispatcher *) pObject;
#ifdef WIN32
	return tdThis->Start();
#endif
#ifdef UNIX
	return (void *) tdThis->Start();
#endif
}

//  Dispatch
//
//  This function provides the main dispatching loop for the Thread Pool Manager. As worker threads
//  become free they will be allocated a new unit of work to do on a priority basis.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//  API Use Notes:
//  
//

void ThreadDispatcher::Dispatch(UCHAR *pTP)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	RQElement				*pRQE;									//  Pointer to a Request Element (RQE)
	RQElement				*pRQEPriority;							//  Pointer to highest priority RQE
	int						iRQEIndex;								//  Index into Request Elements
	int						iRQEPriority;							//  Highest Priority located
	int						iDispWindow;							//  Dispatches in the current window
	BOOL					bCarryOnDispatching;					//  Flag controlling dispatch loop
	UINT					uiBlockedLevels;						//  Request levels that are blocked
	char					szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Dispatch one or more requests to available threads
	pThreadPoolHeader = (Threadpool_Header *) pTP;

	//  Exit is there is no work to do or we are currently blocked
	if ((pThreadPoolHeader->PendRQECount == 0) || (pThreadPoolHeader->TDispState == TDSTATE_BLOCKED))
	{
		pThreadPoolHeader->NoDispNoWork++;							//  Increment the stats counter
		return;
	}

	//  If tracing the dispatcher then show the current settings
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2509T, pThreadPoolHeader->PendRQECount, pThreadPoolHeader->AvailableThreads);
	xeLocal->LogTrace(TRACE_MTD, szMsg, DISPATCHER_THREAD_ID);

	//  Determine if there are any available threads
	if (pThreadPoolHeader->AvailableThreads == 0)
	{
		pThreadPoolHeader->NoDispThreadsBusy++;						//  Increment the stats counter
		return;
	}

	uiBlockedLevels = 0;											//  Nothing blocked
	//  If CMLS is active then compute the blocked levels
	if (pThreadPoolHeader->SchedMode & TPOOL_MODE_CMLS)
	{
		if (pThreadPoolHeader->CurrL0Threads >= pThreadPoolHeader->MaxL0Threads) uiBlockedLevels = uiBlockedLevels | RQATTR_CMLSL0;
		if (pThreadPoolHeader->CurrL1Threads >= pThreadPoolHeader->MaxL1Threads) uiBlockedLevels = uiBlockedLevels | RQATTR_CMLSL1;
		if (pThreadPoolHeader->CurrL2Threads >= pThreadPoolHeader->MaxL2Threads) uiBlockedLevels = uiBlockedLevels | RQATTR_CMLSL2;
	}

	//  Determine if we need to perform priority selection on the available requests
	if (pThreadPoolHeader->AvailableThreads >= pThreadPoolHeader->PendRQECount)
	{
		iDispWindow = pThreadPoolHeader->PendRQECount;
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2510T, pThreadPoolHeader->PendRQECount, pThreadPoolHeader->AvailableThreads);
		xeLocal->LogTrace(TRACE_MT, szMsg, DISPATCHER_THREAD_ID);
		pThreadPoolHeader->BFDispatches++;
		//  Perform a brute force dispatch of all ready requests
		pRQE = (RQElement *) pThreadPoolHeader->pPendRQEPool;
		for (iRQEIndex = 0; iRQEIndex < pThreadPoolHeader->PendRQEPoolSize; iRQEIndex++)
		{
			if (pRQE->RQEState & RQESTATE_READY)
			{
				//  Only dispatch the request if it is not for a blocked level
				if (!(pRQE->RQEAttrs & uiBlockedLevels))
				{
					//  Dispatch this Request Element
					if (!DispatchRQE(pRQE, pTP))
					{
						pThreadPoolHeader->BFDispThreadsBusy++;						//  Increment the stats counter
						return;
					}
					else
					{
						pThreadPoolHeader->DispBruteForce++;						//  Increment the count dispatched

						pThreadPoolHeader->PendRQECount--;							//  and one less request to service
						iDispWindow--;
						if (pThreadPoolHeader->SchedMode & TPOOL_MODE_CMLS)
						{
							//  recompute the blocked mask
							uiBlockedLevels = 0;											//  Nothing blocked
							if (pThreadPoolHeader->CurrL0Threads >= pThreadPoolHeader->MaxL0Threads) uiBlockedLevels = uiBlockedLevels | RQATTR_CMLSL0;
							if (pThreadPoolHeader->CurrL1Threads >= pThreadPoolHeader->MaxL1Threads) uiBlockedLevels = uiBlockedLevels | RQATTR_CMLSL1;
							if (pThreadPoolHeader->CurrL2Threads >= pThreadPoolHeader->MaxL2Threads) uiBlockedLevels = uiBlockedLevels | RQATTR_CMLSL2;
						}
					}
				}
				else pThreadPoolHeader->CMLSDenials++;
			}
			pRQE++;
			if (iDispWindow == 0)
			{
				return;
			}
		}
		return;
	}

	//  Priority selection of available requests
	bCarryOnDispatching = true;
	iDispWindow = 0;
	pThreadPoolHeader->PDispatches++;
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2511T, DISPATCHER_THREAD_ID);

	while (bCarryOnDispatching)
	{
		iRQEPriority = -1;												//  Base Priority
		pRQEPriority = NULL;											//  Best RQE
		
		//  Find the highest priority Request Elelement in the pending pool
		pRQE = (RQElement *) pThreadPoolHeader->pPendRQEPool;
		for (iRQEIndex = 0; iRQEIndex < pThreadPoolHeader->PendRQEPoolSize; iRQEIndex++)
		{
			if (pRQE->RQEState == RQESTATE_READY)
			{
				//  Only consider the request if it is not for a blocked level
				if (!(pRQE->RQEAttrs & uiBlockedLevels))
				{
					//  Higher priority RQE located store it's priority and address
					if (pRQE->RQEPriority > iRQEPriority)
					{
						iRQEPriority = pRQE->RQEPriority;
						pRQEPriority = pRQE;
					}
				}
				else pThreadPoolHeader->CMLSDenials++;
			}
			pRQE++;
		}
		if (pRQEPriority == NULL)
		{
			pThreadPoolHeader->PDispQEmpty++;							//  Increment the stats counter
			return;
		}

		//  Dispatch the highest priority Request Element
		if (!DispatchRQE(pRQEPriority, pTP))
		{
			pThreadPoolHeader->PDispThreadsBusy++;						//  Increment the stats counter
			return;
		}
		pThreadPoolHeader->DispPriority++;								//  Increment the stats counter

		pThreadPoolHeader->PendRQECount--;								//  and one less reqtest to service
		if (pThreadPoolHeader->SchedMode & TPOOL_MODE_CMLS)
		{
			//  recompute the blocked mask
			uiBlockedLevels = 0;											//  Nothing blocked
			if (pThreadPoolHeader->CurrL0Threads >= pThreadPoolHeader->MaxL0Threads) uiBlockedLevels = uiBlockedLevels | RQATTR_CMLSL0;
			if (pThreadPoolHeader->CurrL1Threads >= pThreadPoolHeader->MaxL1Threads) uiBlockedLevels = uiBlockedLevels | RQATTR_CMLSL1;
			if (pThreadPoolHeader->CurrL2Threads >= pThreadPoolHeader->MaxL2Threads) uiBlockedLevels = uiBlockedLevels | RQATTR_CMLSL2;
		}

		iDispWindow++;
		if (iDispWindow == tmpLocal->DispatcherWindowLimit) bCarryOnDispatching = FALSE;
	}
	return;
}

//  DispatchRQE
//
//  This function will dispatch a single request to an available worker thread.
//
//  Parameters
//
//	1  -  RQElement *				-  Pointer to the Request Element to be dispatched
//	2  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//  BOOL			-  TRUE if the request could be dispatched
//					-  FALSE if the request could not be dispatched
//
//  API Use Notes:
//  
//
BOOL ThreadDispatcher::DispatchRQE(RQElement *rqeThis, UCHAR *pTP)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member		*pMyThread;								//  Pointer to the selected thread
	int						iSelectedThread;						//  Thread Selected for dispatch
	char					szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Dispatch the passed RQE to the next available thread
	pThreadPoolHeader = (Threadpool_Header *) pTP;

	//  Find an available thread
	pMyThread = FindAvailableThread(pTP, &iSelectedThread);
	if (pMyThread == NULL) return FALSE;							//  Unable to find a thread to dispatch to

	//  Re-check that the RQE is in the ready state
	if (rqeThis->RQEState != RQESTATE_READY)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2515E, rqeThis, rqeThis->RQEState);
		xeLocal->LogMessage(szMsg, DISPATCHER_THREAD_ID);
		return FALSE;
	}

	//  Copy the RQE to the Thread Member
	memcpy(&pMyThread->rqePosted, rqeThis, sizeof(RQElement));

	//  Update the processed count
	pMyThread->dwReqsProc++;

	//  Mark the thread as posted
	pMyThread->TMemState = WTSTATE_POSTED;

	//  Increment the CMLS lane counts
	if (pThreadPoolHeader->SchedMode & TPOOL_MODE_CMLS)
	{
		IncrementCMLSThreadCount(pTP, rqeThis->RQEAttrs);
		DecrementCMLSPoolCount(pTP, rqeThis->RQEAttrs);
	}

	//  Trace the dispatch
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2512T, rqeThis, iSelectedThread, rqeThis->RQEPriority);
	xeLocal->LogTrace(TRACE_MTD, szMsg, DISPATCHER_THREAD_ID);

	//  Free the Request Elelemnt
	rqeThis->RQEState = RQESTATE_UNALLOCATED;
	return TRUE;
}

//  FindAvailableThread
//
//  This function will locate and return the address of an available Thread Pool Member.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  int						-  ID of the thread that is making the request
//
//  Returns
//
//  BOOL			-  TRUE if the request could be dispatched
//					-  FALSE if the request could not be dispatched
//
//  API Use Notes:
//  
//
Threadpool_Member * ThreadDispatcher::FindAvailableThread(UCHAR * pTP, int *iThread)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	UCHAR					*pStartThread;							//  Pointer to the initial thead pool member
	Threadpool_Member		*pMyThread;								//  Pointer to the selected thread
	int						iInspected;								//  Count of threads inspected
	char					szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Find the next available thread
	pThreadPoolHeader = (Threadpool_Header *) pTP;
	pStartThread = pTP + (sizeof(Threadpool_Header) + (NextThread * sizeof(Threadpool_Member)));
	pMyThread = (Threadpool_Member *) pStartThread;
	iInspected = 0;
	while (iInspected < pThreadPoolHeader->MaxThreads)
	{
		if (pMyThread->TMemState == WTSTATE_RUNNING)
		{
			*iThread = NextThread + 1;
			NextThread++;
			if (NextThread == pThreadPoolHeader->MaxThreads) NextThread = 0;
			return pMyThread;
		}
		iInspected++;
		pMyThread++;
		NextThread++;
		if (NextThread == pThreadPoolHeader->MaxThreads)
		{
			NextThread = 0;
			pMyThread = (Threadpool_Member *) (pTP + sizeof(Threadpool_Header));
		}
	}
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2513T, pThreadPoolHeader->MaxThreads, NextThread, iInspected);
	xeLocal->LogTrace(TRACE_MTD, szMsg, DISPATCHER_THREAD_ID);
	return NULL;
}

//  UpdateAvailableThreadCount
//
//  This function will recompute the available thread count also re-computing the CMLS counts
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadDispatcher::UpdateAvailableThreadCount(UCHAR *pTP)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	UCHAR					*pStartThread;							//  Pointer to the initial thead pool member
	Threadpool_Member		*pMyThread;								//  Pointer to the selected thread
	int						iATCount = 0;							//  Available thread count
	int						iIndex;									//  Generic Index

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	pThreadPoolHeader->CurrL0Threads = 0;
	pThreadPoolHeader->CurrL1Threads = 0;
	pThreadPoolHeader->CurrL2Threads = 0;

	pStartThread = pTP + sizeof(Threadpool_Header);
	pMyThread = (Threadpool_Member *) pStartThread;

	for (iIndex = 0; iIndex < pThreadPoolHeader->MaxThreads; iIndex++)
	{
		if (pMyThread->TMemState == WTSTATE_RUNNING) iATCount++;
		else
		{
			if (pMyThread->TMemState & (WTSTATE_POSTED | WTSTATE_BUSY | WTSTATE_REJOIN))
			{
				if (pThreadPoolHeader->SchedMode & TPOOL_MODE_CMLS) IncrementCMLSThreadCount(pTP, pMyThread->rqePosted.RQEAttrs);
			}
		}
		pMyThread++;
	}

	pThreadPoolHeader->AvailableThreads = iATCount;

	return;
}

//  MonitorDispSempahores
//
//  This function will monitor the per thread request dispatch semaphores, if posted, the request will be
//  posted to the request pending pool and the semaphore will be freed.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadDispatcher::MonitorDispSemaphores(void)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a threadpool member
	int							iIndex;									//  Generic Index

	//  Check if there is anything posted in the Thread Member - if so action it
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header

	for (iIndex = 0; iIndex < (pThreadPoolHeader->MaxThreads + 4); iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) (pThreadPool + sizeof(Threadpool_Header) + (iIndex * sizeof(Threadpool_Member)));
		if (pThreadPoolMember->semReqDisp != REQ_SEM_FREE)
		{
			ProcessDispSemaphore(pThreadPoolMember);
		}
	}
	return;
}

//  ProcessDispSempahore
//
//  This function will move a posted dispatch request to the request pending pool and free the semaphore.
//
//  Parameters
//
//  1  -  Pointer to the Thread Pool Member that has a posted semaphore
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadDispatcher::ProcessDispSemaphore(Threadpool_Member *pTPM)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	RQElement				*pMyRQE;								//  Pointer to the Request Element
	char					szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Post a message to the pending pool
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;			//  Address the header

	//  If the RQE Pending pool is full then skip this (for now)
	if (pThreadPoolHeader->PendRQECount == pThreadPoolHeader->PendRQEPoolSize)
	{
		pThreadPoolHeader->PPFullBackoffs++;
		return;
	}

	//  If CMLS is active and the lane is full then skip this (for now)
	if (pThreadPoolHeader->SchedMode & TPOOL_MODE_CMLS)
	{
		if (pTPM->rqeDisp.RQEAttrs & RQATTR_CMLSLCONSTRAINED)
		{
			if (IsCMLSPoolLaneFull(pThreadPool, pTPM->rqeDisp.RQEAttrs))
			{
				pThreadPoolHeader->PPLaneFullBackoffs++;
				return;
			}
		}
	}

	//  Find and reserve a free RQE in the pending pool
	pMyRQE = FindAndReserveRQE(pThreadPoolHeader);					//  Find and reserve the RQE
	if (pMyRQE == NULL) 
	{
		pThreadPoolHeader->PPFullBackoffs++;
		return;										//  Pool is full - just return & try later
	}

	//  Post information to the request
	pMyRQE->pOObject = pTPM->rqeDisp.pOObject;
	pMyRQE->pXObject = pTPM->rqeDisp.pXObject;
	pMyRQE->pPObject = pTPM->rqeDisp.pPObject;
	pMyRQE->RQEPriority = pTPM->rqeDisp.RQEPriority;
	pMyRQE->RQEAttrs = pTPM->rqeDisp.RQEAttrs;

	//  Trace the event
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2518T, pMyRQE, pMyRQE->pOObject, pMyRQE->pPObject, pMyRQE->pXObject);
	xeLocal->LogTrace(TRACE_MRQ, szMsg);


	//  Update counts and mark the RQE ready to be processed
	pThreadPoolHeader->PendRQECount++;									//  Increment the count in the pool
	pMyRQE->RQEInspectCount = 0;										//  Clear the inspection count
	pMyRQE->RQEState = RQESTATE_READY;									//  Mark the RQE state as Ready
	IncrementCMLSPoolCount(pThreadPool, pTPM->rqeDisp.RQEAttrs);		//  Increment CMLS lane counts
	
	//  Clear the semaphore
	pTPM->semReqDisp = REQ_SEM_FREE;
	return;
}

//  FindAndReserveRQE
//
//  This function will locate, reserve and return the address of a Request Entry from the exec request
//  pool.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  int						-  Thread ID of the requester
//
//  Returns
//
//  RQElement *		-	Pointer to the reserverd RQElement
//					-	NULL an element could not be reserved
//
//  API Use Notes:
//   
RQElement * ThreadDispatcher::FindAndReserveRQE(Threadpool_Header *pTPH)
{
	int				iIndex;												//  Index of RQ elements
	RQElement		*pCandidateRQE;										//  Pointer to RQE slot

	//  Find a free RQE slot in the RQE Pending pool and allocate it
	pCandidateRQE = (RQElement *) pTPH->pPendRQEPool;					//  Point to the first element

	for (iIndex = 0; iIndex < pTPH->PendRQEPoolSize; iIndex++)
	{
		if (pCandidateRQE->RQEState == RQESTATE_UNALLOCATED)
		{
			pCandidateRQE->RQEState = RQESTATE_ALLOCATED;				//  Allocate the entry
			pCandidateRQE->RQEInspectCount = 0;							//  Clear the inspection count
			//  re-check the state variable
			if (pCandidateRQE->RQEState == RQESTATE_ALLOCATED)
			{
				//  Return the selected RQE - keep holding the RQE Pending Pool Mutex
				return pCandidateRQE;
			}
		}
		pCandidateRQE++;
	}
	return NULL;
}

//  MonitorRejoinSempahores
//
//  This function will monitor the per thread rejoin semaphores, if posted, the request will be
//  posted to the rejoin and the semaphore will be freed.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadDispatcher::MonitorRejoinSemaphores(void)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a threadpool member
	int							iIndex;									//  Generic Index

	//  Check if there is anything posted to add to the Rejoin Pool - if so action it
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header

	for (iIndex = 0; iIndex < (pThreadPoolHeader->MaxThreads + 4); iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) (pThreadPool + sizeof(Threadpool_Header) + (iIndex * sizeof(Threadpool_Member)));
		if (pThreadPoolMember->semRejoin != REJ_SEM_FREE)
		{
			ProcessRejoinSemaphore(pThreadPoolMember);
		}
	}
	return;
}

//  ProcessRejoinSempahore
//
//  This function will move a posted rejoin to the request rejoin pool and free the semaphore.
//
//  Parameters
//
//  1  -  Pointer to the Thread Pool Member that has a posted semaphore
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadDispatcher::ProcessRejoinSemaphore(Threadpool_Member *pTPM)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	RQElement				*pMyRQE;								//  Pointer to the Request Element
	char					szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Post a request to the rejoin pool
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;			//  Address the header

	//  Find and reserve a free RQE in the rejoin pool
	pMyRQE = FindAndReserveRejoinRQE(pThreadPoolHeader);			//  Find and reserve the RQE
	if (pMyRQE == NULL) return;										//  Pool is full - just return & try later

	//  Post information to the request
	pMyRQE->pOObject = pTPM->rqeRejoin.pOObject;
	pMyRQE->pXObject = pTPM->rqeRejoin.pXObject;
	pMyRQE->pPObject = pTPM->rqeRejoin.pPObject;
	pMyRQE->RQEPriority = pTPM->rqeRejoin.RQEPriority;
	pMyRQE->RQEAttrs = pTPM->rqeRejoin.RQEAttrs;

	//  Trace the event
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2516T, pMyRQE, pMyRQE->pOObject, pMyRQE->pPObject, pMyRQE->pXObject);
	xeLocal->LogTrace(TRACE_MRQ, szMsg);

	//  Update counts and mark the RQE ready to be processed
	pThreadPoolHeader->RejoinRQECount++;								//  Increment the count in the pool
	pMyRQE->RQEInspectCount = 0;										//  Clear the inspection count
	pMyRQE->RQEState = RQESTATE_READY;									//  Mark the RQE state as Ready

	//  Clear the semaphore
	pTPM->semRejoin = REJ_SEM_FREE;
	return;
}

//  MonitorRejoinPollSempahores
//
//  This function will monitor the per thread rejoin poll semaphores, if posted, the rejoin pool
//  will be checked for any available returns - if not any waiting or executing requests will be checked.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//  The return code from processing is posted into the appropriate thread pool member along with any
//  returned data (If RJR_RETURNED is set).
//  
//
void ThreadDispatcher::MonitorRejoinPollSemaphores(void)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a threadpool member
	int							iIndex;									//  Generic Index

	//  Check if there is anything posted to check fore rejoins - if so action it
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header

	for (iIndex = 0; iIndex < (pThreadPoolHeader->MaxThreads + 4); iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) (pThreadPool + sizeof(Threadpool_Header) + (iIndex * sizeof(Threadpool_Member)));
		if (pThreadPoolMember->semRejoinPoll != REJPOLL_SEM_FREE)
		{
			ProcessRejoinPollSemaphore(pThreadPoolMember);
		}
	}
	return;
}

//  ProcessRejoinPollSempahore
//
//  This function will check the rejoin queues and return any appropriate RQEs.
//  If none are found then executing & pending requests are checked to determine the appropriate return.
//
//  Parameters
//
//  1  -  Pointer to the Thread Pool Member that has a posted semaphore
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadDispatcher::ProcessRejoinPollSemaphore(Threadpool_Member *pTPM)
{
	int						iIndex;									//  Index of RQ elements
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member		*pAnyThread;							//  Pointer to the selected thread
	RQElement				*pCandidateRQE;							//  Pointer to the Request Element
	char					szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Check for & return any ready for rejoin RQEs
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;			//  Address the header

	//  Check the rejoin pool to see if there is a matching element
	pCandidateRQE = (RQElement *) pThreadPoolHeader->pRejoinRQEPool;	//  Point to the first element
	for (iIndex = 0; iIndex < pThreadPoolHeader->RejoinRQEPoolSize; iIndex++)
	{
		if ((pCandidateRQE->RQEState == RQESTATE_READY) && (pCandidateRQE->pOObject == pTPM->rqeRejoinPoll.pOObject))
		{
			//  There is an available element already in the rejoin pool - post the return info
			//  Then clear the semaphore
			pCandidateRQE->RQEState = RQESTATE_ALLOCATED;				//  Prevent it being collected by another
			memcpy(&pTPM->rqeRejoinPoll, pCandidateRQE, sizeof(RQElement));

			//  Trace the event
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2517T, pCandidateRQE, pTPM->rqeRejoinPoll.pOObject, pTPM->rqeRejoinPoll.pPObject, pTPM->rqeRejoinPoll.pXObject);
			xeLocal->LogTrace(TRACE_MRQ, szMsg);

			pCandidateRQE->RQEState = RQESTATE_UNALLOCATED;				//  Free the entry
			pThreadPoolHeader->RejoinRQECount--;						//  Reduce the count in the pool
			pThreadPoolHeader->RJRDequeued++;							//  Maintain the stats
			pTPM->rcRejoinPoll = RJR_RETURNED;
			pTPM->dwRejoinReturned++;
			pTPM->semRejoinPoll = REJPOLL_SEM_FREE;
			return;
		}
		pCandidateRQE++;
	}
	
	//  If there is a request in the RQE pending pool or allocated to a thread then return not ready
	//  First check if there are any executing or posted for the pending pool

	//  Step 1:  Check the worker threads
	pAnyThread = (Threadpool_Member *) (pThreadPool + sizeof(Threadpool_Header));
	for (iIndex = 0; iIndex < pThreadPoolHeader->MaxThreads; iIndex++)
	{
		if ((pAnyThread->TMemState == WTSTATE_BUSY) || (pAnyThread->TMemState == WTSTATE_POSTED) || (pAnyThread->TMemState == WTSTATE_REJOIN))
		{
			if ((pAnyThread->rqePosted.pOObject == pTPM->rqeRejoinPoll.pOObject) && (pAnyThread->rqePosted.RQEAttrs & RQATTR_REJOIN))
			{
				pTPM->dwRejoinNotReady++;
				pTPM->rcRejoinPoll = RJR_NONE_READY;
				pTPM->semRejoinPoll = REJPOLL_SEM_FREE;
				return;
			}
			//  Now check if there is a request posted for the pending execution queue
			if ((pAnyThread->semReqDisp != REQ_SEM_FREE) && (pAnyThread->rqeDisp.pOObject == pTPM->rqeRejoinPoll.pOObject))
			{
				pTPM->dwRejoinNotReady++;
				pTPM->rcRejoinPoll = RJR_NONE_READY;
				pTPM->semRejoinPoll = REJPOLL_SEM_FREE;
				return;
			}
		}
		pAnyThread++;
	}

	//  Step 2:  Check the control thread
	pAnyThread = (Threadpool_Member *) GetMemberAddress(0);

	//  Now check if there is a request posted for the pending execution queue
	if ((pAnyThread->semReqDisp != REQ_SEM_FREE) && (pAnyThread->rqeDisp.pOObject == pTPM->rqeRejoinPoll.pOObject))
	{
		pTPM->dwRejoinNotReady++;
		pThreadPoolHeader->Guard02++;
		pTPM->rcRejoinPoll = RJR_NONE_READY;
		pTPM->semRejoinPoll = REJPOLL_SEM_FREE;
		return;
	}

	//  If the Dipatcher has been blocked then we are in Fast Drain Mode - in which case ignore any
	//  requests that are in the ready pool as they will never be dispatched
	if (pThreadPoolHeader->TDispState == TDSTATE_BLOCKED) 
	{
		pTPM->dwRejoinNone++;
		pTPM->rcRejoinPoll = RJR_NONE_EXIST;
		pTPM->semRejoinPoll = REJPOLL_SEM_FREE;
		return;
	}

	//  Now check the ready pool
	pCandidateRQE = (RQElement *) pThreadPoolHeader->pPendRQEPool;		//  Point to the first element
	for (iIndex = 0; iIndex < pThreadPoolHeader->PendRQEPoolSize; iIndex++)
	{
		if ((pCandidateRQE->RQEState == RQESTATE_READY) && (pCandidateRQE->pOObject == pTPM->rqeRejoinPoll.pOObject) && (pCandidateRQE->RQEAttrs & RQATTR_REJOIN))
		{
			pThreadPoolHeader->RJRBusy++;
			pTPM->dwRejoinNotReady++;
			pTPM->rcRejoinPoll = RJR_NONE_READY;
			pTPM->semRejoinPoll = REJPOLL_SEM_FREE;
			return;
		}
		pCandidateRQE++;
	}

	//  Signal none exist
	pTPM->dwRejoinNone++;
	pTPM->rcRejoinPoll = RJR_NONE_EXIST;
	pTPM->semRejoinPoll = REJPOLL_SEM_FREE;

	//  return to caller
	return;
}

//  FindAndReserveRejoinRQE
//
//  This function will locate, reserve and return the address of a Request Entry from the rejoin
//  pool.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  int						-  Thread ID of the requester
//
//  Returns
//
//  RQElement *		-	Pointer to the reserverd RQElement
//					-	NULL an element could not be reserved
//
//  API Use Notes:
//   
RQElement * ThreadDispatcher::FindAndReserveRejoinRQE(Threadpool_Header *pTPH)
{
	int				iIndex;												//  Index of RQ elements
	RQElement		*pCandidateRQE;										//  Pointer to RQE slot

	//  Find a free RQE slot in the RQE Rejoin pool and allocate it
	pCandidateRQE = (RQElement *) pTPH->pRejoinRQEPool;					//  Point to the first element

	for (iIndex = 0; iIndex < pTPH->RejoinRQEPoolSize; iIndex++)
	{
		if (pCandidateRQE->RQEState == RQESTATE_UNALLOCATED)
		{
			pCandidateRQE->RQEState = RQESTATE_ALLOCATED;				//  Allocate the entry
			pCandidateRQE->RQEInspectCount = 0;							//  Clear the inspection count
			//  re-check the state variable
			if (pCandidateRQE->RQEState == RQESTATE_ALLOCATED)
			{
				//  Return the selected RQE
				return pCandidateRQE;
			}
		}
		pCandidateRQE++;
	}
	return NULL;
}

//  GetMemberAddress
//
//  This function will return the address of the indicated thread pool member
//
//  Parameters
//
//  1		-		int			-  id of the referenced thread 
//
//  Returns
//
//  API Use Notes:
//
//
UCHAR * ThreadDispatcher::GetMemberAddress(int iThreadID)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	UCHAR					*pTPM;									//  Pointer to a Thread Pool Member
	int						iThreadIndex;							//  Thread Index
	
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;			//  Address the header

	//  Adjust the thread ID if it is one of the control or management threads
	switch(iThreadID)
	{
	case CONTROL_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads;
		break;
	case MONITOR_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 1;
		break;
	case SCHEDULER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 2;
		break;
	case DISPATCHER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 3;
		break;
	default:
		iThreadIndex = iThreadID - 1;
	}

	//  Compute the address of the referenced member
	pTPM = pThreadPool + sizeof(Threadpool_Header) + (iThreadIndex * sizeof(Threadpool_Member));
	return pTPM;
}

//  IsCMLSPoolLaneFull
//
//  This function will test the indicated CMLS lane and determine if the Pend Pool is full for this lane
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  UNIT						-  RQE Attributes which contain the lane
//
//  Returns
//
//  BOOL	-	TRUE if the lane is full
//				FALSE if the lane is not full
//
//  API Use Notes:
//
//
BOOL ThreadDispatcher::IsCMLSPoolLaneFull(UCHAR *pTP, UINT Attrs)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	switch(Attrs & (RQATTR_CMLSLCONSTRAINED))
	{
	case RQATTR_CMLSL0:
		if (pThreadPoolHeader->CurrL0PPRQEs >= pThreadPoolHeader->MaxL0PPRQEs) return TRUE;
		break;

	case RQATTR_CMLSL1:
		if (pThreadPoolHeader->CurrL1PPRQEs >= pThreadPoolHeader->MaxL1PPRQEs) return TRUE;
		break;

	case RQATTR_CMLSL2:
		if (pThreadPoolHeader->CurrL2PPRQEs >= pThreadPoolHeader->MaxL2PPRQEs) return TRUE;
		break;
	}
	return FALSE;
}

//  IncrementCMLSPoolCount
//
//  This function will increment the indicated CMLS lane counts
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  UNIT						-  RQE Attributes which contain the lane
//
//  Returns
//
//
//  API Use Notes:
//
//
void ThreadDispatcher::IncrementCMLSPoolCount(UCHAR *pTP, UINT Attrs)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	switch(Attrs & (RQATTR_CMLSLCONSTRAINED))
	{
	case RQATTR_CMLSL0:
		pThreadPoolHeader->CurrL0PPRQEs++;
		break;

	case RQATTR_CMLSL1:
		pThreadPoolHeader->CurrL1PPRQEs++;
		break;

	case RQATTR_CMLSL2:
		pThreadPoolHeader->CurrL2PPRQEs++;
		break;
	}
	return;
}

//  IncrementCMLSThreadCount
//
//  This function will increment the indicated CMLS lane counts
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  UNIT						-  RQE Attributes which contain the lane
//
//  Returns
//
//
//  API Use Notes:
//
//
void ThreadDispatcher::IncrementCMLSThreadCount(UCHAR *pTP, UINT Attrs)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	switch(Attrs & (RQATTR_CMLSLCONSTRAINED))
	{
	case RQATTR_CMLSL0:
		pThreadPoolHeader->CurrL0Threads++;
		break;

	case RQATTR_CMLSL1:
		pThreadPoolHeader->CurrL1Threads++;
		break;

	case RQATTR_CMLSL2:
		pThreadPoolHeader->CurrL2Threads++;
		break;
	}
	return;
}

//  DecrementCMLSPoolCount
//
//  This function will increment the indicated CMLS lane counts
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  UNIT						-  RQE Attributes which contain the lane
//
//  Returns
//
//
//  API Use Notes:
//
//
void ThreadDispatcher::DecrementCMLSPoolCount(UCHAR *pTP, UINT Attrs)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	switch(Attrs & (RQATTR_CMLSLCONSTRAINED))
	{
	case RQATTR_CMLSL0:
		pThreadPoolHeader->CurrL0PPRQEs--;
		break;

	case RQATTR_CMLSL1:
		pThreadPoolHeader->CurrL1PPRQEs--;
		break;

	case RQATTR_CMLSL2:
		pThreadPoolHeader->CurrL2PPRQEs--;
		break;
	}
	return;
}

//  GetCMLSRecommendation
//
//  This function will determine if lane restrictions are in place for the indicated CMLS execution level
//
//  Parameters
//
//  1  -  int		-	The CMLS level 
//
//  Returns
//
//
//  API Use Notes:
//
//
int ThreadDispatcher::GetCMLSRecommendation(UCHAR *pTP, int iLevel, int iThreadID)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	BOOL					bPoolIsAvailable;						//  The pool is available
	BOOL					bThreadIsAvailable;						//  A thread is available

	pThreadPoolHeader = (Threadpool_Header *) pTP;
	bPoolIsAvailable = FALSE;

	//  Firstly determine if there is space in the RQ pool for this lane
	switch(iLevel)
	{
	case RQATTR_CMLSL0:
		if (pThreadPoolHeader->CurrL0PPRQEs >= pThreadPoolHeader->MaxL0PPRQEs) bPoolIsAvailable = TRUE;
		break;

	case RQATTR_CMLSL1:
		if (pThreadPoolHeader->CurrL1PPRQEs >= pThreadPoolHeader->MaxL1PPRQEs) bPoolIsAvailable = TRUE;
		break;

	case RQATTR_CMLSL2:
		if (pThreadPoolHeader->CurrL2PPRQEs >= pThreadPoolHeader->MaxL2PPRQEs) bPoolIsAvailable = TRUE;
		break;
	}
	
	if (!bPoolIsAvailable) return CMLS_LANE_FULL;

	//  Now check if a thread is available for this lane
	bThreadIsAvailable = FALSE;
	switch(iLevel)
	{
	case RQATTR_CMLSL0:
		if (pThreadPoolHeader->CurrL0Threads >= pThreadPoolHeader->MaxL0Threads) bThreadIsAvailable = TRUE;
		break;

	case RQATTR_CMLSL1:
		if (pThreadPoolHeader->CurrL1Threads >= pThreadPoolHeader->MaxL1Threads) bThreadIsAvailable = TRUE;
		break;

	case RQATTR_CMLSL2:
		if (pThreadPoolHeader->CurrL2Threads >= pThreadPoolHeader->MaxL2Threads) bThreadIsAvailable = TRUE;
		break;
	}

	if (!bThreadIsAvailable) return CMLS_LANE_AVAILABLE;
	return CMLS_LANE_AVAILABLE_NOW;
}
