#pragma once
//  ThreadDispatcher
//
//  This class provides the dispatching function for requests for execution that are submitted to the
//  the Thread Pool manager queues.
//
//  Message Identifier:      DXR
//  Message Range:           2501 - 2599
//  Trace area assignment:   TRACE_MT (02)
//  DX Version:				 3.12.0
//  Version:                 3.12.0  Build: 106
//
//  Version History:
//
//  1.0.0		07/06/2003		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.1		30/09/2009		Fixed stats & counter integrity for dispatcher
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.9.0		24/07/2010		Linkage changed for logging & request posting and rejoining
//  3.9.1		02/09/2010		Changes to the ACL Processors
//								Fix for false RJR_NONE_EXIST signals
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2003-2010
//

//  Platform includes
#include "../Platform/PlatBase.h"

//  Application Includes
#include "../ExecEnvironment.h"								//  Runtime
#include "ThreadStructs.h"									//  Thread Structures
#include "ThreadManagerPolicy.h"							//  Thread Manager Policy

//  Local Definitions
#define		IDLE_WAITINTERVAL		100						//  Wait 100 Ms while in the idle state

//  Trace Messages
#define		MSG_DXR2500T			"DXR2500T: The dispatcher thread has entered the not started state."
#define		MSG_DXR2501T			"DXR2501T: The dispatcher thread has entered the running state."
#define		MSG_DXR2502T			"DXR2502T: The dispatcher thread has entered the quitting state."
#define		MSG_DXR2505T			"DXR2505T: A new ThreadDispatcher instance has been created @%p."
#define		MSG_DXR2506T			"DXR2506T: A ThreadDispatcher instance @%p has been destroyed."
#define		MSG_DXR2508T			"DXR2508T: The dispatcher thread has been drained."
#define		MSG_DXR2509T			"DXR2509T: Attempting thread dispatch with Pend Count: %i and Available Threads: %i."
#define		MSG_DXR2510T			"DXR2510T: The thread dispatcher is performing a brute force dispatch for %i RQEs in %i threads."
#define		MSG_DXR2511T			"DXR2511T: The thread dispatcher is performing a prioritiesed dispatch."
#define		MSG_DXR2512T			"DXR2512T: The thread dispatcher has dispatched RQE @%p to Thread %i at priority %i."
#define		MSG_DXR2513T			"DXR2513T: No available threads, Max=%i, Next=%i, Inspected=%i."
#define		MSG_DXR2516T			"DXR2516T: Posted RQE: @%p - O=%p, P=%p, X=%p to the rejoin pool."
#define		MSG_DXR2517T			"DXR2517T: Returning RQE: @%p - O=%p, P=%p, X=%p from the rejoin pool."
#define		MSG_DXR2518T			"DXR2518T: RQE: @%p - O=%p, P=%p, X=%p has been added to the pending pool."

//  Information Messages

//  Error Messages
#define		MSG_DXR2503E			"DXR2503E: The dispatcher thread has detected an invalid state change and is quitting."
#define		MSG_DXR2507E			"DXR2507E: The thread was unable to initialise the Notes runtime."
#define		MSG_DXR2514E			"DXR2514E: The Pending Pool Mutex has been pre-empted by thread: %i."
#define		MSG_DXR2515E			"DXR2515E: RQE @%p will not be dispatched this time as it's state has changed to %i."

class ThreadDispatcher
{
public:
	//  Constructor
	ThreadDispatcher(ExecEnvironment *xeParent, UCHAR *pTP, ThreadManagerPolicy *tmpParent);

	//  Destructor
	~ThreadDispatcher(void);

	//  Public Members
	UCHAR				*pThreadPool;						//  Pointer to the Thread Pool

	//  Public Functions
#ifdef WIN32
	static UINT LNPUBLIC StartWrapper(void * pObject);		//  Wrapper for the Start Dispatcher function
#endif

#ifdef UNIX
	static void * LNPUBLIC StartWrapper(void * pObject);		//  Wrapper for the Start Dispatcher function
#endif

	UINT Start(void);										//  Start the Monitor
	BOOL IsCMLSPoolLaneFull(UCHAR *pTP, UINT Attrs);		//  Test is CMLS Lane in the Pend Pool is full
	int GetCMLSRecommendation(UCHAR *pTP, int iLevel, int iThreadID);	//  Get the CMLS Lane Restrictions

private:
	//  Private Members
	ExecEnvironment				*xeLocal;					//  Runtime
	ThreadManagerPolicy			*tmpLocal;					//  Thread Manager policy
	int NextThread;											//  Next thread to be examined

	//  Private Functions
	void Dispatch(UCHAR * pTP);								//  Main dispatcher function
	BOOL DispatchRQE(RQElement *rqeThis, UCHAR * pTP);		//  Dispatch the passed request element (RQE)
	Threadpool_Member * FindAvailableThread(UCHAR * pTP, int *iThread);	//  Find the next available thread
	void UpdateAvailableThreadCount(UCHAR *pTP);			//  Update the available thread count
	void MonitorDispSemaphores(void);						//  Monitor the Dispatch Semaphores for posted messages
	void ProcessDispSemaphore(Threadpool_Member *pTPM);		//  Process a Dispatch Request Semaphore
	void MonitorRejoinSemaphores(void);						//  Monitor the Rejoin Semaphores for posted messages
	void ProcessRejoinSemaphore(Threadpool_Member *pTPM);			//  Process a Rejoin Semaphore
	void MonitorRejoinPollSemaphores(void);							//  Monitor the Rejoin Poll Semaphores for posted messages
	void ProcessRejoinPollSemaphore(Threadpool_Member *pTPM);		//  Process a Rejoin Poll Semaphore
	RQElement * FindAndReserveRQE(Threadpool_Header *pTPH);			//  Find and reserve an RQE slot
	RQElement * FindAndReserveRejoinRQE(Threadpool_Header *pTPH);	//  Find and reserve an RQE Rejoin slot
	UCHAR *GetMemberAddress(int iThreadID);					//  Get the address of the thread pool member
	void IncrementCMLSPoolCount(UCHAR *pTP, UINT Attrs);	//  Increment the CMLS Lane counts
	void IncrementCMLSThreadCount(UCHAR *pTP, UINT Attrs);	//  Increment the CMLS Lane counts
	void DecrementCMLSPoolCount(UCHAR *pTP, UINT Attrs);	//  Decrement the CMLS Lane counts

};
