//  ThreadManager
//
//  This class provides the root class for the control and management of the multi-threading support
//  environment for Domino eXplorer (DX) applications.
//
//  Message Identifier:      DXR
//  Message Range:           2001 - 2199
//  Trace area assignment:   TRACE_MT (02)
//  DX Version:				 3.14.0
//  Version:                 3.14.0  Build: 108
//
//  Version History:
//
//  1.0.0		07/06/2003		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.0		18/09/2009		New log handler for multithreaded applications
//  3.6.1		30/09/2009		Fixed stats & counter integrity for dispatcher
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.7.0		23/12/2009		Added race auto-correction to Scheduler
//  3.7.1		06/01/2010		Prevent a race condition in Rejoin processing
//  3.7.4		19/03/2010		Fixed various issues with log/repository configuration
//  3.9.0		24/07/2010		Linkage changed for logging & request posting and rejoining
//								A CommandHandler can be hooked into the monitor processing
//  3.9.1		02/09/2010		Changes to the ACL Processors
//								Fix for false RJR_NONE_EXIST signals
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.2		18/11/2010		MTIO support added to DbMover
//  3.10.3		13/12/2010		Db Property Modifier added
//  3.10.10		14/07/2011		Performance Improvements for/to DbCopier
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.12.0		21/12/2011		DXIP introduced
//	3.13.0		20/12/2013		PeekQueues function added
//	3.14.0		19/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2003 - 2015
//

//  Application Includes
#include "ThreadManager.h"

//  Constructor
ThreadManager::ThreadManager(ExecEnvironment *xeParent)
: ThreadPool(NULL)
, ThreadManagerStarted(false)
, tmonMaster(NULL)
, tsMaster(NULL)
, tdMaster(NULL)
, Policy(NULL)
{
	char			szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Save the runtime
	xeLocal = xeParent;

	//  Trace the construction   TRACE_MT
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2021T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg);
}

//  Destructor
ThreadManager::~ThreadManager(void)
{
	char			szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Trace the destruction  TRACE_MT
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2022T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg);
}

//  Start
//
//  This function will start all of the management components that managed by the Thread Manager.
//  Collectively these components provide support for multi-threaded applications.
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if all components were started successfully
//                  FALSE if not all components could be started
//
//  API Use Notes:
//
//
BOOL ThreadManager::Start(void)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Logpool_Header				*pLogPoolHeader;						//  Pointer to tyhe Log Pool Header
	UCHAR						*pTPMs;									//  General pointer
	int							iIndexX;								//  Index into the DBH Mapping pool
	int							iIndexY;								//  Index into the DBH Mapping pool
#ifdef UNIX
	int							iThreadRC;								//  Return code from thread functions
	int							iThreadRetries;							//  Thread start retry count
#endif
	DBHANDLE					*pMappedDBH;							//  Pointer to mapped DB Handle entries
	DBHANDLE					*pMyMappedDBH;							//  Pointer to mapped DB Handle entries
#ifdef WIN32
	HANDLE						*pMappedFH;								//  Pointer to mapped File Handle entries
	HANDLE						*pMyMappedFH;							//  Pointer to mapped File Handle entries
#endif
#ifdef UNIX
	int							*pMappedFH;								//  Pointer to mapped File Handle entries
	int							*pMyMappedFH;							//  Pointer to mapped File Handle entries
#endif
	clock_t						tStartInterval;							//  Timer Interval
	clock_t						tEndInterval;							//  Timer Interval
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer
#ifdef UNIX
	pthread_t					ptThread;								//  Thread Handle
#endif

	//  Start the Thread Manager
	if (ThreadManagerStarted)
	{
		xeLocal->LogMessage(MSG_DXR2001E);
		return FALSE;
	}

	//  Mark the Thread Manager as started to prevent it from being re-entered
	ThreadManagerStarted = TRUE;

	//  Show Thread Manager Starting  
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2005T);

#ifdef UNIX
	//  Initialise the Thread Attributes
	pthread_attr_init(&paThread);
	pthread_attr_setdetachstate(&paThread, PTHREAD_CREATE_DETACHED);      //  Create detached so we don't need to join
#endif

	//  If there was no Policy passed then create a default Policy object
	if (Policy == NULL) SetDefaultPolicy();

	//  Allocate storage for the Thread Pool
	//  NOTE: Four additional thread members are allocated for the control thtead, monitor, scheduler and dispatcher threads
	ThreadPool = (UCHAR *) malloc(sizeof(Threadpool_Header) + ((Policy->MaxThreads + 4) * sizeof(Threadpool_Member)));
	if (ThreadPool == NULL)
	{
		xeLocal->LogMessage(MSG_DXR2002E);
		return FALSE;
	}
	memset(ThreadPool, 0, sizeof(Threadpool_Header) + ((Policy->MaxThreads + 4) * sizeof(Threadpool_Member)));
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;

	//  Perform per member initialisation
	InitMembers(ThreadPool, Policy->MaxThreads + 4);

	//  Allocate storage for the Pending and Rejoin Request Pools
	pThreadPoolHeader->pPendRQEPool = (UCHAR *) malloc(Policy->PendingRQECapacity * sizeof(RQElement));
	if (pThreadPoolHeader->pPendRQEPool == NULL)
	{
		free(ThreadPool);
		xeLocal->LogMessage(MSG_DXR2003E);
		return FALSE;
	}
	memset(pThreadPoolHeader->pPendRQEPool, 0, Policy->PendingRQECapacity * sizeof(RQElement));

	pThreadPoolHeader->pRejoinRQEPool = (UCHAR *) malloc(Policy->RejoinRQECapacity * sizeof(RQElement));
	if (pThreadPoolHeader->pPendRQEPool == NULL)
	{
		free(pThreadPoolHeader->pPendRQEPool);
		free(ThreadPool);
		xeLocal->LogMessage(MSG_DXR2004E);
		return FALSE;
	}
	memset(pThreadPoolHeader->pRejoinRQEPool, 0, Policy->RejoinRQECapacity * sizeof(RQElement));

	//  Allocate the Log Pool
	pThreadPoolHeader->pLogPool = (UCHAR *) malloc(sizeof(Logpool_Header) + (Policy->AsyLogPoolEntries * sizeof(LogEntry)));
	if (pThreadPoolHeader->pLogPool == NULL)
	{
		free(pThreadPoolHeader->pRejoinRQEPool);
		free(pThreadPoolHeader->pPendRQEPool);
		free(ThreadPool);
		xeLocal->LogMessage(MSG_DXR2006E);
		return FALSE;
	}
	memset(pThreadPoolHeader->pLogPool, 0, sizeof(Logpool_Header) + (Policy->AsyLogPoolEntries * (sizeof(LogEntry))));

	//  Allocate storage for and initialise the Database Handle Mapping Pool
	//  Allocate space for MAX THREADS + 4 (0 = Native Handle , and 3 control threads)
	pThreadPoolHeader->pDBHMappingPool = (UCHAR *) malloc((Policy->MaxThreads + 4) * MAX_DBH_CACHE * sizeof(DBHANDLE));
	if (pThreadPoolHeader->pDBHMappingPool == NULL)
	{
		free(pThreadPoolHeader->pLogPool);
		free(pThreadPoolHeader->pRejoinRQEPool);
		free(pThreadPoolHeader->pPendRQEPool);
		free(ThreadPool);
		xeLocal->LogMessage(MSG_DXR2025E);
		return FALSE;
	}

	//  Initialise each entry in the pool to NULLHANDLE
	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;
	for (iIndexX = 0; iIndexX < MAX_DBH_CACHE; iIndexX++)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (Policy->MaxThreads + 4));
		for (iIndexY = 0; iIndexY <= (Policy->MaxThreads + 3); iIndexY++)
		{
			pMyMappedDBH[iIndexY] = NULLHANDLE;
		}
	}

	//  Allocate storage for and initialise the File Handle Mapping Pool
#ifdef WIN32
	pThreadPoolHeader->pFHMappingPool = (UCHAR *) malloc((Policy->MaxThreads + 1) * MAX_FH_CACHE * sizeof(HANDLE));
#endif
#ifdef UNIX
	pThreadPoolHeader->pFHMappingPool = (UCHAR *) malloc((Policy->MaxThreads + 1) * MAX_FH_CACHE * sizeof(int));
#endif
	if (pThreadPoolHeader->pDBHMappingPool == NULL)
	{
		free(pThreadPoolHeader->pDBHMappingPool);
		free(pThreadPoolHeader->pLogPool);
		free(pThreadPoolHeader->pRejoinRQEPool);
		free(pThreadPoolHeader->pPendRQEPool);
		free(ThreadPool);
		xeLocal->LogMessage(MSG_DXR2025E);
		return FALSE;
	}

	//  Initialise each entry in the pool to NULLHANDLE  (NULL for UNIX)
#ifdef WIN32
	pMappedFH = (HANDLE *) pThreadPoolHeader->pFHMappingPool;
#endif
#ifdef UNIX
	pMappedFH = (int *) pThreadPoolHeader->pFHMappingPool;
#endif
	for (iIndexX = 0; iIndexX < MAX_FH_CACHE; iIndexX++)
	{
		pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 1));
		for (iIndexY = 0; iIndexY <= Policy->MaxThreads; iIndexY++)
		{
#ifdef WIN32
			pMyMappedFH[iIndexY] = NULLHANDLE;
#endif
#ifdef UNIX
			pMyMappedFH[iIndexY] = NULL;
#endif
		}
	}

	//  Set the initial settings in the Thread Pool Header
	pThreadPoolHeader->CurrThreads = 0;
	pThreadPoolHeader->SchedMode = Policy->TPSchedMode;
	pThreadPoolHeader->PendRQEPoolSize = Policy->PendingRQECapacity;
	pThreadPoolHeader->RejoinRQEPoolSize = Policy->RejoinRQECapacity;
	pThreadPoolHeader->LogPoolSize = Policy->AsyLogPoolEntries;
	pThreadPoolHeader->MaxThreads = Policy->MaxThreads;
	pThreadPoolHeader->MinThreads = Policy->MinThreads;
	pThreadPoolHeader->TManState = TMSTATE_INIT1;
	pThreadPoolHeader->DBHTableMutex = LOG_MUTEX_FREE;
	pThreadPoolHeader->DBHTMutexAcqs = 0;
	pThreadPoolHeader->DBHTMutexFails = 0;
	pThreadPoolHeader->DBHTMutexWaits = 0;
	pThreadPoolHeader->FHTableMutex = LOG_MUTEX_FREE;
	pThreadPoolHeader->FHTMutexAcqs = 0;
	pThreadPoolHeader->FHTMutexFails = 0;
	pThreadPoolHeader->FHTMutexWaits = 0;
	pThreadPoolHeader->RejoinPoolMutex = LOG_MUTEX_FREE;
	pThreadPoolHeader->RJPMutexAcqs = 0;
	pThreadPoolHeader->RJPMutexFails = 0;
	pThreadPoolHeader->RJPMutexWaits = 0;
	pLogPoolHeader = (Logpool_Header *) pThreadPoolHeader->pLogPool;
	pThreadPoolHeader->LogMutexAcqs = 0;
	pThreadPoolHeader->LogMutexFails = 0;
	pThreadPoolHeader->LogMutexWaits = 0;
	pThreadPoolHeader->PendRQECount = 0;
	if (pThreadPoolHeader->SchedMode & TPOOL_MODE_CMLS) PerformCMLSInits();

	//  Initialise the Application Mutexes
	for (iIndexX = 0; iIndexX <= MAX_APP_MUTEXID; iIndexX++)
	{
		pThreadPoolHeader->AppMutex[iIndexX] = LOG_MUTEX_FREE;
		pThreadPoolHeader->AppMutexAcqs[iIndexX] = 0;
		pThreadPoolHeader->AppMutexFails[iIndexX] = 0;
		pThreadPoolHeader->AppMutexWaits[iIndexX] = 0;
	}

	//  Record the startup time
	OSCurrentTIMEDATE(&pThreadPoolHeader->tdRunStart);

	//  Zero the stats counters
	pThreadPoolHeader->LogMsgDropped = 0;
	pThreadPoolHeader->LogMsgDiscarded = 0;
	pThreadPoolHeader->LogMsgOverflowed = 0;
	pThreadPoolHeader->LogMsgWritten = 0;
	pThreadPoolHeader->LogMsgQueued = 0;
	pThreadPoolHeader->PPFullBackoffs = 0;
	pThreadPoolHeader->PPFullWaits = 0;
	pThreadPoolHeader->PPLaneFullBackoffs = 0;
	pThreadPoolHeader->CMLSDenials = 0;
	pThreadPoolHeader->BFDispatches = 0;
	pThreadPoolHeader->BFDispThreadsBusy = 0;
	pThreadPoolHeader->PDispatches = 0;
	pThreadPoolHeader->NoDispNoWork = 0;
	pThreadPoolHeader->NoDispThreadsBusy = 0;
	pThreadPoolHeader->DispBruteForce = 0;
	pThreadPoolHeader->DispPriority = 0;
	pThreadPoolHeader->PDispQEmpty = 0;
	pThreadPoolHeader->PDispThreadsBusy = 0;
	pThreadPoolHeader->RJRDequeued = 0;
	pThreadPoolHeader->RJRBusy = 0;
	pThreadPoolHeader->Guard01 = 0;
	pThreadPoolHeader->Guard02 = 0;
	pThreadPoolHeader->Guard03 = 0;
	pThreadPoolHeader->Guard04 = 0;
	pThreadPoolHeader->Guard05 = 0;
	pThreadPoolHeader->Guard06 = 0;
	pThreadPoolHeader->Guard07 = 0;
	pThreadPoolHeader->Guard08 = 0;
	pThreadPoolHeader->Guard09 = 0;
	pThreadPoolHeader->Guard10 = 0;

	//  Switch ownership of the Logging Interface to Asynch
	xeLocal->LoggingState = LOGSTATE_SWITCH_ASYNCH;

	//  If there is no persistent log or no current log document then don't flush the current log document
	if (xeLocal->hnLogDoc != NULLHANDLE)
	{
		NSFNoteUpdate(xeLocal->hnLogDoc, UPDATE_FORCE);
		NSFNoteClose(xeLocal->hnLogDoc);
		xeLocal->hnLogDoc = NULLHANDLE;
	}

	//  Create a new Monitor Thread
	tmonMaster = new ThreadMonitor(xeLocal, ThreadPool, Policy);
	
	//  Start the Monitor Thread
	pThreadPoolHeader->TMonState = TXSTATE_INIT;

#ifdef WIN32
	htMonitor = (HANDLE) _beginthreadex(NULL, THREAD_STACK_SIZE, tmonMaster->StartWrapper, tmonMaster, 0, &uitMonitor);
#endif

#ifdef UNIX
	iThreadRetries = 0;										//  Clear the retry counter
	htMonitor = NULL;
	//  Start the thread
	iThreadRC = pthread_create(&ptThread, &paThread, tmonMaster->StartWrapper, tmonMaster);

	//  If EAGAIN is returned then the fail condition is transient (probably short of resources)
	//  so retry the thread start.
	while ((iThreadRC == EAGAIN) && (iThreadRetries < TSTART_RETRY_LIMIT))
	{
		iThreadRetries++;
		iThreadRC = pthread_create(&ptThread, &paThread, tmonMaster->StartWrapper, tmonMaster);
	}

	//  If the thread was started then indicate the fact
	if (iThreadRC == EOK)
	{
		htMonitor = (HANDLE) ptThread;
	}
#endif
	if (htMonitor == NULL)
	{
		xeLocal->LogMessage(MSG_DXR2007E);
#ifdef WIN32
		xeLocal->LogWINAPIError();
#endif
#ifdef UNIX
		xeLocal->GetUNIXAPIError(iThreadRC, szMsg);
		xeLocal->LogMessage(szMsg);
#endif
		free(pThreadPoolHeader->pLogPool);
		free(pThreadPoolHeader->pRejoinRQEPool);
		free(pThreadPoolHeader->pPendRQEPool);
		free(ThreadPool);
		return FALSE;
	}

	//  Wait until the monitor thread has completed initialisation before going further
	while (pThreadPoolHeader->TMonState != TXSTATE_NOTSTARTED)
	{
		SLEEP(CTLINIT_INTERVAL);
	}

	//  Signal the monitor control thread to enter the Started state
	pThreadPoolHeader->TMonState = TXSTATE_START;

	//  Wait until the monitor thread has started before going further
	while (pThreadPoolHeader->TMonState != TXSTATE_RUNNING)
	{
		SLEEP(CTLSTART_INTERVAL);
	}

	//  Show monitor thread started 
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2009T);

	//  Create a new Scheduler Thread
	tsMaster = new ThreadScheduler(xeLocal, ThreadPool, Policy);

	//  Start the Scheduler Thread
	pThreadPoolHeader->TSchedState = TSSTATE_INIT;
#ifdef WIN32
	htScheduler = (HANDLE) _beginthreadex(NULL, THREAD_STACK_SIZE, tsMaster->StartWrapper, tsMaster, 0, &uitScheduler);
#endif

#ifdef UNIX
	iThreadRetries = 0;										//  Clear the retry counter
	htScheduler = NULL;
	//  Start the thread
	iThreadRC = pthread_create(&ptThread, &paThread, tsMaster->StartWrapper, tsMaster);

	//  If EAGAIN is returned then the fail condition is transient (probably short of resources)
	//  so retry the thread start.
	while ((iThreadRC == EAGAIN) && (iThreadRetries < TSTART_RETRY_LIMIT))
	{
		SLEEP(CTLSTART_INTERVAL);
		iThreadRetries++;
		iThreadRC = pthread_create(&ptThread, &paThread, tsMaster->StartWrapper, tsMaster);
	}

	//  If the thread was started then indicate the fact
	if (iThreadRC == EOK)
	{
		htScheduler = (HANDLE) ptThread;
	}
#endif

	if (htScheduler == NULL)
	{
		xeLocal->LogMessage(MSG_DXR2012E);
#ifdef WIN32
		xeLocal->LogWINAPIError();
#endif
#ifdef UNIX
		xeLocal->GetUNIXAPIError(iThreadRC, szMsg);
		xeLocal->LogMessage(szMsg);
#endif
		//  Shut down the Monitor thread
		pThreadPoolHeader->TMonState = TXSTATE_QUIT;						//  Post the quit state
#ifdef WIN32
		WaitForSingleObject(htMonitor, INFINITE);							//  Wait for the thread to close
		CloseHandle(htMonitor);												//  Close the thread handle
#endif

#ifdef UNIX
		//  Wait for the thread to quit
		while (pThreadPoolHeader->TMonState != TXSTATE_DRAINED)
		{
			SLEEP(CTLSTART_INTERVAL);
		}
		pthread_attr_destroy(&paThread);
#endif

		//  Trace the shut down
		xeLocal->LogTrace(TRACE_MT, MSG_DXR2010T);
		delete tmonMaster;													//  Delete the Monitor Object

		//  Free the memory pools
		free(pThreadPoolHeader->pLogPool);
		free(pThreadPoolHeader->pRejoinRQEPool);
		free(pThreadPoolHeader->pPendRQEPool);
		free(ThreadPool);
		return FALSE;
	}

	//  Show scheduler thread initialised
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2013T);

	//  Create a new Thread Dispatcher
	tdMaster = new ThreadDispatcher(xeLocal, ThreadPool, Policy);

	//  Start the Dispatcer Thread
	pThreadPoolHeader->TDispState = TDSTATE_INIT;

#ifdef WIN32
	htDispatcher = (HANDLE) _beginthreadex(NULL, THREAD_STACK_SIZE, tdMaster->StartWrapper, tdMaster, 0, &uitDispatcher);
#endif

#ifdef UNIX
	iThreadRetries = 0;										//  Clear the retry counter
	htDispatcher = NULL;
	//  Start the thread
	iThreadRC = pthread_create(&ptThread, &paThread, tdMaster->StartWrapper, tdMaster);

	//  If EAGAIN is returned then the fail condition is transient (probably short of resources)
	//  so retry the thread start.
	while ((iThreadRC == EAGAIN) && (iThreadRetries < TSTART_RETRY_LIMIT))
	{
		iThreadRetries++;
		iThreadRC = pthread_create(&ptThread, &paThread, tdMaster->StartWrapper, tdMaster);
	}

	//  If the thread was started then indicate the fact
	if (iThreadRC == EOK)
	{
		htDispatcher = (HANDLE) ptThread;
	}
#endif

	if (htDispatcher == NULL)
	{
		xeLocal->LogMessage(MSG_DXR2015E);
#ifdef WIN32
		xeLocal->LogWINAPIError();
#endif
#ifdef UNIX
		xeLocal->GetUNIXAPIError(iThreadRC, szMsg);
		xeLocal->LogMessage(szMsg);
#endif

		//  Drain down the Scheduler thread
		pThreadPoolHeader->TSchedState = TSSTATE_QUIT;						//  Post the quit state
#ifdef WIN32
		WaitForSingleObject(htScheduler, INFINITE);							//  Wait for the thread to close
		CloseHandle(htScheduler);											//  Close the thread handle
#endif

#ifdef UNIX
		//  Wait for the thread to quit
		while (pThreadPoolHeader->TSchedState != TXSTATE_DRAINED)
		{
			SLEEP(CTLSTART_INTERVAL);
		}
#endif

		//  Trace the shut down
		xeLocal->LogTrace(TRACE_MT, MSG_DXR2014T);

		delete tsMaster;													//  Delete the Monitor Object

		//  Shut down the Monitor thread
		pThreadPoolHeader->TMonState = TXSTATE_QUIT;						//  Post the quit state
#ifdef WIN32
		WaitForSingleObject(htMonitor, INFINITE);							//  Wait for the thread to close
		CloseHandle(htMonitor);												//  Close the thread handle
#endif

#ifdef UNIX
		//  Wait for the thread to quit
		while (pThreadPoolHeader->TMonState != TXSTATE_DRAINED)
		{
			SLEEP(CTLSTART_INTERVAL);
		}
		pthread_attr_destroy(&paThread);
#endif

		//  Trace the shut down
		xeLocal->LogTrace(TRACE_MT, MSG_DXR2010T);

		delete tmonMaster;													//  Delete the Monitor Object

		//  Free the memory pools
		free(pThreadPoolHeader->pLogPool);
		free(pThreadPoolHeader->pRejoinRQEPool);
		free(pThreadPoolHeader->pPendRQEPool);
		free(ThreadPool);
		return FALSE;
	}

	//  Show dispatcher thread initialised  
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2016T);

	//  Wait for the Dispatcher & Scheduler control threads to reach the Not Started state
	tStartInterval = xeLocal->RunningTime->getElapsed();
	if (!WaitForControlInit())
	{
		tEndInterval = xeLocal->RunningTime->getElapsed();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2018E, tEndInterval - tStartInterval);
		xeLocal->LogMessage(szMsg, 0);
		Drain();
		return FALSE;
	}

	xeLocal->LogTrace(TRACE_MT, MSG_DXR2071T, 0);

	//  Signal the Dispatcher & Scheduler control threads to enter the Started state
	pThreadPoolHeader->TSchedState = TSSTATE_START;
	pThreadPoolHeader->TDispState = TDSTATE_START;

	//  Wait for the Dispatcher & Scheduler control threads to reach the Running state
	tStartInterval = xeLocal->RunningTime->getElapsed();
	if (!WaitForControlStart())
	{
		tEndInterval = xeLocal->RunningTime->getElapsed();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2019E, tEndInterval - tStartInterval);
		xeLocal->LogMessage(MSG_DXR2019E);
		Drain();
		return FALSE;
	}

	xeLocal->LogTrace(TRACE_MT, MSG_DXR2072T, 0);

	//  For all trace areas publish the master pool addresses
	pTPMs = (UCHAR *) pThreadPoolHeader + sizeof(Threadpool_Header);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2028T, pThreadPoolHeader, pTPMs, pThreadPoolHeader->pPendRQEPool, pThreadPoolHeader->pRejoinRQEPool, pThreadPoolHeader->pLogPool, pThreadPoolHeader->pDBHMappingPool);
	xeLocal->LogTrace(TRACE_MT, szMsg);

	return TRUE;
}

//  Drain
//
//  This function will close all of the management components that managed by the Thread Manager.
//  Collectively these components provide support for multi-threaded applications.
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if all components were closed successfully
//                  FALSE if not all components could be closed
//
//  API Use Notes:
//
//
BOOL ThreadManager::Drain(void)
{
	STATUS						stAPIRC;								//  Return code from Notes API calls
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Logpool_Header				*pLogPoolHeader;						//  Pointer to the Log Pool Header
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;				//  Address the thread pool header
	pLogPoolHeader = (Logpool_Header *) pThreadPoolHeader->pLogPool;	//  Address the log pool haeder

	//  Show Thread Manager is draining
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2033T);

	//  Wait for the Log Pool to drain
	if (pLogPoolHeader->LogPoolEntries > 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2023I, pLogPoolHeader->LogPoolEntries);
		xeLocal->LogVerbose(szMsg);
	}

	while (pLogPoolHeader->LogPoolEntries > 0)
	{
		xeLocal->AppWait(LOGWAIT_INTERVAL);
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR2024I, pThreadPoolHeader->CurrThreads);
	xeLocal->LogVerbose(szMsg);

	//  Shut down all of the worker threads
	while (pThreadPoolHeader->CurrThreads > 0)
	{
		tmonMaster->DrainAnyWorker(ThreadPool, 0);
	}

	//  Wait for the Log Pool to drain again
	if (pLogPoolHeader->LogPoolEntries > 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2023I, pLogPoolHeader->LogPoolEntries);
		xeLocal->LogVerbose(szMsg);
	}

	while (pLogPoolHeader->LogPoolEntries > 0)
	{
		xeLocal->AppWait(LOGWAIT_INTERVAL);
	}

	//  Drain down the Dispatcher thread
	pThreadPoolHeader->TDispState = TDSTATE_QUIT;						//  Post the quit state
#ifdef WIN32
	WaitForSingleObject(htDispatcher, INFINITE);						//  Wait for the thread to close
	CloseHandle(htDispatcher);											//  Close the thread handle
#endif

#ifdef UNIX
		//  Wait for the thread to quit
		while (pThreadPoolHeader->TDispState != TXSTATE_DRAINED)
		{
			SLEEP(CTLSTART_INTERVAL);
		}
#endif

	xeLocal->LogTrace(TRACE_MT, MSG_DXR2017T);

	delete tdMaster;													//  Delete the Dispatcher Object

	//  Drain down the Scheduler thread
	pThreadPoolHeader->TSchedState = TSSTATE_QUIT;						//  Post the quit state
#ifdef WIN32
	WaitForSingleObject(htScheduler, INFINITE);							//  Wait for the thread to close
	CloseHandle(htScheduler);											//  Close the thread handle
#endif

#ifdef UNIX
		//  Wait for the thread to quit
		while (pThreadPoolHeader->TSchedState != TXSTATE_DRAINED)
		{
			SLEEP(CTLSTART_INTERVAL);
		}
#endif

	xeLocal->LogTrace(TRACE_MT, MSG_DXR2014T);

	delete tsMaster;													//  Delete the Scheduler Object

	//  Wait for the Log Pool to drain yet again
	if (pLogPoolHeader->LogPoolEntries > 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2023I, pLogPoolHeader->LogPoolEntries);
		xeLocal->LogVerbose(szMsg);
	}

	while (pLogPoolHeader->LogPoolEntries > 0)
	{
		xeLocal->AppWait(LOGWAIT_INTERVAL);
	}

	//  Finally drain down the monitor thread
	pThreadPoolHeader->TMonState = TXSTATE_QUIT;						//  Post the quit state
#ifdef WIN32
	WaitForSingleObject(htMonitor, INFINITE);							//  Wait for the thread to close
	CloseHandle(htMonitor);												//  Close the thread handle
#endif

#ifdef UNIX
	//  Wait for the thread to quit
	while (pThreadPoolHeader->TMonState != TXSTATE_DRAINED)
	{
		SLEEP(CTLSTART_INTERVAL);
	}
	pthread_attr_destroy(&paThread);
#endif

	//  Re-acquire the logging in the control thread - if logging is enabled
	if ((xeLocal->hdbRepository != NULLHANDLE) && (xeLocal->nidLogDoc != 0))
	{
		stAPIRC = NSFNoteOpen(xeLocal->hdbRepository, xeLocal->nidLogDoc, 0, &xeLocal->hnLogDoc);
		if (ERR(stAPIRC) != NOERROR)
		{
			xeLocal->LoggingState = LOGSTATE_CLOSED;
			std::cout << MSG_DXR2047E << std::endl;
		}
		else xeLocal->LoggingState = LOGSTATE_ACTIVE;
	}
	else xeLocal->LoggingState = LOGSTATE_CLOSED;

	xeLocal->LogTrace(TRACE_MT, MSG_DXR2010T);
	delete tmonMaster;													//  Delete the Monitor Object

	//  Free the Pools
	free(pThreadPoolHeader->pDBHMappingPool);							//  Free the DBH Mapping Pool
	free(pThreadPoolHeader->pFHMappingPool);							//  Free the File Handle Mapping Pool
	free(pThreadPoolHeader->pLogPool);									//  Free the Log Pool
	free(pThreadPoolHeader->pRejoinRQEPool);							//  Free the Rejoin Requests Pool
	free(pThreadPoolHeader->pPendRQEPool);								//  Free the Punding Requests Pool
	free(ThreadPool);

	//  Show shutdown complete
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2011T);
	return TRUE;
}

//  SetDefaultPolicy
//
//  This function will construct a default ThreadManagePolicy object and populate it with generic settings.
//  A default policy is only needed if the policy has not been set by the host application during startup.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::SetDefaultPolicy(void)
{
	//  Create and set the default policy configuration
	Policy = new ThreadManagerPolicy();
	Policy->TPoolPolicy = TPOOL_POLICY_PRESTARTMIN | TPOOL_POLICY_OBEYMAXMIN;
	Policy->MaxThreads = 10;
	Policy->MinThreads = 10;
	Policy->PendingRQECapacity = 100;								//  Pending RQE Pool size
	Policy->RejoinRQECapacity = 100;								//  Rejoin RQE Pool size
	Policy->AsyLogPoolEntries = 100;
	return;
}

//  WaitForControlInit
//
//  This function will wait until the Dispatcher & Scheduler control threads have initialised.
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if the control thread has initialised
//                  FALSE if the control thread did not initialise within the allowed time limit
//
//  API Use Notes:
//
//
BOOL ThreadManager::WaitForControlInit(void)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	UINT						TComState;								//  Combined state
	int							iWaitCount = 0;							//  Count of wait intervals

	//  Wait for the control tasks to initialise
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;				//  Address the header
	TComState = pThreadPoolHeader->TSchedState | pThreadPoolHeader->TDispState;
	while ((TComState != TXSTATE_NOTSTARTED) && (iWaitCount < CTLINIT_LIMIT))
	{
		iWaitCount++;
		xeLocal->AppWait(CTLINIT_INTERVAL);
		TComState = pThreadPoolHeader->TSchedState | pThreadPoolHeader->TDispState;
	}
	if (iWaitCount == CTLINIT_LIMIT) return FALSE;
	return TRUE;
}

//  WaitForControlStart
//
//  This function will wait until the Dispatcher & Scheduler control threads have started.
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if the control thread has started
//                  FALSE if the control thread did not start within the allowed time limit
//
//  API Use Notes:
//
//
BOOL ThreadManager::WaitForControlStart(void)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	UINT						TComState;								//  Combined state
	int							iWaitCount = 0;							//  Count of wait intervals

	//  Wait for the control tasks to start
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;				//  Address the header
	TComState = pThreadPoolHeader->TSchedState | pThreadPoolHeader->TDispState;
	while ((TComState != TXSTATE_RUNNING) && (iWaitCount < CTLSTART_LIMIT))
	{
		iWaitCount++;
		xeLocal->AppWait(CTLSTART_INTERVAL);
		TComState = pThreadPoolHeader->TSchedState | pThreadPoolHeader->TDispState;
	}
	if (iWaitCount == CTLSTART_LIMIT) return FALSE;
	return TRUE;
}

//  CorePostARequest
//
//  This function will post a request into the execution queue for the Thread Pool Manager.
//
//  Parameters
//
//	1  -  int		Flags that determins how the request will be posted. See PXR_ 
//  2  -  void *	Pointer to the owning object for this request, can be NULL if the request will not rejoin
//  3  -  void *	Pointer to the object that will execute the request
//  4  -  void *	Pointer to the parameter object for this request
//  5  -  UINT		Attributes of the request. See RQATTR_
//  6  -  int		The initial priority of the request
//  7  -  int		Thread ID of the requester
//
//  Returns
//
//  BOOL  -			TRUE if the request was posted to the queue successfully
//                  FALSE if the request could not be posted
//
//  API Use Notes:
//  Do Not use this function use the exposed PostARequest() function in MTExecutive
//
BOOL ThreadManager::CorePostARequest(int iFlags, void * pOwner, void * pExecutor, void * pParms, UINT Attrs, int iPriority, int iThreadID)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member		*pThreadPoolMember;						//  Pointer to thread pool member
	RQElement				MyRQE;									//  Local Request Element
	DWORD					dwWaitIncrement;						//  Semaphore wait increment count
	char					szMsg[MAX_MSG + 1];						//  Generic message buffer

	dwWaitIncrement = 0;

	//  Post a request to the request pool
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;				//  Address the header

	//  Get the address of the thread pool member for the posting thread
	pThreadPoolMember = (Threadpool_Member *) GetMemberAddress(iThreadID);

	//  Post information to the request
	MyRQE.pOObject = pOwner;
	MyRQE.pXObject = pExecutor;
	MyRQE.pPObject = pParms;
	MyRQE.RQEPriority = iPriority;
	MyRQE.RQEAttrs = Attrs;

	//  If the wait_if_busy flag is set then wait while the pend pool is full
	if (iFlags & PXR_WAITIF_BUSY)
	{
		while (pThreadPoolHeader->PendRQECount == pThreadPoolHeader->PendRQEPoolSize)
		{
			pThreadPoolHeader->PPFullWaits++;
			//  Check that the Global Permit has not been revoked or Fast Drain Mode set
			if ((!xeLocal->TestGlobalPermit()) || (pThreadPoolHeader->TDispState == TDSTATE_BLOCKED)) return FALSE;
			xeLocal->AppWait(PXR_BUSY_WAIT);					//  Wait
		}

		//  If CMLS is active then there can be a further delay if the constrained lane is full
		if (pThreadPoolHeader->SchedMode & TPOOL_MODE_CMLS)
		{
			if (Attrs & RQATTR_CMLSLCONSTRAINED)
			{
				while (tdMaster->IsCMLSPoolLaneFull(ThreadPool, Attrs))
				{
					pThreadPoolHeader->PPFullWaits++;
					//  Check that the Global Permit has not been revoked or Fast Drain Mode set
					if ((!xeLocal->TestGlobalPermit()) || (pThreadPoolHeader->TDispState == TDSTATE_BLOCKED)) return FALSE;
					xeLocal->AppWait(PXR_BUSY_WAIT);					//  Wait
				}
			}
		}
	}

	//  Wait for the Request Dispatch Semaphore to become free
	while (pThreadPoolMember->semReqDisp != REQ_SEM_FREE)
	{
		dwWaitIncrement = 1;
		pThreadPoolMember->dwReqDispWaitQuanta++;
		//  Check that the Global Permit has not been revoked or Fast Drain Mode set
		if ((!xeLocal->TestGlobalPermit()) || (pThreadPoolHeader->TDispState == TDSTATE_BLOCKED)) return FALSE;
		SLEEP(Policy->DispatchSemFreeWait);
	}
	pThreadPoolMember->dwReqDispWaits = pThreadPoolMember->dwReqDispWaits + dwWaitIncrement;
	
	//  Copy the RQE data to the thread pool member
	memcpy(&pThreadPoolMember->rqeDisp, &MyRQE, sizeof(RQElement));

	//  Increment the count of dispatch requests
	pThreadPoolMember->dwReqDisp++;

	//  Post the semaphore
	pThreadPoolMember->semReqDisp = iThreadID;

	//  Trace the event
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2069T, MyRQE.pOObject, MyRQE.pPObject, MyRQE.pXObject);
	xeLocal->LogTrace(TRACE_MRQ, szMsg);

	return TRUE;
}

//  CorePostALogMessage
//
//  This function will post a log message into the appropriate logger semaphore
//
//  Parameters
//
//  1  -  char *	Pointer to the message to be posted
//  2  -  int		The ID of the thread making the request
//
//  Returns
//
//  API Use Notes:
//
BOOL ThreadManager::CorePostALogMessage(char * szLogMessage, int iThreadID)
{
	Threadpool_Member		*pThreadPoolMember;						//  Pointer to thread pool member
	DWORD					dwWaitIncrement;						//  Semaphore wait increment count

	dwWaitIncrement = 0;

	//  Get the address of the thread pool member for the posting thread
	pThreadPoolMember = (Threadpool_Member *) GetMemberAddress(iThreadID);

#ifdef _DEBUG
	printf("+%s [%i]\n", szLogMessage, iThreadID);
#endif

	//  Wait for the semaphore to become free - there is no timeout on this wait
	while (pThreadPoolMember->semLogMsg != LOG_SEM_FREE)
	{
		dwWaitIncrement = 1;
		pThreadPoolMember->dwLogSemWaitQuanta++;
		SLEEP(Policy->LogSemFreeWait);
	}

	pThreadPoolMember->dwLogSemWaits = pThreadPoolMember->dwLogSemWaits + dwWaitIncrement;

	//  Post the log data into the member
	OSCurrentTIMEDATE(&pThreadPoolMember->tdLogMsg);
	strcpy_s(pThreadPoolMember->szLogMsg, MAX_MSG, szLogMessage);
	pThreadPoolMember->semLogMsg = iThreadID;

	pThreadPoolMember->dwLogMsgsPosted++;

	//  If the caller is the Thread Monitor then post directly to the asynch log pool
	//  this prevents a deadly self-embrace
	if (iThreadID == MONITOR_THREAD_ID)
	{
		tmonMaster->ProcessLogSemaphore(pThreadPoolMember);		
	}

	return TRUE;
}

//  AcquireDBHMutex
//
//  This function will gain control of the mutex that protects the DB Handle Mapping pool structures.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  int						-  ID of the thread on behalf of which the mutex is being acquired
//
//  Returns
//
//  BOOL			-	TRUE if the mutex has been acquired
//					-	FALSE if the mutex could not be acuired within the allowed time
//
//  API Use Notes:
//  
//
BOOL ThreadManager::AcquireDBHTMutex(UCHAR * pTP, int iThreadID)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	int						iWaitCount = 0;							//  Counts of waits on the Runtime Log Mutex

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	//  Increment request count
	pThreadPoolHeader->DBHTMutexAcqs++;

	//  Acquire the Runtime DB Handle Table Mutex on behalf of the thread
	while (pThreadPoolHeader->DBHTableMutex != iThreadID)
	{
#ifdef DXMUTEX_STRENGTH_PUNY
		while ((pThreadPoolHeader->DBHTableMutex != LOG_MUTEX_FREE) && (iWaitCount < Policy->DBHMutexWaitLimit))
		{
			pThreadPoolHeader->DBHTMutexWaits++;
			iWaitCount++;
			SLEEP(Policy->DBHMutexWaitInterval);
		}
		if (iWaitCount == Policy->DBHMutexWaitLimit)
		{
			pThreadPoolHeader->DBHTMutexFails++;
			return FALSE;			//  Timed out
		}

		pThreadPoolHeader->DBHTableMutex = iThreadID;					//  Mark the mutex as owned
#else
#ifdef WIN32
		while ((InterlockedCompareExchange(&pThreadPoolHeader->DBHTableMutex, iThreadID, LOG_MUTEX_FREE) != LOG_MUTEX_FREE) && (iWaitCount < Policy->DBHMutexWaitLimit))
		{
			pThreadPoolHeader->DBHTMutexWaits++;
			iWaitCount++;
			SLEEP(Policy->DBHMutexWaitInterval);
		}
		if (iWaitCount == Policy->DBHMutexWaitLimit)
		{
			pThreadPoolHeader->DBHTMutexFails++;
			return FALSE;				//  Timed out
		}
#endif
#ifdef UNIX
		while ((DXAtomic_cmpxchg(&pThreadPoolHeader->DBHTableMutex, LOG_MUTEX_FREE, iThreadID) != LOG_MUTEX_FREE) && (iWaitCount < Policy->DBHMutexWaitLimit))
		{
			pThreadPoolHeader->DBHTMutexWaits++;
			iWaitCount++;
			SLEEP(Policy->DBHMutexWaitInterval);
		}
		if (iWaitCount == Policy->DBHMutexWaitLimit)
		{
			pThreadPoolHeader->DBHTMutexFails++;
			return FALSE;				//  Timed out
		}
#endif
#endif
	}
	return TRUE;
}

//  CoreAcquireAppMutex
//
//  This function will gain control of one of the application defined mutexes
//
//  Parameters
//
//  1  -  int			-  The ID of the application mutex to be acquired
//  2  -  int			-  The ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the Mutex was acquired
//                  FALSE if the Mutex could not be acquired
//
//  API Use Notes:
//  
//
BOOL ThreadManager::CoreAcquireAppMutex(UCHAR *pTP, int iMutexID, int iThreadID)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	int						iWaitCount = 0;							//  Counts of waits on the Runtime Log Mutex

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	// validate the Mutex ID
	if ((iMutexID < 0) || (iMutexID > MAX_APP_MUTEXID)) return FALSE;

	//  Increment the request count
	pThreadPoolHeader->AppMutexAcqs[iMutexID]++;

	//  Acquire the requested App Mutex on behalf of the thread
	while (pThreadPoolHeader->AppMutex[iMutexID] != iThreadID)
	{
#ifdef DXMUTEX_STRENGTH_PUNY
		while ((pThreadPoolHeader->AppMutex[iMutexID] != LOG_MUTEX_FREE) && (iWaitCount < Policy->APPMutexWaitLimit))
		{
			pThreadPoolHeader->AppMutexWaits[iMutexID]++;
			iWaitCount++;
			SLEEP(Policy->APPMutexWaitInterval);
		}
		if (iWaitCount == Policy->APPMutexWaitLimit)
		{
			pThreadPoolHeader->AppMutexFails[iMutexID]++;
			return FALSE;				//  Timed out
		}
		pThreadPoolHeader->AppMutex[iMutexID] = iThreadID;					//  Mark the mutex as owned
#else
#ifdef WIN32
		while ((InterlockedCompareExchange(&pThreadPoolHeader->AppMutex[iMutexID], iThreadID, LOG_MUTEX_FREE) != LOG_MUTEX_FREE) && (iWaitCount < Policy->APPMutexWaitLimit))
		{
			pThreadPoolHeader->AppMutexWaits[iMutexID]++;
			iWaitCount++;
			SLEEP(Policy->APPMutexWaitInterval);
		}
		if (iWaitCount == Policy->APPMutexWaitLimit)
		{
			pThreadPoolHeader->AppMutexFails[iMutexID]++;
			return FALSE;				//  Timed out
		}
#endif
#ifdef UNIX
		while ((DXAtomic_cmpxchg(&pThreadPoolHeader->AppMutex[iMutexID], LOG_MUTEX_FREE, iThreadID) != LOG_MUTEX_FREE) && (iWaitCount < Policy->APPMutexWaitLimit))
		{
			pThreadPoolHeader->AppMutexWaits[iMutexID]++;
			iWaitCount++;
			SLEEP(Policy->APPMutexWaitInterval);
		}
		if (iWaitCount == Policy->APPMutexWaitLimit)
		{
			pThreadPoolHeader->AppMutexFails[iMutexID]++;
			return FALSE;				//  Timed out
		}
#endif
#endif
	}
	return TRUE;
}

//  CoreFreeAppMutex
//
//  This function will relinquish control of one of the application defined mutexes
//
//  Parameters
//
//  1  -  int			-  The ID of the application mutex to be freed
//  2  -  int			-  The ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the Mutex was freed
//                  FALSE if the Mutex could not be freed
//
//  API Use Notes:
//  
//
BOOL ThreadManager::CoreFreeAppMutex(UCHAR *pTP, int iMutexID, int iThreadID)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	int						iWaitCount = 0;							//  Counts of waits on the Runtime Log Mutex

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	// validate the Mutex ID
	if ((iMutexID < 0) || (iMutexID > MAX_APP_MUTEXID)) return FALSE;

	if(pThreadPoolHeader->AppMutex[iMutexID] == iThreadID)
	{
		pThreadPoolHeader->AppMutex[iMutexID] = LOG_MUTEX_FREE;
		return TRUE;
	}
	return FALSE;
}

//  CoreGetMappedDBH
//
//  This function will return the DB Handle that can be used on the current thread that maps to
//  the database identified by the native DB Handle that was passed.
//
//  Parameters
//
//	1  -  DBHANDLE	Native D/B handle for the requested database
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  DBHANDLE  -		Local (to this thread) DB Handle for the database
//                  NULLHANDLE - A local DB Handle could not be made available
//
//  API Use Notes:
//
//
DBHANDLE ThreadManager::CoreGetMappedDBH(DBHANDLE hdbNative, int iThreadID)
{	
	STATUS				stAPIRC;								//  Return code from Notes API Calls
	int					iIndex;									//  Generic index variable
	int					iIndexX;								//  Index of native DB Handles
	int					iIndexFree;								//  Index of free native DB handle slot
	int					iThreadIndex;							//  Index for the current thread ID
	DBHANDLE			hdbTemp;								//  Temporary DB handle
	DBHANDLE			*pMappedDBH;							//  Pointer to my mapped DB Handle entry
	DBHANDLE			*pMyMappedDBH;							//  Pointer to my mapped DB Handle entry
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  If the action is called on the main application thread (0) then return the native handle
	if (iThreadID == 0) return hdbNative;

	//  Determine if the Native handle has already been cached
	iIndexX = -1;
	iIndexFree = -1;
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;

	//  Compute the Thread Index
	switch(iThreadID)
	{
	case CONTROL_THREAD_ID:
		iThreadIndex = CONTROL_THREAD_ID;
		break;
	case MONITOR_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 1;
		break;
	case SCHEDULER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 2;
		break;
	case DISPATCHER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 3;
		break;
	default:
		iThreadIndex = iThreadID;
		break;
	}

	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;

	//  First perform the search of the DBHandle table without holding the mutex
	//  the search will have to be re-done with the mutex held if a new entry needs to be added
	for (iIndex = 0; iIndex < MAX_DBH_CACHE; iIndex++)
	{
		pMyMappedDBH = pMappedDBH + (iIndex * (Policy->MaxThreads + 4));
		if (pMyMappedDBH[0] == hdbNative)
		{
			iIndexX = iIndex;
			break;
		}
		if ((pMyMappedDBH[0] == NULLHANDLE) && (iIndexFree == -1)) iIndexFree = iIndex;
	}

	//  If there is a row for this native handle then determine if there is already a thread local
	//  handle avaiilable for this thread.
	if (iIndexX != -1)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (Policy->MaxThreads + 4));
		if (pMyMappedDBH[iThreadIndex] != NULLHANDLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2029T, iThreadID, pMyMappedDBH[iThreadIndex], hdbNative);
			xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
			return pMyMappedDBH[iThreadIndex];
		}
	}

	//  Acquire the DB Handle Table Mutex
	if (!AcquireDBHTMutex(ThreadPool, iThreadID))
	{
		xeLocal->LogMessage(MSG_DXR2032E, iThreadID);
		return NULLHANDLE;
	}

	//  Determine if the Native handle has already been cached
	for (iIndex = 0; iIndex < MAX_DBH_CACHE; iIndex++)
	{
		pMyMappedDBH = pMappedDBH + (iIndex * (Policy->MaxThreads + 4));
		if (pMyMappedDBH[0] == hdbNative)
		{
			iIndexX = iIndex;
			break;
		}
		if ((pMyMappedDBH[0] == NULLHANDLE) && (iIndexFree == -1)) iIndexFree = iIndex;
	}

	//  If there is no cache line for this handle then allocate one
	if (iIndexX == -1)
	{
		//  Native handle is not in the cache
		if (iIndexFree == -1)
		{
			//  There are no free slots in the table
			xeLocal->LogMessage(MSG_DXR2026E, iThreadID);
			pThreadPoolHeader->DBHTableMutex = LOG_MUTEX_FREE;
			return NULLHANDLE;
		}
		//  Assign the row in the cache to the native handle
		iIndexX = iIndexFree;
		pMyMappedDBH = pMappedDBH + (iIndexX * (Policy->MaxThreads + 4));
		pMyMappedDBH[0] = hdbNative;
	}

	//  Check if there is already a thread local handle for the requesting thread
	pMyMappedDBH = pMappedDBH + (iIndexX * (Policy->MaxThreads + 4));
	if (pMyMappedDBH[iThreadIndex] != NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2029T, iThreadID, pMyMappedDBH[iThreadIndex], hdbNative);
		xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
		pThreadPoolHeader->DBHTableMutex = LOG_MUTEX_FREE;
		return pMyMappedDBH[iThreadIndex];
	}

	//  Free the Mutex at this point as there is no conflict
	pThreadPoolHeader->DBHTableMutex = LOG_MUTEX_FREE;

	//  Reopen the native handle to create a thread local handle
	stAPIRC = NSFDbReopen(hdbNative, &hdbTemp);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR2027E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Store the thread local handle and return it
	pMyMappedDBH[iThreadIndex] = hdbTemp;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2030T, hdbTemp, hdbNative, iThreadID);
	xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
	return hdbTemp;
}

//  CoreInvalidateNativeDBH
//
//  This function will mark a Native Database Handle as no longer needed in the DB Handle mapping table.
//  This will allow any threads to release the per thread handle opened for the database.
//
//  Parameters
//
//	1  -  DBHANDLE	Native D/B handle that is no longer needed
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the DB Handle was marked as no longer needed
//                  FALSE if the DB Handle could not be marked as no longer needed
//
//  API Use Notes:
//
//
BOOL ThreadManager::CoreInvalidateNativeDBH(DBHANDLE hdbNative, int iThreadID)
{
	int					iIndex;									//  Generic index variable
	int					iIndexX;								//  Index of native DB Handles
	DBHANDLE			*pMappedDBH;							//  Pointer to a mapped DBH entry
	DBHANDLE			*pMyMappedDBH;							//  Pointer to a mapped DBH entry
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Determine if there is a cache row for the Native handle
	iIndexX = -1;
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;
	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;

	for (iIndex = 0; iIndex < MAX_DBH_CACHE; iIndex++)
	{
		pMyMappedDBH = pMappedDBH + (iIndex * (Policy->MaxThreads + 4));
		if (pMyMappedDBH[0] == hdbNative)
		{
			iIndexX = iIndex;
			break;
		}
	}

	//  Handle is not cached - which is perfectly valid
	if (iIndexX == -1) return TRUE;

	//  Mark the entry to be freed by each thread
	pMyMappedDBH = pMappedDBH + (iIndexX * (Policy->MaxThreads + 4));
	pMyMappedDBH[0] = (DBHANDLE) FREEHANDLE;

	//  Trace the action
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2031T, hdbNative);
	xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);

	//  Check for any entries that can be released
	ReleaseAllMappedDBH(iThreadID);
	return TRUE;
}

//  ReleaseAllMappedDBH
//
//  This function will close any locally (thread) mapped DB Handles that can be closed because
//  the native handle has been invalidated.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::ReleaseAllMappedDBH(int iThreadID)
{
	int					iIndexX;								//  Index of native DB Handles
	int					iIndexY;								//  Index of threads
	int					iReleasedEntries;						//  Count of released thread entries
	DBHANDLE			*pMappedDBH;							//  Pointer to a mapped DBH entry
	DBHANDLE			*pMyMappedDBH;							//  Pointer to a mapped DBH entry
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Release any mapped file handles that have been completely freed
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;
	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;

	for (iIndexX = 0; iIndexX < MAX_DBH_CACHE; iIndexX++)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (Policy->MaxThreads + 4));
		if (pMyMappedDBH[0] == (DBHANDLE) FREEHANDLE)
		{
			iReleasedEntries = 0;
			for (iIndexY = 1; iIndexY <= (Policy->MaxThreads + 3); iIndexY++)
			{
				if (pMyMappedDBH[iIndexY] == NULLHANDLE) iReleasedEntries++;
			}
			if (iReleasedEntries == (Policy->MaxThreads + 3))
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2083T, pMyMappedDBH[0], iThreadID);
				xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
				pMyMappedDBH[0] = NULLHANDLE;
			}
		}
	}
	return;
}

//  CoreReleaseMappedDBH
//
//  This function will release any locally mapped DB Handles that are owned by the thread indicated.
//
//  Parameters
//
//  1  -  int		ID of the thread that is making the request
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::CoreReleaseMappedDBH(int iThreadID)
{
	int					iIndexX;								//  Index of native DB Handles
	int					iThreadIndex;							//  Relative thread index
	DBHANDLE			*pMappedDBH;							//  Pointer to a mapped DBH entry
	DBHANDLE			*pMyMappedDBH;							//  Pointer to a mapped DBH entry
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG];							//  Generic message buffer

	//  Compute the Thread Index
	switch(iThreadID)
	{
	case CONTROL_THREAD_ID:
		iThreadIndex = CONTROL_THREAD_ID;
		break;
	case MONITOR_THREAD_ID:
		iThreadIndex = Policy->MaxThreads + 1;
		break;
	case SCHEDULER_THREAD_ID:
		iThreadIndex = Policy->MaxThreads + 2;
		break;
	case DISPATCHER_THREAD_ID:
		iThreadIndex = Policy->MaxThreads + 3;
		break;
	default:
		iThreadIndex = iThreadID;
		break;
	}

	//  Release any mapped file handles that belong to this thread
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;
	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;

	for (iIndexX = 0; iIndexX < MAX_DBH_CACHE; iIndexX++)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (Policy->MaxThreads + 4));
		if (pMyMappedDBH[0] == (DBHANDLE) FREEHANDLE)
		{
			if (pMyMappedDBH[iThreadIndex] != NULLHANDLE)
			{
				NSFDbClose(pMyMappedDBH[iThreadIndex]);
				//  Trace the action
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2041T, pMyMappedDBH[iThreadIndex], iThreadID);
				xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
				pMyMappedDBH[iThreadIndex] = NULLHANDLE;
			}
		}
	}

	//  Check if any entries can be freed in the table
	ReleaseAllMappedDBH(iThreadID);
	return;
}

//  CoreIsRejoinRecommended
//
//  This function will indicate if rejoin processing should be done now or can be deferred.
//
//  Parameters
//
//  1  -  void *	Pointer to the owning object for this request
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if rejoin processing should happen now, FALSE if not
//
//  API Use Notes:
//
//  Do Not use this function use the exposed IsRejoinRecommended() function in MTExecutive
//
BOOL ThreadManager::CoreIsRejoinRecommended(void *pOwner, int iThreadID)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;				//  Address the header

	//  Increment the count of requests
	pThreadPoolHeader->RejoinRecs++;

	//  Check the Pending pool level to see if it is above the trip level
	if (((pThreadPoolHeader->PendRQECount * 100) / pThreadPoolHeader->PendRQEPoolSize) > Policy-> MinPctPPFullRR)
	{
		pThreadPoolHeader->RejoinRecd++;
		return TRUE;
	}

	//  Check the Rejoin pool level to see if it is above the trip level
	if (((pThreadPoolHeader->RejoinRQECount * 100) / pThreadPoolHeader->RejoinRQEPoolSize) > Policy->MinPctRPFullRR)
	{
		pThreadPoolHeader->RejoinRecd++;
		return TRUE;
	}
	return FALSE;
}

//  CoreGetRejoinRequest
//
//  This function will retrieve an Exec Request that has been executed by the Thread Pool Manager and
//  is ready to rejoin the main application flow.
//
//  Parameters
//
//  1  -  void *	Pointer to the owning object for this request
//  2  -  void **	Pointer to a pointer that will be populated with a pointer to the parameter object that is ready
//  3  -  int		ID of the thread that is making the request
//
//  Returns
//
//  int  -			See RJR_ for the values returned
//
//  API Use Notes:
//
//  Do Not use this function use the exposed GetRejoinRequest() function in MTExecutive
//  The function will perform an initial 'fast' check of the pools, for a deep check or to process an
//  element then the routine will post the Rejoin Poll semaphore and have the control threads process
//  the request. 
//  NOTE: The Rejoin Poll semaphore is a two phase semaphore - it is posted by this routine which will wait
//        for it to be marked free again once free the information can be returned to the caller
//
//
int ThreadManager::CoreGetRejoinRequest(void *pOwner, void **pParms, int iThreadID)
{
	int							iIndex;									//  Index of RQ elements
	RQElement					*pCandidateRQE;							//  Pointer to RQE slot
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pMyThread;								//  Pointer to the selected thread
	Threadpool_Member			*pAnyThread;							//  Pointer to the selected thread
	DWORD						dwWaitCount;							//  Wait counter
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;				//  Address the header

	//  Get the address of the current Thread Pool member
	pMyThread = (Threadpool_Member *) GetMemberAddress(iThreadID);

	//  Explicitly clear the returned Parm Object pointer to NULL
	*pParms = NULL;

	//  Increment the poll request counter
	pMyThread->dwRejoinPoll++;

	//  Check the rejoin pool to see if there is a matching element
	pCandidateRQE = (RQElement *) pThreadPoolHeader->pRejoinRQEPool;	//  Point to the first element
	for (iIndex = 0; iIndex < pThreadPoolHeader->RejoinRQEPoolSize; iIndex++)
	{
		if ((pCandidateRQE->RQEState == RQESTATE_READY) && (pCandidateRQE->pOObject == pOwner))
		{
			//  There is an available element already in the rejoin pool - post the rejoin poll semaphore
			//  Wait for the rejoin poll semaphore to become free

			dwWaitCount = 0;
			while (pMyThread->semRejoinPoll != REJPOLL_SEM_FREE)
			{
				dwWaitCount = 1;
				pMyThread->dwRejoinPollWaitQuanta++;
				SLEEP(Policy->RejoinSemFreeWait);
			}
			pMyThread->dwRejoinPollWaits = pMyThread->dwRejoinPollWaits + dwWaitCount;
			
			//  Post the Rejoin Poll semaphore
			pMyThread->rqeRejoinPoll.pOObject = pOwner;
			pMyThread->semRejoinPoll = iThreadID;

			//  Wait until the semaphore is cleared again indicating that the info has been returned
			dwWaitCount = 0;
			while (pMyThread->semRejoinPoll != REJPOLL_SEM_FREE)
			{
				dwWaitCount = 1;
				pMyThread->dwRejoinPollWaitQuanta++;
				SLEEP(Policy->RejoinSemFreeWait);
			}
			pMyThread->dwRejoinPollWaits = pMyThread->dwRejoinPollWaits + dwWaitCount;

			//  Check the return
			if (pMyThread->rcRejoinPoll == RJR_RETURNED)
			{
				//  Return the data to the caller
				*pParms = pMyThread->rqeRejoinPoll.pPObject;
				//  Trace the action
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2042T, pMyThread->rqeRejoinPoll.pPObject, pMyThread->rqeRejoinPoll.pOObject, pMyThread->rqeRejoinPoll.pXObject);
				xeLocal->LogTrace(TRACE_MRQ, szMsg, iThreadID);
			}

			//  return to caller
			return pMyThread->rcRejoinPoll;

		}
		pCandidateRQE++;
	}

	//  If there is a request in the RQE pending pool or allocated to a thread then return not ready
	//  First check if there are any executing or any posted for the pending pool

	//  Step 1:  Check the worker threads
	pAnyThread = (Threadpool_Member *) (ThreadPool + sizeof(Threadpool_Header));
	for (iIndex = 0; iIndex < pThreadPoolHeader->MaxThreads; iIndex++)
	{
		if ((pAnyThread->TMemState == WTSTATE_BUSY) || (pAnyThread->TMemState == WTSTATE_POSTED) || (pAnyThread->TMemState == WTSTATE_REJOIN))
		{
			//  Check for an executing request (or posted for execution or a rejoining request)
			if ((pAnyThread->rqePosted.pOObject == pOwner) && (pAnyThread->rqePosted.RQEAttrs & RQATTR_REJOIN))
			{
				pThreadPoolHeader->RJRBusy++;
				pMyThread->dwRejoinNotReady++;
				return RJR_NONE_READY;
			}
			//  Now check if there is a request posted for the pending execution queue
			if ((pAnyThread->semReqDisp != REQ_SEM_FREE) && (pAnyThread->rqeDisp.pOObject == pOwner))
			{
				pThreadPoolHeader->RJRBusy++;
				pMyThread->dwRejoinNotReady++;
				return RJR_NONE_READY;
			}
		}
		pAnyThread++;
	}

	//  Step 2:  Check the control thread
	pAnyThread = (Threadpool_Member *) GetMemberAddress(0);

	//  Now check if there is a request posted for the pending execution queue
	if ((pAnyThread->semReqDisp != REQ_SEM_FREE) && (pAnyThread->rqeDisp.pOObject == pOwner))
	{
		pThreadPoolHeader->RJRBusy++;
		pThreadPoolHeader->Guard01++;
		pMyThread->dwRejoinNotReady++;
		return RJR_NONE_READY;
	}

	//  If the Dipatcher has been blocked then we are in Fast Drain Mode - in which case ignore any
	//  requests that are in the ready pool as they will never be dispatched
	if (pThreadPoolHeader->TDispState == TDSTATE_BLOCKED) 
	{
		pMyThread->dwRejoinNone++;
		return RJR_NONE_EXIST;
	}

	//  Now check the ready pool
	pCandidateRQE = (RQElement *) pThreadPoolHeader->pPendRQEPool;		//  Point to the first element
	for (iIndex = 0; iIndex < pThreadPoolHeader->PendRQEPoolSize; iIndex++)
	{
		if ((pCandidateRQE->RQEState == RQESTATE_READY) && (pCandidateRQE->pOObject == pOwner) && (pCandidateRQE->RQEAttrs & RQATTR_REJOIN))
		{
			pThreadPoolHeader->RJRBusy++;
			pMyThread->dwRejoinNotReady++;
			return RJR_NONE_READY;
		}
		pCandidateRQE++;
	}

	//  The checks should now be done through the rejoin poll semaphore as the return of 
	//  RJR_NONE_EXIT must be authoratative

	//  Wait for the rejoin poll semaphore to become free
	dwWaitCount = 0;
	while (pMyThread->semRejoinPoll != REJPOLL_SEM_FREE)
	{
		dwWaitCount = 1;
		pMyThread->dwRejoinPollWaitQuanta++;
		SLEEP(Policy->RejoinSemFreeWait);
	}
	pMyThread->dwRejoinPollWaits = pMyThread->dwRejoinPollWaits + dwWaitCount;
			
	//  Post the Rejoin Poll semaphore
	pMyThread->rqeRejoinPoll.pOObject = pOwner;
	pMyThread->semRejoinPoll = iThreadID;

	//  Wait until the semaphore is cleared again indicating that the info has been returned
	dwWaitCount = 0;
	while (pMyThread->semRejoinPoll != REJPOLL_SEM_FREE)
	{
		dwWaitCount = 1;
		pMyThread->dwRejoinPollWaitQuanta++;
		SLEEP(Policy->RejoinSemFreeWait);
	}
	pMyThread->dwRejoinPollWaits = pMyThread->dwRejoinPollWaits + dwWaitCount;

	//  Check the return
	if (pMyThread->rcRejoinPoll == RJR_RETURNED)
	{
		//  Return the data to the caller
		*pParms = pMyThread->rqeRejoinPoll.pPObject;
		//  Trace the action
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2042T, pMyThread->rqeRejoinPoll.pPObject, pMyThread->rqeRejoinPoll.pOObject, pMyThread->rqeRejoinPoll.pXObject);
		xeLocal->LogTrace(TRACE_MRQ, szMsg, iThreadID);
	}

	//  return to caller
	return pMyThread->rcRejoinPoll;
}

//  CorePeekQueues
//
//  This function will return the number of requests that are on all of the request queues for the designated owner.
//
//  Parameters
//
//  1  -  void *	Pointer to the owning object for this request
//  3  -  int		ID of the thread that is making the request
//
//  Returns
//
//  DWORD  -		Count of requests for the designated owner
//
//  API Use Notes:
//
//  Do Not use this function use the exposed PeekQueues() function in MTExecutive
//
//
DWORD ThreadManager::CorePeekQueues(void *pOwner, int iThreadID)
{
	DWORD						RequestCount = 0;						//  Count of requests on the queues
	int							iIndex;									//  Index of RQ elements
	RQElement					*pCandidateRQE;							//  Pointer to RQE slot
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pAnyThread;							//  Pointer to the selected thread
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;				//  Address the header

	//  Step 1:  Check the rejoin pool to see if there is a matching element
	pCandidateRQE = (RQElement *) pThreadPoolHeader->pRejoinRQEPool;	//  Point to the first element
	for (iIndex = 0; iIndex < pThreadPoolHeader->RejoinRQEPoolSize; iIndex++)
	{
		if ((pCandidateRQE->RQEState == RQESTATE_READY) && (pCandidateRQE->pOObject == pOwner))
		{
			RequestCount++;
		}
	}

	//  Step 2:  Check the worker threads
	pAnyThread = (Threadpool_Member *) (ThreadPool + sizeof(Threadpool_Header));
	for (iIndex = 0; iIndex < pThreadPoolHeader->MaxThreads; iIndex++)
	{
		if ((pAnyThread->TMemState == WTSTATE_BUSY) || (pAnyThread->TMemState == WTSTATE_POSTED) || (pAnyThread->TMemState == WTSTATE_REJOIN))
		{
			//  Check for an executing request (or posted for execution or a rejoining request)
			if (pAnyThread->rqePosted.pOObject == pOwner) RequestCount++;

			//  Now check if there is a request posted for the pending execution queue
			if ((pAnyThread->semReqDisp != REQ_SEM_FREE) && (pAnyThread->rqeDisp.pOObject == pOwner)) RequestCount++;
		}
		pAnyThread++;
	}

	//  Step 3:  Check the ready pool
	pCandidateRQE = (RQElement *) pThreadPoolHeader->pPendRQEPool;		//  Point to the first element
	for (iIndex = 0; iIndex < pThreadPoolHeader->PendRQEPoolSize; iIndex++)
	{
		if ((pCandidateRQE->RQEState == RQESTATE_READY) && (pCandidateRQE->pOObject == pOwner))
		{
			RequestCount++;
		}
		pCandidateRQE++;
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR2084T, pOwner, RequestCount);
	xeLocal->LogTrace(TRACE_MRQ, szMsg, iThreadID);

	return RequestCount;
}

//  CoreSetFastDrainMode
//
//  This function will block the dispatcher from starting any new requests and modify the behaviour of
//  the Rejoin functions so that they ignore queued requests..
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if mode was set successfully
//                  FALSE if the mode could not be set
//
//  API Use Notes:
//  Do Not use this function use the exposed SetFastDrainMode() function in MTExecutive
//
//
BOOL ThreadManager::CoreSetFastDrainMode()
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;				//  Address the header
	
	//  Block the Dipatcher
	pThreadPoolHeader->TDispState = TDSTATE_BLOCKED;					//  Set the dispatcher to blocked

	return TRUE;
}

//  CoreShowThreadStats
//
//  This function will display the Thread manager statistics
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::CoreShowThreadStats(int iOpts, int iThreadID)
{
	STATUS						stAPIRC;								//  Return code from Notes API calls
	INTLFORMAT					ifCurrent;								//  International Format options
	TFMT						tfLogTime;								//  Time/Date format options
	WORD						wRetLen;								//  Returned length
	int							iIndex;									//  Generic index
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Logpool_Header				*pLogPoolHeader;						//  Pointer to the Log Pool Header
	char						szTimeStamp[MAXALPHATIMEDATE + 1];		//  Formatted TimeDate
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;				//  Address the thread pool header
	pLogPoolHeader = (Logpool_Header *) pThreadPoolHeader->pLogPool;	//  Address the log pool haeder

	//  Initialise the format options
	OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
	tfLogTime.Date = TDFMT_FULL;
	tfLogTime.Time = TTFMT_FULL;
	tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
	tfLogTime.Structure = TSFMT_DATETIME;

	//  Show the thread manager stats for the current run
	szTimeStamp[0] = '\0';
	stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
									&tfLogTime,
									&pThreadPoolHeader->tdRunStart,
									szTimeStamp,
									sizeof(szTimeStamp),
									&wRetLen);
	if (ERR(stAPIRC) == NOERROR)
	{
		szTimeStamp[wRetLen] = '\0';
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR2049I, szTimeStamp);
	xeLocal->LogMessage(szMsg, iThreadID);
	//  The following are only shown in Debug mode
	if (iOpts & STATS_DEBUG)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2034I, pThreadPoolHeader->LogMsgQueued, pThreadPoolHeader->LogMsgWritten, pLogPoolHeader->LogPoolEntries);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2035I, pThreadPoolHeader->LogMsgDropped, pThreadPoolHeader->LogMsgDiscarded, pThreadPoolHeader->LogMsgOverflowed);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2036I, pThreadPoolHeader->BFDispatches, pThreadPoolHeader->PDispatches);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2037I, pThreadPoolHeader->DispBruteForce, pThreadPoolHeader->DispPriority);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2038I, pThreadPoolHeader->NoDispNoWork, pThreadPoolHeader->NoDispThreadsBusy);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2039I, pThreadPoolHeader->PDispQEmpty, pThreadPoolHeader->PDispThreadsBusy);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2050I, pThreadPoolHeader->BFDispThreadsBusy);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2081I, pThreadPoolHeader->RejoinRecs, pThreadPoolHeader->RejoinRecd);
		xeLocal->LogMessage(szMsg, iThreadID);
	}
	//  Write the dispatch request semaphore stats
	WriteDispSemStats(ThreadPool, pThreadPoolHeader->MaxThreads + 4, iOpts, iThreadID);
	//  The following lines are only shown if debug is set
	if (iOpts & STATS_DEBUG)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2043I, pThreadPoolHeader->RJRDequeued, pThreadPoolHeader->RJRBusy);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2048I, pThreadPoolHeader->PendRQECount, pThreadPoolHeader->RejoinRQECount, pThreadPoolHeader->AvailableThreads);
		xeLocal->LogMessage(szMsg, iThreadID);
	}
	//  Write the log semaphore usage message
	WriteLogSemStats(ThreadPool, pThreadPoolHeader->MaxThreads + 4, iOpts, iThreadID);
	//  Write the rejoin
	WriteRejoinSemStats(ThreadPool, pThreadPoolHeader->MaxThreads + 4, iOpts, iThreadID);
	WriteRejoinPollSemStats(ThreadPool, pThreadPoolHeader->MaxThreads + 4, iOpts, iThreadID);

	if (iOpts & STATS_DEBUG)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2052I, pThreadPoolHeader->PPFullWaits, pThreadPoolHeader->PPFullBackoffs, pThreadPoolHeader->PPLaneFullBackoffs, pThreadPoolHeader->CMLSDenials);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2053I, pThreadPoolHeader->RJPMutexAcqs, pThreadPoolHeader->RJPMutexWaits, pThreadPoolHeader->RJPMutexFails);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2054I, pThreadPoolHeader->DBHTMutexAcqs, pThreadPoolHeader->DBHTMutexWaits, pThreadPoolHeader->DBHTMutexFails);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2080I, pThreadPoolHeader->FHTMutexAcqs, pThreadPoolHeader->FHTMutexWaits, pThreadPoolHeader->FHTMutexFails);
		xeLocal->LogMessage(szMsg, iThreadID);

		for (iIndex = 0; iIndex <= MAX_APP_MUTEXID; iIndex++)
		{
			if (pThreadPoolHeader->AppMutexAcqs[iIndex] > 0)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2055I, iIndex + 1, pThreadPoolHeader->AppMutexAcqs[iIndex], pThreadPoolHeader->AppMutexWaits[iIndex], pThreadPoolHeader->AppMutexFails[iIndex]);
				xeLocal->LogMessage(szMsg, iThreadID);
			}
		}
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2044I, pThreadPoolHeader->Guard01, pThreadPoolHeader->Guard02, pThreadPoolHeader->Guard03, pThreadPoolHeader->Guard04, pThreadPoolHeader->Guard05);
		xeLocal->LogMessage(szMsg, iThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2045I, pThreadPoolHeader->Guard06, pThreadPoolHeader->Guard07, pThreadPoolHeader->Guard08, pThreadPoolHeader->Guard09, pThreadPoolHeader->Guard10);
		xeLocal->LogMessage(szMsg, iThreadID);
	}
	return;
}

//  CoreLogFlush
//
//  This function will trigger a flush of the log data
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::CoreLogFlush()
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Logpool_Header			*pLogPoolHeader;						//  Pointer to the Log Pool Header

	//  Post a message to the Logger queue
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;			//  Address the header
	pLogPoolHeader = (Logpool_Header *) pThreadPoolHeader->pLogPool;

	//  Set the flag in the Logpool Header
	pLogPoolHeader->LogQueueControl = pLogPoolHeader->LogQueueControl | LOG_CTL_FLUSH;
	return;
}

//  CoreDump
//
//  This function will trigger a disgnostic dump of key control blocks to the log
//
//  Parameters
//
//	1		-	int			-  Scope of the dump
//  2       -   int			-  ID of the thread making the request
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::CoreDump(int iDumpScope, int iThreadID)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	UCHAR					*pStartThread;							//  Pointer to the initial thead pool member
	Threadpool_Member		*pMyThread;								//  Pointer to the selected thread
	RQElement				*pRQE;									//  Pointer to a Request Element (RQE)
	int						iSavedLogLevel;							//  Old Log Level
	int						iIndex;									//  Generic Index
	char					szMsg[MAX_MSG + 1];						//  Generic Message buffer

	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;			//  Address the header

	//  Trigger a print of the current stats/settings to the log
	CoreShowThreadStats(0, 0);
	iSavedLogLevel = xeLocal->rsLocal->LogLevel;
	xeLocal->LogSetDebug(TRACE_MT);

	//  Dump the ThreadPool Header
	xeLocal->DumpMemory(ThreadPool, sizeof(Threadpool_Header), "Thread Pool Header", iThreadID);

	//  Dump the Thread Data - if requested
	if ((iDumpScope == DUMP_ALL) || (iDumpScope == DUMP_THREADS))
	{
		pStartThread = (UCHAR *) pThreadPoolHeader + sizeof(Threadpool_Header);
		pMyThread = (Threadpool_Member *) pStartThread;

		for (iIndex = 0; iIndex < pThreadPoolHeader->MaxThreads; iIndex++)
		{
			sprintf_s(szMsg, MAX_MSG, "Thread Pool Member #%i", iIndex + 1);
			xeLocal->DumpMemory(pMyThread, sizeof(Threadpool_Member), szMsg, iThreadID);
			pMyThread++;
		}
	}

	//  Dump the Request Pool Data - if requested
	if ((iDumpScope == DUMP_ALL) || (iDumpScope == DUMP_POOLS))
	{
		//  Firstly dump the Pending Pool
		pRQE = (RQElement *) pThreadPoolHeader->pPendRQEPool;
		for (iIndex = 0; iIndex < pThreadPoolHeader->PendRQEPoolSize; iIndex++)
		{
			sprintf_s(szMsg, MAX_MSG, "Pending Pool Request #%i", iIndex + 1);
			xeLocal->DumpMemory(pRQE, sizeof(RQElement), szMsg, iThreadID);
			pRQE++;
		}

		//  Now dump the Rejoin Pool
		pRQE = (RQElement *) pThreadPoolHeader->pRejoinRQEPool;
		for (iIndex = 0; iIndex < pThreadPoolHeader->RejoinRQEPoolSize; iIndex++)
		{
			sprintf_s(szMsg, MAX_MSG, "Rejoin Pool Request #%i", iIndex + 1);
			xeLocal->DumpMemory(pRQE, sizeof(RQElement), szMsg, iThreadID);
			pRQE++;
		}
	}

	//  Dump the DB Handle Table if requested
	if ((iDumpScope == DUMP_ALL) || (iDumpScope == DUMP_DBHT))
	{
		xeLocal->DumpMemory(pThreadPoolHeader->pDBHMappingPool, (Policy->MaxThreads + 1) * MAX_DBH_CACHE * sizeof(DBHANDLE), "DB Handle Mapping Table", iThreadID);
	}

	//  Reset the log level
	xeLocal->rsLocal->LogLevel = iSavedLogLevel;
	return;
}

//  InitMembers
//
//  This function will intitalise the individual thread entries in the Thread Pool
//
//  Parameters
//
//	1		-		UCHAR *		-  Pointer to the thread pool header
//  2		-		int			-  Count of members to be initialised
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::InitMembers(UCHAR *pTP, int iTCount)
{
	UCHAR						*pTPM;									//  Pointer to a Thread Pool Member
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a Thread Pool Member
	int							iIndex;									//  Generic index

	pTPM = pTP + sizeof(Threadpool_Header);								//  Point to first entry
	for (iIndex = 0; iIndex < iTCount; iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) pTPM;					//  Cast pointer to current entry
		//  Initialise elements for this member
		pThreadPoolMember->semLogMsg = LOG_SEM_FREE;					//  Mark the semaphore as free
		pThreadPoolMember->semReqDisp = REQ_SEM_FREE;					//  Mark the semaphore as free
		pThreadPoolMember->semRejoin = REJ_SEM_FREE;					//  Mark the semaphore as free
		pThreadPoolMember->semRejoinPoll = REJPOLL_SEM_FREE;			//  Mark the semaphore as free
#ifdef DXIP
		//  Initialise the Instrumentation Package Section
		pThreadPoolMember->tToIdle = 0;
		pThreadPoolMember->tToActive = 0;
		pThreadPoolMember->tRetToIdle = 0;
		pThreadPoolMember->EventCount = 0;
		pThreadPoolMember->LastCollection = 0;
		memset(&pThreadPoolMember->CStack[0], 0, sizeof(PCRecRC) * IPC_STACK_SIZE);
#endif
		pTPM = pTPM + sizeof(Threadpool_Member);						//  Point to the next member
	}
	return;
}

//  GetMemberAddress
//
//  This function will return the address of the indicated thread pool member
//
//  Parameters
//
//  1		-		int			-  id of the referenced thread 
//
//  Returns
//
//  API Use Notes:
//
//
UCHAR * ThreadManager::GetMemberAddress(int iThreadID)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	UCHAR					*pTPM;									//  Pointer to a Thread Pool Member
	int						iThreadIndex;							//  Thread Index
	
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;			//  Address the header

	//  Adjust the thread ID if it is one of the control or management threads
	switch(iThreadID)
	{
	case CONTROL_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads;
		break;
	case MONITOR_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 1;
		break;
	case SCHEDULER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 2;
		break;
	case DISPATCHER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 3;
		break;
	default:
		iThreadIndex = iThreadID - 1;
	}

	//  Compute the address of the referenced member
	pTPM = ThreadPool + sizeof(Threadpool_Header) + (iThreadIndex * sizeof(Threadpool_Member));
	return pTPM;
}

//  WriteLogSemStats
//
//  This function will write the accumulated stats from the log semaphore usage
//
//  Parameters
//
//	1		-		UCHAR *		-  Pointer to the thread pool header
//  2		-		int			-  Count of members to be displayed
//  3       -       int			-  Display options STATS_THREAD causes a per thread message to be displayed
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::WriteLogSemStats(UCHAR *pTP, int iTCount, int iOpts, int iThreadID)
{
	UCHAR						*pTPM;									//  Pointer to a Thread Pool Member
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a Thread Pool Member
	int							iIndex;									//  Generic index
	int							iThreadNo;								//  Identity of the thread
	DWORD						dwReqs;									//  Count of requests
	DWORD						dwWaits;								//  Count of Waits
	DWORD						dwWaitQs;								//  Count of Wait Quanta
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	dwReqs = 0;
	dwWaits = 0;
	dwWaitQs = 0;
	iThreadNo = 1;
	pThreadPoolHeader = (Threadpool_Header *) pTP;						//  Address the header

	pTPM = pTP + sizeof(Threadpool_Header);								//  Point to first entry
	for (iIndex = 0; iIndex < iTCount; iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) pTPM;					//  Cast pointer to current entry
		//  Accumulate the counters
		dwReqs = dwReqs + pThreadPoolMember->dwLogMsgsPosted;
		dwWaits = dwWaits + pThreadPoolMember->dwLogSemWaits;
		dwWaitQs = dwWaitQs + pThreadPoolMember->dwLogSemWaitQuanta;
		//  If the STATS_THREAD option is set then show the per thread stats as well
		if (iOpts & STATS_THREAD)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2061I, iThreadNo, pThreadPoolMember->dwLogMsgsPosted, pThreadPoolMember->dwLogSemWaits, pThreadPoolMember->dwLogSemWaitQuanta, pThreadPoolMember->semLogMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		pTPM = pTPM + sizeof(Threadpool_Member);						//  Point to the next member

		//  Compute the next thread ID to be displayed
		switch(iThreadNo)
		{
		case CONTROL_THREAD_ID:
			iThreadNo = MONITOR_THREAD_ID;
			break;
		case MONITOR_THREAD_ID:
			iThreadNo = SCHEDULER_THREAD_ID;
			break;
		case SCHEDULER_THREAD_ID:
			iThreadNo = DISPATCHER_THREAD_ID;
			break;
		case DISPATCHER_THREAD_ID:
			iThreadNo = DISPATCHER_THREAD_ID;
			break;
		default:
			iThreadNo++;
			if (iThreadNo > pThreadPoolHeader->MaxThreads) iThreadNo = CONTROL_THREAD_ID;
			break;
		}
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR2051I, dwReqs, dwWaits, dwWaitQs);
	xeLocal->LogMessage(szMsg, iThreadID);

	return;
}

//  WriteDispSemStats
//
//  This function will write the accumulated stats from the request dispatch semaphore usage
//
//  Parameters
//
//	1		-		UCHAR *		-  Pointer to the thread pool header
//  2		-		int			-  Count of members to be displayed
//  3       -       int			-  Display options STATS_THREAD causes a per thread message to be displayed
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::WriteDispSemStats(UCHAR *pTP, int iTCount, int iOpts, int iThreadID)
{
	UCHAR						*pTPM;									//  Pointer to a Thread Pool Member
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a Thread Pool Member
	int							iIndex;									//  Generic index
	int							iThreadNo;								//  Identity of the thread
	DWORD						dwProc;									//  Count of requests processed
	DWORD						dwReqs;									//  Count of requests
	DWORD						dwWaits;								//  Count of Waits
	DWORD						dwWaitQs;								//  Count of Wait Quanta
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	dwProc = 0;
	dwReqs = 0;
	dwWaits = 0;
	dwWaitQs = 0;
	iThreadNo = 1;
	pThreadPoolHeader = (Threadpool_Header *) pTP;						//  Address the header
	pTPM = pTP + sizeof(Threadpool_Header);								//  Point to first entry
	for (iIndex = 0; iIndex < iTCount; iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) pTPM;					//  Cast pointer to current entry
		//  Accumulate the counters
		dwProc = dwProc + pThreadPoolMember->dwReqsProc;
		dwReqs = dwReqs + pThreadPoolMember->dwReqDisp;
		dwWaits = dwWaits + pThreadPoolMember->dwReqDispWaits;
		dwWaitQs = dwWaitQs + pThreadPoolMember->dwReqDispWaitQuanta;
		//  If the STATS_THREAD option is set then show the per thread stats as well
		if (iOpts & STATS_THREAD)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2062I, iThreadNo, pThreadPoolMember->dwReqsProc, pThreadPoolMember->dwReqDisp, pThreadPoolMember->dwReqDispWaits, pThreadPoolMember->dwReqDispWaitQuanta, pThreadPoolMember->semReqDisp);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		pTPM = pTPM + sizeof(Threadpool_Member);						//  Point to the next member

		//  Compute the next thread ID to be displayed
		switch(iThreadNo)
		{
		case CONTROL_THREAD_ID:
			iThreadNo = MONITOR_THREAD_ID;
			break;
		case MONITOR_THREAD_ID:
			iThreadNo = SCHEDULER_THREAD_ID;
			break;
		case SCHEDULER_THREAD_ID:
			iThreadNo = DISPATCHER_THREAD_ID;
			break;
		case DISPATCHER_THREAD_ID:
			iThreadNo = DISPATCHER_THREAD_ID;
			break;
		default:
			iThreadNo++;
			if (iThreadNo > pThreadPoolHeader->MaxThreads) iThreadNo = CONTROL_THREAD_ID;
			break;
		}
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR2040I, dwProc, dwReqs, dwWaits, dwWaitQs);
	xeLocal->LogMessage(szMsg, iThreadID);

	return;
}

//  WriteRejoinSemStats
//
//  This function will write the accumulated stats from the request dispatch semaphore usage
//
//  Parameters
//
//	1		-		UCHAR *		-  Pointer to the thread pool header
//  2		-		int			-  Count of members to be displayed
//  3       -       int			-  Display options STATS_THREAD causes a per thread message to be displayed
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::WriteRejoinSemStats(UCHAR *pTP, int iTCount, int iOpts, int iThreadID)
{
	UCHAR						*pTPM;									//  Pointer to a Thread Pool Member
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a Thread Pool Member
	int							iIndex;									//  Generic index
	int							iThreadNo;								//  Identity of the thread
	DWORD						dwProc;									//  Count of requests processed
	DWORD						dwReqs;									//  Count of requests
	DWORD						dwWaits;								//  Count of Waits
	DWORD						dwWaitQs;								//  Count of Wait Quanta
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	dwProc = 0;
	dwReqs = 0;
	dwWaits = 0;
	dwWaitQs = 0;
	iThreadNo = 1;
	pThreadPoolHeader = (Threadpool_Header *) pTP;						//  Address the header
	pTPM = pTP + sizeof(Threadpool_Header);								//  Point to first entry
	for (iIndex = 0; iIndex < iTCount; iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) pTPM;					//  Cast pointer to current entry
		//  Accumulate the counters
		dwProc = dwProc + pThreadPoolMember->dwReqsProc;
		dwReqs = dwReqs + pThreadPoolMember->dwRejoin;
		dwWaits = dwWaits + pThreadPoolMember->dwRejoinWaits;
		dwWaitQs = dwWaitQs + pThreadPoolMember->dwRejoinWaitQuanta;
		//  If the STATS_THREAD option is set then show the per thread stats as well
		if (iOpts & STATS_THREAD)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2063I, iThreadNo, pThreadPoolMember->dwReqsProc, pThreadPoolMember->dwRejoin, pThreadPoolMember->dwRejoinWaits, pThreadPoolMember->dwRejoinWaitQuanta, pThreadPoolMember->semRejoin);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		pTPM = pTPM + sizeof(Threadpool_Member);						//  Point to the next member

		//  Compute the next thread ID to be displayed
		switch(iThreadNo)
		{
		case CONTROL_THREAD_ID:
			iThreadNo = MONITOR_THREAD_ID;
			break;
		case MONITOR_THREAD_ID:
			iThreadNo = SCHEDULER_THREAD_ID;
			break;
		case SCHEDULER_THREAD_ID:
			iThreadNo = DISPATCHER_THREAD_ID;
			break;
		case DISPATCHER_THREAD_ID:
			iThreadNo = DISPATCHER_THREAD_ID;
			break;
		default:
			iThreadNo++;
			if (iThreadNo > pThreadPoolHeader->MaxThreads) iThreadNo = CONTROL_THREAD_ID;
			break;
		}
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR2064I, dwProc, dwReqs, dwWaits, dwWaitQs);
	xeLocal->LogMessage(szMsg, iThreadID);

	return;
}

//  WriteRejoinPollSemStats
//
//  This function will write the accumulated stats from the request dispatch semaphore usage
//
//  Parameters
//
//	1		-		UCHAR *		-  Pointer to the thread pool header
//  2		-		int			-  Count of members to be displayed
//  3       -       int			-  Display options STATS_THREAD causes a per thread message to be displayed
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::WriteRejoinPollSemStats(UCHAR *pTP, int iTCount, int iOpts, int iThreadID)
{
	UCHAR						*pTPM;									//  Pointer to a Thread Pool Member
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a Thread Pool Member
	int							iIndex;									//  Generic index
	int							iThreadNo;								//  Identity of the thread
	DWORD						dwProc;									//  Count of requests processed
	DWORD						dwReqs;									//  Count of requests
	DWORD						dwWaits;								//  Count of Waits
	DWORD						dwWaitQs;								//  Count of Wait Quanta
	DWORD						dwRet;									//  Rejoins Returned
	DWORD						dwNotReady;								//  Not ready
	DWORD						dwNone;									//  None exist
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	dwProc = 0;
	dwReqs = 0;
	dwWaits = 0;
	dwWaitQs = 0;
	dwRet = 0;
	dwNotReady = 0;
	dwNone = 0;

	iThreadNo = 1;
	pThreadPoolHeader = (Threadpool_Header *) pTP;						//  Address the header
	pTPM = pTP + sizeof(Threadpool_Header);								//  Point to first entry
	for (iIndex = 0; iIndex < iTCount; iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) pTPM;					//  Cast pointer to current entry
		//  Accumulate the counters
		dwProc = dwProc + pThreadPoolMember->dwReqsProc;
		dwReqs = dwReqs + pThreadPoolMember->dwRejoinPoll;
		dwWaits = dwWaits + pThreadPoolMember->dwRejoinPollWaits;
		dwWaitQs = dwWaitQs + pThreadPoolMember->dwRejoinPollWaitQuanta;
		dwRet = dwRet + pThreadPoolMember->dwRejoinReturned;
		dwNotReady = dwNotReady + pThreadPoolMember->dwRejoinNotReady;
		dwNone = dwNone + pThreadPoolMember->dwRejoinNone;
		//  If the STATS_THREAD option is set then show the per thread stats as well
		if (iOpts & STATS_THREAD)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2065I, iThreadNo, pThreadPoolMember->dwReqsProc, pThreadPoolMember->dwRejoinPoll, pThreadPoolMember->dwRejoinPollWaits, pThreadPoolMember->dwRejoinPollWaitQuanta, pThreadPoolMember->semRejoinPoll);
			xeLocal->LogMessage(szMsg, iThreadID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2067I, iThreadNo, pThreadPoolMember->dwRejoinPoll, pThreadPoolMember->dwRejoinReturned, pThreadPoolMember->dwRejoinNotReady, pThreadPoolMember->dwRejoinNone);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		pTPM = pTPM + sizeof(Threadpool_Member);						//  Point to the next member

		//  Compute the next thread ID to be displayed
		switch(iThreadNo)
		{
		case CONTROL_THREAD_ID:
			iThreadNo = MONITOR_THREAD_ID;
			break;
		case MONITOR_THREAD_ID:
			iThreadNo = SCHEDULER_THREAD_ID;
			break;
		case SCHEDULER_THREAD_ID:
			iThreadNo = DISPATCHER_THREAD_ID;
			break;
		case DISPATCHER_THREAD_ID:
			iThreadNo = DISPATCHER_THREAD_ID;
			break;
		default:
			iThreadNo++;
			if (iThreadNo > pThreadPoolHeader->MaxThreads) iThreadNo = CONTROL_THREAD_ID;
			break;
		}
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR2066I, dwProc, dwReqs, dwWaits, dwWaitQs);
	xeLocal->LogMessage(szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2068I, dwReqs, dwRet, dwNotReady, dwNone);
	xeLocal->LogMessage(szMsg, iThreadID);

	return;
}

//  CoreSetLastOpenError
//
//  This function will store the last open error code so that it can be interrogated by the caller
//
//  Parameters
//
//  1  -  STATUS	-  Last error code
//  2  -  int		-  Thread ID of the requesting thread
//
//  Returns
//
//
//  API Use Notes:
//
//
void ThreadManager::CoreSetLastOpenError(STATUS stOpen, int iThreadID)
{
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a Thread Pool Member

	pThreadPoolMember = (Threadpool_Member *) GetMemberAddress(iThreadID);
	pThreadPoolMember->stLastOpenError = stOpen;
	return;
}

//  CoreGetLastOpenError
//
//  This function will store the last open error code so that it can be interrogated by the caller
//
//  Parameters
//
//  1  -  int		-  Thread ID of the requesting thread
//
//  Returns
//
//	STATUS		-  The last API return code from D/B Open calls
//
//  API Use Notes:
//
//
STATUS ThreadManager::CoreGetLastOpenError(int iThreadID)
{
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a Thread Pool Member

	pThreadPoolMember = (Threadpool_Member *) GetMemberAddress(iThreadID);
	return pThreadPoolMember->stLastOpenError;
}

//  CoreAttachCommandHandler
//
//  This function will attach the passed Command Handler to the Thread Monitor
//
//  Parameters
//
//  1  -  CommandHandler *		-	Pointer to Command Handler to be attached
//
//  Returns
//
//
//  API Use Notes:
//
//
void ThreadManager::CoreAttachCommandHandler(CommandHandler *chApp)
{
	tmonMaster->AttachCommandHandler(chApp);
	return;
}

//  PerformCMLSInits
//
//  This function will perform initialisation functions associated with running in constricted multi-lane mode
//
//  Parameters
//
//
//  Returns
//
//
//  API Use Notes:
//
//
void ThreadManager::PerformCMLSInits()
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;

	//  Set the Max Thread contraints from the Policy
	if (Policy->MaxPctL0Threads == 0) 	Policy->MaxPctL0Threads = 100;
	if (Policy->MaxPctL1Threads == 0) 	Policy->MaxPctL1Threads = 100;
	if (Policy->MaxPctL2Threads == 0) 	Policy->MaxPctL2Threads = 100;

	pThreadPoolHeader->MaxL0Threads = (Policy->MaxPctL0Threads * pThreadPoolHeader->MaxThreads) / 100;
	pThreadPoolHeader->MaxL1Threads = (Policy->MaxPctL1Threads * pThreadPoolHeader->MaxThreads) / 100;
	pThreadPoolHeader->MaxL2Threads = (Policy->MaxPctL2Threads * pThreadPoolHeader->MaxThreads) / 100;

	//  Compute the max entries in the pending pool
	pThreadPoolHeader->MaxL0PPRQEs = (Policy->MaxPctL0Threads * pThreadPoolHeader->PendRQEPoolSize) / 100;
	pThreadPoolHeader->MaxL1PPRQEs = (Policy->MaxPctL1Threads * pThreadPoolHeader->PendRQEPoolSize) / 100;
	pThreadPoolHeader->MaxL2PPRQEs = (Policy->MaxPctL2Threads * pThreadPoolHeader->PendRQEPoolSize) / 100;

	//  Trace Activation
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2070T, pThreadPoolHeader->MaxL0Threads, pThreadPoolHeader->MaxL0PPRQEs, 
					pThreadPoolHeader->MaxL1Threads, pThreadPoolHeader->MaxL1PPRQEs,
					pThreadPoolHeader->MaxL2Threads, pThreadPoolHeader->MaxL2PPRQEs);
	xeLocal->LogTrace(TRACE_CMLS, szMsg, 0);

	//  Clear the current counters
	pThreadPoolHeader->CurrL0Threads = 0;
	pThreadPoolHeader->CurrL1Threads = 0;
	pThreadPoolHeader->CurrL2Threads = 0;
	pThreadPoolHeader->CurrL0PPRQEs = 0;
	pThreadPoolHeader->CurrL1PPRQEs = 0;
	pThreadPoolHeader->CurrL2PPRQEs = 0;

	return;
}

//  CoreGetCMLSRecommendation
//
//  This function will determine if lane restrictions are in place for the indicated CMLS execution level
//
//  Parameters
//
//  1  -  int		-  The CMLS Level (Lane) to be checked
//  2  -  int		-  Thread ID of the requesting thread
//
//  Returns
//
//	int		-  The CMLS dispatch recommendation
//
//  API Use Notes:
//
//
int ThreadManager::CoreGetCMLSRecommendation(int iLevel, int iThreadID)
{
	return tdMaster->GetCMLSRecommendation(ThreadPool, iLevel, iThreadID);
}

//  CoreGetThreadCount
//
//  This function will return the current threadcount
//
//  Parameters
//
//
//  Returns
//
//	int		-  The current count of available threads
//
//  API Use Notes:
//
//
int ThreadManager::CoreGetThreadCount()
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;
	return pThreadPoolHeader->MaxThreads;
}

//  CoreGetMappedFileHandle
//
//  This function will return the File Handle that can be used on the current thread that maps to
//  the file identified by the native file Handle that was passed.
//
//  Parameters
//
//	1  -  HANDLE	- Native file handle for the requested file   (int for UNIX)
//  2  -  char *	- Pointer to the name of the file
//  3  -  DWORD		- Access Flags   (int for UNIX)
//  4  -  DWORD		- Sharing Mode   (omitted for UNIX)
//  5  -  int		ID of the thread that is making the request
//
//  Returns
//
//  HANDLE  -		Local (to this thread) file Handle for the file   (int for UNIX)
//                  NULLHANDLE - A local File Handle could not be made available (NULL for UNIX)
//
//  API Use Notes:
//
//
#ifdef WIN32
HANDLE ThreadManager::CoreGetMappedFileHandle(HANDLE hfNative, char *szFile, DWORD dwAccess, DWORD dwSharing, int iThreadID)
{	
	int					iIndex;									//  Generic index variable
	int					iIndexX;								//  Index of native File Handles
	int					iIndexFree;								//  Index of free native File handle slot
	int					iThreadIndex;							//  Thread Index
	HANDLE				hfTemp;									//  Temporary File handle
	HANDLE				*pMappedFH;								//  Pointer to my mapped File Handle entry
	HANDLE				*pMyMappedFH;							//  Pointer to my mapped File Handle entry
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  If the action is called on the main application thread (0) then return the native handle
	if (iThreadID == 0) return hfNative;

	//  Compute the Thread Index
	switch(iThreadID)
	{
	case CONTROL_THREAD_ID:
		iThreadIndex = CONTROL_THREAD_ID;
		break;
	case MONITOR_THREAD_ID:
		iThreadIndex = Policy->MaxThreads + 1;
		break;
	case SCHEDULER_THREAD_ID:
		iThreadIndex = Policy->MaxThreads + 2;
		break;
	case DISPATCHER_THREAD_ID:
		iThreadIndex = Policy->MaxThreads + 3;
		break;
	default:
		iThreadIndex = iThreadID;
		break;
	}

	//  Determine if the Native handle has already been cached
	iIndexX = -1;
	iIndexFree = -1;
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;
	pMappedFH = (HANDLE *) pThreadPoolHeader->pFHMappingPool;

	//  First perform the search of the File Handle table without holding the mutex
	//  the search will have to be re-done with the mutex held if a new entry needs to be added

	for (iIndex = 0; iIndex < MAX_FH_CACHE; iIndex++)
	{
		pMyMappedFH = pMappedFH + (iIndex * (Policy->MaxThreads + 4));
		if (pMyMappedFH[0] == hfNative)
		{
			iIndexX = iIndex;
			break;
		}
		if ((pMyMappedFH[0] == NULLHANDLE) && (iIndexFree == -1)) iIndexFree = iIndex;
	}

	//  If there is a row for this native handle then determine if there is already a thread local
	//  handle avaiilable for this thread.
	if (iIndexX != -1)
	{
		pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 4));
		if (pMyMappedFH[iThreadIndex] != NULLHANDLE)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2073T, iThreadID, pMyMappedFH[iThreadIndex], hfNative);
			xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
			return pMyMappedFH[iThreadIndex];
		}
	}

	//  Acquire the DB Handle Table Mutex
	if (!AcquireFHTMutex(ThreadPool, iThreadID))
	{
		xeLocal->LogMessage(MSG_DXR2074E, iThreadID);
		return NULLHANDLE;
	}

	//  Determine if the Native handle has already been cached
	for (iIndex = 0; iIndex < MAX_FH_CACHE; iIndex++)
	{
		pMyMappedFH = pMappedFH + (iIndex * (Policy->MaxThreads + 4));
		if (pMyMappedFH[0] == hfNative)
		{
			iIndexX = iIndex;
			break;
		}
		if ((pMyMappedFH[0] == NULLHANDLE) && (iIndexFree == -1)) iIndexFree = iIndex;
	}

	//  If there is no cache line for this handle then allocate one
	if (iIndexX == -1)
	{
		//  Native handle is not in the cache
		if (iIndexFree == -1)
		{
			//  There are no free slots in the table
			xeLocal->LogMessage(MSG_DXR2075E, iThreadID);
			pThreadPoolHeader->FHTableMutex = LOG_MUTEX_FREE;
			return NULLHANDLE;
		}
		//  Assign the row in the cache to the native handle
		iIndexX = iIndexFree;
		pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 4));
		pMyMappedFH[0] = hfNative;
	}

	//  Check if there is already a thread local handle for the requesting thread
	pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 4));
	if (pMyMappedFH[iThreadIndex] != NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2073T, iThreadID, pMyMappedFH[iThreadIndex], hfNative);
		xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
		pThreadPoolHeader->FHTableMutex = LOG_MUTEX_FREE;
		return pMyMappedFH[iThreadIndex];
	}

	//  Free the Mutex at this point as there is no conflict
	pThreadPoolHeader->FHTableMutex = LOG_MUTEX_FREE;

	//  Reopen the native handle to create a thread local handle
	hfTemp = ReopenFile(hfNative, szFile, dwAccess, dwSharing, iThreadID);
	if (hfTemp == NULLHANDLE)
	{
		xeLocal->LogMessage(MSG_DXR2076E, iThreadID);
		return NULLHANDLE;
	}

	//  Store the thread local handle and return it
	pMyMappedFH[iThreadIndex] = hfTemp;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2077T, hfTemp, hfNative, iThreadID);
	xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
	return hfTemp;
}
#endif
#ifdef UNIX
int ThreadManager::CoreGetMappedFileHandle(int fNative, char *szFile, int iFlags, int iThreadID)
{	
	int					iIndex;									//  Generic index variable
	int					iIndexX;								//  Index of native File Handles
	int					iIndexFree;								//  Index of free native File handle slot
	int					fTemp;									//  Temporary File handle
	int					*pMappedFH;								//  Pointer to my mapped File Handle entry
	int					*pMyMappedFH;							//  Pointer to my mapped File Handle entry
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  If the action is called on the main application thread (0) then return the native handle
	if (iThreadID == 0) return fNative;

	//  Determine if the Native handle has already been cached
	iIndexX = -1;
	iIndexFree = -1;
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;
	pMappedFH = (int *) pThreadPoolHeader->pFHMappingPool;

	//  First perform the search of the File Handle table without holding the mutex
	//  the search will have to be re-done with the mutex held if a new entry needs to be added

	for (iIndex = 0; iIndex < MAX_FH_CACHE; iIndex++)
	{
		pMyMappedFH = pMappedFH + (iIndex * (Policy->MaxThreads + 1));
		if (pMyMappedFH[0] == fNative)
		{
			iIndexX = iIndex;
			break;
		}
		if ((pMyMappedFH[0] == NULL) && (iIndexFree == -1)) iIndexFree = iIndex;
	}

	//  If there is a row for this native handle then determine if there is already a thread local
	//  handle avaiilable for this thread.
	if (iIndexX != -1)
	{
		pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 1));
		if (pMyMappedFH[iThreadID] != NULL)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2073T, iThreadID, pMyMappedFH[iThreadID], fNative);
			xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
			return pMyMappedFH[iThreadID];
		}
	}

	//  Acquire the DB Handle Table Mutex
	if (!AcquireFHTMutex(ThreadPool, iThreadID))
	{
		xeLocal->LogMessage(MSG_DXR2074E, iThreadID);
		return NULLHANDLE;
	}

	//  Determine if the Native handle has already been cached
	for (iIndex = 0; iIndex < MAX_FH_CACHE; iIndex++)
	{
		pMyMappedFH = pMappedFH + (iIndex * (Policy->MaxThreads + 1));
		if (pMyMappedFH[0] == fNative)
		{
			iIndexX = iIndex;
			break;
		}
		if ((pMyMappedFH[0] == NULL) && (iIndexFree == -1)) iIndexFree = iIndex;
	}

	//  If there is no cache line for this handle then allocate one
	if (iIndexX == -1)
	{
		//  Native handle is not in the cache
		if (iIndexFree == -1)
		{
			//  There are no free slots in the table
			xeLocal->LogMessage(MSG_DXR2075E, iThreadID);
			pThreadPoolHeader->FHTableMutex = LOG_MUTEX_FREE;
			return NULL;
		}
		//  Assign the row in the cache to the native handle
		iIndexX = iIndexFree;
		pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 1));
		pMyMappedFH[0] = fNative;
	}

	//  Check if there is already a thread local handle for the requesting thread
	pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 1));
	if (pMyMappedFH[iThreadID] != NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2073T, iThreadID, pMyMappedFH[iThreadID], fNative);
		xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
		pThreadPoolHeader->FHTableMutex = LOG_MUTEX_FREE;
		return pMyMappedFH[iThreadID];
	}

	//  Free the Mutex at this point as there is no conflict
	pThreadPoolHeader->FHTableMutex = LOG_MUTEX_FREE;

	//  Reopen the native handle to create a thread local handle
	fTemp = ReopenFile(fNative, szFile, iFlags, iThreadID);
	if (fTemp == NULL)
	{
		xeLocal->LogMessage(MSG_DXR2076E, iThreadID);
		return NULL;
	}

	//  Store the thread local handle and return it
	pMyMappedFH[iThreadID] = fTemp;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2077T, fTemp, fNative, iThreadID);
	xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
	return fTemp;
}
#endif

//  AcquireFHTMutex
//
//  This function will gain control of the mutex that protects the File Handle Mapping pool structures.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  int						-  ID of the thread on behalf of which the mutex is being acquired
//
//  Returns
//
//  BOOL			-	TRUE if the mutex has been acquired
//					-	FALSE if the mutex could not be acuired within the allowed time
//
//  API Use Notes:
//  
//
BOOL ThreadManager::AcquireFHTMutex(UCHAR *pTP, int iThreadID)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	int						iWaitCount = 0;							//  Counts of waits on the Runtime Log Mutex

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	//  Increment request count
	pThreadPoolHeader->FHTMutexAcqs++;

	//  Acquire the Runtime DB Handle Table Mutex on behalf of the thread
	while (pThreadPoolHeader->FHTableMutex != iThreadID)
	{
#ifdef DXMUTEX_STRENGTH_PUNY
		while ((pThreadPoolHeader->FHTableMutex != LOG_MUTEX_FREE) && (iWaitCount < Policy->FHTMutexWaitLimit))
		{
			pThreadPoolHeader->FHTMutexWaits++;
			iWaitCount++;
			SLEEP(Policy->APPMutexWaitInterval);
		}
		if (iWaitCount == Policy->FHTMutexWaitLimit)
		{
			pThreadPoolHeader->FHTMutexFails++;
			return FALSE;			//  Timed out
		}
		pThreadPoolHeader->FHTableMutex = iThreadID;					//  Mark the mutex as owned
#else
#ifdef WIN32
		while ((InterlockedCompareExchange(&pThreadPoolHeader->FHTableMutex, iThreadID, LOG_MUTEX_FREE) != LOG_MUTEX_FREE) && (iWaitCount < Policy->FHTMutexWaitLimit))
		{
			pThreadPoolHeader->FHTMutexWaits++;
			iWaitCount++;
			SLEEP(Policy->FHTMutexWaitInterval);
		}
		if (iWaitCount == Policy->FHTMutexWaitLimit)
		{
			pThreadPoolHeader->FHTMutexFails++;
			return FALSE;				//  Timed out
		}
#endif
#ifdef UNIX
		while ((DXAtomic_cmpxchg(&pThreadPoolHeader->FHTableMutex, LOG_MUTEX_FREE, iThreadID) != LOG_MUTEX_FREE) && (iWaitCount < Policy->FHTMutexWaitLimit))
		{
			pThreadPoolHeader->FHTMutexWaits++;
			iWaitCount++;
			SLEEP(Policy->FHTMutexWaitInterval);
		}
		if (iWaitCount == Policy->FHTMutexWaitLimit)
		{
			pThreadPoolHeader->FHTMutexFails++;
			return FALSE;				//  Timed out
		}
#endif
#endif
	}
	return TRUE;
}

//  ReopenFile
//
//  This function will Reopen an OS file for use as a thread local handle
//
//  Parameters
//
//	1  -  HANDLE					-  Handle of the Native File that is to be reopened   (int for UNIX)
//  2  -  char *					-  Pointer to the File name
//  2  -  DWORD						-  Flags indicationg what access is required in the file
//  3  -  DWORD						-  Flags indicating waht sharing mode should be applied
//  4  -  int						-  ID of the thread on behalf of which the file will be reopened
//
//  Returns
//
//  HANDLE			-	Handle of the reopened file		(int for UNIX)
//					-	NULLHANDLE if the file could not be reopened    (NULL for UNIX)
//
//  API Use Notes:
//  
//
#ifdef WIN32
HANDLE	ThreadManager::ReopenFile(HANDLE hfNative, char *szFile, DWORD dwAccess, DWORD dwSharing, int iThreadID)
{
	HANDLE			hfTemp;									//  Handle for open
	DWORD			dwWinError;								//  Windows error code
	char			szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  Show reopening
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2081T, szFile, iThreadID);
	xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);

	//  Reopen the file 
	hfTemp = CreateFile(szFile, dwAccess, dwSharing, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hfTemp == (HANDLE) HFILE_ERROR)
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2079E, szFile);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Return the handle
	return hfTemp;
}
#endif
#ifdef UNIX
int	ThreadManager::ReopenFile(int hfNative, char *szFile, int iFlags, int iThreadID)
{
	int					fdSrc;									//  File Descriptor for the mapped file handle
	char				szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	fdSrc = open(szFile, O_RDONLY | O_LARGEFILE);
	if (fdSrc == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2079E, szFile);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		fdSrc = NULL;
	}
	return fdSrc;
}
#endif

//  CoreInvalidateNativeFileHandle
//
//  This function will mark a Native File Handle as no longer needed in the File Handle mapping table.
//  This will allow any threads to release the per thread handle opened for the file.
//
//  Parameters
//
//	1  -  HANDLE	Native File handle that is no longer needed   (int for UNIX)
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the File Handle was marked as no longer needed
//                  FALSE if the File Handle could not be marked as no longer needed
//
//  API Use Notes:
//
//
#ifdef WIN32
BOOL ThreadManager::CoreInvalidateNativeFileHandle(HANDLE hfNative, int iThreadID)
{
	int					iIndex;									//  Generic index variable
	int					iIndexX;								//  Index of native DB Handles
	HANDLE				*pMappedFH;								//  Pointer to a mapped File entry
	HANDLE				*pMyMappedFH;							//  Pointer to a mapped File entry
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Determine if there is a cache row for the Native handle
	iIndexX = -1;
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;
	pMappedFH = (HANDLE *) pThreadPoolHeader->pFHMappingPool;

	for (iIndex = 0; iIndex < MAX_FH_CACHE; iIndex++)
	{
		pMyMappedFH = pMappedFH + (iIndex * (Policy->MaxThreads + 1));
		if (pMyMappedFH[0] == hfNative)
		{
			iIndexX = iIndex;
			break;
		}
	}

	//  Handle is not cached - which is perfectly valid
	if (iIndexX == -1) return TRUE;

	//  Mark the entry to be freed by each thread
	pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 1));
	pMyMappedFH[0] = (HANDLE) FREEHANDLE;

	//  Trace the action
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2082T, hfNative);
	xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);

	//  Check for any entries that can be released
	ReleaseAllMappedFileHandles(iThreadID);
	return TRUE;
}
#endif
#ifdef UNIX
BOOL ThreadManager::CoreInvalidateNativeFileHandle(int fNative, int iThreadID)
{
	int					iIndex;									//  Generic index variable
	int					iIndexX;								//  Index of native DB Handles
	HANDLE				*pMappedFH;								//  Pointer to a mapped File entry
	HANDLE				*pMyMappedFH;							//  Pointer to a mapped File entry
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Determine if there is a cache row for the Native handle
	iIndexX = -1;
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;
	pMappedFH = (HANDLE *) pThreadPoolHeader->pFHMappingPool;

	for (iIndex = 0; iIndex < MAX_FH_CACHE; iIndex++)
	{
		pMyMappedFH = pMappedFH + (iIndex * (Policy->MaxThreads + 1));
		if (pMyMappedFH[0] == fNative)
		{
			iIndexX = iIndex;
			break;
		}
	}

	//  Handle is not cached - which is perfectly valid
	if (iIndexX == -1) return TRUE;

	//  Mark the entry to be freed by each thread
	pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 1));
	pMyMappedFH[0] = (HANDLE) FREEHANDLE;

	//  Trace the action
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2082T, fNative);
	xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);

	//  Check for any entries that can be released
	ReleaseAllMappedFileHandles(iThreadID);
	return TRUE;
}
#endif
//  ReleaseAllMappedFileHandles
//
//  This function will close any locally (thread) mapped File Handles that can be closed because
//  the native handle has been invalidated.
//
//  Parameters
//
//  1  -  int						-	ID of the invoking thread
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadManager::ReleaseAllMappedFileHandles(int iThreadID)
{
	int					iIndexX;								//  Index of native DB Handles
	int					iIndexY;								//  Index of threads
	int					iReleasedEntries;						//  Count of released thread entries
	HANDLE				*pMappedFH;								//  Pointer to a mapped File entry
	HANDLE				*pMyMappedFH;							//  Pointer to a mapped File entry
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	//  Release any mapped file handles that have been completely freed
	pThreadPoolHeader = (Threadpool_Header *) ThreadPool;
	pMappedFH = (HANDLE *) pThreadPoolHeader->pFHMappingPool;

	for (iIndexX = 0; iIndexX < MAX_FH_CACHE; iIndexX++)
	{
		pMyMappedFH = pMappedFH + (iIndexX * (Policy->MaxThreads + 1));
		if (pMyMappedFH[0] == (HANDLE) FREEHANDLE)
		{
			iReleasedEntries = 0;
			for (iIndexY = 1; iIndexY <= Policy->MaxThreads; iIndexY++)
			{
				if (pMyMappedFH[iIndexY] == NULLHANDLE) iReleasedEntries++;
			}
			if (iReleasedEntries == Policy->MaxThreads) pMyMappedFH[0] = NULLHANDLE;
		}
	}
	return;
}

