#pragma once
//  ThreadManager
//
//  This class provides the root class for the control and management of the multi-threading support
//  environment for Domino eXplorer (DX) applications.
//
//  Message Identifier:      DXR
//  Message Range:           2001 - 2199
//  Trace area assignment:   TRACE_MT (02)
//  DX Version:				 3.14.0
//  Version:                 3.14.0  Build: 108
//
//  Version History:
//
//  1.0.0		07/06/2003		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.0		18/09/2009		New log handler for multithreaded applications
//  3.6.1		30/09/2009		Fixed stats & counter integrity for dispatcher
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.7.0		23/12/2009		Added race auto-correction to Scheduler
//  3.7.1		06/01/2010		Prevent a race condition in Rejoin processing
//  3.7.4		19/03/2010		Fixed various issues with log/repository configuration
//  3.9.0		24/07/2010		Linkage changed for logging & request posting and rejoining
//								A CommandHandler can be hooked into the monitor processing
//  3.9.1		02/09/2010		Changes to the ACL Processors
//								Fix for false RJR_NONE_EXIST signals
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.2		18/11/2010		MTIO support added to DbMover
//  3.10.3		13/12/2010		Db Property Modifier added
//  3.10.10		14/07/2011		Performance Improvements for/to DbCopier
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.12.0		21/12/2011		DXIP introduced
//	3.13.0		20/12/2013		PeekQueues function added
//	3.14.0		19/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2003 - 2015
//

//  Platform Includes
#include "../Platform/PlatBase.h"

//  Application Includes
#include "../ExecEnvironment.h"
#include "ThreadStructs.h"								//  Thread Structures
#include "ThreadManagerPolicy.h"
#include "ThreadMonitor.h"
#include "ThreadDispatcher.h"
#include "ThreadScheduler.h"

//  Local Definitions
#define			CTLINIT_INTERVAL		100					//  Wait 100 Ms for Control Threads to initialise
#define			CTLINIT_LIMIT			1000				//  1000 Wait intervals are the timeout limit
#define			CTLSTART_INTERVAL		100					//  Wait 100 Ms for Control Threads to start
#define			CTLSTART_LIMIT			1000				//  1000 Wait intervals are the timeout limit
#define			LOGWAIT_INTERVAL		50					//  50 Ms wait for log writes
#define			PXR_BUSY_WAIT			100					//  100 Ms wait if RQE Pool is full
#define			THREAD_STACK_SIZE		24*1024				//  Stack Size
#define			MAX_APP_MUTEXID			9					//  App Mutextes are 0 - 9
#ifdef UNIX
#define			TSTART_RETRY_LIMIT		10					//  10 attempts max to start a thread
#endif
#ifdef UNIX
#include <fcntl.h>											//  Include POSIX File I/O
#endif
#define			MAX_FILE_INFO			1024				//  Max size of file information

//  Flags for Posting an Execution Request
#define			PXR_WAITIF_BUSY		0x00000001				//  Wait if queue is full
#define			PXR_APP_WAIT		0x00000002				//  Use App thread wait

//  Trace Messages
#define			MSG_DXR2005T		"DXR2005T: The thread manager is starting initialisation."
#define			MSG_DXR2009T		"DXR2009T: The Monitor thread has been started."
#define			MSG_DXR2010T		"DXR2010T: The Monitor thread has been shut down."
#define			MSG_DXR2011T		"DXR2011T: The Thread Manager has completed shutdown."
#define			MSG_DXR2013T		"DXR2013T: The Scheduler thread has been initialised."
#define			MSG_DXR2014T		"DXR2014T: The Scheduler thread has been shut down."
#define			MSG_DXR2016T		"DXR2016T: The Dispatcher thread has initialised."
#define			MSG_DXR2017T		"DXR2017T: The Dispatcher thread has been shut down."
#define			MSG_DXR2021T		"DXR2021T: A new ThreadManager instance has been created @%p."
#define			MSG_DXR2022T		"DXR2022T: A ThreadManager instance @0x%p has been destroyed."
#define			MSG_DXR2028T		"DXR2028T: ThreadManager Started: TP@0x%p, TPM@0x%p, PRP@0x%p, RRP@0x%p, LP@0x%p, DBHP@0x%p."
#define			MSG_DXR2033T		"DXR2033T: The thread manager is draining the thread pool."
#define			MSG_DXR2042T		"DXR2042T: Rejoin RQE for parameter object @0x%p has been dequeued (O=@0x%p, X=@0x%p)."
#define			MSG_DXR2069T		"DXR2069T: Posted RQE: O=@0x%p, P=@0x%p, X=@0x%p for adding to the pending pool."
#define			MSG_DXR2070T		"DXR2070T: CMLS Activated with the following constraints. T0: %i, P0: %i, T1: %i, P1: %i, T2: %i, P2: %i."
#define			MSG_DXR2071T		"DXR2071T: Scheduler and Dispatcher threads are now ready to be started.."
#define			MSG_DXR2072T		"DXR2072T: Scheduler and Dispatcher threads are now running, all control threads are active."
#define			MSG_DXR2073T		"DXR2073T: Thread %i using local File handle 0x%p for native handle 0x%p."
#define			MSG_DXR2077T		"DXR2077T: Allocated local File handle %p for native handle 0x%p for thread %i."
#define			MSG_DXR2081T		"DXR2081T: Reopening file '%s' for thread: %i."
#define			MSG_DXR2082T		"DXR2082T: Freeing native File handle 0x%p."
#define			MSG_DXR2084T		"DXR2084T: Peeking request queues for owner: 0x%p shows %i requests found."

#ifdef W64
#define			MSG_DXR2029T		"DXR2029T: Thread %i using local DB handle 0x%i for native handle 0x%i."
#define			MSG_DXR2030T		"DXR2030T: Allocated local DB handle 0x%i for native handle 0x%i for thread %i."
#define			MSG_DXR2031T		"DXR2031T: Freeing native DB handle 0x%i."
#define			MSG_DXR2041T		"DXR2041T: Closed local handle 0x%i by thread %i."
#define			MSG_DXR2083T		"DXR2083T: Released native DB handle 0x%i by thread %i."
#else
#define			MSG_DXR2029T		"DXR2029T: Thread %i using local DB handle 0x%p for native handle 0x%p."
#define			MSG_DXR2030T		"DXR2030T: Allocated local DB handle 0x%p for native handle 0x%p for thread %i."
#define			MSG_DXR2031T		"DXR2031T: Freeing native DB handle 0x%p."
#define			MSG_DXR2041T		"DXR2041T: Closed local handle 0x%p by thread %i."
#define			MSG_DXR2083T		"DXR2083T: Released native DB handle 0x%p by thread %i."
#endif

//  Information Messages
#define			MSG_DXR2023I		"DXR2023I: There are %i log messages pending, waiting for them to be written."
#define			MSG_DXR2024I		"DXR2024I: There are %i worker threads currently active, they will now be closed."
#define			MSG_DXR2034I		"DXR2034I: Log messages queued %d, log messages written %d, current queue size %d."
#define			MSG_DXR2035I		"DXR2035I: Log messages dropped %d, discarded %d, overflowed %d."
#define			MSG_DXR2036I		"DXR2036I: Dispatches by brute force %d, by priority %d." 
#define			MSG_DXR2037I		"DXR2037I: Requests dispatched by brute force %d, by priority %d." 
#define			MSG_DXR2038I		"DXR2038I: Null dispatches because of no work %d, because all threads were busy %d." 
#define			MSG_DXR2039I		"DXR2039I: Priority dispatch windows closed because of no work %d, because all threads were busy %d."
#define			MSG_DXR2040I		"DXR2040I: RQE processed %d requests and made semaphore requests %d, waits %d, wait quanta %d." 
#define			MSG_DXR2043I		"DXR2043I: RQEs rejoined %d, with %d rejoin busy signals."
#define			MSG_DXR2044I		"DXR2044I: Guard exceptions: G1: %i, G2 %i, G3 %i, G4 %i, G5 %i." 
#define			MSG_DXR2045I		"DXR2045I: Guard exceptions: G6: %i, G7 %i, G8 %i, G9 %i, G10 %i." 
#define			MSG_DXR2048I		"DXR2048I: There are %i requests pending execution and %i pending rejoin, %i threads are available."
#define			MSG_DXR2049I		"DXR2049I: The thread manager was started at: %s."
#define			MSG_DXR2050I		"DXR2050I: Brute Force dispatches aborted because all threads were busy %d."
#define			MSG_DXR2051I		"DXR2051I: Log Semaphore requests %i, waits %i, wait quanta %i."
#define			MSG_DXR2052I		"DXR2052I: Pending Pool Full waits %i, backoffs %i, lane backoffs %i, CMLS denials %i."
#define			MSG_DXR2053I		"DXR2053I: Rejoin Pool Mutex requests %i, waits %i, fails %i."
#define			MSG_DXR2054I		"DXR2054I: DB Handle Mapping Table Mutex requests %i, waits %i, fails %i."
#define			MSG_DXR2055I		"DXR2055I: Application Mutex %i requests %i, waits %i, fails %i."
#define			MSG_DXR2061I		"DXR2061I: Thread %i: Log Semaphore requests %i, waits %i, wait quanta %i, current state = %i."
#define			MSG_DXR2062I		"DXR2062I: Thread %i: Processed %d requests and made RQE semaphore requests %d, waits %d, wait quanta %d, current state = %i." 
#define			MSG_DXR2063I		"DXR2063I: Thread %i: Processed %d requests and made rejoin requests %d, waits %d, wait quanta %d, current state = %i." 
#define			MSG_DXR2064I		"DXR2064I: RQE processed %d requests and made rejoin semaphore requests %d, waits %d, wait quanta %d." 
#define			MSG_DXR2065I		"DXR2065I: Thread %i: Processed %d requests and made rejoin poll requests %d, waits %d, wait quanta %d, current state = %i." 
#define			MSG_DXR2066I		"DXR2066I: RQE processed %d requests and rejoin poll semaphore requests %d, waits %d, wait quanta %d." 
#define			MSG_DXR2067I		"DXR2067I: Thread %i: made %d rejoin poll requests %d returned, %d not ready, %d none exist." 
#define			MSG_DXR2068I		"DXR2068I: Total %d rejoin poll requests %d returned, %d not ready, %d none exist." 
#define			MSG_DXR2080I		"DXR2080I: File Handle Mapping Table Mutex requests %i, waits %i, fails %i."
#define			MSG_DXR2081I		"DXR2081I: %d Requests for a rejoin recommendation resulted in %d positive recommendations."

//  Error Messages
#define			MSG_DXR2001E		"DXR2001E: This Thread Manager has already been started."
#define			MSG_DXR2002E		"DXR2002E: Unable to allocate storage for the Thread Pool."
#define			MSG_DXR2003E		"DXR2003E: Unable to allocate storage for the Pending Requests Pool."
#define			MSG_DXR2004E		"DXR2004E: Unable to allocate storage for the Rejoin Requests Pool."
#define			MSG_DXR2006E		"DXR2006E: Unable to allocate storage for the Log Pool."
#define			MSG_DXR2007E		"DXR2007E: The thread manager was unable to start the Monitor thread."
#define			MSG_DXR2012E		"DXR2012E: The thread manager was unable to start the Scheduler thread."
#define			MSG_DXR2015E		"DXR2015E: The thread manager was unable to start the Dispatcher thread."
#define			MSG_DXR2018E		"DXR2018E: The thread manager has timed out waiting for the control threads to initialise, after %i ticks."
#define			MSG_DXR2019E		"DXR2019E: The thread manager has timed out waiting for the control threads to start, after %i ticks."
#define			MSG_DXR2020E		"DXR2020E: The thread manager has timed out waiting for a worker thread to quit."
#define			MSG_DXR2025E		"DXR2025E: Unable to allocate storage for the Database Handle Mapping Pool."
#define			MSG_DXR2026E		"DXR2026E: There are no free entries in the Database Handle Mapping Pool."
#define			MSG_DXR2027E		"DXR2027E: Unable to open a thread local DB handle."
#define			MSG_DXR2032E		"DXR2032E: Unable to acquire the Database Handle Mapping Pool mutex."
#define			MSG_DXR2046E		"DXR2046E: Failed to acquire the Rejoin Pool Mutex, it is currently held by thread: %i, signalling BUSY."
#define			MSG_DXR2047E		"DXR2047E: Failed to re-establish synchronous logging."
#define			MSG_DXR2056E		"DXR2056E: Posting to RQE @%p is unsafe as the Pend Pool mutex is held by %i."
#define			MSG_DXR2057E		"DXR2057E: Posting to RQE @%p is unsafe as the RQE State is %i."
#define			MSG_DXR2058E		"DXR2058E: RQE @%p has been returned to the Pending Pool."
#define			MSG_DXR2059E		"DXR2059E: The Rejoin pool mutex has been pre-empted by Thread: %i."
#define			MSG_DXR2060E		"DXR2060E: RQE @%p is not safe for rejoin, due to RQE state change."
#define			MSG_DXR2074E		"DXR2074E: Unable to acquire the File Handle Mapping Pool mutex."
#define			MSG_DXR2075E		"DXR2075E: There are no free entries in the File Handle Mapping Pool."
#define			MSG_DXR2076E		"DXR2076E: Unable to open a thread local File handle."
#define			MSG_DXR2078E		"DXR2078E: Unable to get the filename to reopen a native file."
#define			MSG_DXR2079E		"DXR2079E: Failed to reopen file '%s' for thread local use."

class ThreadManager
{
public:
	//  Constructor
	ThreadManager(ExecEnvironment *xeParent);

	//  Destructor
	~ThreadManager(void);

	//  Public Members
	HANDLE					htMonitor;					//  Handle to the Monitor Thread
	UINT					uitMonitor;					//  Hnadle of the Monitor Thread Object
	HANDLE					htScheduler;				//  Handle to the Scheduler Thread
	UINT					uitScheduler;				//  Hnadle of the Scheduler Thread Object
	HANDLE					htDispatcher;				//  Handle to the Dispatcher Thread
	UINT					uitDispatcher;				//  Hnadle of the Dispatcher Thread Object
	UCHAR					*ThreadPool;				//  Pointer to the Thread Pool
	ThreadManagerPolicy		*Policy;					//  Thread Manager Policy Object

	//  Public Functions

	//  Request Support Functions
	BOOL CorePostARequest(int iFlags, void * pOwner, void * pExecutor, void * pParms, UINT Attrs, int iPriority, int iThreadID);		//  Post an Exec Request
	BOOL CorePostALogMessage(char * szLogMessage, int iThreadID);			//  Post a log message to the queue
	int CoreGetRejoinRequest(void *pOwner, void **pParms, int iThreadID);   //  Get request that is ready to rejoin
	BOOL CoreIsRejoinRecommended(void *pOwner, int iThreadID);				//  Test if a rejoin poll is indicated
	DWORD CorePeekQueues(void *pOwner, int iThreadID);						//  Peek request queues for owner

	//  Thread Local DBHANDLE Support Functions
	DBHANDLE CoreGetMappedDBH(DBHANDLE hdbNative, int iThreadID);		//  Get a locally mapped Database Handle
	BOOL CoreInvalidateNativeDBH(DBHANDLE hdbNative, int iThreadID);	//  Invalidate a Native Database Handle
	void CoreReleaseMappedDBH(int iThreadID);							//  Release thread DBH mapping entries
	void ReleaseAllMappedDBH(int iThreadID);							//  Release any freed DBH mapping entries

	//  Thread Local File Handle Support Functions
#ifdef WIN32
	HANDLE CoreGetMappedFileHandle(HANDLE hfNative, char *szFile, DWORD dwAccess, DWORD dwSharing, int iThreadID);			//  Get a locally mapped File Handle
	BOOL CoreInvalidateNativeFileHandle(HANDLE hfNative, int iThreadID);	//  Invalidate a Native File Handle
#endif
#ifdef UNIX
	int CoreGetMappedFileHandle(int fNative, char *szFile, int iFlags, int iThreadID);	//  Get a locally mapped File Handle
	BOOL CoreInvalidateNativeFileHandle(int fNative, int iThreadID);		//  Invalidate a Native File Handle
#endif
	void CoreReleaseMappedFileHandle(int iThreadID);						//  Release thread DBH mapping entries
	void ReleaseAllMappedFileHandles(int iThreadID);						//  Release any freed DBH mapping entries

	//  Lightweight Mutex Functions
	BOOL CoreAcquireAppMutex(UCHAR *pTP, int iMutexID, int iThreadID);	//  Acquire the Heap Synch Mutex
	BOOL CoreFreeAppMutex(UCHAR *pTP, int iMutexID, int iThreadID);		//  Free the Heap Synch Mutex

	//  Constrained Multi-Lane Scheduling (CMLS) Functions
	int CoreGetThreadCount(void);								//  Get the current thread count
	int CoreGetCMLSRecommendation(int iLevel, int iThreadID);	//  Get the CMLS Lane Restrictions

	//  Miscellaneous Functions
	BOOL Start(void);											//  Start the Thread Manager
	BOOL Drain(void);											//  Drain the Thread Manager
	BOOL CoreSetFastDrainMode(void);							//  Set fast drain mode
	void CoreShowThreadStats(int iOpts, int iThreadID);			//  Show the thread statistics
	void CoreLogFlush();										//  Force a flush of the log data
	void CoreDump(int iDumpScope, int iThreadID);				//  Perform a diagnostic dump
	STATUS CoreGetLastOpenError(int iThreadID);					//  Get last D/B Open Error
	void CoreSetLastOpenError(STATUS stOpen, int iThreadID);	//  Set last D/B open Error code
	void CoreAttachCommandHandler(CommandHandler *chApp);		//  Attach a CommandHandler

private:
	//  Private Members
	BOOL ThreadManagerStarted;							//  Lock on the Start Function
	ExecEnvironment			*xeLocal;					//  Current Runtime
	ThreadMonitor			*tmonMaster;				//  Master Monitor
	ThreadScheduler			*tsMaster;					//  Master Scheduler
	ThreadDispatcher		*tdMaster;					//  Master Dispatcher
#ifdef UNIX
	pthread_attr_t				paThread;								//  Thread Attributes
#endif

	//  Private Functions
	void SetDefaultPolicy(void);						//  Set default policies
	void PerformCMLSInits(void);						//  Perform CMLS initialisation
	BOOL WaitForControlInit(void);						//  Wait for control threads to initialise
	BOOL WaitForControlStart(void);						//  Wait for control threads to start
	void InitMembers(UCHAR *pTP, int iTCount);			//  Initialise thread members
	void WriteLogSemStats(UCHAR *pTP, int iTCount, int iOpts, int iThreadID);		//  Write the log semaphore stats
	void WriteDispSemStats(UCHAR *pTP, int iTCount, int iOpts, int iThreadID);	//  Write the request dispatch semaphore stats
	void WriteRejoinSemStats(UCHAR *pTP, int iTCount, int iOpts, int iThreadID);	//  Write the Rejoin semaphore stats
	void WriteRejoinPollSemStats(UCHAR *pTP, int iTCount, int iOpts, int iThreadID);	//  Write the Rejoin poll semaphore stats
	BOOL AcquireDBHTMutex(UCHAR *pTP, int iThreadID);	//  Acquire the DB Handle Table mutex
	BOOL AcquireFHTMutex(UCHAR *pTP, int iThreadID);	//  Acquire the File Handle Table mutex
	UCHAR *GetMemberAddress(int iThreadID);				//  Get the address of the thread pool member
#ifdef WIN32
	HANDLE ReopenFile(HANDLE hfNative, char *szFile, DWORD dwAccess, DWORD dwSharing, int iThreadID);	//  Reopen a file
#endif
#ifdef UNIX
	int ReopenFile(int fNative, char *szFile, int iFlags, int iThreadID);			//  Reopen a file
#endif

};
