//  ThreadMonitor
//
//  This class provides monitor functions for the Thread Pool Manager.
//
//  Message Identifier:      DXR
//  Message Range:           2200 - 2399
//  Trace area assignment:   TRACE_MT (02)
//  DX Version:				 3.14.0
//  Version:                 3.14.0  Build: 108
//
//  Version History:
//
//  1.0.0		07/06/2003		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.0		18/09/2009		New log handler for multithreaded applications
//  3.6.1		10/10/2009		Additional controls added
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.7.2		12/01/2010		DbCopier improvements for corrupt folder handling
//								Remote Fixup on DbOpen
//								Logger improvements
//  3.7.4		19/03/2010		Fixed various issues with log/repository configuration
//  3.9.0		24/07/2010		Linkage changed for logging & request posting
//								A CommandHandler can be hooked into the monitor processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.12.0		21/12/2011		DXIP introduced
//	3.14.0		19/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2003 - 2015
//

//  Application Includes
#include "ThreadMonitor.h"

//  Constructor
ThreadMonitor::ThreadMonitor(ExecEnvironment *xeParent, UCHAR *pTP, ThreadManagerPolicy *tmpParent)
: tmpLocal(NULL)
, chApplication(NULL)
{
	char			szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Save the Runtime  & Policy address
	xeLocal = xeParent;
	tmpLocal = tmpParent;
	chApplication = NULL;
	//  Save the Thread Pool address
	pThreadPool = pTP;

	//  Trace the construction 
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2210T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg);
}

//  Destructor
ThreadMonitor::~ThreadMonitor(void)
{
	char			szMsg[MAX_MSG];								//  Generic message buffer

	//  Trace the destruction 
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2211T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg);
}

//  Start
//
//  This function will start the Thread Monitor function of the Thread Pool Manager.
//
//  Parameters
//
//  Returns
//
//  UINT  -			Status indication  
//
//  API Use Notes:
//
//
UINT ThreadMonitor::Start(void)
{
	STATUS						stAPIRC;								//  Return code from Notes API calls
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	int							iThreadDelta;							//  Delta of threads needed
	int							iIndex;									//  Index counter
	int							iStrategyReviewTrigger = 0;				//  Strategy review trigger
	int							iSchedulerPrefsReviewTrigger = 0;		//  Schedule pref review trigger
	int							iAvailableThreadTrigger = 0;			//  Thread availability count trigger
#ifdef DXIP
#ifdef WIN32
	HANDLE						hfRec;									//  Handle of the instrumentation package recording file
	DWORD						dwWinError;								//  Error Code
#endif
#ifdef UNIX
	int							hfRec;									//  Descriptor of the instrumentation package recording file
	ssize_t						stRead;									//  Size read/written
#endif
	UCHAR						*pRecBfr;								//  Recording Buffer
	DWORD						dwWritten;								//  Number of bytes written
	DWORD						dwBytesRecorded;						//  Numbrer of bytes recorded
#endif
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header

	//  Initialise the Notes runtime for the thread
	stAPIRC = NotesInitThread();
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR2212E, MONITOR_THREAD_ID);			//  Show the error
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, MONITOR_THREAD_ID);					//  Show the API error message
		pThreadPoolHeader->TMonState = TXSTATE_DRAINED;
#ifdef WIN32
		_endthreadex(THREAD_RC_ERROR);									//  Terminate the thread
#endif

#ifdef UNIX
		pthread_exit(NULL);
#endif
		return THREAD_RC_ERROR;
	}

	//  Determine if worker threads need to be started
	iThreadDelta = ComputeThreadDelta(pThreadPoolHeader);				//  Compute the delta
	if (iThreadDelta > 0)
	{
		//  Prestart Worker Threads
		for (iIndex = 0; iIndex < iThreadDelta; iIndex++)
		{
			if (!StartWorker(pThreadPoolHeader))
			{
				//  Failed to start a worker thread - return in the initialising state
				xeLocal->LogMessage(MSG_DXR2204E, MONITOR_THREAD_ID);
#ifdef WIN32
				_endthreadex(THREAD_RC_ERROR);									//  Terminate the thread
#endif

#ifdef UNIX
				pthread_exit(NULL);
#endif
				return THREAD_RC_ERROR;
			}
		}
	}

	//  Set the default priority controls
	pThreadPoolHeader->PrioControl = tmpLocal->PriorityPolicy;

	//  Prepare the asynch logger by reopening the Log Document in the current thread
	//  Firstly re-open the Repository in the current thread
	xeLocal->hdbLog = NULLHANDLE;
	if (xeLocal->hdbRepository != NULLHANDLE)
	{
		stAPIRC = NSFDbReopen(xeLocal->hdbRepository, &xeLocal->hdbLog);
		if (ERR(stAPIRC) != NOERROR)
		{	
			std::cout << MSG_DXR2221E << std::endl;					//  Log directly as log is not available
			xeLocal->LoggingState = LOGSTATE_CLOSED;				//  Mark the log as not available
#ifdef WIN32
			_endthreadex(THREAD_RC_ERROR);									//  Terminate the thread
#endif

#ifdef UNIX
			pthread_exit(NULL);
#endif
			return THREAD_RC_ERROR;
		}
	}

	//  Re-open the current log document 
	if (xeLocal->hdbLog != NULLHANDLE)
	{
		stAPIRC = NSFNoteOpen(xeLocal->hdbLog, xeLocal->nidLogDoc, 0, &xeLocal->hnLogDoc);
		if (ERR(stAPIRC) != NOERROR)
		{
			std::cout << MSG_DXR2222E << std::endl;					//  Log directly as log is not available
			xeLocal->LoggingState = LOGSTATE_CLOSED;				//  Mark the log as not available
			NSFDbClose(xeLocal->hdbLog);							//  Close the local log handle
#ifdef WIN32
			_endthreadex(THREAD_RC_ERROR);							//  Terminate the thread
#endif

#ifdef UNIX
			pthread_exit(NULL);
#endif
			return THREAD_RC_ERROR;
		}
	}

	//  Mark the logger as available
	xeLocal->LoggingState = LOGSTATE_ASYNCH;
	LogCommitCycles = 0;

#ifdef UNIX
	//  Initialise the Thread Attributes
	pthread_attr_init(&paThread);
	pthread_attr_setdetachstate(&paThread, PTHREAD_CREATE_DETACHED);      //  Create detached so we don't need to join
#endif

#ifdef DXIP
	//  Build the file name and open the Instrumentation Package Recording file
	pRecBfr = NULL;
	hfRec = StartRecording(MONITOR_THREAD_ID);
#ifdef WIN32
	if (hfRec != NULLHANDLE)
#endif
#ifdef UNIX
	if (hfRec != -1)
#endif
	{
		pRecBfr = (UCHAR *) malloc(IPREC_BFR_SIZE);
		if (pRecBfr == NULL)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2231E, IPREC_BFR_SIZE);
			xeLocal->LogMessage(szMsg, MONITOR_THREAD_ID);
			dwBytesRecorded = (DWORD) strlen(szMsg);
#ifdef WIN32
			WriteFile(hfRec, szMsg, dwBytesRecorded, &dwWritten, NULL);
			CloseHandle(hfRec);
			hfRec = NULLHANDLE;
#endif
#ifdef UNIX
			write(hfRec, szMsg, dwBytesRecorded);
			close(hfRec);
			hfRec = -1;
#endif
		}
		else
		{
			dwBytesRecorded = 0;
			memset(pRecBfr, 0, IPREC_BFR_SIZE);
		}
	}
#endif

	//  The initial state merely idles - servicing only the Log Pool
	pThreadPoolHeader->TMonState = TXSTATE_NOTSTARTED;
	//  Trace the state change 
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2200T, MONITOR_THREAD_ID);				//  Show the state change

	while (pThreadPoolHeader->TMonState == TXSTATE_NOTSTARTED)
	{
		//  Process any posted log messages
		MonitorLogSemaphores();
		MonitorLogPool();
		SLEEP(IDLE_WAITINTERVAL);
	}

	//  Check that the transition requested is to the started state
	if (pThreadPoolHeader->TMonState == TXSTATE_START)
	{
		pThreadPoolHeader->TMonState = TXSTATE_RUNNING;
		xeLocal->LogTrace(TRACE_MT, MSG_DXR2201T, MONITOR_THREAD_ID);				//  Show the state change
	}

	//  Running state loop - monitor the log pool and the worker threads, also compute the dispatcher tactics
	while (pThreadPoolHeader->TMonState == TXSTATE_RUNNING)
	{
		//  Process any posted log messages
		MonitorLogSemaphores();

		//  Process the asynch log pool
		MonitorLogPool();

#ifdef DXIP
		//  Monitor any recording instruments
		MonitorInstruments(hfRec, pRecBfr, &dwBytesRecorded, MONITOR_THREAD_ID);
#endif

		//  If there is a Command Handler attached then invoke the handler to process the message queue (MQ)
		if (chApplication != NULL)
		{
			chApplication->HandleCommands(MONITOR_THREAD_ID);
		}
		iStrategyReviewTrigger++;
		iSchedulerPrefsReviewTrigger++;
		iAvailableThreadTrigger++;

		if (iAvailableThreadTrigger == tmpLocal->MonitorAvailableInterval)
		{
			iAvailableThreadTrigger = 0;							//  Reset the trigger counter
		}

		if (iSchedulerPrefsReviewTrigger == tmpLocal->MonitorPrefReviewCycle)
		{
			//  Re-evaluate Scheduler preferences
			PerformSchedulerPrefsReview((UCHAR *) pThreadPoolHeader);
			iSchedulerPrefsReviewTrigger = 0;						//  Reset the trigger counter
		}

		if (iStrategyReviewTrigger == tmpLocal->MonitorStratReviewCycle)
		{
			//  If in any trace mode then snap the stats
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2213T, pThreadPoolHeader->LogMsgQueued, pThreadPoolHeader->LogMsgWritten);
			xeLocal->LogTrace(TRACE_MT, szMsg, MONITOR_THREAD_ID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2214T, pThreadPoolHeader->LogMsgDropped, pThreadPoolHeader->LogMsgDiscarded, pThreadPoolHeader->LogMsgOverflowed);
			xeLocal->LogTrace(TRACE_MT, szMsg, MONITOR_THREAD_ID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2215T, pThreadPoolHeader->BFDispatches, pThreadPoolHeader->PDispatches);
			xeLocal->LogTrace(TRACE_MT, szMsg, MONITOR_THREAD_ID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2216T, pThreadPoolHeader->DispBruteForce, pThreadPoolHeader->DispPriority);
			xeLocal->LogTrace(TRACE_MT, szMsg, MONITOR_THREAD_ID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2217T, pThreadPoolHeader->NoDispNoWork, pThreadPoolHeader->NoDispThreadsBusy);
			xeLocal->LogTrace(TRACE_MT, szMsg, MONITOR_THREAD_ID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2218T, pThreadPoolHeader->PDispQEmpty, pThreadPoolHeader->PDispThreadsBusy);
			xeLocal->LogTrace(TRACE_MT, szMsg, MONITOR_THREAD_ID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2220T, pThreadPoolHeader->RJRDequeued, pThreadPoolHeader->RJRBusy);
			xeLocal->LogTrace(TRACE_MT, szMsg, MONITOR_THREAD_ID);

			//  CMLS Traces
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2224T, pThreadPoolHeader->PendRQECount, pThreadPoolHeader->RejoinRQECount, pThreadPoolHeader->AvailableThreads);
			xeLocal->LogTrace(TRACE_CMLS, szMsg, MONITOR_THREAD_ID);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2225T, pThreadPoolHeader->PPFullWaits, pThreadPoolHeader->PPFullBackoffs, pThreadPoolHeader->PPLaneFullBackoffs, pThreadPoolHeader->CMLSDenials);
			xeLocal->LogTrace(TRACE_CMLS, szMsg, MONITOR_THREAD_ID);

			//  Re-evaluate the strategy
			PerformStrategyReview((UCHAR *) pThreadPoolHeader);
			iThreadDelta = ComputeThreadDelta(pThreadPoolHeader);   //  Compute thread changes
			while (iThreadDelta != 0)
			{
				//  Change in thread count is needed
				if (iThreadDelta > 0)
				{
					//  Start an additional worker thread
					if (!StartWorker(pThreadPoolHeader))
					{
						//  Failed to start a worker thread - return in the initialising state
						xeLocal->LogMessage(MSG_DXR2204E, MONITOR_THREAD_ID);
					}
					iThreadDelta--;
				}
				else
				{
					//  Stop a worker thread

				}
			}
			iStrategyReviewTrigger = 0;								//  Reset the trigger counter
		}

		CloseMyMappedDBH(MONITOR_THREAD_ID);
		SLEEP(tmpLocal->MonitorWaitCycle);
	}

	//  Close any available local resources
	ForceCloseMyMappedDBH(MONITOR_THREAD_ID);

#ifdef DXIP
#ifdef WIN32
	//  If DXIP is running
	if (hfRec != NULLHANDLE)
	{
		//  Check if there is any recording in the buffer - if so then write it
		if (dwBytesRecorded > 0)
		{
			if (!WriteFile(hfRec, pRecBfr, dwBytesRecorded, &dwWritten, NULL))
			{
				dwWinError = GetLastError();
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2232E, dwBytesRecorded);
				xeLocal->LogMessage(szMsg, MONITOR_THREAD_ID);
				xeLocal->GetWINAPIError(dwWinError, szMsg);
				xeLocal->LogMessage(szMsg, MONITOR_THREAD_ID);
			}
			else
			{
				if (dwBytesRecorded != dwWritten)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR2233E, dwBytesRecorded, dwWritten);
					xeLocal->LogMessage(szMsg, MONITOR_THREAD_ID);
				}
			}
		}
		CloseHandle(hfRec);
		hfRec = NULLHANDLE;
		free(pRecBfr);
		pRecBfr = NULL;
	}
#endif
#ifdef UNIX
	//  If DXIP is running
	if (hfRec != -1)
	{
		//  Check if there is any recording in the buffer - if so then write it
		if (dwBytesRecorded > 0)
		{
			stRead = write(hfRec, pRecBfr, dwBytesRecorded);
			if (stRead == -1)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2232E, dwBytesRecorded);
				xeLocal->LogMessage(szMsg, MONITOR_THREAD_ID);
				xeLocal->GetUNIXAPIError(errno, szMsg);
				xeLocal->LogMessage(szMsg, MONITOR_THREAD_ID);
			}
			else
			{
				dwWritten = stRead;
				if (dwBytesRecorded != dwWritten)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR2233E, dwBytesRecorded, dwWritten);
					xeLocal->LogMessage(szMsg, MONITOR_THREAD_ID);
				}
			}
		}
		close(hfRec);
		hfRec = -1;
		free(pRecBfr);
		pRecBfr = NULL;
	}
#endif
#endif

	//  Check that the transition requested is to the quit state
	if (pThreadPoolHeader->TMonState == TXSTATE_QUIT)
	{
		pThreadPoolHeader->TMonState = TXSTATE_QUITTING;
		xeLocal->LogTrace(TRACE_MT, MSG_DXR2202T, MONITOR_THREAD_ID);				//  Show the state change
	}
	else
	{
		pThreadPoolHeader->TMonState = TXSTATE_QUITTING;
		xeLocal->LogMessage(MSG_DXR2203E, MONITOR_THREAD_ID);				//  Show inconsistent state change
	}

	//  Switch Logging back to the synchronous logger
	//  First drain the log pool

#ifdef UNIX
	//  Initialise the Thread Attributes
	pthread_attr_destroy(&paThread);
#endif
	MonitorLogPool();
	if (xeLocal->hnLogDoc != NULLHANDLE)
	{
		NSFNoteUpdate(xeLocal->hnLogDoc, UPDATE_FORCE);
		NSFNoteClose(xeLocal->hnLogDoc);
		xeLocal->hnLogDoc = NULLHANDLE;
	}
	xeLocal->LoggingState = LOGSTATE_SWITCH_SYNCH;
	if (xeLocal->hdbLog != NULLHANDLE)
	{
		NSFDbClose(xeLocal->hdbLog);
		xeLocal->hdbLog = NULLHANDLE;
	}
	NotesTermThread();
	
	//  Mark the thread as drained and terminate the thread
	pThreadPoolHeader->TMonState = TXSTATE_DRAINED;
#ifdef WIN32
	_endthreadex(THREAD_RC_OK);										//  Terminate the thread
#endif

#ifdef UNIX
	pthread_exit(NULL);
#endif
	return THREAD_RC_OK;
}

//  StartWrapper
//
//  This function is the static wrapper for the Start() function allowing it to be started as an independent thread.
//
//  Parameters
//
//  Returns
//
//  UINT  -			Status indication  
//
//  API Use Notes:
//
//
#ifdef WIN32
UINT LNPUBLIC ThreadMonitor::StartWrapper(void * pObject)
#endif
#ifdef UNIX
void * LNPUBLIC ThreadMonitor::StartWrapper(void * pObject)
#endif
{
	ThreadMonitor		*tmonThis;

	//  Wrapper function calls the Start function in the Object
	tmonThis = (ThreadMonitor *) pObject;
#ifdef WIN32
	return tmonThis->Start();
#endif
#ifdef UNIX
	return (void *) tmonThis->Start();
#endif
}

//  MonitorLogPool
//
//  This function will monitor the asynchronous log pool to see if there are messages that have been posted
//  into the pool. If messages are posted than they will be dequeued in turn and written to the log.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadMonitor::MonitorLogPool(void)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Logpool_Header				*pLogPoolHeader;						//  Pointer to the Log Pool Header
	int							iDQCount;								//  Messages to write
	int							iIndex;									//  Generic index

	//  If the monitor is not yet in charge of the logger then just leave
	if (xeLocal->LoggingState != LOGSTATE_ASYNCH) return;

	//  Check if there is anything posted in the Log Pool - if so action it
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header
	pLogPoolHeader = (Logpool_Header *) pThreadPoolHeader->pLogPool;	//  Address the Log Pool header
	if (pLogPoolHeader->LogPoolEntries == 0) 
	{
		LogCommitCycles++;							//  Increment the commit cycle count
		if (LogCommitCycles >= LOG_UNCOMMIT_LIMIT)
		{
			xeLocal->LogCommit();					//  Commit any dirty log docs
			LogCommitCycles = 0;;
		}
		return;										//  Nothing to do - exit
	}

	LogCommitCycles = 0;												//  Clear the commit counter
	
	//  Default to writing a single log message
	iDQCount = 1;

	// If the thread is quitting or flush is signalled then write all messages
	if ((pThreadPoolHeader->TMonState == TXSTATE_QUITTING) || (pLogPoolHeader->LogQueueControl & LOG_CTL_FLUSH))
	{
		iDQCount = pLogPoolHeader->LogPoolEntries;
	}
	else
	{
		// If the log pool is more that 50% filled then write 3 messages
		if (pLogPoolHeader->LogPoolEntries >= (pThreadPoolHeader->LogPoolSize / 2)) iDQCount = 3;
	}

	//  Dequeue (write) the log messages
	for (iIndex = 0; iIndex < iDQCount; iIndex++)
	{
		WriteLogEntry();
	}

	//  If needed flush the log document
	if (pLogPoolHeader->LogQueueControl & LOG_CTL_FLUSH)
	{
		if (xeLocal->hnLogDoc != NULLHANDLE) NSFNoteUpdate(xeLocal->hnLogDoc, UPDATE_FORCE);				//  Update the note
		pLogPoolHeader->LogQueueControl = pLogPoolHeader->LogQueueControl ^ LOG_CTL_FLUSH;      //  Turn off the signal
	}
	return;
}

//  WriteLogEntry
//
//  This function will write a single message from the asynch log pool to the log document.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadMonitor::WriteLogEntry(void)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Logpool_Header				*pLogPoolHeader;						//  Pointer to the Log Pool Header
	LogEntry					*pLogEntry;								//  Pointer to a log entry
	int							iThreadID;								//  ID of the posting Thread
	TIMEDATE					tdEvent;								//  Event Timestamp
	char						szMsg[MAX_MSG + 1];						//  Message Text

	//  Check if there is anything posted in the Log Pool - if so action it
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header
	pLogPoolHeader = (Logpool_Header *) pThreadPoolHeader->pLogPool;	//  Address the Log Pool header
	if (pLogPoolHeader->LogPoolEntries == 0) return;					//  Nothing to do - exit

	//  Extract the oldest entry in the Log Pool
	pLogEntry = (LogEntry *) (pThreadPoolHeader->pLogPool + sizeof(Logpool_Header) + (pLogPoolHeader->LogQueueHead * sizeof(LogEntry)));
	memcpy(&tdEvent, &pLogEntry->tdEventTime, sizeof(TIMEDATE));
	iThreadID = pLogEntry->LogThread;
	strcpy_s(szMsg, MAX_MSG, pLogEntry->szLogMsg);

	//  Dequeue the entry 
	if (pLogPoolHeader->LogQueueHead == (pThreadPoolHeader->LogPoolSize - 1)) pLogPoolHeader->LogQueueHead = 0;
	else pLogPoolHeader->LogQueueHead++;

	pLogPoolHeader->LogPoolEntries--;

	//  Post the log entry to the Log Document
	xeLocal->LogEvent(szMsg, &tdEvent, iThreadID);
	pThreadPoolHeader->LogMsgWritten++;
	return;
}

//  ComputeThreadDelta
//
//  This function will dynamically compute the difference between the current number of threads running and
//  the estimated number of threads needed - based on the current thread policy constrainys.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//	int				-  The current thread delta
//
//  API Use Notes:
//  
//
int ThreadMonitor::ComputeThreadDelta(Threadpool_Header * pTPH)
{
	//  Compute the number of worker threads that need to be added to or removed from the pool

	//  If we are in the initialisation phase then see if we have a pre-start target
	if (pTPH->TMonState == TXSTATE_INIT)
	{
		if (tmpLocal->TPoolPolicy & TPOOL_POLICY_PRESTARTMIN) return tmpLocal->MinThreads;
		if (tmpLocal->TPoolPolicy & TPOOL_POLICY_PRESTARTTARGET) return tmpLocal->TargetThreads;
		return 0;									//  No threads are pre-started
	}

	return 0;
}

//  StartWorker
//
//  This function will start an additional worker thread.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//	BOOL				-  TRUE if the additional worker thread was started
//						-  FALSE if the worker thread could not be started
//
//  API Use Notes:
//  
//
BOOL ThreadMonitor::StartWorker(Threadpool_Header * pTPH)
{
	WorkerThread				*wtCurrent;							//  Worker Thread Object
	Threadpool_Member			*pMember;							//  Pointer to a Thread Pool entry
	UCHAR						*pTP;								//  Pointer to Thread Pool
#ifdef UNIX
	int							iThreadRC;								//  Return code from thread functions
	int							iThreadRetries;							//  Thread start retry count
#endif
	int							iAllocThreadID;						//  Allocated Thread ID
#ifdef UNIX
	pthread_t					ptThread;							//  Thread Handle
#endif

	//  Start a new worker thread
	iAllocThreadID = GetThreadID((UCHAR *) pTPH);
	if (iAllocThreadID == -1)
	{
		xeLocal->LogMessage(MSG_DXR2205E, MONITOR_THREAD_ID);
		return FALSE;
	}

	//  Create the new Thread Object 
	wtCurrent = new WorkerThread(xeLocal, (UCHAR *) pTPH, tmpLocal, iAllocThreadID);

	//  Set the status and object address in the entry
	pTP = (UCHAR *) pTPH;
	pMember = (Threadpool_Member *) (pTP + sizeof(Threadpool_Header) + ((iAllocThreadID - 1) * sizeof(Threadpool_Member)));
	
	pMember->pTMemObject = wtCurrent;
	pMember->TMemState = WTSTATE_INIT;

	//  Start the worker thread
#ifdef WIN32
	pMember->hThread = (HANDLE) _beginthreadex(NULL, THREAD_STACK_SIZE, wtCurrent->StartWrapper, wtCurrent, 0, &pMember->uiPThreadID);
#endif

#ifdef UNIX
	iThreadRetries = 0;										//  Clear the retry counter
	pMember->hThread = NULL;
	//  Start the thread
	iThreadRC = pthread_create(&ptThread, &paThread, wtCurrent->StartWrapper, wtCurrent);

	//  If EAGAIN is returned then the fail condition is transient (probably short of resources)
	//  so retry the thread start.
	while ((iThreadRC == EAGAIN) && (iThreadRetries < TSTART_RETRY_LIMIT))
	{
		iThreadRetries++;
		iThreadRC = pthread_create(&ptThread, &paThread, wtCurrent->StartWrapper, wtCurrent);
	}

	//  If the thread was started then indicate the fact
	if (iThreadRC == EOK)
	{
		pMember->hThread = (HANDLE) ptThread;
	}
#endif

	if (pMember->hThread == NULL)
	{
		xeLocal->LogMessage(MSG_DXR2206E, MONITOR_THREAD_ID);
		return FALSE;
	}

	//  Wait for the thread to reach the not started state
	if (!WaitForWorkerState(pMember, WTSTATE_NOTSTARTED, MONITOR_THREAD_ID))
	{
		xeLocal->LogMessage(MSG_DXR2207E, MONITOR_THREAD_ID);
		return FALSE;
	}

	//  Start the worker
	pMember->TMemState = WTSTATE_START;

	//  Wait for the thread to get started
	if (!WaitForWorkerState(pMember, WTSTATE_RUNNING, MONITOR_THREAD_ID))
	{
		xeLocal->LogMessage(MSG_DXR2208E, MONITOR_THREAD_ID);
		return FALSE;
	}

	//  Increase the count of current running threads
	pTPH->CurrThreads++;
	return TRUE;
}

//  GetThreadID
//
//  This function will return the next Thread ID that is available in the Thread Pool
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//	int					-  The Thread ID
//						-  -1 There are no free slots in the Thread Pool
//
//  API Use Notes:
//  
//
int ThreadMonitor::GetThreadID(UCHAR * pTP)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pMember;								//  Pointer to a Thread Pool entry
	int							iIndex;									//  Index into the thread pool

	//  Return the first free ThreadID in the Thread Pool
	pThreadPoolHeader = (Threadpool_Header *) pTP;
	pMember = (Threadpool_Member *) (pTP + sizeof(Threadpool_Header));
	for (iIndex = 0; iIndex < pThreadPoolHeader->MaxThreads; iIndex++)
	{
		if (pMember->TMemState == WTSTATE_UNALLOCATED) return iIndex+1;
		pMember++;
	}
	return -1;
}

//  WaitForWorkerState
//
//  This function will wait for a worker thread to achieve the desired state, there are timeout limits for
//  the state change.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  UINT						-  The state being waited for
//  3  -  int						-  ID of the invoking thread
//
//  Returns
//
//	BOOL				-  TRUE if the worker thread has reached the desired state
//						-  FALSE if there was a timeout before the state was achieved
//
//  API Use Notes:
//  
//
BOOL ThreadMonitor::WaitForWorkerState(Threadpool_Member * pTPMember, UINT State, int iThreadID)
{
	int							iWaitCount = 0;							//  Count of wait intervals

	//  Wait for the control tasks to initialise
	while ((pTPMember->TMemState != State) && (iWaitCount < WTSTATE_LIMIT))
	{
		iWaitCount++;

		if (iThreadID == MONITOR_THREAD_ID)
		{
			//  Process any posted log messages (MONITOR Thread Only)
			MonitorLogSemaphores();
			//  Process the asynch log pool
			MonitorLogPool();
		}

		SLEEP(WTSTATE_INTERVAL);
	}
	if (iWaitCount == WTSTATE_LIMIT) return FALSE;
	return TRUE;
}

//  PerformSchedulerPrefsReview
//
//  This function will perform a periodic review of the Scheduler Preferences settings.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadMonitor::PerformSchedulerPrefsReview(UCHAR * pTP)
{
	//  Perform a review and update of the current scheduler preferences

	return;
}

//  PerformStrategyReview
//
//  This function will perform a periodic review of the Thread Pool Manager strategy settings
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadMonitor::PerformStrategyReview(UCHAR * pTP)
{
	//  Perform a review of strategy settings
	return;
}

//  DrainAnyWorker
//
//  This function will shut down a single (anonymous) Worker Thread
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  int						-  ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadMonitor::DrainAnyWorker(UCHAR * pTP, int iThreadID)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pMember;								//  Pointer tothread pool member
	int							iIndex;									//  Index into the thread pool

	//  Shutdown any worker thread
	pThreadPoolHeader = (Threadpool_Header *) pTP;						//  Address the header
	pMember = (Threadpool_Member *) (pTP + sizeof(Threadpool_Header));
	for (iIndex = 0; iIndex < pThreadPoolHeader->MaxThreads; iIndex++)
	{
		if (pMember->TMemState == WTSTATE_RUNNING)
		{
			pMember->TMemState = WTSTATE_QUIT;
			if (!WaitForWorkerState(pMember, WTSTATE_DRAINED, iThreadID))
			{
				xeLocal->LogMessage(MSG_DXR2209E, iThreadID);
			}
			pThreadPoolHeader->CurrThreads--;
			//  Delete the Worker Thread Object
			delete (WorkerThread *) pMember->pTMemObject;
			return;
		}
		pMember++;
	}
	return;
}

//  MonitorLogSempahores
//
//  This function will monitor the per thread log message semaphores, if posted, the message will be
//  posted to the asynchronous log pool and the semaphore will be cancelled.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadMonitor::MonitorLogSemaphores(void)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a threadpool member
	int							iIndex;									//  Generic Index

	//  Check if there is anything posted in the Log Pool - if so action it
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header

	for (iIndex = 0; iIndex < (pThreadPoolHeader->MaxThreads + 4); iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) (pThreadPool + sizeof(Threadpool_Header) + (iIndex * sizeof(Threadpool_Member)));
		if (pThreadPoolMember->semLogMsg != LOG_SEM_FREE)
		{
			ProcessLogSemaphore(pThreadPoolMember);
		}
	}
	return;
}

//  ProcessLogSempahore
//
//  This function will move a posted log message to the asynchronous log pool and free the semaphore.
//
//  Parameters
//
//  1  -  Pointer to the Thread Pool Member that has a posted semaphore
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadMonitor::ProcessLogSemaphore(Threadpool_Member *pTPM)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Logpool_Header			*pLogPoolHeader;						//  Pointer to the Log Pool Header
	LogEntry				*pLogEntry;								//  Pointer to the Log Entry

	//  Post a message to the Logger queue
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;			//  Address the header

	pLogPoolHeader = (Logpool_Header *) pThreadPoolHeader->pLogPool;

	//  Determine if there is no space in the pool
	if (pLogPoolHeader->LogPoolEntries == pThreadPoolHeader->LogPoolSize)
	{
		//  If the request is on behalf ot the monitor thread then clear the semaphore (i.e. lose the message)
		if (pTPM->semLogMsg == MONITOR_THREAD_ID)
		{
			pTPM->semLogMsg = LOG_SEM_FREE;
			pThreadPoolHeader->LogMsgOverflowed++;
		}
		return;
	}

	//  Compute the address of the tail entry
	pLogEntry = (LogEntry *) (pThreadPoolHeader->pLogPool + sizeof(Logpool_Header) + (pLogPoolHeader->LogQueueTail * sizeof(LogEntry)));

	//  Validate the pLogEntry Address
	if (!validateLogEntryAddress(pThreadPool, pLogEntry))
	{
		if (pTPM->semLogMsg == MONITOR_THREAD_ID)
		{
			pTPM->semLogMsg = LOG_SEM_FREE;
			pThreadPoolHeader->Guard02++;
			pThreadPoolHeader->LogMsgDiscarded++;
		}
		return;
	}

	//  Complete the log entry
	memcpy(&pLogEntry->tdEventTime, &pTPM->tdLogMsg, sizeof(TIMEDATE));
	strcpy_s(pLogEntry->szLogMsg, MAX_MSG, pTPM->szLogMsg);
	pLogEntry->LogThread = pTPM->semLogMsg;

	//  Clear the semaphore
	pTPM->semLogMsg = LOG_SEM_FREE;

	//  Increment the tail ponter to the next free entry
	pLogPoolHeader->LogQueueTail++;
	if (pLogPoolHeader->LogQueueTail == pThreadPoolHeader->LogPoolSize) pLogPoolHeader->LogQueueTail = 0;
	pLogPoolHeader->LogPoolEntries++;
	
	//  Update the counts
	pThreadPoolHeader->LogMsgQueued++;
	return;
}

//  validateLogEntryAddress
//
//  This function will validate the computed address of a log entry to ensure that it is within a valid range
//
//  Parameters
//
//	1		-		UCHAR *		-  Pointer to the thread pool header
//  2		-		void *		-  Address of the log entry to be validated
//
//  Returns
//
//  BOOL  -			TRUE if the address is valid
//                  FALSE if the address is invalid
//
//  API Use Notes:
//
//
BOOL ThreadMonitor::validateLogEntryAddress(UCHAR *pTP, void *pEntry)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	UCHAR						*pTest;									//  Address to be tested
	UCHAR						*pLowest;								//  Lowest address
	UCHAR						*pHighest;								//  Highest Address

	pThreadPoolHeader = (Threadpool_Header *) pTP;						//  Address the header
	pTest = (UCHAR *) pEntry;

	//  Compute the lowest and highest addresses
	pLowest = pThreadPoolHeader->pLogPool + sizeof(Logpool_Header);
	pHighest = pLowest + (pThreadPoolHeader->LogPoolSize * sizeof(LogEntry));

	//  Check bounds
	if (pTest < pLowest) return FALSE;
	if (pTest > pHighest) return FALSE;

	return TRUE;
}

//  AttachCommandHandler
//
//  This function will attach the passed Command Handle to the Thread Monitor activities.
//
//  Parameters
//
//  1  -		CommandHandler *		-  Pointer to the Command Handler
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadMonitor::AttachCommandHandler(CommandHandler *chApp)
{
	chApplication = chApp;
	if (chApplication == NULL) xeLocal->LogMessage(MSG_DXR2226I, MONITOR_THREAD_ID);
	else xeLocal->LogMessage(MSG_DXR2223I, MONITOR_THREAD_ID);
	return;
}

//  ReleaseAllMappedDBH
//
//  This function will close any locally (thread) mapped DB Handles that can be closed because
//  the native handle has been invalidated.
//
//  Parameters
//
//  1  -  int				-	ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadMonitor::ReleaseAllMappedDBH(int iThreadID)
{
	int					iIndexX;								//  Index of native DB Handles
	int					iIndexY;								//  Index of threads
	int					iReleasedEntries;						//  Count of released thread entries
	DBHANDLE			*pMappedDBH;							//  Pointer to mapped DB Handle entries
	DBHANDLE			*pMyMappedDBH;							//  Pointer to mapped DB Handle entries
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Release any mapped file handles that have been completely freed
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;
	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;
	for (iIndexX = 0; iIndexX < MAX_DBH_CACHE; iIndexX++)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (pThreadPoolHeader->MaxThreads + 4));
		if (pMyMappedDBH[0] == (DBHANDLE) FREEHANDLE)
		{
			iReleasedEntries = 0;
			for (iIndexY = 1; iIndexY <= (pThreadPoolHeader->MaxThreads + 3); iIndexY++)
			{
				if (pMyMappedDBH[iIndexY] == NULLHANDLE) iReleasedEntries++;
			}
			if (iReleasedEntries == (pThreadPoolHeader->MaxThreads + 3)) 
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2226T, pMyMappedDBH[0], iThreadID);
				xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
				pMyMappedDBH[0] = NULLHANDLE;
			}
		}
	}
	return;
}

//  CloseMyMappedDBH
//
//  This function will close all locally (thread) mapped DB Handles.
//
//  Parameters
//
//  1  -  int				-	ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadMonitor::CloseMyMappedDBH(int iThreadID)
{
	int					iIndexX;								//  Index of native DB Handles
	DBHANDLE			*pMappedDBH;							//  Pointer to a mapped DB Handle
	DBHANDLE			*pMyMappedDBH;							//  Pointer to mapped DB Handle entries
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	int					iThreadIndex;							//  Thread Index
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Close any mapped file handles that belong to this thread
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;

	//  Compute the Thread Index
	switch(iThreadID)
	{
	case CONTROL_THREAD_ID:
		iThreadIndex = CONTROL_THREAD_ID;
		break;
	case MONITOR_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 1;
		break;
	case SCHEDULER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 2;
		break;
	case DISPATCHER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 3;
		break;
	default:
		iThreadIndex = iThreadID;
		break;
	}

	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;
	for (iIndexX = 0; iIndexX < MAX_DBH_CACHE; iIndexX++)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (pThreadPoolHeader->MaxThreads + 4));
		if ((pMyMappedDBH[0] == (DBHANDLE) FREEHANDLE) && (pMyMappedDBH[iThreadIndex] != NULLHANDLE))
		{
			NSFDbClose(pMyMappedDBH[iThreadIndex]);				//  Close the mapped DB handle
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2227T, pMyMappedDBH[iThreadIndex], iThreadID);
			xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
			pMyMappedDBH[iThreadIndex] = NULLHANDLE;			//  Mark the entry closed
		}
	}
	ReleaseAllMappedDBH(iThreadID);										//  Release any completely free rows
	return;
}

//  ForceCloseMyMappedDBH
//
//  This function will force closed all locally (thread) mapped DB Handles.
//
//  Parameters
//
//  1  -  int				-	ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadMonitor::ForceCloseMyMappedDBH(int iThreadID)
{
	int					iIndexX;								//  Index of native DB Handles
	DBHANDLE			*pMappedDBH;							//  Pointer to a mapped DB Handle
	DBHANDLE			*pMyMappedDBH;							//  Pointer to mapped DB Handle entries
	int					iThreadIndex;							//  Thread Index
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Close any mapped file handles that belong to this thread
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;

	//  Compute the Thread Index
	switch(iThreadID)
	{
	case CONTROL_THREAD_ID:
		iThreadIndex = CONTROL_THREAD_ID;
		break;
	case MONITOR_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 1;
		break;
	case SCHEDULER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 2;
		break;
	case DISPATCHER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 3;
		break;
	default:
		iThreadIndex = iThreadID;
		break;
	}

	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;
	for (iIndexX = 0; iIndexX < MAX_DBH_CACHE; iIndexX++)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (pThreadPoolHeader->MaxThreads + 4));
		if (pMyMappedDBH[iThreadIndex] != NULLHANDLE)
		{
			NSFDbClose(pMyMappedDBH[iThreadIndex]);			//  Close the mapped DB handle
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2228T, pMyMappedDBH[iThreadIndex], iThreadID);
			xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
			pMyMappedDBH[iThreadIndex] = NULLHANDLE;			//  Mark the entry closed
		}
	}
	ReleaseAllMappedDBH(iThreadID);										//  Release any completely free rows
	return;
}

#ifdef DXIP
#ifdef WIN32
//  StartRecording
//
//  This function will create and open the Instrumentation Package Recording File
//
//  Parameters
//
//  1  -  int				-	ID of the calling thread
//
//  Returns
//
//  HANDLE		-  The File handle of the recording file or NULLHANDLE if it could not be opened
//
//  API Use Notes:
//
//
HANDLE ThreadMonitor::StartRecording(int iThreadID)
{
	WORD			wRetLen;												//  Returned length
	TIME			tNow;													//  Current time/date and components
	HANDLE			hfRec;													//  Handle of the recording file
	DWORD			dwWinError;												//  Windows Error Code
	char			szRecDir[MAXPATH + 1];									//  Recording File directory
	char			szRecFile[MAXPATH + 1];									//  Name of the recording file
	char			szMsg[MAX_MSG + 1];										//  Generic Message Buffer

	//  Get (or build) the directory name
	szRecFile[0] = '\0';
	wRetLen = OSGetDataDirectory(szRecFile);
	szRecFile[wRetLen] = '\0';
	sprintf_s(szRecDir, MAXPATH, "%s/%s/", szRecFile, IPREC_DIR);
	//  Get the current date/time to format the file name
	OSCurrentTIMEDATE(&tNow.GM);
	TimeGMToLocal(&tNow);
	//  Format the completed file path
	sprintf_s(szRecFile, MAXPATH, "%s%s%s%04i%02i%02i-%02i%02i%02i%s",
						szRecDir,
						IPREC_PREFIX,
						xeLocal->rsLocal->APPName,
						tNow.year,
						tNow.month,
						tNow.day,
						tNow.hour,
						tNow.minute,
						tNow.second,
						IPREC_TYPE);
	//  Open the file
	hfRec = NULLHANDLE;
	hfRec = CreateFile(szRecFile, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hfRec == (HANDLE) HFILE_ERROR)
	{
		dwWinError = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2229E, szRecFile);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetWINAPIError(dwWinError, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Show recording started
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2230I, szRecFile);
	xeLocal->LogMessage(szMsg, iThreadID);

	return hfRec;
}
#endif
#ifdef UNIX
//  StartRecording
//
//  This function will create and open the Instrumentation Package Recording File
//
//  Parameters
//
//  1  -  int				-	ID of the calling thread
//
//  Returns
//
//  int		-  The File descriptor of the recording file or -1 if it could not be opened
//
//  API Use Notes:
//
//
int ThreadMonitor::StartRecording(int iThreadID)
{
	WORD			wRetLen;												//  Returned length
	TIME			tNow;													//  Current time/date and components
	int				hfRec;													//  Descriptor of the recording file
	mode_t			mCreate;												//  File Create Mode
	int				iCloseRC;												//  Return Code from Close
	char			szRecDir[MAXPATH + 1];									//  Recording File directory
	char			szRecFile[MAXPATH + 1];									//  Name of the recording file
	char			szMsg[MAX_MSG + 1];										//  Generic Message Buffer

	//  Get (or build) the directory name
	szRecFile[0] = '\0';
	wRetLen = OSGetDataDirectory(szRecFile);
	szRecFile[wRetLen] = '\0';
	sprintf_s(szRecDir, MAXPATH, "%s/%s/", szRecFile, IPREC_DIR);
	//  Get the current date/time to format the file name
	OSCurrentTIMEDATE(&tNow.GM);
	TimeGMToLocal(&tNow);
	//  Format the completed file path
	sprintf_s(szRecFile, MAXPATH, "%s%s%s%04i%02i%02i-%02i%02i%02i%s",
						szRecDir,
						IPREC_PREFIX,
						xeLocal->rsLocal->APPName,
						tNow.year,
						tNow.month,
						tNow.day,
						tNow.hour,
						tNow.minute,
						tNow.second,
						IPREC_TYPE);
	//  Open the file
	hfRec = -1;

	mCreate = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
	hfRec = open(szRecFile, O_RDWR | O_CREAT | O_EXCL | O_LARGEFILE, mCreate);
	if (hfRec == -1)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2229E, szRecFile);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetUNIXAPIError(errno, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
		return -1;
	}

	//  Show recording started
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2230I, szRecFile);
	xeLocal->LogMessage(szMsg, iThreadID);

	return hfRec;
}
#endif
#endif

#ifdef DXIP
#ifdef WIN32
//  MonitorInstruments
//
//  This function will check each worker thread to see if any Instrumentation Package recordings have been made
//  if any have been made then they will be added to the recording buffer and if that buffer is full it will be written.
//
//  Parameters
//
//  1  -  HANDLE			-   Handle of the recording file
//  2  -  UCHAR *			-   Pointer to the recording buffer
//  3  -  DWORD *			-   Pointer to the number of bytes in the recording buffer
//  4  -  int				-	ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadMonitor::MonitorInstruments(HANDLE hfRec, UCHAR *pRecBfr, DWORD *pBytesRecorded, int iThreadID)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a threadpool member
	PCRecRC						pRecs[IPC_STACK_SIZE];					//  Copy of the recording stack
	DWORD						dwRecEvent;								//  High recorded event number
	DWORD						dwWritten;								//  Count of bytes written
	DWORD						dwWinError;								//  Windows Error Code
	int							iIndex;									//  Generic Index
	int							iRecNo;									//  Record slot number
	char						szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  If inactive then just return
	if (hfRec == NULLHANDLE) return;

	//  Process each Thread Pool Member in turn
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header

	for (iIndex = 0; iIndex < pThreadPoolHeader->MaxThreads; iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) (pThreadPool + sizeof(Threadpool_Header) + (iIndex * sizeof(Threadpool_Member)));
		dwRecEvent = pThreadPoolMember->EventCount;
		if (dwRecEvent > pThreadPoolMember->LastCollection)
		{
			//  Capture the current state of the recording stack
			memcpy(&pRecs[0], &pThreadPoolMember->CStack[0], sizeof(PCRecRC) * IPC_STACK_SIZE);
			
			//  Loop through each of the records processing any that have been recorded since the last collection
			for (iRecNo = 0; iRecNo < IPC_STACK_SIZE; iRecNo++)
			{
				//  Process any unrecorded events
				if (pRecs[iRecNo].PCBase.EventNo > pThreadPoolMember->LastCollection)
				{
					//  Determine if there is insufficient space in the current buffer for the record
					if ((*pBytesRecorded + pRecs[iRecNo].PCBase.Size) > IPREC_BFR_SIZE)
					{
						//  Write and reset the current buffer
						if (!WriteFile(hfRec, pRecBfr, *pBytesRecorded, &dwWritten, NULL))
						{
							dwWinError = GetLastError();
							sprintf_s(szMsg, MAX_MSG, MSG_DXR2232E, *pBytesRecorded);
							xeLocal->LogMessage(szMsg, iThreadID);
							xeLocal->GetWINAPIError(dwWinError, szMsg);
							xeLocal->LogMessage(szMsg, iThreadID);
						}
						else
						{
							if (*pBytesRecorded != dwWritten)
							{
								sprintf_s(szMsg, MAX_MSG, MSG_DXR2233E, *pBytesRecorded, dwWritten);
								xeLocal->LogMessage(szMsg, iThreadID);
							}
						}
						memset(pRecBfr, 0, sizeof(PCRecRC) * IPC_STACK_SIZE);
						*pBytesRecorded = 0;
					}
					//  Add the record into the collection buffer
					memcpy(pRecBfr + *pBytesRecorded, &pRecs[iRecNo].PCBase.Type, pRecs[iRecNo].PCBase.Size);
					*pBytesRecorded += pRecs[iRecNo].PCBase.Size;
				}
			}
			pThreadPoolMember->LastCollection = dwRecEvent;
		}
	}
	return;

}
#endif
#ifdef UNIX
//  MonitorInstruments
//
//  This function will check each worker thread to see if any Instrumentation Package recordings have been made
//  if any have been made then they will be added to the recording buffer and if that buffer is full it will be written.
//
//  Parameters
//
//  1  -  int				-   Descriptor of the recording file
//  2  -  UCHAR *			-   Pointer to the recording buffer
//  3  -  DWORD *			-   Pointer to the number of bytes in the recording buffer
//  4  -  int				-	ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
//
void ThreadMonitor::MonitorInstruments(int hfRec, UCHAR *pRecBfr, DWORD *pBytesRecorded, int iThreadID)
{
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThreadPoolMember;						//  Pointer to a threadpool member
	PCRecRC						pRecs[IPC_STACK_SIZE];					//  Copy of the recording stack
	DWORD						dwRecEvent;								//  High recorded event number
	DWORD						dwWritten;								//  Count of bytes written
	ssize_t						stRead;									//  Size read/written
	int							iIndex;									//  Generic Index
	int							iRecNo;									//  Record slot number
	char						szMsg[MAX_MSG + 1];						//  Generic Message Buffer

	//  If inactive then just return
	if (hfRec == NULLHANDLE) return;

	//  Process each Thread Pool Member in turn
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header

	for (iIndex = 0; iIndex < pThreadPoolHeader->MaxThreads; iIndex++)
	{
		pThreadPoolMember = (Threadpool_Member *) (pThreadPool + sizeof(Threadpool_Header) + (iIndex * sizeof(Threadpool_Member)));
		dwRecEvent = pThreadPoolMember->EventCount;
		if (dwRecEvent > pThreadPoolMember->LastCollection)
		{
			//  Capture the current state of the recording stack
			memcpy(&pRecs[0], &pThreadPoolMember->CStack[0], sizeof(PCRecRC) * IPC_STACK_SIZE);
			
			//  Loop through each of the records processing any that have been recorded since the last collection
			for (iRecNo = 0; iRecNo < IPC_STACK_SIZE; iRecNo++)
			{
				//  Process any unrecorded events
				if (pRecs[iRecNo].PCBase.EventNo > pThreadPoolMember->LastCollection)
				{
					//  Determine if there is insufficient space in the current buffer for the record
					if ((*pBytesRecorded + pRecs[iRecNo].PCBase.Size) > IPREC_BFR_SIZE)
					{
						//  Write and reset the current buffer
						stRead = write(hfRec, pRecBfr, *pBytesRecorded);
						if (stRead == -1)
						{
							sprintf_s(szMsg, MAX_MSG, MSG_DXR2232E, *pBytesRecorded);
							xeLocal->LogMessage(szMsg, iThreadID);
							xeLocal->GetUNIXAPIError(errno, szMsg);
							xeLocal->LogMessage(szMsg, iThreadID);
						}
						else
						{
							dwWritten = stRead;
							if (*pBytesRecorded != dwWritten)
							{
								sprintf_s(szMsg, MAX_MSG, MSG_DXR2233E, *pBytesRecorded, dwWritten);
								xeLocal->LogMessage(szMsg, iThreadID);
							}
						}
						memset(pRecBfr, 0, sizeof(PCRecRC) * IPC_STACK_SIZE);
						*pBytesRecorded = 0;
					}
					//  Add the record into the collection buffer
					memcpy(pRecBfr + *pBytesRecorded, &pRecs[iRecNo].PCBase.Type, pRecs[iRecNo].PCBase.Size);
					*pBytesRecorded += pRecs[iRecNo].PCBase.Size;
				}
			}
			pThreadPoolMember->LastCollection = dwRecEvent;
		}
	}
	return;

}
#endif
#endif
