#pragma once
//  ThreadMonitor
//
//  This class provides monitor functions for the Thread Pool Manager.
//
//  Message Identifier:      DXR
//  Message Range:           2200 - 2399
//  Trace area assignment:   TRACE_MT (02)
//  DX Version:				 3.14.0
//  Version:                 3.14.0  Build: 108
//
//  Version History:
//
//  1.0.0		07/06/2003		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.0		18/09/2009		New log handler for multithreaded applications
//  3.6.1		10/10/2009		Additional controls added
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.7.2		12/01/2010		DbCopier improvements for corrupt folder handling
//								Remote Fixup on DbOpen
//								Logger improvements
//  3.7.4		19/03/2010		Fixed various issues with log/repository configuration
//  3.9.0		24/07/2010		Linkage changed for logging & request posting
//								A CommandHandler can be hooked into the monitor processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.12.0		21/12/2011		DXIP introduced
//	3.14.0		19/01/2015		x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2003 - 2015
//

//  Platform includes
#include "../Platform/PlatBase.h"

//  Return Includes
#include "../ExecEnvironment.h"								//  Basic Runtime
#include "ThreadManagerPolicy.h"							//  Thread Manager Policy
#include "ThreadStructs.h"									//  Thread Structures
#include "WorkerThread.h"									//  Worker Thread
#include "../MTX/CommandHandler.h"							//  Command Handler base class

//  Local Definitions
#define		IDLE_WAITINTERVAL				100				//  Wait 100 Ms while in the idle state
#define		WTSTATE_INTERVAL				50				//  Wait 50 Ms for Worker thread state change
#define		WTSTATE_LIMIT					10				//  Max wait intervals for worker thread state change
#define		THREAD_STACK_SIZE				24*1024			//  Stack Size
#define		LOG_UNCOMMIT_LIMIT				30				//  Log uncommit limit
#ifdef UNIX
#define		TSTART_RETRY_LIMIT				10				//  10 attempts max to start a thread
#endif
#ifdef DXIP
#define		IPREC_DIR						"IBM_TECHNICAL_SUPPORT"
#define		IPREC_PREFIX					"DXIP-"
#define		IPREC_TYPE						".ipr"
#define		IPREC_BFR_SIZE					8*1024			//  Size of the recording buffer
#ifdef UNIX
#include <fcntl.h>											//  Include POSIX File I/O
#endif
#endif

//  Trace Messages
#define		MSG_DXR2200T			"DXR2200T: The monitor thread has entered the not started state."
#define		MSG_DXR2201T			"DXR2201T: The monitor thread has entered the running state."
#define		MSG_DXR2202T			"DXR2202T: The monitor thread has entered the quitting state."
#define		MSG_DXR2210T			"DXR2210T: A new ThreadMonitor instance has been created @%p."
#define		MSG_DXR2211T			"DXR2211T: A ThreadMonitor instance @%p has been destroyed."
#define		MSG_DXR2213T			"DXR2213T: Log messages queued %d, log messages written %d."
#define		MSG_DXR2214T			"DXR2214T: Log messages dropped %d, discarded %d, overflowed %d."
#define		MSG_DXR2215T			"DXR2215T: Dispatches by brute force %d, by priority %d." 
#define		MSG_DXR2216T			"DXR2216T: Requests dispatched by brute force %d, by priority %d." 
#define		MSG_DXR2217T			"DXR2217T: Null dispatches because of no work %d, because all threads were busy %d." 
#define		MSG_DXR2218T			"DXR2218T: Priority dispatch windows closed because of no work %d, because all threads were busy %d." 
#define		MSG_DXR2219T			"DXR2219T: Count of RQEs posted %d, waits for RQE posting %d." 
#define		MSG_DXR2220T			"DXR2220T: RQEs rejoined %d, with %d rejoin busy signals."
#define		MSG_DXR2224T			"DXR2224T: There are %i requests pending execution and %i pending rejoin, %i threads are available."
#define		MSG_DXR2225T			"DXR2225T: Pending Pool full waits %i, backoffs %i, lane backoffs %i, CMLS denials %i."

#ifdef W64
#define		MSG_DXR2226T			"DXR2226T: Released native DB handle %i by thread %i."
#define		MSG_DXR2227T			"DXR2227T: Closed local DB handle %i by thread %i."
#define		MSG_DXR2228T			"DXR2228T: Forced closed local DB handle %i by thread %i."
#else
#define		MSG_DXR2226T			"DXR2226T: Released native DB handle %p by thread %i."
#define		MSG_DXR2227T			"DXR2227T: Closed local DB handle %p by thread %i."
#define		MSG_DXR2228T			"DXR2228T: Forced closed local DB handle %p by thread %i."
#endif

//  Information Messages
#define		MSG_DXR2223I			"DXR2223I: A Command Handler has been attached to the Thread Monitor."
#define		MSG_DXR2226I			"DXR2226I: The current Command Handler has been detached from the Thread Monitor."
#ifdef DXIP
#define		MSG_DXR2230I			"DXR2230I: Instrumentation Package is recording on '%s'."
#endif

//  Error Messages
#define		MSG_DXR2203E			"DXR2203E: The monitor thread has detected an invalid state change and is quitting."
#define		MSG_DXR2204E			"DXR2204E: Unable to start worker threads during initialisation."
#define		MSG_DXR2205E			"DXR2205E: Unable to allocate a new Thread ID, thread pool is full."
#define		MSG_DXR2206E			"DXR2206E: The monitor thread was unable to create a new worker thread."
#define		MSG_DXR2207E			"DXR2207E: The monitor thread was unable to initialise a new worker thread."
#define		MSG_DXR2208E			"DXR2208E: The monitor thread was unable to start a new worker thread."
#define		MSG_DXR2209E			"DXR2209E: The monitor thread was unable to drain a worker thread."
#define		MSG_DXR2212E			"DXR2212E: The thread was unable to initialise the Notes runtime."
#define		MSG_DXR2221E			"DXR2221E: Failed to re-open the repository for asynchronous logging."
#define		MSG_DXR2222E			"DXR2222E: Failed to re-open the current log document  for asynchronous logging."
#ifdef DXIP
#define		MSG_DXR2229E			"DXR2229E: Unable to open the recording file '%s'."
#define		MSG_DXR2231E			"DXR2231E: Failed to allocate %i bytes for Instrumentation Package recording buffer."
#define		MSG_DXR2232E			"DXR2232E: Failed to write %i bytes to the Instrumentation Package recording file."
#define		MSG_DXR2233E			"DXR2233E: Short write detected attempting %i bytes wrote %i on the Instrumentation Package recording file."
#endif

class ThreadMonitor
{
public:
	//  Constructor
	ThreadMonitor(ExecEnvironment *xeParent, UCHAR *pTP, ThreadManagerPolicy *tmpParent);

	//  Destructor
	~ThreadMonitor(void);

	//  Public Members
	UCHAR				*pThreadPool;						//  Pointer to the Thread Pool
	int					LogCommitCycles;					//  Cycles in which to commit log updates

	//  Public Functions
#ifdef WIN32
	static UINT LNPUBLIC StartWrapper(void *pObject);		//  Wrapper for the Start Monitor function
#endif

#ifdef UNIX
	static void * LNPUBLIC StartWrapper(void *pObject);		//  Wrapper for the Start Monitor function
#endif

	UINT Start(void);										//  Start the Monitor
	void DrainAnyWorker(UCHAR *pTP, int iThreadID);			//  Shutdown any worker thread
	void AttachCommandHandler(CommandHandler *chApp);		//  Attach a CommandHandler
	void ProcessLogSemaphore(Threadpool_Member *pTPM);		//  Process a Log Message Semaphore

private:
	//  Private Members
	ExecEnvironment *xeLocal;								//  Multi-Threading Runtime
	ThreadManagerPolicy *tmpLocal;							//  Current Policy object
	CommandHandler  *chApplication;							//  Application provided command handler
#ifdef UNIX
	pthread_attr_t	paThread;								//  Thread Attributes
#endif

	//  Private Functions
	BOOL WaitForWorkerState(Threadpool_Member *pTPMember, UINT State, int iThreadID);		//  Wait for State
	void MonitorLogSemaphores(void);						//  Monitor the Log Semaphores for posted messages
	BOOL validateLogEntryAddress(UCHAR *pTP, void *pEntry);   //  Validate a log entry address
	void MonitorLogPool(void);								//  Monitor the Log Pool for posted messages
	void WriteLogEntry(void);								//  Write a single log entry
	int ComputeThreadDelta(Threadpool_Header * pTPH);		//  Compute the change in thread pool members needed
	BOOL StartWorker(Threadpool_Header * pTPH);				//  Start a new Worker Thread
	int GetThreadID(UCHAR * pTP);							//  Return the first free ThreadID
	void PerformSchedulerPrefsReview(UCHAR * pTP);			//  Review scheduler preferences
	void PerformStrategyReview(UCHAR * pTP);				//  Review strategy
	//  Thread Local Database Handle Functions
	void ReleaseAllMappedDBH(int iThreadID);				//  Release any freed DBH mapping entries
	void CloseMyMappedDBH(int iThreadID);					//  Close any locally mapped DB handles
	void ForceCloseMyMappedDBH(int iThreadID);				//  Force close of all locall mapped DB Handles
#ifdef DXIP
#ifdef WIN32
	HANDLE StartRecording(int iThreadID);					//  Create and open the IP recording file
	void MonitorInstruments(HANDLE hfRec, UCHAR *pRecBfr, DWORD *pBytesRecorded, int iThreadID);		//  Monitor any active instruments
#endif
#ifdef UNIX
	int StartRecording(int iThreadID);						//  Create and open the IP recording file
	void MonitorInstruments(int hfRec, UCHAR *pRecBfr, DWORD *pBytesRecorded, int iThreadID);			//  Monitor any active instruments
#endif
#endif
};
