//  ThreadScheduler
//
//  This class provides the scheduling function for requests for execution that are submitteD to the
//  the Thread Pool manager queues.
//
//  Message Identifier:      DXR
//  Message Range:           2401 - 2499
//  Trace area assignment:   TRACE_MT (02)
//  DX Version:				 3.12.0
//  Version:                 3.12.0  Build: 106
//
//  Version History:
//
//  1.0.0		07/06/2003		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.1		30/09/2009		Fixed stats & counter integrity for dispatcher
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.7.0		23/12/2009		Added race auto-correction to Scheduler
//  3.9.0		24/07/2010		Linkage changed for logging & request posting
//  3.11.1		22/10/2011		VS 2011 Updates
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2003 - 2012
//

//  Application Includes
#include "ThreadScheduler.h"

//  Constructor
ThreadScheduler::ThreadScheduler(ExecEnvironment *xeParent, UCHAR * pTP, ThreadManagerPolicy *tpmParent)
{
	char			szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Save the Runtime and policy address
	xeLocal = xeParent;
	tmpLocal = tpmParent;

	//  Save the Thread Pool address
	pThreadPool = pTP;

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2404T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg);
}

//  Destructor
ThreadScheduler::~ThreadScheduler(void)
{
	char			szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Trace the destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2405T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg);
}

//  Start
//
//  This function will start the scheduling process.
//
//  Parameters
//
//  Returns
//
//  UINT  -			Status of exiting from the scheduling loop
//
//  API Use Notes:
//
//
UINT ThreadScheduler::Start(void)
{
	STATUS					stAPIRC;								//  Return code from Notes API calls
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	int						iPriorityReviewTrigger = 0;				//  Trigger counter for priority review
	int						iHealthCheck = 0;						//  Health check trip counter
	char					szMsg[MAX_MSG + 1];						//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header

	//  Perform any pre-work and mark the thread as not started
	stAPIRC = NotesInitThread();
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR2406E, SCHEDULER_THREAD_ID);			//  Show the error
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, SCHEDULER_THREAD_ID);				//  Show the API error message
		pThreadPoolHeader->TSchedState = TSSTATE_DRAINED;
#ifdef WIN32
		_endthreadex(THREAD_RC_ERROR);									//  Terminate the thread
#endif

#ifdef UNIX
		pthread_exit(NULL);
#endif
		return THREAD_RC_OK;
	}

	pThreadPoolHeader->TSchedState = TSSTATE_NOTSTARTED;
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2400T, SCHEDULER_THREAD_ID);				//  Show the state change

	while (pThreadPoolHeader->TSchedState == TSSTATE_NOTSTARTED)
	{
		SLEEP(IDLE_WAITINTERVAL);
	}

	//  Check that the transition requested is to the started state
	if (pThreadPoolHeader->TSchedState == TSSTATE_START)
	{
		pThreadPoolHeader->TSchedState = TSSTATE_RUNNING;
		xeLocal->LogTrace(TRACE_MT, MSG_DXR2401T, SCHEDULER_THREAD_ID);			//  Show the state change
	}

	//  Running state loop - monitor the Request ready queue adjusting the priority of entries
	while (pThreadPoolHeader->TSchedState == TSSTATE_RUNNING)
	{
		iHealthCheck++;
		iPriorityReviewTrigger++;
		if (iPriorityReviewTrigger == tmpLocal->SchedulerPrioReviewLimit)
		{
			PerformPriorityReview(pThreadPool);						//  Review request priority
			iPriorityReviewTrigger = 0;								//  Reset the trigger
		}
		if (iHealthCheck == tmpLocal->SchedulerHealthCheckLimit)
		{
			PerformSelfHealing(pThreadPool);						//  Perform health review and fix
			iHealthCheck = 0;										//  Reset the trigger
		}

		SLEEP(tmpLocal->SchedulerWaitCycle);
	}

	//  Check that the transition requested is to the quit state
	if (pThreadPoolHeader->TSchedState == TSSTATE_QUIT)
	{
		pThreadPoolHeader->TSchedState = TSSTATE_QUITTING;
		xeLocal->LogTrace(TRACE_MT, MSG_DXR2402T, SCHEDULER_THREAD_ID);				//  Show the state change
	}
	else
	{
		pThreadPoolHeader->TSchedState = TSSTATE_QUITTING;
		xeLocal->LogMessage(MSG_DXR2403E, SCHEDULER_THREAD_ID);						//  Show inconsistent state change
	}

	NotesTermThread();
	
	//  Mark the thread as drained and terminate the thread
	xeLocal->LogTrace(TRACE_MT, MSG_DXR2407T, SCHEDULER_THREAD_ID);					//  Show the state change
	pThreadPoolHeader->TSchedState = TSSTATE_DRAINED;
#ifdef WIN32
	_endthreadex(THREAD_RC_OK);														//  Terminate the thread
#endif

#ifdef UNIX
	pthread_exit(NULL);
#endif
	return THREAD_RC_OK;
}

//  StartWrapper
//
//  This function is the static wrapper for the Start() function allowing it to be started as an independent thread.
//
//  Parameters
//
//  Returns
//
//  UINT  -			Status indication  
//
//  API Use Notes:
//
//

#ifdef WIN32
UINT LNPUBLIC ThreadScheduler::StartWrapper(void * pObject)
#endif
#ifdef UNIX
void * LNPUBLIC ThreadScheduler::StartWrapper(void * pObject)
#endif

{
	ThreadScheduler		*tsThis;

	//  Wrapper function calls the Start function in the Object
	tsThis = (ThreadScheduler *) pObject;
#ifdef WIN32
	return tsThis->Start();
#endif
#ifdef UNIX
	return (void *) tsThis->Start();
#endif
}

//  PerformPriorityReview
//
//  This function will review and adjust the priority of execution requests in the Thread Pool Manager queues.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadScheduler::PerformPriorityReview(UCHAR * pTP)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	RQElement				*pRQE;									//  Pointer to a Request Element (RQE)
	int						iIndex;									//  Index of the requests

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	//  Perform a review of Request priority
	if (pThreadPoolHeader->PendRQECount == 0) return;

	//  See if ageing is in effect
	if (pThreadPoolHeader->PrioControl & PRIO_POLICY_AGERQS)
	{
		pRQE = (RQElement *) pThreadPoolHeader->pPendRQEPool;
		//  Age each RQE in the pending pool
		for (iIndex = 0; iIndex < pThreadPoolHeader->PendRQEPoolSize; iIndex++)
		{
			if (pRQE->RQEState & RQESTATE_READY)
			{
				pRQE->RQEPriority++;
				//  If the controls indicate preference boosting then perform the boost
				if (pThreadPoolHeader->PrioControl & PRIO_POLICY_PREFBOOST)
				{
					if (pRQE->RQEAttrs == pThreadPoolHeader->PrefSchedAttrs)
					{
						pRQE->RQEPriority += tmpLocal->SchedulerProrityBoost;
					}
				}
			}
			pRQE++;
		}
	}
	return;
}

//  PerformSelfHealing
//
//  This function will check the consistency of various counts and states and perform self-healing 
//  corrections as needed.
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//  API Use Notes:
//  
//
void ThreadScheduler::PerformSelfHealing(UCHAR *pTP)
{
	//  There are currently no actions performed in self-healing

	return;
}

