//  WorkerThread
//
//  This class provides the wrapper functions for executing requests for work.
//  
//
//  Message Identifier:      DXR
//  Message Range:           2601 - 2799
//  Trace area assignment:   TRACE_MT (02)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  1.0.0		07/06/2003		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.0		18/09/2009		New log handler for multithreaded applications
//  3.6.1		10/10/2009		Additional controls added
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.9.0		24/07/2010		Linkage changed for logging & request posting and rejoining
//								A CommandHandler can be hooked into the monitor processing
//  3.10.2		18/11/2010		MTIO support added to DbMover
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.12.0		21/12/2011		DXIP Introduced
//  3.14.1		20/01/2015		x64 compatability
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2003 - 2015
//

//  Application Includes
#include "WorkerThread.h"

//  Constructor
WorkerThread::WorkerThread(ExecEnvironment *xeParent, UCHAR * pTP, ThreadManagerPolicy *tmpParent, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Save the Runtime address & Policy
	xeLocal = xeParent;
	tmpLocal = tmpParent;
	//  Save the Thread Pool address
	pThreadPool = pTP;
	//  Set the Thread ID
	ThreadID = iThreadID;

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2611T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
}

//  Destructor
WorkerThread::~WorkerThread(void)
{
	char			szMsg[MAX_MSG + 1];								//  Generic message buffer

	//  Trace the destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2612T, this);
	xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);
}

//  Start
//
//  This function will start the worker thread
//
//  Parameters
//
//  Returns
//
//  UINT  -			Status of exiting from the main execution loop
//
//  API Use Notes:
//
//
UINT WorkerThread::Start(void)
{
	STATUS						stAPIRC;								//  Return code from Notes API calls
	Threadpool_Header			*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	Threadpool_Member			*pThisThread;							//  Pointer to this threadpool member
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;				//  Address the header
	pThisThread = (Threadpool_Member *) (pThreadPool + sizeof(Threadpool_Header) + ((ThreadID - 1) * sizeof(Threadpool_Member)));

	//  Perform any pre-work and mark the thread as not started
	stAPIRC = NotesInitThread();
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->LogMessage(MSG_DXR2613E, ThreadID);					//  Show the error
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, ThreadID);							//  Show the API error message
		pThisThread->TMemState = WTSTATE_DRAINED;	
#ifdef WIN32
		_endthreadex(THREAD_RC_ERROR);									//  Terminate the thread
#endif

#ifdef UNIX
		pthread_exit(NULL);
#endif
		return THREAD_RC_OK;
	}

	pThisThread->TMemState = WTSTATE_NOTSTARTED;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2600T, ThreadID);
	xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);						//  Show the state change

	while (pThisThread->TMemState == WTSTATE_NOTSTARTED)
	{
		SLEEP(IDLE_WAITINTERVAL);
	}

	//  Check that the transition requested is to the started state
	if (pThisThread->TMemState == WTSTATE_START)
	{
		pThisThread->TMemState = WTSTATE_RUNNING;
		pThreadPoolHeader->AvailableThreads++;
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2601T, ThreadID);
		xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);					//  Show the state change
	}

	//  Running state loop - Execute work requests as they are presented
	while ((pThisThread->TMemState == WTSTATE_RUNNING) || (pThisThread->TMemState == WTSTATE_POSTED))
	{
		while (pThisThread->TMemState == WTSTATE_RUNNING)
		{
			CloseMyMappedDBH(ThreadID);
			CloseMyMappedFileHandles();
			SLEEP(tmpLocal->WorkerWaitCycle);
			xeLocal->TestGlobalPermit();
		}

		//  If the thread is posted then there is work to do
		if (pThisThread->TMemState == WTSTATE_POSTED && xeLocal->TestGlobalPermit())
		{
			pThreadPoolHeader->AvailableThreads--;
			ExecuteRequest(pThisThread, pThreadPool);
			CloseMyMappedDBH(ThreadID);
			CloseMyMappedFileHandles();
			DecrementCMLSThreadCount(pThreadPool, pThisThread->rqePosted.RQEAttrs);
			if ((pThisThread->TMemState == WTSTATE_BUSY) || (pThisThread->TMemState == WTSTATE_REJOIN))
			{
				pThisThread->TMemState = WTSTATE_RUNNING;
				pThreadPoolHeader->AvailableThreads++;
			}
		}
	}

	//  Close any available local resources
	ForceCloseMyMappedDBH(ThreadID);
	ForceCloseMyMappedFileHandles();

	//  Check that the transition requested is to the quit state
	if (pThisThread->TMemState == WTSTATE_QUIT)
	{
		pThisThread->TMemState = WTSTATE_QUITTING;
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2602T, ThreadID);
		xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);						//  Show the state change
	}
	else
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2603E, ThreadID, pThisThread->TMemState);
		xeLocal->LogMessage(szMsg, ThreadID);								//  Show the state change
		pThisThread->TMemState = WTSTATE_QUITTING;
	}

	NotesTermThread();
	
	//  Mark the thread as drained and terminate the thread
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2614T, ThreadID);
	xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);				//  Show the state change
	pThisThread->TMemState = WTSTATE_DRAINED;	
#ifdef WIN32
	_endthreadex(THREAD_RC_OK);									//  Terminate the thread
#endif

#ifdef UNIX
	pthread_exit(NULL);
#endif
	return THREAD_RC_OK;
}

//  StartWrapper
//
//  This function is the static wrapper for the Start() function allowing it to be started as an independent thread.
//
//  Parameters
//
//  Returns
//
//  UINT  -			Status indication  
//
//  API Use Notes:
//
//
#ifdef WIN32
UINT LNPUBLIC WorkerThread::StartWrapper(void * pObject)
#endif
#ifdef UNIX
void * LNPUBLIC WorkerThread::StartWrapper(void * pObject)
#endif
{
	WorkerThread		*wtThis;

	//  Wrapper function calls the Start function in the Object
	wtThis = (WorkerThread *) pObject;
#ifdef WIN32
	return wtThis->Start();
#endif

#ifdef UNIX
	return (void *) wtThis->Start();
#endif
}

//  ExecuteRequest
//
//  This function provides the main run function for work. Requests that were placed on the Thread Pool
//  Manager queues end up here for execution in any one of the Worker Threads.
//
//  Parameters
//
//	1  -  Threadpool_Member *		-  Pointer to the Threadpool Member entry for this thread
//	2  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//
//  Returns
//
//  API Use Notes:
//  
//
void WorkerThread::ExecuteRequest(Threadpool_Member *pThisThread, UCHAR *pTP)
{
	Runnable				*runObject;									//  The executable object
	Threadpool_Header		*pThreadPoolHeader;							//  Pointer to the Thread Pool Header
#ifdef DXIP
	PCRecRC					*pSrc;										//  Pointer to a PCRecRC structure
	PCRecRC					*pTgt;										//  Pointer to a PCRecRC structure
	int						iIndex;										//  Generic index
#endif
	char					szMsg[MAX_MSG + 1];							//  Generic message buffer

	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;

	//  Check that we have not lost the ThreadID
	if (ThreadID == 0)
	{
		pThreadPoolHeader->Guard01++;
		ThreadID = ANONYMOUS_WORKER_ID;
	}

#ifdef DXIP
	//  Increment the event counter and record the phase transition timer
	pThisThread->tToActive = xeLocal->RunningTime->getElapsedMillis();
#endif

	//  Execute the posted request
	pThisThread->TMemState = WTSTATE_BUSY;

	//  Show that the request has been accepted
	sprintf_s(szMsg, MAX_MSG, MSG_DXR2604T, ThreadID, pThisThread->rqePosted.pPObject, pThisThread->rqePosted.pXObject);
	xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);

	//  Address the executable object that will process the request
	runObject = (Runnable *) pThisThread->rqePosted.pXObject;

	//  Process the request
	try
	{
		if (runObject->RunObjectSignature != DXRUN_SIGNATURE) 
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2619E, pThisThread->rqePosted.pXObject);
			xeLocal->LogMessage(szMsg, ThreadID);
		}
		else
		{
			runObject->ExecuteThisRequest(pThisThread->rqePosted.pPObject, ThreadID);
		}
		
		//  Return the request to the rejoin pool - if needed
		if (pThisThread->rqePosted.RQEAttrs & RQATTR_REJOIN)
		{
			//  Post the current request to rejoin
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2617T, pThisThread->rqePosted.pPObject);
			xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);
			PostRequestToRejoin(pThisThread, ThreadID);
		}
		else
		{
			//  The request is "Fire and Forget" style
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2617T, pThisThread->rqePosted.pPObject);
			xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);			
		}
	}
	//  Catch Execution exceptions
	catch (DXException& dxeCurrent)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2607E, dxeCurrent.ExceptionType, dxeCurrent.ExceptionCode);
		xeLocal->LogMessage(szMsg, ThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2608E, dxeCurrent.LineNumber, dxeCurrent.OriginatingFunction);
		xeLocal->LogMessage(szMsg, ThreadID);
		if (dxeCurrent.IsPercolated) xeLocal->LogMessage(MSG_DXR2609E, ThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2610E, dxeCurrent.OriginatingFunction, dxeCurrent.LineNumber);
		xeLocal->LogMessage(szMsg, ThreadID);
		if (dxeCurrent.ExceptionType == ERRORTYPE_NOTESAPI)
		{
			xeLocal->GetAPIMessage(dxeCurrent.ExceptionCode, szMsg);
			xeLocal->LogMessage(szMsg, ThreadID);
		}
		if (dxeCurrent.LocalExceptionMessage[0] != '\0')
		{
			xeLocal->LogMessage(dxeCurrent.LocalExceptionMessage, ThreadID);
		}
	}
	//  Catch standard exceptions
	catch (std::exception& exProcess)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2605E, pThisThread->rqePosted.pPObject, pThisThread->rqePosted.pXObject);
		xeLocal->LogMessage(szMsg, ThreadID);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR2606E, exProcess.what());
		xeLocal->LogMessage(szMsg, ThreadID);
	}

#ifdef DXIP
	//  Record the Return to IDLE elapsed timer then shuffle the PC recording stack and insert the new record in the first slot
	pThisThread->tRetToIdle = xeLocal->RunningTime->getElapsedMillis();
	pTgt = (PCRecRC *) &pThisThread->CStack[0];
	pTgt = pTgt + (IPC_STACK_SIZE - 1);
	pSrc = pTgt;
	pSrc--;
	for (iIndex = 0; iIndex < (IPC_STACK_SIZE - 1); iIndex++)
	{
		memcpy(pTgt, pSrc, sizeof(PCRecRC));
	}
	pSrc->PCBase.Type = DXIPR_RQE_COMPLETED;
	pSrc->PCBase.Size = sizeof(PCRecRC);
	pSrc->PCBase.TID = ThreadID;
	pSrc->PCBase.EventNo = pThisThread->EventCount + 1;
	pSrc->PCBase.EventTime = pThisThread->tToIdle;
	pSrc->DPriority = pThisThread->rqePosted.RQEPriority;
	pSrc->Dwell = pThisThread->tToActive - pThisThread->tToIdle;
	pSrc->Mill = pThisThread->tRetToIdle - pThisThread->tToActive;
	//  Cycle the timers
	pThisThread->EventCount++;
	pThisThread->tToIdle = pThisThread->tRetToIdle;
	pThisThread->tToActive = 0;
	pThisThread->tRetToIdle = 0;
#endif
	return;
}

//  ReleaseAllMappedDBH
//
//  This function will close any locally (thread) mapped DB Handles that can be closed because
//  the native handle has been invalidated.
//
//  Parameters
//
//  1  -  int				-	ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
//
void WorkerThread::ReleaseAllMappedDBH(int iThreadID)
{
	int					iIndexX;								//  Index of native DB Handles
	int					iIndexY;								//  Index of threads
	int					iReleasedEntries;						//  Count of released thread entries
	DBHANDLE			*pMappedDBH;							//  Pointer to mapped DB Handle entries
	DBHANDLE			*pMyMappedDBH;							//  Pointer to mapped DB Handle entries
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Release any mapped file handles that have been completely freed
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;
	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;
	for (iIndexX = 0; iIndexX < MAX_DBH_CACHE; iIndexX++)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (pThreadPoolHeader->MaxThreads + 4));
		if (pMyMappedDBH[0] == (DBHANDLE) FREEHANDLE)
		{
			iReleasedEntries = 0;
			for (iIndexY = 1; iIndexY <= (pThreadPoolHeader->MaxThreads + 3); iIndexY++)
			{
				if (pMyMappedDBH[iIndexY] == NULLHANDLE) iReleasedEntries++;
			}
			if (iReleasedEntries == (pThreadPoolHeader->MaxThreads + 3)) 
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR2622T, pMyMappedDBH[0], iThreadID);
				xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
				pMyMappedDBH[0] = NULLHANDLE;
			}
		}
	}
	return;
}

//  CloseMyMappedDBH
//
//  This function will close all locally (thread) mapped DB Handles.
//
//  Parameters
//
//  1  -  int				-	ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
//
void WorkerThread::CloseMyMappedDBH(int iThreadID)
{
	int					iIndexX;								//  Index of native DB Handles
	DBHANDLE			*pMappedDBH;							//  Pointer to a mapped DB Handle
	DBHANDLE			*pMyMappedDBH;							//  Pointer to mapped DB Handle entries
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	int					iThreadIndex;							//  Thread Index
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Close any mapped file handles that belong to this thread
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;

	//  Compute the Thread Index
	switch(iThreadID)
	{
	case CONTROL_THREAD_ID:
		iThreadIndex = CONTROL_THREAD_ID;
		break;
	case MONITOR_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 1;
		break;
	case SCHEDULER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 2;
		break;
	case DISPATCHER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 3;
		break;
	default:
		iThreadIndex = iThreadID;
		break;
	}

	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;
	for (iIndexX = 0; iIndexX < MAX_DBH_CACHE; iIndexX++)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (pThreadPoolHeader->MaxThreads + 4));
		if ((pMyMappedDBH[0] == (DBHANDLE) FREEHANDLE) && (pMyMappedDBH[iThreadIndex] != NULLHANDLE))
		{
			NSFDbClose(pMyMappedDBH[iThreadIndex]);						//  Close the mapped DB handle
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2615T, pMyMappedDBH[iThreadIndex], iThreadID);
			xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
			pMyMappedDBH[iThreadIndex] = NULLHANDLE;					//  Mark the entry closed
		}
	}
	ReleaseAllMappedDBH(iThreadID);										//  Release any completely free rows
	return;
}

//  ForceCloseMyMappedDBH
//
//  This function will force closed all locally (thread) mapped DB Handles.
//
//  Parameters
//
//  1  -  int				-	ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
//
void WorkerThread::ForceCloseMyMappedDBH(int iThreadID)
{
	int					iIndexX;								//  Index of native DB Handles
	DBHANDLE			*pMappedDBH;							//  Pointer to a mapped DB Handle
	DBHANDLE			*pMyMappedDBH;							//  Pointer to mapped DB Handle entries
	int					iThreadIndex;							//  Thread Index
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Close any mapped file handles that belong to this thread
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;

	//  Compute the Thread Index
	switch(iThreadID)
	{
	case CONTROL_THREAD_ID:
		iThreadIndex = CONTROL_THREAD_ID;
		break;
	case MONITOR_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 1;
		break;
	case SCHEDULER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 2;
		break;
	case DISPATCHER_THREAD_ID:
		iThreadIndex = pThreadPoolHeader->MaxThreads + 3;
		break;
	default:
		iThreadIndex = iThreadID;
		break;
	}

	pMappedDBH = (DBHANDLE *) pThreadPoolHeader->pDBHMappingPool;
	for (iIndexX = 0; iIndexX < MAX_DBH_CACHE; iIndexX++)
	{
		pMyMappedDBH = pMappedDBH + (iIndexX * (pThreadPoolHeader->MaxThreads + 4));
		if (pMyMappedDBH[iThreadIndex] != NULLHANDLE)
		{
			NSFDbClose(pMyMappedDBH[iThreadIndex]);			//  Close the mapped DB handle
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2616T, pMyMappedDBH[iThreadIndex], iThreadID);
			xeLocal->LogTrace(TRACE_MT, szMsg, iThreadID);
			pMyMappedDBH[iThreadIndex] = NULLHANDLE;			//  Mark the entry closed
		}
	}
	ReleaseAllMappedDBH(iThreadID);										//  Release any completely free rows
	return;
}

//  PostRequestToRejoin
//
//  This function will post the current request to the Rejoin semaphore to request moving the
//  the request element into the rejoin pool
//
//  Parameters
//
//	1  -  Threadpool_Member *		-  Pointer to the current thread pool member
//  2  -  int						-  ID of the current thread 
//
//  Returns
//
//  API Use Notes:
//  
//
void WorkerThread::PostRequestToRejoin(Threadpool_Member *pThisThread, int iThreadID)
{
	DWORD					dwWaitIncrement;						//  Semaphore wait increment count

	dwWaitIncrement = 0;

	//  Wait for the Rejoin Semaphore to become free
	while (pThisThread->semRejoin != REJ_SEM_FREE)
	{
		dwWaitIncrement = 1;
		pThisThread->dwRejoinWaitQuanta++;
		SLEEP(tmpLocal->RejoinSemFreeWait);
	}
	pThisThread->dwRejoinWaits = pThisThread->dwRejoinWaits + dwWaitIncrement;
	
	//  Copy the RQE data to the thread pool member
	memcpy(&pThisThread->rqeRejoin, &pThisThread->rqePosted, sizeof(RQElement));

	//  Increment the count of dispatch requests
	pThisThread->dwRejoin++;

	//  Post the semaphore
	pThisThread->semRejoin = iThreadID;
	return;
}

//  DecrementCMLSThreadCount
//
//  This function will decrement the indicated CMLS lane counts
//
//  Parameters
//
//	1  -  Threadpool_Header *		-  Pointer to the Threadpool Header
//  2  -  UNIT						-  RQE Attributes which contain the lane
//
//  Returns
//
//
//  API Use Notes:
//
//
void WorkerThread::DecrementCMLSThreadCount(UCHAR *pTP, UINT Attrs)
{
	Threadpool_Header		*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	pThreadPoolHeader = (Threadpool_Header *) pTP;

	switch(Attrs & (RQATTR_CMLSLCONSTRAINED))
	{
	case RQATTR_CMLSL0:
		pThreadPoolHeader->CurrL0Threads--;
		break;

	case RQATTR_CMLSL1:
		pThreadPoolHeader->CurrL1Threads--;
		break;

	case RQATTR_CMLSL2:
		pThreadPoolHeader->CurrL2Threads--;
		break;
	}
	return;
}

//  ReleaseAllMappedFileHandles
//
//  This function will close any locally (thread) mapped File Handles that can be closed because
//  the native handle has been invalidated.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void WorkerThread::ReleaseAllMappedFileHandles(void)
{
	int					iIndexX;								//  Index of native DB Handles
	int					iIndexY;								//  Index of threads
	int					iReleasedEntries;						//  Count of released thread entries
#ifdef WIN32
	HANDLE				*pMappedFH;								//  Pointer to mapped File Handle entries
	HANDLE				*pMyMappedFH;							//  Pointer to mapped File Handle entries
#endif
#ifdef UNIX
	int					*pMappedFH;								//  Pointer to mapped File Handle entries
	int					*pMyMappedFH;							//  Pointer to mapped File Handle entries
#endif
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header

	//  Release any mapped file handles that have been completely freed
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;
#ifdef WIN32
	pMappedFH = (HANDLE *) pThreadPoolHeader->pFHMappingPool;
#endif
#ifdef UNIX
	pMappedFH = (int *) pThreadPoolHeader->pFHMappingPool;
#endif
	for (iIndexX = 0; iIndexX < MAX_FH_CACHE; iIndexX++)
	{
		pMyMappedFH = pMappedFH + (iIndexX * (pThreadPoolHeader->MaxThreads + 1));
#ifdef WIN32
		if (pMyMappedFH[0] == (HANDLE) FREEHANDLE)
		{
			iReleasedEntries = 0;
			for (iIndexY = 1; iIndexY < pThreadPoolHeader->MaxThreads; iIndexY++)
			{
				if (pMyMappedFH[iIndexY] == NULLHANDLE) iReleasedEntries++;
			}
			if (iReleasedEntries == pThreadPoolHeader->MaxThreads) pMyMappedFH[0] = NULLHANDLE;
		}
#endif
#ifdef UNIX
		if (pMyMappedFH[0] == (int) FREEHANDLE)
		{
			iReleasedEntries = 0;
			for (iIndexY = 1; iIndexY < pThreadPoolHeader->MaxThreads; iIndexY++)
			{
				if (pMyMappedFH[iIndexY] == NULL) iReleasedEntries++;
			}
			if (iReleasedEntries == pThreadPoolHeader->MaxThreads) pMyMappedFH[0] = NULL;
		}
#endif
	}
	return;
}

//  CloseMyMappedFileHandles
//
//  This function will close all locally (thread) mapped File Handles.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void WorkerThread::CloseMyMappedFileHandles(void)
{
	int					iIndexX;								//  Index of native DB Handles
#ifdef WIN32
	HANDLE				*pMappedFH;								//  Pointer to mapped File Handle entries
	HANDLE				*pMyMappedFH;							//  Pointer to mapped File Handle entries
#endif
#ifdef UNIX
	int					*pMappedFH;								//  Pointer to mapped File Handle entries
	int					*pMyMappedFH;							//  Pointer to mapped File Handle entries
#endif
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Close any mapped file handles that belong to this thread
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;
#ifdef WIN32
	pMappedFH = (HANDLE *) pThreadPoolHeader->pFHMappingPool;
#endif
#ifdef UNIX
	pMappedFH = (int *) pThreadPoolHeader->pFHMappingPool;
#endif
	for (iIndexX = 0; iIndexX < MAX_FH_CACHE; iIndexX++)
	{
		pMyMappedFH = pMappedFH + (iIndexX * (pThreadPoolHeader->MaxThreads + 1));
#ifdef WIN32
		if ((pMyMappedFH[0] == (HANDLE) FREEHANDLE) && (pMyMappedFH[ThreadID] != NULLHANDLE))
		{
			CloseHandle(pMyMappedFH[ThreadID]);				//  Close the mapped File handle
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2620T, pMyMappedFH[ThreadID], ThreadID);
			xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);
			pMyMappedFH[ThreadID] = NULLHANDLE;				//  Mark the entry closed
		}
#endif
#ifdef UNIX
		if ((pMyMappedFH[0] == (int) FREEHANDLE) && (pMyMappedFH[ThreadID] != NULL))
		{
			close(pMyMappedFH[ThreadID]);				//  Close the mapped File handle
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2620T, pMyMappedFH[ThreadID], ThreadID);
			xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);
			pMyMappedFH[ThreadID] = NULL;				//  Mark the entry closed
		}
#endif
	}
	ReleaseAllMappedFileHandles();										//  Release any completely free rows
	return;
}

//  ForceCloseMyMappedFileHandles
//
//  This function will force closed all locally (thread) mapped File Handles.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void WorkerThread::ForceCloseMyMappedFileHandles(void)
{
	int					iIndexX;								//  Index of native DB Handles
#ifdef WIN32
	HANDLE				*pMappedFH;								//  Pointer to mapped File Handle entries
	HANDLE				*pMyMappedFH;							//  Pointer to mapped File Handle entries
#endif
#ifdef UNIX
	int					*pMappedFH;								//  Pointer to mapped File Handle entries
	int					*pMyMappedFH;							//  Pointer to mapped File Handle entries
#endif
	Threadpool_Header	*pThreadPoolHeader;						//  Pointer to the Thread Pool Header
	char				szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Close any mapped file handles that belong to this thread
	pThreadPoolHeader = (Threadpool_Header *) pThreadPool;
#ifdef WIN32
	pMappedFH = (HANDLE *) pThreadPoolHeader->pFHMappingPool;
#endif
#ifdef UNIX
	pMappedFH = (int *) pThreadPoolHeader->pFHMappingPool;
#endif
	for (iIndexX = 0; iIndexX < MAX_FH_CACHE; iIndexX++)
	{
		pMyMappedFH = pMappedFH + (iIndexX * (pThreadPoolHeader->MaxThreads + 1));
#ifdef WIN32
		if (pMyMappedFH[ThreadID] != NULLHANDLE)
		{
			CloseHandle(pMyMappedFH[ThreadID]);							//  Close the mapped File handle
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2620T, pMyMappedFH[ThreadID], ThreadID);
			xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);
			pMyMappedFH[ThreadID] = NULLHANDLE;							//  Mark the entry closed
		}
#endif
#ifdef UNIX
		if (pMyMappedFH[ThreadID] != NULL)
		{
			close(pMyMappedFH[ThreadID]);				//  Close the mapped File handle
			//  Trace the action
			sprintf_s(szMsg, MAX_MSG, MSG_DXR2620T, pMyMappedFH[ThreadID], ThreadID);
			xeLocal->LogTrace(TRACE_MT, szMsg, ThreadID);
			pMyMappedFH[ThreadID] = NULL;				//  Mark the entry closed
		}
#endif
	}
	ReleaseAllMappedFileHandles();										//  Release any completely free rows
	return;
}

