#pragma once
//  WorkerThread
//
//  This class provides the wrapper functions for executing requests for work.
//  
//
//  Message Identifier:      DXR
//  Message Range:           2601 - 2799
//  Trace area assignment:   TRACE_MT (02)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  1.0.0		07/06/2003		Initial version
//  3.5.3		21/08/2009		Re-based ready for 4.0.0
//  3.6.0		18/09/2009		New log handler for multithreaded applications
//  3.6.1		10/10/2009		Additional controls added
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.9.0		24/07/2010		Linkage changed for logging & request posting and rejoining
//								A CommandHandler can be hooked into the monitor processing
//  3.10.2		18/11/2010		MTIO support added to DbMover
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.12.0		21/12/2011		DXIP Introduced
//  3.14.1		20/01/2015		x64 compatability
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2003 - 2015
//

//  Platform includes
#include "../Platform/PlatBase.h"

//  Application Includes
#include "../ExecEnvironment.h"								//  Runtime
#include "ThreadStructs.h"									//  Thread Structures
#include "ThreadManagerPolicy.h"							//  Thread Manager Policy
#include "../DXException.h"									//  DX Exceptions
#include "Runnable.h"										//  Runnable interface

//  Local Definitions
#define		IDLE_WAITINTERVAL		100						//  Wait 100 Ms while in the idle state
#define		ANONYMOUS_WORKER_ID		800						//  Thread ID for an anonymous worker thread

//  Trace Messages
#define		MSG_DXR2600T			"DXR2600T: Worker thread[%i] has entered the not started state."
#define		MSG_DXR2601T			"DXR2601T: Worker thread[%i] has entered the running state."
#define		MSG_DXR2602T			"DXR2602T: Worker thread[%i] has entered the quitting state."
#define		MSG_DXR2604T			"DXR2604T: Worker thread[%i] has accepted request @%p for processing by @%p."
#define		MSG_DXR2611T			"DXR2611T: A new WorkerThread instance has been created @%p."
#define		MSG_DXR2612T			"DXR2612T: A WorkerThread instance @%p has been destroyed."
#define		MSG_DXR2614T			"DXR2614T: Worker thread[%i] has been drained."
#define		MSG_DXR2617T			"DXR2617T: The current RQE has been allocated in the Rejoin pool for Parm Object %p."
#define		MSG_DXR2618T			"DXR2618T: The current RQE is Fire & Forget for Parm Object %p."
#define		MSG_DXR2620T			"DXR2620T: Closed local File handle %p by thread %i."
#define		MSG_DXR2621T			"DXR2621T: Forced closed local File handle %p by thread %i."

#ifdef W64
#define		MSG_DXR2615T			"DXR2615T: Closed local DB handle %i by thread %i."
#define		MSG_DXR2616T			"DXR2616T: Forced closed local DB handle %i by thread %i."
#define		MSG_DXR2622T			"DXR2622T: Released native DB handle %i by thread %i."
#else
#define		MSG_DXR2615T			"DXR2615T: Closed local DB handle %p by thread %i."
#define		MSG_DXR2616T			"DXR2616T: Forced closed local DB handle %p by thread %i."
#define		MSG_DXR2622T			"DXR2622T: Released native DB handle %p by thread %i."
#endif

//  Information Messages

//  Error Messages
#define		MSG_DXR2603E			"DXR2603E: Worker thread[%i] has detected an invalid state change to %u and is quitting."
#define		MSG_DXR2605E			"DXR2605E: An execption has been thrown while executing request @%p by object @%p."
#define		MSG_DXR2606E			"DXR2606E: %s."
#define		MSG_DXR2607E			"DXR2607E: Explorer exception type: %i, code: %i has been detected."
#define		MSG_DXR2608E			"DXR2608E: The exception occurred at line: %i in module: %s."
#define		MSG_DXR2609E			"DXR2609E: The exception was percolated from a lower level scanner."
#define		MSG_DXR2610E			"DXR2610E: The execption originated in: %s at: %d."
#define		MSG_DXR2613E			"DXR2613E: The thread was unable to initialise the Notes runtime."
#define		MSG_DXR2619E			"DXR2619E: The execution object @%p does not have a valid run signature, execution skipped."

class WorkerThread
{
public:
	//  Constructor
	WorkerThread(ExecEnvironment *xeParent, UCHAR *pTP, ThreadManagerPolicy *tmpParent, int iThreadID);

	//  Destructor
	~WorkerThread(void);

	//  Public Members
	UCHAR				*pThreadPool;						//  Pointer to the Thread Pool
	int					ThreadID;							//  Thread ID

	//  Public Functions
#ifdef WIN32
	static UINT LNPUBLIC StartWrapper(void * pObject);		//  Wrapper for the Start worker function
#endif

#ifdef UNIX
	static void * LNPUBLIC StartWrapper(void * pObject);		//  Wrapper for the Start worker function
#endif

	UINT Start(void);										//  Start the Worker

private:
	//  Private Members
	ExecEnvironment			*xeLocal;						//  Runtime
	ThreadManagerPolicy		*tmpLocal;						//  Thread Manager Policy

	//  Private Functions

	//  Request Execution Functions
	void ExecuteRequest(Threadpool_Member *pThisThread, UCHAR *pTP);		//  Execute posted request
	void PostRequestToRejoin(Threadpool_Member *pThisThread, int iThreadID);		//  Post the current request to the rejoin pool

	//  Thread Local Database Handle Functions
	void ReleaseAllMappedDBH(int iThreadID);				//  Release any freed DBH mapping entries
	void CloseMyMappedDBH(int iThreadID);					//  Close any locally mapped DB handles
	void ForceCloseMyMappedDBH(int iThreadID);				//  Force close of all locall mapped DB Handles

	//  Thread Local File Handle Functions
	void ReleaseAllMappedFileHandles(void);					//  Release any freed file handle mapping entries
	void CloseMyMappedFileHandles(void);					//  Close any locally mapped file handles
	void ForceCloseMyMappedFileHandles(void);				//  Force close of all locall mapped DB Handles

	//  Constrained Multi-Lane Scheduling Functions
	void DecrementCMLSThreadCount(UCHAR *pTP, UINT Attrs);	//  Increment the CMLS Lane counts
};
