//  DbCopier
//
//  This class provides the Database Copy functions used in Domono eXplorer (DX) utilities
//
//  Message Identifier:      DXR
//  Message Range:           5001 - 5499
//  Trace area assignment:   TRACE_DBC (23)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  Build: 109
//
//  Version History:
//
//  1.0.0		10/02/2001		Initial version
//  3.5.3		11/09/2009		Incorporated into DX core
//  3.6.0		20/09/2009		Prepared for QCopy 2.0
//  3.6.1		10/10/2009		Additional controls added
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.7.1		06/01/2010		Prevent a race condition in Rejoin processing
//  3.7.2		12/01/2010		Improve handling of corrupt folders
//  3.7.3		06/02/2010		DbMover & DbCopier improvements
//  3.7.5		30/03/2010		Improved Error Handling in DbCopier
//  3.7.6		08/04/2010		Fixed a problem with error handling
//  3.7.7		10/04/2010		Additional recovery options in DbCopier
//  3.9.0		24/07/2010		Linkage changed for logging & request posting
//								Improved error recovery
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.3		05/01/2011		Pre-Processing actions added
//  3.10.9		07/06/2011		One time replication added
//								Attitional error recovery
//  3.10.10		14/07/2011		Performance Improvements for/to DbCopier
//  3.11.1		22/10/2011		VS 2011 Updates
//  3.11.6		17/12/2011		Elapsed time calculations changed to use ms
//	3.13.0		20/12/2013		Advanced Target Database Properties
//	3.14.0		19/01/2015		x64 compatability fixes
//	3.14.1		24/02/2015		Avoid a panic in DbCopier
//								Recompress Attachments optional
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2001 - 2015
//

//  Application Includes
#include "DbCopier.h"

//  Default Constructor
DbCopier::DbCopier(void)
{
}

//  Constructor
DbCopier::DbCopier(MTExecutive *xeParent, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Save the Runtime Environment
	xeLocal = xeParent;
	//  Save the Creating Thread ID
	OwningThread = iThreadID;
	MultiplexScale = 100;

	//  Trace the construction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5001T, this);
	xeLocal->LogTrace(TRACE_DBC, szMsg, OwningThread);
}

//  Destructor
DbCopier::~DbCopier(void)
{
	char			szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Trace destruction
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5111T, this);
	xeLocal->LogTrace(TRACE_DBC, szMsg, OwningThread);
}

//  CopyThisDb
//
//  This is the synchronous entry point to request copy processing.
//
//  Parameters
//
//  1  -  CopyRequest *		-  Pointer to the CopyRequest for this database
//  2  -  int				-  Thread ID of the requesting thread
//
//  Returns
//
//  BOOL  -			TRUE if the database was processed (copied) successfully
//                  FALSE if the database could not be copied
//
//  API Use Notes:
//
//
BOOL DbCopier::CopyThisDb(CopyRequest *crRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	PartCopyRequest		*pcrqCopyDesign;					//  Design Copy Request
	PartCopyRequest		*pcrqCopyData;						//  Data Copy Request
	PartCopyRequest		*pcrqCopyNotes;						//  Copy notes request
	PartCopyRequest		*pcrqFolders;						//  Folder Maintenance Request
	PartCopyRequest		*pcrqUpgrade;						//  Upgrade Design Request
	PartCopyRequest		*pcrqBuildIndexes;					//  Build View Indexes Request
	PartCopyRequest		*pcrqFTI;							//  Copy the Full Text Index Request
	PartCopyRequest		*pcrqRunAgent;						//  Run the requested agent
	NOTEID				nidAgent;							//  NoteID of the conditioning agent
	WORD				wOpenFlags;							//  Flags to use for open of the source database
	WORD				wDbClass;							//  Database class for target database
	BYTE				Encrypt;							//  Local encrypt strength for the target database
	BOOL				bCopyRecovered;						//  Flag showing that copy was recovered by replication
	DWORD				dwFoldersToMaintain;				//  Folders to be maintained
	clock_t				ct_Start;							//  Copy Start Timestamp
	clock_t				ct_End;								//  Copy End Timestamp
	clock_t				ct_PhaseStart;						//  Phase Start Timestamp
	clock_t				ct_PhaseEnd;						//  Phase End Timestamp
	int					iRejoinStatus;						//  Rejoin request status
	int					iRejoinWaits;						//  Rejoin wait counter
	long				lElapsed;							//  Elapsed Time
	long				lThroughput;						//  Throughput total
	long				lTimeFactor = 1000;					//  Clock tick conversion factor
	char				szLT[2];							//  Qualifier for zero second timing
	char				szGT[2];							//  Qualifier for zero second timing
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	iRejoinWaits = 0;
	bCopyRecovered = FALSE;
	crRQ->bAllowProxyFolderMessages = TRUE;

	//  If the trasaction is being retried then perform any requested restart actions
	if (crRQ->RetryCount > 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5168I, crRQ->RetryCount);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		if (crRQ->RestartActions != 0)
		{
			if (PerformRestartActions(crRQ, iThreadID)) crRQ->RestartActions = 0;
		}
	}
	else
	{
		//  Determine if there any any Pre-Processing action to be done
		if (crRQ->PreProcessActions != 0)
		{
			if (PerformPreProcessActions(crRQ, iThreadID))
			{
				ProxyLogMessage(crRQ, MSG_DXR5178I, iThreadID);
			}
		}
	}

	//  Database Copier Module - report the request
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5002I, crRQ->szSourceDatabase, crRQ->szSourceServer);
	ProxyLogMessage(crRQ, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5003I, crRQ->szTargetDatabase, crRQ->szTargetServer);
	ProxyLogMessage(crRQ, szMsg, iThreadID);

	//  Record the start time
	ct_Start = xeLocal->RunningTime->getElapsedMillis();
	ct_PhaseStart = xeLocal->RunningTime->getElapsedMillis();
	
	//  Open the source database - with a fixup if requested
	wOpenFlags = 0;
	//  If a fixup is required then set the fixup flags
	if (crRQ->ForceFixup) wOpenFlags = DBOPEN_FORCE_FIXUP | DBOPEN_FIXUP_FULL_NOTE_SCAN;
	crRQ->hdbSource = xeLocal->OpenDatabase(crRQ->szSourceServer, crRQ->szSourceDatabase, wOpenFlags, &crRQ->dbidSource, 0);
	if (crRQ->hdbSource == NULLHANDLE)
	{
		ProxyLogMessage(crRQ, MSG_DXR5004E, iThreadID);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		//  If the source database did not exist then do not allow retry
		if (xeLocal->GetLastOpenError(iThreadID) == ERR_NOEXIST) crRQ->AllowRetry = FALSE;
		return FALSE;
	}

	ProxyLogMessage(crRQ, MSG_DXR5007I, iThreadID);

	//  Remove the Force Fixup flag so that if the transaction is retried then fixup will
	//  only be performed if requested.
	crRQ->ForceFixup = FALSE;

	//  Verify the replicaID of the source database
	if (!VerifyReplicaID(crRQ->hdbSource, crRQ, iThreadID))
	{
		ProxyLogMessage(crRQ, MSG_DXR5139E, iThreadID);
		NSFDbClose(crRQ->hdbSource);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		//  Turn off the permit retries for this error
		crRQ->AllowRetry = FALSE;
		return FALSE;
	}

	//  Datermine the class and encryption strength for the target database
	switch(crRQ->ODSLevel)
	{
	case ODS_DFLT:
		wDbClass = DBCLASS_NOTEFILE;
		break;

	case ODS_R6:
		wDbClass = DBCLASS_V6NOTEFILE;
		break;

	case ODS_R8:
		wDbClass = DBCLASS_V8NOTEFILE;
		break;

	case ODS_R85:
		wDbClass = DBCLASS_V85NOTEFILE;
		break;

	case ODS_R9:
		wDbClass = DBCLASS_V9NOTEFILE;
		break;

	default:
		wDbClass = DBCLASS_NOTEFILE;
	}

	//  If the class is not specified (or defaults) then take the class from the source database
	if (wDbClass == DBCLASS_NOTEFILE)
	{
		stAPIRC = NSFDbClassGet(crRQ->hdbSource, &wDbClass);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5198W, crRQ->szSourceDatabase, crRQ->szSourceServer);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
			wDbClass = DBCLASS_NOTEFILE;	
		}
	}

	//  Set the local database encryption strength
	switch(crRQ->LocalEncrypt)
	{
	case ENCRYPT_NONE:
		Encrypt = DBCREATE_ENCRYPT_NONE;
		break;

	case ENCRYPT_SIMPLE:
		Encrypt = DBCREATE_ENCRYPT_SIMPLE;
		break;

	case ENCRYPT_MEDIUM:
		Encrypt = DBCREATE_ENCRYPT_MEDIUM;
		break;

	case ENCRYPT_STRONG:
		Encrypt = DBCREATE_ENCRYPT_STRONG;
		break;

	default:
		Encrypt = DBCREATE_ENCRYPT_NONE;
	}

	//  Create a replica or non-replica target database
	if (crRQ->CopyReplica)
	{
		crRQ->hdbTarget = xeLocal->CreateReplica(crRQ->szTargetServer, 
											crRQ->szTargetDatabase, 
											crRQ->EnableTXLogging, 
											crRQ->hdbSource, 
											&crRQ->dbidTarget,
											wDbClass,
											Encrypt,
											iThreadID);
		if (crRQ->hdbTarget == NULLHANDLE)
		{
			NSFDbClose(crRQ->hdbSource);
			ProxyLogMessage(crRQ, MSG_DXR5005E, iThreadID);
			crRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}
	}
	else
	{
		crRQ->hdbTarget = xeLocal->CreateDatabase(crRQ->szTargetServer, 
											crRQ->szTargetDatabase, 
											crRQ->EnableTXLogging, 
											&crRQ->dbidTarget,
											wDbClass,
											Encrypt,
											iThreadID);
		if (crRQ->hdbTarget == NULLHANDLE)
		{
			NSFDbClose(crRQ->hdbSource);
			ProxyLogMessage(crRQ, MSG_DXR5006E, iThreadID);
			crRQ->ReturnCode = RETURN_INCOMPLETE;
			return FALSE;
		}
	}

	//  Set the restart actions to delete the database
	crRQ->RestartActions = RESTART_DEL_TARGET;

	ProxyLogMessage(crRQ, MSG_DXR5008I, iThreadID);

	//  Modify the Target Database ACL according to the passed settings
	if (crRQ->szGroupToAdd[0] != '\0')
	{
		ConditionACL(crRQ->szGroupToAdd, crRQ, iThreadID);
		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5113T);
	}

	//  Prepare the target database
	if (!prepareTargetDb(crRQ, crRQ->hdbSource, crRQ->hdbTarget, iThreadID))
	{
		ProxyLogMessage(crRQ, MSG_DXR5199E, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		NSFDbClose(crRQ->hdbSource);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		return FALSE;
	}

	//  If background agents are to be disabled in the target then do this now
	if (crRQ->DisableAgents)
	{
		DisableAgents(crRQ, crRQ->hdbTarget, iThreadID);
	}

	//  Check that permits are still in place
	if ((!xeLocal->TestGlobalPermit()) || (!crRQ->PermitRun))
	{
		ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		NSFDbClose(crRQ->hdbSource);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		return FALSE;
	}

	//  Copy processing
	xeLocal->LogTrace(TRACE_DBC, MSG_DXR5114T);

	stAPIRC = IDCreateTable(sizeof(NOTEID), &crRQ->hitFolderTable);		//  Initialise the Folder Table
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5010E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		NSFDbClose(crRQ->hdbSource);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		return FALSE;
	}

	//  Initiate copying of the database design
	pcrqCopyDesign = new PartCopyRequest(crRQ);
	pcrqCopyDesign->CopyAction = COPY_DESIGN;						//  Copy All Design
	pcrqCopyDesign->CompletionStatus = ACTION_INPROGRESS;			//  Mark the design copying in progress
	//  Initiate copying of the data
	pcrqCopyData = new PartCopyRequest(crRQ);
	pcrqCopyData->CopyAction = COPY_DATA;							//  Copy All Data
	pcrqCopyData->CompletionStatus = ACTION_INPROGRESS;				//  Mark the data copying in progress
	pcrqCopyDesign->pcrqCopyDesign = pcrqCopyDesign;
	pcrqCopyDesign->pcrqCopyData = pcrqCopyData;
	pcrqCopyData->pcrqCopyDesign = pcrqCopyDesign;
	pcrqCopyData->pcrqCopyData = pcrqCopyData;

	//  Dispatch requests to start the design & data copy processes

	if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
								this,
								this,
								pcrqCopyDesign,
								RQATTR_FIREANDFORGET | RQATTR_MEGAPRODUCER,
								20,
								iThreadID))
	{
		//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
		ProxyLogMessage(crRQ, MSG_DXR5104E, iThreadID);			
		crRQ->PermitRun = FALSE;
		DrainRequests(crRQ, iThreadID);

		//  Invalidate & Close the source and target databases
		xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
		NSFDbClose(crRQ->hdbSource);
		IDDestroyTable(crRQ->hitFolderTable);
		ProxyLogMessage(crRQ, MSG_DXR5009I, iThreadID);
		crRQ->ReturnCode = RETURN_CERR;
		crRQ->PermitRun = FALSE;
		return FALSE;
	}

	xeLocal->LogTrace(TRACE_DBC, MSG_DXR5115T);

	if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
								this,
								this,
								pcrqCopyData,
								RQATTR_FIREANDFORGET | RQATTR_MEGAPRODUCER,
								20,
								iThreadID))
	{
		//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
		ProxyLogMessage(crRQ, MSG_DXR5104E, iThreadID);			
		crRQ->PermitRun = FALSE;
		DrainRequests(crRQ, iThreadID);

		//  Invalidate & Close the source and target databases
		xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
		NSFDbClose(crRQ->hdbSource);
		IDDestroyTable(crRQ->hitFolderTable);
		ProxyLogMessage(crRQ, MSG_DXR5009I, iThreadID);
		crRQ->ReturnCode = RETURN_CERR;
		crRQ->PermitRun = FALSE;
		return FALSE;
	}

	xeLocal->LogTrace(TRACE_DBC, MSG_DXR5116T);

	//  Wait on the completion of both the Design & the Data processes
	while ((pcrqCopyDesign->CompletionStatus == ACTION_INPROGRESS) || (pcrqCopyData->CompletionStatus == ACTION_INPROGRESS))
	{
		//  Process Rejoins for any sub level requests that might have completed
		while (xeLocal->GetRejoinRequest(crRQ, (void **) &pcrqCopyNotes, iThreadID) == RJR_RETURNED)
		{
			//  Accumulate the results into the top level request
			if (pcrqCopyNotes->CopyAction == COPY_DESIGN_NOTES)
			{
				if (xeLocal->AcquireAppMutex(DESIGN_ACC_MUTEX, iThreadID))
				{
					crRQ->dwDesignDocsCopied += pcrqCopyNotes->dwNotesCopied;
					crRQ->dwDesignStubsCopied += pcrqCopyNotes->dwStubsCopied;
					crRQ->dwDesignAccessErrors += pcrqCopyNotes->dwCopyAccessErrors;
					crRQ->dwDesignOtherErrors += pcrqCopyNotes->dwCopyOtherErrors;
					crRQ->dwDesignRetries += pcrqCopyNotes->dwCopyRetries;
					xeLocal->FreeAppMutex(DESIGN_ACC_MUTEX, iThreadID);
				}
			}
			else
			{
				if (pcrqCopyNotes->CopyAction == COPY_DATA_NOTES)
				{
					if (xeLocal->AcquireAppMutex(DATA_ACC_MUTEX, iThreadID))
					{
						crRQ->dwDataDocsCopied += pcrqCopyNotes->dwNotesCopied;
						crRQ->dwDataStubsCopied += pcrqCopyNotes->dwStubsCopied;
						crRQ->dwDataAccessErrors += pcrqCopyNotes->dwCopyAccessErrors;
						crRQ->dwDataOtherErrors += pcrqCopyNotes->dwCopyOtherErrors;
						crRQ->dwDataRetries += pcrqCopyNotes->dwCopyRetries;
						xeLocal->FreeAppMutex(DATA_ACC_MUTEX, iThreadID);
					}
				}
				else
				{
					//  The returned object is not of a valid type - record the error and clear the pointer
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5147E, pcrqCopyNotes, pcrqCopyNotes->CopyAction);
					ProxyLogMessage(crRQ, szMsg, iThreadID);
					pcrqCopyNotes = NULL;
				}
			}

			//  Discard the returned object
			if (pcrqCopyNotes != NULL)
			{
				//  Check that no non-recovered errors were encountered - if so then abort the copy request
				if (pcrqCopyNotes->dwCopyAccessErrors > 0)
				{
					if (!(crRQ->FaultTolerance & ALLOW_ACCESS_ERRORS))
					{
						//  Abort the copy request
						pcrqCopyNotes->CopyAction = COPY_FREED;
						delete pcrqCopyNotes;
						pcrqCopyNotes = NULL;

						crRQ->PermitRun = FALSE;
						DrainRequests(crRQ, iThreadID);

						//  Mark the transaction incomplete, clean up and exit
						//  Invalidate & Close the source and target databases
						xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
						NSFDbClose(crRQ->hdbTarget);
						xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
						NSFDbClose(crRQ->hdbSource);
						IDDestroyTable(crRQ->hitFolderTable);
						ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
						crRQ->ReturnCode = RETURN_CERR;
						return FALSE;
					}
				}
				if (pcrqCopyNotes->dwCopyOtherErrors > 0)
				{
					if (!(crRQ->FaultTolerance & ALLOW_COPY_ERRORS))
					{
						//  Abort the copy request
						pcrqCopyNotes->CopyAction = COPY_FREED;
						delete pcrqCopyNotes;
						pcrqCopyNotes = NULL;

						crRQ->PermitRun = FALSE;
						DrainRequests(crRQ, iThreadID);

						//  Mark the transaction incomplete, clean up and exit
						//  Invalidate & Close the source and target databases
						xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
						NSFDbClose(crRQ->hdbTarget);
						xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
						NSFDbClose(crRQ->hdbSource);
						IDDestroyTable(crRQ->hitFolderTable);
						ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
						crRQ->ReturnCode = RETURN_CERR;
						return FALSE;
					}
				}
				//  Update the action in memory in case a race causes the freed object to be returned to another thread
				pcrqCopyNotes->CopyAction = COPY_FREED;
				delete pcrqCopyNotes;
				pcrqCopyNotes = NULL;	
			}
		}

		xeLocal->AppWait(PXR_BUSY_WAIT);
	}

	//  If Abort has been signalled then remove the Run Permit flag in the top level request
	//  and perform a drain sequence

	if ((pcrqCopyDesign->CompletionStatus == ACTION_ABORT) || (pcrqCopyData->CompletionStatus == ACTION_ABORT))
	{
		ProxyLogMessage(crRQ, MSG_DXR5170E, iThreadID);

		crRQ->PermitRun = FALSE;
		DrainRequests(crRQ, iThreadID);

		//  Mark the transaction incomplete, clean up and exit
		//  Invalidate & Close the source and target databases
		xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
		IDDestroyTable(crRQ->hitFolderTable);
		NSFDbClose(crRQ->hdbSource);
		ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
		crRQ->ReturnCode = RETURN_CERR;
		return FALSE;
	}

	//  Check for an indication in the Data Copy request that it found an error that it suspects
	//  can be recovered from by using a one-time replication to bring across the data.
	//  If allowed to attempt recovery then do so otherwise fail the transaction.
	if (pcrqCopyData->CompletionStatus == ACTION_RECOVER)
	{
		if ((crRQ->FaultTolerance & RECOVER_RSRC_ERRORS) && (crRQ->CopyReplica))
		{
			//  Attempt recovery by performing a one-time replication
			ReenableReplication(crRQ, iThreadID);
			if (!PerformReplicateDb(crRQ, iThreadID))
			{
				ProxyLogMessage(crRQ, MSG_DXR5194E, iThreadID);

				crRQ->PermitRun = FALSE;
				DrainRequests(crRQ, iThreadID);

				//  Mark the transaction incomplete, clean up and exit
				//  Invalidate & Close the source and target databases
				xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
				NSFDbClose(crRQ->hdbTarget);
				xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
				IDDestroyTable(crRQ->hitFolderTable);
				NSFDbClose(crRQ->hdbSource);
				ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
				crRQ->ReturnCode = RETURN_CERR;
				return FALSE;
			}
			bCopyRecovered = TRUE;
			ProxyLogMessage(crRQ, MSG_DXR5193I, iThreadID);
			DisableReplication(crRQ, iThreadID);

			//  If a one-time replication was requested then cancel it
			crRQ->OneTimeReplicate = FALSE;
		}
		else
		{
			ProxyLogMessage(crRQ, MSG_DXR5191E, iThreadID);

			crRQ->PermitRun = FALSE;
			DrainRequests(crRQ, iThreadID);

			//  Mark the transaction incomplete, clean up and exit
			//  Invalidate & Close the source and target databases
			xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
			NSFDbClose(crRQ->hdbTarget);
			xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
			IDDestroyTable(crRQ->hitFolderTable);
			NSFDbClose(crRQ->hdbSource);
			ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
			crRQ->ReturnCode = RETURN_CERR;
			return FALSE;
		}
	}

	xeLocal->LogTrace(TRACE_DBC, MSG_DXR5117T);

	//  Record any totals accumulated in the high level requests
	if (xeLocal->AcquireAppMutex(DESIGN_ACC_MUTEX, iThreadID))
	{
		crRQ->dwDesignDocsCopied += pcrqCopyDesign->dwNotesCopied;
		crRQ->dwDesignStubsCopied += pcrqCopyDesign->dwStubsCopied;
		crRQ->dwDesignAccessErrors += pcrqCopyDesign->dwCopyAccessErrors;
		crRQ->dwDesignOtherErrors += pcrqCopyDesign->dwCopyOtherErrors;
		crRQ->dwDesignRetries += pcrqCopyDesign->dwCopyRetries;
		xeLocal->FreeAppMutex(DESIGN_ACC_MUTEX, iThreadID);
	}

	if (xeLocal->AcquireAppMutex(DATA_ACC_MUTEX, iThreadID))
	{
		crRQ->dwDataDocsCopied += pcrqCopyData->dwNotesCopied;
		crRQ->dwDataStubsCopied += pcrqCopyData->dwStubsCopied;
		crRQ->dwDataAccessErrors += pcrqCopyData->dwCopyAccessErrors;
		crRQ->dwDataOtherErrors += pcrqCopyData->dwCopyOtherErrors;
		crRQ->dwDataRetries += pcrqCopyData->dwCopyRetries;
		xeLocal->FreeAppMutex(DATA_ACC_MUTEX, iThreadID);
	}

	iRejoinWaits = 0;

	iRejoinStatus = xeLocal->GetRejoinRequest(crRQ, (void **) &pcrqCopyNotes, iThreadID);
	while (iRejoinStatus != RJR_NONE_EXIST)
	{
		if (iRejoinStatus == RJR_RETURNED)
		{
			iRejoinWaits = 0;
			//  Accumulate the results into the top level request
			if (pcrqCopyNotes->CopyAction == COPY_DESIGN_NOTES)
			{
				if (xeLocal->AcquireAppMutex(DESIGN_ACC_MUTEX, iThreadID))
				{
					crRQ->dwDesignDocsCopied += pcrqCopyNotes->dwNotesCopied;
					crRQ->dwDesignStubsCopied += pcrqCopyNotes->dwStubsCopied;
					crRQ->dwDesignAccessErrors += pcrqCopyNotes->dwCopyAccessErrors;
					crRQ->dwDesignOtherErrors += pcrqCopyNotes->dwCopyOtherErrors;
					crRQ->dwDesignRetries += pcrqCopyNotes->dwCopyRetries;
					xeLocal->FreeAppMutex(DESIGN_ACC_MUTEX, iThreadID);
				}
			}
			else
			{
				if (pcrqCopyNotes->CopyAction == COPY_DATA_NOTES)
				{
					if (xeLocal->AcquireAppMutex(DATA_ACC_MUTEX, iThreadID))
					{
						crRQ->dwDataDocsCopied += pcrqCopyNotes->dwNotesCopied;
						crRQ->dwDataStubsCopied += pcrqCopyNotes->dwStubsCopied;
						crRQ->dwDataAccessErrors += pcrqCopyNotes->dwCopyAccessErrors;
						crRQ->dwDataOtherErrors += pcrqCopyNotes->dwCopyOtherErrors;
						crRQ->dwDataRetries += pcrqCopyNotes->dwCopyRetries;
						xeLocal->FreeAppMutex(DATA_ACC_MUTEX, iThreadID);
					}
				}
				else
				{
					//  The returned object is not of a valid type - record the error and clear the pointer
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5147E, pcrqCopyNotes, pcrqCopyNotes->CopyAction);
					ProxyLogMessage(crRQ, szMsg, iThreadID);
					pcrqCopyNotes = NULL;
				}
			}
			//  Clean up the request if it is valid
			if (pcrqCopyNotes != NULL)
			{
				//  Check that no non-recovered errors were encountered - if so then abort the copy request
				if (pcrqCopyNotes->dwCopyAccessErrors > 0)
				{
					if (!(crRQ->FaultTolerance & ALLOW_ACCESS_ERRORS))
					{
						//  Abort the copy request
						pcrqCopyNotes->CopyAction = COPY_FREED;
						delete pcrqCopyNotes;
						pcrqCopyNotes = NULL;

						crRQ->PermitRun = FALSE;
						DrainRequests(crRQ, iThreadID);

						//  Mark the transaction incomplete, clean up and exit
						//  Invalidate & Close the source and target databases
						xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
						NSFDbClose(crRQ->hdbTarget);
						xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
						NSFDbClose(crRQ->hdbSource);
						IDDestroyTable(crRQ->hitFolderTable);
						ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
						crRQ->ReturnCode = RETURN_CERR;
						return FALSE;
					}
				}
				if (pcrqCopyNotes->dwCopyOtherErrors > 0)
				{
					if (!(crRQ->FaultTolerance & ALLOW_COPY_ERRORS))
					{
						//  Abort the copy request
						pcrqCopyNotes->CopyAction = COPY_FREED;
						delete pcrqCopyNotes;
						pcrqCopyNotes = NULL;

						crRQ->PermitRun = FALSE;
						DrainRequests(crRQ, iThreadID);

						//  Mark the transaction incomplete, clean up and exit
						//  Invalidate & Close the source and target databases
						xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
						NSFDbClose(crRQ->hdbTarget);
						xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
						NSFDbClose(crRQ->hdbSource);
						IDDestroyTable(crRQ->hitFolderTable);
						ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
						crRQ->ReturnCode = RETURN_CERR;
						return FALSE;
					}
				}
				//  Update the action in memory in case a race causes the freed object to be returned to another thread
				pcrqCopyNotes->CopyAction = COPY_FREED;
				delete pcrqCopyNotes;
				pcrqCopyNotes = NULL;	
			}
			//  Check if we have too many folders to log to the proxy interface - if so suppress further logging
			if (crRQ->bAllowProxyFolderMessages)
			{
				dwFoldersToMaintain = IDEntries(crRQ->hitFolderTable);
				if (dwFoldersToMaintain > MAX_FOLDER_MSGS) 
				{
					ProxyLogMessage(crRQ, MSG_DXR5176I, iThreadID);
					crRQ->bAllowProxyFolderMessages = FALSE;
				}
			}
		}
		else
		{
			iRejoinWaits++;
			xeLocal->AppWait(PXR_BUSY_WAIT);
		}
		iRejoinStatus = xeLocal->GetRejoinRequest(crRQ, (void **) &pcrqCopyNotes, iThreadID);
	}

	xeLocal->LogTrace(TRACE_DBC, MSG_DXR5118T);

	//  Mark the end time of the copy phase
	ct_PhaseEnd = xeLocal->RunningTime->getElapsedMillis();
	crRQ->dwCopyPhaseElapsed = ComputeElapsed(&ct_PhaseStart, &ct_PhaseEnd);

	//  Show the completion statistics from both design & data copying
	crRQ->dwSizeKB = xeLocal->GetDbAllocatedSize(crRQ->hdbTarget, iThreadID);
	ShowCopyStats(crRQ, iThreadID);

	delete pcrqCopyDesign;
	delete pcrqCopyData;


	//  Check that permits are still in place
	if ((!xeLocal->TestGlobalPermit()) || (!crRQ->PermitRun))
	{
		ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
		xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
		NSFDbClose(crRQ->hdbSource);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		return FALSE;
	}

	//  If folder maintenance is needed then dispatch a request to perform it
	ct_PhaseStart = xeLocal->RunningTime->getElapsedMillis();

	pcrqFolders = new PartCopyRequest(crRQ);
	pcrqFTI = new PartCopyRequest(crRQ);
	if ((crRQ->CopyReplica) || (crRQ->CopyFolders))
	{
		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5119T);
		dwFoldersToMaintain = IDEntries(crRQ->hitFolderTable);
		if ((dwFoldersToMaintain == 0) || (bCopyRecovered))
		{
			ProxyLogMessage(crRQ, MSG_DXR5036I, iThreadID);
		}
		else
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5037I, dwFoldersToMaintain);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
			pcrqFolders->CopyAction = COPY_MAINTAIN_FOLDERS;
			pcrqFolders->CompletionStatus = ACTION_INPROGRESS;
			if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
										this,
										this,
										pcrqFolders,
										RQATTR_FIREANDFORGET | RQATTR_PRODUCER,
										20,
										iThreadID))
			{
				//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
				ProxyLogMessage(crRQ, MSG_DXR5107E, iThreadID);			
				crRQ->PermitRun = FALSE;
				DrainRequests(crRQ, iThreadID);

				//  Invalidate & Close the source and target databases
				xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
				NSFDbClose(crRQ->hdbTarget);
				xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
				NSFDbClose(crRQ->hdbSource);
				IDDestroyTable(crRQ->hitFolderTable);
				ProxyLogMessage(crRQ, MSG_DXR5009I, iThreadID);
				crRQ->ReturnCode = RETURN_CERR;
				crRQ->PermitRun = FALSE;
				return FALSE;
			}
			xeLocal->LogTrace(TRACE_DBC, MSG_DXR5120T);
		}
	}
	else pcrqFolders->CompletionStatus = ACTION_COMPLETED;

	//  Check that permits are still in place
	if ((!xeLocal->TestGlobalPermit()) || (!crRQ->PermitRun))
	{
		ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
		xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
		NSFDbClose(crRQ->hdbSource);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		return FALSE;
	}

	//  If LZ1 recompression was indicated in the request then Peek the request queues until there
	//  are no more transactions intended for the attachment processor.
	if (crRQ->LZ1Compress && crRQ->RecompressAttachments)
	{
		//  Peek the request queues until there are no requests for the attachment owner
		while (xeLocal->PeekQueues(ATT_PROC_OWNER, iThreadID) > 0) xeLocal->AppWait(PXR_BUSY_WAIT);

		//  Show the statistics from the re-compression
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5217I, crRQ->dwNotesRecom, crRQ->dwFilesRecom, crRQ->dwKBSaved);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
	}

	//  If requested dispatch a request to copy the FTI definition from source to target database (and build the Index)
	if ((crRQ->CopyFTI) || (crRQ->CopyAndBuildFTI))
	{
		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5121T);
		pcrqFTI->CopyAction = COPY_FTI;
		pcrqFTI->CompletionStatus = ACTION_INPROGRESS;
		if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
									this,
									this,
									pcrqFTI,
									RQATTR_FIREANDFORGET,
									20,
									iThreadID))
		{
			//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
			ProxyLogMessage(crRQ, MSG_DXR5108E, iThreadID);			
			crRQ->PermitRun = FALSE;
			DrainRequests(crRQ, iThreadID);

			//  Invalidate & Close the source and target databases
			xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
			NSFDbClose(crRQ->hdbTarget);
			xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
			NSFDbClose(crRQ->hdbSource);
			IDDestroyTable(crRQ->hitFolderTable);
			ProxyLogMessage(crRQ, MSG_DXR5009I, iThreadID);
			crRQ->ReturnCode = RETURN_CERR;
			return FALSE;
		}
		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5122T);
	}
	else pcrqFTI->CompletionStatus = ACTION_COMPLETED;

	//  Wait for the Folder maintenance and the FTI Index build processes to complete
	while ((pcrqFolders->CompletionStatus == ACTION_INPROGRESS) || (pcrqFTI->CompletionStatus == ACTION_INPROGRESS))
	{
		xeLocal->AppWait(PXR_BUSY_WAIT);
	}

	ct_PhaseEnd = xeLocal->RunningTime->getElapsedMillis();
	crRQ->dwFolderPhaseElapsed = ComputeElapsed(&ct_PhaseStart, &ct_PhaseEnd);

	xeLocal->LogTrace(TRACE_DBC, MSG_DXR5123T);

	if ((crRQ->CopyReplica) || (crRQ->CopyFolders))
	{
		if (!bCopyRecovered)
		{
			//  Move the totals to the copy request
			crRQ->dwFoldersCopied = pcrqFolders->dwStubsCopied;
			crRQ->dwFolderDocs = pcrqFolders->dwNotesCopied;

			if (pcrqFolders->CompletionStatus == ACTION_COMPLETED) 
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5038I, crRQ->dwFolderPhaseElapsed);
				ProxyLogMessage(crRQ, szMsg, iThreadID);
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5041I, pcrqFolders->dwNotesCopied, pcrqFolders->dwStubsCopied);
				ProxyLogMessage(crRQ, szMsg, iThreadID);
			}
			else 
			{
				if (crRQ->FaultTolerance & ALLOW_FOLDER_ERRORS) ProxyLogMessage(crRQ, MSG_DXR5039W, iThreadID);
				else
				{
					//  Invalidate & Close the source and target databases
					xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
					NSFDbClose(crRQ->hdbTarget);
					xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
					NSFDbClose(crRQ->hdbSource);
					IDDestroyTable(crRQ->hitFolderTable);
					ProxyLogMessage(crRQ, MSG_DXR5190E, iThreadID);
					crRQ->ReturnCode = RETURN_CERR;
					return FALSE;
				}
			}
		}
	}

	//  Delete the Folder maintenance and FTI copy/build part copy requests
	delete pcrqFolders;
	pcrqFolders = NULL;
	delete pcrqFTI;
	pcrqFTI = NULL;

	//  Destroy the Folder ID Table
	IDDestroyTable(crRQ->hitFolderTable);

	ct_PhaseStart = xeLocal->RunningTime->getElapsedMillis();

	//  Check that permits are still in place
	if ((!xeLocal->TestGlobalPermit()) || (!crRQ->PermitRun))
	{
		ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
		xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
		NSFDbClose(crRQ->hdbSource);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		return FALSE;
	}

	//  If required perform Design Refresh/Replace as needed
	if (crRQ->RefreshDesign)
	{
		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5124T);
		pcrqUpgrade = new PartCopyRequest(crRQ);
		pcrqUpgrade->CopyAction = UPGRADE_DESIGN;
		pcrqUpgrade->CompletionStatus = ACTION_INPROGRESS;
		if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
									this,
									this,
									pcrqUpgrade,
									RQATTR_FIREANDFORGET,
									20,
									iThreadID))
		{
			//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
			ProxyLogMessage(crRQ, MSG_DXR5109E, iThreadID);			
			crRQ->PermitRun = FALSE;
			DrainRequests(crRQ, iThreadID);

			//  Invalidate & Close the source and target databases
			xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
			NSFDbClose(crRQ->hdbTarget);
			xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
			NSFDbClose(crRQ->hdbSource);
			ProxyLogMessage(crRQ, MSG_DXR5009I, iThreadID);
			crRQ->ReturnCode = RETURN_CERR;
			return FALSE;
		}

		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5125T);

		//  Wait for the request to complete
		while (pcrqUpgrade->CompletionStatus == ACTION_INPROGRESS)
		{
			xeLocal->AppWait(PXR_BUSY_WAIT);
		}

		ct_PhaseEnd = xeLocal->RunningTime->getElapsedMillis();
		crRQ->dwDesignPhaseElapsed = ComputeElapsed(&ct_PhaseStart, &ct_PhaseEnd);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5136I, crRQ->dwDesignPhaseElapsed);
		ProxyLogMessage(crRQ, szMsg, iThreadID);

		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5126T);
		delete pcrqUpgrade;
	}

	ct_PhaseStart = xeLocal->RunningTime->getElapsedMillis();

	//  Check that permits are still in place
	if ((!xeLocal->TestGlobalPermit()) || (!crRQ->PermitRun))
	{
		ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
		xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
		NSFDbClose(crRQ->hdbSource);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		return FALSE;
	}

	//  If required run the specified agent in the target database
	if (crRQ->szAgentName[0] != '\0')
	{
		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5127T);
		//  Find the NoteID of the specified agent
		stAPIRC = NIFFindDesignNote(crRQ->hdbTarget, crRQ->szAgentName, NOTE_CLASS_FILTER, &nidAgent);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5095E, crRQ->szAgentName);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
		else
		{
			pcrqRunAgent = new PartCopyRequest(crRQ);
			pcrqRunAgent->CopyAction = RUN_AGENT;
			pcrqRunAgent->nidToCopy[0] = nidAgent;
			pcrqRunAgent->CopyNoteCount = 1;
			if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
										this,
										this,
										pcrqRunAgent,
										RQATTR_FIREANDFORGET,
										20,
										iThreadID))
			{
				//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
				ProxyLogMessage(crRQ, MSG_DXR5110E, iThreadID);			
				crRQ->PermitRun = FALSE;
				DrainRequests(crRQ, iThreadID);

				//  Invalidate & Close the source and target databases
				xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
				NSFDbClose(crRQ->hdbTarget);
				xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
				NSFDbClose(crRQ->hdbSource);
				ProxyLogMessage(crRQ, MSG_DXR5009I, iThreadID);
				crRQ->ReturnCode = RETURN_CERR;
				return FALSE;
			}

			xeLocal->LogTrace(TRACE_DBC, MSG_DXR5128T);

			//  Wait for the request to complete
			while (pcrqRunAgent->CompletionStatus == ACTION_INPROGRESS)
			{
				xeLocal->AppWait(PXR_BUSY_WAIT);
			}

			ct_PhaseEnd = xeLocal->RunningTime->getElapsedMillis();
			crRQ->dwAgentPhaseElapsed = ComputeElapsed(&ct_PhaseStart, &ct_PhaseEnd);
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5137I, crRQ->dwAgentPhaseElapsed);
			ProxyLogMessage(crRQ, szMsg, iThreadID);

			xeLocal->LogTrace(TRACE_DBC, MSG_DXR5129T);
			delete pcrqRunAgent;
		}
	}

	ct_PhaseStart = xeLocal->RunningTime->getElapsedMillis();

	//  Check that permits are still in place
	if ((!xeLocal->TestGlobalPermit()) || (!crRQ->PermitRun))
	{
		ProxyLogMessage(crRQ, MSG_DXR5171E, iThreadID);
		xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
		NSFDbClose(crRQ->hdbTarget);
		xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
		NSFDbClose(crRQ->hdbSource);
		crRQ->ReturnCode = RETURN_INCOMPLETE;
		return FALSE;
	}

	//  If requested then build all of the view indexes in the target database
	if (crRQ->BuildIndexes)
	{
		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5130T);
		pcrqBuildIndexes = new PartCopyRequest(crRQ);
		pcrqBuildIndexes->CopyAction = COPY_BUILD_INDEXES;
		PerformCopyAction(pcrqBuildIndexes, iThreadID);					//  Initiate the copy action
		ct_PhaseEnd = xeLocal->RunningTime->getElapsedMillis();
		crRQ->dwIndexPhaseElapsed = ComputeElapsed(&ct_PhaseStart, &ct_PhaseEnd);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5138I, crRQ->dwIndexPhaseElapsed);
		ProxyLogMessage(crRQ, szMsg, iThreadID);

		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5131T);
		delete pcrqBuildIndexes;
	}

	//  If the target database is a replica copy then re-enable replication
	if (crRQ->CopyReplica) ReenableReplication(crRQ, iThreadID);

	//  Get the final size of the target database
	crRQ->dwSizeKB = xeLocal->GetDbAllocatedSize(crRQ->hdbTarget, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5143I, crRQ->dwSizeKB);
	ProxyLogMessage(crRQ, szMsg, iThreadID);

	//  Trace the final dispopsition of the database flags and options
	reportTargetDb(crRQ->hdbTarget, iThreadID);

	//  Invalidate & Close the source and target databases
	xeLocal->InvalidateNativeDBH(crRQ->hdbTarget, iThreadID);
	NSFDbClose(crRQ->hdbTarget);
	xeLocal->InvalidateNativeDBH(crRQ->hdbSource, iThreadID);
	NSFDbClose(crRQ->hdbSource);
	ProxyLogMessage(crRQ, MSG_DXR5009I, iThreadID);

	//  If requested then perform a one time replication of the database
	if (crRQ->OneTimeReplicate)
	{
		ct_PhaseStart = xeLocal->RunningTime->getElapsedMillis();
		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5187T);
		if (!PerformReplicateDb(crRQ, iThreadID))
		{
			crRQ->ReturnCode = RETURN_CERR;
			return FALSE;
		}
		ct_PhaseEnd = xeLocal->RunningTime->getElapsedMillis();
		crRQ->dwReplPhaseElapsed = ComputeElapsed(&ct_PhaseStart, &ct_PhaseEnd);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5188I, crRQ->dwReplPhaseElapsed);
		ProxyLogMessage(crRQ, szMsg, iThreadID);

		xeLocal->LogTrace(TRACE_DBC, MSG_DXR5189T);
	}

	//  Record the ending time
	ct_End = xeLocal->RunningTime->getElapsedMillis();

	//  Document the run statistics
	lElapsed = (ct_End - ct_Start) / lTimeFactor;

	//  Defend against zero divides when total elapsed time is less than a second
	szLT[0] = ' ';
	szGT[0] = ' ';
	szLT[1] = '\0';
	szGT[1] = '\0';
	if (lElapsed == 0)
	{
		szLT[0] = '<';
		szGT[0] = '>';
		lElapsed = 1;
	}

	crRQ->dwTotalElapsed = (DWORD) lElapsed;

	//  Compute and display the throughput in documents per second
	lThroughput = (crRQ->dwDataDocsCopied + crRQ->dwDataStubsCopied + crRQ->dwDesignDocsCopied + crRQ->dwDesignStubsCopied) / crRQ->dwCopyPhaseElapsed;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5060I, szLT, lElapsed, szGT, lThroughput);
	ProxyLogMessage(crRQ, szMsg, iThreadID);

	//  Compute and display the throughput in Kb per second
	lThroughput = crRQ->dwSizeKB / lElapsed;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5061I, szLT, lElapsed, szGT, lThroughput);
	ProxyLogMessage(crRQ, szMsg, iThreadID);

	crRQ->ReturnCode = RETURN_NOERROR;
	return TRUE;
}

//  ExecuteThisRequest
//
//  This is the entry point used by the multi-threading kernel to trigger asynchronous processing in different threadpool threads.
//  This is a wrapper for the PerformCopyAction() function.
//
//  Parameters
//
//  1  -  void *			-  Pointer to the parameter object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::ExecuteThisRequest(void *pReqObject, int iThreadID)
{
	PartCopyRequest		*pcrCurrent;								//  Dispatched Request
	char				szMsg[MAX_MSG + 1];							//  Generic Message Buffer

	//  Cast the Part Copy Request
	pcrCurrent = (PartCopyRequest *) pReqObject;

	//  Trace
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5177T, pcrCurrent, pcrCurrent->CopyAction, pcrCurrent->Parent);
	xeLocal->LogTrace(TRACE_DBC, szMsg, iThreadID);

	//  If this is a Copy Database request - then route the contained request to the synchronous entry point
	if (pcrCurrent->CopyAction == COPY_DATABASE) CopyThisDb(pcrCurrent->Parent, iThreadID);
	else PerformCopyAction(pcrCurrent, iThreadID);
	return;
}

//  PerformCopyAction
//
//  This is the main routing function for all multi-threaded invocations, the function will vector the current
//  request to the appropriate executing functions determined by the Copy Action defined in the request.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformCopyAction(PartCopyRequest * pcrqRQ, int iThreadID)
{
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	if (pcrqRQ == NULL) return;

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5112T, pcrqRQ->CopyAction);
	xeLocal->LogTrace(TRACE_DBC, szMsg, iThreadID);

	//  Check that the permits are still in place
	if ((!xeLocal->TestGlobalPermit()) || (!pcrqRQ->Parent->PermitRun)) 
	{
		pcrqRQ->CompletionStatus = ACTION_ABORT;
		pcrqRQ->Parent->PermitRun = FALSE;
		return;
	}

	//  Vector according to the action requested
	switch(pcrqRQ->CopyAction)
	{
	case COPY_DESIGN:
		PerformCopyDesign(pcrqRQ, iThreadID);
		break;
	case COPY_DATA:
		PerformCopyData(pcrqRQ, iThreadID);
		break;
	case COPY_DESIGN_NOTES:
		PerformCopyNotes(pcrqRQ, iThreadID);
		break;
	case COPY_DATA_NOTES:
		PerformCopyNotes(pcrqRQ, iThreadID);
		break;
	case CHECK_RECOMPRESS:
		PerformCRA(pcrqRQ, iThreadID);
		break;
	case RECOMPRESS_NOTE:
		PerformRecompressInPlace(pcrqRQ, iThreadID);
		break;
	case COPY_MAINTAIN_FOLDERS:
		PerformMaintainFolders(pcrqRQ, iThreadID);
		break;
	case COPY_MAINTAIN_FOLDER:
		PerformMaintainFolder(pcrqRQ, iThreadID);
		break;
	case UPGRADE_DESIGN:
		PerformUpgradeDesign(pcrqRQ, iThreadID);
		break;
	case COPY_BUILD_INDEXES:
		PerformBuildIndexes(pcrqRQ, iThreadID);
		break;
	case COPY_BUILD_INDEX:
		PerformBuildIndex(pcrqRQ, iThreadID);
		break;
	case COPY_FTI:
		PerformCopyFTI(pcrqRQ, iThreadID);
		break;
	case RUN_AGENT:
		PerformRunAgent(pcrqRQ, iThreadID);
		break;
	}
	return;
}

//  PerformCopyDesign
//
//  This function will copy all of the design notes (and profiles) from the source database to the
//  target database. The function uses the generic PerformCopyNotes() asynchronous routine to actually
//  perform the copy.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformCopyDesign(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	int					iWarningsIssued = 0;				//  Number of warnings issued
	DBHANDLE			hdbTLSource;						//  Local Thread handle for the Source database
    TIMEDATE			tdStart;							//  Modified since time
	TIMEDATE			tdFinish;							//  Ending modification time
	DWORD				dwNotesToCopy;						//  Count of notes to copy
	DWORD				dwTotalNotesToCopy;					//  Total Count of notes and profiles to copy
	DWORD				dwScanned;							//  Count of notes returned from the collection
	DWORD				dwFoldersToMaintain;				//  Folder count 
	WORD				wNoteClass;							//  Note types to be retrieved
	BOOL				bGetFirst;							//  IDScan control flag
	NOTEID				nidRetNID;							//  NoteID returned from IDScan
	PartCopyRequest		*pcrqCopyNotes;						//  Request to copy Notes
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5011I, iThreadID);

	//  Get the mapped thread local database handles
	hdbTLSource = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbSource, iThreadID);

	//  Determine which NOTE classes must be copied
	if (pcrqRQ->Parent->CopyPrivates) wNoteClass = NOTE_CLASS_ALLNONDATA ^ NOTE_CLASS_ACL;
	else wNoteClass = (NOTE_CLASS_ALLNONDATA ^ NOTE_CLASS_PRIVATE) ^ NOTE_CLASS_ACL;

	//  Suppress copying the ICON note as this is done while conditioning the target database
	wNoteClass = wNoteClass ^ NOTE_CLASS_ICON;

	TimeConstant(TIMEDATE_WILDCARD, &tdStart);

	//  Get the Modified Note table for all Design Notes - modified up till now
	stAPIRC = NSFDbGetModifiedNoteTable(hdbTLSource, wNoteClass, tdStart, &tdFinish, &pcrqRQ->Parent->hitDesignTable);
	if (ERR(stAPIRC) != NOERROR)
	{
		if (ERR(stAPIRC) == ERR_NO_MODIFIED_NOTES)
		{
			ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5013E, iThreadID);			
			pcrqRQ->CompletionStatus = ACTION_FAILED;
			return;
		}
		else
		{
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5014E, iThreadID);			
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);			
		}

		//  Signal an abort
		pcrqRQ->CompletionStatus = ACTION_ABORT;
		pcrqRQ->Parent->PermitRun = FALSE;
		return;
	}
	
	//  Get the count of notes in the table
	dwNotesToCopy = IDEntries(pcrqRQ->Parent->hitDesignTable);

	//  Add all profile documents to the design note ID table
	pcrqRQ->dbcCopier = this;
	stAPIRC = NSFProfileEnum(hdbTLSource, NULL, 0, WrapperAddProfileToIDT, pcrqRQ, 0);
	if (ERR(stAPIRC) != NOERROR)
	{
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5015E, iThreadID);			
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);			
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5016W, iThreadID);
		iWarningsIssued++;
	}

	//  Get the count of notes and profiles in the table
	dwTotalNotesToCopy = IDEntries(pcrqRQ->Parent->hitDesignTable);

	//  Show the counts
	if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
	{
		dwNotesToCopy = dwTotalNotesToCopy - dwNotesToCopy;
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5017I, dwNotesToCopy);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);			
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5018I, dwTotalNotesToCopy);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);			
	}

	// Copy the design notes
	dwScanned = 0;
	bGetFirst = TRUE;
	pcrqCopyNotes = NULL;

	while (IDScan(pcrqRQ->Parent->hitDesignTable, bGetFirst, &nidRetNID))
	{
		//  Check that the Run Permit flag is still set
		if ((!pcrqRQ->Parent->PermitRun) || (!xeLocal->TestGlobalPermit()))
		{
			pcrqRQ->CompletionStatus = ACTION_FAILED;
			IDDestroyTable(pcrqRQ->Parent->hitDesignTable);
			return;
		}
		bGetFirst = FALSE;
		dwScanned++;

		//  If the copy request does not exist then create it
		if (pcrqCopyNotes == NULL) 
		{
			pcrqCopyNotes = new PartCopyRequest(pcrqRQ->Parent);
			pcrqCopyNotes->CopyAction = COPY_DESIGN_NOTES;
		}

		// Add the current noteid to the copy request
		pcrqCopyNotes->nidToCopy[pcrqCopyNotes->CopyNoteCount] = nidRetNID;
		pcrqCopyNotes->CopyNoteCount++;

		//  Check if the request is full - if so dispatch it
		if (pcrqCopyNotes->CopyNoteCount == COPY_NOTE_MULTIPLEX)
		{
			//  Post a request to execute the copy action
			if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
										pcrqRQ->Parent,
										this,
										pcrqCopyNotes,
										RQATTR_REJOIN,
										15,
										iThreadID))
			{
				//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
				ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5104E, iThreadID);			
				pcrqRQ->CompletionStatus = ACTION_FAILED;
				IDDestroyTable(pcrqRQ->Parent->hitDesignTable);
				//  Signal an abort
				pcrqRQ->CompletionStatus = ACTION_ABORT;
				pcrqRQ->Parent->PermitRun = FALSE;
				return;
			}

			//  Retrieve any completed requests - this is necessary to prevent the RQE Rejoin pool from flooding
			while(xeLocal->GetRejoinRequest(pcrqRQ->Parent, (void **) &pcrqCopyNotes, iThreadID) == RJR_RETURNED)
			{
				//  Accumulate the results
				if (pcrqCopyNotes->CopyAction == COPY_DESIGN_NOTES) 
				{
					if (xeLocal->AcquireAppMutex(DESIGN_ACC_MUTEX, iThreadID))
					{
						pcrqRQ->pcrqCopyDesign->AccumulateTotals(pcrqCopyNotes);
						xeLocal->FreeAppMutex(DESIGN_ACC_MUTEX, iThreadID);
					}
				}
				else 
				{
					if (pcrqCopyNotes->CopyAction == COPY_DATA_NOTES)
					{
						if (xeLocal->AcquireAppMutex(DATA_ACC_MUTEX, iThreadID))
						{
							pcrqRQ->pcrqCopyData->AccumulateTotals(pcrqCopyNotes);
							xeLocal->FreeAppMutex(DATA_ACC_MUTEX, iThreadID);
						}
					}
					else
					{
						//  The returned object is not of a valid type - record the error and clear the pointer
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5147E, pcrqCopyNotes, pcrqCopyNotes->CopyAction);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						pcrqCopyNotes = NULL;
					}
				}
				//  Clean up the request if it is valid
				if (pcrqCopyNotes != NULL)
				{
					//  Check if there were any non-recovered errors noted in the request if so abort the copy
					if (pcrqCopyNotes->dwCopyAccessErrors > 0)
					{
						if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_ACCESS_ERRORS))
						{
							pcrqCopyNotes->CopyAction = COPY_FREED;
							delete pcrqCopyNotes;
							pcrqCopyNotes = NULL;	
							//  Signal an abort
							pcrqRQ->CompletionStatus = ACTION_ABORT;
							pcrqRQ->Parent->PermitRun = FALSE;
							IDDestroyTable(pcrqRQ->Parent->hitDesignTable);
							return;
						}
					}
					if (pcrqCopyNotes->dwCopyOtherErrors > 0)
					{
						if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_COPY_ERRORS))
						{
							pcrqCopyNotes->CopyAction = COPY_FREED;
							delete pcrqCopyNotes;
							pcrqCopyNotes = NULL;	
							//  Signal an abort
							pcrqRQ->CompletionStatus = ACTION_ABORT;
							pcrqRQ->Parent->PermitRun = FALSE;
							IDDestroyTable(pcrqRQ->Parent->hitDesignTable);
							return;
						}
					}
					//  Update the action in memory in case a race causes the freed object to be returned to another thread
					pcrqCopyNotes->CopyAction = COPY_FREED;
					delete pcrqCopyNotes;
					pcrqCopyNotes = NULL;	
				}
			}
			pcrqCopyNotes = NULL;
			//  Check if we have too many folders to log to the proxy interface - if so suppress further logging
			if (pcrqRQ->Parent->bAllowProxyFolderMessages)
			{
				dwFoldersToMaintain = IDEntries(pcrqRQ->Parent->hitFolderTable);
				if (dwFoldersToMaintain > MAX_FOLDER_MSGS) 
				{
					ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5176I, iThreadID);
					pcrqRQ->Parent->bAllowProxyFolderMessages = FALSE;
				}
			}
		}
	}

	//  If there is a request constructed then dispatch it
	if (pcrqCopyNotes != NULL)
	{
		//  Post a request to execute the copy action
		if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
									pcrqRQ->Parent,
									this,
									pcrqCopyNotes,
									RQATTR_REJOIN,
									15,
									iThreadID))
		{
			//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
			ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5104E, iThreadID);			
			pcrqRQ->CompletionStatus = ACTION_FAILED;
			//  Signal an abort
			pcrqRQ->CompletionStatus = ACTION_ABORT;
			pcrqRQ->Parent->PermitRun = FALSE;
			IDDestroyTable(pcrqRQ->Parent->hitDesignTable);
			return;
		}
	}

	//  Retrieve any completed requests - this is necessary to prevent the RQE Rejoin pool from flooding
	while(xeLocal->GetRejoinRequest(pcrqRQ->Parent, (void **) &pcrqCopyNotes, iThreadID) == RJR_RETURNED)
	{
		//  Accumulate the results
		if (pcrqCopyNotes->CopyAction == COPY_DESIGN_NOTES) 
		{
			if (xeLocal->AcquireAppMutex(DESIGN_ACC_MUTEX, iThreadID))
			{
				pcrqRQ->pcrqCopyDesign->AccumulateTotals(pcrqCopyNotes);
				xeLocal->FreeAppMutex(DESIGN_ACC_MUTEX, iThreadID);
			}
		}
		else 
		{
			if (pcrqCopyNotes->CopyAction == COPY_DATA_NOTES)
			{
				if (xeLocal->AcquireAppMutex(DATA_ACC_MUTEX, iThreadID))
				{
					pcrqRQ->pcrqCopyData->AccumulateTotals(pcrqCopyNotes);
					xeLocal->FreeAppMutex(DATA_ACC_MUTEX, iThreadID);
				}
			}
			else
			{
				//  The returned object is not of a valid type - record the error and clear the pointer
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5147E, pcrqCopyNotes, pcrqCopyNotes->CopyAction);
				ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
				pcrqCopyNotes = NULL;
			}
		}
		//  Clean up the request if it is valid
		if (pcrqCopyNotes != NULL)
		{
			//  Check if there were any non-recovered errors noted in the request if so abort the copy
			if (pcrqCopyNotes->dwCopyAccessErrors > 0)
			{
				if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_ACCESS_ERRORS))
				{
					pcrqCopyNotes->CopyAction = COPY_FREED;
					delete pcrqCopyNotes;
					pcrqCopyNotes = NULL;	
					//  Signal an abort
					pcrqRQ->CompletionStatus = ACTION_ABORT;
					pcrqRQ->Parent->PermitRun = FALSE;
					IDDestroyTable(pcrqRQ->Parent->hitDesignTable);
					return;
				}
			}
			if (pcrqCopyNotes->dwCopyOtherErrors > 0)
			{
				if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_COPY_ERRORS))
				{
					pcrqCopyNotes->CopyAction = COPY_FREED;
					delete pcrqCopyNotes;
					pcrqCopyNotes = NULL;	
					//  Signal an abort
					pcrqRQ->CompletionStatus = ACTION_ABORT;
					pcrqRQ->Parent->PermitRun = FALSE;
					IDDestroyTable(pcrqRQ->Parent->hitDesignTable);
					return;
				}
			}
			//  Update the action in memory in case a race causes the freed object to be returned to another thread
			pcrqCopyNotes->CopyAction = COPY_FREED;
			delete pcrqCopyNotes;
			pcrqCopyNotes = NULL;	
		}
		//  Check if we have too many folders to log to the proxy interface - if so suppress further logging
		if (pcrqRQ->Parent->bAllowProxyFolderMessages)
		{
			dwFoldersToMaintain = IDEntries(pcrqRQ->Parent->hitFolderTable);
			if (dwFoldersToMaintain > MAX_FOLDER_MSGS) 
			{
				ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5176I, iThreadID);
				pcrqRQ->Parent->bAllowProxyFolderMessages = FALSE;
			}
		}
	}

	//  Exit - design copying is complete
	if (iWarningsIssued > 0) pcrqRQ->CompletionStatus = ACTION_WARN;
	else pcrqRQ->CompletionStatus = ACTION_COMPLETED;
	IDDestroyTable(pcrqRQ->Parent->hitDesignTable);
	return;
}

//  PerformCopyData
//
//  This function will copy all of the data notes from the source database to the
//  target database. The function uses the generic PerformCopyNotes() asynchronous routine to actually
//  perform the copy.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformCopyData(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	int					iWarningsIssued = 0;				//  Number of warnings issued
	DBHANDLE			hdbTLSource;						//  Local Thread handle for the Source database
	DHANDLE				hitIDTable;							//  Handle of the Note ID Table
    TIMEDATE			tdStart;							//  Modified since time
	TIMEDATE			tdFinish;							//  Ending modification time
	DWORD				dwTotalNotesToCopy;					//  Total Count of notes and profiles to copy
	DWORD				dwSizeKb;							//  Db Size in Kb
	int					iRawMultiplexor;					//  Raw Max notes per request
	int					iMultiplexor;						//  Max notes per request
	DWORD				dwScanned;							//  Count of notes returned from the collection
	DWORD				dwFoldersToMaintain;				//  Count of folders
	WORD				wNoteClass;							//  Note types to be retrieved
	BOOL				bGetFirst;							//  IDScan control flag
	NOTEID				nidRetNID;							//  NoteID returned from IDScan
	PartCopyRequest		*pcrqCopyNotes;						//  Request to copy Notes
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5012I, iThreadID);

	//  Get the mapped thread local database handles
	hdbTLSource = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbSource, iThreadID);

	//  Determine which NOTE classes must be copied
	wNoteClass = NOTE_CLASS_DATA;

	TimeConstant(TIMEDATE_WILDCARD, &tdStart);

	//  Get the Modified Note table for all Data Notes - modified up till now
	stAPIRC = NSFDbGetModifiedNoteTable(hdbTLSource, wNoteClass, tdStart, &tdFinish, &hitIDTable);
	if (ERR(stAPIRC) != NOERROR)
	{
		if (ERR(stAPIRC) == ERR_NO_MODIFIED_NOTES)
		{
			ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5019I, iThreadID);			
			pcrqRQ->CompletionStatus = ACTION_COMPLETED;
			return;
		}
		else
		{
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5020E, iThreadID);			
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);	
			if ((ERR(stAPIRC) == ERR_SPARE_FULL) || (ERR(stAPIRC) == ERR_SRV_LENGTH_MISMATCH))
			{
				//  These specific errors are memory resource problems caused by trying to handle
				//  very large Modified Note tables in the source database. This type of error
				//  is potentially recoverable by using NSFSearch to process the building
				//  of the Notes copy requests. If that fails then we signal to the higher level
				//  processor that recovery may be attempted by using a one-time replication.
				pcrqRQ->dbcCopier = this;
				pcrqRQ->CopyNoteCount = 0;
				pcrqRQ->iProcThread = iThreadID;
				if (!CopyDataUsingSearch(pcrqRQ, iThreadID))
				{
					xeLocal->GetAPIMessage(stAPIRC, szMsg);
					ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5020E, iThreadID);			
					ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);	
					pcrqRQ->Parent->PermitRun = FALSE;
					pcrqRQ->CompletionStatus = ACTION_RECOVER;
					return;
				}
				else
				{
					//  Copy (initiation) has completed - mark the request completed
					pcrqRQ->CompletionStatus = ACTION_COMPLETED;
					return;
				}
			}
			else
			{
				pcrqRQ->Parent->PermitRun = FALSE;
				pcrqRQ->CompletionStatus = ACTION_ABORT;
				return;
			}
		}
	}
	
	//  Get the count of notes in the table
	dwTotalNotesToCopy = IDEntries(hitIDTable);

	//  Show the counts
	if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5021I, dwTotalNotesToCopy);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);			
	}

	//  Get the D/B size and compute the multiplexor (load) per request
	dwSizeKb = xeLocal->GetDbAllocatedSize(hdbTLSource, iThreadID);
	iRawMultiplexor = dwSizeKb / dwTotalNotesToCopy;
	if (iRawMultiplexor <= 1) iRawMultiplexor = COPY_NOTE_MULTIPLEX;
	else
	{
		if (iRawMultiplexor > (COPY_NOTE_MULTIPLEX - COPY_MIN_MULTIPLEX)) iRawMultiplexor = COPY_NOTE_MULTIPLEX - COPY_MIN_MULTIPLEX;
		iRawMultiplexor = COPY_NOTE_MULTIPLEX - iRawMultiplexor;
	}

	//  Scale the raw multiplexor value by the supplied factor
	iMultiplexor = (iRawMultiplexor * MultiplexScale) / 100;
	if (iMultiplexor < COPY_MIN_MULTIPLEX) iMultiplexor = COPY_MIN_MULTIPLEX;

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5197I, iMultiplexor, iRawMultiplexor, xeLocal->GetThreadCount());
	ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);			

	// Copy the data notes
	dwScanned = 0;
	bGetFirst = TRUE;
	pcrqCopyNotes = NULL;

	while (IDScan(hitIDTable, bGetFirst, &nidRetNID))
	{
		//  Check that the Run Permit flag is still set
		if (!pcrqRQ->Parent->PermitRun)
		{
			pcrqRQ->CompletionStatus = ACTION_FAILED;
			IDDestroyTable(hitIDTable);
			return;
		}

		bGetFirst = FALSE;
		dwScanned++;

		//  If the copy request does not exist then create it
		if (pcrqCopyNotes == NULL) 
		{
			pcrqCopyNotes = new PartCopyRequest(pcrqRQ->Parent);
			pcrqCopyNotes->CopyAction = COPY_DATA_NOTES;
		}

		// Add the current noteid to the copy request
		pcrqCopyNotes->nidToCopy[pcrqCopyNotes->CopyNoteCount] = nidRetNID;
		pcrqCopyNotes->CopyNoteCount++;

		//  Check if the request is full - if so dispatch it
		if (pcrqCopyNotes->CopyNoteCount == iMultiplexor)
		{
			//  Post a request to execute the copy action
			if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
										pcrqRQ->Parent,
										this,
										pcrqCopyNotes,
										RQATTR_REJOIN,
										10,
										iThreadID))
			{
				//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
				ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5105E, iThreadID);			
				//  Signal an abort
				pcrqRQ->Parent->PermitRun = FALSE;
				pcrqRQ->CompletionStatus = ACTION_ABORT;
				IDDestroyTable(hitIDTable);
				return;
			}

			//  Check if we need to purge the rejoin pool if not move on to the next request
			if (xeLocal->IsRejoinRecommended(pcrqRQ->Parent, iThreadID))
			{
				//  Retrieve any completed requests - this is necessary to prevent the RQE Rejoin pool from flooding
				while(xeLocal->GetRejoinRequest(pcrqRQ->Parent, (void **) &pcrqCopyNotes, iThreadID) == RJR_RETURNED)
				{
					//  Accumulate the results
					if (pcrqCopyNotes->CopyAction == COPY_DESIGN_NOTES) 
					{
						if (xeLocal->AcquireAppMutex(DESIGN_ACC_MUTEX, iThreadID))
						{
							pcrqRQ->pcrqCopyDesign->AccumulateTotals(pcrqCopyNotes);
							xeLocal->FreeAppMutex(DESIGN_ACC_MUTEX, iThreadID);
						}
					}
					else 
					{
						if (pcrqCopyNotes->CopyAction == COPY_DATA_NOTES)
						{
							if (xeLocal->AcquireAppMutex(DATA_ACC_MUTEX, iThreadID))
							{
								pcrqRQ->pcrqCopyData->AccumulateTotals(pcrqCopyNotes);
								xeLocal->FreeAppMutex(DATA_ACC_MUTEX, iThreadID);
							}
						}
						else
						{
							//  The returned object is not of a valid type - record the error and clear the pointer
							sprintf_s(szMsg, MAX_MSG, MSG_DXR5147E, pcrqCopyNotes, pcrqCopyNotes->CopyAction);
							ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
							pcrqCopyNotes = NULL;
						}
					}
					//  Clean up the request if it is valid
					if (pcrqCopyNotes != NULL)
					{
						//  Check if there were any non-recovered errors noted in the request if so abort the copy
						if (pcrqCopyNotes->dwCopyAccessErrors > 0)
						{
							if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_ACCESS_ERRORS))
							{
								pcrqCopyNotes->CopyAction = COPY_FREED;
								delete pcrqCopyNotes;
								pcrqCopyNotes = NULL;	
								//  Signal an abort
								pcrqRQ->CompletionStatus = ACTION_ABORT;
								pcrqRQ->Parent->PermitRun = FALSE;
								IDDestroyTable(hitIDTable);
								return;
							}
						}
						if (pcrqCopyNotes->dwCopyOtherErrors > 0)
						{
							if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_COPY_ERRORS))
							{
								pcrqCopyNotes->CopyAction = COPY_FREED;
								delete pcrqCopyNotes;
								pcrqCopyNotes = NULL;	
								//  Signal an abort
								pcrqRQ->CompletionStatus = ACTION_ABORT;
								pcrqRQ->Parent->PermitRun = FALSE;
								IDDestroyTable(hitIDTable);
								return;
							}
						}
						//  Update the action in memory in case a race causes the freed object to be returned to another thread
						pcrqCopyNotes->CopyAction = COPY_FREED;
						delete pcrqCopyNotes;
						pcrqCopyNotes = NULL;	
					}
				}
			}
			pcrqCopyNotes = NULL;
			//  Check if we have too many folders to log to the proxy interface - if so suppress further logging
			if (pcrqRQ->Parent->bAllowProxyFolderMessages)
			{
				dwFoldersToMaintain = IDEntries(pcrqRQ->Parent->hitFolderTable);
				if (dwFoldersToMaintain > MAX_FOLDER_MSGS) 
				{
					ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5176I, iThreadID);
					pcrqRQ->Parent->bAllowProxyFolderMessages = FALSE;
				}
			}
		}
	}

	//  If there is a request constructed then dispatch it
	if (pcrqCopyNotes != NULL)
	{
		//  Post a request to execute the copy action
		if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
									pcrqRQ->Parent,
									this,
									pcrqCopyNotes,
									RQATTR_REJOIN,
									10,
									iThreadID))
		{
			//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
			ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5105E, iThreadID);			
			//  Signal an abort
			pcrqRQ->Parent->PermitRun = FALSE;
			pcrqRQ->CompletionStatus = ACTION_ABORT;
			IDDestroyTable(hitIDTable);
			return;
		}
	}

	//  Retrieve any completed requests - this is necessary to prevent the RQE Rejoin pool from flooding
	while(xeLocal->GetRejoinRequest(pcrqRQ->Parent, (void **) &pcrqCopyNotes, iThreadID) == RJR_RETURNED)
	{
		//  Accumulate the results
		if (pcrqCopyNotes->CopyAction == COPY_DESIGN_NOTES) 
		{
			if (xeLocal->AcquireAppMutex(DESIGN_ACC_MUTEX, iThreadID))
			{
				pcrqRQ->pcrqCopyDesign->AccumulateTotals(pcrqCopyNotes);
				xeLocal->FreeAppMutex(DESIGN_ACC_MUTEX, iThreadID);
			}
		}
		else 
		{
			if (pcrqCopyNotes->CopyAction == COPY_DATA_NOTES)
			{
				if (xeLocal->AcquireAppMutex(DATA_ACC_MUTEX, iThreadID))
				{
					pcrqRQ->pcrqCopyData->AccumulateTotals(pcrqCopyNotes);
					xeLocal->FreeAppMutex(DATA_ACC_MUTEX, iThreadID);
				}
			}
			else
			{
				//  The returned object is not of a valid type - record the error and clear the pointer
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5147E, pcrqCopyNotes, pcrqCopyNotes->CopyAction);
				ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
				pcrqCopyNotes = NULL;
			}
		}
		//  Clean up the request if it is valid
		if (pcrqCopyNotes != NULL)
		{
			//  Check if there were any non-recovered errors noted in the request if so abort the copy
			if (pcrqCopyNotes->dwCopyAccessErrors > 0)
			{
				if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_ACCESS_ERRORS))
				{
					pcrqCopyNotes->CopyAction = COPY_FREED;
					delete pcrqCopyNotes;
					pcrqCopyNotes = NULL;	
					//  Signal an abort
					pcrqRQ->CompletionStatus = ACTION_ABORT;
					pcrqRQ->Parent->PermitRun = FALSE;
					IDDestroyTable(hitIDTable);
					return;
				}
			}
			if (pcrqCopyNotes->dwCopyOtherErrors > 0)
			{
				if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_COPY_ERRORS))
				{
					pcrqCopyNotes->CopyAction = COPY_FREED;
					delete pcrqCopyNotes;
					pcrqCopyNotes = NULL;	
					//  Signal an abort
					pcrqRQ->CompletionStatus = ACTION_ABORT;
					pcrqRQ->Parent->PermitRun = FALSE;
					IDDestroyTable(hitIDTable);
					return;
				}
			}
			//  Update the action in memory in case a race causes the freed object to be returned to another thread
			pcrqCopyNotes->CopyAction = COPY_FREED;
			delete pcrqCopyNotes;
			pcrqCopyNotes = NULL;	
		}
		//  Check if we have too many folders to log to the proxy interface - if so suppress further logging
		if (pcrqRQ->Parent->bAllowProxyFolderMessages)
		{
			dwFoldersToMaintain = IDEntries(pcrqRQ->Parent->hitFolderTable);
			if (dwFoldersToMaintain > MAX_FOLDER_MSGS) 
			{
				ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5176I, iThreadID);
				pcrqRQ->Parent->bAllowProxyFolderMessages = FALSE;
			}
		}
	}

	//  Exit - data copying is complete
	if (iWarningsIssued > 0) pcrqRQ->CompletionStatus = ACTION_WARN;
	else pcrqRQ->CompletionStatus = ACTION_COMPLETED;
	IDDestroyTable(hitIDTable);
	return;
}

//  PerformCopyNotes
//
//  This is the generic function that copies notes from the source to the target databese.
//  The copying is performed in chunks - the size of chunks is determined dynamically when the copy proces is started.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformCopyNotes(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLSource;						//  Local Thread handle for the Source database
	DBHANDLE			hdbTLTarget;						//  Local Thread handle for the Target database
	NOTEID				nidNewNID;							//  NoteID of the copied note
	NOTEHANDLE			hnSourceNote;						//  Handle to the source note
	NOTEHANDLE			hnTargetNote;						//  Handle to the target note
	OID					oidNew;								//  OID for a copied note
	BOOL				bNoteIsDeleted;						//  Note is a deletion stub
	BOOL				bNoteIsPrivate;						//  Note is a private design element
	WORD				wNoteType;							//  Note class of copied note
	DWORD				dwOpenFlags;						//  Note Open flags
	DWORD				dwUpdateFlags;						//  Note Update flags
	int					iIndex;								//  Index into the noteids passed
	int					iNetworkErrors;						//  Network Error Count
	PartCopyRequest		*pcrqCRA = NULL;					//  Check recompression request
	BLOCKID				bidScrap;							//  Disposable blockid
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Check that the Run Permit flag is still enabled
	if (!pcrqRQ->Parent->PermitRun)
	{
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Get the mapped thread local database handles
	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);
	hdbTLSource = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbSource, iThreadID);

	//  If this request is copying data notes and LZ1 compression is requested then create the request to
	//  accumulate notes to be checked for attachments to recompress.
	if (pcrqRQ->CopyAction == COPY_DATA_NOTES && pcrqRQ->Parent->LZ1Compress && pcrqRQ->Parent->RecompressAttachments)
	{
		pcrqCRA = new PartCopyRequest(pcrqRQ->Parent);
		pcrqCRA->CopyAction = CHECK_RECOMPRESS;
		pcrqCRA->CopyNoteCount = 0;
	}

	//  Copy each of the requested notes
	for (iIndex = 0; iIndex < pcrqRQ->CopyNoteCount; iIndex++)
	{
		bNoteIsDeleted = FALSE;
		bNoteIsPrivate = FALSE;
		dwOpenFlags = OPEN_RAW_MIME;
		dwUpdateFlags = UPDATE_FORCE | UPDATE_NOREVISION;
		if (pcrqRQ->nidToCopy[iIndex] & RRV_DELETED)
		{
			//  Settings for a deletion stub
			bNoteIsDeleted = TRUE;
			pcrqRQ->nidToCopy[iIndex] = pcrqRQ->nidToCopy[iIndex] ^ RRV_DELETED;
			dwOpenFlags = 0;
			dwUpdateFlags = UPDATE_FORCE | UPDATE_NOREVISION | UPDATE_DELETED;
		}
		else
		{
			if (pcrqRQ->CopyAction == COPY_DESIGN_NOTES)
			{
				//  Set the private design attribute and post the noteid for any folders
				bNoteIsPrivate = IsPrivateDesignNote(pcrqRQ, pcrqRQ->nidToCopy[iIndex], &stAPIRC, iThreadID);
				if (ERR(stAPIRC) != NOERROR)
				{
					//  Treat as a deleted note
					bNoteIsDeleted = TRUE;
					bNoteIsPrivate = FALSE;
					dwOpenFlags = 0;
					dwUpdateFlags = UPDATE_FORCE | UPDATE_NOREVISION | UPDATE_DELETED;
				}
			}
		}

		//  Only process deletion stubs if needed
		if ((!bNoteIsDeleted) || (pcrqRQ->Parent->CopyStubs))
		{
			//  Only process private design notes if needed
			if ((!bNoteIsPrivate) || (pcrqRQ->Parent->CopyPrivates))
			{
				if (pcrqRQ->Parent->CopyReplica)
				{
					//  Make a replica copy of the note
					iNetworkErrors = -1;
					stAPIRC = ERR_TIMEOUT;
					while ((ERR(stAPIRC) == ERR_TIMEOUT) && (iNetworkErrors < NET_RETRY_LIMIT))
					{
						iNetworkErrors++;
						if (bNoteIsDeleted)
						{
							stAPIRC = NSFDbCopyNoteExt(hdbTLSource, 
														NULL, 
														NULL, 
														pcrqRQ->nidToCopy[iIndex], 
														hdbTLTarget, 
														NULL, 
														NULL, 
														dwOpenFlags, 
														dwUpdateFlags, 
														&nidNewNID, 
														&wNoteType);
						}
						else
						{
							stAPIRC = NSFDbCopyNoteExt(hdbTLSource, 
														NULL, 
														&pcrqRQ->Parent->dbidSource, 
														pcrqRQ->nidToCopy[iIndex], 
														hdbTLTarget, 
														NULL, 
														&pcrqRQ->Parent->dbidTarget, 
														dwOpenFlags, 
														dwUpdateFlags, 
														&nidNewNID, 
														&wNoteType);
						}
						if (ERR(stAPIRC) == ERR_SERVER_NOT_RESPONDING) stAPIRC = ERR_TIMEOUT;
					}
					if (ERR(stAPIRC) != NOERROR)
					{
						//  Operation has failed
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5025E, pcrqRQ->nidToCopy[iIndex]);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						xeLocal->GetAPIMessage(stAPIRC, szMsg);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						if (ERR(stAPIRC) == ERR_NOACCESS)
						{
							pcrqRQ->dwCopyAccessErrors++;
						}
						else
						{
							pcrqRQ->dwCopyOtherErrors++;
							//  Request a fixup on the source as part of recovery
							pcrqRQ->Parent->RestartActions = pcrqRQ->Parent->RestartActions | RESTART_ASSERT_FIXUP;
						}
					}
					else
					{
						if (bNoteIsDeleted) pcrqRQ->dwStubsCopied++;
						else
						{
							pcrqRQ->dwNotesCopied++;

							//  If we are checking for recompression of attachments then capture the new NoteID
							if (pcrqCRA != NULL && NSFNoteHasObjects(hnTargetNote, &bidScrap))
							{
								pcrqCRA->nidToCopy[pcrqCRA->CopyNoteCount] = nidNewNID;
								pcrqCRA->CopyNoteCount++;
							}
						}
						if (iNetworkErrors > 0)
						{
							pcrqRQ->dwCopyRetries = pcrqRQ->dwCopyRetries + iNetworkErrors;
							if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
							{
								sprintf_s(szMsg, MAX_MSG, MSG_DXR5026W, pcrqRQ->nidToCopy[iIndex], iNetworkErrors);
								ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
							}
						}
					}
				}
				else
				{
					//  Make a non-replica copy of the note
					iNetworkErrors = -1;
					stAPIRC = ERR_TIMEOUT;
					while ((ERR(stAPIRC) == ERR_TIMEOUT) && (iNetworkErrors < NET_RETRY_LIMIT))
					{
						iNetworkErrors++;
						stAPIRC = NSFNoteOpenExt(hdbTLSource, 
												pcrqRQ->nidToCopy[iIndex], 
												OPEN_RAW_MIME, 
												&hnSourceNote);
						if (ERR(stAPIRC) == ERR_SERVER_NOT_RESPONDING) stAPIRC = ERR_TIMEOUT;
					}
					if (ERR(stAPIRC) != NOERROR)
					{
						//  Operation has failed
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5027E, pcrqRQ->nidToCopy[iIndex]);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						xeLocal->GetAPIMessage(stAPIRC, szMsg);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						if (ERR(stAPIRC) == ERR_NOACCESS)
						{
							pcrqRQ->dwCopyAccessErrors++;
						}
						else
						{
							pcrqRQ->dwCopyOtherErrors++;
							//  Request a fixup on the source as part of recovery
							pcrqRQ->Parent->RestartActions = pcrqRQ->Parent->RestartActions | RESTART_ASSERT_FIXUP;
						}
					}
					else
					{
						if (iNetworkErrors > 0)
						{
							pcrqRQ->dwCopyRetries = pcrqRQ->dwCopyRetries + iNetworkErrors;
							if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
							{
								sprintf_s(szMsg, MAX_MSG, MSG_DXR5028W, pcrqRQ->nidToCopy[iIndex], iNetworkErrors);
								ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
							}
						}

						//  Create an in memory copy of the note
						stAPIRC = NSFNoteCopy(hnSourceNote, &hnTargetNote);
						if (ERR(stAPIRC) != NOERROR)
						{
							sprintf_s(szMsg, MAX_MSG, MSG_DXR5029E, pcrqRQ->nidToCopy[iIndex]);
							ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
							xeLocal->GetAPIMessage(stAPIRC, szMsg);
							ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
							if (ERR(stAPIRC) == ERR_NOACCESS)
							{
								pcrqRQ->dwCopyAccessErrors++;
							}
							else
							{
								pcrqRQ->dwCopyOtherErrors++;
								//  Request a fixup on the source as part of recovery
								pcrqRQ->Parent->RestartActions = pcrqRQ->Parent->RestartActions | RESTART_ASSERT_FIXUP;
							}
						}
						else
						{
							//  If folders are to be maintained then preserve the UNID in the document
							if (!pcrqRQ->Parent->CopyFolders)
							{
								//  Set the new note information
								stAPIRC = NSFDbGenerateOID(hdbTLTarget, 
															&oidNew);
								if (ERR(stAPIRC) != NOERROR)
								{
									sprintf_s(szMsg, MAX_MSG, MSG_DXR5030W, pcrqRQ->nidToCopy[iIndex]);
									ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
								}
								else
								{
									NSFNoteSetInfo(hnTargetNote, _NOTE_OID, &oidNew);
								}
							}
					
							NSFNoteSetInfo(hnTargetNote, _NOTE_ID, NULL);
							NSFNoteSetInfo(hnTargetNote, 
											_NOTE_DB, 
											&hdbTLTarget);

							//  Update the note in the target database
							//  If this is a private design element then set the SECRET SQUIRREL UPDATE FLAG
							if (bNoteIsPrivate) dwUpdateFlags = dwUpdateFlags | SECRET_SQUIRREL_UPDATE_FLAG;
							stAPIRC = ERR_TIMEOUT;
							iNetworkErrors = -1;
							while (stAPIRC == ERR_TIMEOUT)
							{
								iNetworkErrors++;
								stAPIRC = NSFNoteUpdateExtended(hnTargetNote, dwUpdateFlags);
								if (ERR(stAPIRC) == ERR_SERVER_NOT_RESPONDING) stAPIRC = ERR_TIMEOUT;
							}
							if (ERR(stAPIRC) != NOERROR)
							{
								sprintf_s(szMsg, MAX_MSG, MSG_DXR5031E, pcrqRQ->nidToCopy[iIndex]);
								ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
								xeLocal->GetAPIMessage(stAPIRC, szMsg);
								ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
								if (ERR(stAPIRC) == ERR_NOACCESS)
								{
									pcrqRQ->dwCopyAccessErrors++;
								}
								else
								{
									pcrqRQ->dwCopyOtherErrors++;
									//  Request a fixup on the source as part of recovery
									pcrqRQ->Parent->RestartActions = pcrqRQ->Parent->RestartActions | RESTART_ASSERT_FIXUP;
								}
							}
							else
							{
								pcrqRQ->dwNotesCopied++;

								//  If we are checking for recompression of attachments then capture the new NoteID
								if (pcrqCRA != NULL && NSFNoteHasObjects(hnTargetNote, &bidScrap))
								{
									NSFNoteGetInfo(hnTargetNote, _NOTE_ID, &nidNewNID);
									pcrqCRA->nidToCopy[pcrqCRA->CopyNoteCount] = nidNewNID;
									pcrqCRA->CopyNoteCount++;
								}

								//  warn if retries were needed
								if (iNetworkErrors > 0)
								{
									pcrqRQ->dwCopyRetries = pcrqRQ->dwCopyRetries + iNetworkErrors;
									if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
									{
										sprintf_s(szMsg, MAX_MSG, MSG_DXR5032W, pcrqRQ->nidToCopy[iIndex], iNetworkErrors);
										ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
									}
								}
							}
							NSFNoteClose(hnTargetNote);
						}
						NSFNoteClose(hnSourceNote);
					}
				}
			}
		}
	}

	//  If a check attachments request was created and it contains notes to process then dispatch the request
	//  otherwise delete it.
	if (pcrqCRA != NULL)
	{
		if (pcrqCRA->CopyNoteCount > 0)
		{
			//  Dispatch the request
			if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
										ATT_PROC_OWNER,
										this,
										pcrqCRA,
										RQATTR_FIREANDFORGET,
										15,
										iThreadID))
			{
				//  Log a warning
				xeLocal->LogMessage(MSG_DXR5209W, iThreadID);
				delete pcrqCRA;
			}
		}
		else delete pcrqCRA;
	}
	return;
}

//  PerformCRA
//
//  This function checks to see if any notes in passed request have attachments, if so it checks the compression type
//  of each attachment if NONE or HUFFMAN then a request is dispatched to recompress the attachments in that note.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformCRA(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLTarget;						//  Local Thread handle for the Target database
	NOTEHANDLE			hnTargetNote;						//  Handle to the target note
	BLOCKID				bidItem;							//  Blockid of the item
	WORD				wDataType;							//  Data Type
	BLOCKID				bidValue;							//  Blockid of the Item Value
	DWORD				dwDataLen;							//  Data Length
	UCHAR				*pValue;							//  Pointer to the value
	FILEOBJECT			*foSrc;								//  Pointer to the source file object
	FILEOBJECT			foAttachment;						//  File object 
	int					iIndex;								//  Generic index
	BOOL				RecompressThisNote;					//  Compression control
	PartCopyRequest		*pcrqRA;							//  Attachment recompression request
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Check that the Run Permit flag is still enabled
	if (!pcrqRQ->Parent->PermitRun)
	{
		delete pcrqRQ;
		return;
	}

	//  Get the mapped thread local database handles
	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);

	//  Process each note in turn
	for (iIndex = 0; iIndex < pcrqRQ->CopyNoteCount; iIndex++)
	{
		//  Open the target note
		stAPIRC = NSFNoteOpen(hdbTLTarget, pcrqRQ->nidToCopy[iIndex], 0, &hnTargetNote);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5210E, pcrqRQ->nidToCopy[iIndex]);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		else
		{
			//  Only check notes that have objects
			if (NSFNoteHasObjects(hnTargetNote, &bidItem))
			{
				stAPIRC = NSFItemInfo(hnTargetNote,
									  ITEM_NAME_ATTACHMENT, 
									  (WORD) strlen(ITEM_NAME_ATTACHMENT),
									  &bidItem,
									  &wDataType,
									  &bidValue,
									  &dwDataLen);
				if (ERR(stAPIRC) != ERR_ITEM_NOT_FOUND)
				{
					if (ERR(stAPIRC) != NOERROR)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5211E, pcrqRQ->nidToCopy[iIndex]);
						xeLocal->LogMessage(szMsg, iThreadID);
						xeLocal->GetAPIMessage(stAPIRC, szMsg);
						xeLocal->LogMessage(szMsg, iThreadID);
					}
					else
					{
						//  Default to no recompression needed
						RecompressThisNote = FALSE;

						//  Process each attachment until we find one that needs compressing
						while(ERR(stAPIRC) == NOERROR && !RecompressThisNote)
						{
							//  Read the FILEOBJECT for this attachment
							pValue = OSLockBlock(UCHAR, bidValue);
							foSrc = (FILEOBJECT *) (pValue + sizeof(WORD));
							ODSReadMemory(&foSrc, _FILEOBJECT, &foAttachment, 1);
							OSUnlockBlock(bidValue);

							//  Check the compression type
							if (foAttachment.CompressionType == COMPRESS_NONE || foAttachment.CompressionType == COMPRESS_HUFF) RecompressThisNote = TRUE;
							else
							{
								//  Try the next file attachment in this note
								
								stAPIRC = NSFItemInfoNext(hnTargetNote,
													  bidItem,
													  ITEM_NAME_ATTACHMENT, 
													  (WORD) strlen(ITEM_NAME_ATTACHMENT),
													  &bidItem,
													  &wDataType,
													  &bidValue,
													  &dwDataLen);
								if (ERR(stAPIRC) != ERR_ITEM_NOT_FOUND)
								{
									if (ERR(stAPIRC) != NOERROR)
									{
										sprintf_s(szMsg, MAX_MSG, MSG_DXR5211E, pcrqRQ->nidToCopy[iIndex]);
										xeLocal->LogMessage(szMsg, iThreadID);
										xeLocal->GetAPIMessage(stAPIRC, szMsg);
										xeLocal->LogMessage(szMsg, iThreadID);
									}
								}
							}
						}
						if (RecompressThisNote)
						{
							//  Generate a request to recompress the target NOTEID
							pcrqRA = new PartCopyRequest(pcrqRQ->Parent);
							pcrqRA->CopyAction = RECOMPRESS_NOTE;
							pcrqRA->CopyNoteCount = 1;
							pcrqRA->nidToCopy[0] = pcrqRQ->nidToCopy[iIndex];

							//  Dispatch the request
							if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
														ATT_PROC_OWNER,
														this,
														pcrqRA,
														RQATTR_FIREANDFORGET,
														20,
														iThreadID))
							{
								//  Log a warning
								sprintf_s(szMsg, MAX_MSG, MSG_DXR5212W, pcrqRQ->nidToCopy[iIndex]);
								xeLocal->LogMessage(szMsg, iThreadID);
								delete pcrqRA;
							}
						}
					}
				}
			}

			//  Close the target note
			NSFNoteClose(hnTargetNote);
		}
	}

	//  Delete the incoming request
	delete pcrqRQ;

	return;
}

//  PerformRecompress
//
//  This function will recompress the eligible attachemnts in the passed note to use LZ1 compression.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformRecompress(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLTarget;						//  Local Thread handle for the Target database
	NOTEHANDLE			hnTargetNote;						//  Handle to the target note
	BLOCKID				bidItem;							//  Blockid of the item
	WORD				wDataType;							//  Data Type
	BLOCKID				bidValue;							//  Blockid of the Item Value
	DWORD				dwDataLen;							//  Data Length
	UCHAR				*pValue;							//  Pointer to the value
	FILEOBJECT			*foSrc;								//  Pointer to the source file object
	FILEOBJECT			foAttachment;						//  File object
	BLOCKID				bidRA[200];							//  Recompress list
	int					RACount = 0;						//  Count of attachments to be recompressed
	int					iIndex;								//  Generic index
	DWORD				RecompressedCount = 0;				//  Recompression count
	DWORD				BytesSaved = 0;						//  Bytes saved
	DWORD				RCResult;							//  Recompression result
	Attachments			aHandler(xeLocal, iThreadID);		//  Attachment handler
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Check that the Run Permit flag is still enabled
	if (!pcrqRQ->Parent->PermitRun)
	{
		delete pcrqRQ;
		return;
	}

	//  Get the mapped thread local database handles
	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);

	//  Open the target note
	stAPIRC = NSFNoteOpen(hdbTLTarget, pcrqRQ->nidToCopy[0], (WORD) OPEN_RAW_MIME, &hnTargetNote);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5213E, pcrqRQ->nidToCopy[0]);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
	}
	else
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5214T, pcrqRQ->nidToCopy[0]);
		xeLocal->LogTrace(TRACE_DBC, szMsg, iThreadID);

		//  Loop processing each attachment in turn
		stAPIRC = NSFItemInfo(hnTargetNote,
							  ITEM_NAME_ATTACHMENT, 
							  (WORD) strlen(ITEM_NAME_ATTACHMENT),
							  &bidItem,
							  &wDataType,
							  &bidValue,
							  &dwDataLen);
		while(stAPIRC == NOERROR)
		{
			//  Read the FILEOBJECT for this attachment
			pValue = OSLockBlock(UCHAR, bidValue);
			foSrc = (FILEOBJECT *) (pValue + sizeof(WORD));
			ODSReadMemory(&foSrc, _FILEOBJECT, &foAttachment, 1);
			OSUnlockBlock(bidValue);

			//  Check that this is a candidate for recompression to LZ1
			if (foAttachment.CompressionType == COMPRESS_NONE ||foAttachment.CompressionType == COMPRESS_HUFF)
			{
				//  Add the Item blockid to the list of attachments to be recompressed
				bidRA[RACount] = bidItem;
				RACount++;
				if (RACount == 200) RACount--;
			}

			//  Get the next file object
			stAPIRC = NSFItemInfoNext(hnTargetNote,
									  bidItem,
									  ITEM_NAME_ATTACHMENT, 
									  (WORD) strlen(ITEM_NAME_ATTACHMENT),
									  &bidItem,
									  &wDataType,
									  &bidValue,
									  &dwDataLen);
		}

		//  Check the loop termination state
		if (stAPIRC != ERR_ITEM_NOT_FOUND)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5211E, pcrqRQ->nidToCopy[0]);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		else
		{
			//  Recompress each attachment in turn
			for (iIndex = 0; iIndex < RACount; iIndex++)
			{
				RCResult = aHandler.RecompressAttachment(hnTargetNote, bidRA[iIndex], COMPRESS_LZ1, iThreadID);
				if (RCResult == MAXDWORD)
				{
					//  Notify the error
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5125E, iIndex + 1, pcrqRQ->nidToCopy[0]);
					ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
					NSFNoteClose(hnTargetNote);
					delete pcrqRQ;
					return;
				}
				if (RCResult > 0) RecompressedCount++;
				BytesSaved += RCResult;
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5218T, RCResult, BytesSaved);
				xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
			}

			//  If any attachments were successfully recompressed then update the note with the new attachments
			if (RecompressedCount > 0)
			{
				stAPIRC = NSFNoteUpdate(hnTargetNote, UPDATE_FORCE);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5126E, pcrqRQ->nidToCopy[0]);
					ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
					xeLocal->GetAPIMessage(stAPIRC, szMsg);
					xeLocal->LogMessage(szMsg, iThreadID);
				}
				else
				{
					//  Update the counts in the parent request
					pcrqRQ->Parent->dwNotesRecom++;
					pcrqRQ->Parent->dwFilesRecom += RecompressedCount;
					pcrqRQ->Parent->dwKBSaved += (BytesSaved / 1024);
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5219T, BytesSaved, pcrqRQ->Parent->dwKBSaved);
					xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
				}
			}
		}
	}

	//  Close the target note
	NSFNoteClose(hnTargetNote);

	//  Delete the request
	delete pcrqRQ;

	return;
}

//  PerformRecompressInPLace
//
//  This function will recompress the eligible attachemnts in the passed note to use LZ1 compression.
//  The function recompresses each attachment in turn replacing the content of the compressed object
//  if possible (and improves the compression).
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformRecompressInPlace(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLTarget;						//  Local Thread handle for the Target database
	NOTEHANDLE			hnTargetNote;						//  Handle to the target note
	BLOCKID				bidItem;							//  Blockid of the item
	WORD				wDataType;							//  Data Type
	BLOCKID				bidValue;							//  Blockid of the Item Value
	DWORD				dwDataLen;							//  Data Length
	UCHAR				*pValue;							//  Pointer to the value
	FILEOBJECT			*foSrc;								//  Pointer to the source file object
	FILEOBJECT			foAttachment;						//  File object
	BLOCKID				bidRA[200];							//  Recompress list
	int					RACount = 0;						//  Count of attachments to be recompressed
	int					iIndex;								//  Generic index
	DWORD				RecompressedCount = 0;				//  Recompression count
	DWORD				BytesSaved = 0;						//  Bytes saved
	DWORD				RCResult;							//  Recompression result
	Attachments			aHandler(xeLocal, iThreadID);		//  Attachment handler
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Mark off the frame
	MARK_FRAME("PerformRecompressInPlace()", &stAPIRC, szMsg);

	//  Check that the Run Permit flag is still enabled
	if (!pcrqRQ->Parent->PermitRun)
	{
		delete pcrqRQ;
		return;
	}

	//  Get the mapped thread local database handles
	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);

	//  Open the target note
	stAPIRC = NSFNoteOpen(hdbTLTarget, pcrqRQ->nidToCopy[0], (WORD) OPEN_RAW_MIME, &hnTargetNote);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5213E, pcrqRQ->nidToCopy[0]);
		xeLocal->LogMessage(szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg, iThreadID);
	}
	else
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5214T, pcrqRQ->nidToCopy[0]);
		xeLocal->LogTrace(TRACE_DBC, szMsg, iThreadID);

		//  Loop processing each attachment in turn
		stAPIRC = NSFItemInfo(hnTargetNote,
							  ITEM_NAME_ATTACHMENT, 
							  (WORD) strlen(ITEM_NAME_ATTACHMENT),
							  &bidItem,
							  &wDataType,
							  &bidValue,
							  &dwDataLen);
		while(stAPIRC == NOERROR)
		{
			//  Read the FILEOBJECT for this attachment
			pValue = OSLockBlock(UCHAR, bidValue);
			foSrc = (FILEOBJECT *) (pValue + sizeof(WORD));
			ODSReadMemory(&foSrc, _FILEOBJECT, &foAttachment, 1);
			OSUnlockBlock(bidValue);

			//  Check that this is a candidate for recompression to LZ1
			if (foAttachment.CompressionType == COMPRESS_NONE || foAttachment.CompressionType == COMPRESS_HUFF)
			{
				//  Add the Item blockid to the list of attachments to be recompressed
				bidRA[RACount] = bidItem;
				RACount++;
				if (RACount == 200) RACount--;
			}

			//  Get the next file object
			stAPIRC = NSFItemInfoNext(hnTargetNote,
									  bidItem,
									  ITEM_NAME_ATTACHMENT, 
									  (WORD) strlen(ITEM_NAME_ATTACHMENT),
									  &bidItem,
									  &wDataType,
									  &bidValue,
									  &dwDataLen);
		}

		//  Check the loop termination state
		if (stAPIRC != ERR_ITEM_NOT_FOUND)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5211E, pcrqRQ->nidToCopy[0]);
			xeLocal->LogMessage(szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			xeLocal->LogMessage(szMsg, iThreadID);
		}
		else
		{
			//  Recompress each attachment in turn
			for (iIndex = 0; iIndex < RACount; iIndex++)
			{
				RCResult = aHandler.RecompressAttachmentInPlace(hnTargetNote, bidRA[iIndex], COMPRESS_LZ1, iThreadID);
				if (RCResult == RETURN_RECOMPRESS_ERROR)
				{
					//  Notify the error
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5125E, iIndex + 1, pcrqRQ->nidToCopy[0]);
					ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
					NSFNoteClose(hnTargetNote);
					delete pcrqRQ;
					return;
				}
				if (RCResult > RETURN_NOT_RECOMPRESSED) RecompressedCount++;
				BytesSaved += RCResult;
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5218T, RCResult, BytesSaved);
				xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
			}

			//  If any attachments were successfully recompressed then update the note with the new attachments
			if (RecompressedCount > 0)
			{
				stAPIRC = NSFNoteUpdate(hnTargetNote, UPDATE_NOREVISION);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5126E, pcrqRQ->nidToCopy[0]);
					ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
					xeLocal->GetAPIMessage(stAPIRC, szMsg);
					xeLocal->LogMessage(szMsg, iThreadID);
				}
				else
				{
					//  Update the counts in the parent request
					pcrqRQ->Parent->dwNotesRecom++;
					pcrqRQ->Parent->dwFilesRecom += RecompressedCount;
					pcrqRQ->Parent->dwKBSaved += (BytesSaved / 1024);
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5219T, BytesSaved, pcrqRQ->Parent->dwKBSaved);
					xeLocal->LogTrace(TRACE_ATT, szMsg, iThreadID);
				}
			}
		}
	}

	//  Close the target note
	NSFNoteClose(hnTargetNote);

	//  Delete the request
	delete pcrqRQ;

	return;
}

//  PerformMaintainFolder
//
//  This function will populate a single folder in the target database based on the conmtents of the
//  folder in the source database.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformMaintainFolder(PartCopyRequest * pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLSource;						//  Thread Local DB Handle
	DBHANDLE			hdbTLTarget;						//  Thread Local DB Handle
	NOTEHANDLE			hnFolder;							//  Handle of the Folder Note
	DHANDLE				hitTargetFolder;					//  Handle of the target folder Note ID Table
	DHANDLE				hitSourceFolder;					//  Handle of the source folder Note ID Table
	NOTEID				nidRetNID;							//  NoteID returned from IDScan
	NOTEID				nidNewNID;							//  NoteID of the note in the target database	
	NOTEID				nidTargetFolder;					//  NoteID of the target folder
	OID					oidEntry;							//  OID of the folder entry
	OID					oidFolder;							//  OID of the source fo;der
	UNID				uidTarget;							//  UNID of note in the source and target database
	TIMEDATE			tdMod;								//  Modification timestamp of the folder entry
	WORD				wRetLen;							//  Returned string length
	WORD				wNoteClass;							//  Class of note of the folder entry
	BOOL				bGetFirst;							//  IDScan control flag
	BOOL				bEntryAdded;						//  Flag indicating if folder entry was added
	BOOL				bPurgeFolder;						//  Flag to trigger purging of the target folder
	int					iWarningsIssued = 0;				//  Number of warnings issued
	DWORD				dwFolderEntries;					//  Count of document entries in the target folder
	DWORD				dwFolderStubs;						//  Count of deletion stubs in the target folder
	char				szFolderName[MAXPATH];				//  Name of the folder
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Check that the Run Permit flag is still enabled
	if (!pcrqRQ->Parent->PermitRun)
	{
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Get the mapped thread local database handles
	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);
	hdbTLSource = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbSource, iThreadID);

	//  Open the folder note to get the name of the folder
	stAPIRC = NSFNoteOpen(hdbTLSource, pcrqRQ->nidToCopy[0], 0, &hnFolder);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5042E, pcrqRQ->nidToCopy[0]);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	wRetLen = NSFItemGetText(hnFolder, FIELD_TITLE, szFolderName, MAXPATH); 
	szFolderName[wRetLen] = '\0';
	NSFNoteClose(hnFolder);

	//  Show the folder we are processing
	if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5043I, szFolderName, pcrqRQ->nidToCopy[0]);
		if (pcrqRQ->Parent->bAllowProxyFolderMessages) ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		else xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Create the ID table for the target folder notes
	stAPIRC = IDCreateTable(sizeof(NOTEID), &hitTargetFolder);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5044E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	bPurgeFolder = FALSE;

	//  Get the note ID table from the source folder
	//  NOTE: There are some cases where this will return an error because of a malformed empty source folder
	//  in these cases create a dummy empty source ID table and continue (with a warning).
	stAPIRC = NSFFolderGetIDTable(hdbTLSource, hdbTLSource, pcrqRQ->nidToCopy[0], DB_GETIDTABLE_VALIDATE, &hitSourceFolder);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5045W, pcrqRQ->nidToCopy[0]);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);

		//  Attempt (2) - try using the unvalidated NoteID Table

		bPurgeFolder = TRUE;							//  Trigger purging of the target folder

		sprintf_s(szMsg, MAX_MSG, MSG_DXR5148I, pcrqRQ->nidToCopy[0]);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);

		stAPIRC = NSFFolderGetIDTable(hdbTLSource, hdbTLSource, pcrqRQ->nidToCopy[0], 0, &hitSourceFolder);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5045W, pcrqRQ->nidToCopy[0]);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);

			//  Attempt (3) - we will attempt to build a Note ID table from the native object
			//  stored in the Folder note itself.
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5149I, szFolderName);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);

			hitSourceFolder = BuildFolderIDTable(pcrqRQ, hdbTLSource, pcrqRQ->nidToCopy[0], iThreadID);

			//  If we still don't have the contents of the Folder then
			//  Create a dummy ID table for the source folder notes
			if (hitSourceFolder == NULLHANDLE)
			{
				stAPIRC = IDCreateTable(sizeof(NOTEID), &hitSourceFolder);
				if (ERR(stAPIRC) != NOERROR)
				{
					ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5046E, iThreadID);
					xeLocal->GetAPIMessage(stAPIRC, szMsg);
					ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
					IDDestroyTable(hitTargetFolder);
					pcrqRQ->CompletionStatus = ACTION_FAILED;
					return;
				}
			}
		}
	}
	
	//  Iterate the source folder ID Table compute the target noteID and add the new entry to the 
	//  target folder ID table.
	bGetFirst = TRUE;
	dwFolderEntries = 0;
	dwFolderStubs = 0;

	while (IDScan(hitSourceFolder, bGetFirst, &nidRetNID))
	{
		bGetFirst = FALSE;
		dwFolderEntries++;

		//  Skip any notes that are deleted
		if (nidRetNID & RRV_DELETED)
		{
			dwFolderStubs++;
		}
		else
		{
			//  Get the UNID of the document in the source database from it's NotesID
			stAPIRC = NSFDbGetNoteInfo(hdbTLSource, nidRetNID, &oidEntry, &tdMod, &wNoteClass);
			if (ERR(stAPIRC) != NOERROR)
			{
				//  Warn that the folder entry will not be processed
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5047W, nidRetNID, pcrqRQ->nidToCopy[0]);
				ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
				xeLocal->GetAPIMessage(stAPIRC, szMsg);
				ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
				dwFolderStubs++;
			}
			else
			{
				uidTarget.File = oidEntry.File;
				uidTarget.Note = oidEntry.Note;

				//  Find the NoteID of the document with that UNID in the target database
				stAPIRC = NSFDbGetNoteInfoByUNID(hdbTLTarget, &uidTarget, &nidNewNID, NULL, NULL, NULL);
				if (ERR(stAPIRC) != NOERROR)
				{
					//  Folder entry has not yet been marked deleted
					if ((ERR(stAPIRC) == ERR_NOT_FOUND) || (ERR(stAPIRC) == ERR_NOTE_DELETED))
					{
						dwFolderStubs++;
					}
					else
					{
						//  Warn that the folder entry will not be processed
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5048W, nidRetNID, pcrqRQ->nidToCopy[0]);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						xeLocal->GetAPIMessage(stAPIRC, szMsg);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						dwFolderStubs++;
					}
				}
				else
				{
					//  Add the NoteID for the document in the target database to the NoteID table
					//  of the target folder
					stAPIRC = IDInsert(hitTargetFolder, nidNewNID, &bEntryAdded);
					if (ERR(stAPIRC) != NOERROR)
					{
						//  Warn that the folder entry was not processed
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5049W, nidRetNID, pcrqRQ->nidToCopy[0]);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						xeLocal->GetAPIMessage(stAPIRC, szMsg);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						dwFolderStubs++;
					}
					else
					{
						if (!bEntryAdded)
						{
						//  Warn that the folder entry was not processed
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5050E, nidRetNID, pcrqRQ->nidToCopy[0]);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						xeLocal->GetAPIMessage(stAPIRC, szMsg);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						dwFolderStubs++;
						}
						else pcrqRQ->dwNotesCopied++;
					}
				}
			}
		}
	}

	//  Destroy the source folder table
	IDDestroyTable(hitSourceFolder);

	//  Locate the NoteID of the target folder
	stAPIRC = NSFDbGetNoteInfo(hdbTLSource, pcrqRQ->nidToCopy[0], &oidFolder, &tdMod, &wNoteClass);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5050E, pcrqRQ->nidToCopy[0]);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		IDDestroyTable(hitTargetFolder);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}
	uidTarget.File = oidFolder.File;
	uidTarget.Note = oidFolder.Note;

	stAPIRC = NSFDbGetNoteInfoByUNID(hdbTLTarget, &uidTarget, &nidTargetFolder, NULL, NULL, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5051E, pcrqRQ->nidToCopy[0]);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		IDDestroyTable(hitTargetFolder);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Attempt to purge any (invalid) existing folder content
	if (bPurgeFolder)
	{
		stAPIRC = PurgeFolderContent(pcrqRQ, hdbTLTarget, nidTargetFolder, iThreadID);
		//  stAPIRC = FolderDocRemoveAll(hdbTLTarget, NULLHANDLE, nidTargetFolder, 0);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5159E, nidTargetFolder);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		}
	}

	//  Add the target NoteID table to the target folder
	stAPIRC = FolderDocAdd(hdbTLTarget, NULLHANDLE, nidTargetFolder, hitTargetFolder, 0);
	if (ERR(stAPIRC) == ERR_RSP_ADDTO_HIERFOLD)
	{
		//  The NoteID table contains response documents and the folder has a response hierarchy
		//  the response documents must be filtered from the table before adding them to the folder.
		//  Folders with a respose hierarchy dynamically include respose documents.
		stAPIRC = FilteredFolderDocAdd(pcrqRQ, pcrqRQ->Parent->hdbTarget, nidTargetFolder, hitTargetFolder, iThreadID);
	}
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5052E, pcrqRQ->nidToCopy[0]);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		IDDestroyTable(hitTargetFolder);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Show the copy stats
	if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5059I, pcrqRQ->dwNotesCopied, pcrqRQ->nidToCopy[0], dwFolderStubs);
		if (pcrqRQ->Parent->bAllowProxyFolderMessages) ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		else xeLocal->LogMessage(szMsg, iThreadID);
	}

	//  Destroy the target folder table
	IDDestroyTable(hitTargetFolder);
	pcrqRQ->CompletionStatus = ACTION_COMPLETED;
	return;
}

//  PerformUpgradeDesign
//
//  This function will upgrade the design of the target database to the template that is specified.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformUpgradeDesign(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLTarget;						//  Thread Local Database Handle
	char				cDbInfo[NSF_INFO_SIZE];				//  D/B Info Buffer
	char				szOriginalTemplate[MAX_TEMPLATE];	//  Original template name
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer


	//  Ensure that the initial completion state is set
	pcrqRQ->CompletionStatus = ACTION_INPROGRESS;

	//  Get the mapped thread local database handles
	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);

	//  Get the D/B info Buffer
	stAPIRC = NSFDbInfoGet(hdbTLTarget, cDbInfo);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5073E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Parse out the current template that the database inherits from
	NSFDbInfoParse(cDbInfo, INFOPARSE_DESIGN_CLASS, szOriginalTemplate, MAX_TEMPLATE);

	//  Determine if the template name needs to be changed
	if (pcrqRQ->Parent->szTemplateName[0] != '\0')
	{
		if (strcmp(pcrqRQ->Parent->szTemplateName, szOriginalTemplate) != 0)
		{
			NSFDbInfoModify(cDbInfo, INFOPARSE_DESIGN_CLASS, pcrqRQ->Parent->szTemplateName);

			//  Update the info in the database
			stAPIRC = NSFDbInfoSet(hdbTLTarget, cDbInfo);
			if (ERR(stAPIRC) != NOERROR)
			{
				ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5074E, iThreadID);
				xeLocal->GetAPIMessage(stAPIRC, szMsg);
				ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
				pcrqRQ->CompletionStatus = ACTION_FAILED;
				return;
			}
			if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
			{
				strcpy_s(szOriginalTemplate,MAX_TEMPLATE,  pcrqRQ->Parent->szTemplateName);
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5075I, pcrqRQ->Parent->szTemplateName);
				ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
			}
		}
	}

	//  Force a design refresh
	stAPIRC = DesignRefresh("", hdbTLTarget, DESIGN_FORCE, NULL, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5076E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Show design refresh has been completed
	if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5077I, szOriginalTemplate);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
	}
	pcrqRQ->CompletionStatus = ACTION_COMPLETED;
	return;
}

//  PerformBuildIndexes
//
//  This function will build all of the view indexes in the target database.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformBuildIndexes(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	PartCopyRequest		*pcrqBuildIndex;					//  Request to build a single index
	HCOLLECTION			hcDesign;							//  Handle to the Design collection
	COLLECTIONPOSITION	cpDesign;							//  Collection Position in the Design collection
	DHANDLE				hnidDesign;							//  Handle for the returned design collection
	NOTEID				*nidDesign;							//  Design Note IDs
	DWORD				dwNotesFound;						//  Number of design notes found
	DWORD				dwIndex;							//  Index in the design collection
	DWORD				dwViewsBuilt;						//  Count of views built
	WORD				wSignalFlag;						//  NIF signals
	OID					oidDesign;							//  OID of the design note
	TIMEDATE			tdMod;								//  Last modified timestamp
	WORD				wNoteClass;							//  Note Class of the design note
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Build all of the indexes in the requested database

	//  Open the design collection
	stAPIRC = NIFOpenCollection(pcrqRQ->Parent->hdbTarget, pcrqRQ->Parent->hdbTarget, NOTE_CLASS_DESIGN | NOTE_ID_SPECIAL, 0L, NULLHANDLE, &hcDesign, NULLHANDLE, NULL, NULLHANDLE, NULLHANDLE);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5078E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Read the design collection
	cpDesign.Level = 0;
	cpDesign.Tumbler[0] = 0;

	stAPIRC = NIFReadEntries(hcDesign,						// handle to this collection
							&cpDesign,						// where to start in collection
							NAVIGATE_NEXT,					// order to use when skipping
							1L,								// number to skip
							NAVIGATE_NEXT,					// order to use when reading
							0xFFFFFFFF,						// max number to read 
							READ_MASK_NOTEID,				// info we want 
							&hnidDesign,					// handle to info buffer (return)
							NULL,							// length of info buffer (return)
							NULL,							// entries skipped (return)
							&dwNotesFound,					// entries read (return)
							&wSignalFlag);					// share warning and more signal flag (return)

	if (ERR(stAPIRC) != NOERROR)
	{
		//  Bad RC from NIFReadEntries
		NIFCloseCollection(hcDesign);
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5079E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}
									
	if (hnidDesign == NULLHANDLE)
	{
		//  No handle to returned collection
		NIFCloseCollection(hcDesign);
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5080E, iThreadID);
		return;
	}

	//  See if there is any design to process
	if (dwNotesFound == 0)
	{
		NIFCloseCollection(hcDesign);
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5081E, iThreadID);
		return;
	}

	dwViewsBuilt = 0;
	nidDesign = (NOTEID *) OSLockObject(hnidDesign);

	for (dwIndex = 0; dwIndex < dwNotesFound; dwIndex++)
	{
		//  Determine if the design note is a View
		stAPIRC = NSFDbGetNoteInfo(pcrqRQ->Parent->hdbTarget, nidDesign[dwIndex], &oidDesign, &tdMod, &wNoteClass);
		if (ERR(stAPIRC) != NOERROR)
		{
			ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5079E, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
			OSUnlock(hnidDesign);
			OSMemFree(hnidDesign);
			NIFCloseCollection(hcDesign);
			pcrqRQ->CompletionStatus = ACTION_FAILED;
			return;
		}
		if (wNoteClass == NOTE_CLASS_VIEW)
		{
			//  For each view - build the Index
			pcrqBuildIndex = new PartCopyRequest(pcrqRQ->Parent);
			pcrqBuildIndex->CopyAction = COPY_BUILD_INDEX;
			pcrqBuildIndex->nidToCopy[0] = nidDesign[dwIndex];
			pcrqBuildIndex->CopyNoteCount = 1;

			//  Send the request for processing
			PerformCopyAction(pcrqBuildIndex, iThreadID);

			//  Record the stats
			if (pcrqBuildIndex->CompletionStatus == ACTION_COMPLETED) dwViewsBuilt++;

			//  Delete the request
			delete pcrqBuildIndex;
		}
	}
	OSUnlock(hnidDesign);
	OSMemFree(hnidDesign);
	NIFCloseCollection(hcDesign);

	//  Show the results
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5083I, dwViewsBuilt);
	ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);

	pcrqRQ->CompletionStatus = ACTION_COMPLETED;
	return;
}

//  PerformBuildIndex
//
//  This function will build a single view index in the target database
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformBuildIndex(PartCopyRequest *pcrqRQ, int iThreadID)
{
	DBHANDLE		hdbTLTarget;							//   Local D/B Handle

	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);
	xeLocal->BuildIndex(hdbTLTarget, pcrqRQ->nidToCopy[0], iThreadID);
	pcrqRQ->CompletionStatus = ACTION_COMPLETED;
	return;
}

//  PerformCopyFTI
//
//  This function will copy the Full Text Index (FTI) settings from the source to the target database
//  and will then build the Full Text Index.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformCopyFTI(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLSource;						//  Thread Local Database Handle
	DBHANDLE			hdbTLTarget;						//  Thread Local Database Handle
	DHANDLE				hObjectBuffer;						//  Handle of the Full Text Index definition object
	NOTEID				nidObjRRV;							//  RRV of the FTI definition object in the target database
	TIMEDATE			tdLastIndexed;						//  Timestamp of the last update to the FTI
	DWORD				dwObjID;							//  Special Note ID of the FTI Definition Object
	DWORD				dwObjSize;							//  Size of the FTI Definition Object
	DWORD				dwTargetObjSize;					//  Target Database object size
	WORD				wObjectClass;						//  Class of the returned object
	WORD				wObjectPrivs;						//  Privileges of the returned object
	//  Fix for 8.5.3 extension of structure - Thanks to DN.
	union {
		FT_INDEX_STATS		fisNew;								//  Full Text Index stats from Index update
		DWORD				Merges;								//  Count of merges
		DWORD				MergeMsec;							//  ms spent merging
	} fisX;

	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Make sure the status is set
	pcrqRQ->CompletionStatus = ACTION_INPROGRESS;

	//  Get the mapped thread local database handles
	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);
	hdbTLSource = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbSource, iThreadID);

	//  Determine if the source database is indexed
	stAPIRC = FTGetLastIndexTime(hdbTLSource, &tdLastIndexed);
	if (ERR(stAPIRC) != NOERROR)
	{
		if (ERR(stAPIRC) == ERR_FT_NOT_INDEXED)
		{
			if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
			{
				ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5084I, iThreadID);
			}
			pcrqRQ->CompletionStatus = ACTION_COMPLETED;
			return;
		}
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5085E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Get the size of the FTI definition object in the Source Database
	dwObjID = NOTE_ID_SPECIAL | FTI_DEF_OBJ_ID;
	stAPIRC = NSFDbGetObjectSize(hdbTLSource, dwObjID, OBJECT_UNKNOWN, &dwObjSize, &wObjectClass, &wObjectPrivs);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5086E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Determine if the FTI Definition object already exists in the target database
	dwObjID = NOTE_ID_SPECIAL | FTI_DEF_OBJ_ID;
	stAPIRC = NSFDbGetObjectSize(hdbTLTarget, dwObjID, OBJECT_UNKNOWN, &dwTargetObjSize, NULL, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		if (ERR(stAPIRC) != ERR_SPECIAL_ID)
		{
			ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5087E, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
			pcrqRQ->CompletionStatus = ACTION_FAILED;
			return;
		}
	}
	else
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5088E, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Read the FTI Definition object from the source database
	dwObjID = NOTE_ID_SPECIAL | FTI_DEF_OBJ_ID;
	stAPIRC = NSFDbReadObject(hdbTLSource, dwObjID, 0, dwObjSize, &hObjectBuffer);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5089E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Create the FTI Definition object in the target database
	stAPIRC = NSFDbAllocObject(hdbTLTarget, dwObjSize, wObjectClass, 0, &nidObjRRV);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5090E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		OSMemFree(hObjectBuffer);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	// Write the object from the source to the target database
	stAPIRC = NSFDbWriteObject(hdbTLTarget, nidObjRRV, hObjectBuffer, 0, dwObjSize);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5091E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		OSMemFree(hObjectBuffer);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	OSMemFree(hObjectBuffer);

	// Allocate the object as a special D/B Object
	dwObjID = SPECIAL_FTI_ID;
	stAPIRC = xeLocal->SetSpecialNoteID(hdbTLTarget, (WORD) SPECIAL_FTI_ID, nidObjRRV, iThreadID);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5092E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Enable FT Indexing in the target database
	stAPIRC = NSFDbSetOptions(hdbTLTarget, DBOPTION_FT_INDEX, DBOPTION_FT_INDEX);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5093E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}	

	//  Log the enabling of full text indexing
	if ((xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE) || (!pcrqRQ->Parent->CopyAndBuildFTI))
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5094I, iThreadID);
	}

	//  If requested build the new Full Text Index
	if (pcrqRQ->Parent->CopyAndBuildFTI)
	{
		stAPIRC = FTIndex(hdbTLTarget, FT_INDEX_REINDEX, NULL, (FT_INDEX_STATS *) &fisX);
		if (ERR(stAPIRC) != NOERROR)
		{
			ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5102E, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
			pcrqRQ->CompletionStatus = ACTION_FAILED;
			return;
		} 

		//  Show the stats from the build
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5103I, fisX.fisNew.DocsAdded, fisX.fisNew.BytesIndexed);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
	}
	pcrqRQ->CompletionStatus = ACTION_COMPLETED;
	return;
}

//  PerformRunAgent
//
//  This function will execute the designated agent in the target database.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformRunAgent(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLTarget;						//  Thread Local Database Handle
	HAGENT				haAgent;							//  Handle of the agent
	HAGENTCTX			hactxRunContext;					//  Run context for the agent
	DHANDLE				hOutput;							//  Handle to any agent output
	DWORD				dwOutputSize;						//  Size of any output captured
	DWORD				dwIndex;							//  Index into the output buffer
	int					iSegLen;							//  Length of the extracted text segment
	char				*pOutput;							//  Pointer to the output text
	char				*pSegment;							//  Pointer to the text segment
	char				szLogSeg[MAX_MSG];					//  Agent output segment
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Get the mapped thread local database handles
	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);

	//  Open the Agent
	stAPIRC = AgentOpen(hdbTLTarget, pcrqRQ->nidToCopy[0], &haAgent);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5096E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		return;
	}

	//  Create an Agent Run context for the agent
	stAPIRC = AgentCreateRunContext(haAgent, NULL, AGENT_SECURITY_OFF, &hactxRunContext);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5097E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		AgentClose(haAgent);
		return;
	}

	//  Setup redirection of the agent output to an in-memory buffer
	stAPIRC = AgentRedirectStdout(hactxRunContext, AGENT_REDIR_MEMORY);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5098E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		AgentDestroyRunContext(hactxRunContext);
		AgentClose(haAgent);
		return;
	}

	//  Run the agent
	stAPIRC = AgentRun(haAgent, hactxRunContext, 0, 0);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5099E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		pcrqRQ->CompletionStatus = ACTION_FAILED;
		AgentDestroyRunContext(hactxRunContext);
		AgentClose(haAgent);
		return;
	}

	//  Get any output that was generated by the agent and display it in the log
	AgentQueryStdoutBuffer(hactxRunContext, &hOutput, &dwOutputSize);
	if (dwOutputSize > 0)
	{
		pOutput = OSLock(char, hOutput);
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5100I, iThreadID);
		iSegLen = 0;
		memset(szLogSeg, 0, MAX_MSG);
		pSegment = szLogSeg;
		for (dwIndex = 0; dwIndex < dwOutputSize; dwIndex++)
		{
			if (*pOutput > '\16')
			{
				*pSegment = *pOutput;
				pSegment++;
				iSegLen++;
				if (iSegLen > (MAX_MSG - 20))
				{
					*pSegment = '\0';
				}
			}
			else
			{
				*pSegment = '\0';
			}
			pOutput++;
			if (*pSegment == '\0')
			{
				if (iSegLen > 0)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5101I, szLogSeg);
					ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
					iSegLen = 0;
					memset(szLogSeg, 0, MAX_MSG);
					pSegment = szLogSeg;
				}
			}
		}

		//  If there is any residue in the log segment then write it
		if (iSegLen > 0)
		{
			*pSegment = '\0';
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5101I, szLogSeg);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		}
		OSUnlock(hOutput);
	}

	AgentDestroyRunContext(hactxRunContext);
	AgentClose(haAgent);
	pcrqRQ->CompletionStatus = ACTION_COMPLETED;
	return;
}

//  IsPrivateDesignNote
//
//  This function will determine if the passed note (identified by NOTEID) is a private design note.
//
//  Parameters
//
//  1  -  PartCopyRequest * -  Pointer to the current Part Copy Request
//  2  -  NOTEID			-  NOTEID of the design note
//  3  -  STATUS *			-  Pointer to a status to record any open error
//  4  -  int				-  Identity of the thread that is making the call
//
//  Returns
//  BOOL  -  TRUE if the note is a private design element
//           FALSE if the note is NOT a private design element
//
//  API Use Notes:
//
//
BOOL DbCopier::IsPrivateDesignNote(PartCopyRequest *pcrqRQ, NOTEID nidNote, STATUS *stOpenErr, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLSource;						//  Local Thread mapped source database handle
	NOTEHANDLE			hnNote;								//  Handle of the note
	BOOL				bEntryAdded;						//  Table maintenance flag
	WORD				wNoteClass;							//  Note Class
	WORD				wRetLen;							//  Length of returned flags
	int					iIndex;								//  Index into the design flags
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer
	char				szFlags[MAX_FLAGS + 1];				//  Design Flags

	*stOpenErr = NOERROR;

	//  Map the database handle to use
	hdbTLSource = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbSource, iThreadID);

	//  Determine if the note is a private design note - also post any folders for later processing
	stAPIRC = NSFNoteOpen(hdbTLSource, nidNote, 0, &hnNote);
	if (ERR(stAPIRC) != NOERROR)
	{
		*stOpenErr = ERR(stAPIRC);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5022E, nidNote);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);			
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		return FALSE;
	}
	//  Determine if the note is a folder
	NSFNoteGetInfo(hnNote, _NOTE_CLASS, &wNoteClass);
	if (wNoteClass & NOTE_CLASS_VIEW)
	{
		wRetLen = NSFItemGetText(hnNote, DESIGN_FLAGS, szFlags, sizeof(szFlags));
		if (wRetLen > 0)
		{
			for (iIndex = 0; iIndex < wRetLen; iIndex++)
			{
				if (szFlags[iIndex] == DESIGN_FLAG_FOLDER_VIEW)
				{
					//  Post the noteID to the Folder table for later processing
					stAPIRC = IDInsert(pcrqRQ->Parent->hitFolderTable, nidNote, &bEntryAdded);
					if (ERR(stAPIRC) != NOERROR)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5023E, nidNote);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
						xeLocal->GetAPIMessage(stAPIRC, szMsg);
						ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
					}
					else
					{
						if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
						{
							sprintf_s(szMsg, MAX_MSG, MSG_DXR5024I, nidNote);
							if (pcrqRQ->Parent->bAllowProxyFolderMessages) ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
							else xeLocal->LogMessage(szMsg, iThreadID);
						}
					}
				}
			}
		}		
	}

	NSFNoteClose(hnNote);

	//  Check for private design note
	if (wNoteClass & NOTE_CLASS_PRIVATE) return TRUE;
	return FALSE;
}

//  PerformMaintainFolders
//
//  This function will trigger the copying of the content of every folder in the target database.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::PerformMaintainFolders(PartCopyRequest *pcrqRQ, int iThreadID)
{
	int					iWarningsIssued = 0;				//  Number of warnings issued
	int					iRejoinStatus;						//  Return code from rejoining requests
	BOOL				bGetFirst;							//  IDScan control flag
	NOTEID				nidRetNID;							//  NoteID returned from IDScan
	PartCopyRequest		*pcrqMaintainFolder;				//  Request to copy Notes
	DWORD				dwFoldersScanned;					//  Folders Scanned
	DWORD				dwFoldersProcessed;					//  Folders Processed
	DWORD				dwFolderEntries;					//  Number of Folder entries
	char				szMsg[MAX_MSG];						//  Generic message buffer

	pcrqRQ->CompletionStatus = ACTION_INPROGRESS;

	//  Dispatch a request to maintain each folder in turn
	bGetFirst = TRUE;
	dwFoldersScanned = 0;
	dwFoldersProcessed = 0;
	dwFolderEntries = 0;

	while (IDScan(pcrqRQ->Parent->hitFolderTable, bGetFirst, &nidRetNID))
	{
		bGetFirst = FALSE;
		dwFoldersScanned++;

		if (nidRetNID & RRV_DELETED)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5040I, nidRetNID);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		}
		else
		{
			pcrqMaintainFolder = new PartCopyRequest(pcrqRQ->Parent);
			pcrqMaintainFolder->CopyAction = COPY_MAINTAIN_FOLDER;
			pcrqMaintainFolder->nidToCopy[0] = nidRetNID;
			pcrqMaintainFolder->CopyNoteCount = 1;

			//  Post a request to execute the copy action
			if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
										pcrqRQ->Parent,
										this,
										pcrqMaintainFolder,
										RQATTR_REJOIN,
										10,
										iThreadID))
			{
				//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
				ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5106E, iThreadID);			
				pcrqRQ->CompletionStatus = ACTION_FAILED;
				return;
			}

			//  Retrieve any completed requests - this is necessary to revent the RQE rejoin pool from flooding
			while(xeLocal->GetRejoinRequest(pcrqRQ->Parent, (void **) &pcrqMaintainFolder, iThreadID) == RJR_RETURNED)
			{
				//  Accumulate the results
				dwFoldersProcessed++;
				dwFolderEntries += pcrqMaintainFolder->dwNotesCopied;
				if (pcrqMaintainFolder->CompletionStatus == ACTION_WARN) iWarningsIssued++;
				
				//  Clean up the request
				delete pcrqMaintainFolder;
			}
			pcrqMaintainFolder = NULL;	
		}
	}

	//  Wait until all of the posted requests have completed
	iRejoinStatus = xeLocal->GetRejoinRequest(pcrqRQ->Parent, (void **) &pcrqMaintainFolder, iThreadID);
	while (iRejoinStatus != RJR_NONE_EXIST)
	{
		if (iRejoinStatus == RJR_RETURNED)
		{
			//  Accumulate the results
			dwFoldersProcessed++;
			dwFolderEntries += pcrqMaintainFolder->dwNotesCopied;
			if (pcrqMaintainFolder->CompletionStatus == ACTION_WARN) iWarningsIssued++;
			
			//  Clean up the request
			delete pcrqMaintainFolder;
			pcrqMaintainFolder = NULL;	
		}
		else
		{
			xeLocal->AppWait(PXR_BUSY_WAIT);
		}
		iRejoinStatus = xeLocal->GetRejoinRequest(pcrqRQ->Parent, (void **) &pcrqMaintainFolder, iThreadID);
	}

	//  Pass back the processing statistics
	pcrqRQ->dwNotesCopied = dwFolderEntries;
	pcrqRQ->dwStubsCopied = dwFoldersProcessed;
	pcrqRQ->CompletionStatus = ACTION_COMPLETED;
	return;
}

//  FilteredFolderDocAdd
//
//  This function will add document entries to the specified folder while removing any response documents
//  if there is a respons hierarchy in the folder.
//
//  Parameters
//
//  1  -  PartCopyRequest * -  Pointer to the current Part Copy Request
//  2  -  DBHANDLE			-  Handle of the target database
//  3  -  NOTEID			-  NOTEID of the target folder
//  4  -  DHANDLE			-  Handle of the Note ID Table with the notes to be added to the folder
//  5  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  STATUS			-  API Return Code
//
//  API Use Notes:
//
//
STATUS DbCopier::FilteredFolderDocAdd(PartCopyRequest *pcrqRQ, DBHANDLE hdbTarget, NOTEID nidTargetFolder, DHANDLE hitDocs, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLTarget;						//  Thread Local DB Handle
	NOTEHANDLE			hnFolderEntry;						//  Handle of the document
	DHANDLE				hitResponses;						//  Handle of the response documents
	NOTEID				nidEntry;							//  NoteID of folder entry
	NOTEID				nidParent;							//  NoteID of any parent (main) document
	BOOL				bGetFirst;							//  IDScan control flag
	BOOL				bEntryAdded;						//  Flag indicating if folder entry was added
	DWORD				dwFolderEntries;					//  Number of entries in the folder
	DWORD				dwResponses;						//  Number of responses removed
	DWORD				dwDeleted;							//  Count of deleted folder entries
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	dwFolderEntries = 0;
	dwResponses = 0;
	dwDeleted = 0;

	//  Map the database handle to use
	hdbTLTarget = xeLocal->GetMappedDBH(hdbTarget, iThreadID);

	//  Create a new table to contain the response documents
	stAPIRC = IDCreateTable(sizeof(NOTEID), &hitResponses);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5053E, iThreadID);
		return(ERR(stAPIRC));
	}

	bGetFirst = TRUE;

	//  Iterate the folder contents collecting any responses
	while (IDScan(hitDocs, bGetFirst, &nidEntry))
	{
		dwFolderEntries++;
		bGetFirst = FALSE;

		//  Skip the note if it has been deleted
		if (nidEntry & RRV_DELETED)
		{
			dwDeleted++;
		}
		else
		{
			//  Open the note pointed to by the folder entry
			stAPIRC = NSFNoteOpenExt(hdbTLTarget, nidEntry, OPEN_RAW_MIME, &hnFolderEntry);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5054E, nidEntry);
				ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
				IDDestroyTable(hitResponses);
				return(ERR(stAPIRC));
			}

			//  Determine if the note is a response
			NSFNoteGetInfo(hnFolderEntry, _NOTE_PARENT_NOTEID, &nidParent);
			NSFNoteClose(hnFolderEntry);

			//  If the document is a response then add it to the table
			if (nidParent != 0)
			{
				dwResponses++;
				stAPIRC = IDInsert(hitResponses, nidEntry, &bEntryAdded);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5055E, nidEntry);
					ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
					IDDestroyTable(hitResponses);
					return(ERR(stAPIRC));
				}
			}
		}
	}

	//  If any responses were found then remove those note IDs from the table
	if (IDEntries(hitResponses) == 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5056I, nidTargetFolder);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
	}
	else
	{
		stAPIRC = IDDeleteTable(hitDocs, hitResponses);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5057E, nidTargetFolder);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
			IDDestroyTable(hitResponses);
			return(ERR(stAPIRC));
		}
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5058I, dwResponses, nidTargetFolder);
	ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);

	//  Add the filtered table to the folder
	IDDestroyTable(hitResponses);
	stAPIRC = FolderDocAdd(hdbTLTarget, NULLHANDLE, nidTargetFolder, hitDocs, 0);
	return stAPIRC;
}

//  WrapperAddProfileToIDT
//
//  This function is a callback from enumerating the profile notes in the database, these are added to the 
//  Note ID Table of design notes.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the source database
//  2  -  void *			-  Pointer to this instance of the DbCopier class
//  3  -  char *			-  Pointer to the profile name
//  4  -  WORD				-  Length of the profile name
//  5  -  char *			-  Pointer to the profile user name
//  6  -  WORD				-  Length of the profile user name
//  7  -  NOTEID			-  Note ID of the profile note
//
//  Returns
//
//  STATUS			-  API Return Code
//
//  API Use Notes:
//
//
STATUS LNPUBLIC DbCopier::WrapperAddProfileToIDT(DBHANDLE hdbSrc, void *lpDBC, char *sProfName, WORD wProfNameLen, char *sUserName, WORD wUserNameLen, NOTEID nidProf)
{
	PartCopyRequest		*pcrqRQ;
	DbCopier			*dbcCurrent;

	pcrqRQ = (PartCopyRequest *) lpDBC;
	dbcCurrent = (DbCopier *) pcrqRQ->dbcCopier;
	return dbcCurrent->AddProfileToIDT(pcrqRQ, nidProf);
}

//  AddProfileToIDT
//
//  This function will add an individual profile note ID to the Note ID Table of design notes.
//
//  Parameters
//
//  1  -  PartCopyRequest * -  Pointer to the current Part Copy Request
//  2  -  NOTEID			-  Note ID of the profile note
//
//  Returns
//
//  STATUS			-  API Return Code
//
//  API Use Notes:
//
//
STATUS DbCopier::AddProfileToIDT(PartCopyRequest *pcrqRQ, NOTEID nidProf)
{
	BOOL			bRetFlag;

	return IDInsert(pcrqRQ->Parent->hitDesignTable, nidProf, &bRetFlag);
}

//  DisableReplication
//
//  This function will Disable replication in the target database.
//
//  Parameters
//
//  1  -  CopyRequest *     -  Pointer to the current Copy Request
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
void DbCopier::DisableReplication(CopyRequest *crRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
    DBREPLICAINFO		riReplicaSettings;					//  Replication settings
	DBHANDLE			hdbTLTarget;						//  Thread Local D/B handle
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	hdbTLTarget = xeLocal->GetMappedDBH(crRQ->hdbTarget, iThreadID);

	stAPIRC = NSFDbReplicaInfoGet(hdbTLTarget, &riReplicaSettings);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5062E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		return;
	}

	riReplicaSettings.Flags = riReplicaSettings.Flags | REPLFLG_DISABLE;

	stAPIRC = NSFDbReplicaInfoSet(hdbTLTarget, &riReplicaSettings);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5063E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		return;
	}

	ProxyLogMessage(crRQ, MSG_DXR5192I, iThreadID);
	return;
}

//  ReenableReplication
//
//  This function will Re-enable replication in the target database.
//
//  Parameters
//
//  1  -  CopyRequest *     -  Pointer to the current Copy Request
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
void DbCopier::ReenableReplication(CopyRequest *crRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
    DBREPLICAINFO		riReplicaSettings;					//  Replication settings
	DBHANDLE			hdbTLTarget;						//  Thread Local D/B handle
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	hdbTLTarget = xeLocal->GetMappedDBH(crRQ->hdbTarget, iThreadID);

	stAPIRC = NSFDbReplicaInfoGet(hdbTLTarget, &riReplicaSettings);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5062E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		return;
	}

	riReplicaSettings.Flags = riReplicaSettings.Flags ^ REPLFLG_DISABLE;

	stAPIRC = NSFDbReplicaInfoSet(hdbTLTarget, &riReplicaSettings);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5063E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		return;
	}

	ProxyLogMessage(crRQ, MSG_DXR5064I, iThreadID);
	return;
}

//  ConditionACL
//
//  This function will add the specified group to the ACL and set on or off the enforce consistent ACL setting.
//
//  Parameters
//
//  1  -  char *			-  Pointer to the group name to be added to the ACL
//  2  -  BOOL				-  The required setting of the enforce consistent ACL attribute
//  3  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::ConditionACL(char *szGroup, CopyRequest *crRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLTarget;						//  Thread Local D/B Handle
	DHANDLE				haclCurrent;						//  Handle of the Database ACL
	ACL_PRIVILEGES		apNULL = {0,0,0,0,0,0,0,0,0,0};		//  Default ACL privileges
	DWORD				dwACLFlags;							//  ACL Flag Settings
	WORD				wAccessFlags;						//  ACL Entry Access Flags
	BOOL				bACLUpdated;						//  Switch indicating if ACL has been updated
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	hdbTLTarget = xeLocal->GetMappedDBH(crRQ->hdbTarget, iThreadID);

	//  Read the current ACL
	stAPIRC = NSFDbReadACL(hdbTLTarget, &haclCurrent);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5065E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		return;
	}

	bACLUpdated = FALSE;

	//  If there was an ACL group specified then process it
	if (szGroup[0] != '\0')
	{
		//  Get the Privilege array for all roles in the database
		SetAllRoles(haclCurrent, &apNULL);

		//  Add the requested group to the target database ACL as manager
		wAccessFlags = ACL_FLAG_GROUP | ACL_FLAG_CREATE_FOLDER | ACL_FLAG_CREATE_LOTUSSCRIPT | ACL_FLAG_PUBLICREADER | ACL_FLAG_PUBLICWRITER;
		stAPIRC = ACLAddEntry(haclCurrent, szGroup, ACL_LEVEL_MANAGER, &apNULL, wAccessFlags);
		if (ERR(stAPIRC) == ERR_ACL_INLIST)
		{
			//  Entry was already in the list - issue warning
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5066W, szGroup);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
		else
		{
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5067E, szGroup);
				ProxyLogMessage(crRQ, szMsg, iThreadID);
				xeLocal->GetAPIMessage(stAPIRC, szMsg);
				ProxyLogMessage(crRQ, szMsg, iThreadID);
			}
			else
			{
				if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE) 
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5068I, szGroup);
					ProxyLogMessage(crRQ, szMsg, iThreadID);
				}
				bACLUpdated = TRUE;
			}
		}
	}

	//  If requested then turn off the 'enforce consistent ACL' setting
	if (crRQ->ConsistentACLOff)
	{
		//  Read the ACL Flags
		stAPIRC = ACLGetFlags(haclCurrent, &dwACLFlags);
		if (ERR(stAPIRC) != NOERROR)
		{
			ProxyLogMessage(crRQ, MSG_DXR5069E, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
		else
		{
			//  Check if the enforce consistent ACL flag is on
			if (dwACLFlags & ACL_UNIFORM_ACCESS)
			{
				//  Turn the flag off
				dwACLFlags = (dwACLFlags ^ ACL_UNIFORM_ACCESS);
			
				stAPIRC = ACLSetFlags(haclCurrent, dwACLFlags);
				if (ERR(stAPIRC) != NOERROR)
				{
					ProxyLogMessage(crRQ, MSG_DXR5070E, iThreadID);
					xeLocal->GetAPIMessage(stAPIRC, szMsg);
					ProxyLogMessage(crRQ, szMsg, iThreadID);
				}
				else
				{
					if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE) ProxyLogMessage(crRQ, MSG_DXR5071I, iThreadID); 
					bACLUpdated = TRUE;
				}
			}
		}
	}

	//  Update the ACL in the database - if necessary
	if (bACLUpdated)
	{
		stAPIRC = NSFDbStoreACL(hdbTLTarget, haclCurrent, 0l, 0);
		if (ERR(stAPIRC) != NOERROR)
		{
			ProxyLogMessage(crRQ, MSG_DXR5072E, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
	}
	return;
}

//  ShowCopyStats
//
//  This function will show the throughput stats from the copy.
//
//  Parameters
//
//  1  -  CopyRequest *		-  Pointer to the CopyRequest object for this operation
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::ShowCopyStats(CopyRequest *crRQ, int iThreadID)
{
	DWORD				dwSizeUsed;							//  Calculated size for throughput calcs
	long				lThroughput;						//  Throughput
	long				lTimeFactor = 1000;		//  Clock tick conversion factor
	char				szLT[2];							//  Qualifier for zero second timing
	char				szGT[2];							//  Qualifier for zero second timing
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Design stats
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5132I, crRQ->dwDesignDocsCopied,
								crRQ->dwDesignStubsCopied,
								crRQ->dwDesignRetries,
								crRQ->dwDesignAccessErrors,
								crRQ->dwDesignOtherErrors,
									crRQ->dwCopyPhaseElapsed);
	ProxyLogMessage(crRQ, szMsg, iThreadID);

	//  Data stats
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5133I, crRQ->dwDataDocsCopied,
								crRQ->dwDataStubsCopied,
								crRQ->dwDataRetries,
								crRQ->dwDataAccessErrors,
								crRQ->dwDataOtherErrors,
								crRQ->dwCopyPhaseElapsed);
	ProxyLogMessage(crRQ, szMsg, iThreadID);

	//  Compute and display the copy phase only throughput figures
	//  Defend against zero divides when total elapsed time is less than a second
	szLT[0] = ' ';
	szGT[0] = ' ';
	szLT[1] = '\0';
	szGT[1] = '\0';
	if (crRQ->dwCopyPhaseElapsed == 0)
	{
		szLT[0] = '<';
		szGT[0] = '>';
		crRQ->dwCopyPhaseElapsed = 1;
	}

	//  Compute and display the throughput in documents per second
	lThroughput = (crRQ->dwDataDocsCopied + crRQ->dwDataStubsCopied + crRQ->dwDesignDocsCopied + crRQ->dwDesignStubsCopied) / crRQ->dwCopyPhaseElapsed;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5134I, szLT, crRQ->dwCopyPhaseElapsed, szGT, lThroughput);
	ProxyLogMessage(crRQ, szMsg, iThreadID);

	//  Compute and display the throughput in Kb per second
	dwSizeUsed = crRQ->dwSizeKB;
	lThroughput = dwSizeUsed / crRQ->dwCopyPhaseElapsed;
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5135I, szLT, crRQ->dwCopyPhaseElapsed, szGT, lThroughput);
	ProxyLogMessage(crRQ, szMsg, iThreadID);

	return;
}

//  ProxyLogMessage
//
//  The defulat implementation of this virtual function writes the logged messages to the normal runtime logger.
//
//  Parameters
//
//  1  -  CopyRequest *		-  Pointer to the current copy request
//  2  -  char *			-  Pointer to the message to be logged
//  3  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL					-  TRUE if the log action succeeded
//							-  FALSE if the log action failed
//
//  API Use Notes:
//
//
BOOL DbCopier::ProxyLogMessage(CopyRequest *crRQ, char *szMsg, int iThreadID)
{
	return xeLocal->LogMessage(szMsg, iThreadID);
}

//  ComputeElapsed
//
//  This function will compute and return an elapsed time in seconds from two clock values
//
//  Parameters
//
//  1  -  clock_t *			-  Start Time
//  2  -  clock_t *			-  End Time
//
//  Returns
//  DWORD					-  Elapsed time in seconmds
//
//  API Use Notes:
//
//
DWORD DbCopier::ComputeElapsed(clock_t *tStart, clock_t *tEnd)
{
	long				lElapsed;							//  Computed elapsed time
	long				lTimeFactor = 1000;		//  Ticks/Seconds conversion factor

	lElapsed = (long) (*tEnd - *tStart) / lTimeFactor;
	return (DWORD) lElapsed;
}

//  VerifyReplicaID
//
//  This function will (if needed) verify the replica ID of the source database against the ID provided
//  in the copy request.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the source database
//  2  -  CopyRequest *		-  Pointer to the current CopyRequest object
//  3  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL					-  TRUE if the replica ID match (or the check is not needed)
//							-  FALSE if the replica ID does not match the one in the request
//
//  API Use Notes:
//
BOOL DbCopier::VerifyReplicaID(DBHANDLE hdbSrc, CopyRequest *crCurrent, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API calls
    DBREPLICAINFO		riReplicaSettings;					//  Replica Info from the source database
	char				szREPID[MAX_REPID + 1];				//  Replica ID
	char				szSuppliedID[MAX_REPID + 1];		//  The supplied replica ID
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer
	
	//  Check if the replica ID needs checking
	if (crCurrent->szSourceDBReplicaID[0] == '\0')
	{
		xeLocal->LogVerbose(MSG_DXR5140I, iThreadID);
		return TRUE;
	}

	//  Verify the replica ID
	stAPIRC = NSFDbReplicaInfoGet(hdbSrc, &riReplicaSettings);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crCurrent, MSG_DXR5141E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crCurrent, szMsg, iThreadID);
		return FALSE;
	}

	sprintf_s(szREPID, MAX_REPID, FORMAT_REPID, riReplicaSettings.ID.Innards[1], riReplicaSettings.ID.Innards[0]);

	//  Format the supplied replica ID
	if (strlen(crCurrent->szSourceDBReplicaID) == 16)
	{
		strcpy_s(szSuppliedID, MAX_REPID, crCurrent->szSourceDBReplicaID);
		szSuppliedID[8] = ':';
		szSuppliedID[9] = '\0';
		strcat_s(szSuppliedID, MAX_REPID, crCurrent->szSourceDBReplicaID + 8);
	}
	else strcpy_s(szSuppliedID, MAX_REPID, crCurrent->szSourceDBReplicaID);

	if (strcmp(szSuppliedID, szREPID) != 0)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5142E, szREPID, crCurrent->szSourceDBReplicaID);
		ProxyLogMessage(crCurrent, szMsg, iThreadID);
		return FALSE;
	}
	return TRUE;
}

//  DisableAgents
//
//  This function will disable the backgroud agnets in the target database
//
//  Parameters
//
//  1  -  CopyRequest *		-  Pointer to the current Copy Request
//  2  -  DBHANDLE			-  Handle of the target database
//  3  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::DisableAgents(CopyRequest *crRQ, DBHANDLE hdbTarget, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API calls
    DBREPLICAINFO		riReplicaSettings;					//  Replica Info from the source database
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Get the replication information
	stAPIRC = NSFDbReplicaInfoGet(hdbTarget, &riReplicaSettings);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5141E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		return;
	}

	//  Disable agents
	riReplicaSettings.Flags = riReplicaSettings.Flags | REPLFLG_NO_CHRONOS;

	//  Set the replication info settings again
	stAPIRC = NSFDbReplicaInfoSet(hdbTarget, &riReplicaSettings);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5141E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		return;
	}

	ProxyLogMessage(crRQ, MSG_DXR5144I, iThreadID);
	return;
}

//  SetAllRoles
//
//  This function update an array of PRIVILEGES with the roles that are available in a D/B ACL.
//
//  Parameters
//
//  1  -  DHANDLE			-  Handle of the database ACL
//  2  -  ACL_PRIVILEGES *	-  Pointer to the privileges array that will be updated
//
//  Returns
//
//  int		-  Count of the roles that were set in the array
//
//  API Use Notes:
//
//
int DbCopier::SetAllRoles(DHANDLE hACL, ACL_PRIVILEGES *apCurrent)
{
	STATUS				stAPIRC;							//  Return code from Notes API calls
	WORD				wPrivNum;							//  The Privilege Number to Get/Set
	int					iRolesSet = 0;						//  Count of the roles set
	ACL_PRIVILEGES		apLocal = {0,0,0,0,0,0,0,0,0,0};	//  Privilege array
	char				szRoleName[ACL_PRIVSTRINGMAX];		//  Returned Role Name

	wPrivNum = 5;											//  Privilege number of the first Role 

	memcpy(&apLocal, apCurrent, sizeof(ACL_PRIVILEGES));	//  Copy the supplied priv array

	stAPIRC = ACLGetPrivName(hACL, wPrivNum, szRoleName);
	while ((ERR(stAPIRC) == NOERROR) &&( strlen(szRoleName) > 0))
	{
		ACLSetPriv(apLocal, wPrivNum);
		iRolesSet++;
		wPrivNum++;
		stAPIRC = ACLGetPrivName(hACL, wPrivNum, szRoleName);
	}
	if (iRolesSet > 0)
	{
		memcpy(apCurrent, &apLocal, sizeof(ACL_PRIVILEGES)); //  Copy back the priv array
	}
	return iRolesSet;
}

//  BuildFolderIDTable
//
//  This function will build a vaid note ID table for a folder from the native ID table object
//
//  Parameters
//
//  1  -  PartCopyRequest * -  Pointer to the Current Part Copy Request
//  2  -  DBHANDLE			-  Handle of the source database
//  3  -  NOTEID			-  Note ID of the Folder Note
//  4  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  HANDLE		-  HANDLE of the validated Note ID Table
//
//  API Use Notes:
//
//
DHANDLE DbCopier::BuildFolderIDTable(PartCopyRequest *pcrqRQ, DBHANDLE hdbSource, NOTEID nidFolder, int iThreadID)
{
	STATUS					stAPIRC;							//  Return code from notesAPI calls
	NOTEHANDLE				hnFolder;							//  Handle to the folder note
	BLOCKID					bidItem;							//  Block ID of the item
	BLOCKID					bidValue;							//  Block ID of the item value
	WORD					wItemType;							//  Item Type
	DWORD					dwValueLen;							//  Value Length
	UCHAR					*pValue;							//  Pointer to the item value
	UCHAR					*pObjDesc;							//  Pointer to the Object Descriptor
	DHANDLE					hIDT;								//  Handle to the object
	OBJECT_DESCRIPTOR		odIDT;								//  Object Descriptor
	BOOL					bGetFirst;							//  Control for table iteration
	BOOL					bEntryAdded;						//  Entry was added to the table
	DHANDLE					hVIDT;								//  Validated Note ID Table
	DWORD					dwNativeFolderEntries;				//  Count of entries in the native ID table
	DWORD					dwValidatedFolderEntries;			//  Count of entries in the vaidated ID table
	DWORD					dwInvalidFolderEntries;				//  Count of invalid entries from the native ID table
	DWORD					dwFolderStubs;						//  Count of deletion stubs
	DWORD					dwAddFailed;						//  Count of additions that failed
	NOTEID					nidRetNID;							//  Returned Note ID
	OID						oidEntry;							//  OID of the folder entry
	TIMEDATE				tdMod;								//  Modification timestamp of the folder entry
	WORD					wNoteClass;							//  Class of note of the folder entry
	char					szMsg[MAX_MSG];						//  Generic Message Buffer

	//  Step 1 - open the folder note and obtain the Object Descriptor for the $FolderIDTable object
	stAPIRC = NSFNoteOpenExt(hdbSource, nidFolder, OPEN_WITH_FOLDERS, &hnFolder);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5042E, nidFolder);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Attempt to find the $FolderIDTable item
	stAPIRC = NSFItemInfo(hnFolder, VIEW_FOLDER_IDTABLE, (WORD) strlen(VIEW_FOLDER_IDTABLE),
							&bidItem,
							&wItemType,
							&bidValue,
							&dwValueLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5150E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		NSFNoteClose(hnFolder);
		return NULLHANDLE;
	}
	
	//  Check that the data type is TYPE_OBJECT
	if (wItemType != TYPE_OBJECT)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5151E, iThreadID);
		NSFNoteClose(hnFolder);
		return NULLHANDLE;
	}

	//  Lock and obtain a pointer to the object
	pValue = OSLockBlock(UCHAR, bidValue);

	//  Extract the object descriptor - use ODSReadMemory as the data is canonical
	pObjDesc = pValue + sizeof(WORD);
	ODSReadMemory(&pObjDesc, _OBJECT_DESCRIPTOR, &odIDT, 1);

	//  Free the block
	OSUnlockBlock(bidValue);

	//  Close the folder note as we no longer need it
	NSFNoteClose(hnFolder);

	//  Step 2 - Read the $FolderIDTable object into memory
	stAPIRC = NSFDbReadObject(hdbSource, odIDT.RRV, 0, MAXDWORD, &hIDT);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5052E, odIDT.RRV);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Step 3 - Create an empty ID table to hold the validated IDs found in the object
	stAPIRC = IDCreateTable(sizeof(NOTEID), &hVIDT);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5053E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		return NULLHANDLE;		
	}

	//  Step 4 - Iterate over the Native table and add valid documents to the validated table
	bGetFirst = TRUE;
	dwNativeFolderEntries = 0;
	dwValidatedFolderEntries = 0;
	dwInvalidFolderEntries = 0;
	dwFolderStubs = 0;
	dwAddFailed = 0;

	while (IDScan(hIDT, bGetFirst, &nidRetNID))
	{
		bGetFirst = FALSE;
		dwNativeFolderEntries++;

		//  Skip any notes that are deleted
		if (nidRetNID & RRV_DELETED)
		{
			dwFolderStubs++;
		}
		else
		{
			//  Get the UNID of the document in the source database from it's NotesID
			stAPIRC = NSFDbGetNoteInfo(hdbSource, nidRetNID, &oidEntry, &tdMod, &wNoteClass);
			if (ERR(stAPIRC) != NOERROR)
			{
				//  Warn (TRACE) that the note will be dropped from the vaidated table
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5154E, nidRetNID);
				xeLocal->LogTrace(TRACE_DBC, szMsg);
				xeLocal->GetAPIMessage(stAPIRC, szMsg);
				xeLocal->LogTrace(TRACE_DBC, szMsg);
				dwInvalidFolderEntries++;
			}
			else
			{
				//  Note is ok .... add it to the validated table
				stAPIRC = IDInsert(hVIDT, nidRetNID, &bEntryAdded);
				if (ERR(stAPIRC) != NOERROR)
				{	
					sprintf_s(szMsg, MAX_MSG, MSG_DXR5154E, nidRetNID);
					xeLocal->LogTrace(TRACE_DBC, szMsg);
					xeLocal->GetAPIMessage(stAPIRC, szMsg);
					xeLocal->LogTrace(TRACE_DBC, szMsg);
					dwAddFailed++;
				}
				else
				{
					if (!bEntryAdded)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR5154E, nidRetNID);
						xeLocal->LogTrace(TRACE_DBC, szMsg);
						dwAddFailed++;
					}
					else dwValidatedFolderEntries++;
				}
			}
		}
	}

	//  Free the native ID Table
	OSMemFree(hIDT);

	//  Scan completed - report on the results
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5156I, dwNativeFolderEntries, dwFolderStubs);
	ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5157I, dwInvalidFolderEntries, dwAddFailed);
	ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5158I, dwValidatedFolderEntries);
	ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);

	return hVIDT;
}

//  PurgeFolderContent
//
//  This function will purge any content from an existing folder, this is used as a prt of the recovery
//  mechanism for damaged folders.
//
//  Parameters
//
//  1  -  PartCopyRequest * -  Pointer to the current Part Copy Request
//  2  -  DBHANDLE			-  Handle of the target database
//  3  -  NOTEID			-  Note ID of the Folder Note
//  4  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  STATUS  -				NOERROR - purged ok, <other> API error status
//
//  API Use Notes:
//
//
STATUS DbCopier::PurgeFolderContent(PartCopyRequest *pcrqRQ, DBHANDLE hdbTarget, NOTEID nidFolder, int iThreadID)
{
	STATUS					stAPIRC;							//  Return code from notesAPI calls
	NOTEHANDLE				hnFolder;							//  Handle to the folder note
	BLOCKID					bidItem;							//  Block ID of the item
	BLOCKID					bidValue;							//  Block ID of the item value
	WORD					wItemType;							//  Item Type
	DWORD					dwValueLen;							//  Value Length
	UCHAR					*pValue;							//  Pointer to the item value
	UCHAR					*pObjDesc;							//  Pointer to the Object Descriptor
	WORD					wObjClass;							//  Object Class
	WORD					wObjPrivs;							//  Object Privileges
	DWORD					dwObjSize;							//  Object Size
	DWORD					dwIDTSize;							//  Size of the ID Table
	DHANDLE					hEIDT;								//  Empty Note ID Table
	OBJECT_DESCRIPTOR		odIDT;								//  Object Descriptor
	BOOL					bContinueDumping;					//  Loop control
	int						iFolderObj;							//  Folder Object Number
	char					szMsg[MAX_MSG];						//  Generic Message Buffer

	//  Step 1 - open the folder note and obtain the Object Descriptor for the $FolderIDTable object
	stAPIRC = NSFNoteOpenExt(hdbTarget, nidFolder, OPEN_WITH_FOLDERS, &hnFolder);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5042E, nidFolder);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		return ERR_FOLDER_CORRUPT;
	}

	//  Attempt to find the $FolderIDTable item
	stAPIRC = NSFItemInfo(hnFolder, VIEW_FOLDER_IDTABLE, (WORD) strlen(VIEW_FOLDER_IDTABLE),
							&bidItem,
							&wItemType,
							&bidValue,
							&dwValueLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5150E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		NSFNoteClose(hnFolder);
		return ERR_FOLDER_CORRUPT;
	}
	
	//  Check that the data type is TYPE_OBJECT
	if (wItemType != TYPE_OBJECT)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5151E, iThreadID);
		NSFNoteClose(hnFolder);
		return ERR_FOLDER_CORRUPT;
	}

	//  Lock and obtain a pointer to the object
	pValue = OSLockBlock(UCHAR, bidValue);

	//  Extract the object descriptor - use ODSReadMemory as the data is canonical
	pObjDesc = pValue + sizeof(WORD);
	ODSReadMemory(&pObjDesc, _OBJECT_DESCRIPTOR, &odIDT, 1);

	//  Free the block
	OSUnlockBlock(bidValue);


	//  Step 2 - get the size of the referenced object
	stAPIRC = NSFDbGetObjectSize(hdbTarget, odIDT.RRV, OBJECT_UNKNOWN, &dwObjSize, &wObjClass, &wObjPrivs);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5160E, odIDT.RRV);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		NSFNoteClose(hnFolder);
		return ERR_FOLDER_CORRUPT;
	}

	//  Step 3 - Create an empty ID table to hold the validated IDs found in the object
	stAPIRC = IDCreateTable(sizeof(NOTEID), &hEIDT);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5053E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		NSFNoteClose(hnFolder);
		return ERR_FOLDER_CORRUPT;		
	}

	//  Get the size of the enpty Note ID Table
	dwIDTSize = IDTableSize(hEIDT);

	//  Step 4 - If the object is smaller than the empty ID table then reallocate it
	if (dwObjSize < dwIDTSize)
	{
		stAPIRC = NSFDbReallocObject(hdbTarget, odIDT.RRV, dwIDTSize);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5161E, odIDT.RRV, dwIDTSize);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
			NSFNoteClose(hnFolder);
			IDDestroyTable(hEIDT);
			return ERR_FOLDER_CORRUPT;
		}
	}

	//  Step 5 - Write the empty Note ID Table to the referenced object
	stAPIRC = NSFDbWriteObject(hdbTarget, odIDT.RRV, hEIDT, 0, dwIDTSize);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5162E, odIDT.RRV, dwIDTSize);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		NSFNoteClose(hnFolder);
		IDDestroyTable(hEIDT);
		return ERR_FOLDER_CORRUPT;
	}

	IDDestroyTable(hEIDT);

	//  Step 6 - Loop through deleting all of the $FolderObject[n] items

	bContinueDumping = TRUE;
	iFolderObj = 0;
	while (bContinueDumping)
	{
		bContinueDumping = PurgeFolderObject(pcrqRQ, hnFolder, iFolderObj, iThreadID);
		iFolderObj++;
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5163I, iFolderObj - 1);
	ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);

	//  Update the note
	stAPIRC = NSFNoteUpdateExtended(hnFolder, UPDATE_FORCE);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5164E, nidFolder);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
		NSFNoteClose(hnFolder);
		return ERR_FOLDER_CORRUPT;
	}

	//  Close the folder note
	NSFNoteClose(hnFolder);

	return NOERROR;	
}

//  PurgeFolderObject
//
//  This function will purge an individual $FolderObject from the Folder Note
//
//  Parameters
//
//  1  -  PartCopyRequest * -  Pointer to the currebt Part Copy Request
//  2  -  NOTEHANDLE		-  Handle of the Folder Note
//  3  -  int				-  Number siffix of the object to be removed
//  4  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL  -				TRUE if the object was removed and false if not
//
//  API Use Notes:
//
//
BOOL DbCopier::PurgeFolderObject(PartCopyRequest *pcrqRQ, NOTEHANDLE hnFolder, int iObjNo, int iThreadID)
{
	STATUS					stAPIRC;							//  Return code from notes API calls
	DBHANDLE				hdbTLTarget;						//  Thread Local D/B Handle
	BLOCKID					bidItem;							//  Block ID of the item
	BLOCKID					bidValue;							//  Block ID of the item value
	WORD					wItemType;							//  Item Type
	DWORD					dwValueLen;							//  Value Length
	UCHAR					*pValue;							//  Pointer to the item value
	UCHAR					*pObjDesc;							//  Pointer to the Object Descriptor
	OBJECT_DESCRIPTOR		odIDT;								//  Object Descriptor
	char					szObjName[MAX_ELEMENT];				//  Folder Object Name
	char					szMsg[MAX_MSG];						//  Generic Message Buffer

	hdbTLTarget = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbTarget, iThreadID);

	if (iObjNo == 0) strcpy_s(szObjName, MAX_ELEMENT, VIEW_FOLDER_OBJECT);
	else sprintf_s(szObjName, MAX_ELEMENT, "%s%i", VIEW_FOLDER_OBJECT, iObjNo);

	//  Attempt to find the $FolderObject item
	stAPIRC = NSFItemInfo(hnFolder, szObjName, (WORD) strlen(szObjName),
							&bidItem,
							&wItemType,
							&bidValue,
							&dwValueLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5165E, szObjName); 
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	//  Lock and obtain a pointer to the object
	pValue = OSLockBlock(UCHAR, bidValue);

	//  Extract the object descriptor - use ODSReadMemory as the data is canonical
	pObjDesc = pValue + sizeof(WORD);
	ODSReadMemory(&pObjDesc, _OBJECT_DESCRIPTOR, &odIDT, 1);

	//  Free the block
	OSUnlockBlock(bidValue);

	//  Free the object that the descriptor is pointing to
	stAPIRC = NSFDbFreeObject(hdbTLTarget, odIDT.RRV);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5167E, szObjName, odIDT.RRV); 
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;

	}

	stAPIRC = NSFItemDeleteByBLOCKID(hnFolder, bidItem);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5166E, szObjName); 
		xeLocal->LogMessage(szMsg);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		xeLocal->LogMessage(szMsg);
		return FALSE;
	}

	return TRUE;
}

//  PerformRestartActions
//
//  This function will perform any actions necessary to restart a request
//
//  Parameters
//
//  1  -  CopyRequest *		-  Pointer to the Copy Request object
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL  -				TRUE if actions were carried out
//
//  API Use Notes:
//
//
BOOL DbCopier::PerformRestartActions(CopyRequest *crRQ, int iThreadID)
{
	STATUS			stAPIRC;									//  API return code
	char			szMsg[MAX_MSG + 1];							//  Generic Error message

	sprintf_s(szMsg, MAX_MSG, MSG_DXR5172T, crRQ->RestartActions);
	xeLocal->LogTrace(TRACE_DBC, szMsg, iThreadID);

	//  See if the target database needs to be deleted
	if (crRQ->RestartActions & RESTART_DEL_TARGET)
	{
		stAPIRC = xeLocal->DeleteDatabase(crRQ->szTargetServer, crRQ->szTargetDatabase, FALSE, iThreadID);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5169W, crRQ->szTargetDatabase, crRQ->szTargetServer);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
		else
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5174I, crRQ->szTargetDatabase, crRQ->szTargetServer);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
	}
	
	//  See if fixup should be asserted on the source database
	if (crRQ->RestartActions & RESTART_ASSERT_FIXUP)
	{
		crRQ->ForceFixup = TRUE;
		ProxyLogMessage(crRQ, MSG_DXR5175I, iThreadID);
	}

	xeLocal->LogTrace(TRACE_DBC, MSG_DXR5173T, iThreadID);
	return TRUE;
}

//  PerformPreProcessActions
//
//  This function will perform any actions necessary to complete before the request can be processed
//
//  Parameters
//
//  1  -  CopyRequest *		-  Pointer to the Copy Request object
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL  -				TRUE if actions were carried out
//
//  API Use Notes:
//
//
BOOL DbCopier::PerformPreProcessActions(CopyRequest *crRQ, int iThreadID)
{
	STATUS			stAPIRC;									//  API return code
	char			szMsg[MAX_MSG + 1];							//  Generic Error message

	//  See if the target database needs to be deleted
	if (crRQ->PreProcessActions & PRESTART_DEL_TARGET)
	{
		stAPIRC = xeLocal->DeleteDatabase(crRQ->szTargetServer, crRQ->szTargetDatabase, FALSE, iThreadID);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5179W, crRQ->szTargetDatabase, crRQ->szTargetServer);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
		else
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5180I, crRQ->szTargetDatabase, crRQ->szTargetServer);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
	}
	
	return TRUE;
}

//  PerformReplicateDb
//
//  This function will perform a one-time replication between the source and target databases.
//
//  Parameters
//
//  1  -  CopyRequest *		-  Pointer to the Copy Request object
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL  -				TRUE if actions were carried out
//
//  API Use Notes:
//
//
BOOL DbCopier::PerformReplicateDb(CopyRequest *crRQ, int iThreadID)
{
	STATUS					stAPIRC;						//  Return code from Notes API calls
	REPLSERVSTATS			rsActivity;						//  Replication Stats
	WORD					wReplOpts;						//  Replication Options
	char					szServer[MAXUSERNAME + 1];		//  Server to replicate with
	char					szTarget[MAXPATH + 1];			//  Local target D/B Name
	char					szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Determine the direction of the replication PULL/PUSH
	wReplOpts = 0;
	if ((crRQ->szSourceServer[0] == '\0') || 
		(strcmpi(crRQ->szSourceServer, "LOCAL") == 0) || 
		(strcmpi(crRQ->szSourceServer, xeLocal->AbbreviatedUser) == 0))
	{
		//  Source is Local -  check that the destination is not also local
		if ((crRQ->szTargetServer[0] == '\0') || 
			(strcmpi(crRQ->szTargetServer, "LOCAL") == 0) || 
			(strcmpi(crRQ->szTargetServer, xeLocal->AbbreviatedUser) == 0))
		{
			//  Target also local, this is an error
			ProxyLogMessage(crRQ, MSG_DXR5181E, iThreadID);
			return FALSE;
		}

		strcpy_s(szServer, MAXUSERNAME, crRQ->szTargetServer);
		strcpy_s(szTarget, MAXPATH, crRQ->szSourceDatabase);
		wReplOpts = REPL_OPTION_SEND_NOTES;						//  PUSH only
	}
	else
	{
		//  Source is Remote -  check that the destination is not also Remote
		if ((crRQ->szTargetServer[0] == '\0') || 
			(strcmpi(crRQ->szTargetServer, "LOCAL") == 0) || 
			(strcmpi(crRQ->szTargetServer, xeLocal->AbbreviatedUser) == 0))
		{
			strcpy_s(szServer, MAXUSERNAME, crRQ->szSourceServer);
			strcpy_s(szTarget, MAXPATH, crRQ->szTargetDatabase);
			wReplOpts = REPL_OPTION_RCV_NOTES;					//  PULL only
		}
		else
		{
			//  Target also local, this is an error
			ProxyLogMessage(crRQ, MSG_DXR5182E, iThreadID);
			return FALSE;
		}
	}

	//  Replicate the Endpoint database with the Depot Server
	stAPIRC = ReplicateWithServer(NULL, 
								szServer, 
								wReplOpts, 
								1, 
								szTarget, 
								&rsActivity);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5183E, szTarget, szServer);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		return FALSE;
	}

	//  Show the results
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5184I, szTarget, szServer);
	ProxyLogMessage(crRQ, szMsg, iThreadID);
	if (wReplOpts & REPL_OPTION_RCV_NOTES)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5185I, rsActivity.Pull.NotesAdded, rsActivity.Pull.NotesDeleted, rsActivity.Pull.NotesUpdated, rsActivity.Pull.NumberErrors);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
	}
	else
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5186I, rsActivity.Push.NotesAdded, rsActivity.Push.NotesDeleted, rsActivity.Push.NotesUpdated, rsActivity.Push.NumberErrors);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
	}

	return TRUE;
}

//  DrainRequests
//
//  This function will drain and discard all requests that are on the asynchronous execution queue
//
//  Parameters
//
//  1  -  CopyRequest * -  Pointer to he top level copy request
//  2  -  int			-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
void DbCopier::DrainRequests(CopyRequest *crRQ, int iThreadID)
{
	PartCopyRequest		*pcrqCopyNotes;						//  Copy notes request
	int					iRejoinStatus;						//  Status from Rejoin 

	iRejoinStatus = xeLocal->GetRejoinRequest(crRQ, (void **) &pcrqCopyNotes, iThreadID);
	while (iRejoinStatus != RJR_NONE_EXIST)
	{
		if (iRejoinStatus == RJR_RETURNED)
		{
			//  Clean up the request if it is valid
			if (pcrqCopyNotes != NULL)
			{
				//  Update the action in memory in case a race causes the freed object to be returned to another thread
				pcrqCopyNotes->CopyAction = COPY_FREED;
				delete pcrqCopyNotes;
				pcrqCopyNotes = NULL;	
			}
		}
		else
		{
			xeLocal->AppWait(PXR_BUSY_WAIT);
		}
		iRejoinStatus = xeLocal->GetRejoinRequest(crRQ, (void **) &pcrqCopyNotes, iThreadID);
	}
	return;
}

//  CopyDataUsingSearch
//
//  This function will copy all of the data notes from the source database to the
//  target database. The function uses the generic PerformCopyNotes() asynchronous routine to actually
//  perform the copy. This method will attempt to use NSFSearch to generate the actual copy notes 
//  requests.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
//
BOOL DbCopier::CopyDataUsingSearch(PartCopyRequest *pcrqRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	DBHANDLE			hdbTLSource;						//  Local Thread handle for the Source database
	WORD				wNoteClass;							//  Note types to be retrieved
	DWORD				dwTotalNotesToCopy;					//  Total Count of notes and profiles to copy
	TIMEDATE			tdStart;							//  Modified since time
	TIMEDATE			tdFinish;							//  Ending modification time
	PartCopyRequest		*pcrqCopyNotes;						//  Request to copy Notes
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Get the mapped thread local database handles
	hdbTLSource = xeLocal->GetMappedDBH(pcrqRQ->Parent->hdbSource, iThreadID);
	if (hdbTLSource == NULLHANDLE) return FALSE;

	//  Setup to perform the D/B Search
	//  Determine which NOTE classes must be retrieved
	wNoteClass = NOTE_CLASS_DATA;
	TimeConstant(TIMEDATE_WILDCARD, &tdStart);

	//  Log the fact that we are starting the search
	ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5195I, iThreadID);			
	//  Get the count of notes in the table
	dwTotalNotesToCopy = xeLocal->GetDataDocumentCount(hdbTLSource, iThreadID);
	//  Show the counts
	if (xeLocal->rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR5021I, dwTotalNotesToCopy);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);			
	}

	//  Invoke the DbSearch
	stAPIRC = NSFSearch(hdbTLSource,
						NULLHANDLE,
						NULL,
						0,
						wNoteClass,
						&tdStart,
						WrapperProcessSearchResult,
						pcrqRQ,
						&tdFinish);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5196E, iThreadID);			
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);			
		return FALSE;
	}

	//  Drain any completed rejoin requests from the Rejoin queue
	while(xeLocal->GetRejoinRequest(pcrqRQ->Parent, (void **) &pcrqCopyNotes, iThreadID) == RJR_RETURNED)
	{
		//  Accumulate the results
		if (pcrqCopyNotes->CopyAction == COPY_DESIGN_NOTES) 
		{
			if (xeLocal->AcquireAppMutex(DESIGN_ACC_MUTEX, iThreadID))
			{
				pcrqRQ->pcrqCopyDesign->AccumulateTotals(pcrqCopyNotes);
				xeLocal->FreeAppMutex(DESIGN_ACC_MUTEX, iThreadID);
			}
		}
		else 
		{
			if (pcrqCopyNotes->CopyAction == COPY_DATA_NOTES)
			{
				if (xeLocal->AcquireAppMutex(DATA_ACC_MUTEX, iThreadID))
				{
					pcrqRQ->pcrqCopyData->AccumulateTotals(pcrqCopyNotes);
					xeLocal->FreeAppMutex(DATA_ACC_MUTEX, iThreadID);
				}
			}
			else
			{
				//  The returned object is not of a valid type - record the error and clear the pointer
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5147E, pcrqCopyNotes, pcrqCopyNotes->CopyAction);
				ProxyLogMessage(pcrqRQ->Parent, szMsg, iThreadID);
				pcrqCopyNotes = NULL;
			}
		}
		//  Clean up the request if it is valid
		if (pcrqCopyNotes != NULL)
		{
			//  Check if there were any non-recovered errors noted in the request if so abort the copy
			if (pcrqCopyNotes->dwCopyAccessErrors > 0)
			{
				if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_ACCESS_ERRORS))
				{
					pcrqCopyNotes->CopyAction = COPY_FREED;
					delete pcrqCopyNotes;
					pcrqCopyNotes = NULL;	
					//  Signal an abort
					pcrqRQ->CompletionStatus = ACTION_ABORT;
					pcrqRQ->Parent->PermitRun = FALSE;
					return FALSE;
				}
			}
			if (pcrqCopyNotes->dwCopyOtherErrors > 0)
			{
				if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_COPY_ERRORS))
				{
					pcrqCopyNotes->CopyAction = COPY_FREED;
					delete pcrqCopyNotes;
					pcrqCopyNotes = NULL;	
					//  Signal an abort
					pcrqRQ->CompletionStatus = ACTION_ABORT;
					pcrqRQ->Parent->PermitRun = FALSE;
					return FALSE;
				}
			}
			//  Update the action in memory in case a race causes the freed object to be returned to another thread
			pcrqCopyNotes->CopyAction = COPY_FREED;
			delete pcrqCopyNotes;
			pcrqCopyNotes = NULL;	
		}
	}
	return TRUE;
}

//  WrapperProcessSearchResult
//
//  This function is the wrapper that is invoked as the callback from the NSFSearch() call.
//  It sets up addressability and calls the real ProcessSearchResult routine.
//
//  Parameters
//
//  1  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//  2  -  SEARCH_MATCH *	-  Poibter to the SERACH_MATCH strucure for the search result
//  3  -  ITEM_TABLE *		-  NULL - no summary items are returned
//
//  Returns
//
//	STATUS		-  NOERROR to continue the search, any other will terminate the search
//
//  API Use Notes:
//
//
STATUS DbCopier::WrapperProcessSearchResult(void *ObjPtr, SEARCH_MATCH *smCurrent, ITEM_TABLE *itCurrent)
{
	PartCopyRequest		*pcrqCopyData;						//  Request to copy Data
	DbCopier			*dbcCurrent;						//  The Db Copier

	pcrqCopyData = (PartCopyRequest *) ObjPtr;
	dbcCurrent = (DbCopier *) pcrqCopyData->dbcCopier;

	return dbcCurrent->ProcessSearchResult(smCurrent, pcrqCopyData);
}

//  ProcessSearchResult
//
//  This function is invoked as the callback from the NSFSearch() call.
//  It builds up a collection of Note IDs and then dispatches a Copy Notes request.
//
//  Parameters
//
//  1  -  SEARCH_MATCH *	-  Poibter to the SERACH_MATCH strucure for the search result
//  2  -  PartCopyRequest *	-  Pointer to the PartCopyRequest object for this scheduling
//
//  Returns
//
//	STATUS		-  NOERROR to continue the search, any other will terminate the search
//
//  API Use Notes:
//
//
STATUS DbCopier::ProcessSearchResult(SEARCH_MATCH *smHit, PartCopyRequest *pcrqRQ)
{
	int						iIndex;						//  Generic index
	SEARCH_MATCH			smCurrent;					//  Local copy of search match structure
	PartCopyRequest			*pcrqCopyNotes;				//  Copy Notes Request
	char					szMsg[MAX_MSG + 1];			//  Generic message buffer

	//  Copy the passed structures to local storage in case they are not word aligned
	memcpy((char *) &smCurrent, (char *) smHit, sizeof(SEARCH_MATCH));

	//  Check that the current item was matched
	if (!smCurrent.SERetFlags & SE_FMATCH) return NOERROR;

	//  Add the current match result to the passed request
	pcrqRQ->nidToCopy[pcrqRQ->CopyNoteCount] = smCurrent.ID.NoteID;
	pcrqRQ->CopyNoteCount++;

	//  If the request is not full then just return to process the next one
	if (pcrqRQ->CopyNoteCount < COPY_NOTE_MULTIPLEX) return NOERROR;

	//  Build a new Copy NOtes Request
	pcrqCopyNotes = new PartCopyRequest(pcrqRQ->Parent);
	pcrqCopyNotes->CopyAction = COPY_DATA_NOTES;
	pcrqCopyNotes->CopyNoteCount = 0;
	for (iIndex = 0; iIndex < COPY_NOTE_MULTIPLEX; iIndex++)
	{
		pcrqCopyNotes->nidToCopy[iIndex] = pcrqRQ->nidToCopy[iIndex];
		pcrqCopyNotes->CopyNoteCount++;
		pcrqRQ->nidToCopy[iIndex] = 0;
		pcrqRQ->CopyNoteCount--;
	}

	//  Now dispatch the request
	if (!xeLocal->PostARequest(PXR_WAITIF_BUSY | PXR_APP_WAIT,
								pcrqRQ->Parent,
								this,
								pcrqCopyNotes,
								RQATTR_REJOIN,
								10,
								pcrqRQ->iProcThread))
	{
		//  Should not occur (SNO) Error - as the call should block in the case of any resource shortage
		ProxyLogMessage(pcrqRQ->Parent, MSG_DXR5105E, pcrqRQ->iProcThread);			
		//  Signal an abort
		pcrqRQ->Parent->PermitRun = FALSE;
		pcrqRQ->CompletionStatus = ACTION_ABORT;
		return ERR_MQ_TIMEOUT;
	}

	//  Check for and drain any requests that are ready in the rejoin queue
	while(xeLocal->GetRejoinRequest(pcrqRQ->Parent, (void **) &pcrqCopyNotes, pcrqRQ->iProcThread) == RJR_RETURNED)
	{
		//  Accumulate the results
		if (pcrqCopyNotes->CopyAction == COPY_DESIGN_NOTES) 
		{
			if (xeLocal->AcquireAppMutex(DESIGN_ACC_MUTEX, pcrqRQ->iProcThread))
			{
				pcrqRQ->pcrqCopyDesign->AccumulateTotals(pcrqCopyNotes);
				xeLocal->FreeAppMutex(DESIGN_ACC_MUTEX, pcrqRQ->iProcThread);
			}
		}
		else 
		{
			if (pcrqCopyNotes->CopyAction == COPY_DATA_NOTES)
			{
				if (xeLocal->AcquireAppMutex(DATA_ACC_MUTEX, pcrqRQ->iProcThread))
				{
					pcrqRQ->pcrqCopyData->AccumulateTotals(pcrqCopyNotes);
					xeLocal->FreeAppMutex(DATA_ACC_MUTEX, pcrqRQ->iProcThread);
				}
			}
			else
			{
				//  The returned object is not of a valid type - record the error and clear the pointer
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5147E, pcrqCopyNotes, pcrqCopyNotes->CopyAction);
				ProxyLogMessage(pcrqRQ->Parent, szMsg, pcrqRQ->iProcThread);
				pcrqCopyNotes = NULL;
			}
		}
		//  Clean up the request if it is valid
		if (pcrqCopyNotes != NULL)
		{
			//  Check if there were any non-recovered errors noted in the request if so abort the copy
			if (pcrqCopyNotes->dwCopyAccessErrors > 0)
			{
				if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_ACCESS_ERRORS))
				{
					pcrqCopyNotes->CopyAction = COPY_FREED;
					delete pcrqCopyNotes;
					pcrqCopyNotes = NULL;	
					//  Signal an abort
					pcrqRQ->CompletionStatus = ACTION_ABORT;
					pcrqRQ->Parent->PermitRun = FALSE;
					return ERR_NET_CMD_CANCELLED;
				}
			}
			if (pcrqCopyNotes->dwCopyOtherErrors > 0)
			{
				if (!(pcrqRQ->Parent->FaultTolerance & ALLOW_COPY_ERRORS))
				{
					pcrqCopyNotes->CopyAction = COPY_FREED;
					delete pcrqCopyNotes;
					pcrqCopyNotes = NULL;	
					//  Signal an abort
					pcrqRQ->CompletionStatus = ACTION_ABORT;
					pcrqRQ->Parent->PermitRun = FALSE;
					return ERR_NET_CMD_CANCELLED;
				}
			}
			//  Update the action in memory in case a race causes the freed object to be returned to another thread
			pcrqCopyNotes->CopyAction = COPY_FREED;
			delete pcrqCopyNotes;
			pcrqCopyNotes = NULL;	
		}
	}
	return NOERROR;
}

//  prepareTargetDb
//
//  This function will prepare the target database for copying, by applying the needed srettings.
//
//  Parameters
//
//	1  -  CopyRequest *		-  Pointer to the copy request
//  2  -  DBHANDLE			-  Handle of the source database
//  3  -  DBHANDLE			-  Handle of the target database
//  4  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  BOOL					-  TRUE if the database was prepared
//							-  FALSE if the database cannot be prepared
//
//  API Use Notes:
//
BOOL DbCopier::prepareTargetDb(CopyRequest *crRQ, DBHANDLE hdbSrc, DBHANDLE hdbTgt, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	NOTEHANDLE			hnSrcIcon,hnTgtIcon;				//  Handles to the source and target Db Icon notes
	NOTEID				nidIcon;							//  Note ID of the Icon Note
	WORD				wClass;								//  Note class
	WORD				wRetLen;							//  Length of the flags
	DWORD				DBO1,DBO2,DBO3,DBO4;				//  Very Extended Database options
	char				szFlags[26 + 26 + 10 + 20];			//  Database Flags
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Get the Db Icon Note from the source database
	stAPIRC = NSFNoteOpen(hdbSrc, NOTE_ID_SPECIAL + NOTE_CLASS_ICON, 0, &hnSrcIcon);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5200W, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		hnSrcIcon = NULLHANDLE;
	}

	//  No Icon note should exist in the target database - however if one does then we delete it
	stAPIRC = NSFNoteOpen(hdbTgt, NOTE_ID_SPECIAL + NOTE_CLASS_ICON, 0, &hnTgtIcon);
	if (ERR(stAPIRC) == NOERROR)
	{
		NSFNoteGetInfo(hnTgtIcon, _NOTE_ID, &nidIcon);
		NSFNoteClose(hnTgtIcon);
		stAPIRC = NSFNoteDelete(hdbTgt, nidIcon, UPDATE_FORCE);
		if (ERR(stAPIRC) != NOERROR)
		{
			ProxyLogMessage(crRQ, MSG_DXR5201E, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
			NSFNoteClose(hnSrcIcon);
			return FALSE;
		}
	}

	//  Create a new database icon note in the target database
	stAPIRC = NSFNoteCreate(hdbTgt, &hnTgtIcon);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5202E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		NSFNoteClose(hnSrcIcon);
		return FALSE;
	}

	wClass = NOTE_CLASS_ICON | NOTE_CLASS_DEFAULT;
	NSFNoteSetInfo(hnTgtIcon, _NOTE_CLASS, &wClass);
	
	//  Build up the content of the Icon note
	//  Start with the database icon
	copyIcon(hnSrcIcon, hnTgtIcon, crRQ, iThreadID);

	//  Now process the database title
	copyIconTextItem(hnSrcIcon, hnTgtIcon, FIELD_TITLE, crRQ, iThreadID);

	//  Now process the database flags
	wRetLen = NSFItemGetText(hnSrcIcon, DESIGN_FLAGS, szFlags, (26 + 26 + 10 + 15));
	szFlags[wRetLen] = '\0';

	//  Trace the source database flags and associated database options
	DBO1 = DBO2 = DBO3 = DBO4 = 0;
	stAPIRC = xeLocal->GetDbOptions(hdbSrc, &DBO1, &DBO2, &DBO3, &DBO4);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5220T, szFlags, DBO1, DBO2, DBO3, DBO4);
	xeLocal->LogTrace(TRACE_DBC, szMsg, iThreadID);

	//  Make sure that the large unktable flag is set on
	wRetLen = setFlag(CHFLAG_LARGE_UNKTABLE, szFlags);

	//  The LZ1 compression setting is determined by the setting in the request
	if (crRQ->LZ1Compress) wRetLen = setFlag(CHFLAG_LZ1, szFlags);
	else wRetLen = removeFlag(CHFLAG_LZ1, szFlags);
	szFlags[wRetLen] = '\0';

	//  Set the flags in the Icon Note
	stAPIRC = NSFItemSetText(hnTgtIcon, DESIGN_FLAGS, szFlags, wRetLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5206E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
	}

	//  Set the DAOS flag according to the settings in the request
	if (crRQ->EnableDAOS) stAPIRC = NSFItemSetText(hnTgtIcon, DAOS_ICON_ITEM, DAOS_ICON_ENABLED, DAOS_ICON_LEN);
	else stAPIRC = NSFItemSetText(hnTgtIcon, DAOS_ICON_ITEM, DAOS_ICON_DISABLED, DAOS_ICON_LEN);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5207E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
	}

	//  Copy significant settings from the source to the target ICON note
	//  The following are conditional copies the item is only copied if it exists in the source ICON note
	copyIconTextItem(hnSrcIcon, hnTgtIcon, LAUNCH_XPAGE_RUNONSERVER, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_NAVIGATOR_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_WEB_NAVIGATOR_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_WEBPAGE_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_OUTLINE_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_FRAMESET_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_COMPAPP_PAGE_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_COMPAPP_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_WEB_FRAMESET_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_WEB_PAGE_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_XPAGE_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_WEB_XPAGE_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, LAUNCH_VIEWNAME_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, LAUNCH_WEB_VIEWNAME_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, LAUNCH_DOCLINK_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, LAUNCH_WEB_DOCLINK_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, LAUNCH_WEB_DOCLINK_ANCHOR_ITEM, crRQ, iThreadID);

	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_LANGUAGE, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DEFAULT_COLLATION, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DB_THEME_NAME, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DB_DEFAULT_JSLIB, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, DB_NEW_HTML, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, CS_AUTO_ICON_ITEM, crRQ, iThreadID);
	
	copyIconTextItem(hnSrcIcon, hnTgtIcon, ICON_TEMPLATESERVERNAME_ITEM, crRQ, iThreadID);
	copyIconTextItem(hnSrcIcon, hnTgtIcon, ICON_TEMPLATEFILENAME_ITEM, crRQ, iThreadID);
	copyIconTDItem(hnSrcIcon, hnTgtIcon, ICON_TEMPLATEMODTIME_ITEM, crRQ, iThreadID);
	copyIconTDItem(hnSrcIcon, hnTgtIcon, ICON_PIRCREFRESHMODTIME_ITEM, crRQ, iThreadID);

	NSFNoteClose(hnSrcIcon);

	//  Update the Icon Note
	stAPIRC = NSFNoteUpdate(hnTgtIcon, UPDATE_FORCE);
	NSFNoteClose(hnTgtIcon);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5208E, iThreadID);
		xeLocal->GetAPIMessage(stAPIRC, szMsg);
		ProxyLogMessage(crRQ, szMsg, iThreadID);
		return FALSE;
	}

	//  Trace the flags and Database Options in the target database
	DBO1 = DBO2 = DBO3 = DBO4 = 0;
	stAPIRC = xeLocal->GetDbOptions(hdbTgt, &DBO1, &DBO2, &DBO3, &DBO4);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5221T, szFlags, DBO1, DBO2, DBO3, DBO4);
	xeLocal->LogTrace(TRACE_DBC, szMsg, iThreadID);

	return TRUE;
}

//  reportTargetDb
//
//  This function will report the state of the database flags and options
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the target database
//  2  -  int				-  Identity of the thread that is making the call
//
//  Returns
//
//  API Use Notes:
//
void DbCopier::reportTargetDb(DBHANDLE hdbTgt, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	NOTEHANDLE			hnTgtIcon;							//  Handles to the target Db Icon notes
	WORD				wRetLen;							//  Length of the flags
	DWORD				DBO1,DBO2,DBO3,DBO4;				//  Very Extended Database options
	char				szFlags[26 + 26 + 10 + 20];			//  Database Flags
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Get the Db Icon Note from the Target database
	stAPIRC = NSFNoteOpen(hdbTgt, NOTE_ID_SPECIAL + NOTE_CLASS_ICON, 0, &hnTgtIcon);
	if (ERR(stAPIRC) != NOERROR) return;

	//  Now get the database flags
	wRetLen = NSFItemGetText(hnTgtIcon, DESIGN_FLAGS, szFlags, (26 + 26 + 10 + 15));
	szFlags[wRetLen] = '\0';
	NSFNoteClose(hnTgtIcon);

	//  Trace the flags and Database Options in the target database
	DBO1 = DBO2 = DBO3 = DBO4 = 0;
	stAPIRC = xeLocal->GetDbOptions(hdbTgt, &DBO1, &DBO2, &DBO3, &DBO4);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR5221T, szFlags, DBO1, DBO2, DBO3, DBO4);
	xeLocal->LogTrace(TRACE_DBC, szMsg, iThreadID);

	return;
}

//  setFlag
//
//  This function will set the specified flag in the passed flag string
//
//  Parameters
//
//	1  -  char				-  Flag value to be set (if it not already)
//  2  -  char *			-  Pointer to the null terminated flag string
//
//  Returns
//
//  WORD					-  New length of the flags string
//
//  API Use Notes:
//
WORD DbCopier::setFlag(char Flag, char *szFlags)
{
	char				*pFlag;										//  Pointer into the flag string
	WORD				wLen;										//  Length of the flag string

	//  See if the flag is already present in the flag string
	pFlag = szFlags;
	wLen = 0;
	while (*pFlag != Flag && *pFlag != '\0') 
	{
		wLen++;
		pFlag++;
	}

	if (*pFlag == Flag) return WORD(strlen(szFlags));

	//  Add the flag to the end of the flag string
	*pFlag = Flag;
	pFlag++;
	*pFlag = '\0';

	return wLen + 1;
}

//  removeFlag
//
//  This function will unset the specified flag in the passed flag string
//
//  Parameters
//
//	1  -  char				-  Flag value to be unset (if it is already)
//  2  -  char *			-  Pointer to the null terminated falg string
//
//  Returns
//
//  WORD					-  New length of the flags string
//
//  API Use Notes:
//
WORD DbCopier::removeFlag(char Flag, char *szFlags)
{
	char				*pFlag;										//  Pointer into the flag string
	char				*pNextFlag;									//  Pointer into the flag string
	WORD				wLen;										//  Length of the flag string

	//  See if the flag is not present in the flag string
	pFlag = szFlags;
	wLen = 0;
	while (*pFlag != Flag && *pFlag != '\0') 
	{
		wLen++;
		pFlag++;
	}

	if (*pFlag == '\0') return wLen;

	//  Remove the flag to the end of the flag string
	pNextFlag = pFlag + 1;
	while (*pFlag != '\0')
	{
		*pFlag = *pNextFlag;
		pFlag++;
		pNextFlag++;
		wLen++;
	}

	return wLen - 1;
}

//  copyIcon
//
//  This function will copy a database icon from source to target note
//
//  Parameters
//
//	1  -  NOTEHANDLE			-  Handle of the source note
//  2  -  NOTEHANDLE			-  Handle of the target note
//
//  Returns
//
//  API Use Notes:
//
//	The function implements a different method for x64 due to a "Panic: lookup handle out of range" error that
//  is triggered on 8.5.2 FP4 (x64) in NSFItemCopy().
// 
void DbCopier::copyIcon(NOTEHANDLE hnSrc, NOTEHANDLE hnTgt, CopyRequest *crRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	BLOCKID				bidItem,bidValue;					//  Item and item value block IDs
	WORD				wType;								//  Item type
	DWORD				dwLen;								//  Item length
#ifdef W64
	BYTE *				pIcon;								//  Pointer to the Icon in memory
#endif
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

#ifdef W64
	//  Widows 64 bit build
	stAPIRC = NSFItemInfo(hnSrc, DESK_ICON_ITEM, (WORD) strlen(DESK_ICON_ITEM), &bidItem, &wType, &bidValue, &dwLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5203I, iThreadID);
	}
	else
	{
		//  Get a pointer to the ICON data
		pIcon = OSLockBlock(BYTE, bidValue);
		if (pIcon != NULL)
		{
			//  Adjust the pointer and length to drop the leading data type word
			pIcon += sizeof(WORD);
			dwLen -= sizeof(WORD);

			//  Append the ICON to the target note, we use NSFItemAppend rather than creating a BLOCKID and
			//  using NSFItemAppendByBLOCKID because we suspect that would use the same code path as NSFItemCopy.
			stAPIRC = NSFItemAppend(hnTgt, 
									ITEM_SUMMARY, 
									DESK_ICON_ITEM, 
									(WORD) strlen(DESK_ICON_ITEM),
									TYPE_ICON,
									pIcon,
									dwLen);
			OSUnlockBlock(bidValue);
			if (ERR(stAPIRC) != NOERROR)
			{
				ProxyLogMessage(crRQ, MSG_DXR5204E, iThreadID);
				xeLocal->GetAPIMessage(stAPIRC, szMsg);
				ProxyLogMessage(crRQ, szMsg, iThreadID);
			}
		}
	}
#else
	//  32 bit builds
	stAPIRC = NSFItemInfo(hnSrc, DESK_ICON_ITEM, (WORD) strlen(DESK_ICON_ITEM), &bidItem, &wType, &bidValue, &dwLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		ProxyLogMessage(crRQ, MSG_DXR5203I, iThreadID);
	}
	else
	{
		//  Copy the ICON to the target database icon note
		stAPIRC = NSFItemCopy(hnTgt, bidItem);
		if (ERR(stAPIRC) != NOERROR)
		{
			ProxyLogMessage(crRQ, MSG_DXR5204E, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
	}
#endif

	//  Return to caller
	return;
}

//  copyIconTextItem
//
//  This function will copy a text item from the source to the target icon notes.
//
//  Parameters
//
//	1  -  NOTEHANDLE			-  Handle of the source note
//  2  -  NOTEHANDLE			-  Handle of the target note
//	3  -  char *				-  Pointer to the item name to be copied
//	4  -  CopyRequest *			-  Pointer to the copy request
//	5  -  int					-  ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
void DbCopier::copyIconTextItem(NOTEHANDLE hnSrc, NOTEHANDLE hnTgt, char *szItemName, CopyRequest *crRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	WORD				wRetLen;							//  Length of the returned text
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Only copy items that exist on the source icon note
	if (NSFItemIsPresent(hnSrc, szItemName, WORD(strlen(szItemName))))
	{
		//  Retrieve the text value
		wRetLen = NSFItemGetText(hnSrc, szItemName, szMsg, MAX_MSG);
		szMsg[wRetLen] = '\0';

		//  Set the text value in the target icon note
		stAPIRC = NSFItemSetText(hnTgt, szItemName, szMsg, wRetLen);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR5205E, szItemName);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
			xeLocal->GetAPIMessage(stAPIRC, szMsg);
			ProxyLogMessage(crRQ, szMsg, iThreadID);
		}
	}

	//  Return to caller
	return;
}

//  copyIconTDItem
//
//  This function will copy a TIMEDATE item from the source to the target icon notes.
//
//  Parameters
//
//	1  -  NOTEHANDLE			-  Handle of the source note
//  2  -  NOTEHANDLE			-  Handle of the target note
//	3  -  char *				-  Pointer to the item name to be copied
//	4  -  CopyRequest *			-  Pointer to the copy request
//	5  -  int					-  ID of the calling thread
//
//  Returns
//
//  API Use Notes:
//
void DbCopier::copyIconTDItem(NOTEHANDLE hnSrc, NOTEHANDLE hnTgt, char *szItemName, CopyRequest *crRQ, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API Calls
	TIMEDATE			tdValue;							//  TIMEDATE value
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Only copy items that exist on the source icon note
	if (NSFItemIsPresent(hnSrc, szItemName, WORD(strlen(szItemName))))
	{
		//  Retrieve the TIMEDATE value
		if (NSFItemGetTime(hnSrc, szItemName, &tdValue))
		{
			//  Set the TIMEDATE value in the target icon note
			stAPIRC = NSFItemSetTime(hnTgt, szItemName, &tdValue);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR5205E, szItemName);
				ProxyLogMessage(crRQ, szMsg, iThreadID);
				xeLocal->GetAPIMessage(stAPIRC, szMsg);
				ProxyLogMessage(crRQ, szMsg, iThreadID);
			}
		}
	}

	//  Return to caller
	return;
}
