#pragma once
//  DesignManager for DX based applications
//
//  This class provides the Design Management functions used in Domono eXplorer (DX) utilities
//
//  Message Identifier:      DXR
//  Message Range:           5500 - 5999
//  Trace area assignment:   TRACE_DM (24)
//  DX Version:				 3.14.0
//  Version:                 2.3.0  Build: 120
//
//  Version History:
//
//  28/03/2011		-  2.0.1		-  Updated for compatability with Resource Loader
//  20/06/2011		-  2.1.0		-  Tidy up for unix
//  22/10/2011		-  2.1.1		-  VS 2011 Updates
//  04/11/2011		-  2.2.0		-  Migration of D/B conditioning from kernel to Design Manager
//	21/01/2015		-  2.3.0		-  x64 compatability fixes
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2001 - 2015
//

//  Platform Includes
#include "../Platform/PlatBase.h"

//  Runtime Includes
#include "../ExecEnvironment.h"
#include "../DXGlobals.h"
#include "DesignManifest.h"
#include "DesignManifestLine.h"
#include "../DXResourceLoader.h"

#include <viewfmt.h>
#include <nls.h>
#include <fsods.h>
#include <niferr.h>

//  Trace Messages
#define		MSG_DXR5501T		"DXR5501T: A new DesignManager instance has been created @%p."
#define		MSG_DXR5502T		"DXR5502T: A DesignManager instance  @%p has been destroyed."
#define		MSG_DXR5505T		"DXR5505T: Parsing '%s' at state: %i."
#define		MSG_DXR5518T		"DXR5518T: Element: %s of type: %i has been parsed."
#define		MSG_DXR5522T		"DXR5522T: Attempting to update design element: %s in the repository."
#define		MSG_DXR5529T		"DXR5529T: Configuring connection to host: %s."
#define		MSG_DXR5531T		"DXR5531T: Configuring request for resource: %s."
#define		MSG_DXR5538T		"DXR5538T: Returning %ld bytes to the XML processor from cache."
#define		MSG_DXR5541T		"DXR5541T: Returning %ld bytes to the XML processor from a new cache."
#define		MSG_DXR5542T		"DXR5542T: End of content stream has been processed normally."
#define		MSG_DXR5543T		"DXR5543T: Returning %ld bytes to the XML processor from uncached response data."
#define		MSG_DXR5564T		"DXR5564T: Check for newly imported Design Element '%s' found Note ID: 0x%.4X."
#define		MSG_DXR5565T		"DXR5565T: Search for Design Element '%s' found Note ID: 0x%.4X."
#define		MSG_DXR5570T		"DXR5570T: XML download returned a total of %d bytes."
#define		MSG_DXR5572T		"DXR5572T: XMLReader hit stage %d."
#define		MSG_DXR5579T		"DXR5579T: Processing column %d of %d."
#define		MSG_DXR5580T		"DXR5580T: Column %d has as item name '%s'."
#define		MSG_DXR5582T		"DXR5582T: Rebuilding selection formula in view '%s'."
#define		MSG_DXR5595T		"DXR5595T: Column %d is hidden."
#define		MSG_DXR5596T		"DXR5596T: Column %d has a zero length formula."
#define		MSG_DXR5597T		"DXR5597T: Column %d has a decompiled formula length of %d."
#define		MSG_DXR5598T		"DXR5598T: Column %d has a compiled formula length of %d."
#define		MSG_DXR5599T		"DXR5599T: Formula buffer contains %d bytes."
#define		MSG_DXR5600T		"DXR5600T: View selection formula is '%s'."
#define		MSG_DXR5601T		"DXR5601T: Column formula."
#define		MSG_DXR5602T		"DXR5602T: View should display response docs hierarchically."
#define		MSG_DXR5603T		"DXR5603T: View should display replication conflicts."
#define		MSG_DXR5604T		"DXR5602T: Added view summary item to display response docs hierarchically."
#define		MSG_DXR5605T		"DXR5603T: Added view summary item to display replication conflicts."
#define		MSG_DXR5619T		"DXR5619T: Found design note 0x%.4X that matched search for '%s'."
#define		MSG_DXR5621T		"DXR5621T: Manifest Line: Name '%s', Type %d, UNID '%s', Options %.4X."
#define		MSG_DXR5622T		"DXR5622T: Found %d partial matches in design collection."
#define		MSG_DXR5623T		"DXR5623T: Read %d partial matches in design collection."
#define		MSG_DXR5624T		"DXR5624T: Found Title(0) of '%s' in design Note ID: 0x%.4X."
#define		MSG_DXR5626T		"DXR5626T: Freeing DXL Import log memory."
#define		MSG_DXR5630T		"DXR5630T: Translated element name from source '%s' to LMBCS '%s'."
#define		MSG_DXR5631T		"DXR5631T: Design element has title '%s'."
#define		MSG_DXR5633T		"DXR5633T: Design collection contains %i documents."
#define		MSG_DXR5634T		"DXR5635T: Candidate note 0x%.4X does not match note class."
#define		MSG_DXR5635T		"DXR5636T: Candidate note 0x%.4X does not match supplied UNID."

//  Information Messages
#define		MSG_DXR5506I		"DXR5506I: Design Manager is loading the manifest (%s)."
#define		MSG_DXR5507I		"DXR5507I: Design Manifest (%s) has been successfully loaded."
#define		MSG_DXR5517I		"DXR5517I: The Design Manifest has been loaded with %i sets and %i elements."
#define		MSG_DXR5519I		"DXR5519I: Processing a request to apply Design Set: %s to the repository."
#define		MSG_DXR5521I		"DXR5521I: Completed a request to apply Design Set: %s to the repository."
#define		MSG_DXR5544I		"DXR5544I: Design Manager completed, %i Design Sets were applied, %i Design Elements were added and %i were skipped."
#define		MSG_DXR5545I		"DXR5545I: Skipping Design Element '%s', already in the repository."
#define		MSG_DXR5556I		"DXR5556I: Current Domino build number is %i."
#define		MSG_DXR5560I		"DXR5560I: Manifest located, processing %d bytes."
#define		MSG_DXR5568I		"DXR5568I: Design Element '%s' completed import process."
#define		MSG_DXR5640I		"DXR5640I: Design Element '%s' (Note ID: 0x%.4X) has been signed by '%s'."

//  Warning Messages
#define		MSG_DXR5546W		"DXR5546W: Post-processing failed for Design Element '%s'."
#define		MSG_DXR5555W		"DXR5555W: Pre-processing failed for Design Element '%s', skipping element."
#define		MSG_DXR5557W		"DXR5557W: The current version of Domino does not import DXL reliably (use 7.0.3 and up)."
#define		MSG_DXR5611W		"DXR5611W: Design element '%s' was not found using design collection search of %i docs."
#define		MSG_DXR5637W		"DXR5637W: Attempting to fix database design collection by creating collection note."
#define		MSG_DXR5638W		"DXR5638W: Attempting to fix database design collection by creating collection."
#define		MSG_DXR5644W		"DXR5644W: Refresh of the design collection failed."

//  Error Messages
#define		MSG_DXR5503E		"DXR5503E: Unable to locate the DesignManifest resource in this executable."
#define		MSG_DXR5504E		"DXR5504E: Unable to parse the DesignManifest, Design manager is inactive."
#define		MSG_DXR5508E		"DXR5508E: The Design Manifest line parser has returned a permanent error."
#define		MSG_DXR5509E		"DXR5509E: The Design Manifest line parser could not recognise an XML line."
#define		MSG_DXR5510E		"DXR5510E: The Manifest definition is missing from the design manifest."
#define		MSG_DXR5511E		"DXR5511E: A DesignSet definition is missing from the design manifest."
#define		MSG_DXR5512E		"DXR5512E: A DesignSet close definition is missing from the design manifest."
#define		MSG_DXR5513E		"DXR5513E: An empty DesignSet (containing no elements) is invalid."
#define		MSG_DXR5514E		"DXR5514E: A DesignElement is missing the URL definition."
#define		MSG_DXR5515E		"DXR5515E: A DesignElement close is missing from the design manifest."
#define		MSG_DXR5516E		"DXR5516E: A Design Manifest cannot be defined without any Design Sets."
#define		MSG_DXR5520E		"DXR5520E: Design Set '%s' could not be located in the loaded design manifest."
#define		MSG_DXR5522E		"DXR5522E: Unable to create a valid DXL Importer."
#define		MSG_DXR5523E		"DXR5523E: DXLCreateImporter call failed."
#define		MSG_DXR5524E		"DXR5524E: Unable to set a DXL Importer property %d to %d."
#define		MSG_DXR5525E		"DXR5525E: DXLImport of Design Element '%s' failed."
#define		MSG_DXR5526E		"DXR5526E: Unable to get the DXL Importer Log Handle."
#define		MSG_DXR5530E		"DXR5530E: Failed to open a design note (Note ID: 0x%.4X) for post processing."
#define		MSG_DXR5531E		"DXR5531E: Failed to read the $AssistInfo structure item from an agent note (Note ID: 0x%.4X)."
#define		MSG_DXR5532E		"DXR5532E: Failed to update a design note (Note ID: 0x%.4X) after post processing."
#define		MSG_DXR5547E		"DXR5547E: Unable to locate the icon data carrier design note."
#define		MSG_DXR5548E		"DXR5548E: Unable to open the icon data carrier design note."
#define		MSG_DXR5549E		"DXR5549E: Unable to open the database icon note."
#define		MSG_DXR5550E		"DXR5550E: Unable to remove the icon bitmap item from the icon note."
#define		MSG_DXR5551E		"DXR5551E: Unable to get icon bitmap info from the carrier note."
#define		MSG_DXR5552E		"DXR5552E: Unable to copy the icon bitmap item to the icon note."
#define		MSG_DXR5553E		"DXR5553E: Unable to delete the icon data carrier design note."
#define		MSG_DXR5554E		"DXR5554E: Unable to delete the '%s' note."
#define		MSG_DXR5558E		"DXR5558E: Unable to allocate memory for the Manifest Buffer."
#define		MSG_DXR5559E		"DXR5559E: The Design Manifest '%s' could not be loaded, design update abandoned."
#define		MSG_DXR5561E		"DXR5561E: Unable to find newly imported Design Element '%s'."
#define		MSG_DXR5562E		"DXR5562E: Failed to determine NoteID to sign Design Element '%s'."
#define		MSG_DXR5563E		"DXR5563E: Failed to '%s' Design Element '%s' with NoteID 0x%.4X."
#define		MSG_DXR5566E		"DXR5566E: Failed to check for existing Design Element '%s' [UNID: '%s']."
#define		MSG_DXR5567E		"DXR5567E: Failed to find match for required Design Element '%s' [UNID: '%s']."
#define		MSG_DXR5569E		"DXR5569E: Failed to download XML data for Design Element '%s'."
#define		MSG_DXR5571E		"DXR5571E: XML import of Design Element '%s' is not completed or been cleaned up."
#define		MSG_DXR5573E		"DXR5573E: XML import of Design Element '%s' suffered import errors."
#define		MSG_DXR5574E		"DXR5574E: Post processing failed to update Design Element '%s'."
#define		MSG_DXR5576E		"DXR5576E: LS Compile Error on line %d of '%s': '%s'."
#define		MSG_DXR5577E		"DXR5577E: Failed to get info about item '%s' on Design Element '%s'."
#define		MSG_DXR5578E		"DXR5578E: View format version is not equal to %d on Design Element '%s'."
#define		MSG_DXR5581E		"DXR5581E: Failed to decompile the formula for column %d on Design Element '%s'."
#define		MSG_DXR5583E		"DXR5583E: Formula for column %d is longer than %d characters on Design Element '%s'."
#define		MSG_DXR5584E		"DXR5584E: Selection formula is longer than %d characters on Design Element '%s'."
#define		MSG_DXR5585E		"DXR5585E: Selection formula is not of type FORMULA on Design Element '%s'."
#define		MSG_DXR5586E		"DXR5586E: Failed to compile base selection formula on Design Element '%s'."
#define		MSG_DXR5587E		"DXR5587E: Failed to append item name of column %d to selection formula on Design Element '%s'."
#define		MSG_DXR5588E		"DXR5588E: Failed to merge formula from column %d to selection formula on Design Element '%s'."
#define		MSG_DXR5589E		"DXR5589E: Failed to append '%s' to selection formula on Design Element '%s'."
#define		MSG_DXR5590E		"DXR5590E: Failed to get size of selection formula for Design Element '%s'."
#define		MSG_DXR5591E		"DXR5591E: Failed to append item '%s' to view note of Design Element '%s'."
#define		MSG_DXR5592E		"DXR5592E: Failed to fix view selection formula of Design Element '%s'."
#define		MSG_DXR5593E		"DXR5593E: Failed to decompile the existing selection formula on Design Element '%s'."
#define		MSG_DXR5594E		"DXR5594E: Failed to removing existing item '%s' from Design Element '%s'."
#define		MSG_DXR5613E		"DXR5613E: Failed to open design collection in database."
#define		MSG_DXR5614E		"DXR5614E: Failed to find design collection entries partially matching '%s'."
#define		MSG_DXR5615E		"DXR5615E: Failed to read design collection entries partially matching '%s'."
#define		MSG_DXR5616E		"DXR5616E: Read of design collection entries partially matching '%s' returned a NULL Handle!"
#define		MSG_DXR5617E		"DXR5617E: Failed to open candidate (Note ID: 0x%.4X) that partially matched search for '%s'."
#define		MSG_DXR5618E		"DXR5618E: Multiple design elements match '%s'!"
#define		MSG_DXR5620E		"DXR5620E: Failed to close design collection in database."
#define		MSG_DXR5625E		"DXR5625E: Failed to get DXL import log memory handle for '%s'."
#define		MSG_DXR5628E		"DXR5628E: Failed to get NLS_INFO to convert element name into LMBCS."
#define		MSG_DXR5629E		"DXR5629E: Failed to translate element name into LMBCS."
#define		MSG_DXR5632E		"DXR5632E: Failed to close candidate note 0x%.4X that partially matched search for '%s'."
#define		MSG_DXR5636E		"DXR5636E: Failed to create/rebuild design collection in database."
#define		MSG_DXR5639E		"DXR5639E: Failed to resize the the Manifest memory to %i bytes."
#define		MSG_DXR5641E		"DXR5641E: Failed to open the local name and address book."
#define		MSG_DXR5642E		"DXR5642E: Unable to find the Note ID of the design collection in the NAB."
#define		MSG_DXR5643E		"DXR5643E: Unable to update the new design collection note in the repository."
#define		MSG_DXR5645E		"DXR5645E: Unable to create a new note for the design collection."
#define		MSG_DXR5646E		"DXR5646E: Unable to add the '%s' item to the design collection note."
#define		MSG_DXR5647E		"DXR5647E: Unable to open the existing design collection note."
#define		MSG_DXR5648E		"DXR5648E: Unable to find the BLOCKID of the '%s' in the design collection."
#define		MSG_DXR5649E		"DXR5649E: Unable to copy the '%s' item to the new design collection."
#define		MSG_DXR5650E		"DXR5650E: Failed to open the default view in the database."
#define		MSG_DXR5651E		"DXR5651E: Failed to get the NoteIDs of view notes in the database."
#define		MSG_DXR5652E		"DXR5652E: Failed to get the NoteID of the first view note in the database."
#define		MSG_DXR5653E		"DXR5653E: Unable to create a default view note in the database."
#define		MSG_DXR5654E		"DXR5654E: Unable to write the default view title."
#define		MSG_DXR5655E		"DXR5655E: Failed to compile the default view selection formula."
#define		MSG_DXR5656E		"DXR5656E: Formula error is at Line: %i, Column: %i, Offset: %i, Length: %i."
#define		MSG_DXR5657E		"DXR5657E: Failed to compile the default view column formula."
#define		MSG_DXR5658E		"DXR5658E: Unable to add summary item '%s' to the default view note."
#define		MSG_DXR5659E		"DXR5659E: Unable to merge the formula for '%s' into the default view formula."
#define		MSG_DXR5660E		"DXR5660E: Unable to get the size of the merged default view selection formula."
#define		MSG_DXR5661E		"DXR5661E: Unable to append the Formula item (%i bytes) to the default view note."
#define		MSG_DXR5662E		"DXR5662E: Failed to allocate a %i byte buffer for the default view VIEWFORMAT."
#define		MSG_DXR5663E		"DXR5663E: Unable to append the Viewformat item (%i bytes) to the default view note."
#define		MSG_DXR5664E		"DXR5664E: Unable to write the default view note to the database."
#define		MSG_DXR5665E		"DXR5665E: Failed to mark Note ID: 0x%.4X as the default view note in the database."

//  Local Definitions

#define		DM_BUILD						120
#define		DM_IDENTITY						"DXCommon Design Manager 2.3.0 (build: %i)"
#define		MAX_LINE						256								//  Max line size
#define		DUMMY_CARRIER_ELEMENT			"DX-Dummy-Carrier"				//  Dummy carrier for ICONDATA
#define		MANIFEST_BFR_SIZE				20*1024							//  20k Buffer for Manifest
#define		IMPORT_WAIT_INTERVAL			3*1000							//  Wait 3 seconds for backend
#define		MAX_COLUMN_NAME_SIZE			128
#define		MAX_COLUMN_FORMULA_SIZE			64000
#define		MAX_FORMULA_LINE				32000
#define		DEF_VIEW_TITLE					""								//  Default View Title
#define		DEF_VIEW_SELECT					"@All"							//  Default view selection formula
#define		DEF_VIEW_C0NAME					"$0"							//  Programmatic name for column 0
#define		DEF_VIEW_C0FORMULA				"@DocNumber"					//  Formula value for column 0
#define		DEF_VIEW_C0TITLE				"#"								//  Title value for column 0
#define		DC_VIEW_VER						"J"								//  View version
#define		DC_VIEW_VERV2					"4"								//  View designer version
#define		DC_VIEW_CLASSES					"3724"							//  View Classes

//  Parser States
#define		PSTATE_NOT_STARTED				1				//  Parsing not started
#define		PSTATE_MANIFEST_DEFINED			2				//  Root Manifest object is defined
#define		PSTATE_SET_OPEN					3				//  DesignSet is open
#define		PSTATE_ELEMENT_INDEF			4				//  Element is being defined
#define		PSTATE_ELEMENT_URLDEF			5				//  Element has had the URL added

class DesignManager
{
public:
	//  Constructor
	DesignManager(ExecEnvironment *xeParent);

	//  Destructor
	~DesignManager(void);

	//  Public members
	DesignManifest *dmRoot;									//  Design manifest root
	int				iDMDesignImportOption;					//  XML Design Element Import Options
	BOOL			bDMReplaceDbProperties;					//  XML import replace db properties
	BOOL			bDMReplicaRequiredForReplaceOrUpdate;	//  Replica Info required for XML import
	int				iDMInputValidationOption;				//  XML Input Validation Option
	BOOL			bDMExitOnFirstFatalError;				//  Flag to indicate if xml import should exit on first fatal error
	int				iDMUnknownTokenLogOption;				//  Unknown XML Token Log Option
	BOOL			bRepairDbErrors;						//  Flag to indicate if we should attempt fixes to known db errors 

	//  Public functions
	BOOL LoadManifest(char *szManifestID);					//  Load the design manifest
	void DestroyManifest(void);								//  Destroy the manifest
	BOOL ApplyAllDesignSets(void);							//  Apply all design sets to the repository
	BOOL ApplyAllDesignSets(DBHANDLE hDB);					//  Apply all design sets to a Database
	BOOL ApplyDesignSet(char *szSetName);					//  Apply a design set to the repository
	BOOL ApplyDesignSet(char *szSetName, DBHANDLE hDB);		//  Apply a design set to a Database
	void GetBuildInfo_s(char *szBuild, int iSize);			//  Get the build identity
	void GetDefaultDbTitle_s(char *szTitle, int iSize);		//  Get the default (constructed) Db Title
	BOOL ValidateDesign(DBHANDLE hDB);						//  Validate and fix the database design

	STATUS static LNCALLBACK WrapperLSCompileError(LSCOMPILE_ERR_INFO *pInfo, void *pContext);

private:
	//  Private Members
	ExecEnvironment *xeLocal;								//  Execution Environment
	DXResourceLoader *rlLocal;								//  Local Resource Loader
	DXLIMPORTHANDLE dihImporter;							//  DXL Importer Handle
	int iSetsApplied;										//  Count of design sets applied
	int iElementsAdded;										//  Count of design elements added
	int iElementsSkipped;									//  Count of design elements skipped
	char szCurrentReadURL[MAX_DM_URL + 1];					//  Current URL to Read

	//  Private Functions
	BOOL ParseManifest(char *pManifest);					//  Parse the Design Manifest resource
	DesignSet * FindDesignSet(char *szSetName);				//  Find a named Design Set
	BOOL ApplyDesignElement(DesignElement *deCurrent, DBHANDLE hDB, HCOLLECTION hDesignColl);		//  Apply a Design Element to the repository
	BOOL CreateImporter(void);								//  Create a DXL Importer instance
	BOOL PostProcessElement(DesignElement *deCurrent, WORD wNoteClass, DBHANDLE hDB, HCOLLECTION hDesignColl, NOTEID nidDE);		//  Post processing for some design elements
	BOOL PreProcessElement(DesignElement *deCurrent, DBHANDLE hDB);			//  Pre processing for some design elements
	BOOL CoreApplyDesignSet(char *szSetName, DBHANDLE hDB); //  Apply a design set to a database
	BOOL CoreApplyAllDesignSets(DBHANDLE hdb);				//  Apply all designsets to a database
	BOOL DumpXMLImportErrors(DesignElement *deCurrent);     //  Process the DXL Importer log
	BOOL FindDesignElementNote(DesignElement *deCurrent, WORD wNoteClass, DBHANDLE hDB, HCOLLECTION hDesignColl, NOTEID *nidDesign);
	STATUS DumpLSCompileError(LSCOMPILE_ERR_INFO *pInfo, DesignManager *dmCurrent);
	BOOL FixViewSelectionFormula(DesignElement *deCurrent, DBHANDLE hDB, NOTEHANDLE hnDesign);
	void FettleAgents(DBHANDLE hdbDB, NOTEID nidAgent, int iThreadID);		//  Apply post import fixes to agents
	BOOL CopyDesignCollection(DBHANDLE hdbX, int iThreadID);				//  Copy the design collection from names.nsf
	NOTEID FindAnyView(DBHANDLE hDB, int iThreadID);						//  Find the note id of any view
	BOOL WriteDefaultView(DBHANDLE hdbX, int iThreadID);					//  Write a default view note into the database
};
