//  ExecEnvironment
//
//  This class provides all of the common runtime support for Domino eXplorer (DX) applications.
//
//  Message Identifier:      DXR
//  Message Range:           0001 - 0499
//  Trace area assignment:   TRACE_CORERT (01)
//  DX Version:				 3.14.1
//  Version:                 3.14.1  build: 109
//
//  Version History:
//
//  1.0.0		10/02/2001		Initial version
//  3.5.3		21/08/2009		Re-based on DX 3.x.x
//  3.6.0		18/09/2009		New log handler for multithreaded applications
//  3.6.1		10/10/2009		Additional controls added
//  3.6.2		20/11/2009		Unix compat changes plus minor fixes/enhancements
//  3.6.3		19/12/2009		Enhancements to DbMover
//  3.6.4		23/12/2009		Additional fixes to DbMover
//  3.7.0		23/12/2009		Added race auto-correction to MT support
//  3.7.1		29/12/2009		AppWait() updated to eliminate overhead
//  3.7.2		12/01/2010		DbCopier improvements for corrupt folder handling
//								Remote Fixup on DbOpen
//								Logger improvements
//  3.7.3		06/02/2010		DbMover & DbCopier improvements
//  3.7.4		19/03/2010		Fixed various issues with log/repository configuration
//  3.7.5		30/03/2010		Improved Error Handling in DbCopier
//  3.7.6		08/04/2010		Fixed a problem with error handling
//  3.7.7		10/04/2010		Additional recovery options in DbCopier
//  3.8.0		15/05/2010		DCF stub calls added to core
//								Load & Execute functions added
//  3.8.3       25/08/2010		Prevent escape % characters being passed to console logger
//  3.8.4		01/07/2010		(spl) Enhancements to OSLoadProg functions
//								Changed return code from simple BOOL to int
//								Added code to get RC of program if we are not running in detached mode
//  3.9.0		24/07/2010		Linkage changed for logging & request posting
//								Additional functions added GenUniqueID(), populateUNID() & xtoi()
//  3.9.1		02/09/2010		Changes to the ACL Processors
//								Fix for false RJR_NONE_EXIST signals
//  3.9.2		20/09/2010		UNID handling routines added
//								Fixes for ACL Processor
//								Avoided a recursion in DX processing
//  3.10.0		01/10/2010		Wildcard & multi-thread execution added to DX processing
//								DXNoteScanner incorporated in kernel
//  3.10.1		15/10/2010		Asynch Transaction Handler added
//  3.10.2		18/11/2010		MTIO support added to DbMover
//  3.10.3		13/12/2010		Db Property Modifier added
//								Debug Helper added
//  3.10.4		14/01/2011		Command handler supports automated command injection
//  3.10.5		18/02/2011		DX changes to support reporting on D/Bs that cannot be opened
//  3.10.6		19/03/2011		Upload capability added to DXResourceLoader
//  3.10.7		17/05/2011		Fixed an issue with ACL processing
//  3.10.8		30/05/2011		Transaction scheduling added to transaction handler
//  3.10.9		04/06/2011		Some tidying up for Linux
//  3.10.10		14/07/2011		Performance Improvements for/to DbCopier
//  3.10.11		21/07/2011		ObjectCache added to Kernel
//  3.10.12		25/08/2011		Fix timing reporting in OpenDatabase
//  3.11.0		29/09/2011		NamedEntity support added
//  3.11.1		22/10/2011		VS 2011 Readiness
//  3.11.2		04/11/2011		Migration of D/B conditioning from kernel to Design Manager
//  3.11.3		12/11/2011		Fixes to Explorer and ACL Processor
//  3.11.4		27/11/2011		Adopt TimeBasedArray, fixes for ACL processor
//  3.11.5		30/11/2011		Explorer recursion improvements
//								Document count changes for R5 servers
//								Transaction scheduler last run time caclulation changed
//								ObjectCache coherence improvements
//								Additional duplicate match messages in NEResolver
//  3.11.6		10/12/2011		OpenDatabase handles Replica IDs
//								Log writer improvements
//  3.12.0		21/12/2011		DXIP introduced
//	3.13.0		20/12/2013		Additional D/B Create options
//	3.14.0		19/01/2015		x64 compatability fixes
//	3.14.1		24/02/2015		Avoid a panic in DbCopier
//
//  Copyright: Hadleigh Marshall (Netherlands) b.v. 2001 - 2015
//

//  Application Includes
#include "ExecEnvironment.h"

//  Default constructor (only used for constructing derived classes)
ExecEnvironment::ExecEnvironment(void)
: IsInitialised(FALSE)
, rsLocal(0)
, hdbRepository(NULLHANDLE)
, RepositoryCreated(FALSE)
, LoggingState(LOGSTATE_NOTOPEN)
, hnLogDoc(NULLHANDLE)
, LogLines(0)
, bNRTInitialised(FALSE)
, hDefDescriptor(NULLHANDLE)
, hmqProcessor(NULLHANDLE)
, LogLinesPending(0)
, DCFGlobalPermit(TRUE)
, RunningTime(NULL)
{
}

//  Constructor
ExecEnvironment::ExecEnvironment(RunSettings* rsIn, int argc, char * argv[])
: IsInitialised(FALSE)
, rsLocal(0)
, hdbRepository(NULLHANDLE)
, RepositoryCreated(FALSE)
, LoggingState(LOGSTATE_NOTOPEN)
, hnLogDoc(NULLHANDLE)
, LogLines(0)
, bNRTInitialised(FALSE)
, hDefDescriptor(NULLHANDLE)
, hmqProcessor(NULLHANDLE)
, LogLinesPending(0)
, DCFGlobalPermit(TRUE)
, RunningTime(NULL)
{
	//  Initialise the runtime environment
	CanonicalUser[0] = '\0';
	AbbreviatedUser[0] = '\0';
	CommonUser[0] = '\0';
	rsLocal = rsIn;									//  Save the Run Settings

	IsInitialised = InitEnvironment(argc, argv);	//  Perform the Runtime initialisation
}

//  Destructor
ExecEnvironment::~ExecEnvironment(void)
{
	//  If logging is still active the close it down
	if (LoggingState == LOGSTATE_ACTIVE)
	{
		LogMessage(MSG_DXR0016W, 0);
		CloseLogDoc();
		LoggingState = LOGSTATE_CLOSED;
	}

	//  Close the environment
	Close();
}

//  InitEnvironment
//
//  This function carries out all of the initialisation functions needed to initialise a Domino eXplorer (DX)
//  application either as a standalone application or as a Domio Server Add-In. All options that control how
//  the runtime environment is initialised are determined by the current RunSettings object.
//
//  Parameters
//
//  1  -  int				-  Count of parameters that were used to invoke the application entry point
//  2  -  char *[]			-  Array of string parameters that were used o invoke the application
//
//  Returns
//
//  BOOL  -			TRUE if the initialisation was successful
//                  FALSE if the initialisation was not successful
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::InitEnvironment(int argc, char *argv[])
{
	STATUS			stAPIRC;						//  Return call from Notes API calls
	HMODULE			hmProcessor;					//  Handle to this process module
	int				iMQSuffix;						//  Message Queue Suffix for multiple instances
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer
	char			szAppName[MAXAPPNAME + 1];		//  Application Name

	//  Check for NULL RunSettings passed
	if (rsLocal == NULL)
	{
		RunningOnServer = FALSE;
		LogMessage(MSG_DXR0061S, 0);
		return FALSE;
	}

	//  If not already created the create the default ElapsedTimer
	if (RunningTime == NULL) RunningTime = new ElapsedTimer();

	//  Initialise the Notes runitime
	NotesInitExtended(argc, argv);

	//  Determine if we are running on a server
	RunningOnServer = OSGetEnvironmentString("ServerKeyFileName", szMsg, MAX_MSG);

	//  Show that Notes Runtime is initialised
	bNRTInitialised = TRUE;
	tLastLogWrite = 0;

	//  Create the API Package Descriptor object
	pkgLocal = new APIPackages();

	LogTrace(TRACE_CORERT, MSG_DXR0062T, 0);

	//  Record the start runtime
	OSCurrentTIMEDATE(&tdStarted);

	//  If we are running as a Server Addin task then perform additional initialisation
	if (rsLocal->RunningAsAddin)
	{
		AddInQueryDefaults(&hmProcessor, &hDefDescriptor);
		if (hDefDescriptor != NULLHANDLE) AddInDeleteStatusLine(hDefDescriptor);
		hDefDescriptor = AddInCreateStatusLine(rsLocal->APPName);
		AddInSetDefaults(hmProcessor, hDefDescriptor);
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0024I, rsLocal->APPTitle, rsLocal->APPName, rsLocal->APPVer);
		AddInLogMessageText(szMsg, NOERROR);
		sprintf_s(szMsg, MAX_MSG, INIT_STATUS_LINE, rsLocal->APPTitle, rsLocal->APPVer);
		AddInSetStatusText(szMsg);
		LogTrace(TRACE_CORERT, MSG_DXR0063T, 0);

		if (rsLocal->NeedsMQ)
		{
			//  NOTE: MQ Queue names are forced to uppercase as some interfaces convert the queue name
			//        to uppercase. 
			//
			//  Open the message queue for this Addin Task - if multiple instances of the Add-In can
			//  be run on the same server then a unique queue name for the instance is generated
			if (rsLocal->AllowMultipleAddins)
			{
				iMQSuffix = 1;
				TrimToUpper_s(szAppName, MAXAPPNAME, rsLocal->APPName);
				stAPIRC = ERR_DUPLICATE_MQ;
				while (stAPIRC == ERR_DUPLICATE_MQ)
				{
					sprintf_s(szQName, MAXAPPNAME, "%s%s%i", TASK_QUEUE_PREFIX, szAppName, iMQSuffix);
					stAPIRC = MQCreate(szQName, 0, 0);
					if ((ERR(stAPIRC) != NOERROR) && (ERR(stAPIRC) != ERR_DUPLICATE_MQ))
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR0025E, szQName);
						LogMessage(szMsg, 0);
						GetAPIMessage(stAPIRC, szMsg);
						LogMessage(szMsg, 0);
						if (ERR(stAPIRC) == ERR_MQ_QUITTING)
						{
							//  The MQ has a quit already pending - clean it up
							LogMessage(MSG_DXR0062I, 0);
							DrainMQ(szQName);
						}
						return FALSE;
					}
					iMQSuffix++;
				}
				LogTrace(TRACE_CORERT, MSG_DXR0064T, 0);

				//  Now reset the status line so that it uses the MQ name for the application ID
				if (hDefDescriptor != NULLHANDLE) AddInDeleteStatusLine(hDefDescriptor);
				hDefDescriptor = AddInCreateStatusLine(szQName + 3);
				AddInSetDefaults(hmProcessor, hDefDescriptor);
				sprintf_s(szMsg, MAX_MSG, INIT_STATUS_LINE, rsLocal->APPTitle, rsLocal->APPVer);
				AddInSetStatusText(szMsg);
				LogTrace(TRACE_CORERT, MSG_DXR0065T, 0);
			}
			else
			{
				//  Use the application name as the MQ queue name.
				TrimToUpper_s(szAppName, MAXAPPNAME, rsLocal->APPName);
				sprintf_s(szQName, MAXAPPNAME, "%s%s", TASK_QUEUE_PREFIX, szAppName);
				stAPIRC = MQCreate(szQName, 0, 0);
				if (ERR(stAPIRC) != NOERROR)
				{
					sprintf_s(szMsg, MAX_MSG, MSG_DXR0025E, szQName);
					LogMessage(szMsg, 0);
					GetAPIMessage(stAPIRC, szMsg);
					LogMessage(szMsg, 0);
					return FALSE;
				}
				LogTrace(TRACE_CORERT, MSG_DXR0064T, 0);
			}
			//  Open the application queue name
			stAPIRC = MQOpen(szQName, 0, &hmqProcessor);
			if (ERR(stAPIRC) != NOERROR)
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR0026E, szQName);
				LogMessage(szMsg, 0);
				GetAPIMessage(stAPIRC, szMsg);
				LogMessage(szMsg, 0);
				return FALSE;
			}
			LogTrace(TRACE_CORERT, MSG_DXR0066T, 0);

			//  Show the queue name that is in use
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0027I, szQName + 3);
			LogMessage(szMsg, 0);
		}
	}

	//  Show that the Notes Runtime has been initialised
	LogVerbose(MSG_DXR0001I, 0);

	//  Get the current user name for the logger and/or repository create functions
	if (!GetCurrentUser())
	{
		return FALSE;
	}
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0067T, AbbreviatedUser);
	LogTrace(TRACE_CORERT, szMsg, 0);

	//  Check for a persistent log has been requested - but no repository
	if ((rsLocal->NoRepository) && (!rsLocal->NoAppLog)) rsLocal->NoRepository = FALSE;

	//  Open the repository database
	if (!rsLocal->NoRepository)
	{
		//  If a repository is needed - but none was specified then use the local log.nsf
		if (rsLocal->szRepDb[0] == '\0')
		{
			rsLocal->CreateRepository = FALSE;
			rsLocal->szRepServer[0] = '\0';
			strcpy_s(rsLocal->szRepDb, MAXPATH, DEFAULT_REPOSITORY);
			LogMessage(MSG_DXR0094W, 0);
		}

		//  Open/Create the repository
		if (rsLocal->CreateRepository) hdbRepository = CreateRepository(rsLocal->szRepServer, rsLocal->szRepDb, TRUE, 0);
		else hdbRepository = OpenDatabase(rsLocal->szRepServer, rsLocal->szRepDb);
		if (hdbRepository == NULLHANDLE)
		{
			LogMessage(MSG_DXR0003S, 0);
			return FALSE;
		}
		LogTrace(TRACE_CORERT, MSG_DXR0068T, 0);
	}

	//  Start the logger
	if (!rsLocal->NoAppLog)
	{
		if (!StartLogger())
		{
			LogMessage(MSG_DXR0019W, 0);
		}
	}
	LogTrace(TRACE_CORERT, MSG_DXR0069T, 0);

	//  Mark the runtime environment as initialised
	IsInitialised = TRUE;
	LogVerbose(MSG_DXR0017I, 0);
	LogTrace(TRACE_CORERT, MSG_DXR0070T, 0);
	return TRUE;
}

//  Close
//
//  This function will close down all runtime support capabilities that were initialised for this application,
//  capabilities include the logging interface and the Notes runtime.
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if all functions were closed successfully
//                  FALSE if not all functions could be closed
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::Close(void)
{
	//  This is a wrapper for the CoreClose() function
	return CoreClose();
}

//  CoreClose
//
//  This function will close down all runtime support capabilities that were initialised for this application,
//  capabilities include the logging interface and the Notes runtime.
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if all functions were closed successfully
//                  FALSE if not all functions could be closed
//
//  API Use Notes:
//
//

BOOL ExecEnvironment::CoreClose(void)
{
	STATUS		stAPIRC;										//  Return code from API calls
	char		szAPIMsg[MAX_MSG + 1];							//  API Error message

	//  Close the runtime environment
	LogTrace(TRACE_CORERT, MSG_DXR0071T, 0);

	//  If logging is active the close the logger
	if (LoggingState == LOGSTATE_ACTIVE)
	{
		LogMessage(MSG_DXR0018I, 0);
		CloseLogDoc();
		LoggingState = LOGSTATE_CLOSED;
		LogTrace(TRACE_CORERT, MSG_DXR0072T, 0);
	}

	//  Close the repository database
	if (hdbRepository != NULLHANDLE)
	{
		stAPIRC = NSFDbClose(hdbRepository);
		if (ERR(stAPIRC) != NOERROR)
		{
				LogMessage(MSG_DXR0005E, 0);
				GetAPIMessage(stAPIRC, szAPIMsg);
				LogMessage(szAPIMsg, 0);
		}
		else
		{
			hdbRepository = NULLHANDLE;
			LogVerbose(MSG_DXR0004I, 0);
		}
	}

	//  Delete the API Package Object
	if (pkgLocal != NULL)
	{
		delete pkgLocal;
		pkgLocal = NULL;
	}

	//  Terminate the Notes Runtime
	if (bNRTInitialised)
	{
		//  If we are running as a Server Addin Task then close the Message Queue and delete the status line
		if (rsLocal->RunningAsAddin)
		{
			if (hmqProcessor != NULLHANDLE) MQClose(hmqProcessor, 0);
			if (hDefDescriptor != NULLHANDLE) AddInDeleteStatusLine(hDefDescriptor);
		}

		NotesTerm();
		bNRTInitialised = FALSE;
		LogVerbose(MSG_DXR0002I, 0);
	}

	//  Delete the default Elapsed Timer
	if (RunningTime != NULL)
	{
		delete RunningTime;
		RunningTime = NULL;
	}
	return TRUE;
}

//  OpenDatabase
//
//  This function will open a Notes Database or Directory.
//
//  Parameters
//
//  1  -  char *			-  The server name where the database resides (or "Local")
//  2  -  char *			-  The path to the database relative to the Notes Data Directory
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be opened
//						NULLHANDLE if the database could not be opened
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::OpenDatabase(char *szServer, char *szDatabase)
{
	//  This function is a wrapper for the CoreOpenDatabase() function
	return CoreOpenDatabase(szServer, szDatabase, 0, 0);
}

//  OpenDatabase
//
//  This function will open a Notes Database or Directory, this variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the database resides (or "Local")
//  2  -  char *			-  The path to the database relative to the Notes Data Directory
//  3  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be opened
//						NULLHANDLE if the database could not be opened
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::OpenDatabase(char *szServer, char *szDatabase, int iThreadID)
{
	//  This function is a wrapper for the CoreOpenDatabase() function
	return CoreOpenDatabase(szServer, szDatabase, 0, iThreadID);
}

//  OpenDatabase
//
//  This function will open a Notes Database or Directory, this variant returns the MODE of the Db/Directory
//
//  Parameters
//
//  1  -  char *			-  The server name where the database resides (or "Local")
//  2  -  char *			-  The path to the database relative to the Notes Data Directory
//  3  -  USHORT *			-  Pointer to a USHORT where the database mode will be returned
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be opened
//						NULLHANDLE if the database could not be opened
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::OpenDatabase(char *szServer, char *szDatabase, USHORT *usMode)
{
	//  This function is a wrapper for the CoreOpenDatabase() function
	return CoreOpenDatabase(szServer, szDatabase, 0, usMode, 0);
}

//  OpenDatabase
//
//  This function will open a Notes Database or Directory, this variant returns the MODE of the Db/Directory.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the database resides (or "Local")
//  2  -  char *			-  The path to the database relative to the Notes Data Directory
//  3  -  USHORT *			-  Pointer to a USHORT where the database mode will be returned
//  4  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be opened
//						NULLHANDLE if the database could not be opened
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::OpenDatabase(char *szServer, char *szDatabase, USHORT *usMode, int iThreadID)
{
	//  This function is a wrapper for the CoreOpenDatabase() function
	return CoreOpenDatabase(szServer, szDatabase, 0, usMode, iThreadID);
}

//  OpenDatabase
//
//  This function will open a Notes Database or Directory, this variant returns the DBID of the database that was opened.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the database resides (or "Local")
//  2  -  char *			-  The path to the database relative to the Notes Data Directory
//  3  -  WORD				-  Flags to be used to open the database
//  4  -  DBID *			-  Ponter to a DBID structure where the DBID will be returned
//  5  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be opened
//						NULLHANDLE if the database could not be opened
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::OpenDatabase(char *szServer, char *szDatabase, WORD wOpenFlags, DBID *dbidDB, int iThreadID)
{
	//  This function is a wrapper for the CoreOpenDatabase() function
	return CoreOpenDatabase(szServer, szDatabase, wOpenFlags, dbidDB, iThreadID);
}

//  CoreOpenDatabase
//
//  This function will open a Notes Database or Directory.
//
//  Parameters
//
//  1  -  char *			-  The server name where the database resides (or "Local")
//  2  -  char *			-  The path to the database relative to the Notes Data Directory
//  3  -  WORD				-  Flags to be used when opening the database
//  4  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be opened
//						NULLHANDLE if the database could not be opened
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CoreOpenDatabase(char *szServer, char *szDatabase, WORD wOpenFlags, int iThreadID)
{
	STATUS			stAPIRC;									//  API return code
	WORD			wRetLen;									//  Returned sizes
	BOOL			bServerIsRemote = FALSE;					//  Default to Local Server
	BOOL			bDoPostOpenFixup = FALSE;					//  Perform a fixup after opening
	DBHANDLE		hdbThis;									//  Database Handle
	TIMEDATE		tdDataMod;									//  Timestamp of last data modification
	TIMEDATE		tdNonDataMod;								//  Timedtamp of last non-data modification
	BUILDVERSION	bvServer;									//  Build Version of the target server
	DWORD			dwSrcToTarg = 0;							//  Source to target latency
	DWORD			dwTargToSrc = 0;							//  Target to Source Latency
	WORD			wRmtVersion = 0;							//  Remote server Domino version
	WORD			wAdjOpenFlags;								//  Adjusted open flags
	clock_t			ctStartOpen;								//  Start Time for the open
	clock_t			ctEndOpen;									//  End Time of the open
	long			lElapsed;									//  Elapsed Time
	long			lTimeFactor = CLOCKS_PER_SEC;				//  Clock tick conversion factor
	char			szAPIMsg[MAX_MSG + 1];						//  API Error message
	char			szFullDbName[MAXPATH + 1];					//  Full Database Name
	char			szCanonicalServer[MAXUSERNAME + 1];			//  Canonical server name
	char			szAbbrevServer[MAXUSERNAME + 1];			//  Abbreviated server name
	char			szFmtRepID[MAX_REPID + 1];					//  Formatted Replica ID
	char			szDFileName[MAXPATH + 1];					//  Derived FileName

	//  Adjust the time conversion factor to report in milliseconds
	lTimeFactor = lTimeFactor / 1000;
	if (lTimeFactor == 0) lTimeFactor = 1;

	//  Open a Domino Database/Directory
	DNAbbreviate(0L, NULL, szServer, szAbbrevServer, MAXUSERNAME, &wRetLen);
	TrimToUpper_s(szAPIMsg, MAX_MSG, szAbbrevServer);

	//  Check if the specified server is the current server - if so covert to a local open
	if (strcmpi(szAPIMsg, AbbreviatedUser) == 0) szAPIMsg[0] = '\0';

	//  Start the timer for the open
	ctStartOpen = RunningTime->getElapsed();

	//  If the filename requested is a Replica ID then get the filename
	if (IsThisAReplicaID(szDatabase, szFmtRepID))
	{
		//  Lookup the filename in the catalog
		if (!LookupCatalog(szServer, szFmtRepID, szDFileName, iThreadID))
		{
			//  Show that the D/B Open failed
			sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0014E, szDatabase);
			LogMessage(szAPIMsg, iThreadID);
			GetAPIMessage(ERR_NOEXIST, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			SetLastOpenError(ERR_NOEXIST, iThreadID);
			return NULLHANDLE;
		}
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0150I, szFmtRepID, szDFileName, szServer);
		LogVerbose(szAPIMsg, iThreadID);
	}
	else strcpy_s(szDFileName, MAXPATH, szDatabase);

	if (strlen(szAPIMsg) == 0 || (strcmp(szAPIMsg, "LOCAL") == 0))
	{
		strcpy_s(szFullDbName, MAXPATH, szDFileName);
	}
	else
	{
		bServerIsRemote = TRUE;
		//  canonicalize the server name and build the NET path for the database
		stAPIRC = DNCanonicalize(0L, NULL, szServer, szCanonicalServer, MAXUSERNAME, &wRetLen);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0012S, szServer);
			LogMessage(szAPIMsg, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			SetLastOpenError(ERR(stAPIRC), iThreadID);
			return NULLHANDLE;
		}

		//  Obtain the server latency
		stAPIRC = NSFGetServerLatency(szCanonicalServer, SERVER_TIMEOUT, &dwSrcToTarg, &dwTargToSrc, &wRmtVersion);

		stAPIRC = OSPathNetConstruct(NULL, szCanonicalServer, szDFileName, szFullDbName);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0013S, szCanonicalServer, szDatabase);
			LogMessage(szAPIMsg, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			SetLastOpenError(ERR(stAPIRC), iThreadID);
			return NULLHANDLE;
		}
	}

	// If the database is on a remote server and a fixup has been requested then remove the fixup options
	// from the open flags but set a flag to do a fixup after the database is opened.
	// The fixup options are not supported on a remote database.

	wAdjOpenFlags = wOpenFlags;
	if (bServerIsRemote)
	{
		if (wAdjOpenFlags & (OPEN_FIXUP_FLAGS))
		{
			//  Switch off the Fixup bits
			wAdjOpenFlags = wAdjOpenFlags & (~(OPEN_FIXUP_FLAGS));
			//  Show that a post open fixup is required
			bDoPostOpenFixup = TRUE;
		}
	}

	//  Now open the database
	stAPIRC = NSFDbOpenExtended(szFullDbName, wAdjOpenFlags, NULLHANDLE, NULL, &hdbThis, &tdDataMod, &tdNonDataMod);
	//  If a post open fixup is requested and the open failed because a consistence check is needed
	//  then accept the error and process the fixup
	if ((ERR(stAPIRC) == ERR_NEEDS_FIXUP) && bServerIsRemote)
	{
		//  Ensure that post open fixup is requested
		bDoPostOpenFixup = TRUE;
		hdbThis = NULLHANDLE;
		stAPIRC = NOERROR;
	}
	if (ERR(stAPIRC) != NOERROR)
	{
		//  Show that the D/B Open failed
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0014E, szFullDbName);
		LogMessage(szAPIMsg, iThreadID);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, iThreadID);
		SetLastOpenError(ERR(stAPIRC), iThreadID);
		return NULLHANDLE;
	}

	//  If a post open fixup is required then perform it now
	if (bDoPostOpenFixup)
	{
		//  Re-open the database with a fixup
		hdbThis = ReopenWithFixup(hdbThis, szFullDbName, iThreadID);
		if (ERR(stAPIRC) != NOERROR)
		{
			//  Show that the D/B Open failed
			sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0076E, szFullDbName);
			LogMessage(szAPIMsg, iThreadID);
			SetLastOpenError(ERR(stAPIRC), iThreadID);
			return NULLHANDLE;
		}
	}
	ctEndOpen = RunningTime->getElapsed();

	//  Log the fact that the database has been opened
	sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0015I, szFullDbName);
	LogVerbose(szAPIMsg, iThreadID);

	//  Format the open timing message
	lElapsed = (ctEndOpen - ctStartOpen) / lTimeFactor;
	
	//  Get the domino version serving the database
	stAPIRC = NSFDbGetMajMinVersion(hdbThis, &bvServer);
	if (ERR(stAPIRC) != NOERROR) memset(&bvServer, 0, sizeof(BUILDVERSION));
	if (bvServer.QMUNumber > 128) bvServer.QMUNumber = 0;
	if (bvServer.FixpackNumber > 128) bvServer.FixpackNumber = 0;
	bvServer.FixpackNumber = bvServer.FixpackNumber + bvServer.QMUNumber;
	if (!bServerIsRemote) NSFDbGetBuildVersion(hdbThis, &wRmtVersion);

	sprintf_s(szCanonicalServer, MAXUSERNAME,
								 FORMAT_DVERSION,
								 bvServer.MajorVersion,
								 bvServer.MinorVersion,
								 bvServer.QMRNumber,
								 bvServer.FixpackNumber,
								 bvServer.HotfixNumber,
								 wRmtVersion);

	if (bServerIsRemote)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0074I, szServer, szCanonicalServer, lElapsed, dwSrcToTarg + dwTargToSrc, dwSrcToTarg, dwTargToSrc);
	}
	else
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0075I, AbbreviatedUser, szCanonicalServer, lElapsed);
	}
	LogVerbose(szAPIMsg, iThreadID);

	//  Return the handle to the open database
	SetLastOpenError(NOERROR, iThreadID);
	return hdbThis;
}

//  CoreOpenDatabase
//
//  This function will open a Notes Database or Directory. This variant will return the Database mode.
//
//  Parameters
//
//  1  -  char *			-  The server name where the database resides (or "Local")
//  2  -  char *			-  The path to the database relative to the Notes Data Directory
//  3  -  WORD				-  Flags to be used when opening the database
//  4  -  USHORT *			-  Pointer to a USHORT where the database mode will be returned
//  5  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be opened
//						NULLHANDLE if the database could not be opened
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CoreOpenDatabase(char *szServer, char *szDatabase, WORD wOpenFlags, USHORT *usMode, int iThreadID)
{
	STATUS			stAPIRC;									//  API return code
	DBHANDLE		hdbTemp;									//  Handle of D/B or Directory
	char			szAPIMsg[MAX_MSG + 1];						//  API Error message

	//  This variant of CoreOpenDatabase also returns the Mode of the opened database
	hdbTemp = CoreOpenDatabase(szServer, szDatabase, wOpenFlags, iThreadID);
	if (hdbTemp != NULLHANDLE)
	{
		//  Get the Mode of the Database (or Directory)
		stAPIRC = NSFDbModeGet(hdbTemp, usMode);				//  Get the Mode
		if (ERR(stAPIRC) != NOERROR)
		{
			NSFDbClose(hdbTemp);								//  Close the Database
			hdbTemp = NULLHANDLE;
			LogMessage(MSG_DXR0021E, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
		}
	}
	else
	{
		*usMode = 0;
	}
	return hdbTemp;
}

//  CoreOpenDatabase
//
//  This function will open a Notes Database or Directory, this variant returns the DBID of the database that was opened.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the database resides (or "Local")
//  2  -  char *			-  The path to the database relative to the Notes Data Directory
//  3  -  WORD				-  Flags to be used to open the database
//  4  -  DBID *			-  Ponter to a DBID structure where the DBID will be returned
//  5  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be opened
//						NULLHANDLE if the database could not be opened
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CoreOpenDatabase(char *szServer, char *szDatabase, WORD wOpenFlags, DBID *dbidDB,int iThreadID)
{
	STATUS			stAPIRC;									//  API return code
	DBHANDLE		hdbTemp;									//  Handle of D/B or Directory
	char			szAPIMsg[MAX_MSG + 1];						//  API Error message

	//  This variant of CoreOpenDatabase also returns the DBID of the opened database
	hdbTemp = CoreOpenDatabase(szServer, szDatabase, wOpenFlags, iThreadID);
	if (hdbTemp != NULLHANDLE)
	{
		//  Get the DBID of the Database 
		stAPIRC = NSFDbIDGet(hdbTemp, dbidDB);
		if (ERR(stAPIRC) != NOERROR)
		{
			NSFDbClose(hdbTemp);								//  Close the Database
			hdbTemp = NULLHANDLE;
			LogMessage(MSG_DXR0052E, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
		}
	}
	return hdbTemp;
}

//  LookupCatalog
//
//  This function will lookup the passed replica ID in the catalog on the indicated server, if it can find an entry it will
//  return the filename of the database with the replica ID on the indicated server.
//
//  Parameters
//
//  1  -  char *			-  The server name where the database resides (or "Local")
//  2  -  char *			-  The requested Replica ID
//  3  -  char *			-  The path to the database relative to the Notes Data Directory
//  4  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  BOOL	-			TRUE if the Replica ID was found on the target serve, FALSE if not.
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::LookupCatalog(char *szServer, char *szRepID, char *szFile, int iThreadID)
{
	STATUS				stAPIRC;									//  Return code from notes API calls
	DBHANDLE			hdbCatalog;									//  Handle of the database catalog
	NOTEID				nidView;									//  Note ID of the by Replica ID View
	HCOLLECTION			hcLUView;									//  Handle of the lookup view
	COLLECTIONPOSITION	cpView;										//  Position in the view
	DWORD				dwMatches;									//  Count of matching notes
	BOOL				bKeepReading;								//  Loop control
	WORD				wSignalFlags;								//  Signal Flags from NIF Read
	DWORD				dwSkipsNeeded;								//  Skips needed on the read
	DWORD				dwSkipped;									//  Count of entries skipped
	DWORD				dwRead;										//  Count of entries read
	DHANDLE				hRetBfr;									//  Handle of the returned buffer
	WORD				wBfrLen;									//  Length of the returned buffer
	WORD				wDataType;									//  Data type of an item in the summary buffer
	WORD				wDataLen;									//  Length of an entry in the summary buffer
	char				*pData;										//  Pointer to a value in the summary buffer
	UCHAR				*pSumBfr;									//  Pointer to the summary buffer
	ITEM_TABLE			*pItemTable;								//  Typed pointer to the summary buffer header
	char				szCommonTarget[MAXUSERNAME + 1];			//  Common target server name
	char				szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  Open the database catalog on the requested server
	hdbCatalog = OpenDatabase(szServer, ADMIN_CATALOG5_FILE, iThreadID);
	if (hdbCatalog == NULLHANDLE) return FALSE;

	//  Find the view to lookup by Replica ID
	stAPIRC = NIFFindDesignNote(hdbCatalog, CATALOG_LU_REPID, NOTE_CLASS_VIEW, &nidView);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0142E, CATALOG_LU_REPID);
		LogMessage(szMsg, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		NSFDbClose(hdbCatalog);
		return FALSE;
	}

	//  Open the collection for the lookup by replica ID view
	stAPIRC = NIFOpenCollection(hdbCatalog,
								hdbCatalog,
								nidView,
								0,
								NULLHANDLE,
								&hcLUView,
								NULL,
								NULL,
								NULL,
								NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0143E, CATALOG_VIEW_REPLICA_ID);
		LogMessage(szMsg, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		NSFDbClose(hdbCatalog);
		return FALSE;
	}

	// Locate the matching notes in the view
	stAPIRC = NIFFindByName(hcLUView, szRepID, FIND_EQUAL, &cpView, &dwMatches);
	if (ERR(stAPIRC) == ERR_NOT_FOUND)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0144E, szRepID, szServer);
		LogMessage(szMsg, iThreadID);
		NIFCloseCollection(hcLUView);
		NSFDbClose(hdbCatalog);
		return FALSE;
	}
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0145E, szRepID, szServer);
		LogMessage(szMsg, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		NIFCloseCollection(hcLUView);
		NSFDbClose(hdbCatalog);
		return FALSE;
	}

	//  Get the common name of the target server to match against entries
	if ((strlen(szServer) == 0) || (strcmpi(szServer, "LOCAL") == 0))
	{
		strcpy_s(szCommonTarget, MAXUSERNAME, CommonUser);
	}
	else
	{
		stAPIRC = DNAbbreviate(0L, NULL, szServer, szCommonTarget, MAXUSERNAME, &wDataLen);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0149E, szServer);
			LogMessage(szMsg, iThreadID);
			GetAPIMessage(stAPIRC, szMsg);
			LogMessage(szMsg, iThreadID);
			NIFCloseCollection(hcLUView);
			NSFDbClose(hdbCatalog);
			return FALSE;
		}
		szCommonTarget[wDataLen] = '\0';
		pData = szCommonTarget;
		while (*pData != '\0')
		{
			if (*pData == '/') *pData = '\0';
			else pData++;
		}
	}

	//  Loop processing the summary buffers for the notes following the determined position
	
	bKeepReading = TRUE;
	dwSkipsNeeded = 0;
	while (bKeepReading)
	{
		stAPIRC = NIFReadEntries(hcLUView,
								 &cpView,
								 NAVIGATE_NEXT,
								 dwSkipsNeeded,
								 NAVIGATE_NEXT,
								 1L,
								 READ_MASK_SUMMARY,
								 &hRetBfr,
								 &wBfrLen,
								 &dwSkipped,
								 &dwRead,
								 &wSignalFlags);
		dwSkipsNeeded = 1;
		if (ERR(stAPIRC) != NOERROR)
		{
			LogMessage(MSG_DXR0146E, iThreadID);
			GetAPIMessage(stAPIRC, szMsg);
			LogMessage(szMsg, iThreadID);
			bKeepReading = FALSE;
		}
		else
		{
			if (hRetBfr == NULLHANDLE) bKeepReading = FALSE;
			else
			{
				if (dwRead == 0) bKeepReading = FALSE;
				else
				{
					//  Lock the summary buffer
					pSumBfr = (UCHAR *) OSLockObject(hRetBfr);

					//  Cast a pointer to the ITEM_TABLE header of the summary buffer
					pItemTable = (ITEM_TABLE *) pSumBfr;
					//  Check the count - if we have reached a next category then cop out
					if (pItemTable->Items == 1) bKeepReading = FALSE;
					else
					{
						//  Firstly check that we are still on the same Replica ID
						if (!NSFLocateSummaryValue(pSumBfr, CATALOG_REPID_COL, &pData, &wDataLen, &wDataType))
						{
							sprintf_s(szMsg, MAX_MSG, MSG_DXR0147E, CATALOG_REPID_COL);
							LogMessage(szMsg, iThreadID);
							DumpMemoryExt(pSumBfr, wBfrLen, "Summary Buffer", iThreadID);
							bKeepReading = FALSE;
						}
						else
						{
							if (wDataType != TYPE_TEXT)
							{
								sprintf_s(szMsg, MAX_MSG, MSG_DXR0148E, CATALOG_REPID_COL, wDataType);
								LogMessage(szMsg, iThreadID);
								DumpMemoryExt(pSumBfr, wBfrLen, "Summary Buffer", iThreadID);
								bKeepReading = FALSE;
							}
							else
							{
								//  Check the Replica ID
								if (memicmp(pData, szRepID, strlen(szRepID)) == 0)
								{
									//  Now check if this is the entry for the target server
									if (!NSFLocateSummaryValue(pSumBfr, CATALOG_SERVER_COL, &pData, &wDataLen, &wDataType))
									{
										sprintf_s(szMsg, MAX_MSG, MSG_DXR0147E, CATALOG_SERVER_COL);
										LogMessage(szMsg, iThreadID);
										DumpMemoryExt(pSumBfr, wBfrLen, "Summary Buffer", iThreadID);
										bKeepReading = FALSE;
									}
									else
									{
										//  Check the type returned
										if (wDataType != TYPE_TEXT)
										{
											sprintf_s(szMsg, MAX_MSG, MSG_DXR0148E, CATALOG_SERVER_COL, wDataType);
											LogMessage(szMsg, iThreadID);
											DumpMemoryExt(pSumBfr, wBfrLen, "Summary Buffer", iThreadID);
											bKeepReading = FALSE;
										}
										else
										{
											if (memicmp(pData, szCommonTarget, strlen(szCommonTarget)) == 0)
											{
												//  Server name has matched - get the path name to return to the caller
												if (!NSFLocateSummaryValue(pSumBfr, CATALOG_ITEM_PATHNAME, &pData, &wDataLen, &wDataType))
												{
													sprintf_s(szMsg, MAX_MSG, MSG_DXR0147E, CATALOG_ITEM_PATHNAME);
													LogMessage(szMsg, iThreadID);
													DumpMemoryExt(pSumBfr, wBfrLen, "Summary Buffer", iThreadID);
													bKeepReading = FALSE;
												}
												else
												{
													strncpy_s(szFile, MAXPATH, pData, wDataLen);
													szFile[wDataLen] = '\0';
													OSUnlockObject(hRetBfr);
													OSMemFree(hRetBfr);
													NIFCloseCollection(hcLUView);
													NSFDbClose(hdbCatalog);
													return TRUE;
												}
											}
										}
									}
								}
							}
						}
					}
					OSUnlockObject(hRetBfr);
				}
				OSMemFree(hRetBfr);
			}
		}
	}

	NIFCloseCollection(hcLUView);
	NSFDbClose(hdbCatalog);
	return FALSE;
}

//  GetAPIMessage
//
//  This function will return the Notes API message string for the passed API error code.
//
//  Parameters
//
//  1  -  STATUS			-  The Notes API Error code
//  2  -  char *			-  Pointer to the string buffer where the message will be returned
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::GetAPIMessage(STATUS stAPICode, char * szMsg)
{
	char		szRemote[5];							//  Local or Remote indicator
	char		szPkgID[MAX_PKGID + 1];
	char		szAPIMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Indicate if the message is local or remote
	if (stAPICode & STS_REMOTE) strcpy_s(szRemote, 5, "RMT");
	else strcpy_s(szRemote, 5, "LCL");

	//  Get an API Error Message 
	OSLoadString(NULLHANDLE, ERR(stAPICode), szAPIMsg, MAX_MSG);

	//  Get a Package ID string
	pkgLocal->GetPackageID(stAPICode, szPkgID, MAX_PKGID);

	//  Format the Error Line
	sprintf_s(szMsg, MAX_MSG, LOG_API_MSG, szRemote, ERR(stAPICode), szPkgID, szAPIMsg);
}

//  LogMessage
//
//  This function will write a log message to the currently active logger(s).
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string buffer containing the message to be written
//
//  Returns
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::LogMessage(char *szMsg)
{
	//  This function is a wrapper for the CoreLogMessage() function
	return CoreLogMessage(szMsg, NULL, 0);
}

//  LogMessage
//
//  This function will write a log message to the currently active logger(s). This variant is for use by
//  multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string buffer containing the message to be written
//  2  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  BOOL		- TRUE if the message was logged, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::LogMessage(char *szMsg, int iThreadID)
{
	//  This function is a wrapper for the CoreLogMessage() function
	return CoreLogMessage(szMsg, NULL, iThreadID);
}

//  CoreLogMessage
//
//  This function will write a log message to the currently active logger(s). This variant is for use by
//  multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string buffer containing the message to be written
//  2  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  BOOL		- TRUE if the message was logged, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::CoreLogMessage(char *szMsg, TIMEDATE *tdEvt, int iThreadID)
{
	char			szIMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Write a log message
	if (iThreadID != 0) sprintf_s(szIMsg, MAX_MSG, "%s [%i]", szMsg, iThreadID);
	else strcpy_s(szIMsg, MAX_MSG, szMsg);

	//  If log status is ok then write the message & Application Logging is allowed
	if ((LoggingState >= LOGSTATE_ACTIVE) && (!rsLocal->NoAppLog))
	{
		if (!LogEvent(szIMsg, tdEvt, iThreadID))
		{
			//  Event logging failed - terminate the event log
			LoggingState = LOGSTATE_CLOSED;
			LogMessage(MSG_DXR0007E, 0);
			CloseLogDoc();
		}
	}

	//  If requested or logging not available then echo the message to the console
	if ((rsLocal->EchoLog) || (LoggingState < LOGSTATE_ACTIVE))
	{
		//  If the application is running on a Server then echo to the Domino Log
		if ((RunningOnServer) && (bNRTInitialised))
		{
			//  Eliminate Escapae characters
			RemoveEscapes(szMsg, MAX_MSG+1);
			AddInLogMessageText(szMsg, NOERROR);
		}
		else
		{
			RemoveEscapes(szMsg, MAX_MSG+1);
			std::cout << szMsg << std::endl;
		}
	}
	return TRUE;
}

//  StartLogger
//
//  This function will initialise the logging interface desired by the application, by default this is
//  an event style log in the repository database.
//
//  Parameters
//
//  Returns
//
//  BOOL		- TRUE if the logger was started, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::StartLogger(void)
{
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Start the Logger
	hnLogDoc = NULLHANDLE;							//  Coerce starting a new log document
	LoggingState = LOGSTATE_ACTIVE;					//  Use the active logging path

	// Write the start message to the log
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0006I, rsLocal->APPTitle, rsLocal->APPName, rsLocal->APPVer);
	LogMessage(szMsg, 0);
	return TRUE;
}

//  LogEvent
//
//  This function will write a log message into the current event log document. The function is a warpper
//  for the CoreLogEvent() function for which it will generate a default event time stamp.
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string buffer containing the message to be written
//
//  Returns
//
//  BOOL		- TRUE if the message was logged, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::LogEvent(char *szMsg, TIMEDATE *tdEvt, int iThreadID)
{
	TIMEDATE		tdCurrent;							//  Current timedate used to decorate the message
	char			szIMsg[MAX_MSG + 1];				//  Generic Mesage Buffer

	//  Format the message if needed
	if (iThreadID != 0) sprintf_s(szIMsg, MAX_MSG,"%s [%i]", szMsg, iThreadID);
	else strcpy_s(szIMsg, MAX_MSG, szMsg);

	//  Create an event time stamp
	if (tdEvt == NULL) OSCurrentTIMEDATE(&tdCurrent);
	else memcpy(&tdCurrent, tdEvt, sizeof(TIMEDATE));
	return CoreLogEvent(szIMsg, &tdCurrent);
}

//  CoreLogEvent
//
//  This function will write a log message into the current event log document.
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string buffer containing the message to be written
//  2  -  TIMEDATE *		-  Pointer to the event timestamp for this log entry
//
//  Returns
//
//  BOOL		- TRUE if the message was logged, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::CoreLogEvent(char *szMsg, TIMEDATE *tdEvt)
{
	STATUS			stAPIRC;							//  Return code from Notes API calls
	INTLFORMAT		ifCurrent;							//  International settings to use
	TFMT			tfLogTime;							//  Format for the text version of the timestamp
	WORD			wRetLen;							//  Returned length
	clock_t			tLogInterval;						//  Interval since last log write
	clock_t			tLogStaleLimit;						//  Limit to determine that a log document is stale
	char			szTimeStamp[MAXALPHATIMEDATE + 1];	//  Formatted timestamp
	char			szMessageString[MAX_MSG + 1];		//  Generic message buffer

	//  Log a message using the standard Domino event style

	if ((LoggingState == LOGSTATE_ACTIVE) || (hdbLog != NULLHANDLE))
	{
		//  Check to see if the current log document is full, if so, switch to a new one
		LogLines++;
		if (LogLines > LOGLINES_MAX)
		{
			CloseLogDoc();
		}

		//  If a current log document does not exist then create a new one
		if (hnLogDoc == NULLHANDLE)
		{
			if (!CreateLogDoc())
			{
				return FALSE;
			}
			tLastLogWrite = 0;
		}
	}
	//  Write the event to the current log document

	//  Firstly format the TimeStamp for the Log Entry
	OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
	tfLogTime.Date = TDFMT_FULL;
	tfLogTime.Time = TTFMT_FULL;
	tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
	tfLogTime.Structure = TSFMT_DATETIME;
	stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
									&tfLogTime,
									tdEvt,
									szTimeStamp,
									sizeof(szTimeStamp),
									&wRetLen);
	if (ERR(stAPIRC) != NOERROR)
	{
		LoggingState = LOGSTATE_CLOSED;
		LogMessage(MSG_DXR0010E, 0);
		GetAPIMessage(stAPIRC, szMessageString);
		LogMessage(szMessageString, 0);
		return FALSE;
	}
	szTimeStamp[wRetLen] = 0;

	//  Format the decorated log line
	sprintf_s(szMessageString, MAX_MSG, "%s: %s", szTimeStamp, szMsg);

	if ((LoggingState == LOGSTATE_ACTIVE) || (hdbLog != NULLHANDLE))
	{
		//  Append the line to the current log document
		stAPIRC = NSFItemAppendTextList(hnLogDoc,
										LOG_EVENT_LIST,
										szMessageString,
										(WORD) strlen(szMessageString),
										TRUE);
		if (ERR(stAPIRC) != NOERROR)
		{
			LoggingState = LOGSTATE_CLOSED;
			LogMessage(MSG_DXR0011E, 0);
			GetAPIMessage(stAPIRC, szMessageString);
			LogMessage(szMessageString, 0);
			return FALSE;
		}

		//  Increment the count of non-flushed lines and force an update if they exceed the max
		LogLinesPending++;
		if (LogLinesPending > MAX_PENDING_LOGLINES)
		{
			NSFNoteUpdate(hnLogDoc, UPDATE_FORCE);
			LogLinesPending = 0;
			tLastLogWrite = RunningTime->getElapsed();
		}
		else
		{
			//  Determine if the log document is stale and should be updated
			tLogInterval = RunningTime->getElapsed();
			tLogInterval -= tLastLogWrite;
			tLogStaleLimit = MAX_STALE_LOG_MINS * 60;
			tLogStaleLimit = tLogStaleLimit * CLOCKS_PER_SEC;
			if (tLogInterval >= tLogStaleLimit)
			{
				NSFNoteUpdate(hnLogDoc, UPDATE_FORCE);
				LogLinesPending = 0;
				tLastLogWrite = RunningTime->getElapsed();
			}
		}
	}
	else std::cout << szMessageString << std::endl;
	return TRUE;
}

//  LogFlush
//
//  This function will force an update (write) of the current log document.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogFlush(void)
{
	if (hnLogDoc != NULLHANDLE) NSFNoteUpdate(hnLogDoc, UPDATE_FORCE);				//  Update the note
	return;
}

//  LogCommit
//
//  This function will force an update (write) of the current log document if it is needed
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogCommit(void)
{
	if (hnLogDoc != NULLHANDLE)
	{
		if (LogLinesPending > 0)
		{
			NSFNoteUpdate(hnLogDoc, UPDATE_FORCE);
			LogLinesPending = 0;
		}
	}
	return;
}

//  CloseLogDoc
//
//  This function will close the current log document, updating the document with the end time.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::CloseLogDoc(void)
{
	STATUS			stAPIRC;						//  Return code from Notes API Calls
	TIMEDATE		tdCurrent;						//  Ending timestamp for this log document

	//  Close the current log document
	if (hnLogDoc != NULLHANDLE)
	{
		OSCurrentTIMEDATE(&tdCurrent);						//  Set the current TimeStamp
		stAPIRC = NSFItemAppend(hnLogDoc, ITEM_SUMMARY, LOG_FINISH_TIME, (WORD) strlen(LOG_FINISH_TIME),
								TYPE_TIME, &tdCurrent, (DWORD) sizeof(TIMEDATE));
		NSFNoteUpdate(hnLogDoc, UPDATE_FORCE);				//  Update the note
		NSFNoteClose(hnLogDoc);								//  Close the note
		hnLogDoc = NULLHANDLE;								//  Show that the note is closed
	}
	return;
}

//  CreateLogDoc
//
//  This function will create and initialise a new log document
//
//  Parameters
//
//  Returns
//
//  BOOL			-	TRUE if the document was created, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::CreateLogDoc(void)
{
	STATUS			stAPIRC;									//  API return code
	TIMEDATE		tdCurrent;									//  Timestamp
	char			szAPIMsg[MAX_MSG + 1];						//  API Error message

	//  Create a new log document
	if (LoggingState > LOGSTATE_ACTIVE) stAPIRC = NSFNoteCreate(hdbLog, &hnLogDoc);
	else stAPIRC = NSFNoteCreate(hdbRepository, &hnLogDoc);
	if (ERR(stAPIRC) != NOERROR)
	{
		LoggingState = LOGSTATE_CLOSED;
		LogMessage(MSG_DXR0008E, 0);
		return FALSE;
	}

	//  Write the identification fields to the new log document
	stAPIRC = NSFItemAppend(hnLogDoc, ITEM_SUMMARY, "Form", (WORD) strlen("Form"),
							TYPE_TEXT, LOG_FORM_NAME, (DWORD) strlen(LOG_FORM_NAME));
	if (ERR(stAPIRC) != NOERROR)
	{
		LoggingState = LOGSTATE_CLOSED;
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0009E, "Form");
		LogMessage(szAPIMsg, 0);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, 0);
		return FALSE;
	}

	//  Format the identification for these event records "Application title(Common Username)"
	sprintf_s(szAPIMsg, MAX_MSG, "%s(%s)", rsLocal->APPTitle, CommonUser);
	stAPIRC = NSFItemAppend(hnLogDoc, ITEM_SUMMARY, LOG_FORM_SERVER, (WORD) strlen(LOG_FORM_SERVER),
							TYPE_TEXT, szAPIMsg, (DWORD) strlen(szAPIMsg));
	if (ERR(stAPIRC) != NOERROR)
	{
		LoggingState = LOGSTATE_CLOSED;
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0009E, LOG_FORM_SERVER);
		LogMessage(szAPIMsg, 0);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, 0);
		return FALSE;
	}

	OSCurrentTIMEDATE(&tdCurrent);						//  Set the current TimeStamp
	stAPIRC = NSFItemAppend(hnLogDoc, ITEM_SUMMARY, LOG_START_TIME, (WORD) strlen(LOG_START_TIME),
							TYPE_TIME, &tdCurrent, (DWORD) sizeof(TIMEDATE));
	if (ERR(stAPIRC) != NOERROR)
	{
		LoggingState = LOGSTATE_CLOSED;
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0009E, LOG_START_TIME);
		LogMessage(szAPIMsg, 0);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, 0);
		return FALSE;
	}

	//  Update (save) the empty log document and store it's NoteID
	NSFNoteUpdate(hnLogDoc, UPDATE_FORCE);
	NSFNoteGetInfo(hnLogDoc, _NOTE_ID, &nidLogDoc);

	LogLines = 0;
	LogLinesPending = 0;
	return TRUE;
}

//  GetCurrentUser
//
//  This function will populate the user name entry in the runtime object with the currently active user name.
//  It will build the Canonical, Abbreviated and Common variations of the name.
//
//  Parameters
//
//  Returns
//
//  BOOL			-	TRUE if the name variants were set, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::GetCurrentUser(void)
{
	STATUS			stAPIRC;									//  API return code
	int				iIndex;										//  Generic index
	char			szAPIMsg[MAX_MSG + 1];						//  API Error message

	//  Get the current user name
	stAPIRC = SECKFMGetUserName(CanonicalUser);					//  Get the user name
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0020E, 0);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, 0);
		return FALSE;
	}

	if (strcmp(CanonicalUser, "") == 0)
	{
		//  Local unauthenticated user
		strcpy_s(CanonicalUser, MAXUSERNAME, "Local User");
		strcpy_s(AbbreviatedUser, MAXUSERNAME, "Local User");
		strcpy_s(CommonUser, MAXUSERNAME, "Local User");
	}

	//  Build the abbreviated version
	stAPIRC = DNAbbreviate(0L, NULL, CanonicalUser, AbbreviatedUser, MAXUSERNAME, NULL);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0020E, 0);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, 0);
		return FALSE;
	}

	//  Get the common name
	for (iIndex = 0; iIndex < (WORD) strlen(AbbreviatedUser); iIndex++)
	{
		if ((AbbreviatedUser[iIndex] == '\0') || (AbbreviatedUser[iIndex] == '/'))
		{
			CommonUser[iIndex] = '\0';
		}
		else
		{
			CommonUser[iIndex] = AbbreviatedUser[iIndex];
		}
	}
	return TRUE;
}

//  DumpMemory
//
//  This is a logging function that can be called (in Debug mode) to provide a Dump of a memory are in the log.
//
//  Parameters
//
//	1 - void *			-  Pointer to the area of memory that is to be dumped
//  2 - int				-  Length of the memory area to be dumped
//  3 - char *			-  String that describes the memory are being dumped
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::DumpMemory(void *lpMem, int iLen, char *szSymbol)
{
	//  This function is a wrapper for the DumpMemory() (MT) function
	DumpMemory(lpMem, iLen, szSymbol, 0);
	return;
}

//  DumpMemory
//
//  This is a logging function that can be called (in Debug mode) to provide a Dump of a memory are in the log.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//	1 - void *			-  Pointer to the area of memory that is to be dumped
//  2 - int				-  Length of the memory area to be dumped
//  3 - char *			-  String that describes the memory are being dumped
//  4 - int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::DumpMemory(void *lpMem, int iLen, char *szSymbol, int iThreadID)
{
	char			szMsg[MAX_MSG + 1];							//  Generic message buffer

	//  If the log mode is lower than TRACE then silently suppress the dump
	if (rsLocal->LogLevel < LOGLEVEL_TRACE) return;

	//  If the log mode is TRACE then supress the dump (with a messge)
	if (rsLocal->LogLevel == LOGLEVEL_TRACE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0023I, szSymbol, lpMem);
		LogMessage(szMsg, iThreadID);
		return;
	}

	//  This function is a wrapper for the CoreDumpMemory() function
	CoreDumpMemory(lpMem, iLen, szSymbol, iThreadID);
	return;
}

//  DumpMemoryExt
//
//  This is a logging function that can be called to provide a Dump of an area of memory.
//
//  Parameters
//
//	1 - void *			-  Pointer to the area of memory that is to be dumped
//  2 - int				-  Length of the memory area to be dumped
//  3 - char *			-  String that describes the memory are being dumped
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::DumpMemoryExt(void *lpMem, int iLen, char *szSymbol)
{
	//  This function is a wrapper for the CoreDumpMemory() function
	CoreDumpMemory(lpMem, iLen, szSymbol, 0);
}

//  DumpMemoryExt
//
//  This is a logging function that can be called to provide a Dump of an area of memory
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//	1 - void *			-  Pointer to the area of memory that is to be dumped
//  2 - int				-  Length of the memory area to be dumped
//  3 - char *			-  String that describes the memory are being dumped
//  4 - int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::DumpMemoryExt(void *lpMem, int iLen, char *szSymbol, int iThreadID)
{
	//  This function is a wrapper for the CoreDumpMemory() function
	CoreDumpMemory(lpMem, iLen, szSymbol, iThreadID);
}

//  CoreDumpMemory
//
//  This is a logging function that can be called (in Debug mode) to provide a Dump of a memory are in the log.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//	1 - void *			-  Pointer to the area of memory that is to be dumped
//  2 - int				-  Length of the memory area to be dumped
//  3 - char *			-  String that describes the memory are being dumped
//  4 - int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::CoreDumpMemory(void *lpMem, int iLen, char *szSymbol, int iThreadID)
{
	//  Dump a portion of memory to the log
	int				iResidue;									//  Count of bytes left to dump
	int				iIndex;										//  Index counter
	int				iOffset = 0;								//  Offset in the memory buffer
	char			szMsg[MAX_MSG + 1];							//  Generic message buffer
	char			szHexSeg1[30];								//  First segment of Hexadecimal
	char			szHexSeg2[30];								//  Second segment of Hexadecimal
	char			szRawSeg1[10];								//  First segment of Ascii data
	char			szRawSeg2[10];								//  Second segment of Ascii data
	char			szTemp[20];									//  Temporary format area
	char			cLocalBuffer[20];							//  Local buffer
	char			*BfrPtr;									//  Pointer into the memory buffer
	UCHAR			*LocalBfrPtr;								//  Pointer into the local buffer


	//  Show start of the dump
	sprintf_s(szMsg, MAX_MSG, START_DUMP, szSymbol, iLen, iLen, lpMem);
	while (strlen(szMsg) < 75)
	{
		strcat_s(szMsg, MAX_MSG, "-");
	}
	LogMessage(szMsg, iThreadID);
	
	//  Dump the complete lines
	iResidue = iLen;
	BfrPtr = (char *) lpMem;
	while (iResidue > 0)
	{
		szHexSeg1[0] = '\0';
		szHexSeg2[0] = '\0';
		szRawSeg1[8] = '\0';
		szRawSeg2[8] = '\0';
		//  Copy the data to the local buffer
		if (iResidue >= 16)
		{
			memcpy(cLocalBuffer, BfrPtr, 16);
		}
		else
		{
			memset(cLocalBuffer, 0, 16);
			memcpy(cLocalBuffer, BfrPtr, iResidue);
		}
		LocalBfrPtr = (UCHAR *) cLocalBuffer;

		for (iIndex = 0; iIndex < 16; iIndex++)
		{
			sprintf_s(szTemp, 20, "%02X ", (unsigned short) *LocalBfrPtr);
			if (iIndex < 8)
			{
				strcat_s(szHexSeg1, 30, szTemp);
				if ((*LocalBfrPtr < 33) || (*LocalBfrPtr > 126))
				{
					szRawSeg1[iIndex] = '.';
				}
				else
				{
					szRawSeg1[iIndex] = *LocalBfrPtr;
				}
			}
			else
			{
				strcat_s(szHexSeg2, 30, szTemp);
				if ((*LocalBfrPtr < 33) || (*LocalBfrPtr > 126))
				{
					szRawSeg2[iIndex - 8] = '.';
				}
				else
				{
					szRawSeg2[iIndex - 8] = *LocalBfrPtr;
				}
			}
			LocalBfrPtr++;
		}
		sprintf_s(szMsg, MAX_MSG, DUMP_LINE, iOffset, szHexSeg1, szHexSeg2, szRawSeg1, szRawSeg2);
		LogMessage(szMsg, iThreadID);
		iOffset = iOffset + 16;
		iResidue = iResidue - 16;
		BfrPtr = BfrPtr + 16;
	}

	//  Mark the end of the dump in the log
	sprintf_s(szMsg, MAX_MSG, END_DUMP, szSymbol, iLen, iLen, lpMem);
	while (strlen(szMsg) < 75)
	{
		strcat_s(szMsg, MAX_MSG, "-");
	}
	LogMessage(szMsg, iThreadID);
	return;
}

#ifdef WIN32
//  LogWINAPIError
//
//  This function will write the latest Windows API error to the currently active log.
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogWINAPIError(void)
{
	//  This function is a wrapper for the CoreLogWINAPIError() function
	CoreLogWINAPIError(0);
}
#endif

#ifdef WIN32
//  LogWINAPIError
//
//  This function will write the latest Windows API error to the currently active log. This variant is for
//  multi-threaded applications.
//
//  Parameters
//
//  1 - int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogWINAPIError(int iThreadID)
{
	//  This function is a wrapper for the CoreLogWINAPIError() function
	CoreLogWINAPIError(iThreadID);
}
#endif

#ifdef WIN32
//  GetWINAPIError
//
//  This function will retirn the formatted error message for the latest Windows API error.
//
//  Parameters
//
//	1 - DWORD			-  The Windows API Error Code
//  2 - char *			-  Pointer to a string buffer where the formatted error message will be returned
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::GetWINAPIError(DWORD dwErrorCode, char *szMsgBuffer)
{
	DWORD				dwMsgLen;						//  Length of message
	DWORD				dwWECode;						//  Alternate Error Code
	char				szMsg[MAX_MSG + 1];				//  OS message buffer

	//  This function will return a formatted windows API error message
	dwMsgLen = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, dwErrorCode, 0, szMsg, MAX_MSG, NULL);
	if (dwMsgLen == 0)
	{
		//  Error message could not be retrieved - try substituting the failure of formatting the message
		dwWECode = GetLastError();
		dwMsgLen = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, dwWECode, 0, szMsg, MAX_MSG, NULL);
	}
	sprintf_s(szMsgBuffer, MAX_MSG, MSG_DXR0022E, dwErrorCode, szMsg);
	return;
}
#endif

#ifdef UNIX
//  GetUNIXAPIError
//
//  This function will retirn the formatted error message for the latest API error.
//
//  Parameters
//
//	1 - int				-  The API Error Code
//  2 - char *			-  Pointer to a string buffer where the formatted error message will be returned
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::GetUNIXAPIError(int iErrorCode, char *szMsgBuffer)
{
	//  Return the formatted error code
	sprintf_s(szMsgBuffer, MAX_MSG, MSG_DXR0022E, iErrorCode, strerror(iErrorCode));
	return;
}
#endif

#ifdef WIN32
//  CoreLogWINAPIError
//
//  This function will write the latest Windows API error to the currently active log. This variant is for
//  multi-threaded applications.
//
//  Parameters
//
//  1 - int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::CoreLogWINAPIError(int iThreadID)
{
	DWORD				dwErrorCode;					//  OS Error Code
	DWORD				dwMsgLen;						//  Length of message
	DWORD				dwWECode;						//  Alternate Error Code
	char				szMsg[MAX_MSG + 1];				//  OS message buffer
	char				szFullMsg[MAX_MSG + 1];			//  Generic message buffer

	//  This function will log a windows API error
	dwErrorCode = GetLastError();
	dwMsgLen = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, dwErrorCode, 0, szMsg, MAX_MSG, NULL);
	if (dwMsgLen == 0)
	{
		//  Error message could not be retrieved - try substituting the failure of formatting the message
		dwWECode = GetLastError();
		dwMsgLen = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, dwWECode, 0, szMsg, MAX_MSG, NULL);
	}
	sprintf_s(szFullMsg, MAX_MSG, MSG_DXR0022E, dwErrorCode, szMsg);
	LogMessage(szFullMsg, iThreadID);
	return;
}
#endif

//  BuildIndexes
//
//  This function will (re)Build all of the view indexes in the current repository databasse.
//
//  Parameters
//
//  Returns
//
//  BOOL		-  TRUE if all indexes were rebuilt
//				-  FALSE if not all indexes could be rebuilt
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::BuildIndexes(void)
{
	//  Build all of the indexes in the repository
	return CoreBuildIndexes(hdbRepository, 0);
}

//  BuildIndexes
//
//  This function will (re)Build all of the view indexes in the specified database.
//
//  Parameters
//  1 - DBHANDLE		- Handle of the database where the view indexes will be rebuilt
//
//  Returns
//
//  BOOL		-  TRUE if all indexes were rebuilt
//				-  FALSE if not all indexes could be rebuilt
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::BuildIndexes(DBHANDLE hdbIX)
{
	//  This function is a wrapper for the CoreBuildIndexes() function.
	return CoreBuildIndexes(hdbIX, 0);
}

//  BuildIndexes
//
//  This function will (re)Build all of the view indexes in the specified database. This variant is for
//  use by multi-threaded applications.
//
//  Parameters
//
//  1 - DBHANDLE		- Handle of the database where the view indexes will be rebuilt
//  2 - int				- The ID of the thread from which the open request is being made
//
//  Returns
//  BOOL		-  TRUE if all indexes were rebuilt
//				-  FALSE if not all indexes could be rebuilt
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::BuildIndexes(DBHANDLE hdbIX, int iThreadID)
{
	//  This function is a wrapper for the CoreBuildIndexes() function.
	return CoreBuildIndexes(hdbIX, iThreadID);
}

//  CoreBuildIndexes
//
//  This function will (re)Build all of the view indexes in the specified database. This variant is for
//  use by multi-threaded applications.
//
//  Parameters
//
//  1 - DBHANDLE		- Handle of the database where the view indexes will be rebuilt
//  2 - int				- The ID of the thread from which the open request is being made
//
//  Returns
//
//  BOOL		-  TRUE if all indexes were rebuilt
//				-  FALSE if not all indexes could be rebuilt
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::CoreBuildIndexes(DBHANDLE hdbIX, int iThreadID)
{
	STATUS				stAPIRC;						//  Status code returmed from Notes API calls
	HCOLLECTION			hcDesign;						//  Handle to the Design collection
	COLLECTIONPOSITION	cpDesign;						//  Collection Position in the Design collection
	DHANDLE				hnidDesign;						//  Handle for the returned design collection
	NOTEID				*nidDesign;						//  Design Note IDs
	DWORD				dwNotesFound;					//  Number of design notes found
	DWORD				dwIndex;						//  Index in the design collection
	WORD				wSignalFlag;					//  NIF signals
	OID					oidDesign;						//  OID of the design note
	TIMEDATE			tdMod;							//  Last modified timestamp
	WORD				wNoteClass;						//  Note Class of the design note
	char				szMsg[MAX_MSG + 1];				//  Generic message buffer
	
	//  Build all of the indexes in the requested database

	//  Open the design collection
	stAPIRC = NIFOpenCollection(hdbIX, hdbIX, NOTE_CLASS_DESIGN | NOTE_ID_SPECIAL, 0L, NULLHANDLE, &hcDesign, NULLHANDLE, NULL, NULLHANDLE, NULLHANDLE);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0028E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	//  Read the design collection
	cpDesign.Level = 0;
	cpDesign.Tumbler[0] = 0;

	stAPIRC = NIFReadEntries(hcDesign,						// handle to this collection
							&cpDesign,						// where to start in collection
							NAVIGATE_NEXT,					// order to use when skipping
							1L,								// number to skip
							NAVIGATE_NEXT,					// order to use when reading
							0xFFFFFFFF,						// max number to read 
							READ_MASK_NOTEID,				// info we want 
							&hnidDesign,					// handle to info buffer (return)
							NULL,							// length of info buffer (return)
							NULL,							// entries skipped (return)
							&dwNotesFound,					// entries read (return)
							&wSignalFlag);					// share warning and more signal flag (return)

	if (ERR(stAPIRC) != NOERROR)
	{
		//  Bad RC from NIFReadEntries
		NIFCloseCollection(hcDesign);
		LogMessage(MSG_DXR0029E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return FALSE;
	}
									
	if (hnidDesign == NULLHANDLE)
	{
		//  No handle to returned collection
		NIFCloseCollection(hcDesign);
		LogMessage(MSG_DXR0030E, iThreadID);
		return FALSE;
	}

	//  See if there is any design to process
	if (dwNotesFound == 0)
	{
		NIFCloseCollection(hcDesign);
		LogMessage(MSG_DXR0030E, iThreadID);
		return FALSE;
	}

	nidDesign = (NOTEID *) OSLockObject(hnidDesign);

	for (dwIndex = 0; dwIndex < dwNotesFound; dwIndex++)
	{
		//  Determine if the design note is a View
		stAPIRC = NSFDbGetNoteInfo(hdbIX, nidDesign[dwIndex], &oidDesign, &tdMod, &wNoteClass);
		if (ERR(stAPIRC) != NOERROR)
		{
			LogMessage(MSG_DXR0031E, iThreadID);
			GetAPIMessage(stAPIRC, szMsg);
			LogMessage(szMsg, iThreadID);
			OSUnlock(hnidDesign);
			OSMemFree(hnidDesign);
			NIFCloseCollection(hcDesign);
			return FALSE;
		}
		if (wNoteClass == NOTE_CLASS_VIEW)
		{
			//  For each view - build the Index
			BuildIndex(hdbIX, nidDesign[dwIndex], iThreadID);	
		}
	}
	OSUnlock(hnidDesign);
	OSMemFree(hnidDesign);
	NIFCloseCollection(hcDesign);
	return TRUE;
}

//  BuildIndex
//
//  This function will (re)Build a single view index in the specified database. This variant is for
//  use by multi-threaded applications.
//
//  Parameters
//
//  1 - DBHANDLE		- Handle of the database where the view indexes will be rebuilt
//  2 - NOTEID			- NOTEID of the view index that is to be rebuilt
//  3 - int				- The ID of the thread from which the open request is being made
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::BuildIndex(DBHANDLE hdbIX, NOTEID nidView, int iThreadID)
{
	STATUS					stAPIRC;					//  Status code returned from Notes API calls
	HCOLLECTION				hcView;						//  Collection handle for the view
	NOTEHANDLE				hnView;						//  Handle to the open view note
	OID						oidDesign;					//	OID of the design note
	TIMEDATE				tdMod;						//  Modification date of the design note
	WORD					wNoteClass;					//  Note Class of the design note
	WORD					wRetLen;					//  Length of flags returned
	int						iIndex;						//  Index into the design flags
	char					szFlags[MAX_FLAGS + 1];		//  Design flags from the view note
	char					szMsg[MAX_MSG + 1];			//  Generic Message Buffer

	//  If the note is deleted then return
	if (nidView & RRV_DELETED) return;

	//  Check that the noteid passed refers to a view
	stAPIRC = NSFDbGetNoteInfo(hdbIX, nidView, &oidDesign, &tdMod, &wNoteClass);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0057E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return;
	}

	//  Check that the note is not a folder
	stAPIRC = NSFNoteOpen(hdbIX, nidView, 0, &hnView);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0058E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return;
	}
	wRetLen = NSFItemGetText(hnView, DESIGN_FLAGS, szFlags, sizeof(szFlags));
	if (wRetLen > 0)
	{
		for (iIndex = 0; iIndex <wRetLen; iIndex++)
		{
			if (szFlags[iIndex] == DESIGN_FLAG_FOLDER_VIEW)
			{
				NSFNoteClose(hnView);
				return;
			}
		}
	}
	NSFNoteClose(hnView);

	//  Open the collection for the specified view and force a build of the index
	stAPIRC = NIFOpenCollection(hdbIX, hdbIX, nidView, OPEN_REBUILD_INDEX, NULLHANDLE, &hcView, NULLHANDLE, NULL, NULLHANDLE, NULLHANDLE);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0032E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return;
	}
	NIFCloseCollection(hcView);
	return;
}

//  CreateRepository
//
//  This function will create a repository database on-demand, optionally a default pattern ACL will
//  be set in the new repository.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  BOOL				-  Switch that if TRUE indicates that the default ACL pattern should be set
//  4  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the repository if it could be created
//						NULLHANDLE if the repository could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateRepository(char *szServer, char *szDatabase, BOOL bSetACL, int iThreadID)
{
	//  This function is a wrapper for the CreateRepository() function
	return CreateRepository(szServer, szDatabase, DBCLASS_NOTEFILE, bSetACL, iThreadID);
}

//  CreateRepository
//
//  This function will create a repository database on-demand, optionally a default pattern ACL will
//  be set in the new repository. The DBCLASS of the repository is determined by the caller.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  WORD				-  The DBCLASS to be set for the repository
//  4  -  BOOL				-  Switch that if TRUE indicates that the default ACL pattern should be set
//  5  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the repository if it could be created
//						NULLHANDLE if the repository could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateRepository(char *szServer, char *szDatabase, WORD wDBClass, BOOL bSetACL, int iThreadID)
{
	//  This function is a wrapper for the CreateRepository() function
	return CreateRepository(szServer, szDatabase, wDBClass, FALSE, bSetACL, iThreadID);
}

//  CreateRepository
//
//  This function will create a repository database on-demand, optionally a default pattern ACL will
//  be set in the new repository. The DBCLASS of the repository is determined by the caller.
//  Transaction logging is enabled/disabled as requested by the caller.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  WORD				-  The DBCLASS to be set for the repository
//  4  -  BOOL				-  Switch that if TRUE will enable Transaction logging on the repository
//  5  -  BOOL				-  Switch that if TRUE indicates that the default ACL pattern should be set
//  6  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the repository if it could be created
//						NULLHANDLE if the repository could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateRepository(char *szServer, char *szDatabase, WORD wDBClass, BOOL bTXLog, BOOL bSetACL, int iThreadID)
{
	//  This function is a wrapper for the CreateRepository() function
	return CreateRepository(szServer, szDatabase, wDBClass, bTXLog, bSetACL, DBCREATE_ENCRYPT_NONE, iThreadID);
}

//  CreateRepository
//
//  This function will create a repository database on-demand, optionally a default pattern ACL will
//  be set in the new repository. The DBCLASS of the repository is determined by the caller.
//  Transaction logging is enabled/disabled as requested by the caller.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  WORD				-  The DBCLASS to be set for the repository
//  4  -  BOOL				-  Switch that if TRUE will enable Transaction logging on the repository
//  5  -  BOOL				-  Switch that if TRUE indicates that the default ACL pattern should be set
//	6  -  BYTE				-  Encryption strength for the repository
//  7  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the repository if it could be created
//						NULLHANDLE if the repository could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateRepository(char *szServer, char *szDatabase, WORD wDBClass, BOOL bTXLog, BOOL bSetACL, BYTE Encrypt, int iThreadID)
{
	STATUS					stAPIRC;						//  Status code returned from Notes API calls
	DBHANDLE				hdbNew;							//  DB Handle to the new Repository database
	WORD					wRetLen;						//  Returned String Lengths
	WORD					wCreateOptions;					//  Db creation option flags
	char					szCanonicalServer[MAXUSERNAME + 1];	//  Canonical Server Name
	char					szFullDbName[MAXPATH + 1];		//  Net Path to the database
	char					szAPIMsg[MAX_MSG + 1];			//  API message buffer

	//  Construct the db name to open
	szCanonicalServer[0] = '\0';
	TrimToUpper_s(szAPIMsg, MAX_MSG, szServer);
	if (strlen(szAPIMsg) == 0 || (strcmp(szAPIMsg, "LOCAL") == 0))
	{
		strcpy_s(szFullDbName, MAXPATH, szDatabase);
	}
	else
	{
		//  Canonicalize the server name and build the NET path for the database
		stAPIRC = DNCanonicalize(0L, NULL, szServer, szCanonicalServer, MAXUSERNAME, &wRetLen);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0012S, szServer);
			LogMessage(szAPIMsg, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			return NULLHANDLE;
		}

		stAPIRC = OSPathNetConstruct(NULL, szCanonicalServer, szDatabase, szFullDbName);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0013S, szCanonicalServer, szDatabase);
			LogMessage(szAPIMsg, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			return NULLHANDLE;
		}
	}

	//  Set the options for creating the repository
	if (bTXLog) wCreateOptions = DBCREATE_MAX_SPECIFIED;
	else wCreateOptions = DBCREATE_MAX_SPECIFIED + DBCREATE_DISABLE_TXN_LOGGING;
	
	//  Create the repository database
	stAPIRC = NSFDbCreateExtended(szFullDbName,
									wDBClass,
									FALSE,
									wCreateOptions,
									Encrypt,
									0);
	if (ERR(stAPIRC) != NOERROR)
	{
		if (ERR(stAPIRC) == ERR_EXISTS)
		{
			if (rsLocal->CreateOnDemand)
			{
				//  Not a failure - database is created if needed
				sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0050W, szDatabase, szServer);
				LogMessage(szAPIMsg, iThreadID);
				return OpenDatabase(szServer, szDatabase, iThreadID);
			}
			LogMessage(MSG_DXR0033E, iThreadID);
			return NULLHANDLE;
		}
		//  Create failed
		LogMessage(MSG_DXR0034E, iThreadID);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, iThreadID);
		return NULLHANDLE;
	}

	RepositoryCreated = TRUE;									//  Mark the repository created on this run

	if (rsLocal->LogLevel >= LOGLEVEL_VERBOSE)
	{
		sprintf_s(szAPIMsg, MAX_MSG, MSG_DXR0035I, szDatabase, szServer);
		LogMessage(szAPIMsg, iThreadID);
	}

	hdbNew = OpenDatabase(szServer, szDatabase, iThreadID);
	if (hdbNew == NULLHANDLE) return NULLHANDLE;

	//  Set Default ACL in the new database
	if (bSetACL)
	{
		if (!SetDefaultACL(hdbNew, szCanonicalServer, iThreadID))
		{
			LogMessage(MSG_DXR0036E, iThreadID);
			NSFDbClose(hdbNew);
			return NULLHANDLE;
		}
	}

	//  Close and re-open the repository database
	NSFDbClose(hdbNew);
	hdbNew = OpenDatabase(szServer, szDatabase, iThreadID);
	if (hdbNew == NULLHANDLE) return NULLHANDLE;

	return hdbNew;
}


//  CreateDatabase
//
//  This function will create a database on-demand.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateDatabase(char *szServer, char *szDatabase)
{
	//  This is the pulic function to create a database
	return CreateRepository(szServer, szDatabase, DBCLASS_NOTEFILE, FALSE, TRUE, 0);
}

//  CreateDatabase
//
//  This function will create a database on-demand. This variant is for multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateDatabase(char *szServer, char *szDatabase, int iThreadID)
{
	//  This is the pulic function to create a database
	return CreateRepository(szServer, szDatabase, DBCLASS_NOTEFILE, FALSE, TRUE, iThreadID);
}

//  CreateDatabase
//
//  This function will create a database on-demand. This variant is for multi-threaded applications.
//  Transaction logging can be enabled/disabled according to the callers request.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  BOOL				-  Switch that if TRUE will enable Transaction logging on the repository
//  4  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateDatabase(char *szServer, char *szDatabase, BOOL bTXLog, int iThreadID)
{
	//  This is the pulic function to create a database
	return CreateRepository(szServer, szDatabase, DBCLASS_NOTEFILE, bTXLog, TRUE, iThreadID);
}

//  CreateDatabase
//
//  This function will create a database on-demand. This variant is for multi-threaded applications.
//  Transaction logging can be enabled/disabled according to the callers request.
//  The DBID of the database is returned.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  BOOL				-  Switch that if TRUE will enable Transaction logging on the repository
//  4  -  DBID *			-  Pointer to the DBID structure that will be completed
//  5  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateDatabase(char *szServer, char *szDatabase, BOOL bTXLog, DBID *dbidDB, int iThreadID)
{
	STATUS			stAPIRC;									//  API return code
	DBHANDLE		hdbTemp;									//  Handle of D/B or Directory
	char			szAPIMsg[MAX_MSG + 1];						//  API Error message

	//  This is the pulic function to create a database
	hdbTemp = CreateRepository(szServer, szDatabase, DBCLASS_NOTEFILE, bTXLog, TRUE, iThreadID);
	if (hdbTemp != NULLHANDLE)
	{
		//  Get the DBID of the Database (or Directory)
		stAPIRC = NSFDbIDGet(hdbTemp, dbidDB);
		if (ERR(stAPIRC) != NOERROR)
		{
			NSFDbClose(hdbTemp);								//  Close the Database
			hdbTemp = NULLHANDLE;
			LogMessage(MSG_DXR0052E, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			return hdbTemp;
		}	
	}
	return hdbTemp;
}

//  CreateDatabase
//
//  This function will create a database on-demand. This variant is for multi-threaded applications.
//  Transaction logging can be enabled/disabled according to the callers request.
//  The DBID of the database is returned.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  BOOL				-  Switch that if TRUE will enable Transaction logging on the repository
//  4  -  DBID *			-  Pointer to the DBID structure that will be completed
//  5  -  WORD				-  DBCLASS_ class for the database to be created
//  6  -  BYTE				-  Local encryption strength for the database
//  7  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateDatabase(char *szServer, char *szDatabase, BOOL bTXLog, DBID *dbidDB, WORD wDbClass, BYTE Encrypt, int iThreadID)
{
	STATUS			stAPIRC;									//  API return code
	DBHANDLE		hdbTemp;									//  Handle of D/B or Directory
	char			szAPIMsg[MAX_MSG + 1];						//  API Error message

	//  This is the pulic function to create a database
	hdbTemp = CreateRepository(szServer, szDatabase, wDbClass, bTXLog, TRUE, Encrypt, iThreadID);
	if (hdbTemp != NULLHANDLE)
	{
		//  Get the DBID of the Database (or Directory)
		stAPIRC = NSFDbIDGet(hdbTemp, dbidDB);
		if (ERR(stAPIRC) != NOERROR)
		{
			NSFDbClose(hdbTemp);								//  Close the Database
			hdbTemp = NULLHANDLE;
			LogMessage(MSG_DXR0052E, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			return hdbTemp;
		}	
	}
	return hdbTemp;
}


//  CreateReplica
//
//  This function will create a new Replica database of the passed database handle.
//  The DBID of the database is returned.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  DBHANDLE			-  Handle of the database that is to be replicated
//  4  -  DBID *			-  Pointer to the DBID structure that will be completed
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateReplica(char *szServer, char *szDatabase, DBHANDLE hdbSrc, DBID *dbidDB)
{
	STATUS					stAPIRC;						//  Status code returned from Notes API calls
	WORD					wDBClass;						//  Class of the source database
	char					szAPIMsg[MAX_MSG + 1];			//  API message buffer

	//  Get the D/B Class of the source database
	stAPIRC = NSFDbClassGet(hdbSrc, &wDBClass);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0056E, 0);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, 0);
		return NULLHANDLE;
	}

	//  This function is a wrapper for the CoreCreateReplica() function
	return CoreCreateReplica(szServer, szDatabase, FALSE, hdbSrc, dbidDB, wDBClass, DBCREATE_ENCRYPT_NONE, 0);
}

//  CreateReplica
//
//  This function will create a new Replica database of the passed database handle.
//  The DBID of the database is returned. This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  DBHANDLE			-  Handle of the database that is to be replicated
//  4  -  DBID *			-  Pointer to the DBID structure that will be completed
//  5  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateReplica(char *szServer, char *szDatabase, DBHANDLE hdbSrc, DBID *dbidDB, int iThreadID)
{
	STATUS					stAPIRC;						//  Status code returned from Notes API calls
	WORD					wDBClass;						//  Class of the source database
	char					szAPIMsg[MAX_MSG + 1];			//  API message buffer

	//  Get the D/B Class of the source database
	stAPIRC = NSFDbClassGet(hdbSrc, &wDBClass);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0056E, iThreadID);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, iThreadID);
		return NULLHANDLE;
	}

	//  This function is a wrapper for the CoreCreateReplica() function
	return CoreCreateReplica(szServer, szDatabase, FALSE, hdbSrc, dbidDB, wDBClass, DBCREATE_ENCRYPT_NONE, iThreadID);
}

//  CreateReplica
//
//  This function will create a new Replica database of the passed database handle.
//  The DBID of the database is returned. Transaction logging can be enabled/disabled.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  BOOL				-  Switch that if TRUE will enable Transaction logging on the repository
//  4  -  DBHANDLE			-  Handle of the database that is to be replicated
//  5  -  DBID *			-  Pointer to the DBID structure that will be completed
//  6  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateReplica(char *szServer, char *szDatabase, BOOL bTXLog, DBHANDLE hdbSrc, DBID *dbidDB, int iThreadID)
{
	STATUS					stAPIRC;						//  Status code returned from Notes API calls
	WORD					wDBClass;						//  Class of the source database
	char					szAPIMsg[MAX_MSG + 1];			//  API message buffer

	//  Get the D/B Class of the source database
	stAPIRC = NSFDbClassGet(hdbSrc, &wDBClass);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0056E, iThreadID);
		GetAPIMessage(stAPIRC, szAPIMsg);
		LogMessage(szAPIMsg, iThreadID);
		return NULLHANDLE;
	}

	//  This function is a wrapper for the CoreCreateReplica() function
	return CoreCreateReplica(szServer, szDatabase, bTXLog, hdbSrc, dbidDB, wDBClass, DBCREATE_ENCRYPT_NONE, iThreadID);
}

//  CreateReplica
//
//  This function will create a new Replica database of the passed database handle.
//  The DBID of the database is returned. Transaction logging can be enabled/disabled.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  BOOL				-  Switch that if TRUE will enable Transaction logging on the repository
//  4  -  DBHANDLE			-  Handle of the database that is to be replicated
//  5  -  DBID *			-  Pointer to the DBID structure that will be completed
//	6  -  WORD				-  Database class of the replica to be created
//	7  -  BYTE				-  Local encryption strength for the database to be created
//  8  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CreateReplica(char *szServer, char *szDatabase, BOOL bTXLog, DBHANDLE hdbSrc, DBID *dbidDB, WORD wDbClass, BYTE Encrypt, int iThreadID)
{
	//  This function is a wrapper for the CoreCreateReplica() function
	return CoreCreateReplica(szServer, szDatabase, bTXLog, hdbSrc, dbidDB, wDbClass, Encrypt, iThreadID);
}

//  CoreCreateReplica
//
//  This function will create a new Replica database of the passed database handle.
//  The DBID of the database is returned. Transaction logging can be enabled/disabled.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  char *			-  The server name where the repository will reside (or "Local")
//  2  -  char *			-  The path to the repository relative to the Notes Data Directory
//  3  -  BOOL				-  Switch that if TRUE will enable Transaction logging on the repository
//  4  -  DBHANDLE			-  Handle of the database that is to be replicated
//  5  -  DBID *			-  Pointer to the DBID structure that will be completed
//	6  -  WORD				-  Database class of the replica to be created
//	7  -  BYTE				-  Local encryption strength for the database to be created
//  8  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  DBHANDLE  -			Handle to the database if it could be created
//						NULLHANDLE if the database could not be created
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::CoreCreateReplica(char *szServer, char *szDatabase, BOOL bTXLog, DBHANDLE hdbSrc, DBID *dbidDB, WORD wDbClass, BYTE Encrypt, int iThreadID)
{
	STATUS			stAPIRC;									//  API return code
	DBHANDLE		hdbTemp;									//  Handle of D/B or Directory
    DBREPLICAINFO   riReplicaSettings;							//  Replication Information 
	char			szAPIMsg[MAX_MSG + 1];						//  API Error message

	//  This variant of CreateReplica also returns the DBID of the opened database
	//  Create the database
	hdbTemp = CreateRepository(szServer, szDatabase, wDbClass, bTXLog, FALSE, Encrypt, iThreadID);
	if (hdbTemp != NULLHANDLE)
	{
		//  Set the ReplicaID for the database
		stAPIRC = NSFDbReplicaInfoGet(hdbSrc, &riReplicaSettings);
		if (ERR(stAPIRC) != NOERROR)
		{
			NSFDbClose(hdbTemp);								//  Close the Database
			hdbTemp = NULLHANDLE;
			LogMessage(MSG_DXR0053E, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			return hdbTemp;
		}	

		//  Disable replication in the target database
		riReplicaSettings.Flags = riReplicaSettings.Flags | REPLFLG_DISABLE;

		stAPIRC = NSFDbReplicaInfoSet(hdbTemp, &riReplicaSettings);
		if (ERR(stAPIRC) != NOERROR)
		{
			NSFDbClose(hdbTemp);								//  Close the Database
			hdbTemp = NULLHANDLE;
			LogMessage(MSG_DXR0054E, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			return hdbTemp;
		}

		//  Copy the ACL from the source database to the target
		stAPIRC = NSFDbCopyACL(hdbSrc, hdbTemp);
		if (ERR(stAPIRC) != NOERROR)
		{
			NSFDbClose(hdbTemp);								//  Close the Database
			hdbTemp = NULLHANDLE;
			LogMessage(MSG_DXR0055E, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			return hdbTemp;
		}

		//  Get the DBID of the Database (or Directory)
		stAPIRC = NSFDbIDGet(hdbTemp, dbidDB);
		if (ERR(stAPIRC) != NOERROR)
		{
			NSFDbClose(hdbTemp);								//  Close the Database
			hdbTemp = NULLHANDLE;
			LogMessage(MSG_DXR0052E, iThreadID);
			GetAPIMessage(stAPIRC, szAPIMsg);
			LogMessage(szAPIMsg, iThreadID);
			return hdbTemp;
		}	
	}
	return hdbTemp;
}

//  SetSpecialNoteID
//
//  This function will expose the undocumented NSFDbSetSpecialNoteID() function.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database where the special Note ID is to be set
//  2  -  WORD				-  The Special ID that is to be set
//  3  -  NOTEID			-  ID of the Note that is to be assigned the Special ID
//  4  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  STATUS  -			Notes API Return Code
//						NOERROR if the function completed
//
//  API Use Notes:
//
//  NOTE:  This function uses undocumented and unsupported notes API calls
//  NOTE:  Use this function with great care
//
STATUS ExecEnvironment::SetSpecialNoteID(DBHANDLE hdbTarget, WORD wSpecialID, NOTEID nidReal, int iThreadID)
{
	typedef STATUS LNPUBLIC NSFPROC(DBHANDLE, WORD, NOTEID);	//  Typedef for the undocumented entry point

	STATUS			stAPIRC;									//  API return code
	HMODULE			hmNotesEPS;									//  Handle to the notes API library
#ifdef UNIX
	void			*hDLL;										//  DLL Handle
#endif
	NSFPROC			*lpNotesEntry;								//  Entry point for undocumented function
	STATUS (LNCALLBACKPTR ProcAddress)(VOID *);					//  Forward definition for the entry point
	char			szFullLibName[MAXPATH + 1];					//  Full name of the notes API DLL
	char			szMsg[MAX_MSG + 1];							//  Generic Error buffer

	//  Load the library containing the needed entry point
	OSGetExecutableDirectory(szFullLibName);
	strcat_s(szFullLibName, MAXPATH, PLATFORM_APILIB_NAME);
	stAPIRC = OSLoadLibrary(szFullLibName, (DWORD) 0, &hmNotesEPS, &ProcAddress);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0059E, PLATFORM_APILIB_NAME);
		LogMessage(szMsg, iThreadID);
		return(stAPIRC);
	}

#ifdef UNIX
	//  Reload the library natively to obtain a native handle
	hDLL = dlopen(szFullLibName, RTLD_LAZY);
	if (hDLL == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0059E, PLATFORM_APILIB_NAME);
		LogMessage(szMsg, iThreadID);
		OSFreeLibrary(hmNotesEPS);
		return(ERR_LIBLOAD2);
	}
	lpNotesEntry = (NSFPROC *) dlsym(hDLL, NSF_SN_EPNAME);
#endif

#ifdef WIN32
	//  Get the address of the entry point for NSFDbSetSpecialNoteID
	//  NOTE: The explicit casting of the HMODULE back to HINSTANCE is needed because the Notes API
	//  redefines HMODULE as a generic HANDLE.
	lpNotesEntry = (NSFPROC *) GetProcAddress((HINSTANCE) hmNotesEPS, NSF_SN_EPNAME);
#endif

	if (lpNotesEntry == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0060E, NSF_SN_EPNAME, PLATFORM_APILIB_NAME);
		LogMessage(szMsg, iThreadID);
		OSFreeLibrary(hmNotesEPS);
#ifdef UNIX
		dlclose(hDLL);
#endif
		return(ERR_NOT_FOUND);
	}

	//  Call the undocumented entry point
	stAPIRC = (lpNotesEntry) (hdbTarget, wSpecialID, nidReal);
#ifdef UNIX
	dlclose(hDLL);
#endif
	OSFreeLibrary(hmNotesEPS);
	return stAPIRC;
}

//  SetDefaultACL
//
//  This function will set a default pattern ACL in the passed database.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database that is to have a default ACL set
//  2  -  char *			-  The server name where the database resides
//  3  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  BOOL	-			TRUE if the ACL was set
//						FALSE if the ACL could not be set
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::SetDefaultACL(DBHANDLE hdbNew, char *szServer, int iThreadID)
{
	STATUS					stAPIRC;							//  Status code returned from Notes API calls
	DHANDLE					hACL;								//  Handle to the ACL
	ACL_PRIVILEGES			privsNull = {0,0,0,0,0,0,0,0,0,0};	//  Default privileges
	char					szMsg[MAX_MSG + 1];					//  Generic message buffer

	//  Set the default ACL in a new database
	stAPIRC = ACLCreate(&hACL);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0037E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	//  Add The current user as Manager - note for a fresh ACL this should be already done
	if (RunningOnServer)
	{
		stAPIRC = ACLAddEntry(hACL, CanonicalUser, ACL_LEVEL_MANAGER, &privsNull, ACL_FLAG_SERVER|ACL_FLAG_ADMIN_SERVER);
	}
	else
	{
		stAPIRC = ACLAddEntry(hACL, CanonicalUser, ACL_LEVEL_MANAGER, &privsNull, ACL_FLAG_PERSON);
	}
	if (ERR(stAPIRC) == ERR_ACL_INLIST)
	{
		//  Entry already in the list - update it
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0051T, CanonicalUser);
		LogTrace(TRACE_CORERT, szMsg, iThreadID);
		stAPIRC = ACLUpdateEntry(hACL, CanonicalUser, ACL_UPDATE_LEVEL|ACL_UPDATE_FLAGS, 
								CanonicalUser, ACL_LEVEL_MANAGER,&privsNull, ACL_FLAG_PERSON);
	}
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0038E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	//  Set Default and Anonymous with No Access
	stAPIRC = ACLUpdateEntry(hACL, NULL, ACL_UPDATE_LEVEL|ACL_UPDATE_FLAGS, 
							NULL, ACL_LEVEL_NOACCESS, &privsNull, ACL_FLAG_PUBLICREADER|ACL_FLAG_PUBLICWRITER|ACL_FLAG_NOREPLICATE);
	if ((ERR(stAPIRC) != NOERROR) && (ERR(stAPIRC) != ERR_ACL_INLIST))
	{
		LogMessage(MSG_DXR0039W, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
	}
	stAPIRC = ACLAddEntry(hACL, ANONYMOUS_ACL_USER, ACL_LEVEL_NOACCESS, 
						&privsNull, ACL_FLAG_PERSON|ACL_FLAG_PUBLICREADER|ACL_FLAG_PUBLICWRITER|ACL_FLAG_NOREPLICATE);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0040W, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
	}

	//  If the database is on a server - add the server as the Admin Server if it is dfferent form the current user
	if (szServer[0] != '\0')
	{
		if (strcmpi(szServer, CanonicalUser) != 0)
		{
			stAPIRC = ACLAddEntry(hACL, szServer, ACL_LEVEL_MANAGER, &privsNull, ACL_FLAG_SERVER|ACL_FLAG_ADMIN_SERVER);
			if (ERR(stAPIRC) != NOERROR)
			{
				LogMessage(MSG_DXR0041W, iThreadID);
				GetAPIMessage(stAPIRC, szMsg);
				LogMessage(szMsg, iThreadID);
			}
		}
	}

	//  Set the LocalDomainServers & LocalDomainAdmins
	stAPIRC = ACLAddEntry(hACL, LDS_ACL_GROUP, ACL_LEVEL_MANAGER, &privsNull, ACL_FLAG_SERVER|ACL_FLAG_GROUP);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0042W, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
	}
	stAPIRC = ACLAddEntry(hACL, LDA_ACL_GROUP, ACL_LEVEL_MANAGER, &privsNull, ACL_FLAG_PERSON|ACL_FLAG_GROUP);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0043W, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
	}

	//  Update ther ACL
	stAPIRC = NSFDbStoreACL(hdbNew, hACL, 0, 1);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0044E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return FALSE;
	}
	return TRUE;
}

//  GetBuildInfo
//
//  This function will return the current build information for this Runtime.
//
//  Parameters
//
//  2  -  char *			-  Pointer to a string buffer that will receive the 
//  3  -  int				-  The size of the buffer where the build info will be returned
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::GetBuildInfo_s(char * szBuild, int iSize)
{
	sprintf_s(szBuild, iSize, RT_IDENTITY, RT_BUILD);
	return;
}

//  GetDbTitle
//
//  This function will set the D/B Title in the passed database.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database that is to have the title set
//  2  -  char *			-  The string buffer that will be filled with the title
//  3  -  int				-  Identity of the thread making the call
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::GetDbTitle(DBHANDLE hDB, char *szTitle, int iThreadID)
{
	STATUS				stAPIRC;						//  Return code from Notes API Calls
	char				szInfoBuffer[NSF_INFO_SIZE + 1];	//  Database Info Buffer
	char				szMsg[MAX_MSG + 1];				//  Generic message buffer

	//  Set the title of the specified database
	stAPIRC = NSFDbInfoGet(hDB, szInfoBuffer);			//  Get the current Info Buffer
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0045E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return;
	}

	//  Get the title in the Info Buffer
	NSFDbInfoParse(szInfoBuffer, INFOPARSE_TITLE, szTitle, NSF_INFO_SIZE - 1);
	return;
}

//  SetDbTitle
//
//  This function will set the D/B Title in the passed database.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database that is to have the title set
//  2  -  char *			-  The string that will be set as the title
//
//  Returns
//
//  BOOL	-			TRUE if the Title was set
//						FALSE if the Title could not be set
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::SetDbTitle(DBHANDLE hDB, char * szTitle)
{
	STATUS				stAPIRC;						//  Return code from Notes API Calls
	NOTEHANDLE			hnDBICON;						//  Handle to Db Icon note
	char				szInfoBuffer[NSF_INFO_SIZE + 1];	//  Database Info Buffer
	char				szMsg[MAX_MSG + 1];				//  Generic message buffer

	//  Set the title of the specified database
	stAPIRC = NSFDbInfoGet(hDB, szInfoBuffer);			//  Get the current Info Buffer
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0045E);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg);
		return FALSE;
	}

	//  Modify the title in the Info Buffer
	NSFDbInfoModify(szInfoBuffer, INFOPARSE_TITLE, szTitle);

	//  Update the D/B info buffer in the database
	stAPIRC = NSFDbInfoSet(hDB, szInfoBuffer);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0046E);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg);
		return FALSE;
	}

	//  Now update the title in the DbIcon note
	stAPIRC = NSFNoteOpen(hDB, NOTE_ID_SPECIAL+NOTE_CLASS_ICON, 0, &hnDBICON);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0047E);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg);
		return FALSE;
	}
	stAPIRC = NSFItemSetText(hnDBICON, FIELD_TITLE, szTitle, MAXWORD);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0048E);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg);
		return FALSE;
	}
	stAPIRC = NSFNoteUpdate(hnDBICON, UPDATE_FORCE);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0049E);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg);
		return FALSE;
	}
	NSFNoteClose(hnDBICON);
	return TRUE;
}

//  SetRepositoryTitle
//
//  This function will set the D/B Title in the current repository.
//
//  Parameters
//
//  1  -  char *			-  The string that will be set as the title
//
//  Returns
//
//  BOOL	-			TRUE if the Title was set
//						FALSE if the Title could not be set
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::SetRepositoryTitle(char * szTitle)
{
	//  Set the title of the current repository
	return SetDbTitle(hdbRepository, szTitle);
}

//  AppWait
//
//  This function will suspend the application main thread for the number of MilliSeconds specified.
//  The function will continue to monitor for log messages posted by other application threads for
//  the duration of the wait.
//
//  Parameters
//
//  1  -  int			-  The number of MilliSeconds that the application is to wait
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::AppWait(int iWaitMS)
{
	int					iWaitCycles;							//  Number of cycles to wait for
	int					iIndex;									//  Loop index counter
	clock_t				ct_Start;								//  Wait Start Timestamp
	clock_t				ct_Current;								//  Wait current timestamp
	long				lElapsed;								//  Elapsed Time
	long				lTimeFactor = CLOCKS_PER_SEC;			//  Clock tick conversion factor

	lTimeFactor = lTimeFactor / 1000;							//  Factor for milliseconds

	//  Bale out if the wait interval is zero
	if (iWaitMS <= 0) return;

	//  If the total wait is < 50 Ms then do this in a single wait interval
	if (iWaitMS < SINGLE_WAIT_QUANTUM)
	{
#ifdef WIN32
		if (rsLocal->RunningAsAddin)
		{
			//  Do not use AddInIdleDelay if there is a quit pending
			if (MQIsQuitPending(hmqProcessor)) SLEEP(iWaitMS);
			else AddInIdleDelay(iWaitMS);
		}
		else SLEEP(iWaitMS);
#endif
#ifdef UNIX
		SLEEP(iWaitMS);
#endif
		return;
	}

	//  Record the start time of the wait interval
	ct_Start = RunningTime->getElapsed();

	//  Calculate the number of wait cycles
	iWaitCycles = (iWaitMS / PSEUDO_WAIT_CHUNK) + 1;				//  Round up to the next chunk interval
	
	//  Perform the wait loop
	for (iIndex = 0; iIndex < iWaitCycles; iIndex++)
	{
#ifdef WIN32
		if (rsLocal->RunningAsAddin)
		{
			//  Do not use AddInIdleDelay if there is a quit pending
			if (MQIsQuitPending(hmqProcessor)) SLEEP(iWaitMS);
			else AddInIdleDelay(iWaitMS);
		}
		else SLEEP(iWaitMS);
#endif
#ifdef UNIX
		SLEEP(iWaitMS);
#endif
		//  Check if the wait time interval has expired
		ct_Current = RunningTime->getElapsed();
		lElapsed = (ct_Current - ct_Start) / lTimeFactor;
		if (lElapsed >= iWaitMS) return;
	}
	return;
}

//  LogTrace
//
//  This function will log the passed message only if the application is being run in Trace or Debug mode
//  and if a focus area was specified that the message belongs to the specified focus area.
//
//  Parameters
//
//  1  -  int				-  The ID of the focus area that this trace message belongs to
//  2  -  char *			-  Pointer to the string buffer containing the message to be written
//
//  Returns
//
//  BOOL					-  TRUE if the log message was logged or ignored
//							   FALSE if the message could not be logged
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::LogTrace(int iTraceArea, char *szMsg)
{
	char		szTMsg[MAX_MSG + 11];					//  Formatted Trace Message

	//  Detrmine if trace level logging is enabled
	if ((rsLocal->LogLevel != LOGLEVEL_TRACE) && (rsLocal->LogLevel != LOGLEVEL_DEBUG)) return TRUE;

	//  Determine if the source trace area is in the current tracing scope
	if ((rsLocal->TraceArea != iTraceArea) && (rsLocal->TraceArea != 0)) return TRUE;

	//  Format the trace message
	sprintf_s(szTMsg, MAX_MSG, TRACE_FORMATTER, szMsg, iTraceArea);
	return LogMessage(szTMsg);
}

//  LogTrace
//
//  This function will log the passed message only if the application is being run in Trace or Debug mode
//  and if a focus area was specified that the message belongs to the specified focus area.
//  This variant is for use by multi-threaded applications.
//
//  Parameters
//
//  1  -  int				-  The ID of the focus area that this trace message belongs to
//  2  -  char *			-  Pointer to the string buffer containing the message to be written
//  3  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  BOOL					-  TRUE if the log message was logged or ignored
//							   FALSE if the message could not be logged
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::LogTrace(int iTraceArea, char *szMsg, int iThreadID)
{
	char		szTMsg[MAX_MSG + 11];					//  Formatted Trace Message

	//  Detrmine if trace level logging is enabled
	if ((rsLocal->LogLevel != LOGLEVEL_TRACE) && (rsLocal->LogLevel != LOGLEVEL_DEBUG)) return TRUE;

	//  Determine if the source trace area is in the current tracing scope
	if ((rsLocal->TraceArea != iTraceArea) && (rsLocal->TraceArea != 0)) return TRUE;

	//  Format the trace message
	sprintf_s(szTMsg, MAX_MSG, TRACE_FORMATTER, szMsg, iTraceArea);
	return LogMessage(szTMsg, iThreadID);
}

//  LogVerboses
//
//  This function will log the passed message only if the application is being run in Verbose or higher mode.
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string buffer containing the message to be written
//
//  Returns
//
//  BOOL					-  TRUE if the log message was logged or ignored
//							   FALSE if the message could not be logged
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::LogVerbose(char *szMsg)
{
	//  Do nothing if the logging level is less than verbose
	if (rsLocal->LogLevel < LOGLEVEL_VERBOSE) return TRUE;
	return LogMessage(szMsg);
}

//  LogVerboses
//
//  This function will log the passed message only if the application is being run in Verbose or higher mode.
//  This variant is for use by multi-threaded applications.
//  
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string buffer containing the message to be written
//  2  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  BOOL					-  TRUE if the log message was logged or ignored
//							   FALSE if the message could not be logged
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::LogVerbose(char *szMsg, int iThreadID)
{
	//  Do nothing if the logging level is less than verbose
	if (rsLocal->LogLevel < LOGLEVEL_VERBOSE) return TRUE;
	return LogMessage(szMsg, iThreadID);
}

//  TrimToUpper
//
//  This function will trim the passed string and fold it to upper case.
//  
//
//  Parameters
//
//  1  -  char *			-  Pointer to the string buffer containing the source string
//  2  -  int				-  Size of the output buffer
//  3  -  char *			-  Pointer to the string buffer where the resulting string will be placed
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::TrimToUpper_s(char * szDest, int iDest_Size, char * szSrc)
{
	int		iIndex;
	int		iDestIndex = 0;
	char	cValue;

	//  Trim and fold to uppercase

	for (iIndex = 0; iIndex < (int) strlen(szSrc); iIndex++)
	{
		cValue = szSrc[iIndex];
		if (cValue != ' ')
		{
			szDest[iDestIndex] = toupper(cValue);
			iDestIndex++;
		}
		if ((cValue == '\0') || (iDestIndex == iDest_Size))
		{
			szDest[iDestIndex] = '\0';
			return;
		}
	}
	
	szDest[iDestIndex] = '\0';
}

//  DrainMQ
//
//  This function will attempt to drain the MQ message queue
//  
//
//  Parameters
//  1  -  char *			-  Pointer to the string containg the message queue name
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::DrainMQ(char *szQName)
{
	STATUS			stAPIRC;									//  Return call from Notes API calls
	char			szMsg[MAX_MSG + 1];							//  Generic Message Buffer

	//  Open the application queue name
	stAPIRC = MQOpen(szQName, 0, &hmqProcessor);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0026E, szQName);
		LogMessage(szMsg, 0);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, 0);
		return;
	}
	LogTrace(TRACE_CORERT, MSG_DXR0066T, 0);

	//  Test for Quit
	if (MQIsQuitPending(hmqProcessor))
	{
		LogMessage(MSG_DXR0063I, 0); 
	}

	MQClose(hmqProcessor, 0);	
	return;
}

//  GetIconFlags
//
//  This function will return the flags set in the DB Icon note of the passed database
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database 
//  2  -  char *			-  The string buffer where the flags will be returned
//  3  -  int				-  Length of the string buffer
//
//  Returns
//
//  STATUS	-			NOERROR if the Flags were returned
//						        Other status codes returned by Notes
//
//  API Use Notes:
//
//
STATUS ExecEnvironment::GetIconFlags(DBHANDLE hDB, char *szFlags, int iMaxLen)
{
	return GetIconFlags(hDB, szFlags, iMaxLen, 0);
}

//  GetIconFlags
//
//  This function will return the flags set in the DB Icon note of the passed database
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database 
//  2  -  char *			-  The string buffer where the flags will be returned
//  3  -  int				-  Length of the string buffer
//  4  -  int				-  Thread ID of the caller
//
//  Returns
//
//  STATUS	-			NOERROR if the Flags were returned
//						        Other status codes returned by Notes
//
//  API Use Notes:
//
//
STATUS ExecEnvironment::GetIconFlags(DBHANDLE hDB, char *szFlags, int iMaxLen, int iThreadID)
{
	STATUS				stAPIRC;						//  Return code from Notes API Calls
	NOTEHANDLE			hnDBICON;						//  Handle to Db Icon note
	WORD				wRetLen;						//  Returned length of the flags
	char				szMsg[MAX_MSG + 1];				//  Generic message buffer

	//  Now update the title in the DbIcon note
	stAPIRC = NSFNoteOpen(hDB, NOTE_ID_SPECIAL+NOTE_CLASS_ICON, 0, &hnDBICON);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0047E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return stAPIRC;
	}

	//  Get the Flag Values
	wRetLen = NSFItemGetText(hnDBICON, ICON_FLAGS, szFlags, (WORD) iMaxLen);
	szFlags[wRetLen] = '\0';

	NSFNoteClose(hnDBICON);
	return NOERROR;
}

//  IsIconFlagSet
//
//  This function will indicate if the passed flag value is set in the DB Icon flags of the passed database
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database 
//  2  -  char				-  The value of the flag that is to be tested for
//
//  Returns
//
//  BOOL	-				TRUE if the Flag value is set
//						    FALSE if the Flag value is not set
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::IsIconFlagSet(DBHANDLE hDB, char cFlag)
{
	return IsIconFlagSet(hDB, cFlag, 0);
}

//  IsIconFlagSet
//
//  This function will indicate if the passed flag value is set in the DB Icon flags of the passed database
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database 
//  2  -  char				-  The value of the flag that is to be tested for
//  3  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  BOOL	-				TRUE if the Flag value is set
//						    FALSE if the Flag value is not set
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::IsIconFlagSet(DBHANDLE hDB, char cFlag, int iThreadID)
{
	STATUS				stAPIRC;						//  Return code from Notes API Calls
	char				*pFlag;							//  Pointer into the Flags
	char				szFlags[MAX_DESC + 1];			//  Flags

	stAPIRC = GetIconFlags(hDB, szFlags, MAX_DESC);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0064E, iThreadID);
		return FALSE;
	}

	pFlag = szFlags;
	
	//  Test for the requested flag setting
	while (*pFlag != '\0')
	{
		if (*pFlag == cFlag) return TRUE;
		pFlag++;
	}
	return FALSE;
}

//  CoreSetLogLevel
//
//  This function will set the current logging level according to the settings passed in the call
//
//  Parameters
//
//  1  -  int				-  The logging level to be set
//  2  -  int				-  The trace area to be set for Trace & Debug logging levels
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::CoreSetLogLevel(int iLevel, int iTraceArea)
{
	//  Validate the level
	if ((iLevel < LOGLEVEL_TERSE) || (iLevel > LOGLEVEL_DEBUG)) return;

	//  Set the level and the trace area
	rsLocal->TraceArea = iTraceArea;
	rsLocal->LogLevel = iLevel;
	return;
}

//  LogSetNormal
//
//  This function will set the current logging level to Normal
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogSetNormal(void)
{
	CoreSetLogLevel(LOGLEVEL_NORMAL, TRACE_ALL);
	return;
}

//  LogSetVerbose
//
//  This function will set the current logging level to Verbose
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogSetVerbose(void)
{
	CoreSetLogLevel(LOGLEVEL_VERBOSE, TRACE_ALL);
	return;
}

//  LogSetTrace
//
//  This function will set the current logging level to Trace
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogSetTrace(int iTraceArea)
{
	CoreSetLogLevel(LOGLEVEL_TRACE, iTraceArea);
	return;
}

//  LogSetDebug
//
//  This function will set the current logging level to Debug
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogSetDebug(int iTraceArea)
{
	CoreSetLogLevel(LOGLEVEL_DEBUG, iTraceArea);
	return;
}

//  LogGetLevel
//
//  This function will return the current logging level
//
//  Parameters
//
//  Returns
//
//  int - the current logging level
//
//  API Use Notes:
//
//
int ExecEnvironment::LogGetLevel(void)
{
	return rsLocal->LogLevel;
}

//  LogPushSettings
//
//  This function will save the current logging settings
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogPushSettings(void)
{
	SavedLogLevel = rsLocal->LogLevel;
	SavedTraceArea = rsLocal->TraceArea;
	return;
}

//  LogPopSettings
//
//  This function will restore the saved logging settings
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogPopSettings(void)
{
	rsLocal->LogLevel = SavedLogLevel;
	rsLocal->TraceArea = SavedTraceArea;
	return;
}

//  CoreSetLogEcho
//
//  This function will set the current logging echo state
//
//  Parameters
//
//  1  -  BOOL				-  The logging echo setting
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::CoreSetLogEcho(BOOL bEchoState)
{
	if (bEchoState) rsLocal->EchoLog = TRUE;
	else rsLocal->EchoLog = FALSE;
	return;
}

//  LogIsEchoing
//
//  This function will return the current logging echo state
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE if the log is being echoed, FALSE if not
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::LogIsEchoing()
{
	return rsLocal->EchoLog;
}

//  LogSetEchoOn
//
//  This function will set the current logging echo state to on
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogSetEchoOn(void)
{
	CoreSetLogEcho(TRUE);
	return;
}

//  LogSetEchoOff
//
//  This function will set the current logging echo state to off
//
//  Parameters
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogSetEchoOff(void)
{
	CoreSetLogEcho(FALSE);
	return;
}

//  GetDataDocumentCount
//
//  This function will return the count of data documents in the database
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database 
//  2  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  DWORD	-				-1 is the count could not be determined, otherwise the count
//
//  API Use Notes:
//
//  NOTE: This call uses undocumented Notes API calls, use with care.
//
DWORD ExecEnvironment::GetDataDocumentCount(DBHANDLE hDB, int iThreadID)
{
	STATUS				stAPIRC;									//  Return code from Notes API Calls
	DHANDLE				hitData;									//  Handle of the Modified Note Table
	INFOBLK				ibDB;										//  Infoblk structure
	BUILDVERSION		bvServer;									//  Server Build Version
	TIMEDATE			tdSince;									//  Start of modified time window
	TIMEDATE			tdUntil;									//  End of modified time window
	char				szMsg[MAX_MSG + 1];							//  Generic message buffer
	
	//  For servers below R6 use the traditional method for obtaining the count
	stAPIRC = NSFDbGetMajMinVersion(hDB, &bvServer);
	if (ERR(stAPIRC) != NOERROR) bvServer.MajorVersion = 6;			//  Default to R6 if we did not get a version		
	if (bvServer.MajorVersion < 6)
	{
		ibDB.dwDataDocs = -1;
		TimeConstant(TIMEDATE_WILDCARD, &tdSince);
		stAPIRC = NSFDbGetModifiedNoteTable(hDB, NOTE_CLASS_DATA, tdSince, &tdUntil, &hitData);
		if (ERR(stAPIRC) != NOERROR)
		{
			if (ERR(stAPIRC) == ERR_NO_MODIFIED_NOTES)
			{
				ibDB.dwDataDocs = 0;
			}
			else
			{
				LogMessage(MSG_DXR0140E, iThreadID);
				GetAPIMessage(stAPIRC, szMsg);
				LogMessage(szMsg, iThreadID);
			}
			return ibDB.dwDataDocs;
		}
		ibDB.dwDataDocs = IDEntries(hitData);
		IDDestroyTable(hitData);
		return ibDB.dwDataDocs;
	}

	//  Fill the infoblk for the requested database
	stAPIRC = GetDbInfoBlock(hDB, &ibDB, iThreadID);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0073E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return -1;
	}

	//  Return the count
	return ibDB.dwDataDocs;
}

//  GetDbInfoBlock
//
//  This function will return the information block for the specified database
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database 
//  2  -  INFOBLK *			-  Pointer to an INFOBLK structure that will be filled by the call
//  3  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  STATUS	-				Notes API Status - NOEEROR if the call suxcceeds
//
//  API Use Notes:
//
//  NOTE:  This function uses undocumented and unsupported notes API calls
//  NOTE:  Use this function with great care
//
STATUS ExecEnvironment::GetDbInfoBlock(DBHANDLE hDB, INFOBLK *ibInfo, int iThreadID)
{
	typedef STATUS LNPUBLIC NSFPROC(DBHANDLE, WORD, WORD, DHANDLE *);	//  Typedef for the undocumented entry point

	STATUS			stAPIRC;									//  API return code
	HMODULE			hmNotesEPS;									//  Handle to the notes API library
	DHANDLE			hIB;										//  Returned handle to the Info Block
	void			*pIB;										//  Pointer to the info block in memory
#ifdef UNIX
	void			*hDLL;										//  DLL Handle
#endif
	NSFPROC			*lpNotesEntry;								//  Entry point for undocumented function
	STATUS (LNCALLBACKPTR ProcAddress)(VOID *);					//  Forward definition for the entry point
	char			szFullLibName[MAXPATH + 1];					//  Full name of the notes API DLL
	char			szMsg[MAX_MSG + 1];							//  Generic Error buffer

	//  Load the library containing the needed entry point
	OSGetExecutableDirectory(szFullLibName);
	strcat_s(szFullLibName, MAXPATH, PLATFORM_APILIB_NAME);
	stAPIRC = OSLoadLibrary(szFullLibName, (DWORD) 0, &hmNotesEPS, &ProcAddress);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0059E, PLATFORM_APILIB_NAME);
		LogMessage(szMsg, iThreadID);
		return(stAPIRC);
	}

#ifdef UNIX
	//  Reload the library natively to obtain a native handle
	hDLL = dlopen(szFullLibName, RTLD_LAZY);
	if (hDLL == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0059E, PLATFORM_APILIB_NAME);
		LogMessage(szMsg, iThreadID);
		OSFreeLibrary(hmNotesEPS);
		return(ERR_LIBLOAD2);
	}
	lpNotesEntry = (NSFPROC *) dlsym(hDLL, NSF_GDI_EPNAME);
#endif

#ifdef WIN32
	//  Get the address of the entry point for NSFGetDbInfo
	//  NOTE: The explicit casting of the HMODULE back to HINSTANCE is needed because the Notes API
	//  redefines HMODULE as a generic HANDLE.
	lpNotesEntry = (NSFPROC *) GetProcAddress((HINSTANCE) hmNotesEPS, NSF_GDI_EPNAME);
#endif

	if (lpNotesEntry == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0060E, NSF_GDI_EPNAME, PLATFORM_APILIB_NAME);
		LogMessage(szMsg, iThreadID);
		OSFreeLibrary(hmNotesEPS);
#ifdef UNIX
		dlclose(hDLL);
#endif
		return(ERR_NOT_FOUND);
	}

	//  Call the undocumented entry point
	stAPIRC = (lpNotesEntry) (hDB, (WORD) DBINFO_FLAG_VALUE, 0, &hIB);
	OSFreeLibrary(hmNotesEPS);
#ifdef UNIX
	dlclose(hDLL);
#endif

	//  If there was a bad return code the pass it back
	if (ERR(stAPIRC) != NOERROR)
	{
		return stAPIRC;
	}

	//  Lock the handle and fill the INFOBLK structure with the returned data
	pIB = OSLockObject(hIB);

	//  Copy the info to a local structure
	memcpy(ibInfo, pIB, sizeof(INFOBLK));

	//  Free the Memory Block
	OSUnlockObject(hIB);
	OSMemFree(hIB);
	
	return NOERROR;
}

//  GetDbOptions
//
//  This function will return the (Very Extended) Database Options (4 DWORDS).
//
//  Parameters
//
//  1  -  DBHANDLE				-  Handle of the database to be reported on
//  2  -  DWORD *				-  Pointer to a DWORD to contain the DBOPTIONS
//  3  -  DWORD *				-  Pointer to a DWORD to contain the DBOPTIONS2
//  4  -  DWORD *				-  Pointer to a DWORD to contain the DBOPTIONS3
//  5  -  DWORD *				-  Pointer to a DWORD to contain the DBOPTIONS4
//
//  Returns
//
//  STATUS		-		Status returned by the call
//
//  API Use Notes:
//
//  NOTE:  This function uses undocumented and unsupported notes API calls
//  NOTE:  Use this function with great care
//
STATUS ExecEnvironment::GetDbOptions(DBHANDLE hdbX, DWORD *dwO1, DWORD *dwO2, DWORD *dwO3, DWORD *dwO4)
{
	typedef STATUS LNPUBLIC NSFPROC(DBHANDLE, OptionsArray *);	//  Typedef for the undocumented entry point

	STATUS				stAPIRC;									//  Return code from Notes API calls
	OptionsArray		dbOpts;										//  Db Options Array Structure
	HMODULE				hmNotesEPS;									//  Handle to the notes API library
#ifdef UNIX
	void				*hDLL;										//  DLL Handle
#endif
	NSFPROC				*lpNotesEntry;								//  Entry point for undocumented function
	STATUS (LNCALLBACKPTR ProcAddress)(VOID *);						//  Forward definition for the entry point
	char				szFullLibName[MAXPATH + 1];					//  Full name of the notes API DLL
	char				szMsg[MAX_MSG + 1];							//  Generic Error buffer

	//  Initialise the return values
	stAPIRC = NOERROR;
	*dwO1 = 0;
	*dwO2 = 0;
	*dwO3 = 0;
	*dwO4 = 0;

	//  Load the library containing the needed entry point
	OSGetExecutableDirectory(szFullLibName);
	strcat_s(szFullLibName, MAXPATH, PLATFORM_APILIB_NAME);
	stAPIRC = OSLoadLibrary(szFullLibName, (DWORD) 0, &hmNotesEPS, &ProcAddress);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0059E, PLATFORM_APILIB_NAME);
		std::cout << szMsg << std::endl;
		return(stAPIRC);
	}

#ifdef UNIX
	//  Reload the library natively to obtain a native handle
	hDLL = dlopen(szFullLibName, RTLD_LAZY);
	if (hDLL == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0059E, PLATFORM_APILIB_NAME);
		std::cout << szMsg << std::endl;
		OSFreeLibrary(hmNotesEPS);
		return(ERR_LIBLOAD2);
	}
	lpNotesEntry = (NSFPROC *) dlsym(hDLL, NSF_GETDBO_EPNAME);
#endif

#ifdef WIN32
	//  Get the address of the entry point for NSFGetDbFileSummary
	//  NOTE: The explicit casting of the HMODULE back to HINSTANCE is needed because the Notes API
	//  redefines HMODULE as a generic HANDLE.
	lpNotesEntry = (NSFPROC *) GetProcAddress((HINSTANCE) hmNotesEPS, NSF_GETDBO_EPNAME);
#endif

	if (lpNotesEntry == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0060E, NSF_GETDBO_EPNAME, PLATFORM_APILIB_NAME);
		std::cout << szMsg << std::endl;
		OSFreeLibrary(hmNotesEPS);
#ifdef UNIX
		dlclose(hDLL);
#endif
		return(ERR_NOT_FOUND);
	}

	//  Call the undocumented entry point
	stAPIRC = (lpNotesEntry) (hdbX, &dbOpts);
	OSFreeLibrary(hmNotesEPS);
#ifdef UNIX
	dlclose(hDLL);
#endif

	//  Return the individual options
	if (ERR(stAPIRC) == NOERROR)
	{
		*dwO1 = dbOpts.Options1;
		*dwO2 = dbOpts.Options2;
		*dwO3 = dbOpts.Options3;
		*dwO4 = dbOpts.Options4;
	}

	return stAPIRC;
}

//  GetFileSummary
//
//  This function will return a handle to the database file summary buffer
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database
//  2  -  WORD				-  Flags - set to zero
//  3  -  HANDLE *			-  Pointer to memory where the handle to the object will be returned
//  4  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  STATUS	-				Notes API Status - NOEEROR if the call suxcceeds
//
//  API Use Notes:
//
//  NOTE:  This function uses undocumented and unsupported notes API calls
//  NOTE:  Use this function with great care
//
STATUS ExecEnvironment::GetFileSummary(DBHANDLE hdbX, WORD wFlags, DHANDLE *hFS, int iThreadID)
{
	typedef STATUS LNPUBLIC NSFPROC(DBHANDLE, WORD, DHANDLE *);	//  Typedef for the undocumented entry point

	STATUS			stAPIRC;									//  API return code
	HMODULE			hmNotesEPS;									//  Handle to the notes API library
#ifdef UNIX
	void			*hDLL;										//  DLL Handle
#endif
	NSFPROC			*lpNotesEntry;								//  Entry point for undocumented function
	STATUS (LNCALLBACKPTR ProcAddress)(VOID *);					//  Forward definition for the entry point
	char			szFullLibName[MAXPATH + 1];					//  Full name of the notes API DLL
	char			szMsg[MAX_MSG + 1];							//  Generic Error buffer

	//  Load the library containing the needed entry point
	OSGetExecutableDirectory(szFullLibName);
	strcat_s(szFullLibName, MAXPATH, PLATFORM_APILIB_NAME);
	stAPIRC = OSLoadLibrary(szFullLibName, (DWORD) 0, &hmNotesEPS, &ProcAddress);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0059E, PLATFORM_APILIB_NAME);
		LogMessage(szMsg, iThreadID);
		return(stAPIRC);
	}

#ifdef UNIX
	//  Reload the library natively to obtain a native handle
	hDLL = dlopen(szFullLibName, RTLD_LAZY);
	if (hDLL == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0059E, PLATFORM_APILIB_NAME);
		LogMessage(szMsg, iThreadID);
		OSFreeLibrary(hmNotesEPS);
		return(ERR_LIBLOAD2);
	}
	lpNotesEntry = (NSFPROC *) dlsym(hDLL, NSF_GFS_EPNAME);
#endif

#ifdef WIN32
	//  Get the address of the entry point for NSFGetDbFileSummary
	//  NOTE: The explicit casting of the HMODULE back to HINSTANCE is needed because the Notes API
	//  redefines HMODULE as a generic HANDLE.
	lpNotesEntry = (NSFPROC *) GetProcAddress((HINSTANCE) hmNotesEPS, NSF_GFS_EPNAME);
#endif

	if (lpNotesEntry == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0060E, NSF_GFS_EPNAME, PLATFORM_APILIB_NAME);
		LogMessage(szMsg, iThreadID);
		OSFreeLibrary(hmNotesEPS);
#ifdef UNIX
		dlclose(hDLL);
#endif
		return(ERR_NOT_FOUND);
	}

	//  Call the undocumented entry point
	stAPIRC = (lpNotesEntry) (hdbX, wFlags, hFS);
	OSFreeLibrary(hmNotesEPS);
#ifdef UNIX
	dlclose(hDLL);
#endif	
	return stAPIRC;
}

//  ReopenWith Fixup
//
//  This function will re-open a remote database, after performing a fixup.
//  If the database could not be opened initially then it uses a fake last fixup timestamp.
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database to be re-opened
//  2  -  char *			-  The full net path of the database
//  3  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  DBHANDLE	-			-  The new Database Handle
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::ReopenWithFixup(DBHANDLE hdbX, char *szFullDbName, int iThreadID)
{
	STATUS			stAPIRC;						//  return code from Notes API calls
	DBHANDLE		hdbFixed;						//  Handle to the re-opened database
	DHANDLE			hResponse;						//  Handle to server command response
	TIMEDATE		tdMod1;							//  Data Mod timestamp
	TIMEDATE		tdMod2;							//  Design Mod timestamp
	TIMEDATE		tdOriginalMod1;					//  Original Data Mod timestamp
	TIMEDATE		tdOriginalMod2;					//  Original Design Mod timestamp
	TIMEDATE		tdLastFixup;					//  Timestamp of the last fixup performed on this database
	TIMEDATE		tdLatestFixup;					//  Timestamp of the latest fixup
	BOOL			bWaitForFixup;					//  Waiting for Fixup
	BOOL			bFixupInProgress;				//  Fixup in progress has been detected
	LONG			lTDiff;							//  Difference in timestamps
	int				iFixupTimeoutCounter;			//  Fixup Timeout Counter
	int				iFixupProgressCounter;			//  Fixup in Progress counter
	char			*pResponse;						//  Pointer to the response string
	char			szPath[MAXPATH + 1];			//  Path of the database
	char			szRmtServer[MAXUSERNAME + 1];	//  Remote Server Name
	char			szRmtPath[MAXPATH + 1];			//  Remote D/B Name
	char			szMsg[MAX_MSG + 1];				//  Generic Message Buffer
	char			szRmtCmd[MAX_MSG + 1];			//  Remote command for server

	if (hdbX != NULLHANDLE)
	{
		//  Get the path to the current database
		stAPIRC = NSFDbPathGet(hdbX, NULL, szPath);
		if (ERR(stAPIRC) != NOERROR)
		{
			LogMessage(MSG_DXR0077E, iThreadID);
			GetAPIMessage(stAPIRC, szMsg);
			LogMessage(szMsg, iThreadID);
			SetLastOpenError(ERR(stAPIRC), iThreadID);
			NSFDbClose(hdbX);
			return NULLHANDLE;
		}

		//  Get the Data Modification timestamp and non-Data Modification timestamp for later
		stAPIRC = NSFDbModifiedTime(hdbX, &tdOriginalMod1, &tdOriginalMod2);
		if (ERR(stAPIRC) != NOERROR)
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0078E, szPath);
			LogMessage(szMsg, iThreadID);
			GetAPIMessage(stAPIRC, szMsg);
			LogMessage(szMsg, iThreadID);
			SetLastOpenError(ERR(stAPIRC), iThreadID);
			NSFDbClose(hdbX);
			return NULLHANDLE;
		}

		//  Now get the last fixup timestamp for the database
		if (!GetLastFixupTime(hdbX, &tdLastFixup, iThreadID))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0078E, szPath);
			LogMessage(szMsg, iThreadID);
			SetLastOpenError(ERR_NEEDS_FIXUP, iThreadID);
			NSFDbClose(hdbX);
			return NULLHANDLE;
		}

		//  Log the timestamps
		LogTimestamps(&tdOriginalMod1, &tdOriginalMod2, &tdLastFixup, iThreadID);

		//  Check if the Fixup is necessary - this uses that same logic as is used in fixup to avoid 
		//  doing fixups when it is not needed. It is important to do this check here as we do not
		//  want to attempt a fixup which will not execute and have us waiting for it to start/complete. 
		//  If the last fixup timestamp is later than both the Data Modification timestamp and the 
		//  non-Data modification timestamp then skip the fixup and return the unreopened D/B handle.
		lTDiff = TimeDateDifference(&tdLastFixup, &tdOriginalMod1);
		if (lTDiff > 0)
		{
			//  Last fixup is later than Data Mod time - now try the non data mod time
			lTDiff = TimeDateDifference(&tdLastFixup, &tdOriginalMod2);
			if (lTDiff > 0)
			{	
				//  Fixup is not needed - log the fact and return the original D/B handle
				LogMessage(MSG_DXR0079I, iThreadID);
				SetLastOpenError(NOERROR, iThreadID);
				return hdbX;
			}
		}

		//  Close the database in preparation for the fixup
		NSFDbClose(hdbX);
	}
	else
	{	
		//  The database could not be opened - use a dummy timestamp for the last fixup time
		OSCurrentTIMEDATE(&tdLastFixup);
		strcpy_s(szPath, MAXPATH, szFullDbName);
		LogMessage(MSG_DXR0107I, iThreadID);
	}

	//  Parse the returned path into a server and file name
	stAPIRC = OSPathNetParse(szPath, NULL, szRmtServer, szRmtPath);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0080E, szPath);
		LogMessage(szMsg, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		SetLastOpenError(ERR(stAPIRC), iThreadID);
		return NULLHANDLE;
	}

	//  Format the fixup command
	sprintf_s(szRmtCmd, MAX_MSG, FORMAT_FIXUP_CMD, szRmtPath);
	
	//  Issue the Fixup command on the remote server
	stAPIRC = NSFRemoteConsole(szRmtServer, szRmtCmd, &hResponse);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0081E, szRmtCmd, szRmtServer);
		LogMessage(szMsg, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		SetLastOpenError(ERR(stAPIRC), iThreadID);
		return NULLHANDLE;
	}

	//  Show and dispose of the response
	pResponse = OSLock(char, hResponse);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0082I, szRmtServer, pResponse);
	LogVerbose(szMsg, iThreadID);
	OSUnlock(hResponse);
	OSMemFree(hResponse);

	//  Wait until the command has executed
	bWaitForFixup = TRUE;
	bFixupInProgress = FALSE;
	iFixupTimeoutCounter = 0;
	iFixupProgressCounter = 0;

	while (bWaitForFixup)
	{
		//  Re-Open the database and obtain the database IDs
		stAPIRC = NSFDbOpenExtended(szPath, 0L, NULLHANDLE, NULL, &hdbFixed, &tdMod1, &tdMod2);
		if (ERR(stAPIRC) == ERR_DBOFFLINE)
		{
			//  The fixup is in progress - we now wait for it to complete
			bFixupInProgress = TRUE;
			//  Check that we have not exhausted the fixup progress timeout
			if (iFixupProgressCounter > FIXUP_PROGRESS_LIMIT)
			{
				//  Abandon the attempt
				LogMessage(MSG_DXR0083E, iThreadID);
				return NULLHANDLE;
			}
			AppWait(FIXUP_WAIT_INTERVAL);
			iFixupProgressCounter++;
		}
		else
		{
			if ((ERR(stAPIRC) != NOERROR) && (ERR(stAPIRC) != ERR_NEEDS_FIXUP))
			{
				sprintf_s(szMsg, MAX_MSG, MSG_DXR0084E, szPath);
				LogMessage(szMsg, iThreadID);
				GetAPIMessage(stAPIRC, szMsg);
				LogMessage(szMsg, iThreadID);
				SetLastOpenError(ERR(stAPIRC), iThreadID);
				return NULLHANDLE;
			}
			else
			{
				if (ERR(stAPIRC) == ERR_NEEDS_FIXUP)
				{
					//  Check that we have not exceeded the limits for waiting for fixup to start
					if (iFixupTimeoutCounter > FIXUP_WAIT_LIMIT)
					{
						//  Abandon the attempt
						LogMessage(MSG_DXR0087E, iThreadID);
						SetLastOpenError(ERR_NEEDS_FIXUP, iThreadID);
						return NULLHANDLE;
					}
					AppWait(FIXUP_WAIT_INTERVAL);
					iFixupTimeoutCounter++;
				}
				else
				{
					//  If fixup in progress was detected then it's now safe to leave the wait loop
					if (bFixupInProgress)
					{
						sprintf_s(szMsg, MAX_MSG, MSG_DXR0085I, szRmtPath, szRmtServer);
						LogMessage(szMsg, iThreadID);
						NSFDbClose(hdbFixed);
						bWaitForFixup = FALSE;
					}
					else
					{
						//  Get the Last Fixup timestamp to check if it's later than the original
						if (!GetLastFixupTime(hdbFixed, &tdLatestFixup, iThreadID))
						{
							LogMessage(MSG_DXR0086E, iThreadID);
							NSFDbClose(hdbFixed);
							SetLastOpenError(ERR_NEEDS_FIXUP, iThreadID);
							return NULLHANDLE;
						}

						NSFDbClose(hdbFixed);
						//  Log the new timestamp
						LogTimestamps(NULL, NULL, &tdLatestFixup, iThreadID);

						//  Determine if the Timestamp has changed in which case the fixup has completed
						if (!TimeDateEqual(&tdLastFixup, &tdLatestFixup))
						{
							sprintf_s(szMsg, MAX_MSG, MSG_DXR0085I, szRmtPath, szRmtServer);
							LogMessage(szMsg, iThreadID);
							bWaitForFixup = FALSE;
						}
						else
						{
							//  Check that we have not exceeded the limits for waiting for fixup to start
							if (iFixupTimeoutCounter > FIXUP_WAIT_LIMIT)
							{
								//  Abandon the attempt
								LogMessage(MSG_DXR0087E, iThreadID);
								SetLastOpenError(ERR_NEEDS_FIXUP, iThreadID);
								return NULLHANDLE;
							}
							AppWait(FIXUP_WAIT_INTERVAL);
							iFixupTimeoutCounter++;
						}
					}
				}
			}
		}
	}

	//  Reopen the database - or keep trying to while Fixup is on progress
	stAPIRC = ERR_DBOFFLINE;

	while (ERR(stAPIRC) == ERR_DBOFFLINE)
	{
		stAPIRC = NSFDbOpenExtended(szPath, 0L, NULLHANDLE, NULL, &hdbFixed, &tdMod1, &tdMod2);
		if (ERR(stAPIRC) == ERR_DBOFFLINE)
		{
			//  Check that we have not exhausted the fixup progress timeout
			if (iFixupProgressCounter > FIXUP_PROGRESS_LIMIT)
			{
				//  Abandon the attempt
				LogMessage(MSG_DXR0083E, iThreadID);
				SetLastOpenError(ERR_NEEDS_FIXUP, iThreadID);
				return NULLHANDLE;
			}
			AppWait(FIXUP_WAIT_INTERVAL);
			iFixupProgressCounter++;
		}
	}
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0088E, szPath);
		LogMessage(szMsg, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		SetLastOpenError(ERR(stAPIRC), iThreadID);
		return NULLHANDLE;
	}
	SetLastOpenError(NOERROR, iThreadID);
	return hdbFixed;
}

//  GetLastFixupTime
//
//  This function will return the last fixup timestamp for the requested database
//
//  Parameters
//
//  1  -  DBHANDLE			-  Handle of the database to be re-opened
//  2  -  TIMEDATE *		-  Pointer to the TIMEDATE structure that will be filled
//  3  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  BOOL	-				-  TRUE if the Fixup Timestamp was populated
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::GetLastFixupTime(DBHANDLE hdbX, TIMEDATE *tdLastFixup, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API calls
	DHANDLE				hFS;								//  Hande to the returned summary buffer
	ITEM_TABLE			*itSummary;							//  Pointer to the summary buffer
	TIMEDATE			*tdSBLFU;							//  Returned pointer to the TimeDate
	WORD				wRetLen;							//  Length of item returned
	WORD				wType;								//  Data type of the returmed item
	char				szMsg[MAX_MSG + 1];					//  Generic Message Buffer

	//  Get the Database Summary buffer for the passed D/B handle
	stAPIRC = GetFileSummary(hdbX, 0, &hFS, iThreadID);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0089E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	//  Lock the returned handle to dereference a pointer to the summary buffer
	itSummary = (ITEM_TABLE *) OSLockObject(hFS);

	//  Locate the Last Fixup Timestamp in the summary buffer
	if (NSFLocateSummaryValue(itSummary, DBDIR_LASTFIXUP_ITEM, &tdSBLFU, &wRetLen, &wType))
	{
		//  Return the timedate
		memcpy(tdLastFixup, tdSBLFU, sizeof(TIMEDATE));
		//  Unlock and free the memory
		OSUnlock(hFS);
		OSMemFree(hFS);
		return TRUE;
	}
	else
	{
		LogMessage(MSG_DXR0090E, iThreadID);
		//  Unlock and free the memory
		OSUnlock(hFS);
		OSMemFree(hFS);
		return FALSE;
	}
}

//  LogTimeStamps
//
//  This function will write certain critical D/B timestamps to the current log
//
//  Parameters
//
//  1  -  TIMEDATE *		-  Pointer to the D/B data update TIMEDATE structure
//  2  -  TIMEDATE *		-  Pointer to the D/B design update TIMEDATE structure
//  3  -  TIMEDATE *		-  Pointer to the D/B last fixup TIMEDATE structure
//  4  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::LogTimestamps(TIMEDATE *tdDMod, TIMEDATE *tdNDMod, TIMEDATE *tdLFU,  int iThreadID)
{
	STATUS				stAPIRC;						//  Return code from Notes API calls
	INTLFORMAT			ifCurrent;						//  International Format options
	TFMT				tfLogTime;						//  Time/Date format options
	WORD				wRetLen;						//  Returned length
	char				szTimeStamp[MAXALPHATIMEDATE + 1];  //  Formatted TimeDate
	char				szMsg[MAX_MSG + 1];				//  Generic Message Buffer

	//  Initialise the format options
	OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
	tfLogTime.Date = TDFMT_FULL;
	tfLogTime.Time = TTFMT_FULL;
	tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
	tfLogTime.Structure = TSFMT_DATETIME;

	//  If the passed values are not NULL then log them
	if (tdDMod != NULL)
	{
		stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
										&tfLogTime,
										tdDMod,
										szTimeStamp,
										sizeof(szTimeStamp),
										&wRetLen);
		if (ERR(stAPIRC) == NOERROR)
		{
			szTimeStamp[wRetLen] = '\0';
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0091I, 
						szTimeStamp,
						tdDMod->Innards[1],
						tdDMod->Innards[0]);
			LogMessage(szMsg, iThreadID);
		}
	}

	if (tdNDMod != NULL)
	{
		stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
										&tfLogTime,
										tdNDMod,
										szTimeStamp,
										sizeof(szTimeStamp),
										&wRetLen);
		if (ERR(stAPIRC) == NOERROR)
		{
			szTimeStamp[wRetLen] = '\0';
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0092I, 
						szTimeStamp,
						tdNDMod->Innards[1],
						tdNDMod->Innards[0]);
			LogMessage(szMsg, iThreadID);
		}
	}

	if (tdLFU != NULL)
	{
		stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
										&tfLogTime,
										tdLFU,
										szTimeStamp,
										sizeof(szTimeStamp),
										&wRetLen);
		if (ERR(stAPIRC) == NOERROR)
		{
			szTimeStamp[wRetLen] = '\0';
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0093I, 
						szTimeStamp,
						tdLFU->Innards[1],
						tdLFU->Innards[0]);
			LogMessage(szMsg, iThreadID);
		}
	}

	return;
}

//  DeleteDatabase
//
//  This function will delete the specified database (optionally it will take the D/B offline first)
//
//  Parameters
//
//  1  -  char *			-  Pointer to the server name (normally abbreviated)
//  2  -  char *			-  Pointer to the database path name (relative to the notes data directory)
//  3  -  BOOL				-  If TRUE then the database is taken offline & deleted (only valid for local D/Bs)
//  4  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  API Use Notes:
//
//
STATUS ExecEnvironment::DeleteDatabase(char *szServer, char *szDatabase, BOOL bMakeOffline, int iThreadID)
{
	STATUS			stAPIRC;									//  API return code
	WORD			wRetLen;									//  Returned sizesversion factor
	char			szAPIMsg[MAX_MSG + 1];						//  API Error message
	char			szFullDbName[MAXPATH + 1];					//  Full Database Name
	char			szCanonicalServer[MAXUSERNAME + 1];			//  Canonical server name

	//  Open a Domino Database/Directory
	TrimToUpper_s(szAPIMsg, MAX_MSG, szServer);

	//  Check if the specified server is the current server - if so covert to a local open
	if (strcmpi(szAPIMsg, AbbreviatedUser) == 0) szAPIMsg[0] = '\0';


	if (strlen(szAPIMsg) == 0 || (strcmp(szAPIMsg, "LOCAL") == 0))
	{
		strcpy_s(szFullDbName, MAXPATH, szDatabase);
	}
	else
	{
		if (bMakeOffline)
		{
			//  Invalid - return a suitable RC
			return ERR_NOT_LOCAL;
		}
		//  canonicalize the server name and build the NET path for the database
		stAPIRC = DNCanonicalize(0L, NULL, szServer, szCanonicalServer, MAXUSERNAME, &wRetLen);
		if (ERR(stAPIRC) != NOERROR)
		{
			return stAPIRC;
		}

		stAPIRC = OSPathNetConstruct(NULL, szCanonicalServer, szDatabase, szFullDbName);
		if (ERR(stAPIRC) != NOERROR)
		{
			return stAPIRC;
		}
	}
	
	//  Determine which method to use for the delete
	if (bMakeOffline)
	{
		stAPIRC = NSFTakeDatabaseOffline(szFullDbName, DEFAULT_OFFLINE_WAIT, 0);
	}
	else
	{
		stAPIRC = NSFDbDelete(szFullDbName);
	}
	return stAPIRC;
}

//  AssertGlobalPermit
//
//  This function will assert the DCF Global Permit Flag
//
//  Parameters
//
//  Returns
//
//  BOOL	-		TRUE if the flag was changed to TRUE, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::AssertGlobalPermit(void)
{
	if (!DCFGlobalPermit)
	{
		DCFGlobalPermit = TRUE;
		return TRUE;
	}
	return FALSE;
}

//  RevokeGlobalPermit
//
//  This function will revoke the DCF Global Permit Flag
//
//  Parameters
//
//  Returns
//
//  BOOL	-		TRUE if the flag was changed to FALSE, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::RevokeGlobalPermit(void)
{
	if (DCFGlobalPermit)
	{
		DCFGlobalPermit = FALSE;
		return TRUE;
	}
	return FALSE;
}

//  TestGlobalPermit
//
//  This function will the current value of the DCF Global Permit Flag
//
//  Parameters
//
//  Returns
//
//  BOOL	-		The current value of the Global Permit flag
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::TestGlobalPermit(void)
{
	//  Check for a quit signal
	if (hmqProcessor != NULLHANDLE)
	{
		//  Revoke the global permit
		if (AddInShouldTerminate()) DCFGlobalPermit = FALSE;
	}
	return DCFGlobalPermit;
}

//  OSLoadSrvCommand
//
//  This function will load and execute the server command that is defined in the passed parameter block.
//
//  Parameters
//
//  1  -  OSPROG_DEF *		-  Pointer to the program definition block for the server command
//  2  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  int	-		Indicates whether function was successful or type of error
//
//  API Use Notes:
//
//  This function is available for Windows only, other O/S implementations should use the Notes API OSLoadProgram function.
//  The function is supplied to get round a bug in certain Domino O/S version combinations that breaks the OSLoadProgram
//  function when it launches an application and waits for completion.
//
#ifdef WIN32
int ExecEnvironment::OSLoadSrvCommand(OSPROG_DEF *ospProg, int  iThreadID)
{
     OSGetDataDirectory(ospProg->szWorkingDir);
     return  OSLoadProgram(ospProg, iThreadID);
}
#endif

//  OSLoadProgram
//
//  This function will load and execute the specified program. This function provides the same capability as
//  the Notes API OSLoadProgram, but this one works correctly on Windows.
//
//  Parameters
//
//  1  -  OSPROG_DEF *		-  Pointer to the program definition block for the server command
//  2  -  int				-  The Thread ID of the caller
//
//  Returns
//
//  int	-		Indicates whether function was successful or type of error
//
//  API Use Notes:
//
//  This function is available for Windows only, other O/S implementations should use the Notes API OSLoadProgram function.
//  The function is supplied to get round a bug in certain Domino O/S version combinations that breaks the OSLoadProgram
//  function when it launches an application and waits for completion.
//
#ifdef WIN32
int ExecEnvironment::OSLoadProgram(OSPROG_DEF *ospProg, int  iThreadID)
{
    char						*lpAppName = NULL;						//  Name of the executable
    char						*lpWorkingDir = NULL;					//  Name of the working directory
    char						*lpAppCmdLine = NULL;					//  The command line to be used
    STARTUPINFO					siProcess;								//  Program Start Information Block
    PROCESS_INFORMATION			piProcess;								//	Process Information Block
    char						szMsg[MAX_MSG + 1];						//  Generic Message Buffer
     
	ospProg->dwReturnCode = 0;
	ospProg->dwProcessId = 0;

	// Clear the information blocks
	ZeroMemory( &siProcess, sizeof (siProcess) );
	siProcess.cb = sizeof (siProcess);
	ZeroMemory( &piProcess, sizeof (piProcess) );

	//  Set up local pointers
	lpAppName = ospProg->szAppName;
	lpWorkingDir = ospProg->szWorkingDir;
	lpAppCmdLine = ospProg->szAppCmdLine;

	//  Trace the load
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0095T, ospProg->szAppName, ospProg->szWorkingDir);
	LogTrace(TRACE_CORERT, szMsg, iThreadID);
	sprintf_s(szMsg, MAX_MSG, MSG_DXR0096T, ospProg->szAppCmdLine, ospProg->dwFlags);
	LogTrace(TRACE_CORERT, szMsg, iThreadID);
	
	// Start the child process. 
	if (!CreateProcess(lpAppName,							// No module name (use command line)
						lpAppCmdLine,						// Command line
						NULL,								// Process handle not inheritable
						NULL,								// Thread handle not inheritable
						FALSE,								// Set handle inheritance to FALSE
						ospProg->dwFlags,					// No creation flags
						NULL,								// Use parent's environment block
						lpWorkingDir,						// Use parent's starting directory 
						&siProcess,							// Pointer to STARTUPINFO structure
						&piProcess))						// Pointer to PROCESS_INFORMATION structure
	{
		ospProg->dwReturnCode = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0097E, ospProg->szAppName, ospProg->dwReturnCode);
		LogMessage(szMsg, iThreadID);
		GetWINAPIError(ospProg->dwReturnCode, szMsg);
		LogMessage(szMsg, iThreadID);
		return RC_OSPROG_CP;
	}

	//  If the program is going to run detached - then we can return to the caller
	if  (ospProg->dwFlags && DETACHED_PROCESS)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0098T, ospProg->szAppName);
		LogTrace(TRACE_CORERT, szMsg, iThreadID);
		ospProg->dwProcessId = piProcess.dwProcessId;
		CloseHandle(piProcess.hProcess);
		CloseHandle(piProcess.hThread);
		return RC_OSPROG_SA;
	}

	// Wait until child process exits.
	WaitForSingleObject(piProcess.hProcess, INFINITE);

	// Close thread handle
	CloseHandle( piProcess.hThread );

	// Get exit code from executable
	if (!GetExitCodeProcess (piProcess.hProcess, &ospProg->dwReturnCode))
	{
		ospProg->dwReturnCode = GetLastError();
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0099E, ospProg->szAppName, ospProg->dwReturnCode);
		LogMessage(szMsg, iThreadID);
		GetWINAPIError(ospProg->dwReturnCode, szMsg);
		LogMessage(szMsg, iThreadID);
		CloseHandle( piProcess.hProcess);
		return RC_OSPROG_GECP;
	}

	// Close process handle
	CloseHandle( piProcess.hProcess);

	sprintf_s(szMsg, MAX_MSG, MSG_DXR0100T, ospProg->szAppName, ospProg->szAppCmdLine, ospProg->dwReturnCode );
	LogTrace(TRACE_CORERT, szMsg, iThreadID);

	return RC_OSPROG_OK;
} 
#endif

//  GetDbAllocatedSize
//
//  This function will return the allocated size of a database in KB.
//
//  Parameters
//
//  1  -  DBHANDLE               -  Handle of the database 
//  2  -  int                    -  The Thread ID of the caller
//
//  Returns
//
//  DWORD     -                    -1 is the size could not be determined, otherwise the size in Kb
//
//  API Use Notes:
//
//
DWORD ExecEnvironment::GetDbAllocatedSize(DBHANDLE hDB, int  iThreadID)
{
	DWORD				dwSizeUsed;							//  Size Used
	DWORD				dwSizeFree;							//  Size Free
	DWORD				dwGSize;							//  Granule Size
     
	//  Get the space used
	NSFDbSpaceUsageScaled(hDB, &dwSizeUsed, &dwSizeFree, &dwGSize);
	if (dwGSize < BYTES_PER_KB)
	{
		dwGSize = BYTES_PER_KB / dwGSize;
		dwSizeUsed = dwSizeUsed / dwGSize;
	}
	else
	{
		dwGSize = dwGSize / BYTES_PER_KB;
		dwSizeUsed = dwSizeUsed * dwGSize;
	}

     //  Return the count
     return  dwSizeUsed;
}

//  GetDbFreeSpace
//
//  This function will return the free space in a database in KB.
//
//  Parameters
//
//  1  -  DBHANDLE               -  Handle of the database 
//  2  -  int                    -  The Thread ID of the caller
//
//  Returns
//
//  DWORD     -                    -1 is the free space could not be determined, otherwise the free space in Kb
//
//  API Use Notes:
//
//
DWORD ExecEnvironment::GetDbFreeSpace(DBHANDLE hDB, int  iThreadID)
{
	DWORD				dwSizeUsed;							//  Size Used
	DWORD				dwSizeFree;							//  Size Free
	DWORD				dwGSize;							//  Granule Size
     
	//  Get the space used
	NSFDbSpaceUsageScaled(hDB, &dwSizeUsed, &dwSizeFree, &dwGSize);
	if (dwGSize < BYTES_PER_KB)
	{
		dwGSize = BYTES_PER_KB / dwGSize;
		dwSizeFree = dwSizeFree / dwGSize;
	}
	else
	{
		dwGSize = dwGSize / BYTES_PER_KB;
		dwSizeFree = dwSizeFree * dwGSize;
	}

     //  Return the count
     return  dwSizeFree;
}

//  RemoveEscapes
//
//  This function will eliminate all escape (%) charactes from the passed string
//
//  Parameters
//
//  1  -  char *				-  Pointer to the string to be sanitized
//  2  -  int                   -  size of the string buffer to be sanitized
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::RemoveEscapes(char *szMsg, int iStr_Size)
{
	char	*pChar;									//  Pointer to character in the string
	int		iIndex = 0;								//  Index into the string

	if (iStr_Size == 0) return;
	pChar = szMsg;
	while (*pChar != '\0')
	{
		iIndex++;
		if (iIndex == iStr_Size)
		{
			*pChar = '\0';							//  Properly terminate the string
			return;
		}
		if (*pChar == '%')
		{
			*pChar = ' ';							//  Convert to a blank
		}
		pChar++;
	}
}

//  GenUniqueID
//
//  This function runs @Unique() to generate a random, unique text value
//
//  Parameters
//
//  1  -  char * -  Pointer to the target character buffer
//  2  -  int -  Size of the target character buffer
//  3  -  int -  The Thread ID of the caller
//
//  Returns
//
//  BOOL - TRUE if the action was successful, otherwise FALSE
//
//  API Use Notes:
//
BOOL ExecEnvironment::GenUniqueID(char *szUniqueID, int iUniqueIDLen, int iThreadID)
{
	STATUS			stAPIRC;							//  Return code from Notes API calls
	STATUS			stRetCompileError;					//  Return code from formula compilation
	FORMULAHANDLE	hRetFormula;						//  Handle to the compiled formula
	HCOMPUTE		hCompute;							//  Handle to the executable formula
	DHANDLE			hResult;							//  Handle to the execution result
	USHORT			usEntries;							//  Result descriptor
	USHORT			usLen;								//  Result descriptor
	WORD			wRetFormulaLength;					//  Formula length
	WORD			wRetCompileErrorLine;				//  Compile error info
	WORD			wRetCompileErrorColumn;				//  Compile error info
	WORD			wRetCompileErrorOffset;				//  Compile error info
	WORD			wRetCompileErrorLength;				//  Compile error info
	WORD			wRetResultLen;						//  Result length
	WORD			wDataType;							//  Result data type
	char			*pFormula;							//  Pointer to compiled formula
	char			*pResult;							//  Pointer to formula result
	char			szMsg[MAX_MSG + 1];					//  Generic Message Buffer

	szUniqueID[0] = '\0';

	//  Compile the formula - not the formula is escaped so that it is language independent
	stAPIRC = NSFFormulaCompile(NULL, 
		NULL, 
		FORMULA_GENUNIQUEID, 
		(WORD) strlen(FORMULA_GENUNIQUEID), 
		&hRetFormula,
		&wRetFormulaLength,
		&stRetCompileError,
		&wRetCompileErrorLine,
		&wRetCompileErrorColumn,
		&wRetCompileErrorOffset,
		&wRetCompileErrorLength);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0101E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	pFormula = NULL;
	pFormula = (char *) OSLockObject(hRetFormula);
	if (pFormula == NULL)
	{
		LogMessage(MSG_DXR0102E, iThreadID);
		return FALSE;
	}

	//  Generate an execution context for executing the formula
	stAPIRC = NSFComputeStart(0, pFormula, &hCompute);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0103E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		OSUnlockObject(hRetFormula);
		OSMemFree(hRetFormula);
		return FALSE;
	}

	//  Evaluate the formula
	stAPIRC = NSFComputeEvaluate(hCompute, NULL, &hResult, &wRetResultLen, 0, 0, 0);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0104E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		NSFComputeStop(hCompute);
		OSUnlockObject(hRetFormula);
		OSMemFree(hRetFormula);
		return FALSE;
	}

	pResult = (char *) OSLockObject(hResult);

	//  Get the datatype of the result
	wDataType = *(WORD *) pResult;

	//  Set pResult to point to the actual data
	pResult += sizeof(WORD); 

	//  Reduce length of result by 1
	wRetResultLen -= sizeof(WORD);

	//  Get UniqueID Text
	switch (wDataType)
	{
	case TYPE_TEXT:
		strncpy_s(szUniqueID, iUniqueIDLen, pResult, (rsize_t) wRetResultLen);
		szUniqueID[wRetResultLen] = '\0';
		break;

	case TYPE_TEXT_LIST:

		usEntries = *(USHORT *) pResult;
		pResult += sizeof(USHORT);

		usLen = *(USHORT *) pResult;
		pResult += ( usEntries * sizeof(USHORT));

		strncpy_s(szUniqueID, iUniqueIDLen, pResult, (rsize_t) usLen);
		szUniqueID[usLen] = '\0';
		break;

	default:
		OSUnlockObject(hResult);
		OSMemFree(hResult);
		OSUnlockObject(hRetFormula);
		OSMemFree(hRetFormula);
		NSFComputeStop(hCompute);
		return FALSE;
	}

	//  Unlock the handle to the formula result and free the memory
	OSUnlockObject(hResult);
	OSMemFree(hResult);

	//  Unlock the handle to the compiled formula and free the memory
	OSUnlockObject(hRetFormula);
	OSMemFree(hRetFormula);

	//  Call NSFComputeStop to mark the end of the formula evaluation
	stAPIRC = NSFComputeStop(hCompute);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0105E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		szUniqueID[0] = '\0';
		return FALSE;
	}
	return TRUE;
}

//  xtoi
//
//  This function will convert the passed hexadecimal string to an integer value
//
//  Parameters
//
//  1  -  char * -  Pointer to the hexadecimal string to be converted
//
//  Returns
//
//  int - value contained in the hexadecimal string
//
//  API Use Notes:
//
int ExecEnvironment::xtoi(char *szHex)
{
	int		iOffset = 0;
	int		iAccum = 0;
	int		iDigit;

	while (szHex[iOffset] > '\0')
	{
		if ((szHex[iOffset] >= '0') && (szHex[iOffset] <= '9'))
		{
			iDigit = szHex[iOffset] - '0';
		}
		else if ((szHex[iOffset] >= 'a') && (szHex[iOffset] <= 'f'))
		{
			iDigit = szHex[iOffset] - 'a' + 10;
		}
		else if ((szHex[iOffset] >= 'A') && (szHex[iOffset] <= 'F'))
		{
			iDigit = szHex[iOffset] - 'A' + 10;
		}
		else iDigit = 0;

		iAccum = (iAccum << 4) + iDigit;
		iOffset++;
	}
	return iAccum;
}

//  populateUNID
//
//  This function will populate a UNID from the passed hexadecimal string
//
//  Parameters
//
//  1  -  char * -  Pointer to the hexadecimal string to be converted
//  2  -  UNID * -  Pointer to the UNID structure that is to be populated
//
//  Returns
//
//
//  API Use Notes:
//
//
void ExecEnvironment::populateUNID(char *szUNID, UNID *uidFormatted)
{
	int			iOff1;
	int			iOff2;
	int			iOff3;
	char		szUNIDBit[9];								//  Work buffer

	//  Clear the UNID values
	uidFormatted->File.Innards[1] = 0;
	uidFormatted->File.Innards[0] = 0;
	uidFormatted->Note.Innards[1] = 0;
	uidFormatted->Note.Innards[0] = 0;

	switch(strlen(szUNID))
	{
	case 32:
		iOff1 = 8;
		iOff2 = 16;
		iOff3 = 24;
		break;
	case 33:
		iOff1 = 8;
		iOff2 = 17;
		iOff3 = 25;
		break;
	case 35:
		iOff1 = 9;
		iOff2 = 18;
		iOff3 = 27;
		break;
	default:
		iOff1 = 8;
		iOff2 = 16;
		iOff3 = 24;
	}

	//  Take a chunk of the UNID String at a time and populate the appropriate member
	memcpy(&szUNIDBit, szUNID, 8);
	szUNIDBit[8] = '\0';
	uidFormatted->File.Innards[1] = xtoi(szUNIDBit);
	
	memcpy(&szUNIDBit, szUNID + iOff1, 8);
	szUNIDBit[8] = '\0';
	uidFormatted->File.Innards[0] = xtoi(szUNIDBit);

	memcpy(&szUNIDBit, szUNID + iOff2, 8);
	szUNIDBit[8] = '\0';
	uidFormatted->Note.Innards[1] = xtoi(szUNIDBit);
	
	memcpy(&szUNIDBit, szUNID + iOff3, 8);
	szUNIDBit[8] = '\0';
	uidFormatted->Note.Innards[0] = xtoi(szUNIDBit);
	return;
}

//  ShowStats
//
//  This function will show statistics from the Runtime - this version is minimal
//  the function is overloaded in the MTExecutive which has more stats
//
//  Parameters
//
//  1  -  int		-  options describing what statistics to show
//  2  -  int		-  Thread ID of the requesting thread
//
//  Returns
//
//
//  API Use Notes:
//
//
void ExecEnvironment::ShowStats(int iOpts, int iThreadID)
{
	STATUS						stAPIRC;								//  Return code from Notes API calls
	INTLFORMAT					ifCurrent;								//  International Format options
	TFMT						tfLogTime;								//  Time/Date format options
	WORD						wRetLen;								//  Returned length
	char						szTimeStamp[MAXALPHATIMEDATE + 1];		//  Formatted TimeDate
	char						szMsg[MAX_MSG + 1];						//  Generic message buffer

	//  Initialise the format options
	OSGetIntlSettings(&ifCurrent, sizeof(ifCurrent));
	tfLogTime.Date = TDFMT_FULL;
	tfLogTime.Time = TTFMT_FULL;
	tfLogTime.Zone = TZFMT_ALWAYS;                           //  Always show Zone
	tfLogTime.Structure = TSFMT_DATETIME;

	//  Show the thread manager stats for the current run
	szTimeStamp[0] = '\0';
	stAPIRC = ConvertTIMEDATEToText(&ifCurrent,
									&tfLogTime,
									&tdStarted,
									szTimeStamp,
									sizeof(szTimeStamp),
									&wRetLen);
	if (ERR(stAPIRC) == NOERROR)
	{
		szTimeStamp[wRetLen] = '\0';
	}

	sprintf_s(szMsg, MAX_MSG, MSG_DXR0106I, szTimeStamp);
	LogMessage(szMsg, iThreadID);
	AddInLogMessageText(szMsg, NOERROR);
	return;
}

//  SetLastOpenError
//
//  This function will store the last open error code so that it can be interrogated by the caller
//
//  Parameters
//
//  1  -  STATUS	-  Last error code
//  2  -  int		-  Thread ID of the requesting thread
//
//  Returns
//
//
//  API Use Notes:
//
//
void ExecEnvironment::SetLastOpenError(STATUS stOpen, int iThreadID)
{
	stLastOpenError = stOpen;
}

//  GetLastOpenError
//
//  This function will store the last open error code so that it can be interrogated by the caller
//
//  Parameters
//
//
//  Returns
//
//	STATUS		-  The last API return code from D/B Open calls
//
//  API Use Notes:
//
//
STATUS ExecEnvironment::GetLastOpenError()
{
	return GetLastOpenError(0);
}

//  GetLastOpenError
//
//  This function will store the last open error code so that it can be interrogated by the caller
//
//  Parameters
//
//  1  -  int		-  Thread ID of the requesting thread
//
//  Returns
//
//	STATUS		-  The last API return code from D/B Open calls
//
//  API Use Notes:
//
//
STATUS ExecEnvironment::GetLastOpenError(int iThreadID)
{
	return stLastOpenError;
}

//  GetACLRoles
//
//  This function will get the privilege mask and string bufer for all roles that are defined in the database.
//
//  Parameters
//
//  1  -  HANDLE			-  Handle of the ACL for which roles are to be identified
//  2  -  ACL_PRIVILEGES *	-  Pointer to the ACL_PRIVILEGES array where the full set of priv bits will be returned
//  3  -  char *			-  Pointer to a buffer that will contain all of the role names from the ACL
//  4  -  int				-  The ID of the thread from which the open request is being made
//
//  Returns
//
//  BOOL	-			TRUE if the ACL was parsed
//						FALSE if the ACL could not be parsed
//
//  API Use Notes:
//
//  The buffer to contain the ACL roles must be (ACL_PRIVSTRINGMAX * ACL_PRIVCOUNT) + 1 bytes.
//
BOOL ExecEnvironment::GetACLRoles(DHANDLE hACL, ACL_PRIVILEGES *pAllPrivs, char *szRoles, int iThreadID)
{
	STATUS					stAPIRC;						//  Return code from Notes API calls
	int						iIndex;							//  Generic index
	ACL_PRIVILEGES			apAll;							//  All Privs build buffer
	char					szMsg[MAX_MSG + 1];				//  Generic Message Buffer
	char					szRole[ACL_PRIVSTRINGMAX + 1];	//  Buffer for Role name

	szRoles[0] = '\0';										//  Clear the roles buffer
	memset(pAllPrivs, 0, sizeof(ACL_PRIVILEGES));			//  Clear the privileges bit array

	//  Loop through testing all of the possible privelege numbers available
	for (iIndex = ACL_BITPRIVCOUNT; iIndex < ACL_PRIVCOUNT; iIndex++)
	{
		stAPIRC = ACLGetPrivName(hACL, iIndex, szRole);
		if ((ERR(stAPIRC) != NOERROR) && (ERR(stAPIRC) != ERR_ACL_NOENTRY))
		{
			sprintf_s(szMsg, MAX_MSG, MSG_DXR0108E, iIndex);
			LogMessage(szMsg, iThreadID);
			GetAPIMessage(stAPIRC, szMsg);
			LogMessage(szMsg, iThreadID);
			return FALSE;
		}
		if ((szRole[0] != '\0') && (ERR(stAPIRC) == NOERROR))
		{
			sprintf_s(szRoles, (ACL_PRIVSTRINGMAX * ACL_PRIVCOUNT), "%s%s", szRoles, szRole);
			ACLSetPriv(apAll, iIndex);
		}
		else
		{
			sprintf_s(szRoles, (ACL_PRIVSTRINGMAX * ACL_PRIVCOUNT), "%s[]", szRoles);
		}
	}

	//  Return the privilege bit array
	memcpy(pAllPrivs, &apAll, sizeof(ACL_PRIVILEGES));
	return TRUE;
}

//  GetNoteUNID
//
//  This function will get the UNID from a notehandle as a UNID structure and/or a string
//
//  Parameters
//
//  1  -  NOTEHANDLE -  Handle of open note
//  2  -  UNID * -  Pointer to an optional target UNID structure
//  3  -  char * -  Pointer to an optional target string buffer
//  4  -  int -  Length of optional target string buffer
//  5  -  BOOL -  If TRUE, a decorated UNID string will be returned (i.e. "OF: ....:.... - ON: ....:....")
//  6  -  int -  ID of calling thread
//
//  Returns
//
//	BOOL			-  TRUE is the UNID was obtained, otherwise FALSE
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::GetNoteUNID(NOTEHANDLE hnSource, UNID *unidTarget, char *szTarget, int iTargetLen, BOOL bExtString, int iThreadID)
{
	OID			oidSource;												//  OID of source note
	UNID		unidSource;												//  UNID of source note
	char		szMsg[MAX_MSG + 1];										//  Generic message buffer

	if (hnSource == NULLHANDLE)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0109E, __FUNCTION__);
		LogMessage(szMsg, iThreadID);
		return FALSE;
	}

	NSFNoteGetInfo(hnSource, _NOTE_OID, &oidSource);

	memcpy(&unidSource.File.Innards[0], &oidSource.File.Innards[0], sizeof(DWORD));
	memcpy(&unidSource.File.Innards[1], &oidSource.File.Innards[1], sizeof(DWORD));
	memcpy(&unidSource.Note.Innards[0], &oidSource.Note.Innards[0], sizeof(DWORD));
	memcpy(&unidSource.Note.Innards[1], &oidSource.Note.Innards[1], sizeof(DWORD));

	if ( unidTarget != NULL )
	{
		memcpy(&unidTarget->File.Innards[0], &unidSource.File.Innards[0], sizeof(DWORD));
		memcpy(&unidTarget->File.Innards[1], &unidSource.File.Innards[1], sizeof(DWORD));
		memcpy(&unidTarget->Note.Innards[0], &unidSource.Note.Innards[0], sizeof(DWORD));
		memcpy(&unidTarget->Note.Innards[1], &unidSource.Note.Innards[1], sizeof(DWORD));
	}

	if ( szTarget != NULL )
	{
		GetUNIDString(unidSource, szTarget, iTargetLen, bExtString , iThreadID);
		if ( szTarget[0] == '\0' ) return FALSE;
	}
	return TRUE;
}


//  GetUNIDString
//
//  This function will return a string representation of a UNID structure
//
//  Parameters
//
//  1  -  UNID -  Source UNID structure
//  2  -  char * -  Pointer to the target string buffer
//  3  -  int -  Length of the target string buffer
//  4  -  BOOL -  If TRUE, a decorated UNID string will be returned (i.e. "OF: ....:.... - ON: ....:....")
//  5  -  int -  ID of calling thread
//
//  Returns
//
//
//  API Use Notes:
//
//
void ExecEnvironment::GetUNIDString(UNID unidSource, char *szTarget, int iTargetLen, BOOL bExtString, int iThreadID)
{
	char		szMsg[MAX_MSG + 1];										//  Generic message buffer

	if (szTarget == NULL)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0110E, __FUNCTION__);
		LogMessage(szMsg, iThreadID);
		return;
	}

	szTarget[0] = '\0';
	if (bExtString)
	{
		sprintf_s(szTarget, 
				  iTargetLen, 
				  FORMAT_UNID_STRING_EXT, 
				  unidSource.File.Innards[1], 
				  unidSource.File.Innards[0], 
				  unidSource.Note.Innards[1], 
				  unidSource.Note.Innards[0]);
	}
	else
	{
		sprintf_s(szTarget, 
				  iTargetLen, 
				  FORMAT_UNID_STRING, 
				  unidSource.File.Innards[1], 
				  unidSource.File.Innards[0], 
				  unidSource.Note.Innards[1], 
				  unidSource.Note.Innards[0]);
	}
	return;
}

//  MatchesPattern
//
//  This function will match the first string against the wilcard pattern in the second string.
//  This function variant IS CASE SENSITIVE.
//
//  Parameters
//
//  1  -  char *		- Pointer to the string to be matched
//  2  -  char *		- Pointer to the pattern to be matched against
//
//  Returns
//
//  BOOL	-	TRUE if the string matches the pattern
//				FALSE if the string does not match the pattern
//
//  Notes:
//
BOOL ExecEnvironment::MatchesPattern(const char *szString, const char *szPattern)
{
	//  Perform case sensitive pattern matching

	const char *pString = NULL;					//  Pointer into the String
	const char *pPattern = NULL;				//  Pointer into the Pattern

	pString = szString;
	pPattern = szPattern;

	//  Base comparison allow substitution '?' wildcards - breakout if a "*" is encountered
	while ((*pString) && (*pPattern != '*')) 
	{
		if ((*pString != *pPattern) && (*pPattern != '?')) 
		{
			return FALSE;							//  mismatch
		}
		pPattern++;
		pString++;
	}
	
	//  If there are more characters in the string them perform a wildcard '*' realignment
	if (*pString)
	{
		//  Search for the next position in the string that matches the next part of the pattern
		pPattern++;
		if (*pPattern)
		{
			//  Eliminate any superfluous wilcards
			while ((*pPattern) && ((*pPattern == '*') || (*pPattern == '?')))
			{
				pPattern++;
			}
			if (*pPattern)
			{
				while ((*pString) && (*pString != *pPattern))
				{
					*pString++;
				}

				//  String match position has been re-established
				if (*pString)
				{
					//  recursively perform a wildcard match on the remainder of the strings
					return MatchesPattern(pString, pPattern);
				}
			}
			else
			{
				//  Pattern is exhausted - therefore it is a match
				return TRUE;
			}
		}
		else
		{
			//  Pattern is exhausted - therefore it is a match
			return TRUE;
		}
	}

	//  End of the string is encountered - only wildcards are allowed to remain in the pattern
	while (*pPattern)
	{
		if ((*pPattern != '*') && (*pPattern != '?')) return FALSE;
		pPattern++;
	}
	
	//  Pattern is exhausted - match is positive
	return TRUE;
}

//  MatchesPatterni
//
//  This function will match the first string against the wilcard pattern in the second string.
//  This function variant IS NOT CASE SENSITIVE.
//
//  Parameters
//
//  1  -  char *		- Pointer to the string to be matched
//  2  -  char *		- Pointer to the pattern to be matched against
//
//  Returns
//
//  BOOL	-	TRUE if the string matches the pattern
//				FALSE if the string does not match the pattern
//
//  Notes:
//
BOOL ExecEnvironment::MatchesPatterni(const char *szString, const char *szPattern)
{
	//  Perform case insensitive pattern matching

	const char *pString = NULL;					//  Pointer into the String
	const char *pPattern = NULL;				//  Pointer into the Pattern

	pString = szString;
	pPattern = szPattern;

	//  Base comparison allow substitution '?' wildcards - breakout if a "*" is encountered
	while ((*pString) && (*pPattern != '*')) 
	{
		if ((toupper(*pString) != toupper(*pPattern)) && (*pPattern != '?')) 
		{
			return FALSE;							//  mismatch
		}
		pPattern++;
		pString++;
	}
	
	//  If there are more characters in the string them perform a wildcard '*' realignment
	if (*pString)
	{
		//  Search for the next position in the string that matches the next part of the pattern
		pPattern++;
		if (*pPattern)
		{
			//  Eliminate any superfluous wilcards
			while ((*pPattern) && ((*pPattern == '*') || (*pPattern == '?')))
			{
				pPattern++;
			}
			if (*pPattern)
			{
				while ((*pString) && (toupper(*pString) != toupper(*pPattern)))
				{
					*pString++;
				}

				//  String match position has been re-established
				if (*pString)
				{
					//  recursively perform a wildcard match on the remainder of the strings
					return MatchesPatterni(pString, pPattern);
				}
			}
			else
			{
				//  Pattern is exhausted - therefore it is a match
				return TRUE;
			}
		}
		else
		{
			//  Pattern is exhausted - therefore it is a match
			return TRUE;
		}
	}

	//  End of the string is encountered - only wildcards are allowed to remain in the pattern
	while (*pPattern)
	{
		if ((*pPattern != '*') && (*pPattern != '?')) return FALSE;
		pPattern++;
	}
	
	//  Pattern is exhausted - match is positive
	return TRUE;
}

//  IsThisADatabase
//
//  This function will determine if the filespac pointed to is a valid file type for a database
//
//  Parameters
//
//  1  -  char *			-  Pointer to the filespec to be checked
//
//  Returns
//
//  BOOL					-  TRUE if it is a valid database type
//							-  FALSE if it is not a Notes database type
//
//  API Use Notes:
//
//  valid types .N?F, .NSn, .box
//
BOOL ExecEnvironment::IsThisADatabase(char *szFileSpec)
{
	char		*szAtom;										//  Pointer to the last .Atom in the spec
	char		*pSpecChar;										//  Pointer to character in the file spec

	szAtom = NULL;

	if (szFileSpec == NULL) return FALSE;						//  Safety check	
	pSpecChar = szFileSpec;

	while (*pSpecChar != '\0')
	{
		if (*pSpecChar == '.') szAtom = pSpecChar;				//  Find the last .Atom in the string
		pSpecChar++;
	}
	if (szAtom == NULL) return FALSE;							//  Probably a directory name

	if (szAtom[0] != '.') return FALSE;
	if (szAtom[1] == '\0') return FALSE;
	if (toupper(szAtom[1]) != 'N')
	{
		if (strcmpi(szAtom, ".BOX") == 0) return TRUE;
		return FALSE;
	}
	if (szAtom[2] == '\0') return FALSE;
	if (szAtom[3] == '\0') return FALSE;
	if (szAtom[4] != '\0') return FALSE;
	if (toupper(szAtom[3]) == 'F') return TRUE;								//  NSF, NTF etc.
	if (toupper(szAtom[2]) == 'S')
	{
		if (toupper(szAtom[3] == 'G')) return TRUE;							//  NSG
		if ((szAtom[3] >= '1') && (szAtom[3] <= '9')) return TRUE;			//  NSn
	}
	return FALSE;
}

//  IsThisAReplicaID
//
//  This function will determine if the filename passed is a Replica ID, if so it will return the normally formatted version.
//
//  Parameters
//
//  1  -  char *			-  Pointer to the filename to be checked
//  2  -  char *			-  Pointer to the buffer to contain the normally formatted ID
//
//  Returns
//
//  BOOL					-  TRUE if it is a Replica ID
//							-  FALSE if it is not a ReplicaID
//
//  API Use Notes:
//
//  valid formats
//
//  XXXXXXXXXXXXXXXX or XXXXXXXX:XXXXXXXX or __XXXXXXXXXXXXXXXX.nsf
//
BOOL ExecEnvironment::IsThisAReplicaID(char *szFileName, char *szFRID)
{
	char			*pSource;								//  Pointer into source string
	char			*pTarget;								//  Pointer into target string
	int				iIndex;									//  Generic index

	pSource = szFileName;
	pTarget = szFRID;

	if (*pSource == '_') pSource++;
	if (*pSource == '_') pSource++;

	//  Loop processing the first 8 characters of the (possible) ReplicaID
	for (iIndex = 0; iIndex < 8; iIndex++)
	{
		if ((*pSource >= '0') && (*pSource <= '9'))
		{
			*pTarget = *pSource;
			pTarget++;
		}
		else
		{
			if ((*pSource >= 'A') && (*pSource <= 'F'))
			{
				*pTarget = *pSource;
				pTarget++;
			}
			else
			{
				if ((*pSource >= 'a') && (*pSource <= 'f'))
				{
					*pTarget = toupper(*pSource);
					pTarget++;
				}
				else
				{
					szFRID[0] = '\0';
					return FALSE;
				}
			}
		}
		pSource++;
	}

	if (*pSource == ':') pSource++;
	*pTarget = ':';
	pTarget++;

	//  Loop processing the next 8 characters of the (possible) ReplicaID
	for (iIndex = 0; iIndex < 8; iIndex++)
	{
		if ((*pSource >= '0') && (*pSource <= '9'))
		{
			*pTarget = *pSource;
			pTarget++;
		}
		else
		{
			if ((*pSource >= 'A') && (*pSource <= 'F'))
			{
				*pTarget = *pSource;
				pTarget++;
			}
			else
			{
				if ((*pSource >= 'a') && (*pSource <= 'f'))
				{
					*pTarget = toupper(*pSource);
					pTarget++;
				}
				else
				{
					szFRID[0] = '\0';
					return FALSE;
				}
			}
		}
		pSource++;
	}

	//  Check for valid termination sequence
	*pTarget = '\0';
	if ((*pSource == '\0') || (*pSource == '.')) return TRUE;
	szFRID[0] = '\0';
	return FALSE;
}

//  SetFastDrainMode
//
//  This function is overridden in MTExecutive to provide quick shutdown paths.
//
//  Parameters
//
//  Returns
//
//  BOOL  -			TRUE 
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::SetFastDrainMode(void)
{
	return TRUE;	
}

//  Panic
//
//  This function will induce a panic condition that will cause an NSD in the current client/server.
//
//  Parameters
//
//  1  -  char *		- Module Name
//  2  -  int			- Line Number
//  3  -  char *		- Reason for the Panic
//  4  -  void *		- First pointer of interest (usually the invoking object ... i.e. this)
//  5  -  void *		- Second pointer of interest
//  6  -  void *        - Third pointer of interest
//  7  -  int			- ID of the invoking Thread
//
//  Returns
//
//  Notes:
//
void ExecEnvironment::Panic(char *szModule, int iLineNo, char *szReason, void *p1, void *p2, void *p3, int iThreadID)
{
	char		szMsg[MAX_MSG + 1];								//  Disgnostic Message Buffer

	sprintf_s(szMsg, MAX_MSG, MSG_DXR0111S, szModule, iLineNo, szReason, iThreadID);
	LogMessage(szMsg, iThreadID);

	//  Induce a panic using the breakpoint interrupt
#ifdef WIN32
#ifdef W64
	OSLock(UCHAR, NULLHANDLE);
#else
	__asm int 3
#endif
#endif
#ifdef UNIX
	asm("int $3");
#endif
}

//  getDominoMemory
//
//  Function to allocate Domino memory of correct type for loading resources
//
//  Parameters
//
//  1  -  DWORD			- Size of the memory to be allocated
//  2  -  int			- ID of the invoking Thread
//
//  Returns
//
//  MEMHANDLE	-  Handle of the Domino Memory Buffer allocated
//                 NULLHANDLE if the allocation request failed.
//

MEMHANDLE ExecEnvironment::getDominoMemory(DWORD dwSize, int iThreadID)
{
	STATUS			stAPIRC;					//  Return code from Notes API calls
	MEMHANDLE		hMemBuffer;					//  Handle to the allocated memory buffer
	char			szMsg[MAX_MSG + 1];			//  Generic Message Buffer

	//  Allocate the Domino memory buffer
	stAPIRC = OSMemoryAllocate(MEM_GROWABLE, dwSize, &hMemBuffer);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0112E, dwSize);
		LogMessage(szMsg, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return NULLHANDLE;
	}

	//  Return the handle to the allocated memory
	return hMemBuffer;
}

//  resizeDominoMemory
//
//  Function to re-allocate Domino memory to the desired size
//
//  Parameters
//
//  1  -  MEMHANDLE - Handle to the allocated memory buffer
//  2  -  DWORD		- New size of the memory buffer
//  3  -  int			- ID of the invoking Thread
//
//  Returns
//
//  MEMHANDLE	-  Handle of the Domino Memory Buffer that was resized
//                 NULLHANDLE if the re-allocation request failed.
//
MEMHANDLE ExecEnvironment::resizeDominoMemory(MEMHANDLE hMem, DWORD dwSize, int iThreadID)
{
	STATUS			stAPIRC;					//  Return code from Notes API calls
	char			szMsg[MAX_MSG + 1];			//  Generic Message Buffer

	//  Reallocate the Domino memory buffer
	stAPIRC = OSMemoryReallocate(hMem, dwSize);
	if (ERR(stAPIRC) != NOERROR)
	{
		sprintf_s(szMsg, MAX_MSG, MSG_DXR0113E, dwSize);
		LogMessage(szMsg, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		//  Free the original memory buffer
		OSMemoryFree(hMem);
		return NULLHANDLE;
	}

	//  Return the handle to the allocated memory
	return hMem;
}

//  GetMappedDBH
//
//  This function will return the DB Handle passed
//
//  Parameters
//
//	1  -  DBHANDLE	Native D/B handle for the requested database
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  DBHANDLE  -		Local (to this thread) DB Handle for the database
//                  NULLHANDLE - A local DB Handle could not be made available
//
//  API Use Notes:
//
//
DBHANDLE ExecEnvironment::GetMappedDBH(DBHANDLE hdbNative, int IthreadID)
{
	return hdbNative;
}

//  InvalidateNativeDBH
//
//  This function merely returns TRUE, it is overridden by the MTExecutive implementation.
//
//  Parameters
//
//	1  -  DBHANDLE	Native D/B handle that is no longer needed
//  2  -  int		ID of the thread that is making the request
//
//  Returns
//
//  BOOL  -			TRUE if the DB Handle was marked as no longer needed
//                  FALSE if the DB Handle could not be marked as no longer needed
//
//  API Use Notes:
//
//
BOOL ExecEnvironment::InvalidateNativeDBH(DBHANDLE hdbNative, int iThreadID)
{
	return TRUE;
}

//  GetReplicaID
//
//  This function will format the replica ID in the passed buffer for a given database (handle)
//
//  Parameters
//
//	1  -  DBHANDLE	Handle of the database
//  2  -  char *    Pointer to the buffer that will contain the decorated replica ID
//  3  -  int		Size of the buffer
//  4  -  int		ID of the thread that is making the request
//
//  Returns
//
//  API Use Notes:
//
//
void ExecEnvironment::GetReplicaID(DBHANDLE hDB, char *szRepID, int iLen, int iThreadID)
{
	STATUS				stAPIRC;							//  Return code from Notes API calls
    DBREPLICAINFO		riReplicaSettings;					//  Replica Info from the source database
	char				szFRepID[MAX_REPID];				//  Formatted Replica ID
	char				szMsg[MAX_MSG + 1];					//  Generic message buffer
	
	//  Safety Chacks
	szRepID[0] = '\0';
	if (hDB == NULLHANDLE) return;
	if (iLen < MAX_REPID) return;

	//  Get the Replication Information Structure
	stAPIRC = NSFDbReplicaInfoGet(hDB, &riReplicaSettings);
	if (ERR(stAPIRC) != NOERROR)
	{
		LogMessage(MSG_DXR0140E, iThreadID);
		GetAPIMessage(stAPIRC, szMsg);
		LogMessage(szMsg, iThreadID);
		return;
	}

	//  Format the replica ID
	sprintf_s(szFRepID, MAX_REPID, FORMAT_REPID, riReplicaSettings.ID.Innards[1], riReplicaSettings.ID.Innards[0]);

	//  Return the decorated replica ID to the caller
	strcpy_s(szRepID, MAX_REPID, szFRepID);
	return;
}

